! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Boundary condition treatment routines for fluid simulation
!! Details on implementation can be found in \ref tem_bc_module 
!!
!! \author Manuel Hasert, Kannan Masilamani
?? include 'header/lbm_macros.inc'
module mus_bc_fluid_module

  ! include treelm modules
  use env_module,               only: rk
  use tem_param_module,         only: div1_3, csInv, cs, cs2inv
  use tem_spacetime_fun_module, only: tem_spacetime_for
  use tem_time_module,          only: tem_time_type
  use treelmesh_module,         only: treelmesh_type
  use tem_var_system_module,    only: tem_var_system_type
  use tem_property_module,      only: prp_missingNeigh
  use tem_subTree_type_module,  only: tem_subTree_type
  use tem_ini_condition_module, only: tem_ini_condition_type

  ! include musubi modules
  use mus_bc_header_module,     only: boundary_type, glob_boundary_type
  use mus_scheme_layout_module, only: mus_scheme_layout_type 
  use mus_field_prop_module,    only: mus_field_prop_type
  use mus_pdf_module,           only: pdf_global_type
  use mus_variable_module,      only: mus_derVarPos_type
  use mus_global_module,        only: param_type
  use mus_physics_module,       only: mus_physics_type

  implicit none

  private

  public :: inlet_ubb
  public :: inlet_eq

  public :: outlet_pab
  public :: outlet_eq
  public :: outlet_dnt
  public :: outlet_expol
  public :: outlet_zero_prsgrd

  public :: outlet_nrbc
  public :: outlet_nrbc_eq

contains
!******************************************************************************!
  !> Inlet Velocity Equilibrium type boundary conditions
  !!
  !! The incoming densities are set to the equilibrium distribution with 
  !! macroscopic values:
  !! - density is extrapolated (0 or 1st order) from the flow domain
  !!   + 0-order: so we simply use the density of the boundary element itself
  !! - velocity is used as defined in the configuration file
  !!
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'inlet', 
  !!    kind = 'inlet_eq', 
  !!    velocityX = 0.06, 
  !!    velocityY = 0.0, 
  !!    velocityZ = 0.0 }
  !!\endverbatim
  !!
  ! subroutine inlet_eq
?? copy :: bc_routineHeader( inlet_eq )
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    real(kind=rk) :: rho
    real(kind=rk) :: ux(globBC%nElems(iLevel)), uy(globBC%nElems(iLevel)),     &
      &              uz(globBC%nElems(iLevel))
    integer :: iELem, iDir
    real(kind=rk) :: input_loc(4*globBC%nElems(iLevel))
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !--------------------------------------------------------------------------- 
 
    ! Get velocity which might be defined to vary in space and time
    ux = tem_spacetime_for( me      = me%bc_states%velocity(1)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = timing,                                  &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%conf )

    uy = tem_spacetime_for( me      = me%bc_states%velocity(2)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = timing,                                  &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%conf )

    uz = tem_spacetime_for( me      = me%bc_states%velocity(3)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = timing,                                  &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%conf )
    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    ux = ux/physics%fac( iLevel )%vel
    uy = uy/physics%fac( iLevel )%vel
    uz = uz/physics%fac( iLevel )%vel

    input_loc = 0.0_rk
    ! Calculate the density of current element
    do iElem = 1, globBC%nElems(iLevel)
      rho = 0._rk
      do iDir = 1,layout%stencil(1)%QQ
        rho = rho + bcBuffer(                                                  &
   & ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer( iElem ), nScalars ) )
      enddo 
      input_loc( iElem ) = rho
    enddo
    do iElem = 1, globBC%nElems(iLevel)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 1 ) = ux(iElem)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 2 ) = uy(iElem)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 3 ) = uz(iElem)
    end do
    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )
    ! Calculate the equilibrium distribution
    call varSys%derive( derVarPos%equilFromMacro )%do(                         &
      &                input   = input_loc,                                    &
      &                bound   = bound,                                        &
      &                tree    = tree,                                         &
      &                subTree = subTree,                                      &
      &                requVar = varSys%Variable%val(derVarPos%equilFromMacro),&
      &                globSys = varSys,                                       &
      &                stencil = layout%stencil(1),                            &
      &                conf    = varSys%solSpec_conf,                          &
      &                res     = fEq )

    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(                         &
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) )   &
        & then

      do iDir = 1,layout%stencil(1)%QQ
        ! Write the values
        ! The bitmask points into the incoming direction into the flow domain, 
        !  which actually we want to update
        ! * For PUSH, we write to the incoming position, 
        !   as the kernel reads it from there without propagation.
        ! * For PULL, we need to write to the inverse direction, as the kernel 
        !   performs a bounce back before reading it. 
        !   However, this bounced back direction actually comes from the 
        !   non-existent boundary element and would point into the incoming 
        !   direction, so the value has to be treated and set as if it points 
        !   really into the incoming direction.
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          ! Depending on PUSH or pull, use + or - for cxDir, because directions 
          ! are inverted
          state(                                                               & 
?FETCH?( varPos(iDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT) &
            & = fEq( ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) )
        endif
      enddo
      endif  
    enddo

  end subroutine inlet_eq
!******************************************************************************!


!******************************************************************************!
  !> Outlet Pressure Equilibrium type boundary conditions
  !!
  !! The incoming densities are set to the equilibrium distribution with 
  !! macroscopic values used from:
  !! * Velocity: Linear extrapolation of velocity of precollision of current 
  !!   time step
  !! * Pressure as defined in the configuration file
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'outlet', 
  !!    kind = 'outlet_eq', 
  !!    density = 1.0
  !!     }
  !!\endverbatim
  !!
  !! This is described in the paper: 
  !! S. Izquierdo, P. Martínez-Lera, and N. Fueyo, 
  !! Analysis of open boundary effects in unsteady lattice Boltzmann simulations
  !! Comput. Math. Applications, vol. 58, no. 5, pp. 914–921, Sep. 2009.
  !!
  !! The velocity is extrapolated (1st order) from the flow domain
  !! See equation (6)
  !!
  !subroutine outlet_eq
?? copy :: bc_routineHeader( outlet_eq )
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    ! need to compute precollision velocity of neighbor element
    ! Density on boundary element
    real(kind=rk) :: rhoB( globBC%nElems(iLevel))
    ! Velocity on boundary element
    real(kind=rk) :: uxB(3*globBC%nElems(iLevel))
    ! Vel on frst neighbor element
    real(kind=rk) :: uxB_1(3*globBC%nElems(iLevel))
    ! Vel on scnd neighbor element
    real(kind=rk) :: uxB_2(3*globBC%nElems(iLevel))
    real(kind=rk) :: omega       ! Relaxation parameter
    integer :: iDir              ! Direction counter
    integer :: iElem             ! Element counter
    real(kind=rk) :: input_loc(4*globBC%nElems(iLevel))
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !--------------------------------------------------------------------------- 

    ! get density value from the definition in the lua file
    rhoB = tem_spacetime_for( me      = me%bc_states%pressure%spacetime,       &
      &                       treeIDs = glob%levelDesc( iLevel )%total(globBC% &
      &                                               elems(iLevel)%elem(:) ), &
      &                       tree    = tree,                                  &
      &                       time    = timing,                                &
      &                       n       = globBC%nElems(iLevel),                 &
      &                       conf    = params%conf )
    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    rhoB = rhoB/physics%fac( iLevel )%press
    ! And convert pressure to density
    rhoB = cs2inv*rhoB

    omega = fieldProp%fluid%omLvl( iLevel )
 
    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )
    !KM: velocity is taken from precollision of current
    call varSys%derive( derVarPos%velocity )%do(                               &
        &                  input   = me%elems(iLevel)%neighBufferPre(1,:),     &
        &                  bound   = bound,                                    &
        &                  tree    = tree,                                     &
        &                  subTree = subTree,                                  &
        &                  requVar = varSys%Variable%val( derVarPos%velocity ),&
        &                  globSys = varSys,                                   &
        &                  stencil = layout%stencil(1),                        &
        &                  conf    = varSys%solSpec_conf,                      &
        &                  res     = uxB_1 )

    call varSys%derive( derVarPos%velocity )%do(                               &
        &                  input   = me%elems(iLevel)%neighBufferPre(2,:),     &
        &                  bound   = bound,                                    &
        &                  tree    = tree,                                     &
        &                  subTree = subTree,                                  &
        &                  requVar = varSys%Variable%val( derVarPos%velocity ),&
        &                  globSys = varSys,                                   &
        &                  stencil = layout%stencil(1),                        &
        &                  conf    = varSys%solSpec_conf,                      & 
        &                  res     = uxB_2 )
    ! time step for extrapolation
    ! Extrapolate velocities on the boundary node
    ! !! Note here, that the position is q=0.5
    ! Other interpolation formula required for different q
    ! Zero order extrapolation
!    uxB = (uxB_1)
    ! MH: Linear extrapolation
    !     Compare with 
    uxB = 2.0_rk*uxB_1 - 1.0_rk*uxB_2
    ! Calculate the equilibrium distribution
!    input_loc = (/ rhoB(iElem), uxB/)
    input_loc = 0.0_rk
    input_loc( 1:globBC%nElems(iLevel) ) = rhoB
    do iElem = 1, globBC%nElems(iLevel)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 1) = uxB((iElem-1)*3+1)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 2) = uxB((iElem-1)*3+2)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 3) = uxB((iElem-1)*3+3)
    end do

    ! \todo: Passing the complete treeID list here is false, only the elements
    !        for this boundary have to be passed
    call varSys%derive( derVarPos%equilFromMacro )%do(                         &
        &            input   = input_loc,                                      &
        &            bound   = bound,                                          &
        &            tree    = tree,                                           &
        &            subTree = subTree,                                        &
        &            requVar = varSys%Variable%val( derVarPos%equilFromMacro ),&
        &            globSys = varSys,                                         &
        &            stencil = layout%stencil(1),                              &
        &            conf    = varSys%solSpec_conf,                            & 
        &            res     = fEq )


    ! Run over all elements in list for this boundary
    do iElem = 1, globBC%nElems( iLevel )
     do iDir = 1,layout%stencil(1)%QQ
        ! the bitmask points into the incoming direction, which must
        ! be updated. As the kernel reads this value with FETCH,
        ! we store it to the FETCH position. 
        ! This results for 
        ! - PUSH in storing the density to where the bitmask 
        !   points, i.e. incoming
        ! - PULL storing the density pointing outwards, but as 
        !   this density is being bounced back before collision in 
        !   the kernel, this is the correct position
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          ! Now assign the values
  state( ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)  = &
          & fEq( ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) )
        endif
      enddo
    enddo

  end subroutine outlet_eq
!******************************************************************************!


!******************************************************************************!
!> Outlet Pressure do-nothing boundary
!! Here: Linear extrapolation of velocity of precollistion of current time step
!! is used
!!
!! This is taken from the paper: 
!! M. Junk and Z. Yang, 
!! Asymptotic Analysis of Lattice Boltzmann Outflow Treatments
!! Commun. Comput. Phys., pp. 1–11, 2011.
!! KM: This boundary is stable only for omega <=1.925. 
!! This boundary condition prescribes density = 1.0 at outlet as default
!!
!! Algorithm used in this boundary condition:
!! fEq(1,u) and fEq(rho,u) are computed using macroscopic values from current element
!! In fNeq, pre-collision of current time step is used
!! for non-normal direction, extrapolated velocity from two neighbor element 
!! at current time step is used
!!
!  subroutine outlet_dnt
?? copy :: bc_routineHeader( outlet_dnt )
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    real(kind=rk) :: fEq0( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    real(kind=rk) :: fNEq( layout%stencil(1)%QQ )
    real(kind=rk) :: visc    ! Temporary density for transient 
    real(kind=rk) :: rho     ! Temporary density for transient 
    real(kind=rk) :: rho0    ! reference density 
    real(kind=rk) :: uxB(3)  ! Velocity on boundary element
    real(kind=rk) :: uxB_1(3*globBC%nElems(iLevel)) ! Vel on first neigh element
    real(kind=rk) :: uxB_2(3*globBC%nElems(iLevel)) ! Vel on scond neigh element
    real(kind=rk) :: uxB_3(3*globBC%nElems(iLevel)) ! Vel on third neigh element
    real(kind=rk) :: omega   ! Relaxation parameter
    integer :: iDir          ! Direction counter
    integer :: iElem         ! Element counter
    real(kind=rk) :: input_loc(4*globBC%nElems(iLevel))
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !--------------------------------------------------------------------------- 

    omega = fieldProp%fluid%omLvl( iLevel )
    visc  = fieldProp%fluid%kine_viscosityLB
    rho0 = fieldProp%fluid%rho0

    input_loc = 0.0_rk
    do iElem = 1, globBC%nElems(iLevel)
      rho = 0._rk
      do iDir = 1,layout%stencil(1)%QQ
        fTmp(?IDX?(iDir, iElem, layout%stencil(1)%QQ)) = bcBuffer(             &
      & ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
        rho = rho + fTmp(?IDX?(iDir, iElem, layout%stencil(1)%QQ))
      enddo 
      input_loc( iElem ) = rho
    enddo

    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )

    ! Get local and neighbor velocity
    call varSys%derive( derVarPos%velocity )%do(                               &
      &                    input   = fTmp,                                     &
      &                    bound   = bound,                                    &
      &                    tree    = tree,                                     &
      &                    subTree = subTree,                                  &
      &                    requVar = varSys%Variable%val( derVarPos%velocity ),&
      &                    globSys = varSys,                                   &
      &                    stencil = layout%stencil(1),                        &
      &                    conf    = varSys%solSpec_conf,                      &
      &                    res     = uxB_1 )

    !KM: velocity is taken from precollision of current
    ! time step for extrapolation
    call varSys%derive( derVarPos%velocity )%do(                               &
      &                    input   = me%elems(iLevel)%neighBufferPre(1,:),     &
      &                    bound   = bound,                                    &
      &                    tree    = tree,                                     &
      &                    subTree = subTree,                                  &
      &                    requVar = varSys%Variable%val( derVarPos%velocity ),&
      &                    globSys = varSys,                                   &
      &                    stencil = layout%stencil(1),                        &
      &                    conf    = varSys%solSpec_conf,                      &
      &                    res     = uxB_2 )

    call varSys%derive( derVarPos%velocity )%do(                               &
      &                    input   = me%elems(iLevel)%neighBufferPre(2,:),     &
      &                    bound   = bound,                                    &
      &                    tree    = tree,                                     &
      &                    subTree = subTree,                                  &
      &                    requVar = varSys%Variable%val( derVarPos%velocity ),&
      &                    globSys = varSys,                                   &
      &                    stencil = layout%stencil(1),                        &
      &                    conf    = varSys%solSpec_conf,                      &
      &                    res     = uxB_3 )

    do iElem = 1, globBC%nElems(iLevel)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 1) = uxB_1((iElem-1)*3+1)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 2) = uxB_1((iElem-1)*3+2)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 3) = uxB_1((iElem-1)*3+3)
    end do
    ! \todo: Passing the complete treeID list here is false, only the elements
    !        for this boundary have to be passed
    ! Calculate the equilibrium distribution
    ! KM: post collision macroscopic values are used to compute fEq
    call varSys%derive( derVarPos%equilFromMacro )%do(                         &
      &              input   = input_loc,                                      &
      &              bound   = bound,                                          &
      &              tree    = tree,                                           &
      &              subTree = subTree,                                        &
      &              requVar = varSys%Variable%val( derVarPos%equilFromMacro ),&
      &              globSys = varSys,                                         &
      &              stencil = layout%stencil(1),                              &
      &              conf    = varSys%solSpec_conf,                            &
      &              res     = fEq )

    ! Calculate the reference equilibrium distribution
    input_loc( 1:globBC%nElems(iLevel) ) = rho0

    call varSys%derive( derVarPos%equilFromMacro )%do(                         &
      &              input   = input_loc,                                      &
      &              bound   = bound,                                          &
      &              tree    = tree,                                           &
      &              subTree = subTree,                                        &
      &              requVar = varSys%Variable%val( derVarPos%equilFromMacro ),&
      &              globSys = varSys,                                         &
      &              stencil = layout%stencil(1),                              &
      &              conf    = varSys%solSpec_conf,                            &
      &              res     = fEq0 )


    ! Run over all elements in list for this boundary
    do iElem = 1, globBC%nElems(iLevel)

      ! Extrapolate velocities on the boundary node
      ! !! Note here, that the position is q=0.5
      ! Other interpolation formula required for different q
      ! Zero order extrapolation
!      uxB = uxB_1((iElem-1)*3+1 : iElem*3)
      ! MH: Linear extrapolation
      !     Compare with 
      uxB = 1.5_rk*uxB_2((iElem-1)*3+1 : iElem*3)                              &
        & - 0.5_rk*uxB_3((iElem-1)*3+1 : iElem*3)

      do iDir = 1,layout%stencil(1)%QQ
        fNEq(iDir) = fTmp(?IDX?(iDir, iElem, layout%stencil(1)%QQ))            &
          &        - fEq(?IDX?(iDir, iElem, layout%stencil(1)%QQ)) 
      enddo   

      ! Treat all directions, but actually we only want to treat 
      ! non-normal directions
      ! Equation (3.2b)
      do iDir = 1,layout%stencil(1)%QQ
        ! the bitmask points into the incoming direction, which must
        ! be updated. As the kernel reads this value with FETCH,
        ! we store it to the FETCH position. 
        ! This results for 
        ! - PUSH in storing the density to where the bitmask 
        !   points, i.e. incoming
        ! - PULL storing the density pointing outwards, but as 
        !   this density is being bounced back before collision in 
        !   the kernel, this is the correct position
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
!          fluidNeighbor = &
! &?ElemIDX?(neigh( ?NGPOS?( varPos( iDir ), globBC%elems(iLevel)%elem(iElem),nScalars )),nScalars)
          ! update the incoming velocity direction 
          state( ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)  = &
         ! This bc is stable only using FETCH with inverse direction of bitmask
         ! For PULL this means, get the outgoing one from upstream neighbor
         ! For PUSH this means, get the outgoing one from local element
         ! KM: Using FETCH gives stable result
         ! \todo test for different testcase for stability
            & state(                                                           &
&?FETCH?(varPos(layout%stencil(1)%cxDirInv(iDir)),globBC%elems(iLevel)%elem(iElem),nScalars,neigh),currT) &
!    &    fTmp(?IDX?(layout%stencil(1)%cxDirInv(iDir), iElem, layout%stencil(1)%QQ))  &
            & + 6._rk*layout%weight( iDir )                                    &
            !extrapolated velocity is used here to remove fluctuations
            & * ( layout%stencil(1)%cxDir( 1, iDir )*uxB(1)                    &
            & +   layout%stencil(1)%cxDir( 2, iDir )*uxB(2)                    &
            & +   layout%stencil(1)%cxDir( 3, iDir )*uxB(3))
        endif
      enddo

      ! then overwrite the normal direction with special treatment
      ! Equation (3.2a)
      iDir = globBC%elems( iLevel )%normalInd( iElem )
state( ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)  = &
      ! KM:we use fEq2 since this equilibrium has to be computed from 
      ! reference density
        & fEq0(?IDX?(iDir, iElem, layout%stencil(1)%QQ))                       &
        & - (visc*omega -1._rk)*fNeq( layout%stencil(1)%cxDirInv( iDir ))
    enddo

  end subroutine outlet_dnt
!******************************************************************************!


!******************************************************************************!
!> Outlet Pressure extrapolation boundary
!!
!! This is taken from the paper: 
!! M. Junk and Z. Yang, 
!! Asymptotic Analysis of Lattice Boltzmann Outflow Treatments
!! Commun. Comput. Phys., pp. 1–11, 2011.
!!
!! This boundary condition prescribes density = 1.0 at outlet as default
!!
!  subroutine outlet_expol
?? copy :: bc_routineHeader( outlet_expol )
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    real(kind=rk) :: fTmp_1
    real(kind=rk) :: fTmp_2
    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    real(kind=rk) :: fEq0( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    real(kind=rk) :: fNEq( layout%stencil(1)%QQ )
    real(kind=rk) :: rho     ! Temporary density for transient 
    real(kind=rk) :: rho0    ! reference density 
    real(kind=rk) :: uxB(3*globBC%nElems(iLevel)) ! Velocity on boundary elem
    integer :: iDir              ! Direction counter
    integer :: iElem             ! Element counter
    real(kind=rk) :: input_loc(4*globBC%nElems(iLevel))
    real(kind=rk) :: rhoB(globBC%nElems(iLevel))
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !---------------------------------------------------------------------------
    ! get transient density value with reference value from me%bc_states%pressure
    rhoB = tem_spacetime_for( me      = me%bc_states%pressure%spacetime,       &
      &                       treeIDs = glob%levelDesc( iLevel )%total(globBC% &
      &                                               elems(iLevel)%elem(:) ), &
      &                       tree    = tree,                                  &
      &                       time    = timing,                                &
      &                       n       = globBC%nElems(iLevel),                 &
      &                       conf    = params%conf )
    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    rhoB = rhoB/physics%fac( iLevel )%press
    ! And convert pressure to density
    rhoB = cs2inv*rhoB

    rho0 = fieldProp%fluid%rho0
    input_loc = 0._rk
    ! Run over all elements in list for this boundary
    do iElem = 1, globBC%nElems(iLevel)
      rho = 0._rk
      do iDir = 1,layout%stencil(1)%QQ
        fTmp(?IDX?(iDir, iElem, layout%stencil(1)%QQ)) = bcBuffer(             &
      & ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
        ! Get local density
        rho = rho + fTmp(?IDX?(iDir, iElem, layout%stencil(1)%QQ))
      enddo
      input_loc( iElem ) = rho
    enddo

    ! Get local velocity
    uxB = 0._rk
    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )

    !KM: velocity is taken from postcollision of current
    call varSys%derive( derVarPos%velocity )%do(                               &
      &                    input   = fTmp,                                     &
      &                    bound   = bound,                                    &
      &                    tree    = tree,                                     &
      &                    subTree = subTree,                                  &
      &                    requVar = varSys%Variable%val( derVarPos%velocity ),&
      &                    globSys = varSys,                                   &
      &                    stencil = layout%stencil(1),                        &
      &                    conf    = varSys%solSpec_conf,                      &
      &                    res     = uxB )

    do iElem = 1, globBC%nElems(iLevel)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 1) = uxB((iElem-1)*3+1)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 2) = uxB((iElem-1)*3+2)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 3) = uxB((iElem-1)*3+3)
    end do
    ! Calculate the equilibrium distribution
    ! KM: we use post collision macroscopic values and pdf to 
    ! compute equilibrium and fNeq 
    ! Calculate local equilibrium (use local rho and uxB )
    call varSys%derive( derVarPos%equilFromMacro )%do(                         &
      &              input   = input_loc,                                      &
      &              bound   = bound,                                          &
      &              tree    = tree,                                           &
      &              subTree = subTree,                                        &
      &              requVar = varSys%Variable%val( derVarPos%equilFromMacro ),&
      &              globSys = varSys,                                         &
      &              stencil = layout%stencil(1),                              &
      &              conf    = varSys%solSpec_conf,                            &
      &              res     = fEq )

    !MH: using not rho0 but a defined density as the inlet density
    !input_loc( 1:globBC%nElems(iLevel) ) = rho0
    input_loc( 1:globBC%nElems(iLevel) ) = rhoB(1:globBC%nELems(iLevel))
    ! Calculate reference equilibrium (use rho0 and uxB )
    call varSys%derive( derVarPos%equilFromMacro )%do(                         &
      &              input   = input_loc,                                      &
      &              bound   = bound,                                          &
      &              tree    = tree,                                           &
      &              subTree = subTree,                                        &
      &              requVar = varSys%Variable%val( derVarPos%equilFromMacro ),&
      &              globSys = varSys,                                         &
      &              stencil = layout%stencil(1),                              &
      &              conf    = varSys%solSpec_conf,                            &
      &              res     = fEq0 )

    ! Run over all elements in list for this boundary
    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest( glob%levelDesc(iLevel)%property(                        &
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh ))then
      ! Calculate the non-equilirbium part with the local fEq
      do iDir = 1,layout%stencil(1)%QQ
        fNEq(iDir) = fTmp(?IDX?(iDir, iElem, layout%stencil(1)%QQ))            &
          &  - fEq(?IDX?(iDir, iElem, layout%stencil(1)%QQ)) 
      enddo   

      ! Treat all directions, but actually we only want to treat 
      ! non-normal directions
      ! Equation (3.2b)
      do iDir = 1,layout%stencil(1)%QQ
        ! the bitmask points into the incoming direction, which must
        ! be updated. As the kernel reads this value with FETCH,
        ! we store it to the FETCH position. 
        ! This results for 
        ! - PUSH in storing the density to where the bitmask 
        !   points, i.e. incoming
        ! - PULL storing the density pointing outwards, but as 
        !   this density is being bounced back before collision in 
        !   the kernel, this is the correct position
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
        ! KM: pre-collision of current time step must be used for extrapolation
          fTmp_1 = me%elems( iLevel )%neighBufferPre( 1,                       &
            & ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) )
          fTmp_2 = me%elems( iLevel )%neighBufferPre( 2,                       &
            & ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) )

          ! update the incoming velocity direction  
state( ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)  = &
            !linear extrapolation for wall at half-position 
            1.5_rk*fTmp_1 - 0.5_rk*fTmp_2
        endif
      enddo

      ! then overwrite the normal direction with special treatment
      ! Equation (3.2a)
      iDir = globBC%elems( iLevel )%normalInd( iElem )
      state( ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)  = &
      ! KM:we use fEq2 since this equilibrium has to be computed from reference density
        & fEq0(?IDX?(iDir, iElem, layout%stencil(1)%QQ))                       &
        &  + fNeq( layout%stencil(1)%cxDirInv( iDir ))

      endif 

    enddo

  end subroutine outlet_expol
!******************************************************************************!


!******************************************************************************!
!> Inlet Velocity Bounce Back boundary condition
!!
!! This is taken from the paper: 
!! [1] S. Izquierdo and N. Fueyo, “Characteristic  non-reflecting 
!! boundary  conditions  for  open  boundaries  in  lattice 
!!  Boltzmann  methods,” Physical Review E, vol. 78, no. 46707, 2008
!! The incoming densities \f$ \bar{f_{alpha}} \f$ are reconstructed from the
!! already reflected densities, enforcing the given velocity me\%ubb\%velocity.
!! [2] M. Junk and Z. Yang. One-point boundary condition for the lattice Boltzmann method,
!! Physical Review E, vol 72, issu 8, year 2005.
!!
!! This boundary condition has error of 1st order pressure and 2nd order velocity
!! only if wall is located exactly at q=1/2 else the accuracy of both pressure
!! and velocity reduce by order 1.
!!
!  subroutine inlet_ubb
?? copy :: bc_routineHeader( inlet_ubb)
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: eqPlus
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: ux(globBC%nElems(iLevel))
    real(kind=rk) :: uy(globBC%nElems(iLevel))
    real(kind=rk) :: uz(globBC%nElems(iLevel))
    integer :: iELem, iDir
    !--------------------------------------------------------------------------- 
      
    !Multiply transient velocity with spatial velocity. Spatial term includes
    !maxvalue
    ux = tem_spacetime_for( me      = me%bc_states%velocity(1)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = timing,                                  &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%conf )

    uy = tem_spacetime_for( me      = me%bc_states%velocity(2)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = timing,                                  &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%conf )

    uz = tem_spacetime_for( me      = me%bc_states%velocity(3)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = timing,                                  &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%conf )
    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    ux = ux/physics%fac( iLevel )%vel
    uy = uy/physics%fac( iLevel )%vel
    uz = uz/physics%fac( iLevel )%vel

    !element loop
    do iElem = 1, globBC%nElems( iLevel )
      if( .not. btest( glob%levelDesc(iLevel)%property(                        &
        &              globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh))then

        !store state in temp array
        !KM:produce no checker board effect if we use pre-collision values of 
        !current time step
        do iDir = 1,layout%stencil(1)%QQ
          !KM: Fetch gives precollision value at current time step and it produce
          ! no fluctuation at inlet.
          !SAVE use post collision values and produce fluctuation. 
          !\todo [1] and [2] use post collision values so we need to validate this
          !KM: It was found by Kartick with his testcase, SAVE is more stable then FETCH
          !fTmp( iDir ) = state( ?SAVE?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)
          fTmp( iDir ) = bcBuffer(                                               &
        & ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
        enddo 
       
        do iDir = 1,layout%stencil(1)%QQ
          ! Write the values
          if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then
            ! Accourding to the bounce back rule, 
            ! Subtract the equlibrium if its computed from alpha-(inverse of
            ! bitmask) or add the equlibrium if its computed in the direction of
            ! bitmask.
            ! In the Izquiredo paper, this is rho0, the reference density
            eqPlus = layout%weight( iDir )*6._rk*fieldProp%fluid%rho0            &
              &    * ( layout%stencil(1)%cxDir( 1, iDir )*ux(iElem)              &
              &    +   layout%stencil(1)%cxDir( 2, iDir )*uy(iElem)              &
              &    +   layout%stencil(1)%cxDir( 3, iDir )*uz(iElem))
 state(?FETCH?(varPos(iDir), globBC%elems(iLevel)%elem(iElem), nScalars, neigh), currT) = &
            ! We need to get post-collision pdf in direction
            ! alpha-, which is the inverse direction of bitmask
            ! For PULL this means, get the outgoing one, as this is the one which will be bounced back
            ! For PUSH this means, get the already bounced back pdf back, so take the incoming
              & fTmp( layout%stencil(1)%cxDirInv( iDir ) ) + eqPlus
            ! MH: Please check this density!
          endif
        enddo
      endif
    
    enddo
  end subroutine inlet_ubb
!******************************************************************************!


!******************************************************************************!
!> Outlet boundary conditions with zero pressure gradient.
!! \author Kannan Masilamani
!! 
!! These boundary conditions use the neighbor cells density and velocity in the
!! equilibrium function. It is not necessary to specify density at boundary in
!! the lua configuration file \f$ f = f^{eq}(\rho_{b-1},u_{b-1}) \f$
!!
!! MH: Kannan, please update them
!!
!  subroutine outlet_zero_prsgrd
?? copy :: bc_routineHeader( outlet_zero_prsgrd)
    !---------------------------------------------------------------------------
    ! defining local variables
    real(kind=rk) :: rho, ux(3*globBC%nElems(iLevel)), usq
    integer :: iELem, iDir
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !---------------------------------------------------------------------------

    do iElem = 1, globBC%nElems(iLevel)
      !Calculate local density and velocity moments
      do iDir = 1,layout%stencil(1)%QQ
        fTmp(?IDX?(iDir, iElem, layout%stencil(1)%QQ)) = bcBuffer(             &
      & ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
      enddo
    enddo
    ! Get local velocity
    !KM: velocity is taken from postcollision of current
    ux  = 0._rk
    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )
    call varSys%derive( derVarPos%velocity )%do(                               &
      &                    input   = fTmp,                                     &
      &                    bound   = bound,                                    &
      &                    tree    = tree,                                     &
      &                    subTree = subTree,                                  &
      &                    requVar = varSys%Variable%val( derVarPos%velocity ),&
      &                    globSys = varSys,                                   &
      &                    stencil = layout%stencil(1),                        &
      &                    conf    = varSys%solSpec_conf,                      & 
      &                    res     = ux )

    do iElem = 1, globBC%nElems(iLevel)
      !Calculate local density and velocity moments
      rho = 0._rk
      do iDir = 1,layout%stencil(1)%QQ
        rho = rho + fTmp(?IDX?(iDir, iElem, layout%stencil(1)%QQ))
      enddo

      usq = 0.0_rk
      usq = ux((iElem-1)*3+1)*ux((iElem-1)*3+1)                                &
        & + ux((iElem-1)*3+2)*ux((iElem-1)*3+2)                                &
        & + ux((iElem-1)*3+3)*ux((iElem-1)*3+3) 

      do iDir = 1,layout%stencil(1)%QQ
        ! Write the values
        if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then
          state( ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ),nScalars, neigh ), currT)  =   &
            &  layout%weight( iDir )*(2._rk*rho+9._rk*fieldProp%fluid%rho0*( ( &
            &            layout%stencil(1)%cxDir( 1, iDir )*ux((iElem-1)*3+1)  &
            &         +  layout%stencil(1)%cxDir( 2, iDir )*ux((iElem-1)*3+2)  &
            &         +  layout%stencil(1)%cxDir( 3, iDir )*ux((iElem-1)*3+3)  &
            &         )**2                                                     &
            &         -  div1_3*usq))                                          &
          ! We need to get post-collision pdf in direction
          ! alpha-, which is the inverse direction of bitmask
          ! For PULL this means, get the outgoing one, as this is the one which will be bounced back
          ! For PUSH this means, get the already bounced back pdf back, so take the incoming
            &         - fTmp(                                                  &
        & ?IDX?(layout%stencil(1)%cxDirInv(iDir), iElem, layout%stencil(1)%QQ) )   
        endif
      enddo
    enddo

  end subroutine outlet_zero_prsgrd
!******************************************************************************!


!******************************************************************************!
!> Outlet Pressure Bounce Back boundary condition 
!! \author Manuel Hasert
!!
!! An anti-bounce back is performed for the outgoing links, so that the 
!! incoming densities are set with the defined reference density while 
!! maintaining the correct non-equilibrium part.
!! The velocity is extrapolated from two neighboring fluid elements (boundary element
!! itself + one fluid neighbor). The position of the wall is hence between the boundary 
!! element and the (non-existing) solid element
!! Based on the Bouzidi Boundary conditions, the current implementation is 
!! taken from the paper: 
!! S. Izquierdo and N. Fueyo, “Characteristic  non-reflecing  boundary
!!  conditions  for  open  boundaries  in  lattice  Boltzmann
!!  methods,” Physical Review E, vol. 78, no. 46707, 2008.
!!
!! MH: I deactivated the correction term as it involves information from the
!! old time step. In parallel, this can not be provided.
!!
!subroutine outlet_pab
?? copy :: bc_routineHeader( outlet_pab)
    !---------------------------------------------------------------------------
    real(kind=rk) :: rhoB(globBC%nElems(iLevel)) ! Density on boundary element
    real(kind=rk) :: rhoF    ! density on fluid element 
    real(kind=rk) :: uxF(3)  ! Velocity on fluid element
    real(kind=rk) :: uxB(3)  ! Velocity on boundary element
    ! Velocity on first neighbor element
    real(kind=rk) :: uxB_1(3*globBC%nElems(iLevel) )
    ! Velocity on second neighbor element
    real(kind=rk) :: uxB_2(3*globBC%nElems(iLevel) )
    real(kind=rk) :: usq, usqF        ! Squared velocity magnitude 
    real(kind=rk) :: fPlusFluid       ! f Plus for anti-bounceback correction
    ! Equilibrium Plus for anti-bounceback correction
    real(kind=rk) :: fEqPlus
    ! Equlibrium Plus from the old time step
    real(kind=rk) :: fEqPlusFluid
    real(kind=rk) :: fTmpPost( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    real(kind=rk) :: omega       ! Relaxation parameter
    integer :: iDir              ! Direction counter
    integer :: iElem             ! Element counter
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !---------------------------------------------------------------------------

    ! get transient density value with reference value from me%bc_states%pressure
    rhoB = tem_spacetime_for( me      = me%bc_states%pressure%spacetime,       &
      &                       treeIDs = glob%levelDesc( iLevel )%total(globBC% &
      &                                               elems(iLevel)%elem(:) ), &
      &                       tree    = tree,                                  &
      &                       time    = timing,                                &
      &                       n       = globBC%nElems(iLevel),                 &
      &                       conf    = params%conf )
    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    rhoB = rhoB/physics%fac( iLevel )%press
    ! And convert pressure to density
    rhoB = cs2inv*rhoB

    omega = fieldProp%fluid%omLvl( iLevel )

    do iElem = 1, globBC%nElems( iLevel )
      do iDir = 1,layout%stencil(1)%QQ
        fTmpPost(?IDX?(iDir, iElem, layout%stencil(1)%QQ)) = bcBuffer(         &
      & ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
      enddo
    enddo

    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )
    ! Get local and neighbor velocity
    call varSys%derive( derVarPos%velocity )%do(                               &
      &                    input   = fTmpPost,                                 &
      &                    bound   = bound,                                    &
      &                    tree    = tree,                                     &
      &                    subTree = subTree,                                  &
      &                    requVar = varSys%Variable%val( derVarPos%velocity ),&
      &                    globSys = varSys,                                   &
      &                    stencil = layout%stencil(1),                        &
      &                    conf    = varSys%solSpec_conf,                      &
      &                    res     = uxB_1 )

    call varSys%derive( derVarPos%velocity )%do(                               &
      &                    input   = me%elems(iLevel)%neighBufferPre(1,:),     &
      &                    bound   = bound,                                    &
      &                    tree    = tree,                                     &
      &                    subTree = subTree,                                  &
      &                    requVar = varSys%Variable%val( derVarPos%velocity ),&
      &                    globSys = varSys,                                   &
      &                    stencil = layout%stencil(1),                        &
      &                    conf    = varSys%solSpec_conf,                      &
      &                    res     = uxB_2 )

    ! Run over all elements in list for this boundary
    do iElem = 1, globBC%nElems( iLevel )
      !KM: Deactivated correction term
      rhoF = 0.0_rk
      do iDir = 1,layout%stencil(1)%QQ
        ! previous time step precollision to compute only density and velocity  
!        fTmpPre(iDir) = state(                                                 &
!          &  ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem(iElem), nScalars, neigh ),&
!          &  lastT) 
        rhoF = rhoF + fTmpPost( ?IDX?(iDir, iElem, layout%stencil(1)%QQ) )
      enddo

!      call varSys%derive( derVarPos%velocity )%do(                             &
!        &              input = fTmpPre, nElems = 1, tree = tree,               &
!        &              requVar = varSys%Variable%val( derVarPos%velocity ),    &
!        &              globSys = varSys, stencil = layout%stencil(1),          &
!        &              res = uxBOld,                                           &
!        &              conf = varSys%solSpec_conf )

      ! Extrapolate velocities on the boundary node
      ! !! Note here, that the position is q=0.5
      ! Other interpolation formula required for different q
      uxB = 1.5_rk*uxB_1((iElem-1)*3+1 : iElem*3)                              &
        & - 0.5_rk*uxB_2((iElem-1)*3+1 : iElem*3)
      uxF = uxB_1((iElem-1)*3+1 : iElem*3)

      usq = uxB(1)*uxB(1) + uxB(2)*uxB(2) + uxB(3)*uxB(3) 
      usqF = uxF(1)*uxF(1) + uxF(2)*uxF(2) + uxF(3)*uxF(3) 
      
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then
          ! Calc the correction term and the Dirichlet (Eq) term
          fEqPlus = layout%weight( iDir )*rhoB(iElem)                          & 
            &     + 4.5_rk*layout%weight( iDir )*fieldProp%fluid%rho0*((       &
            &       layout%stencil(1)%cxDir( 1, layout%stencil(1)%             &
            &                                          cxDirInv(iDir ))*uxB(1) &
            &     + layout%stencil(1)%cxDir( 2, layout%stencil(1)%             &
            &                                          cxDirInv(iDir ))*uxB(2) &
            &     + layout%stencil(1)%cxDir( 3, layout%stencil(1)%             &
            &                                          cxDirInv(iDir ))*uxB(3) &
            &     )**2                                                         &  
            &     - div1_3*usq)                
          fEqPlusFluid = layout%weight( iDir )*rhoF                            &
            &          + 4.5_rk*layout%weight( iDir )*fieldProp%fluid%rho0*((  &
            &            layout%stencil(1)%cxDir( 1, layout%stencil(1)%        &
            &                                          cxDirInv(iDir ))*uxF(1) &
            &          + layout%stencil(1)%cxDir( 2, layout%stencil(1)%        &
            &                                          cxDirInv(iDir ))*uxF(2) &
            &          + layout%stencil(1)%cxDir( 3, layout%stencil(1)%        &
            &                                          cxDirInv(iDir ))*uxF(3) &
            &          )**2                                                    & 
            &          - div1_3*usqF)                

          fPlusFluid = 0.5_rk*( &
    &     fTmpPost( ?IDX?(iDir, iElem, layout%stencil(1)%QQ) ) &  
    &   + fTmpPost( ?IDX?(layout%stencil(1)%cxDirInv(iDir), iElem, layout%stencil(1)%QQ) ) )  

          ! Now assign the values
          ! Actually, we should take fPlusLAst and fEqPlusLast.
          ! But this introduces errors in parallel
          state( ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)  = &
          ! antibounceback term
!    &   - state( ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)    &
          ! We need to get post-collision pdf in direction
          ! alpha-, which is the inverse direction of bitmask
          ! For PULL this means, get the outgoing one, as this is the one which will be bounced back
          ! For PUSH this means, get the already bounced back pdf back, so take the incoming
    &   - fTmpPost( ?IDX?(layout%stencil(1)%cxDirInv(iDir), iElem, layout%stencil(1)%QQ) ) &  
    &   + 2._rk*fEqPlus     &                          ! Dirichlet pressure term
    &   + (2._rk - omega )*( fPlusFluid - fEqPlusFluid ) ! Correction term
        endif
      enddo
    enddo

  end subroutine outlet_pab
!******************************************************************************!


!******************************************************************************!
  !> Characteristic-based non-reflective open boundary conditions 
  !!
  !! These boundary conditions are taken from the paper:  
  !!   S. Izquierdo and N. Fueyo, 
  !!   "Characteristic nonreﬂecting boundary 
  !!   conditions for open boundaries in lattice Boltzmann methods," 
  !!   Physical Review E, vol. 78, no. 46707, 2008.
  !! 
  !! > Note: unstable behavior for high omega relaxation parameters.
  !! >       It is recommended to use a sponge layer with a 
  !! >       \ref mus_aux_module::mus_setspatialomega "spatial omega" 
  !! >       where the omega is decreased towards the boundaries to increase 
  !! >       stability
  !!
  !! \author Manuel Hasert
  !!
  !  subroutine outlet_nrbc
?? copy :: bc_routineHeader( outlet_nrbc )
    !---------------------------------------------------------------------------
    real(kind=rk) :: fEqPlus, fPlus
    real(kind=rk) :: usq, omega, rhoTmp, rhoTmp_2
    real(kind=rk) :: rhoRef(globBC%nElems( iLevel ))
    real(kind=rk) :: Ma, swap, sigma, kappa, K_mod
    real(kind=rk) :: L1, L2, L3, L4, L5
    real(kind=rk) :: dudx, dvdx, dwdx, drhodx
    real(kind=rk) :: rhoN1Prev, uN1Prev(3), rhoN2Prev, uN2Prev(3)
    real(kind=rk) :: rhoLodi, uLodi(3), csMod
    real(kind=rk) :: fTmpPost( layout%stencil(1)%QQ )
    real(kind=rk) :: fTmpPost_2( layout%stencil(1)%QQ )
    real(kind=rk) :: rho0ref
    integer :: iELem, iDir
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !---------------------------------------------------------------------------
    rhoRef = tem_spacetime_for( me      = me%bc_states%pressure%spacetime,     &
      &                         treeIDs = glob%levelDesc( iLevel )%total(      &
      &                                         globBC%elems(iLevel)%elem(:)), &
      &                         tree    = tree,                                &
      &                         time    = timing,                              &
      &                         n       = globBC%nElems(iLevel),               &
      &                         conf    = params%conf )
    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    rhoRef = rhoRef/physics%fac( iLevel )%press
    ! And convert pressure to density
    rhoRef = cs2inv*rhoRef

    omega = fieldProp%fluid%omLvl( iLevel )
    sigma = me%nrbc%sigma 
    kappa = me%nrbc%kappa  ! =1 for D2Q9, =1.66666??? for D3Q19
    csMod = me%nrbc%cs_mod
    ! transient density value

    rho0ref = fieldProp%fluid%rho0
    do iElem = 1, globBC%nElems( iLevel )
      rhoLodi = me%elems( iLevel )%lodi( 1,   1, iElem )
      uLodi   = me%elems( iLevel )%lodi( 2:4, 1, iElem )
      ! density and velocity of my first neighbor from last iteration
      rhoN1Prev = me%elems( iLevel )%lodi( 1,   2, iElem )
      uN1Prev   = me%elems( iLevel )%lodi( 2:4, 2, iElem )
      ! density and velocity of my secnd neighbor from last iteration
      rhoN2Prev = me%elems( iLevel )%lodi( 1,   3, iElem )
      uN2Prev   = me%elems( iLevel )%lodi( 2:4, 3, iElem )
      
      ! Use a swapping variable to treat +x and -x directions
      swap = 1._rk
      ! x-direction
      if( globBC%elems( iLevel )%normal( 1, iElem ) .ne. 0 ) then 
        swap = 1._rk*(-real( globBC%elems( iLevel )%normal( 1, iElem ),kind=rk))
        
        ! LODI Treatment
        Ma   = abs(uLodi(1))*csInv ! X direction
        ! Equation (17) for the linear relaxation model constant
        ! lodi_length= distance outlet-inlet
        K_mod = swap*sigma*(1.d0 - Ma*Ma)*cs/(me%nrbc%lodi_length)
      
        ! calculate spatial derivatives according to (15) which will be used to
        ! compute the characteristics
        ! This is a quadratic extrapolation
        drhodx =  swap*div1_3*(8.d0*rhoLodi  - 9.d0*rhoN1Prev  + rhoN2Prev)
        dudx   =  swap*div1_3*(8.d0*uLodi(1) - 9.d0*uN1Prev(1) + uN2Prev(1))
        dvdx   =  swap*div1_3*(8.d0*uLodi(2) - 9.d0*uN1Prev(2) + uN2Prev(2))
        dwdx   =  swap*div1_3*(8.d0*uLodi(3) - 9.d0*uN1Prev(3) + uN2Prev(3))
      
        ! wave amplitudes
        ! Incoming wave L1 is approached with a linear relaxation model (16)
        L1 = K_mod*csMod*csMod*(rhoLodi-rhoRef(iElem))
        L2 = (uLodi(1))*dvdx  ! equation (9) second row
        L3 = (uLodi(1))*dwdx  ! equation (9) second row adapted for z-direction
        ! equation (9) third row with EOS equation (3) used
        L4 = (uLodi(1))*csMod*csMod*drhodx*(1.d0-1.d0/kappa)
        ! equation (9) fourth row with EOS equation (3) used
        ! MH: Here rhoLodi should be replaced with rho0
        L5 = ( uLodi(1) + swap * csMod )                                       &
          &* ( csMod * csMod * drhodx / kappa + rho0ref * swap * csMod * dudx )
      
        ! LODI Equations (18)
        rhoLodi  = rhoLodi  - (L4+0.5d0*(L5+L1))/(csMod*csMod)
        ! Eq. (18b) for x-velocity
        uLodi(1) = uLodi(1) - ((L5-L1))/(rho0Ref*swap*csMod)*0.5d0
        uLodi(2) = uLodi(2) - L2  ! equation (18c) for y-direction
        uLodi(3) = uLodi(3) - L3  ! equation (18c) for z-direction
      else
        ! Treat y-direction
        if( globBC%elems( iLevel )%normal( 2, iElem ) .ne. 0 ) then 
          swap = 1._rk*(-real( globBC%elems( iLevel )%normal(2,iElem),kind=rk))
          
          ! LODI Treatment
          Ma   = abs(uLodi(2))*csInv ! X direction
          ! lodi_length= distance outlet-inlet
          K_mod = swap*sigma*(1.d0 - Ma*Ma)*cs/(me%nrbc%lodi_length)
        
          ! calculate derivatives
          ! This is a quadratic extrapolation
          drhodx =  swap*div1_3*(8.d0*rhoLodi  - 9.d0*rhoN1Prev  + rhoN2Prev)
          dudx   =  swap*div1_3*(8.d0*uLodi(1) - 9.d0*uN1Prev(1) + uN2Prev(1))
          dvdx   =  swap*div1_3*(8.d0*uLodi(2) - 9.d0*uN1Prev(2) + uN2Prev(2))
          dwdx   =  swap*div1_3*(8.d0*uLodi(3) - 9.d0*uN1Prev(3) + uN2Prev(3))
        
          ! wave amplitudes
          L1 = K_mod*csMod*csMod*(rhoLodi-rhoRef(iElem))
          L2 = ( uLodi(2) ) * dudx
          L3 = ( uLodi(2) ) * dwdx
          L4 = ( uLodi(2) ) * csMod * csMod * drhodx * ( 1.d0 - 1.d0 / kappa )
          L5 = ( uLodi(2) + swap * csMod )                                     &
            &* ( csMod * csMod * drhodx / kappa                                &
            &+   rho0ref * swap * csMod * dvdx )
        
          ! LODI Equations
          rhoLodi  = rhoLodi  - (L4+0.5d0*(L5+L1))/(csMod*csMod)
          uLodi(1) = uLodi(1) - L2 
          uLodi(2) = uLodi(2) - ((L5-L1))/(rho0ref*swap*csMod)*0.5d0
          uLodi(3) = uLodi(3) - L3 
        else
          ! Treat z-direction
          if( globBC%elems( iLevel )%normal( 3, iElem ) .ne. 0 ) then 
            swap = 1._rk*(-real( globBC%elems(iLevel)%normal(3,iElem),kind=rk ))
            
            ! LODI Treatment
            Ma   = abs(uLodi(3))*csInv ! X direction
            !lodi_length= distance outlet-inlet
            K_mod = swap*sigma*(1.d0 - Ma*Ma)*cs/(me%nrbc%lodi_length)
          
            ! calculate derivatives
           ! This is a quadratic extrapolation
            drhodx =  swap*div1_3*(8.d0*rhoLodi  - 9.d0*rhoN1Prev  + rhoN2Prev)
            dudx   =  swap*div1_3*(8.d0*uLodi(1) - 9.d0*uN1Prev(1) + uN2Prev(1))
            dvdx   =  swap*div1_3*(8.d0*uLodi(2) - 9.d0*uN1Prev(2) + uN2Prev(2))
            dwdx   =  swap*div1_3*(8.d0*uLodi(3) - 9.d0*uN1Prev(3) + uN2Prev(3))
          
            ! wave amplitudes
            L1 = K_mod*csMod*csMod*(rhoLodi-rhoRef(iElem))
            L2 = ( uLodi(3) ) * dudx
            L3 = ( uLodi(3) ) * dvdx
            L4 = ( uLodi(3) ) * csMod * csMod * drhodx * ( 1.d0 - 1.d0 / kappa )
            L5 = ( uLodi(3) +swap * csMod )                                    &
              &* ( csMod * csMod * drhodx / kappa                              &
              &+   rho0ref * swap * csMod * dwdx )
          
            ! LODI Equations
            rhoLodi  = rhoLodi  - (L4+0.5d0*(L5+L1))/(csMod*csMod)
            uLodi(1) = uLodi(1) - L2 
            uLodi(2) = uLodi(2) - L3
            uLodi(3) = uLodi(3) - ((L5-L1))/(rho0ref*swap*csMod)*0.5d0
          endif
        endif
      endif

      ! Now store the previous variables for next iteration
      me%elems( iLevel )%lodi( 1,   1, iElem ) = rhoLodi
      me%elems( iLevel )%lodi( 2:4, 1, iElem ) = uLodi 
      ! density and velocity of my first neighbor
      ! MH: Velocity and Density should be calculated with access as SAVE
      ! For pull, this is fine however, as IDX and SAVE coincide
      rhoTmp = 0.0_rk
      rhoTmp_2 = 0.0_rk
      do iDir = 1,layout%stencil(1)%QQ
      ! MH: These should be calculated with a SAVE access!
        fTmpPost( iDir ) = bcBuffer(                                           &
      & ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
        fTmpPost_2(iDir) = me%elems( iLevel )%neighBufferPost( 1,              &
          &                  ?IDX?( iDir, iElem, layout%stencil(1)%QQ ))
        rhoTmp = rhoTmp + fTmpPost(iDir)
        rhoTmp_2 = rhoTmp_2 + fTmpPost_2(iDir)
      enddo
      !density
      me%elems( iLevel )%lodi( 1,   2, iElem ) = rhoTmp
      ! initialize the lower and upper bound
      bound(1) = 1
      bound(2) = 1
      !velocity
    ! \todo: Passing the complete treeID list here is false, only the elements
    !        for this boundary have to be passed
      call varSys%derive( derVarPos%velocity )%do(                             &
        &                  input   = fTmpPost,                                 &
        &                  bound   = bound,                                    &
        &                  tree    = tree,                                     &
        &                  subTree = subTree,                                  &
        &                  requVar = varSys%Variable%val( derVarPos%velocity ),&
        &                  globSys = varSys,                                   &
        &                  stencil = layout%stencil(1),                        &
        &                  conf    = varSys%solSpec_conf,                      &
        &                  res     = me%elems(iLevel)%lodi(2:4,2,iElem) )

      ! density and velocity of my secnd neighbor
      me%elems( iLevel )%lodi( 1,   3, iElem )  = rhoTmp_2
      call varSys%derive( derVarPos%velocity )%do(                             &
        &                  input   = fTmpPost_2,                               &
        &                  bound   = bound,                                    &
        &                  tree    = tree,                                     &
        &                  subTree = subTree,                                  &
        &                  requVar = varSys%Variable%val( derVarPos%velocity ),&
        &                  globSys = varSys,                                   &
        &                  stencil = layout%stencil(1),                        &
        &                  conf    = varSys%solSpec_conf,                      &
        &                  res     = me%elems(iLevel)%lodi(2:4,3,iElem) )

      usq = uLodi(1)*uLodi(1) + uLodi(2)*uLodi(2) + uLodi(3)*uLodi(3) 
      ! Now do the anti-bounce back
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then
          ! MH: This calculation must be the same as for the outlet_pab,
          !     except that we are using the Lodi values
          ! Calc the correction term and the Dirichlet (Eq) term
          ! MH: It makes no difference to use either iDir or inv(iDir)
          fEqPlus = layout%weight(iDir)*rhoLodi                                &
            &     + 4.5_rk*layout%weight(iDir)*fieldProp%fluid%rho0*((         &
            &       layout%stencil(1)%cxDir( 1, layout%stencil(1)%             &
            &                                        cxDirInv(iDir ))*uLodi(1) &
            &     + layout%stencil(1)%cxDir( 2, layout%stencil(1)%             &
            &                                        cxDirInv(iDir ))*uLodi(2) &
            &     + layout%stencil(1)%cxDir( 3, layout%stencil(1)%             &
            &                                        cxDirInv(iDir ))*uLodi(3) &
            &     )**2 - div1_3*usq)                

          fPlus = 0.5_rk*(                                                     &
&  state( ?FETCH?( varPos( iDir ), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT) &
&+ state( ?FETCH?( varPos( layout%stencil(1)%cxDirInv( iDir )), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ) &
&           , currT))

          ! Now assign the values
state( ?FETCH?( varPos( iDir ), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT)  =     &
          ! antibounceback term
  & - state( ?FETCH?( varPos( iDir ), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT)  &
  & + 2._rk*fEqPlus !     &                                      ! Dirichlet pressure term
       ! MH: Correciton term increases error!
!       + (2._rk - omega )*( fPlus - fEqPlus )                      ! Correction term
        endif
      enddo
    enddo
 
  end subroutine outlet_nrbc
!******************************************************************************!


!******************************************************************************!
  !> Characteristic-based non-reflective open boundary conditions 
  !!
  !! These boundary conditions are taken from the paper:  
  !!   S. Izquierdo and N. Fueyo, 
  !!   "Characteristic nonreﬂecting boundary 
  !!   conditions for open boundaries in lattice Boltzmann methods," 
  !!   Physical Review E, vol. 78, no. 46707, 2008.
  !! 
  !! > Note: unstable behavior for high omega relaxation parameters.
  !! >       It is recommended to use a sponge layer with a 
  !! >       \ref mus_aux_module::mus_setspatialomega "spatial omega" 
  !! >       where the omega is decreased towards the boundaries to increase 
  !! >       stability
  !!
  !! \author Manuel Hasert
  !!
  !  subroutine outlet_nrbc_eq
?? copy :: bc_routineHeader( outlet_nrbc_eq )
    !---------------------------------------------------------------------------
    real(kind=rk) :: usq, omega, rhoTmp, rhoTmp_2
    real(kind=rk) :: rhoRef(globBC%nElems( iLevel ))
    real(kind=rk) :: Ma, swap, sigma, kappa, K_mod
    real(kind=rk) :: L1, L2, L3, L4, L5
    real(kind=rk) :: dudx, dvdx, dwdx, drhodx
    real(kind=rk) :: rhoN1Prev, uN1Prev(3), rhoN2Prev, uN2Prev(3)
    real(kind=rk) :: rhoLodi, uLodi(3), csMod
    real(kind=rk) :: fEq( layout%stencil(1)%QQ )
    real(kind=rk) :: fTmpPost( layout%stencil(1)%QQ )
    real(kind=rk) :: fTmpPost_2( layout%stencil(1)%QQ )
    real(kind=rk) :: input_loc(4), rho0ref
    integer :: iELem, iDir
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !---------------------------------------------------------------------------
    rhoRef = tem_spacetime_for( me      = me%bc_states%pressure%spacetime,     &
      &                         treeIDs = glob%levelDesc( iLevel )%total(      &
      &                                      globBC%elems(iLevel)%elem(:)),    &
      &                         tree    = tree,                                &
      &                         time    = timing,                              &
      &                         n       = globBC%nElems(iLevel),               &
      &                         conf    = params%conf )
    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    rhoRef = rhoRef/physics%fac( iLevel )%press
    ! And convert pressure to density
    rhoRef = cs2inv*rhoRef

    omega = fieldProp%fluid%omLvl( iLevel )
    sigma = me%nrbc%sigma 
    kappa = me%nrbc%kappa  ! =1 for D2Q9, D3Q19
    csMod = me%nrbc%cs_mod
    ! transient density value

    rho0ref = fieldProp%fluid%rho0
    do iElem = 1, globBC%nElems( iLevel )
      rhoLodi = me%elems( iLevel )%lodi( 1,   1, iElem )
      uLodi   = me%elems( iLevel )%lodi( 2:4, 1, iElem )
      ! density and velocity of my first neighbor from last iteration
      rhoN1Prev = me%elems( iLevel )%lodi( 1,   2, iElem )
      uN1Prev   = me%elems( iLevel )%lodi( 2:4, 2, iElem )
      ! density and velocity of my secnd neighbor from last iteration
      rhoN2Prev = me%elems( iLevel )%lodi( 1,   3, iElem )
      uN2Prev   = me%elems( iLevel )%lodi( 2:4, 3, iElem )
      ! MH: Playing with rho0 for (in)compressible formulations
      !rho0ref = rhoRef(iElem)
      
      ! Use a swapping variable to treat +x and -x directions
      swap = 1._rk
      ! x-direction
      if( globBC%elems( iLevel )%normal( 1, iElem ) .ne. 0 ) then 
        ! Use a swapping variable to treat +x and -x directions
        swap = 1._rk*(-real( globBC%elems( iLevel )%normal( 1, iElem ),kind=rk))
        
        ! LODI Treatment
        Ma   = abs(uLodi(1))*csInv ! X direction
        ! Equation (17) for the linear relaxation model constant
        ! lodi_length= distance outlet-inlet
        K_mod = swap*sigma*(1.d0 - Ma*Ma)*cs/(me%nrbc%lodi_length)
      
        ! calculate spatial derivatives according to (15) which will be used to
        ! compute the characteristics
        ! This is a quadratic extrapolation
        drhodx =  swap*div1_3*(8.d0*rhoLodi  - 9.d0*rhoN1Prev  + rhoN2Prev)
        dudx   =  swap*div1_3*(8.d0*uLodi(1) - 9.d0*uN1Prev(1) + uN2Prev(1))
        dvdx   =  swap*div1_3*(8.d0*uLodi(2) - 9.d0*uN1Prev(2) + uN2Prev(2))
        dwdx   =  swap*div1_3*(8.d0*uLodi(3) - 9.d0*uN1Prev(3) + uN2Prev(3))
      
        ! wave amplitudes
        ! Incoming wave L1 is approached with a linear relaxation model (16)
        L1 = K_mod*csMod*csMod*(rhoLodi-rhoRef(iElem))
        L2 = (uLodi(1))*dvdx  ! equation (9) second row
        L3 = (uLodi(1))*dwdx  ! equation (9) second row adapted for z-direction
        ! equation (9) third row with EOS equation (3) used
        L4 = (uLodi(1))*csMod*csMod*drhodx*(1.d0-1.d0/kappa)
        ! equation (9) fourth row with EOS equation (3) used
        ! MH: Here rhoLodi should be replaced with rho0
        L5 = ( uLodi(1) + swap * csMod )                                       &
          &* ( csMod * csMod * drhodx / kappa                                  &
          &+   rho0ref * swap * csMod * dudx )
      
        ! LODI Equations (18)
        rhoLodi  = rhoLodi  - (L4+0.5d0*(L5+L1))/(csMod*csMod)
        ! Eq. (18b) for x-velocity
        uLodi(1) = uLodi(1) - ((L5-L1))/(rho0Ref*swap*csMod)*0.5d0
        uLodi(2) = uLodi(2) - L2  ! equation (18c) for y-direction
        uLodi(3) = uLodi(3) - L3  ! equation (18c) for z-direction
      else
        ! Treat y-direction
        if( globBC%elems( iLevel )%normal( 2, iElem ) .ne. 0 ) then 
          swap = 1._rk*(-real( globBC%elems( iLevel )%normal(2,iElem),kind=rk ))
          
          ! LODI Treatment
          Ma   = abs(uLodi(2))*csInv ! X direction
          ! lodi_length= distance outlet-inlet
          K_mod = swap*sigma*(1.d0 - Ma*Ma)*cs/(me%nrbc%lodi_length)
        
          ! calculate derivatives
          ! This is a quadratic extrapolation
          drhodx =  swap*div1_3*(8.d0*rhoLodi  - 9.d0*rhoN1Prev  + rhoN2Prev)
          dudx   =  swap*div1_3*(8.d0*uLodi(1) - 9.d0*uN1Prev(1) + uN2Prev(1))
          dvdx   =  swap*div1_3*(8.d0*uLodi(2) - 9.d0*uN1Prev(2) + uN2Prev(2))
          dwdx   =  swap*div1_3*(8.d0*uLodi(3) - 9.d0*uN1Prev(3) + uN2Prev(3))
        
          ! wave amplitudes
          L1 = K_mod*csMod*csMod*(rhoLodi-rhoRef(iElem))
          L2 = (uLodi(2))*dudx
          L3 = (uLodi(2))*dwdx
          L4 = (uLodi(2))*csMod*csMod*drhodx*(1.d0-1.d0/kappa)
          L5 = ( uLodi(2) + swap * csMod )                                     &
            &* ( csMod * csMod * drhodx / kappa                                &
            &+   rho0ref * swap * csMod * dvdx )
        
          ! LODI Equations
          rhoLodi  = rhoLodi  - (L4+0.5d0*(L5+L1))/(csMod*csMod)
          uLodi(1) = uLodi(1) - L2 
          uLodi(2) = uLodi(2) - ((L5-L1))/(rho0ref*swap*csMod)*0.5d0
          uLodi(3) = uLodi(3) - L3 
        else
          ! Treat z-direction
          if( globBC%elems( iLevel )%normal( 3, iElem ) .ne. 0 ) then 
            swap = 1._rk*(-real(globBC%elems( iLevel )%normal(3,iElem),kind=rk))
            
            ! LODI Treatment
            Ma   = abs(uLodi(3))*csInv ! X direction
            !lodi_length= distance outlet-inlet
            K_mod = swap*sigma*(1.d0 - Ma*Ma)*cs/(me%nrbc%lodi_length)
          
            ! calculate derivatives
           ! This is a quadratic extrapolation
            drhodx =  swap*div1_3*(8.d0*rhoLodi  - 9.d0*rhoN1Prev  + rhoN2Prev)
            dudx   =  swap*div1_3*(8.d0*uLodi(1) - 9.d0*uN1Prev(1) + uN2Prev(1))
            dvdx   =  swap*div1_3*(8.d0*uLodi(2) - 9.d0*uN1Prev(2) + uN2Prev(2))
            dwdx   =  swap*div1_3*(8.d0*uLodi(3) - 9.d0*uN1Prev(3) + uN2Prev(3))
          
            ! wave amplitudes
            L1 = K_mod*csMod*csMod*(rhoLodi-rhoRef(iElem))
            L2 = (uLodi(3))*dudx
            L3 = (uLodi(3))*dvdx
            L4 = (uLodi(3))*csMod*csMod*drhodx*(1.d0-1.d0/kappa)
            L5 = ( uLodi(3) + swap * csMod )                                   &
              &* ( csMod * csMod * drhodx / kappa                              &
              &+   rho0ref * swap * csMod * dwdx )
          
            ! LODI Equations
            rhoLodi  = rhoLodi  - (L4+0.5d0*(L5+L1))/(csMod*csMod)
            uLodi(1) = uLodi(1) - L2 
            uLodi(2) = uLodi(2) - L3
            uLodi(3) = uLodi(3) - ((L5-L1))/(rho0ref*swap*csMod)*0.5d0
          endif
        endif
      endif

      ! Now store the previous variables for next iteration
      me%elems( iLevel )%lodi( 1,   1, iElem ) = rhoLodi
      me%elems( iLevel )%lodi( 2:4, 1, iElem ) = uLodi 
      ! density and velocity of my first neighbor
      ! MH: Velocity and Density should be calculated with access as SAVE
      ! For pull, this is fine however, as IDX and SAVE coincide
      rhoTmp = 0.0_rk
      rhoTmp_2 = 0.0_rk
      do iDir = 1,layout%stencil(1)%QQ
      ! MH: These should be calculated with a SAVE access!
        fTmpPost( iDir ) = bcBuffer(                                           &
      & ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
        fTmpPost_2(iDir) = me%elems( iLevel )%neighBufferPost( 1,              &
          &                  ?IDX?( iDir, iElem, layout%stencil(1)%QQ ))
        rhoTmp = rhoTmp + fTmpPost(iDir)
        rhoTmp_2 = rhoTmp_2 + fTmpPost_2(iDir)
      enddo
      !density
      me%elems( iLevel )%lodi( 1,   2, iElem ) = rhoTmp
      ! initialize the lower and upper bound
      bound(1) = 1
      bound(2) = 1
      !velocity
      call varSys%derive( derVarPos%velocity )%do(                             &
        &                  input   = fTmpPost,                                 &
        &                  bound   = bound,                                    &
        &                  tree    = tree,                                     &
        &                  subTree = subTree,                                  &
        &                  requVar = varSys%Variable%val( derVarPos%velocity ),&
        &                  globSys = varSys,                                   &
        &                  stencil = layout%stencil(1),                        &
        &                  conf    = varSys%solSpec_conf,                      &
        &                  res     = me%elems(iLevel)%lodi(2:4,2,iElem) )

      ! density and velocity of my secnd neighbor
      me%elems( iLevel )%lodi( 1,   3, iElem )  = rhoTmp_2
    ! \todo: Passing the complete treeID list here is false, only the elements
    !        for this boundary have to be passed
      call varSys%derive( derVarPos%velocity )%do(                             &
        &                  input   = fTmpPost_2,                               &
        &                  bound   = bound,                                    &
        &                  tree    = tree,                                     &
        &                  subTree = subTree,                                  &
        &                  requVar = varSys%Variable%val( derVarPos%velocity ),&
        &                  globSys = varSys,                                   &
        &                  stencil = layout%stencil(1),                        &
        &                  conf    = varSys%solSpec_conf,                      &
        &                  res     = me%elems(iLevel)%lodi(2:4,3,iElem) )

      ! Calculate the equilibrium distribution
      input_loc = (/ rhoLodi, uLodi(1), uLodi(2), uLodi(3) /)
    ! \todo: Passing the complete treeID list here is false, only the elements
    !        for this boundary have to be passed
      call varSys%derive( derVarPos%equilFromMacro )%do(                       &
        &              input   = fTmpPost_2,                                   &
        &              bound   = bound,                                        &
        &              tree    = tree,                                         &
        &              subTree = subTree,                                      &
        &              requVar = varSys%Variable%val(derVarPos%equilFromMacro),&
        &              globSys = varSys,                                       &
        &              stencil = layout%stencil(1),                            &
        &              conf    = varSys%solSpec_conf,                          &
        &              res     = fEq )

      usq = uLodi(1)*uLodi(1) + uLodi(2)*uLodi(2) + uLodi(3)*uLodi(3) 
      ! Now do the anti-bounce back
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then
          ! Now assign the values
 state( ?FETCH?( varPos( iDir ), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT)  = &
          fEq( iDir )
        endif
      enddo
    enddo
 
  end subroutine outlet_nrbc_eq
!******************************************************************************!


end module mus_bc_fluid_module
!******************************************************************************!
