! See copyright notice in the COPYRIGHT file.
?? include 'header/lbm_macros.inc'
!******************************************************************************!
!> Boundary condition wall treatment routines
!!
!! This module contains higher order wall treatments
!! A detailed description on the implementation details are given 
!! in \ref tem_bc_module
!!
module mus_bc_fluid_wall_module

  ! include treelm modules
  use env_module,               only: rk
  use tem_tools_module,         only: tem_write
  use tem_time_module,          only: tem_time_type
  use treelmesh_module,         only: treelmesh_type
  use tem_var_system_module,    only: tem_var_system_type
  use tem_subTree_type_module,  only: tem_subTree_type
  use tem_ini_condition_module, only: tem_ini_condition_type

  ! include musubi modules
  use mus_bc_header_module,     only: boundary_type, glob_boundary_type
  use mus_scheme_layout_module, only: mus_scheme_layout_type 
  use mus_field_prop_module,    only: mus_field_prop_type
  use mus_pdf_module,           only: pdf_global_type
  use mus_variable_module,      only: mus_derVarPos_type
  use mus_global_module,        only: param_type
  use mus_physics_module,       only: mus_physics_type

  implicit none

  private

  public :: wall_bouzidi
  public :: wall_multiReflection
  public :: slip_wall

contains

!******************************************************************************!
  !> Higher order no-slip wall linear interpolation bounce back boundary 
  !! conditions.
  !! \author Manuel Hasert, Jiaxing Qi
  !!
  !! An interpolation is performed, which can exactly set the wall to a position
  !! between the fluid and the solid node. The geometric information is stored
  !! in the qVal, which gives the proportion of the wall position by the real
  !! position between fluid and solid node. \n
  !! After the paper [1]  M. Bouzidi, M. Firdaouss, and P. Lallemand,
  !! “Momentum transfer of a Boltzmann-lattice fluid with boundaries,”
  !! Physics of Fluids, vol. 13, no. 11, pp. 3452–3459, Nov. 2001 \n
  !! Equations 5 (linear interpolation). quadratic is not there yet.
  !!
  !  subroutine wall_bouzidi
?? copy :: bc_routineHeader( wall_bouzidi )
    !---------------------------------------------------------------------------
    ! fIn:        incoming link of boundary element
    ! fOut:       outgoing link of boundary element
    ! fOutNeigh:  outgoing link of neighbor of boundary element
    real(kind=rk) :: fIn, fOut, fOutNeigh
    integer :: iElem, iDir, invDir
    real(kind=rk) :: qVal
    !---------------------------------------------------------------------------

    ! Loop over all boundary elements in this level
    do iElem = 1, globBC%nElems( iLevel )
      ! Loop over all links except central one.
      do iDir = 1, layout%stencil(1)%QQN
        ! bitmask points inverse to boundary direction incoming Dir (iDir)
        ! Whereas qVal is stored in outgoing Dir ( invDir )
        if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then

          ! Only treat the required directions
          invDir = layout%stencil(1)%cxDirInv( iDir )
          qVal = globBC%elems(iLevel)%qVal( invDir, iElem )

          ! --------------------------------------------------------------------
          ! JQ: for PULL, should update f_outgoing( invDir ), as in kernel it
          !               will bounce back (stream)
          !     for PUSH, should update f_incoming( iDir   ), as in kernel it
          !               will stay (relax), iDir = neigh( invDir )
          !               so we should use SAVE( invDir )
          if( qVal >= 0.5_rk ) then

            fIn  = state(                                                      &
& ?SAVE?( varPos(iDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT)
            fOut = state(                                                      &
& ?SAVE?( varPos(invDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT)
            state(                                                             &
& ?SAVE?( varPos(invDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT) = &
              &  (1._rk - 0.5_rk/qVal) * fIn + (0.5_rk/qVal) * fOut

          ! --------------------------------------------------------------------
          ! JQ: for PULL, should update f_outgoing( invDir ), as in kernel it
          !               will bounce back (stream)
          !     for PUSH, should update f_incoming( iDir   ), as in kernel it
          !               will stay (relax), iDir = neigh( invDir )
          !     so we should use SAVE( invDir ), i.e. update post-collision value
          else ! qVal < 0.5

            fOutNeigh  = state(                                                &
& ?FETCH?( varPos(invDir), globBC%elems(iLevel)%elem(iElem), nScalars, neigh ), currT)
            fOut       = state(                                                &
& ?SAVE?( varPos(invDir), globBC%elems(iLevel)%elem(iElem), nScalars, neigh ), currT)
            state(                                                             &
& ?SAVE?( varPos(invDir), globBC%elems(iLevel)%elem(iElem), nScalars, neigh ), currT) = &
              &   (1._rk - 2._rk*qVal) * fOutNeigh + (2._rk*qVal) * fOut

          endif ! qVal > 0.5 or qVal < 0.5
        endif ! BitMask
      enddo ! iDir
    enddo ! iElem

  end subroutine wall_bouzidi
!******************************************************************************!


!******************************************************************************!
  !> Multi reflection wall boundary conditions
  !! \author Manuel Hasert
  !!
  !! After the paper [1]  C. Pan, L. Luo, and C. Miller, 
  !! "An evaluation of lattice Boltzmann schemes for porous medium flow 
  !! simulation", Computers & Fluids, vol. 35, no. 8, pp. 898–909, Sep.2006.
  !!
  !! Result: Tested with mini channel. Result: Instabilities from corner nodes.
  !! -> need to downgrade the corner nodes
  !! MH: Update: Since the re-definition of the 
  !!
  !  subroutine wall_multiReflection
?? copy :: bc_routineHeader( wall_multiReflection )
    !---------------------------------------------------------------------------
    integer :: iELem, iDir
    integer, parameter :: one_point = 2
    integer, parameter :: two_point = 2
    integer, parameter :: three_point = 3
    integer :: mr_type
    integer :: fluidNeighbor   ! position of the fluid neigbor in total list
    integer :: fluidNeighbor2  ! position of second fluid neighbor in total list
    real(kind=rk) :: qVal      ! current q-value for distance of wall
    real(kind=rk) :: qValSq    ! q-Value^2
    real(kind=rk) :: fPostR_xA_tn
    real(kind=rk) :: fPostR_xD_tn
    real(kind=rk) :: fPostR_xE_tn
    real(kind=rk) :: fPreR__xA_tn1
    real(kind=rk) :: fPreR__xD_tn1
    real(kind=rk) :: fPostL_xA_tn
    real(kind=rk) :: fPostL_xD_tn
    real(kind=rk) :: R_term          ! anti-bounce back term
    real(kind=rk) :: k1,k2,k3,k4,k5,k6 ! Multiplicators 
    !---------------------------------------------------------------------------
    call tem_write('MH: Warning: I didnt adapt these BCs to the new bitmasks! Wrong for sure')
    ! select kind of multi-reflection boundary
    mr_type = three_point
    mr_type = one_point

    ! Linear interpolation
    do iElem = 1, globBC%nElems( iLevel )
      do iDir = 1,layout%stencil(1)%QQ
        ! iDir points out of the domain
        if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then
          ! Only treat the required directions
          qVal   = globBC%elems(iLevel)%qVal( iDir, iElem )
          qValSq =  qVal*qVal
          ! equation 15 in [1]
          k1 = 1._rk
          k2 = (1._rk - 2._rk*(qVal + qValSq))/(1._rk+qValSq)
          k3 = qValSq/(1._rk+qValSq)
          k4 = -k2
          k5 =  k3
          k6 = 1._rk/(4._rk*(1._rk+qVal)*(1._rk+qVal))

          ! Get the fluid neighbor xD (x-c)
          fluidNeighbor =                                                      &
& ?ElemIDX?(neigh( ?NGPOS?( varPos( layout%stencil(1)%cxDirInv( iDir ) ), globBC%elems(iLevel)%elem( iElem ),nScalars )),nScalars )
          ! Fluid neighbor xE (x-2*c)
          fluidNeighbor2 =                                                     &
& ?ElemIDX?(neigh( ?NGPOS?( varPos( layout%stencil(1)%cxDirInv( iDir ) ), fluidNeighbor,nScalars )),nScalars )

        select case( mr_type)
        case( two_point )
          fPostR_xA_tn  = state(                                               &
& ?SAVE?( varPos( iDir ), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT)
          fPreR__xA_tn1 = state(                                               &
& ?FETCH?( varPos( iDir ), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT)
          fPreR__xD_tn1 = state(                                               &
& ?FETCH?( varPos(iDir), fluidNeighbor, nScalars, neigh ), currT)
          fPostL_xA_tn  = state(                                               &
& ?SAVE?( varPos( layout%stencil(1)%cxDirInv( iDir )), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT)
          fPostL_xD_tn  = state(                                               &
& ?SAVE?( varPos( layout%stencil(1)%cxDirInv( iDir )), fluidNeighbor, nScalars, neigh ), currT)
          R_term = k1*fPostR_xA_tn + k2*fPreR__xA_tn1 + k3*fPreR__xD_tn1       &
            &    + k4*fPostL_xA_tn + k5*fPostL_xD_tn
        case( three_point ) 
          fPostR_xA_tn  = state(                                               &
& ?SAVE?( varPos(iDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT)
          fPostR_xD_tn  = state(                                               &
& ?SAVE?( varPos(iDir), fluidNeighbor, nScalars, neigh ), currT)
          fPostR_xE_tn  = state(                                               &
& ?SAVE?( varPos(iDir), fluidNeighbor2, nScalars, neigh ), currT)
          fPostL_xA_tn  = state(                                               &
& ?SAVE?( varPos( layout%stencil(1)%cxDirInv( iDir )), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT)
          fPostL_xD_tn  = state(                                               &
& ?SAVE?( varPos( layout%stencil(1)%cxDirInv( iDir )), fluidNeighbor, nScalars, neigh ), currT)
          R_term = k1*fPostR_xA_tn + k2*fPostR_xD_tn + k3*fPostR_xE_tn         &
            &    + k4*fPostL_xA_tn + k5*fPostL_xD_tn

        end select

          ! iDir points outwards, and by SAVE we can update the incoming 
          ! (opposite) density for PUSH
          ! Independent of the streaming algorithm(PUSH,PULL) we need to 
          ! determine the neighbor in the direction opposite of the outgoing 
          ! boundary i = iDir, i' = stencil%cxDirInv( iDir )
          !          f_i'(x) = 2*q*f_i(x) + (1-2q)*f_i(x-c_i) 
          ! But we have already bounced back intrinsically, so we take the 
          ! first f_i = f_i' and also have to trace back the f_i(x-c_i)
          state(                                                               &
& ?SAVE?( varPos(iDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT) &
            & = R_term  !+ CorrectionTerm 
          ! \todo  calculate the correction term for multi-reflection
        endif
      enddo
    enddo
 
  end subroutine wall_multiReflection
!******************************************************************************!


!******************************************************************************!
  !> slip-wall boundary condition. Slip defined by a slip factor
  !!
  !! \li Normal velocity,\f$ u_n = 0 \f$
  !! \li Tangential velocity, \f$ \frac{\partial u_t}{\partial n} = 0 \f$
  !! \li Pressure, \f$ \frac{\partial P}{\partial n} = 0 \f$
  !! For slip-wall boundary, the slip factor will be multiplied by the velocity
  !! if slip factor = 1, then it is full/free-slip and if slip factor = 0, then
  !! it is no-slip  
  !!
  !! \todo KM: Currently, free-slip boundary works only for axis-parallel planes.
  !!           Need to extend it for arbitary geometries
  !!
  !  subroutine slip_wall
?? copy :: bc_routineHeader( slip_wall )
    !--------------------------------------------------------------------------- 
    ! defining local variables
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    real(kind=rk) :: uB(3*globBC%nElems(iLevel)) ! Velocity on boundary element
    real(kind=rk) :: uBTmp(3)
    integer :: iELem, iDir, bndNormalDir
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !--------------------------------------------------------------------------- 
    uB = 0.0_rk
    do iElem = 1, globBC%nElems(iLevel)
      do iDir = 1,layout%stencil(1)%QQ
        fTmp(?IDX?(iDir, iElem, layout%stencil(1)%QQ)) = bcBuffer(             &
& ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ) )
      enddo 
    enddo

    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )
    ! Get local velocity
    call varSys%derive( derVarPos%velocity )%do(                               &
      &                  input   = fTmp,                                       &
      &                  bound   = bound,                                      &
      &                  tree    = tree,                                       &
      &                  subTree = subTree,                                    &
      &                  requVar = varSys%Variable%val( derVarPos%velocity ),  &
      &                  globSys = varSys,                                     &
      &                  stencil = layout%stencil(1),                          &
      &                  conf    = varSys%solSpec_conf,                        &
      &                  res     = uB )
     
    do iElem = 1, globBC%nElems(iLevel)

!      rho = 0.0_rk
!      ! Get local density
!      do iDir = 1,layout%stencil(1)%QQ
!        rho = rho + fTmp(?IDX?(iDir, iElem, layout%stencil(1)%QQ))
!      enddo 
      uBTmp = uB((iElem-1)*3+1 : iELem*3) * me%slip%fac
      bndNormalDir = layout%stencil(1)%cxDirInv( globBC%elems( iLevel )%       &
        &                                                   normalInd( iElem ))
      if( abs(layout%stencil(1)%cxDir( 1, bndNormalDir )) .eq. 1)              &
        &                                                     uBTmp(1) = 0.0_rk
      if( abs(layout%stencil(1)%cxDir( 2, bndNormalDir )) .eq. 1)              &
        &                                                     uBTmp(2) = 0.0_rk
      if( abs(layout%stencil(1)%cxDir( 3, bndNormalDir )) .eq. 1)              &
        &                                                     uBTmp(3) = 0.0_rk
      
      do iDir = 1,layout%stencil(1)%QQ
        ! Write the values
        if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then
          ! Depending on PUSH or pull, use + or - for cxDir, because directions
          ! are inverted
          state(                                                               &
& ?FETCH?(varPos(iDir), globBC%elems(iLevel)%elem(iElem), nScalars, neigh), currT) = &
          ! We need to get post-collision pdf in direction
          ! alpha- outgoing direction, which is the inverse direction of bitmask
          ! For PULL this means, get the outgoing one, as this is the one which
          ! will be bounced back
          ! For PUSH this means, get the already bounced back pdf back, so take
          ! the incoming
            & fTmp(                                                            &
& ?IDX?(layout%stencil(1)%cxDirInv(iDir), iElem, layout%stencil(1)%QQ))        &
            &       - layout%weight( iDir )*6._rk*fieldProp%fluid%rho0         &
            &       * ( layout%stencil(1)%cxDir( 1, layout%stencil(1)%         &
            &                                       cxDirInv( iDir ))*uBTmp(1) &
            &       +   layout%stencil(1)%cxDir( 2, layout%stencil(1)%         &
            &                                       cxDirInv( iDir ))*uBTmp(2) &
            &       +   layout%stencil(1)%cxDir( 3, layout%stencil(1)%         &
            &                                       cxDirInv( iDir ))*uBTmp(3))
        endif
      enddo
    enddo !iElem

  end subroutine slip_wall
!******************************************************************************!


end module mus_bc_fluid_wall_module
!******************************************************************************!
