! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module contains general boundary routines 
!!
!! like initializing boundary, setting boundary at every time step for each 
!! field and other general boundary routines
!!
!! \author Manuel Hasert
?? include 'header/lbm_macros.inc'
module mus_bc_general_module

  ! include treelm modules
  use env_module,              only: rk, globalMaxLevels, long_k
  use tem_param_module,        only: cs
  use tem_tools_module,        only: tem_write
  use treelmesh_module,        only: treelmesh_type
  use tem_var_system_module,   only: tem_var_system_type
  use tem_bc_module,           only: tem_bc_state_type
  use tem_bc_prop_module,      only: tem_bc_prop_type
  use tem_topology_module,     only: tem_levelOF
  use tem_subTree_type_module, only: tem_subTree_type
  use tem_debug_module,        only: tem_debug

  ! include musubi modules
  use mus_field_module,                 only: mus_field_type
  use mus_field_prop_module,            only: mus_field_prop_type
  use mus_scheme_layout_module,         only: mus_scheme_layout_type
  use mus_pdf_module,                   only: pdf_data_type, pdf_global_type,  &
    &                                         pdf_globInfo_type 
  use mus_bc_header_module,             only: boundary_type, glob_boundary_type
  use mus_bc_fluid_module,              only: inlet_ubb, inlet_eq, outlet_nrbc,&
    &                                         outlet_nrbc_eq, outlet_dnt,      &
    &                                         outlet_expol, outlet_eq,         &
    &                                         outlet_pab, outlet_zero_prsgrd
  use mus_bc_fluid_wall_module,         only: wall_bouzidi, slip_wall,         &
    &                                         wall_multiReflection
  use mus_bc_fluid_experimental_module, only: outlet_bouzidi, test_bcNeighs,   &
    &                                         outlet_extrapolation,            &
    &                                         outlet_extrapolation_microscopic
  use mus_variable_module,              only: mus_derVarPos_type
  use mus_global_module,                only: param_type
  use mus_physics_module,               only: mus_physics_type

  implicit none

  private

  public :: set_boundary
  public :: init_boundary

contains


!******************************************************************************!
  !> Call the functions associated with each boundary condition
  !! 
  !! Loop over each field and Run over all the boundary conditions for 
  !! current iLevel and call the function pointer. 
  !! The function pointer was before assigned in init_boundary
  !! This routine is being called from the 
  !! \ref mus_control_module::do_recursive_multischeme "main control routine"
  !! before the compute (advection_relaxation) kernel call
  subroutine set_boundary( field, pdf, glob, tree, iLevel, pdfGlobal,          &
    &                      params, layout, varSys, derVarPos, globBC,          &
    &                      physics )
    !--------------------------------------------------------------------------- 
    !> fluid parameters and properties
    type( mus_field_type ), intent(inout) :: field(:)
    !> state and neighbor arrays for all iLevels
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !>gloabl type contains iLevel descriptor
    type( pdf_global_type ), intent(in) :: glob
    !> global treelm mesh 
    type( treelmesh_type ), intent(in) ::tree
    !> the iLevel on which this boundary was invoked
    integer, intent(in) :: iLevel
    !> global information 
    type( pdf_globInfo_type), intent(in) :: pdfGlobal 
    !> global parameters
    type(param_type),intent(in) :: params   
    !> stencil layout information
    type( mus_scheme_layout_type ), intent(in) ::layout
    !> scheme variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> position of derived quantities in varsys
    type( mus_derVarPos_type ), intent(in) :: derVarPos(:)
    !> scheme global boundary type
    type( glob_boundary_type ), intent(in) :: globBC(:)
    !> scheme global boundary type
    type( mus_physics_type ), intent(in) :: physics
    !------------------------------------------------------------------------ 
    integer :: iField
    integer :: iBnd   !< counter for the boundary condition
    !------------------------------------------------------------------------ 

    if ( pdfGlobal%nBCs .gt. 0 ) then
      call fill_neighBuffer(                                                   &
        &              currState = pdf( iLevel )%state( :, pdf(iLevel)%nNow ), &
        &              bcBuffer  = pdf( iLevel )%bcBuffer,                     &
        &              neigh     = pdf( iLevel )%neigh,                        &
        &              glob      = glob,                                       &
        &              globBC    = globBC,                                     &
        &              pdfGlobal = pdfGlobal,                                  &
        &              field     = field,                                      &
        &              varSys    = varSys,                                     &
        &              layout    = layout,                                     &
        &              iLevel    = iLevel )                                          

      ! loop over fields
      do iField = 1, varSys%nVars
      ! Treat all boundary conditions
        do iBnd = 1, pdfGlobal%nBCs         
          call tem_debug( level = 10,                                          &
            &             msg   = 'setting bnd'// trim( globBC( iBnd )%label ))
          if( associated( field(iField)%bc( iBnd )%fnct ))   then
              call field(iField)%bc( iBnd )%fnct(                              &
                &      state     = pdf( iLevel )%state( :, : ),                &
                ! we want to write into the current time step
                ! the field pointers have been inverted, so the newest
                ! information is stored in nNow
                ! So we want to proivde the information for the coming
                ! kernel update, which reads the info from nNow with FETCH
                &      currT        = pdf(iLevel)%nNow,                        &
                &      lastT        = pdf(iLevel)%nNext,                       &
                &      bcBuffer     = pdf( iLevel )%bcBuffer,                  &
                &      globBC       = globBC( iBnd ),                          &
                &      glob         = glob,                                    &
                &      tree         = tree,                                    &
                &      iLevel       = iLevel,                                  &
                &      timing       = pdfGlobal%timing,                        &
                &      params       = params,                                  &
                &      physics      = physics,                                 &
                &      IC_fields    = field(:)%ic,                             &
                &      iField       = iField,                                  &
                &      neigh        = pdf( iLevel )%neigh,                     &
                &      layout       = layout,                                  &
                &      fieldProp    = field(iField)%fieldProp,                 &
                &      varSys       = varSys,                                  &
                &      derVarPos    = derVarPos(iField),                       &
                &      derVarPosAll = derVarPos(:),                            &
                &      varPos       = varSys%Variable%val(iField)%varPos,      &
                &      nScalars     = varSys%nScalars )
          end if
        end do
      end do  
    end if

  end subroutine set_boundary
!******************************************************************************!


!******************************************************************************!
  ! > This routine call init_boundary to initialize boundary for each field 
  !!
  subroutine init_boundary( field, pdf, tree, glob, pdfGlobal, layout, varSys, &
    &                       derVarPos, globBC, bc_prop )
    !---------------------------------------------------------------------------
    !> fluid parameters and properties
    type( mus_field_type ), intent(in) :: field(:)
    !> global fluid parameters 
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !> global treelm mesh
    type( treelmesh_type ), intent(in) ::tree
    !> global info type
    type( pdf_globInfo_type ), intent(in) :: pdfGlobal
    !> scheme layout type
    type( mus_scheme_layout_type ), intent(in) ::layout
    !> global pdf type
    type( pdf_global_type ), intent(in) :: glob
    !> scheme variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> position of derived quantities in varsys
    type( mus_derVarPos_type ), intent(in) :: derVarPos(:)
    !> scheme global boundary type
    type( glob_boundary_type ), intent(inout) :: globBC(:)
    !> boundary property type
    type( tem_bc_prop_type ) :: bc_prop
    !---------------------------------------------------------------------------
    integer :: iField
    !---------------------------------------------------------------------------

    if ( pdfGlobal%nBCs .gt. 0 ) then
      do iField = 1, size(field)
        call init_boundary_single(                                             &
          &                    bc        = field(iField)%bc,                   &
          &                    pdf       = pdf,                                &
          &                    tree      = tree,                               &
          &                    fieldProp = field(iField)%fieldProp,            &
          &                    glob      = glob,                               &
          &                    pdfGlobal = pdfGlobal,                          &
          &                    layout    = layout,                             &
          &                    varPos    = varSys%Variable%val(iField)%varPos, &
          &                    nScalars  = varSys%nScalars,                    &
          &                    varSys    = varSys,                             &
          &                    dervarPos = derVarPos(iField),                  &
          &                    globBC    = globBC,                             &
          &                    bc_prop   = bc_prop )
      end do
    end if

  end subroutine init_boundary
!******************************************************************************!


!******************************************************************************!
  !> This subroutine sets the right boundary conditions for the different
  !! boundaries.
  !!
  subroutine init_boundary_single(bc, pdf, tree, fieldProp, glob, pdfGlobal,   &
    &            layout, varPos, nScalars, varSys, derVarPos, globBC, bc_prop )
    !--------------------------------------------------------------------------- 
    type( boundary_type ) :: bc(:) !< global array boundary type
    !> contains global state vector
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !> global treelm mesh
    type( treelmesh_type ), intent(in) ::tree
    !> fluid parameters
    type( mus_field_prop_type ), intent(in) :: fieldProp
    !> global info type
    type( pdf_globInfo_type ), intent(in) :: pdfGlobal
    !> scheme layout type
    type( mus_scheme_layout_type ), intent(in) ::layout
    !> global pdf type
    type( pdf_global_type ), intent(in) :: glob
    !> varPos of current field variable
    integer, intent(in) :: varPos(:)
    !> number of scalars in global system
    integer, intent(in) :: nScalars
    !> scheme variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> position of derived quantities in varsys
    type( mus_derVarPos_type ), intent(in) :: derVarPos
    !> scheme global boundary type
    type( glob_boundary_type ), intent(inout) :: globBC(:)
    !> boundary property type
    type( tem_bc_prop_type ) :: bc_prop
    !--------------------------------------------------------------------------- 
    integer :: iBnd
    !--------------------------------------------------------------------------- 

    !> The boundary conditions were read in the order of config.lua
    !! the order in the boundary condition description file however 
    !! might be different.
    !! We have to match the labels.
    call tem_write( 'Setting up the boundary conditions', pdfGlobal%nBCs )

    do iBnd = 1, pdfGlobal%nBCs
      if( trim( bc( iBnd )%label ) .ne. trim( globBC( iBnd )%label )) then
        call tem_write( 'Error: The boundary with label from the mesh:'        &
          &              //trim( globBC(iBnd)%label ) )
        call tem_write( 'does not match the one specified in the lua file:'    &
          &              //trim(bc( iBnd  )%label))
        call tem_write( 'Stopping ...' )
        stop
      end if
      call tem_write( 'Initializing BC with label:'//trim(bc( iBnd  )%label))
      select case (trim(bc( iBnd  )%BC_kind))
      case('wall')
      case('wall_linearInterpolation')
        bc( iBnd )%fnct => wall_bouzidi
        ! Here we have to allocate and set the q-values if it is not
        ! provided by seeder and set the qVal from musubi.lua
        if(.not. globBC(iBnd)%hasQVal) then
          call init_qVals( bc        = bc( iBnd ),                             &
            &              tree      = tree,                                   &
            &              bc_prop   = bc_prop,                                &
            &              pdfGlobal = pdfGlobal,                              &
            &              layout    = layout,                                 &
            &              globBC    = globBC(iBnd),                           &
            &              iBnd      = int(iBnd, kind=long_k) )
        endif
      case('wall_multireflection','wall_mr')
        bc( iBnd )%fnct => wall_multiReflection
        ! Here we have to allocate and set the q-values
        !call init_qVals( bc = bc( iBnd ), tree = tree, bc_prop = bc_prop,      &
        !  &              pdfGlobal = pdfGlobal, layout = layout,               &
        !  &              globBC = globBC(iBnd), iBnd = int(iBnd, kind=long_k) )
      case('slip_wall')
        bc( iBnd )%fnct => slip_wall
      case('inlet_ubb')
        bc( iBnd )%fnct => inlet_ubb
      case('inlet_eq')
        bc( iBnd )%fnct => inlet_eq
      case('outlet_nrbc','outlet_nrbc_eq')
        select case (trim(bc( iBnd  )%BC_kind))
        case( 'outlet_nrbc' )
        bc( iBnd )%fnct => outlet_nrbc
        case( 'outlet_nrbc_eq' )
        bc( iBnd )%fnct => outlet_nrbc_eq
        end select
        call init_nrbc( bc        = bc(iBnd),                                  &
          &             pdf       = pdf,                                       &
          &             fieldProp = fieldProp,                                 &
          &             layout    = layout,                                    &
          &             pdfGlobal = pdfGlobal,                                 &
          &             varPos    = varPos,                                    &
          &             nScalars  = nScalars,                                  &
          &             varSys    = varSys,                                    &
          &             derVarPos = derVarPos,                                 &
          &             globBC    = globBC(iBnd),                              &
          &             tree      = tree ) 

      case('outlet_zfg','outlet_extrapolation_microscopic')
        bc( iBnd )%fnct => outlet_extrapolation_microscopic
        call tem_write( 'Microscopic extrapolation. Zero Gradient for pdfs')

      case('outlet_dnt')
        bc( iBnd )%fnct => outlet_dnt
      case('outlet_expol')
        bc( iBnd )%fnct => outlet_expol
      case('outlet_eq')
        bc( iBnd )%fnct => outlet_eq
      case('outlet_pab')
        bc( iBnd )%fnct => outlet_pab 
      case('outlet_bouzidi')
        bc( iBnd )%fnct => outlet_bouzidi 
      case('outlet_zero_prsgrd')
        bc( iBnd )%fnct => outlet_zero_prsgrd 
      case('test')
        bc( iBnd )%fnct => test_bcNeighs
      case default 
        call tem_write('BC type '//trim(bc( iBnd  )%label)//'not found')
      end select
    end do

  end subroutine init_boundary_single
!******************************************************************************!



!******************************************************************************!
  !> JQ: assign qVal to corresponding BC and level-wise
  !! todo: first check if qVal is given by files,
  !! if not, try to receive qVal from lua file.
  subroutine init_qVals( bc, tree, bc_prop, pdfGlobal, layout, globBC, iBnd )
    !---------------------------------------------------------------------------
    type( boundary_type ) :: bc !< global array boundary type
    !> fluid tree from mesh
    type( treelmesh_type ), intent(in)        :: tree
    !> boundary information from mesh
    type( tem_bc_prop_type ), intent(in)      :: bc_prop
    !> contains pdf global information
    type( pdf_globInfo_type ), intent(in)     :: pdfGlobal
    type( mus_scheme_layout_type), intent(in) :: layout
    type( glob_boundary_type ), intent(inout) :: globBC
    integer(kind=long_k), intent(in)          :: iBnd
    !---------------------------------------------------------------------------
    integer, allocatable :: nBnds(:) ! number of bc elements in each level
    integer(kind=long_k) :: bID
    integer :: iElem, iLevel, level, iDir, iQQQ, iElem_qVal
    logical :: found
    !---------------------------------------------------------------------------

    allocate( nBnds( globalMaxLevels ))
    nBnds = 0
    ! element pointer to qVal array in bc_prop
    iElem_qVal = 0

    ! 2. allocate qVal array level-wise
    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      allocate( globBC%elems( iLevel )%qVal(                                   &
        &       layout%stencil(1)%QQN, globBC%nElems( iLevel )))
      globBC%elems( iLevel )%qVal = -1._rk
    end do

    ! loop over all boundary elements
    do iElem = 1, tree%property(1)%nElems

      found = .false.

      ! count elements that has prp_hasQVal property
      iElem_qVal = iElem_qval + 1

      ! run over each direction in the stencil
      do iDir = 1, layout%stencil(1)%QQN
        ! Stencil order of directions might differ from the treelm numbering
        ! iDir is musubi numbering
        ! iQQQ is treelm numbering.
        iQQQ = layout%stencil(1)%map( iDir )
        ! Get the boundary ID as stored in the mesh
        bID = bc_prop%boundary_ID( iQQQ, iElem )

        if( bID == iBnd ) then
          level = tem_LevelOf( tree%treeID( tree%property(1)%ElemID( iElem )))

          if( .not. found ) then
            nBnds( level ) = nBnds( level ) + 1
            found = .true.
          end if

          globBC%elems( level )%qVal( iDir, nBnds(level))  = bc%qVal
        end if
      end do ! iDir
    end do ! iElem

  end subroutine init_qVals
!******************************************************************************!


!******************************************************************************!
  !> Initialize the values required for the characteristic boundary conditions
  !!
  !! Simply calculate the macroscopic values from the current pdf distributions
  !! in the boundary element and its neighbor elements and store it to the 
  !! %nrbc%lodi field
  !! NOTE:   
  !! To be consistent, the same calculation procedure of the LODI values need to
  !! be performed as in the 
  !! \ref mus_bc_fluid_module::outlet_nrbc "NRBC routine" itself
  !!
  subroutine init_nrbc( bc, pdf, fieldProp, layout, pdfGlobal, varPos,         &
    &                   nScalars, varSys, derVarPos, globBC, tree )
    !---------------------------------------------------------------------------
    !> global array boundary type
    type( boundary_type ) :: bc
    !> global fluid parameters 
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !> fluid parameters
    type( mus_field_prop_type ), intent(in) :: fieldProp
    type( mus_scheme_layout_type), intent(in) :: layout
    type( pdf_globInfo_type ), intent(in) :: pdfGlobal 
    !> varPos of current field variable
    integer, intent(in) :: varPos(:)
    !> number of scalars in global system
    integer, intent(in) :: nScalars
    !> scheme variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> position of derived quantities in varsys
    type( mus_derVarPos_type ), intent(in) :: derVarPos
    !> scheme global boundary type
    type( glob_boundary_type ), intent(in) :: globBC
    !> global treelm mesh
    type( treelmesh_type ), intent(in) ::tree
    !---------------------------------------------------------------------------
    integer :: iLevel, iElem, iDir
    ! Neighbor element variables
    integer :: iNeigh, neighPos
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !---------------------------------------------------------------------------

    if( bc%nrbc%lodi_length .eq. 0._rk) &
        bc%nrbc%lodi_length = 1.0_rk
    call tem_write( 'Initializing NRBC with parameters:' )
    call tem_write( '    kappa:         ',bc%nrbc%kappa)
    call tem_write( '    sigma:         ',bc%nrbc%sigma)
    call tem_write( '    Char. length:  ',bc%nrbc%lodi_length)
    bc%nrbc%cs_mod = cs*sqrt( bc%nrbc%kappa )

    ! Assign the initial values to the lodi variables
    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      allocate( bc%elems( iLevel )%lodi( 4, 3, globBC%nElems( iLevel )))
      ! reset 
      do iElem = 1, globBC%nElems( iLevel )
        ! adopt initial density from the fluid cells
        bc%elems( iLevel )%lodi( :, 1, iElem) = 0._rk
        do iDir = 1,layout%stencil(1)%QQ
          fTmp(iDir) = pdf(iLevel)%state(                                      &
& ?SAVE?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars,pdf(iLevel)%neigh ), &
& pdf( iLevel )%nNext ) 

          bc%elems( iLevel )%lodi( 1, 1, iElem ) =                             &
            &               bc%elems( iLevel )%lodi( 1, 1, iElem ) + fTmp(iDir) 
        end do    
          
        ! initialize the lower and upper bound
        bound(1) = 1
        bound(2) = 1
        ! adapt initial velocities from the fluid cells
        call varSys%derive( derVarPos%velocity )%do(                           &
          &                input   = fTmp,                                     &
          &                bound   = bound,                                    &
          &                tree    = tree,                                     &
          &                subTree = subTree,                                  &
          &                requVar = varSys%Variable%val( derVarPos%velocity ),&
          &                globSys = varSys,                                   &
          &                stencil = layout%stencil(1),                        &
          &                conf    = varSys%solSpec_conf,                      &
          &                res     = bc%elems(iLevel)%lodi(2:4,1,iElem) )

        do iNeigh = 1,2
          ! Get the values for the neighbors as well
          neighPos = bc%elems( iLevel )%neigh( iNeigh, iElem )
          bc%elems( iLevel )%lodi( :, 1+iNeigh, iElem) = 0._rk
          if( neighPos .gt. 0 ) then
            do iDir = 1,layout%stencil(1)%QQ
              fTmp(iDir) = pdf(iLevel)%state(                                  &
& ?SAVE?( varPos(iDir), neighPos, nScalars, pdf(iLevel)%neigh ), pdf( iLevel )%nNext ) 
              bc%elems( iLevel )%lodi( 1, 1+iNeigh, iElem ) =                  &
                &    bc%elems( iLevel )%lodi( 1, 1+iNeigh, iElem ) + fTmp(iDir)
            end do

            ! velocities
            call varSys%derive( derVarPos%velocity )%do(                       &
              &            input   = fTmp,                                     &
              &            bound   = bound,                                    &
              &            tree    = tree,                                     &
              &            subTree = subTree,                                  &
              &            requVar = varSys%Variable%val( derVarPos%velocity ),&
              &            globSys = varSys,                                   &
              &            stencil = layout%stencil(1),                        &
              &            conf    = varSys%solSpec_conf,                      &
              &            res     = bc%elems(iLevel)%lodi(2:4,1+iNeigh,iElem))

          end if
        end do
      end do
    end do

  end subroutine init_nrbc
!******************************************************************************!


!******************************************************************************!
  !> This routine will fill BC buffer with state values of boundary element
  !! of all fields from current time step
  !!
  !! \todo KM: Please give reasons why the access patterns are used
  !! Access to state array
  !! ---------------------
  !! * bcBuffer:        SAVE 
  !! * NeighBufferPre: FETCH
  !! * NeighBufferPost: SAVE 
  !!
  subroutine fill_neighBuffer( currState, bcBuffer, neigh, glob, globBC,       &
    &                          pdfGlobal, field, varSys, layout, iLevel )
    !--------------------------------------------------------------------------- 
    !> Current state vector of iLevel  
    real(kind=rk),intent(in) :: currState(:) 
    !> state values of boundary elements of all fields of iLevel
    real(kind=rk),intent(inout) :: bcBuffer(:) 
    !> connectivity array corresponding to state vector
    integer,intent(in) :: neigh(:)       
    !>gloabl type contains iLevel descriptor
    type( pdf_global_type ), intent(in) :: glob
    !> scheme global boundary type
    type( glob_boundary_type ), intent(in) :: globBC(:)
    !> global information 
    type( pdf_globInfo_type), intent(in) :: pdfGlobal 
    !> fluid parameters and properties
    type( mus_field_type ), intent(inout) :: field(:)
    !> scheme variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> stencil layout information
    type( mus_scheme_layout_type ), intent(in) ::layout
    !> the iLevel on which this boundary was invoked
    integer, intent(in) :: iLevel
    !---------------------------------------------------------------------------
    integer :: iField, iElem, iDir, iBnd, iNeigh
    integer :: pos, elemPos
    !---------------------------------------------------------------------------

    do iField = 1, varSys%nVars
      do iElem = 1, glob%levelDesc( iLevel )%bc_elemBuffer%nVals 
        elemPos =  glob%levelDesc( iLevel )%bc_elemBuffer%val( iElem )
        do iDir = 1, layout%stencil(1)%QQ
          pos = varSys%Variable%val( iField )%varPos( iDir )
          bcBuffer( ?IDX?( pos, iElem, varSys%nScalars ) ) = currState( &
& ?SAVE?( pos, elemPos, varSys%nScalars, neigh ) )
          ! MH: Why do we use SAVE access here?
          ! KM: Store final result of compute kernel 
        end do !iDir
      end do !iElem
    end do !iField

    !fill boundary neighbor pre- and post-collision state values 
    do iField = 1, varSys%nVars
      do iBnd = 1, pdfGlobal%nBCs
        do iNeigh = 1, field( iField )%bc( iBnd )%nNeighs
          do iElem = 1, globBC( iBnd )%nElems( iLevel )
            elemPos =  field( iField )%bc( iBnd )%elems( iLevel )%             &
              &                                          neigh( iNeigh, iElem )
            do iDir = 1, layout%stencil(1)%QQ
              pos = varSys%Variable%val( iField )%varPos( iDir )
              !pre-collision neigh
              field( iField )%bc( iBnd )%elems( iLevel )%neighBufferPre(       &          
& iNeigh, ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) ) = currState(            & 
& ?FETCH?( pos, elemPos, varSys%nScalars, neigh ) )
              !post-collision neigh  
              field( iField )%bc( iBnd )%elems( iLevel )%neighBufferPost(      &          
& iNeigh, ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) ) = currState(            & 
& ?SAVE?( pos, elemPos, varSys%nScalars, neigh ) )
            end do !iDir
          end do !iElem
        end do !iNeigh
      end do !iBnd
    end do !iField

  end subroutine fill_neighBuffer
!******************************************************************************!


end module mus_bc_general_module
!******************************************************************************!
