! See copyright notice in the COPYRIGHT file.
?? include 'header/lbm_macros.inc'
!******************************************************************************!
!> This module provides the definition and methods for
!! BGK advection relaxation scheme.
module mus_bgk_module

  ! include treelm modules
  use env_module,               only: rk
  use tem_var_system_module,    only: tem_var_system_type
  use tem_param_module,         only: div1_3, div1_36, div1_8, div3_4h, div1_4,&
    &                                 div3_8, div9_16, div3_16, cs2inv, cs4inv
  use tem_spacetime_fun_module, only: tem_spacetime_fun_type, tem_spacetime_for
  use treelmesh_module,         only: treelmesh_type
  use tem_comm_env_module,      only: process
  use tem_tools_module,         only: tem_write

  ! include musubi modules
  use mus_field_prop_module,    only: mus_field_prop_type
  use mus_scheme_layout_module, only: mus_scheme_layout_type
  use mus_pdf_module,           only: pdf_globInfo_type
  use mus_scheme_type_module,   only: mus_scheme_type
  use mus_global_module,        only: param_type
  use mus_solSpecHelpers_module, only: getConversionFac

  implicit none

  private

  public :: bgk_advRel_d3q7
  public :: bgk_advRel_explicit
  public :: bgk_advRel_explicit_incomp
  public :: bgk_advRel_forcing
  public :: bgk_advRel_nonNewtonian_incomp

contains

!******************************************************************************!
  !> Advection relaxation routine for the d3q7 BGK model.
  !! 
  subroutine bgk_advRel_d3q7( iScheme, fieldProp, inState, outState, neigh,    &
    &    nStart, nEnd, level, layout, pdfGlobal, params, scheme, varSys, tree )
    !---------------------------------------------------------------------------
    !> current scheme
    integer, intent(in) :: iScheme
    !> global fluid
    type(mus_field_prop_type), intent(in) :: fieldProp(:)
    !> input pdf vector
    real(kind=rk), dimension(:), intent(in)  :: inState
    !> output pdf vector
    real(kind=rk), dimension(:), intent(out) :: outState
    !> connectivity vector 
    integer, dimension(:), intent(in) :: neigh
    !> start positions in pdf vector 
    integer, intent(in) :: nStart
    !> end positions in pdf vector 
    integer, intent(in) :: nEnd
    !> current level
    integer,intent(in) :: level
    !> current layout
    type(mus_scheme_layout_type), intent(in) :: layout
    !> global info type
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    !> global parameters
    type(param_type),intent(in) :: params
    !> array of all schemes
    type( mus_scheme_type ), intent(in) :: scheme(:)
    !> variable system definition
    type(tem_var_system_type), intent(in) :: varSys
    !> treelm mesh information
    type( treelmesh_type ), intent(in) :: tree
    !---------------------------------------------------------------------------
    integer :: iElem      ! voxel element counter
    real(kind=rk) pdfTmp(layout%stencil(1)%QQ)    ! temporary local pdf values
    real(kind=rk) rho     ! local density
    real(kind=rk) inv_rho ! inverse local density
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) u_z     ! local z-velocity
    real(kind=rk) usq     ! square velocity
    real(kind=rk) ux_sq, uy_sq, uz_sq
    ! derived constants
    real(kind=rk) sum1, sum_usq, sum_ux, sum_ux_sq, sum_uy, sum_uy_sq,         &
      &           sum_uz, sum_uz_sq
    real(kind=rk) omega, cmpl_o
    !---------------------------------------------------------------------------

    ! some global variables
    omega = fieldProp(1)%fluid%omLvl( level )
    cmpl_o  = 1._rk - omega

    nodeloop: do iElem=nStart,nEnd       
      !> First load all local values into temp array
      !! Generic! PUSH+PULL is possible
      pdfTmp(  1 ) = inState( ?FETCH?(  1, iElem, varSys%nScalars, neigh ))
      pdfTmp(  2 ) = inState( ?FETCH?(  2, iElem, varSys%nScalars, neigh ))
      pdfTmp(  3 ) = inState( ?FETCH?(  3, iElem, varSys%nScalars, neigh ))
      pdfTmp(  4 ) = inState( ?FETCH?(  4, iElem, varSys%nScalars, neigh ))
      pdfTmp(  5 ) = inState( ?FETCH?(  5, iElem, varSys%nScalars, neigh ))
      pdfTmp(  6 ) = inState( ?FETCH?(  6, iElem, varSys%nScalars, neigh ))
      pdfTmp(  7 ) = inState( ?FETCH?(  7, iElem, varSys%nScalars, neigh ))
 
 
      ! local density
      rho = pdfTmp( 1) + pdfTmp( 2) + pdfTmp( 3) &
        & + pdfTmp( 4) + pdfTmp( 5) + pdfTmp( 6) &
        & + pdfTmp( 7) 

      ! local x-, y- and z-velocity
      u_x = pdfTmp( 4) - pdfTmp( 1)
                         
      u_y = pdfTmp( 5) - pdfTmp( 2)
                         
      u_z = pdfTmp( 6) - pdfTmp( 3)

      ! inverse local density
      inv_rho = 1._rk / rho

      ! transfer moments to velocities
      u_x = u_x * inv_rho
      u_y = u_y * inv_rho
      u_z = u_z * inv_rho

      ! square velocity and derived constants

      ux_sq = u_x*u_x
      uy_sq = u_y*u_y
      uz_sq = u_z*u_z
      usq = ux_sq + uy_sq + uz_sq
     

      outState( ?SAVE?( 7,iElem,varSys%nScalars,neigh )) =                     &
        &                      pdfTmp(7)*cmpl_o+omega*rho*( div1_4-div3_8*usq )

      sum_usq = div3_16*usq*omega*rho
      sum1 = div1_8*omega*rho

      sum_ux_sq = div9_16*ux_sq*omega*rho
      sum_ux = div3_8*u_x*omega*rho

      sum_uy_sq = div9_16*uy_sq*omega*rho
      sum_uy = div3_8*u_y*omega*rho
      
      sum_uz_sq = div9_16*uz_sq*omega*rho
      sum_uz = div3_8*u_z*omega*rho

      outState( ?SAVE?( 1,iElem,varSys%nScalars,neigh )) =                     &
        &                       pdfTmp(1) *cmpl_o+sum1-sum_ux+sum_ux_sq-sum_usq
      outState( ?SAVE?( 2,iElem,varSys%nScalars,neigh )) =                     &
        &                       pdfTmp(2) *cmpl_o+sum1-sum_uy+sum_uy_sq-sum_usq
      outState( ?SAVE?( 3,iElem,varSys%nScalars,neigh )) =                     &
        &                       pdfTmp(3) *cmpl_o+sum1-sum_uz+sum_uz_sq-sum_usq
      outState( ?SAVE?( 4,iElem,varSys%nScalars,neigh )) =                     &
        &                       pdfTmp(4) *cmpl_o+sum1+sum_ux+sum_ux_sq-sum_usq
      outState( ?SAVE?( 5,iElem,varSys%nScalars,neigh )) =                     &
        &                       pdfTmp(5) *cmpl_o+sum1+sum_uy+sum_uy_sq-sum_usq
      outState( ?SAVE?( 6,iElem,varSys%nScalars,neigh )) =                     &
        &                       pdfTmp(6) *cmpl_o+sum1+sum_uz+sum_uz_sq-sum_usq

    end do nodeloop

  end subroutine bgk_advRel_d3q7
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the BGK model.
  !! 
  subroutine bgk_advRel_forcing( iScheme, fieldProp, inState, outState, neigh, &
    &    nStart, nEnd, level, layout, pdfGlobal, params, scheme, varSys, tree )
    !---------------------------------------------------------------------------
    !> current scheme
    integer, intent(in) :: iScheme
    !> global fluid
    type(mus_field_prop_type), intent(in) :: fieldProp(:)
    !> input pdf vector
    real(kind=rk), dimension(:), intent(in)  :: inState
    !> output pdf vector
    real(kind=rk), dimension(:), intent(out) :: outState
    !> connectivity vector 
    integer, dimension(:), intent(in) :: neigh
    !> start positions in pdf vector 
    integer, intent(in) :: nStart
    !> end positions in pdf vector 
    integer, intent(in) :: nEnd
    !> current level
    integer,intent(in) :: level
    !> current layout
    type(mus_scheme_layout_type), intent(in) :: layout
    !> global info type
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    !> global parameters
    type(param_type),intent(in) :: params
    !> array of all schemes
    type( mus_scheme_type ), intent(in) :: scheme(:)
    !> variable system definition
    type(tem_var_system_type), intent(in) :: varSys
    !> treelm mesh information
    type( treelmesh_type ), intent(in) :: tree
    !---------------------------------------------------------------------------
    integer :: iElem, iDir                       ! voxel element counter
    real(kind=rk) pdfTmp( layout%stencil(1)%QQ ) ! temporary local pdf values
    real(kind=rk) rho     ! local density
    real(kind=rk) inv_rho ! inverse local density
    real(kind=rk) force( layout%stencil(1)%QQ )  ! local x-velocity
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) u_z     ! local z-velocity
    real(kind=rk) usq     ! square velocity
    ! derived constants
    real(kind=rk) usqn, usqn_o1, usqn_o2 
    real(kind=rk) omega_2, cmpl_o, omega
    real(kind=rk) coeff_1, coeff_2
    real(kind=rk) ui1, ui3, ui10, ui11, ui12, ui13
    real(kind=rk) fac_1, fac_2, fac_3, fac_4, fac_9, fac_10, fac_11, fac_12,   &
      &           fac_13
    real(kind=rk) sum1_1, sum1_2, sum2_1, sum2_2, sum3_1, sum3_2, sum4_1,      &
      &           sum4_2, sum9_1, sum9_2, sum10_1, sum10_2, sum11_1, sum11_2,  &
      &           sum12_1, sum12_2, sum13_1, sum13_2
    !---------------------------------------------------------------------------

    ! some global variables
    omega = fieldProp(1)%fluid%omLvl( level )
    omega_2 = 2._rk * omega
    cmpl_o  = 1._rk - omega

!$omp do schedule(static)
    nodeloop: do iElem=nStart,nEnd       
      ! First load all local values into temp array
      ! Generic! PUSH+PULL is possible
      pdfTmp(  1 ) = inState( ?FETCH?(  1, iElem, varSys%nScalars, neigh ))
      pdfTmp(  2 ) = inState( ?FETCH?(  2, iElem, varSys%nScalars, neigh ))
      pdfTmp(  3 ) = inState( ?FETCH?(  3, iElem, varSys%nScalars, neigh ))
      pdfTmp(  4 ) = inState( ?FETCH?(  4, iElem, varSys%nScalars, neigh ))
      pdfTmp(  5 ) = inState( ?FETCH?(  5, iElem, varSys%nScalars, neigh ))
      pdfTmp(  6 ) = inState( ?FETCH?(  6, iElem, varSys%nScalars, neigh ))
      pdfTmp(  7 ) = inState( ?FETCH?(  7, iElem, varSys%nScalars, neigh ))
      pdfTmp(  8 ) = inState( ?FETCH?(  8, iElem, varSys%nScalars, neigh ))
      pdfTmp(  9 ) = inState( ?FETCH?(  9, iElem, varSys%nScalars, neigh ))
      pdfTmp( 10 ) = inState( ?FETCH?( 10, iElem, varSys%nScalars, neigh ))
      pdfTmp( 11 ) = inState( ?FETCH?( 11, iElem, varSys%nScalars, neigh ))
      pdfTmp( 12 ) = inState( ?FETCH?( 12, iElem, varSys%nScalars, neigh ))
      pdfTmp( 13 ) = inState( ?FETCH?( 13, iElem, varSys%nScalars, neigh ))
      pdfTmp( 14 ) = inState( ?FETCH?( 14, iElem, varSys%nScalars, neigh ))
      pdfTmp( 15 ) = inState( ?FETCH?( 15, iElem, varSys%nScalars, neigh ))
      pdfTmp( 16 ) = inState( ?FETCH?( 16, iElem, varSys%nScalars, neigh ))
      pdfTmp( 17 ) = inState( ?FETCH?( 17, iElem, varSys%nScalars, neigh ))
      pdfTmp( 18 ) = inState( ?FETCH?( 18, iElem, varSys%nScalars, neigh ))
      pdfTmp( 19 ) = inState( ?FETCH?( 19, iElem, varSys%nScalars, neigh ))
 
 
 
      ! local density
      rho = pdfTmp(18) + pdfTmp( 5) + pdfTmp(16)                               &
        & + pdfTmp( 1) + pdfTmp(15) + pdfTmp( 2)                               &
        & + pdfTmp(17) + pdfTmp( 4) + pdfTmp( 6)                               &
        & + pdfTmp(14) + pdfTmp(10) + pdfTmp(13)                               &
        & + pdfTmp( 8) + pdfTmp( 3) + pdfTmp(12)                               &
        & + pdfTmp( 9) + pdfTmp(11) + pdfTmp( 7)                               &
        & + pdfTmp(19)

      ! local x-, y- and z-velocity
!      u(1:3) = getMomentumForSubset( pdfTmp, fluid, level, layout )
      u_x = pdfTmp(18) + pdfTmp(17) + pdfTmp( 4)                               &
        & + pdfTmp(14) + pdfTmp(12)                                            &
        & - pdfTmp(16) - pdfTmp( 1) - pdfTmp(15)                               &
        & - pdfTmp(13) - pdfTmp(11)
                        
      u_y = pdfTmp(18) + pdfTmp( 5)  + pdfTmp(16)                              &
        &  + pdfTmp(10) + pdfTmp( 9)                                           &
        &  - pdfTmp(15) - pdfTmp( 2 ) - pdfTmp(17)                             &
        &  - pdfTmp( 8) - pdfTmp( 7)
                        
      u_z = pdfTmp( 6) + pdfTmp(14) + pdfTmp(10)                               &
        & + pdfTmp(13) + pdfTmp( 8)                                            & 
        & - pdfTmp( 3) - pdfTmp(12) - pdfTmp( 9)                               &
        & - pdfTmp(11) - pdfTmp( 7)

      ! inverse local density
      inv_rho = 1._rk / rho

      ! transfer moments to velocities
      u_x = u_x * inv_rho
      u_y = u_y * inv_rho
      u_z = u_z * inv_rho

      ! square velocity and derived constants
      usq = u_x*u_x + u_y*u_y + u_z*u_z
      usqn = div1_36 * (1._rk - 1.5_rk * usq) * rho

      ! Calculate the force for all directions
      do iDir = 1, layout%stencil(1)%QQ
        force( iDir ) = ( 1._rk -0.5_rk*omega )*layout%Weight( iDir )          &
          &           * ((( layout%stencil(1)%cxDir( 1, iDir ) - u_x  )        &
          &           * cs2inv + ( layout%stencil(1)%cxDir( 1, iDir ) * u_x    &
          &           +            layout%stencil(1)%cxDir( 2, iDir ) * u_y    &
          &           +            layout%stencil(1)%cxDir( 3, iDir ) * u_z )  &
          &           *     cs4inv*layout%stencil(1)%cxDir( 1, iDir ))         &
          &           * fieldProp(1)%fluid%force( 1 )                          &
          &           + (( layout%stencil(1)%cxDir( 2, iDir ) - u_y ) * cs2inv &
          &           +  ( layout%stencil(1)%cxDir( 2, iDir ) * u_x            &
          &           +    layout%stencil(1)%cxDir( 2, iDir ) * u_y            &
          &           +    layout%stencil(1)%cxDir( 3, iDir ) * u_z )          &
          &           * cs4inv*layout%stencil(1)%cxDir( 2, iDir ))             &
          &           * fieldProp(1)%fluid%force( 2 )                          &
          &           + (( layout%stencil(1)%cxDir( 3, iDir ) - u_z ) * cs2inv &
          &           +  ( layout%stencil(1)%cxDir( 1, iDir ) * u_x            &
          &           +    layout%stencil(1)%cxDir( 2, iDir ) * u_y            &
          &           +    layout%stencil(1)%cxDir( 3, iDir ) * u_z )          &
          &           * cs4inv*layout%stencil(1)%cxDir( 3, iDir ))             &
          &           * fieldProp(1)%fluid%force( 3 ))
      end do

      outState( ?SAVE?( 19,iElem,varSys%nScalars,neigh ))                      &
        &                   = pdfTmp(19)*cmpl_o+omega*rho*( div1_3-0.5d0*usq ) &
        &                   + pdfGlobal%dtLB( level )*force( 19 )

      coeff_1 = div1_8 * omega * rho

      usqn_o1 = omega * usqn

      ui1     =  u_x + u_y
      fac_1   = coeff_1 * ui1
      sum1_1  = fac_1 * div3_4h
      sum1_2  = fac_1 * ui1 + usqn_o1

      outState( ?SAVE?( 18,iElem,varSys%nScalars,neigh ))                      &
        &                                 = pdfTmp(18) *cmpl_o+sum1_1 +sum1_2  &
        &                                 + pdfGlobal%dtLB( level )*force( 18 )
      outState( ?SAVE?( 15,iElem,varSys%nScalars,neigh ))                      &
        &                                 = pdfTmp(15) *cmpl_o-sum1_1 +sum1_2  &
        &                                 + pdfGlobal%dtLB( level )*force( 15 )

      ui3     = -u_x + u_y
      fac_3   = coeff_1 * ui3
      sum3_1  = fac_3 * div3_4h
      sum3_2  = fac_3 * ui3 + usqn_o1

      outState( ?SAVE?( 16,iElem,varSys%nScalars,neigh ))                      &
        &                                 = pdfTmp(16) *cmpl_o+sum3_1 +sum3_2  &
        &                                 + pdfGlobal%dtLB( level )*force( 16 )

      outState( ?SAVE?( 17,iElem,varSys%nScalars,neigh ))                      &
        &                                 = pdfTmp(17) *cmpl_o-sum3_1 +sum3_2  &
        &                                 + pdfGlobal%dtLB( level )*force( 17 )


      ui10    =  u_x + u_z
      fac_10  = coeff_1 * ui10
      sum10_1 = fac_10 * div3_4h
      sum10_2 = fac_10 * ui10 + usqn_o1

      outState( ?SAVE?( 14,iElem,varSys%nScalars,neigh ))                      &
        &                                 = pdfTmp(14)*cmpl_o+sum10_1+sum10_2  &
        &                                 + pdfGlobal%dtLB( level )*force( 14 )

      outState( ?SAVE?( 11,iElem,varSys%nScalars,neigh ))                      &
        &                                 = pdfTmp(11)*cmpl_o-sum10_1+sum10_2  &
        &                                 + pdfGlobal%dtLB( level )*force( 11 )


      ui12    = -u_x + u_z
      fac_12  = coeff_1 * ui12
      sum12_1 = fac_12 * div3_4h
      sum12_2 = fac_12 * ui12 + usqn_o1

      outState( ?SAVE?( 13,iElem,varSys%nScalars,neigh ))                      &
        &                                 = pdfTmp(13)*cmpl_o+sum12_1+sum12_2  &
        &                                 + pdfGlobal%dtLB( level )*force( 13 )

      outState( ?SAVE?( 12,iElem,varSys%nScalars,neigh ))                      &
        &                                 = pdfTmp(12)*cmpl_o-sum12_1+sum12_2  &
        &                                 + pdfGlobal%dtLB( level )*force( 12 )


      ui11    =  u_y + u_z
      fac_11  = coeff_1 * ui11
      sum11_1 = fac_11 * div3_4h
      sum11_2 = fac_11 * ui11 + usqn_o1

      outState( ?SAVE?( 10,iElem,varSys%nScalars,neigh ))                      &
        &                                 = pdfTmp(10)*cmpl_o+sum11_1+sum11_2  &
        &                                 + pdfGlobal%dtLB( level )*force( 10 )

      outState( ?SAVE?( 7,iElem,varSys%nScalars,neigh ))                       &
        &                                  = pdfTmp(7)*cmpl_o-sum11_1+sum11_2  &
        &                                  + pdfGlobal%dtLB( level )*force( 7 )


      ui13    = -u_y + u_z
      fac_13  = coeff_1 * ui13
      sum13_1 = fac_13 * div3_4h
      sum13_2 = fac_13 * ui13 + usqn_o1

      outState( ?SAVE?( 8,iElem,varSys%nScalars,neigh ))                       &
        &                                  = pdfTmp(8)*cmpl_o+sum13_1+sum13_2  &
        &                                  + pdfGlobal%dtLB( level )*force( 8 )

      outState( ?SAVE?( 9,iElem,varSys%nScalars,neigh ))                       &
        &                                  = pdfTmp(9)*cmpl_o-sum13_1+sum13_2  &
        &                                  + pdfGlobal%dtLB( level )*force( 9 )


      coeff_2 = div1_8 * omega_2 * rho
      usqn_o2 = omega_2 * usqn

      fac_2   = coeff_2 * u_y
      sum2_1  = fac_2 * div3_4h
      sum2_2  = fac_2 * u_y + usqn_o2

      outState( ?SAVE?( 5,iElem,varSys%nScalars,neigh ))                       &
        &                                  = pdfTmp(5) *cmpl_o+sum2_1 +sum2_2  &
        &                                  + pdfGlobal%dtLB( level )*force( 5 )

      outState( ?SAVE?( 2,iElem,varSys%nScalars,neigh ))                       &
        &                                  = pdfTmp(2) *cmpl_o-sum2_1 +sum2_2  &
        &                                  + pdfGlobal%dtLB( level )*force( 2 )


      fac_4   = coeff_2 * u_x
      sum4_1  = fac_4 * div3_4h
      sum4_2  = fac_4 * u_x + usqn_o2

      outState( ?SAVE?( 1,iElem,varSys%nScalars,neigh ))                       &
        &                                  = pdfTmp(1) *cmpl_o-sum4_1 +sum4_2  &
        &                                  + pdfGlobal%dtLB( level )*force( 1 )

      outState( ?SAVE?( 4,iElem,varSys%nScalars,neigh ))                       &
        &                                  = pdfTmp(4) *cmpl_o+sum4_1 +sum4_2  &
        &                                  + pdfGlobal%dtLB( level )*force( 4 )


      fac_9   = coeff_2 * u_z
      sum9_1  = fac_9 * div3_4h
      sum9_2  = fac_9 * u_z + usqn_o2

      outState( ?SAVE?( 6,iElem,varSys%nScalars,neigh ))                       &
        &                                  = pdfTmp(6) *cmpl_o+sum9_1 +sum9_2  &
        &                                  + pdfGlobal%dtLB( level )*force( 6 )

      outState( ?SAVE?( 3,iElem,varSys%nScalars,neigh ))                       &
        &                                  = pdfTmp(3) *cmpl_o-sum9_1 +sum9_2  &
        &                                  + pdfGlobal%dtLB( level )*force( 3 )


    end do nodeloop
!$omp end do nowait

  end subroutine bgk_advRel_forcing
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the 
  !! BGK model with an explicit calculation of all equilibrium 
  !! quantities. Slow and simple. This routine should only be
  !! used for testing purposes
  !!
  subroutine bgk_advRel_explicit( iScheme, fieldProp, inState, outState, neigh,&
    &    nStart, nEnd, level, layout, pdfGlobal, params, scheme, varSys, tree ) 
    !---------------------------------------------------------------------------
    !> current scheme
    integer, intent(in) :: iScheme
    !> global fluid
    type(mus_field_prop_type), intent(in) :: fieldProp(:)
    !> input pdf vector
    real(kind=rk), dimension(:), intent(in)  :: inState
    !> output pdf vector
    real(kind=rk), dimension(:), intent(out) :: outState
    !> connectivity vector 
    integer, dimension(:), intent(in) :: neigh
    !> start positions in pdf vector 
    integer, intent(in) :: nStart
    !> end positions in pdf vector 
    integer, intent(in) :: nEnd
    !> current level
    integer,intent(in) :: level
    !> current layout
    type(mus_scheme_layout_type), intent(in) :: layout
    !> global info type
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    !> global parameters
    type(param_type),intent(in) :: params
    !> array of all schemes
    type( mus_scheme_type ), intent(in) :: scheme(:)
    !> variable system definition
    type(tem_var_system_type), intent(in) :: varSys
    !> treelm mesh information
    type( treelmesh_type ), intent(in) :: tree
    !---------------------------------------------------------------------------
    integer :: iElem, iDir                       ! voxel element counter
    ! temporary distribution variables
    real(kind=rk) pdfTmp(layout%stencil(1)%QQ)
    real(kind=rk) ux(3)   ! local velocity
    real(kind=rk) rho     ! local density
    real(kind=rk) inv_rho ! inverse local density
    real(kind=rk) usq     ! square velocity
    ! derived constants
    ! equilibrium calculation variables
    real(kind=rk) ucx
    real(kind=rk) eqState(layout%stencil(1)%QQ) 
    real(kind=rk) omega
    !---------------------------------------------------------------------------
    
    omega = fieldProp(1)%fluid%omLvl( level )
    nodeloop: do iElem=nStart,nEnd       
      !> Generic fetching step:
      !! Streaming for pull
      !! Local copy for push
      ux = 0._rk
      rho = 0._rk
      do iDir = 1, layout%stencil(1)%QQ
        pdfTmp( iDir ) = inState( ?FETCH?( iDir, iElem, varSys%nScalars, neigh))
      end do
 
      ! local density
      do iDir = 1,layout%stencil(1)%QQ
        rho = rho + inState( ?FETCH?( iDir, iElem, varSys%nScalars, neigh ))
      end do
 
      ! local x-, y- and z-velocity
      do iDir = 1,layout%stencil(1)%QQ
        ux(:) = ux(:) + inState( ?FETCH?(iDir, iElem, varSys%nScalars, neigh)) &
          &   * layout%stencil(1)%cxDir( :, iDir )
      end do

      ! inverse local density
      inv_rho = 1._rk / rho

      ! transfer moments to velocities
      ux(:) = ux(:) * inv_rho

      ! square velocity and derived constants
      usq = ux(1)*ux(1) + ux(2)*ux(2) + ux(3)*ux(3)

      do iDir = 1,layout%stencil(1)%QQ

        !> Pre-calculate velocitiy terms
        ucx = real(layout%stencil(1)%cxDir( 1, iDir ), kind=rk)*ux(1)          &
          & + real(layout%stencil(1)%cxDir( 2, iDir ), kind=rk)*ux(2)          &
          & + real(layout%stencil(1)%cxDir( 3, iDir ), kind=rk)*ux(3) 

        !> Calculate equilibrium distribution functions fEq
        eqState(iDir) = layout%weight(iDir)*rho*( 1.0_rk                       &
          &           +  ucx*cs2inv                                            & 
          &           +  ucx*ucx*cs2inv*cs2inv*0.5d0                           & 
          &           -  usq*0.5d0*cs2inv )

        !> Relaxation     
        outState( ?SAVE?( iDir,iElem,varSys%nScalars,neigh )) =                &
          &            pdfTmp( iDir ) -omega*( pdfTmp(iDir ) - eqState( iDir )) 
      end do  !< iDir

    end do nodeloop

  end subroutine bgk_advRel_explicit
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the non-Newtonian flow
  !! BGK model with an explicit calculation of all equilibrium 
  !! quantities. Slow and simple. This routine should only be
  !! used for testing purposes
  !! Jiaxing Qi
  !!
  subroutine bgk_advRel_nonNewtonian_incomp( iScheme, fieldProp, inState, outState,&
    & neigh, nStart, nEnd, level, layout, pdfGlobal, params, scheme, varSys,   &
    & tree ) 
    !---------------------------------------------------------------------------
    !> current scheme
    integer, intent(in) :: iScheme
    !> global fluid
    type(mus_field_prop_type), intent(in) :: fieldProp(:)
    !> input pdf vector
    real(kind=rk), dimension(:), intent(in)  :: inState
    !> output pdf vector
    real(kind=rk), dimension(:), intent(out) :: outState
    !> connectivity vector 
    integer, dimension(:), intent(in) :: neigh
    !> start positions in pdf vector 
    integer, intent(in) :: nStart
    !> end positions in pdf vector 
    integer, intent(in) :: nEnd
    !> current level
    integer,intent(in) :: level
    !> current layout
    type(mus_scheme_layout_type), intent(in) :: layout
    !> global info type
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    !> global parameters
    type(param_type),intent(in) :: params
    !> array of all schemes
    type( mus_scheme_type ), intent(in) :: scheme(:)
    !> variable system definition
    type(tem_var_system_type), intent(in) :: varSys
    !> treelm mesh information
    type( treelmesh_type ), intent(in) :: tree
    !---------------------------------------------------------------------------
    integer :: iElem, iDir
    ! temporary distribution variables
    real(kind=rk) pdfTmp(layout%stencil(1)%QQ)
    real(kind=rk) ux(3)   ! local velocity
    real(kind=rk) rho     ! local density
    real(kind=rk) rho0     ! referece density
    real(kind=rk) inv_rho0 ! inverse local density
    real(kind=rk) usq      ! square velocity
    ! derived constants
    ! equilibrium calculation variables
    real(kind=rk) ucx
    real(kind=rk) eqState(layout%stencil(1)%QQ)
    real(kind=rk) omega
    !---------------------------------------------------------------------------
    ! parameters for non-Newtonian
    real(kind=rk) strainRate(6)
    real(kind=rk) shearRate ! second invariant of strainRate
    real(kind=rk) visc0, viscInf    ! relaxation time, i.e. inverse of omega
    real(kind=rk) viscPhy, viscLB
    real(kind=rk) lambda ! critical strain rate
    real(kind=rk) a, n, n_1_a ! constant params
    real(kind=rk) nonEq(layout%stencil(1)%QQ)
    real(kind=rk) :: stenCxDir(6, layout%stencil(1)%QQ) ! stencil cxDir
    real(kind=rk) :: convSR(tree%global%minLevel:tree%global%maxLevel)
    real(kind=rk) :: convVisc(tree%global%minLevel:tree%global%maxLevel)
    real(kind=rk) :: densityPhy = 1000.0_rk
    !---------------------------------------------------------------------------

    ! parameters for non_newtonian model
    ! todo: these parameters should be read from lua file
    visc0 = 0.16_rk
    viscInf = 0.0035_rk
    a = 0.64_rk
    n = 0.2128_rk
    lambda = 8.2_rk
    n_1_a = (n - 1._rk) / a

    ! omega = fieldProp(1)%fluid%omLvl( level )
    rho0 = fieldProp(1)%fluid%rho0
    ! inverse reference density
    inv_rho0 = 1._rk / rho0


    ! conversion factor for Shear Rate (or Strain Rate)
    convSR(:) = getConversionFac( conf = varSys%solSpec_conf, &
      &                            facName = 'strainRate', &
      &                            nLevels = tree%global%maxLevel-tree%global%minLevel+1)

    ! conversion factor for kinematic viscosity
    ! todo: add a conversion factor for dynamics viscosity
    convVisc(:) = getConversionFac( conf = varSys%solSpec_conf, &
      &                            facName = 'viscosity', &
      &                            nLevels = tree%global%maxLevel-tree%global%minLevel+1)

    ! convert stencil cxDir into real type
    stenCxDir(1, :) =   real(layout%stencil(1)%cxDir(1, :), kind=rk)                     &
      &               * real(layout%stencil(1)%cxDir(1, :), kind=rk)
    stenCxDir(2, :) =   real(layout%stencil(1)%cxDir(2, :), kind=rk)                     &
      &               * real(layout%stencil(1)%cxDir(2, :), kind=rk)
    stenCxDir(3, :) =   real(layout%stencil(1)%cxDir(3, :), kind=rk)                     &
      &               * real(layout%stencil(1)%cxDir(3, :), kind=rk)
    stenCxDir(4, :) =   real(layout%stencil(1)%cxDir(1, :), kind=rk)                     &
      &               * real(layout%stencil(1)%cxDir(2, :), kind=rk)
    stenCxDir(5, :) =   real(layout%stencil(1)%cxDir(2, :), kind=rk)                     &
      &               * real(layout%stencil(1)%cxDir(3, :), kind=rk)
    stenCxDir(6, :) =   real(layout%stencil(1)%cxDir(1, :), kind=rk)                     &
      &               * real(layout%stencil(1)%cxDir(3, :), kind=rk)

    iElemLoop: do iElem = nStart, nEnd
      ! get omega for this element
      omega = scheme(iScheme)%pdf(level)%aux(1,iElem)
      ! initialize
      strainRate = 0.0_rk

      !> Generic fetching step:
      !! Streaming for pull
      !! Local copy for push
      ux = 0._rk
      rho = 0._rk
      do iDir = 1, layout%stencil(1)%QQ
        pdfTmp( iDir ) = inState( ?FETCH?( iDir, iElem, varSys%nScalars, neigh))
      end do
 
      ! local density
      do iDir = 1,layout%stencil(1)%QQ
        rho = rho + inState( ?FETCH?( iDir, iElem, varSys%nScalars, neigh ))
      end do
 
      ! local x-, y- and z-velocity
      do iDir = 1,layout%stencil(1)%QQ
        ux(:) = ux(:) + inState( ?FETCH?(iDir, iElem, varSys%nScalars, neigh)) &
          &   * layout%stencil(1)%cxDir( :, iDir )
      end do

      ! transfer moments to velocities
      ux(:) = ux(:) * inv_rho0

      ! square velocity and derived constants
      usq = ux(1)*ux(1) + ux(2)*ux(2) + ux(3)*ux(3)

      do iDir = 1,layout%stencil(1)%QQ

        ! Pre-calculate velocitiy terms
        ucx = real(layout%stencil(1)%cxDir( 1, iDir ), kind=rk)*ux(1)          &
          & + real(layout%stencil(1)%cxDir( 2, iDir ), kind=rk)*ux(2)          &
          & + real(layout%stencil(1)%cxDir( 3, iDir ), kind=rk)*ux(3) 

        ! Calculate equilibrium distribution functions fEq
        eqState(iDir) = layout%weight(iDir)*( rho + rho0*(                     &
          & +  ucx*cs2inv                                                      &
          & +  ucx*ucx*cs2inv*cs2inv*0.5_rk                                    &
          & -  usq*0.5_rk*cs2inv ))

        ! get nonEq
        nonEq(iDir) = pdfTmp(iDir) - eqState(iDir)

        ! Relaxation
        outState( ?SAVE?( iDir,iElem,varSys%nScalars,neigh )) =                &
          &            pdfTmp(iDir) - omega * nonEq(iDir)

        ! calculate strain rate
        strainRate(1) = strainRate(1) + nonEq(iDir) * stenCxDir(1, iDir) ! strainRate_xx
        strainRate(2) = strainRate(2) + nonEq(iDir) * stenCxDir(2, iDir) ! strainRate_yy
        strainRate(3) = strainRate(3) + nonEq(iDir) * stenCxDir(3, iDir) ! strainRate_zz
        strainRate(4) = strainRate(4) + nonEq(iDir) * stenCxDir(4, iDir) ! strainRate_xy
        strainRate(5) = strainRate(5) + nonEq(iDir) * stenCxDir(5, iDir) ! strainRate_yz
        strainRate(6) = strainRate(6) + nonEq(iDir) * stenCxDir(6, iDir) ! strainRate_xz

      end do  ! iDir

      ! Calculate strain rate
      strainRate = -1.5_rk * omega / rho0 * strainRate

      ! Calculate physical shear rate
      shearRate = strainRate(1)*strainRate(1) + &
        &         strainRate(2)*strainRate(2) + &
        &         strainRate(3)*strainRate(3) + &
        &         strainRate(4)*strainRate(4) + &
        &         strainRate(5)*strainRate(5) + &
        &         strainRate(6)*strainRate(6)
      shearRate = 2._rk * sqrt(shearRate) * convSR(level)

      ! get physical dynamic viscosity
      viscPhy = viscInf + (visc0-viscInf) * ( 1+(lambda*shearRate)**a )** n_1_a

      ! get LB kinematic viscosity
      viscLB = viscPhy / densityPhy / convVisc(level)
      ! get LB omega
      omega = 1._rk / ( 0.5_rk + 3 * viscLB )

      ! save omega
      ! scheme(iScheme)%pdf(level)%aux(1,iElem) = omega
    end do iElemLoop

  end subroutine bgk_advRel_nonNewtonian_incomp
!******************************************************************************!

!******************************************************************************!
  !> Advection relaxation routine for the
  !! BGK model with an explicit calculation of all equilibrium 
  !! quantities. Slow and simple. This routine should only be
  !! used for testing purposes
  !!
  subroutine bgk_advRel_explicit_incomp( iScheme, fieldProp, inState, outState,&
    & neigh, nStart, nEnd, level, layout, pdfGlobal, params, scheme, varSys,   &
    & tree ) 
    !---------------------------------------------------------------------------
    !> current scheme
    integer, intent(in) :: iScheme
    !> global fluid
    type(mus_field_prop_type), intent(in) :: fieldProp(:)
    !> input pdf vector
    real(kind=rk), dimension(:), intent(in)  :: inState
    !> output pdf vector
    real(kind=rk), dimension(:), intent(out) :: outState
    !> connectivity vector 
    integer, dimension(:), intent(in) :: neigh
    !> start positions in pdf vector 
    integer, intent(in) :: nStart
    !> end positions in pdf vector 
    integer, intent(in) :: nEnd
    !> current level
    integer,intent(in) :: level
    !> current layout
    type(mus_scheme_layout_type), intent(in) :: layout
    !> global info type
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    !> global parameters
    type(param_type),intent(in) :: params
    !> array of all schemes
    type( mus_scheme_type ), intent(in) :: scheme(:)
    !> variable system definition
    type(tem_var_system_type), intent(in) :: varSys
    !> treelm mesh information
    type( treelmesh_type ), intent(in) :: tree
    !---------------------------------------------------------------------------
    integer :: iElem, iDir                       ! voxel element counter
    ! temporary distribution variables
    real(kind=rk) pdfTmp(layout%stencil(1)%QQ)
    real(kind=rk) ux(3)   ! local velocity
    real(kind=rk) rho     ! local density
    real(kind=rk) rho0     ! referece density
    real(kind=rk) inv_rho0 ! inverse local density
    real(kind=rk) usq      ! square velocity
    ! derived constants
    ! equilibrium calculation variables
    real(kind=rk) ucx
    real(kind=rk) eqState(layout%stencil(1)%QQ) 
    real(kind=rk) omega
    !---------------------------------------------------------------------------
    
    omega = fieldProp(1)%fluid%omLvl( level )
    rho0 = fieldProp(1)%fluid%rho0
    ! inverse local density
    inv_rho0 = 1._rk / rho0

    nodeloop: do iElem=nStart,nEnd       
      !> Generic fetching step:
      !! Streaming for pull
      !! Local copy for push
      ux = 0._rk
      rho = 0._rk
      do iDir = 1, layout%stencil(1)%QQ
        pdfTmp( iDir ) = inState( ?FETCH?( iDir, iElem, varSys%nScalars, neigh))
      end do
 
      ! local density
      do iDir = 1,layout%stencil(1)%QQ
        rho = rho + inState( ?FETCH?( iDir, iElem, varSys%nScalars, neigh ))
      end do
 
      ! local x-, y- and z-velocity
      do iDir = 1,layout%stencil(1)%QQ
        ux(:) = ux(:) + inState( ?FETCH?(iDir, iElem, varSys%nScalars, neigh)) &
          &   * layout%stencil(1)%cxDir( :, iDir )
      end do

      ! transfer moments to velocities
      ux(:) = ux(:) * inv_rho0

      ! square velocity and derived constants
      usq = ux(1)*ux(1) + ux(2)*ux(2) + ux(3)*ux(3)

      do iDir = 1,layout%stencil(1)%QQ

        !> Pre-calculate velocitiy terms
        ucx = real(layout%stencil(1)%cxDir( 1, iDir ), kind=rk)*ux(1)          &
          & + real(layout%stencil(1)%cxDir( 2, iDir ), kind=rk)*ux(2)          &
          & + real(layout%stencil(1)%cxDir( 3, iDir ), kind=rk)*ux(3) 

        !> Calculate equilibrium distribution functions fEq
        eqState(iDir) = layout%weight(iDir)*( rho + rho0*(                     &
          & +  ucx*cs2inv                                                      & 
          & +  ucx*ucx*cs2inv*cs2inv*0.5d0                                     & 
          & -  usq*0.5d0*cs2inv ))

        !> Relaxation     
        outState( ?SAVE?( iDir,iElem,varSys%nScalars,neigh )) =                &
          &            pdfTmp( iDir ) -omega*( pdfTmp(iDir ) - eqState( iDir )) 

      end do  ! iDir

    end do nodeloop

  end subroutine bgk_advRel_explicit_incomp
!******************************************************************************!


!******************************************************************************!
  !> Test wrapper routine for BGK advection relaxation 
  !! uses pre-calculated values to validate the routine
  !! Stops code after test.
  !! These values can be checked in the validation folder 
  !! within the maple sheet MRT_ben.mw 
  !!
  subroutine test_bgk_advRel( inState, outState, neigh, nStart, nEnd, omega,   &
    &                   test, layout, pdfGlobal, params, scheme, varSys, tree ) 
    !---------------------------------------------------------------------------
    implicit none

    !> input pdf vector
    real(kind=rk), dimension(:), intent(inout) :: inState
    !> output pdf vector
    real(kind=rk), dimension(:), intent(out)   :: outState
    !> connectivity vector 
    integer, dimension(:), intent(in)          :: neigh
    !> start positions in pdf vector 
    integer, intent(in) :: nStart
    !> end positions in pdf vector 
    integer, intent(in) :: nEnd
    !> Relaxation parameter 
    real(kind=rk), intent(in)  :: omega
    !> 
    logical, intent(in)  :: test
    !> current layout
    type(mus_scheme_layout_type), intent(in) :: layout
    !> global info type
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    !> global parameters
    type(param_type),intent(in) :: params
    !> array of all schemes
    type( mus_scheme_type ), intent(in) :: scheme(:)
    !> variable system definition
    type(tem_var_system_type), intent(in) :: varSys
    !> global treelmesh information
    type( treelmesh_type ), intent(in) :: tree
    !---------------------------------------------------------------------------
    ! a reference solution
    real(kind=rk), dimension(layout%stencil(1)%QQ)  :: refSolution
    ! momentum of incoming state
    real(kind=rk), dimension(3)   :: uxIn
    ! density of incoming state
    real(kind=rk)                 :: rhoIn
    ! momentum of outgoing state
    real(kind=rk), dimension(3)   :: uxOut
    ! density of outgoing state
    real(kind=rk)                 :: rhoOut
    integer :: iDir,iElem
    !---------------------------------------------------------------------------

    call tem_write( "Test routine" )
    call tem_write( "Validation of BGK relaxation routine." )
    call tem_write( "Using standard set of pdf values on each element" )
    !> set incoming reference values 
    do iElem = nStart, nEnd
      inState( ?IDX?( 19,iElem,varSys%nScalars )) = 0.332333333503913_rk
      inState( ?IDX?(  4,iElem,varSys%nScalars )) = 0.60909350699940e-1_rk
      inState( ?IDX?(  1,iElem,varSys%nScalars )) = 0.49868427134697e-1_rk
      inState( ?IDX?(  5,iElem,varSys%nScalars )) = 0.101307711335621_rk
      inState( ?IDX?(  2,iElem,varSys%nScalars )) = 0.55555555555500e-1_rk
      inState( ?IDX?(  6,iElem,varSys%nScalars )) = 0.60909350699940e-1_rk
      inState( ?IDX?(  3,iElem,varSys%nScalars )) = 0.50368426964117e-1_rk
      inState( ?IDX?( 18,iElem,varSys%nScalars )) = 0.18918438141908e-1_rk
      inState( ?IDX?( 15,iElem,varSys%nScalars )) = 0.27777777777770e-1_rk
      inState( ?IDX?( 17,iElem,varSys%nScalars )) = 0.27777777777770e-1_rk
      inState( ?IDX?( 16,iElem,varSys%nScalars )) = 0.10882810381340e-1_rk
      inState( ?IDX?( 14,iElem,varSys%nScalars )) = 0.33464906155991e-1_rk
      inState( ?IDX?( 11,iElem,varSys%nScalars )) = 0.22923982420168e-1_rk
      inState( ?IDX?( 12,iElem,varSys%nScalars )) = 0.27694444458659e-1_rk
      inState( ?IDX?( 13,iElem,varSys%nScalars )) = 0.27694444458659e-1_rk
      inState( ?IDX?( 10,iElem,varSys%nScalars )) = 0.18818739533272e-1_rk
      inState( ?IDX?(  7,iElem,varSys%nScalars )) = 0.27777777777770e-1_rk
      inState( ?IDX?(  9,iElem,varSys%nScalars )) = 0.10900357542234e-1_rk
      inState( ?IDX?(  8,iElem,varSys%nScalars )) = 0.27777777777770e-1_rk
    end do

    !> Set element to test
    iElem = 115
    call tem_write( "Checking element number",iElem )

    !> Calculate local density
    rhoIn = 0.0
    do iDir = 1,layout%stencil(1)%QQ
        rhoIn = rhoIn + inState( ?IDX?( iDir,iElem,varSys%nScalars ))
    end do

    !> Calculate local momenta   
    uxIn(1) = inState( ?IDX?( 18,iElem,varSys%nScalars ))                      &
      &     + inState( ?IDX?( 17,iElem,varSys%nScalars ))                      & 
      &     + inState( ?IDX?(  4,iElem,varSys%nScalars ))                      &
      &     + inState( ?IDX?( 14,iElem,varSys%nScalars ))                      &
      &     + inState( ?IDX?( 12,iElem,varSys%nScalars ))                      &
      &     - inState( ?IDX?( 16,iElem,varSys%nScalars ))                      &
      &     - inState( ?IDX?(  1,iElem,varSys%nScalars ))                      &
      &     - inState( ?IDX?( 15,iElem,varSys%nScalars ))                      &
      &     - inState( ?IDX?( 13,iElem,varSys%nScalars ))                      &
      &     - inState( ?IDX?( 11,iElem,varSys%nScalars ))
                                                                    
    uxIn(2) = inState( ?IDX?( 18,iElem,varSys%nScalars ))                      &
      &     + inState( ?IDX?(  5,iElem,varSys%nScalars ))                      &
      &     + inState( ?IDX?( 16,iElem,varSys%nScalars ))                      &
      &     + inState( ?IDX?( 10,iElem,varSys%nScalars ))                      &
      &     + inState( ?IDX?(  9,iElem,varSys%nScalars ))                      &
      &     - inState( ?IDX?( 15,iElem,varSys%nScalars ))                      &
      &     - inState( ?IDX?(  2,iElem,varSys%nScalars ))                      &
      &     - inState( ?IDX?( 17,iElem,varSys%nScalars ))                      &
      &     - inState( ?IDX?(  8,iElem,varSys%nScalars ))                      &
      &     - inState( ?IDX?(  7,iElem,varSys%nScalars ))
                                                                    
    uxIn(3) = inState( ?IDX?(  6,iElem,varSys%nScalars ))                      &
      &     + inState( ?IDX?( 14,iElem,varSys%nScalars ))                      &
      &     + inState( ?IDX?( 10,iElem,varSys%nScalars ))                      &
      &     + inState( ?IDX?( 13,iElem,varSys%nScalars ))                      &
      &     + inState( ?IDX?(  8,iElem,varSys%nScalars ))                      & 
      &     - inState( ?IDX?(  3,iElem,varSys%nScalars ))                      &
      &     - inState( ?IDX?( 12,iElem,varSys%nScalars ))                      &
      &     - inState( ?IDX?(  9,iElem,varSys%nScalars ))                      &
      &     - inState( ?IDX?( 11,iElem,varSys%nScalars ))                      &
      &     - inState( ?IDX?(  7,iElem,varSys%nScalars ))
                                                       
    !> Transfer momenta to velocites                   
    uxIn = uxIn/rhoIn                                  
                                                       
    call tem_write( "density ", rhoIn )
    call tem_write( " ux ", uxIn )

    !> run collision routine
!    call bgk_advRel(inState,outState,neigh,nStart,nEnd,1.9_rk)
!<DEBUG
    !> test outgoing values against fixed, validated values
    rhoOut = 0.0
    do iDir = 1,layout%stencil(1)%QQ
        rhoOut = rhoOut + outState( ?IDX?( iDir,1,varSys%nScalars ))
    end do

    !> Calculate local momenta   
    uxOut(1) = outState( ?IDX?( 18,iElem,varSys%nScalars ))                    &
      &      + outState( ?IDX?( 17,iElem,varSys%nScalars ))                    & 
      &      + outState( ?IDX?(  4,iElem,varSys%nScalars ))                    &   
      &      + outState( ?IDX?( 14,iElem,varSys%nScalars ))                    &
      &      + outState( ?IDX?( 12,iElem,varSys%nScalars ))                    &
      &      - outState( ?IDX?( 16,iElem,varSys%nScalars ))                    &
      &      - outState( ?IDX?(  1,iElem,varSys%nScalars ))                    &
      &      - outState( ?IDX?( 15,iElem,varSys%nScalars ))                    &   
      &      - outState( ?IDX?( 13,iElem,varSys%nScalars ))                    &
      &      - outState( ?IDX?( 11,iElem,varSys%nScalars ))
                                                                       
    uxOut(2) = outState( ?IDX?( 18,iElem,varSys%nScalars ))                    &
      &      + outState( ?IDX?(  5,iElem,varSys%nScalars ))                    &
      &      + outState( ?IDX?( 16,iElem,varSys%nScalars ))                    &   
      &      + outState( ?IDX?( 10,iElem,varSys%nScalars ))                    &
      &      + outState( ?IDX?(  9,iElem,varSys%nScalars ))                    &
      &      - outState( ?IDX?( 15,iElem,varSys%nScalars ))                    &
      &      - outState( ?IDX?(  2,iElem,varSys%nScalars ))                    &
      &      - outState( ?IDX?( 17,iElem,varSys%nScalars ))                    &   
      &      - outState( ?IDX?(  8,iElem,varSys%nScalars ))                    &
      &      - outState( ?IDX?(  7,iElem,varSys%nScalars ))
                                                                       
    uxOut(3) = outState( ?IDX?(  6,iElem,varSys%nScalars ))                    &
      &      + outState( ?IDX?( 14,iElem,varSys%nScalars ))                    &
      &      + outState( ?IDX?( 10,iElem,varSys%nScalars ))                    &   
      &      + outState( ?IDX?( 13,iElem,varSys%nScalars ))                    &
      &      + outState( ?IDX?(  8,iElem,varSys%nScalars ))                    & 
      &      - outState( ?IDX?(  3,iElem,varSys%nScalars ))                    &
      &      - outState( ?IDX?( 12,iElem,varSys%nScalars ))                    &
      &      - outState( ?IDX?(  9,iElem,varSys%nScalars ))                    &   
      &      - outState( ?IDX?( 11,iElem,varSys%nScalars ))                    &
      &      - outState( ?IDX?(  7,iElem,varSys%nScalars ))
   
    !> Transfer momenta to velocites
    uxOut = uxOut/rhoOut

    call tem_write( "density ", rhoOut )
    call tem_write( " ux ", uxOut )

    !> Reference solution calculated in Maple sheet
    !! according to above stated input pdfs and omega=1.9
    refSolution(19) =0.3285435784864783050_rk
    refSolution( 4) =0.0595870400570539943_rk
    refSolution( 1) =0.0507661370927727021_rk
    refSolution( 5) =0.0115977255629410997_rk
    refSolution( 2) =0.0564724519780500020_rk
    refSolution( 6) =0.0593742831110539993_rk
    refSolution( 3) =0.0504943024882947011_rk
    refSolution(18) =0.0391771715702827969_rk
    refSolution(15) =0.0236737918800069974_rk
    refSolution(17) =0.0332179855570070015_rk
    refSolution(16) =0.0371956960207940057_rk
    refSolution(14) =0.0322876166306080986_rk
    refSolution(11) =0.0232121748948488013_rk
    refSolution(12) =0.0274764531672068990_rk
    refSolution(13) =0.0272809920732068993_rk
    refSolution(10) =0.0391622445750551967_rk
    refSolution( 7) =0.0237645972690069957_rk
    refSolution( 9) =0.0372672634479893974_rk
    refSolution( 8) =0.0331098843160069975_rk

    ! Exit the program
    stop

  end subroutine test_bgk_advRel 
!******************************************************************************!


end module mus_bgk_module
!******************************************************************************!
