! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Routines and parameter definitions for the standard D3Q19 model
!! \author Manuel Hasert, Kannan Masilamani
?? include 'header/lbm_macros.inc'
module mus_d3q27_module

  ! include treelm modules
  use env_module,            only: rk
  use tem_var_system_module, only: tem_var_system_type
  use tem_param_module,      only: div1_3, div1_36, div1_8, div3_4h
  use treelmesh_module,      only: treelmesh_type
  use tem_tools_module,      only: tem_write

  ! include musubi modules
  use mus_field_prop_module,    only: mus_field_prop_type
  use mus_scheme_layout_module, only: mus_scheme_layout_type
  use mus_pdf_module,           only: pdf_globInfo_type
  use mus_scheme_type_module,   only: mus_scheme_type
  use mus_global_module,        only: param_type

  implicit none
  private
  public :: bgk_advRel_d3q27
  !=============================================================================
  ! D3Q19 flow model
  !=============================================================================
  !> Definition of the discrete velocity set

  integer, parameter :: QQ  = 27   !> number of pdf directions
  integer, parameter :: QQN = 26   !> number of neighbors

  integer,parameter :: q__W     = 1     !< west             x-
  integer,parameter :: q__S     = 2     !< south            y-
  integer,parameter :: q__B     = 3     !< bottom           z-
  integer,parameter :: q__E     = 4     !< east             x+
  integer,parameter :: q__N     = 5     !< north            y+
  integer,parameter :: q__T     = 6     !< top              z+
  integer,parameter :: q_BS     = 7     !<                  z-,y-
  integer,parameter :: q_TS     = 8     !<                  z+,y-
  integer,parameter :: q_BN     = 9     !<                  z-,y+
  integer,parameter :: q_TN     = 10    !<                  z+,y+
  integer,parameter :: q_BW     = 11    !<                  x-,z-
  integer,parameter :: q_BE     = 12    !<                  x+,z-
  integer,parameter :: q_TW     = 13    !<                  x-,z+
  integer,parameter :: q_TE     = 14    !<                  x+,z+
  integer,parameter :: q_SW     = 15    !<                  y-,x-
  integer,parameter :: q_NW     = 16    !<                  y+,x-
  integer,parameter :: q_SE     = 17    !<                  y-,x+
  integer,parameter :: q_NE     = 18    !<                  y+,x+
  integer,parameter :: qBSW     = 19    !<                  x-,z-
  integer,parameter :: qTSW     = 20    !<                  x+,z-
  integer,parameter :: qBNW     = 21    !<                  x-,z+
  integer,parameter :: qTNW     = 22    !<                  x+,z+
  integer,parameter :: qBSE     = 23    !<                  y-,x-
  integer,parameter :: qTSE     = 24    !<                  y+,x-
  integer,parameter :: qBNE     = 25    !<                  y-,x+
  integer,parameter :: qTNE     = 26    !<                  y+,x+
  integer,parameter :: q__0     = 27    !< rest density is last

contains

!******************************************************************************!
  !> Advection relaxation routine for the D3Q27 model with BGK.
  !! 
  !! Streaming is defined in the header/lbm_macros.inc in STREAM_PUSH = .false.
  !! This routine ONLY contains pure computation
  !! and should be implemented in a very efficient way 
  !! \f$ f_\alpha(x_i+e_{\alpha,i},t+1) = 
  !! f_\alpha(x_i,t) - \omega(f_\alpha(x_i,t)-f^{eq}_{\alpha}(x_i,t)) \f$  
  !!
  subroutine bgk_advRel_d3q27( iScheme, fieldProp, inState, outState,          &
    &                          neigh, nStart, nEnd, level, layout, pdfGlobal,  &
    &                          params, scheme, varSys, tree )
    !---------------------------------------------------------------------------
    !> current scheme
    integer, intent(in) :: iScheme
    !> global fluid
    type(mus_field_prop_type), intent(in) :: fieldProp(:)
    !> input pdf vector
    real(kind=rk), dimension(:), intent(in)  :: inState
    !> output pdf vector
    real(kind=rk), dimension(:), intent(out) :: outState
    !> connectivity vector 
    integer, dimension(:), intent(in) :: neigh
    !> start positions in pdf vector 
    integer, intent(in) :: nStart
    !> end positions in pdf vector 
    integer, intent(in) :: nEnd
    !> current level
    integer,intent(in) :: level
    !> current layout
    type(mus_scheme_layout_type), intent(in) :: layout
    !> global info type
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    !> global parameters
    type(param_type),intent(in) :: params
    !> array of all schemes
    type( mus_scheme_type ), intent(in) :: scheme(:)
    !> variable system definition
    type(tem_var_system_type), intent(in) :: varSys
    !> treelm mesh information
    type( treelmesh_type ), intent(in) :: tree
    !---------------------------------------------------------------------------
    integer :: iElem ! element counter
    real(kind=rk) pdfTmp(layout%stencil(1)%QQ) ! temporary local pdf values
    real(kind=rk) rho     ! local density
    real(kind=rk) inv_rho ! inverse local density
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) u_z     ! local z-velocity
    real(kind=rk) usq     ! square velocity
    real(kind=rk) temp0   ! reference temperature
    ! derived constants
    real(kind=rk) usqn, usqn_o1, usqn_o2 
    real(kind=rk) omega_2, cmpl_o, omega
    real(kind=rk) coeff_1, coeff_2
    real(kind=rk) ui1, ui3, ui10, ui11, ui12, ui13
    real(kind=rk) :: fac_1, fac_2, fac_3, fac_4, fac_9, fac_10, fac_11, fac_12,&
      &              fac_13
    real(kind=rk) :: sum1_1, sum1_2, sum2_1, sum2_2, sum3_1, sum3_2, sum4_1,   &
      &              sum4_2, sum9_1, sum9_2, sum10_1, sum10_2, sum11_1,        &
      &              sum11_2, sum12_1, sum12_2, sum13_1, sum13_2
    !---------------------------------------------------------------------------

    ! read the relaxation parameter omega for the current level
    omega   = fieldProp(1)%fluid%omLvl( level )
    temp0   = fieldProp(1)%fluid%temp0
    ! pre-calculate partial collision constants
    omega_2 = 2._rk * omega
    cmpl_o  = 1._rk - omega
    call tem_write('Warning: This routine is not tested at all. Please do not use. Manuel')

!$omp do schedule(static)

    nodeloop: do iElem=nStart,nEnd       
      !> First load all local values into temp array
      !! Generic! PUSH+PULL is possible
      ! Flexible array access
      ! pdfPos = scheme%pdfPos 
      ! pdfTmp( Q__W ) = inState(?FETCH?( 
      !     sys%Variable%val(pdfPos)%varPos(Q__W)), iElem, layout%stencil(1)%QQ, neigh ))
      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, varSys%nScalars, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, varSys%nScalars, neigh ))
      pdfTmp( Q__B ) = inState(?FETCH?( Q__B, iElem, varSys%nScalars, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, varSys%nScalars, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, varSys%nScalars, neigh ))
      pdfTmp( Q__T ) = inState(?FETCH?( Q__T, iElem, varSys%nScalars, neigh ))
      pdfTmp( Q_BS ) = inState(?FETCH?( Q_BS, iElem, varSys%nScalars, neigh ))
      pdfTmp( Q_TS ) = inState(?FETCH?( Q_TS, iElem, varSys%nScalars, neigh ))
      pdfTmp( Q_BN ) = inState(?FETCH?( Q_BN, iElem, varSys%nScalars, neigh ))
      pdfTmp( Q_TN ) = inState(?FETCH?( Q_TN, iElem, varSys%nScalars, neigh ))
      pdfTmp( Q_BW ) = inState(?FETCH?( Q_BW, iElem, varSys%nScalars, neigh ))
      pdfTmp( Q_BE ) = inState(?FETCH?( Q_BE, iElem, varSys%nScalars, neigh ))
      pdfTmp( Q_TW ) = inState(?FETCH?( Q_TW, iElem, varSys%nScalars, neigh ))
      pdfTmp( Q_TE ) = inState(?FETCH?( Q_TE, iElem, varSys%nScalars, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, varSys%nScalars, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, varSys%nScalars, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, varSys%nScalars, neigh ))
      pdfTmp( QBSW ) = inState(?FETCH?( QBSW, iElem, varSys%nScalars, neigh ))
      pdfTmp( QTSW ) = inState(?FETCH?( QTSW, iElem, varSys%nScalars, neigh ))
      pdfTmp( QBNW ) = inState(?FETCH?( QBNW, iElem, varSys%nScalars, neigh ))
      pdfTmp( QTNW ) = inState(?FETCH?( QTNW, iElem, varSys%nScalars, neigh ))
      pdfTmp( QBSE ) = inState(?FETCH?( QBSE, iElem, varSys%nScalars, neigh ))
      pdfTmp( QTSE ) = inState(?FETCH?( QTSE, iElem, varSys%nScalars, neigh ))
      pdfTmp( QBNE ) = inState(?FETCH?( QBNE, iElem, varSys%nScalars, neigh ))
      pdfTmp( QTNE ) = inState(?FETCH?( QTNE, iElem, varSys%nScalars, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, varSys%nScalars, neigh ))
 
 
      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__T)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_TN) + pdfTmp(Q_TW)                         &
        & + pdfTmp(Q_TS) + pdfTmp(Q__B) + pdfTmp(Q_BE)                         &
        & + pdfTmp(Q_BN) + pdfTmp(Q_BW) + pdfTmp(Q_BS)                         &
        & + pdfTmp(QBSW) + pdfTmp(QTSW) + pdfTmp(QBNW)  + pdfTmp(QTNW)         &
        & + pdfTmp(QBSE) + pdfTmp(QTSE) + pdfTmp(QBNE)  + pdfTmp(QTNE)         &
        & + pdfTmp(Q__0)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(Q_NE) + pdfTmp(Q_SE) + pdfTmp(Q__E)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_BE)                                        &
        & + pdfTmp(QBSE) + pdfTmp(QTSE) + pdfTmp(QBNE)  + pdfTmp(QTNE)         &
        & - pdfTmp(QBSW) - pdfTmp(QTSW) - pdfTmp(QBNW)  - pdfTmp(QTNW)         &
        & - pdfTmp(Q_NW) - pdfTmp(Q__W) - pdfTmp(Q_SW)                         &
        & - pdfTmp(Q_TW) - pdfTmp(Q_BW)
                         
      u_y = pdfTmp(Q_NE) + pdfTmp(Q__N)  + pdfTmp(Q_NW)                        &
        & + pdfTmp(Q_TN) + pdfTmp(Q_BN)                                        &
        & - pdfTmp(QBSW) - pdfTmp(QTSW) + pdfTmp(QBNW)  + pdfTmp(QTNW)         &
        & - pdfTmp(QBSE) - pdfTmp(QTSE) + pdfTmp(QBNE)  + pdfTmp(QTNE)         &
        & - pdfTmp(Q_SW) - pdfTmp(Q__S ) - pdfTmp(Q_SE)                        &
        & - pdfTmp(Q_TS) - pdfTmp(Q_BS)
                         
      u_z = pdfTmp(Q__T) + pdfTmp(Q_TE) + pdfTmp(Q_TN)                         &
        & + pdfTmp(Q_TW) + pdfTmp(Q_TS)                                        &
        & - pdfTmp(QBSW) + pdfTmp(QTSW) - pdfTmp(QBNW)  +  pdfTmp(QTNW)        &
        & - pdfTmp(QBSE) + pdfTmp(QTSE) - pdfTmp(QBNE)  +  pdfTmp(QTNE)        &
        & - pdfTmp(Q__B) - pdfTmp(Q_BE) - pdfTmp(Q_BN)                         &
        & - pdfTmp(Q_BW) - pdfTmp(Q_BS)
      ! inverse local density
      inv_rho = 1._rk / rho

      ! transfer moments to velocities
      u_x = u_x * inv_rho
      u_y = u_y * inv_rho
      u_z = u_z * inv_rho
      !MH: Checked up to here
      !temp = 

      ! square velocity and derived constants
      usq  = u_x*u_x + u_y*u_y + u_z*u_z
      usqn = div1_36 * (1._rk - 1.5_rk * usq) * rho

      outState(?SAVE?(Q__0,iElem,varSys%nScalars,neigh)) =                     &
        &                     pdfTmp(Q__0)*cmpl_o+omega*rho*(div1_3-0.5_rk*usq)

      coeff_1 = div1_8 * omega * rho

      usqn_o1 = omega * usqn

      ui1     =  u_x + u_y
      fac_1   = coeff_1 * ui1
      sum1_1  = fac_1 * div3_4h
      sum1_2  = fac_1 * ui1 + usqn_o1

      outState(?SAVE?(Q_NE,iElem,varSys%nScalars,neigh)) =                     &
        &                                   pdfTmp(Q_NE) *cmpl_o+sum1_1 +sum1_2
      outState(?SAVE?(Q_SW,iElem,varSys%nScalars,neigh)) =                     &
        &                                   pdfTmp(Q_SW) *cmpl_o-sum1_1 +sum1_2

      ui3     = -u_x + u_y
      fac_3   = coeff_1 * ui3
      sum3_1  = fac_3 * div3_4h
      sum3_2  = fac_3 * ui3 + usqn_o1

      outState(?SAVE?(Q_NW,iElem,varSys%nScalars,neigh)) =                     &
        &                                   pdfTmp(Q_NW) *cmpl_o+sum3_1 +sum3_2
      outState(?SAVE?(Q_SE,iElem,varSys%nScalars,neigh)) =                     &
        &                                   pdfTmp(Q_SE) *cmpl_o-sum3_1 +sum3_2

      ui10    =  u_x + u_z
      fac_10  = coeff_1 * ui10
      sum10_1 = fac_10 * div3_4h
      sum10_2 = fac_10 * ui10 + usqn_o1

      outState(?SAVE?(Q_TE,iElem,varSys%nScalars,neigh)) =                     &
        &                                   pdfTmp(Q_TE)*cmpl_o+sum10_1+sum10_2
      outState(?SAVE?(Q_BW,iElem,varSys%nScalars,neigh)) =                     &
        &                                   pdfTmp(Q_BW)*cmpl_o-sum10_1+sum10_2

      ui12    = -u_x + u_z
      fac_12  = coeff_1 * ui12
      sum12_1 = fac_12 * div3_4h
      sum12_2 = fac_12 * ui12 + usqn_o1

      outState(?SAVE?(Q_TW,iElem,varSys%nScalars,neigh)) =                     &
        &                                   pdfTmp(Q_TW)*cmpl_o+sum12_1+sum12_2
      outState(?SAVE?(Q_BE,iElem,varSys%nScalars,neigh)) =                     &
        &                                   pdfTmp(Q_BE)*cmpl_o-sum12_1+sum12_2

      ui11    =  u_y + u_z
      fac_11  = coeff_1 * ui11
      sum11_1 = fac_11 * div3_4h
      sum11_2 = fac_11 * ui11 + usqn_o1

      outState(?SAVE?(Q_TN,iElem,varSys%nScalars,neigh)) =                     &
        &                                   pdfTmp(Q_TN)*cmpl_o+sum11_1+sum11_2
      outState(?SAVE?(Q_BS,iElem,varSys%nScalars,neigh)) =                     &
        &                                   pdfTmp(Q_BS)*cmpl_o-sum11_1+sum11_2

      ui13    = -u_y + u_z
      fac_13  = coeff_1 * ui13
      sum13_1 = fac_13 * div3_4h
      sum13_2 = fac_13 * ui13 + usqn_o1

      outState(?SAVE?(Q_TS,iElem,varSys%nScalars,neigh)) =                     &
        &                                   pdfTmp(Q_TS)*cmpl_o+sum13_1+sum13_2
      outState(?SAVE?(Q_BN,iElem,varSys%nScalars,neigh)) =                     &
        &                                   pdfTmp(Q_BN)*cmpl_o-sum13_1+sum13_2

      coeff_2 = div1_8 * omega_2 * rho
      usqn_o2 = omega_2 * usqn

      fac_2   = coeff_2 * u_y
      sum2_1  = fac_2 * div3_4h
      sum2_2  = fac_2 * u_y + usqn_o2

      outState(?SAVE?(Q__N,iElem,varSys%nScalars,neigh )) =                    &
        &                                    pdfTmp(Q__N)*cmpl_o+sum2_1 +sum2_2
      outState(?SAVE?(Q__S,iElem,varSys%nScalars,neigh )) =                    &
        &                                    pdfTmp(Q__S)*cmpl_o-sum2_1 +sum2_2

      fac_4   = coeff_2 * u_x
      sum4_1  = fac_4 * div3_4h
      sum4_2  = fac_4 * u_x + usqn_o2

      outState(?SAVE?(Q__W,iElem,varSys%nScalars,neigh )) =                    &
        &                                    pdfTmp(Q__W)*cmpl_o-sum4_1 +sum4_2
      outState(?SAVE?(Q__E,iElem,varSys%nScalars,neigh )) =                    &
        &                                    pdfTmp(Q__E)*cmpl_o+sum4_1 +sum4_2

      fac_9   = coeff_2 * u_z
      sum9_1  = fac_9 * div3_4h
      sum9_2  = fac_9 * u_z + usqn_o2

      outState(?SAVE?(Q__T,iElem,varSys%nScalars,neigh )) =                    &
        &                                    pdfTmp(Q__T)*cmpl_o+sum9_1 +sum9_2
      outState(?SAVE?(Q__B,iElem,varSys%nScalars,neigh )) =                    &
        &                                    pdfTmp(Q__B)*cmpl_o-sum9_1 +sum9_2

    enddo nodeloop
!$omp end do nowait

  end subroutine bgk_advRel_d3q27
!******************************************************************************!


end module mus_d3q27_module
!******************************************************************************!
