! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module provides the MUSUBI specific functions for calculating 
!! macroscopic quantities from the state variables for incompressible LBM 
!! models.
!!
!! The depending common interface between MUSUBI and ATELES is defined in the
!! tem_derived_module. The functionality for accessing a variable from the state
!! and evaluating a lua function are also provided in the tem_derived module.
!!
!! \author Kannan Masilamani
!!
?? include 'header/lbm_macros.inc'
?? include 'treelm/source/deriveMacros.inc'
module mus_derQuanIncomp_module

  ! include treelm modules
  use tem_param_module,        only: div1_2, div1_3, div1_54, div1_9, div3_4,  &
    &                                sqrt3, cs2inv, cs2
  use env_module,              only: rk, long_k, globalMaxLevels
  use tem_var_system_module,   only: tem_var_system_type
  use tem_variable_module,     only: tem_variable_type
  use tem_stencil_module,      only: tem_stencilHeader_type
  use tem_topology_module,     only: tem_levelOf
  use tem_tools_module,        only: tem_write
  use tem_time_module,         only: tem_time_type
  use treelmesh_module,        only: treelmesh_type
  use tem_subTree_type_module, only: tem_subTree_type, tem_treeIDfrom_subTree

  ! include musubi modules
  use mus_solSpecHelpers_module, only: getWeights, getFieldVariable

  ! include aotus modules
  use aotus_module, only: flu_State

  implicit none

  private

  public :: derivePressureIncomp_FromState
  public :: deriveVelocityIncomp_FromState
  public :: deriveVelMagIncomp_FromState
  public :: deriveEquilIncomp_FromState
  public :: deriveNonEquilIncomp_FromState
  public :: deriveEquilIncomp_FromMacro
  public :: deriveKineEnerIncomp_FromState
  public :: deriveShearMagIncomp_FromState
  public :: deriveStrainRateIncomp_FromState

contains


!******************************************************************************!
!       Subroutines with common interface for the function pointers            !
!******************************************************************************!

!******************************************************************************!
  !> Calculate the velocity of a given element number with the given input
  !!        vector (sum up all values) for incompressible model
  !! \f$ \rho = 1.0 \f$
  !! \f$ \vec{u} = \sum(f_i c_i) \f$
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(deriveVelocityIncomp_FromState)
    !---------------------------------------------------------------------------
    integer :: iElem
    integer :: nElems
    real(kind=rk) :: rho0
    integer :: iDep, iDir, pos, dep, iComp
    real(kind=rk) :: vel( requVar%nComponents )
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    !---------------------------------------------------------------------------
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    !get reference density
    rho0 = getFieldVariable( conf      = globSys%solSpec_conf,                 &
      &                      varLabel  = trim(requVar%label),                  &
      &                      varName   = 'velMag',                             &
      &                      fieldVar  = 'rho0',                               &
      &                      fieldProp = 'fluid' )

    do iElem = 1, nElems
      ! initialize the local velocity for every element
      vel = 0._rk
      do iDep = 1, size( globSys%variable%val( requVar%mapping )%varDep )
        ! get the position of the depending variable in the global varSys
        dep = globSys%variable%val( requVar%mapping )%varDep( iDep )
      
        do iDir = 1, globSys%variable%val( dep )%nComponents
          ! get the position of the current direction of the depending variable
          ! in the global system (= position in the input)
          pos = globSys%variable%val( dep )%varPos( iDir )
          do iComp = 1,requVar%nComponents
            vel(iComp) = vel(iComp) + input( (iElem - 1) *                     &
              &                              globSys%nScalars + pos )          &
              & * real(stencil%cxDir( requVar%sub_index(iComp), iDir ), kind=rk)
          end do !icomp
        end do !idir
      end do !idep
      ! copy the values to the right position in the inout array input
      local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        &    = vel / rho0 
    end do

  end subroutine deriveVelocityIncomp_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the pressure variation of a given set of elements for the 
  !! incompressible Navier-Stokes equations
  !!
  !! Pressure variation calculation according to the isentropic equation of 
  !! state for the LBM \f$ p = (\rho - \rho0) c_s^2 \f$
  !! with the calculation of density as in deriveDensity_FromState
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(derivePressureIncomp_FromState)
    !---------------------------------------------------------------------------
    integer :: iDep, dep, iElem, buf_start, buf_end, nElems
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: dens, rho0
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    !get reference density
    rho0 = getFieldVariable( conf      = globSys%solSpec_conf,                 &
      &                      varLabel  = trim(requVar%label),                  &
      &                      varName   = 'pressure',                           &
      &                      fieldVar  = 'rho0',                               &
      &                      fieldProp = 'fluid' )

    do iElem = 1, nElems
      dens = 0.0_rk
      do iDep = 1, size( globSys%variable%val( requVar%mapping )%varDep )
        ! get the position of the depending variable in the global varSys
        dep = globSys%variable%val( requVar%mapping )%varDep( iDep )
      
        ! set the start and end position in the state array
        buf_start = (iElem - 1)*globSys%nScalars                               &
          &   + globSys%variable%val( dep )%varPos( 1 )
      
        buf_end   = (iElem - 1)*globSys%nScalars                               &
          &   + globSys%variable%val( dep )%varPos(                            &
          &                            globSys%variable%val( dep )%nComponents )
      
        ! derive the density and store it in local_res
        ! derive the density and store it in the input
        dens = dens + sum( input( buf_start : buf_end ))
      
      end do  ! iDep
      local_res( iElem ) = (dens)*cs2
    end do  ! iElem

  end subroutine derivePressureIncomp_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the velocity magnitude of a given element number with the given
  !! input vector (sum up all values)
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(deriveVelMagIncomp_FromState)
    !---------------------------------------------------------------------------
    integer :: iElem, vel_pos, pdf_pos, pdf_nComps, nElems
    integer :: buf_start, buf_end
    integer :: localBound(2)
    real(kind=rk) :: vel( 3 ) ! velocity results
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    !---------------------------------------------------------------------------
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! position of vel in glob system
    vel_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )
    ! position of pdf in glob system
    pdf_pos = globSys%variable%val( vel_pos )%varDep( 1 )
    pdf_nComps = globSys%variable%val( pdf_pos )%nComponents

    do iElem = 1, nElems
      ! set the bounds to be iElem call is done elementwise
      localBound(1) = iElem
      localBound(2) = iElem
      buf_start = (iElem - 1)*globSys%nScalars                                 &
        &   + globSys%variable%val( pdf_pos )%varPos( 1 )
      buf_end   = (iElem - 1)*globSys%nScalars                                 &
        &   + globSys%variable%val( pdf_pos )%varPos( pdf_nComps )

      ! First calculate velocity vector, by calling its function pointer
      ! save results to vel array
      call globSys%derive( vel_pos )%do(                                       &
        &                      input   = input( buf_start:buf_end ),           &
        &                      bound   = localBound,                           &
        &                      tree    = tree,                                 &
        &                      subTree = subTree,                              &
        &                      requVar = globSys%variable%val(vel_pos),        &
        &                      globSys = globSys,                              &
        &                      stencil = stencil,                              &
        &                      conf    = conf,                                 &
        &                      res     = vel )

      ! Calculate the square root of velocity vector
      local_res( iElem ) = sqrt( vel(1)*vel(1) + vel(2)*vel(2) + vel(3)*vel(3) )
    enddo ! iElem

  end subroutine deriveVelMagIncomp_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the shear stress magnitude of a given element number with the 
  !! given input vector (sum up all values)
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(deriveShearMagIncomp_FromState)
    !---------------------------------------------------------------------------
    integer :: iElem, tau_pos, pdf_pos, pdf_nComps, nElems
    integer :: buf_start, buf_end
    integer :: localBound(2)
    real(kind=rk) :: tau( 6 ) ! shear stress result
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    !---------------------------------------------------------------------------
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! position of shear stress in glob system
    tau_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )
    ! position of pdf in glob system
    pdf_pos = globSys%variable%val( tau_pos )%varDep( 1 )
    pdf_nComps = globSys%variable%val( pdf_pos )%nComponents

    do iElem = 1, nElems
      ! set the bounds to be iElem call is done elementwise
      localBound(1) = iElem
      localBound(2) = iElem
      buf_start = (iElem - 1)*globSys%nScalars                                 &
        &   + globSys%variable%val( pdf_pos )%varPos( 1 )
      buf_end   = (iElem - 1)*globSys%nScalars                                 &
        &   + globSys%variable%val( pdf_pos )%varPos( pdf_nComps )
      
      ! First calculate velocity vector, by calling its function pointer
      ! save results to tau array
      call globSys%derive( tau_pos )%do(                                       &
        &                      input   = input( buf_start:buf_end ),           &
        &                      bound   = localBound,                           &
        &                      tree    = tree,                                 &
        &                      subTree = subTree,                              &
        &                      requVar = globSys%variable%val(tau_pos),        &
        &                      globSys = globSys,                              &
        &                      stencil = stencil,                              &
        &                      conf    = conf,                                 &
        &                      res     = tau,                                  &
        &                      bary    = reshape((/bary( iElem, :)/),(/1,3/)), &
        &                      time    = time )
      
      ! Calculate the square root of shear vector
      ! local_res( iElem ) = sqrt( sum( tau(:)*tau(:) )) 
      ! Von mises-criterion (see wikipedia article on von Mises yield criterion)
      ! wss = (tau_xx-tau_yy)**2 + (tau_yy-tau_zz)**2 + (tau_xx-tau_zz)**2  &
      ! &   + 6._rk*( tau_xy*tau_xy + tau_yz*tau_yz + tau_xz*tau_xz)
      ! wss = sqrt(0.5_rk*wss )
      local_res( iElem ) = sqrt( 0.5_rk*(                                      &
        &   (tau(1)-tau(2))**2 + (tau(2)-tau(3))**2 + (tau(1)-tau(3))**2       &
        &   + 6._rk*( tau(4)*tau(4) + tau(5)*tau(5) + tau(6)*tau(6))           &
        &   ))
    enddo ! iElem

  end subroutine deriveShearMagIncomp_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the equlibrium of a given element number with the given input
  !! state vector for incompressible model
  !! \f$ \rho0 = 1.0 \f$
  !! \f$ f_eq = w_\alpha*(\rho + \rho0( (u . c_i) + (u . c_i)^2 - u^2 )  \f$)
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(deriveEquilIncomp_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: dens !< temporary array for density
    real(kind=rk) :: vel(3) !< temporary array for velocity
    integer :: iDir, iElem, dep, nComp, nElems
    integer :: posBgn, posEnd
    real(kind=rk) :: ucx, usq
    real(kind=rk) :: rho0
    real(kind=rk) :: weights(stencil%QQ) ! weights read from spec_string
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    !get reference density
    rho0 = getFieldVariable( conf      = globSys%solSpec_conf,                 &
      &                      varLabel  = trim(requVar%label),                  &
      &                      varName   = 'equilFromState',                     &
      &                      fieldVar  = 'rho0',                               &
      &                      fieldProp = 'fluid' )

    ! get the weights from the solver specific character
    weights = getWeights( globSys%solSpec_conf, stencil )

    ! get the position of the depending variable in the global varSys
    ! it is always for only one field.
    dep = globSys%variable%val( requVar%mapping )%varDep( 1 )
    nComp = globSys%variable%val(dep)%nComponents

    !> Calculation of density and velocity
    do iElem = 1, nElems
      dens = 0._rk
      vel = 0._rk
      
      ! set the begin and end position in the state array to compute density
      posBgn = (iElem-1) * globSys%nScalars                                    &
        &   + globSys%variable%val( dep )%varPos( 1 )
      posEnd = (iElem-1) * globSys%nScalars                                    &
        &   + globSys%variable%val( dep )%varPos( nComp )
      dens = sum( input( posBgn:posEnd ) )
      
      !> Velocity calculation
      do iDir = 1, stencil%QQ
        vel = vel + input( (iElem - 1) * globSys%nScalars +                    &
          & globSys%variable%val(dep)%varPos( iDir ) )                         &
          &   * real(stencil%cxDir( 1:3, iDir ), kind=rk)
      end do
      vel = vel / rho0
      
      ! Calculate the square of velocity
      usq = vel(1) * vel(1) + vel(2) * vel(2) + vel(3) * vel(3)
      
      do iDir = 1, stencil%QQ
        ! Velocity times lattice unit velocity
        ucx =    real( stencil%cxDir( 1, iDir ), kind=rk) * vel(1)             &
          &    + real( stencil%cxDir( 2, iDir ), kind=rk) * vel(2)             &
          &    + real( stencil%cxDir( 3, iDir ), kind=rk) * vel(3)
      
        ! calculate equilibrium density
        local_res( (iElem-1)*stencil%QQ + iDir ) =                             &
          &   weights( iDir ) * ( dens + rho0 * (                              &
          &                              ucx * cs2inv                          &
          &                            + ucx * ucx * cs2inv * cs2inv * 0.5_rk  &
          &                            - usq * cs2inv * 0.5_rk ) )
      end do ! iDir
    end do ! iElem

  end subroutine deriveEquilIncomp_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the equlibrium of a given element number with the given input
  !! vector.
  !!
  !! input array contains density and velocity in the following order
  !! |dens_1 | dens_2 | ... | dens_nElems | u_x_1 | u_y_1 | u_z_1 | ....
  !! .. | u_x_nElems | u_y_nElems | u_z_nElems |
  !! \f$ \rho0 = 1.0 \f$
  !! \f$ f_eq = w_\alpha*(\rho + \rho0( (u . c_i) + (u . c_i)^2 - u^2 )  \f$)
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(deriveEquilIncomp_FromMacro)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: ucx, usq
    real(kind=rk) :: rho0
    integer :: iDir, iElem, posVel, offVel, nElems
    real(kind=rk) :: weights(stencil%QQ) ! weights read from spec_string
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    !get reference density
    rho0 = getFieldVariable( conf      = globSys%solSpec_conf,                 &
      &                      varLabel  = trim(requVar%label),                  &
      &                      varName   = 'equilFromState',                     &
      &                      fieldVar  = 'rho0',                               &
      &                      fieldProp = 'fluid' )

    ! get the weights from the solver specific character
    weights = getWeights( globSys%solSpec_conf, stencil )

    offVel = nElems   !< offset of velocity components in input array

    do iElem = 1, nElems
      ! square of velocity
      posVel = (iElem-1) * 3 + offVel ! position of this element's vel in input array
      usq =   input( posVel+1 ) * input( posVel+1 )                            &
        &   + input( posVel+2 ) * input( posVel+2 )                            &
        &   + input( posVel+3 ) * input( posVel+3 )
      
      do iDir = 1, stencil%QQ
      
        ! velocity times lattice unit velocity
        ucx =    real( stencil%cxDir( 1, iDir ), kind=rk)*input( posVel+1 )    &
          &    + real( stencil%cxDir( 2, iDir ), kind=rk)*input( posVel+2 )    &
          &    + real( stencil%cxDir( 3, iDir ), kind=rk)*input( posVel+3 )
      
        ! calculate equilibrium density
        local_res( (iElem-1)*stencil%QQ + iDir ) = weights( iDir )             &
          &         * ( input( iElem ) + rho0 * ( ucx*cs2inv                   &
          &                                    + ucx*ucx*cs2inv*cs2inv*0.5_rk  &
          &                                    - usq*cs2inv*0.5_rk ) )
      
      enddo ! iDir
    end do ! iElem

  end subroutine deriveEquilIncomp_FromMacro
!******************************************************************************!


!******************************************************************************!
  !> Calculate the Non-Equlibrium of given elements with given input state
  !! array.
  !!
?? copy :: derive_headtxt(deriveNonEquilIncomp_FromState)
    !---------------------------------------------------------------------------
    !> local variables
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: dens !< temporary array for density
    real(kind=rk) :: vel(3) !< temporary array for velocity
    integer :: iDir, iElem, dep, nComp, nElems
    integer :: posBgn, posEnd
    real(kind=rk) :: ucx, usq
    real(kind=rk) :: rho0
    real(kind=rk) :: weights(stencil%QQ) ! weights read from spec_string
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    !get reference density
    rho0 = getFieldVariable( conf      = globSys%solSpec_conf,                 &
      &                      varLabel  = trim(requVar%label),                  &
      &                      varName   = 'equilFromState',                     &
      &                      fieldVar  = 'rho0',                               &
      &                      fieldProp = 'fluid' )

    ! get the weights from the solver specific character
    weights = getWeights( globSys%solSpec_conf, stencil )

    ! get the position of the depending variable in the global varSys
    ! it is always for only one field.
    dep = globSys%variable%val( requVar%mapping )%varDep( 1 )
    nComp = globSys%variable%val(dep)%nComponents
    !> Calculation of density and velocity
    do iElem = 1, nElems
      dens = 0._rk
      vel = 0._rk
      
      ! set the begin and end position in the state array to compute density
      posBgn = (iElem-1) * globSys%nScalars                                    &
        &   + globSys%variable%val( dep )%varPos( 1 )                          
      posEnd = (iElem-1) * globSys%nScalars                                    &
        &   + globSys%variable%val( dep )%varPos( nComp )
      dens = sum( input( posBgn:posEnd ) )
      
      ! Calculate velocity
      do iDir = 1, stencil%QQ
        vel = vel + input( (iElem - 1) * globSys%nScalars +                    &
          & globSys%variable%val(dep)%varPos( iDir ) )                         &
          &   * real(stencil%cxDir( 1:3, iDir ), kind=rk)
      end do
      vel = vel / rho0
      
      ! Calculate the square of velocity
      usq = vel(1) * vel(1) + vel(2) * vel(2) + vel(3) * vel(3)
      
      do iDir = 1, stencil%QQ
        ! Velocity times lattice unit velocity
        ucx =    real( stencil%cxDir( 1, iDir ), kind=rk) * vel(1)             &
          &    + real( stencil%cxDir( 2, iDir ), kind=rk) * vel(2)             &
          &    + real( stencil%cxDir( 3, iDir ), kind=rk) * vel(3)
      
        ! calculate Non-Equilibrium density
        local_res( (iElem-1)*stencil%QQ + iDir ) =                             &
          &   input( (iElem-1)*stencil%QQ + iDir )                             &
          & - weights( iDir ) * ( dens + rho0 * (                              &
          &                              ucx * cs2inv                          &
          &                            + ucx * ucx * cs2inv * cs2inv * 0.5_rk  &
          &                            - usq * cs2inv * 0.5_rk ) )
      end do ! iDir
    end do ! iElem

  end subroutine deriveNonEquilIncomp_FromState
!******************************************************************************!

!******************************************************************************!
?? copy :: derive_headtxt(deriveStrainRateIncomp_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: stenCxDir(6, stencil%QQ) ! stencil cxDir
    real(kind=rk) :: tau(requVar%nComponents), nonEq(stencil%QQ)
    integer       :: iDir, iElem, iLevel, buf_start, buf_end, nElems
    integer       :: nonEq_pos, pdf_pos, pdf_nComps
    real(kind=rk) :: omega( globalMaxLevels), iOmega, rho0
    integer(kind=long_k), allocatable :: treeID(:)
    integer :: localBound(2)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1
    ! get the treeIDs
    call tem_treeIDfrom_subTree( subTree, tree, treeID, bound )

    ! Get all omegas from all levels
    omega(:) = getFieldVariable( conf = globSys%solSpec_conf,                  &
      &                       varLabel = trim(requVar%label) ,                 &
      &                       varName = 'shearstress',                         &
      &                       fieldVar = 'omLvl',                              &
      &                       fieldProp = 'fluid', nVals = globalMaxLevels )

    !get reference density
    rho0 = getFieldVariable( conf      = globSys%solSpec_conf,                 &
      &                      varLabel  = trim(requVar%label),                  &
      &                      varName   = 'velMag',                             &
      &                      fieldVar  = 'rho0',                               &
      &                      fieldProp = 'fluid' )

    ! convert stencil cxDir into real type
    stenCxDir(1, :) =   real(stencil%cxDir(1, :), kind=rk)                     &
      &               * real(stencil%cxDir(1, :), kind=rk)
    stenCxDir(2, :) =   real(stencil%cxDir(2, :), kind=rk)                     &
      &               * real(stencil%cxDir(2, :), kind=rk)
    stenCxDir(3, :) =   real(stencil%cxDir(3, :), kind=rk)                     &
      &               * real(stencil%cxDir(3, :), kind=rk)
    stenCxDir(4, :) =   real(stencil%cxDir(1, :), kind=rk)                     &
      &               * real(stencil%cxDir(2, :), kind=rk)
    stenCxDir(5, :) =   real(stencil%cxDir(2, :), kind=rk)                     &
      &               * real(stencil%cxDir(3, :), kind=rk)
    stenCxDir(6, :) =   real(stencil%cxDir(1, :), kind=rk)                     &
      &               * real(stencil%cxDir(3, :), kind=rk)

    ! position of pdf in glob system
    pdf_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )
    pdf_nComps = globSys%variable%val( pdf_pos )%nComponents
    ! position of nonEquilibrium in glob system
    nonEq_pos = globSys%variable%val( requVar%mapping )%varDep( 2 )

    do iElem = 1, nElems
      ! set the bounds to be iElem call is done elementwise
      localBound(1) = iElem
      localBound(2) = iElem
      nonEq = 0._rk

      ! set the begin and end position in the input state array
      buf_start = (iElem - 1)*globSys%nScalars                &
        &   + globSys%variable%val( pdf_pos )%varPos( 1 )
      buf_end   = (iElem - 1)*globSys%nScalars                &
        &   + globSys%variable%val( pdf_pos )%varPos( pdf_nComps )

      ! Calculate nonEquilibrium density
      ! for PULL, this is post-collision Non-Eq
      ! for PUSH, this is  pre-collision Non-Eq
      call globSys%derive( nonEq_pos )%do(                         &
        &                    input   = input(buf_start:buf_end),   &
        &                    bound   = localBound,                 &
        &                    tree    = tree,                       &
        &                    subTree = subTree,                    &
        &                    requVar = globSys%variable%val(nonEq_pos) ,&
        &                    globSys = globSys,                    &
        &                    stencil = stencil,                    &
        &                    conf    = conf,                       &
        &                    res     = nonEq  )

      ! get my current level
      iLevel = tem_levelOf( treeID( iElem ) )
      ! get the correct omega value
      iOmega = omega( iLevel )

      ! Shear stress requires pre-collision NonEq
      ! Convert post-coll to pre-coll by a factor of 1/(1-omega), if use PULL
?? if( .not. STREAM_PUSH ) then
      if( iOmega /= 1._rk ) then
        nonEq = nonEq / ( 1._rk - iOmega )
      else
        nonEq = 0._rk
      endif
?? endif

      tau = 0._rk
      do iDir = 1, stencil%QQ
        tau(1) = tau(1) + nonEq(iDir) * stenCxDir(1, iDir) ! tau_xx
        tau(2) = tau(2) + nonEq(iDir) * stenCxDir(2, iDir) ! tau_yy
        tau(3) = tau(3) + nonEq(iDir) * stenCxDir(3, iDir) ! tau_zz
        tau(4) = tau(4) + nonEq(iDir) * stenCxDir(4, iDir) ! tau_xy
        tau(5) = tau(5) + nonEq(iDir) * stenCxDir(5, iDir) ! tau_yz
        tau(6) = tau(6) + nonEq(iDir) * stenCxDir(6, iDir) ! tau_xz
      enddo ! iDir

      local_res( (iElem-1)*6+1 : iElem*6 ) = -1.5_rk * iOmega / rho0 * tau(:)
    enddo ! iElem

  end subroutine deriveStrainRateIncomp_FromState
!******************************************************************************!

!******************************************************************************!
  !> Calculate the kinetic energy as
  !!  E = 1/2 rho ( ux^2 + uy^2 +uz^2 )
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(deriveKineEnerIncomp_FromState)
    !---------------------------------------------------------------------------
    !> local variables
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: dens ! temporary array for density
    real(kind=rk) :: vel(3) ! temporary array for velocity
    integer :: iDir, iElem, dep, nComp, nElems
    integer :: buf_start, buf_end
    real(kind=rk) :: weights(stencil%QQ) ! weights read from spec_string
    !---------------------------------------------------------------------------
  
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1
  
    ! get the weights from the solver specific character
    weights = getWeights( globSys%solSpec_conf, stencil )

    !get reference density
    dens = getFieldVariable( conf      = globSys%solSpec_conf,                 &
      &                      varLabel  = trim(requVar%label),                  &
      &                      varName   = 'kinetic_energy',                     &
      &                      fieldVar  = 'rho0',                               &
      &                      fieldProp = 'fluid' )
  
    ! get the position of the depending variable in the global varSys
    ! it is always for only one field.
    dep = globSys%variable%val( requVar%mapping )%varDep( 1 )
    nComp = globSys%variable%val(dep)%nComponents
    !> Calculation of density and velocity
    do iElem = 1, nElems
      vel = 0._rk
      
      ! set the begin and end position in the state array to compute density
      buf_start = (iElem-1) * globSys%nScalars                                 &
        &   + globSys%variable%val( dep )%varPos( 1 )
      buf_end = (iElem-1) * globSys%nScalars                                   &
        &   + globSys%variable%val( dep )%varPos( nComp )
      
      !> Velocity calculation
      do iDir = 1, stencil%QQ
        vel = vel + input( (iElem - 1) * globSys%nScalars +                    &
          & globSys%variable%val(dep)%varPos( iDir ) )                         &
          &   * real(stencil%cxDir( 1:3, iDir ), kind=rk)
      end do
      vel = vel / dens
      
      local_res( iElem ) = 0.5_rk*dens*(                                       &
        &       vel(1)*vel(1) + vel(2)*vel(2) + vel(3)*vel(3) )
    end do ! iElem
  
  end subroutine deriveKineEnerIncomp_FromState
!******************************************************************************!


end module mus_derQuanIncomp_module
!******************************************************************************!
