! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module provides all possible 'pure variables' (= no prefixes) for a
!!  given kind of simulation.
!!
!! IMPORTANT NOTE: When you create a new variable, do not forget to add this 
!! variable
!! name into mus_derVarPos_type and also in mus_build_derVarPos routine.
!!
!! \author Simon Zimny
!
module mus_variable_module

  ! include treelm modules
  use env_module,                   only: rk, globalMaxLevels, long_k, labelLen
  use tem_aux_module,               only: tem_abort
  use tem_tools_module,             only: tem_write, tem_horizontalSpacer
  use tem_var_system_module,        only: tem_var_system_type, tem_var_append
  use tem_stencil_module,           only: tem_stencilHeader_type
  use tem_derived_module,           only: evalSpacetime, evalDiff
  use tem_var_module,               only: tem_var_type
  use tem_meshInfo_module,          only: tem_initMeshInfoList
  use tem_variable_module,          only: tem_variable_type

  ! include musubi modules
  use mus_derQuan_module,           only: deriveDensity_FromState,             &
    &                                     derivePressure_FromState,            &
    &                                     deriveVelocity_FromState,            &
    &                                     deriveEquil_FromState,               &
    &                                     deriveNonEquil_FromState,            &
    &                                     deriveEquil_FromMacro,               &
    &                                     deriveNonEquil_FromMacro,            &
    &                                     deriveEquil_ps_fromMacro,            &
    &                                     deriveKineEner_FromState,            &
    &                                     deriveShearStress_FromState,         &
    &                                     deriveShearMag_FromState,            &
    &                                     deriveWSS_FromState,                 &
    &                                     deriveTemp_FromState,                &
    &                                     deriveVelMag_FromState,              &
    &                                     deriveStrainRate_FromState          
  use mus_derQuanIncomp_module,     only: deriveVelocityIncomp_FromState,      &
    &                                     derivePressureIncomp_FromState,      &
    &                                     deriveVelMagIncomp_FromState,        &
    &                                     deriveEquilIncomp_FromState,         &
    &                                     deriveNonEquilIncomp_FromState,      &
    &                                     deriveEquilIncomp_FromMacro,         &
    &                                     deriveShearMagIncomp_FromState,      &
    &                                     deriveKineEnerIncomp_FromState,      &
    &                                     deriveStrainRateIncomp_FromState
  use mus_derQuanPhysics_module,    only: derivePressurePhy_FromState,         &
    &                                     deriveDensityPhy_FromState,          &
    &                                     deriveNumDensityPhy_FromState,       &
    &                                     deriveMoleFluxPhy_FromState,         &
    &                                     deriveVelocityPhy_FromState,         &  
    &                                     deriveWSSPhy_FromState,              &  
    &                                     deriveKineEnerPhy_FromState,         &
    &                                     deriveShearStressPhy_FromState,      &
    &                                     deriveShearMagPhy_FromState,         &
    &                                     deriveVelMagPhy_FromState,           &
    &                                     deriveTempPhy_FromState,             &
    &                                     deriveStrainRatePhy_FromState

  implicit none
  private

  public :: mus_derVarPos_type
  public :: mus_init_varSys
  public :: mus_build_derVarList
  public :: mus_kill_derVarList
  public :: mus_store_derVarPos
  public :: mus_init_variables

  !> This type stores the position of each variable in the global sys
  type mus_derVarPos_type
    integer :: density = -1
    integer :: numDensity = -1
    integer :: pressure = -1
    integer :: velocity = -1
    integer :: equilvelFromState = -1
    integer :: equilvelFromMacro = -1
    integer :: velMag = -1
    integer :: shearStress = -1
    integer :: strainRate = -1
    integer :: shearMag    = -1
    integer :: wss = -1
    integer :: temperature = -1
    integer :: equilFromState = -1
    integer :: equilFromMacro = -1
    integer :: nonEquilFromState = -1
    integer :: omega = -1
    integer :: molefrac = -1
    integer :: moleflux = -1
    integer :: kineticEnergy = -1
  end type mus_derVarPos_type

  ! pure variables, their dependencies and regular function pointers without prefixes
  ! in case one of the following is a state variable direct access will be granted

  ! density
  type(tem_var_type), SAVE :: density_var
  ! number density (density/molecular_weight)
  type(tem_var_type), SAVE :: numDensity_var
  ! velocity
  type(tem_var_type), SAVE :: velocity_var
  ! equilibrium velocity from macro
  type(tem_var_type), SAVE :: equilvelFromMacro_var
  ! equilibrium velocity from state
  type(tem_var_type), SAVE :: equilvelFromState_var
  ! pressure
  type(tem_var_type), SAVE :: pressure_var
  ! equilibrium from state
  type(tem_var_type), SAVE :: equilFromState_var
  ! equilibrium from state
  type(tem_var_type), SAVE :: nonEquilFromState_var
  ! equilibrium from Macroscopic variable
  type(tem_var_type), SAVE :: equilFromMacro_var
  ! omega
  type(tem_var_type), SAVE :: omega_var
  ! shearStress
  type(tem_var_type), SAVE :: shearStress_var
  ! shearStress magnitude
  type(tem_var_type), SAVE :: shearMag_var
  ! wall shear stress
  type(tem_var_type), SAVE :: wss_var
  ! mole fraction
  type(tem_var_type), SAVE :: molefrac_var
  ! mole flux
  type(tem_var_type), SAVE :: moleflux_var
  ! velocity magnitude
  type(tem_var_type), SAVE :: velMag_var
  ! difference bewteen spacetime function and another quantity
  type(tem_var_type), SAVE :: diff_var
  ! kinetic energy E = 1/2 rho sqrt( ux^2 + uy^2 +uz^2 )
  type(tem_var_type), SAVE :: kineticEnergy_var
  ! temperature
  type(tem_var_type), SAVE :: temperature_var
  ! shear strain
  type(tem_var_type), SAVE :: strainRate_var

  ! variable physical units
  ! density
  type(tem_var_type), SAVE :: densityPhy_var
  ! number density (density/molecular_weight)
  type(tem_var_type), SAVE :: numDensityPhy_var
  ! pressure
  type(tem_var_type), SAVE :: pressurePhy_var
  ! velocity
  type(tem_var_type), SAVE :: velocityPhy_var
  ! velocity magnitude
  type(tem_var_type), SAVE :: velMagPhy_var
  ! shearStress
  type(tem_var_type), SAVE :: shearStressPhy_var
  ! shear Magnitude
  type(tem_var_type), SAVE :: shearMagPhy_var
  ! wall shear stress
  type(tem_var_type), SAVE :: wssPhy_var
  ! kinetic energy E = 1/2 rho sqrt( ux^2 + uy^2 +uz^2 )
  type(tem_var_type), SAVE :: kineticEnergyPhy_var
  ! temperature
  type(tem_var_type), SAVE :: temperaturePhy_var
  ! mole flux
  type(tem_var_type), SAVE :: molefluxPhy_var
  ! shear strain
  type(tem_var_type), SAVE :: strainRatePhy_var

  contains

!******************************************************************************!
  !> This routine initialize all derive quantity variables with theit label,
  !! nComponents and dependencies.
  !!
  !! Only the function pointer for density is set here since it is
  !! generic for all different scheme kinds.
  !! All other function pointers are initialized based on their scheme kind
  !! in mus_build_dervarList
  !!
  subroutine mus_init_variables( )
    !---------------------------------------------------------------------------

    ! density
    density_var%variable%mapping = 1
    density_var%variable%label = 'density'
    density_var%variable%nComponents = 1
    if( .not. allocated(density_var%variable%depVars) ) then
      allocate(density_var%variable%depVars(1))
    end if
    density_var%variable%depVars(1) = 'pdf'
    density_var%do => deriveDensity_FromState

    ! density
    densityPhy_var%variable%mapping = 1
    densityPhy_var%variable%label = 'density_phy'
    densityPhy_var%variable%nComponents = 1
    if( .not. allocated(densityPhy_var%variable%depVars) ) then
      allocate(densityPhy_var%variable%depVars(1))
    end if
    densityPhy_var%variable%depVars(1) = 'density'
    densityPhy_var%do => deriveDensityPhy_FromState

    ! numDensity
    numDensityPhy_var%variable%mapping = 1
    numDensityPhy_var%variable%label = 'numDensity_phy'
    numDensityPhy_var%variable%nComponents = 1
    if( .not. allocated(numDensityPhy_var%variable%depVars) ) then
      allocate(numDensityPhy_var%variable%depVars(1))
    end if
    numDensityPhy_var%variable%depVars(1) = 'numDensity'
    numDensityPhy_var%do => deriveNumDensityPhy_FromState

    ! Pressure
    pressure_var%variable%mapping = 1
    pressure_var%variable%label = 'pressure'
    pressure_var%variable%nComponents = 1
    if( .not. allocated(pressure_var%variable%depVars) ) then
      allocate(pressure_var%variable%depVars(1))
    end if
    pressure_var%variable%depVars(1) = 'pdf'

    pressurePhy_var%variable%mapping = 1
    pressurePhy_var%variable%label = 'pressure_phy'
    pressurePhy_var%variable%nComponents = 1
    if( .not. allocated(pressurePhy_var%variable%depVars) ) then
      allocate(pressurePhy_var%variable%depVars(1))
    end if
    pressurePhy_var%variable%depVars(1) = 'pressure'
    pressurePhy_var%do => derivePressurePhy_FromState

    ! velocity
    velocity_var%variable%label = 'velocity'
    velocity_var%variable%nComponents = 3
    if( .not. allocated(velocity_var%variable%depVars) ) then
      allocate(velocity_var%variable%depVars(1))
    end if
    velocity_var%variable%depVars(1) = 'pdf'

    ! equilibrium velocity for multispecies
    equilvelFromState_var%variable%label = 'equilvelFromState'
    equilvelFromState_var%variable%nComponents = 3
    if( .not. allocated(equilvelFromState_var%variable%depVars) ) then
      allocate(equilvelFromState_var%variable%depVars(1))
    end if
    equilvelFromState_var%variable%depVars(1) = 'pdf'

    ! equilibrium velocity for multispecies
    equilvelFromMacro_var%variable%label = 'equilvelFromMacro'
    equilvelFromMacro_var%variable%nComponents = 3
    if( .not. allocated(equilvelFromMacro_var%variable%depVars) ) then
      allocate(equilvelFromMacro_var%variable%depVars(1))
    end if
    equilvelFromMacro_var%variable%depVars(1) = 'pdf'
    !\todo KM: 
    !equilFromMacro_var%variable%depVars(1) = 'density'
    !equilFromMacro_var%variable%depVars(2) = 'velocity'

    velocityPhy_var%variable%label = 'velocity_phy'
    velocityPhy_var%variable%nComponents = 3
    if( .not. allocated(velocityPhy_var%variable%depVars) ) then
      allocate(velocityPhy_var%variable%depVars(1))
    end if
    velocityPhy_var%variable%depVars(1) = 'velocity'
    velocityPhy_var%do => deriveVelocityPhy_FromState

    ! Temperature
    temperature_var%variable%mapping = 1
    temperature_var%variable%label = 'temperature'
    temperature_var%variable%nComponents = 1
    if( .not. allocated(temperature_var%variable%depVars) ) then
      allocate(temperature_var%variable%depVars(1))
    end if
    temperature_var%variable%depVars(1) = 'pdf'
    temperature_var%do => deriveTemp_FromState

    temperaturePhy_var%variable%mapping = 1
    temperaturePhy_var%variable%label = 'temperature_phy'
    temperaturePhy_var%variable%nComponents = 1
    if( .not. allocated(temperaturePhy_var%variable%depVars) ) then
      allocate(temperaturePhy_var%variable%depVars(1))
    end if
    temperaturePhy_var%variable%depVars(1) = 'temperature'
    temperaturePhy_var%do => deriveTempPhy_FromState


    ! equilibrium
    equilFromState_var%variable%label = 'equilFromState'
    if( .not. allocated(equilFromState_var%variable%depVars) ) then
      allocate(equilFromState_var%variable%depVars(1))
    end if
    equilFromState_var%variable%depVars(1) = 'pdf'

    ! non-equilibrium from pdf
    nonEquilFromState_var%variable%label = 'nonEquilFromState'
    if( .not. allocated(nonEquilFromState_var%variable%depVars) ) then
      allocate(nonEquilFromState_var%variable%depVars(2))
    end if
    nonEquilFromState_var%variable%depVars(1) = 'pdf'
    nonEquilFromState_var%variable%depVars(2) = 'equilFromState'

    equilFromMacro_var%variable%label = 'equilFromMacro'
    if( .not. allocated(equilFromMacro_var%variable%depVars) ) then
      allocate(equilFromMacro_var%variable%depVars(1))
    end if
    equilFromMacro_var%variable%depVars(1) = 'pdf'
    !equilFromMacro_var%variable%depVars(1) = 'density'
    !equilFromMacro_var%variable%depVars(2) = 'velocity'

    ! omega
    omega_var%variable%label = 'omega'
    omega_var%variable%nComponents = 1
    if( .not. allocated(omega_var%variable%depVars) ) then
      allocate(omega_var%variable%depVars(1))
    end if
    omega_var%variable%depVars(1) = 'pdf'
!    omega_var%do => accessOmega_fromFieldProp2

    ! shear stress
    shearStress_var%variable%label = 'shearstress'
    shearStress_var%variable%nComponents = 6
    if( .not. allocated(shearStress_var%variable%depVars) ) then
      allocate(shearStress_var%variable%depVars(1))
    end if
    shearStress_var%variable%depVars(1) = 'nonEquilFromState'

    shearStressPhy_var%variable%label = 'shearstress_phy'
    shearStressPhy_var%variable%nComponents = 6
    if( .not. allocated(shearStressPhy_var%variable%depVars) ) then
      allocate(shearStressPhy_var%variable%depVars(1))
    end if
    shearStressPhy_var%variable%depVars(1) = 'shearstress'
    shearStressPhy_var%do    => deriveShearStressPhy_FromState

    ! shear magnitude
    shearMag_var%variable%label = 'shearMag'
    shearMag_var%variable%nComponents = 1
    if(.not. allocated(shearMag_var%variable%depVars) ) &
      allocate(shearMag_var%variable%depVars(1))
    shearMag_var%variable%depVars(1) = 'shearstress'

    shearMagPhy_var%variable%label = 'shearMag_phy'
    shearMagPhy_var%variable%nComponents = 1
    if(.not. allocated(shearMagPhy_var%variable%depVars) ) &
      allocate(shearMagPhy_var%variable%depVars(1))
    shearMagPhy_var%variable%depVars(1) = 'shearMag'
    shearMagPhy_var%do       => deriveShearMagPhy_FromState

    ! wall shear stress
    wss_var%variable%label = 'wss'
    wss_var%variable%nComponents = 1
    if( .not. allocated(wss_var%variable%depVars) ) then
      allocate(wss_var%variable%depVars(1))
    end if
    wss_var%variable%depVars(1) = 'shearstress'

    wssPhy_var%variable%label = 'wss_phy'
    wssPhy_var%variable%nComponents = 1
    if( .not. allocated(wssPhy_var%variable%depVars) ) then
      allocate(wssPhy_var%variable%depVars(1))
    end if
    wssPhy_var%variable%depVars(1) = 'wss'
    wssPhy_var%do  => deriveWSSPhy_FromState

    ! shear strain
    strainRate_var%variable%label = 'strainRate'
    strainRate_var%variable%nComponents = 6
    if( .not. allocated(strainRate_var%variable%depVars) ) then
      allocate(strainRate_var%variable%depVars(2))
    end if
    strainRate_var%variable%depVars(1) = 'pdf' ! to calculate density
    strainRate_var%variable%depVars(2) = 'nonEquilFromState'

    strainRatePhy_var%variable%label = 'strainRate_phy'
    strainRatePhy_var%variable%nComponents = 6
    if( .not. allocated(strainRatePhy_var%variable%depVars) ) then
      allocate(strainRatePhy_var%variable%depVars(1))
    end if
    strainRatePhy_var%variable%depVars(1) = 'strainRate'
    strainRatePhy_var%do    => deriveStrainRatePhy_FromState

    ! velocity magnitude
    velMag_var%variable%label = 'velMag'
    velMag_var%variable%nComponents = 1
    if(.not. allocated(velMag_var%variable%depVars) ) &
      allocate(velMag_var%variable%depVars(1))
    velMag_var%variable%depVars(1) = 'velocity'

    velMagPhy_var%variable%label = 'velMag_phy'
    velMagPhy_var%variable%nComponents = 1
    if(.not. allocated(velMagPhy_var%variable%depVars) ) &
      allocate(velMagPhy_var%variable%depVars(1))
    velMagPhy_var%variable%depVars(1) = 'velMag'
    velMagPhy_var%do         => deriveVelMagPhy_FromState

    ! kinetic energy
    kineticEnergy_var%variable%label = 'kinetic_energy'
    kineticEnergy_var%variable%nComponents = 1
    if( .not. allocated(kineticEnergy_var%variable%depVars) ) then
      allocate(kineticEnergy_var%variable%depVars(1))
    end if 
    kineticEnergy_var%variable%depVars(1) = 'pdf'

    ! kinetic energy in physical units
    kineticEnergyPhy_var%variable%label = 'kinetic_energy_phy'
    kineticEnergyPhy_var%variable%nComponents = 1
    if( .not. allocated(kineticEnergyPhy_var%variable%depVars) ) then
      allocate(kineticEnergyPhy_var%variable%depVars(1))
    end if
    kineticEnergyPhy_var%variable%depVars(1) = 'kinetic_energy'
    kineticEnergyPhy_var%do  => deriveKineEnerPhy_FromState

    ! difference between two scalar variables
    diff_var%variable%label = 'difference'
    diff_var%variable%nComponents = 1
    diff_var%do => evalDiff

  end subroutine mus_init_variables
!******************************************************************************!


!******************************************************************************!
  !> This routine build the derived variables depends on the scheme kind
  !!
  subroutine mus_build_derVarList( derVarList, schemeKind, stencil )
    !---------------------------------------------------------------------------
    !> array of all possible derived variables for this scheme kind
    !! (without prefixes)
    type(tem_var_type), allocatable, intent(inout) :: derVarList(:)
    character(len=*) :: schemeKind !< defines scheme kind
    type(tem_stencilHeader_type), intent(in) :: stencil
    !---------------------------------------------------------------------------

    !set nComponents of any variable which are dependent on scheme
    equilFromState_var%variable%nComponents = stencil%QQ
    equilFromMacro_var%variable%nComponents = stencil%QQ
    nonEquilFromState_var%variable%nComponents = stencil%QQ

    call tem_write('Building the derived variable list ...')
    ! build the right variable list depending on the scheme kind
    select case( trim( schemeKind ))
      case('lbm')
        velocity_var%do          => deriveVelocity_FromState
        velMag_var%do            => deriveVelMag_FromState
        equilFromState_var%do    => deriveEquil_FromState
        nonEquilFromState_var%do => deriveNonEquil_FromState
        equilFromMacro_var%do    => deriveEquil_fromMacro
        shearStress_var%do       => deriveShearStress_FromState
        shearMag_var%do          => deriveShearMag_FromState
        wss_var%do               => deriveWSS_FromState
        kineticEnergy_var%do     => deriveKineEner_FromState
        pressure_var%do          => derivePressure_FromState
        strainRate_var%do        => deriveStrainRate_FromState

        ! create derVarList
        allocate( derVarList( 23 ))
        derVarList = (/ density_var, velocity_var, pressure_var,               &
          &             equilFromState_var, equilFromMacro_var,                &
          &             nonEquilFromState_var,                                 &
          &             shearStress_var, shearMag_var, wss_var, velMag_var,    &
          &             strainRate_var,                                        &
          &             kineticEnergy_var, temperature_var, densityPhy_var,    &
          &             pressurePhy_var, velocityPhy_var, velMagPhy_var,       &
          &             shearStressPhy_var, shearMagPhy_var, wssPhy_var,       &
          &             strainRatePhy_var,                                     &
          &             kineticEnergyPhy_var, temperaturePhy_var /)

      case('lbm_incomp')
        velocity_var%do          => deriveVelocityIncomp_FromState
        velMag_var%do            => deriveVelMagIncomp_FromState
        equilFromState_var%do    => deriveEquilIncomp_FromState
        nonEquilFromState_var%do => deriveNonEquilIncomp_FromState
        equilFromMacro_var%do    => deriveEquilIncomp_fromMacro
        shearStress_var%do       => deriveShearStress_FromState
        shearMag_var%do          => deriveShearMagIncomp_FromState
        wss_var%do               => deriveWSS_FromState
        kineticEnergy_var%do     => deriveKineEnerIncomp_FromState
        pressure_var%do          => derivePressureIncomp_FromState
        strainRate_var%do        => deriveStrainRateIncomp_FromState

        ! create derVarList
        allocate( derVarList( 21 ))
        derVarList = (/ density_var, velocity_var, pressure_var,               &
          &             equilFromState_var, nonEquilFromState_var,             &
          &             equilFromMacro_var, shearStress_var, wss_var,          &
          &             shearMag_var, velMag_var, kineticEnergy_var,           &
          &             strainRate_var, strainRatePhy_var,                     &
          &             densityPhy_var, pressurePhy_var, velocityPhy_var,      &
          &             velMagPhy_var,                                         &
          &             shearStressPhy_var, shearMagPhy_var, wssPhy_var,       &
          &             kineticEnergyPhy_var /)

      case default
        call tem_write( ' The selected scheme kind is unknown '//trim(schemeKind) )
        call tem_abort()
    end select

  end subroutine mus_build_derVarList
!******************************************************************************!


!******************************************************************************!
  !> This routine kills the derived variables
  !!
  subroutine mus_kill_derVarList( derVarList )
    !---------------------------------------------------------------------------
    !> array of all possible derived variables for this scheme kind
    !! (without prefixes)
    type(tem_var_type), allocatable, intent(inout) :: derVarList(:)
    !---------------------------------------------------------------------------
    deallocate( derVarList )

  end subroutine mus_kill_derVarList
!******************************************************************************!


!******************************************************************************!
!> Store the position of each variable in the global system in the derVarPos
  subroutine mus_store_derVarPos( derVarPos, var_sys, label )
    !---------------------------------------------------------------------------
    type(mus_derVarPos_type), intent(inout) :: derVarPos(:)
    !> global variable system
    type(tem_var_system_type), intent(in) :: var_sys
    !> array of field label prefix. Size=nFields
    character(len=*), intent(in) :: label(:)
    !------------------------------------------------------------------------
    integer :: iField, iVar, nVars
    character(len=labelLen) :: label_loc
    !------------------------------------------------------------------------
    nVars = var_sys%nVars
    if ( var_Sys%nVars > 1 ) nVars = var_sys%nVars + 1

    call tem_write( 'storing derived quantities position ')
    do iVar = var_sys%nVars + 1, var_sys%Variable%nVals
      do iField = 1, nVars
        if ( iField > var_Sys%nVars ) then
          label_loc = ''
        else
          label_loc = trim(label(iField))
        endif  
        if ( trim(label_loc)//'density' .eq. &
             & trim(var_sys%Variable%val(iVar)%label) ) then
           derVarPos(iField)%density = iVar
        else if ( trim(label_loc)//'numDensity' .eq. &
             & trim(var_sys%Variable%val(iVar)%label) ) then
           derVarPos(iField)%numDensity = iVar
        else if ( trim(label_loc)//'pressure' .eq. &
             & trim(var_sys%Variable%val(iVar)%label) ) then
           derVarPos(iField)%pressure = iVar
        else if ( trim(label_loc)//'velocity' .eq. &
             & trim(var_sys%Variable%val(iVar)%label) ) then
           derVarPos(iField)%velocity = iVar
        else if ( trim(label_loc)//'equilvelFromState' .eq. &
             & trim(var_sys%Variable%val(iVar)%label) ) then
           derVarPos(iField)%equilvelFromState = iVar
        else if ( trim(label_loc)//'equilvelFromMacro' .eq. &
             & trim(var_sys%Variable%val(iVar)%label) ) then
           derVarPos(iField)%equilvelFromMacro = iVar
        else if ( trim(label_loc)//'equilFromState' .eq. &
             & trim(var_sys%Variable%val(iVar)%label) ) then
           derVarPos(iField)%equilFromState = iVar
        else if ( trim(label_loc)//'nonEquilFromState' .eq. &
             & trim(var_sys%Variable%val(iVar)%label) ) then
           derVarPos(iField)%nonEquilFromState = iVar
        else if ( trim(label_loc)//'equilFromMacro' .eq. &
             & trim(var_sys%Variable%val(iVar)%label) ) then
           derVarPos(iField)%equilFromMacro = iVar
        else if ( trim(label_loc)//'omega' .eq. &
             & trim(var_sys%Variable%val(iVar)%label) ) then
           derVarPos(iField)%omega = iVar
        else if ( trim(label_loc)//'shearstress' .eq. &
             & trim(var_sys%Variable%val(iVar)%label) ) then
           derVarPos(iField)%shearstress = iVar
        else if ( trim(label_loc)//'wss' .eq. &
             & trim(var_sys%Variable%val(iVar)%label) ) then
           derVarPos(iField)%wss = iVar
        else if ( trim(label_loc)//'velMag' .eq. &
             & trim(var_sys%Variable%val(iVar)%label) ) then
           derVarPos(iField)%velMag = iVar
        else if ( trim(label_loc)//'kinetic_energy' .eq. &
             & trim(var_sys%Variable%val(iVar)%label) ) then
           derVarPos(iField)%kineticEnergy = iVar
        else if ( trim(label_loc)//'strainRate' .eq. &
             & trim(var_sys%Variable%val(iVar)%label) ) then
           derVarPos(iField)%strainRate = iVar
        endif
      enddo
    end do

  end subroutine mus_store_derVarPos
!******************************************************************************!


!******************************************************************************!
  !> Register all variables in derVarList to var_sys. \n
  !! append labels as prefix.
  !!
  subroutine mus_init_varSys( derVarList, var_sys, label, tVar )
    !---------------------------------------------------------------------------
    !> array of all possible derived variables for this scheme kind
    !! (without prefixes)
    type(tem_var_type), intent(inout) :: derVarList(:)
    !> global variable system
    type(tem_var_system_type), intent(inout) :: var_sys
    !> array of field label prefix. Size=nFields
    character(len=*), intent(in) :: label(:)
    !> additional variable defined in the lua file
    type( tem_variable_type ), intent(in) :: tVar(:)
    !---------------------------------------------------------------------------
    ! local variables
    integer :: pos, iVar, iField
    character(len=labelLen) :: buffer, buffer_depVars
    type(tem_var_type), allocatable :: meshInfoList(:)
    !---------------------------------------------------------------------------

    call tem_write( 'Initializing derived quantities ')
    ! Run over all variables
    do iVar = 1, size( derVarList )
      ! Register all derived variables in the global varSys
      ! Run over all the field labels
      do iField = 1, var_sys%nVars
        ! Use field as a prefix for all the variables defined herefore
        write(buffer,'(a)') trim(label(iField))
        ! ... and append them to the global system with the prefix of the field
        !     in the label as well as in the dependencies
        call tem_var_append(                                                   &
          &       sys         = var_sys,                                       &
          &       varName     = trim(buffer)//derVarList(iVar)%variable%label, &
          &       nComponents = derVarList(iVar)%variable%nComponents,         &
          &       pos         = pos,                                           &
          &       depVar      = trim(buffer)//derVarList(iVar)%variable%depVars)

        ! check if the variable has been added (only appended if requirements are
        ! fullfilled)
        if(pos .eq. 0 )then
          write( buffer_depVars,*) trim(derVarList(iVar)%variable%depVars(1))
          call tem_write('------------------------------------------------')
          call tem_write('The following variable has not been appended to the list')
          call tem_write('label: '//trim(buffer)//                             &
            &             trim(derVarList(iVar)%variable%label))
          call tem_write('dep: '//trim(buffer)//trim(buffer_depVars))
          call tem_write('------------------------------------------------')
        ! check if the variable to append is already a state variable
        else if( pos .gt. var_sys%nVars )then
          var_sys%derive( pos )%do => derVarList( iVar )%do
          var_sys%variable%val( pos )%mapping = pos
        end if
      enddo ! loop labels

    end do ! loop derVarList

    ! Append space time variables to variable system
    do iVar = 1, size(tVar)
      if ( trim(tVar(iVar)%st_fun(1)%fun_kind) .ne. 'none') then
        call tem_var_append( sys = var_sys, var = tVar( iVar ), pos = pos)
        ! check if the variable has been added (only appended if requirements are
        ! fullfilled)
        if(pos .eq. 0 )then
          write( buffer_depVars,*) trim(tVar(iVar)%depVars(1))
          call tem_write('------------------------------------------------')
          call tem_write('The following variable has not been appended to '//  &
            &            'the list')
          call tem_write('label: '//trim(buffer)//trim(tVar(iVar)%label))
          call tem_write('dep: '//trim(buffer)//trim(buffer_depVars))
          call tem_write('------------------------------------------------')
        ! check if the variable to append is already a state variable
        else if( pos .gt. var_sys%nVars )then
          var_sys%derive(pos)%do => evalSpacetime
          var_sys%variable%val( pos )%mapping = pos
        end if
      endif
    enddo

    ! add diff_var variable in the end of the variable system since it is
    ! generic and does not need to be defined for each fields
    ! and it is need for all scheme kinds
    call tem_var_append( sys         = var_sys,                                &
      &                  varName     = diff_var%variable%label,                &
      &                  nComponents = diff_var%variable%nComponents,          &
      &                  pos         = pos )
    ! check if the variable to append is already a state variable
    if( pos .gt. var_sys%nVars )then
      var_sys%derive( pos )%do  => diff_var%do
    end if

    ! add the variables for mesh information
    call tem_initMeshInfoList( meshInfoList )
    do iVar = 1, size(meshInfoList)
      call tem_var_append( sys = var_sys, var = meshInfoList( iVar )%variable, &
        &                  pos = pos )
      ! check if the variable has been added (only appended if requirements are
      ! fullfilled)
      if(pos .eq. 0 )then
        write( buffer_depVars,*) trim(meshInfoList(iVar)%variable%depVars(1))
        call tem_write('------------------------------------------------')
        call tem_write('The following variable has not been appended to '//    &
          &            'the list')
        call tem_write('label: '//trim(buffer)//                               &
          &            trim(meshInfoList(iVar)%variable%label))
        call tem_write('dep: '//trim(buffer)//trim(buffer_depVars))
        call tem_write('------------------------------------------------')
      ! check if the variable to append is already a state variable
      else if( pos .gt. var_sys%nVars )then
        var_sys%derive( pos )%do => meshInfoList( iVar )%do
        var_sys%variable%val( pos )%mapping = pos
      end if
    enddo


  end subroutine mus_init_varSys
!******************************************************************************!


end module mus_variable_module
!******************************************************************************!
