! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module contains the datatypes for the geometry increase table within the
!! lua configuration file. It also contains the routines to load various
!! variables, parameters from the lua file which are then used in performing
!! geometry changes
!!
!! \author Kartik Jain
!!
module mus_geomIncrHead_module

  ! include treelm modules
  use tem_timeControl_module, only: tem_timeControl_type, tem_load_timeControl
  use tem_tools_module,       only: tem_write, tem_horizontalSpacer
  use tem_depend_module,      only: tem_load_depend, tem_depend_type

  ! include aotus modules
  use flu_binding,       only: flu_State
  use aot_table_module,  only: aot_get_val, aot_table_open, aot_table_close,   &
    &                          aot_table_length
  use aot_vector_module, only: aot_get_val

  ! include musubi modules
  use mus_scheme_type_module, only: mus_scheme_type
  use mus_global_module,      only: param_type

  implicit none
  private

  public :: mus_geomIncrHead_type
  public :: mus_geomIncrHead_load

  type mus_geomIncrHead_type
    logical :: active     = .false.  !< geometry increment active?
    logical :: solidify   = .false.  !< if true solidify, if false fluidify
    logical :: fluidify   = .false.
    logical :: proximity  = .false.  !< Check proximity condition if true
    type(tem_timeControl_type) :: time
    type(tem_depend_type), allocatable :: depend(:)
  end type mus_geomIncrHead_type

contains

!******************************************************************************!
  !> Read all the necessary information for the geometry increase from the lua
  !! config file. This routine basically provides as a wrapper to the routine
  !! which reads single values
  !!
  subroutine mus_geomIncrHead_load( me, params, schemes, conf, parent,         &
    &                               dynamicGeom)
    !---------------------------------------------------------------------------
    type( mus_geomIncrHead_type), allocatable, intent(inout) :: me(:)
    !> global parameter type
    type( param_type ), intent(in) :: params
    type( mus_scheme_type ), intent(in) :: schemes(:)
    type( flu_state ) :: conf
    integer, optional, intent(in) :: parent
    logical, intent(inout)        :: dynamicGeom
    !---------------------------------------------------------------------------
    integer :: tc_handle, sub_handle 
    integer :: nGeomIncrs 
    integer :: iGInc
    !---------------------------------------------------------------------------

    ! Attempt to open the geomIncr table (within another table, if a parent is
    ! given ) 
    if(present(parent))then
      call aot_table_open( L       = conf,                                     &
        &                  parent  = parent,                                   &
        &                  thandle = tc_handle,                                &
        &                  key     = 'geomIncr' )
    else
      call aot_table_open( L       = conf,                                     &
        &                  thandle = tc_handle,                                &
        &                  key     = 'geomIncr' )
    end if

    ! Check if geomIncr is actually defined
    if (tc_handle .ne. 0) then
      ! Set the dynamicGeom flag to true 
      dynamicGeom = .true.
      ! check whether there are other members inside geomIncr
      call aot_table_open( L       = conf,                                     &
        &                  parent  = tc_handle,                                &
        &                  thandle = sub_handle,                               &
        &                  pos     = 1)
      ! If there is only one member in geomIncr, call the load routine once
      if(sub_handle .eq. 0) then
        allocate(me(1))
        call aot_table_close( L = conf, thandle = sub_handle)
        call mus_geomIncrHead_load_single( me             = me(1),             &
          &                                refTimeControl = params%timeControl,&
          &                                schemes        = schemes,           &
          &                                conf           = conf,              &
          &                                thandle        = tc_handle )
      else
        ! multiple definitions inside geomIncr
        call aot_table_close( L = conf, thandle = sub_handle)
        nGeomIncrs = aot_table_length( L = conf, thandle = tc_handle)
        allocate(me(nGeomIncrs))

        ! Read the sub tables individually
        do iGInc = 1, nGeomIncrs
          call aot_table_open( L       = conf,                                 &
            &                  parent  = tc_handle,                            &
            &                  thandle = sub_handle,                           &
            &                  pos     = iGInc )
          call mus_geomIncrHead_load_single(                                   &
            &                           me             = me( iGInc ),          &
            &                           refTimeControl = params%timeControl,   &
            &                           schemes        = schemes,              &
            &                           conf           = conf,                 &
            &                           thandle        = sub_handle )
        end do
      end if
    end if

    call aot_table_close(L=conf, thandle=tc_handle)

  end subroutine mus_geomIncrHead_load
!******************************************************************************!


!******************************************************************************!
  !> Reads various parameters from the lua file defined for geometry increase
  !! This routine reads single values and is wrapped around in another function
  !! where it is called multiple times as required
  !!
  subroutine mus_geomIncrHead_load_single( me, refTimeControl, schemes, conf,  &
    &                                      thandle )
    !---------------------------------------------------------------------------
    type( mus_geomIncrHead_type),intent(inout)  :: me
    type( tem_timeControl_type ), intent(in) :: refTimeControl
    type( mus_scheme_type), intent(in) :: schemes(:)
    type( flu_state ), intent(in) :: conf
    integer, intent(in) :: thandle 
    !---------------------------------------------------------------------------
    integer :: iError
    !---------------------------------------------------------------------------
   
    me%active = .true.
    ! Read the solidify, fluidify and proximity flags
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               val     = me%solidify,                                   &
      &               ErrCode = iError,                                        &
      &               key     = 'solidify',                                    &
      &               default = .false.)

    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               val     = me%fluidify,                                   &
      &               ErrCode = iError,                                        &
      &               key     = 'fluidify',                                    &
      &               default = .false.)

    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               val     = me%proximity,                                  &
      &               ErrCode = iError,                                        &
      &               key     = 'proximity',                                   &
      &               default = .false. )

    call tem_load_depend( medepend = me%depend,                                &
      &                   conf     = conf,                                     &
      &                   parent   = thandle )

    ! load time control to perform geomIncr
    call tem_load_timeControl( conf           = conf,                          &
      &                        parent         = thandle,                       &
      &                        refTimeControl = reftimeControl,                &
      &                        me             = me%time )

    ! If geomIncr is activited, output its configuration to console
    if( me%active ) then
      call tem_write('Geometry Increment ACTIVATED! ')
      if( me%solidify ) then
        call tem_write('  Solidification is activited!')
      else
        call tem_write('  Fluidification is activited!')
      endif
      if( me%proximity) then
        call tem_write('  Proximity is activited.')
      else
        call tem_write('  Proximity is activited.')
      endif
      call tem_horizontalSpacer()
    endif

  end subroutine mus_geomIncrHead_load_single
!******************************************************************************!


end module mus_geomIncrHead_module
!******************************************************************************!
