! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module contains the definition of geometry type

!! \author Kartik Jain

module mus_geom_module

  ! include treelm modules
  use treelmesh_module,        only: treelmesh_type
  use env_module,              only: rk, long_k, globalMaxLevels
  use tem_balance_module,      only: tem_balance_type
  use tem_bc_prop_module,      only: tem_bc_prop_type
  use tem_aux_module,          only: tem_abort
  use tem_tools_module,        only: tem_write
  use tem_topology_module,     only: tem_levelOf
  use tem_construction_module, only: tem_levelDesc_type

  ! include musubi modules
  use mus_geomIncrHead_module, only: mus_geomIncrHead_type
  use mus_global_module,       only: param_type
  use mus_pdf_module,          only: pdf_globInfo_type, pdf_global_type
  
  implicit none
  private

  public mus_geom_type
  public mus_geom_setWeight
  public mus_geom_setLevelWeight
  public mus_geom_cleanWeight

  !> Geometric information and definitions
  type mus_geom_type
    !> tree data type
    type(treelmesh_type) :: tree
    !> boundary information as stored on disk
    type( tem_bc_prop_type )    :: boundary
    !> The header type containing all the geometry increase information
    type( mus_geomIncrHead_type), allocatable :: geomIncr(:)
    !> Logical to define whether geometry increase is active or not
    logical :: dynamicGeom = .false.
    ! todo: SZ: description is missing
    real(kind=rk), allocatable :: weight(:)
    integer(kind=long_k), allocatable :: offset(:)
    real(kind=rk) :: levelWeight(globalMaxLevels)
  end type mus_geom_type

contains

!******************************************************************************!
  !> set the weight for each level ( initial load balancing )
  !!
  subroutine mus_geom_setLevelWeight( params, geometry, balance, levelWeight )
    !---------------------------------------------------------------------------
    !> parameters
    type( param_type ), intent(inout)    :: params
    !> geometry information
    type( mus_geom_type ), intent(inout) :: geometry
    !> load balancing info
    type( tem_balance_type ), intent(in) :: balance
    !> weights for each level
    real(kind=rk), intent(out)           :: levelWeight(globalMaxLevels)
    !---------------------------------------------------------------------------
    integer :: iLevel
    !---------------------------------------------------------------------------
    ! Set level weight
    levelWeight = 1._rk
    if( trim( balance%balanceKind ) == 'none' ) then
      call tem_write(' Enforcing same number of elements on all processes.' )
    else
      call tem_write(' Initial load balancing with level weight: ',            &
        &             real(params%scaleFactor ))
      do iLevel = 1, globalMaxLevels
        levelWeight( iLevel ) = real(params%scaleFactor,kind=rk)**( iLevel ) 
      end do
    end if
  end subroutine mus_geom_setLevelWeight
!******************************************************************************!


!******************************************************************************!
  !> set weights for each element in the treeIDlist
  !!
  subroutine mus_geom_setWeight( params, geometry, balance, glob,         &
    &                            pdfGlobal )
    !---------------------------------------------------------------------------
    !> parameters
    type( param_type ),       intent(inout) :: params
    !> geometry information
    type( mus_geom_type ),    intent(inout) :: geometry
    !> load balancing information
    type( tem_balance_type ), intent(in)    :: balance
    !> Glob type 
    type(pdf_global_type), intent(in), optional      :: glob
    !> PdfGlobal type
    type( pdf_globInfo_type ), intent(in), optional  :: pdfGlobal
    !---------------------------------------------------------------------------
    integer :: iElem, iLevel, refLevel, level
    integer :: elemPos
    !---------------------------------------------------------------------------
    allocate( geometry%weight( geometry%tree%nElems ))
    refLevel = geometry%tree%global%minLevel

    call tem_write('Setting weights for load balancing...')
    select case( trim(balance%balanceKind ))
    case( 'simple' )
      call tem_write('  Using base weight:  ', real( params%scaleFactor ))
      do iElem = 1, geometry%tree%nElems
        level = tem_levelOf( geometry%tree%treeID( iElem ))
        ! set weight to fac^( level - minlevel ) 
        !                 acoustic      diffusive
        !  ------------------------------------------------
        !  fac =         |   2              4
        !  ------------------------------------------------
        !  minLevel      |   1              1
        !  minLevel + 1  |   2              4
        !  minLevel + 2  |   4              16
        geometry%weight( iElem ) =                                             & 
          &              real( params%scaleFactor, kind=rk )**( level-refLevel )
      enddo

    ! Setting level wise weights for dynamic load balancing
    case('levelwise')
      call tem_write('Using levelwise weights'  )
      do iElem = 1, geometry%tree%nElems
        level = tem_levelOf( geometry%tree%treeID( iElem ))
        geometry%weight( iElem ) = params%level_weight( level )
      enddo
    ! As a special case, we treat the boundary elements with additional
    ! weights. The boundary computations especially in the case of complex
    ! higher order boundaries is expensive and causes load imbalance. Weights
    ! which were computed before are just added to elements which are
    ! boundary. GLOB AND PDFGLOBAL HAVE TO BE PRESENT FOR THIS OPERATION 
    case('boundary')
      call tem_write('Using levelwise weights and boundary weights')
      ! First set the level weights for all elements
      do iElem = 1, geometry%tree%nElems
        level = tem_levelOf( geometry%tree%treeID( iElem ))
        geometry%weight( iElem ) = params%level_weight( level )
      enddo
      ! Set the additional weights for boundary elements
      if( present(glob) .and. present(pdfGlobal) ) then
        do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
          if( glob%levelDesc(iLevel)%bc_elemBuffer%nVals .gt. 0 ) then
            do iElem = 1, glob%levelDesc(iLevel)%bc_elemBuffer%nVals 
              elemPos = glob%levelDesc( iLevel )%bc_elemBuffer%val( iElem )
              geometry%weight( elemPos ) = params%level_weight_bnd( iLevel )
            end do
          end if ! nBCs>0
        end do ! iLevel
      end if ! IF present

    case default
      call tem_write( ' Given load balancing not found in Musubi'//            &
        &             trim(balance%balanceKind ))
      call tem_abort
    end select

  end subroutine mus_geom_setWeight
!******************************************************************************!


!******************************************************************************!
  !> delete weight array
  !!
  subroutine mus_geom_cleanWeight( geometry )
    !---------------------------------------------------------------------------
    !> geometry information
    type( mus_geom_type ), intent(inout)  :: geometry
    !---------------------------------------------------------------------------
    deallocate( geometry%weight)

  end subroutine mus_geom_cleanWeight
!******************************************************************************!


end module mus_geom_module
!******************************************************************************!
