! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
?? include 'header/lbm_macros.inc'
!> \brief Interpolation of flow quantities between different grid levels
!!
!! Ghost elements are employed at grid level interfaces to provide valid
!! pdf values to the neighboring fluid elements. This way, the solvers can 
!! act on elements of the same size only, treating the levels successively.
!! Target elements are the ghost elements, which have to be filled with 
!! valid values.
!! Source elements are the fluid elements from other levels, from where to 
!! take the input values for the interpolation.
!! The target ghost elements on the target level have corresponding source 
!! fluid elements on the source level. 
!!
!! \ref For a detailed description of the grid 
!!
module mus_interpolate_compact_module

  ! include treelm modules
  use env_module,            only: rk
  use tem_element_module,    only: eT_GhostFromCoarserElem,                    &
    &                              eT_ghostFromFinerElem
  use tem_geometry_module,   only: tem_intp_bilinear
  use tem_grow_array_module, only: grw_intArray_type
  use tem_var_system_module, only: tem_var_system_type
  use tem_logging_module,    only: tem_log, tem_toStr

  ! include musubi modules
  use mus_field_prop_module,         only: mus_field_prop_type
  use tem_param_module,              only: cs2inv, cs2
  use mus_pdf_module,                only: pdf_data_type, pdf_global_type
  use mus_scheme_layout_module,      only: mus_scheme_layout_type
  use mus_scheme_type_module,        only: mus_scheme_type
  use tem_math_module,               only: tem_intp_trilinearReduced
  use mus_interpolate_header_module, only: interpolation_type,                 &
    &                                      determine_timeLayer
  use mus_variable_module,           only: mus_derVarPos_type

  implicit none

  private

  public :: unitTest_interpolate_quadraticCompact
  public :: fillMyGhostsFromFiner_quadraticCompact2d
  public :: fillFinerGhostsFromMe_quadraticCompact2d
  public :: fillMyGhostsFromFiner_quadraticCompact3d
  public :: fillFinerGhostsFromMe_quadraticCompact3d
  
  integer, parameter :: QmomentumX = 1
  integer, parameter :: QmomentumY = 2
  integer, parameter :: QmomentumZ = 3
  integer, parameter :: QPxx       = 4
  integer, parameter :: QPyy       = 5
  integer, parameter :: QPzz       = 6
  integer, parameter :: QPxy       = 7
  integer, parameter :: QPyz       = 8
  integer, parameter :: QPzx       = 9
  integer, parameter :: c_x  = 1 
  integer, parameter :: c_y  = 2 
  integer, parameter :: c_z  = 3 
  
 contains

!******************************************************************************!
  !> Interpolation with momentum gradients from the stress tensor
  !! source points to the target position located at targetCoord. 
  !! The source points are arranged in a square from (0,0,0)x(1,1,1)
  !! The order of the source points are according to the morton curve
  !! See tem_param for the explicit ordering
  !!
  pure function mus_interpolate_quadraticCompact3d_leastSq( srcMom,            &
    &                            targetCoord, Pab, nSourceElems, matrix )      &
    &                            result( phi )
    !---------------------------------------------------------------------------
    !>
    integer, intent(in) :: nSourceElems
    !> source values of the momentum on the square corners
    real(kind=rk), intent(in) :: srcMom(3,nSourceElems)
    !>
    real(kind=rk), intent(in) :: targetCoord(3)
    !> source values of the square corners
    real(kind=rk), intent(in) :: Pab(6,nSourceElems)
    !> relaxation time for the non-eq moments
    real(kind=rk), intent(in) :: matrix(:,:)
    !> result: interpolated value
    real(kind=rk) :: phi(3)
    !---------------------------------------------------------------------------
    ! Coefficients
    real(kind=rk) :: a(10), b(10), c(10)
    ! right and left hand side of the equation system
    real(kind=rk) :: rhs(9*nSourceElems), lhs(30)
    !---------------------------------------------------------------------------

    ! We extract momentum information completely on the view of the source
    ! coordinate system. Set the right hand side of the equation system
    rhs( 0*nSourceElems+1:1*nSourceElems)  = srcMom( 1, 1:nSourceElems)
    rhs( 1*nSourceElems+1:2*nSourceElems)  = srcMom( 2, 1:nSourceElems)
    rhs( 2*nSourceElems+1:3*nSourceElems)  = srcMom( 3, 1:nSourceElems)
    rhs( 3*nSourceElems+1:4*nSourceElems)  =    Pab( 1, 1:nSourceElems)
    rhs( 4*nSourceElems+1:5*nSourceElems)  =    Pab( 2, 1:nSourceElems)
    rhs( 5*nSourceElems+1:6*nSourceElems)  =    Pab( 3, 1:nSourceElems)
    rhs( 6*nSourceElems+1:7*nSourceElems)  =    Pab( 4, 1:nSourceElems)
    rhs( 7*nSourceElems+1:8*nSourceElems)  =    Pab( 5, 1:nSourceElems)
    rhs( 8*nSourceElems+1:9*nSourceElems)  =    Pab( 6, 1:nSourceElems)
    ! Solve the problem, where b = rhs, x = coefficients
    ! A*x = b
    ! overdetermined, solve the least Square fit problem
    ! (A^T)A*x = (A^T)b
    ! x = ((A^T)A)^-1*(A^T)b
    ! Solve linear system of equation with inverted matrix
    lhs = matmul( matrix, rhs )
 
    a(1:10) = lhs( 1:10)
    b(1:10) = lhs(11:20)
    c(1:10) = lhs(21:30)
    ! Evaluate the bubble function with the above calculated coefficients
    ! m_pi(x) = a1+a2*xcoord+a3*ycoord+a4*zcoord 
    !             +a5*x^2 + a6*y^2 + a7*z^2
    !             +a8*x*y + a9*y*z + a10*x*z
    phi(1) = a(1) + a( 2)*targetCoord(c_x)                                     &
                  + a( 3)*targetCoord(c_y)                                     &
                  + a( 4)*targetCoord(c_z)                                     &
                  + a( 5)*targetCoord(c_x)*targetCoord(c_x)                    &
                  + a( 6)*targetCoord(c_y)*targetCoord(c_y)                    &
                  + a( 7)*targetCoord(c_z)*targetCoord(c_z)                    &
                  + a( 8)*targetCoord(c_x)*targetCoord(c_y)                    &
                  + a( 9)*targetCoord(c_y)*targetCoord(c_z)                    &
                  + a(10)*targetCoord(c_z)*targetCoord(c_x)                  
    phi(2) = b(1) + b( 2)*targetCoord(c_x)                                     &
                  + b( 3)*targetCoord(c_y)                                     &
                  + b( 4)*targetCoord(c_z)                                     &
                  + b( 5)*targetCoord(c_x)*targetCoord(c_x)                    &
                  + b( 6)*targetCoord(c_y)*targetCoord(c_y)                    &
                  + b( 7)*targetCoord(c_z)*targetCoord(c_z)                    &
                  + b( 8)*targetCoord(c_x)*targetCoord(c_y)                    &
                  + b( 9)*targetCoord(c_y)*targetCoord(c_z)                    &
                  + b(10)*targetCoord(c_z)*targetCoord(c_x)                  
    phi(3) = c(1) + c( 2)*targetCoord(c_x)                                     &
                  + c( 3)*targetCoord(c_y)                                     &
                  + c( 4)*targetCoord(c_z)                                     &
                  + c( 5)*targetCoord(c_x)*targetCoord(c_x)                    &
                  + c( 6)*targetCoord(c_y)*targetCoord(c_y)                    &
                  + c( 7)*targetCoord(c_z)*targetCoord(c_z)                    &
                  + c( 8)*targetCoord(c_x)*targetCoord(c_y)                    &
                  + c( 9)*targetCoord(c_y)*targetCoord(c_z)                    &
                  + c(10)*targetCoord(c_z)*targetCoord(c_x)                  

  end function mus_interpolate_quadraticCompact3d_leastSq
!******************************************************************************!


!******************************************************************************!
  !> Interpolation with momentum gradients from the stress tensor
  !! source points to the target position located at targetCoord. 
  !! The source points are arranged in a square from (0,0)x(1,1)
  !! The order of the source points are according to the morton curve
  !!  1      2      3      4
  !! (0,0); (1,0); (0,1); (1,1)
  !! 
  pure function mus_interpolate_quadraticCompact2d_leastSq( srcMom,            &
    &                                targetCoord, Pab, nSourceElems, matrix )  &
    &                                result( phi )
    !---------------------------------------------------------------------------
    !>
    integer, intent(in) :: nSourceElems          
    !> source values of the momentum on the square corners
    real(kind=rk), intent(in) :: srcMom(2,nSourceElems)
    !> interpolation location within the square
    real(kind=rk), intent(in) :: targetCoord(2)
    !> source values of the square corners
    real(kind=rk), intent(in) :: Pab(3,nSourceElems)
    !> relaxation time for the non-eq moments
    real(kind=rk), intent(in) :: matrix(:,:)
    !> interpolated value
    real(kind=rk) :: phi(2)
    !---------------------------------------------------------------------------
    real(kind=rk) :: a(6), b(6) ! Coefficients
    real(kind=rk) :: rhs(5*nSourceElems), lhs(12) ! Coefficients
    !---------------------------------------------------------------------------
    ! We extract momentum information completely on the view of the source
    ! coordinate system. Set the right hand side of the equation system
    rhs( 0*nSourceElems+1:1*nSourceElems)  = srcMom( 1, 1:nSourceElems)
    rhs( 1*nSourceElems+1:2*nSourceElems)  = srcMom( 2, 1:nSourceElems)
    rhs( 2*nSourceElems+1:3*nSourceElems)  =    Pab( 1, 1:nSourceElems)
    rhs( 3*nSourceElems+1:4*nSourceElems)  =    Pab( 2, 1:nSourceElems)
    rhs( 4*nSourceElems+1:5*nSourceElems)  =    Pab( 3, 1:nSourceElems)

    ! Solve the problem, where b = rhs, x = coefficients
    ! A*x = b
    ! overdetermined, solve the least Square fit problem
    ! (A^T)A*x = (A^T)b
    ! x = ((A^T)A)^-1*(A^T)b
    ! Solve linear system of equation with inverted matrix
    lhs = matmul( matrix, rhs )
 
    a(1:6) = lhs(1:6)
    b(1:6) = lhs(7:12)
    ! Evaluate the bubble function with the above calculated coefficients
    ! m_pi(x) = a1+a2*xcoord+a3*ycoord+a4*xcoord*xcoord+a5*ycoord*ycoord+a6*xcoord*ycoord;
    phi(1) = a(1) + a(2)*targetCoord(c_x) + a(3)*targetCoord(c_y)              &
                  + a(4)*targetCoord(c_x)*targetCoord(c_x)                     &
                  + a(5)*targetCoord(c_y)*targetCoord(c_y)                     &
                  + a(6)*targetCoord(c_x)*targetCoord(c_y)

    ! m_pi(y) = b1+b2*xcoord+b3*ycoord+b4*xcoord*xcoord+b5*ycoord*ycoord+b6*xcoord*ycoord;
    phi(2) = b(1) + b(2)*targetCoord(c_x) + b(3)*targetCoord(c_y)              &
                  + b(4)*targetCoord(c_x)*targetCoord(c_x)                     &
                  + b(5)*targetCoord(c_y)*targetCoord(c_y)                     &
                  + b(6)*targetCoord(c_x)*targetCoord(c_y)

  end function mus_interpolate_quadraticCompact2d_leastSq
!******************************************************************************!


!******************************************************************************!
  !>  
  !!  
  subroutine fillFinerGhostsFromMe_quadraticCompact3d( pdf, glob, fieldProp,   &
    &                    level, scheme, layout, intp, ind, varSys, derVarPos ) 
    !---------------------------------------------------------------------------
    !> global fluid parameters 
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !> global parameters 
    type( pdf_global_type ), intent(inout) :: glob
    !> the current scheme 
    type( mus_scheme_type ), intent(in)   :: scheme
    !> fluid and species parameters
    type( mus_field_prop_type ), intent(in) ::  fieldProp(:)
    !> level
    integer, intent(in)   :: level
    !> the layout used
    type( mus_scheme_layout_type), intent(in)   :: layout
    !> interpolation method info
    type( interpolation_type ), intent(in)   :: intp
    !> indirection list
    type( grw_intArray_type ), intent(in)   :: ind
    !> global variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> required variable system which maps to global varsys
    type( mus_derVarPos_type ), intent(in) :: derVarPos(:)
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iVal           ! value counter
    integer :: iElem          ! element counter for outer loop
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    real(kind=rk) :: targetPdf( layout%stencil(1)%QQ ) ! pdf to reconstruct
    real(kind=rk) :: sourcePdf( layout%stencil(1)%QQ ) ! pdf to reconstruct from
    ! moments of the source elements' pdf
    real(kind=rk) :: sourceMom( layout%stencil(1)%QQ, intp%nMaxSources )
    ! interpolated moments
    real(kind=rk) :: targetMom( layout%stencil(1)%QQ )
    real(kind=rk) :: targetCoord(3) ! coordinates to interpolate to
    ! Coordinates in rotated reference cube
    real(kind=rk), parameter :: targetCoord_trafo(1:3) =                       &
      &                                          (/0.25_rk, 0.25_rk, 0.25_rk /)
    ! shear stress scaling factor
    real(kind=rk) :: omfac
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to write to
    integer :: t_source ! which time layer to read from
    real(kind=rk) :: sourceEq( layout%stencil(1)%QQ ) ! source equilirbium part 
    real(kind=rk) :: tMom( layout%stencil(1)%QQ )  ! temp moment calculation
    integer :: iCase ! child case for rotation of the unit cube
    !---------------------------------------------------------------------------

    sourceLevel = level
    targetLevel = level + 1
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))

    ! First calculate all the required moments for all the source elements
    do iSourceElem = 1, glob%levelDesc( targetLevel )%sourceFromCoarser%nVals
      ! Get the source element's position in the state vector
      sourceElem = glob%levelDesc( targetLevel )%                              &
        &                                  sourceFromCoarser%val( iSourceElem )
      iVar = 1
      do iVal = 1, layout%stencil(1)%QQ
        sourcePdf( iVal )  = pdf( sourceLevel )%state(                         &
  &  ?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh ),&
  &      t_source  )
      enddo
      ! Calculate the raw moments form the pdf
      tMom( : ) = matmul( layout%moment%toMoments%A, sourcePdf)

      ! And calculate the equilibrium distribution from the macroscopic 
      ! quantities in tMom
      sourceEq = getEquilibrium( tMom( 1),    & ! density
                      (/ tMom( 2)/tMom(1),    & ! velX
                         tMom( 3)/tMom(1),    & ! velY
                         tMom( 4)/tMom(1)/),  & ! velZ
                         layout )

      ! remove pi(0) parts from the second order moment
      ! that is ((rho u_\alpha * rho u_\beta) / rho)
      ! and additionally rho*cs2 from diagonal entries
      ! Pab = sum( c_ia*c_ib *f_i ) - kronecker_ab*(rho*cs2) - rho*u_a*u_b
      tMom( 5 ) = tMom( 5 ) - cs2*tMom( 1) & !Pxx
                   - tMom( 2)*tMom( 2)/tMom( 1)
      tMom( 6 ) = tMom( 6 ) - cs2*tMom( 1) & !Pyy
                   - tMom( 3)*tMom( 3)/tMom( 1)
      tMom( 7 ) = tMom( 7 ) - cs2*tMom( 1) & !Pzz
                   - tMom( 4)*tMom( 4)/tMom( 1)
      tMom( 8 ) = tMom( 8 )                & !Pxy
                   - tMom( 2)*tMom( 3)/tMom( 1)
      tMom( 9 ) = tMom( 9 )                & !Pyz
                   - tMom( 3)*tMom( 4)/tMom( 1)
      tMom(10 ) = tMom(10 )                & !Pzx
                   - tMom( 2)*tMom( 4)/tMom( 1)
      ! Transfer momentum to velocity
      tMom( 2:4 ) = tMom( 2:4 )/ tMom( 1 )

      ! Set the further higher moments to their equilibrium distribution
      tMom(11 ) = get_moment( sourceEq, layout, (/2,1,0/) ) !mxxy
      tMom(12 ) = get_moment( sourceEq, layout, (/2,0,1/) ) !mxxz
      tMom(13 ) = get_moment( sourceEq, layout, (/1,2,0/) ) !myyx
      tMom(14 ) = get_moment( sourceEq, layout, (/0,2,1/) ) !myyz
      tMom(15 ) = get_moment( sourceEq, layout, (/1,0,2/) ) !mzzx
      tMom(16 ) = get_moment( sourceEq, layout, (/0,1,2/) ) !mzzy
      tMom(17 ) = get_moment( sourceEq, layout, (/2,2,0/) ) !mxxyy
      tMom(18 ) = get_moment( sourceEq, layout, (/0,2,2/) ) !myyzz
      tMom(19 ) = get_moment( sourceEq, layout, (/2,0,2/) ) !mzzxx      
      ! Store all the moments to the current source element's position
      ! in the buffer
      glob%levelDesc( targetLevel )%intpBufFromCoarser( :, iSourceElem ) = tMom
    enddo
 

    ! Treat all fine target elements:
    !do iElem = 1, pdf( targetLevel )%nElems_ghostFromCoarser
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      ! Read the target element treeId
      targetElem = iElem + glob%levelDesc( targetLevel )%offset( 1, eT_ghostFromCoarserElem )
      ! get the target coordinates 
      targetCoord = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%coord

      ! Find out how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      nSourceElems = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%elem%nVals

      !Compute rho and velocity for each field and interpolate pdf field wise
      do iVar = 1, varSys%nVars
        omfac = (1._rk - 0.5*fieldProp( iVar )%fluid%omLvl( sourceLevel ))     &
          &   / (1._rk - 0.5*fieldProp( iVar )%fluid%omLvl( targetLevel )) 

        ! Now loop over all fine source elements for this target:
        do iSourceElem = 1, nSourceElems
          ! Get the source element in the levelwise buffer
          sourceElem = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%  &
            &                                     elemBuffer%val( iSourceElem )
          ! and read its entries
          sourceMom(:, iSourceElem ) = glob%levelDesc( targetLevel )%          &
            &                               intpBufFromCoarser( :, sourceElem )
        end do   
        ! interpolate the density linearly 
        targetMom(1)  = tem_intp_trilinearReduced(                             &
          &                    sourceMom(1,1:nSourceElems), targetCoord_trafo )
        
        ! get the rotation case from the child number of the target element
        iCase = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%childNum

        targetMom(2:4) = mus_interpolate_quadraticCompact3d_leastSq(           &
          &             sourceMom(2:4, 1:nSourceElems),                        &
          &             targetCoord, sourceMom(5:10, 1:nSourceElems),          &
          &             nSourceElems,                                          &
          &             intp%fromCoarser( targetLevel )%matrixInv( iCase )%A )
        !do iVal = 2,4
        !targetMom(iVal) = tem_intp_trilinearReduced( sourceMom(iVal,1:nSourceElems), &
        !  &                                  targetCoord_trafo)
        !enddo

        ! linear interpolation for higher moments
        do iVal = 5,19
          targetMom(iVal) = tem_intp_trilinearReduced( sourceMom(              &
            &                         iVal,1:nSourceElems), targetCoord_trafo )
        enddo

        ! Rescale the shear stress moments
        targetMom(5:10) = targetMom(5:10)*omfac

        ! and add back the pi(0) part to the tensor
        ! No need to rescale higher order moments as they were set
        ! to equilibrium
        targetMom( 5 ) = targetMom( 5 ) + cs2*targetMom( 1) & !Pxx
                     + targetMom( 2)*targetMom( 2)/targetMom( 1)
        targetMom( 6 ) = targetMom( 6 ) + cs2*targetMom( 1) & !Pyy
                     + targetMom( 3)*targetMom( 3)/targetMom( 1)
        targetMom( 7 ) = targetMom( 7 ) + cs2*targetMom( 1) & !Pzz
                     + targetMom( 4)*targetMom( 4)/targetMom( 1)
        targetMom( 8 ) = targetMom( 8 )                     & !Pxy
                     + targetMom( 2)*targetMom( 3)/targetMom( 1)
        targetMom( 9 ) = targetMom( 9 )                     & !Pyz
                     + targetMom( 3)*targetMom( 4)/targetMom( 1)
        targetMom(10 ) = targetMom(10 )                     & !Pzx
                     + targetMom( 2)*targetMom( 4)/targetMom( 1)

        ! Transfer back velocity field to momentum field
        targetMom(2:4) = targetMom(2:4)*targetMom(1)

        ! transform the moments back to the pdfs
        targetPdf = matmul( layout%moment%toPdf%A, targetMom )

        ! Now write the resulting pdf in the current direction to the target
        ! Element position
        do iVal = 1, layout%stencil(1)%QQ
        pdf( targetLevel )%state( &
         &?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), targetElem, varSys%nScalars, pdf(targetLevel)%neigh ),&
         & t_target  ) &
         & = targetPdf(iVal)
         enddo
      enddo ! iVar loop
    enddo

  end subroutine fillFinerGhostsFromMe_quadraticCompact3d
!******************************************************************************!


!******************************************************************************!
  !> 
  !! 
  subroutine fillMyGhostsFromFiner_quadraticCompact3d( pdf, glob, fieldProp,   &
    &                     level, scheme, layout, intp, ind, varSys, derVarPos ) 
    !---------------------------------------------------------------------------
    !> global fluid parameters 
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !> global parameters 
    type( pdf_global_type ), intent(inout) :: glob
    !> the current scheme 
    type( mus_scheme_type ), intent(in)   :: scheme
    !> fluid and species parameters 
    type( mus_field_prop_type ), intent(in) ::  fieldProp(:)
    !> level
    integer, intent(in)   :: level
    !> the layout used
    type( mus_scheme_layout_type), intent(in)   :: layout
    !> interpolation method info
    type( interpolation_type ), intent(in)   :: intp
    !> indirection list
    type( grw_intArray_type ), intent(in)   :: ind
    !> global variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> required variable system which maps to global varsys
    type( mus_derVarPos_type ), intent(in) :: derVarPos(:)
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iVal           ! value counter
    integer :: iElem          ! element counter for outer loop
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    real(kind=rk) :: targetPdf( layout%stencil(1)%QQ ) ! pdf to reconstruct
    real(kind=rk) :: sourcePdf( layout%stencil(1)%QQ ) ! pdf to reconstruct from
    ! moments of the source elements' pdf
    real(kind=rk) :: sourceMom( layout%stencil(1)%QQ, intp%nMaxSources )
    ! interpolated moments
    real(kind=rk) :: targetMom( layout%stencil(1)%QQ )
    ! coordinates to interpolate to. always fixed for ghost-from-finer
    real(kind=rk), parameter :: targetCoord(3)  = (/ 0.5_rk, 0.5_rk, 0.5_rk /)
    ! shear stress scaling factor
    real(kind=rk) :: omfac
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to write to
    integer :: t_source ! which time layer to read from
    real(kind=rk) :: sourceEq( layout%stencil(1)%QQ ) ! source equilirbium part
    real(kind=rk) :: tMom( layout%stencil(1)%QQ )  ! temp moment calculation
    integer :: iCase ! child case for rotation of the unit cube
    !---------------------------------------------------------------------------

    targetLevel = level
    sourceLevel = level+1
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))
    

    ! First calculate all the required moments for all the source elements
    do iSourceElem = 1, glob%levelDesc( targetLevel )%sourceFromFiner%nVals
      ! Get the source element's position in the state vector
      sourceElem = glob%levelDesc( targetLevel )%sourceFromFiner%              &
        &                                                    val( iSourceElem )
      iVar = 1
      do iVal = 1, layout%stencil(1)%QQ
        sourcePdf( iVal )  = pdf( sourceLevel )%state( &
  &  ?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh ),&
  &      t_source  )
      enddo
      ! Calculate the raw moments form the pdf
      tMom( : ) = matmul( layout%moment%toMoments%A, sourcePdf)

      ! And calculate the equilibrium distribution from the macroscopic 
      ! quantities in tMom
      sourceEq = getEquilibrium( tMom( 1),    & ! density
                      (/ tMom( 2)/tMom(1),    & ! velX
                         tMom( 3)/tMom(1),    & ! velY
                         tMom( 4)/tMom(1)/),  & ! velZ
                         layout )

      ! remove pi(0) parts from the second order moment
      ! that is ((rho u_\alpha * rho u_\beta) / rho)
      ! and additionally rho*cs2 from diagonal entries
      ! Pab = sum( c_ia*c_ib *f_i ) - kronecker_ab*(rho*cs2) - rho*u_a*u_b
      tMom( 5 ) = tMom( 5 ) - cs2*tMom( 1) & !Pxx
                   - tMom( 2)*tMom( 2)/tMom( 1)
      tMom( 6 ) = tMom( 6 ) - cs2*tMom( 1) & !Pyy
                   - tMom( 3)*tMom( 3)/tMom( 1)
      tMom( 7 ) = tMom( 7 ) - cs2*tMom( 1) & !Pzz
                   - tMom( 4)*tMom( 4)/tMom( 1)
      tMom( 8 ) = tMom( 8 )                & !Pxy
                   - tMom( 2)*tMom( 3)/tMom( 1)
      tMom( 9 ) = tMom( 9 )                & !Pyz
                   - tMom( 3)*tMom( 4)/tMom( 1)
      tMom(10 ) = tMom(10 )                & !Pzx
                   - tMom( 2)*tMom( 4)/tMom( 1)
      ! Transfer momentum to velocity
      tMom( 2:4 ) = tMom( 2:4 )/ tMom( 1 )

      ! Set the further higher moments to their equilibrium distribution
      tMom(11 ) = get_moment( sourceEq, layout, (/2,1,0/) ) !mxxy
      tMom(12 ) = get_moment( sourceEq, layout, (/2,0,1/) ) !mxxz
      tMom(13 ) = get_moment( sourceEq, layout, (/1,2,0/) ) !myyx
      tMom(14 ) = get_moment( sourceEq, layout, (/0,2,1/) ) !myyz
      tMom(15 ) = get_moment( sourceEq, layout, (/1,0,2/) ) !mzzx
      tMom(16 ) = get_moment( sourceEq, layout, (/0,1,2/) ) !mzzy
      tMom(17 ) = get_moment( sourceEq, layout, (/2,2,0/) ) !mxxyy
      tMom(18 ) = get_moment( sourceEq, layout, (/0,2,2/) ) !myyzz
      tMom(19 ) = get_moment( sourceEq, layout, (/2,0,2/) ) !mzzxx      
      ! Store all the moments to the current source element's position in
      ! the buffer
      glob%levelDesc( targetLevel )%intpBufFromFiner( :, iSourceElem ) = tMom
    enddo
 
    ! Treat all coarse target elements:
    !do iElem = 1, pdf( level )%nElems_ghostFromFiner   
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      ! Read the target element treeId
      targetElem = iElem + glob%levelDesc( level )%                            &
        &                                    offset( 1, eT_ghostFromFinerElem )
      ! Find out how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      ! We take the compute stencil as a reference and omit the (1,1,1)
      ! corner element
      nSourceElems = glob%levelDesc( targetLevel )%                            &
        &                                       depFromFiner( iElem)%elem%nVals
       
      ! Compute sourceRho and velocity for each field and interpolate pdf
      ! field wise
      do iVar = 1, varSys%nVars

        omfac = (1._rk - 0.5*fieldProp( iVar )%fluid%omLvl( sourceLevel ))     &
          &   / (1._rk - 0.5*fieldProp( iVar )%fluid%omLvl( targetLevel ))
        ! Now loop over all fine source elements for this target:
        do iSourceElem = 1, nSourceElems
          ! Get the source element in the levelwise buffer
          sourceElem = glob%levelDesc( targetLevel )%depFromFiner( iElem )%    &
            &                                     elemBuffer%val( iSourceElem )
          ! and read its entries
          sourceMom(:, iSourceElem ) = glob%levelDesc( targetLevel )%          &
            &                                 intpBufFromFiner( :, sourceElem )
        enddo
        ! interpolate the density linearly 
        targetMom(1) = sum(sourceMom(1,1:nSourceElems))                        &
          &          / real(nSourceElems,kind=rk)
        ! the source element enumeration always is the case for childNum 4 
        iCase = 8 ! In the last case, the elements should always be numbered
        ! according to the morton curve
        
        ! quadratic interpolation for velX, velY, velZ
        targetMom(2:4) = mus_interpolate_quadraticCompact3d_leastSq(           &
          &                 sourceMom(2:4, 1:nSourceElems),                    &
          &                 targetCoord, sourceMom(5:10, 1:nSourceElems),      &
          &                 nSourceElems,                                      &
          &                 intp%fromFiner( targetLevel )%matrixInv( iCase )%A )
        ! linear interpolation for higher moments
        do iVal = 5,19
          targetMom(iVal) = sum(sourceMom(iVal,1:nSourceElems))                &
            &             / real(nSourceElems,kind=rk)
        enddo

        ! shear stress part has to be scaled
        targetMom(5:10) = targetMom(5:10)*omfac
        ! and add back the pi(0) part to the tensor
        ! No need to rescale higher order moments as they were set to
        ! equilibrium
        targetMom( 5 ) = targetMom( 5 ) + cs2*targetMom( 1) & !Pxx
                     + targetMom( 2)*targetMom( 2)/targetMom( 1)
        targetMom( 6 ) = targetMom( 6 ) + cs2*targetMom( 1) & !Pyy
                     + targetMom( 3)*targetMom( 3)/targetMom( 1)
        targetMom( 7 ) = targetMom( 7 ) + cs2*targetMom( 1) & !Pzz
                     + targetMom( 4)*targetMom( 4)/targetMom( 1)
        targetMom( 8 ) = targetMom( 8 )                     & !Pxy
                     + targetMom( 2)*targetMom( 3)/targetMom( 1)
        targetMom( 9 ) = targetMom( 9 )                     & !Pyz
                     + targetMom( 3)*targetMom( 4)/targetMom( 1)
        targetMom(10 ) = targetMom(10 )                     & !Pzx
                     + targetMom( 2)*targetMom( 4)/targetMom( 1)

        ! Transfer back velocity field to momentum field
        targetMom(2:4) = targetMom(2:4)*targetMom(1)

        ! transform the moments back to the pdfs
        targetPdf = matmul( layout%moment%toPdf%A, targetMom )

        ! Now write the resulting pdf in the current direction to the
        ! target Element position
        do iVal = 1, layout%stencil(1)%QQ
          pdf( targetLevel )%state( &
   &?SAVE?( varSys%Variable%val(iVar)%varPos( iVal ), targetElem, varSys%nScalars, pdf(targetLevel)%neigh ),&
         & t_target   ) &
         & = targetPdf( iVal ) 
        enddo ! ivar loop
      enddo ! ivar loop
    enddo

  end subroutine fillMyGhostsFromFiner_quadraticCompact3d
!******************************************************************************!


!******************************************************************************!
  !> 
  !! 
  subroutine fillMyGhostsFromFiner_quadraticCompact2d( pdf, glob, fieldProp,   &
    &                     level, scheme, layout, intp, ind, varSys, derVarPos ) 
    !---------------------------------------------------------------------------
    !> global fluid parameters 
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !> global parameters 
    type( pdf_global_type ), intent(inout) :: glob
    !> the current scheme 
    type( mus_scheme_type ), intent(in)   :: scheme
    !> fluid and species parameters 
    type( mus_field_prop_type ), intent(in) ::  fieldProp(:)
    !> level
    integer, intent(in)   :: level
    !> the layout used
    type( mus_scheme_layout_type), intent(in)   :: layout
    !> interpolation method info
    type( interpolation_type ), intent(in)   :: intp
    !> indirection list
    type( grw_intArray_type ), intent(in)   :: ind
    !> global variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> required variable system which maps to global varsys
    type( mus_derVarPos_type ), intent(in) :: derVarPos(:)
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iVal           ! value counter
    integer :: iElem          ! element counter for outer loop
    integer :: iDir           ! link counter for stencil loop
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    real(kind=rk) :: targetPdf( layout%stencil(1)%QQ ) ! pdf to reconstruct
    real(kind=rk) :: sourcePdf( layout%stencil(1)%QQ ) ! pdf to reconstruct from
    ! moments of the source elements' pdf
    real(kind=rk) :: sourceMom( layout%stencil(1)%QQ, intp%nMaxSources )
    ! interpolated moments
    real(kind=rk) :: targetMom( layout%stencil(1)%QQ )
    ! coordinates to interpolate to. always fixed for ghost-from-finer
    real(kind=rk), parameter :: targetCoord(3) = (/ 0.5_rk, 0.5_rk, 0.5_rk /)
    ! shear stress scaling factor
    real(kind=rk) :: omfac
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to write to
    integer :: t_source ! which time layer to read from
    real(kind=rk) :: sourceEq( layout%stencil(1)%QQ )  ! source equilirbium part
    integer :: iCase ! child case for rotation of the unit cube
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) usq, ucx ! square velocity
    !---------------------------------------------------------------------------

    targetLevel = level
    sourceLevel = level+1
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))
 
    ! Treat all coarse target elements:
    !do iElem = 1, pdf( level )%nElems_ghostFromFiner   
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      ! Read the target element treeId
      targetElem = iElem + glob%levelDesc( level )%                            &
        &                                    offset( 1, eT_ghostFromFinerElem )
      ! Find out how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      nSourceElems = glob%levelDesc( level )%depFromFiner( iElem )%elem%nVals
       
      ! Compute sourceRho and velocity for each field and interpolate pdf
      ! field wise
      do iVar = 1, varSys%nVars

        omfac = (1._rk - 0.5*fieldProp( iVar )%fluid%omLvl( sourceLevel ))     &
          &   / (1._rk - 0.5*fieldProp( iVar )%fluid%omLvl( targetLevel ))
        ! Now loop over all fine source elements for this target:
        do iSourceElem = 1, nSourceElems
          ! Get the source element's treeId
          sourceElem = glob%levelDesc( targetLevel )%depFromFiner( iElem )%    &
            &                                           elem%val( iSourceElem )
          !sourceElem = glob%levelDesc( targetLevel )%depFromFiner( iElem )%elem%val( iSourceElem )
          ! Read the pdf values from the state vector into sourcePdf
          do iVal = 1, layout%stencil(1)%QQ
            sourcePdf(iVal)  = pdf( sourceLevel )%state( &
   & ?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh ),&
         &    t_source )
          enddo
          ! Calculate the raw moments form the pdf
          sourceMom( :,iSourceElem ) = matmul( layout%moment%toMoments%A, sourcePdf)
!          ! And calculate the equilibrium distribution from the macroscopic 
!          ! quantities in sourceMom
!          input_loc = (/ sourceMom( 1, iSourceElem),    & ! density
!                       & sourceMom( 2, iSourceElem)/sourceMom(1, iSourceElem),   & ! velX
!                       & sourceMom( 3, iSourceElem)/sourceMom(1, iSourceElem),   & ! velY
!                       & 0.0_rk /) ! velZ
!
!          call varSys%derive( derVarPos(iVar)%equilFromMacro )%do(                         &
!            &              input = input_loc, nElems = 1, iVar = 1,                  &
!            &              requVar = varSys%Variable%val( derVarPos(iVar)%equilFromMacro ),&
!            &              globSys = varSys, stencil = layout%stencil(1),            &
!            &              res = sourceEq,                           &
!            &              conf = varSys%solSpec_conf )

          ! velX, velY
          u_x = sourceMom( 2, iSourceElem)/sourceMom(1, iSourceElem)
          u_y = sourceMom( 3, iSourceElem)/sourceMom(1, iSourceElem)

          ! square velocity and derived constants
          usq  = u_x*u_x + u_y*u_y

          do iDir = 1, layout%stencil(1)%QQ
            ! velocity times lattice unit velocity
            ucx = dble( layout%stencil(1)%cxDir( 1, iDir ) )*u_x               &
              & + dble( layout%stencil(1)%cxDir( 2, iDir ) )*u_y 
            ! calculate equilibrium density
            sourceEq( iDir ) = layout%weight( iDir )                           &
              &              * sourceMom(1, iSourceElem)                       &
              &              * ( 1._rk + ucx * cs2inv + ucx * ucx * cs2inv     &
              &              * cs2inv * 0.5_rk - usq * cs2inv * 0.5_rk )
          end do

          ! remove pi(0) parts from the second order moment
          ! that is ((rho u_\alpha * rho u_\beta) / rho)
          ! and additionally rho*cs2 from diagonal entries
          ! Pab = sum( c_ia*c_ib *f_i ) - kronecker_ab*(rho*cs2) - rho*u_a*u_b
          sourceMom( 4, iSourceElem ) = sourceMom( 4, iSourceElem )            &
            &          - cs2*sourceMom( 1, iSourceElem)                        &
            &          - sourceMom( 2, iSourceElem)*sourceMom( 2, iSourceElem) &
            &          / sourceMom( 1, iSourceElem)
          sourceMom( 5, iSourceElem ) = sourceMom( 5, iSourceElem )            &
            &          - cs2*sourceMom( 1, iSourceElem)                        &
            &          - sourceMom( 3, iSourceElem)*sourceMom( 3, iSourceElem) &
            &          / sourceMom( 1, iSourceElem)
          sourceMom( 6, iSourceElem ) = sourceMom( 6, iSourceElem )            &
            &          - sourceMom( 2, iSourceElem)*sourceMom( 3, iSourceElem) &
            &          / sourceMom( 1, iSourceElem)
          ! Transfer momentum to velocity
          sourceMom( 2:3, iSourceElem ) = sourceMom( 2:3, iSourceElem )        &
            &                           / sourceMom( 1, iSourceElem )

          ! Set the further higher moments to their equilibrium distribution
          sourceMom( 7, iSourceElem ) = get_moment( sourceEq, layout, (/2,1,0/))
          sourceMom( 8, iSourceElem ) = get_moment( sourceEq, layout, (/1,2,0/))
          sourceMom( 9, iSourceElem ) = get_moment( sourceEq, layout, (/2,2,0/))
        enddo
        ! interpolate the density linearly 
        targetMom(1) = sum(sourceMom(1,1:nSourceElems))                        &
          &          / real(nSourceElems,kind=rk)
        ! the source element enumeration always is the case for childNum 4 
        iCase = 4
        
        ! quadratic interpolation for momentumX, momentumY
        targetMom(2:3) = mus_interpolate_quadraticCompact2d_leastSq(           &
          &              sourceMom(2:3, 1:nSourceElems),                       &
          &              targetCoord(1:2), sourceMom(4:6, 1:nSourceElems),     &
          &              nSourceElems,                                         &
          &              intp%fromFiner( targetLevel )%matrixInv( iCase )%A )
        ! linear interpolation for higher moments
        do iVal = 4,9
          targetMom(iVal) = sum(sourceMom(iVal,1:nSourceElems))                &
            &             / real(nSourceElems,kind=rk)
        enddo

        ! shear stress part has to be scaled with fac
        targetMom(4:6) = targetMom(4:6)*omfac
        ! and add back the pi(0) part to the tensor
        ! No need to rescale higher order moments as they were set to
        ! equilibrium
        targetMom(4) = targetMom(4) + cs2*targetMom(1)                         &
          &          + targetMom(2)*targetMom(2)/targetMom(1)
        targetMom(5) = targetMom(5) + cs2*targetMom(1)                         &
          &          + targetMom(3)*targetMom(3)/targetMom(1)
        targetMom(6) = targetMom(6)                                            &
          &          + targetMom(2)*targetMom(3)/targetMom(1)
        ! Transfer back velocity field to momentum field
        targetMom(2:3) = targetMom(2:3)*targetMom(1)

        ! transform the moments back to the pdfs
        targetPdf = matmul( layout%moment%toPdf%A, targetMom )

        ! Now write the resulting pdf in the current direction to the target
        ! Element position
        do iVal = 1, layout%stencil(1)%QQ
          pdf( targetLevel )%state( &
   &?SAVE?( varSys%Variable%val(iVar)%varPos( iVal ), targetElem, varSys%nScalars, pdf(targetLevel)%neigh ),&
         & t_target   ) &
         & = targetPdf( iVal ) 
        enddo ! ivar loop
      enddo ! ivar loop
    enddo

  end subroutine fillMyGhostsFromFiner_quadraticCompact2d
!******************************************************************************!



!******************************************************************************!
  !>  
  !!  
  subroutine fillFinerGhostsFromMe_quadraticCompact2d( pdf, glob, fieldProp, level, &
   &                               scheme, layout, intp, ind, varSys, derVarPos ) 
    !---------------------------------------------------------------------------
    !> global fluid parameters 
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !> global parameters 
    type( pdf_global_type ), intent(inout) :: glob
    !> the current scheme 
    type( mus_scheme_type ), intent(in)   :: scheme
    !> fluid and species parameters 
    type( mus_field_prop_type ), intent(in) ::  fieldProp(:)
    !> level
    integer, intent(in)   :: level
    !> the layout used
    type( mus_scheme_layout_type), intent(in)   :: layout
    !> interpolation method info
    type( interpolation_type ), intent(in)   :: intp
    !> indirection list
    type( grw_intArray_type ), intent(in)   :: ind
    !> global variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> required variable system which maps to global varsys
    type( mus_derVarPos_type ), intent(in) :: derVarPos(:)
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iVal           ! value counter
    integer :: iElem          ! element counter for outer loop
    integer :: iDir           ! link counter for stencil loop
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    real(kind=rk) :: targetPdf( layout%stencil(1)%QQ ) ! pdf to reconstruct
    real(kind=rk) :: sourcePdf( layout%stencil(1)%QQ ) ! pdf to reconstruct from
    ! moments of the source elements' pdf
    real(kind=rk) :: sourceMom( layout%stencil(1)%QQ, intp%nMaxSources )
    ! interpolated moments
    real(kind=rk) :: targetMom( layout%stencil(1)%QQ )
    real(kind=rk) :: targetCoord(3) ! coordinates to interpolate to
    ! Coordinates in rotated reference cube
    real(kind=rk) :: targetCoord_trafo(3)
    ! shear stress scaling factor
    real(kind=rk) :: omfac
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to write to
    integer :: t_source ! which time layer to read from
    real(kind=rk) :: sourceEq( layout%stencil(1)%QQ ) ! source equilirbium part
    integer :: iCase ! child case for rotation of the unit cube
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) usq, ucx     ! square velocity
    !---------------------------------------------------------------------------

    sourceLevel = level
    targetLevel = level + 1
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))

    ! Treat all fine target elements:
    !do iElem = 1, pdf( targetLevel )%nElems_ghostFromCoarser
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      !test = .false.
      ! Read the target element treeId
      targetElem = iElem + glob%levelDesc( targetLevel )%                      &
        &                                  offset( 1, eT_ghostFromCoarserElem )
      ! get the target coordinates 
      targetCoord = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%coord
      targetCoord_trafo = (/0.25_rk, 0.25_rk, 0._rk /)
      ! Find out how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      !nSourceElems = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%elem%nVals
      !MH: D2Q9
      nSourceElems =4

      !Compute rho and velocity for each field and interpolate pdf field wise
      do iVar = 1, varSys%nVars
        omfac = (1._rk - 0.5*fieldProp( iVar )%fluid%omLvl( sourceLevel ))     &
          &   / (1._rk - 0.5*fieldProp( iVar )%fluid%omLvl( targetLevel )) 

        ! Now loop over all fine source elements for this target:
        do iSourceElem = 1, nSourceElems
          ! Get the source element's treeId
          sourceElem = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%  &
            &                                           elem%val( iSourceElem )
          ! get the pdf values for the source elements from the state vector
          ! and store in sourcePdf
          do iVal = 1, layout%stencil(1)%QQ
            sourcePdf( iVal )  = pdf( sourceLevel )%state( &
  &  ?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh ),&
         &    t_source  )
          enddo

          ! Calculate the raw moments from the pdfs
          sourceMom( :,iSourceElem ) = matmul( layout%moment%toMoments%A,      &
            &                                                       sourcePdf )
!          ! And calculate the equilibrium distribution from the macroscopic 
!          ! quantities in sourceMom
!          input_loc = (/ sourceMom( 1, iSourceElem),    & ! density
!                       & sourceMom( 2, iSourceElem)/sourceMom(1, iSourceElem),   & ! velX
!                       & sourceMom( 3, iSourceElem)/sourceMom(1, iSourceElem),   & ! velY
!                       & 0.0_rk /) ! velZ
!
!          call varSys%derive( derVarPos(iVar)%equilFromMacro )%do(                         &
!            &              input = input_loc, nElems = 1, iVar = 1,                  &
!            &              requVar = varSys%Variable%val( derVarPos(iVar)%equilFromMacro ),&
!            &              globSys = varSys, stencil = layout%stencil(1),            &
!            &              res = sourceEq,                           &
!            &              conf = varSys%solSpec_conf )

          ! velX, velY
          u_x = sourceMom( 2, iSourceElem)/sourceMom(1, iSourceElem)
          u_y = sourceMom( 3, iSourceElem)/sourceMom(1, iSourceElem)

          ! square velocity and derived constants
          usq  = u_x*u_x + u_y*u_y

          do iDir = 1, layout%stencil(1)%QQ
            ! velocity times lattice unit velocity
            ucx = dble( layout%stencil(1)%cxDir( 1, iDir ) )*u_x               &
              & + dble( layout%stencil(1)%cxDir( 2, iDir ) )*u_y 
            ! calculate equilibrium density
            sourceEq( iDir ) = layout%weight( iDir )*sourceMom(1, iSourceElem) &
              &              * ( 1._rk + ucx*cs2inv + ucx * ucx * cs2inv       &
              &              * cs2inv*0.5_rk - usq * cs2inv * 0.5_rk )
          end do

          ! remove pi(0) parts from the second order moment
          sourceMom( 4, iSourceElem ) = sourceMom( 4, iSourceElem )            &
            &          - cs2*sourceMom( 1, iSourceElem)                        &
            &          - sourceMom( 2, iSourceElem)*sourceMom( 2, iSourceElem) &
            &          / sourceMom( 1, iSourceElem)
          sourceMom( 5, iSourceElem ) = sourceMom( 5, iSourceElem )            &
            &          - cs2*sourceMom( 1, iSourceElem)                        &
            &          - sourceMom( 3, iSourceElem)*sourceMom( 3, iSourceElem) &
            &          / sourceMom( 1, iSourceElem)
          sourceMom( 6, iSourceElem ) = sourceMom( 6, iSourceElem )            &
            &          - sourceMom( 2, iSourceElem)*sourceMom( 3, iSourceElem) &
            &          / sourceMom( 1, iSourceElem)
          ! Set the further higher moments to equilibrium
          sourceMom( 7, iSourceElem ) = get_moment( sourceEq, layout, (/2,1,0/))
          sourceMom( 8, iSourceElem ) = get_moment( sourceEq, layout, (/1,2,0/))
          sourceMom( 9, iSourceElem ) = get_moment( sourceEq, layout, (/2,2,0/))
          ! Transfer momentum field to velocity field
          sourceMom( 2:3, iSourceElem ) = sourceMom( 2:3, iSourceElem )        &
            &                           / sourceMom( 1, iSourceElem )
        end do   
        ! interpolate the density linearly 
        targetMom(1)  = tem_intp_bilinear( sourceMom(1,1:nSourceElems),        &
          &                                             targetCoord_trafo(1:2))
        
        ! get the rotation case from the child number of the target element
        iCase = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%childNum

        targetMom(2:3) = mus_interpolate_quadraticCompact2d_leastSq(           &
          &              sourceMom(2:3, 1:nSourceElems),                       &
          &              targetCoord(1:2), sourceMom(4:6, 1:nSourceElems),     &
          &              nSourceElems,                                         &
          &              intp%fromCoarser( targetLevel )%matrixInv( iCase )%A )
        ! linear interpolation for higher moments
        targetMom(4:9) = tem_intp_bilinear( sourceMom(4:9,1:nSourceElems),     &
          &                                          targetCoord_trafo(1:2), 6)

        ! Rescale the shear stress moments
        targetMom(4:6) = targetMom(4:6)*omfac
        ! and add back the pi(0) part to the tensor
        targetMom(4) = targetMom(4) + cs2*targetMom(1)                         &
          &          + targetMom(2)*targetMom(2)/targetMom(1)
        targetMom(5) = targetMom(5) + cs2*targetMom(1)                         &
          &          + targetMom(3)*targetMom(3)/targetMom(1)
        targetMom(6) = targetMom(6)                                            &
          &          + targetMom(2)*targetMom(3)/targetMom(1)
        ! Transfer back velocity field to momentum field
        targetMom(2:3) = targetMom(2:3)*targetMom(1)

        ! Transform back into pdf space
        targetPdf = matmul( layout%moment%toPdf%A, targetMom )

        ! Now write the resulting pdf in the current direction to the target
        ! Element position
        do iVal = 1, layout%stencil(1)%QQ
        pdf( targetLevel )%state( &
         &?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), targetElem, varSys%nScalars, pdf(targetLevel)%neigh ),&
         & t_target  ) &
         & = targetPdf(iVal)
         enddo
      enddo ! ivar loop
    enddo

  end subroutine fillFinerGhostsFromMe_quadraticCompact2d
!******************************************************************************!


!******************************************************************************!
  !>
  !!
  subroutine unitTest_interpolate_quadraticCompact( matrix, iCase, passed )
    !---------------------------------------------------------------------------
    !>
    real(kind=rk), intent(in) :: matrix(:,:)
    !>
    integer, intent(in)  :: iCase
    !>
    logical, intent(out) :: passed 
    !---------------------------------------------------------------------------
    ! non-equilibrium part of the resulting pdf
    real(kind=rk) :: sourceMom( 9, 4 )
    real(kind=rk) :: targetMom( 9 ) ! non-equilibrium part of the resulting pdf
    real(kind=rk) :: refX(2)        ! reference values
    real(kind=rk) :: error(4)       ! error values
    integer :: nSourceElems, SE, SW, NE, NW
    !---------------------------------------------------------------------------
    passed = .false.
    error  = 0._rk
    nSourceElems = 4
    call tem_log( level = 1,                                                   &
      &           msg   = ' Performing unit test for interpolation moment '//  &
      &                   'gradient routine' )
    call tem_log( level = 1,                                                   &
      &           msg   = ' Current case for child '//trim( tem_toStr( iCase )))

    select case( iCase )
    case( 1, 5)
      SW = 4
      SE = 3
      NW = 2
      NE = 1
    case( 2, 6)
      SW = 2
      SE = 4
      NW = 1
      NE = 3
    case( 3, 7)
      SW = 3
      SE = 1
      NW = 4
      NE = 2
    case( 4, 8)
      SW = 1
      SE = 2
      NW = 3
      NE = 4
    end select

    sourceMom = 0._rk
    sourceMom( 1, : ) = 1._rk
    sourceMom( 3, : ) = 0.1_rk
?? if( STREAM_PUSH ) then 
      sourceMom( 6, SW ) = -0.074074074074074_rk  ! SW
      sourceMom( 6, SE ) = -0.037037037037037_rk  ! SE
      sourceMom( 6, NW ) =  0.074074074074074_rk  ! NW
      sourceMom( 6, NE ) =  0.037037037037037_rk  ! NE
?? else
      sourceMom( 6, SW ) =  0.059259259259259_rk  ! SW
      sourceMom( 6, SE ) =  0.029629629629630_rk  ! SE
      sourceMom( 6, NW ) = -0.059259259259259_rk  ! NW
      sourceMom( 6, NE ) = -0.029629629629630_rk  ! NE
?? endif
    refX(1) = 0.05625_rk
    refX(2) = 0.1_rk
    targetMom(1:2) = mus_interpolate_quadraticCompact2d_leastSq(               &
      &                                      sourceMom(2:3, 1:nSourceElems),   &
      &                (/0.25_rk, 0.25_rk/), sourceMom(4:6, 1:nSourceElems),   &
      &                nSourceElems, matrix )
    error(1:2) = (refX(1:2)-targetMom(1:2))/refX(1:2)*100._rk
?? if( STREAM_PUSH ) then 
?? else
    call tem_log( level = 6,                                                   &
      &           msg   = 'X  reference result '//trim( tem_toStr( refX(1) ))//&
      &                   ' intp '//trim( tem_toStr( targetMom(1) ))//         &
      &                   ' Error [%] '//trim( tem_toStr( error(1) )))
    call tem_log( level = 6,                                                   &
      &           msg   = 'Y  reference result '//trim( tem_toStr( refX(2) ))//&
      &                   ' intp '//trim( tem_toStr( targetMom(2) ))//         &
      &                   ' Error [%] '//trim( tem_toStr( error(2) )))
?? endif

    refX(1) = 0.075_rk
    refX(2) = 0.1_rk
    targetMom(1:2) = mus_interpolate_quadraticCompact2d_leastSq(               &
      &                                      sourceMom(2:3, 1:nSourceElems),   &
      &                  (/0.5_rk, 0.5_rk/), sourceMom(4:6, 1:nSourceElems),   &
      &                  nSourceElems, matrix )
    error(3:4) = (refX(1:2)-targetMom(1:2))/refX(1:2)*100._rk

?? if( STREAM_PUSH ) then 
?? else
    call tem_log( level = 6,                                                   &
      &           msg   = 'X  reference result '//trim( tem_toStr( refX(1) ))//&
      &                   ' intp '//trim( tem_toStr( targetMom(1) ))//         &
      &                   ' Error [%] '//trim( tem_toStr( error(3) )))
    call tem_log( level = 6,                                                   &
      &           msg   = 'Y  reference result '//trim( tem_toStr( refX(2) ))//&
      &                   ' intp '//trim( tem_toStr( targetMom(2) ))//         &
      &                   ' Error [%] '//trim( tem_toStr( error(4) )))
?? endif

    if( abs( maxval( error ) ) .gt. 0.00001_rk ) then
      passed = .false.
    else
      passed = .true.
    endif 

  end subroutine unitTest_interpolate_quadraticCompact
!******************************************************************************!


!******************************************************************************!
  !> set moments of arbitary moment
  !!
  pure function get_moment( pdf, layout, expX ) result( mom )
    !---------------------------------------------------------------------------
    !> stencil layout
    type( mus_scheme_layout_type), intent(in)   :: layout
    !> source values of the momentum on the square corners
    real(kind=rk), intent(in) :: pdf(layout%stencil(1)%QQ)
    !> exponents of the moments
    integer, intent(in) :: expX(3)
    !> interpolated value
    real(kind=rk) :: mom
    !---------------------------------------------------------------------------
    integer :: iVal         
    !---------------------------------------------------------------------------

    mom = 0._rk
    do iVal = 1, layout%stencil(1)%QQ
      mom = mom                                                                &
        & + (real(layout%stencil(1)%cxDir(1,iVal),kind=rk))**expX(1)           &
        & * (real(layout%stencil(1)%cxDir(2,iVal),kind=rk))**expX(2)           &
        & * (real(layout%stencil(1)%cxDir(3,iVal),kind=rk))**expX(3)           &
        & *  pdf( iVal )
    enddo

  end function get_moment
!******************************************************************************!


!******************************************************************************!
  !> Calculate the equilibrium distribution function in all directions
  !!
  !! The equilibrium distribution function is:\n
  !! \f$ f^{eq}_i = w_i \rho ( 1 + \frac{\vec c_i \cdot \vec u}{c^2_s}
  !!                      + \frac{ {(\vec c_i \cdot \vec u)}^2}{2c^4_s}
  !!                      - \frac{\vec u \cdot \vec u}{2c^2_s}) \f$\n
  !! where \f$w_i\f$ is the weight in each direction,\n
  !! \f$\rho\f$ is the macroscopic value of density,\n
  !! \f$c_s\f$ is the speed of sound,\n
  !! \f$\vec c_i\f$ is the lattice unit velocity in each direction,\n
  !! \f$\vec u\f$ is the macroscopic value of velocity.
  !!
  pure function getEquilibrium( density, velocity, layout ) result( equil )
    !---------------------------------------------------------------------------
    !> scheme layout
    type(mus_scheme_layout_type), intent(in) :: layout
    !>
    real(kind=rk), intent(in) :: density
    !>
    real(kind=rk), intent(in) :: velocity(3)
    !> return value
    real(kind=rk)             :: equil(layout%stencil(1)%QQ)
    !---------------------------------------------------------------------------
    ! local variables
    real(kind=rk) :: ucx, usq
    integer :: iDir
    !---------------------------------------------------------------------------

    ! square of velocity
    usq = velocity(1) * velocity(1)                                            &
      & + velocity(2) * velocity(2)                                            &
      & + velocity(3) * velocity(3)

    do iDir = 1, layout%stencil(1)%QQ

      ! velocity times lattice unit velocity
      ucx = real( layout%stencil(1)%cxDir( 1, iDir ), kind=rk)*velocity(1)     &
        & + real( layout%stencil(1)%cxDir( 2, iDir ), kind=rk)*velocity(2)     &
        & + real( layout%stencil(1)%cxDir( 3, iDir ), kind=rk)*velocity(3) 

      ! calculate equilibrium density
      equil( iDir ) = layout%weight( iDir ) * density * ( 1.d0 + ucx*cs2inv    &
        &           + ucx*ucx*cs2inv*cs2inv*0.5d0                              &
        &           - usq*cs2inv*0.5d0 )

    enddo

  end function getEquilibrium
!******************************************************************************!


end module mus_interpolate_compact_module
!******************************************************************************!
