! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Interpolation of flow quantities between different grid levels
!!
!! \section intp_main Interpolation 
!! The routines defined here, fill up the ghost elements with valid data.
!! Ghost elements are employed at grid level interfaces to provide valid
!! pdf values to the neighboring fluid elements. This way, the solvers can 
!! act on elements of the same size only, treating the levels successively.
!! Target elements are the ghost elements, which have to be filled with 
!! valid values.
!! Source elements are the fluid elements from other levels, from where to 
!! take the input values for the interpolation.
!! The target ghost elements on the target level have corresponding source 
!! fluid elements on the source level. 
!!
!! \ref tem_topology_module For a detailed description of the grid 
!! 
!! \section workflow Workflow
!! 
!! Each interpolation routine acts on a list of ghost elements.
!! This list contains pointers to the position in the total list.
!! For each of these ghost elements, the source elements are identified.
!! Before that, the sourceLevel is identified. However, the code is restricted
!! to work with a level jump of only one level, so the sourceLevel is 
!! for sourceLevel = targetLevel+1
!! sourceLevel = targetLevel-1
!!
!! For an overview over implemented interpolation methods, see
!! \ref intp_methods "Interpolation methods"
!!
!!\author Manuel Hasert
!!
?? include 'header/lbm_macros.inc'
module mus_interpolate_debug_module

  ! include treelm modules
  use env_module,            only: rk
  use tem_comm_env_module,   only: process
  use tem_element_module,    only: eT_GhostFromCoarserElem,eT_ghostFromFinerElem
  use tem_grow_array_module, only: grw_intArray_type
  use tem_tools_module,      only: tem_write
  use tem_var_system_module, only: tem_var_system_type
  use tem_logging_module,    only: tem_toStr, tem_log

  ! include musubi modules
  use mus_field_prop_module,         only: mus_field_prop_type
  use mus_interpolate_header_module, only: interpolation_type,                 &
    &                                      determine_timeLayer
  use mus_pdf_module,                only: pdf_data_type, pdf_global_type
  use mus_scheme_layout_module,      only: mus_scheme_layout_type
  use mus_scheme_type_module,        only: mus_scheme_type
  use mus_variable_module,           only: mus_derVarPos_type

  implicit none

  private

  public :: do_nothing
  public :: fillMyGhostsFromFiner_average
  public :: fillFinerGhostsFromMe_average
  public :: fillMyGhostsFromFiner_debug
  public :: fillFinerGhostsFromMe_debug
  public :: fillMyGhostsFromFiner_copyFirst
  public :: fillFinerGhostsFromMe_copyFirst

 contains

!******************************************************************************!
  !> Fill GhostFromFiner elements on my level with debug value 
  !!
  subroutine do_nothing ( pdf, glob, fieldProp, level,                         &
    &                     scheme,  layout, intp, ind, varSys, derVarPos ) 
    !---------------------------------------------------------------------------
    !> global fluid parameters 
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !> global parameters 
    type( pdf_global_type ), intent(inout) :: glob
    !> the current scheme 
    type( mus_scheme_type ), intent(in)   :: scheme
    !> fluid and species parameters 
    type( mus_field_prop_type ), intent(in) ::  fieldProp(:)
    !> level
    integer, intent(in)   :: level
    !> the layout used
    type( mus_scheme_layout_type), intent(in)   :: layout
    !> interpolation method info
    type( interpolation_type ), intent(in)   :: intp
    !> indirection list
    type( grw_intArray_type ), intent(in)   :: ind
    !> global variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> required variable system which maps to global varsys
    type( mus_derVarPos_type ), intent(in) :: derVarPos(:)
    !---------------------------------------------------------------------------
 
  end subroutine do_nothing
!******************************************************************************!


!******************************************************************************!
  !> Interpolation routine that is based on a simple weighted average of source
  !! nodes
  !!
  !! This is the interpolation fine-> coarse. It does not need weights, as the 
  !! distances source <-> target are equal for all source nodes. Thus, we can 
  !! add all source pdf's and divide by the number of source nodes.
  !!
  subroutine fillMyGhostsFromFiner_Average( pdf, glob, fieldProp, level,       &
    &                            scheme, layout, intp, ind, varSys, derVarPos )
    !---------------------------------------------------------------------------
    !> global fluid parameters 
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !> global parameters 
    type( pdf_global_type ), intent(inout) :: glob
    !> the current scheme 
    type( mus_scheme_type ), intent(in)   :: scheme
    !> fluid and species parameters 
    type( mus_field_prop_type ), intent(in) ::  fieldProp(:)
    !> level
    integer, intent(in)   :: level
    !> the layout used
    type( mus_scheme_layout_type), intent(in)   :: layout
    !> interpolation method info
    type( interpolation_type ), intent(in)   :: intp
    !> indirection list
    type( grw_intArray_type ), intent(in)   :: ind
    !> global variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> required variable system which maps to global varsys
    type( mus_derVarPos_type ), intent(in) :: derVarPos(:)
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iDir           ! current direction (discrete velocity) for loop
    integer :: iElem          ! current target element (for outer loop)
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    real(kind=rk) :: tVal     ! temporary value: accumulator for pdf density
    integer :: pos  ! position of field component in the state vector
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to work on?
    integer :: t_source ! which time layer to work on?
    !--------------------------------------------------------------------------
    targetLevel = level
    sourceLevel = level+1
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))
  
    ! Treat every coarse target element:
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )

      ! find out the treeId of the current target element that we are treating 
      ! in this iteration
      targetElem = iElem + glob%levelDesc( targetLevel )%                      &
        &                                    offset( 1, eT_ghostFromFinerElem )
      ! find out, how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      nSourceElems = glob%levelDesc( targetLevel )%depFromFiner( iElem )%      &
        &                                                            elem%nVals
     
      ! Now that we know the target element, loop over all nVars and
      ! discrete velocity directions for each field variable 
      do iVar = 1, varSys%nVars
        do iDir = 1, layout%stencil(1)%QQ
          pos = varSys%Variable%val( iVar )%varPos( iDir )
          tVal = 0._rk
          !... and all the fine source elements...
          do iSourceElem = 1, nSourceElems
            ! Find out the source element's treeId
            sourceElem = glob%levelDesc( targetLevel )%depFromFiner( iElem )   &
              &                                       %elem%val( iSourceElem )
            ! Accumulate all source pdf's in tVal. We are just adding up values
            ! here...
            tVal = tVal + pdf( sourceLevel )%state(                            &
         & ?SAVE?( pos, sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh ), &
         &    t_source )
          enddo
          
          ! Now we have all pdf's for this discrete velocity (iDir) added up.
          ! Divide this by the number of sources and write the result to the
          ! target element.
          pdf( level )%state(                                                  &
            & ?SAVE?( pos, targetElem, varSys%nScalars, pdf( level)%neigh ),   &
            & t_target ) = tVal / real( nSourceElems, kind=rk )
        enddo
      enddo
    enddo
  end subroutine fillMyGhostsFromFiner_Average
!******************************************************************************!


!******************************************************************************!
  !> Interpolation routine that is based on a simple weighted average of 
  !! source nodes
  !!
  !! This is the interpolation coarse-> fine. Weights are needed here, as 
  !! the distances source <-> target are different for the source nodes.
  !!
  subroutine fillFinerGhostsFromMe_Average( pdf, glob, fieldProp, level,       &
    &                           scheme,   layout, intp, ind, varSys, derVarPos) 
    !---------------------------------------------------------------------------
    !> global fluid parameters 
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !> global parameters 
    type( pdf_global_type ), intent(inout) :: glob
    !> the current scheme 
    type( mus_scheme_type ), intent(in)   :: scheme
    !> fluid and species parameters 
    type( mus_field_prop_type ), intent(in) ::  fieldProp(:)
    !> level
    integer, intent(in)   :: level
    !> the layout used
    type( mus_scheme_layout_type), intent(in)   :: layout
    !> interpolation method info
    type( interpolation_type ), intent(in)   :: intp
    !> indirection list
    type( grw_intArray_type ), intent(in)   :: ind
    !> global variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> required variable system which maps to global varsys
    type( mus_derVarPos_type ), intent(in) :: derVarPos(:)
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iDir           ! current direction (discrete velocity) for loop
    integer :: iElem          ! current target element (for outer loop)
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    real(kind=rk) :: tVal     ! temporary value: accumulator for pdf density
    real(kind=rk) :: fTmp     ! temporary pdf value
    ! weight for source elements depending on distance to target element
    real(kind=rk) :: weight(intp%nMaxSources)  
    integer :: pos  ! position of field component in the state vector
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to work on?
    integer :: t_source ! which time layer to work on?
    logical :: error
    !--------------------------------------------------------------------------

    sourceLevel = level
    targetLevel = level + 1
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))
    
    ! Treat every coarse target element:
    !do iElem = 1, pdf( targetLevel )%nElems_ghostFromCoarser 
    do indElem = 1, ind%nVals !pdf( targetLevel )%nElems_ghostFromCoarser
      iElem = ind%val( indElem )
      ! find out the treeId of the current target element that we are treating 
      ! in this iteration
      targetElem = iElem + glob%levelDesc( targetLevel )                       &
        &                                 %offset( 1, eT_ghostFromCoarserElem )
      ! find out, how many coarse source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      nSourceElems = glob%levelDesc( targetLevel )%depFromCoarser( iElem )     &
        &                                                           %elem%nVals

      ! Get the weights for all source elements. Weight of each element 
      ! is basically 1/distance.
      ! The weights are normalized, such that sum(weights)=1
      ! We do not calculate weights on the fly, because the calculation of 
      ! element positions and distances is expensive.
      weight( 1:nSourceElems ) = 1._rk/real(nSourceElems,kind=rk)
      if( abs(1._rk- sum(weight(1:nSourceElems))) .gt. 1.E-15_rk ) then
        call tem_log( level = 3,                                               &
          &           msg   = trim( tem_toStr( process%rank ))//               &
          &                   ' error for tID '//trim( tem_toStr(              &
          &                glob%levelDesc( targetLevel )%total( targetElem ))))
        call tem_log( level = 3,                                               &
          &           msg   = trim( tem_toStr( process%rank ))//               &
          &                   ' sum of weight is wrong '//                     &
          &                   trim( tem_toStr( sum( weight(1:nSourceElems)))))
      end if
      ! Now that we know the target element,
      !... and all the coarse source elements..
      ! loop over all nVars and
      ! discrete velocity directions for each field variable 
      error = .false.
      do iVar = 1, varSys%nVars
        do iDir = 1, layout%stencil(1)%QQ
          pos = varSys%Variable%val( iVar )%varPos( iDir )
          tVal = 0._rk
          do iSourceElem = 1, nSourceElems 
            ! Find out the source element's treeId
            sourceElem = glob%levelDesc( targetLevel )%depFromCoarser( iElem ) &
              &                                       %elem%val( iSourceElem)
            fTmp = pdf( sourceLevel )%state(                                   &
           ?SAVE?( pos, sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh ), &
         &    t_source )
            if( fTmp .lt. 0._rk ) then
              error = .true.
            end if

            ! Accumulate all source pdf's with weighting in tVal
            tVal = tVal + weight(iSourceElem) * fTmp
          enddo
          
          ! Now we have all pdf's for this discrete velocity (iDir) weighted 
          ! and added up. The weights add up to 1, so we don't have to 
          ! divide here by anything. 
          pdf( targetLevel )%state(                                            &
           ?SAVE?( pos, targetElem, varSys%nScalars, pdf(targetLevel)%neigh ), &
         & t_target ) = tVal
        enddo
      enddo
      if( error ) then
        call tem_write( trim( tem_toStr( process%rank ))//' error for tID '//  &
          &             trim( tem_toStr( glob%levelDesc( targetLevel )%        &
          &                                              total( targetElem ))))
        call tem_write( trim( tem_toStr( process%rank ))//                     &
          &             ' negative fTmp in sourceElem '//                      &
          &             trim( tem_toStr( glob%levelDesc( sourceLevel )%        &
          &                                              total( sourceElem ))))
      end if
    enddo

  end subroutine fillFinerGhostsFromMe_Average
!******************************************************************************!


!******************************************************************************!
  !> Fill GhostFromFiner elements on my level with debug value 
  !!
  subroutine fillMyGhostsFromFiner_Debug ( pdf, glob, fieldProp, level,        &
    &                           scheme, layout, intp, ind, varSys, derVarPos ) 
    !---------------------------------------------------------------------------
    !> global fluid parameters 
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !> global parameters 
    type( pdf_global_type ), intent(inout) :: glob
    !> the current scheme 
    type( mus_scheme_type ), intent(in)   :: scheme
    !> fluid and species parameters 
    type( mus_field_prop_type ), intent(in) ::  fieldProp(:)
    !> level
    integer, intent(in)   :: level
    !> the layout used
    type( mus_scheme_layout_type), intent(in)   :: layout
    !> interpolation method info
    type( interpolation_type ), intent(in)   :: intp
    !> indirection list
    type( grw_intArray_type ), intent(in)   :: ind
    !> global variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> required variable system which maps to global varsys
    type( mus_derVarPos_type ), intent(in) :: derVarPos(:)
    !---------------------------------------------------------------------------
    integer :: sourceLevel      ! level of source elements
    integer :: sourceElem       ! treeId of current source element
    integer :: targetLevel      ! level of target elements
    integer :: targetElem       ! treeId of current source element
    integer :: iDir             ! current direction (discrete velocity) for loop
    integer :: iElem            ! current target element (for outer loop)
    integer :: indElem          ! element counter for indirection list
    integer :: pos  ! position of field component in the state vector
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to work on?
    integer :: t_source ! which time layer to work on?
    !--------------------------------------------------------------------------
 
    targetLevel = level
    sourceLevel = level+1
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))

    ! Treat every coarse target element:
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      ! find out the treeId of the current target element that we are treating 
      ! in this iteration
      targetElem = iElem + glob%levelDesc( targetLevel )                       &
        &                      %offset( 1, eT_ghostFromFinerElem )
      ! find out the treeId of the only source element from which we will 
      ! copy the pdf values
      sourceElem = glob%levelDesc( targetLevel )%depFromFiner( iElem )         &
        &                                       %elem%val(1)

      ! Now that we know the target element, loop over all nVars and
      ! discrete velocity directions for each field variable 
      do iVar = 1, varSys%nVars
        do iDir = 1, layout%stencil(1)%QQ
          pos = varSys%Variable%val( iVar )%varPos( iDir )
          ! ... and simply copy the pdf from source to target.
          pdf( targetLevel )%state(                                            &
         & ?SAVE?( pos, targetElem, varSys%nScalars, pdf(targetLevel)%neigh ), &
         & t_target ) = intp%debugValue
        enddo
      enddo
    enddo
  end subroutine fillMyGhostsFromFiner_Debug
!******************************************************************************!


!******************************************************************************!
  !> Fill the ghost elements on the finer level with debug value
  !!
  subroutine fillFinerGhostsFromMe_Debug( pdf, glob, fieldProp, level,         &
    &                           scheme, layout, intp, ind, varSys, derVarPos ) 
    !---------------------------------------------------------------------------
    !> global fluid parameters 
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !> global parameters 
    type( pdf_global_type ), intent(inout) :: glob
    !> the current scheme 
    type( mus_scheme_type ), intent(in)   :: scheme
    !> fluid and species parameters 
    type( mus_field_prop_type ), intent(in) ::  fieldProp(:)
    !> level
    integer, intent(in)   :: level
    !> the layout used
    type( mus_scheme_layout_type), intent(in)   :: layout
    !> interpolation method info
    type( interpolation_type ), intent(in)   :: intp
    !> indirection list
    type( grw_intArray_type ), intent(in)   :: ind
    !> global variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> required variable system which maps to global varsys
    type( mus_derVarPos_type ), intent(in) :: derVarPos(:)
    !---------------------------------------------------------------------------
    integer :: sourceLevel      ! level of source elements
    integer :: sourceElem       ! treeId of current source element
    integer :: targetLevel      ! level of target elements
    integer :: targetElem       ! treeId of current source element
    integer :: iDir             ! current direction (discrete velocity) for loop
    integer :: iElem            ! current target element (for outer loop)
    integer :: indElem          ! element counter for indirection list
    integer :: pos  ! position of field component in the state vector
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to work on?
    integer :: t_source ! which time layer to work on?
    !--------------------------------------------------------------------------
    sourceLevel = level
    targetLevel = level + 1
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))

    ! Treat every fine target element:
    !do iElem = 1, pdf( targetLevel )%nElems_ghostFromCoarser 
    do indElem = 1, ind%nVals !pdf( targetLevel )%nElems_ghostFromCoarser
      iElem = ind%val( indElem )
      ! Find out treeIds of current target element and the single source element 
      ! that we copy from
      sourceElem = glob%levelDesc( targetLevel )%depFromCoarser( iElem )       &
        &                                       %elem%val( 1)
      targetElem = iElem + glob%levelDesc( targetLevel )                       &
        &                      %offset( 1, eT_ghostFromCoarserElem )

      ! Now that we know the target element, loop over all nVars and
      ! discrete velocity directions for each field variable 
      do iVar = 1, varSys%nVars
        do iDir = 1, layout%stencil(1)%QQ
          pos = varSys%Variable%val( iVar )%varPos( iDir )
            ! ... and simply copy the pdf from source to target.
          pdf( targetLevel )%state(                                            &
         &  ?SAVE?( pos, targetElem, varSys%nScalars, pdf(targetLevel)%neigh ),&
         & t_target ) = intp%debugValue
        enddo
      enddo
    enddo

  end subroutine fillFinerGhostsFromMe_Debug
!******************************************************************************!


!******************************************************************************!
  !> This simple method does not perform actual interpolation, but simply copies
  !! one of the source element's pdf values to the target element. It is mostly
  !! there for debugging purposes, the quality is not good enough for
  !! production use! 
  !!
  subroutine fillMyGhostsFromFiner_CopyFirst ( pdf, glob, fieldProp, level,    &
    &                           scheme, layout, intp, ind, varSys, derVarPos ) 
    !---------------------------------------------------------------------------
    !> global fluid parameters 
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !> global parameters 
    type( pdf_global_type ), intent(inout) :: glob
    !> the current scheme 
    type( mus_scheme_type ), intent(in)   :: scheme
    !> fluid and species parameters 
    type( mus_field_prop_type ), intent(in) ::  fieldProp(:)
    !> level
    integer, intent(in)   :: level
    !> the layout used
    type( mus_scheme_layout_type), intent(in)   :: layout
    !> interpolation method info
    type( interpolation_type ), intent(in)   :: intp
    !> indirection list
    type( grw_intArray_type ), intent(in)   :: ind
    !> global variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> required variable system which maps to global varsys
    type( mus_derVarPos_type ), intent(in) :: derVarPos(:)
    !---------------------------------------------------------------------------
    integer :: sourceLevel      ! level of source elements
    integer :: sourceElem       ! treeId of current source element
    integer :: targetLevel      ! level of target elements
    integer :: targetElem       ! treeId of current source element
    integer :: iDir             ! current direction (discrete velocity) for loop
    integer :: iElem            ! current target element (for outer loop)
    integer :: indElem          ! element counter for indirection list
    integer :: pos  ! position of field component in the state vector
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to work on?
    integer :: t_source ! which time layer to work on?
    !--------------------------------------------------------------------------
 
    targetLevel = level
    sourceLevel = level+1
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))

    ! Treat every coarse target element:
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )

      ! find out the treeId of the current target element that we are treating 
      ! in this iteration
      targetElem = iElem + glob%levelDesc( targetLevel )                       &
        &                      %offset( 1, eT_ghostFromFinerElem )
      ! find out the treeId of the only source element from which we will   
      ! copy the pdf values
      sourceElem = glob%levelDesc( targetLevel )%depFromFiner( iElem )         &
        &                                       %elem%val(1)

      ! Now that we know the target element, loop over all nVars and
      ! discrete velocity directions for each field variable 
      do iVar = 1, varSys%nVars
        do iDir = 1, layout%stencil(1)%QQ
          pos = varSys%Variable%val( iVar )%varPos( iDir )
          ! ... and simply copy the pdf from source to target.
          pdf( level )%state( &
         & ?SAVE?( pos, targetElem, varSys%nScalars, pdf(targetLevel)%neigh ), &
         & t_target ) = pdf( sourceLevel )%state(                              &
         & ?SAVE?( pos, sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh ), &
         & t_source )
        enddo
      enddo
    enddo
  end subroutine fillMyGhostsFromFiner_CopyFirst
!******************************************************************************!


!******************************************************************************!
  !> This simple method does not perform actual interpolation, but simply copies
  !! one of the source element's pdf values to the target element. It is mostly
  !! there for debugging purposes, the quality is not good enough for
  !! production use! 
  !!
  subroutine fillFinerGhostsFromMe_CopyFirst( pdf, glob, fieldProp, level,     &
    &                           scheme, layout, intp, ind, varSys, derVarPos ) 
    !---------------------------------------------------------------------------
    !> global fluid parameters 
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !> global parameters 
    type( pdf_global_type ), intent(inout) :: glob
    !> the current scheme 
    type( mus_scheme_type ), intent(in)   :: scheme
    !> fluid and species parameters 
    type( mus_field_prop_type ), intent(in) ::  fieldProp(:)
    !> level
    integer, intent(in)   :: level
    !> the layout used
    type( mus_scheme_layout_type), intent(in)   :: layout
    !> interpolation method info
    type( interpolation_type ), intent(in)   :: intp
    !> indirection list
    type( grw_intArray_type ), intent(in)   :: ind
    !> global variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> required variable system which maps to global varsys
    type( mus_derVarPos_type ), intent(in) :: derVarPos(:)
    !---------------------------------------------------------------------------
    integer :: sourceLevel      ! level of source elements
    integer :: sourceElem       ! treeId of current source element
    integer :: targetLevel      ! level of target elements
    integer :: targetElem       ! treeId of current source element
    integer :: iDir             ! current direction (discrete velocity) for loop
    integer :: iElem            ! current target element (for outer loop)
    integer :: indElem          ! element counter for indirection list
    integer :: pos  ! position of field component in the state vector
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to work on?
    integer :: t_source ! which time layer to work on?
    !--------------------------------------------------------------------------
    sourceLevel = level
    targetLevel = level + 1
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))

    ! Treat every fine target element:
    !do iElem = 1, pdf( targetLevel )%nElems_ghostFromCoarser 
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      ! Find out treeIds of current target element and the single source 
      ! element that we copy from
      sourceElem = glob%levelDesc( targetLevel )%depFromCoarser( iElem )       &
        &                                       %elem%val( 1)
      targetElem = iElem + glob%levelDesc( targetLevel )                       &
        &                      %offset( 1, eT_ghostFromCoarserElem )

      ! Now that we know the target element, loop over all nVars and
      ! discrete velocity directions for each field variable 
      do iVar = 1, varSys%nVars
        do iDir = 1, layout%stencil(1)%QQ
          pos = varSys%Variable%val( iVar )%varPos( iDir )
            ! ... and simply copy the pdf from source to target.
          pdf( targetLevel )%state(                                            &
         & ?SAVE?( pos, targetElem, varSys%nScalars, pdf(targetLevel)%neigh),  &
         & t_target ) = pdf( sourceLevel )%state(                              &
         & ?SAVE?( iDir, sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh), &
         & t_source )
        enddo
      enddo
    enddo

  end subroutine fillFinerGhostsFromMe_CopyFirst
!******************************************************************************!


end module mus_interpolate_debug_module
!******************************************************************************!

