! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Interpolation header to load confiugration and type definition
!! \author Manuel Hasert
?? include 'header/lbm_macros.inc'
module mus_interpolate_header_module

  ! include aotus modules
  use aotus_module,     only: flu_State, aot_get_val
  use aot_table_module, only: aot_table_open, aot_table_close, aot_table_length

  ! include treelm modules
  use env_module,            only: rk, labelLen
  use tem_aux_module,        only: tem_abort
  use tem_tools_module,      only: tem_write
  use tem_geometry_module,   only: tem_baryOfId, tem_intp_bilinear,            &
    &                              tem_intp_trilinear
  use tem_grow_array_module, only: grw_intArray_type
  use treelmesh_module,      only: treelmesh_type
  use tem_var_system_module, only: tem_var_system_type
  use tem_param_module,      only: cs2inv, cs2
  use tem_math_module,       only: matrix_type
  use tem_logging_module,    only: tem_log, tem_toStr

  ! include musubi modules
  use mus_field_prop_module,    only: mus_field_prop_type 
  use mus_global_module,        only: param_type
  use mus_pdf_module,           only: pdf_data_type, pdf_global_type,          &
    &                                 pdf_globInfo_type
  use mus_scheme_layout_module, only: mus_scheme_layout_type
  use mus_scheme_type_module,   only: mus_scheme_type
  use mus_moments_module,       only: intf_get_MomentFromPdf,                  &
    &                                 intf_get_PdfFromMoment
  use mus_variable_module,      only: mus_derVarPos_type

  implicit none

  private

  public :: interpolation_type
  public :: mus_load_interpolate
  public :: determine_timeLayer
  public :: refinement_type

  !> refinement (coarse to fine or fine to coarse) information
  type refinement_type
    !> original matrix of the linear system of equation
    type(matrix_type), allocatable :: matrix(:)
    !> inverted matrix to solve linear system of equation
    type(matrix_type), allocatable :: matrixInv(:)
  end type refinement_type

  !> definition of the used interpolation method
  type interpolation_type
    !> name of this interpolation method
    character(len=labelLen) :: label
    !> debug interpolation fillup value
    real(kind=rk) :: debugValue = 1.0_rk
    !> Interpolation test by comparing against the initial condition
    logical :: testInterpolation = .false.
    logical :: testEachElement   = .false.
    logical :: testFluids  = .false.
    logical :: requireWeights = .false.
    logical :: noIntpFromFiner  = .false.
    logical :: noIntpFromCoarser= .false.
    
    !> high order interpolation fine to coarse for ghostFromFiner elements
    procedure( interpolateRoutine ),nopass,pointer :: fillMineFromFiner =>     &
      &                                                                  null() 
    !> low order interpolation fine to coarse for ghostFromFiner elements
    procedure( interpolateRoutine ),nopass,pointer :: fillMineFromFiner_low => &
      &                                                                  null() 
    !> high order interpolation coarse to fine for ghostFromCoarser elements
    procedure( interpolateRoutine ),nopass,pointer :: fillFinerFromMe => null()
    !> low order interpolation coarse to fine for ghostFromCoarser elements
    procedure( interpolateRoutine ),nopass,pointer :: fillFinerFromMe_low =>   &
      &                                                                  null() 

    !> how many source elements are required by interpolation scheme
    integer :: nSourcesFromCoarser_required
    integer :: nSourcesFromFiner_required
    !> how many source elements are required for compact interpolation
    integer :: nSourcesCompactFromCoarser_required
    integer :: nSourcesCompactFromFiner_required

    !> Which time layer to fill
    integer :: timeLayer
    !> Interpolation matrix definition for ghostfromFiner
    type(refinement_type), allocatable :: fromFiner(:)
    !> Interpolation matrix definition for ghostfromCoarser
    type(refinement_type), allocatable :: fromCoarser(:)

    !> function pointer to transform the pdf to moment space
    procedure( intf_get_MomentFromPdf ),nopass,pointer :: get_moment => null() 
    !> function pointer to transform the moment space to pdfs
    procedure( intf_get_PdfFromMoment ),nopass,pointer :: get_pdf => null() 
    procedure( solve_linearSystem ),nopass,pointer     :: solve => null() 

    !> maximum number of sources for the interpolation scheme
    integer :: nMaxSources = 0
    !> is the interpolation scheme acting on a compact stencil
    logical :: compact
    !> order of the interpolatino scheme
    integer :: order
    !> require moment calculation?
    logical :: requireMoments = .false.
    !> pointing indices from the compact directions to the stencil directions
    !! for all eight cases.
    !! size: 1st dim: stencil(1)%QQ 2nd dim: nCases = 8
    integer, allocatable :: compactInd(:,:)
  end type interpolation_type

!******************************************************************************!
  abstract interface
    !> This is the interface for all interpolation methods that 
    !  can be called from outside.
    subroutine interpolateRoutine(  pdf, glob, fieldProp, level, scheme,       &
      &                   layout, intp, ind, varSys, derVarPos )
      import :: pdf_data_type, pdf_global_type, mus_field_prop_type,           &
        & mus_scheme_layout_type, interpolation_type, tem_var_system_type,     &
        & mus_derVarPos_type, grw_intArray_type, mus_scheme_type
      !> pdf state arrays
      type( pdf_data_type ), intent(inout) :: pdf(:) 
      !> pdf state properties
      type( pdf_global_type ), intent(inout) :: glob 
      !> fluid and species parameters
      type( mus_field_prop_type ), intent(in) ::  fieldProp(:)
      !> the layout used
      type( mus_scheme_layout_type ), intent(in)   :: layout
      !> the current scheme 
      type( mus_scheme_type ), intent(in)   :: scheme
      !> interpolation method info
      type( interpolation_type ), intent(in)   :: intp
      !> indirection list
      type( grw_intArray_type ), intent(in)   :: ind
      !> my refinement level
      integer,intent(in) :: level
      !> global variable system
      type( tem_var_system_type ), intent(in) :: varSys
      !> required variable system which maps to global varsys
      type( mus_derVarPos_type ), intent(in) :: derVarPos(:)
    end subroutine interpolateRoutine
  
    !> Function to calculate the pdf from moments
    function solve_linearSystem( mom, targetCoord, layout, matrix, nVals )     &
      &                                                           result( res )
      import :: mus_scheme_layout_type, rk
      !> Stencil layout
      type( mus_scheme_layout_type ), intent(in) :: layout
      !> target coordinates
      real(kind=rk), intent(in) :: targetCoord(3)
      !> moment input
      real(kind=rk), intent(in) :: mom( layout%stencil(1)%QQ )
      !> number of resulting values
      integer, intent(in) :: nVals
      !> Transformation matrix
      real(kind=rk), intent(in), optional :: matrix( layout%stencil(1)%QQ,     &
        &                                                layout%stencil(1)%QQ )
      !> results obtained by the solution of linear system
      real(kind=rk) :: res( nVals )
    end function
  end interface
!******************************************************************************!

  contains

!******************************************************************************!
  !>
  !!
  function determine_timeLayer( reqTimeLayer, pdf ) result( tLayer )
    !---------------------------------------------------------------------------
    !> global fluid parameters 
    type( pdf_data_type ), intent(inout) :: pdf
    !> which time layer is requested? 
    !!   1 = Next as in kernel (default)
    !!   2 = Now for updating second time layer
    integer :: reqTimeLayer
    !> which time layer to work on?
    integer :: tLayer
    !---------------------------------------------------------------------------

    if( reqTimeLayer == 2) then
      tLayer = pdf%nNow 
    else
      tLayer = pdf%nNext
    endif

  end function determine_timeLayer
!******************************************************************************!


!******************************************************************************!
  !> Read in the type of interpolation scheme
  !!
  !! \verbatim
  !! interpolation_method = 'linear'               -- simple definition
  !! interpolation_method = 'quadraticCompact'     -- simple definition
  !! interpolation_method = {'debug', value = 1.} -- definition in a table
  !! \endverbatim
  !!
  subroutine mus_load_interpolate( me, conf, pdfGlobal, parent, key )
    !---------------------------------------------------------------------------
    !> interpolation type to load info to
    type(interpolation_type),intent(out) :: me
    !> lua state to load from
    type(flu_state) :: conf
    !> global information
    type( pdf_globInfo_type ), intent(in) :: pdfGlobal
    !> optional parent table to load from
    integer, optional, intent(in) :: parent
    !> optional key
    character(len=*), optional, intent(in) :: key
    !---------------------------------------------------------------------------
    character(len=32) :: localKey
    integer :: iError, nEntries
    integer :: intp_handle
    !---------------------------------------------------------------------------
    if( present( key )) then
      localKey = key
    else
      localKey = 'interpolation_method'
    endif
    me%requireMoments = .false.
  
    ! if more than one variable is defined in a table then
    ! ref_value should also be a table with reference value for each variable
    ! defined in the variable table. 
    if( present( parent )) then 
      call aot_table_open( L=conf, thandle=intp_handle, key=localKey,          &
        &                  parent= parent )
    else
      call aot_table_open( L=conf, thandle=intp_handle, key=localKey)
    endif
      
    ! not a table, just the name of the interpolation method
    if (intp_handle .eq. 0) then                     
      call aot_get_val(L=conf, key=localKey,                                   &
        &              val=me%label,                                           &
        &              ErrCode=iError, default='quadratic')
    else 
      ! interpolation is defined as a table with optional additional information
      ! Get size of table
      nEntries = aot_table_length(L=conf, thandle=intp_handle)
      call aot_get_val(L=conf, thandle=intp_handle,                            &
        &              val=me%label, ErrCode=iError,                           &
        &              key = 'method', default='quadratic')

      call aot_get_val(L=conf, thandle=intp_handle,                            &
        &              val=me%testInterpolation, ErrCode=iError,               &
        &              key = 'test', default=.false.)
      if( me%testInterpolation ) then
        call tem_write(' Activated the Interpolation test.')
        me%requireMoments = .true.
        call aot_get_val(L=conf, thandle=intp_handle,                          &
          &              val=me%testEachElement, ErrCode=iError,               &
          &              key = 'testEach', default=.false.)
        if( me%testEachElement   ) then
        call tem_write(' Testing each element... ')
        end if
        call aot_get_val(L=conf, thandle=intp_handle,                          &
          &              val=me%testFluids, ErrCode=iError,                    &
          &              key = 'testFluids', default=.false.)
        if( me%testFluids ) then
        call tem_write(' Testing fluid element... ')
        end if
      end if
      call aot_get_val(L=conf, thandle=intp_handle,                            &
        &              val=me%noIntpFromFiner, ErrCode=iError,                 &
        &              key = 'noIntpFromFiner', default=.false.)

      call aot_get_val(L=conf, thandle=intp_handle,                            &
        &              val=me%noIntpFromCoarser, ErrCode=iError,               &
        &              key = 'noIntpFromCoarser', default=.false.)

      if( trim(me%label) == 'debug' ) then
        call aot_get_val(L=conf, thandle=intp_handle,                          &
          &              val=me%debugValue, ErrCode=iError,                    &
          &              key = 'value', default=1.0_rk)
        call tem_write(' Interpolation value', real(me%debugValue ))
      endif
    endif 
    call aot_table_close(L=conf, thandle=intp_handle)    
    call tem_write( 'Level interpolation: '// trim( me%label ))

    ! Determine, how many elements are required
    me%nSourcesFromFiner_required   = 0
    me%nSourcesFromCoarser_required = 0
    me%compact = .true.
    select case (trim(me%label) )
      case('none')
        me%nSourcesFromFiner_required     = 0
        me%nSourcesFromCoarser_required   = 0
      case('debug')
        me%nSourcesFromFiner_required     = 1
        me%nSourcesFromCoarser_required   = 1
      case('compact')                    
        me%nSourcesFromFiner_required     = 1
        me%nSourcesFromCoarser_required   = 1
      case('copyfirst')                  
        me%nSourcesFromFiner_required     = 1
        me%nSourcesFromCoarser_required   = 1
      case('average')                    
        me%requireWeights = .true.
        me%nSourcesFromFiner_required     = 1
        me%nSourcesFromCoarser_required   = 1
      case('eqneq', 'linear')                      
        me%label = 'linear'
        me%requireWeights = .true.
        ! \todo: SZ: check if the moments are required??
        me%requireMoments = .true.
        me%nSourcesFromFiner_required     = 1
        me%nSourcesFromCoarser_required   = 1
      case('quadraticCompact')
        me%requireWeights = .true.
        me%requireMoments = .true.
        if( pdfGlobal%nDims == 2 ) then
          me%nSourcesFromFiner_required   = 4
          me%nSourcesFromCoarser_required = 4
        elseif( pdfGlobal%nDims == 3) then
          me%nSourcesFromFiner_required   = 8
          me%nSourcesFromCoarser_required = 7
        elseif( pdfGlobal%nDims == 1) then
          me%nSourcesFromFiner_required   = 2
          me%nSourcesFromCoarser_required = 2
        endif
      case('quadratic', 'quadraticExtended','quadraticReduced')
        me%requireMoments = .true.
        if( trim(me%label) == 'quadraticExtended') then 
          me%label = 'quadratic'
        endif
        me%compact = .false.
        if( pdfGlobal%nDims == 1 ) then
          me%nSourcesFromFiner_required   = 2
          me%nSourcesFromCoarser_required = 3
        elseif( pdfGlobal%nDims == 2 ) then
          me%nSourcesFromFiner_required   = 4
          me%nSourcesFromCoarser_required = 9
        elseif( pdfGlobal%nDims == 3) then
          me%nSourcesFromFiner_required   = 8
          me%nSourcesFromCoarser_required = 19
        endif
!      default case
!        call tem_write(' No valid interpolation method selected')
    end select
    if( pdfGlobal%nDims == 2 ) then
      me%nSourcesCompactFromFiner_required     = 4
      me%nSourcesCompactFromCoarser_required   = 4
    elseif( pdfGlobal%nDims == 3) then
      me%nSourcesCompactFromFiner_required     = 8
      me%nSourcesCompactFromCoarser_required   = 7
    elseif( pdfGlobal%nDims == 1) then
      me%nSourcesCompactFromFiner_required     = 2
      me%nSourcesCompactFromCoarser_required   = 2
    endif    
    call tem_log( level = 1,                                                   &
      &           msg   = 'Setting for interpolation scheme: '//trim(me%label))
    call tem_log( level = 1,                                                   &
      &           msg   = '    from coarser intp: require #: '//               &
      &                   trim( tem_toStr( me%nSourcesFromCoarser_required )))
    call tem_log( level = 1,                                                   &
      &           msg   = '      from finer intp: require #: '//               &
      &                   trim( tem_toStr( me%nSourcesFromFiner_required )))
 
  end subroutine mus_load_interpolate
!******************************************************************************!


end module mus_interpolate_header_module
!******************************************************************************!
