! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Interpolation of flow quantities between different grid levels
!!
!! \section intp_main Interpolation 
!! The routines defined here, fill up the ghost elements with valid data.
!! Ghost elements are employed at grid level interfaces to provide valid
!! pdf values to the neighboring fluid elements. This way, the solvers can 
!! act on elements of the same size only, treating the levels successively.
!! Target elements are the ghost elements, which have to be filled with 
!! valid values.
!! Source elements are the fluid elements from other levels, from where to 
!! take the input values for the interpolation.
!! The target ghost elements on the target level have corresponding source 
!! fluid elements on the source level. 
!!
!! \ref tem_topology_module For a detailed description of the grid 
!! 
!! \section workflow Workflow
!! 
!! Each interpolation routine acts on a list of ghost elements.
!! This list contains pointers to the position in the total list.
!! For each of these ghost elements, the source elements are identified.
!! Before that, the sourceLevel is identified. However, the code is restricted
!! to work with a level jump of only one level, so the sourceLevel is 
!! for sourceLevel = targetLevel+1
!! sourceLevel = targetLevel-1
!!
!! For an overview over implemented interpolation methods, see
!! \ref intp_methods "Interpolation methods"
!!
!!\author Manuel Hasert
?? include 'header/lbm_macros.inc'
module mus_interpolate_module

  ! include treelm modules
  use env_module,            only: rk
  use treelmesh_module,      only: treelmesh_type
  use tem_aux_module,        only: tem_abort
  use tem_tools_module,      only: tem_write
  use tem_debug_module,      only: main_debug

  ! include musubi modules
  use mus_field_prop_module,            only: mus_field_prop_type
  use mus_global_module,                only: param_type
  use mus_interpolate_header_module,    only: interpolation_type
  use mus_interpolate_debug_module,     only: fillFinerGhostsFromMe_average,   &
    &                                         fillMyGhostsFromFiner_average,   &
    &                                         fillFinerGhostsFromMe_copyFirst, &
    &                                         fillMyGhostsFromFiner_copyFirst, &
    &                                         fillFinerGhostsFromMe_debug,     &
    &                                         fillMyGhostsFromFiner_debug,     &
    &                                         do_nothing
  use mus_interpolate_linear_module,    only:                                  &
    &                                   fillFinerGhostsFromMe_linear,          &
    &                                   fillMyGhostsFromFiner_linear,          &
    &                                   fillFinerGhostsFromMe_linearDiffusive, &
    &                                   fillMyGhostsFromFiner_linearDiffusive, &
    &                                   fillMyGhostsFromFiner_linearIncomp,    &
    &                                   fillFinerGhostsFromMe_linearIncomp,    &
    &                                   fillMyGhostsFromFiner_linMoment3d,     &
    &                                   fillMyGhostsFromFiner_linMoment2d
  use mus_interpolate_compact_module,   only:                                  &
    &                                fillFinerGhostsFromMe_quadraticCompact2d, &
    &                                fillMyGhostsFromFiner_quadraticCompact2d, &
    &                                fillFinerGhostsFromMe_quadraticCompact3d, &
    &                                fillMyGhostsFromFiner_quadraticCompact3d
  use mus_interpolate_quadratic_module, only:                                  &
    &                              fillFinerGhostsFromMe_quadratic2d,          &
    &                              fillFinerGhostsFromMe_quadratic2dIncomp,    &
    &                              fillFinerGhostsFromMe_quadratic2dDiffusive, &
    &                              fillFinerGhostsFromMe_quadratic3d,          &
    &                              fillFinerGhostsFromMe_quadratic3dIncomp,    &
    &                              fillFinerGhostsFromMe_quadratic3dDiffusive, &
    &                              mus_init_intpMatrices
  use mus_pdf_module,                   only: pdf_global_type,                 &
    &                                         pdf_globInfo_type
  use mus_scheme_layout_module,         only: mus_scheme_layout_type
  use mus_scheme_type_module,           only: mus_scheme_type

  implicit none

  private

  public :: mus_init_interpolate

  contains
!******************************************************************************!
  !> This subroutine initialzes the interpolation
  !!
  !! - setting the fillMineFromFiner and fillFinerFromMe function pointers
  !!   to the interpolation function chosen by the user and
  !! - pre-calculating the weights based on the distances between source
  !!   and target nodes
  !! - allocate buffers if required
  !! - define interpolation order
  !!
  !! \ref intp_main "Interpolation algorithm"
  !!
  subroutine mus_init_interpolate( intp, fieldProp, glob, tree,                &
    &                              scheme, pdfGlobal, layout, params)
    !---------------------------------------------------------------------------
    !> interpolation type
    type(interpolation_type), intent(inout) :: intp
    !> fluid and species parameters 
    type( mus_field_prop_type ), intent(in) ::  fieldProp(:)
    !>
    type( mus_scheme_type ), intent(in) ::  scheme
    !> fluid parameters 
    type( pdf_global_type ), intent(inout) :: glob
    !> treelm mesh 
    type( treelmesh_type ), intent(in) :: tree
    !> simulation info 
    type( pdf_globInfo_type ), intent(in) :: pdfGlobal
    !> the stencil layout representation
    type( mus_scheme_layout_type ), intent(in) :: layout
    !>
    type( param_type ), intent(in) :: params
    !---------------------------------------------------------------------------
    integer :: iLevel, iField, nVals
    logical :: require_buffer
    !---------------------------------------------------------------------------
    ! set the theoretical maximum of source elements
    intp%nMaxSources = layout%stencil(1)%QQ
    require_buffer = .false.
    select case( trim(params%scaling))
    case( 'acoustic')
      if( trim(scheme%header%kind) .eq. 'lbm_incomp' ) then
        intp%fillFinerFromMe_low     => fillFinerGhostsFromMe_linearIncomp
        intp%fillMineFromFiner_low   => fillMyGhostsFromFiner_linearIncomp
      else
        intp%fillFinerFromMe_low     => fillFinerGhostsFromMe_linear
        intp%fillMineFromFiner_low   => fillMyGhostsFromFiner_linear
      end if
    case( 'diffusive')
      intp%fillFinerFromMe_low     => fillFinerGhostsFromMe_linearDiffusive
      intp%fillMineFromFiner_low   => fillMyGhostsFromFiner_linearDiffusive
    end select

    call tem_write( 'Setting up the interpolation method ...'//trim(intp%label))
    call tem_write( '        nMaxSourceElems:   ', intp%nMaxSources )
    select case (trim( intp%label ))
    case('none')
      call tem_write('   do nothing + do nothing ')
      intp%fillMineFromFiner => do_nothing
      intp%fillFinerFromMe   => do_nothing
      intp%order = -1
    case('debug')
      call tem_write('   debug + debug ')
      intp%fillMineFromFiner => fillMyGhostsFromFiner_Debug
      intp%fillFinerFromMe   => fillFinerGhostsFromMe_Debug
      intp%order = 0
    case('compact', 'compactfc')
      call tem_write('   average frF + copyfirst frC')
      intp%fillMineFromFiner => fillMyGhostsFromFiner_Average
      intp%fillFinerFromMe   => fillFinerGhostsFromMe_CopyFirst   
      intp%order = 0
    case('compactff')
      call tem_write('   copyfirst frF + average frC')
      intp%fillMineFromFiner => fillMyGhostsFromFiner_CopyFirst
      intp%fillFinerFromMe   => fillFinerGhostsFromMe_Average
      intp%order = 0
    case('copyfirst')
      call tem_write('   copyfirst + copyfirst ')
      intp%fillMineFromFiner => fillMyGhostsFromFiner_CopyFirst 
      intp%fillFinerFromMe   => fillFinerGhostsFromMe_CopyFirst   
      intp%order = 0
    case('average')
      call tem_write('   average + average ')
      intp%fillMineFromFiner => fillMyGhostsFromFiner_Average
      intp%fillFinerFromMe   => fillFinerGhostsFromMe_Average
      intp%order = 0
    case('check')
      call tem_write('   average + FF: lin ')
      intp%fillMineFromFiner => fillMyGhostsFromFiner_linearIncomp
      intp%fillFinerFromMe   => fillFinerGhostsFromMe_Average
      intp%order = 0
    case('linear')
      select case( trim(params%scaling))
      case( 'acoustic')
        if( trim(scheme%header%kind) .eq. 'lbm_incomp' ) then
          call tem_write('   linearIncomp + linearIncomp')
          intp%fillMineFromFiner => fillMyGhostsFromFiner_linearIncomp
          intp%fillFinerFromMe   => fillFinerGhostsFromMe_linearIncomp
        else
          call tem_write('   linear + linear')
          intp%fillMineFromFiner => fillMyGhostsFromFiner_linear
          intp%fillFinerFromMe   => fillFinerGhostsFromMe_linear
        end if
      case( 'diffusive')
        call tem_write('   linearDiffusive + linearDiffusive')
        intp%fillMineFromFiner => fillMyGhostsFromFiner_linearDiffusive
        intp%fillFinerFromMe   => fillFinerGhostsFromMe_linearDiffusive
      end select
      intp%order = 1
    case('quadraticCompact','quadratic')
      intp%order = 2
      require_buffer = .true.
      if( trim( intp%label ) == 'quadraticCompact') then
        call tem_write('   compact quadratic + compact quadratic ')
        if( pdfGlobal%nDims == 2 ) then
          intp%fillMineFromFiner => fillMyGhostsFromFiner_quadraticCompact2d
          intp%fillFinerFromMe   => fillFinerGhostsFromMe_quadraticCompact2d
        elseif( pdfGlobal%nDims == 3) then
          intp%fillMineFromFiner => fillMyGhostsFromFiner_quadraticCompact3d
          intp%fillFinerFromMe   => fillFinerGhostsFromMe_quadraticCompact3d
        endif
      else
        call tem_write('   extended quadratic + linear interpolation')
        select case( trim(params%scaling ))
        case('acoustic') 
          if( pdfGlobal%nDims == 2 ) then
            if( trim(scheme%header%kind) .eq. 'lbm_incomp' ) then
              call tem_write(' Choosing incompressible interpolation')
              intp%fillMineFromFiner => fillMyGhostsFromFiner_linearIncomp
              intp%fillFinerFromMe   => fillFinerGhostsFromMe_quadratic2dIncomp
            else
              call tem_write(' Choosing linear 2d + quadratic 2d ')
              intp%fillMineFromFiner => fillMyGhostsFromFiner_linMoment2d
              intp%fillFinerFromMe   => fillFinerGhostsFromMe_quadratic2d
            end if
          elseif( pdfGlobal%nDims == 3) then
            if( trim(scheme%header%kind) .eq. 'lbm_incomp' ) then
              call tem_write(' Choosing incompressible interpolation')
              intp%fillMineFromFiner => fillMyGhostsFromFiner_linearIncomp
              intp%fillFinerFromMe   => fillFinerGhostsFromMe_quadratic3dIncomp
            else
              call tem_write(' Choosing linear 3d + quadratic 3d ')
              intp%fillMineFromFiner => fillMyGhostsFromFiner_linMoment3d
              intp%fillFinerFromMe   => fillFinerGhostsFromMe_quadratic3d
            endif
          endif
        case('diffusive')
          call tem_write('Diffusive scaling quadratic interpolation')
          if( pdfGlobal%nDims == 2 ) then
            intp%fillMineFromFiner => fillMyGhostsFromFiner_linearDiffusive
            intp%fillFinerFromMe   => fillFinerGhostsFromMe_quadratic2dDiffusive
          elseif( pdfGlobal%nDims == 3) then
            intp%fillMineFromFiner => fillMyGhostsFromFiner_linearDiffusive
            intp%fillFinerFromMe   => fillFinerGhostsFromMe_quadratic3dDiffusive
          endif
        end select
      endif
      ! Prepare the inversion matrices
      allocate( intp%fromFiner(   pdfGLobal%minLevel:pdfGLobal%maxLevel ))
      allocate( intp%fromCoarser( pdfGLobal%minLevel:pdfGLobal%maxLevel ))
      !\todo iField loop!
      iField = 1
      call tem_write('   Init interpolation matrices ...')
      do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
        if( iLevel .lt. pdfGlobal%maxLevel ) then
        call mus_init_intpMatrices(                                            &
          &           me          = intp%fromCoarser( iLevel+1 ),              &
          &           intp        = intp,                                      &
          &           layout      = layout,                                    &
          &           dep         = glob%levelDesc( iLevel+1 )%depFromCoarser, &
          &           glob        = glob,                                      &
          &           fieldProp   = fieldProp( iField ),                       &
          &           compact     = intp%compact,                              &
          &           debug       = main_debug%debugDependencies,              &
          &           targetLevel = iLevel+1,                                  &
          &           sourceLevel = iLevel )                            
        endif
        if( iLevel .lt. pdfGlobal%maxLevel ) then 
        call mus_init_intpMatrices(                                            &
          &                me          = intp%fromFiner( iLevel ),             &
          &                intp        = intp,                                 &
          &                layout      = layout,                               &
          &                dep         = glob%levelDesc( iLevel )%depFromFiner,&
          &                glob        = glob,                                 &
          &                fieldProp   = fieldProp( iField ),                  &
          ! Always do a compact interpolation fromFiner.
          ! accuracy seems to be enough.
          &                compact     = .true.,                               &
          &                debug       = main_debug%debugDependencies,         &
          &                targetLevel = iLevel,                               &
          &                sourceLevel = iLevel+1 )
        endif
      enddo
      call tem_write('   Done.')
    case default 
      call tem_write('Interpolation method '//trim(intp%label)//' not found')
      if( tree%global%minLevel .ne. tree%global%maxLevel ) then
        call tem_write('This is fatal for multilevel. Stopping.')
        call tem_abort
      endif
    end select

    if( require_buffer ) then
      nVals = layout%stencil(1)%QQ
      do iLevel =  pdfGlobal%minLevel, pdfGlobal%maxLevel
        if( allocated( glob%levelDesc( iLevel )%depFromCoarser )) then
          if( size( glob%levelDesc( iLevel )%depFromCoarser ) .gt. 0 ) then
            allocate( glob%levelDesc( iLevel )%intpBufFromCoarser(   &
              &       nVals, glob%levelDesc( iLevel )%sourceFromCoarser%nVals ))
          endif
        endif
        if( allocated( glob%levelDesc( iLevel )%depFromFiner )) then
          if( size( glob%levelDesc( iLevel )%depFromFiner ) .gt. 0 ) then
            allocate( glob%levelDesc( iLevel )%intpBufFromFiner(   &
              &       nVals, glob%levelDesc( iLevel )%sourceFromFiner%nVals ))
          endif
        endif
      enddo
    endif

    if( intp%noIntpFromFiner ) then
      intp%fillMineFromFiner => do_nothing
    end if
    if( intp%noIntpFromCoarser ) then
      intp%fillFinerFromMe   => do_nothing
    end if

    call tem_write( 'Done with interpolation method.')

  end subroutine mus_init_interpolate
!******************************************************************************!


end module mus_interpolate_module
!******************************************************************************!
!> \page intp_methods Interpolation methods
!!
!! \section intp_average Averaging interpolation
!!\verbatim interpolation_method = 'linear'
!!\endverbatim
!! Simple averaging of all quantities with weighted summation
!! \section intp_linear Linear interpolation
!! Linear interpolation of the first two moments (density and momentum) and 
!! separation of equilibrium and non-equilibrium part to account for correct 
!! shear stress calculation
!! This fine->coarse interpolation method is based on Dazhi Yu et. al. (2002):
!! "A multi-block lattice Boltzmann method for viscous fluid flows".
!! The basic idea is to split up the source pdfs into equilibrium- and 
!! non-equilibrum- parts, then interpolate both parts with average interpolation,
!! and apply a corrective scaling factor to the non-equilibrium part before
!! adding both parts up and writing it to the target element.
!! Values are interpolated in a linear fashion by multiplying the quantities
!! of the source elements by weighting factors, which were determined in 
!! \link mus_construction_module::mus_intp_complete_fromcoarserdep  completing the
!! fromCoarser dependencies \endlink
!! in terms of the distance of the source element's barycenters from the
!! target element's barycenters.
!! 
!! \section intp_quadratic Quadratic interpolation
!! Quadratic interpolation based on the compute stencil 
!! \section intp_quadratic_compact Quadratic interpolation with compact stencil
!! Quadratic interpolation of the moment field with derivative information from 
!! the shear stress moments. Only a compact stencil is required.
!!
