! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Auxiliary functionality 
!!
!! \author Manuel Hasert 
!!
module mus_aux_module

  ! include musubi modules
  use mus_comm_module,               only: mus_exchange
  use mus_pdf_module,                only: pdf_globInfo_type
  use mus_global_module,             only: param_type
  use mus_geom_module,               only: mus_geom_type 
  use mus_tools_module,              only: perform_checks, mus_time_update,    &
    &                                      check_streaming_layout,             &
    &                                      mus_time_homogenize
  use mus_tracking_module,           only: tracker
  use mus_restart_module,            only: mus_readRestart, mus_writeRestart
  use mus_bc_general_module,         only: init_boundary
  use mus_interpolate_header_module, only: interpolation_type
  use mus_interpolate_module,        only: mus_init_interpolate
  use mus_interpolate_verify_module, only: mus_testInterpolation
  use mus_moments_module, only: mus_init_moments
  use mus_scheme_module, only: mus_init_sources
  use mus_scheme_type_module, only: mus_scheme_type
  use mus_scheme_header_module, only: mus_scheme_header_type
  use mus_geomIncr_module, only: mus_geomIncr, mus_init_geomIncr
  use mus_geom_module, only: mus_geom_setWeight
  use mus_interpolate_tools_module, only: debug_dependencies, dump_intpLists

  ! include treelm modules
  use env_module,             only: rk
  use treelmesh_module,       only: treelmesh_type, unload_tem
  use tem_aux_module,         only: tem_unit_close
  use tem_balance_module,     only: tem_balance_evaluateImbalance
  use tem_comm_env_module,    only: process
  use tem_param_module,       only: cs2inv, cs2
  use tem_math_module,        only: invert_matrix
  use tem_restart_module,     only: tem_init_restart, tem_restart_finalize
  use tem_spatial_module,     only: tem_spatial_for
  use tem_tools_module,       only: tem_write, tem_horizontalSpacer
  use tem_balance_module,     only: tem_balance
  use tem_tracking_module,    only: tem_init_tracker, tem_tracking_finalize
  use tem_transient_module,   only: tem_transient_for
  use tem_timeControl_module, only: tem_time_activeTimeStep,                   &
    &                               tem_time_updateTrigger,                    &
    &                               tem_time_determineEnd,                     &
    &                               tem_timeControlToInt
  use tem_timer_module,       only: timerData, tem_addTimer
  use tem_debug_module,       only: main_debug, dbgUnit
  use tem_status_module,      only: tem_stat_run_terminate
  use tem_logging_module,     only: tem_log, tem_toStr

  ! include aotus modules
  use aotus_module, only: flu_State, aot_get_val, close_config

  implicit none
  private

  public :: check_flow_status
  public :: mus_init_aux
  public :: mus_finalize
  public :: mus_setParameters
  public :: mus_balance

contains

!******************************************************************************!
  !> Check, output and write restart file at each time step
  subroutine check_flow_status( scheme, params, geometry, pdfGlobal)
    !--------------------------------------------------------------------------
    !> containers for the different schemes
    type(mus_scheme_type), intent(inout)   :: scheme(:)
    !> global flow quantity properties
    type(pdf_globInfo_type), intent(inout) :: pdfGlobal
    !> geometry infomation
    type(mus_geom_type),intent(in)         :: geometry
    !> global parameters
    type(param_type),intent(inout)         :: params
    !--------------------------------------------------------------------------
    integer :: iScheme !< scheme counter
    !--------------------------------------------------------------------------

    ! Call the geometry increment module which  performs solidification 
    ! or liquification based on certain criteria read from the Lua file
    if( geometry%dynamicGeom ) then
      call mus_geomIncr(geometry = geometry, scheme = scheme, pdfGlobal = pdfGlobal, &
        &               commPattern = params%commPattern)
    end if ! Time for geom incr
    ! MH: Currently using the dt from the first scheme \todo add correct dt. Which one?
    ! update the time counters. MH: checked. Please dont move
    ! Increasing with the smallest time step (maxLevel)
    call mus_time_update( me = pdfGlobal%timing,                               &
      &           dt = scheme(1)%physics%dtLvl( geometry%tree%global%maxLevel ))

    do iScheme = 1, pdfGlobal%nSchemes
      if( scheme( iScheme )%trackingControl%active ) then
        call tracker( scheme = scheme(iScheme), pdf = scheme(iScheme)%pdf,     &
          &           track = scheme(iScheme)%track,                           &
          &           params = params, layout = scheme(iScheme)%layout,        &
          &           glob_tree = geometry%tree,                               &
          &           pdfGlobal = pdfGlobal )
      end if

      ! Perform run-time checks for total density, NaNs, stop file, etc.
      call perform_checks( pdfGlobal, params, scheme( iScheme ), iScheme )
     
    end do ! iScheme

    if( params%restart%controller%writeRestart ) then
      if( tem_time_activeTimeStep( timing = pdfGlobal%timing,                 &
       &         timeControl = params%restart%controller%timeControl,         &
       &         reqInt = params%scaleFactor**(pdfGlobal%maxLevel - pdfGlobal%minLevel ))&
       &  .or.                                                                &
       ! if the code terminates soon ...
       &  (params%stat%bits(tem_stat_run_terminate) .and. (                 &
       ! ... then make sure it is really the last iteration due to recursiveness
       &   mod( pdfGlobal%timing%nIters,   &
       &   params%scaleFactor**(pdfGlobal%maxLevel - pdfGlobal%minLevel )) .eq. 0)) &
       &  .or.                                    &
       &  tem_time_determineEnd( pdfGlobal%timing, params%timeControl )) then
        call tem_write( 'Writing restart...')
        call tem_write( '        Restart iTime', pdfGLobal%timing%iTime)
        call tem_write( '        Restart tTime', pdfGLobal%timing%rTime)
        call mus_writeRestart( pdfGlobal = pdfGlobal,                          &
          &                    restart   = params%restart,                     &
          &                    scheme    = scheme,                             &
          &                    tree      = geometry%tree,                      &
          &                    timing    = pdfGlobal%timing )
      end if
      call tem_time_updateTrigger(me  = params%restart%controller%timeControl, &
        &                         now = pdfGlobal%timing)
    end if ! write Restart

  end subroutine check_flow_status
!******************************************************************************!



!******************************************************************************!
  !> This routine performs the load balancing for multilevel simulations. The
  !! weights are calculated on the basis of levelwise run time, which are then
  !! fed to sparta for calculation of splitting positions. Restart files are 
  !! saved and the simulation is restarted with the newly distributed mesh
  subroutine mus_balance( scheme, pdfGlobal, params, geometry)
    !---------------------------------------------------------------------------
    !> The scheme type
    type(mus_scheme_type), intent(inout)   :: scheme(:)
    !> The current scheme index
!    integer, intent(in)                    :: iScheme
    !> global flow quantity properties
    type(pdf_globInfo_type), intent(inout) :: pdfGlobal
    !> global parameters
    type(param_type), intent(inout)        :: params
    !> geometry infomation
    type(mus_geom_type),intent(inout)      :: geometry
    !---------------------------------------------------------------------------
    integer :: iLevel
    !---------------------------------------------------------------------------

    if( .not. allocated(params%level_weight) ) &
      & allocate ( params%level_weight(pdfGlobal%minLevel:pdfGlobal%maxLevel) )
    if( .not. allocated(params%level_weight_bnd) ) &
      & allocate ( params%level_weight_bnd(pdfGlobal%minLevel:pdfGlobal%maxLevel) )

    ! Here weights are assigned depending on user request. In case of level wise
    ! weights, only levelwise weights are calculated whereas in case of
    ! boundary, the weights are calculated which are based on boundary time and
    ! then passed on to set weight routine for special treatment

    select case( trim(params%balance%balanceKind) )
    case( 'levelwise' )
      ! Interpolation times for different levels were recorded earlier
      ! Total interpolation time at each level is to be divided by 
      ! the number of elements at that level which will serve as weights
      do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
        if( scheme(1)%glob%levelDesc(iLevel)%nElems_fluid .gt. 0) then
          params%level_weight(iLevel) = ( params%timer(iLevel)%value(params%   &
            &                 lw_tIntp(iLevel) )                               &
            &               + params%timer(iLevel)%value(params%               &
            &                 lw_tCompute(iLevel) )                            & 
            &               / scheme( 1 )%glob%levelDesc(iLevel)%nElems_fluid ) 
        end if
        write(*,*) 'weights', params%level_weight(iLevel)
      end do
    case( 'boundary' )
      ! In case of boundaries, level weights are calculated as above in addition
      ! to the specific boundary weights
      do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
        if( scheme(1)%glob%levelDesc(iLevel)%nElems_fluid .gt. 0) then
          params%level_weight(iLevel) = ( params%timer(iLevel)%value(params%   &
            &                 lw_tIntp(iLevel) )                               &
            &               + params%timer(iLevel)%value(params%               &
            &                 lw_tCompute(iLevel) )                            & 
            &               / scheme( 1 )%glob%levelDesc(iLevel)%nElems_fluid ) 

          params%level_weight_bnd(iLevel) = params%timer(iLevel)%value(params% &
            &                               lw_tSetBnd(iLevel) )               &
            &               / scheme( 1 )%glob%levelDesc(iLevel)%nElems_fluid  
        end if
      end do
    case default
      call tem_write( ' Given load balancing not found in Musubi'//            &
        &             trim(params%balance%balanceKind ))
    end select

    ! Set the weights of each element according to levelwise weight obtained
    call mus_geom_setWeight ( params    = params,                              &
      &                       geometry  = geometry,                            &
      &                       balance   = params%balance,                      &
      &                       glob      = scheme(1)%glob,                      &
      &                       pdfGlobal = pdfGlobal )

    ! Call the sparta to get new splitting positions wrt new weights
    call tem_balance( weight = geometry%weight, myPart = process%rank,         &
      &               nParts = process%comm_size,                              &
      &               offset = geometry%offset, comm = process%comm )

    ! Write a restart file
    call tem_write( 'Writing restart for dynamic load balancing ...')
    call tem_write( '        Restart iTime', pdfGLobal%timing%iTime)
    call tem_write( '        Restart tTime', pdfGLobal%timing%rTime)
    call mus_writeRestart( pdfGlobal = pdfGlobal,                              &
      &                    restart   = params%restart,                         &
      &                    scheme    = scheme,                                 &
      &                    tree      = geometry%tree,                          &
      &                    timing    = pdfGlobal%timing )

    ! Unload the tree
    call unload_tem( me = geometry%tree )


  end subroutine mus_balance
!******************************************************************************!


!******************************************************************************!
  !> Init auxiliary features such as boundaries, restart and the tracker
  subroutine mus_init_aux( scheme, params, geometry, intp, pdfGlobal, dynamic )
    !---------------------------------------------------------------------------
    !> containers for the different schemes
    type(mus_scheme_type) :: scheme(:)
    !> geometry infomation
    type(mus_geom_type)   :: geometry
    !> global parameters
    type(param_type),intent(inout) :: params   
    !> Interpolation information
    type(interpolation_type) :: intp      
    !> global flow quantity properties
    type(pdf_globInfo_type) :: pdfGlobal
    logical, optional, intent(in) :: dynamic
    !---------------------------------------------------------------------------
    integer :: iScheme, iLevel, iTimeLayer, iTrack
    integer :: tLayerMin, tLayerMax
    logical :: dyn_loadbal
    !---------------------------------------------------------------------------
    ! Set the dyn_loadbal if dynamic load balancing is required
    if( present(dynamic) ) then
      dyn_loadbal = dynamic
    else
      dyn_loadbal = .false.
    end if

    ! verify correct settings for the streaming layout
    call check_streaming_layout( pdfGlobal )
    ! \todo dt is taken from the first scheme. There should be a global dt

    ! When the restart is read from separate restart table, we start the 
    ! simulation from the time step given in restart file. In the case,
    ! when restart is read from initial condition table, the simulation start
    ! time step is taken as the one defined in configuration file
    call mus_time_homogenize( me = pdfGlobal%timing,                           &
      &                       dt = scheme(1)%physics%dtLvl(                    &
      &                            geometry%tree%global%maxLevel ),            &
      &            fromRealToInt = (.not.params%timeControl                    &
      &                             %useIterations),                           &
      &              readRestart = params%restart%controller                   &
      &                             %readRestart )
    call tem_timeControlToInt( me = params%timeControl,                        &
      &                       dt = scheme(1)%physics%dtLvl(                    &
      &                            geometry%tree%global%maxLevel ))
    ! De-activate the check density for time step 1
    call tem_time_updateTrigger(me  = params%timeControl, &
      &                         now = pdfGlobal%timing)

    if( params%restart%controller%writeRestart  .or.                         &
        params%restart%controller%readRestart )  then

      ! initialize the restart
      call tem_write( 'Initializing restart...')
      call tem_init_restart( me      = params%restart,                         &
        &                    solver  = params%solver,                          &
        &                    nSys    = pdfGlobal%nSchemes,                     &
        &                    varSys  = scheme(:)%varSys,                       &
        &                    tree    = geometry%tree,                          &
        &                    dynamic = dynamic )
      call tem_timeControlToInt( me  = params%restart%controller%timeControl,  &
        &                        dt  = scheme(1)%physics%dtLvl(                &
        &                            geometry%tree%global%maxLevel ))
    end if

    !\todo: no general read option available in restart read
    if(params%restart%controller%readRestart .or. dyn_loadbal) then                             
      call tem_write( 'Reading restart...')
      call mus_readRestart( pdfGlobal = pdfGlobal,                             &
        &                   restart   = params%restart,                        &
        &                   scheme    = scheme,                                &
        &                   tree      = geometry%tree )
      call tem_write( 'Done reading restart.')
    end if !restart?

    ! allocate levelwise timer and handles, this will be used in 
    ! dynamic load  balancing
    if ( .not. dyn_loadbal ) then
      allocate (params%timer(pdfGlobal%minLevel:pdfGlobal%maxLevel) )
      allocate (params%lw_tIntp(pdfGlobal%minLevel:pdfGlobal%maxLevel) )
      allocate (params%lw_tCompute(pdfGlobal%minLevel:pdfGlobal%maxLevel) )
      allocate (params%lw_tSetBnd(pdfGlobal%minLevel:pdfGlobal%maxLevel) )
    end if

    ! scheme loop to initialize tracking, boundary, interpolation and output
    do iScheme = 1, pdfGlobal%nSchemes

      call mus_setSpatialOmega( scheme    = scheme( iScheme ),                 &
        &                       pdfGlobal = pdfGlobal,                         &
        &                       tree      = geometry%tree,                     &
        &                       params    = params )

      do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
        call mus_setParameters( scheme( iScheme ), pdfGlobal, iLevel, params )
        ! add a levelwise timer for load balancing
        call tem_addTimer( me          = params%timer(iLevel),                 &
          &                timerHandle = params%lw_tIntp(iLevel),              &
          &                timerName   = 'tIntp' )

        call tem_addTimer( me          = params%timer(iLevel),                 &
          &                timerHandle = params%lw_tCompute(iLevel),           &   
          &                timerName   = 'tCompute' )

        call tem_addTimer( me          = params%timer(iLevel),                 &
          &                timerHandle = params%lw_tSetBnd(iLevel),            &  
          &                timerName   = 'tSetBnd' )
      end do

      if( scheme( iScheme )%sourceActive ) then
        ! initialize the source terms for this field
        call mus_init_sources( me    = scheme(iScheme),                        &
          &                    tree  = geometry%tree )
      end if
      
      ! tracking objects
      call tem_write( 'Initializing tracker...')
      call tem_init_tracker( me = scheme(iScheme)%track,                       &
        &                    tCtrl = scheme(iScheme)%trackingControl,          & 
        &                    levelPointer = pdfGlobal%levelPointer,            &
        &                    tree = geometry%tree,                             &
        &                    solver = params%solver,                           &
        &                    varSys = scheme( iScheme )%varSys)
      do iTrack = 1, size( scheme( iScheme )%track )
        call tem_timeControlToInt( me = scheme( iScheme )%track( iTrack )%     &
          &                             header%timeControl,                    &
          &                       dt = scheme(1)%physics%dtLvl(                &
          &                            geometry%tree%global%maxLevel ))
      end do

      call tem_write( 'Initializing interpolation...')

      if( pdfGlobal%minLevel .ne. pdfGlobal%maxlevel ) then
        ! initialize the interpolation
        call mus_init_interpolate(                                             &
          &                  intp      = intp,                                 &
          &                  glob      = scheme(iScheme)%glob,                 &
          &                  scheme    = scheme( iScheme ),                    &
          &                  fieldProp = scheme( iScheme )%field(:)%fieldProp, &
          &                  tree      = geometry%tree,                        &
          &                  layout    = scheme( iScheme )%layout,             &
          &                  pdfGlobal = pdfGlobal, params = params )
      end if
      if( main_debug%debugDependencies) then
          call debug_dependencies( pdfGlobal, intp, scheme( iScheme )%layout,  &
            &                      scheme( iScheme )%glob, geometry%tree )  
          call dump_intpLists( pdfGlobal, intp, scheme( iScheme )%layout,      &
            &                  scheme( iScheme )%glob, geometry%tree )  
        end if

      ! Fill all elements (ghost, halo) with valid values from fluid elements
      ! for both time layers now and next
      tLayerMin = 1
      if( params%init_allTimeLayers ) then
        ! Initialize all time layers
        tLayerMax = 2
      else
        ! Initialize only nNext ( iTimeLayer = 1 )
        tLayerMax = 1
      end if

      do iTimeLayer = tLayerMax, tLayerMin, -1
        call fillHelperElementsFineToCoarse( scheme     = scheme(iScheme),     &
          &                                  geometry   = geometry,            &
          &                                  params     = params,              &
          &                                  iLevel     = pdfGlobal%minLevel,  &
          &                                  intp       = intp,                &
          &                                  pdfGlobal  = pdfGlobal,           &
          &                                  iTimeLayer = iTimeLayer  )
        call fillHelperElementsCoarseToFine( scheme     = scheme(iScheme),     &
          &                                  geometry   = geometry,            &
          &                                  params     = params,              &
          &                                  iLevel     = pdfGlobal%minLevel,  &
          &                                  intp       = intp,                &
          &                                  pdfGlobal  = pdfGlobal,           &
          &                                  iTimeLayer = iTimeLayer  )
      end do

      if( intp%testInterpolation ) then
        call tem_write('Testing  for the interpolation method.')
        do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
          call mus_testInterpolation( scheme     = scheme(iScheme),            &
            &                         tree       = geometry%tree,              &
            &                         params     = params,                     &
            &                         iLevel     = iLevel,                     &
            &                         intp       = intp,                       &
            &                         pdfGlobal  = pdfGlobal,                  &
            &                         iTimeLayer = iTimeLayer  )
        end do
      end if
      ! And set back the time layer to standard to interpolate the correct
      ! time layer in the interpolation module
      intp%timeLayer = 1


      if ( scheme(iScheme)%trackingControl%active ) then
        call tracker( scheme    = scheme(iScheme),                             &
          &           pdf       = scheme(iScheme)%pdf,                         &
          &           track     = scheme(iScheme)%track,                       &
          &           params    = params,                                      &
          &           layout    = scheme(iScheme)%layout,                      &
          &           glob_tree = geometry%tree,                               &
          &           pdfGlobal = pdfGlobal )
      end if
  
      ! initialize each field boundary by looping over the field inside
      ! init_boundary_field
      call tem_write( 'Initializing boundary conditions...')
      call init_boundary( field     = scheme(iScheme)%field,                   &
        &                 pdf       = scheme(iScheme)%pdf,                     &
        &                 tree      = geometry%tree,                           &
        &                 glob      = scheme(iScheme)%glob,                    &
        &                 pdfGlobal = pdfGlobal,                               &
        &                 layout    = scheme(iScheme)%layout,                  &
        &                 varSys    = scheme(iScheme)%varSys,                  &
        &                 derVarPos = scheme( iScheme)%derVarPos,              &
        &                 globBC    = scheme( iScheme )%globBC,                &
        &                 bc_prop   = geometry%boundary )

    end do !scheme loop

    if( geometry%dynamicGeom ) then
      call mus_init_geomIncr( geometry  = geometry,                            &
        &                     scheme    = scheme,                              &
        &                     pdfGlobal = pdfGlobal )
    end if ! Time for geom incr

    if( params%restart%controller%writeRestart ) then
      if(  pdfGlobal%timing%iTime .ge.                                         &
         &   int(params%restart%controller%timeControl%iMin                    &
         &   + pdfGlobal%timing%iStart)) then   
        call tem_write( 'Writing restart...')
        call mus_writeRestart( pdfGlobal = pdfGlobal,                          &
          &                    restart   = params%restart,                     &
          &                    scheme    = scheme(:),                          &
          &                    tree      = geometry%tree,                      &
          &                    timing    = pdfGlobal%timing )
        call tem_time_updateTrigger(                                           &
          &                       me  = params%restart%controller%timeControl, &
          &                       now = pdfGlobal%timing )
      end if
    end if

    pdfGlobal%timing%iTime  = pdfGlobal%timing%iTime
    pdfGlobal%timing%iStart = pdfGlobal%timing%iTime
    pdfGlobal%timing%iStop  = params%timeControl%iMax

    call tem_write("Starting MAIN loop from  iMin:",                           &
      &             params%timeControl%iMin, before = 1 ) 
    call tem_write("                         tMin:",                           &
      &             real(params%timeControl%tMin ))
    call tem_write("Starting MAIN loop until iMax:",                           &
      &             params%timeControl%iMax, before = 0 ) 
    call tem_write("                         tMax:",                           &
      &             params%timeControl%tMax )
    call tem_horizontalSpacer( after = 1 )

  end subroutine mus_init_aux
!******************************************************************************!


!******************************************************************************!
  !> Recursively fill all the helper elements (i.e. ghost, halo) with valid
  !! information from the fluid elements.
  !! 
  !! This step is required before each run of the simulation. It would be
  !! possible to fill also the helper elements with the initial conditions.
  !! However, we are only able to fill the fluid elements with valid data
  !! (restart files have no information about the helper elements)
  !!
  recursive subroutine fillHelperElementsFineToCoarse( scheme, geometry,       &
    &                       params,iLevel, intp, pdfGlobal, iTimeLayer  )
    !---------------------------------------------------------------------------
    !> containers for the scheme
    type(mus_scheme_type) :: scheme
    !> geometry infomation
    type(mus_geom_type)   :: geometry
    !> global parameters
    type(param_type)      :: params  
    !> Interpolation information
    type(interpolation_type) :: intp      
    !> global flow quantity properties
    type(pdf_globInfo_type) :: pdfGlobal
    !> Level counter variable
    integer, intent(in) :: iLevel
    !> which time layer to fill
    integer, intent(in) :: iTimeLayer
    !---------------------------------------------------------------------------
    integer :: timeLayer ! which time layer to fill
    !---------------------------------------------------------------------------

    if( iTimeLayer == 1 ) then
      timeLayer = scheme%pdf( iLevel )%nNext
    else
      timeLayer = scheme%pdf( iLevel )%nNow 
    end if

      if( iLevel .lt. pdfGlobal%maxLevel ) then
        call fillHelperElementsFineToCoarse( scheme, geometry, params,         &
          &                             iLevel+1, intp, pdfGlobal, iTimeLayer )
        !\todo loop over field, pass varPos, nScalars
        intp%timeLayer = iTimeLayer
        call mus_exchange(                                                     &
          &     recv    = scheme%glob%levelDesc( iLevel )%recvbufferFromFiner, &
          &     send    = scheme%glob%levelDesc( iLevel )%sendbufferFromFiner, &
          &     state   = scheme%pdf( iLevel )%state( :, timeLayer ),          &
          &     pattern = params%commPattern,                                  &
          &     level   = iLevel )
        call intp%fillMineFromFiner_low(                                       &
          &      pdf       = scheme%pdf,                                       &
          &      glob      = scheme%glob,                                      &
          &      scheme    = scheme,                                           &
          &      fieldProp = scheme%field(:)%fieldProp,                        &
          &      level     = iLevel,                                           &
          &      layout    = scheme%layout,                                    &
          &      intp      = intp,                                             &
          &      ind       = scheme%glob%levelDesc( iLevel)%intpFromFiner_low, &
          &      varSys    = scheme%varSys,                                    &
          &      derVarPos = scheme%derVarPos )

        call intp%fillMineFromFiner(                                           &
          &      pdf       = scheme%pdf,                                       &
          &      glob      = scheme%glob,                                      &
          &      scheme    = scheme,                                           &
          &      fieldProp = scheme%field(:)%fieldProp,                        &
          &      level     = iLevel,                                           &
          &      layout    = scheme%layout,                                    &
          &      intp      = intp,                                             &
          &      ind       = scheme%glob%levelDesc(iLevel)%intpFromFiner_high, &
          &      varSys    = scheme%varSys,                                    &
          &      derVarPos = scheme%derVarPos  )
        call mus_exchange(                                                     &
          &     recv    = scheme%glob%levelDesc( iLevel )%recvbufferFromFiner, &
          &     send    = scheme%glob%levelDesc( iLevel )%sendbufferFromFiner, &
          &     state   = scheme%pdf( iLevel )%state( :, timeLayer ),          &
          &     pattern = params%commPattern,                                  &
          &     level   = iLevel )

      end if

      call mus_exchange( recv    = scheme%glob%levelDesc( iLevel )%recvbuffer, &
        &                send    = scheme%glob%levelDesc( iLevel )%sendbuffer, &
        &                state   = scheme%pdf( iLevel )%state( :, timeLayer ), &
        &                pattern = params%commPattern,                         &
        &                level   = iLevel )
                                                                              
  end subroutine fillHelperElementsFineToCoarse
!******************************************************************************!


!******************************************************************************!
  !> Recursively fill all the helper elements (i.e. ghost, halo) with valid
  !! information from the fluid elements.
  !! 
  !! This step is required before each run of the simulation. It would be
  !! possible to fill also the helper elements with the initial conditions.
  !! However, we are only able to fill the fluid elements with valid data 
  !! (restart files have no information about the helper elements)
  !!
  recursive subroutine fillHelperElementsCoarseToFine( scheme, geometry,       &
    &                       params,iLevel, intp, pdfGlobal, iTimeLayer  )
    !---------------------------------------------------------------------------
    !> containers for the scheme
    type(mus_scheme_type) :: scheme
    !> geometry infomation
    type(mus_geom_type)   :: geometry
    !> global parameters
    type(param_type)      :: params  
    !> Interpolation information
    type(interpolation_type) :: intp      
    !> global flow quantity properties
    type(pdf_globInfo_type) :: pdfGlobal
    !> Level counter variable
    integer, intent(in) :: iLevel
    !> which time layer to fill
    integer, intent(in) :: iTimeLayer
    !---------------------------------------------------------------------------
    integer :: timeLayer ! which time layer to fill
    !---------------------------------------------------------------------------

    if( iTimeLayer == 1 ) then
      timeLayer = scheme%pdf( iLevel )%nNext
    else
      timeLayer = scheme%pdf( iLevel )%nNow 
    end if

    if( iLevel .gt. pdfGlobal%minLevel ) then
      call mus_exchange(                                                       &
        &     recv    = scheme%glob%levelDesc( iLevel )%recvbufferFromCoarser, &
        &     send    = scheme%glob%levelDesc( iLevel )%sendbufferFromCoarser, &
        &     state   = scheme%pdf( iLevel )%state( :, timeLayer ),            &
        &     pattern = params%commPattern,                                    &
        &     level   = iLevel )
    end if

    if( iLevel .lt. pdfGlobal%maxLevel ) then
      intp%timeLayer = iTimeLayer
      call intp%fillFinerFromMe_low(                                           &
        &    pdf       = scheme%pdf,                                           &
        &    glob      = scheme%glob,                                          &
        &    scheme    = scheme,                                               &
        &    fieldProp = scheme%field(:)%fieldProp,                            &
        &    level     = iLevel,                                               &
        &    intp      = intp,                                                 &
        &    ind       = scheme%glob%levelDesc( iLevel+1)%intpFromCoarser_low, &
        &    layout    = scheme%layout,                                        &
        &    varSys    = scheme%varSys,                                        &
        &    derVarPos = scheme%derVarPos )

      call intp%fillFinerFromMe(                                               &
        &   pdf       = scheme%pdf,                                            &
        &   glob      = scheme%glob,                                           &
        &   scheme    = scheme,                                                &
        &   fieldProp = scheme%field(:)%fieldProp,                             &
        &   level     = iLevel,                                                &
        &   intp      = intp,                                                  &
        &   ind       = scheme%glob%levelDesc( iLevel+1)%intpFromCoarser_high, &
        &   layout    = scheme%layout,                                         &
        &   varSys    = scheme%varSys,                                         &
        &   derVarPos = scheme%derVarPos  )  
      
      call fillHelperElementsCoarseToFine( scheme, geometry, params,           &
        &                               iLevel+1, intp, pdfGlobal, iTimeLayer )

    end if

  end subroutine fillHelperElementsCoarseToFine
!******************************************************************************!


!******************************************************************************!
  !> Close auxiliary stuff such as restart and the tracker
  !!
  subroutine mus_finalize( tree, scheme, params )
    !---------------------------------------------------------------------------
    !> tree information
    type(treelmesh_type), intent(in) :: tree
    !> containers for the scheme
    type(mus_scheme_type)            :: scheme(:)
    !> global parameters
    type(param_type),intent(inout)   :: params
    !---------------------------------------------------------------------------
    integer           :: iScheme, iTimer
    character(len=40) :: timerLabel
    real(kind=rk)     :: mainloopTime, timerVal, commTime, cpuCost
    !---------------------------------------------------------------------------

    ! close musubi main configuration file
    call close_config(params%conf)

    if( params%restart%controller%writeRestart .or.                            &
      & params%restart%controller%readRestart ) then
      call tem_restart_finalize( params%restart )
    end if

    ! Evaluate the load imbalance
    if( params%balance%active ) then
      ! Get the mainloop time
      do iTimer = 1, timerData%nTimers
        timerVal   = timerData%value( iTimer )
        timerLabel = 'time'//trim(timerData%label( iTimer ))
        !! \todo the identification, of which entry is really the main loop 
        !!  has to be done in a more general way
        if( trim(timerData%label( iTimer )) == 'MainLoop' ) then
          mainloopTime = timerVal
        elseif( trim(timerData%label( iTimer )) == 'Comm' ) then
          commTime = timerVal
        endif
      enddo
      ! The actual computational cost is the total mainloop time 
      ! minus the time it spent inside the communication routines
      cpuCost = mainLoopTime - commTime
      call tem_balance_evaluateImbalance( me      = params%balance,            &
        &                                 CPUcost = cpuCost,                   &
        &                                 nParts  = process%comm_size,         &
        &                                 comm    = process%comm )
    endif

    do iScheme = 1, size( scheme )
      ! Check, if any open units have to be closed
      ! KJ: Moved this unit close stuff here from tem_restart as
      ! we are now defining solSpec unit in tem_var_system_module
      ! instead of tem_restartHeader_type
      if ( tree%global%myPart .eq. 0) then
        if( scheme(iScheme)%varSys%solSpecUnit .ge. 0 ) then
          call tem_unit_close( scheme(iScheme)%varSys%solSpecUnit  )
        end if
      end if
      call tem_tracking_finalize( scheme( iScheme )%track )
      ! close the solver specific string
      call close_config(L = scheme(iScheme)%varSys%solSpec_conf)
    end do

  end subroutine mus_finalize
!******************************************************************************!


!******************************************************************************!
  !> Set parameters
  !!
  subroutine mus_setParameters( scheme, pdfGlobal, iLevel, params )
    !---------------------------------------------------------------------------
    !> scheme type
    type(mus_scheme_type) :: scheme
    !> global info type
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    !> level
    integer, intent(in)     :: iLevel
    !> global parameters
    type(param_type),intent(in) :: params   
    !---------------------------------------------------------------------------
    !call tem_log( level = 10,                                                  &
    !  & msg = 'Initializing relaxation parameters for scheme: '//              &
    !  &       trim(scheme%header%kind) )

    select case(trim(scheme%header%kind))
    case('lbm', 'lbm_incomp')
      ! \todo KM: must set omega at each time step only if transient is 
      ! active
      call setParameters_lbm( scheme = scheme, iField = 1,                     &
        &                     pdfGlobal = pdfGlobal, iLevel = iLevel,          &
        &                     params = params ) 
    end select  
    
    !call tem_log( level = 10,                                                  &
    !  & msg = 'Done initializing relaxation parameters.')
  end subroutine mus_setParameters
!******************************************************************************!


!******************************************************************************!
  !> Set parameters for lbm and lbm_incompressible schemes
  subroutine setParameters_lbm( scheme, iField, pdfGlobal, iLevel, params )
    !--------------------------------------------------------------------------
    type(mus_scheme_type) :: scheme      !< Scheme type
    integer, intent(in) :: iField !< field counter
    type(pdf_globInfo_type), intent(in) :: pdfGlobal !< global info type
    integer, intent(in)     :: iLevel    !<  level
    !> global parameters
    type(param_type),intent(in) :: params   
    !--------------------------------------------------------------------------
    real(kind=rk) :: fac !< ramping multiplication factor
    real(kind=rk) :: omega !< the shear relaxation parameter
    !--------------------------------------------------------------------------
    ! get the transient multiplication factor
    fac = tem_transient_for( scheme%field( iField )%fieldprop            &
      &           %fluid%omega_ramping, pdfGlobal%timing, params%conf )
    !! Relaxation parameter for each level
    !! kine_viscosity = dt*c^2*(1/omega - 0.5)
    !!    =>   omega = dt / (viscosity/c^2 + dt/2)
    !! omega for each level is stored at fluid%omLvl
    omega  = 1._rk  /                                                    &
     &        (cs2inv*scheme%field( iField )%fieldProp                   &
     &        %fluid%visc( iLevel ) + 0.5_rk )

    omega = omega * fac
    scheme%field(iField )%fieldProp%fluid%omLvl( iLevel ) = omega
    if( scheme%header%relaxation(1:3) == 'mrt' ) then
      if( .not. allocated( scheme%field(iField)%fieldProp%fluid          &
        &          %mrt( iLevel )%s_mrt))                                &
        & allocate( scheme%field(iField)%fieldProp%fluid                 &
        &          %mrt( iLevel )%s_mrt( scheme%layout%stencil(1)%QQ ))
      ! Set the relaxation parameters (only the non-relaxing modes)
      scheme%field(iField)%fieldProp%fluid%mrt(iLevel )                  & 
              %s_mrt(1:scheme%layout%stencil(1)%QQ) = omega

      ! set bulk omega with multiplication factor
      if( trim(scheme%header%relaxation) .ne. 'mrt_bgk' ) then
        select case(trim(scheme%header%layout))
        case('d2q9' )
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel )              & 
                  %s_mrt(1) = 0._rk 
                !%s_mrt(1) = omega
          ! Set the bulk viscosity
          ! \todo KM->MH: From where relaxation matrix values are taken
          scheme%field(iField)%fieldProp%fluid%mrt( iLevel )%s_mrt(2) =  &
            & fac*scheme%field(iField)%fieldProp%fluid%omLvl_bulk(iLevel)
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel )              & 
            &     %s_mrt(3) = 1.14_rk
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel )              & 
            &     %s_mrt(4) = 0._rk 
                  !%s_mrt(4) = omega
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel )              & 
            &     %s_mrt(5) = 1.92_rk
                  !%s_mrt(5) = 1._rk
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel )              & 
            &     %s_mrt(6) = 0._rk 
                 !%s_mrt(6) = omega
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel )              & 
            &     %s_mrt(7) = 1.92_rk
                  !%s_mrt(7) = 1._rk
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel )              & 
            &     %s_mrt(8:9) = omega
        case('d3q19')
          ! KM : We can choose between [-2, 0] for rest
          ! \ref
          ! J. Tölke, S. Freudiger, and M. Krafczyk, 
          ! "An adaptive scheme using hierarchical grids for 
          ! lattice Boltzmann multi-phase flow simulations," 
          ! Comput. Fluids, vol. 35, pp. 820–830, 2006.
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel )              & 
            &     %s_mrt(:) = 1._rk
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel )              & 
            &     %s_mrt(14:16) = omega
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel )              & 
            &     %s_mrt(10) = omega
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel )              & 
            &     %s_mrt(12) = omega
          ! Bulk viscosity
          scheme%field(iField)%fieldProp%fluid%mrt( iLevel )%s_mrt(2) =  &
            & fac*scheme%field(iField)%fieldProp%fluid%omLvl_bulk(iLevel)
        end select
      end if ! mrt_bgk?
    end if ! mrt?

  end subroutine setParameters_lbm
!******************************************************************************!


!******************************************************************************!
  !> This routine sets the spatial value for the omega for 
  !! spatial omega definition
  !!
  !! \verbatim 
  !! -- Definition of the spatial function for how to vary omega 
  !! -- in the simulation domain
  !! function spatialFunction(x,y,z)
  !!    if(  x > xmax) then
  !!      res =w_max
  !!    elseif( x < xmin ) then
  !!      res =  w_min
  !!    else
  !!      slope = (w_max-w_min)/(cutoff_min-cutoff_max)
  !!      res =  r*slope + w_max - cutoff_min*slope
  !!    end
  !!    return res
  !! end
  !! 
  !! fluid = { omega_spatial = {predefined = 'lua_fun',
  !!                            omega_spatial = spatialFunction, store = true }
  !! identify = { relaxation = 'bgk_omega_spatial'} 
  !! \endverbatim
  !!
  subroutine mus_setSpatialOmega( scheme, tree, pdfGlobal, params )
    !--------------------------------------------------------------------------
    !> scheme to work on
    type( mus_scheme_type ), intent(inout) :: scheme
    !> global treelm mesh 
    type( treelmesh_type ), intent(in) ::tree
    !> pdf global info type
    type( pdf_globInfo_type ) :: pdfGlobal
    !> global parameters
    type(param_type),intent(in) :: params   
    !---------------------------------------------------------------------------
    integer :: iLevel, nElems, iField
    integer :: iVel
    real(kind=rk) :: nu_LB
    !---------------------------------------------------------------------------
    
    select case(trim(scheme%header%kind))
    case('lbm', 'lbm_incomp')
      iField = 1
      if( scheme%field(iField)%fieldprop%fluid%active ) then
        if( scheme%field(iField)%fieldProp%fluid%viscSpatial%toStore ) then
          call tem_write(' Storing spatial omega ...')
          ! loop over each refinement iLevels
          do iLevel  = pdfGlobal%minLevel, pdfGlobal%maxLevel
            nu_LB   = scheme%field(iField)%fieldprop%fluid%visc( iLevel )
            nElems = scheme%pdf( iLevel )%nElems_solve
            allocate( scheme%field(iField)%fieldprop%fluid%viscSpatial%        &
              &                               stored( iLevel )%val( nElems ))  

            scheme%field(iField)%fieldprop%fluid%viscSpatial%                  &
              &                            stored( iLevel )%val( : ) = 0.0_rk  
            ! get spatial term for all elements in iLevel
            ! in terms of viscosity
            scheme%field(iField)%fieldprop%fluid%viscSpatial%                  &
              &                      stored( iLevel )%val = tem_spatial_for(   &
              & me      = scheme%field(iField)%fieldprop%fluid%viscSpatial,    &
              & treeIds = scheme%glob%levelDesc( iLevel )%total( 1:nElems ),   &
              & tree    = tree,                                                &
              & n       = nElems,                                              &
              & conf     = params%conf ) 
            ! and now translate the viscosity in physical units to lattice
            ! omega multiplication factor
            ! The sponge factor gives the multiplication factor of the
            ! physical viscosity
            ! The factor store here in viscSpatial is given as a
            ! multiplication to the original omega (which might be ramped as
            ! well in time -> multiply). However, the relationship between
            ! viscosity and omega is non-linear with nu = 1/3(1/omega - 1/2)
            ! omega_sponged = factor * omega
            !  -> factor = omega_sponged / omega
            ! omega = 1/(3*nu_phys*dt/dx^2 + 1/2 )
            ! or with lattice viscosity
            ! omega = 1/(3*nu_LB + 1/2)
            ! the factor for converting the physical viscosity to lattice
            ! viscosity is collected in q = dt/dx^2
            scheme%field(iField)%fieldprop%fluid%viscSpatial%                  &
              &                                       stored( iLevel )%val     &
              &  = (3._rk*nu_LB + 0.5_rk) / (3._rk*nu_LB                       &
              &  * scheme%field(iField)%fieldprop%fluid%viscSpatial%           &
              &                                       stored( iLevel )%val     &
              &  + 0.5_rk )
          end do
        endif
        do iVel = 1,3
          if( scheme%field(iField)%fieldProp%fluid%                            &
            &                               velSpatial(iVel)%toStore ) then
            call tem_write(' Storing velocity field...')
            ! loop over each refinement iLevels
            do iLevel  = pdfGlobal%minLevel, pdfGlobal%maxLevel
              nElems = scheme%pdf( iLevel )%nElems_solve
              allocate( scheme%field(iField)%fieldprop%fluid%velSpatial(iVel)  &
                &             %stored( iLevel )%val( nElems ))
              scheme%field(iField)%fieldprop%fluid%velSpatial(iVel)            &
                &                 %stored( iLevel )%val( : ) = 0.0_rk
              ! get spatial term all elements in iLevel
              scheme%field(iField)%fieldprop%fluid%velSpatial(iVel)            &
                &                  %stored( iLevel )%val = tem_spatial_for(    &
                & me      = scheme%field(iField)%ic%ini_state(1+iVel),         &
                & treeIds = scheme%glob%levelDesc(iLevel)%total( 1:nElems ),   &
                & tree    = tree,                                              &
                & n       = nElems,                                            &
                & conf    = params%conf ) 
              scheme%field(iField)%fieldprop%fluid%velSpatial(iVel)%           &
                &                                     stored( iLevel )%val     &
                & = scheme%field(iField)%fieldprop%fluid%velSpatial(iVel)%     &
                &                                     stored( iLevel )%val     &
                & / scheme%physics%fac( iLevel )%vel
            end do
          endif
        end do
      endif
    end select  

  end subroutine mus_setSpatialOmega
!******************************************************************************!

end module mus_aux_module
!******************************************************************************!
