! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Module containing subroutines to prepare buffers for transformation of
!! quantities (tracking) and the restart/tracking IO.
!!
!! \author Simon Zimny 
!!
!! To save memory, the transformation of quantities and the IO-functionality of 
!! MUSUBI is done in linear chunks. The size of the chunk can be defined in the
!! lua input file and is defined by the function [load_env_params]. The default
!! size is set to be 8MB.
!!
!! [load_env_params]: @ref env_module#tem_load_env_params "tem_load_env_params"
!!
?? include 'header/lbm_macros.inc'

module mus_buffer_module

  ! include treelm modules
  use mpi
  use env_module,            only: rk, long_k
  use treelmesh_module,      only: treelmesh_type
  use tem_topology_module,   only: tem_LevelOf
  use tem_restart_module,    only: tem_restart_type
  use tem_var_system_module, only: tem_var_system_type, tem_var_sys_debug
  use tem_debug_module,      only: tem_debug, main_debug
  use tem_logging_module,    only: tem_toStr

  ! include musubi modules
  use mus_scheme_type_module, only: mus_scheme_type
  use mus_pdf_module,         only: pdf_globInfo_type, pdf_data_type

  implicit none
  private

  public :: mus_rest_serialize
  public :: mus_rest_unserialize
  public :: mus_trans_serialize

contains

!******************************************************************************!
  !> Preparation of the state data for writing it to disk for restart
  !! 
  !! Serialize the information from the (level-wise) state vector (scheme%state) 
  !! into chunks for writing it in original treeIDlist order to disk (which is 
  !! not sorted by levels, but by the space-filling curve)
  !! The data is stored like follows (e.g. 2 varSys, 2 elements):
  !!
  !! \verbatim
  !!   Elem = 1     Elem = 2    | Elem = 1    Elem = 2  
  !! ---------------------------|------------------------
  !! | stateVars | stateVars    | stateVars | stateVars |
  !! ---------------------------|------------------------
  !!          varSys1           |        varSys2        
  !! \endverbatim
  !! 
  subroutine mus_rest_serialize( scheme, pdfGlobal, restart, tree, chunk,      &
    &                            iChunk )
    !---------------------------------------------------------------------------
    !> scheme type containing the different state vectors
    type(mus_scheme_type), intent(inout)   :: scheme(:)
    !> global pdf info
    type(pdf_globInfo_type), intent(inout) :: pdfGlobal
    !> restart information  
    type(tem_restart_type), intent(inout)  :: restart
    !> global tree data type
    type(treelmesh_type), intent(in)       :: tree
    !> global chunk to fill 
    real(kind=rk), intent(inout)           :: chunk(:)
    !> current chunk ID
    integer, intent(in)                    :: iChunk
    !---------------------------------------------------------------------------
    integer :: iSys, iVar, iElem, iLevel, iComp ! counter variables
    integer :: iIndex  ! amount of written variable components per varsys
    integer :: elemOff ! Offset in the overall number of elements
    integer :: nComp   ! max number of components
    integer :: nScalars
    integer, allocatable :: varPos(:)
    !---------------------------------------------------------------------------

    ! Number of elements written by previous chunks.
    elemOff = ((iChunk-1)*restart%chunkSize)

    ! set the number of elements that are on the stack
    ! number of elements which have to be written out within chunk iChunk
    restart%nChunkElems = min(restart%chunkSize, tree%nElems-elemOff)

    ! Global counter for all words put into the buffer.
    iIndex = 0
    ! Run over all variable systems.
    do iSys = 1, restart%nSys    
      if(allocated(varPos)) deallocate(varPos)
      allocate(varPos(scheme(iSys)%layout%stencil(1)%QQ))
      nScalars = restart%varSys(iSys)%nScalars

      ! debug output
      if( main_debug%debugRestart )then
        call tem_var_sys_debug( sys   = restart%varSys( iSys ),                &
          &                     level = 1 )
      end if

      ! For the current chunk (given from the caller), treat the corresponding
      ! elements from the tree ID list
      do iElem=elemOff+1,elemOff+restart%nChunkElems
        ! On which level lies the current element
        iLevel = tem_LevelOf( tree%treeID( iElem ))

        ! nVars gives the number of non derived variables in the var system
        ! Run over the variables defined in the var system
        do iVar = 1, restart%varSys( iSys )%nVars

          nComp = restart%varSys( iSys )%variable%val( iVar )%nComponents
          varPos = restart%varSys(iSys)%variable%val( iVar )%varPos
          ! ... treat each component
          do iComp = varPos(1), varPos(nComp)
           iIndex = iIndex + 1
           chunk(iIndex) = scheme(iSys)%pdf( iLevel )%state(                   &
&   ?IDX?( iComp, pdfGlobal%levelPointer( iElem ), nScalars, scheme(iSys)%pdf( iLevel )%neigh ), &
             &   scheme(iSys)%pdf( iLevel )%nNext)

            ! debug output
            if( main_debug%debugRestart )then
              call tem_debug( level = 1,                                       &
                &             msg   = 'iElem: '//trim( tem_toStr( iELem ))//   &
                &                     ' iVar: '//trim( tem_toStr( iVar )))
              call tem_debug( level = 1,                                       &
                &             msg   = 'icomp: '//trim( tem_toStr( iComp ))//   &
                &                     ' chunk(iIndex): '//                     &
                &                     trim( tem_toStr( chunk( iIndex ))))
            end if

          end do  ! iComp
        end do  ! iVar
      end do  ! iElem
    end do  ! iSys

  end subroutine mus_rest_serialize
!******************************************************************************!


!******************************************************************************!
  !> This subroutine unserializes the read data and stores it in the state-
  !! vector to perform a restart.
  !!
  subroutine mus_rest_unserialize( scheme, pdfGlobal, restart, tree, chunk,    &
    &                              iChunk, iSys_idx )
    !---------------------------------------------------------------------------
    !> scheme type containing the different state vectors
    type(mus_scheme_type), intent(inout)   :: scheme(:) 
    !> global pdf info      
    type(pdf_globInfo_type), intent(inout) :: pdfGlobal
    !> restart information  
    type(tem_restart_type), intent(inout)  :: restart
    !> global tree data type
    type(treelmesh_type), intent(in)       :: tree
    !> buffer array to fill
    real(kind=rk), intent(out)             :: chunk(:)
    !> current chunk ID
    integer, intent(in)                    :: iChunk
    !> the index of varsys
    integer, optional, intent(in)          :: iSys_idx
    !---------------------------------------------------------------------------
    integer :: iSys, iVar, iElem, iLevel, iComp ! counter variables
    integer :: iIndex       ! amount of written variable components per varsys
    integer :: nComp        ! max number of components
    integer :: elemOff      ! offset in the overall number of elements
    integer :: nScalars
    integer, allocatable :: varPos(:)
    !--------------------------------------------------------------------------- 

    ! Number of elements read so far in previous chunks.
    elemOff = ((iChunk-1)*restart%chunkSize)

    restart%nChunkElems = min(restart%chunkSize, tree%nElems-elemOff)

    ! For all variable systems...
    iIndex = 0
    do iSys = 1, restart%nSys   
      ! debug output
      if( main_debug%debugRestart )then
        call tem_var_sys_debug( sys   = restart%varSys( iSys ),                &
          &                     level = 1 )
      end if

      ! Actually this loop over the elems should be moved into and an 
      ! array of treeIDs should be passed to the derive subroutine
      if(allocated(varPos)) deallocate(varPos)
      allocate(varPos(scheme(iSys)%layout%stencil(1)%QQ))
      
      nScalars = restart%varSys(iSys)%nScalars
      do iElem =  elemOff + 1, elemOff + restart%nChunkElems
        iLevel = tem_LevelOf( tree%treeID( iElem ))
        ! nVars gives the number of non derived variables in the var system
        do iVar = 1, restart%varSys(iSys)%nVars

          nComp = restart%varSys(iSys)%variable%val( iVar )%nComponents
          varPos = restart%varSys(iSys)%variable%val( iVar )%varPos
          do iComp = varPos(1), varPos(nComp)
            ! fill the state vector with the values from the chunk
            iIndex = iIndex + 1
            !KJ: iSys_idx will be present only in the case when restart is being 
            !read from the initial conditions
            if ( present(iSys_idx)) then
              if(  iSys_idx .eq. iSys ) then
              scheme(iSys)%pdf(iLevel)%state(                                  &
              ! MH: Change this (from SAVE) to IDX, to be consistent with
              ! mus_rest_serialize
              ! This should not affect anything in the PULL algorithm as
              ! SAVE=IDX
            &   ?IDX?( iComp, pdfGlobal%levelPointer( iElem ), nScalars, scheme(iSys)%pdf( iLevel )%neigh ), &
              ! MH: Only set nNext (is being switched to nNow before the compute
              ! kernel, which reads from nNow)
            &   scheme(iSys)%pdf(iLevel)%nNext ) = chunk(iIndex)
              end if 
            !If iSys is not present, state array is populated for all nSys
            else
              scheme(iSys)%pdf(iLevel)%state(                                  &
            &   ?IDX?( iComp, pdfGlobal%levelPointer( iElem ), nScalars, scheme(iSys)%pdf( iLevel )%neigh ), &
            &   scheme(iSys)%pdf(iLevel)%nNext ) = chunk(iIndex)
            end if

            ! debug output
            if( main_debug%debugRestart )then
              call tem_debug( level = 1,                                       &
                &             msg   = 'iElem: '//trim( tem_toStr( iELem ))//   &
                &                     ' iVar: '//trim( tem_toStr( iVar )))
              call tem_debug( level = 1,                                       &
                &             msg   = 'icomp: '//trim( tem_toStr( iComp ))//   &
                &                     ' chunk(iIndex): '//                     &
                &                     trim( tem_toStr( chunk( iIndex ))))
            end if

          end do  ! iComp
        end do  ! iVar   
      end do  ! iElem

    end do  ! iSys

  end subroutine mus_rest_unserialize
!******************************************************************************!


!******************************************************************************!
  !> This subroutine serializes a single state array for a given set of elements
  !! to perform transformations on.
  !!
  !! The data is stored like follows (e.g. globSys, 2 elements):
  !!
  !! \verbatim
  !!   Elem = 1     Elem = 2   
  !! ----------------------------
  !! | stateVars | stateVars    |
  !! ----------------------------
  !!          globSys           
  !! \endverbatim
  !! 
  subroutine mus_trans_serialize( pdf, globSys, chunk, ichunk, chunkSize,      &
    &                             nChunkElems, treeID, elemPos, preColl )
    !---------------------------------------------------------------------------
    !> pdf datatype containing state vector of all levels and neigh array
    type(pdf_data_type), intent(in)   :: pdf(:)
    !> the buffer array to fill
    real(kind=rk), intent(out) :: chunk(:)
    !> current chunk ID
    integer, intent(in)        :: iChunk
    !> maximum number of elements that fit into the chunk
    integer, intent(in)        :: chunkSize 
    !> actual number of elements in the chunk
    integer, intent(out)       :: nChunkElems 
    !> array of treeIDs (elements) to serialize the data
    integer(kind=long_k), intent(in) :: treeID(:)
    !> global varSys for the current scheme
    type(tem_var_system_type), intent(in) :: globSys 
    !> array element position in the state vector
    integer, intent(in) :: elemPos(:)
    !> use pre collision values? else post collision
    logical, optional, intent(in)          :: preColl 
    !---------------------------------------------------------------------------
    integer :: iVar, iElem, iLevel, iComp, iIndex, iPos ! counter variables
    integer :: elemOff  ! element offset
    integer :: nComp    ! total number of components
    logical :: loc_preColl ! local logical to decide pre or post collision
    !---------------------------------------------------------------------------

    ! check if preColl is present, 
    ! default: preColl = false -> use post collision
    if(present(preColl))then
      loc_preColl = preColl
    else
      loc_preColl = .false.
    end if

    ! Number of elements read so far in previous chunks.
    elemOff = ((iChunk-1)*chunkSize)

    ! number of elements written to THIS chunk
    nChunkElems = min(chunkSize, size(treeID)-elemOff)

    iIndex = 0
    do iElem =  elemOff+1, elemOff+nChunkElems
      iLevel = tem_LevelOf( treeID( iElem ))
      iPos = elemPos( iElem )
      ! nVars gives the number of non derived variables in the var system
      do iVar = 1, globSys%nVars
        nComp = globSys%variable%val( iVar )%nComponents

        do iComp = globSys%variable%val( iVar )%varPos(1),                     &
            &      globSys%variable%val( iVar )%varPos(nComp)
          ! fill the state vector with the values from the chunk
          iIndex = iIndex + 1

          ! decide wether we have to use IDX or FETCH to access the values
          if( loc_preColl)then
            ! access with FETCH for wss, shearstress
            chunk(iIndex) = pdf( iLevel )%state(                               &
              & ?FETCH?( iComp, iPos, globSys%nScalars, pdf( iLevel )%neigh ), &
              & pdf( iLevel )%nNext)
          else 
            ! access cell local with IDX
            chunk(iIndex) = pdf( iLevel )%state(                               &
              & ?IDX?( iComp, iPos, globSys%nScalars, pdf( iLevel )%neigh ),   &
              & pdf( iLevel )%nNext)
          end if
        end do  ! iComp
      end do    ! iVar
    end do     ! iElem

  end subroutine mus_trans_serialize
!******************************************************************************!


end module mus_buffer_module
!******************************************************************************! 
