! See copyright notice in the COPYRIGHT file.
!*****************************************************************************!
!> In this module, all parameter files are read in 
!! as lua script or a sample configuration is being loaded
!!
!! Possible Parameter configuration
!! - \ref General Parameters
!! - \ref mus_scheme_layout_module "Scheme Definitions"
!! - \ref tem_debug_module "Debug Parameters"
!!
!! 
module mus_config_module

  ! include musubi modules
  use mus_global_module,             only: param_type
  use mus_geom_module,               only: mus_geom_type, mus_geom_setWeight,  &
    &                                      mus_geom_cleanWeight,               &
    &                                      mus_geom_setLevelWeight
  use mus_interpolate_header_module, only: interpolation_type,                 &
    &                                      mus_load_interpolate
  use mus_pdf_module,                only: pdf_globInfo_type
  use mus_scheme_type_module,        only: mus_scheme_type
  use mus_scheme_module,             only: mus_load_schemes
  use mus_tools_module,              only: mus_time_init
  use mus_geomIncrHead_module,       only: mus_geomIncrHead_load
  use mus_physicsLB_module,          only: mus_setLBdxdt
  use mus_physics_module,            only: mus_create_funcStr

  ! include treelm modules
  use env_module,             only: pathLen, io_buffer_size
  use treelmesh_module,       only: load_tem, unload_tem
  use tem_aux_module,         only: tem_abort, tem_open_distconf
  use tem_property_module,    only: prp_hasBnd, prp_hasQVal
  use tem_balance_module,     only: tem_balance, tem_balance_loadConfig
  use tem_bc_prop_module,     only: tem_bc_prop_type, load_tem_BC_prop,        &
    &                               tem_empty_BC_prop, load_tem_bc_qVal
  use tem_comm_env_module,    only: process
  use tem_debug_module,       only: main_debug
  use tem_restart_module,     only: tem_load_restart
  use tem_time_module,        only: tem_time_reset
  use tem_timeControl_module, only: tem_load_timeControl
  use tem_timer_module,       only: tem_addTimer, tem_startTimer, tem_stopTimer
  use tem_tools_module,       only: tem_write, tem_horizontalSpacer 
  use tem_status_module,      only: tem_stat_runEnd
  use tem_geometry_module,    only: tem_setEffBoundingBox
  use tem_general_module,     only: tem_general_load

  ! include aotus modules
  use aot_path_module, only: aot_path_type, assignment(=), aot_init_path
  use aotus_module,    only: flu_State, aot_get_val, close_config,             &
    &                        open_config_chunk

  implicit none

  private

  public :: mus_load_config
  public :: mus_open_config

contains

!******************************************************************************!
  !> Read in LUA parameter file 
  !! See http://www.lua.org for a reference on how to use
  !! Lua is a scripting language in itself which allows 
  !! more complex parameter files including comments 
  !! And load / create the mesh depending on the configuration
  subroutine mus_load_config( geometry, boundary, params, scheme, pdfGlobal,   &
    &                         intp )
    !---------------------------------------------------------------------------
    !> Treelmesh data
    type( mus_geom_type ), intent(out)  :: geometry
    !> boundary data from mesh
    type( tem_bc_prop_type ), intent(out):: boundary
    !> Global parameters
    type( param_type ),intent(inout)     :: params
    !> scheme type
    type( mus_scheme_type ), allocatable :: scheme(:)
    !> global information
    type( pdf_globInfo_type )            :: pdfGlobal
    !> interpolation method
    type( interpolation_type )           :: intp
    !---------------------------------------------------------------------------
    type(flu_State) :: conf
    character(len=PathLen) :: filename
    integer :: iError, iProp
    type( aot_path_type ) :: init_path
    logical :: found_bnd
    !---------------------------------------------------------------------------
    ! Initialize solver run end for each process 
    params%stat%bits(tem_stat_runEnd) = .false.

    call tem_startTimer( params%loadMeshTimer ) 
    call tem_time_reset( pdfGlobal%timing )

    ! Get filename from command line argument
    call get_command_argument( 1,filename )
    if( trim(filename) .eq. '' ) then
      filename = 'musubi.lua'
    end if

    params%solver%configFile = filename
    call tem_write( "Loading configuration file "//trim( filename ))
    call tem_horizontalSpacer()

    ! open musubi config file and solver specific lua functions as chunk
    call mus_open_config( conf = conf, filename = filename )

    ! Set aotus initial path 
    call aot_init_path(me = init_path, Filename = filename)

    ! load general information
    call tem_general_load( conf        = conf,                                 &
      &                    rank        = process%rank,                         &
      &                    commpattern = params%commpattern )

    ! Load the IO Buffer size
    call tem_write('Size of the IO Buffer: ', io_buffer_size)

    ! Load debugging settings
    if( main_debug%checkNans ) then
      call tem_addTimer( params%isNanTimer, 'isNan')
    end if

    ! Load the control routine
    call aot_get_val( L       = conf,                                          &
      &               key     = 'control_routine',                             &
      &               val     = params%controlRoutine,                         &
      &               ErrCode = iError,                                        &
      &               default = 'standard' )
    call tem_write( 'Choosing Control Routine:'// trim( params%controlRoutine ))

    ! Decide, which scaling
    call aot_get_val( L       = conf,                                          &
      &               key     = 'scaling',                                     &
      &               val     = params%scaling,                                &
      &               ErrCode = iError,                                        &
      &               default = 'acoustic')
    call tem_horizontalSpacer
    call tem_write( 'Setting SCALING to:'// trim( params%scaling ))
    call tem_horizontalSpacer
    select case( trim( params%scaling ))
      case('acoustic')
        pdfGlobal%nNesting = 1
        params%scaleFactor = 2
      case('diffusive')
        pdfGlobal%nNesting = 3
        params%scaleFactor = 4
      case default 
        call tem_write(' Error: wrong scaling chosen!   ')
        call tem_write(' Choose either acoustic or diffusive   ')
        call tem_abort()
    end select

    ! Load the simulation name
    call aot_get_val( L       = conf,                                          &
      &               key     = 'simulation_name',                             &
      &               val     = params%solver%simName,                         &
      &               ErrCode = iError,                                        &
      &               default = 'simulation')
    call tem_write( 'Simulation Name: '// trim( params%solver%simName ))

    ! Communicate reduced set of links only? 
    call aot_get_val( L       = conf,                                          &
      &               key     = 'comm_reduced',                                &
      &               val     = pdfGlobal%comm_reduced,                        &
      &               ErrCode = iError,                                        &
      &               default = .true.)

    ! initialize all elements? this includes not only fluid but also ghost and
    ! halo
    call aot_get_val( L       = conf,                                          &
      &               key     = 'init_allElems',                               &
      &               val     = params%init_allElems,                          &
      &               ErrCode = iError,                                        &
      &               default = .false.)
    if( params%init_allElems ) then
      call tem_write( 'WARNING: Initializing all (including helper '//         &
        &             'elements) with valid data' )
      call tem_write( '         Only activate this for debugging purposes!!  ')
    end if

    ! initialize all time layers?
    call aot_get_val( L       = conf,                                          &
      &               key     = 'init_allTimeLayers',                          &
      &               val     = params%init_allTimeLayers,                     &
      &               ErrCode = iError,                                        &
      &               default = .true.)
    if( params%init_allTimeLayers ) then
      call tem_write( 'NOTE:    Initializing all Time layers with valid data.')
    else
      call tem_write( 'WARNING: Only initializing nNext time layer.')
      call tem_write( '         Boundaries using old time layer info will '//  &
        &             'not work and might cause a crash in the beginning' )
    end if

    ! Get the global maximum time
    call tem_load_timeControl( conf    = conf,                                 &
      &                        me      = params%timeControl )
    call mus_time_init( me = pdfGlobal%timing, timeControl = params%timeControl)

    !---------------------------------------------------------------------------
    !  Load mesh                                                               !
    !---------------------------------------------------------------------------
    ! Get load balancing config
    call tem_balance_loadConfig( conf          = conf,                         &
      &                          me            = params%balance,               &
      &                          defaultActive = .true.,                       &
      &                          globalTime    = params%timeControl )
    ! Set the initial weights
    call mus_geom_setLevelWeight( params      = params,                        &
      &                           geometry    = geometry,                      &
      &                           balance     = params%balance,                &
      &                           levelWeight = geometry%levelWeight )
    
    ! First check, if we are starting from a restart
    ! KJ: Now the restart is read from initial conditions, which are loaded
    ! much later than the config in the process flow

    call tem_load_restart( me             = params%restart,                    &
      &                    conf           = conf,                              &
      &                    tree           = geometry%tree,                     &
      &                    levelWeight    = geometry%levelWeight,              &
      &                    timing         = pdfGlobal%timing,                  &
      &                    refTimeControl = params%timeControl,                &
      &                    proc           = process )

    if( .not. params%restart%controller%readRestart ) then
      call tem_write(' No restart given. Loading mesh file')
      ! load the tree from the mesh = '' definition in case no restartRead is 
      ! given in this case the mesh from the restart header is read (in 
      ! tem_load_restart)
      call load_tem( me          = geometry%tree,                              &
        &            conf        = conf,                                       &
        &            myPart      = process%rank,                               &
        &            levelWeight = geometry%levelWeight,                       &
        &            nParts      = process%comm_size,                          &
        &            comm        = process%comm )
    else
      ! If there is a restart, the timings in the params type have to be updated
      ! to those read from the restart
      params%timeControl%iMin = pdfGlobal%timing%iTime
      params%timeControl%tMin = pdfGlobal%timing%rTime

    endif 

    ! calculate the effective origin and effective length
    ! KM :\todo get effective bounding box from seeder header
    call tem_setEffBoundingBox( geometry%tree )

    ! Do an initial load-balancing
    if( params%balance%active ) then
      call tem_horizontalSpacer
      call tem_write('Load balancing ...' )
      call tem_write('  Number of elements before load balancing',             &
        &            geometry%tree%nElems )
      ! Invoke the load balancing based on the geometry
      call mus_geom_setWeight( params   = params,                              &
        &                      geometry = geometry,                            &
        &                      balance  = params%balance )
      call tem_balance( weight = geometry%weight,                              &
        &               myPart = process%rank,                                 &
        &               nParts = process%comm_size,                            &
        &               offset = geometry%offset,                              &
        &               comm   = process%comm )
      ! unload tree ...
      call unload_tem( me = geometry%tree ) 
      ! ... and reload it with the offsets obtained from the balancing algorithm
      call load_tem( me     = geometry%tree,                                   &
        &            conf   = conf,                                            &
        &            myPart = process%rank,                                    &
        &            nParts = process%comm_size,                               &
        &            comm   = process%comm,                                    &
        &            offset = geometry%offset )
      call mus_geom_cleanWeight( geometry )
      call tem_write('  Number of elements after load balancing',              &
        &            geometry%tree%nElems )
      call tem_write('Finished load balancing.' )
      call tem_horizontalSpacer
    endif
    !---------------------------------------------------------------------------
    ! Done loading mesh.                                                       !
    !---------------------------------------------------------------------------
   
    !-------------- boundary conditions ----------------------------------------
    ! load boundary information from mesh
    ! generate / load the mesh
    ! per default, load the mesh from the location specified in the lua file
    found_bnd = .false.
    ! Check if there is any boundary property
    do iProp = 1, geometry%tree%global%nProperties
      if( geometry%tree%global%property( iProp )%bitpos .eq. prp_hasBnd ) then
        if( geometry%tree%global%property( iProp )%nElems .gt. 0 ) then
          found_bnd = .true.
        end if
      end if
    end do
    do iProp = 1, geometry%tree%global%nProperties
      if( geometry%tree%global%property( iProp )%nElems > 0 ) then
        select case( geometry%tree%global%property( iProp )%bitpos )
        case( prp_hasBnd )
          ! Load boundary condition from disk
          call tem_write('Loading boundary data from directory "'              &
            &             //trim(geometry%tree%global%dirname)//'"')
          call load_tem_BC_prop(                                               &
            &            me       = boundary,                                  &
            &            offset   = geometry%tree%Property(1)%Offset,          &
            &            nElems   = geometry%tree%Property(1)%nElems,          &
            &            basename = trim(geometry%tree%global%dirname)//'bnd', &
            &            mypart   = process%rank,                              &
            &            comm     = process%comm )
          call tem_write('Done, reading the boundary!')

        case( prp_hasQVal )
          ! Load qVal from disk
          ! prp_hasQVal is the 2nd property in mesh
          call tem_write('Loading qVal data from directory "'                  &
            &             //trim(geometry%tree%global%dirname)//'"')
          call load_tem_BC_qVal(                                               &
            &           me       = boundary,                                   &
            &           offset   = geometry%tree%Property(2)%Offset,           &
            &           nElems   = geometry%tree%Property(2)%nElems,           &
            &           basename = trim(geometry%tree%global%dirname)//'qval', &
            &           mypart   = process%rank,                               &
            &           comm     = process%comm )
          call tem_write('Done, reading the qVal!')
        end select ! property( iProp )%bitpos
      endif
    enddo ! iProp
    if( .not. found_bnd ) then
      ! No boundary conditions available!
      call tem_empty_BC_prop( boundary )
    endif
    !-------------- end boundary conditions ------------------------------------

    ! Initialize the LB time and space steps (for multilevel)
    pdfGlobal%minLevel     = geometry%tree%global%minLevel
    pdfGlobal%maxLevel     = geometry%tree%global%maxLevel
    pdfGlobal%nLevels      = pdfGlobal%maxLevel - pdfGlobal%minLevel + 1
    pdfGlobal%nBCs         = geometry%boundary%nBCtypes
    call mus_setLBdxdt( pdfGlobal = pdfGlobal, params    = params )

    !\todo: SZ: after restarting this call overwrites the solver specific header
    !           this has to be modified (task for the restart update, Kartik)
    call mus_load_schemes( me        = scheme,                                 &
      &                    boundary  = boundary,                               &
      &                    conf      = conf,                                   &
      &                    params    = params,                                 &
      &                    pdfGlobal = pdfGlobal,                              &
      &                    tree      = geometry%tree,                          &
      &                    path      = init_path )
    if( geometry%tree%global%minLevel .ne. geometry%tree%global%maxLevel ) then
      ! get the interpolation method.
      call mus_load_interpolate( conf = conf, me = intp, pdfGlobal = pdfGlobal)
    endif
    call tem_horizontalSpacer()
 
    !> If density check interval is less than 0, set it to be global tmax /10
    params%timeControl%iInterval = max( params%timeControl%iInterval, 1 )

    ! Load solidification/fluidification settings
    call mus_geomIncrHead_load( me          = geometry%geomIncr,               &
      &                         schemes     = scheme,                          &
      &                         params      = params,                          &
      &                         conf        = conf,                            &
      &                         dynamicGeom = geometry%dynamicGeom )

    ! closing the config file (reading afterwards not possible withput open
    ! the file again)
    call close_config( conf )
    
    call tem_stopTimer( params%loadMeshTimer )

  end subroutine mus_load_config
!******************************************************************************!


!******************************************************************************!
!> This routine loads musubi specific lua function from string and musubi
!! input configuration file
  subroutine mus_open_config( conf, filename )
    !--------------------------------------------------------------------------
    type(flu_State) :: conf !< lua state 
    character(len=*), intent(in) :: filename
    !--------------------------------------------------------------------------
    character(len=2048) :: fun_str !contains function to be used in musubi 
    !--------------------------------------------------------------------------

    call mus_create_funcStr( fun_str = fun_str )

    !pre-loading musubi specific lua functions using same conf used to load
    !main musubi config file
    call open_config_chunk(L = conf, chunk = fun_str) 

    ! Open the lua config
    call tem_open_distconf(L = conf, filename = trim(filename))

  end subroutine mus_open_config
!******************************************************************************!


end module mus_config_module
!******************************************************************************!

