! See copyright notice in the COPYRIGHT file.
!!*****************************************************************************!
!> \ref mus_construct "Creation of the data structures" 
!! from the information in the configuration and 
!! from the mesh read from disk
!!
!!\author Manuel Hasert 
?? include 'header/lbm_macros.inc'
module mus_construction_module

  ! include treelm modules
  use mpi
  use env_module,              only: rk, long_k, globalMaxLevels, eps
  use treelmesh_module,        only: treelmesh_type
  use tem_comm_env_module,     only: process
  use tem_comm_module,         only: tem_communication_type,                   &
    &                                tem_commpattern_type
  use tem_param_module,        only: childPosition, parentNeighbor, PI,        &
    &                                tem_rotationMatrix, qDirName
  use tem_topology_module,     only: tem_LevelOf, tem_ChildNumber,             &
    &                                tem_FirstIdAtLevel
  use tem_geometry_module,     only: tem_CoordOfId, tem_IdOfCoord,             &
    &                                tem_determine_discreteVector,             &
    &                                tem_ElemSizeLevel
  use tem_tools_module,        only: tem_write, tem_horizontalSpacer
  use tem_grow_array_module,   only: grw_intArray_type, init, append, destroy, &
    &                                empty
  use tem_dyn_array_module,    only: dyn_intArray_type, init, append, expand,  &
    &                                destroy, empty
  use tem_property_module,     only: tem_property_type, prp_solid,             &
    &                                prp_missingNeigh, prp_hasQVal
  use tem_bc_prop_module,      only: tem_bc_prop_type
  use tem_construction_module, only: tem_init_elemLevels, tem_find_allElements,&
    &                                tem_build_verticalDependencies,           &
    &                                tem_build_horizontalDependencies,         &
    &                                tem_eTypeOfTotalPos,                      &
    &                                tem_cleanup_arrays, tem_levelDesc_type,   &
    &                                tem_treeIDinTotal, tem_elemList_dump,     &
    &                                tem_levelNeighbor_type,                   &
    &                                tem_dumpTreeIDlists, tem_dumpPropBits,    &
    &                                tem_debugDependencies
  use tem_element_module,      only: eT_haloElem, eT_ghostFromCoarserElem,     &
    &                                eT_ghostFromFinerElem
  use tem_timer_module,        only: tem_startTimer, tem_stopTimer,            &
    &                                tem_getTimerVal
  use tem_stencil_module,      only: tem_stencilHeader_type,                   &
    &                                tem_stencil_map_ToTreelmDef
  use tem_debug_module,        only: tem_reportStatus, main_debug, dbgUnit
  use tem_reduction_module,    only: tem_GetRealBoundingCube
  use tem_aux_module,          only: tem_abort
  use tem_logging_module,      only: tem_log, tem_toStr
  use tem_subTree_module,      only: tem_write_debugMesh

  ! include musubi modules
  use mus_bc_header_module,          only: boundary_type, glob_boundary_type
  use mus_debug_module,              only: debug_normals, dump_elemDependency
  use mus_global_module,             only: param_type
  use mus_geom_module,               only: mus_geom_type
  use mus_moments_module,            only: mus_init_moments
  use mus_interpolate_header_module, only: interpolation_type
  use mus_scheme_layout_module,      only: mus_scheme_layout_type
  use mus_scheme_type_module,        only: mus_scheme_type
  use mus_pdf_module,                only: pdf_globInfo_type, pdf_data_type
  use mus_statistics_module,         only: mus_stat_type
  use mus_debug_module,              only: debug_dependenciesFromFiner,        &
    &                                      debug_dependenciesFromCoarser,      &
    &                                      debug_connectivity
  use mus_comm_module,               only: mus_init_longBuffers  

  implicit none

  private

  public :: mus_construct
  public :: mus_construct_cleanup

contains

!******************************************************************************!
  !> Initialize Musubi data strucutres based on data provided by Treelm
  !!
  !! Load the mesh and boundary conditions for this process from disk.
  !! Get the level-wise treeID lists and create the required ghost and halo 
  !! elements.
  !! 
  !! This is achieved by a two-folded identification of elements.
  !! 
  !! -# the theoretically required elements are collected based on 
  !! \ref mus_scheme_layout_module::mus_scheme_layout_type "stencil information"
  !! The \ref tem_construction_module::tem_findneighbor "find neighbor routine" 
  !! performs this task for compute fluid elements.
  !! For boundaries which require information from neighbor elements, these 
  !! required \ref tem_topology_module "treeIDs" are collected into the 
  !! \ref mus_bc_header_module
  !! "boundary element type"
  !! -# All required elements are created in the 
  !! \ref tem_construction_module::tem_createleveldescriptor 
  !! "Level Descriptor creation routine"
  !!
  !! \section mus_construction_additionalTasks Additional Tasks
  !! - receive \ref tem_construction_module::tem_buildhorizontaldependencies 
  !! "horizontal"
  !! (within a level for the element updates) 
  !! - and \ref tem_construction::tem_buildverticaldependencies "vertical" 
  !! dependencies (between levels for ghost-interpolations).
  !! - The main state vector and the neighbor lists on which the kernel then
  !!    acts is created
  !! - The MPI buffers are created.
  !! - For each \ref mus_scheme_module "Scheme", the 
  !! \ref tem_construction_module::level_descriptor_type "Level Descriptor" 
  !! is created
  !!
  !! \section mus_construction_result Result 
  !!
  !! After this routine, all data structures for starting the main loop of the
  !! solver are allocated and ready.
  !! 
  subroutine mus_construct( params, geometry, scheme, pdfGlobal, intp )
    !---------------------------------------------------------------------------
    !> global information
    type( pdf_globInfo_type ), intent(inout) :: pdfGlobal
    !> run-time Parameters
    type( param_type ), intent(inout)        :: params
    !> geometric information
    type( mus_geom_type ), intent(inout)     :: geometry
    !> scheme information including fluid, boundary and flow information 
    type( mus_scheme_type ), intent(inout)   :: scheme(:)
    !> interpolation information
    type( interpolation_type ), intent(inout) :: intp
    !---------------------------------------------------------------------------
    integer :: iLevel      ! counter for level
    integer :: iStencil    ! lists with different stencils
    integer :: nStencils   ! number of total stencils for a scheme
    integer :: iScheme     ! counter for different schemes
    integer :: iField      ! counter for different fields
    integer :: stencilPos  ! position of the stencil in the scheme
    !---------------------------------------------------------------------------

    call tem_reportStatus( text  = 'before construction',                      &
      &                    level = 6 )
    call tem_startTimer( params%initLvlDTimer ) 
    
    ! Output basic mesh infos to the screen
    call tem_write( 'Mesh Information' )
    call tem_horizontalSpacer
    call tem_write( 'Mesh name:                  '                             &
      &             //trim(geometry%tree%global%label) )
    call tem_write( 'local number of elements:   ', geometry%tree%nElems )
    call tem_write( 'total number of elements:   ',                            &
      &             geometry%tree%global%nElems ) 
    call tem_write( 'minimum Level of the tree:  ',                            &
      &             geometry%tree%global%minLevel )
    call tem_write( 'maximum Level of the tree:  ',                            &
      &             geometry%tree%global%maxLevel )
    call tem_write( 'bounding cube origin:       ',                            &
      &             geometry%tree%global%origin )
    call tem_write( 'bounding cube size:         ',                            &
      &  real( geometry%tree%global%BoundingCubeLength ))
    call tem_write( 'coarsest dx                 ',                            &
      &  real(tem_ElemSizeLevel( geometry%tree, geometry%tree%global%minLevel)))
    call tem_write( 'finest dx                   ',                            &
      &  real(tem_ElemSizeLevel( geometry%tree, geometry%tree%global%maxLevel)))
    call tem_horizontalSpacer
    call tem_write( 'Boundary information' )
    if( size(geometry%tree%Property) .gt. 0 )                                  &
    call tem_write( 'nElems with boundaries:  ',                               &
     &              geometry%tree%Property(1)%nElems )
    call tem_horizontalSpacer( after  = 1 )

    ! For all the schemes, set up the data structures for the kernel
    ! \todo: SZ: update this subroutine for multiple fields
    do iScheme = 1, pdfGlobal%nSchemes

      ! initialize the moments for each scheme
      call mus_init_moments( me             = scheme( iScheme )%layout%moment, &
        &                    layout         = scheme( iScheme )%layout,        &
        &                    requireMoments =                                  &
        &                      (intp%requireMoments .or.                       &
        &                      (scheme(iScheme)%header%relaxation(1:3) .eq.    &
        &                                                              'mrt')) )

      ! Determine the size of the stencils:
      ! allocate the arrays: one for each boundary type + one for 
      ! the scheme representation
      nStencils = geometry%boundary%nBCtypes * scheme(iScheme)%nFields + 1

      ! check nBC in mesh and lua, for each filed
      do iField = 1, scheme( iScheme )%nFields
        if( geometry%boundary%nBCtypes .ne.                                    &
            & size( scheme( iScheme )%field( iField )%bc )) then
          call tem_write( 'The number of boundary conditions specified'   //   &
            & ' in '//trim( params%solver%configFile )//' for field '//        &
            & trim( scheme( iScheme )%field( iField )%label )// ' in Scheme '//&
            & trim( scheme( iScheme )%header%label )//' does not correspond'// &
            & ' to the ones specified in the mesh' )
          call tem_write( 'Mesh File, nBC : ', geometry%boundary%nBCtypes )
          call tem_write( 'Lua File, nBC  : ',                                 &
            & size( scheme( iScheme )%field( iField )%bc ) )
          call tem_abort()
        end if
      end do ! iField

      scheme( iScheme )%glob%nElems_total      = geometry%tree%nElems

      call tem_write( 'Starting to initialize the geometry for the solver ...' )
      call tem_horizontalSpacer()
      call tem_write( 'Identifying neighbors ...' )

      call tem_stencil_map_toTreelmDef(                                        &
        &                            me = scheme( iScheme )%layout%stencil(1))
  
      ! build the bc lists once for each scheme
      call build_BClists( scheme    = scheme( iScheme ),                       &
        &                 tree      = geometry%tree,                           &
        &                 boundary  = geometry%boundary,                       &
        &                 property  = geometry%tree%Property,                  &
        &                 pdfGlobal = pdfGlobal )

      scheme( iScheme )%layout%stencil(1)%nElems = geometry%tree%nElems

      ! Treat the boundary lists: in addition to the normal stencil treatment
      ! one has to find all those elements along the element list in the stencil
      ! type, which are boundary elements and give them the special
      ! boundary treatment
      do iField = 1, scheme( iScheme )%nFields
        do iStencil = 1, pdfGlobal%nBCs
          stencilPos = (iField-1)*pdfGlobal%nBCs + iStencil + 1
          call mus_findIndividualNeighbor(                                     &
            &      globbc  = scheme( iScheme )%globbc( iStencil ),             &
            &      bc      = scheme( iScheme )%field( iField )%bc( iStencil ), &
            &      tree    = geometry%tree,                                    &
            &      layout  = scheme( iScheme )%layout,                         &
            &      stencil = scheme( iScheme )%layout%stencil( stencilPos ))
        enddo
      enddo
 
      call tem_reportStatus( text  = 'before init_elemL',                      &
        &                    level = 6 )
      ! find the neighbors for all the elements which use this stencil
      call tem_init_elemLevels(                                                &
        &                me             = scheme( iScheme )%glob%levelDesc,    &
        &                boundary       = geometry%boundary,                   &
        &                tree           = geometry%tree,                       &
        &                computeStencil = scheme( iScheme )%layout%stencil(:))
      call tem_reportStatus( text  = 'before find_allElements',                &
        &                    level = 6 )

      ! Create level-wise element lists including fluid, ghost and halo elements
      ! This is the heart of the topologic data structure creation and 
      ! takes most of the compute time, depending on the mesh you have
      call tem_write('Creating level descriptor ...')
      call tem_find_allElements(                                               &
        &                  tree            = geometry%tree,                    &
        &                  levelDesc       = scheme( iScheme )%glob%levelDesc, &
        &                  levelPointer    = pdfGlobal%levelPointer,           &
        &                  computeStencil  = scheme(iScheme)%layout%stencil,   &
        &                  commPattern     = params%commPattern,               &
        &                  cleanup         = .true.,                           &
        &                  reqNesting      = pdfGlobal%nNesting,               &
        &                  dumpTreeIDlists = main_debug%dumpTreeIDlists )

      call tem_reportStatus( text  = 'before horizontal dep',                  &
        &                    level = 6 )
      ! For each of the neighbor lists create the horizontal (neighbor)
      ! relations
      do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
        allocate( scheme( iScheme )%glob%levelDesc( iLevel )%neigh( nStencils ))
      enddo
      do iStencil = 1, nStencils
        call tem_build_horizontalDependencies(                                 &
          &       iStencil       = iStencil,                                   &
          &       levelDesc      = scheme( iScheme )%glob%levelDesc,           &
          &       tree           = geometry%tree,                              &
          &       computeStencil = scheme( iScheme )%layout%stencil(iStencil), &
          &       check          = main_debug%checkDependencies )
      enddo

      !KM: Communuicate property after setting missing Neighbor in 
      !tem_build_horizontalDependencies routine
      if(pdfGlobal%minLevel .ne. pdfGlobal%maxLevel) then
        call tem_write('WARNING: Might require special treatment to')
        call tem_write('         to communicate property for multilevel')
      endif

      do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
        call communicate_property(                                             &
          &   send     = scheme( iScheme )%glob%levelDesc( iLevel )%sendbuffer,&
          &   recv     = scheme( iScheme )%glob%levelDesc( iLevel )%recvbuffer,&
          &   property = scheme( iScheme )%glob%levelDesc( iLevel )%property,  &
          &   flag     = iLevel,                                               &
          &   pattern  = params%commPattern )
        call communicate_property(                                             &
          &   send     = scheme( iScheme )%glob%levelDesc( iLevel )%           &
          &                                            sendbufferFromCoarser,  &
          &   recv     = scheme( iScheme )%glob%levelDesc( iLevel )%           &
          &                                            recvbufferFromCoarser,  &
          &   property = scheme( iScheme )%glob%levelDesc( iLevel )%property,  &
          &   flag     = iLevel,                                               &
          &   pattern  = params%commPattern )
        call communicate_property(                                             &
          &   send     = scheme( iScheme )%glob%levelDesc( iLevel )%           &
          &                                            sendbufferFromFiner,    &
          &   recv     = scheme( iScheme )%glob%levelDesc( iLevel )%           &
          &                                            recvbufferFromFiner,    &
          &   property = scheme( iScheme )%glob%levelDesc( iLevel )%property,  &
          &   flag     = iLevel,                                               &
          &   pattern  = params%commPattern )
      enddo    
     
      call tem_horizontalSpacer( after = 1 )

      ! Update the positions stored in the bc%elem lists to the positions in the 
      ! levelwise total list
      call update_BClists( scheme    = scheme( iScheme ),                      &
        &                  pdfGlobal = pdfGlobal )

      call tem_reportStatus( text  = 'before vertical dep',                    &
        &                    level = 6 )
      ! Get vertical dependencies (between levels for interpolation) 
      call tem_build_verticalDependencies(                                     &
        &        levelDesc = scheme( iScheme )%glob%levelDesc,                 &
        &        minlevel  = pdfGlobal%minLevel )
        
      call tem_reportStatus( text  = 'cleanup',                                &
        &                    level = 6 )
      !MH: \todo{ this has to be de-activated for adaptive / dynamic mesh
      !           runs}
      ! Delete unrequired arrays
      call tem_cleanup_arrays( levelDesc = scheme( iScheme )%glob%levelDesc )

      call tem_reportStatus( text  = 'after cleanup',                          &
        &                    level = 6 )

      call calculate_nElems_scheme( pdfGlobal = pdfGlobal,                     &
        &                           scheme    = scheme( iScheme ),             &
        &                           iScheme   = iScheme )

      do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
        ! allocate the state array 
        allocate( scheme( iScheme )%pdf( iLevel )%state(                       &
          & scheme( iScheme )%pdf( iLevel )%nElems_local *                     &
          & scheme( iScheme )%varSys%nScalars, 2 ))
        ! allocate the connectivity array
        !\todo KM: use single neighbor connectivity for all fields/pdfs
        allocate( scheme( iScheme )%pdf( iLevel )%neigh(                       &
          & scheme( iScheme )%pdf( iLevel )%nElems_local *                     &
          & scheme( iScheme )%varSys%nScalars ))
      end do


      ! reconstruct connectivity vector pdf( iLevel )%neigh including 
      ! bounce back rules
      call reconstruct_connectivity( scheme    = scheme( iScheme ),            &
        &                            pdfGlobal = pdfGlobal )

      call tem_reportStatus( text  = 'before update deps',                     &
        &                    level = 6 )
      ! Only if multilevel
      if( pdfGlobal%minLevel .ne.pdfGlobal%maxLevel ) then
        if( trim(intp%label) .ne. 'compact' ) then
        ! Add additional source elements for fromCoarser ghost elements
        call mus_intp_complete_fromCoarserDep(                                 &
          &                        intp      = intp,                           &
          &                        scheme    = scheme( iScheme ),              &
          &                        pdfGlobal = pdfGlobal )
      
        ! Update the dependencies fromFiner (in 2d, 4 of 8 get removed)
        call mus_intp_update_fromFinerDep(                                     &
          &                        intp      = intp,                           &
          &                        scheme    = scheme( iScheme ),              &
          &                        pdfGlobal = pdfGlobal )
        end if
      end if
      
      
  
      call set_halo_commLinks( scheme( iScheme ), pdfGlobal )

      call tem_reportStatus( text  = 'before init levelB',                     &
        &                    level = 6 )
      ! Create communication buffers
      pdfGLobal%stat%nLinks_comm  = 0
      pdfGLobal%stat%nLinks_total = 0
      do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
        !    for multi-level, still all the links have to be communicated
        !    so, check for regular Buffer, if the max and min level is the same
        call init_levelBuffers(                                                &
          & send       = scheme( iScheme )%glob%levelDesc( iLevel )%sendbuffer,&
          & recv       = scheme( iScheme )%glob%levelDesc( iLevel )%recvbuffer,&
          & pdf        = scheme( iScheme )%pdf(iLevel),                        &
          & requireAll = .false.,                                              &
          & pattern    = params%commPattern,                                   &
          & scheme     = scheme( iScheme ),                                    &
          & stat       = pdfGLobal%stat,                                       &
          & levelDesc  = scheme( iScheme )%glob%levelDesc( iLevel ),           &
          & layout     = scheme( iScheme )%layout )
        
        call init_levelBuffers(                                                &
          & send       = scheme( iScheme )%glob%levelDesc( iLevel )%           &
          &                                             sendbufferFromCoarser, &
          & recv       = scheme( iScheme )%glob%levelDesc( iLevel )%           &
          &                                             recvbufferFromCoarser, &
          & pdf        = scheme( iScheme )%pdf(iLevel),                        &
          & requireAll = .false.,                                              &
          & pattern    = params%commPattern,                                   &
          & scheme     = scheme( iScheme ),                                    &
          & stat       = pdfGLobal%stat,                                       &
          & levelDesc  = scheme( iScheme )%glob%levelDesc( iLevel ),           &
          & layout     = scheme( iScheme )%layout )
        
        call init_levelBuffers(                                                &
          & send       = scheme( iScheme )%glob%levelDesc( iLevel )%           &
          &                                               sendbufferFromFiner, &
          & recv       = scheme( iScheme )%glob%levelDesc( iLevel )%           &
          &                                               recvbufferFromFiner, &
          & pdf        = scheme( iScheme )%pdf(iLevel),                        &
          & requireAll = .false.,                                              &
          & pattern    = params%commPattern,                                   &
          & scheme     = scheme( iScheme ),                                    &
          & stat       = pdfGLobal%stat,                                       &
          & levelDesc  = scheme( iScheme )%glob%levelDesc( iLevel ),           &
          & layout     = scheme( iScheme )%layout )
    enddo

      ! Debug the level lists, dependencies, buffers ...
      ! \todo: DEB: update these debug calls
      if( main_debug%dumpTreeIDlists ) then
        call tem_write( 'Debugging ...     ', 0, before = 2, after = 1)
        call tem_log( level = 1,                                               &
          &           msg   = 'Debugging ...     0' )
        do iLevel = geometry%tree%global%minlevel,                           &
          &                                     geometry%tree%global%maxLevel
          call tem_elemList_dump(                                            &
            &  me         = scheme( iScheme )%glob%levelDesc( iLevel )%elem, &
            !MH: just setting the unit to 5 beacuse need to hand over a scalar. please check!!
            &  nUnit      = main_debug%logger%funit(5),                         &
            &  string     = '  writing treeID list ',                        &
            &  stencil    = .true.,                                          &
            &  stencilPos = .true.,                                          &
            &  compact    = .true. )
        end do
      end if
      if( main_debug%dumpPropBits )                                          &
      call tem_dumpPropBits( levelDesc = scheme( iScheme )%glob%levelDesc,   &
        &                    minlevel  = pdfGlobal%minLevel)

      if( main_debug%dumpAuxLists ) then
      ! Write out detailed information about the boundary lists
      !call debug_boundaries(tree, glob, pdf, glob%levelDesc, boundary, bc ) 

      ! Info about the connectivity array (neighbors!)
      call debug_connectivity( scheme    = scheme( iScheme ),                &
        &                      pdfGlobal = pdfGlobal )
      end if

      call tem_write_debugMesh( globtree  = geometry%tree,                   &
        &                       levelDesc = scheme(iScheme)%glob%levelDesc )

      if( main_debug%dumpDependencies) then
        ! Check the dependencies from finer
        call debug_dependenciesFromFiner( pdfGlobal,                         &
              scheme( iScheme )%pdf, scheme( iScheme )%glob )                
        call debug_dependenciesFromCoarser( pdfGlobal,                       &
              scheme( iScheme )%pdf, scheme( iScheme )%glob )
        if ( size( geometry%tree%Property ) > 0 ) then
          call debug_normals( pdfGlobal = pdfGlobal,                         &
            &                 tree      = geometry%tree,                     &
            &                 scheme    = scheme( iScheme) )
        end if
      end if
    end do ! SchemeLoop

    call calc_commAmount( pdfGlobal = pdfGlobal )

    call tem_stopTimer( params%initLvlDTimer ) 
    call tem_horizontalSpacer( before = 1 )
    ! assign the init time to the runtime type
    params%timeControl%runTime%init = tem_getTimerVal(params%initLvlDTimer )
    call tem_write( 'Done initializing geometry. Took [s]',                    &
      &  real( params%timeControl%runTime%init ))
    call tem_horizontalSpacer( )
    call tem_horizontalSpacer( after  = 1 )

    call tem_reportStatus( text  = 'after the construction',                   &
      &                    level = 6 )

  end subroutine mus_construct
!******************************************************************************!


!******************************************************************************!
  !> Initialize the communication buffers for a single level
  subroutine communicate_property( send, recv, property, flag, pattern )
    !--------------------------------------------------------------------------- 
    !> Communication structure to send and receive property
    type(tem_communication_type), intent(inout) :: send, recv
    !> property to communicate
    integer(long_k), intent(inout) :: property(:)
    !> communication flag
    integer, intent(in) :: flag
    !>
    type(tem_commPattern_type), intent(in) :: pattern
    !---------------------------------------------------------------------------
    call mus_init_longBuffers( send    = send,                                 &
      &                        recv    = recv,                                 &
      &                        pattern = pattern )
    call pattern%exchange_long( send         = send,                           &
      &                         recv         = recv,                           &
      &                         state        = property,                       & 
      &                         message_flag = flag )

  end subroutine communicate_property  
!******************************************************************************!


!******************************************************************************!
  !> Initialize the communication buffers for a single level
  subroutine init_levelBuffers( send, recv, pdf, pattern, levelDesc, layout,   &
    &                           requireAll, scheme, stat )
    !---------------------------------------------------------------------------
    !> Communication structure to initialize
    type(tem_communication_type) :: send, recv
    !>
    type( pdf_data_type), intent(in) :: pdf
    !>
    type( tem_commpattern_type ), intent(in)  :: pattern
    !>
    type( tem_levelDesc_type ), intent(in):: levelDesc
    !>
    type( mus_scheme_layout_type ), intent(in) :: layout
    !> Statistics
    type( mus_stat_type ), intent(inout) :: stat
    !---------------------------------------------------------------------------
    type( mus_scheme_type ), intent(in)   :: scheme !< fluid, bnd and flow info
    !> different place to take values from for interpolation
    logical :: requireAll
    !---------------------------------------------------------------------------

    ! only initialize recvBuffers and then communicate the desired
    ! values to the neighbor processes.
    call init_recvBuffers(me         = recv,                                   &
      &                   pdf        = pdf,                                    &
      &                   nDOFs      = layout%stencil(1)%QQ,                   &
      &                   requireAll = requireAll,                             &
      &                   levelDesc  = levelDesc,                              &
      &                   stat       = stat,                                   &
      &                   scheme     = scheme,                                 &
      &                   pattern    = pattern )

    call init_sendBuffers(send       = send,                                   &
      &                   recv       = recv,                                   &
      &                   pdf        = pdf,                                    &
      &                   nDOFs      = layout%stencil(1)%QQ,                   &
      &                   requireAll = requireAll,                             &
      &                   scheme     = scheme,                                 &
      &                   pattern    = pattern )

  end subroutine init_levelBuffers
!******************************************************************************!


!******************************************************************************!
  !> Calculate level-wise, total and local as well as global numbers of elements
  subroutine calculate_nElems_scheme( pdfGlobal, scheme, iScheme )
    !---------------------------------------------------------------------------
    !>
    type( pdf_globInfo_type), intent(inout) :: pdfGlobal
    !>
    type( mus_scheme_type )              :: scheme
    !> number of current scheme
    integer, intent(in) :: iScheme
    !--------------------------------------------------------------------------- 
    integer :: iLevel         !< level of source elements
    integer(kind=long_k), allocatable           :: ElemCount(:)
    integer :: iErr
    integer :: nElems_solve
    !---------------------------------------------------------------------------
    ! Count the amount of different element types
    ! counting levelwise elements ...
    scheme%glob%nElems_localFluid = 0
    scheme%glob%nElems_localSolve = 0
    call tem_horizontalSpacer
    call tem_log( level = 3,                                                   &
      &           msg   = 'Elements per level for scheme '//                   &
      &                   trim( tem_toStr( iScheme )))
      
    !call tem_write('WARNING: Not computing on ghostFromFiner!!')
    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      scheme%glob%nElems_localFluid = scheme%glob%nElems_localFluid            &
        &          + scheme%glob%levelDesc(iLevel)%nElems_fluid
      scheme%glob%nElems_localSolve = scheme%glob%nElems_localSolve            &
        &          + scheme%glob%levelDesc(iLevel)%nElems_fluid                &
        &          + scheme%glob%levelDesc(iLevel)%nElems_ghostFromCoarser
      scheme%pdf( iLevel )%nElems_fluid                                        &
        &          = scheme%glob%levelDesc( iLevel )%nElems_fluid
      scheme%pdf( iLevel )%nElems_ghostFromCoarser                             &
        &          = scheme%glob%levelDesc( iLevel )%nElems_ghostFromCoarser
      scheme%pdf( iLevel )%nElems_ghostFromFiner                               &
        &          = scheme%glob%levelDesc( iLevel )%nElems_ghostFromFiner
      scheme%pdf( iLevel )%nElems_ghost                                        &
        &          = scheme%glob%levelDesc( iLevel )%nElems_ghostFromCoarser   &
        &          + scheme%glob%levelDesc( iLevel )%nElems_ghostFromFiner     
      scheme%pdf( iLevel )%nElems_halo                                         &
        &          = scheme%glob%levelDesc( iLevel )%nElems_halo
      scheme%pdf( iLevel )%nElems_local                                        &
        &          = scheme%glob%levelDesc( iLevel )%nElems_fluid              &
        &          + scheme%glob%levelDesc( iLevel )%nElems_ghostFromFiner     &
        &          + scheme%glob%levelDesc( iLevel )%nElems_ghostFromCoarser   &
        &          + scheme%glob%levelDesc( iLevel )%nElems_halo               
      nElems_solve = scheme%glob%levelDesc( iLevel )%nElems_fluid              &
        &          + scheme%glob%levelDesc( iLevel )%nElems_ghostFromFiner     &
        &          + scheme%glob%levelDesc( iLevel )%nElems_ghostFromCoarser

      ! compute only fluid nodes if single level only
      if ( pdfGlobal%minLevel .eq. pdfGlobal%maxLevel ) then
        scheme%pdf( iLevel )%nElems_solve = scheme%pdf( iLevel )%nElems_fluid
      else  
        !scheme%pdf( iLevel )%nElems_solve = scheme%pdf( iLevel )%nElems_local
        scheme%pdf( iLevel )%nElems_solve = nElems_solve
      end if

      ! some debug output
      call tem_log( level = 4,                                                 &
        &           msg   = '  Level '//trim( tem_toStr( iLevel )))
      call tem_log( level = 4,                                                 &
        &           msg   = '    fluid '//                                     &
        &                 trim( tem_toStr( scheme%pdf( iLevel )%nElems_fluid )))
      call tem_log( level = 4,                                                 &
        &           msg   = '    ghostFromCoarser '//                          &
        &      trim( tem_toStr( scheme%pdf( iLevel )%nElems_ghostFromCoarser )))
      call tem_log( level = 4,                                                 &
        &           msg   = '    ghostFromFiner  '//                           &
        &        trim( tem_toStr( scheme%pdf( iLevel )%nElems_ghostFromFiner )))
      call tem_log( level = 4,                                                 &
        &           msg   = '    halo             '//                          &
        &                  trim( tem_toStr( scheme%pdf( iLevel )%nElems_halo )))
    end do                         
    call tem_horizontalSpacer
    ! done counting levelwise  elements

    ! Calculate the total number of fluid elements in ALL LEVELS and processes
    allocate( ElemCount( pdfGlobal%maxLevel ))
    do iLevel = pdfGlobal%minLevel,pdfGlobal%MaxLevel
      ElemCount( iLevel ) = scheme%pdf( iLevel )%nElems_solve
    end do
    call mpi_allreduce( ElemCount, scheme%glob%nElems_solveLevel(:),           &
        pdfGlobal%maxLevel, mpi_integer8, mpi_sum, process%comm, iErr )
    do iLevel = pdfGlobal%minLevel,pdfGlobal%MaxLevel
      ElemCount( iLevel ) = scheme%pdf( iLevel )%nElems_local
    end do
    call mpi_allreduce( ElemCount, scheme%glob%nElems_allLevel(:),             &
         pdfGlobal%maxLevel, mpi_integer8, mpi_sum, process%comm, iErr )
    do iLevel = pdfGlobal%minLevel,pdfGlobal%MaxLevel
      ElemCount( iLevel ) = scheme%pdf( iLevel )%nElems_fluid
    end do
    call mpi_allreduce( ElemCount, scheme%glob%nElems_totalLevel(:),           &
         pdfGlobal%maxLevel, mpi_integer8, mpi_sum, process%comm, iErr )
    deallocate(ElemCount)
    scheme%glob%nElems_totalFluid = 0_long_k
    scheme%glob%nElems_totalSolve = 0_long_k
    scheme%glob%nElems_overall =                                               &
      &     sum( scheme%glob%nElems_allLevel(                                  &
      &          pdfGlobal%minLevel:pdfGlobal%maxLevel ))
    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      scheme%glob%nElems_totalFluid = scheme%glob%nElems_totalFluid            &
                             + scheme%glob%nElems_totalLevel( iLevel )
      scheme%glob%nElems_totalSolve = scheme%glob%nElems_totalSolve            &
                             + scheme%glob%nElems_solveLevel( iLevel )
      call tem_write( ' Level            :', iLevel )
      call tem_write( '  number of Fluids:', scheme%glob%nElems_totalLevel( iLevel ))
    end do
    call tem_write( 'Numbers of elements for scheme', iScheme )
    call tem_write( 'Total number of Fluids:', scheme%glob%nElems_totalFluid )
    call tem_write( 'Total number to Solve:', scheme%glob%nElems_totalSolve )
    call tem_write( 'Total number overall:', scheme%glob%nElems_overall )

    call tem_write( '' )

  end subroutine calculate_nElems_scheme
!******************************************************************************!


!******************************************************************************!
  !> Calculate level-wise, total and local as well as global numbers of elements
  subroutine calc_commAmount( pdfGlobal )
    !---------------------------------------------------------------------------
    !>
    type( pdf_globInfo_type), intent(inout) :: pdfGlobal
    !--------------------------------------------------------------------------- 
    integer :: iErr
    integer :: nLinks_comm, nLinks_total
    !---------------------------------------------------------------------------

    if ( process%comm_size .gt. 1 ) then 
      call mpi_reduce( pdfGlobal%stat%nLinks_comm, nLinks_comm, 1, mpi_integer,&
        &              mpi_sum, 0, process%comm, iErr )
      call mpi_reduce( pdfGlobal%stat%nLinks_total, nLinks_total, 1,           &
        &              mpi_integer, mpi_sum, 0, process%comm, iErr )
      pdfGlobal%stat%nLinks_comm  = nLinks_comm
      pdfGlobal%stat%nLinks_total = nLinks_total

      if( nLinks_total .gt. 0 ) then 
        call tem_write( 'Communication amount compressed to [%]:',             &
          &          real( nLinks_comm )/real( nLinks_total )*100.)
      endif
    endif
    
  end subroutine calc_commAmount
!******************************************************************************!

 
!******************************************************************************!
  !> Complete the required interpolation dependencies and calculate the 
  !! distances for the weights.
  !! 
  !! The depdendencies for ghostFromCoarser elements initially only include its 
  !! parent element.
  !! Additional dependencies Y for the element x are found which were not 
  !! included in the construction routine before hand. However, they are 
  !! defined to be inherently part of the stencil and hence exist locally and 
  !! can simply be added to the dependencies. Two different ways are 
  !! implemented. 
  !! - compact interpolation stencil
  !! \verbatim 
  !!  Y       Y   
  !!                             
  !!        x   o              
  !!  Y       O                     
  !!        o   o
  !! \endverbatim
  !! - extended interpolation stencil
  !!   the complete compute stencil is included into the dependency list
  !! \verbatim 
  !!  Y       Y       Y
  !!                             
  !!        x   o              
  !!  Y       O       Y            
  !!        o   o
  !!
  !!  Y       Y       Y
  !! \endverbatim
  !! 
  !! Then, calculate the weights based on  barycenter distance of source and
  !! target elements.
  !!
  !! Note: The view here is according to how the interpolation will be done.
  !!       The coarse level does the interpolation for the finer level.
  !!       Thus, we access the information from the levelDesc with the
  !!       targetLevel and sourceLevel perspective.
  !!
  subroutine mus_intp_complete_fromCoarserDep( intp, scheme, pdfGlobal )
    !---------------------------------------------------------------------------
    !>
    type( mus_scheme_type )              :: scheme
    !>
    type( interpolation_type ), intent(inout) :: intp      
    !>
    type( pdf_globInfo_type), intent(in) :: pdfGlobal

    !--------------------------------------------------------------------------- 
    integer(kind=long_k) :: nTreeID, tOffset, curTreeID
    integer :: sourceLevel  ! level of source elements
    integer :: sourceElem   ! treeId of current source element
    integer :: targetLevel  ! level of target elements
    integer :: targetElem   ! treeId of current source element
    integer :: iElem        ! current target element (for outer loop)
    integer :: iSourceElem  ! current source element (for inner loop)
    integer :: nSourceElems ! number of source elements for the current target
    ! maximum number of neighbor elements
    ! Compact stencil -> 8, quadraticExtended = stencil%QQ
    integer :: nMaxElems 
    integer :: nNeighElems
    integer :: elemPos
    integer :: childNum, iNeigh
    integer :: neighOffset(3), x(4), parentPos( 3,8)
    integer :: iStencilElem, iLoop
    integer :: foundDirInStencil
    logical :: force_compact
    logical :: completed
    logical :: validElement
    integer :: dPos, iPos ! position, where a source element was stored
    ! weight for each source element depending on distance to target element
    real(kind=rk),allocatable :: weight(:)
    real(kind=rk) :: targetBary(3) ! barycenter of current target element
    ! resulting distance of bary centers from target to uncle
    real(kind=rk) :: distance(3)
    integer, allocatable :: stencil(:,:)
    ! intermediate storage
    type( grw_intArray_type ) :: tElem 
    type( dyn_intArray_type ) :: tSourceFromCoarser
    logical :: verb
    real(kind=rk) :: sumWeight
    integer(kind=long_k ) :: targetTID
    !---------------------------------------------------------------------------
    verb = .false.
    if( intp%compact ) then
      nMaxElems = max( 1, intp%nSourcesCompactFromCoarser_required)
    else
      nMaxElems = scheme%layout%stencil(1)%QQ
    end if

!    call map_compactToComputeStencil( intp = intp, &
!      &            stencil = scheme%layout%stencil(1)) 
    
    allocate( stencil( 3, nMaxElems ))
    allocate( weight( nMaxElems ))

    do sourceLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel - 1 
      targetLevel = sourceLevel + 1
      ! initialize the levelwise source element list for the buffer 
      call init( me = scheme%glob%levelDesc( targetLevel )%sourceFromCoarser,  &
        &        unique = .true. )
      call init( me = scheme%glob%levelDesc( targetLevel )%intpFromCoarser_high)
      call init( me = scheme%glob%levelDesc( targetLevel )%intpFromCoarser_low )

      ! Treat all the fromCoarser elements
      do iElem = 1, scheme%pdf( targetLevel )%nElems_ghostFromCoarser
        ! Get the target element which has to be updated
        targetElem = iElem + scheme%glob%levelDesc( targetLevel )%             &
          &                                offset( 1, eT_ghostFromCoarserElem )
        targetTID = scheme%glob%levelDesc( targetLevel )%total( targetElem )

!        call tem_debug( level = 10,                                            &
!          &             msg   = 'found the targetTID' )

        ! Re-assign initial nMaxElems definition
        nNeighElems = nMaxElems
        force_compact = .false.
        completed = .false.
        ! Loop twice: first, try to find the extended stencil
        !             second: if not all elements were found, try to find the
        !                     compact stencil
        do iLoop = 1, 2
          ! If all source elements were found in the first round, then exit.
          ! No second loop required.
          if( completed ) exit
          if( iLoop ==1 ) then
            ! Store intermediate element dependencies (store complete grw type)
            tElem = scheme%glob%levelDesc( targetLevel )%                      &
              &                                    depFromCoarser( iElem )%elem
            ! And the list of all so far detected source elements (might 
            ! be deleted)
            tSourceFromCoarser = scheme%glob%levelDesc( targetLevel )%         &
              &                                               sourceFromCoarser
            if( tSourceFromCoarser%nvals .gt. 0 ) then
!              call tem_debug( level = 10,                                      &
!                &             msg   = '1st loop: shelving srcFromCoarser'//    &
!                &               trim( tem_toStr( val = tSourceFromCoarser%     &
!!                &                     val( 1: tSourceFromCoarser%nvals ),      &
!                &                                sep = '' )))
            end if
          end if
          ! If not all source elements were found in the first round, it means 
          ! that the compact neighborhood must be reconstructed.
          if( iLoop == 2 ) then
!            call tem_debug( level = 10,                                        &
!              &             msg   = '2nd loop: ' )
            ! If 2nd loop: use compat stencil definition
            nNeighElems = intp%nSourcesCompactFromCoarser_required
            force_compact = .true.
            ! re-create the lists for this element
            call destroy( me = scheme%glob%levelDesc( targetLevel )%           &
              &                              depFromCoarser(iElem)%elemBuffer )
            call destroy( me = scheme%glob%levelDesc( targetLevel )%           &
              &                                    depFromCoarser(iElem)%elem )
            call destroy( me = scheme%glob%levelDesc( targetLevel )%           &
              &                                             sourceFromCoarser )
            ! recover the original element dependencies of before the first round
            ! replace current empty dyn list with the one stored above 
            scheme%glob%levelDesc( targetLevel )%depFromCoarser( iElem )%elem  &
              &                                                         = tElem
            ! recover list of all so far detected source elements from stored
            ! above
            scheme%glob%levelDesc( targetLevel )%sourceFromCoarser             &
              &                                            = tSourceFromCoarser
          end if

          ! initialize array for storing the positions of the source elements in
          ! the levelwise buffer
          call init( me     = scheme%glob%levelDesc( targetLevel )%            &
            &                               depFromCoarser(iElem)%elemBuffer,  &
            &        length = nNeighElems )
          stencil = 0
          
          ! Run over all the source elements which are there
          if( scheme%glob%levelDesc( targetLevel )%depFromCoarser( iElem )     &
            &            %elem%nVals .gt. 1) then
            call tem_write( 'Warning: encountered more than one parent '//     &
              &             'elements, for which we are completing the '//     &
              &             'dependencies')
            call tem_write( ' depFromCoarser '//trim( tem_toStr(               &
              &  val = scheme%glob%levelDesc( targetLevel )%depFromCoarser     &
              &        ( iElem )%elem%val( 1:scheme%glob%levelDesc             &
              &        ( targetLevel )%depFromCoarser( iElem )%elem%nVals ),   &
              &  sep = '' )))
          end if
         
          !--------------- treat   parent element (first source) --------------
          ! There is only one source element, namely the parent one.
          ! Therefore I need to fill up the neighbors of the parent according
          ! to the stencil
          iSourceElem = 1
          ! Reset the source element entries
          nSourceElems = scheme%glob%levelDesc( targetLevel )                  &
            &                               %depFromCoarser( iElem )%elem%nVals

!          call tem_debug( level = 10,                                          &
!            &             msg   = 'Initial entries in deps: '//                &
!            &                     trim( tem_toStr( nSourceElems )))
!          call tem_debug( level = 10,                                          &
!            &             msg   = 'content: '//                                &
!            &   trim( tem_toStr( val = scheme%glob%levelDesc( targetLevel )%   &
!            &         depFromCoarser( iElem )%elem%val( 1: nSourceElems ),     &
!            &                    sep = '' )))

          scheme%glob%levelDesc( targetLevel )                                 &
            &          %depFromCoarser( iElem )%elem%val( nSourceElems+1:) = -1
          weight = 0._rk
          ! start with treating the parent element, which is located at the 
          ! first position in the dependency list
          ! Later on, proceed to siblings of parent in the target element's 
          ! directions from its parent element
          sourceElem = scheme%glob%levelDesc( targetLevel )                    &
            &                  %depFromCoarser( iElem )%elem%val( iSourceElem )
          ! add the parent to the list of source elements on this level
          ! delete parent element from the list
          call empty( me = scheme%glob%levelDesc( targetLevel )                &
            &                                   %depFromCoarser( iElem )%elem )
          !--------------- treated parent element (first source) ---------------
!          call tem_debug( level = 10,                                          &
!            &             msg   = trim( tem_toStr( iLoop ))//                  &
!!            &                     'content of dep after adding parent'//       &
!            &                     trim( tem_toStr(                             &
!            & val = scheme%glob%levelDesc( targetLevel )%depFromCoarser        &
!            &       ( iElem )%elem%val( 1:scheme%glob%levelDesc                &
!            &       ( targetLevel )%depFromCoarser( iElem )%elem%nVals ),      &
!            & sep = '' )))
         
          !--------------- treat next source elements --------------------------
          ! Get treeID of first source element
          nTreeID = scheme%glob%levelDesc( sourceLevel )%total( sourceElem )
          x = tem_CoordOfId( nTreeID ) 
          tOffset = tem_FirstIdAtLevel( x(4) )
         
          ! target (ghost) treeID, to find out into which direction the
          ! neighbors have to be searched
          curTreeID = scheme%glob%leveldesc( targetlevel )%total( targetElem )
          childNum = tem_ChildNumber( curTreeID )
          ! Store the child number 
          scheme%glob%levelDesc( targetLevel )%depFromCoarser( iElem )%childNum&
            &                                                        = childNum
         
          ! barycenter computation:
          ! Distance from one barycenter to the next = 1
          targetBary = 0.25_rk * real(childPosition(childNum, 1:3), kind=rk)
          ! Store the relative position for later interpolation.
          ! We store the position relative to the parent one.
         
          if( intp%compact .or. force_compact ) then
          ! Here we store the position relative to the enclosing elements on 
          ! the parent level. The children on the right, which have local coord 
          ! -0.25 become 0.75, children on the left with +0.25 become 0.25
            scheme%glob%levelDesc( targetLevel )%depFromCoarser( iElem )       &
              &                           %coord(:) = abs( targetBary - 0.5_rk )
          else
            scheme%glob%levelDesc( targetLevel )%depFromCoarser( iElem )       &
              &                                           %coord(:) = targetBary
          end if
          ! Now loop over the maximum number of elements = computeStencil
          do iNeigh = 1, nNeighElems
            if( intp%compact .or. force_compact ) then
!              foundDirInStencil = get_DirFrom
              ! If interpolation stencil /= compute stencil, try to match the 
              ! offsets
              !\todo: move this into a function, as it is needed above in 
              ! mapping compact->ext
              ! dir = map_compactToStencil( iNeigh, childNum,                  &
              !   &                                  scheme%layout%stencil(1) )
              x = tem_coordOfId( int(iNeigh, kind=long_k))*2 -1
              parentPos(:, iNeigh)  =                                          &
                &           matmul(tem_rotationMatrix(:,:, ChildNum ), x(1:3) )
              neighOffset  = (parentPos(:,iNeigh)  - parentPos(:, 1))/2
!              call tem_debug( level = 10,                                      &
!                &             msg   = trim( tem_toStr( iLoop ))//              &
!                &                     ' original coord '//                     &
!                &                     trim( tem_toStr( val = x(1:3),           &
!                &                                      sep = '' )) )
!              call tem_debug( level = 10,                                      &
!                &             msg   = trim( tem_toStr( iLoop ))//              &
!                &                     ' parentOffset   '//                     &
!                &                     trim( tem_toStr( val=parentPos(:,iNeigh),&
!                &                                      sep='' )))
!              call tem_debug( level = 10,                                      &
!                &             msg   = trim( tem_toStr( iLoop ))//              &
!                &                     ' neighOffset    '//                     &
!                &                     trim( tem_toStr( val = neighOffset,      &
!                &                                      sep = '' )))
!              call tem_debug( level = 10,                                      &
!                &             msg   = trim( tem_toStr( iLoop ))//              &
!                &                     ' content of dep '//                     &
!                &                     trim( tem_toStr(                         &
!                & val = scheme%glob%levelDesc( targetLevel )%depFromCoarser    &
!                &       (iElem )%elem%val( 1:scheme%glob%levelDesc             &
!                &       ( targetLevel )%depFromCoarser( iElem )%elem%nVals ),  &
!                & sep = '' )))

              ! Match the found neighOffset with the stencil. Is this direction
              ! available?
              foundDirInStencil = -1
              do iStencilElem = 1, scheme%layout%stencil(1)%QQ
                if( neighOffset(1) .eq. scheme%layout%stencil(1)%cxDir( 1,     &
                  &                                       iStencilElem ) .and. &
                  & neighOffset(2) .eq. scheme%layout%stencil(1)%cxDir( 2,     &
                  &                                       iStencilElem ) .and. &
                  & neighOffset(3) .eq. scheme%layout%stencil(1)%cxDir( 3,     &
                  &                                        iStencilElem )) then
                  foundDirInStencil = iStencilElem
                end if
              end do
            else
              ! If interpolation stencil = compute stencil, simply take the 
              ! offsets
              neighOffset = scheme%layout%stencil(1)%cxDir( :, iNeigh )
!              call tem_debug( level = 10,                                      &
!                &             msg   = trim( tem_toStr( iLoop ))//              &
!                &                     ' Get offset from stencil '//            &
!                &                     trim( tem_toStr( iNeigh )))
              foundDirInStencil = iNeigh
            end if
            
            if( foundDirInStencil .gt. 0) then
              ! Get the corresponding treeID from the stencil neighbor
              ! information
              distance(1:3) = real(neighOffset(1:3),kind=rk) - targetBary
              if(foundDirInStencil.eq.scheme%layout%Stencil(1)%restPosition)then
                ! it's the parent element. -> no neighbor info required or
                ! available
                elemPos = sourceElem 
              else
                ! it's a neihgbor of the parent element -> 
                ! access neighbor information.
                elemPos = scheme%glob%levelDesc( sourceLevel )%neigh(1)%       &
                  &                   nghElems( foundDirInStencil, sourceElem )
              end if
              if( elemPos .gt. 0 ) then
!                call tem_debug( level = 10,                                    &
!!                  &             msg   = trim( tem_toStr( iLoop ))//            &
!                  &                     ' valid Elem '//                       &
!                  &                     trim( tem_toStr(                       &
!                  &     scheme%glob%levelDesc( sourceLevel )%total( elemPos ))))

                validElement = .true. 
              else
                validElement = .false.
!                call tem_debug( level = 10,                                    &
!                  &             msg   = trim( tem_toStr( iLoop ))//            &
!                  &                     ' INVALID Elem '//                     &
!                  &                     trim( tem_toStr(                       &
!                  &     scheme%glob%levelDesc( sourceLevel )%total( elemPos ))))
              end if
              if( validElement ) then
                ! Add the found element to the dependency element list 
                call append( me  = scheme%glob%levelDesc( targetLevel )%       &
                  &                              depFromCoarser( iElem )%elem, &
                  &          val = elemPos )
                dPos = scheme%glob%levelDesc( targetLevel )%depFromCoarser(    &
                  &                                          iElem )%elem%nVals
                ! Store to the levelwise source element list
                call append( me  = scheme%glob%levelDesc( targetLevel )%       &
                  &                                         sourceFromCoarser, &
                  &          val = elemPos,                                    &
                  &          pos = iPos )
                ! ... and store the position in the levelwise source element 
                ! list for each ghost dep
                call append( me  = scheme%glob%levelDesc( targetLevel )%       &
                  &                         depFromCoarser( iElem )%elemBuffer,&
                  &          val = iPos )
!                call tem_debug( level = 10,                                    &
!                  &             msg   = 'tgtID loop '//                        &
!                  &                     trim( tem_toStr( iLoop ))//            &
!                  &                     trim( tem_toStr( iNeigh ))//           &
!                  &                     ' dPos '//trim( tem_toStr( dPos ))//   &
!                  &                     ' nMaxElems '//                        &
!                  &                     trim( tem_toStr( nMaxElems )))
                if( intp%compact ) then
                  stencil( :, dPos ) = (parentPos( :, iNeigh) +1)/2
                else
                  stencil(:,dPos ) = neighOffset
                end if
                weight( dPos ) = (1._rk-abs(distance(1)))                      &
                  &            * (1._rk-abs(distance(2)))                      &
                  &            * (1._rk-abs(distance(3)))
              end if
            end if
          end do
          nSourceElems = scheme%glob%levelDesc( targetLevel )%depFromCoarser(  &
            &                                                iElem )%elem%nVals
         
          if ( nSourceElems .lt. intp%nSourcesFromCoarser_required ) then
            completed = .false.
          else
            completed = .true.
          end if
        end do
        if( .not. completed .and. iLoop == 1) then
          call tem_write( 'I got here although it shouldnt have happened '//   &
            &             'inside intp_update_FromCoarser' )
          call tem_abort
        end if
        if ( nSourceElems .lt. intp%nSourcesFromCoarser_required ) then
          ! not enough source elements -> assign to lower order interpolation
          ! scheme
          call append( me  = scheme%glob%levelDesc( targetLevel )%             &
            &                                             intpFromCoarser_low, &
            &          val = iElem )
        else
          ! enough source elements -> assign to higher order interpolation
          ! scheme
          call append( me  = scheme%glob%levelDesc( targetLevel )%             &
            &                                            intpFromCoarser_high, &
            &          val = iElem )
        end if
        allocate( scheme%glob%levelDesc( targetLevel )%depFromCoarser( iElem )%&
          &                                         stencil( 3, nSourceELems ))
        scheme%glob%levelDesc( targetLevel )%depFromCoarser( iElem )%          &
          &         stencil( :, 1:nSourceELems ) = stencil( :, 1:nSourceElems )
        ! Now I have found all the neighbors there are.
        ! Continue with a calculation of the weights
        allocate( scheme%glob%levelDesc( targetLevel )%depFromCoarser( iElem ) &
          &             %weight( 1:nSourceElems ))
        
        ! Normalize weight, so that sum(weight) = 1
        sumWeight = sum( weight( 1:nSourceElems ))
        weight( 1:nSourceElems ) = weight( 1:nSourceElems ) / sumWeight
        
        ! ... and store, but in the order according to the rotated coordinate 
        ! system
        scheme%glob%levelDesc( targetLevel )%depFromCoarser( iElem )%          &
          &                 weight( 1:nSourceElems ) = weight( 1:nSourceElems )
        if( intp%requireWeights .or. force_compact) then
          if( abs( sum( scheme%glob%levelDesc( targetLevel )%                  &
            &    depFromCoarser( iElem )%weight( 1:nSourceElems )) - 1._rk )   &
            &      .ge. eps ) then

            call tem_write('Error: sum of interpolation weights not 1!')
            call tem_abort
          end if
        end if

      end do ! source%elem loop
      call tem_write(' fromCoarser Interpolation source element information')
      call tem_write('   targetLevel:', targetLevel)
      call tem_write('   # number of ghost elements: ',                        &
        &               scheme%pdf( targetLevel )%nElems_ghostFromCoarser )
      call tem_write('   # higher order elements:',                            &
        &     scheme%glob%levelDesc( targetLevel )%intpFromCoarser_high%nVals )
      call tem_write('   # lower order elements: ',                            &
        &      scheme%glob%levelDesc( targetLevel )%intpFromCoarser_low%nVals )

    end do ! level loop

end subroutine mus_intp_complete_fromCoarserDep
!******************************************************************************!


!******************************************************************************!
  !> Update the dependencies for ghostfromFiner
  !! 
  subroutine mus_intp_update_fromFinerDep( intp, scheme, pdfGlobal )
    !---------------------------------------------------------------------------
    !>
    type( mus_scheme_type )              :: scheme
    !>
    type( interpolation_type ), intent(in) :: intp      
    !>
    type( pdf_globInfo_type), intent(in) :: pdfGlobal
    !---------------------------------------------------------------------------
    integer :: sourceLevel         ! level of source elements
    integer :: targetLevel         ! level of target elements
    integer :: iElem               ! current target element (for outer loop)
    integer :: iSourceElem         ! current source element
    integer :: nSourceElems        ! number of source elements
    integer :: dPos                ! storage position of source element
    !---------------------------------------------------------------------------

    do targetLevel = pdfGlobal%minLevel, pdfGLobal%maxLevel-1
      call init( me = scheme%glob%levelDesc( targetLevel )%sourceFromFiner,    &
        &        unique = .true. )
      call init( me = scheme%glob%levelDesc( targetLevel )%intpFromFiner_high )
      call init( me = scheme%glob%levelDesc( targetLevel )%intpFromFiner_low )
      sourceLevel = targetLevel + 1
      do iElem = 1, scheme%pdf( targetLevel )%nElems_ghostFromFiner
        if( pdfGlobal%nDims .eq. 2 ) then
          ! in two dimensional cases, remove the additional source elements
          ! for the fromFiner ghosts.
          ! There are always eight, provided by tem_construction routine.
          ! For the interpolation we only require 4.
          ! Only consider cases where more than 4 source elements are present
          if( scheme%glob%levelDesc( targetLevel )%depFromFiner( iElem )%      &
            &                                          elem%nVals .gt. 4 ) then
            scheme%glob%levelDesc( targetLevel )%depFromFiner( iElem )%        &
              &                                                  elem%nVals = 4
          end if
        elseif( pdfGlobal%nDims .eq. 1 ) then
          ! Accordingly in one dimensions: only 2 elements required to
          ! interpolate from finer
          if( scheme%glob%levelDesc( targetLevel )%depFromFiner( iElem )%      &
            &                                          elem%nVals .gt. 2 ) then
            scheme%glob%levelDesc( targetLevel )%depFromFiner( iElem )%        &
              &                                                  elem%nVals = 2
          end if
        end if
        nSourceElems = scheme%glob%levelDesc( targetLevel )%                   &
          &                                    depFromFiner( iElem )%elem%nVals
        call init( me     = scheme%glob%levelDesc( targetLevel )%              &
          &                                    depFromFiner(iElem)%elemBuffer, &
          &        length = 8 )
        do iSourceElem = 1, scheme%glob%levelDesc( targetLevel )%              &
          &                                    depFromFiner( iElem )%elem%nVals
          call append( me  = scheme%glob%levelDesc( targetLevel )%             &
            &                                                 sourceFromFiner, &
            &          val = scheme%glob%levelDesc( targetLevel )              &
            &                  %depFromFiner( iElem )%elem%val( iSourceElem ), &
            &          pos = dPos )
          call append( me  = scheme%glob%levelDesc( targetLevel )%             &
            &                                depFromFiner( iElem )%elemBuffer, &
            &          val = dPos )
        end do
        if ( nSourceElems .lt. intp%nSourcesFromFiner_required ) then
          ! not enough source elements -> assign to lower order interpolation 
          ! scheme
          call append( me  = scheme%glob%levelDesc( targetLevel )%             &
            &                                               intpFromFiner_low, &
            &          val = iElem )
        else
          ! enough source elements -> assign to higher order interpolation
          ! scheme
          call append( me  = scheme%glob%levelDesc( targetLevel )%             &
            &                                              intpFromFiner_high, &
            &          val = iElem )
        end if
      end do ! source%elem loop
      call tem_write(' fromFiner Interpolation source element information')
      call tem_write('   targetLevel:', targetLevel)
      call tem_write('   # higher order elements:',                            &
        &        scheme%glob%levelDesc( targetLevel )%intpFromFiner_high%nVals )
      call tem_write('   # lower order elements: ',                            &
        &        scheme%glob%levelDesc( targetLevel )%intpFromFiner_low%nVals )
    end do

  end subroutine mus_intp_update_fromFinerDep
!******************************************************************************!


!******************************************************************************!
  !> Construct the propagation stencil for each element
  !!
  !! the bounce back rules have to be applied here 
  !!
  subroutine reconstruct_connectivity( scheme, pdfGlobal )
    !---------------------------------------------------------------------------
    !>
    type( pdf_globInfo_type ), intent(in) :: pdfGlobal
    !>
    type( mus_scheme_type ) :: scheme
    !---------------------------------------------------------------------------
    integer :: iElem  ! element counter for current element in treeID list
    integer :: iLevel ! level counter
    integer :: iField ! field counter
    integer :: idx_idir ! current direction in global varsys
    integer :: iDir ! counter for current direction in iElem
    integer :: invDir ! inverse direction for iDir
    ! result: from which direction to take the pdf from (inverse for bounceback)
    integer :: GetFromDir
    ! result: from which element to take the pdf from (local for bounceback) 
    integer :: GetFromPos
    integer :: neighPos      ! position of current neighElem in treeID list
    integer(kind=long_k)  :: neighProp, elemProp
    !---------------------------------------------------------------------------

    call tem_write( 'Reconstructing connectivity ...' )

     levelLoop: do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
       elemLoop: do iElem = 1, scheme%pdf( iLevel )%nElems_local
         fieldLoop: do iField = 1, scheme%nFields
           ! get the element property
           elemProp = scheme%glob%levelDesc( iLevel )%property( iElem )
    
           ! Loop over every link
           neighLoop: do iDir  = 1, scheme%layout%stencil(1)%QQ 
             ! check if the direction equals to the rest direction
             ! if available the rest direction is at position QQ of the stencil
             ! (e.g. the d3q6 stencil does not have a rest direction at all)
             if ( iDir .gt. scheme%layout%stencil(1)%QQN  ) then
               ! Rest density takes from itself, iElem and iDir
               GetFromPos = iElem
               GetFromDir = scheme%varSys%Variable%val(iField)%varPos( iDir )
             else ! iDir not rest density
     
               ! For push we need to find different neighbors than for pull
               ! -> ngDir
               neighPos = scheme%glob%levelDesc( iLevel )%neigh(1)%            &
                 &  nghElems( ?NgDir?( iDir, scheme%layout%stencil(1) ), iElem )

               neighProp = 0_long_k
               ! JQ: get neighbor's property
               if( neighPos .gt. 0 ) then
                 neighProp = scheme%glob%levelDesc( iLevel )%property(neighPos)
               end if

              !KM: Treat missing neighbor element as solid
              ! neighPos < 0 : neighbor is a boundary
              ! neighPos = 0 : neighbor is nonExisting element
              ! neighProp solid: neighbor was solidified
              ! elemProp solid: myself was solidified
              if( (neighPos .le. 0) .or. (btest(neighProp, prp_solid))         &
                &  .or. (btest(elemProp, prp_solid))                           &
                &  .or. (btest(neighProp, prp_missingNeigh))                   &
                &  .or. (btest(elemProp, prp_missingNeigh))                    &
                & ) then
                ! 1. Fluid element has boundary in current link direction
                ! 2. neighbor element has fluidify property (still solid now)
                ! Get inverse direction of LOCAL element
                invDir = scheme%layout%stencil(1)%cxDirInv( iDir )
                GetFromDir = scheme%varSys%Variable%val(iField)%varPos(invDir)
                GetFromPos = iElem
                ! I set the source to the target link for halo cells 
                ! with no neighbor in that direction 
              else
                ! neighElem = scheme%glob%levelDesc( iLevel )%total( neighPos )
                GetFromDir = scheme%varSys%Variable%val(iField)%varPos(iDir)
                GetFromPos = neighPos
              end if ! has no boundary -> regular treatment
            end if ! not the 0 velocity

            idx_idir = scheme%varSys%Variable%val(iField)%varPos(iDir)
            scheme%pdf( iLevel )%neigh(                                        &
              &          ?NGPOS?( idx_idir, iElem, scheme%varSys%nScalars )) = &
              &          ?IDX?( GetFromDir, GetFromPos, scheme%varSys%nScalars )

          end do neighLoop
        end do fieldLoop
      end do elemLoop
    end do levelLoop
    call tem_write( 'Done with connectivity.' )

  end subroutine reconstruct_connectivity
!******************************************************************************!


!******************************************************************************!
  !> Select the halo elements which require all links
  !!
  !! Determine which pdfs to communicate 
  !!
  subroutine set_halo_commLinks( scheme, pdfGlobal )
    !---------------------------------------------------------------------------
    !> Global information
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    !> scheme information including fluid, boundary and flow information 
    type(mus_scheme_type), intent(inout) :: scheme
    !---------------------------------------------------------------------------
    integer :: iElem, iLevel, iSourceElem
    integer :: iBnd, iField, iNeigh, iProc
    integer :: targetElem, sourceElem, targetLevel, sourceLevel
    integer :: nSourceElems
    integer :: elemPos, haloOffset
    logical :: comm_reduced
    !---------------------------------------------------------------------------
    comm_reduced = pdfGlobal%comm_reduced
    ! Allocate the halo comm data requirement array (logicals)
    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      allocate(scheme%pdf(iLevel)%haloRequired(scheme%pdf(iLevel)%nElems_halo))
      scheme%pdf(iLevel)%haloRequired(:) = .false.
    end do
    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      if( .not. comm_reduced ) then
        ! Communicate all links of all halos
        scheme%pdf(iLevel)%haloRequired(:) = .true.
      else ! Choose which ones are really required
        !-----------------------------------------------------------------------
        ! GhostFromFiner
        ! check if there are halo deps
        if( iLevel .lt. pdfGlobal%maxLevel ) then
          targetLevel = iLevel 
          sourceLevel = iLevel + 1
          haloOffset =scheme%glob%levelDesc( sourceLevel )%offset(1,eT_haloElem)

          do iElem = 1, scheme%pdf( targetLevel )%nElems_ghostFromFiner
            targetElem = iElem + scheme%glob%levelDesc( targetLevel )          &
              &                      %offset( 1, eT_ghostFromFinerElem )
            nSourceElems = scheme%glob%levelDesc( targetLevel )%               &
              &                                depFromFiner( iElem )%elem%nVals
            do iSourceElem = 1, nSourceElems
              sourceElem = scheme%glob%levelDesc( targetLevel )%               &
                &                 depFromFiner( iElem )%elem%val( iSourceElem )
              call set_requiredLink(                                           &
                &          haloElem  = sourceElem,                             &
                &          offset    = haloOffset,                             &
                &          linkField = scheme%pdf( sourceLevel )%haloRequired )
              ! Check the stencil neighbors of the element as well
             ! do iDir = 1, scheme%layout%stencil(1)%QQN
             !   elemPos  = scheme%glob%levelDesc( iLevel )%neigh(1)%   &
             !     & nghElems( ?NgDir?( iDir, scheme%layout%stencil(1) ), iElem)
             !   if ( elemPos .gt. 0 ) then
             !     if( elemPos .gt. scheme%glob%levelDesc( sourceLevel )%      &
             !       &                            offset(1, eT_haloElem )) then
             !   scheme%pdf( sourceLevel )%haloRequired( elemPos  &
             !   &  - haloOffset) = .true.
             !     end if 
             !   end if 
             ! end do ! IDir
            end do ! iSourceElem
          end do ! iElem
        end if !iLevel < maxLevel
        ! GhostFromFiner
        !-----------------------------------------------------------------------

        !-----------------------------------------------------------------------
        ! GhostFromCoarser
        ! check if there are halo deps
        if( iLevel .lt. pdfGlobal%maxLevel ) then
          targetLevel = iLevel + 1
          sourceLevel = iLevel
          haloOffset =scheme%glob%levelDesc( sourceLevel )%offset(1,eT_haloElem)
          do iElem = 1, scheme%pdf( targetLevel )%nElems_ghostFromCoarser
            targetElem = iElem + scheme%glob%levelDesc( targetLevel )          &
              &                      %offset( 1, eT_ghostFromCoarserElem )
            nSourceElems = scheme%glob%levelDesc( targetLevel )                &
              &            %depFromCoarser( iElem )%elem%nVals
            do iSourceElem = 1, nSourceElems
              sourceElem = scheme%glob%levelDesc( targetLevel )                &
                &          %depFromCoarser( iElem )%elem%val( iSourceElem )
              ! Check the type of the source element. 
              ! If it is a halo -> mark as require all links
              call set_requiredLink(                                           &
                &          haloElem  = sourceElem,                             &
                &          offset    = haloOffset,                             &
                &          linkField = scheme%pdf( sourceLevel )%haloRequired )
              ! stencil neighbors of the gfC
             ! do iDir = 1, scheme%layout%stencil(1)%QQN
             !   elemPos  = scheme%glob%levelDesc( iLevel )%neigh(1)%   &
             !     &        nghElems(                                          &
             !     &         ?NgDir?( iDir, scheme%layout%stencil(1) ), iElem )
             !   if ( elemPos .gt. 0 ) then
             !     if( elemPos .gt. haloOffset) then
             !   scheme%pdf( sourceLevel )%haloRequired( elemPos  &
             !   &  - haloOffset) = .true.
             !     end if 
             !   end if ! elemPos > 0
             ! end do ! iDir 
            end do ! iSourceElem
          end do ! iElem
        end if !level<maxLevel
        ! GhostFromCoarser
        !-----------------------------------------------------------------------

        haloOffset =scheme%glob%levelDesc( iLevel )%offset(1, eT_haloElem ) 
        !-----------------------------------------------------------------------
        ! Run over recvbufferFromCoarser
        do iProc = 1, scheme%glob%levelDesc(iLevel)%recvBufferFromCoarser%nProcs
          do iElem = 1, scheme%glob%levelDesc( iLevel )%recvBufferFromCoarser% &
              &                                             nElemsProc( iProc )
            elemPos = scheme%glob%levelDesc( iLevel )%recvBufferFromCoarser%   &
              &                                   elemPos( iProc )%val( iElem )
            call set_requiredLink(                                             &
              &                 haloElem  = elemPos,                           &
              &                 offset    = haloOffset,                        &
              &                 linkField = scheme%pdf( iLevel )%haloRequired )
          end do
        end do
        !-----------------------------------------------------------------------

        !-----------------------------------------------------------------------
        ! Run over recvbufferFromFiner
        do iProc = 1, scheme%glob%levelDesc( iLevel )%recvBufferFromFiner%nProcs
          do iElem = 1, scheme%glob%levelDesc( iLevel )%recvBufferFromFiner%   &
            &                                               nElemsProc( iProc )
            elemPos = scheme%glob%levelDesc( iLevel )%recvBufferFromFiner%     &
              &                                   elemPos( iProc )%val( iElem )
            call set_requiredLink(                                             &
              &                 haloElem  = elemPos,                           &
              &                 offset    = haloOffset,                        &
              &                 linkField = scheme%pdf( iLevel )%haloRequired )
          end do
        end do
        !-----------------------------------------------------------------------

        !-----------------------------------------------------------------------
        ! Run over all the boundary neighbors
        ! If they require neighbor infos, it is usually integral quantities 
        ! ( macrosopic)  and thus require all links
        do iField = 1, scheme%nFields
          do iBnd = 1, pdfGlobal%nBCs
            do iElem = 1, scheme%globBC( iBnd )%nElems( iLevel )
              do iNeigh = 1, scheme%field( iField )%bc( iBnd )%nNeighs
                ! Get the neighbor source element
                sourceElem = scheme%field( iField )%bc( iBnd )                 &
                  &                     %elems( iLevel )%neigh( iNeigh, iElem )
                call set_requiredLink(                                         &
                  &             haloElem  = sourceElem,                        &
                  &             offset    = haloOffset,                        &
                  &             linkField = scheme%pdf( iLevel )%haloRequired )
              end do ! iNeigh
            end do ! iElem
          end do ! iBnd
        end do  ! iField
        ! Boundary neighbors
        !-----------------------------------------------------------------------
      end if ! comm_reduced?

    end do ! iLevel

  end subroutine set_halo_commLinks
!******************************************************************************!


!******************************************************************************!
  !> 
  !!
  subroutine set_requiredLink( haloElem, offset, linkField )
    !---------------------------------------------------------------------------
    !>
    integer, intent(in) :: haloElem
    !>
    integer, intent(in) :: offset
    !>
    logical, intent(inout) :: linkField(:)
    !---------------------------------------------------------------------------
    if( haloElem .gt. offset ) then
      ! haloRequired is only nElems_halo long, so remove the preceeding
      ! elem nums
      linkField( haloElem - offset ) = .true.
    end if

  end subroutine set_requiredLink
!******************************************************************************!

  
!******************************************************************************!
  !> Update the neighbor stencil positions for the Boundaries
  !!
  !! Before, positions in the original treeIDs were assigned to 
  !! the boundary element lists. Update these now with level-wise positions in 
  !! the \ref tem_construction_module::level_descriptor_type "total list".
  !! When no element is found, the value in nghElems is 0. This has to be
  !! treated in this routine. If an empty neighbor relation is encountered, the 
  !! last valid neighbor is used for all other neighbors
  !!
  subroutine update_BClists( scheme, pdfGlobal )
    !---------------------------------------------------------------------------
    !> Global information
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    !> scheme information including fluid, boundary and flow information 
    type(mus_scheme_type), intent(inout) :: scheme
    !---------------------------------------------------------------------------
    integer :: iElem, iLevel, iBnd, iCount, iNeigh, iField
    integer :: noNeighFound, neighVal, elemPos
    integer :: iPos
    logical :: wallBnd(scheme%nFields)
    !---------------------------------------------------------------------------
    ! 1. Update the element positions from the treeID to now the 
    ! levelDesc%total list
    ! 1a. Create dynamic array of bc_elemBuffer with position of 
    ! treeID in the levelDescriptor%total list and store the position
    ! in globBC%elems(iLevel)%elembuffer
    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      call init( me     = scheme%glob%levelDesc( iLevel )%bc_elemBuffer,       &
        &        unique = .true. )
      do iBnd = 1, pdfGlobal%nBCs
        ! set to true if BC kind specified for this boundary is wall
        wallBnd = .false.
        do iField = 1 ,scheme%nFields
          if(trim(scheme%field(iField)%bc(iBnd)%BC_kind) .eq. 'wall') &
            & wallBnd(iField) = .true.
        end do
        
        ! if BC kind of all fields are wall then do nothing for this boundary
        ! i.e do not add this boundary to elemBuffer or bc_elemBuffer or 
        ! neigh buffers since wall boundaries are treated explicitly as
        ! bounce back via pdf%neigh array
        if (.not. all(wallBnd)) then
          allocate( scheme%globBC( iBnd )%elems( iLevel )%elemBuffer(          &
            &                          scheme%globBC( iBnd )%nElems( iLevel )))

          !allocate neighBuffer pre-collision and postcollision for each 
          !boundary in each field
          do iField = 1, scheme%nFields
              allocate( scheme%field( iField )%bc( iBnd )%                     &
                & elems( iLevel )%neighBufferPre(                              &
                &         scheme%field( iField )%bc( iBnd )%nNeighs,           &
                &         scheme%globBC( iBnd )%nElems( iLevel )*              &
                &         scheme%layout%stencil(1)%QQ ) )
              allocate( scheme%field( iField )%bc( iBnd )%                     &
                & elems( iLevel )%neighBufferPost(                             &
                &         scheme%field( iField )%bc( iBnd )%nNeighs,           &
                &         scheme%globBC( iBnd )%nElems( iLevel )*              &
                &         scheme%layout%stencil(1)%QQ ) )
              scheme%field(iField)%bc(iBnd)%elems(iLevel)%neighBufferPre =     &
                &                                                        -1.0_rk
              scheme%field(iField)%bc(iBnd)%elems(iLevel)%neighBufferPost=     &
                &                                                        -1.0_rk
          end do

          !create bc_elemBuffer
          do iElem = 1, scheme%globBC( iBnd )%nElems( iLevel )
            elemPos = pdfGlobal%levelPointer( scheme%globBC( iBnd )%           &
              &                                  elems( iLevel )%elem( iElem ))

            scheme%globBC( iBnd )%elems( iLevel )%elem( iElem ) = elemPos

            !store elemPos in bcElemBuffer
            call append( me  = scheme%glob%levelDesc( iLevel )%bc_elemBuffer,  &
              &          val = elemPos,                                        &
              &          pos = iPos )
            !store 
            scheme%globBC( iBnd )%elems( iLevel )%elemBuffer( iElem ) = iPos
          end do
        end if ! not all fields are wall Bnd
      end do ! iBnd
      write(dbgUnit(10),*) 'bc_elemBuffer nVals: ', &
        &                  scheme%glob%levelDesc( iLevel )%bc_elemBuffer%nVals

      !allocate bcBuffer to store state vector values of all boundary elements
      ! for all fields in bcBuffer
      allocate( scheme%pdf( iLevel )%bcBuffer(                                 &
        &              scheme%glob%levelDesc( iLevel )%bc_elemBuffer%nVals     &
        &              *scheme%varSys%nScalars ) )
      scheme%pdf(iLevel)%bcBuffer = -1.0_rk  
    end do !iLevel

    ! 2. build up neigh information for higher order boundaries
    ! loop over the field to update array boundary in each field
    do iField = 1, scheme%nFields
      do iBnd = 1, pdfGlobal%nBCs
        ! 2. Copy the neighboring information to the bc%elem%neigh
        if( scheme%field( iField )%bc( iBnd )%nNeighs .gt. 0 ) then
          !call tem_write( '   updating boundary ', iBnd )
          ! count up the position where to store the neighborhood
          iCount = 0
          do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
            !call tem_write( '     on Level        ', iLevel )
            allocate( scheme%field( iField )%bc( iBnd )%elems( iLevel )%neigh(   &
              &                      scheme%field( iField )%bc( iBnd )%nNeighs,  &
              &                      scheme%globBC( iBnd )%nElems( iLevel ) ) )
            do iElem = 1, scheme%globBC( iBnd )%nElems( iLevel )
              iCount = iCount + 1
              noNeighFound = -1
              do iNeigh = 1, scheme%field( iField )%bc( iBnd )%nNeighs
                if( noNeighFound .lt. 0 ) then
                  ! So far, all neighbors have been found. 
                  neighVal = scheme%glob%levelDesc( iLevel )%neigh( (iField-1)*&
                    &    pdfGlobal%nBCs + iBnd + 1 )%nghElems( iNeigh, iCount ) 
                  ! No neighbor found. set to?
                  if( neighVal .lt. 1 ) then
                    noNeighFound = iNeigh
                  end if
                end if
                if( noNeighFound .lt. 0 ) then
                  ! regular neighbor: assign
                  scheme%field( iField )%bc( iBnd )%elems( iLevel )%           &
                    & neigh( iNeigh, iElem ) = neighVal
                else
                  if( noNeighFound .eq. 1 ) then
                    ! The first neighbor already fails, so save simply current 
                    ! bnd elem pos (itself as neighbor)
                    scheme%field( iField )%bc( iBnd )%elems( iLevel )%         &
                      &                                    neigh( :, iElem ) = &
                      &     scheme%globBC( iBnd )%elems( iLevel )%elem( iElem ) 
                  else
                    ! Neighbor fails, so take last neighbor which was valid
                    scheme%field( iField )%bc( iBnd )%elems( iLevel )%         &
                      &                                    neigh( :, iElem ) = &
                      &    scheme%glob%levelDesc( iLevel )%neigh( (iField-1) * &
                      &      pdfGlobal%nBCs + iBnd + 1 )%nghElems(             &
                      &                                noNeighFound-1, iCount ) 
                  end if
                end if
              end do
            end do  ! iElem
          end do  ! iLevel
        end if !nNeigh
      end do ! iBnd  
    end do ! iField  
    
  end subroutine update_BClists
!******************************************************************************!


!******************************************************************************!
  !> Assemble the level-wise list of elements which adhere to the boundary 
  !! conditions.
  !!
  !! The boundaries will then be treated for each level one by one, running over
  !! the list of elements.
  !! The bitmasks are set for the directions pointing into the domain.
  !! In the LBM, the incoming densities have to be updated. 
  !! As boundaries are being set before the kernel, the state arrays have to be
  !! stored to the FETCH position
  !!
  subroutine build_BClists( scheme, tree, boundary, property, pdfGlobal )
    !---------------------------------------------------------------------------
    !> fluid tree from mesh
    type( treelmesh_type ), intent(in)       :: tree
    !> boundary information from mesh
    type( tem_bc_prop_type ), intent(in)     :: boundary
    !> contains pdf global information
    type( pdf_globInfo_type ), intent(in)    :: pdfGlobal
    !> element property type
    type( tem_property_type ), intent(inout) :: property(:)
    !> scheme type
    type( mus_scheme_type ), intent(inout)   :: scheme
    !---------------------------------------------------------------------------
    ! number of elem for each BC on each lvl
    integer, allocatable :: nBnds(:,:)
    integer(kind=long_k) :: bID
    integer :: iElem, iLevel, level, iMusBnd, iDir, iMeshBnd, iMeshDir
    integer :: iField
    integer :: weight, length
    logical :: found( boundary%nBCtypes ) 
    integer :: map_BCs( boundary%nBCtypes ) 
    integer :: nBCs_local, iElem_qVal
    ! which property to read from in the treelmesh
    integer :: iProperty
    ! For each boundary, determine if there are q-values attached
    logical :: hasQval_loc(boundary%nBCtypes)
    !---------------------------------------------------------------------------
    iProperty = 1

    call tem_write( 'Creating boundary lists ...' )
    allocate( nBnds( globalMaxLevels, boundary%nBCtypes )) 
    allocate( scheme%globBC( boundary%nBCtypes ) )
      
    if( boundary%nBCtypes .gt. 0 ) then

      ! Initialize Qval info
      ! boundary%hasQval is allocated only if prp_hasQval is set in seeder
      ! so initialize them first and use hasQval_loc in rest of the routine
      hasQval_loc = .false.
      if( allocated(boundary%hasQval) ) then 
        do iMeshBnd = 1, pdfGlobal%nBCs
          hasQVal_loc(iMeshBnd) = boundary%hasQval(iMeshBnd)
        enddo
      endif  

      nBnds  = 0
      map_BCs= -1

      ! 1. Identify the number of boundary condition elements of each BC type
      !    and identify the necessary directions to treat.
      do iElem = 1, property( iProperty )%nElems 
        found = .false.
        level = tem_LevelOf( tree%treeID( property( iProperty )%ElemID(iElem)))
        do iDir = 1, scheme%layout%stencil(1)%QQN
          iMeshDir = scheme%layout%stencil(1)%map( iDir )
          ! the index in boundary_ID does not correspond to the stencil!
          bID = boundary%boundary_ID( iMeshDir, iElem )
          if( bID .gt. 0  ) then
            if( bID .gt. pdfGlobal%nBCs ) then
              call tem_write( ' found BID:     ', bID )
              call tem_write( ' where the highest should be ', pdfGlobal%nBCs )
              call tem_write( ' at element     ', iElem )
              call tem_write( ' found higher bID than nr of bcs exist.' )
            end if
            if (found ( bID ) .eqv. .false. ) then
              ! Increase the counter only, if this bID has not been 
              ! identified before at iElem
              nBnds( level, bID ) = nBnds( level, bID )+1
              found( bID ) = .true.
            end if
          end if
        end do
      end do

      ! Counting number of boundaries on local partition
      nBCs_local = 0
      do iMeshBnd = 1, pdfGlobal%nBCs
        scheme%globBC( iMeshBnd )%label = boundary%BC_label( iMeshBnd )
        scheme%globBC( iMeshBnd )%hasQval = hasQval_loc( iMeshBnd )
        ! allocate bc states for each boundary with size of nFields
        ! to collect bc states of all fields of this particular boundary 
        ! into one global list. Needed when boundary condition is
        ! dependent on all fields.
        allocate(scheme%globBC(iMeshBnd)%bc_states(scheme%nFields) )
        do iField = 1, scheme%nFields
          scheme%globBC(iMeshBnd)%bc_states(iField) = &
            & scheme%field(iField)%bc(iMeshBnd)%bc_states
        end do
        if( maxval( nBnds( pdfGlobal%minLevel:pdfGlobal%maxLevel, iMeshBnd ))  &
          &                                                        .gt. 0 )then
          ! Ok, there are elements for this boundary
          nBCs_local = nBCs_local + 1
          ! Map the iMeshBnd-th boundary to the index in the global number of 
          ! BCs
          map_BCs( nBCs_local ) = iMeshBnd
        end if
        ! Sum up the elements of the local partition for each iMeshBnd
        scheme%globBC( iMeshBnd )%nElems_local = sum(                          &
          &           nBnds( pdfGlobal%minLevel:pdfGlobal%maxLevel, iMeshBnd ))
      end do

      ! 2. Allocate the BC lists
      do iMusBnd = 1, pdfGlobal%nBCs
        ! Load the global boundary number from the local one
        ! KM: map_BCS are not used anywhere, Do we need them?
        iMeshBnd = map_BCs( iMusBnd )
        scheme%globBC( iMusBnd )%nElems( : ) = nBnds( :, iMusBnd )
        do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
          ! in gcc-ompi/4.7... this setting caused probems in the build mode
          ! assigning the whole array like done above works
          allocate( scheme%globBC( iMusBnd )%elems( iLevel )%elem(             &
            &                                        nBnds( iLevel, iMusBnd )))
          scheme%globBC( iMusBnd )%elems( iLevel )%elem(:) = 0
          allocate( scheme%globBC( iMusBnd )%elems( iLevel )%bitmask(          &
            &           scheme%layout%stencil(1)%QQ, nBnds( iLevel, iMusBnd )))
          ! allocate qVal for each level for boudnary which has qVals        
          if( hasQVal_loc( iMusBnd )) then
            allocate( scheme%globBC( iMusBnd )%elems( iLevel )%qVal(           &
              &         scheme%layout%stencil(1)%QQ, nBnds( iLevel, iMusBnd )))
          end if

          allocate( scheme%globBC( iMusBnd )%elems( iLevel )%normal(           &
            &                                     3, nBnds( iLevel, iMusBnd )))
          allocate( scheme%globBC( iMusBnd )%elems( iLevel )%normalInd(        &
            &                                        nBnds( iLevel, iMusBnd )))
          if( nBnds( iLevel, iMusBnd ) .gt. 0) then
            call tem_log(                                                      &
              &         level = 10,                                            &
              &         msg   = ' BC '//trim(scheme%globBC( iMusBnd )%label)// &
              &    trim( tem_toStr( scheme%globBC( iMusBnd )%nElems( iLevel ))))
            call tem_log(                                                      &
              &         level = 10,                                            &
              &         msg   = ' BC '//trim(scheme%globBC( iMusBnd )%label)// &
              &                 ' on level: '//trim( tem_toStr( iLevel )))
            ! Reset the bitmask 
            scheme%globBC( iMusBnd )%elems( iLevel )%bitmask =  .false.
            !> Reset the normals
            scheme%globBC( iMusBnd )%elems( iLevel )%normal = 0
          end if
        end do  
      end do

      ! reset the boundary number counter
      nBnds = 0

      ! Initialize element count for boundary with qVal
      iElem_qVal = 0

      !> 3. Assign the BC lists
      ! Run over all the elements with the property boundary
      ! and check each direction.
      ! Assign all common boundaries to the level-wise representation
      do iElem = 1, property( iProperty )%nElems 
        ! Count the elem with qVal property to access the qVal from
        ! boundary%qVal
        if( btest(tree%elemPropertyBits(                                       &
          &             property(iProperty)%ElemID(iElem)), prp_hasQVal) ) then
          iElem_qVal = iElem_qVal + 1          
        endif

        found = .false.
        level = tem_LevelOf( tree%treeID( property( iProperty )%               &
          &                                                   ElemID( iElem )))
        ! run over each direction in the stencil
        do iDir = 1, scheme%layout%stencil(1)%QQN
          ! Stencil order of directions might differ from
          ! the treelm numbering. -> look up correct treelm direction number
          iMeshDir = scheme%layout%stencil(1)%map( iDir )
          ! Get the boundary ID as stored in the mesh
          bID = boundary%boundary_ID( iMeshDir, iElem ) 
          ! Exclude no-boundary or periodic boundaries
          if( bID .gt. 0  ) then
            ! If this boundary ID was not found for current element yet,
            ! assign to list
            if (found ( bID ) .eqv. .false. ) then
              ! count up and add, only if it has not been added before for iElem
              ! this is before all the helper elements have been identified
              ! therefore only the pointers to the original treeID list are
              ! added. This is done later in the update_BC routine.
              ! Wouldn't it be better to have a dynamic array here instead of 
              ! doing it manually
              nBnds( level, bID ) = nBnds( level, bID )+1

              if( .not. allocated( scheme%globBC(bID)%elems(level)%elem))then
!                call tem_debug( level = 1,                                     &
!                  &             msg   = 'Error: bc not allctd lvl '//          &
!                  &                     trim( tem_toStr( level ))//            &
!                  &                     ' bid '//trim( tem_toStr( bID )))
              end if
              ! Assign the position of the treeID to the preliminary boundary
              ! lists
              scheme%globBC( bID )%elems( level )%elem( nBnds( level, bID )) = &
                &                         property( iProperty )%ElemID( iElem ) 
              found( bID ) = .true.
            end if
            ! 
            ! Set the bitmask for the incoming directions which have to be 
            ! updated
            scheme%globBC( bID )%elems( level )%bitmask(                       &
              ! Set bitmask for incoming direction, that is opposite
              ! of the one stored in the treelmesh
              & scheme%layout%stencil(1)%cxDirInv(iDir),                       &
              ! nBnds points to current boudnary element
              & nBnds( level, bID ) ) = .true.

            if( hasQVal_loc( bID )) then
              ! Set the qvalue from the boundary array
              ! q-values stored in outgoing direction
              scheme%globBC( bID )%elems( level )%qVal( iDir, nBnds(level,bID))&
                &                       = boundary%qVal( iMeshDir, iElem_qVal ) 
            end if

            ! Get length of currently treated direction and determine
            ! If it is one of the main coordinate axis
            length = (scheme%layout%stencil(1)%cxDir( 1, iDir )                &
              &      *scheme%layout%stencil(1)%cxDir( 1, iDir )                &
              &     + scheme%layout%stencil(1)%cxDir( 2, iDir )                &
              &      *scheme%layout%stencil(1)%cxDir( 2, iDir )                &
              &     + scheme%layout%stencil(1)%cxDir( 3, iDir )                &
              &      *scheme%layout%stencil(1)%cxDir( 3, iDir )) 
            ! For the normals, we weight the orthogonal directions 
            ! of the stencil more heavily
            if( length == 1) then 
              weight = 2  ! axis orthogonal weighs twice as much as all others
            else
              weight = 1  ! non-axis orthogonal 
            end if
            ! Add this boundary direction to the sum for the normal vector
            ! use the minus, as we want to go into the flow domain, and the 
            ! stencil is pointing outside
            scheme%globBC( bID )%elems( level )%normal(:, nBnds( level, bID )) &
      &  =  scheme%globBC( bID )%elems( level )%normal(:, nBnds( level, bID )) &
      &  -  weight*scheme%layout%stencil(1)%cxDir( :, iDir )
          end if
        end do  ! iDir
      end do ! iElem

      ! 4. Normalize the normal vectors
      !    and choose the corresponding prevailing direction from the stencil
      do iMusBnd = 1, pdfGlobal%nBCs
        do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
          do iElem = 1, scheme%globBC( iMusBnd )%nElems( iLevel )
            call tem_determine_discreteVector(                                 &
              &   scheme%globBC( iMusBnd )%elems( iLevel )%normal( :, iElem ), &
              &   scheme%layout%prevailDir ) 
            ! Find the index in the stencil corresponding to the normal 
            ! direction
            do iDir = 1, scheme%layout%stencil(1)%QQ
              if( scheme%layout%stencil(1)%cxDir(1, iDir ) .eq.                &
                & scheme%globBC( iMusBnd )%elems( iLevel )%normal( 1, iElem )  &
                & .and.                                                        &
                & scheme%layout%stencil(1)%cxDir(2, iDir ) .eq.                &
                & scheme%globBC( iMusBnd )%elems( iLevel )%normal( 2, iElem )  &
                & .and. scheme%layout%stencil(1)%cxDir(3, iDir ) .eq.          &
                & scheme%globBC( iMusBnd )%elems( iLevel )%normal( 3, iElem )) &
                & then
                scheme%globBC( iMusBnd )%elems( iLevel )%normalInd( iElem ) =  &
                  &                                                        iDir
              end if
            end do
          end do  ! iElem
        end do  ! iLevel
      end do ! iMusBnd  
    end if ! nBCs > 0 ?
   
  end subroutine build_BClists
!******************************************************************************!


!******************************************************************************!
  !> Create the communication buffers
  !!
  !! Assign the positions in state vectors, where to fetch from or save to 
  !! Solver specific part to identify the actual data locations
  !! to send use within the communication buffers.
  !! \todo: this should be pattern specific, with typed exchange we
  !!        do not need to store the pos, as it is done inside MPI
  !!        with the defined datatype
  !!
  !! Access to state array: SAVE 
  !! ---------------------------
  !! Why? Because we need to work on the pdfs on which the
  !! compute kernel was working. The kernel stored to SAVE
  !! All actions taking place on the state array after the kernel
  !! need to be performed with SAVE access
  !!
  subroutine init_recvBuffers( me, pdf, nDOFs, pattern, requireAll, levelDesc, &
    &                          scheme, stat )
    !---------------------------------------------------------------------------
    !>
    type(tem_communication_type), intent(inout) :: me
    !>
    type(tem_commPattern_type), intent(in) :: pattern
    !>
    type(pdf_data_type), intent(in) :: pdf
    !> how many links per element? QQ!
    integer, intent(in) :: nDOFs
    !> requires complete element information or only required links? 
    logical, intent(in) :: requireAll
    !> Level descriptor
    type(tem_levelDesc_type), intent(in):: levelDesc
    !> Scheme information on fluid, boundary and flow properties
    type( mus_scheme_type ), intent(in)   :: scheme 
    !> Statistics
    type(mus_stat_type), intent(inout) :: stat
    !---------------------------------------------------------------------------
    ! positions, in state vectors, of links with fluid neighbors
    integer, allocatable :: pos(:)
    integer :: iElem ! count variable for elements
    integer :: iDir, iProc, elemPos
    integer :: counter ! number of total links that acturally needs to update
    integer :: iField, nScalars, varPos(nDOFs)
    ! Include the current direction for the communciation
    logical :: includeInComm
    integer :: nghElem ! neighbor element position in total list
    !---------------------------------------------------------------------------

    nScalars = scheme%varSys%nScalars
    allocate(pos(maxval(me%nElemsProc*nScalars)))

    allocate( me%buf_real( me%nProcs ))

    do iProc = 1, me%nProcs
      ! Allocate the bitmask for the current process( later to be communicated
      ! to the provider proc)
      allocate( me%buf_real( iProc )%bitMask( nScalars, me%nElemsProc( iProc )))

      counter = 0
      do iElem = 1, me%nElemsProc( iProc )
        ! Now assign the exact positions in the sendBuf

        elemPos = me%elemPos( iProc )%val( iElem )
        ! Reset bitmasks
        me%buf_real( iProc )%bitMask(:,iElem) = .false.
        ! Here the check for the bitmasks has to be performed 
!        if( .not. requireAll ) then
        do iField = 1, scheme%nFields
          varPos = scheme%varSys%Variable%val( iField )%varPos
          do iDir = 1, nDOFs
            includeInComm = .false.
            ! Now check all the neighbors of the current halo element
            ! - If we encounter fluid or ghost element, then the direction is
            !   required to exchange for      -> bitmask( iDir, iElem ) = .true.
            ! - If a halo element is ecountered, there is no need for 
            !   communication                -> bitmask( iDir, iElem ) = .false.
            ! Get the element position of the neighbor element
            ! By accessing the neighbor array, get the position of the neighbor
            ! element's direction and reconstruct the neighbor element position
            ! in the total List
            nghElem =                                                          &
& ?ElemIDX?( pdf%neigh( ?NGPOS?( varPos(?NgDir?(iDir,scheme%layout%stencil(1))), elemPos, nScalars )), nScalars)
            ! Now check, what kind of element it is. Must be anything but halo
            if( nghElem .le. levelDesc%offset( 1, eT_haloElem ) .or.           &
              &                                               requireAll ) then
              includeInComm = .true.
            else
            end if
            ! If it's a ghostHalo, then communicate all
            if( elemPos .gt. levelDesc%offset( 1, eT_haloElem ) ) then
              if( pdf%haloRequired( elemPos - levelDesc%                       &
                &                               offset( 1, eT_haloElem ))) then
                includeInComm = .true.
              end if
            end if

            ! If current neighbor is a ghosthalo
            if( nghElem .gt. levelDesc%offset( 1, eT_haloElem ) ) then
              if( pdf%haloRequired( nghElem  - levelDesc%                      &
                &                               offset( 1, eT_haloElem ))) then
                includeInComm = .true.
              end if
            end if

            if( includeInComm ) then
              stat%nLinks_comm = stat%nLinks_comm + 1
              counter = counter + 1
              ! ? SAVE ? ... for regular halos
              ! ? IDX ? ... for ghost-halos
              ! - > These two are the same for PULL!! so, no difference.
              ! Why SAVE? Because we need to work on the pdfs on which the
              ! compute kernel was working. The kernel stored to SAVE
              ! All actions taking place on the state array after the kernel
              ! need to be performed with SAVE access
              pos(counter) = ?SAVE?( varPos(iDir), elemPos, nScalars, pdf%neigh)
              me%buf_real( iProc )%bitMask( varPos(iDir), iElem ) = .true.
            endif
            stat%nLinks_total = stat%nLinks_total + 1
          end do ! iDir
        end do ! field loop
      end do ! iElem

      ! copy position array to me%pos, allocate me%val array
      call pattern%initBuf_real( me%buf_real( iProc ), pos, counter )

    end do ! iProc
    deallocate(pos)

  end subroutine init_recvBuffers
!******************************************************************************!


!******************************************************************************!
  !> Create the communication buffers
  !!
  !! Receive buffers were before created. Now receive the buffers as
  !! send buffers from the remote processes.
  !! Assign the positions, where to fetch from and where to save to 
  !! Solver specific part to identify the actual data locations
  !! to send use within the communication buffers.  
  !! \todo: this should be pattern specific, with typed exchange we
  !!        do not need to store the pos, as it is done inside MPI
  !!        with the defined datatype
  !! 
  !! Access to state array: SAVE 
  !! ---------------------------
  !! Why? Because we need to work on the pdfs on which the
  !! compute kernel was working. The kernel stored to SAVE
  !! All actions taking place on the state array after the kernel
  !! need to be performed with SAVE access
  !! 
  subroutine init_sendBuffers( send, recv, pdf, nDOFs, pattern, requireAll,    &
    &                          scheme )
    !---------------------------------------------------------------------------
    !>
    type(tem_communication_type), intent(inout) :: send, recv
    !>
    type( pdf_data_type), intent(in) :: pdf
    !>
    type(tem_commPattern_type), intent(in) :: pattern
    !> how many links per element? QQ!
    integer, intent(in) :: nDOFs
    !> if true, then take not SAVE but IDX
    logical, intent(in) :: requireAll
    !> fluid, boundary and flow information 
    type( mus_scheme_type ), intent(in)   :: scheme
    !---------------------------------------------------------------------------
    ! Element positions
    integer, allocatable :: pos(:)
    integer :: iElem ! count variable for elements
    integer :: iDir, iProc, counter, elemPos
    !------ MPI variables
    ! send%nProcs + recv%nProcs )
    integer,allocatable :: rq_handle(:)
    ! mpi_status_size, recv%nProcs + send%nProcs ) 
    integer,allocatable :: status(:,:)
    integer :: iErr
    integer :: message_flag
    integer :: iField, nScalars, varPos
    !---------------------------------------------------------------------------
    nScalars = scheme%varSys%nScalars
    ! Choosing random message flag
    message_flag = 1
    allocate( pos(maxval(send%nElemsProc*nScalars)))
    allocate( rq_handle( send%nProcs+ recv%nProcs)) 
    allocate( status( mpi_status_size, send%nProcs+ recv%nProcs))
    allocate( send%buf_real( send%nProcs ))

    ! First start receive communication of the 'send' bitmasks from the target 
    ! Procs. Consumer view: receive from the target Process
    do iProc = 1, send%nProcs
      allocate( send%buf_real(iProc)%bitMask( nScalars, send%nElemsProc(iProc)))
      ! dbg: reset bitmask
      send%buf_real( iProc )%bitMask = .false.
      call mpi_irecv(                              &
        &      send%buf_real( iProc )%bitMask,     & ! me
        &      send%nElemsProc( iProc )*nScalars,  & ! me size
        &      MPI_LOGICAL,                        & ! data type
        &      send%proc( iProc ),                 & ! target me
        &      message_flag,                       & ! flag
        &      process%comm,                       & ! communicator
        &      rq_handle(iProc),                   & ! handle
        &      iErr  )                               ! error status

    end do

    ! then send the 'recv' bitmask arrays to the neighbor procs
    ! Provider view: send to the source processes
    do iProc = 1, recv%nProcs
      call mpi_isend(                              &
       &      recv%buf_real( iProc )%bitMask,      & ! me
       &      recv%nElemsProc( iProc )*nScalars,   & ! me size
       &      MPI_LOGICAL,                         & ! data type
       &      recv%proc( iProc ),                  & ! target me
       &      message_flag,                        & ! flag
       &      process%comm,                        & ! communicator
       &      rq_handle( iProc + send%nProcs ),    & ! handle
       &      iErr                    )              ! error status

    end do ! iProc

    ! Wait for above communications to complete
    call mpi_waitall(               &
      & send%nProcs+recv%nProcs,    & ! total number of comm.'s to wait for
      & rq_handle,                  & ! request handles
      & status,                     & ! mpi status
      & iErr         )                ! error status
    !--------------------------------------------------------------------------
    ! Status: We received the bitmasks from the Provider processes,
    !         from which we can now construct the correct send buffers
    !--------------------------------------------------------------------------

    ! Then assign the send buffer positions
    do iProc = 1, send%nProcs
      counter = 0
      do iElem = 1, send%nElemsProc( iProc )
        !> Now assign the exact positions in the sendBuf
        elemPos = send%elemPos( iProc )%val( iElem )
        do iField = 1, scheme%nFields
          do iDir = 1, nDOFs
            varPos = scheme%varSys%Variable%val( iField )%varPos(iDir)
            if( send%buf_real( iProc )%bitMask( varPos, iElem )) then
              counter = counter + 1
              ! ? SAVE ? ... for regular halos
              ! ? IDX ? ... for ghost-halos
              ! -> These two are the same for PULL!! so, no difference.
              ! Why SAVE? Because we need to work on the pdfs on which the
              ! compute kernel was working. The kernel stored to SAVE
              ! All actions taking place on the state array after the kernel
              ! need to be performed with SAVE access
              pos( counter ) = ?SAVE?( varPos, elemPos, nScalars, pdf%neigh ) 
            endif
          end do
        end do!field loop
      end do !< iElem

      call pattern%initBuf_real( send%buf_real( iProc ), pos, counter )
      ! Bitmasks not necessary any more. deallocate.
      deallocate( send%buf_real( iProc )%bitMask )
    end do !< iProc

    ! Deallocate recv bitmasks as well
    do iProc = 1, recv%nProcs
      deallocate( recv%buf_real( iProc )%bitMask )
    end do

    deallocate(pos)

  end subroutine init_sendBuffers
!******************************************************************************!


!******************************************************************************!
  !> subroutine to find neighbours of element with individual (for each element)
  !! stencil definitions 
  subroutine mus_findIndividualNeighbor( globBC, bc, tree, layout, stencil )
    !---------------------------------------------------------------------------
    !> boundaries for the elements with bnd property set 
    type(glob_boundary_type), intent(in) :: globBC
    !> field boundary with boundary neighbor info
    type(boundary_type), intent(inout) :: bc
    !> subset of tree ids for which the neighbours will be specified
    type(treelmesh_type), intent(in)   :: tree
    !> scheme layout
    type( mus_scheme_layout_type ), intent(in) :: layout
    !> Stencil definition
    type(tem_stencilHeader_type), intent(inout) :: stencil
    !---------------------------------------------------------------------------
    integer              :: iElem, iPos, iLevel, iCount
    integer              :: normal(3)
    integer              :: iNeigh
    !---------------------------------------------------------------------------
    stencil%useAll = .false.
    stencil%requireNeighNeigh = .true.
    stencil%QQ     = bc%nNeighs 
    stencil%QQN    = bc%nNeighs 
    stencil%nElems = globBC%nElems_local
    allocate( stencil%elem( stencil%nElems ))
    allocate( stencil%elemLvl( tree%global%minLevel: tree%global%maxLevel))
    allocate( stencil%cxDir( 3, stencil%QQN ))
    ! Reset the stencil, so that if nothing is defined, no additional 
    ! elements are recovered
    stencil%cxDir(:,:) = 0
    iCount = 0
    do iLevel = tree%global%minLevel, tree%global%maxLevel
      call init( me = stencil%elemLvl( iLevel ), length = 8 )
      do iElem = 1, globBC%nElems( iLevel )   
        iCount = iCount + 1
        ! Get the position of the treeID in the primal mesh
        iPos = globBC%elems( iLevel )%elem( iElem )
        stencil%elem( iCount ) = iPos
        call append( me = stencil%elemLvl( iLevel ), val = iPos )
        ! Get the normal direction of this boundary
        ! only for higher order boundaries
        if( bc%nNeighs .gt. 0) then
          normal(:) = globBC%elems( iLevel )%normal( :, iElem )
          ! And store the normal direction into the stencil 
          ! QQN stands for the number of neighbors
          ! Here it means, how many neighbors in the normal direction
          do iNeigh = 1, stencil%QQN
            ! Check if there is a boundary here! ??MH: do we have to do that
            ! here?
            ! I think it should be added, that we check for a boundary in the
            ! current search direction.
            !
            ! ?: what is the purpose of multiplying iStencilElem with normal
            !    direction?
            ! !: We are going along the normal direction for more than 
            ! one neighbor in one direction (required for higher order boundary)
            stencil%cxDir( :, iNeigh ) = stencil%cxDir( :, iNeigh )            &
              &                        + normal(:)*iNeigh
          enddo ! iStencilElem
        endif
      enddo ! iElem
    enddo ! iLevel 
    
    ! KM: QUICK HACK: Average the normal of each boundary element to obtain 
    ! stencil for boundary
    if( bc%nNeighs .gt. 0) then
      do iNeigh = 1, stencil%QQN
        call tem_determine_discreteVector( stencil%cxDir( : , iNeigh ),        &
          &                                                 layout%prevailDir ) 
      end do
    endif

  end subroutine mus_findIndividualNeighbor
!******************************************************************************!

!******************************************************************************!
  !>  map the compact stencil to the compute stencil
  !!
  !! The result of this routine is the compact interpolation index for each of
  !! the eight possible occurring cases for fromCoarser interpolations.
  !! There are eight child positions and hence eight different compact stencil
  !! configurations.
  !! For each configuration, the neighbor coordinates need to be known.
  !! MH: Note that this is currently only being used in
  !! intp_quadratic_compactReduced
  !!
  subroutine map_compactToComputeStencil( intp, stencil )
    !---------------------------------------------------------------------------
    !> boundaries for the elements with bnd property set 
    type(interpolation_type), intent(inout) :: intp
    !> Stencil definition
    type(tem_stencilHeader_type), intent(inout) :: stencil
    !---------------------------------------------------------------------------
    integer :: iCase, childNum, iNeigh, iRow
    integer :: neighOffset(3), x(4), parentPos( 3,8)
    integer :: foundDirInStencil
    integer :: iStencilElem
    !---------------------------------------------------------------------------
    ! size: number of compact stencil entries, 8 (=number of child cases )
    ! only to be done for noncompact interpolation schemes
    allocate( intp%compactInd( intp%nSourcesCompactFromCoarser_required, 8 ))
    intp%compactInd = -1
    ! map compact to compute stencil 
    do iCase = 1, 8
      childNum = iCase
      call tem_log( level = 1,                                                 &
        &           msg   = 'icase'//trim( tem_toStr( icase )))
      ! For each case, identify all the compact neighbors
      do iNeigh = 1, 8
        x = tem_coordOfId( int(iNeigh, kind=long_k))*2 -1
        parentPos(:, iNeigh)  = matmul(tem_rotationMatrix(:,:, ChildNum),x(1:3))
        neighOffset  = (parentPos(:,iNeigh)  - parentPos(:, 1))/2

        call tem_log( level = 1,                                               &
          &           msg   = 'coordinate of neigh '//                         &
          &                   trim(tem_toStr( val = x, sep = ', ' )))
        call tem_log( level = 1,                                               &
          &           msg   = 'parentPos          '//                          &
          &                   trim(tem_toStr( val = parentPos(:, iNeigh),      &
          &                                   sep = ', ' )))
        call tem_log( level = 1,                                               &
          &           msg   = trim( tem_toStr( iNeigh ))// ' neighOff '//      &
          &                   trim(tem_toStr( val = neighOffset, sep = ', ' )))

        ! Match the found neighOffset with the stencil. Is this direction
        ! available?
        foundDirInStencil = -1
        do iStencilElem = 1, stencil%QQ

          call tem_log( level = 1,                                             &
            &           msg   = trim( tem_toStr( iStencilElem ))// ' vs '//    &
            &                   trim(tem_toStr(                                &
            &                          val = stencil%cxDir( :, iStencilElem ), &
            &                          sep = ', ' )))

          if( neighOffset(1) .eq. stencil%cxDir( 1, iStencilElem ) .and.       &
            & neighOffset(2) .eq. stencil%cxDir( 2, iStencilElem ) .and.       &
            & neighOffset(3) .eq. stencil%cxDir( 3, iStencilElem )) then
            foundDirInStencil = iStencilElem
            call tem_log( level = 1,                                           &
              &           msg   = 'found '//trim(tem_toStr(foundDirInStencil )))
          end if
        end do
        call tem_log( level = 1,                                               &
          &           msg   = 'iNeigh '//trim( tem_toStr( iNeigh ))//          &
          &                   'found dir '//trim(tem_toStr( foundDirInStencil)))
        call tem_log( level = 1,                                               &
          &           msg   = 'iNeigh '//trim( tem_toStr( iNeigh))//           &
          &                   'req '//trim( tem_toStr( intp%                   &
          &                            nSourcesCompactFromCoarser_required )))
        
        do iRow = 1, size( intp%compactInd, 2)
          call tem_log( level = 1,                                             &
            &           msg = trim( tem_toStr( val = intp%compactInd(:,iRow ), &
            &                                  sep = '' )))
        end do

        if( iNeigh .le. intp%nSourcesCompactFromCoarser_required )             &
          &                intp%compactInd( iNeigh, iCase ) = foundDirInStencil
      end do
      if( minval( intp%compactInd( 1:intp%nSourcesCompactFromCoarser_required, &
        &                                                iCase )) .lt. 1 ) then
        call tem_log( level = 1,                                               &
          &           msg   = 'found '//trim( tem_toStr( minval(               &
          &intp%compactInd(1:intp%nSourcesCompactFromCoarser_required,iCase)))))
        call tem_log( level = 1,                                               &
          &           msg   = 'at '//trim( tem_toStr(                          &
          & val = minloc( intp%compactInd(1:intp%                              &
          &               nSourcesCompactFromCoarser_required,iCase)),         &
          & sep = '' )))
        call tem_write( 'Error: did not find all the required compact '//      &
          &             'neighbors in the given stencil' )
        call tem_abort
      end if
    end do
 
  end subroutine map_compactToComputeStencil
!****************************************************************************!
  
  
!****************************************************************************!
  !> This subroutine acts as a destructor for the construct routine
  subroutine mus_construct_cleanup( scheme, intp )
    !------------------------------------------------------------------------ 
    !> scheme information including fluid, boundary and flow information 
    type( mus_scheme_type ), allocatable, intent(inout)   :: scheme(:)
    type( interpolation_type ), intent(inout) :: intp
    !------------------------------------------------------------------------ 
    if( allocated(scheme) )           deallocate ( scheme )
    if( allocated(intp%fromFiner) )   deallocate ( intp%fromFiner )
    if( allocated(intp%fromCoarser) ) deallocate ( intp%fromCoarser)

  end subroutine mus_construct_cleanup
  !****************************************************************************!

end module mus_construction_module
!******************************************************************************!
