! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> In this module, the control structure for computing each time step is
!! established.
!!
!! Various different control structures can be defined and set by function 
!! pointers.
!! The current main routine is recursive_multilevel which recursively computes
!! the new time step for all the schemes defined on all levels, starting from 
!! the coarsest.
!! 
module mus_control_module

  ! include treelm modules
  use env_module,         only: pathLen
  use tem_timer_module,   only: tem_startTimer, tem_stopTimer
  use tem_debug_module,   only: tem_debug, main_debug
  use tem_logging_module, only: tem_toStr

  ! include musubi modules
  use mus_pdf_module,                only: pdf_globInfo_type
  use mus_global_module,             only: param_type
  use mus_geom_module,               only: mus_geom_type 
  use mus_comm_module,               only: mus_exchange
  use mus_bc_general_module,         only: set_boundary
  use mus_aux_module,                only: check_flow_status,                  &
    &                                      mus_setParameters
  use mus_interpolate_header_module, only: interpolation_type
  use mus_scheme_type_module,        only: mus_scheme_type
  use mus_source_module,             only: mus_update_source
  use mus_debug_module,              only: dump_debug_Info
!  use mus_geomIncr_module,           only: fillMinePropFromFiner,              &
!    &                                      fillFinerPropFromME

  implicit none

  private

  public :: mus_control_type
  public :: mus_init_control

  !> Datatype containing control information loaded from lua file
  !! It includes information about which control routine to use
  !! e.g. highly efficient one without function pointers
  type mus_controlHeader_type
    character(len=PathLen) :: label
  end type mus_controlHeader_type

  !> Datatype containing mapping of control routines to function pointers
  type mus_control_type
    procedure( controlRoutine ), pointer :: do_computation => null()
  end type mus_control_type
 
  abstract interface
    !> Interface describes the main control routine which does computation
    !! set boundary and check flow status
    subroutine controlRoutine( me, scheme, geometry, params, intp, iLevel,     &
      &                        pdfGlobal )
      import :: mus_scheme_type, mus_geom_type, param_type, &
        & interpolation_type, pdf_globInfo_type, mus_control_type
      !> self control type
      class( mus_control_type ) :: me
      !> containers for the different schemes
      type( mus_scheme_type ), intent(inout) :: scheme(:)
      !> geometry infomation
      type( mus_geom_type ), intent(in) :: geometry
      !> global parameters
      type( param_type ), intent(inout) :: params   
      !> Interpolation control
      type( interpolation_type ), intent(in) :: intp    
      integer, intent(in) :: iLevel     !< Level counter variable
      !> global flow quantity properties
      type( pdf_globInfo_type ), intent(inout) :: pdfGlobal
    end subroutine controlRoutine
  end interface
 
  contains

!******************************************************************************!
  !> This routines sets the function pointer to main control routine 
  !!
  !! control routine is chosen based on the type of simulation.
  !! like single level, multi-level because multilevel requires 
  !! recursive routine
  !!
  !! in the lua file you can now select 
  !! \verbatim control_routine = '...'\endverbatim
  !! where possible values are currently
  !! - `benchmark`: strongly reduced control routine with only single level,
  !!                no sources, etc. 
  !! mainly for benchmarking
  !! - `multischeme`: full multilevel, multischeme routine
  !! - if nothing is given, the full multilevel, multischeme routine is chosen
  subroutine mus_init_control( params, me )
    !---------------------------------------------------------------------------
    !> global parameters
    type( param_type ), intent(in) :: params   
    !> contains function pointer to point control routine
    type( mus_control_type ), intent(out) :: me
    !---------------------------------------------------------------------------

    select case( trim( params%controlRoutine ))
    case('debug')
      me%do_computation => do_recursive_debug
    case('benchmark','fast')
      me%do_computation => do_fast_singleScheme
    case('multischeme', 'multilevel')
      me%do_computation => do_recursive_multiScheme
    case default !       'standard'
      me%do_computation => do_recursive_multiScheme
    end select

  end subroutine mus_init_control
!******************************************************************************!

 
!******************************************************************************!
  !> Main control routine: Update the time step for all levels
  !!
  recursive subroutine do_recursive_multischeme( me, scheme, geometry, params, &
    &                                            intp, iLevel, pdfGlobal )
    !--------------------------------------------------------------------------
    !> self control type
    class( mus_control_type ) :: me
    !> containers for the different schemes
    type( mus_scheme_type ), intent(inout)  :: scheme(:)   
    !> geometry infomation
    type( mus_geom_type ), intent(in)        :: geometry         
    !> global parameters
    type( param_type ), intent(inout)       :: params       
    !> Interpolation control
    type( interpolation_type ), intent(in) :: intp          
    !> global flow quantity properties
    type( pdf_globInfo_type ), intent(inout)  :: pdfGlobal   
    integer, intent(in) :: iLevel     !< Level counter variable
    !--------------------------------------------------------------------------
    integer :: iScheme    !< Scheme counter variable
    integer :: depScheme    !< depending scheme
    !--------------------------------------------------------------------------

    ! For each level, run over all the schemes defined
    do iScheme = 1, pdfGlobal%nSchemes
!$omp single
      ! swap double buffer index for current level
      scheme( iScheme )%pdf( iLevel )%nNow                                     &
        &                   = mod( scheme( iScheme )%pdf( iLevel )%nNow ,2 ) + 1
      scheme( iScheme )%pdf( iLevel )%nNext                                    &
        &                   = mod( scheme( iScheme )%pdf( iLevel )%nNext,2 ) + 1
      !set boundary for each field in current scheme
      call tem_startTimer( params%setBndTimer )
      call tem_startTimer( me = params%timer(ilevel),                          &
        &                  timerHandle = params%lw_tSetBnd(iLevel) )
      call set_boundary( field     = scheme( iScheme )%field,                  &
        &                pdf       = scheme( iScheme )%pdf,                    &
        &                glob      = scheme( iScheme )%glob,                   &
        &                tree      = geometry%tree,                            &
        &                iLevel    = iLevel,                                   &
        &                pdfGlobal = pdfGlobal,                                &
        &                params    = params,                                   &
        &                layout    = scheme( iScheme )%layout,                 &
        &                physics   = scheme( iScheme )%physics,                &
        &                varSys    = scheme( iScheme )%varSys,                 &
        &                derVarPos = scheme( iScheme)%derVarPos,               &
        &                globBC    = scheme( iScheme )%globBC )
      call tem_stopTimer( params%setBndTimer )
      call tem_stopTimer( me = params%timer(iLevel),                           &
                          timerHandle = params%lw_tSetBnd(iLevel) )
      
      ! Update parameters 
      call mus_setParameters( scheme( iScheme ), pdfGlobal, iLevel, params )

      ! Compute current scheme of current level
      call tem_startTimer( params%advRelTimer )
      call tem_startTimer( me = params%timer(ilevel),                          &
        &                  timerHandle = params%lw_tCompute(iLevel) )
!$omp end single

      call scheme( iScheme )%compute(                                          &
        &  iScheme   = iScheme,                                                &
        &  fieldProp = scheme( iScheme )%field(:)%fieldProp,                   &
        &  inState   = scheme( iScheme )%pdf(iLevel)%state(:,                  &
        &                         scheme( iScheme )%pdf(iLevel)%nNow),         &
        &  outState  = scheme( iScheme )%pdf(iLevel)%state(:,                  &
        &                         scheme( iScheme )%pdf(iLevel)%nNext),        &
        &  neigh     = scheme( iScheme )%pdf(iLevel)%neigh(:),                 &
        &  nStart    = 1,                                                      &
        &  nEnd      = scheme( iScheme )%pdf(iLevel)%nElems_solve,             &
        &  level     = iLevel,                                                 &
        &  layout    = scheme( iScheme )%layout,                               &
        &  pdfGlobal = pdfGlobal,                                              &
        &  params    = params,                                                 &
        &  scheme    = scheme,                                                 &
        &  tree      = geometry%tree,                                          &
        &  varSys    = scheme( iScheme )%varSys )
      call tem_stopTimer( params%advRelTimer )
      call tem_stopTimer( me = params%timer(iLevel),                           &
                          timerHandle = params%lw_tCompute(iLevel) )
    end do ! scheme loop


    ! Check, if we are on the finest level
    if( iLevel .lt. pdfGlobal%maxLevel ) then
      ! perform interpolation and recursive calls
      call do_computeFinerIntpandExchange( me, scheme, geometry, params,       &
        &                                  intp, iLevel, pdfGlobal )
    end if

!$omp single
    ! Communicate the halo elements of each scheme on current level 
    do iScheme = 1, pdfGlobal%nSchemes
      call tem_startTimer( params%commTimer )
      call mus_exchange(                                                       &
        &  send    = scheme( iScheme )%glob%levelDesc(iLevel)%sendbuffer,      &
        &  recv    = scheme( iScheme )%glob%levelDesc(iLevel)%recvbuffer,      &
        &  state   = scheme( iScheme )%pdf(iLevel)%state(:,                    &
        &                                scheme( iScheme )%pdf(iLevel)%nNext), &
        &  pattern = params%commPattern,                                       &
        &  level   = iLevel)
      if( iLevel .gt. pdfGlobal%minLevel ) then
        call mus_exchange(                                                     &
          & send    = scheme( iScheme )%glob%levelDesc(iLevel)%                &
          &                                             sendbufferFromCoarser, &
          & recv    = scheme( iScheme )%glob%levelDesc(iLevel)%                &
          &                                             recvbufferFromCoarser, &
          & state   = scheme( iScheme )%pdf(iLevel)%state(:,scheme( iScheme )% &
          &                                                pdf(iLevel)%nNext), &
          & pattern = params%commPattern,                                      &
          & level   = iLevel)
      end if
      call tem_stopTimer( params%commTimer )
    end do ! scheme loop
 
    if( iLevel .lt. pdfGlobal%maxLevel ) then
      ! Now when all coarser ghosts are present, we can interpolate the other
      ! way round.
      ! Interpolate the ghost elements on the finer level with data provided 
      ! from current level (This only needs to be done once in two finer time
      ! steps)
      call do_intpCoarserandExchange( scheme, params, intp, iLevel, pdfGlobal, &
        &                             geometry )   
    end if ! if not on finest level

    do iScheme = 1, pdfGlobal%nSchemes
      call tem_startTimer(params%sourceTimer)
      if( allocated(scheme( iScheme )%depend) )then
        depScheme = scheme( iScheme )%depend( 1 )%useScheme
        ! update the source terms for each field and the global one for the
        ! scheme
        call mus_update_source(                                                &
          &                  me            = scheme(iScheme)%source,           &
          &                  pdf           = scheme(iScheme)%pdf,              &
          &                  depPdf        = scheme(depScheme)%pdf,            &
          &                  time          = pdfGlobal%timing,                 &
          &                  global_varSys = scheme(iScheme)%varSys,           &
          &                  global_tree   = geometry%tree,                    &
          &                  stencil       = scheme(iScheme)%layout%stencil(1))
      else
        ! update the source terms for each field and the global one for the
        ! scheme
        call mus_update_source(                                                &
          &                  me            = scheme(iScheme)%source,           &
          &                  pdf           = scheme(iScheme)%pdf,              &
          &                  time          = pdfGlobal%timing,                 &
          &                  global_varSys = scheme(iScheme)%varSys,           &
          &                  global_tree   = geometry%tree,                    &
          &                  stencil       = scheme(iScheme)%layout%stencil(1))
      end if
      call tem_stopTimer(params%sourceTimer)
    end do

    if( iLevel .eq. pdfGlobal%maxLevel ) then
      ! check if some action has to be taken:
      ! output, tracking, global reduction operations, increase timers
      call check_flow_status( scheme    = scheme,                              &
        &                     params    = params,                              &
        &                     geometry  = geometry,                            &
        &                     pdfGlobal = pdfGlobal )
    end if 
!$omp end single

  end subroutine do_recursive_multischeme
!******************************************************************************!


!******************************************************************************!
!> This routine does computation twice at fine level, interpolate and 
!! exchange the datas between process at fine level and coaser level
  recursive subroutine do_computeFinerIntpandExchange( me, scheme,             &
    &                               geometry, params, intp, iLevel, pdfGlobal )
    !---------------------------------------------------------------------------
    class( mus_control_type ) :: me !< contains control routine function pointer
    !> containers for the different schemes
    type( mus_scheme_type ), intent(inout)  :: scheme(:)   
    !> geometry infomation
    type( mus_geom_type ), intent(in)        :: geometry         
    !> global parameters
    type( param_type ), intent(inout)       :: params       
    !> Interpolation control
    type( interpolation_type ), intent(in) :: intp          
    !> global flow quantity properties
    type( pdf_globInfo_type ), intent(inout)  :: pdfGlobal   
    integer, intent(in) :: iLevel     !< Level counter variable
    !---------------------------------------------------------------------------
    integer :: iScheme
    !> diffusive or acoustic nesting loop
    integer :: iNestingLoop
    character(len=33) :: buffer
    !---------------------------------------------------------------------------
 
    ! Perform the number of nested time steps on the finer level according to 
    ! the scaling ( diffusive = 4, acoustic = 2 )
    do iNestingLoop = 1, pdfGlobal%nNesting
      ! Compute and communicate
      call me%do_computation( scheme, geometry, params, intp, iLevel+1,        &
        &                     pdfGlobal )
    end do
    ! Compute and communicate
    call me%do_computation( scheme, geometry, params, intp, iLevel+1,          &
      &                     pdfGlobal )

!$omp single
    ! Interpolate the coarse ghost elements from the finer ones
    do iScheme = 1, pdfGlobal%nSchemes
      call tem_startTimer( params%intpTimer_low )
      call tem_startTimer( me = params%timer(ilevel),                          &
        &                  timerHandle = params%lw_tIntp(iLevel) )

      call intp%fillMineFromFiner_low(                                         &
        &                  pdf       = scheme( iScheme )%pdf,                  &
        &                  glob      = scheme( iScheme )%glob,                 &
        &                  scheme    = scheme( iScheme ),                      &   
        &                  fieldProp = scheme( iScheme )%field(:)%fieldProp,   &
        &                  level     = iLevel,                                 &
        &                  intp      = intp,                                   &
        &                  ind       = scheme( iScheme )%glob%                 &
        &                                 levelDesc(iLevel)%intpFromFiner_low, &
        &                  layout    = scheme( iScheme )%layout,               &
        &                  varSys    = scheme( iScheme )%varSys,               &
        &                  derVarPos = scheme( iScheme )%derVarPos  )  
      call tem_stopTimer( params%intpTimer_low )

      call tem_startTimer( params%intpTimer_hgh )
      call intp%fillMineFromFiner(                                             &
        &                  pdf       = scheme( iScheme )%pdf,                  &
        &                  glob      = scheme( iScheme )%glob,                 &
        &                  scheme    = scheme( iScheme ),                      &   
        &                  fieldProp = scheme( iScheme )%field(:)%fieldProp,   &
        &                  level     = iLevel,                                 &
        &                  intp      = intp,                                   &
        &                  ind       = scheme( iScheme )%glob%                 &
        &                                levelDesc(iLevel)%intpFromFiner_high, &
        &                  layout    = scheme( iScheme )%layout,               &
        &                  varSys    = scheme( iScheme )%varSys,               &
        &                  derVarPos    = scheme( iScheme )%derVarPos)
      call tem_stopTimer( params%intpTimer_hgh )
      call tem_stopTimer( me = params%timer(iLevel),                           &
                          timerHandle = params%lw_tIntp(iLevel) )

      if ( main_debug%checkEachAlgorithmicStep ) then
        buffer=' after fill Mine FromFiner       '
        call dump_debug_info( buffer, params, geometry, scheme, pdfGlobal,     &
          &                   iLevel, iScheme )
      end if

      ! JQ: fill my coarse elements prop from finer
!      call fillMinePropFromFiner( glob = scheme( iScheme )%glob, &
!        &                         level = iLevel )

      ! Exchange the coarse ghost elements
      call tem_startTimer( params%commTimer )

      ! Exchange the coarse ghost elements
      call mus_exchange(                                                       &
        &             send    = scheme( iScheme )%glob%levelDesc(iLevel)%      &
        &                                                 sendbufferFromFiner, &
        &             recv    = scheme( iScheme )%glob%levelDesc(iLevel)%      &
        &                                                 recvbufferFromFiner, &
        &             state   = scheme( iScheme )%pdf(iLevel)%state(:,scheme(  &
        &                                        iScheme )%pdf(iLevel)%nNext), &
        &             pattern = params%commPattern,                            &
        &             level   = iLevel)
      if ( main_debug%checkEachAlgorithmicStep ) then
        buffer=' after exch Mine FromFiner       '
        call dump_debug_info( buffer, params, geometry, scheme, pdfGlobal,     &
          &                   iLevel, iScheme )
      end if
      call tem_stopTimer( params%commTimer )
    end do
!$omp end single
   
  end subroutine do_computeFinerIntpandExchange
!******************************************************************************!


!******************************************************************************!
!> This routine does interpolate pdf from fine to coarser and 
!! exchange the datas between process at coarser level
  subroutine do_intpCoarserandExchange( scheme, params, intp, iLevel,          &
    &                                   pdfGlobal, geometry )
    !---------------------------------------------------------------------------
    !> containers for the different schemes
    type( mus_scheme_type ), intent(inout)  :: scheme(:)   
    !> global parameters
    type( param_type ), intent(inout)       :: params       
    !> Interpolation control
    type( interpolation_type ), intent(in) :: intp          
    !> global flow quantity properties
    type( pdf_globInfo_type ), intent(in)  :: pdfGlobal   
    !> geometry infomation
    type( mus_geom_type ), intent(in)        :: geometry         
    !> Level counter variable
    integer, intent(in) :: iLevel
    !---------------------------------------------------------------------------
    integer :: iScheme
    character(len=33) :: buffer
    !---------------------------------------------------------------------------

    do iScheme = 1, pdfGlobal%nSchemes
      call tem_startTimer( params%intpTimer_low )
      call intp%fillFinerFromMe_low(                                           &
        &               pdf       = scheme( iScheme )%pdf,                     &
        &               glob      = scheme( iScheme )%glob,                    & 
        &               scheme    = scheme( iScheme ),                         &   
        &               fieldProp = scheme( iScheme )%field(:)%fieldProp,      &
        &               level     = iLevel,                                    &
        &               intp      = intp,                                      &
        &               ind       = scheme( iScheme )%glob%                    &
        &                            levelDesc( iLevel+1)%intpFromCoarser_low, &
        &               layout    = scheme( iScheme )%layout,                  &
        &               varSys    = scheme( iScheme )%varSys,                  &  
        &               derVarPos = scheme( iScheme )%derVarPos )
      call tem_stopTimer( params%intpTimer_low )

      call tem_startTimer( params%intpTimer_hgh )
      call intp%fillFinerFromMe(                                               &
        &               pdf       = scheme( iScheme )%pdf,                     &
        &               glob      = scheme( iScheme )%glob,                    & 
        &               scheme    = scheme( iScheme ),                         &   
        &               fieldProp = scheme( iScheme )%field(:)%fieldProp,      &
        &               level     = iLevel,                                    &
        &               intp      = intp,                                      &
        &               ind       = scheme( iScheme )%glob%                    &
        &                           levelDesc( iLevel+1)%intpFromCoarser_high, &
        &               layout    = scheme( iScheme )%layout,                  &
        &               varSys    = scheme( iScheme )%varSys,                  &  
        &               derVarPos = scheme( iScheme )%derVarPos ) 
      call tem_stopTimer( params%intpTimer_hgh )
      if ( main_debug%checkEachAlgorithmicStep ) then
        buffer=' after fillFinerFromMe           '
        call dump_debug_info( buffer, params, geometry, scheme, pdfGlobal,     &
          &                   iLevel+1, iScheme )
      end if

      ! JQ: fill finer property from me
!      call fillFinerPropFromMe( glob = scheme( iScheme )%glob, &
!        &                       level = iLevel )
        
      call tem_startTimer( params%commTimer )
      ! Exchange the fine ghost elements
      call mus_exchange(                                                       &
        &            send    = scheme( iScheme )%glob                          &
        &                       %levelDesc(iLevel+1)%sendbufferFromCoarser,    &
        &            recv    = scheme( iScheme )%glob                          &
        &                       %levelDesc(iLevel+1)%recvbufferFromCoarser,    &
        &            state   = scheme( iScheme )%pdf(iLevel+1)%                &
        &                      state(:,scheme( iScheme )%pdf(iLevel+1)%nNext), &
        &            pattern = params%commPattern,                             &
        &            level   = iLevel)
      call tem_stopTimer( params%commTimer )
      if ( main_debug%checkEachAlgorithmicStep ) then
        buffer=' after exchFinerFromMe           '
        call dump_debug_info( buffer, params, geometry, scheme, pdfGlobal,     &
          &                   iLevel+1, iScheme )
      end if
    end do
  end subroutine do_intpCoarserandExchange
!******************************************************************************!


!******************************************************************************!
  !> Control routine for an optimized workflow with reduced functionality.
  !!
  !! No sources, no multilevel, no multischeme. 
  !! Use for benchmarking
  !!
  subroutine do_fast_singleScheme( me, scheme, geometry, params, intp, iLevel, &
    &                              pdfGlobal )
    !---------------------------------------------------------------------------
    !> self control type
    class( mus_control_type ) :: me
    !> containers for the different schemes
    type( mus_scheme_type ), intent(inout)  :: scheme(:)   
    !> geometry infomation
    type( mus_geom_type ), intent(in)        :: geometry         
    !> global parameters
    type( param_type ), intent(inout)       :: params       
    !> Interpolation control
    type( interpolation_type ), intent(in) :: intp          
    !> global flow quantity properties
    type( pdf_globInfo_type ), intent(inout)  :: pdfGlobal
    !> Level counter variable
    integer, intent(in) :: iLevel
    !---------------------------------------------------------------------------
    integer, parameter  :: iScheme = 1   ! Scheme counter variable
    !---------------------------------------------------------------------------
!$omp single
    ! swap double buffer index for current level
    scheme( iScheme )%pdf( iLevel )%nNow                                       &
      &                   = mod( scheme( iScheme )%pdf( iLevel )%nNow ,2 ) + 1
    scheme( iScheme )%pdf( iLevel )%nNext                                      &
      &                   = mod( scheme( iScheme )%pdf( iLevel )%nNext,2 ) + 1
    !set boundary for each field in current scheme
    call tem_startTimer( params%setBndTimer )
    call tem_startTimer( me = params%timer(ilevel),                          &
      &                  timerHandle = params%lw_tSetBnd(iLevel) )
    call set_boundary( field     = scheme( iScheme )%field,                    &
      &                pdf       = scheme( iScheme )%pdf,                      &
      &                glob      = scheme( iScheme )%glob,                     &
      &                tree      = geometry%tree,                              &
      &                iLevel    = iLevel,                                     &
      &                pdfGlobal = pdfGlobal,                                  &
      &                params    = params,                                     &
      &                layout    = scheme( iScheme )%layout,                   &
      &                physics   = scheme( iScheme )%physics,                  &
      &                varSys    = scheme( iScheme )%varSys,                   &
      &                derVarPos = scheme( iScheme)%derVarPos,                 &
      &                globBC    = scheme( iScheme )%globBC )
    call tem_stopTimer( params%setBndTimer )
    call tem_stopTimer( me = params%timer(iLevel),                           &
                        timerHandle = params%lw_tSetBnd(iLevel) )

    ! Update parameters, relaxation time .etc
    call mus_setParameters( scheme( iScheme ), pdfGlobal, iLevel, params )

    ! Compute current scheme of current level
    call tem_startTimer( params%advRelTimer )
    call tem_startTimer( me = params%timer(ilevel),                          &
      &                  timerHandle = params%lw_tCompute(iLevel) )
!$omp end single

    call scheme( iScheme )%compute(                                            &
      &           iScheme   = iScheme,                                         &
      &           fieldProp = scheme( iScheme )%field(:)%fieldProp,            &
      &           inState   = scheme( iScheme )%pdf(iLevel)%state(:,           &
      &                                  scheme( iScheme )%pdf(iLevel)%nNow),  &
      &           outState  = scheme( iScheme )%pdf(iLevel)%state(:,           &
      &                                  scheme( iScheme )%pdf(iLevel)%nNext), &
      &           neigh     = scheme( iScheme )%pdf(iLevel)%neigh(:),          &
      &           nStart    = 1,                                               &
      &           nEnd      = scheme( iScheme )%pdf(iLevel)%nElems_solve,      &
      &           level     = iLevel,                                          &
      &           layout    = scheme( iScheme )%layout,                        &
      &           pdfGlobal = pdfGlobal,                                       &
      &           params    = params,                                          &
      &           scheme    = scheme,                                          &
      &           tree      = geometry%tree,                                   &
      &           varSys    = scheme( iScheme )%varSys )
    call tem_stopTimer( params%advRelTimer )
    call tem_stopTimer( me = params%timer(iLevel),                           &
                        timerHandle = params%lw_tCompute(iLevel) )

!$omp single
    ! Communicate the halo elements of each scheme on current level 
    call tem_startTimer( params%commTimer )
    call mus_exchange(                                                         &
      &         send    = scheme( iScheme )%glob%levelDesc(iLevel)%sendbuffer, &
      &         recv    = scheme( iScheme )%glob%levelDesc(iLevel)%recvbuffer, &
      &         state   = scheme( iScheme )%pdf(iLevel)%state(:,               &
      &                                  scheme( iScheme )%pdf(iLevel)%nNext), &
      &         pattern = params%commPattern,                                  &
      &         level   = iLevel)
    call tem_stopTimer( params%commTimer )

    ! check if some action has to be taken:
    ! output, tracking, global reduction operations, increase timers
    call check_flow_status( scheme    = scheme,                                &
      &                     params    = params,                                &
      &                     geometry  = geometry,                              &
      &                     pdfGlobal = pdfGlobal )
!$omp end single

  end subroutine do_fast_singleScheme
!******************************************************************************!


!******************************************************************************!
  !> Main control routine: Update the time step for all levels
  !!
  recursive subroutine do_recursive_debug( me, scheme, geometry, params,       &
    &                                      intp, iLevel, pdfGlobal )
    !---------------------------------------------------------------------------
    !> self control type
    class( mus_control_type ) :: me
    !> containers for the different schemes
    type( mus_scheme_type ), intent(inout)  :: scheme(:)   
    !> geometry infomation
    type( mus_geom_type ), intent(in)        :: geometry         
    !> global parameters
    type( param_type ), intent(inout)       :: params       
    !> Interpolation control
    type( interpolation_type ), intent(in) :: intp          
    !> global flow quantity properties
    type( pdf_globInfo_type ), intent(inout)  :: pdfGlobal
    !> Level counter variable
    integer, intent(in) :: iLevel
    !---------------------------------------------------------------------------
    integer :: iScheme    ! Scheme counter variable
    integer :: depScheme    ! depending scheme
    character(len=33) :: buffer
    !---------------------------------------------------------------------------
    if( iLevel .eq. pdfGlobal%minLevel ) then
      call tem_debug( level = 1,                                               &
        &             msg   = ' TIMESTEP '//                                   &
        &                     trim( tem_toStr( pdfGlobal%timing%iTime )))
    endif
    ! For each level, run over all the schemes defined
    do iScheme = 1, pdfGlobal%nSchemes
!$omp single
      write(buffer,'(a,i2,a,i3)')' before advrel L = ', iLevel, 't = ',        &
        &                                                pdfGlobal%timing%iTime
      call dump_debug_info( buffer, params, geometry, scheme, pdfGlobal,       &
        &                   iLevel, iScheme )
      ! swap double buffer index for current level
      scheme( iScheme )%pdf( iLevel )%nNow                                     &
        &                   = mod( scheme( iScheme )%pdf( iLevel )%nNow ,2 ) + 1
      scheme( iScheme )%pdf( iLevel )%nNext                                    &
        &                   = mod( scheme( iScheme )%pdf( iLevel )%nNext,2 ) + 1
      !set boundary for each field in current scheme
      call tem_startTimer( params%setBndTimer )
      call tem_startTimer( me = params%timer(ilevel),                          &
        &                  timerHandle = params%lw_tSetBnd(iLevel) )
      call set_boundary( field     = scheme( iScheme )%field,                  &
        &                pdf       = scheme( iScheme )%pdf,                    &
        &                glob      = scheme( iScheme )%glob,                   &
        &                tree      = geometry%tree,                            &
        &                iLevel    = iLevel,                                   &
        &                pdfGlobal = pdfGlobal,                                &
        &                params    = params,                                   &
        &                physics   = scheme( iScheme )%physics,                &
        &                layout    = scheme( iScheme )%layout,                 &
        &                varSys    = scheme( iScheme )%varSys,                 &
        &                derVarPos = scheme( iScheme)%derVarPos,               &
        &                globBC    = scheme( iScheme )%globBC )
      call tem_stopTimer( params%setBndTimer )
      call tem_stopTimer( me = params%timer(iLevel),                           &
                          timerHandle = params%lw_tSetBnd(iLevel) )
      
      ! Update parameters 
      call mus_setParameters( scheme( iScheme ), pdfGlobal, iLevel, params )

      ! Compute current scheme of current level
      call tem_startTimer( params%advRelTimer )
      call tem_startTimer( me = params%timer(ilevel),                          &
        &                  timerHandle = params%lw_tCompute(iLevel) )
!$omp end single

      call scheme( iScheme )%compute(                                          &
        &         iScheme   = iScheme,                                         &
        &         fieldProp = scheme( iScheme )%field(:)%fieldProp,            &
        &         inState   = scheme( iScheme )%pdf(iLevel)%state(:,           &
        &                                scheme( iScheme )%pdf(iLevel)%nNow),  &
        &         outState  = scheme( iScheme )%pdf(iLevel)%state(:,           &
        &                                scheme( iScheme )%pdf(iLevel)%nNext), &
        &         neigh     = scheme( iScheme )%pdf(iLevel)%neigh(:),          &
        &         nStart    = 1,                                               &
        &         nEnd      = scheme( iScheme )%pdf(iLevel)%nElems_solve,      &
        &         level     = iLevel,                                          &
        &         layout    = scheme( iScheme )%layout,                        &
        &         pdfGlobal = pdfGlobal,                                       &
        &         params    = params,                                          &
        &         scheme    = scheme,                                          &
        &         tree      = geometry%tree,                                   &
        &         varSys    = scheme( iScheme )%varSys )
      call tem_stopTimer( params%advRelTimer )
      call tem_stopTimer( me = params%timer(iLevel),                           &
                          timerHandle = params%lw_tCompute(iLevel) )
      write(buffer,'(a,i2,a,i3)')' after advrel L = ', iLevel, 't = ',         &
        &                                                pdfGlobal%timing%iTime


      call dump_debug_info( buffer, params, geometry, scheme, pdfGlobal,       &
        &                   iLevel, iScheme )

    end do ! scheme loop


    ! Check, if we are on the finest level
    if( iLevel .lt. pdfGlobal%maxLevel ) then
      ! perform interpolation and recursive calls
      call do_computeFinerIntpandExchange( me, scheme, geometry, params,       &
        &                                  intp, iLevel, pdfGlobal )
    end if

!$omp single
    ! Communicate the halo elements of each scheme on current level 
    do iScheme = 1, pdfGlobal%nSchemes
      call tem_startTimer( params%commTimer )
      call mus_exchange(                                                       &
        &  send    = scheme( iScheme )%glob%levelDesc(iLevel)%sendbuffer,      &
        &  recv    = scheme( iScheme )%glob%levelDesc(iLevel)%recvbuffer,      &
        &  state   = scheme( iScheme )%pdf(iLevel)%state(:,scheme( iScheme )%  &
        &                                                  pdf(iLevel)%nNext), &
        &  pattern = params%commPattern,                                       &
        &  level   = iLevel)
      write(buffer,'(a,i2)')' after exchange buffer L      ', iLevel
      call dump_debug_info( buffer, params, geometry, scheme, pdfGlobal,       &
        &                   iLevel, iScheme )
      call tem_stopTimer( params%commTimer )
      if( iLevel .gt. pdfGlobal%minLevel ) then
        call tem_startTimer( params%commTimer )
        call mus_exchange(                                                     &
          & send    = scheme( iScheme )%glob%levelDesc(iLevel)%                &
          &                                             sendbufferFromCoarser, &
          & recv    = scheme( iScheme )%glob%levelDesc(iLevel)%                &
          &                                             recvbufferFromCoarser, &
          & state   = scheme( iScheme )%pdf(iLevel)%state(:,scheme( iScheme )% &
          &                                                pdf(iLevel)%nNext), &
          & pattern = params%commPattern,                                      &
          & level   = iLevel)
        if ( main_debug%checkEachAlgorithmicStep ) then
          write(buffer,'(a,i2)')' after exch gFC buffer L      ', iLevel
          call dump_debug_info( buffer, params, geometry, scheme, pdfGlobal,   &
            &                   iLevel, iScheme )
        end if
        call tem_stopTimer( params%commTimer )
      end if
    end do ! scheme loop
 
    if( iLevel .lt. pdfGlobal%maxLevel ) then
      ! Now when all coarser ghosts are present, we can interpolate the other
      ! way round.
      ! Interpolate the ghost elements on the finer level with data provided 
      ! from current level (This only needs to be done once in two finer time 
      ! steps)
      call do_intpCoarserandExchange( scheme, params, intp, iLevel, pdfGlobal, &
        &                             geometry )   
     ! buffer=' after fillFinerFromMe           '
     ! iScheme = 1
     ! call dump_debug_info( buffer, params, geometry, scheme, pdfGlobal, iLevel, iScheme )
    end if ! if not on finest level

    do iScheme = 1, pdfGlobal%nSchemes
      call tem_startTimer(params%sourceTimer)
      if( allocated(scheme( iScheme )%depend))then
        depScheme = scheme( iScheme )%depend( 1 )%useScheme
        ! update the source terms for each field and the global one for the
        ! scheme
        call mus_update_source(                                                &
          &                   me            = scheme(iScheme)%source,          &
          &                   pdf           = scheme(iScheme)%pdf,             &
          &                   depPdf        = scheme(depScheme)%pdf,           &
          &                   time          = pdfGlobal%timing,                &
          &                   global_varSys = scheme(iScheme)%varSys,          &
          &                   global_tree   = geometry%tree,                   &
          &                   stencil       = scheme(iScheme)%layout%stencil(1))
      else
        ! update the source terms for each field and the global one for the
        ! scheme
        call mus_update_source(                                                &
          &                   me            = scheme(iScheme)%source,          &
          &                   pdf           = scheme(iScheme)%pdf,             &
          &                   time          = pdfGlobal%timing,                &
          &                   global_varSys = scheme(iScheme)%varSys,          &
          &                   global_tree   = geometry%tree,                   &
          &                   stencil       = scheme(iScheme)%layout%stencil(1))  
      end if
      call tem_stopTimer(params%sourceTimer)
    end do

    if( iLevel .eq. pdfGlobal%maxLevel ) then
      ! check if some action has to be taken:
      ! output, tracking, global reduction operations, increase timers
      call check_flow_status( scheme    = scheme,                              &
        &                     params    = params,                              &
        &                     geometry  = geometry,                            &
        &                     pdfGlobal = pdfGlobal )
    end if 
!$omp end single

  end subroutine do_recursive_debug
!******************************************************************************!


end module mus_control_module
!******************************************************************************!


!******************************************************************************!
!>\page computationProcedure Description of the multi-level algorithm  
!!
!!\section multilevel_introduction Introductory comments
!! Here we describe the recursive processing of the level-wise sorted list of 
!! elements through the kernels.
!! From the mesh on disk, which was stored as treeIDs, a level-wise collection
!! of element lists are created before the computation can be started. These
!! lists include ghost and halo cells to account for elements which are missing
!! locally but are calculated on a remote partition as well as missing neighbor
!! elements due to level interfaces. 
!! A detailed description of the mesh generation process and the employed data
!! structures can be found where.
!!
!!
!! All levels are updated starting from the coarsest, 
!! updating all schemes. For each scheme the following procedure is performed.
!! First, the double buffer index is swapped and the boundaries are set.
!! Then the flow field is computed for the current time step (for this level and this scheme).
!! Next, communication with adjacent partitions for current level is performed.
!! If multiple levels are present, a recursive part is entered.
!! On a level n, 2*(n-1) times the time steps of the coarsest level have to be computed.
!! Therefore, the compute routine is called twice for the next finer level.
!! Afterwards, the ghost elements have to be updated again to provide valid values for 
!! the fluid elements. First the coarse ghost elements are updated from fluid elements 
!! on the finer level, then they have to be communicated.
!! Same applies for the finer ghost elements.
!! After all schemes and levels are done, a routine handling output, check and restart
!! is called and the time step is updated.
!!\verbatim
!! recursive subroutine compute( iLevel )
!!   compute_kernel
!!   if( iLevel < maxLevel)
!!     recursive call compute( iLevel+1 )
!!     comm FromCoarser
!!     recursive call compute( iLevel+1 )
!!     fillFromFiner
!!   endif
!!   comm FromFiner
!!   comm 
!!   if( iLevel < maxLevel)
!!     fillFromCoarser( iLevel+1)
!!     comm FromCoarser( iLevel+1)
!!   endif
!! end subroutine compute   
!!\endverbatim
!!
!! \image html multilevel_recursivealgorithm.png
