! See copyright notice in the COPYRIGHT file.
?? include 'header/lbm_macros.inc'
!******************************************************************************!
!> Collection of debugging functions
!!
!! 
module mus_debug_module

  ! include treelm subroutines
  use env_module,              only: long_k
  use tem_tools_module,        only: tem_write
  use tem_timer_module,        only: tem_startTimer, tem_stopTimer
  use tem_param_module,        only: qDirName
  use tem_construction_module, only: tem_levelDesc_type
  use treelmesh_module,        only: treelmesh_type
  use tem_bc_prop_module,      only: tem_bc_prop_type
  use tem_element_module,      only: eT_ghostFromCoarserELem,                  &
    &                                eT_ghostFromFinerElem
  use tem_property_module,     only: prp_hasBnd
  use tem_topology_module,     only: tem_firstIdAtLevel
  use tem_geometry_module,     only: tem_CoordOfId, tem_IdOfCoord,             &
    &                                tem_baryOfId
  use tem_debug_module,        only: tem_debug, main_debug
  use tem_logging_module,      only: tem_toStr

  ! include musubi subroutines
  use mus_aux_module,                only: check_flow_status
  use mus_pdf_module,                only: pdf_data_type, pdf_global_type,     &
    &                                      pdf_globInfo_type
  use mus_global_module,             only: param_type
  use mus_geom_module,               only: mus_geom_type
  use mus_comm_module,               only: mus_exchange
  use mus_bc_general_module,         only: set_boundary
  use mus_scheme_type_module,        only: mus_scheme_type
  use mus_debug_tools_module,        only: dumpPdfAll
  use mus_interpolate_header_module, only: interpolation_type

  implicit none
  private

  public :: dump_elemDependency
  public :: debug_dependenciesFromFiner
  public :: debug_dependenciesFromCoarser
  public :: debug_connectivity
  public :: debug_normals
  public :: dump_debug_info

  contains

!******************************************************************************!
  !> Main control routine: Update the time step for all levels
  !!
  !! All levels are updated starting from the coarsest, 
  !! updating all schemes. For each scheme the following procedure is performed.
  !! First, the double buffer index is swapped and the boundaries are set.
  !! Then the flow field is computed for the current time step (for this level 
  !! and this scheme). Next, communication with adjacent partitions for current 
  !! level is performed If multiple levels are present, a recursive part is 
  !! entered. On a level n, 2*(n-1) times the time steps of the coarsest level 
  !! have to be computed. Therefore, the compute routine is called twice for the
  !! next finer level. Afterwards, the ghost elements have to be updated again 
  !! to provide valid values for the fluid elements. First the coarse ghost
  !! elements are updated from fluid elements on the finer level, then they
  !! have to be communicated. Same applies for the finer ghost elements. After
  !! all schemes and levels are done, a routine handling output, check and 
  !! restart is called and the time step is updated.
  !!
  recursive subroutine debug_recursive_multilevel( scheme, geometry, params,   &
    &                                              intp, iLevel, pdfGlobal )
    !---------------------------------------------------------------------------
    !> containers for the different schemes
    type( mus_scheme_type )    :: scheme(:)
    !> geometry infomation
    type( mus_geom_type )      :: geometry
    !> global parameters
    type( param_type )         :: params
    !> Interpolation control
    type( interpolation_type ) :: intp
    !> global flow quantity properties
    type( pdf_globInfo_type )  :: pdfGlobal
    !---------------------------------------------------------------------------
    integer :: iLevel     ! Level counter variable
    integer :: iLoop      ! nesting loops for diffusive or acoustic scaling
    integer :: iScheme    ! Scheme counter variable
    character(len=34) :: buffer
    !---------------------------------------------------------------------------

    call tem_write( ' TIMESTEP ',pdfGlobal%timing%iTime )
    if( iLevel .eq. pdfGlobal%minLevel ) then
      call tem_debug( level = 1,                                               &
        &             msg   = ' TIMESTEP '//                                   &
        &                     trim( tem_toStr( pdfGlobal%timing%iTime )))
    endif
    ! For each level, run over all the schemes defined
    do iScheme = 1, pdfGlobal%nSchemes
      ! swap double buffer index for current level
!$omp single
      scheme( iScheme )%pdf( iLevel )%nNow  =                                  &
        &                     mod( scheme( iScheme )%pdf( iLevel )%nNow ,2 ) + 1
      scheme( iScheme )%pdf( iLevel )%nNext =                                  &
        &                     mod( scheme( iScheme )%pdf( iLevel )%nNext,2 ) + 1

!      buffer=' before advRel                   '
!      call dump_debug_info( buffer, params, geometry, scheme, pdfGlobal, iLevel, iScheme )
      call tem_startTimer( params%setBndTimer )
      call tem_startTimer( me = params%timer(ilevel),                          &
        &                  timerHandle = params%lw_tSetBnd(iLevel) )
      call set_boundary( field     = scheme( iScheme )%field,                  & 
        &                pdf       = scheme( iScheme )%pdf,                    &
        &                glob      = scheme( iScheme )%glob,                   &
        &                tree      = geometry%tree,                            &
        &                iLevel    = iLevel,                                   & 
        &                pdfglobal = pdfGlobal,                                &
        &                params    = params,                                   &
        &                physics   = scheme( iScheme )%physics,                &
        &                layout    = scheme( iScheme )%layout,                 &
        &                varSys    = scheme( iScheme )%varSys,                 &
        &                derVarPos = scheme( iScheme )%derVarPos,              &
        &                globBC    = scheme( iScheme )%globBC )
      call tem_stopTimer( params%setBndTimer )
      call tem_stopTimer( me = params%timer(iLevel),                           &
                          timerHandle = params%lw_tSetBnd(iLevel) )


      ! Compute current scheme of current level
      call tem_startTimer( params%advRelTimer )
      call tem_startTimer( me = params%timer(ilevel),                          &
        &                  timerHandle = params%lw_tCompute(iLevel) )
!$omp end single
      ! \todo: SZ: update this subroutine for multiple fields
      call scheme( iScheme )%compute(                                          &
        &  iScheme   = iScheme,                                                &
        &  fieldProp = scheme( iScheme )%field(:)%fieldProp,                   &
        &  inState   = scheme( iScheme )%pdf(iLevel)%state(:,scheme(           &
        &                                     iScheme )%pdf(iLevel)%nNow),     &
        &  outState  = scheme( iScheme )%pdf(iLevel)%state(:,scheme(           &
        &                                    iScheme )%pdf(iLevel)%nNext),     &
        &  neigh     = scheme( iScheme )%pdf(iLevel)%neigh(:),                 &
        &  nStart    = 1,                                                      &
        &  nEnd      = scheme( iScheme )%pdf(iLevel)%nElems_solve,             &
        &  level     = iLevel,                                                 &
        &  layout    = scheme( iScheme )%layout,                               &
        &  pdfGlobal = pdfGlobal,                                              &
        &  params    = params,                                                 &
        &  scheme    = scheme,                                                 &
        &  tree      = geometry%tree,                                          &
        &  varSys    = scheme( iScheme )%varSys ) 
      call tem_stopTimer( params%advRelTimer )
      call tem_stopTimer( me = params%timer(iLevel),                           &
                          timerHandle = params%lw_tCompute(iLevel) )
      buffer=' after  advRel                   '
      call dump_debug_info( buffer, params, geometry, scheme, pdfGlobal,       &
        &                   iLevel, iScheme )

      ! Check, if we are on the finest level
      if( iLevel .lt. pdfGlobal%maxLevel ) then
        ! If not, perform interpolation and recursive calls
        do iLoop = 1, pdfGlobal%nNesting 
          ! Perform two time steps on the next finer grid levels
          call debug_recursive_multilevel( scheme, geometry, params, intp,     &
            &                              iLevel+1, pdfGlobal )
          ! Exchange the fine ghost elements
          call mus_exchange(                                                   &
            &    send    = scheme( iScheme )%glob%levelDesc(iLevel+1)%         &
            &                                           sendbufferFromCoarser, &
            &    recv    = scheme( iScheme )%glob%levelDesc(iLevel+1)%         &
            &                                           recvbufferFromCoarser, &
            &    state   = scheme( iScheme )%pdf(iLevel+1)%state(:,scheme(     &
            &                                  iScheme )%pdf(iLevel+1)%nNext), &
            &    pattern = params%commPattern,                                 &
            &    level    = iLevel)
        end do
        call debug_recursive_multilevel( scheme, geometry, params, intp,       &
          &                              iLevel+1, pdfGlobal )

        ! Interpolate the coarse ghost elements from the finer ones
        call intp%fillMineFromFiner_low(                                       &
          &      pdf       = scheme( iScheme )%pdf,                            &
          &      glob      = scheme( iScheme )%glob,                           &
          &      scheme    = scheme( iScheme ),                                &
          &      fieldProp = scheme( iScheme )%field(:)%fieldProp,             &
          &      level     = iLevel,                                           &
          &      intp      = intp,                                             &
          &      ind       = scheme( iScheme )%glob%levelDesc( iLevel)%        &
          &                                                 intpFromFiner_low, &
          &      layout    = scheme( iScheme )%layout,                         &
          &      varSys    = scheme( iScheme )%varSys,                         & 
          &      derVarPos = scheme( iScheme )%derVarPos)

        ! Interpolate the coarse ghost elements from the finer ones
        call intp%fillMineFromFiner(                                           &
          &      pdf       = scheme( iScheme )%pdf,                            &
          &      glob      = scheme( iScheme )%glob,                           &
          &      scheme    = scheme( iScheme ),                                &
          &      fieldProp = scheme( iScheme )%field(:)%fieldProp,             &
          &      level     = iLevel,                                           &
          &      intp      = intp,                                             &
          &      ind       = scheme( iScheme )%glob%levelDesc( iLevel)%        &
          &                                                intpFromFiner_high, &
          &      layout    = scheme( iScheme )%layout,                         &
          &      varSys    = scheme( iScheme )%varSys,                         & 
          &      derVarPos = scheme( iScheme )%derVarPos)

        call tem_startTimer( params%commTimer )
      buffer=' after intp MineFromFiner        '
      call dump_debug_info( buffer, params, geometry, scheme, pdfGlobal,       &
        &                   iLevel, iScheme )

        ! Exchange the coarse ghost elements
        call mus_exchange(                                                     &
          &   send    = scheme( iScheme )%glob%levelDesc(iLevel)%              &
          &                                               sendbufferFromFiner, &
          &   recv    = scheme( iScheme )%glob%levelDesc(iLevel)%              &
          &                                               recvbufferFromFiner, &
          &   state   = scheme( iScheme )%pdf(iLevel)%state(:,scheme(          &
          &                                      iScheme )%pdf(iLevel)%nNext), &
          &   pattern = params%commPattern,                                    &
          &   level   = iLevel)
        call tem_stopTimer( params%commTimer )
!      buffer=' after exch FromFiner            '
!      call dump_debug_info( buffer, params, geometry, scheme, pdfGlobal, iLevel, iScheme )
      endif

!$omp single
      ! Communicate the halo elements of current scheme on current level 
      call tem_startTimer( params%commTimer )
      call mus_exchange(                                                       &
        &   send    = scheme( iScheme )%glob%levelDesc(iLevel)%sendbuffer,     &
        &   recv    = scheme( iScheme )%glob%levelDesc(iLevel)%recvbuffer,     &
        &   state   = scheme( iScheme )%pdf(iLevel)%state(:,scheme( iScheme )% &
        &                                                  pdf(iLevel)%nNext), &
        &   pattern = params%commPattern,                                      &
        &   level   = iLevel)
      call tem_stopTimer( params%commTimer )
!      buffer=' after exchange buffer           '
!      call dump_debug_info( buffer, params, geometry, scheme, pdfGlobal, iLevel, iScheme )
!$omp end single
 
      if( iLevel .lt. pdfGlobal%maxLevel ) then
        ! Now when all coarser ghosts are present, we can interpolate the other
        ! way round. Interpolate the ghost elements on the finer level with data
        ! provided from current level (This only needs to be done once in two 
        ! finer time steps)
        call intp%fillFinerFromMe_low(                                         &
          &    pdf       = scheme( iScheme )%pdf,                              &
          &    glob      = scheme( iScheme )%glob,                             & 
          &    scheme    = scheme( iScheme ),                                  &
          &    fieldProp = scheme( iScheme )%field(:)%fieldProp,               &
          &    level     = iLevel,                                             &
          &    intp      = intp,                                               &
          &    ind       = scheme( iScheme )%glob%levelDesc( iLevel)%          &
          &                                               intpFromCoarser_low, &
          &    layout    = scheme( iScheme )%layout,                           &
          &    varSys    = scheme( iScheme )%varSys,                           & 
          &    derVarPos = scheme( iScheme )%derVarPos)

        call intp%fillFinerFromMe(                                             &
          &    pdf       = scheme( iScheme )%pdf,                              &
          &    glob      = scheme( iScheme )%glob,                             & 
          &    scheme    = scheme( iScheme ),                                  &
          &    fieldProp = scheme( iScheme )%field(:)%fieldProp,               &
          &    level     = iLevel,                                             &
          &    intp      = intp,                                               &
          &    ind       = scheme( iScheme )%glob%levelDesc( iLevel)%          &
          &                                              intpFromCoarser_high, &
          &    layout    = scheme( iScheme )%layout,                           &
          &    varSys    = scheme( iScheme )%varSys,                           & 
          &    derVarPos = scheme( iScheme )%derVarPos)

        call tem_startTimer( params%commTimer )
      buffer=' after fillFinerFromMe           '
      call dump_debug_info( buffer, params, geometry, scheme, pdfGlobal,       &
        &                   iLevel, iScheme )

        ! Exchange the fine ghost elements
        call mus_exchange(                                                     &
          &        send    = scheme( iScheme )%glob%levelDesc(iLevel+1)%       &
          &                                             sendbufferFromCoarser, &
          &        recv    = scheme( iScheme )%glob%levelDesc(iLevel+1)%       &
          &                                             recvbufferFromCoarser, &
          &        state   = scheme( iScheme )%pdf(iLevel+1)%state(:,scheme(   &
          &                                    iScheme )%pdf(iLevel+1)%nNext), &
          &        pattern = params%commPattern,                               &
          &        level   = iLevel)
        call tem_stopTimer( params%commTimer )
!      buffer=' after exch FromCoarser          '
!      call dump_debug_info( buffer, params, geometry, scheme, pdfGlobal, iLevel, iScheme )

      else  ! we are on finest level
!$omp single
        ! check if some action has to be taken:
        ! output, restart, tracking, global reduction operations
!        call check_flow_status( scheme( iScheme ), params, geometry, pdfGlobal, iScheme )
!
!        if( iScheme .eq. pdfGlobal%nSchemes )then
!          ! The last scheme updates then the global time index
!          pdfGlobal%iTime = pdfGlobal%iTime + 1 
!        end if
!$omp end single
      endif ! if not on finest level
    end do ! scheme loop

  end subroutine debug_recursive_multilevel
!******************************************************************************!


!******************************************************************************!
  !> Write required debug info to debug file, output, restart, tracking etc. and
  !! increase time counter.
  !!
  subroutine dump_debug_info( buffer, params, geometry, scheme, pdfGlobal,     &
    &                         iLevel, iScheme, iTime )
    !---------------------------------------------------------------------------
    character(len=*), intent(in)  :: buffer
    !> global parameters
    type( param_type )            :: params
    !> geometry infomation
    type( mus_geom_type )         :: geometry
    !> containers for the different schemes
    type( mus_scheme_type )       :: scheme(:)
    !> global flow quantity properties
    type( pdf_globInfo_type )     :: pdfGlobal
    !> Level counter variable
    integer                       :: iLevel
    !> Scheme counter variable
    integer                       :: iScheme
    !> Level counter variable
    integer, intent(in), optional :: iTime
    !---------------------------------------------------------------------------
 
    call tem_debug( level = 1,                                                 &
      &             msg   = trim( buffer )//' t='//                            &
      &                     trim( tem_toStr( pdfGlobal%timing%rTime )))
    call tem_debug( level = 1,                                                 &
      &             msg   = '            level      '//                        &
      &                     trim( tem_toStr( iLevel )))

    if( main_debug%dumpLevelwiseState       ) then
      call dumpPdfAll( pdf       = scheme( iScheme )%pdf,                      &
        &              glob      = scheme( iScheme )%glob,                     &
        &              layout    = scheme( iScheme )%layout,                   &
        &              iLevel    = iLevel,                                     &
        &              text      = buffer,                                     &
        &              iTime     = iTime,                                      &
        &              dumpHalos = main_debug%dumpHaloState )
    endif
    ! \todo: SZ: change this call such that not all schemes are checked for
    ! debugging
    if( main_debug%checkEachAlgorithmicStep ) then
      call tem_debug( level = 1,                                               &
        &             msg   = '  debug '//trim( buffer )//' t='//              &
        &                     trim( tem_toStr( pdfGlobal%timing%rTime ))//     &
        &                     ' t='//trim( tem_toStr( pdfGlobal%timing%iTime )))
      call check_flow_status( scheme, params, geometry, pdfGlobal )
    endif

  end subroutine dump_debug_info
!******************************************************************************!


!******************************************************************************!
  !>
  !!
  subroutine dumpBufferFromFiner( scheme, text, iLevel )
    !---------------------------------------------------------------------------
    type( mus_scheme_type ) :: scheme
    character(len=*) :: text
    !> Level counter variable
    integer, intent(in) :: iLevel
    !---------------------------------------------------------------------------
    integer :: iElem, elemPos, iDir, pos, iField
    !---------------------------------------------------------------------------

    call tem_debug( level = 1,                                                 &
      &             msg   = 'bufferFromFiner:   '//trim( text ))
    do iField = 1, scheme%nFields
      call tem_debug( level = 1,                                               &
        &             msg   = 'field: '//trim( tem_toStr( iField )))
      do iElem = 1, scheme%pdf( iLevel )%nElems_GhostFromFiner
        elemPos = iElem + scheme%glob%levelDesc( iLevel )%                     &
          &                                  offset(1, eT_ghostFromFinerElem)
        call tem_debug( level = 1,                                             &
          &             msg   = 'TID '//trim( tem_toStr( scheme%glob%          &
          &                     levelDesc( iLevel )%total( elemPos ))))
        do iDir = 1, scheme%layout%stencil(1)%QQ
          pos = scheme%varSys%Variable%val( iField )%varPos(iDir)
          call tem_debug( level = 1,                                           &
            &             msg   = qDirName( iDir )//trim( tem_toStr(           &
            &                               scheme%pdf( iLevel )%state(        &
    ?SAVE?( pos, elemPos, scheme%varSys%nScalars, scheme%pdf( iLevel )%neigh ) &
                                           ,scheme%pdf( iLevel )%nNext ))))
        enddo
      enddo
    enddo
  end subroutine dumpBufferFromFiner
!******************************************************************************!


!******************************************************************************!
  !>
  !!
  subroutine dumpBufferFromCoarser( scheme, text, iLevel )
    !---------------------------------------------------------------------------
    type( mus_scheme_type ) :: scheme
    character(len=*) :: text
    !> Level counter variable
    integer, intent(in) :: iLevel
    !---------------------------------------------------------------------------
    integer :: iElem, elemPos, iDir, pos, iField
    !---------------------------------------------------------------------------

    call tem_debug( level = 1,                                                 &
      &             msg   = 'bufferFromCoarser: '//trim( text ))
    do iField = 1, scheme%nFields
      call tem_debug( level = 1,                                               &
        &             msg   = 'field: '//trim( tem_toStr( iField )))
      do iElem = 1, scheme%pdf( iLevel+1 )%nElems_GhostFromCoarser
        elemPos = iElem + scheme%glob%levelDesc( iLevel+1 )%                 &
          &                                offset(1, eT_ghostFromCoarserElem)
        call tem_debug( level = 1,                                             &
          &             msg   = 'TID '//trim( tem_toStr( scheme%glob%          &
          &                     levelDesc( iLevel+1 )%total( elemPos ))))
        do iDir = 1, scheme%layout%stencil(1)%QQ 
          pos = scheme%varSys%Variable%val( iField )%varPos(iDir)
          call tem_debug( level = 1,                                           &
            &             msg   = qDirName( iDir )//trim( tem_toStr(           &
            &                               scheme%pdf( iLevel+1 )%state(      &
  ?SAVE?( pos, elemPos, scheme%varSys%nScalars, scheme%pdf( iLevel+1 )%neigh ) &
                                           ,scheme%pdf( iLevel+1 )%nNext ))))
        enddo
      enddo
    enddo
  end subroutine dumpBufferFromCoarser
!******************************************************************************!


!******************************************************************************!
  !> Debug the dependencies for the ghostFromFiner elements
  !!
  subroutine debug_dependenciesFromFiner( pdfGlobal, pdf, glob ) 
    !---------------------------------------------------------------------------
    !> global information
    type( pdf_globInfo_type ), intent(inout) :: pdfGlobal
    !> global pdf parameters 
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !> global parameters
    type( pdf_global_type ), intent(in) :: glob
    !---------------------------------------------------------------------------
    integer :: sourceLevel  ! level of source elements
    integer :: sourceElem   ! treeId of current source element
    integer :: targetLevel  ! level of target elements
    integer :: targetElem   ! treeId of current source element
    integer :: iElem        ! current target element (for outer loop)
    integer :: level
    integer :: iSourceElem  ! current source element (for inner loop)
    integer :: nSourceElems ! number of source elements for the current target
    character(len=256) :: buffer
    !---------------------------------------------------------------------------
 
    call tem_debug( level = 1,                                                 &
      &             msg   =  'debug depFromFiner' )

    do level = pdfGlobal%minLevel, pdfGlobal%maxLevel
      ! Treat every coarse target element:
      do iElem = 1, pdf( level )%nElems_ghostFromFiner
        ! find out which is the next finer level
        sourceLevel = glob%levelDesc( level )%depFromFiner( iElem )%           &
          &                                                     dependencyLevel
        ! call the current level "targetLevel" just for clarity
        targetLevel = level
        if( sourceLevel - targetLevel .ne. 1 ) then
          call tem_debug( level = 1,                                           &
            &             msg   = 'error in level difference on level '//      &
            &                     trim( tem_toStr( level))//' elem '//         &
            &                     trim( tem_toStr( iElem))//' sourcel '//      &
            &                     trim( tem_toStr( sourceLevel )))
          stop
        endif

        ! find out the treeId of the current target element that we are
        ! treating in this iteration
        targetElem = iElem + glob%levelDesc( targetLevel )%                    &
          &                                  offset( 1, eT_ghostFromFinerElem )
        ! find out, how many fine source elements we have for interpolation.
        ! Usually 8, but can differ at corners, obstacles, boundaries...
        nSourceELems = glob%levelDesc( targetLevel )%depFromFiner( iElem )%    &
          &                                                          elem%nVals
        write( buffer,'(a, i8)')  'FromFiner ', glob%levelDesc( targetLevel )% &
          &                                                 total( targetElem )
        !... and all the fine source elements...
        do iSourceElem = 1, nSourceElems
          ! Find out the source element's treeId
          sourceElem = glob%levelDesc( targetLevel )%depFromFiner( iElem )%    &
            &                                           elem%val( iSourceElem )
          if( sourceElem .gt. 0 ) then
            write( buffer, '(2a,i8)' )  trim(buffer), ' ', glob%levelDesc(     &
              &                               sourceLevel )%total( sourceElem )
          else
            write( buffer, '(2a,i8)' )  trim(buffer), 'x', sourceElem
          endif
        enddo
        call tem_debug( level = 1,                                             &
          &             msg   = trim( buffer ))
      enddo
    enddo

  end subroutine debug_dependenciesFromFiner
!******************************************************************************!


!******************************************************************************!
  !> Interpolation routine that is based on a simple weighted average of source
  !! nodes. This is the interpolation coarse-> fine. Weights are needed here,
  !! as the distances source <-> target are different for the source nodes.
  !!
  subroutine debug_dependenciesFromCoarser( pdfGlobal, pdf, glob) 
    !---------------------------------------------------------------------------
    !> global information
    type( pdf_globInfo_type ), intent(inout) :: pdfGlobal
    !> global fluid parameters 
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !> global parameters 
    type( pdf_global_type ), intent(in) :: glob
    !---------------------------------------------------------------------------
    integer :: sourceLevel   ! level of source elements
    integer :: sourceElem    ! treeId of current source element
    integer :: targetLevel   ! level of target elements
    integer :: targetElem    ! treeId of current source element
    integer :: iElem         ! current target element (for outer loop)
    integer :: iSourceElem   ! current source element (for inner loop)
    integer :: nSourceElems  ! number of source elements for the current target
    integer :: level
    character(len=256) :: buffer
    !---------------------------------------------------------------------------

    call tem_debug( level = 1,                                                 &
      &             msg   =  'debug depFromCoarser' )

    do level = pdfGlobal%minLevel, pdfGlobal%maxLevel-1
      sourceLevel = level
      targetLevel = level + 1 !\todo: why not use depFromFiner?
      
      ! Treat every coarse target element:
      do iElem = 1, pdf( targetLevel )%nElems_ghostFromCoarser 
        ! find out the treeId of the current target element that we are
        ! treating in this iteration
        targetElem = iElem + glob%levelDesc( targetLevel )%                    &
          &                                offset( 1, eT_ghostFromCoarserElem )
        ! find out, how many coarse source elements we have for interpolation.
        ! Usually 8, but can differ at corners, obstacles, boundaries...
        nSourceElems = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%  &
          &                                                          elem%nVals
    
        write( buffer,'(a, i8)')  'FromCoarser ', glob%levelDesc(targetLevel)% &
          &                                                 total( targetElem )
        !... and all the fine source elements...
        do iSourceElem = 1, nSourceElems
          ! Find out the source element's treeId
          sourceElem = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%  &
            &                                           elem%val( iSourceElem )

          if( sourceElem .gt. 0 ) then
            write( buffer, '(2a,i8)' )  trim(buffer), ' ', glob%levelDesc(     &
              &                               sourceLevel )%total( sourceElem )
          else
            write( buffer, '(2a,i8)' )  trim(buffer), 'x', sourceElem
          endif
        enddo
        call tem_debug( level = 1,                                             &
          &             msg   = trim( buffer ))
      enddo
    enddo

  end subroutine debug_dependenciesFromCoarser
!******************************************************************************!


!******************************************************************************!
  !> 
  !!
  subroutine dump_elemDependency( levelDesc, ghostElem, fromCoarser,           &
    &                             targetLevel, minlevel )
    !---------------------------------------------------------------------------
    integer, intent(in) :: minlevel
    type( tem_levelDesc_type ), intent(in) :: levelDesc(minlevel:)
    integer, intent(in) :: ghostElem
    integer, intent(in) :: targetLevel 
    logical, intent(in) :: fromCoarser
    !---------------------------------------------------------------------------
    integer :: sourceLevel   ! level of source elements
    integer :: sourceElem    ! treeId of current source element
    integer :: targetElem    ! treeId of current source element
    integer :: iSourceElem   ! current source element (for inner loop)
    integer :: nSourceElems  ! number of source elements for the current target
    character(len=256) :: buffer
    !---------------------------------------------------------------------------

    ! find out the treeId of the current target element that we are treating in
    ! this iteration
    targetElem = ghostElem + levelDesc( targetLevel )%                         &
      &                                    offset( 1, eT_ghostFromCoarserElem )
    sourceLevel = targetLevel - 1
    ! find out, how many coarse source elements we have for interpolation.
    ! Usually 8, but can differ at corners, obstacles, boundaries...
    nSourceElems = levelDesc( targetLevel )%depFromCoarser( ghostElem )%       &
      &                                                              elem%nVals
    
    write( buffer,'(a, i8)')  'FromCoarser ', levelDesc( targetLevel )%        &
      &                                                     total( targetElem )
    !... and all the fine source elements...
    do iSourceElem = 1, nSourceElems
      ! Find out the source element's treeId
      sourceElem = levelDesc( targetLevel )%depFromCoarser( ghostElem )%       &
        &                                               elem%val( iSourceElem )
   
      if( sourceElem .gt. 0 ) then
        write( buffer, '(2a,i8)' )  trim(buffer), ' ', levelDesc(              &
          &                                   sourceLevel )%total( sourceElem )
      else
        write( buffer, '(2a,i8)' )  trim(buffer), 'x', sourceElem
      endif
    enddo
    call tem_debug( level = 1,                                                 &
      &             msg   = trim( buffer ))

  end subroutine dump_elemDependency
!******************************************************************************!


!******************************************************************************!
  !> Debug the finding of neighbors
  !!
  subroutine debug_find_neighbor( nSides, boundary, treeidsubset,              &
    &                             propertylistsubset, nElemsSubset )
    !---------------------------------------------------------------------------
    integer, intent(in)                  :: nSides
    type( tem_bc_prop_type ), intent(in) :: boundary
    integer(kind=long_k), intent(in)     :: treeidsubset(:)
    integer(kind=long_k), intent(in)     :: propertylistsubset(:)
    integer, intent(in)                  :: nElemsSubset
    !---------------------------------------------------------------------------
    integer              :: iElem
    integer(kind=long_k) :: myProperty
    integer              :: x(4)
    integer :: nElemsBnd, iDirection
    logical :: hasBoundary, hasBndInDir
    integer(kind=long_k) :: tOffset, origID    ! List tracking pointer
    character(len=1024) :: buffer
    !---------------------------------------------------------------------------
    nElemsBnd = 0

    call tem_write( 'Debugging neighbor search  ...' )
    call tem_debug( level = 1,                                                 &
      &             msg   = '' )
    call tem_debug( level = 1,                                                 &
      &             msg   = 'List of treeIDs' )
    call tem_debug( level = 1,                                                 &
      &             msg   = trim( tem_toStr( val = int( treeidsubset ),        &
      &                                      sep = ',' )))
    call tem_debug( level = 1,                                                 &
      &             msg   = '' )
    call tem_debug( level = 1,                                                 &
      &             msg   = 'Debugging neighbor search ... ' )
    call tem_debug( level = 1,                                                 &
      &             msg   = '' )

    elemLoop: do iElem = 1, nElemsSubset

      x = tem_CoordOfId( treeidsubset( iElem ))
      call tem_debug( level = 1,                                               &
        &             msg   = 'treeID '//                                      &
        &                     trim( tem_toStr( int( treeIDsubset( iElem )))))
      tOffset = tem_FirstIDAtLevel( x(4) )
      origID = tem_IdOfCoord( x, tOffset ) !< MH: was passiert hier?
      myProperty = propertylistsubset( iElem )
      hasBoundary = .false.

      !
      ! First check, if I have boundaries
      ! If yes, there might be no neighbors
      if( btest( myProperty, prp_hasBnd )) then
        hasBoundary = .true.
        nElemsBnd = nElemsBnd +1
      endif
      !
      ! start searching for neighbors stored in nTreeID
      ! and create linked lists in neighborList
      !
      !
      buffer = ''
      directionLoop: do iDirection = 1, nSides
        hasBndInDir = .false.
        if( hasBoundary ) then
          if( boundary%boundary_ID(( iDirection ),nElemsBnd ) .gt. 0) then
            hasBndInDir = .true. !< PUSH
          endif
        endif
        !
        ! If there is NO boundary in the current direction
        if( .not. hasBndInDir ) then
!          write(buffer,'(3a,i4)') trim(buffer), qDirName(iDirection), &
!          '  ',int(neighbor(ielem)%elemId(iDirection))
        else
          write(buffer,'(3a,i4)') trim( buffer ), qDirName( iDirection ),' bnd '
        endif
      end do directionLoop
      call tem_debug( level = 1,                                               &
        &             msg   = trim( buffer ))
    end do elemLoop
    call tem_debug( level = 1,                                                 &
      &             msg   = '' )
    call tem_debug( level = 1,                                                 &
      &             msg   = 'Finished debugging neighbor search.' )
    call tem_debug( level = 1,                                                 &
      &             msg   = '----------------------------------------------'// &
      &                     '---------' )
    call tem_debug( level = 1,                                                 &
      &             msg   = '' )


  end subroutine debug_find_neighbor
!******************************************************************************!


!******************************************************************************!
  !> \brief Dump the neighbors found for the boundaries with neighbor elements
  !!
  subroutine debug_normals( tree, scheme, pdfGlobal )
    !---------------------------------------------------------------------------
    type( treelmesh_type ), intent(in)    :: tree  
    type(pdf_globInfo_type), intent(in)   :: pdfGlobal
    type( mus_scheme_type ) :: scheme
    !--------------------------------------------------------------------------- 
    integer :: iElem, iLevel, iBnd, iNeigh
    integer(kind=long_k) :: nTreeID
    !--------------------------------------------------------------------------- 

    call tem_debug( level = 1,                                                 &
      &             msg   = 'Debugging normals...' )
    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      do iBnd = 1, pdfGlobal%nBCs
        call tem_debug( level = 1,                                             &
          &             msg   = 'Boundary ID '//trim( tem_toStr( iBnd))//      &
          &                     ' at Level '//trim( tem_toStr( iLevel )))
        call tem_debug( level = 1,                                             &
          &             msg   = 'following tIDs with neighbor relations' )
        do iElem = 1, scheme%globBC( iBnd )%nElems( iLevel )
          nTreeID = scheme%glob%levelDesc( iLevel )%total(                     &
            &              scheme%globBC( iBnd )%elems( iLevel )%elem( iElem ))
          call tem_debug( level = 1,                                           &
            &             msg   = 'tID '//trim( tem_toStr( nTreeID ))//        &
            &                     ' pos '//trim( tem_toStr(                    &
            &                       val       = tem_BaryOfId( tree, nTreeID ), &
            &                       sep       = ',' )))
          call tem_debug( level = 1,                                           &
            &             msg   = 'normal '//trim( tem_toStr(                  &
            &               val       = scheme%globBC( iBnd )%elems( iLevel )% &
            &                                              normal( :, iElem ), &
            &               sep       = ',' ))//                               &
            &                     ' nrmInd '//trim( tem_toStr(                 &
            &                           scheme%globBC( iBnd )%elems( iLevel )% &
            &                                           normalInd( iElem )))// &
            &                     ' nNeighs '//trim( tem_toStr(                &
            &                           scheme%field(1)%bc( iBnd )%nNeighs )))


          do iNeigh = 1, scheme%field(1)%bc( iBnd )%nNeighs
            call tem_debug( level = 1,                                         &
              &             msg   = '   '//trim( tem_toStr( iNeigh ))//        &
              &                     ' neigh '//trim( tem_toStr(                &
              &        scheme%glob%levelDesc( iLevel )%total( scheme%field(1)% &
              &        bc( iBnd )%elems( iLevel )%neigh( iNeigh, iElem ))))//  &
              &                     ' '//trim( tem_toStr(                      &
              &          val       = tem_baryOfId( tree, scheme%glob%          &
              &                        levelDesc( iLevel )%total( scheme%      &
              &                        field(1)%bc( iBnd )%elems( iLevel )%    &
              &                                     neigh( iNeigh, iElem ))),  &
              &          sep       = ',' )))
          enddo
        enddo
      enddo
    enddo

  end subroutine debug_normals
!******************************************************************************!


!******************************************************************************!
  !> Detailed Debug output to the neighbor / connectivity list
  !!
  !! currently the dubug connectivity is done only for 1st field
  subroutine debug_connectivity( scheme, pdfGlobal )
    !---------------------------------------------------------------------------
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    type(mus_scheme_type)               :: scheme
    !--------------------------------------------------------------------------- 
    integer(kind=long_k) :: iElem
    integer :: iLevel, iDir, iField
    integer :: iQQ, varPos(scheme%layout%stencil(1)%QQ), nScalars
    !--------------------------------------------------------------------------- 

    iQQ = scheme%layout%stencil(1)%QQ
    ! take varpos of 1st field
    nScalars = scheme%varSys%nScalars

    call tem_debug( level = 1,                                                 &
      &             msg   = '' )
    call tem_debug( level = 1,                                                 &
      &             msg   = ' Debugging connectivity for field 1' )
    call tem_debug( level = 1,                                                 &
      &             msg   = '' )
    call tem_debug( level = 1,                                                 &
      &             msg   = 'NEIGHBOR from '//                                 &
      &                     trim( tem_toStr( pdfGlobal%minLevel ))//' to '//   &
      &                     trim( tem_toStr( pdfGlobal%maxLevel )))
    do iLevel= pdfGlobal%minLevel, pdfGlobal%maxLevel
      call tem_debug( level = 1,                                               &
        &             msg   = '' )
      call tem_debug( level = 1,                                               &
        &             msg   = 'LEVEL '//trim( tem_toStr( iLevel )))
      call tem_debug( level = 1,                                               &
        &             msg   = 'nElems of fluid '//                             &
        &                 trim( tem_toStr( scheme%pdf( iLevel )%nElems_fluid )))
      call tem_debug( level = 1,                                               &
        &             msg   = 'nElems of solve '//                             &
        &                 trim( tem_toStr( scheme%pdf( iLevel )%nElems_solve )))
      call tem_debug( level = 1,                                               &
        &             msg   = 'nElems of halo '//                              &
        &                 trim( tem_toStr( scheme%pdf( iLevel )%nElems_halo )))
      call tem_debug( level = 1,                                               &
        &             msg   = 'nElems of ghost '//                             &
        &                 trim( tem_toStr( scheme%pdf( iLevel )%nElems_ghost )))
      call tem_debug( level = 1,                                               &
        &             msg   = 'nElems of local '//                             &
        &                 trim( tem_toStr( scheme%pdf( iLevel )%nElems_local )))
      call tem_debug( level = 1,                                               &
        &             msg   = 'size of total '//                               &
        &       trim( tem_toStr( size(scheme%glob%levelDesc( iLevel )%total ))))
      call tem_debug( level = 1,                                               &
        &             msg   = 'size of state '//                               &
        &              trim( tem_toStr( size(scheme%pdf( iLevel )%state(:,1)))))
      call tem_debug( level = 1,                                               &
        &             msg   = 'size of neigh '//                               &
        &                  trim( tem_toStr( size(scheme%pdf( iLevel )%neigh ))))
      do iElem = 1, scheme%pdf( iLevel )%nElems_local
        call tem_debug( level = 1,                                             &
          &             msg   = 'IELEM '//trim( tem_toStr( iElem ))//          &
          &                     ' treeID '//trim( tem_toStr(                   &
          &            int(scheme%glob%levelDesc( iLevel )%total( iElem ))))// &
          &                     ' coord '//trim( tem_toStr(                    &
          &              val = tem_CoordOfId( scheme%glob%levelDesc( iLevel )% &
          &                                                    total( iElem )),&
          &              sep = ',' )))
        do iField = 1, scheme%nFields  
         varPos = scheme%varSys%Variable%val( iField )%varPos
         call tem_debug( level = 1,                                            &
           &             msg   = 'Field '//trim( tem_toStr( iField )))
          do iDir = 1, iQQ
            if( scheme%pdf( iLevel )%neigh                                     &
              &( ?NGPOS?( varPos(iDir), iElem, nScalars ) )  .le. 0 ) then
              call tem_debug( level = 1,                                       &
                &             msg   = 'ERROR in '//qDirName( iDir ))
            else
              call tem_debug( level = 1,                                       &
                &             msg   = trim( tem_toStr( varPos( iDir )))//      &
                &                     ' '//trim( qDirName( iDir ))//           &
                &                     ' tID '//trim( tem_toStr(                &
                &                   int(scheme%glob%levelDesc( iLevel )%total( &
  &?ElemIDX?(scheme%pdf( iLevel )%neigh( ?NGPOS?( varPos(iDir), iElem, nScalars )), nScalars )))))//&
                &                     ' crd '//trim( tem_toStr(                &
                &val = tem_CoordOfId(   scheme%glob%levelDesc( iLevel )%total( &
  &?ElemIDX?(scheme%pdf( iLevel )%neigh( ?NGPOS?( varPos(iDir), iElem, nScalars )), nScalars ))),&
                &sep = ',' )))
            endif
          enddo
        enddo
      enddo
    enddo
    call tem_debug( level = 1,                                                 &
      &             msg   = ' End Debugging connectivity ' )
    call tem_debug( level = 1,                                                 &
      &             msg   = '----------------------------------------------'// &
      &                     '------' )

  end subroutine debug_connectivity
!******************************************************************************!


end module mus_debug_module
!******************************************************************************!
