! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Collection of debugging functions for writing stuff to files
!! \author Manuel Hasert
?? include 'header/lbm_macros.inc'
module mus_debug_tools_module

  ! include treelm modules
  use env_module,          only: long_k
  use tem_element_module,  only: eT_fluidElem, eT_ghostFromFinerElem,          &
    &                            eT_ghostFromCoarserElem, eT_haloElem
  use tem_debug_module,    only: tem_debug

  ! include musubi modules
  use mus_pdf_module,           only: pdf_data_type, pdf_global_type
  use mus_scheme_layout_module, only: mus_scheme_layout_type

  implicit none
  private

  public :: dumpPdfAll
  public :: dumpPdfSingleElem
  public :: dumpPdfSingleElemSAVE

  contains
!******************************************************************************!
  !> write all pdf entries of a single element to the debug unit
  !!
  subroutine dumpPdfSingleElem( elem, pdf, layout, iLevel, text )
    !---------------------------------------------------------------------------
    type( pdf_data_type ), intent(inout) :: pdf(:)   !< pdf state
    type( mus_scheme_layout_type), intent(in)   :: layout !< the layout used
    character(len=*),optional :: text  !< descriptive text as title 
    integer, intent(in) :: elem       !< element position in totalList
    integer, intent(in) :: iLevel     !< Level counter variable
    !---------------------------------------------------------------------------
    integer :: elemPos, iQQ, iDir
    character(len=1024) :: buffer
    !---------------------------------------------------------------------------

    elemPos = elem
    iQQ = layout%stencil(1)%QQ

    if(present(text))                                                          &
      write(buffer,*)  trim(text)

!    write(buffer,'(a,i8)')  ''
    do iDir = 1, iQQ 
      write(buffer,'(2a,en17.7)') trim(buffer),' ',pdf( iLevel )%state(        &
        & ?IDX?( iDir, elemPos, iQQ, pdf( iLevel)%neigh), pdf( iLevel)%nNext ) 
    enddo
    call tem_debug( level = 1,                                                 &
      &             msg   = trim( buffer ))

  end subroutine dumpPdfSingleElem
!******************************************************************************!


!******************************************************************************!
  !> write all pdf entries of a single element to the debug unit
  !! with SAVE access to the state array
  !!
  subroutine dumpPdfSingleElemSAVE( elem, pdf, glob, layout, iLevel, text )
    !---------------------------------------------------------------------------
    type( pdf_data_type ), intent(inout) :: pdf(:) !< pdf state
    type( pdf_global_type ), intent(in) :: glob    !< global parameters 
    type( mus_scheme_layout_type), intent(in)   :: layout !< the layout used
    character(len=*),optional :: text !< descriptive text as title
    integer, intent(in) :: elem       !< element position in totalList
    integer, intent(in) :: iLevel     !< Level counter variable
    !---------------------------------------------------------------------------
    integer :: elemPos, iQQ, iDir
    integer(kind=long_k) :: nTreeID
    character(len=1024) :: buffer
    !---------------------------------------------------------------------------

    elemPos = elem
    iQQ = layout%stencil(1)%QQ
    nTreeID = glob%levelDesc( iLevel)%total( elemPos )

    if(present(text)) &
      write(buffer,*)  trim(text)

    write(buffer,'(a,i8)')  'sgl ',nTreeID
    do iDir = 1, iQQ 
      write(buffer,'(2a,en17.7)') trim(buffer),' ',pdf( iLevel )%state(        &
        & ?SAVE?( iDir, elemPos, iQQ, pdf( iLevel)%neigh), pdf( iLevel)%nNext ) 
    enddo
    call tem_debug( level = 1,                                                 &
      &             msg   = trim( buffer ))

  end subroutine dumpPdfSingleElemSAVE
!******************************************************************************!


!******************************************************************************!
  !> write all pdf entries of a all elements of the pdf array for a given level
  !! to the debug unit
  !!
  subroutine dumpPdfAll( pdf, glob, layout, iLevel, text, iTime, dumpHalos )
    !---------------------------------------------------------------------------
    type( pdf_data_type ), intent(inout) :: pdf(:) !< pdf state 
    type( pdf_global_type ), intent(in) :: glob    !< global parameters
    type( mus_scheme_layout_type), intent(in)   :: layout !< the layout used
    logical, optional, intent(in)   :: dumpHalos
    character(len=*) :: text
    integer, intent(in) :: iLevel     !< Level counter variable
    integer, intent(in), optional :: iTime      !< Level counter variable
    !---------------------------------------------------------------------------
    integer :: iElem, elemPos, iType, nElems, iQQ, iDir, iTLayer
    integer :: nTypes
    character(len=1024) :: buffer
    logical    :: loc_dumpHalos
    !---------------------------------------------------------------------------

    iQQ = layout%stencil(1)%QQ
    iTLayer = pdf( iLevel )%nNext
    if( present( iTime ) ) then
      iTLayer = iTime
    endif
    if( present( dumpHalos ) ) then
      loc_dumpHalos = dumpHalos
    else
      loc_dumpHalos = .true.
    endif
    if( loc_dumpHalos ) then
      nTypes = 4
    else
      nTypes = 3
    end if

    call tem_debug( level = 1,                                                 &
      &             msg   = trim( text ))
    do iType = 1, nTypes
      select case( iType ) 
      case( eT_fluidElem )
        nElems = pdf( iLevel)%nElems_Fluid 
        call tem_debug( level = 1,                                             &
          &             msg   = 'FLUID' )
      case( eT_ghostFromCoarserElem )
        nElems = pdf( iLevel)%nElems_ghostFromCoarser
        call tem_debug( level = 1,                                             &
          &             msg   = 'GHOSTFROMCOARSER' )
      case( eT_ghostFromFinerElem )
        nElems = pdf( iLevel)%nElems_ghostFromFiner
        call tem_debug( level = 1,                                             &
          &             msg   = 'GHOSTFROMFINER' )
      case( eT_haloElem )
        nElems = pdf( iLevel)%nElems_halo
        call tem_debug( level = 1,                                             &
          &             msg   = 'HALO' )
      end select

      do iElem = 1, nElems
        elemPos = iElem + glob%levelDesc( iLevel )%offset( 1, iType )
        write(buffer,'(i8)')  glob%levelDesc( iLevel)%total( elemPos )
        do iDir = 1, iQQ 
          write(buffer,'(2a,en17.7)') trim(buffer),' ',pdf( iLevel )%state(   &
            & ?SAVE?( iDir, elemPos, iQQ, pdf( iLevel)%neigh), iTLayer ) 
        enddo
        call tem_debug( level = 1,                                             &
          &             msg   = trim( buffer ))
      enddo
    enddo
  end subroutine dumpPdfAll
!******************************************************************************!


end module mus_debug_tools_module
!******************************************************************************!
