! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module provides functions for calculating macroscopic quantities
?? include 'header/lbm_macros.inc'
module mus_derivedQuantities_module2
  
  ! include treelm modules
  use env_module,               only: rk, globalMaxLevels, long_k, labelLen,   &
    &                                 pathlen
  use treelmesh_module,         only: treelmesh_type
  use tem_tools_module,         only: tem_write, tem_horizontalSpacer
  use tem_var_system_module,    only: tem_var_system_type
  use tem_stencil_module,       only: tem_stencilHeader_type
  use tem_aux_module,           only: tem_abort
  use tem_source_module,        only: tem_sourceMap_type
  use tem_time_module,          only: tem_time_type 
  use tem_spacetime_fun_module, only: tem_spacetime_for
  use tem_param_module,         only: div1_3, div3_4, cs2inv, div1_9, div1_54, & 
    &                                 cs2, cs4inv, div1_2, sqrt3

  ! include musubi modules
  use mus_pdf_module,           only: pdf_globInfo_type
  use mus_field_prop_module,    only: mus_field_prop_type
  use mus_scheme_layout_module, only: mus_scheme_layout_type
  use mus_moments_module,       only: get_moment

  implicit none
  private

  public :: getDensity, getMomentum, getVelocity
  public :: getEquilibriumIncomp, getEquilibrium
  public :: getWSS2D, getWSS, getShearStress, getShearStressTensor
  public :: getShearRateTensor_diffusive
  public :: getShearRateTensor_acoustic 
  public :: getDensityPre, getVelocityPre
  public :: getEqDistr, getVelocityDir
  public :: set_pdfDiffusive
  public :: set_pdfAcoustic 
  public :: getNEq_diffusive
  public :: getNEq_acoustic
  public :: convPrePost
  public :: getShearstressTensorIncomp_forPdfSubset
  public :: setThirdOrder_diffusive

  interface getVelocityDir
    module procedure getVelocityDir_noForce
    module procedure getVelocityDir_Force
  end interface

  interface getShearStressTensor
    module procedure getShearStressTensor_forElemFromState
    module procedure getShearStressTensor_forPdfSubset
    module procedure getShearStressTensorIncomp_forPdfSubset
  end interface

  interface getShearRateTensor_diffusive
    module procedure getShearRateTensor_diffusive_forElemFromState
    module procedure getShearRateTensor_diffusive_forPdfSubset
  end interface

  interface getShearRateTensor_acoustic 
    !module procedure getShearRateTensor_acoustic_forElemFromState
    module procedure getShearRateTensor_acoustic_forPdfSubset
  end interface

  interface getEquilibrium
    module procedure getEquilibrium_forElemfromState
    module procedure getEquilibrium_fromMacrVal
    module procedure getEquilibrium_forPdfSubset
  end interface

  interface getDensityPre
    module procedure getDensity_forElemfromStatePre
  end interface

  interface getVelocityPre
    module procedure getVelocity_forElemFromState_noForcePre
  end interface

  interface getDensity
    module procedure getDensity_forElemfromState
    module procedure getDensity_forPdfSubset
  end interface

  interface getVelocity
    module procedure getVelocity_forElemFromState_noForce
    module procedure getVelocity_forElemFromState_Force
    module procedure getVelocity_forPdfSubset
  end interface

  interface getMomentum
    module procedure getMomentum_forElemFromState_noForce
    module procedure getMomentum_forElemFromState_Force
    module procedure getMomentum_forPdfSubset
  end interface
contains



!!******************************************************************************!
!  !> access the omega from fluidProp%fluid for given elemPos which defined 
!  !! sourcelevel.
!  !!
!  !! use this function only for standard lbm simulation
!  subroutine accessOmega_fromFieldProp( state, neigh, elemPos, fieldProp, varSys, &
!    &          globSys, layout, res, iVar, varDep, varMap, sub_index, sourceMap, &
!    &          treeID, time, tree)
!    !------------------------------------------------------------------------ 
!    !> state vector all elements in a level with all fields
!    real(kind=rk), intent(in)    :: state(:)
!    !> neighbor array
!    integer, intent(in), optional :: neigh(:)
!    !> position of element in the state array
!    integer, intent(in) :: elemPos
!    !> requested variable system 
!    type( tem_var_system_type ), optional, intent(in) :: varSys 
!    !> global variable system 
!    type( tem_var_system_type ), intent(in) :: globSys
!    !> information about the field
!    type( mus_field_prop_type ), intent(in) :: fieldprop(:)
!    !> the scheme layout representation used in this scheme with stencil
!    type( mus_scheme_layout_type ), intent(in) :: layout 
!    !> calculated / derived quantities
!    real(kind=rk), intent(out)  :: res(:)
!    !> which variable of the re variable system to output? optional!
!    integer, optional, intent(in) :: iVar
!    !> array which points to dependend state variable of 
!    !! requested variable iVar at position varSys%mapping(iVar)
!    !! in the global variable system
!    integer, intent(in) :: varDep(:)
!    !> global list of all quantities
!    type(mus_varMap_type), optional, intent(inout) :: varMap(:)
!    !> an optional subset of the quantities to calculate
!    integer, optional, intent(in) :: sub_index(:)
!
!    !> the current treeID
!    integer(kind=long_k), optional, intent(in) :: treeID
!    !> information on the sources (including pointers to the lua functions)
!    type( tem_sourceMap_type ), optional, intent(in) :: sourceMap
!    !> current timing information
!    type(tem_time_type), optional, intent(in) :: time
!    !> tree used by spacetime function
!    type( treelmesh_type ), optional, intent(in) :: tree
!    !------------------------------------------------------------------------ 
!    !local variable
!    integer :: iDir, pos, iDep
!    !------------------------------------------------------------------------ 
!    
!    res = 0._rk
!
!    res = fieldProp( iVar )%fluid%omLvl( elemPos )
!
!  end subroutine accessOmega_FromFieldProp
!!******************************************************************************!


!!******************************************************************************!
!  !> call the spacetime function (provided by lua ) of variable,
!  !! and return values to res
!  subroutine callSpaceTimeFun( state, neigh, elemPos, fieldProp,  &
!    &         varSys, globSys, layout, res, iVar, varDep, varMap, &
!    &         sub_index, sourceMap, treeID, time, tree ) 
!    !------------------------------------------------------------------------ 
!    !> state vector all elements in a level with all fields
!    real(kind=rk), intent(in)    :: state(:)
!    !> neighbor array
!    integer, intent(in), optional :: neigh(:)
!    !> position of element in the state array
!    integer, intent(in) :: elemPos
!    !> requested variable system
!    type( tem_var_system_type ), optional, intent(in) :: varSys 
!    !> global variable system
!    type( tem_var_system_type ), intent(in) :: globSys
!    !> information about the field
!    type( mus_field_prop_type ), intent(in) :: fieldprop(:)
!    !> the scheme layout representation used in this scheme with stencil
!    type( mus_scheme_layout_type ), intent(in) :: layout 
!    !> which variable of the re variable system to output? optional!
!    integer, optional, intent(in) :: iVar
!    !> array which points to dependend state variable of 
!    !! requested variable iVar at position varSys%mapping(iVar)
!    !! in the global variable system
!    integer, intent(in) :: varDep(:)
!    !> global list of all quantities
!    type(mus_varMap_type), optional, intent(inout) :: varMap(:)
!    !> an optional subset of the quantities to calculate
!    integer, optional, intent(in) :: sub_index(:)
!    !> calculated / derived quantities
!    real(kind=rk), intent(out)  :: res(:)
!
!    !> the current treeID
!    integer(kind=long_k), optional, intent(in) :: treeID
!    !> information on the sources (including pointers to the lua functions)
!    type( tem_sourceMap_type ), optional, intent(in) :: sourceMap
!    !> current timing information
!    type( tem_time_type ), optional, intent(in) :: time
!    !> tree used by spacetime function
!    type( treelmesh_type ), optional, intent(in) :: tree
!    !------------------------------------------------------------------------ 
!    integer(kind=long_k), allocatable :: tmp_treeID(:)
!    integer :: iFun
!    !-------------------------------------------------------------------------
!    allocate( tmp_treeID( 1 ))
!    tmp_treeID( 1 ) = treeID
!    !------------------------------------------------------------------------ 
!
!    res = 0._rk
!    do iFun = 1, size( varSys%Variable%val( iVar )%st_fun )
!      res = res + tem_spacetime_for(                                    &
!        &             me = varSys%Variable%val( iVar )%st_fun(iFun),    &
!        &             treeIDs = tmp_treeID, time = time, tree = tree,   &
!        &             n = varSys%Variable%val( iVar )%nComponents )
!    end do
!
!  end subroutine callSpaceTimeFun
!!******************************************************************************!
!
!!******************************************************************************!
! !> Return the solver state of a variable of a given elemPos
! !! 
! !! run over all the components of the variable and get the position of 
! !! variable in the global dynamic variable array and from that variable
! !! get varPos of direction and use this direction to access the currect field
! !! in the state array. Then, add the value res, which will be returned
!  subroutine access_stateVar( state, neigh, elemPos, fieldProp, varSys, &
!    &          globSys, layout, res, iVar, varDep, varMap, sub_index, sourceMap, &
!    &          treeID, time, tree )
!    !-------------------------------------------------------------------------
!    !> state vector all elements in a level with all fields
!    real(kind=rk), intent(in)    :: state(:)
!    !> neighbor array
!    integer, intent(in), optional :: neigh(:)
!    !> position of element in the state array
!    integer, intent(in) :: elemPos
!    !> requested variable system 
!    type( tem_var_system_type ), optional, intent(in) :: varSys 
!    !> global variable system 
!    type( tem_var_system_type ), intent(in) :: globSys
!    !> information about the field
!    type( mus_field_prop_type ), intent(in) :: fieldprop(:)
!    !> the scheme layout representation used in this scheme with stencil
!    type( mus_scheme_layout_type ), intent(in) :: layout 
!    !> calculated / derived quantities
!    real(kind=rk), intent(out)  :: res(:)
!    !> which variable of the re variable system to output? optional!
!    integer, optional, intent(in) :: iVar
!    !> array which points to dependend state variable of 
!    !! requested variable iVar at position varSys%mapping(iVar)
!    !! in the global variable system
!    integer, intent(in) :: varDep(:)
!    !> global list of all quantities
!    type(mus_varMap_type), optional, intent(inout) :: varMap(:)
!    !> an optional subset of the quantities to calculate
!    integer, optional, intent(in) :: sub_index(:)
!
!    !> the current treeID
!    integer(kind=long_k), optional, intent(in) :: treeID
!    !> information on the sources (including pointers to the lua functions)
!    type( tem_sourceMap_type ), optional, intent(in) :: sourceMap
!    !> current timing information
!    type(tem_time_type), optional, intent(in) :: time
!    !> tree used by spacetime function
!    type( treelmesh_type ), optional, intent(in) :: tree
!    !-------------------------------------------------------------------------
!    integer :: iDir, pos
!    !-------------------------------------------------------------------------
!    do iDir = 1, varSys%Variable%val( iVar )%nComponents
!      pos = globSys%Variable%val( varSys%variable%val( iVar )%mapping )%varPos( iDir )
!      res( iDir ) = state( ?IDX?( pos, elemPos, globSys%nScalars ) )
!    enddo
!
!   end subroutine access_stateVar
!!******************************************************************************!
!
!
!!******************************************************************************!
! !> Return the solver state of a variable of a given elemPos
! !! 
! !! run over all the components of the variable and get the position of 
! !! variable in the global dynamic variable array and from that variable
! !! get varPos of direction and use this direction to access the currect field
! !! in the state array. Then, add the value res, which will be returned
!  subroutine evaluateSources_single( state, neigh, elemPos, fieldProp, varSys, &
!    &          globSys, layout, res, iVar, varDep, varMap, sub_index, sourceMap, &
!    &          treeID, time, tree )
!    !-------------------------------------------------------------------------
!    !> state vector all elements in a level with all fields
!    real(kind=rk), intent(in)    :: state(:)
!    !> neighbor array
!    integer, intent(in), optional :: neigh(:)
!    !> position of element in the state array
!    integer, intent(in) :: elemPos
!    !> requested variable system 
!    type( tem_var_system_type ), optional, intent(in) :: varSys 
!    !> global variable system 
!    type( tem_var_system_type ), intent(in) :: globSys
!    !> information about the field
!    type( mus_field_prop_type ), intent(in) :: fieldprop(:)
!    !> the scheme layout representation used in this scheme with stencil
!    type( mus_scheme_layout_type ), intent(in) :: layout 
!    !> calculated / derived quantities
!    real(kind=rk), intent(out) :: res(:)
!    !> which variable of the re variable system to output? optional!
!    integer, optional, intent(in) :: iVar
!    !> array which points to dependend state variable of 
!    !! requested variable iVar at position varSys%mapping(iVar)
!    !! in the global variable system
!    integer, intent(in) :: varDep(:)
!    !> global list of all quantities
!    type(mus_varMap_type), optional, intent(inout) :: varMap(:)
!    !> an optional subset of the quantities to calculate
!    integer, optional, intent(in) :: sub_index(:)
!
!    !> the current treeID
!    integer(kind=long_k), optional, intent(in) :: treeID
!    !> information on the sources (including pointers to the lua functions)
!    type( tem_sourceMap_type ), optional, intent(in) :: sourceMap
!    !> current timing information
!    type(tem_time_type), optional, intent(in) :: time
!    !> tree used by spacetime function
!    type( treelmesh_type ), optional, intent(in) :: tree
!    !-------------------------------------------------------------------------
!    integer :: iSourceVar, pos, iRes
!    integer :: sourcePos
!    integer :: treeIDpos
!    integer :: iFun
!    real(kind=rk), allocatable :: luaRes(:)
!    integer(kind=long_k), allocatable :: local_treeID(:)
!    !-------------------------------------------------------------------------
!    allocate( local_treeID( 1 ))
!    allocate( luaRes( 1 ))
!    local_treeID( 1 ) = treeID
!
!    ! initialize result array with 0
!    res = 0
!    ! get the position of the requested variable in the global varSys
!    pos = varSys%variable%val( iVar )%mapping
!
!    ! get the positions in the list of source variables to catch all
!    ! sources of the same type (e.g. all injections)
!    ! loop over the source variables and check wether the treeID is
!    ! part of the different sources 
!    ! * if yes -> evaluate the function and add the result to res
!    ! * if no -> add 0
!
!    do iSourceVar = 1, sourceMap%source( pos )%map( sourceMap%map( pos ))%nVals 
!      do iFun = 1, size( sourceMap%source( pos )%sourceVar( sourceMap%source( pos )% &
!          &              map( sourceMap%map( pos ))%val( iSourceVar ))%st_fun )
!        ! check if the treeID is part of the tree of the current source variable
!        treeIDpos = tem_PosOfId( treeID, &
!          &     sourceMap%source( pos )%sourceVar(              &
!          &       sourceMap%source( pos )%map( sourceMap%map( pos ))%val( iSourceVar )) &
!          &     %st_fun(iFun)%tree%treeID )
! 
!        if(treeIDpos .gt. 0)then
!          luaRes = tem_spacetime_for(                                                      &
!            &    me =  sourceMap%source( pos )%sourceVar( sourceMap%source(                &
!            &            pos )%map( sourceMap%map( pos ))%val( iSourceVar ))%st_fun(iFun), & 
!            &    treeIDs = local_treeID,                                                   &
!            &    time = time,                                                              &
!            &    tree = sourceMap%source( pos )%sourceVar( sourceMap%source(               &
!            &       pos )%map( sourceMap%map( pos ))%val( iSourceVar ))%st_fun(iFun)%tree, &
!            &    n = 1 )
!          do iRes = 1, size(res)
!            res(iRes) = res(iRes) + luaRes(1)
!          end do
!        end if
!      end do
!    end do
!
!   end subroutine evaluateSources_single
!!******************************************************************************!
!
!!******************************************************************************!
!  !> Calculate the molefraction of a given element number with the given state 
!  !!        vector (sum up all values)
!  subroutine deriveMolfrac_forElemFromState( state, neigh, elemPos, fieldProp, &
!    & varSys, globSys, layout, res, iVar, varDep, varMap, sub_index , sourceMap, &
!    &          treeID, time, tree )
!    !------------------------------------------------------------------------ 
!    !> state vector all elements in a level with all fields
!    real(kind=rk), intent(in)    :: state(:)
!    !> neighbor array
!    integer, intent(in), optional :: neigh(:)
!    !> position of element in the state array
!    integer, intent(in) :: elemPos
!    !> requested variable system 
!    type( tem_var_system_type ), optional, intent(in) :: varSys 
!    !> global variable system 
!    type( tem_var_system_type ), intent(in) :: globSys
!    !> information about the field
!    type( mus_field_prop_type ), intent(in) :: fieldprop(:)
!    !> the scheme layout representation used in this scheme with stencil
!    type( mus_scheme_layout_type ), intent(in) :: layout 
!    !> calculated / derived quantities
!    real(kind=rk), intent(out)  :: res(:)
!    !> which variable of the re variable system to output? optional!
!    integer, optional, intent(in) :: iVar
!    !> array which points to dependend state variable of 
!    !! requested variable iVar at position varSys%mapping(iVar)
!    !! in the global variable system
!    integer, intent(in) :: varDep(:)
!    !> global list of all quantities
!    type(mus_varMap_type), optional, intent(inout) :: varMap(:)
!    !> an optional subset of the quantities to calculate
!    integer, optional, intent(in) :: sub_index(:)
!
!    !> the current treeID
!    integer(kind=long_k), optional, intent(in) :: treeID
!    !> information on the sources (including pointers to the lua functions)
!    type( tem_sourceMap_type ), optional, intent(in) :: sourceMap
!    !> current timing information
!    type(tem_time_type), optional, intent(in) :: time
!    !> tree used by spacetime function
!    type( treelmesh_type ), optional, intent(in) :: tree
!    !------------------------------------------------------------------------ 
!    !local variable
!    integer :: iDir, pos, iDep, iField
!    real(kind=rk) :: rho(size(fieldProp)), ptol
!    !------------------------------------------------------------------------ 
!    
!    res = 0._rk
!    rho = 0._rk
!    do iField = 1, size(fieldProp)
!      do iDir = 1, layout%stencil(1)%QQ
!        pos = globSys%Variable%val( iField )%varPos( iDir )
!        rho( iField ) = rho( iField ) &
!          & + state( ?IDX?( pos, elemPos, globSys%nScalars ) )
!      enddo
!    !partial pressure
!    !p_s = rho_s*mratio_s/3
!      rho( iField ) = rho( iField )*fieldProp( iField )%species%molWeigRatio/3.0_rk
!    enddo
!    !total partial pressure
!    ptol = sum(rho)
!   
!    do iDep = 1, size(varDep)
!      res( 1 ) = res( 1 ) + rho(varDep( iDep ))/ptol
!    end do
!  
!
!  end subroutine deriveMolfrac_forElemFromState
!!******************************************************************************!



!******************************************************************************!
  !> Calculate the density of a given subset of pdfs 
  !!        vector (sum up all values)
  !!
  pure function getDensity_forPdfSubset( subset, stencil, varPos ) result( res )
    !--------------------------------------------------------------------------- 
    type(tem_stencilHeader_type), intent(in) :: stencil
    real(kind=rk), intent(in) :: subset(:)
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    real(kind=rk)             :: res !< return value
    !--------------------------------------------------------------------------- 
    ! local variables
    integer :: iDir
    !--------------------------------------------------------------------------- 

    res = 0._rk
    do iDir = 1,stencil%QQ
      res = res + subset( varPos( iDir ) )
    enddo

  end function getDensity_forPdfSubset
!******************************************************************************!


!******************************************************************************!
  !> Calculate the density of a given element number with the given state 
  !!        vector (sum up all values)
  !!
  pure function getDensity_forElemFromStatePre( state, elem, stencil, varPos,  &
    &                                            nScalars, neigh ) result( res )
    !---------------------------------------------------------------------------
    type(tem_stencilHeader_type), intent(in) :: stencil
    real(kind=rk), intent(in) :: state(:)
    integer, intent(in)       :: elem
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    integer, intent(in) :: nScalars !< number of scalars in global system
    !> connectivity array corresponding to state vector
    integer,intent(in) :: neigh(:)       
    real(kind=rk)             :: res !< return value
    !---------------------------------------------------------------------------
    ! local variables
    integer :: iDir
    !---------------------------------------------------------------------------

    res = 0._rk
    do iDir = 1,stencil%QQ
      res = res + state( ?FETCH?( varPos(iDir), elem, nScalars, neigh ))
    enddo

  end function getDensity_forElemFromStatePre
!******************************************************************************!


!******************************************************************************!
  !> Calculate the density of a given element number with the given state 
  !!        vector (sum up all values)
  !!
  pure function getDensity_forElemFromState( state, elem, stencil, varPos,     &
    &                                        nScalars ) result( res )
    !--------------------------------------------------------------------------- 
    type(tem_stencilHeader_type), intent(in) :: stencil
    real(kind=rk), intent(in) :: state(:)
    integer, intent(in)       :: elem
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    integer, intent(in) :: nScalars !< number of scalars in global system
    real(kind=rk)             :: res !< return value
    !--------------------------------------------------------------------------- 
    ! local variables
    integer :: iDir
    !--------------------------------------------------------------------------- 

    res = 0._rk
    do iDir = 1,stencil%QQ
      res = res + state( ?IDX?( varPos(iDir), elem, nScalars ))
    enddo

  end function getDensity_forElemFromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the velocity in all only one direction specified with coordDir
  !!
  pure function getVelocityDir_noForce( coordDir, state, elem, stencil,        &
    &                                   varPos, nScalars ) result( vel )
    !--------------------------------------------------------------------------- 
    type(tem_stencilHeader_type), intent(in) :: stencil
    real(kind=rk), intent(in) :: state(:)
    integer, intent(in)       :: coordDir
    integer, intent(in)       :: elem
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    integer, intent(in) :: nScalars !< number of scalars in global system
    real(kind=rk)             :: vel !< return value
    !--------------------------------------------------------------------------- 
    ! local variables
    real(kind=rk) :: dens
    integer       :: iDir
    !--------------------------------------------------------------------------- 

    vel = 0._rk
    dens = 0._rk
    do iDir = 1,stencil%QQ
      vel = vel  + state( ?IDX?( varPos(iDir), elem, nScalars, ))  & 
              &   *real(stencil%cxDir( coordDir, iDir ), kind=rk) 
      dens= dens + state( ?IDX?( varPos(iDir), elem, nScalars, ))
    enddo
    vel = vel/dens
    
  end function getVelocityDir_noForce
!******************************************************************************!


!******************************************************************************!
  !> Calculate the velocity in all only one direction specified with coordDir
  !!
  pure function getVelocityDir_Force( coordDir, state, elem, fieldProp,        &
    &                pdfGlobal, level, stencil, varPos, nScalars ) result( vel )
    !--------------------------------------------------------------------------- 
    type( tem_stencilHeader_type ), intent(in) :: stencil
    type(mus_field_prop_type), intent(in) :: fieldProp   !< Global fluid 
    type( pdf_globInfo_type), intent(in) :: pdfGlobal !< global pdf info
    real(kind=rk), intent(in)    :: state(:)
    integer, intent(in)          :: level
    integer, intent(in)          :: coordDir
    integer, intent(in)          :: elem
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    integer, intent(in) :: nScalars !< number of scalars in global system
    real(kind=rk)                :: vel !< return value
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: dens
    integer       :: iDir
    !--------------------------------------------------------------------------- 

    vel = 0._rk
    dens = 0._rk
    do iDir = 1,stencil%QQ
      ! Forcing Term
      vel = vel  + state( ?IDX?( varPos(iDir), elem, nScalars, ))              &
        & * real(stencil%cxDir( coordDir, iDir ), kind=rk)                     &
        & + pdfGlobal%dtLB( level )*0.5_rk*fieldProp%fluid%force( coordDir )
      dens= dens + state( ?IDX?( varPos(iDir), elem, nScalars, ))
    enddo
    vel = vel/dens
    
  end function getVelocityDir_Force
!******************************************************************************!


!******************************************************************************!
  !> Calculate the velocity in all 3 directions 
  !!        from a subset given, ordered according to the stencil
  !!        
  pure function getVelocity_forPdfSubset( subset, stencil, varPos )            &
    &                             result( vel )
    !--------------------------------------------------------------------------- 
    type(tem_stencilHeader_type), intent(in) :: stencil !< stencil information
    real(kind=rk), intent(in) :: subset(:) !< complete state of one level
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    real(kind=rk)             :: vel(3)    !< return value 
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: dens
    integer       :: iDir
    !--------------------------------------------------------------------------- 

    vel = 0._rk
    dens = 0._rk
    do iDir = 1,stencil%QQ
      vel(:) = vel(:) + subset( varPos(iDir) ) &
           &   *real(stencil%cxDir( :, iDir ), kind=rk) 
      dens   = dens   + subset( varPos(iDir) )
    enddo
    vel = vel/dens
    
  end function getVelocity_forPdfSubset
!******************************************************************************!


!******************************************************************************!
  !> Calculate the velocity in all 3 directions  
  !! from the element indicated (elem), 
  !! reading the pdf (state information) from the state array.
  !! state array includes all the pdfs of all elements.
  !! The access to the state array has to be done via the generic
  !! access macro IDX, as we want to access post-collision values.
  !! herefore, we IDX assuming we use only PULL operation for interpolation
  !!
  pure function getVelocity_forElemFromState_noForcePre( state, elem, stencil, &
    &                                    varPos, nScalars, neigh ) result( vel )
    !--------------------------------------------------------------------------- 
    type(tem_stencilHeader_type), intent(in) :: stencil !< stencil information
    real(kind=rk), intent(in) :: state(:)  !< complete state of one level
    !> element index, for which to calc velocity
    integer, intent(in)       :: elem
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    integer, intent(in) :: nScalars !< number of scalars in global system
    !> connectivity array corresponding to state vector
    integer,intent(in) :: neigh(:)       
    real(kind=rk)             :: vel(3)    !< return value 
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: dens
    integer       :: iDir
    !--------------------------------------------------------------------------- 

    vel = 0._rk
    dens = 0._rk
    do iDir = 1,stencil%QQ
      vel(:) = vel(:) + state( ?FETCH?( varPos(iDir), elem, nScalars, neigh )) &
           &   *real(stencil%cxDir( :, iDir ), kind=rk) 
      dens   = dens   + state( ?FETCH?( varPos(iDir), elem, nScalars, neigh ))
    enddo
    vel = vel/dens
    
  end function getVelocity_forElemFromState_noForcePre
!******************************************************************************!


!******************************************************************************!
  !> Calculate the velocity in all 3 directions  
  !! from the element indicated (elem), 
  !! reading the pdf (state information) from the state array.
  !! state array includes all the pdfs of all elements.
  !! The access to the state array has to be done via the generic
  !! access macro IDX, as we want to access post-collision values.
  !! herefore, we IDX assuming we use only PULL operation for interpolation
  !!
  pure function getVelocity_forElemFromState_noForce( state, elem, stencil,    &
    &                                           varPos, nScalars ) result( vel )
    !--------------------------------------------------------------------------- 
    type(tem_stencilHeader_type), intent(in) :: stencil !< stencil information
    real(kind=rk), intent(in) :: state(:)  !< complete state of one level
    !> element index, for which to calc velocity
    integer, intent(in)       :: elem
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    integer, intent(in) :: nScalars !< number of scalars in global system
    real(kind=rk)             :: vel(3)    !< return value 
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: dens
    integer       :: iDir
    !--------------------------------------------------------------------------- 

    vel = 0._rk
    dens = 0._rk
    do iDir = 1,stencil%QQ
      vel(:) = vel(:) + state( ?IDX?( varPos(iDir), elem, nScalars, ))         &
        &    * real(stencil%cxDir( :, iDir ), kind=rk) 
      dens   = dens   + state( ?IDX?( varPos(iDir), elem, nScalars, ))
    enddo

    vel = vel/dens
    
  end function getVelocity_forElemFromState_noForce
!******************************************************************************!


!******************************************************************************!
  !> Calculate the velocity in all 3 directions 
  !!
  pure function getVelocity_forElemFromState_Force( state, elem, fieldProp,    &
    &                pdfGlobal, level, stencil, varPos, nScalars ) result( vel )
    !--------------------------------------------------------------------------- 
    real(kind=rk), intent(in)    :: state(:) !< state array of a level
    type(tem_stencilHeader_type), intent(in) :: stencil 
    type(mus_field_prop_type), intent(in) :: fieldProp   !< Global fluid 
    type( pdf_globInfo_type), intent(in) :: pdfGlobal !< global pdf info
    integer, intent(in)          :: level
    integer, intent(in)          :: elem
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    integer, intent(in) :: nScalars !< number of scalars in global system
    real(kind=rk)                :: vel(3) !< return value
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: dens
    integer       :: iDir
    !--------------------------------------------------------------------------- 

    vel = 0._rk
    dens = 0._rk
    do iDir = 1,stencil%QQ
      vel(:) = vel(:) + state( ?IDX?( varPos(iDir), elem, nScalars, ))         &
        &    * real(stencil%cxDir( :, iDir ), kind=rk)                         &
      ! Forcing Term
        &    + pdfGlobal%dtLB( level )*0.5_rk*fieldProp%fluid%force( : )

      dens   = dens   + state( ?IDX?( varPos(iDir), elem, nScalars, ))
    enddo
    vel = vel/dens
    
  end function getVelocity_forElemFromState_Force
!******************************************************************************!


!******************************************************************************!
  !> Calculate the momentum in all 3 directions 
  !!
  pure function getMomentum_forElemFromState_noForce( state, elem, stencil,    &
    &                                           varPos, nScalars ) result( mom )
    !--------------------------------------------------------------------------- 
    type(tem_stencilHeader_type), intent(in) :: stencil
    real(kind=rk), intent(in) :: state(:)
    integer, intent(in)       :: elem
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    integer, intent(in) :: nScalars !< number of scalars in global system
    real(kind=rk)             :: mom(3) !< return value
    !--------------------------------------------------------------------------- 
    integer :: iDir
    !--------------------------------------------------------------------------- 

    mom = 0._rk
    do iDir = 1,stencil%QQ
      mom(:) = mom(:) + state( ?IDX?( varPos(iDir), elem, nScalars, ))         &
        &    * real(stencil%cxDir( :, iDir ), kind=rk )
    enddo

  end function getMomentum_forElemFromState_noForce
!******************************************************************************!


!******************************************************************************!
  !> Calculate the momentum in all 3 directions with force 
  !!
  pure function getMomentum_forElemFromState_Force( state, elem, fieldProp,    &
    &                 pdfGlobal, level, stencil, varPos, nScalars) result( mom )
    !--------------------------------------------------------------------------- 
    type(tem_stencilHeader_type), intent(in) :: stencil
    type(mus_field_prop_type), intent(in) :: fieldProp   !< Global fluid 
    real(kind=rk), intent(in)    :: state(:)
    type( pdf_globInfo_type), intent(in) :: pdfGlobal !< global pdf info
    integer, intent(in)          :: level
    integer, intent(in)          :: elem
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    integer, intent(in) :: nScalars !< number of scalars in global system
    real(kind=rk) :: mom(3) !< return value
    !--------------------------------------------------------------------------- 
    integer :: iDir
    !--------------------------------------------------------------------------- 

    mom = 0._rk
    do iDir = 1,stencil%QQ
      mom(:) = mom(:) + state( ?IDX?( varPos(iDir), elem,nScalars, ))          &
        &    * real(stencil%cxDir( :, iDir ), kind=rk)                         &
      ! Forcing Term
        &    + pdfGlobal%dtLB( level )*0.5_rk*fieldProp%fluid%force( : )
    enddo

  end function getMomentum_forElemFromState_Force
!******************************************************************************!


!******************************************************************************!
  !> Calculate the momentum in all 3 directions  for a subset of pdfs
  !!        the subset is provided and must be of the size of DOFs per element
  !!
  pure function getMomentum_forPdfSubset( subset, stencil, varPos )            &
    &                                                              result( mom )
    !--------------------------------------------------------------------------- 
    real(kind=rk), intent(in)     :: subset(:)
    type(tem_stencilHeader_type), intent(in) :: stencil 
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    real(kind=rk) :: mom(3) !< return value
    !--------------------------------------------------------------------------- 
    integer :: iDir
    !--------------------------------------------------------------------------- 

    mom = 0._rk
    do iDir = 1,stencil%QQ
      mom(:) = mom(:) + subset( varPos( iDir ))                                &
        &             * real(stencil%cxDir( :, iDir ), kind=rk) 
    enddo

  end function getMomentum_forPdfSubset
!******************************************************************************!


!******************************************************************************!
  !> Calculate the momentum in all 3 directions  for a subset of pdfs
  !!        the subset is provided and must be of the size of DOFs per element
  !!
  pure function getMomentum_forPdfSubset_force( subset, fieldProp, pdfGlobal,  &
    &                                     level, stencil, varPos ) result( mom )
    !--------------------------------------------------------------------------- 
    type(mus_field_prop_type), intent(in) :: fieldProp   !< Global fluid
    type(tem_stencilHeader_type), intent(in) :: stencil 
    type( pdf_globInfo_type), intent(in) :: pdfGlobal !< global pdf info
    real(kind=rk), intent(in)     :: subset(:)
    integer, intent(in)           :: level
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    real(kind=rk) :: mom(3)!< return value
    !--------------------------------------------------------------------------- 
    integer :: iDir
    !------------------------------------------------------------------------ 

    mom = 0._rk
    do iDir = 1,stencil%QQ
      mom(:) = mom(:) + subset( varPos( iDir ))                                &
        &             * real(stencil%cxDir( :, iDir ), kind=rk)                &
      ! Forcing Term
      ! Forcing Term
         &     + pdfGlobal%dtLB( level )*0.5_rk*fieldProp%fluid%force( : )
    enddo

  end function getMomentum_forPdfSubset_force
!******************************************************************************!


!******************************************************************************!
  !> Get the equilibrium distribution in the specified direction iDir
  !!
  pure function getEqDistr( iDir, rho, vel, layout ) result( fEq )
    !--------------------------------------------------------------------------- 
    type(mus_scheme_layout_type), intent(in) :: layout !scheme layout
    real(kind=rk), intent(in) :: rho
    real(kind=rk), intent(in) :: vel(3)
    real(kind=rk) :: fEq!< return value
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: usq, ucx, cxReal(3)
    integer, intent(in) :: iDir
    !--------------------------------------------------------------------------- 

    cxReal=real(layout%stencil(1)%cxDir( :, iDir ), kind=rk)
    usq = vel(1)*vel(1) + vel(2)*vel(2) + vel(3)*vel(3)
    ucx = cxReal(1)*vel(1) + cxReal(2)*vel(2) + cxReal(3)*vel(3) 

    fEq = layout%weight( iDir )*rho*( 1.d0 +                                   &
      &   ucx*cs2inv                                                           &
      & + ucx*ucx*cs2inv*cs2inv*0.5d0                                          &
      & - usq*0.5d0*cs2inv )

  end function getEqDistr
!******************************************************************************!


!******************************************************************************!
  !> Calculate the equilibrium distribution function in all directions
  !!
  !! The equilibrium distribution function is:\n
  !! \f$ f^{eq}_i = w_i \rho ( 1 + \frac{\vec c_i \cdot \vec u}{c^2_s}
  !!                      + \frac{ {(\vec c_i \cdot \vec u)}^2}{2c^4_s}
  !!                      - \frac{\vec u \cdot \vec u}{2c^2_s}) \f$\n
  !! where \f$w_i\f$ is the weight in each direction,\n
  !! \f$\rho\f$ is the macroscopic value of density,\n
  !! \f$c_s\f$ is the speed of sound,\n
  !! \f$\vec c_i\f$ is the lattice unit velocity in each direction,\n
  !! \f$\vec u\f$ is the macroscopic value of velocity.
  !!
  pure function getEquilibrium_fromMacrVal( density, velocity, layout )        &
    &                                                            result( equil )
    !--------------------------------------------------------------------------- 
    type(mus_scheme_layout_type), intent(in) :: layout !scheme layout
    real(kind=rk), intent(in) :: density
    real(kind=rk), intent(in) :: velocity(3)
    real(kind=rk)             :: equil(layout%stencil(1)%QQ) !< return value
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: ucx, usq
    integer :: iDir
    !--------------------------------------------------------------------------- 

    ! square of velocity
    usq = velocity(1)*velocity(1) + velocity(2)*velocity(2)                    &
      & + velocity(3)*velocity(3)

    do iDir = 1, layout%stencil(1)%QQ

      ! velocity times lattice unit velocity
      ucx = real( layout%stencil(1)%cxDir( 1, iDir ), kind=rk)*velocity(1)     &
        & + real( layout%stencil(1)%cxDir( 2, iDir ), kind=rk)*velocity(2)     &
        & + real( layout%stencil(1)%cxDir( 3, iDir ), kind=rk)*velocity(3) 

      ! calculate equilibrium density
      equil( iDir ) = layout%weight( iDir ) * density * ( 1.d0 + ucx*cs2inv    &
        &           + ucx*ucx*cs2inv*cs2inv*0.5d0                              &
        &           - usq*cs2inv*0.5d0 )

    enddo

  end function getEquilibrium_fromMacrVal
!******************************************************************************!


!******************************************************************************!
  !> Calculate the equilibrium distribution function in all directions
  !!
  !! The equilibrim distribution function is:\n
  !! \f$ f^{eq}_i = w_i \rho ( 1 + \frac{\vec c_i \cdot \vec u}{c^2_s}
  !!                      + \frac{ {(\vec c_i \cdot \vec u)}^2}{2c^4_s}
  !!                      - \frac{\vec u \cdot \vec u}{2c^2_s}) \f$\n
  !! where \f$w_i\f$ is the weight in each direction,\n
  !! \f$\rho\f$ is the macroscopic value of density,\n
  !! \f$c_s\f$ is the speed of sound,\n
  !! \f$\vec c_i\f$ is the lattice unit velocity in each direction,\n
  !! \f$\vec u\f$ is the macroscopic value of velocity.
  !!
  pure function getEquilibrium_forPdfSubset( subset, layout, varPos )          &
    &                                                            result( equil )
    !--------------------------------------------------------------------------- 
    type(mus_scheme_layout_type), intent(in) :: layout !scheme layout
    real(kind=rk), intent(in) :: subset(:)  !< pdf array
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    real(kind=rk)             :: equil(layout%stencil(1)%QQ) !< return value
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: rho, vel(3)
    real(kind=rk) :: ucx, usq
    integer :: iDir
    !--------------------------------------------------------------------------- 

    rho = getDensity_forPdfSubset( subset, layout%stencil(1), varPos )
    vel(:) = getVelocity_forPdfSubset( subset, layout%stencil(1), varPos )
    ! square of velocity
    usq = vel(1)*vel(1) + vel(2)*vel(2) + vel(3)*vel(3)

    do iDir = 1, layout%stencil(1)%QQ

      ! velocity times lattice unit velocity
      ucx = real( layout%stencil(1)%cxDir( 1, iDir ), kind=rk)*vel(1)          &
        & + real( layout%stencil(1)%cxDir( 2, iDir ), kind=rk)*vel(2)          &
        & + real( layout%stencil(1)%cxDir( 3, iDir ), kind=rk)*vel(3) 

      ! calculate equilibrium density
      equil( iDir ) = layout%weight( iDir ) * rho * ( 1.d0 + ucx*cs2inv        &
        &           + ucx*ucx*cs2inv*cs2inv*0.5d0                              &
        &           - usq*cs2inv*0.5d0 )

    enddo

  end function getEquilibrium_forPdfSubset
!******************************************************************************!


!******************************************************************************!
  !> Calculate the equilibrium distribution function in all directions
  !!
  !! The equilibrim distribution function is:\n
  !! \f$ f^{eq}_i = w_i \rho ( 1 + \frac{\vec c_i \cdot \vec u}{c^2_s}
  !!                      + \frac{ {(\vec c_i \cdot \vec u)}^2}{2c^4_s}
  !!                      - \frac{\vec u \cdot \vec u}{2c^2_s}) \f$\n
  !! where \f$w_i\f$ is the weight in each direction,\n
  !! \f$\rho\f$ is the macroscopic value of density,\n
  !! \f$c_s\f$ is the speed of sound,\n
  !! \f$\vec c_i\f$ is the lattice unit velocity in each direction,\n
  !! \f$\vec u\f$ is the macroscopic value of velocity.
  !!
  pure function getEquilibrium_forElemfromState( state, elem, layout,          &
    &                                  varPos, nScalars, neigh ) result( equil )
    !--------------------------------------------------------------------------- 
    type(mus_scheme_layout_type), intent(in) :: layout !scheme layout
    real(kind=rk), intent(in) :: state(:)   !< pdf array
    integer, intent(in)       :: elem       !< treeID of the target element
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    integer, intent(in) :: nScalars !< number of scalars in global system
    integer, intent(in) :: neigh(:)   !< connectivity vector
    real(kind=rk)             :: equil(layout%stencil(1)%QQ) !< return value
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: rho, vel(3)
    real(kind=rk) :: ucx, usq
    integer :: iDir
    !--------------------------------------------------------------------------- 
    rho = getDensity( state, elem, layout%stencil(1), varPos, nScalars )
    vel(:) = getVelocity( state, elem, layout%stencil(1), varPos, nScalars )

    ! square of velocity
    usq = vel(1)*vel(1) + vel(2)*vel(2) + vel(3)*vel(3)

    do iDir = 1, layout%stencil(1)%QQ

      ! velocity times lattice unit velocity
      ucx = real( layout%stencil(1)%cxDir( 1, iDir ), kind=rk)*vel(1)          &
        & + real( layout%stencil(1)%cxDir( 2, iDir ), kind=rk)*vel(2)          &
        & + real( layout%stencil(1)%cxDir( 3, iDir ), kind=rk)*vel(3) 

      ! calculate equilibrium density
      equil( iDir ) = layout%weight( iDir ) * rho * ( 1.d0 + ucx*cs2inv        &
        &           + ucx*ucx*cs2inv*cs2inv*0.5d0                              &
        &           - usq*cs2inv*0.5d0 )

    enddo

  end function getEquilibrium_forElemfromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the viscous shear stress (exclude pressure)
  !! \author Manuel Hasert
  !! This function returns a one-dimensional array with 6 entries:
  !! tau(1:6) = [ Sxx, Syy, Szz, Sxy, Syz, Sxz ]
  !!
  !! The formula is:\n
  !! \f$ \tau_{\alpha \beta}=
  !! -(1-\frac{\omega}{2}) \sum_{i} f^{neq}_{i} c_{i\alpha} c_{i\beta} \f$\n
  !! where \f$ \tau_{\alpha \beta}\f$ is the stress
  !! in the \f$\beta\f$-direction on a face normal to the \f$\alpha\f$-axis,\n
  !! \f$ f^{neq}_i = f_i - f^{eq}_i\f$ is the non-equilibirium density.\n
  !! For more information, please refer to:\n
  !! Krueger T, Varnik F, Raabe D. Shear stress in lattice Boltzmann
  !! simulations. Physical Review E. 2009;79(4):1-14.
  !!
  pure function getShearStressTensorIncomp_forPdfSubset( subset, omega, layout,&
    &                                rho0 ) result( tau )
    !--------------------------------------------------------------------------- 
    type(mus_scheme_layout_type), intent(in) :: layout !scheme layout
    real(kind=rk), intent(in) :: subset(:)  !< pdf array
    real(kind=rk), intent(in) :: omega, rho0
    real(kind=rk)  :: tau(6) !< Shear stress includes xx, xy, xz, yy, yz, zz
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: equil(layout%stencil(1)%QQ), nonEquil, conv, density,vel(3)
    integer       :: iDir
    !--------------------------------------------------------------------------- 
    
    ! JQ: check if omega equals to one. if yes, shear stress can NOT compute.
    conv = convPrePost( omega )
    density = sum(subset)
    do iDir = 1, 3
      vel(iDir)  = sum(subset(:)*real(layout%stencil(1)%cxDir(iDir,:),kind=rk))
    end do
    ! Get the equilibrium density which is required by shear stress
    equil(:) = getEquilibriumIncomp( density, vel, layout, rho0 )

    tau = 0._rk
    
    do iDir = 1, layout%stencil(1)%QQ

      ! NonEquilibrium density should be computed from Pre-Collision pdf
      ! Either use FETCH here, or use post-coll and convert it to pre-coll.
      nonEquil = subset( iDir ) - equil( iDir )

      ! JQ: use post-collision and convert it to pre-collision
      nonEquil = nonEquil * conv

      ! tau_xx
      tau(1) = tau(1) + nonequil * real( layout%stencil(1)%cxDir( 1, iDir )    &
        &                        * layout%stencil(1)%cxDir( 1, iDir ), kind=rk)
      ! tau_yy
      tau(2) = tau(2) + nonequil * real( layout%stencil(1)%cxDir( 2, iDir )    &
        &                        * layout%stencil(1)%cxDir( 2, iDir ), kind=rk)
      ! tau_zz
      tau(3) = tau(3) + nonequil * real( layout%stencil(1)%cxDir( 3, iDir )    &
        &                        * layout%stencil(1)%cxDir( 3, iDir ), kind=rk) 
      ! tau_xy
      tau(4) = tau(4) + nonequil * real( layout%stencil(1)%cxDir( 1, iDir )    &
        &                        * layout%stencil(1)%cxDir( 2, iDir ), kind=rk)
      ! tau_yz
      tau(5) = tau(5) + nonequil * real( layout%stencil(1)%cxDir( 2, iDir )    &
        &                        * layout%stencil(1)%cxDir( 3, iDir ), kind=rk)  
      ! tau_xz
      tau(6) = tau(6) + nonequil * real( layout%stencil(1)%cxDir( 1, iDir )    &
        &                        * layout%stencil(1)%cxDir( 3, iDir ), kind=rk)

    enddo

    tau(:) = ( omega * .5_rk - 1._rk ) * tau(:)

  end function getShearstressTensorIncomp_forPdfSubset
!******************************************************************************!


!******************************************************************************!
  !> Calculate the viscous shear stress (exclude pressure)
  !! \author Manuel Hasert
  !! This function returns a one-dimensional array with 6 entries:
  !! tau(1:6) = [ Sxx, Syy, Szz, Sxy, Syz, Sxz ]
  !!
  !! The formula is:\n
  !! \f$ \tau_{\alpha \beta}=
  !! -(1-\frac{\omega}{2}) \sum_{i} f^{neq}_{i} c_{i\alpha} c_{i\beta} \f$\n
  !! where \f$ \tau_{\alpha \beta}\f$ is the stress
  !! in the \f$\beta\f$-direction on a face normal to the \f$\alpha\f$-axis,\n
  !! \f$ f^{neq}_i = f_i - f^{eq}_i\f$ is the non-equilibirium density.\n
  !! For more information, please refer to:\n
  !! Krueger T, Varnik F, Raabe D. Shear stress in lattice Boltzmann
  !! simulations. Physical Review E. 2009;79(4):1-14.
  !!
  function getShearStressTensor_forPdfSubset( subset, omega, layout,           &
    &                                                     varPos ) result( tau )
    !--------------------------------------------------------------------------- 
    type(mus_scheme_layout_type), intent(in) :: layout !scheme layout
    real(kind=rk), intent(in) :: subset(:)  !< pdf array
    real(kind=rk), intent(in) :: omega
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    real(kind=rk)  :: tau(6) !< Shear stress includes xx, xy, xz, yy, yz, zz
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: equil(layout%stencil(1)%QQ), nonEquil, conv
    integer       :: iDir
    !--------------------------------------------------------------------------- 
    
    ! JQ: check if omega equals to one. if yes, shear stress can NOT compute.
    conv = convPrePost( omega )

    ! Get the equilibrium density which is required by shear stress
    equil(:) = getEquilibrium( subset, layout, varPos)

    tau = 0._rk
    
    do iDir = 1, layout%stencil(1)%QQ

      ! NonEquilibrium density should be computed from Pre-Collision pdf
      ! Either use FETCH here, or use post-coll and convert it to pre-coll.
      nonEquil = subset( iDir ) - equil( iDir )

      ! JQ: use post-collision and convert it to pre-collision
      nonEquil = nonEquil * conv

      ! tau_xx
      tau(1) = tau(1) + nonequil * real( layout%stencil(1)%cxDir( 1, iDir )    &
        &                        * layout%stencil(1)%cxDir( 1, iDir ), kind=rk)
      ! tau_yy
      tau(2) = tau(2) + nonequil * real( layout%stencil(1)%cxDir( 2, iDir )    &
        &                        * layout%stencil(1)%cxDir( 2, iDir ), kind=rk)
      ! tau_zz
      tau(3) = tau(3) + nonequil * real( layout%stencil(1)%cxDir( 3, iDir )    &
        &                        * layout%stencil(1)%cxDir( 3, iDir ), kind=rk) 
      ! tau_xy
      tau(4) = tau(4) + nonequil * real( layout%stencil(1)%cxDir( 1, iDir )    &
        &                        * layout%stencil(1)%cxDir( 2, iDir ), kind=rk)
      ! tau_yz
      tau(5) = tau(5) + nonequil * real( layout%stencil(1)%cxDir( 2, iDir )    &
        &                        * layout%stencil(1)%cxDir( 3, iDir ), kind=rk)  
      ! tau_xz
      tau(6) = tau(6) + nonequil * real( layout%stencil(1)%cxDir( 1, iDir )    &
        &                        * layout%stencil(1)%cxDir( 3, iDir ), kind=rk)

    enddo

    tau(:) = ( omega * .5_rk - 1._rk ) * tau(:)

  end function getShearstressTensor_forPdfSubset
!******************************************************************************!


!******************************************************************************!
  !> Calculate the viscous shear stress (exclude pressure)
  !! \author Manuel Hasert
  !! This function returns a one-dimensional array with 6 entries:
  !! tau(1:6) = [ Sxx, Syy, Szz, Sxy, Syz, Sxz ]
  !!
  !! The formula is:\n
  !! \f$ \tau_{\alpha \beta}=
  !! -(1-\frac{\omega}{2}) \sum_{i} f^{neq}_{i} c_{i\alpha} c_{i\beta} \f$\n
  !! where \f$ \tau_{\alpha \beta}\f$ is the stress
  !! in the \f$\beta\f$-direction on a face normal to the \f$\alpha\f$-axis,\n
  !! \f$ f^{neq}_i = f_i - f^{eq}_i\f$ is the non-equilibirium density.\n
  !! For more information, please refer to:\n
  !! Krueger T, Varnik F, Raabe D. Shear stress in lattice Boltzmann
  !! simulations. Physical Review E. 2009;79(4):1-14.
  !!
  pure function getShearStressTensor_forElemFromState( state, neigh, elem,     &
    &                            omega, layout, varPos, nScalars ) result( tau )
    !--------------------------------------------------------------------------- 
    type(mus_scheme_layout_type), intent(in) :: layout !scheme layout
    real(kind=rk), intent(in) :: state(:)  !< pdf array
    real(kind=rk), intent(in) :: omega
    integer, intent(in) :: neigh(:)   !< connectivity vector
    integer, intent(in)       :: elem   !< treeID of target element
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    integer, intent(in) :: nScalars !< number of scalars in global system
    real(kind=rk)  :: tau(6) !< Shear stress includes xx, xy, xz, yy, yz, zz
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: equil(layout%stencil(1)%QQ), nonEquil, conv
    integer       :: iDir
    !--------------------------------------------------------------------------- 
    ! Get conversion factor for pre/post-collision 
    conv = convPrePost( omega )

    ! Get the equilibrium density which is required by shear stress
    equil(:) = getEquilibrium( state, elem, layout, varPos, nScalars, neigh )

    tau = 0._rk
    
    do iDir = 1, layout%stencil(1)%QQ

      ! NonEquilibrium density should be computed from Pre-Collision pdf
      ! Either use FETCH here, or use post-coll and convert it to pre-coll.
      nonEquil = state( ?SAVE?( varPos(iDir), elem, nScalars, neigh ) )        &
        &      - equil( iDir )

      ! JQ: use post-collision and convert it to pre-collision
      nonEquil = nonEquil * conv

      ! tau_xx
      tau(1) = tau(1) + nonequil * real( layout%stencil(1)%cxDir( 1, iDir )    &
        &                        * layout%stencil(1)%cxDir( 1, iDir ), kind=rk)
      ! tau_yy
      tau(2) = tau(2) + nonequil * real( layout%stencil(1)%cxDir( 2, iDir )    &
        &                        * layout%stencil(1)%cxDir( 2, iDir ), kind=rk)
      ! tau_zz
      tau(3) = tau(3) + nonequil * real( layout%stencil(1)%cxDir( 3, iDir )    &
        &                        * layout%stencil(1)%cxDir( 3, iDir ), kind=rk) 
      ! tau_xy
      tau(4) = tau(4) + nonequil * real( layout%stencil(1)%cxDir( 1, iDir )    &
        &                        * layout%stencil(1)%cxDir( 2, iDir ), kind=rk)
      ! tau_yz
      tau(5) = tau(5) + nonequil * real( layout%stencil(1)%cxDir( 2, iDir )    &
        &                        * layout%stencil(1)%cxDir( 3, iDir ), kind=rk)  
      ! tau_xz
      tau(6) = tau(6) + nonequil * real( layout%stencil(1)%cxDir( 1, iDir )    &
        &                        * layout%stencil(1)%cxDir( 3, iDir ), kind=rk)

    enddo

    tau(:) = ( omega * .5_rk - 1._rk ) * tau(:)

  end function getShearstressTensor_forElemFromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the viscous shear stress (exclude pressure)
  !! \author Manuel Hasert
  !! This function returns a one-dimensional array with 6 entries:
  !! tau(1:6) = [ Sxx, Syy, Szz, Sxy, Syz, Sxz ]
  !!
  !! The formula is:\n
  !! \f$ \tau_{\alpha \beta}=
  !! -(1-\frac{\omega}{2}) \sum_{i} f^{neq}_{i} c_{i\alpha} c_{i\beta} \f$\n
  !! where \f$ \tau_{\alpha \beta}\f$ is the stress
  !! in the \f$\beta\f$-direction on a face normal to the \f$\alpha\f$-axis,\n
  !! \f$ f^{neq}_i = f_i - f^{eq}_i\f$ is the non-equilibirium density.\n
  !! For more information, please refer to:\n
  !! Krueger T, Varnik F, Raabe D. Shear stress in lattice Boltzmann
  !! simulations. Physical Review E. 2009;79(4):1-14.
  !!
  pure function getShearRateTensor_acoustic_forPdfSubset( subset, omega,       &
    &                            layout, incompressible ) result( sourceS )
    !--------------------------------------------------------------------------- 
    type(mus_scheme_layout_type), intent(in) :: layout !scheme layout
    real(kind=rk), intent(in) :: subset(:)  !< pdf array
    real(kind=rk), intent(in) :: omega
    real(kind=rk) :: sourceS(3,3)
    logical, intent(in), optional :: incompressible
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: vel(3) ! velocity components
    real(kind=rk) :: diagVal, conv, rho, rho0, viscLB
    real(kind=rk) :: fTmp(layout%stencil(1)%QQ), rhoD
    integer       :: iDir, iVal, jVal, order(3)
    logical       :: incomp
    !--------------------------------------------------------------------------- 
    if( present( incompressible )) then
      incomp = incompressible
    else
      incomp = .false.
    end if
    conv = convPrePost( omega )

    rho0 = 1._rk
    Rho = 0._rk
    Vel = 0._rk
    do iDir = 1, layout%stencil(1)%QQ
      fTmp( iDir ) = subset( iDir )
      Rho = Rho + fTmp( iDir )
      vel(:) = vel(:) + fTmp( iDir )                                           &
        &             * real( layout%stencil(1)%cxDir(:,iDir),kind=rk)
    end do
    if( .not. incomp ) then
      vel = vel/rho
      rhoD = rho
    else
      rhoD = rho0
    end if

    ! Get the second velocity moments of the source element's pdf
    sourceS( :,: ) = 0._rk
    do jVal = 1, layout%stencil(1)%nDims
      do iVal = 1, layout%stencil(1)%nDims
        order = 0
        order( iVal ) = order( iVal) +1
        order( jVal ) = order( jVal) +1
        if( iVal == jVal ) then
          diagVal = cs2*Rho 
        else
          diagVal = 0._rk
        end if
        sourceS(iVal, jVal ) = get_moment( fTmp, layout, order )

        ! Now compute the shear stress from the second moment
        sourceS( iVal, jVal ) =  sourceS( iVal, jVal )                         &
          &                   - (rhoD*Vel(iVal )*Vel( jVal) + diagVal )
      end do
    end do
    viscLB = 1._rk/3._rk*(1._rk/omega -0.5_rk )
    ! Convert from post-to pre-collision value for the shear stress
    sourceS(:,:) = -( 1._rk - omega*0.5_rk )*sourceS(:,:)*conv/viscLB/rho0
    
  end function getShearRateTensor_acoustic_forPdfSubset
!******************************************************************************!


!******************************************************************************!
  !> Calculate the viscous shear stress (exclude pressure)
  !! \author Manuel Hasert
  !! This function returns a one-dimensional array with 6 entries:
  !! tau(1:6) = [ Sxx, Syy, Szz, Sxy, Syz, Sxz ]
  !!
  !! The formula is:\n
  !! \f$ \tau_{\alpha \beta}=
  !! -(1-\frac{\omega}{2}) \sum_{i} f^{neq}_{i} c_{i\alpha} c_{i\beta} \f$\n
  !! where \f$ \tau_{\alpha \beta}\f$ is the stress
  !! in the \f$\beta\f$-direction on a face normal to the \f$\alpha\f$-axis,\n
  !! \f$ f^{neq}_i = f_i - f^{eq}_i\f$ is the non-equilibirium density.\n
  !! For more information, please refer to:\n
  !! Krueger T, Varnik F, Raabe D. Shear stress in lattice Boltzmann
  !! simulations. Physical Review E. 2009;79(4):1-14.
  !!
  pure function getShearRateTensor_diffusive_forPdfSubset( subset, omega,      &
    &                                                 layout ) result( sourceS )
    !--------------------------------------------------------------------------- 
    !> scheme layout
    type(mus_scheme_layout_type), intent(in) :: layout
    !> pdf array
    real(kind=rk), intent(in) :: subset(:)
    !>
    real(kind=rk), intent(in) :: omega
    !>
    real(kind=rk) :: sourceS(3,3)
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: vel(3) ! velocity components
    real(kind=rk) :: diagVal, conv, rho
    real(kind=rk) :: fTmp(layout%stencil(1)%QQ)
    integer       :: iDir, iVal, jVal, order(3)
    !--------------------------------------------------------------------------- 
    conv = convPrePost( omega )

    Rho = 0._rk
    Vel = 0._rk
    do iDir = 1, layout%stencil(1)%QQ
      fTmp( iDir ) = subset( iDir )
      Rho = Rho + fTmp( iDir )
      vel(:) = vel(:) + fTmp( iDir )                                           &
        &             * real( layout%stencil(1)%cxDir(:,iDir),kind=rk)
    end do

    ! Get the second velocity moments of the source element's pdf
    sourceS( :,: ) = 0._rk
    do jVal = 1, layout%stencil(1)%nDims
      do iVal = 1, layout%stencil(1)%nDims
        order = 0
        order( iVal ) = order( iVal) +1
        order( jVal ) = order( jVal) +1
        if( iVal == jVal ) then
          diagVal = cs2*Rho
        else
          diagVal = 0._rk
        end if
        sourceS(iVal, jVal ) = get_moment( fTmp, layout, order )

        ! Now compute the shear stress from the second moment
        sourceS( iVal, jVal ) =                                                &
          &                     (Vel(iVal )*Vel( jVal)                         &
          &                      - sourceS( iVal, jVal )                       &
          &                      + diagVal                                     &
          &                     )
      end do
    end do
    sourceS = sourceS*omega*cs2inv
    ! Convert from post-to pre-collision value for the shear stress
    sourceS(:,:) = sourceS(:,:)*conv
    
  end function getShearRateTensor_diffusive_forPdfSubset
!******************************************************************************!


!******************************************************************************!
  !> Calculate the viscous shear stress (exclude pressure)
  !! \author Manuel Hasert
  !! This function returns a one-dimensional array with 6 entries:
  !! tau(1:6) = [ Sxx, Syy, Szz, Sxy, Syz, Sxz ]
  !!
  !! The formula to compute this can be found in Junk 2005 
  !! after A.15 for the second velocity moment Pi(2)
  !! \Pi(2) = u(1) x u(1) + cs^2*rho(2)*I - \nu *S(1)
  !! with the identity matrix I
  pure function getShearRateTensor_diffusive_forElemFromState(                 &
    &    state, neigh, elem, omega, layout, varPos, nScalars ) result( sourceS )
    !--------------------------------------------------------------------------- 
    type(mus_scheme_layout_type), intent(in) :: layout !scheme layout
    real(kind=rk), intent(in) :: state(:)  !< pdf array
    real(kind=rk), intent(in) :: omega
    integer, intent(in) :: neigh(:)   !< connectivity vector
    integer, intent(in)       :: elem   !< treeID of target element
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    integer, intent(in) :: nScalars !< number of scalars in global system
    real(kind=rk) :: sourceS(3,3)
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: vel(3) ! velocity components
    real(kind=rk) :: diagVal, conv, rho
    real(kind=rk) :: fTmp(layout%stencil(1)%QQ)
    integer       :: iDir, iVal, jVal, order(3)
    !--------------------------------------------------------------------------- 
    conv = convPrePost( omega )

    Rho = 0._rk
    Vel = 0._rk
    do iDir = 1, layout%stencil(1)%QQ
      fTmp( iDir ) = state( ?SAVE?( varPos(iDir), elem, nScalars, neigh ) ) 
      Rho = Rho + fTmp( iDir )
      Vel(:) = Vel(:) + fTmp( iDir )                                           &
        &             * real( layout%stencil(1)%cxDir(:,iDir),kind=rk)
    end do

    ! Get the second velocity moments of the source element's pdf
    sourceS( :,: ) = 0._rk
    do jVal = 1, layout%stencil(1)%nDims
      do iVal = 1, layout%stencil(1)%nDims
        order = 0
        order( iVal ) = order( iVal) +1
        order( jVal ) = order( jVal) +1
        if( iVal == jVal ) then
          diagVal = cs2*Rho
        else
          diagVal = 0._rk
        end if
        sourceS(iVal, jVal ) = get_moment( fTmp, layout, order )

        ! Now compute the shear stress from the second moment
        sourceS( iVal, jVal ) =                                                &
          &                     (Vel(iVal )*Vel( jVal)                         &
          &                      - sourceS( iVal, jVal )                       &
          &                      + diagVal                                     &
          &                     )
      end do
    end do
    sourceS = sourceS*omega*cs2inv
    ! Convert from post-to pre-collision value for the shear stress
    sourceS(:,:) = sourceS(:,:)*conv
    
  end function getShearRateTensor_diffusive_forElemFromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the viscous shear stress (exclude pressure)
  !! \author Jiaxing Qi
  !!
  !! The formula is:\n
  !! \f$ \tau_{\alpha \beta}=
  !! -(1-\frac{\omega}{2}) \sum_{i} f^{neq}_{i} c_{i\alpha} c_{i\beta} \f$\n
  !! where \f$ \tau_{\alpha \beta}\f$ is the stress
  !! in the \f$\beta\f$-direction on a face normal to the \f$\alpha\f$-axis,\n
  !! \f$ f^{neq}_i = f_i - f^{eq}_i\f$ is the non-equilibirium density.\n
  !! For more information, please refer to:\n
  !! Krueger T, Varnik F, Raabe D. Shear stress in lattice Boltzmann
  !! simulations. Physical Review E. 2009;79(4):1-14.
  !!
  function getShearstressPost( state, neigh, elem, omega, layout, varPos,      &
    &                          nScalars ) result( tau )
    !--------------------------------------------------------------------------- 
    type(mus_scheme_layout_type), intent(in) :: layout !scheme layout
    real(kind=rk), intent(in) :: state(:)  !< pdf array
    real(kind=rk), intent(in) :: omega
    integer, intent(in) :: neigh(:)   !< connectivity vector
    integer, intent(in) :: elem   !< treeID of target element
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    integer, intent(in) :: nScalars !< number of scalars in global system
    real(kind=rk)  :: tau(6) !< Shear stress includes xx, xy, xz, yy, yz, zz
    !--------------------------------------------------------------------------- 
    ! local variables
    real(kind=rk) :: equil(layout%stencil(1)%QQ), nonEquil
    real(kind=rk) :: rho, vel(3)
    integer       :: iDir
    !--------------------------------------------------------------------------- 
    rho = 0._rk
    vel = 0._rk
    do iDir = 1, layout%stencil(1)%QQ
      rho = rho + state( ?SAVE?( varPos(iDir), elem, nScalars, neigh ) )
      vel(:) = vel(:) + state( ?SAVE?( varPos(iDir), elem, nScalars, neigh ) ) &
        &             * real( layout%stencil(1)%cxDir(:, iDir) ,kind=rk)
    end do
    
    equil(:) = getEquilibrium( rho, vel, layout )
    ! Get the equilibrium density which is required by shear stress
!    equil(:) = getEquilibrium( state, elem, layout, varPos, nScalars, neigh )

    tau = 0._rk
    
    do iDir = 1, layout%stencil(1)%QQ

      ! NonEquilibrium density should be computed from Pre-Collision pdf
      ! Either use FETCH here, or use post-coll and convert it to pre-coll.
      nonEquil = state( ?SAVE?( varPos(iDir), elem, nScalars, neigh ) )        &
        &      - equil( iDir )
      ! tau_xx
      tau(1) = tau(1) + nonequil * real( layout%stencil(1)%cxDir( 1, iDir )    &
        &                        * layout%stencil(1)%cxDir( 1, iDir ), kind=rk)
      ! tau_yy
      tau(2) = tau(2) + nonequil * real( layout%stencil(1)%cxDir( 2, iDir )    &
        &                        * layout%stencil(1)%cxDir( 2, iDir ), kind=rk)
      ! tau_zz
      tau(3) = tau(3) + nonequil * real( layout%stencil(1)%cxDir( 3, iDir )    &
        &                        * layout%stencil(1)%cxDir( 3, iDir ), kind=rk)
      ! tau_xy
      tau(4) = tau(4) + nonequil * real( layout%stencil(1)%cxDir( 1, iDir )    &
        &                        * layout%stencil(1)%cxDir( 2, iDir ), kind=rk)
      ! tau_yz
      tau(5) = tau(5) + nonequil * real( layout%stencil(1)%cxDir( 2, iDir )    &
        &                        * layout%stencil(1)%cxDir( 3, iDir ), kind=rk)
      ! tau_xz
      tau(6) = tau(6) + nonequil * real( layout%stencil(1)%cxDir( 1, iDir )    &
        &                        * layout%stencil(1)%cxDir( 3, iDir ), kind=rk)

    enddo

    tau(:) = ( omega * .5_rk - 1._rk ) * tau(:)

  end function getShearStressPost
!******************************************************************************!


!******************************************************************************!
  !> Calculate the viscous shear stress (exclude pressure)
  !! \author Jiaxing Qi
  !!
  !! The formula is:\n
  !! \f$ \tau_{\alpha \beta}=
  !! -(1-\frac{\omega}{2}) \sum_{i} f^{neq}_{i} c_{i\alpha} c_{i\beta} \f$\n
  !! where \f$ \tau_{\alpha \beta}\f$ is the stress
  !! in the \f$\beta\f$-direction on a face normal to the \f$\alpha\f$-axis,\n
  !! \f$ f^{neq}_i = f_i - f^{eq}_i\f$ is the non-equilibirium density.\n
  !! For more information, please refer to:\n
  !! Krueger T, Varnik F, Raabe D. Shear stress in lattice Boltzmann
  !! simulations. Physical Review E. 2009;79(4):1-14.
  !!
  pure function getShearstress( state, neigh, elem, omega, layout, varPos,     &
    &                           nScalars ) result( tau )
    !--------------------------------------------------------------------------- 
    type(mus_scheme_layout_type), intent(in) :: layout !scheme layout
    real(kind=rk), intent(in) :: state(:)  !< pdf array
    real(kind=rk), intent(in) :: omega
    integer, intent(in) :: neigh(:)   !< connectivity vector
    integer, intent(in) :: elem   !< treeID of target element
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    integer, intent(in) :: nScalars !< number of scalars in global system
    real(kind=rk)  :: tau(6) !< Shear stress includes xx, xy, xz, yy, yz, zz
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: equil(layout%stencil(1)%QQ), nonEquil, conv
    integer       :: iDir
    !--------------------------------------------------------------------------- 
    conv = convPrePost( omega )
    
    ! Get the equilibrium density which is required by shear stress
    equil(:) = getEquilibrium( state, elem, layout, varPos, nScalars, neigh )

    tau = 0._rk
    
    do iDir = 1, layout%stencil(1)%QQ

      ! NonEquilibrium density should be computed from Pre-Collision pdf
      ! Either use FETCH here, or use post-coll and convert it to pre-coll.
      nonEquil = state( ?SAVE?( varPos(iDir), elem, nScalars, neigh ) )        &
        &      - equil( iDir )

      ! JQ: use post-collision and convert it to pre-collision
      nonEquil = nonEquil * conv

      ! tau_xx
      tau(1) = tau(1) + nonequil * real( layout%stencil(1)%cxDir( 1, iDir )    &
        &                        * layout%stencil(1)%cxDir( 1, iDir ), kind=rk)
      ! tau_yy
      tau(2) = tau(2) + nonequil * real( layout%stencil(1)%cxDir( 2, iDir )    &
        &                        * layout%stencil(1)%cxDir( 2, iDir ), kind=rk)
      ! tau_zz
      tau(3) = tau(3) + nonequil * real( layout%stencil(1)%cxDir( 3, iDir )    &
        &                        * layout%stencil(1)%cxDir( 3, iDir ), kind=rk)
      ! tau_xy
      tau(4) = tau(4) + nonequil * real( layout%stencil(1)%cxDir( 1, iDir )    &
        &                        * layout%stencil(1)%cxDir( 2, iDir ), kind=rk)
      ! tau_yz
      tau(5) = tau(5) + nonequil * real( layout%stencil(1)%cxDir( 2, iDir )    &
        &                        * layout%stencil(1)%cxDir( 3, iDir ), kind=rk)
      ! tau_xz
      tau(6) = tau(6) + nonequil * real( layout%stencil(1)%cxDir( 1, iDir )    &
        &                        * layout%stencil(1)%cxDir( 3, iDir ), kind=rk)

    enddo

    tau(:) = ( omega * .5_rk - 1._rk ) * tau(:)

  end function getShearstress
!******************************************************************************!


!******************************************************************************!
  !> Calculate the wall shear stress (WSS)
  !! \author: Jiaxing Qi
  !!
  !! \TODO: add formule and reference!
  function getWSS( state, neigh, elem, omega, layout, varPos, nScalars )       &
    &                                                             result( wss )
    !--------------------------------------------------------------------------- 
    type(mus_scheme_layout_type), intent(in) :: layout !scheme layout
    real(kind=rk), intent(in) :: state(:)
    integer, intent(in) :: neigh(:)   !< connectivity vector
    real(kind=rk), intent(in) :: omega
    integer, intent(in)       :: elem
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    integer, intent(in) :: nScalars !< number of scalars in global system
    real(kind=rk) :: wss !< return value
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: tau(6) ! shear stress components
    real(kind=rk) :: a0, a1, a2
    real(kind=rk) :: q_stress, q_stress_cube, q_term, r_stress, d_stress
    real(kind=rk) :: theta_stress, cos_term, shear_1, shear_2, shear_3
    !--------------------------------------------------------------------------- 
    wss = 0._rk
    tau(:) = getShearstress(state, neigh, elem, omega, layout, varPos, nScalars)
!    call tem_write('WARNING: This is still the old order of the shear stress!!')
!    call tem_write('PLease change in getWSS !')
    a2 = -1.d0 * ( tau(1) + tau(4) + tau(6) )
    a1 = tau(1) * tau(4) + tau(4) * tau(6) + tau(1) * tau(6)                   &
      &- tau(2) ** 2 - tau(3) ** 2 - tau(5) ** 2
    a0 = -1.d0 * ( tau(1)*tau(4)*tau(6) + 2.d0*tau(2)*tau(3)*tau(5)            &
      &- tau(1)*tau(5)**2 - tau(4)*tau(3)**2 - tau(6)*tau(2)**2 )

    q_stress = div1_9 * ( 3.d0*a1 - a2*a2 )
    if ( q_stress .lt. 0.d0 ) then
      q_term = 2.d0 * sqrt( -q_stress )
    else
      q_term = 0.d0
    end if
    r_stress = div1_54 * ( 9.d0*a2*a1 - 27.d0*a0 - 2.d0*a2*a2*a2 )
    q_stress_cube = q_stress * q_stress * q_stress
    d_stress = q_stress_cube + r_stress * r_stress

    if ( d_stress .lt. 0.d0 ) then
      ! the solutions are real and unequal
      theta_stress = div1_3 * acos( r_stress / sqrt( -q_stress_cube ))

      cos_term = div3_4 * q_term * cos( theta_stress )

      ! three values of maximum shear stress
      shear_1 = abs( sqrt3 * div1_2 * q_term * sin( theta_stress ))

      shear_2 = abs( cos_term - shear_1*div1_2 )

      shear_3 = abs( cos_term + shear_1*div1_2 )

      !! choose the max shear as Wall Shear Stress(wss)
      wss = max(shear_1, shear_2, shear_3)
    end if

  end function getWSS
!******************************************************************************!


!******************************************************************************!
  !> Calculate the wall shear stress (WSS)
  !! \author: Jiaxing Qi
  !!
  !! \TODO: add formule and reference!
  function getWSS_vonMises( state, neigh, elem, omega, layout, varPos,         &
    &                       nScalars ) result( wss )
    !--------------------------------------------------------------------------- 
    type(mus_scheme_layout_type), intent(in) :: layout !scheme layout
    real(kind=rk), intent(in) :: state(:)
    integer, intent(in) :: neigh(:)   !< connectivity vector
    real(kind=rk), intent(in) :: omega
    integer, intent(in)       :: elem
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    integer, intent(in) :: nScalars !< number of scalars in global system
    real(kind=rk) :: wss !< return value
    !---------------------------------------------------------------------------
    ! local variables
    real(kind=rk) :: tau(6) ! shear stress components
    !---------------------------------------------------------------------------
    wss = 0._rk
    tau(:) = getShearstressPost( state, neigh, elem, omega, layout, varPos,    &
      &                                                              nScalars )
    ! Check wikipedia article on von Mises stress criterion
    wss = (tau(1)-tau(2))**2 + (tau(2)-tau(3))**2 + (tau(1)-tau(3))**2         &
      & + 6._rk*( tau(4)*tau(4) + tau(5)*tau(5) + tau(6)*tau(6))
    wss = sqrt(0.5_rk*wss )

  end function getWSS_vonMises
!******************************************************************************!


!******************************************************************************!
  !> Calculate the wall shear stress (WSS) 2D
  !! \author: Jiaxing Qi
  !!
  !! \TODO: add formule and reference!
  function getWSS2D( state, neigh, elem, omega, layout, varPos, nScalars )     &
    &                                                             result( wss )
    !--------------------------------------------------------------------------- 
    type(mus_scheme_layout_type), intent(in) :: layout !scheme layout
    real(kind=rk), intent(in) :: state(:)
    integer, intent(in) :: neigh(:)   !< connectivity vector
    real(kind=rk), intent(in) :: omega
    integer, intent(in)       :: elem
    integer, intent(in) :: varPos(:) !< varPos of current field variable
    integer, intent(in) :: nScalars !< number of scalars in global system
    real(kind=rk) :: wss !< return value
    !---------------------------------------------------------------------------
    ! local variables
    real(kind=rk) :: tau(6) !< shear stress components
    real(kind=rk) :: a0, a1
    !---------------------------------------------------------------------------
    wss = 0._rk
    call tem_write('WARNING: This is still the old order of the shear stress!!')
    call tem_write('Please change in getWSS !')
    tau(:) = getShearstress(state, neigh, elem, omega, layout, varPos, nScalars)

    a1 = -1.0_rk * ( tau(1) + tau(4) )
    a0 = tau(1) * tau(4) - tau(2) * tau(2)

    wss = sqrt( a1 * a1 - 4._rk * a0 )

  end function getWSS2D
!******************************************************************************!


!******************************************************************************!
  !> Setting the non-equilibrium part based on the acoustic scaling
  !!
  !! According to [1] J. Latt, "Hydrodynamic limit of lattice Boltzmann
  !! equations," unige.ch, Jan. 2007.
  !! 
  !! here we set
  !! f = fEq + fNeq
  !! with the tensor contraction : as 
  !! fNeq = -t_i/(2*cs^4*(1-0.5*omega))*rho*nu*(grad(u) 
  !!      + grad(u)T):(Lambda)*t_i
  !! and
  !! Lambda = ( c_\alpha c_\beta  -cs^2 \delta_{\alpha\beta})
  !!
  !! Note that the viscosity has to equal the definition in the Junk paper
  !! visc = kappa*cs^2/omega
  !! The input is the strain rate tensor 
  !! S_\alpha\beta = (d_\alpha u_\beta + d_\beta u_alpha ) 
  !!               = sigma_{\alpha\beta}/viscosity
  function getNEq_acoustic( layout, omega, Sxx, rho0 ) result( nEq )
    !---------------------------------------------------------------------------
    type( mus_scheme_layout_type ), intent(in) :: layout
    real(kind=rk), intent(in) :: omega
    real(kind=rk), intent(in) :: rho0
    !> Stress tensor divided by viscosity:
    !! Contains the derivatives of the velocity
    !! S_\alpha\beta = (d_\alpha u_\beta + d_\beta u_alpha )
    real(kind=rk), intent(in) :: Sxx(3,3)
    real(kind=rk) :: nEq(layout%stencil(1)%QQ)
    !---------------------------------------------------------------------------
    integer :: iVal, jVal     !< current direction (discrete velocity) for loop
    integer :: iDir
    real(kind=rk) :: targetS(3,3)
    real(kind=rk) :: diagVal, conv
    real(kind=rk) :: viscosity
    !---------------------------------------------------------------------------
    ! Determine lattice viscosity from omega
    viscosity = cs2*(1._rk/omega - 0.5_rk )
    ! First fill the shear stress tensor from the linearized form Sxx
    ! Sxx = [ Sxx, Syy, Szz, Sxy, Syz, Sxz ]
    targetS(:,:) = Sxx(:,:)
    ! Recover the non-equilibrium part as
    nEq(:) = 0._rk
    do iDir = 1,layout%stencil(1)%QQ
      diagVal = 0._rk
      do jVal = 1, layout%stencil(1)%nDims
        do iVal = 1,  layout%stencil(1)%nDims
          Neq( iDir ) = Neq( iDir ) + targetS( iVal, jVal )                    &
          &     *real( layout%stencil(1)%cxDir( iVal,iDir)                     &
          &           *layout%stencil(1)%cxDir( jVal,iDir), kind=rk)
        end do                                                                   
        diagVal = diagVal +  targetS( jVal, jVal ) 
      end do
      ! Up to now we have Q:S
      ! but we still need nu*rho*Q:S
      Neq( iDir ) = viscosity*rho0*( Neq( iDir ) - cs2 * diagVal )
      ! Multiply the pre-factor                                                 
      Neq( iDir ) = -layout%weight( iDir )*cs4inv*0.5_rk*Neq( iDir )           &
        &           /(1._rk -0.5_rk*omega)
    end do
    conv = convPrePost( omega )
    ! Convert from post to pre-collision
    Neq = Neq/conv

  end function getNEq_acoustic
!******************************************************************************!


!******************************************************************************!
  !> Setting the non-equilibrium part based on the diffusive scaling
  !!
  !! This is done according to Junk 2005, Eq. (20)
  !! here we set
  !! f = fEq + fNeq
  !! with 
  !! fNeq = -1/(2kappa*cs^4)*nu*(grad(u) + grad(u)T):(Lambda)*t_i
  !! and
  !! Lambda = ( c_\alpha c_\beta  -1/nDims*sum_gamma( c_gamma*c_gamma )
  !!                                                      \delta_{\alpha\beta})
  !!
  !! Note that the viscosity has to equal the definition in the Junk paper
  !! visc = kappa*cs^2/omega
  !! The input is the strain rate tensor 
  !! S_\alpha\beta = (d_\alpha u_\beta + d_\beta u_alpha ) 
  !!               = sigma_{\alpha\beta}/viscosity
  !!
  function getNEq_diffusive( layout, omega, Sxx ) result( nEq )
    !---------------------------------------------------------------------------
    type( mus_scheme_layout_type ), intent(in) :: layout
    real(kind=rk), intent(in) :: omega
    !> Stress tensor divided by viscosity:
    !! Contains the derivatives of the velocity
    !! S_\alpha\beta = (d_\alpha u_\beta + d_\beta u_alpha )
    real(kind=rk), intent(in) :: Sxx(3,3)
    real(kind=rk) :: nEq(layout%stencil(1)%QQ)
    !---------------------------------------------------------------------------
    integer :: iVal, jVal     !< current direction (discrete velocity) for loop
    integer :: iDir, cx2sum
    real(kind=rk) :: targetS(3,3)
    real(kind=rk) :: diagVal, conv
    !---------------------------------------------------------------------------
    ! WARNING: MH: 
    ! This routine was duplicated to the interpolation routines!!
    ! Please make sure to adapt any changes there!
    ! Duplication reason: no function inlining for out-of-module routines with
    ! gfortran.
    !
    ! First fill the shear stress tensor from the linearized form Sxx
    ! Sxx = [ Sxx, Syy, Szz, Sxy, Syz, Sxz ]
    targetS = Sxx
    ! Recover the non-equilibrium part as
    ! f_neq = -t_i/cs^2/omega*S(1):Lambda
    nEq(:) = 0._rk
    do iDir = 1,layout%stencil(1)%QQ
      diagVal = 0._rk
      cx2Sum = 0
      do iVal = 1, 3
        cx2Sum = cx2Sum + layout%stencil(1)%cxDir( iVal,iDir)**2
      end do
      do jVal = 1, layout%stencil(1)%nDims
        do iVal = 1,  layout%stencil(1)%nDims
          Neq( iDir ) = Neq( iDir ) + targetS( iVal, jVal )                    &
          &     *real( layout%stencil(1)%cxDir( iVal,iDir)                     &
          &           *layout%stencil(1)%cxDir( jVal,iDir), kind=rk)                 
        end do                                                                   
        diagVal = diagVal + targetS( jVal, jVal )                              &
          &     *real(cx2sum, kind=rk)     
      end do
      ! The diagonal entries enforce Tr(S) = 0. with div(u) = 0 for the
      ! incompressible NSE resulting from the diffusive scaling
      Neq( iDir ) = Neq( iDir )                                                &
        &     -1._rk/real( layout%stencil(1)%nDims, kind=rk)                   &
        &     *diagVal                                                          
      ! Multiply the pre-factor                                                 
      Neq( iDir ) = -layout%weight( iDir )*cs4inv*0.5_rk*Neq( iDir )           &
        &           /omega*cs2 !*visc; visc = kappa*cs^2/omega
    end do

    conv =  convPrePost( omega )
    ! Convert from post to pre-collision
     Neq = Neq/conv

  end function getNEq_diffusive
!******************************************************************************!


!******************************************************************************!
  !> Set third order for diffusive scaling (for push algorithm?!)
  !! Have a look at the initial condition file for TGV in
  !! Hasert/2013/multilevel_diffusive/ @make ic
  !! Note that this is only for a 2d taylor green test case in the region 
  !! x = y = 0:2pi
  !! with reference velocity u0 = 1
  !!
  function setThirdOrder_diffusive( layout, omega, x, dx, dt ) result( f3 )
    !---------------------------------------------------------------------------
    type( mus_scheme_layout_type ), intent(in) :: layout
    real(kind=rk), intent(in) :: omega
    real(kind=rk), intent(in) :: x(3), dx, dt
    real(kind=rk) :: f3(layout%stencil(1)%QQ)
    !---------------------------------------------------------------------------
    integer :: iDir
    real(kind=rk) :: cx(3), nu, u0
    real(kind=rk) :: A, BI, BII, CI, CII, D
    !---------------------------------------------------------------------------
    ! Reference velocity
    u0 = 1._rk
    ! A term
    nu = 1._rk/3._rk*(1._rk/omega - 0.5_rk)   ! Viscosity in LB
    do iDir = 1, layout%stencil(1)%QQ
      cx = real(layout%stencil(1)%cxDir(:,iDir),kind=rk)
      ! k = 1
      A = u0**2*nu*cs2inv*cs2inv*0.5_rk*(cx(1)*sin(2._rk*x(1))                 &
                                        +cx(2)*sin(2._rk*x(2)))*(dt**3/dx**3)
      BI  = 6._rk*nu/omega*u0*(-cx(1)*sin(x(1))*cos(x(2))+cx(2)*cos(x(1))      &
        & * sin(x(2)))*(dx*dt)
      BII = 3._rk/omega*u0**2*(cx(1)*cos(x(1))*sin(x(1))+cx(2)*cos(x(2))       &
        & * sin(x(2))) &
           *(dt**2/dx)
      CI  = -9._rk/omega*u0**2*(cx(1)**2 - cx(2)**2)                           &
        & * (cx(1)*sin(x(1))*cos(x(1))*(cos(x(2)))**2                          &
        & - cx(2)*sin(x(2))*cos(x(2))*(cos(x(1)))**2 )*(dt**2/dx)
      CII = 3._rk/omega*u0**2*( cx(1)*sin(x(1))*cos(x(1))*(cos(x(2))**2        &
        & - sin(x(2))**2) + cx(2)*sin(x(2))*cos(x(2))*(cos(x(1))**2            &
        & - sin(x(1))**2))*(dt**2/dx)
      D   = cs2inv*(1._rk/(omega**2)-0.5_rk/omega)*u0*( sin(x(1))*cos(x(2))    &
        & * cx(1) + cos(x(1))*sin(x(2))*cx(2))*(cx(2)**2 - cx(1)**2)*(dt*dx)

      f3(iDir ) = layout%weight( iDir )*(A + BI + BII + CI + CII + D)
      ! A, BI: error smaller, BII, CI: error larger-, CII: error larger+, 
      ! D: error ! little larger-
      !f3(iDir ) = layout%weight( iDir )*(A)
    end do
    ! ... and convert for pre/post collision correction
    f3 = f3/convPrePost( omega )

  end function setThirdOrder_diffusive
!******************************************************************************!


!******************************************************************************!
  !> Conversion factor betwen the pre- and post-collision quantity for the shear
  !! stress
  !!
  !! How to use this pre-factor?
  !!  shearPhys   = convPrePost(omega)*omega*cs2inv*shearLB_postColl
  !! WARNING: Duplicated this routine to the interpolation modules, because 
  !! Gfortran could not inline functions from other modules
  !! -> performance impact.
  !!
  pure function convPrePost( omega ) result( conv )
    !---------------------------------------------------------------------------
    !>
    real(kind=rk), intent(in) :: omega
    !>
    real(kind=rk) :: conv
    !---------------------------------------------------------------------------
??  if( STREAM_PUSH ) then 
      conv = 1._rk
??  else
      if( omega /= 1._rk ) then
        conv = 1._rk / ( 1._rk - omega ) ! conversion factor
      else
        conv = 0._rk
      endif 
    !conv = 1._rk
    !call tem_write('deactivated the conversion of the shear stress in neq_diffusive')
??  endif 
  end function convPrePost
!******************************************************************************!


!******************************************************************************!
  !> Calculate the equilibrium distribution function in all directions
  !! This is the incompressible formulation with reference density rho0
  !!
  !! The equilibrium distribution function is:\n
  !! \f$ f^{eq}_i = w_i  ( \rho + \frac{\vec c_i \cdot \vec u}{c^2_s}
  !!                      + \frac{ {(\vec c_i \cdot \vec u)}^2}{2c^4_s}
  !!                      - \frac{\vec u \cdot \vec u}{2c^2_s}) \f$\n
  !! where \f$w_i\f$ is the weight in each direction,\n
  !! \f$\rho = \sum_i f_i\f$ is the macroscopic density,\n
  !! \f$c_s\f$ is the speed of sound,\n
  !! \f$\vec c_i\f$ is the lattice unit velocity in each direction,\n
  !! \f$\vec u = \sum_i c_i f_i\f$ is the macroscopic value of velocity.
  !!
  pure function getEquilibriumIncomp( density, velocity, layout, rho0 )        &
    &result( equil )
    !---------------------------------------------------------------------------
    type(mus_scheme_layout_type), intent(in) :: layout !scheme layout
    real(kind=rk), intent(in) :: density
    real(kind=rk), intent(in) :: rho0
    real(kind=rk), intent(in) :: velocity(3)
    real(kind=rk)             :: equil(layout%stencil(1)%QQ) !< return value
    !---------------------------------------------------------------------------
    ! local variables
    real(kind=rk) :: ucx, usq
    integer :: iDir
    !---------------------------------------------------------------------------

    ! square of velocity
    usq = velocity(1)*velocity(1)                                              &
      & + velocity(2)*velocity(2)                                              &
      & + velocity(3)*velocity(3)
    !usq = sum( velocity(:)**2 )

    do iDir = 1, layout%stencil(1)%QQ

      ! velocity times lattice unit velocity
      ucx = real( layout%stencil(1)%cxDir( 1, iDir ), kind=rk)*velocity(1)     &
        & + real( layout%stencil(1)%cxDir( 2, iDir ), kind=rk)*velocity(2)     &
        & + real( layout%stencil(1)%cxDir( 3, iDir ), kind=rk)*velocity(3) 

      ! calculate equilibrium density
      equil( iDir ) = layout%weight( iDir )                                    &
        &           *( density +  rho0*( ucx*cs2inv                            &
        &           + ucx*ucx*cs2inv*cs2inv*0.5d0                              &
        &           - usq*cs2inv*0.5d0 ))

    enddo

  end function getEquilibriumIncomp
!******************************************************************************!


!******************************************************************************!
  !> Calculate the distribution function in all directions
  !! by using the fEq + fNeq
  !!
  !! The input vector must look like
  !! 2D:
  !! mom = [ rho, ux, uy, Sxx, Syy, Sxy ]
  !! mom = [ rho, ux, uy, uz, Sxx, Syy, Szz, Sxy, Syz, Sxz ]
  !!
  function set_pdfDiffusive( layout, omega, rho0, mom )                        &
    &                                                             result( pdf )
    !--------------------------------------------------------------------------- 
    type(mus_scheme_layout_type), intent(in) :: layout !scheme layout
    real(kind=rk), intent(in) :: mom(layout%stencil(1)%QQ) !< input moments
    real(kind=rk), intent(in) :: omega
    real(kind=rk), intent(in) :: rho0
    real(kind=rk)             :: pdf(layout%stencil(1)%QQ) !< return value
    !---------------------------------------------------------------------------
    real(kind=rk) :: fEq(layout%stencil(1)%QQ)  ! eq distribution
    real(kind=rk) :: fnEq(layout%stencil(1)%QQ) ! neq distribution
    !--------------------------------------------------------------------------- 
    ! Compute the equilibrium
    if( layout%stencil(1)%nDims .eq. 3 ) then
      fEq(:) = getEquilibriumIncomp( density = mom(1),                         &
        &     velocity = mom(2:4),                                             &
        &     rho0     = rho0, layout = layout )                             
      ! Compute the non-equilibrium
      fNeq = getNeq_diffusive( layout = layout,                                &
        &  omega = omega, Sxx = [                                              &
        &               mom(5), mom(8), mom(10),                               & 
        &               mom(8), mom(6), mom(9),                                & 
        &               mom(10), mom(9), mom(7) ] )
    else
      fEq(:) = getEquilibriumIncomp( density = mom(1),                         &
        &     velocity = [ mom(2), mom(3), 0._rk ],                            &
        &     rho0 = rho0, layout = layout )                             
      fNeq = getNeq_diffusive( layout = layout,                                &
        &  omega = omega,                                                      &
        &  Sxx = [ mom(4), mom(6), 0._rk,                                      &
        &          mom(6), mom(5), 0._rk,                                      &
        &          0._rk,  0._rk,  0._rk ] )
    end if
    pdf(:) = fEq(:) + fNeq(:)

  end function set_pdfDiffusive
!******************************************************************************!


!******************************************************************************!
  !> Calculate the distribution function in all directions
  !! by using the fEq + fNeq
  !!
  !! The input vector must look like
  !! 2D:
  !! mom = [ rho, ux, uy, Sxx, Syy, Sxy ]
  !! mom = [ rho, ux, uy, uz, Sxx, Syy, Szz, Sxy, Syz, Sxz ]
  !!
  function set_pdfAcoustic( layout, omega, rho0, mom, incompressible )         &
    &                                                             result( pdf )
    !---------------------------------------------------------------------------
    type(mus_scheme_layout_type), intent(in) :: layout !scheme layout
    real(kind=rk), intent(in) :: mom(layout%stencil(1)%QQ) !< input moments
    real(kind=rk), intent(in) :: omega
    real(kind=rk), intent(in) :: rho0
    real(kind=rk)             :: pdf(layout%stencil(1)%QQ) !< return value
    logical, intent(in), optional :: incompressible
    !---------------------------------------------------------------------------
    real(kind=rk) :: fEq(layout%stencil(1)%QQ)  !< eq distribution
    real(kind=rk) :: fnEq(layout%stencil(1)%QQ) !< neq distribution
    real(kind=rk) :: moms(layout%stencil(1)%QQ) !< neq distribution
    logical :: incomp
    !--------------------------------------------------------------------------- 
    ! Compute the equilibrium
    if( present( incompressible )) then
      incomp = incompressible
    else
      incomp = .false.
    end if

    moms = mom
    if( layout%stencil(1)%nDims .eq. 2 ) then
      ! Set for 2d problems the z-velocity to zero
      moms(4) = 0._rk
    end if
    if( incomp ) then
      fEq(:) = getEquilibriumIncomp( density = mom(1),                         &
        &     velocity = moms(2:4),                                            &
        &     rho0     = rho0, layout = layout )                             
    else
      fEq(:) = getEquilibrium( density = mom(1),                               &
        &     velocity = moms(2:4),                                            &
        &     layout = layout )                             
    end if
    if( layout%stencil(1)%nDims .eq. 3 ) then
      ! Compute the non-equilibrium
      fNeq = getNeq_acoustic( layout = layout,                                 &
        &  omega = omega, Sxx = [                                              &
        &               mom(5), mom(8), mom(10),                               & 
        &               mom(8), mom(6), mom(9),                                & 
        &               mom(10), mom(9), mom(7) ], rho0 =rho0 )
    else
      fNeq = getNeq_acoustic( layout = layout,                                 &
        &  omega = omega,                                                      &
        &  Sxx = [ mom(4), mom(6), 0._rk,                                      &
        &          mom(6), mom(5), 0._rk,                                      &
        &          0._rk,  0._rk,  0._rk ], rho0 = rho0 )
    end if

    pdf(:) = fEq(:) + fNeq(:)

  end function set_pdfAcoustic
!******************************************************************************!


end module mus_derivedQuantities_module2
!******************************************************************************!
