! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module contains information about all fields like fluid, 
!! species, temperature etc.
!!
!! \author Kannan Masilamani, Simon Zimny
!!
!! [mus_field_prop]: @ref mus_field_prop_module "mus_field_prop_module"
!!
module mus_field_module

  ! include treelm modules
  use env_module,               only: labelLen
  use treelmesh_module,         only: treelmesh_type
  use tem_bc_prop_module,       only: tem_bc_prop_type
  use tem_tools_module,         only: tem_write, tem_horizontalSpacer
  use tem_aux_module,           only: tem_abort, tem_CheckLabel
  use tem_ini_condition_module, only: tem_ini_condition_type, tem_load_ic
  use tem_source_module,        only: tem_source_type, tem_load_source
  use tem_restart_module,       only: tem_restart_type

  ! include aotus modules
  use aotus_module,     only: flu_State, aot_get_val, aoterr_NonExistent
  use aot_table_module, only: aot_table_open, aot_table_close, aot_table_length
  use aot_path_module,  only: aot_path_type, assignment(=), aot_path_addNode
  use aot_out_module,   only: aot_out_type, aot_out_open_table,                &
    &                         aot_out_close_table, aot_out_val

  ! include musubi modules
  use mus_bc_header_module,     only: boundary_type, mus_load_bc
  use mus_scheme_header_module, only: mus_scheme_header_type
  use mus_field_prop_module,    only: mus_field_prop_type, mus_field_prop_out, &
    &                                 mus_load_field_prop
  use mus_pdf_module,           only: pdf_globInfo_type
  use mus_physics_module,       only: mus_physics_type

  implicit none

  private

  public :: mus_field_type
  public :: mus_load_fields, mus_field_out

  !> This type contains all information on fields with ic and bc
  !! Example fields: fluid, species etc.
  !! Each field contains one initial condition and array of 
  !! boundary conditions
  !!
  type mus_field_type
    !> field label. Should be unique for each field
    character(len=labelLen) :: label
    !> lbm parameters (fluid and species) for field
    type( mus_field_prop_type ) :: fieldProp
    !> array of fieldboundary types for each field
    type( boundary_type ), allocatable ::bc(:)
    !> initialization case, one initial condition for each field
    type( tem_ini_condition_type ) :: ic  
    logical :: setThirdOrder = .false.
    !> An instance of restart type 
    type( tem_restart_type ) :: restart
  end type mus_field_type

  !> Interface for dumping a single field or a set of fields in a file in lua 
  !! format.
  !!
  interface mus_field_out
    module procedure mus_field_out_vec
    module procedure mus_field_out_scal
  end interface mus_field_out
   
contains

!******************************************************************************!
  !> Subroutine to load the field table from the lua configuration file.
  !!
  !! If field table is not defined than load bc, ic, fluid, species from scheme
  !! table. If scheme table is not defined than load field variables from
  !! config parent.
  !!
  subroutine mus_load_fields( me, nFields, boundary, conf, path,      &
    &                         parent, pdfGlobal, schemeHeader, source,         &
    &                         physics )  
    !--------------------------------------------------------------------------
    !> array of field type
    type( mus_field_type ), allocatable, intent(out) :: me(:)
    integer, intent(out) :: nFields !< number of fields defined in lua file
    type( tem_bc_prop_type ), intent(in):: boundary !< boundary data from mesh
    type( flu_State ), intent(inout) :: conf !< flu state
    type( aot_path_type ),intent(in) :: path !< lua path
    type( pdf_globInfo_type) :: pdfGlobal !< global pdf info
    !> identifier of the scheme
    type( mus_scheme_header_type ), intent(in) :: schemeHeader
    !> parent handle if scheme table is defined
    integer, intent(in), optional :: parent
    !> array of source information
    type( tem_source_type ), allocatable, intent(out) :: source(:)
    !> physics type to convert physics to lattice unit or vice versa
    type( mus_physics_type ), intent(in) :: physics
    !---------------------------------------------------------------------------
    ! counter variables
    integer :: iField, iError 
    ! aotus handles
    integer :: field_handle, field_sub_handle
    ! aotus paths for field and field sub table
    type( aot_path_type ) :: fields_path, field_sub_path 
    !---------------------------------------------------------------------------

    call tem_horizontalSpacer()
    call tem_write( 'Loading the fields ...' )
    fields_path = path
    if(present(parent)) then
      call aot_table_open( L       =conf,                                      &
        &                  parent  = parent,                                   &
        &                  thandle = field_handle,                             &
        &                  key     = 'field' ) 
    else
      call aot_table_open( L       = conf,                                     &
        &                  thandle = field_handle,                             &
        &                  key     = 'field' )
    endif
  
    !get field path
    call aot_path_addNode( me       = fields_path,                             &
      &                    NodeType = 'table',                                 &
      &                    key      = 'field' ) 
    
    ! allocate the fields
    nFields = 1
    allocate( me( nFields ) )
    me(1)%label = ''

    ! allocate the sources (size: nFields + 1, for the entire scheme)
    allocate( source( nFields + 1 ))

    if ( field_handle .eq. 0 ) then
      if (present(parent)) then
        ! field table is not defined load field variable from scheme 
        call tem_write( 'No field table defined.'//                            &
          &             'Loading field variable from scheme' )
  
        call mus_load_field_single( me = me(1),                                &
          &                         nFields = nFields,                         &
          &                         boundary = boundary,                       &
          &                         conf = conf,                               &
          &                         path = path,                               &
          &                         pdfGlobal = pdfGlobal,                     &
          &                         parent = parent,                           &
          &                         source = source( 1 ),                      &
          &                         physics = physics,                         &
          &                         schemeHeader = schemeHeader )

      ! field table is not defined load field variable from config 
      else
        call tem_write( 'No field table and scheme table are defined.'//       &
          &             'Loading field variable from config' )
  
        call mus_load_field_single( me = me(1),                                &
          &                         nFields = nFields,                         &
          &                         boundary = boundary,                       &
          &                         conf = conf,                               &
          &                         path = path,                               &
          &                         pdfGlobal = pdfGlobal,                     &
          &                         source = source( 1 ),                      &
          &                         physics = physics,                         &
          &                         schemeHeader = schemeHeader )
      endif ! present parent?
    else ! check whether field is a single table or multiple table

      call aot_table_open( L       = conf,                                     &
        &                  parent  = field_handle,                             & 
        &                  thandle = field_sub_handle,                         &
        &                  pos     = 1 )
      if ( field_sub_handle .eq. 0 ) then
      ! field is a single table
        call aot_table_close( L=conf, thandle=field_sub_handle )
        call tem_write( 'Field is a single table' )

        call aot_get_val( L       = conf,                                      &
          &               thandle = field_handle,                              &
          &               key     = 'label',                                   &
          &               val     = me(1)%label,                               &
          &               default = '',                                        &
          &               ErrCode = iError )
        if( trim(me(1)%label) .ne. '')                                         &
          & me(1)%label = trim( me(1)%label )//'_'
        call mus_load_field_single( me = me(1),                                &
          &                         nFields = nFields,                         &
          &                         boundary = boundary,                       &
          &                         conf = conf,                               &
          &                         path = fields_path,                        &
          &                         parent = field_handle,                     &
          &                         pdfGlobal = pdfGlobal,                     &
          &                         source = source( 1 ),                      &
          &                         physics = physics,                         &
          &                         schemeHeader = schemeHeader )

      else ! field is multiple table
        call aot_table_close( L=conf, thandle=field_sub_handle )
        nFields = aot_table_length( L=conf, thandle=field_handle )
        call tem_write( 'Number of fields: ', nFields )
        ! reallocate the fields
        deallocate( me )
        allocate( me( nFields ))
        ! reallocate the sources
        deallocate( source )
        allocate( source( nFields + 1 ))
        do iField = 1, nFields
          call aot_table_open( L       = conf,                                 &
            &                  parent  = field_handle,                         & 
            &                  thandle = field_sub_handle,                     &
            &                  pos     = iField )

          ! update the lua path
          field_sub_path = fields_path
          call aot_path_addNode( me       = field_sub_path,                    &
            &                    NodeType = 'table',                           &
            &                    pos      = iField ) 

          call tem_write( 'Properties for Field ', iField )
          call aot_get_val( L       = conf,                                    &
            &               thandle = field_sub_handle,                        &
            &               key     = 'label',                                 &
            &               val     = me(iField)%label,                        &
            &               default = '',                                      &
            &               ErrCode = iError )

          ! add underscore here to differentiate state variables and derirved 
          ! variables of each field
          if( trim(me(iField)%label) .ne. '' )                                 &
            &          me(iField)%label = trim( me(iField)%label )//'_'

          if (btest(iError, aoterr_NonExistent)) then
            call tem_write( 'field label is not specified' )
            call tem_write( 'field label is neccessary when nFields>1' )
            call tem_abort()
          end if  

          call tem_horizontalSpacer()
          call mus_load_field_single( me = me(iField),                         &
            &                         nFields = nFields,                       &
            &                         boundary = boundary,                     &
            &                         conf = conf,                             &
            &                         path = field_sub_path,                   &
            &                         pdfGlobal = pdfGlobal,                   &
            &                         parent = field_sub_handle,               &
            &                         source = source( iField ),               &
            &                         physics = physics,                       &
            &                         schemeHeader = schemeHeader )

          ! Check, if the names of the fields are unique. If not, add an  
          ! increasing number
          call tem_checkLabel( label = me(:)%label, nLabels = iField )

          call aot_table_close( L=conf, thandle=field_sub_handle )
        end do ! number of fields
      end if ! if field table is present
    end if

    call aot_table_close( L=conf, thandle=field_handle )

  end subroutine mus_load_fields
!******************************************************************************!
  

!******************************************************************************!
  !> load a single field table
  !!
  subroutine mus_load_field_single( me, nFields, boundary, conf, path, parent, &
    &            pdfGlobal, source, physics, schemeHeader )  
    !--------------------------------------------------------------------------
    !> field type
    type( mus_field_type ), intent(inout) :: me
    integer, intent(in) :: nFields !< number of fields defined in lua file
    !> boundary data from mesh
    type( tem_bc_prop_type ), intent(in):: boundary 
    !> flu state
    type( flu_State ), intent(inout) :: conf
    !> lua path
    type( aot_path_type ),intent(in) :: path
    !> parent handle if scheme table is defined
    integer, intent(in), optional :: parent
    type( pdf_globInfo_type) :: pdfGlobal !< global pdf info
    !> source information for the current field
    type( tem_source_type ), intent(out) :: source
    !> physics type to convert physics to lattice unit or vice versa
    type( mus_physics_type ), intent(in) :: physics
    !> identifier of the scheme
    type( mus_scheme_header_type ), intent(in) :: schemeHeader
    !--------------------------------------------------------------------------
    character(len=labelLen), allocatable :: ic_states(:)
    !--------------------------------------------------------------------------

    call tem_write('Field label '//trim(me%label))
    ! load field properties
    call mus_load_field_prop( me           = me%fieldProp,                     &
      &                       nFields      = nFields,                          &
      &                       conf         = conf,                             &
      &                       pdfGlobal    = pdfGlobal,                        &
      &                       parent       = parent,                           &
      &                       path         = path,                             &
      &                       physics      = physics,                          &
      &                       schemeHeader = schemeHeader )

    ! load the global source variables
    call tem_load_source( me     = source,                                     &
      &                   conf   = conf,                                       &
      &                   parent = parent,                                     &
      &                   key    = 'source',                                   &
      &                   path   = path )

    ! load boundary                      
    call mus_load_bc( me       = me%bc,                                        &
      &               boundary = boundary,                                     &
      &               conf     = conf,                                         &
      &               path     = path,                                         &
      &               parent   = parent )

    !load scheme specific field
    select case(trim(schemeHeader%kind))
    case('lbm', 'lbm_incomp')
      allocate(ic_states(10))
      ic_states = (/ 'pressure ', 'velocityX', 'velocityY', 'velocityZ',       &
        &            'Sxx      ', 'Syy      ', 'Szz      ', 'Sxy      ',       &
        &            'Syz      ', 'Sxz      ' /)
    case default
          call tem_write('The selected scheme kind model is not &
             & supported to initialize ic state names: '&
             &//trim( schemeHeader%kind ))
          call tem_abort()
    end select
    ! Read initial condition
    call tem_load_ic( me=me%ic, conf=conf,                                     &
      &               parent=parent, key = 'initial_condition',                &
      &               path = path,                                             &
      &               StateName = ic_states )

    call mus_load_ic_prop( me     = me,                                        &
      &                    key    = 'initial_condition',                       &
      &                    conf   = conf,                                      &
      &                    parent = parent,                                    &
      &                    path   = path )

    call tem_horizontalSpacer()

  end subroutine mus_load_field_single
!******************************************************************************!


!******************************************************************************!
  !> load fluid properties like fluid and species table from
  !! lua file based on the scheme kind
  !! KM: \todo what is the purpose of this routine? what is thirdOrder in 
  !! initial condition
  !!
  subroutine mus_load_ic_prop( me, conf, parent, path, key )
    !---------------------------------------------------------------------------
    !> field property type
    type( mus_field_type ), intent(inout) :: me
    !> flu state
    type( flu_State ), intent(inout) :: conf 
    !> parent lua handle
    integer, intent(in), optional :: parent 
    !> aotus lua path
    type( aot_path_type ),intent(in) :: path 
    character(len=*), optional :: key
    !---------------------------------------------------------------------------
    character(len=labelLen) :: locKey
    integer :: ic_table, iError
    !---------------------------------------------------------------------------
    call tem_write('Loading IC properties ..')
    if( present( key )) then
      locKey = key
    else
      locKey = 'initial_condition'
    end if
    if ( present(parent) ) then
      call aot_table_open(L=conf, parent=parent, thandle=ic_table, key=locKey)
    else
      call aot_table_open(L=conf, thandle=ic_table, key=locKey)
    end if
    if (ic_table == 0) then
      ! Interpret the initial_condition entry as name for the restart file,
      ! if it is not a table.
      call tem_write( 'Initial condition NOT a table!')
    else
      ! First check if a restart file is present:
      call aot_get_val( L       = conf,                                        &
        &               thandle = ic_table,                                    &
        &               key     = 'setThirdOrder',                             &
        &               val     = me%setThirdOrder,                            &
        &               default = .false.,                                     &
        &               ErrCode = iError )
    end if
    call aot_table_close(L=conf, thandle=ic_table)
    if( me%setThirdOrder ) then
      call tem_write('  Setting third order in initial conditions')
    end if

  end subroutine mus_load_ic_prop
!******************************************************************************!


!******************************************************************************!
  !> write array of fields into a lua file
  !!
  subroutine mus_field_out_vec( me, conf, schemeHeader )
    !---------------------------------------------------------------------------
    !> array of field type
    type( mus_field_type ), intent(in) :: me(:)
    !> aotus out type
    type( aot_out_type ), intent(inout) :: conf
    !> identifier of the scheme
    type( mus_scheme_header_type ), intent(in) :: schemeHeader
    !---------------------------------------------------------------------------
    integer :: iField ! counter variable
    !---------------------------------------------------------------------------

    call aot_out_val( put_conf = conf,                                         &
      &               val      = size(me),                                     &
      &               vname    = 'nFields' )
   
    ! write out the field prefixes for more than one field
    if(size(me) .gt. 1) then
      call aot_out_val( put_conf     = conf,                                   &
        &               val          = me(:)%label,                            &
        &               vname        = 'fieldPrefixes',                        &
        &               max_per_line = 3 )
    end if
    do iField = 1, size(me)
      call mus_field_out_scal( me           = me( iField ),                    &
        &                      conf         = conf,                            &
        &                      schemeHeader = schemeHeader )
    end do
    
  end subroutine mus_field_out_vec
!******************************************************************************!


!******************************************************************************!
  !> write single field into a lua file
  !!
  subroutine mus_field_out_scal( me, conf, schemeHeader )
    !---------------------------------------------------------------------------
    !> single field type
    type( mus_field_type ), intent(in) :: me
    !> aotus out type
    type( aot_out_type ), intent(inout) :: conf
    !> identifier of the scheme
    type( mus_scheme_header_type ), intent(in) :: schemeHeader
    !---------------------------------------------------------------------------

    call aot_out_open_table(put_conf = conf, tname = trim(me%label)//'field')
    call mus_field_prop_out( me           = me%fieldProp,                      &
      &                      conf         = conf,                              &
      &                      schemeHeader = schemeHeader )
    call aot_out_close_table(put_conf = conf)

  end subroutine mus_field_out_scal
!******************************************************************************!


end module mus_field_module
!******************************************************************************!
