! See copyright notice in the COPYRIGHT file.
?? include 'header/lbm_macros.inc'
!******************************************************************************!
!> Flow-related routines such as initialization of the flow field and parameters
!!
module mus_flow_module

  ! include treelm modules
  use env_module,              only: rk, io_buffer_size, long_k, eps, single_k,&
    &                                eps_single_k
  use tem_param_module,        only: cs2inv, cs2
  use tem_aux_module,          only: tem_abort
  use tem_tools_module,        only: tem_write
  use tem_geometry_module,     only: tem_baryOfId, tem_idOfCoord
  use tem_spatial_module,      only: tem_spatial_for
  use tem_restart_module,      only: tem_restart_readHeader,                   &
    &                                tem_init_read_restart,                    &
    &                                tem_restart_openRead_single,              &
    &                                tem_restart_readData_single,              &
    &                                tem_restart_closeRead,                    &
    &                                tem_restart_finalize_single
  use tem_depend_module,       only: tem_init_depend
  use tem_comm_env_module,     only: process
  use tem_subTree_type_module, only: tem_subTree_type, tem_local_subTree_from, &
    &                                tem_destroy_subTree
  use tem_logging_module,      only: tem_log, tem_toStr

  ! include musubi modules
  use mus_global_module,             only: param_type
  use mus_geom_module,               only: mus_geom_type
  use mus_pdf_module,                only: pdf_globInfo_type
  use mus_moments_module,            only: get_moment
  use mus_scheme_type_module,        only: mus_scheme_type
  use mus_derivedQuantities_module2, only: getNEq_diffusive,                   &
    &                                      getNEq_acoustic, convPrePost,       &
    &                                      setThirdOrder_diffusive
  use mus_config_module,             only: mus_open_config
  use mus_buffer_module,             only: mus_rest_unserialize
  use mus_field_module,              only: mus_field_type
  use mus_interpolate_header_module, only: interpolation_type
  use mus_initLBM_module, only: mus_init_advRel_LBM
  use mus_initLBMIncomp_module, only: mus_init_advRel_LBMIncomp

  implicit none

  private

  public :: mus_init_flow
  public :: init_advRel

contains 

!******************************************************************************!
  !> Initialize  the relaxation parameters and choose the relaxation model
  !!
  subroutine init_advRel( scheme )
    !---------------------------------------------------------------------------
    !> scheme type 
    type(mus_scheme_type) :: scheme
    !---------------------------------------------------------------------------
    integer :: iField
    !---------------------------------------------------------------------------
    if( index( scheme%header%relaxation, 'fixedVel') .gt. 0 ) then
      call tem_write(' FixedVelocity: setting spatial velocity distribution to be stored')
      do iField = 1, scheme%nFields
        scheme%field(iField)%fieldProp%fluid%velSpatial(:)%toStore = .true.
      end do
    end if

    select case( trim( scheme%header%kind ) )
    case( 'lbm' )
      call mus_init_advRel_lbm( scheme )
    case( 'lbm_incomp' )
      call mus_init_advRel_lbmIncomp( scheme )
    case default  
      call tem_write( 'The selected scheme kind model is not supported: '//trim( scheme%header%kind ))
      call tem_abort()
    end select  

    call tem_write( 'Choosing scheme kind: '//trim( scheme%header%kind ))
    call tem_write( 'Choosing relaxation model: '//trim( scheme%header%relaxation ))

  end subroutine init_advRel
!******************************************************************************!



!******************************************************************************!
  !> Initialize flow field 
  !!
  subroutine mus_init_flow( scheme, geometry, params, pdfGlobal, intp )
    !---------------------------------------------------------------------------
    !> Scheme type
    type(mus_scheme_type), allocatable :: scheme(:)
    !> Global parameters
    type(param_type), intent(inout) :: params
    !> Geometry information
    type(mus_geom_type), intent(in) :: geometry
    !> global info type
    type(pdf_globInfo_type) :: pdfGlobal
    !> information on interpolation
    type(interpolation_type) :: intp
    !---------------------------------------------------------------------------
    integer :: iScheme, nElems, iField
    integer :: iLevel
    !---------------------------------------------------------------------------

    ! open musubi config file and solver specific lua functions as chunk
    call mus_open_config( conf = params%conf,                                  &
      &                   filename = params%solver%configFile )

    call tem_write( " Initialize fluid initial conditions..." )
    !Normally one would expect a call to open config here. However, with the
    !implementation of dynamic load balancing we have moved that call to
    !mus_load_config. It was always there but the conf was closed in the end and
    !reopened here. This does not make much sense so this strategy has been
    !followed. 

    do iScheme = 1, pdfGlobal%nSchemes
      !> Set equilibrium distribution
      call tem_write( " Initialzing flow on scheme ", iScheme )
      do iLevel = pdfGlobal%minLevel, pdfGlobal%maxlevel
        call tem_write('Initializing level ', iLevel )
        ! For debugging purposes, set complete flow field to invalid
        scheme( iScheme )%pdf(iLevel)%state(:,:) =  -1000000.0_rk
        ! Set number of elements to fill (only fluid or also helper elements?)           
        if( .not. params%init_allElems ) then
          nElems =  scheme(iScheme)%pdf( iLevel )%nElems_fluid
        else
          nElems =  scheme(iScheme)%pdf( iLevel )%nElems_solve
        end if 
        if( nElems .ne. scheme(iScheme)%pdf( iLevel )%nElems_fluid ) then
          call tem_write('') 
          call tem_write('WARNING!!!!')
          call tem_write( 'For debugging purposes, in addition to fluid '//    &
            &             'elements')
          call tem_write( 'more than fluid elements are filled with Initial '//&
            &             'values')
          call tem_write('') 
        end if
 
        select case( trim(scheme( iScheme )%header%kind) )
        case( 'lbm', 'lbm_incomp')
          do iField = 1, scheme(iScheme)%nFields
            ! If we are restarting the simulation, the flow parameters are
            ! initialized from restart file and we don't have to calculate
            ! find chunksize and number of chunks required for initialzation
            if ( .not. scheme(iScheme)%field(iField)%ic%isRestart ) then
              ! Invoke the routine which initializes flow with calculated
              ! quantities
              call mus_init_lbm_withFlowParams( me = scheme(iScheme),          &
                &                geometry  = geometry,                         &
                &                params    = params,                           &
                &                pdfGlobal = pdfGlobal,                        &
                &                nElems    = nElems,                           &
                &                iField    = iField,                           &
                &                iLevel    = iLevel,                           &
                &                field     = scheme(iScheme)%field(iField) )
            else
              ! Invoke the routine which initializes flow from restart file
              call mus_init_withRestart( scheme   = scheme,                        &
                &                        geometry = geometry,                  &
                &                        params   = params,                    &
                &                        pdfGlobal= pdfGlobal,                 &
                &                        field = scheme(iScheme)%field(iField) )
            endif ! not restart
          enddo !ifield  
        case default
          call tem_write('The selected scheme kind model is not '//            &
             & 'supported to initialize flow with equilibrium function: '//    &
             & trim( scheme( iScheme )%header%kind ))
          call tem_abort()
        end select     
      end do !ilevel

      select case( trim( scheme(iScheme)%header%kind ) )
        ! Check the shear stress only for lbm fluid
        case('lbm', 'lbm_incomp')
          if( intp%requireMoments ) then
            call mus_checkShearStress( scheme    = scheme( iScheme ),          &
              &                        geometry  = geometry,                   &
              &                        params    = params,                     &
              &                        pdfGlobal = pdfGlobal )
          else
            call tem_write('WARNING: Check for shear stress is not available')
            call tem_write('         for interpolation schemes not using ')
            call tem_write('         moments')
          end if
      end select  

      ! Choose the advection relaxation scheme
      call init_advRel( scheme = scheme(iScheme) )

    end do !schemes

  end subroutine mus_init_flow 
!******************************************************************************!


!******************************************************************************!
  !> Initialize the flow from calculated quantitites like density, pressure etc.
  !!
  subroutine mus_init_lbm_withFlowParams( me, geometry, params, pdfGlobal, Field,  &
    &                                 iField, nElems, iLevel )
    !--------------------------------------------------------------------------
    !> Scheme type
    type(mus_scheme_type), intent(inout) :: me
    !> Global parameters
    type(param_type), intent(in) :: params
    !> Geometry information
    type(mus_geom_type), intent(in) :: geometry
    !> global info type
    type(pdf_globInfo_type) :: pdfGlobal
    !> Field type
    type(mus_field_type), intent(inout) :: field
    !> Field index
    integer, intent(in)                 :: iField
    !> Number of Elements
    integer, intent(in)                 :: nElems
    !> Level index
    integer, intent(in)                 :: iLevel
    !---------------------------------------------------------------------------
    integer :: iDir, idx_iDir, iElem
    real(kind=rk), allocatable :: fEq(:), fnEq(:), press(:)
    real(kind=rk), allocatable :: xc(:,:), ux(:), uy(:), uz(:), rho(:)
    ! shear rate tensor 
    real(kind=rk), allocatable :: Sxx(:,:)
    real(kind=rk),allocatable :: input_loc(:)
    integer :: iChunk, nChunks, chunkSize, nChunkElems, elemPos
    integer :: tLayerMin, tLayerMax
    integer :: elemoff
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !---------------------------------------------------------------------------
    if( params%init_allTimeLayers ) then
      tLayerMin = 1
      tLayerMax = size( me%pdf( iLevel )%state, 2 )
    else
      tLayerMin = me%pdf( iLevel )%nNext
      tLayerMax = me%pdf( iLevel )%nNext
      call tem_write('') 
      call tem_write('WARNING!!!!')
      call tem_write('For debugging purposes, only one time layer' // &
      &  'being initialized.')
      call tem_write('') 
    end if

    ! find chunksize and number of chunks required for initialzation
    chunkSize = io_buffer_size/me%layout%stencil(1)%QQ
    nChunks = ceiling( real(nElems, kind=rk)/real(chunkSize, kind=rk))

    do iChunk = 1, nChunks
      ! Number of elements read so far in previous chunks.
      elemOff = ( (iChunk-1)*chunksize )
      nChunkElems = min(chunkSize, nElems - elemOff)

      !allocate memory for density, vel, eq, coord
      allocate(xc(nChunkElems, 3))
      allocate(rho(nChunkElems))
      allocate(press(nChunkElems))
      allocate(ux(nChunkElems))
      allocate(uy(nChunkElems))
      allocate(uz(nChunkElems))
      allocate(Sxx(6,nChunkElems))
      allocate(input_loc(nChunkElems*4))
      ! Equilibrium part of the pdf
      allocate(fEq(me%layout%stencil(1)%QQ*nChunkElems))
      ! non-Equilibrium part of the pdf
      allocate(fnEq(me%layout%stencil(1)%QQ*nChunkElems))
      fnEq = 0._rk
      
      do iElem = 1, nChunkElems
        ! Calculate the coordinates
        elemPos = elemOff + iElem
        xc(iElem,1:3) = tem_BaryOfId( geometry%tree,                           &
          & me%glob%levelDesc( iLevel )%total( elemPos ) )
      end do
      
      press = tem_spatial_for( conf  = params%conf,                            &
        &                      me    = field%ic%ini_state(1),                  &
        &                      coord = xc,                                     &
        &                      n     = nChunkElems )
      ux = tem_spatial_for( conf  = params%conf,                               &
        &                   me    = field%ic%ini_state(2),                     &
        &                   coord = xc,                                        &
        &                   n     = nChunkElems )
      uy = tem_spatial_for( conf  = params%conf,                               &
        &                   me    = field%ic%ini_state(3),                     &
        &                   coord = xc,                                        &
        &                   n     = nChunkElems )
      uz = tem_spatial_for( conf  = params%conf,                               &
        &                   me    = field%ic%ini_state(4),                     &
        &                   coord = xc,                                        &
        &                   n     = nChunkElems )
      ! Read in the shear rate tensor
      ! This corresponds to S = grad(u) + grad(u)^T
      Sxx = 0._rk
      Sxx(1,:) = tem_spatial_for( conf  = params%conf,                         &
        &                         me    = field%ic%ini_state(5),               &
        &                         coord = xc,                                  &
        &                         n     = nChunkElems )
      Sxx(2,:) = tem_spatial_for( conf  = params%conf,                         &
        &                         me    = field%ic%ini_state(6),               &
        &                         coord = xc,                                  &
        &                         n     = nChunkElems )
      Sxx(3,:) = tem_spatial_for( conf  = params%conf,                         &
        &                         me    = field%ic%ini_state(7),               &
        &                         coord = xc,                                  &
        &                         n     = nChunkElems )
      Sxx(4,:) = tem_spatial_for( conf  = params%conf,                         &
        &                         me    = field%ic%ini_state(8),               &
        &                         coord = xc,                                  &
        &                         n     = nChunkElems )
      Sxx(5,:) = tem_spatial_for( conf  = params%conf,                         &
        &                         me    = field%ic%ini_state(9),               &
        &                         coord = xc,                                  &
        &                         n     = nChunkElems )
      Sxx(6,:) = tem_spatial_for( conf  = params%conf,                         &
        &                         me    = field%ic%ini_state(10),              &
        &                         coord = xc,                                  &
        &                         n     = nChunkElems )

      ! If physics conversion is active then convert from physics to LB
      !if( me%physics%active ) then
      !call tem_write(' Conversion of initial conditions from physics to LB')
      ! 1. velocity 
      ux = ux/me%physics%fac( iLevel )%vel
      uy = uy/me%physics%fac( iLevel )%vel
      uz = uz/me%physics%fac( iLevel )%vel
      !if( trim(params%scaling) .eq. 'diffusive' .or. &
      !  & trim(me%header%kind) .eq. 'lbm_incomp' ) then
      ! For the incompressible formulation, add the reference
      ! density
      ! 2. density / pressure
      rho = press*cs2inv/me%physics%fac( iLevel )%press
      !rho = rho + field%fieldProp%fluid%rho0
      !else
      !  rho = press*cs2inv
      !end if
      ! 3. shear rate
      Sxx = Sxx/( me%physics%fac( iLevel )%strainRate )
      !end if
     
      !linearized the input for equilibrium function pointer
      !with following order
      !|dens_1 | dens_2 | .. | dens_nElems | ux_1 | uy_1 | uz_1 | ..
      !!.. | ux_nelems| uy_nelems| uz_nelems|
      input_loc = 0._rk
      input_loc(1:nChunkElems) = rho

      do iElem = 1, nChunkElems
        ! MH: Fix for the intel compiler:
        ! entries have to be assigned one by one instead of 
        ! assigning them as an array of (/ux, uy, uz/)
        input_loc( nChunkElems + (iElem-1)*3 + 1) = ux( iElem )
        input_loc( nChunkElems + (iElem-1)*3 + 2) = uy( iElem )
        input_loc( nChunkElems + (iElem-1)*3 + 3) = uz( iElem )
      end do

      ! initialize the lower and upper bound
      bound(1) = 1
      bound(2) = nChunkElems
      ! Calculate the equilibrium distribution
      call me%varSys%derive( me%derVarPos(iField)%equilFromMacro )%do(         &
        &              input   = input_loc,                                    &
        &              bound   = bound,                                        &
        &              tree    = geometry%tree,                                &
        &              subTree = subTree,                                      &
        &              requVar = me%varSys%Variable%val(                       &
        &                                me%derVarPos(iField)%equilFromMacro), &
        &              globSys = me%varSys,                                    &
        &              stencil = me%layout%stencil(1),                         &
        &              conf    = me%varSys%solSpec_conf,                       & 
        &              res     = fEq )


      do iElem = 1, nChunkElems
        elemPos = elemOff + iElem
        select case( trim( me%header%kind ) )
        case('lbm', 'lbm_incomp')
          select case( trim( params%scaling ))
          case('acoustic')
            ! Acoustic fNeq with defining the stress tensor instead of the 
            ! shear rate tensor in the lua file
            fNeq( (iElem-1)*me%layout%stencil(1)%QQ + 1                        &
            &  : (iElem)*me%layout%stencil(1)%QQ) =                            &
            &  getnEq_acoustic( layout = me%layout,                            &
            &                   rho0   = field%fieldProp%fluid%rho0,           &
            &                   omega  = field%fieldProp%fluid%omLvl(iLevel),  &
            &                   Sxx    = [ Sxx(1,iElem), Sxx(4,iElem),         &
            &                              Sxx(6,iElem), Sxx(4,iElem),         &
            &                              Sxx(2,iElem), Sxx(5,iElem),         &
            &                              Sxx(6,iElem ), Sxx(5,iElem),        &
            &                              Sxx(3,iElem) ] )  
          case('diffusive')
            ! Diffusive non-equilibrium part with the shear rate tensor
            ! given as an input
            fNeq( (iElem-1)*me%layout%stencil(1)%QQ + 1                        &
            &  : (iElem)*me%layout%stencil(1)%QQ) =                            &
            &  getnEq_diffusive( layout = me%layout,                           &
            &                    omega  = field%fieldProp%fluid%omLvl(iLevel), &
            &                    Sxx    = [ Sxx(1,iElem), Sxx(4,iElem),        &
            &                               Sxx(6,iElem), Sxx(4,iElem),        &
            &                               Sxx(2,iElem), Sxx(5,iElem),        &
            &                               Sxx(6,iElem), Sxx(5,iElem),        &
            &                               Sxx(3,iElem) ] )  
            if( field%setThirdOrder ) then
            fNeq( (iElem-1)*me%layout%stencil(1)%QQ+1:(iElem)                  &
              &                                      *me%layout%stencil(1)%QQ) &
              & = fNeq( (iElem-1)*me%layout%stencil(1)%QQ+1:(iElem)            &
              &                                      *me%layout%stencil(1)%QQ) &
              & +  setThirdOrder_diffusive(                                    &
              &                  layout = me%layout,                           &
              &                  omega  = field%fieldProp%fluid%omLvl(iLevel), &
              &                  x      = xc(iElem,:),                         &
            ! uLB Velocity in lattice units = u0Phys *dt/dx , u0Phys = 1 for TGV
              &                  dx     = me%physics%dxLvl( iLevel ),          &
              &                  dt     = me%physics%dtLvl( iLevel ) )
            end if
          end select
        end select

        do iDir =1,me%layout%stencil(1)%QQ 
          idx_iDir = me%varSys%Variable%val(iField)%varPos( iDir )
          me%pdf( iLevel )%state(                                              &
          ! In general we want to set for each element a valid entry.
          ! This was  IDX, but for PULL, it doesnt change anything
          ! for PUSH however, we need to set it to save!
          !KM: initialize both state vectors for boundaries
& ?IDX?( idx_iDir, elemPos, me%varSys%nScalars, me%pdf(iLevel)%neigh),         &
          !\todo need to check for multilevel
            &   tLayerMin:tLayerMax )                                          &
            & = fEq((iElem-1)*me%layout%stencil(1)%QQ + iDir)                  &
            & + fNeq((iElem-1)*me%layout%stencil(1)%QQ + iDir)

        end do !idir
      end do !ielem

      ! deallocate memory for next chunk  
      deallocate(xc)
      deallocate(rho)
      deallocate(press)
      deallocate(ux)
      deallocate(uy)
      deallocate(uz)
      deallocate(Sxx)
      deallocate(input_loc)
      deallocate(fEq)
      deallocate(fnEq)

    end do ! chunk

  end subroutine mus_init_lbm_withFlowParams
!******************************************************************************!

!******************************************************************************!
  !> Initialize flow field 
  !!
  subroutine mus_init_withRestart( scheme, geometry, params, pdfGlobal, Field )
    !---------------------------------------------------------------------------
    !> Scheme type
    type(mus_scheme_type), allocatable :: scheme(:)
    !> Global parameters
    type(param_type), intent(inout) :: params
    !> Geometry information
    type(mus_geom_type), intent(in) :: geometry
    !> global info type
    type(pdf_globInfo_type) :: pdfGlobal
    !> Field type
    type(mus_field_type), intent(inout) :: Field
    !---------------------------------------------------------------------------
    integer :: iChunk
    integer :: offset, iSys, nDofs, nChunkElems_res, elemoff
    ! array for chunkwise memory bounded output
    real(kind=rk), allocatable :: chunk(:)
    !---------------------------------------------------------------------------
    allocate( chunk( io_buffer_size ))
    ! The file is read in ini condition type. Assign the file name to
    ! restart controller
    field%restart%controller%readFileName = field%ic%RestartFile

    ! Load the restart header for this field
    call tem_restart_readHeader( me       = field%restart,                     &
      &                          tree     = geometry%tree,                     &
      &                          timing   = pdfGlobal%timing,                  &
      &                          proc     = process,                           &
      &                          readMesh = .false. )
    
    ! Initialize the depend, to get the position of restart varsys
    call tem_init_depend( medepend    = field%ic%depend,                       &
      &                   varSys      = scheme(:)%varSys,                      &
      &                   resSyslabel = 'restart',                             &
      &                   label_list  = scheme(:)%header%label,                &
      &                   iScheme     = 0 )

    ! Initialize the restart for this variable system. The scheme on 
    ! which the requested system is dependent is passed as an input
    ! to find the position of varsys in the global system
    !\todo: depno: KJ: multiple depend tables are not making much sense here
    ! it has to be discussed and resolved, right now passing first 
    ! instance of depend array
    iSys = field%ic%depend(1)%useScheme

    call tem_init_read_restart( me     = field%restart,                        &
      &                         solver = params%solver,                        &
      &                         varSys = field%restart%varSys,                 &
      &                         tree   = geometry%tree,                        &
      &                         iSys   = iSys )

    ! Open the restart file and set the processes' view on the file
    !\todo: depno
    call tem_restart_openRead_single( me = field%restart, iSys = iSys )

    ! Read the data from restart file into the chunk array
    ! first calculate the offset which is dependent on the positon
    ! of the variable system

!    elemOff = 0
    call tem_write( 'Reading restart...' )

    do iChunk = 1, field%restart%nChunks
      
      elemOff = ((iChunk-1)*field%restart%chunkSize )
      field%restart%nChunkElems = min( field%restart%chunkSize,                &
        &      geometry%tree%nElems-elemOff)

      nDofs = field%restart%nDofs
      nChunkElems_res = field%restart%nChunkElems
      if( iSys .eq. 1 ) then
        offset = 1
      else
        offset = 1 + sum( field%restart%varSys                                 &
          &                           (1:iSys)%nScalars*nChunkElems_res*nDofs )
      end if

      call tem_restart_readData_single( restart = field%restart,               &
        &                               chunk   = chunk,                       &
        &                               iSys    = iSys,                        &
        &                               offset  = offset )

      ! \todo: KJ: Invoke the unserialization routine to put the data in chunk
      ! in state arrays. This is quite a work around at this moment
      ! and the index of chunk needs to be thought of in detail
      
      call mus_rest_unserialize( scheme    = scheme,                           &
        &                        pdfGlobal = pdfGlobal,                        &
        &                        restart   = field%restart,                    &
        &                        tree      = geometry%tree, chunk=chunk,       &
        &                        iChunk    = iChunk,                           &
        &                        iSys_idx  = iSys )
      
!      elemOff = elemOff + field%restart%nChunkElems

    end do !nChunks within restart

    !close the restart file
    call tem_restart_closeRead( me = field%restart, iSys_idx = iSys)
    
    ! Free the datatypes by finalizing the restart
    call tem_restart_finalize_single(me = field%restart, iSys = iSys)

    deallocate( chunk )

  end subroutine mus_init_withRestart
!******************************************************************************!


!******************************************************************************!
  !> test the shear stress setting with the non-eq part and re-computing it with
  !! the derive routine
  !!
  subroutine mus_checkShearStress( scheme, geometry, params, pdfGlobal )
    !---------------------------------------------------------------------------
    !> Scheme type
    type(mus_scheme_type) :: scheme
    !> Global parameters
    type(param_type), intent(in) :: params
    !> Geometry information
    type(mus_geom_type), intent(in) :: geometry
    !> global info type
    type(pdf_globInfo_type) :: pdfGlobal
    !---------------------------------------------------------------------------
    integer :: iField
    integer :: iLevel, iVal, jVal
    real(kind=rk), allocatable :: fEq(:), fnEq(:), fTmp(:), fnEqDerived(:)
    real(kind=rk), allocatable :: tMom(:)
    ! Physical quantities
    real(kind=rk) :: pPhRef, uxPhRef(3), mSxxPhRef(3,3), SxxPhRef(6)
    real(kind=rk) :: pPhres(1), uxPhres(3), SxxPhres(6), strPhRes(6)
    ! Lattice  quantities
    real(kind=rk) :: rhoLBref, uxLBref(3), mSxxLBref(3,3), SxxLBref(6)
    real(kind=rk) :: rhoLBres, uxLBres(3), SxxLBres(6), strLBres(6)
    real(kind=rk) :: errpPh, errVel(3), errSxx(6), Pixx0(6)
    real(kind=rk) :: input_loc(6), omega, rho0, viscLB, rho0div
    integer :: nStresses, iShearMin, iShearMax
    integer(kind=long_k) :: targetID
    logical :: checkShear
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !---------------------------------------------------------------------------
    if( process%rank .eq. 0 ) then
      call tem_write('Checking shear stress computation for '//                &
        &                                                 trim(params%scaling))
      checkShear = .true.

      do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel

        targetID = tem_idOfCoord( [1,1,1,iLevel] )

        ! build the subTree from the global information and the targetTreeID
        subTree%global = geometry%tree%global
        call tem_local_subTree_from( me = subTree, treeID = (/targetID/))

        iField = 1
        omega = scheme%field( iField )%fieldProp%fluid%omLvl( iLevel )
        rho0  = scheme%field( iField )%fieldProp%fluid%rho0
        viscLB = cs2*(1._rk/omega-0.5_rk)
        if( allocated( fTmp )) deallocate( fTmp )
        if( allocated( fNeq )) deallocate( fNeq )
        if( allocated( fEq ))  deallocate( fEq )
        if( allocated( fnEqDerived ))  deallocate( fnEqDerived )
        if( allocated( tMom ))  deallocate( tMom )
        allocate( fTmp( scheme%layout%stencil(1)%QQ ))
        allocate(  fEq( scheme%layout%stencil(1)%QQ ))
        allocate( fNeq( scheme%layout%stencil(1)%QQ ))
        allocate( fnEqDerived( scheme%layout%stencil(1)%QQ ))
        allocate( tMom( scheme%layout%stencil(1)%QQ ))
        call random_number(pPhRef) 
        uxPhRef  = 0._rk
        mSxxPhRef = 0._rk
        nStresses = max( 3*(pdfGlobal%nDims-1), 1 )
        iShearMin = pdfGlobal%nDims+2
        iShearMax = iShearMin + nStresses - 1
        do jVal = 1, pdfGlobal%nDims
          call random_number( uxPhRef( jVal ))
        end do
        do jVal = 1, pdfGlobal%nDims
          do iVal = 1, pdfGlobal%nDims
            call random_number( mSxxPhRef( iVal, jVal ))
          end do
        end do
        mSxxPhRef(1,2) = mSxxPhRef(2,1)
        mSxxPhRef(2,3) = mSxxPhRef(3,2)
        mSxxPhRef(1,3) = mSxxPhRef(3,1)

        ! Physical to LB units
        mSxxLBref = mSxxPhRef / scheme%physics%fac( iLevel )%strainRate
        uxLBref   = uxPhRef  / scheme%physics%fac( iLevel )%vel
        rhoLBref  = pPhRef  / scheme%physics%fac( iLevel )%press*cs2inv
        if( trim( scheme%header%kind ) .eq. 'lbm') then
          rhoLBref  = rhoLBref + rho0
        end if
        ! Matrices to linearized
        SxxPhRef(1) = mSxxPhRef(1,1)
        SxxPhRef(2) = mSxxPhRef(2,2)
        SxxPhRef(3) = mSxxPhRef(3,3)
        SxxPhRef(4) = mSxxPhRef(1,2)
        SxxPhRef(5) = mSxxPhRef(2,3)
        SxxPhRef(6) = mSxxPhRef(1,3)
        SxxLBRef(1) = mSxxLBRef(1,1)
        SxxLBRef(2) = mSxxLBRef(2,2)
        SxxLBRef(3) = mSxxLBRef(3,3)
        SxxLBRef(4) = mSxxLBRef(1,2)
        SxxLBRef(5) = mSxxLBRef(2,3)
        SxxLBRef(6) = mSxxLBRef(1,3)
     
        input_loc = 0._rk
        input_loc(1) = rhoLBref
        input_loc(2:4) = uxLBref(:)
     
        ! initialize the lower and upper bound
        bound(1) = 1
        bound(2) = 1

        ! Calculate the equilibrium distribution
        call scheme%varSys%derive(scheme%derVarPos(iField)%equilFromMacro)%do( &
          &          input   = input_loc,                                      &
          &          bound   = bound,                                          &
          &          tree    = geometry%tree,                                  & 
          &          subTree = subTree,                                        &
          &          requVar = scheme%varSys%Variable%val(                     &
          &                         scheme%derVarPos(iField)%equilFromMacro),  &
          &          globSys = scheme%varSys,                                  &
          &          stencil = scheme%layout%stencil(1),                       &
          &          conf    = scheme%varSys%solSpec_conf,                     &
          &          res     = fEq )

        ! Calculate non-equilibrium
        select case( trim( params%scaling ))
        case('acoustic')
          ! Acoustic fNeq with defining the stress tensor instead of the 
          ! shear rate tensor in the lua file
          fNeq(1:scheme%layout%stencil(1)%QQ) =                                &
            &  getnEq_acoustic(                                                &
            &              layout = scheme%layout,                             &
            &              rho0   = scheme%field(iField)%fieldProp%fluid%rho0, &
            &              omega  = omega,                                     &
            &              Sxx    = mSxxLBref  )
        case('diffusive')
          ! Diffusive non-equilibrium part with the shear rate tensor
          ! given as an input
          fNeq( 1:scheme%layout%stencil(1)%QQ) =                               &
            &  getnEq_diffusive(                                               &
            &     layout = scheme%layout,                                      &
            &     omega  = scheme%field(iField)%fieldProp%fluid%omLvl(iLevel), &
            &     Sxx    = mSxxLBref )
        end select
     
        fTmp = fEq + fNeq
        ! Double check the non-equilibrium part
        ! -> recompute and compare against fnEq
        call scheme%varSys%derive(scheme%derVarPos(iField)%                    &
          &                                             nonEquilfromState)%do( &
          &      input   = fTmp,                                               &
          &      bound   = bound,                                              &
          &      tree    = geometry%tree,                                      &
          &      subTree = subTree,                                            &
          &      requVar = scheme%varSys%Variable%val(                         &
          &                      scheme%derVarPos(iField)%nonEquilFromState ), &
          &      globSys = scheme%varSys,                                      &
          &      stencil = scheme%layout%stencil(1),                           &
          &      conf    = scheme%varSys%solSpec_conf,                         &
          &      res     = fNeqDerived )
        call tem_log( level = 10,                                              &
          &           msg   = 'fNeq  Error abs:  '//                           &
          &                  trim( tem_toStr(maxval(abs((fNeq-fNeqDerived))))))
        if( maxval(abs((fNeq-fNeqDerived))) .gt. 1000._rk*eps ) then
          call tem_log( level = 1,                                             &
            &           msg   = 'Deviation between the nEq-computation in '//  &
            &                   'the derive module and a manual computation '//&
            &                   'by fNeq = f-fEq. Please check!' )
          call tem_log( level = 1,                                             &
            &           msg   = 'fNeq    '//                                   &
            &                   trim( tem_toStr( val = fNeq, sep = '' )))
          call tem_log( level = 1,                                             &
            &           msg   = 'fNeqDer '//                                   &
            &                   trim( tem_toStr( val = fNeqDerived, sep ='' )))
          call tem_log( level = 1,                                             &
            &           msg   = 'maxError'//                                   &
            &                trim( tem_toStr( maxval(abs((fNeq-fNeqDerived))))))
        end if

        ! get all the moments form the pdf by matrix multiplication
        tMom(:) = matmul( scheme%layout%moment%toMoments%A, fTmp)
        if( trim( scheme%header%kind ) .eq. 'lbm' ) then
          rho0div = 1._rk / tMom(1)
        else
          rho0div = rho0
        end if
        ! Compare the matrix multiplication result to a direct extraction of the
        ! moment from get_moment
        if( scheme%layout%stencil(1)%nDims .eq. 2 ) then
          Pixx0( 1 ) = get_moment( fEq, scheme%layout, (/2,0,0/) )
          Pixx0( 2 ) = get_moment( fEq, scheme%layout, (/0,2,0/) )
          Pixx0( 3 ) = get_moment( fEq, scheme%layout, (/1,1,0/) )
          tMom( 4 ) = tMom( 4 ) - cs2*tMom( 1) - tMom( 2)*tMom( 2)*rho0div
          tMom( 5 ) = tMom( 5 ) - cs2*tMom( 1) - tMom( 3)*tMom( 3)*rho0div  
          tMom( 6 ) = tMom( 6 ) - tMom( 2)*tMom( 3)*rho0div  
          ! Get the shear stress
          tMom(4:6) = -( 1._rk - omega*0.5_rk )*tMom(4:6)*convPrePost( omega )
          ! Convert shear stress to shear rate
          tMom(4:6) = tMom(4:6)/viscLB/rho0
          ! Convert LB shear rate to physical shear rate
          tMom(4:6) = tMom(4:6)*scheme%physics%fac( iLevel )%strainRate
        else
          Pixx0( 1 ) = get_moment( fEq, scheme%layout, (/2,0,0/) )
          Pixx0( 2 ) = get_moment( fEq, scheme%layout, (/0,2,0/) )
          Pixx0( 3 ) = get_moment( fEq, scheme%layout, (/0,0,2/) )
          Pixx0( 4 ) = get_moment( fEq, scheme%layout, (/1,1,0/) )
          Pixx0( 5 ) = get_moment( fEq, scheme%layout, (/0,1,1/) )
          Pixx0( 6 ) = get_moment( fEq, scheme%layout, (/1,0,1/) )
          tMom( 5 ) = tMom( 5 ) - cs2*tMom( 1) - tMom( 2)*tMom( 2)*rho0div
          tMom( 6 ) = tMom( 6 ) - cs2*tMom( 1) - tMom( 3)*tMom( 3)*rho0div     
          tMom( 7 ) = tMom( 7 ) - cs2*tMom( 1) - tMom( 4)*tMom( 4)*rho0div  
          tMom( 8 ) = tMom( 8 ) - tMom( 2)*tMom( 3)*rho0div  
          tMom( 9 ) = tMom( 9 ) - tMom( 3)*tMom( 4)*rho0div  
          tMom(10 ) = tMom(10 ) - tMom( 4)*tMom( 2)*rho0div  
          ! Get the shear stress
          tMom(5:10) = -( 1._rk - omega*0.5_rk )*tMom(5:10)*convPrePost( omega )
          ! Convert shear stress to shear rate
          tMom(5:10) = tMom(5:10)/viscLB/rho0
          ! Convert LB shear rate to physical shear rate
          tMom(5:10) = tMom(5:10)*scheme%physics%fac( iLevel )%strainRate
        end if

        ! Compute from the given distribution fTmp the initial quantities again
        call scheme%varSys%derive( scheme%derVarPos(iField)%pressure )%do(     &
          &                    input   = fTmp,                                 &
          &                    bound   = bound,                                &
          &                    tree    = geometry%tree,                        &
          &                    subTree = subTree,                              &
          &                    requVar = scheme%varSys%Variable%val(           &
          &                               scheme%derVarPos(iField)%pressure ), &
          &                    globSys = scheme%varSys,                        &
          &                    stencil = scheme%layout%stencil(1),             &
          &                    conf    = scheme%varSys%solSpec_conf,           &
          &                    res     = pPhres )
        if( trim( scheme%header%kind ) .eq. 'lbm' ) then
          pPhRes    = pPhRes - cs2*rho0
        end if
        pPhRes    = pPhRes*scheme%physics%fac( iLevel )%press
        rhoLBres  = sum(fTmp)
     
        call scheme%varSys%derive( scheme%derVarPos(iField)%velocity )%do(     &
          &                     input   = fTmp,                                &
          &                     bound   = bound,                               &
          &                     tree    = geometry%tree,                       &
          &                     subTree = subTree,                             &
          &                     requVar = scheme%varSys%Variable%val(          &
          &                               scheme%derVarPos(iField)%velocity ), &
          &                     globSys = scheme%varSys,                       &
          &                     stencil = scheme%layout%stencil(1),            &
          &                     conf    = scheme%varSys%solSpec_conf,          & 
          &                     res     = uxLBres(:) )
        uxPhRes = uxLBres*scheme%physics%fac( iLevel )%vel
     
        call scheme%varSys%derive( scheme%derVarPos(iField)%shearstress )%do(  &
          &                       input   = fTmp,                              &
          &                       bound   = bound,                             &
          &                       tree    = geometry%tree,                     &
          &                       subTree = subTree,                           &
          &                       requVar = scheme%varSys%Variable%val(        &
          &                            scheme%derVarPos(iField)%shearstress ), &
          &                       globSys = scheme%varSys,                     &
          &                       stencil = scheme%layout%stencil(1),          &
          &                       conf    = scheme%varSys%solSpec_conf,        &
          &                       res     = StrLBres(:) )
        sxxLBres    = strLBres/viscLB/rho0
        sxxPhRes    = SxxLBres*scheme%physics%fac( iLevel )%strainRate
        strPhres    = strLBres*scheme%physics%fac( iLevel )%press
     
        !-----------------------------------------------------------------------
        !   Error evaluation
        !-----------------------------------------------------------------------
        errpPh = 0._rk
        errVel = 0._rk
        errSxx = 0._rk
        if( abs(pPhRef) .gt. eps ) then
          errpPh = abs((pPhRef-pPhRes(1))/pPhRef)
        end if
        do jVal = 1,pdfGlobal%nDims
          if( abs(uxPhRef(jVal)) .gt. eps ) then
            errVel(jVal) = abs((uxPhRef( jVal ) -uxPhRes( jVal ))/uxPhRef(jVal))
          end if
        end do
        do jVal = 1,size( sxxPhRef  )
          if( abs(sxxPhRef(jVal)) .gt. eps ) then
            errSxx(jVal) = abs((SxxPhRef(jVal) -SxxPhRes(jVal))/SxxPhRef(jVal))
          end if
        end do

        call tem_log( level = 10,                                              &
          &           msg   = '    input data' )
        call tem_log( level = 10,                                              &
          &           msg   = '   rho        ux                            '// &
          &                   ' Sxx      ' )
        call tem_log( level = 10,                                              &
          &           msg   = '-----------Physical units-------------------'// &
          &                   '---------' )
        call tem_log( level = 10,                                              &
          &           msg   = 'ref '//                                         &
          &                   trim( tem_toStr( pPhRef ))//                     &
          &                   trim( tem_toStr( val = uxPhRef, sep = '' ))//    &
          &                   trim( tem_toStr( val = SxxPhRef,sep = '' )))
        call tem_log( level = 10,                                              &
          &           msg   = 'val '//                                         &
          &                   trim( tem_toStr( val = pPhRes,  sep = '' ))//    &
          &                   trim( tem_toStr( val = uxPhRef, sep = '' ))//    &
          &                   trim( tem_toStr( val = SxxPhRef,sep = '' )))
        call tem_log( level = 10,                                              &
          &           msg   = 'val '//                                         &
          &                   trim( tem_toStr( val = pPhRes,  sep = '' ))//    &
          &                   trim( tem_toStr( val = uxPhRes, sep = '' ))//    &
          &                   trim( tem_toStr( val = tMom(iShearMin:iShearMax),&
          &                                    sep = ',' )))
        call tem_log( level = 10,                                              &
          &           msg   = 'err '//                                         &
          &                   trim( tem_toStr( errpPh ))//                     &
          &                   trim( tem_toStr( val = errVel, sep = '' ))//     &
          &                   trim( tem_toStr( val = errSxx, sep = '' )))
        call tem_log( level = 10,                                              &
          &           msg   = '-----------Lattice units--------------------'// &
          &                   '---------' )
        call tem_log( level = 10,                                              &
          &           msg   = 'ref '//                                         &
          &                   trim( tem_toStr( rhoLBRef ))//                   &
          &                   trim( tem_toStr( val = uxLBRef, sep = '' ))//    &
          &                   trim( tem_toStr( val = SxxLBRef, sep = '' )))
        call tem_log( level = 10,                                              &
          &           msg   = 'val '//                                         &
          &                   trim( tem_toStr( rhoLBRes ))//                   &
          &                   trim( tem_toStr( val = uxLBRes, sep = '' ))//    &
          &                   trim( tem_toStr( val = SxxLBRes, sep = '' )))

        if( maxval( abs( [errpPh, errVel, errSxx] )) .gt. 100000._rk*eps ) then
          call tem_log( level = 1,                                             &
            &           msg   = 'max error:   '//                              &
            &        trim( tem_toStr( maxval(abs( [errpPh, errVel, errSxx] )))))
          call tem_log( level = 1,                                             &
            &           msg   = 'error limit: '//                              &
            &                   trim( tem_toStr( eps*100000._rk )))
          call tem_write('error in shear computation. abort.')
          call tem_write('Larger deviation than allowed!')
          if(trim(params%scaling) == 'diffusive' ) then
          else
            !call tem_abort
          end if
        end if
        call tem_destroy_subTree(subTree)
      end do
    end if

  end subroutine mus_checkShearStress
!******************************************************************************!

  
end module  mus_flow_module
!******************************************************************************!
