! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module keeps all information about the fluid
!!
!! In this module, all infos about the fluid is collected. 
!! The file type \ref fluid_type "fluid type" contains all relevant information.
!! This includes physical parameters such as the viscosity and reference 
!! density. Also, LBM-specific parameters such as the relaxation rates are
!! defined.
!!
module mus_fluid_module

  ! include treelm modules
  use env_module,           only: rk, single_k, eps_single_k, globalMaxLevels
  use treelmesh_module,     only: treelmesh_type
  use tem_tools_module,     only: tem_write, tem_horizontalSpacer
  use tem_aux_module,       only: tem_abort
  use tem_param_module,     only: cs2, cs2inv
  use tem_transient_module, only: tem_transient_type, load_transient
  use tem_spatial_module,   only: tem_spatial_type, load_spatial  
  use tem_debug_module,     only: main_debug

  ! include aotus modules
  use aotus_module,     only: flu_State, aot_get_val, aoterr_Fatal,            &
    &                         aoterr_NonExistent, aoterr_WrongType
  use aot_table_module, only: aot_table_open, aot_table_close
  use aot_path_module,  only: aot_path_type, assignment(=), aot_path_addNode
  use aot_out_module,   only: aot_out_type, aot_out_val, aot_out_open_table,   &
    &                         aot_out_close_table

  ! include musubi modules
  use mus_physics_module, only: mus_physics_type
  use mus_pdf_module,     only: pdf_globInfo_type

  implicit none
  private

  public :: fluid_type
  public :: mus_load_fluid
  public :: mus_fluid_out
  public :: set_omegas

  ! Large eddy turbulence model type
  type les_fluid_type
    real(kind=rk) :: c_s = 0.17_rk     !< Smagorinsky constant
  end type les_fluid_type

  ! MRT relaxation model type
  type mrt_fluid_type
    !> Relaxation times for each moment
    real(kind=rk),allocatable :: s_mrt(:)
  end type mrt_fluid_type

  !> collection of properties of the fluid
  type fluid_type
    logical :: active = .false.       !< is this object a fluid?
    !> main relaxation parameter for setting viscosity
    real(kind=rk) :: omega = 1.7_rk 
    !> level-wise omegas which is a multiplication of omega and omega_ramping
    real(kind=rk) :: omLvl(globalMaxLevels)
    !> level-wise bulk omegas 
    real(kind=rk) :: omLvl_bulk(globalMaxLevels)
    !> transient omega for ramping etc.
    type(tem_transient_type) :: omega_ramping
    real(kind=rk) :: om_bulk              !< bulk relaxation parameter
    real(kind=rk) :: kine_viscosityLB    !< lattice kinematic shear viscosity 
    real(kind=rk) :: kine_viscosity    !< physical kinematic shear viscosity 
    !> lattice kinematic shear viscosity for each level
    real(kind=rk) :: visc(globalMaxLevels)
    !> lattice bulk viscosity for each level
    real(kind=rk) :: bulkVisc(globalMaxLevels)
    real(kind=rk) :: bulk_viscosityLB     !< lattice bulk viscosity
    real(kind=rk) :: bulk_viscosity     !< physical bulk viscosity
    real(kind=rk) :: rho0            !< reference lattice density
    real(kind=rk) :: temp0           !< reference lattice temperature
    real(kind=rk) :: force(3)        !< forcing term
    real(kind=rk) :: kappa = 1._rk  !< isentropic heat coefficient
    !> properties for the mrt relaxation model
    type( mrt_fluid_type ) :: mrt( globalMaxLevels ) 
    type( les_fluid_type ) :: les    !< properties for the les model
    !> spatial viscosity definition, e.g. for sponge layers 
    type(tem_spatial_type) :: viscSpatial
    !> spatial velocity definition for fixed velocity initial condition iterations
    type(tem_spatial_type) :: velSpatial(3)
  end type fluid_type

contains

!******************************************************************************!
  !> Read in the fluid property from the LUA parameter file 
  subroutine mus_load_fluid( me, conf, parent, path, pdfGlobal, physics )
    !--------------------------------------------------------------------------
    !> fluid type 
    type( fluid_type ),intent(out) :: me
    !> lua state
    type( flu_state )              :: conf
    !> parent handle
    integer, intent(in), optional  :: parent
    !> lua path to transient lua function
    type(aot_path_type), intent(in) :: path
    !> global pdf info
    type( pdf_globInfo_type) :: pdfGlobal
    !> physics type to convert physics to lattice unit or vice versa
    type( mus_physics_type ), intent(in) :: physics
    !--------------------------------------------------------------------------
    integer :: fluid_table
    integer :: iError 
    integer :: vecError(3)
    type(aot_path_type) :: fluid_path !< lua path to transient lua function
    !---------------------------------------------------------------------------
    
    fluid_path = path

    ! if fluid informations in scheme table parentHandle /= 0
    if( present(parent) ) then 
      call aot_table_open( L       = conf,                                     &
        &                  parent  = parent,                                   &
        &                  thandle = fluid_table,                              &
        &                  key     = 'fluid' )
    else
      call aot_table_open( L=conf, thandle = fluid_table, key = 'fluid' )
    end if
    call aot_path_addNode( me = fluid_path, NodeType = 'table', key = 'fluid' ) 
    ! Need to look for:
    ! * active (mandatory)
    ! * boundary_type (default 1)
    ! * fileformat (default binary)
    if ( fluid_table .eq. 0 ) then
      call tem_write(' No fluid table defined')
      call tem_abort()
    endif
    call tem_write( ' Loading fluid informations' )
    ! Activating the current fluid object
    me%active = .true.
    ! Omega ramping
    call load_transient( me     = me%omega_ramping,                            &
      &                  conf   = conf,                                        &
      &                  parent = fluid_table,                                 &
      &                  key    = 'omega_ramping',                             &
      &                  path   = fluid_path )

    me%viscSpatial%toStore = .false.
    call load_spatial( me      = me%viscSpatial,                               &
      &                conf    = conf,                                         &
      &                parent  = fluid_table,                                  &
      &                key     = 'visc_spatial',                               &
      &                path    = fluid_path )

    me%velSpatial(:)%toStore = .false.

    ! load omega
    call aot_get_val(L       = conf,                                           &
      &              thandle = fluid_table,                                    &
      &              key     = 'omega',                                        &
      &              val     = me%omega,                                       &
      &              ErrCode = iError )

    if (btest(iError, aoterr_NonExistent)) then 
      ! omega not defined, try load shear viscosity
      call tem_write( 'omega not defined. Load kine_shear_viscosity')
      call aot_get_val(L = conf, thandle = fluid_table,                         &
        &              key = 'kine_shear_viscosity',                            &
        &              val = me%kine_viscosity,                                 &
        &              ErrCode = iError)
      if (btest(iError, aoterr_NonExistent)) then 
        call tem_write( 'ERROR: specify either omega or kine_shear_viscosity')
        call tem_abort()
      end if
      ! convert physical viscosity to LB viscosity, and calculate omega
      me%kine_viscosityLB = me%kine_viscosity/physics%fac(pdfGlobal%minLevel)%visc
      me%omega = 1._rk / ( cs2inv*me%kine_viscosityLB + 0.5_rk )
    else
      ! calculate LB viscosity from omega, and convert it to physical viscosity
      me%kine_viscosityLB = ( 1._rk/me%omega - 0.5_rk )*cs2
      me%kine_viscosity = me%kine_viscosityLB*physics%fac(pdfGlobal%minLevel)%visc
    endif

    call aot_get_val( L       = conf,                                          &
      &               thandle = fluid_table,                                   &
      &               key     = 'les_cs',                                      &
      &               val     = me%les%c_s,                                    &
      &               default = 0.17_rk,                                       &
      &               ErrCode = iError )

    ! get bulk viscosity and convert it to LB
    call aot_get_val( L       = conf,                                          &
      &               thandle = fluid_table,                                   &
      &               key     = 'bulk_viscosity',                              &
      &               val     = me%bulk_viscosityLB,                           &
      &               ErrCode = iError )
    me%bulk_viscosityLB = me%bulk_viscosity/physics%fac(pdfGlobal%minLevel)%visc

    ! calculate bulk viscosity from kinetic viscosity
    if (btest(iError, aoterr_NonExistent)) then
      me%bulk_viscosityLB = 2._rk/3._rk*me%kine_viscosityLB
      me%bulk_viscosity = me%bulk_viscosityLB                                  &
        &               * physics%fac(pdfGlobal%minLevel)%visc
    endif

    ! compute bulk omega
    me%om_bulk = 2.0_rk/(9.0_rk*me%bulk_viscosityLB + 1.0_rk)

    call aot_get_val( L       = conf,                                          &
      &               thandle = fluid_table,                                   &
      &               key     = 'force',                                       &
      &               val     = me%force,                                      &
      &               default = [0._rk,0._rk,0._rk],                           &
      &               ErrCode = vecError )

    ! convert force to lattice
    me%force = me%force / physics%fac(pdfGlobal%minLevel)%force

    if( maxval( abs( me%force )) .gt. 0._rk )                                  &
      &     call tem_write( ' Forcing: ', me%force )

    ! get reference density
    call aot_get_val( L       = conf,                                          &
      &               thandle = fluid_table,                                   &
      &               key     = 'rho0',                                        &
      &               val     = me%rho0,                                       &
      &               default = 1.0_rk,                                        &
      &               ErrCode = iError )

    ! convert rho0 from physical unit to LB unit
    me%rho0 = me%rho0 / physics%rho0

    ! get temperate
    call aot_get_val( L       = conf,                                          &
      &               thandle = fluid_table,                                   &
      &               key     = 'temp0',                                       &
      &               val     = me%temp0,                                      &
      &               default = 1.0_rk/3._rk,                                  &
      &               ErrCode = iError )

    ! convert temperature to LB unit
    if(physics%temp0 .gt. -1.0_rk) then
      me%temp0 = me%temp0/physics%temp0
    endif

    call tem_write( ' Fluid properties          ' )
    call tem_write( '   Relaxation Parameter Omega:',                          &
      &             real( me%omega ))
    call tem_write( '   Reference density rho0:    ',                          &
      &             real( me%rho0 ))
    call tem_write( '   Shear viscosityPhysical:   ',                          &
      &             real( me%kine_viscosity ))
    call tem_write( '   Shear viscosityLB:         ',                          &
      &             real( me%kine_viscosityLB ))
    call tem_write( '   Bulk viscosity physical:   ',                          &
      &             real( me%bulk_viscosity ))
    call tem_write( '   Bulk viscosityLB:          ',                          &
      &             real( me%bulk_viscosityLB ))
    call tem_write( '   Bulk / Shear relation      ',                          &
      &             real( me%bulk_viscosityLB/me%kine_viscosityLB ))
    call tem_write( '   Smagorinsky Cs             ',                          &
      &             real( me%les%c_s ))
    if( trim(me%omega_ramping%kind) .ne. 'none' )                              &
      & call tem_write( '   Omega ramping active!      ')

    call aot_table_close( L=conf, thandle=fluid_table )
    call tem_horizontalSpacer()

    call tem_write('   Setting omegas and visosities for each level')
    call set_omegas( fluid = me, pdfGlobal = pdfGLobal, physics = physics )

    ! Check if omega is larger than 1.99 or smaller than 1.0
    ! Check if omega is close to 1.0, which may cause shear stress to be 0
    ! double check viscosity
    call check_omegas( fluid = me, pdfGlobal = pdfGLobal, physics = physics )

  end subroutine mus_load_fluid
!******************************************************************************!


!******************************************************************************!
  !> Set the omegas according to the time step setting
  subroutine set_omegas( fluid, pdfGlobal, physics )
    !--------------------------------------------------------------------------
    type( fluid_type ),intent(inout) :: fluid !< fluid type 
    type( pdf_globInfo_type) :: pdfGlobal !< global pdf info
    !> physics type to convert physLB unit
    type( mus_physics_type ), intent(in) :: physics
    !--------------------------------------------------------------------------
    integer :: iLevel, minLevel
    !--------------------------------------------------------------------------
    minLevel = pdfGlobal%minLevel
    fluid%visc = 0._rk

    ! set LB viscosity and omega for each level
    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      fluid%visc( iLevel ) = fluid%kine_viscosityLB *                          &
        &                     physics%fac( minLevel )%visc /                   &
        &                     physics%fac( iLevel )%visc

      !!KM: Omega used in compute kernel is set in mus_setParameters
      fluid%omLvl( iLevel ) = 1._rk / (cs2inv*fluid%visc( iLevel ) + 0.5_rk )
      call tem_write('    level ', iLevel )
      call tem_write('    omega ', real( fluid%omLvl( iLevel )))

      !bulk omega
      fluid%bulkVisc( iLevel ) = fluid%bulk_viscosityLB *                      &
        &                     physics%fac( minLevel )%visc /                   &
        &                     physics%fac( iLevel )%visc
      fluid%omLvl_bulk( iLevel ) = 2.0_rk/(9.0_rk * fluid%bulkVisc( iLevel )   &
        &                        + 1.0_rk ) 
      call tem_write('    bulk omega ', real( fluid%omLvl_bulk( iLevel )))
    end do

  end subroutine set_omegas
!******************************************************************************!


!******************************************************************************!
  !> Check if omega is in the range of 1.0 - 1.99.\n
  !! check if omega is close to 1.0, which may cause shear stress to be 0.\n
  !! And cross check whether omega at each level is set correctly
  !! by computing physical viscosity from omega and check it with 
  !! specified physical kinematic viscosity.\n
  !! nu = (1/omega - 0.5)*cs2
  subroutine check_omegas( fluid, pdfGlobal, physics )
    !---------------------------------------------------------------------------
    type( fluid_type ),intent(inout) :: fluid !< fluid type 
    type( pdf_globInfo_type) :: pdfGlobal !< global pdf info
    !> physics type to convert physLB unit
    type( mus_physics_type ), intent(in) :: physics
    !---------------------------------------------------------------------------
    integer :: iLevel
    real(kind=rk) :: viscPhy_loc, error_rel
    !--------------------------------------------------------------------------

    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      if( fluid%omLvl(iLevel) .gt. 1.99_rk ) then
        call tem_write(' Attention: Omega on level: ', iLevel )
        call tem_write('            is chosen close to the stability limit.')
        call tem_write('            Consider using a higher resolutoin')
        call tem_write('            or a special relaxation model.')
      elseif( fluid%omLvl(iLevel) .lt. 1.0_rk) then
        call tem_write(' Attention: Omega on level: ', iLevel )
        call tem_write('            is chosen to be under-relaxing.')
        call tem_write('            This might cause non-physical behavior.')
      elseif( abs(fluid%omLvl(iLevel) - 1.0_rk ) < 1000 * eps_single_k ) then
        call tem_write(' Attention: Omega on level: ', iLevel )
        call tem_write('            is chosen close to 1.0.')
        call tem_write('            This cause shear stress (strain) to be 0.')
      endif

      ! calculate physical viscosity from omega
      viscPhy_loc = (1.0_rk/fluid%omLvl(iLevel) - 0.5_rk) * cs2 &
        &         * physics%fac(iLevel)%visc

      ! calculate relative error
      error_rel = (fluid%kine_viscosity-viscPhy_loc) / fluid%kine_viscosity
      call tem_write( ' ATTENTION: Relative error between defined physical'// &
        &             '  viscosity to')
      call tem_write( ' computed physical viscosity on level: ', ilevel)
      call tem_write( '                                  is : ', error_rel)
      if( abs(real(error_rel, kind=single_k)) > eps_single_k) then 
        call tem_write( 'Error: Physical kinematic viscosity computed from omega')
        call tem_write( 'does not match with specified physical kinematic viscosity')
        call tem_write( ' Physical kinematic shear viscosity: ' ,fluid%kine_viscosity)
        call tem_write( ' Calculated phy kinematic shear viscosity: ', viscPhy_loc)
        call tem_write( ' Defined omega: ', fluid%omLvl(iLevel))
        call tem_abort()
      end if
    end do ! iLevel

  end subroutine check_omegas
!******************************************************************************!


!******************************************************************************!
  !> write fluid prop into a lua file
  !!
  subroutine mus_fluid_out( me, conf )
    !---------------------------------------------------------------------------
    !> single fluid type
    type( fluid_type ), intent(in) :: me
    type(aot_out_type) :: conf
    !---------------------------------------------------------------------------

    call aot_out_open_table( put_conf = conf, tname = 'fluid' )

    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'rho0',                                       &
      &               val      = me%rho0 )
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'omega',                                      &
      &               val      = me%omega )
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'bulk_viscosityLB',                           &
      &               val      = me%bulk_viscosityLB )
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'kine_viscosityLB',                          &
      &               val      = me%kine_viscosityLB )
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'omLvl',                                      &
      &               val      = me%omLvl )

    call aot_out_close_table( put_conf = conf )
 
  end subroutine mus_fluid_out
!******************************************************************************!


end module mus_fluid_module
!******************************************************************************!
