! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module provides the definition and methods for
!! boundarys.
module mus_global_module

  ! This module has to be created on the fly by the waf environment

  ! include treelm modules
  use env_module,              only: labelLen, rk
  use tem_aux_module,          only: tem_print_execInfo, utc_date_string
  use tem_revision_module,     only: tem_solver_revision
  use tem_comm_module,         only: tem_commPattern_type
  use tem_restart_module,      only: tem_restart_type
  use tem_solveHead_module,    only: tem_solveHead_type
  use tem_timeControl_module,  only: tem_timeControl_type
  use tem_timer_module,        only: tem_addTimer, tem_timer_type
  use tem_tools_module,        only: tem_write
  use tem_balance_module,      only: tem_balance_type
  use tem_status_module,       only: tem_status_type

  ! include aotus modules
  use flu_binding, only: flu_state

  implicit none

  private

  public  mus_init_global 
  public  param_type


  !> Global parameter type defintion, filled with  
  type param_type
    !> aotus lua conf state to load main configuration file
    !! and musubi specific lua functions as chunk
    type(flu_state) :: conf
    ! Handles for timing objects
    !> timer handle for advection relaxation
    integer :: advrelTimer    
    !> timer handle for communication
    integer :: commTimer 
    !> timer handle for interpolation
    integer :: intpTimer_low, intpTimer_hgh 
    !> timer handle for the complete mainloop
    integer :: mainloopTimer
    !> timer handle for initialising the levelDescriptor
    integer :: initLvlDTimer
    !> timer handle for loading / creating the mesh and config
    integer :: loadMeshTimer
    !> timer handle for the output (restart dump)
    integer :: outputTimer
    !> timer handle for source terms
    integer :: sourceTimer
    !> timer handle for checking nans
    integer :: isNanTimer
    !> timer handle for setboundary terms
    integer :: setBndTimer
!> timer handle for levelwise interpolation
    integer, allocatable :: lw_tIntp(:)
    !> timer handle for levelwise computation
    integer, allocatable :: lw_tCompute(:)
    !> timer handle for levelwise set boundary
    integer, allocatable :: lw_tSetBnd(:)
    !> Array of timers for level wise timing
    type(tem_timer_type), allocatable :: timer(:)
    !> weights obtained for load balancing for each level
    real(kind=rk), allocatable :: level_weight(:)
    !> weights obtained for load balancing for each level boundary
    real(kind=rk), allocatable :: level_weight_bnd(:)
    !> time control information, about when to do what
    type(tem_timeControl_type) :: timeControl 
    !> solver Head
    type(tem_solveHead_type) :: solver
    !> global restart type for all schemes
    type(tem_restart_type) :: restart
    !> mpi communication pattern type
    type(tem_commPattern_type)    :: commPattern
    character(len=16) :: revision
    !> time control for the global solver behavior
    type(tem_balance_type) :: balance
    !> type of the control routine
    character(len=labelLen) :: controlRoutine
    !> initialize all elements with valid entries?
    !! This should only be activated for debugging, 
    !! as it needs to be ensured that all helper elements
    !! are filled by communication and interpolation instead
    !! of filling initial values (consider restart!)
    logical :: init_allElems = .false.
    !> Initialize both time layers nNext and nNow during the initialization
    !! phase? This is needed for boundary conditions requiring old time step
    !! information. For debugging multilevel, it should be turned off.
    logical :: init_allTimeLayers = .true.
    !> type of the scaling:
    !! Either 
    !! * acoustic 
    !! * diffusive
    character(len=labelLen) :: scaling
    !> Temporal scaling factor for the scaling. Acoustic = 2, Diffusive = 4
    integer :: scaleFactor
    !> An instance of status type to track status bits
    type(tem_status_type) :: stat
  end type param_type

contains

!******************************************************************************!
  !> Global initialization  routine for whatever 
  !!
  subroutine mus_init_global( params )
    !---------------------------------------------------------------------------
    type(param_type), intent(inout)  :: params
    !---------------------------------------------------------------------------
    character(len=26) :: dat_string
    !---------------------------------------------------------------------------

    params%solver%solver  = 'Musubi'
    params%solver%version = 'v1.0'
    params%revision = tem_solver_revision

    call tem_write("                                                              ")
    call tem_write(" .___  ___.  __    __       _______. __    __  .______    __  ")
    call tem_write(" |   \/   | |  |  |  |     /       ||  |  |  | |   _  \  |  | ")
    call tem_write(" |  \  /  | |  |  |  |    |   (----`|  |  |  | |  |_)  | |  | ")
    call tem_write(" |  |\/|  | |  |  |  |     \   \    |  |  |  | |   _  <  |  | ")
    call tem_write(" |  |  |  | |  `--'  | .----)   |   |  `--'  | |  |_)  | |  | ")
    call tem_write(" |__|  |__|  \______/  |_______/     \______/  |______/  |_"   &
                    //trim(params%solver%version) )
    call tem_write("                                                              ")
    call tem_write(" (C) 2012 German Research School for Simulation Sciences")
    call tem_write("                                                              ")
    call tem_print_execInfo()
    call tem_write("                                                              ")
    dat_string = utc_date_string()
    call tem_write("Run at: "//dat_string)
    call tem_write("                                                              ")
    call tem_write("                                                              ")

    call tem_addTimer( params%mainloopTimer, 'MainLoop' )
    call tem_addTimer( params%advrelTimer,   'Comp'     )
    call tem_addTimer( params%setBndTimer,   'SetBnd'   )
    call tem_addTimer( params%commTimer,     'Comm'     )
    call tem_addTimer( params%intpTimer_hgh, 'Intp_hgh' )
    call tem_addTimer( params%intpTimer_low, 'Intp_low' )
    call tem_addTimer( params%loadMeshTimer, 'LoadMesh' )
    call tem_addTimer( params%initLvlDTimer, 'InitLvlD' )
    call tem_addTimer( params%outputTimer,   'Output'   )
    call tem_addTimer( params%sourceTimer,   'source'   )

  end subroutine mus_init_global
!******************************************************************************!


end module mus_global_module
!******************************************************************************!

