! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module deals with the calculation of moments from pdfs
!!
module mus_moments_module

  ! include treelm modules
  use env_module,         only: rk
  use tem_aux_module,     only: tem_abort
  use tem_tools_module,   only: tem_write
  use tem_math_module,    only: invert_matrix
  use tem_debug_module,   only: tem_debug
  use tem_logging_module, only: tem_log, tem_toStr

  ! include musubi modules
  use mus_moments_type_module,  only: mus_moment_type
  use mus_scheme_layout_module, only: mus_scheme_layout_type

  implicit none
  private

  public :: get_moment
  public :: get_momentVector
  public :: get_momentCentralized
  public :: get_momentVectorCentralized
  public :: mus_init_moments
  public :: init_transformation_matrix
  public :: intf_get_MomentFromPdf
  public :: intf_get_PdfFromMoment
  public :: set_momentIndices     

  abstract interface
    !> Function to calculate the moment from pdfs
    function intf_get_MomentFromPdf( pdf, layout, matrix ) result( mom )
      import :: mus_scheme_layout_type, rk
      !> Stencil layout
      type( mus_scheme_layout_type ), intent(in) :: layout
      !> Pdf state input
      real(kind=rk), intent(in) :: pdf(layout%stencil(1)%QQ)
      !> Transformation matrix
      real(kind=rk), intent(in), optional :: matrix( layout%stencil(1)%QQ,     &
        &                                            layout%stencil(1)%QQ )
      !> moments, obtained by transformation of pdf values
      real(kind=rk) :: mom( layout%stencil(1)%QQ )
    end function

    !> Function to calculate the pdf from moments
    function intf_get_PdfFromMoment( mom, layout, matrix ) result( pdf )
      import :: mus_scheme_layout_type, rk
      !> Stencil layout
      type( mus_scheme_layout_type ), intent(in) :: layout
      !> moment input
      real(kind=rk), intent(in) :: mom( layout%stencil(1)%QQ )
      !> Transformation matrix
      real(kind=rk), intent(in), optional :: matrix( layout%stencil(1)%QQ,     &
        &                                            layout%stencil(1)%QQ )
      !> Pdf output, obtained by back-transformation from moment space via
      !! matrix
      real(kind=rk) :: pdf(layout%stencil(1)%QQ)
    end function
  end interface

 contains

!******************************************************************************!
  !> Initialize the moment space
  !!
  subroutine mus_init_moments( me, layout, requireMoments )
    !--------------------------------------------------------------------------- 
    type( mus_moment_type ) :: me
    type( mus_scheme_layout_type ) :: layout
    logical, intent(in) :: requireMoments
    !--------------------------------------------------------------------------- 
    logical :: testUnit
    !--------------------------------------------------------------------------- 
    if( requireMoments ) then
      testUnit = .false.

      call tem_write('Initializing the moment space ...', layout%stencil(1)%QQ)
      ! Set both entries for the matrix to the stencil size (quadratic matrix)
      me%toMoments%nEntries = layout%stencil(1)%QQ
      me%toPdf%nEntries = layout%stencil(1)%QQ
      if( allocated( me%toMoments%A )) deallocate( me%toMoments%A )
      if( allocated( me%toPdf%A )) deallocate( me%toPdf%A )
      allocate( me%toMoments%A( me%toMoments%nEntries(1),                      &
        &                       me%toMoments%nEntries(2) ))
      allocate( me%toPdf%A( me%toPdf%nEntries(1), me%toPdf%nEntries(2) ))
      ! Allocate and reset the moment labels
      allocate( me%momLabel(layout%stencil(1)%QQ))
      me%momLabel = ''
      call init_transformation_matrix( layout, me, me%toMoments%A, me%toPdf%A )    
      if( testUnit ) &
      call test_transformation_matrix( layout, me%toMoments%A, me%toPdf%A )
    endif

  end subroutine mus_init_moments
!******************************************************************************!


!******************************************************************************!
  !> test the transformation matrices with the known macroscopic moments
  !! density and momentum
  !!
  subroutine test_transformation_matrix( layout, toMoment, toPdf )
    !---------------------------------------------------------------------------
    type(mus_scheme_layout_type), intent(in) :: layout
    real(kind=rk), intent(inout) :: toMoment( layout%stencil(1)%QQ,            &
      &                                       layout%stencil(1)%QQ )
    real(kind=rk), intent(inout) :: toPdf( layout%stencil(1)%QQ,               &
      &                                    layout%stencil(1)%QQ )
    !---------------------------------------------------------------------------
    real(kind=rk) :: tPdf(layout%stencil(1)%QQ)
    real(kind=rk) :: tMoments(layout%stencil(1)%QQ)
    real(kind=rk) :: mx, my, mz, rho
    integer       :: iVal
    !---------------------------------------------------------------------------
    do iVal = 1, layout%stencil(1)%QQ
      tPdf( iVal ) = real(iVal, kind=rk)*0.1_rk
    enddo
    tMoments = matmul( toMoment, tPdf )
    ! Check zero and first moments
    rho = sum( tPdf )
    mx  = sum( tPdf(:)*real(layout%stencil(1)%cxDir(1,:), kind=rk))
    my  = sum( tPdf(:)*real(layout%stencil(1)%cxDir(2,:), kind=rk))
    mz  = sum( tPdf(:)*real(layout%stencil(1)%cxDir(3,:), kind=rk))
    call tem_debug( level = 1,                                                 &
      &             msg   = 'pdf     '//trim( tem_toStr( val       = tPdf,     &
      &                                                  sep       = ',' )))
    call tem_debug( level = 1,                                                 &
      &             msg   = 'moments '//trim( tem_toStr( val       = tMoments, &
      &                                                  sep       = ',' )))
    call tem_debug( level = 1,                                                 &
      &             msg   = 'ref Val '//                                       &
      &                     trim( tem_toStr( rho ))//' '//                     &
      &                     trim( tem_toStr( mx ))//' '//                      &
      &                     trim( tem_toStr( my ))//' '//                      &
      &                     trim( tem_toStr( mz ))//' ' )
    ! and transform back to pdf space
    tPdf = matmul( toPdf, tMoments)
    call tem_debug( level = 1,                                                 &
      &             msg   = 'pdf transformed back'//                           &
      &                     trim( tem_toStr( val       = tPdf,                 &
      &                                      sep       = ',' )))

  end subroutine test_transformation_matrix
!******************************************************************************!


!******************************************************************************!
  !> Intialize the matrix to convert the complete pdf space to moments and back
  !!
  subroutine init_transformation_matrix( layout, me, toMoment, toPdf )
    !---------------------------------------------------------------------------
    type(mus_scheme_layout_type), intent(in) :: layout
    type( mus_moment_type ), intent(inout) :: me
    real(kind=rk), intent(inout) :: toMoment( layout%stencil(1)%QQ,            &
      &                                       layout%stencil(1)%QQ )
    real(kind=rk), intent(inout) :: toPdf( layout%stencil(1)%QQ,               &
      &                                    layout%stencil(1)%QQ )
    !---------------------------------------------------------------------------
    integer :: iVal   
    !---------------------------------------------------------------------------
    ! Set the moments one after another on the matrix. 
    ! The moment description might come from some place else.
    ! Lua? -> check mus_scheme_type_module the type(mus_momentSpace_type)
    select case( trim(layout%stencil(1)%label) )
    case('d2q9')
      me%momLabel(1) = 'density'
      toMoment( 1,:) = get_momentVector( layout, (/0,0,0/)) ! m0
      me%momLabel(2) = 'momX'
      toMoment( 2,:) = get_momentVector( layout, (/1,0,0/)) ! mx
      me%momLabel(3) = 'momY'
      toMoment( 3,:) = get_momentVector( layout, (/0,1,0/)) ! my
      me%momLabel(4) = 'Sxx'
      toMoment( 4,:) = get_momentVector( layout, (/2,0,0/)) ! mxx
      me%momLabel(5) = 'Syy'
      toMoment( 5,:) = get_momentVector( layout, (/0,2,0/)) ! myy
      me%momLabel(6) = 'Sxy'
      toMoment( 6,:) = get_momentVector( layout, (/1,1,0/)) ! mxy
      me%momLabel(7) = 'mxxy'
      toMoment( 7,:) = get_momentVector( layout, (/2,1,0/)) ! mxxy
      me%momLabel(8) = 'mxyy'
      toMoment( 8,:) = get_momentVector( layout, (/1,2,0/)) ! mxyy
      me%momLabel(9) = 'mxxyy'
      toMoment( 9,:) = get_momentVector( layout, (/2,2,0/)) ! mxxyy
    case('d2q9_reference')
      toMoment( 1,:) = (/ 1._rk,  1._rk,  1._rk,  1._rk,  1._rk,  1._rk,  1._rk,  1._rk,  1._rk /)
      toMoment( 2,:) = (/-1._rk,  0._rk,  1._rk,  0._rk, -1._rk, -1._rk,  1._rk,  1._rk,  0._rk /)
      toMoment( 3,:) = (/ 0._rk, -1._rk,  0._rk,  1._rk, -1._rk,  1._rk, -1._rk,  1._rk,  0._rk /)
      toMoment( 4,:) = (/ 1._rk,  0._rk,  1._rk,  0._rk,  1._rk,  1._rk,  1._rk,  1._rk,  0._rk /)
      toMoment( 5,:) = (/ 0._rk,  1._rk,  0._rk,  1._rk,  1._rk,  1._rk,  1._rk,  1._rk,  0._rk /)
      toMoment( 6,:) = (/ 0._rk,  0._rk,  0._rk,  0._rk,  1._rk, -1._rk, -1._rk,  1._rk,  0._rk /)
      toMoment( 7,:) = (/ 0._rk,  0._rk,  0._rk,  0._rk, -1._rk,  1._rk, -1._rk,  1._rk,  0._rk /)
      toMoment( 8,:) = (/ 0._rk,  0._rk,  0._rk,  0._rk, -1._rk, -1._rk,  1._rk,  1._rk,  0._rk /)
      toMoment( 9,:) = (/ 0._rk,  0._rk,  0._rk,  0._rk,  1._rk,  1._rk,  1._rk,  1._rk,  0._rk /)
    case('d3q19')
      me%momLabel(1) = 'density'
      toMoment( 1,:) = get_momentVector( layout, (/0,0,0/)) ! m0
      me%momLabel(2) = 'momX   '
      toMoment( 2,:) = get_momentVector( layout, (/1,0,0/)) ! mx
      me%momLabel(3) = 'momY   '
      toMoment( 3,:) = get_momentVector( layout, (/0,1,0/)) ! my
      me%momLabel(4) = 'momZ   '
      toMoment( 4,:) = get_momentVector( layout, (/0,0,1/)) ! mz
      me%momLabel(5) = 'momXX  '
      toMoment( 5,:) = get_momentVector( layout, (/2,0,0/)) ! mxx
      me%momLabel(6) = 'momYY  '
      toMoment( 6,:) = get_momentVector( layout, (/0,2,0/)) ! myy
      me%momLabel(7) = 'momZZ  '
      toMoment( 7,:) = get_momentVector( layout, (/0,0,2/)) ! mzz
      me%momLabel(8) = 'momXY  '
      toMoment( 8,:) = get_momentVector( layout, (/1,1,0/)) ! mxy
      me%momLabel(9) = 'momYZ  '
      toMoment( 9,:) = get_momentVector( layout, (/0,1,1/)) ! myz
      me%momLabel(10) = 'momXZ'
      toMoment(10,:) = get_momentVector( layout, (/1,0,1/)) ! mzx
      me%momLabel(11) = 'momXXY '
      toMoment(11,:) = get_momentVector( layout, (/2,1,0/)) ! mxxy
      me%momLabel(12) = 'momXXZ '
      toMoment(12,:) = get_momentVector( layout, (/2,0,1/)) ! mxxz
      me%momLabel(13) = 'momYYX '
      toMoment(13,:) = get_momentVector( layout, (/1,2,0/)) ! myyx
      me%momLabel(14) = 'momYYZ '
      toMoment(14,:) = get_momentVector( layout, (/0,2,1/)) ! myyz
      me%momLabel(15) = 'momZZX '
      toMoment(15,:) = get_momentVector( layout, (/1,0,2/)) ! mzzx
      me%momLabel(16) = 'momZZY '
      toMoment(16,:) = get_momentVector( layout, (/0,1,2/)) ! mzzy
      me%momLabel(17) = 'momXXYY'
      toMoment(17,:) = get_momentVector( layout, (/2,2,0/)) ! mxxyy
      me%momLabel(18) = 'momYYZZ'
      toMoment(18,:) = get_momentVector( layout, (/0,2,2/)) ! myyzz
      me%momLabel(19) = 'momZZXX'
      toMoment(19,:) = get_momentVector( layout, (/2,0,2/)) ! mzzxx
    case default
      call tem_write('Moment definition for this stencil has not been defined yet.')
      call tem_abort
    end select
    call tem_log( level = 2,                                                   &
      &           msg   = 'matrix size'//trim(tem_toStr(size(toMoment,1)))//   &
      &                   ' '//trim( tem_toStr( size( toMoment, 2))))
    do iVal = 1, size( toMoment,2)
      call tem_log( level = 2,                                                 &
        &           msg   = trim( tem_toStr( val=toMoment(:,iVal), sep=',')))
    end do
    toPdf = invert_matrix( toMoment )
   
  end subroutine init_transformation_matrix
!******************************************************************************!


!******************************************************************************!
  !> The integer moment vector for a given cxDir and order.
  !!
  !! Assuming 0**0 = 1 here.
  !!
  pure function mus_iMomVector( cxDir, expX, QQ ) result(iMom)
    !---------------------------------------------------------------------------
    !>
    integer, intent(in) :: QQ
    !>
    integer, intent(in) :: cxDir(:,:)
    !>
    integer, intent(in) :: expX(3)
    !>
    integer :: iMom(QQ)
    !---------------------------------------------------------------------------
    integer :: nIndices
    integer :: iVal
    integer :: iX
    integer :: X_idx(3)
    integer :: xx(3)
    !---------------------------------------------------------------------------

    nIndices = 0
    do iX=1,3
      if (expX(iX) > 0) then
        nIndices = nIndices + 1
        X_idx(nIndices) = iX
        xx(nIndices) = expX(iX)
      end if
    end do

    select case(nIndices)
    case(0)
      imom = 1
    case(1)
      do iVal=1,QQ
        imom(iVal) = cxDir(X_idx(1),iVal)**xx(1)
      end do
    case(2)
      do iVal=1,QQ
        imom(iVal) = cxDir(X_idx(1),iVal)**xx(1)                               &
          &        * cxDir(X_idx(2),iVal)**xx(2)
      end do
    case(3)
      do iVal=1,QQ
        imom(iVal) = cxDir(1,iVal)**xx(1)                                      &
          &        * cxDir(2,iVal)**xx(2)                                      &
          &        * cxDir(3,iVal)**xx(3)
      end do
    end select

  end function mus_iMomVector
!******************************************************************************!


!******************************************************************************!
  !> The integer moment vector for a given cxDir and order.
  !!
  !! Assuming 0**0 = 1 here.
  !!
  pure function mus_rMomVector(cxDir, expX, QQ) result(rMom)
    !---------------------------------------------------------------------------
    integer, intent(in) :: QQ
    real(kind=rk), intent(in) :: cxDir(:,:)
    integer, intent(in) :: expX(3)
    real(kind=rk) :: rMom(QQ)
    !---------------------------------------------------------------------------
    integer :: nIndices
    integer :: iVal
    integer :: iX
    integer :: X_idx(3)
    integer :: xx(3)
    !---------------------------------------------------------------------------

    nIndices = 0
    do iX=1,3
      if (expX(iX) > 0) then
        nIndices = nIndices + 1
        X_idx(nIndices) = iX
        xx(nIndices) = expX(iX)
      end if
    end do

    select case(nIndices)
    case(0)
      rmom = 1.0_rk
    case(1)
      do iVal=1,QQ
        rmom(iVal) = cxDir(X_idx(1),iVal)**xx(1)
      end do
    case(2)
      do iVal=1,QQ
        rmom(iVal) = cxDir(X_idx(1),iVal)**xx(1)                               &
          &        * cxDir(X_idx(2),iVal)**xx(2)
      end do
    case(3)
      do iVal=1,QQ
        rmom(iVal) = cxDir(1,iVal)**xx(1)                                      &
          &        * cxDir(2,iVal)**xx(2)                                      &
          &        * cxDir(3,iVal)**xx(3)
      end do
    end select

  end function mus_rMomVector
!******************************************************************************!


!******************************************************************************!
  !> set moments of arbitary moment
  !!
  pure function get_moment( pdf, layout, expX ) result( mom )
    !---------------------------------------------------------------------------
    !> stencil layout
    type( mus_scheme_layout_type), intent(in)   :: layout
    !> source values of the momentum on the square corners
    real(kind=rk), intent(in) :: pdf(layout%stencil(1)%QQ)
    integer, intent(in) :: expX(3)      !< exponents of the moments
    real(kind=rk) :: mom                !< interpolated value
    !---------------------------------------------------------------------------

    !HK: assuming 0**0 should be 1 here
    mom = sum( pdf * mus_iMomVector(cxDir = layout%stencil(1)%cxDir,           &
      &                            expX   = expX,                              &
      &                            QQ     = layout%stencil(1)%QQ) )

  end function get_moment
!******************************************************************************!


!******************************************************************************!
  !> get the moment vector to calculate the moment from the pdf
  !!
  pure function get_momentVector( layout, expX ) result( mom )
    !---------------------------------------------------------------------------
    !> stencil layout
    type( mus_scheme_layout_type), intent(in)   :: layout
    ! source values of the momentum on the square corners
    !> exponents of the moments
    integer, intent(in) :: expX(3)
    !> moment vector
    real(kind=rk) :: mom(layout%stencil(1)%QQ)
    !---------------------------------------------------------------------------

    !HK: assuming 0**0 should be 1 here
    Mom = real( mus_iMomVector(cxDir = layout%stencil(1)%cxDir,                &
      &                        expX  = expX,                                   &
      &                        QQ    = layout%stencil(1)%QQ),                  &
      &         kind = rk )

    !>\todo HK: There occure 0**0 here, what is supposed to
    !!          evaluate to in this case?
!!    do iVal = 1, layout%stencil(1)%QQ
!!      mom( iVal ) = &
!!      real(layout%stencil(1)%cxDir(1,iVal)**expX(1)           &
!!       *   layout%stencil(1)%cxDir(2,iVal)**expX(2)           &
!!       *   layout%stencil(1)%cxDir(3,iVal)**expX(3), kind=rk) 
!!    end do

  end function get_momentVector
!******************************************************************************!


!******************************************************************************!
  !> get centralized moment of arbitrary order
  !!
  pure function get_momentCentralized( pdf, vel, layout, expX ) result( mom )
    !---------------------------------------------------------------------------
    !> stencil layout
    type( mus_scheme_layout_type), intent(in)   :: layout
    !> source values of the momentum on the square corners
    real(kind=rk), intent(in) :: pdf(layout%stencil(1)%QQ)
    !> exponents of the moments
    integer, intent(in) :: expX(3)
    !> exponents of the moments
    real(kind=rk), intent(in) :: vel(3)
    !> interpolated value
    real(kind=rk) :: mom
    !---------------------------------------------------------------------------
!    integer :: iVal         
    !---------------------------------------------------------------------------
    mom = sum(pdf * get_momentVectorCentralized(vel, layout, expX))

!HK: What about 0**0 here?
!!    mom = 0._rk
!!    do iVal = 1, layout%stencil(1)%QQ
!!      mom = mom +  &
!!        (real(layout%stencil(1)%cxDir(1,iVal),kind=rk)-vel(1))**expX(1) &
!!       *(real(layout%stencil(1)%cxDir(2,iVal),kind=rk)-vel(2))**expX(2) &
!!       *(real(layout%stencil(1)%cxDir(3,iVal),kind=rk)-vel(3))**expX(3) &
!!       * pdf( iVal )
!!    enddo
  end function get_momentCentralized
!*******************************************************************************


!*******************************************************************************
  !> get the moment vector to calculate the moment from the pdf
  !!
  pure function get_momentVectorCentralized( vel, layout, expX ) result( mom )
    !---------------------------------------------------------------------------
    !> stencil layout
    type( mus_scheme_layout_type), intent(in)   :: layout
    !> source values of the momentum on the square corners
    !! exponents of the moments
    integer, intent(in) :: expX(3)
    !>
    real(kind=rk), intent(in) :: vel(3)
    !> moment vector
    real(kind=rk) :: mom(layout%stencil(1)%QQ)
    !---------------------------------------------------------------------------
    integer :: iVal         
    real(kind=rk) :: c_cxDir(3,layout%stencil(1)%QQ)
    !---------------------------------------------------------------------------

    do iVal=1,layout%stencil(1)%QQ
      c_cxDir(:,iVal) = layout%stencil(1)%cxDir(:,iVal) - vel
    end do

    !HK: assuming 0**0 should be 1 here
    Mom = mus_rMomVector(cxDir = c_cxDir,                                      &
      &                  expX  = expX,                                         &
      &                  QQ    = layout%stencil(1)%QQ)

!HK: What should be done for 0**0 here?
!!    do iVal = 1, layout%stencil(1)%QQ
!!      mom( iVal ) = &
!!        (real(layout%stencil(1)%cxDir(1,iVal),kind=rk)-vel(1))**expX(1) &
!!       *(real(layout%stencil(1)%cxDir(2,iVal),kind=rk)-vel(2))**expX(2) &
!!       *(real(layout%stencil(1)%cxDir(3,iVal),kind=rk)-vel(3))**expX(3) 
!!    enddo
  end function get_momentVectorCentralized
!******************************************************************************!


!******************************************************************************!
  !> set indices for accessing the pressure, velocity and the shear from a 1d
  !! vector
  !!
  subroutine set_momentIndices( nDims, iPress, iVelMin, iVelMax,               &
    &                           iStressMin, iStressMax )
    !---------------------------------------------------------------------------
    !> number of dimensions
    integer, intent(in) :: nDims
    !> index for the pressure / density
    integer, intent(out) :: iPress
    !> starting index for velocity
    integer, intent(out) :: iVelMin
    !> ending index for velocity
    integer, intent(out) :: iVelMax
    !> starting index for shear
    integer, intent(out) :: iStressMin
    !> ending index for shear
    integer, intent(out) :: iStressMax
    !---------------------------------------------------------------------------

    ! Pressure / density is always first entry
    iPress = 1
    ! Indices from where / to where to read / store the velocity quantities
    iVelMin = 2
    iVelMax = nDims+1
    ! Indices from where / to where to read / store the stress   quantities
    iStressMin = iVelMax+1
    iStressMax = iStressMin+max(3*(nDims-1)-1, 0)

  end subroutine set_momentIndices
!******************************************************************************!


end module mus_moments_module
!******************************************************************************!
