!******************************************************************************!
!> This module keeps all information about the nonNewtonian models
!!
!!
module mus_nonNewtonian_module

  ! include treelm modules
  use env_module,           only: rk, single_k, eps_single_k, globalMaxLevels
  use treelmesh_module,     only: treelmesh_type
  use tem_tools_module,     only: tem_write, tem_horizontalSpacer
  use tem_aux_module,       only: tem_abort
  use tem_param_module,     only: cs2, cs2inv
  use tem_transient_module, only: tem_transient_type, load_transient
  use tem_spatial_module,   only: tem_spatial_type, load_spatial  
  use tem_debug_module,     only: main_debug

  ! include aotus modules
  use aotus_module,     only: flu_State, aot_get_val, aoterr_Fatal,            &
    &                         aoterr_NonExistent, aoterr_WrongType
  use aot_table_module, only: aot_table_open, aot_table_close
  use aot_path_module,  only: aot_path_type, assignment(=), aot_path_addNode
  use aot_out_module,   only: aot_out_type, aot_out_val, aot_out_open_table,   &
    &                         aot_out_close_table

  ! include musubi modules
  use mus_physics_module, only: mus_physics_type
  use mus_pdf_module,     only: pdf_globInfo_type

  implicit none
  private

  public :: mus_nonNewtonian_type
  public :: mus_load_nonNewtonian
  public :: mus_out_nonNewtonian

  type mus_nonNewtonian_type
    logical :: active = .false.
    real(kind=rk) :: visc0
    real(kind=rk) :: viscInf
    real(kind=rk) :: lambda
    real(kind=rk) :: a
    real(kind=rk) :: n
  end type mus_nonNewtonian_type

contains

!******************************************************************************!
  !> Read in the nonNewtonian from the LUA parameter file
  subroutine mus_load_nonNewtonian( me, conf, parent, path )
    !--------------------------------------------------------------------------
    !> nonNewtonian type
    type( mus_nonNewtonian_type ),intent(out) :: me
    !> lua state
    type( flu_state ) :: conf
    !> parent handle
    integer, intent(in), optional  :: parent
    !> lua path to transient lua function
    type(aot_path_type), intent(in) :: path
    !--------------------------------------------------------------------------
    integer :: nonNwt_table
    integer :: iError 
    type(aot_path_type) :: nonNwt_path !< lua path to transient lua function
    CHARACTER(LEN=12) :: nonNwt_table_str
    !---------------------------------------------------------------------------

    nonNwt_path = path
    nonNwt_table_str = "nonNewtonian"

    ! if nonNewtonian informations in scheme table parentHandle /= 0
    if( present(parent) ) then 
      call aot_table_open( L       = conf,                                     &
        &                  parent  = parent,                                   &
        &                  thandle = nonNwt_table,                             &
        &                  key     = nonNwt_table_str )
    else
      call aot_table_open( L=conf, thandle = nonNwt_table, key = nonNwt_table_str )
    end if

    call aot_path_addNode( me = nonNwt_path, NodeType = 'table', &
      &                    key = nonNwt_table_str )

    if ( nonNwt_table .eq. 0 ) then
      ! no table found
      call tem_write(' No nonNewtonian table defined')
    else
      ! Activating the current fluid object
      me%active = .true.
      ! load visc0
      call aot_get_val(L       = conf,                                           &
        &              thandle = nonNwt_table,                                   &
        &              key     = 'visc0',                                        &
        &              val     = me%visc0,                                       &
        &              default = 0.16_rk,                                       &
        &              ErrCode = iError )
  
      ! load viscInf
      call aot_get_val( L       = conf,                                          &
        &               thandle = nonNwt_table,                                  &
        &               key     = 'viscInf',                                     &
        &               val     = me%viscInf,                                    &
        &               default = 0.0035_rk,                                     &
        &               ErrCode = iError )
  
      ! load lambda
      call aot_get_val( L       = conf,                                          &
        &               thandle = nonNwt_table,                                  &
        &               key     = 'lambda',                                     &
        &               val     = me%viscInf,                                    &
        &               default = 8.2_rk,                                     &
        &               ErrCode = iError )
  
      call aot_get_val( L       = conf,                                          &
        &               thandle = nonNwt_table,                                  &
        &               key     = 'a',                                     &
        &               val     = me%a,                                    &
        &               default = 0.64_rk,                                     &
        &               ErrCode = iError )
  
      call aot_get_val( L       = conf,                                          &
        &               thandle = nonNwt_table,                                  &
        &               key     = 'n',                                     &
        &               val     = me%n,                                    &
        &               default = 0.2128_rk,                                     &
        &               ErrCode = iError )
  
      call tem_write( ' nonNewtonian fluid parameters:' )
      call tem_write( '   visc0: ', me%visc0 )
    endif

    call aot_table_close( L=conf, thandle=nonNwt_table )
    call tem_horizontalSpacer()

  end subroutine mus_load_nonNewtonian
!******************************************************************************!


!******************************************************************************!
!******************************************************************************!


!******************************************************************************!
!******************************************************************************!


!******************************************************************************!
  !> write fluid prop into a lua file
  !!
  subroutine mus_out_nonNewtonian( me, conf )
    !---------------------------------------------------------------------------
    !> single fluid type
    type( mus_nonNewtonian_type ), intent(in) :: me
    type(aot_out_type) :: conf
    !---------------------------------------------------------------------------

    call aot_out_open_table( put_conf = conf, tname = 'nonNewtonian' )

    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'visc0',                                      &
      &               val      = me%visc0 )
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'viscInf',                                    &
      &               val      = me%viscInf )
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'lambda',                                     &
      &               val      = me%lambda )
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'a',                                          &
      &               val      = me%a )
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'n',                                          &
      &               val      = me%n )

    call aot_out_close_table( put_conf = conf )
 
  end subroutine mus_out_nonNewtonian
!******************************************************************************!


end module mus_nonNewtonian_module
!******************************************************************************!
