! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!?? include 'header/lbm_macros.inc'
!> Definitions for the main state and neighbor arrays
!!
module mus_pdf_module

  ! include treelm modules
  use env_module,              only: rk, globalMaxLevels, long_k
  use tem_construction_module, only: tem_levelDesc_type
  use tem_comm_module,         only: tem_communication_type
  use tem_time_module,         only: tem_time_type

  ! include musubi modules
  use mus_statistics_module, only: mus_stat_type

  implicit none
  private

  public :: pdf_data_type
  public :: pdf_global_type
  public :: pdf_globInfo_type

  !> This data type contains the Musubi-specific arrays which 
  !! are fed into the kernel. The solver updates the state vector
  !! and finds the position of the neighbor vectors by looking it 
  !! up in the neigh array
  type pdf_data_type
    ! Per level element numbers
    !> number of local fluid elements (inner+exposed)
    integer :: nElems_fluid
    !> fluid elements + ghost elements (elements for solver)
    integer :: nElems_solve
    !> number of halo elements (from remote processes)
    integer :: nElems_halo
    !> number of ghost elements (from other levels)
    integer :: nElems_ghost
    !> number of ghost elements (from other levels)
    integer :: nElems_ghostFromCoarser
    !> number of ghost elements (from other levels)
    integer :: nElems_ghostFromFiner
    !> number of local elements (inner+exposed+ghost)
    integer :: nElems_local
    !> number of inner fluid elements 
    integer :: nElems_inner
    !> number of exposed fluid elements    
    integer :: nElems_exposed

    !> which buffer to use for current time step
    integer :: nNow  = 1
    !> which buffer to use for next time step 
    integer :: nNext = 2

    !> Data vector containing the pdf state 
    !! Double buffering requires two arrays
    !! Size: scheme%varSys%nScalars * nElems_local, 2
    real(kind=rk),allocatable,dimension(:,:) :: state

    !> Additional values for each element
    !! Size: glob%nAuxVars, nElems_local
    real(kind=rk),allocatable,dimension(:,:) :: aux

    !> Connectivity array 
    !! Points to where to send respective pdfs 
    !! Access in a neigh way
    integer,allocatable,dimension(:) :: neigh

    !> containing state vector values of elements which has boundary
    real(kind=rk),allocatable, dimension(:) :: bcBuffer

    !> MPI Buffer for communication on the same level
    type( tem_communication_type ) :: buffer
    !> MPI Buffer for communication to finer level
    type( tem_communication_type ) :: buffer_coarseToFine
    !> MPI Buffer for communication to coarser level
    type( tem_communication_type ) :: buffer_fineToCoarse
    !> Which halo requires all informations?
    logical, allocatable :: haloRequired(:)

  end type pdf_data_type

  !> Definition of a block containing the 1D vector and more information 
  type :: pdf_global_type
    !> number of total fluid elements (inner+exposed elements)
    integer(kind=long_k) :: nElems_totalFluid
    !> number of to-solve elements (inner+exposed+ghost elements)
    integer(kind=long_k) :: nElems_totalSolve
    !> number of total fluid elements per level
    integer(kind=long_k) :: nElems_totalLevel( globalMaxLevels ) 
    !> number of total elements per level
    integer(kind=long_k) :: nElems_allLevel( globalMaxLevels )
    !> number of elements which have to be solved
    integer(kind=long_k) :: nElems_solveLevel( globalMaxLevels ) 

    !> Local number of fluid elements
    integer :: nElems_localFluid
    !> Local number of elements to solve
    integer :: nElems_localSolve
    !> number of total local elements (inner+exposed+ghost elements)
    integer(kind=long_k) :: nElems_overall
    !> number of total local elements (inner+exposed+ghost elements)
    integer(kind=long_k) :: nElems_total

    real(kind=rk) :: totalDens !< overall integrated density

    !> index of the first fluid cell in the 
    !! local partition using global coordinates
    integer :: globalPosLocal

    !> Detailed information about the level-wise treeID lists
    type(tem_levelDesc_type),allocatable  :: levelDesc(:)

  end type pdf_global_type

  !> This datatype contains all scheme independent informations
  type :: pdf_globInfo_type
    !> Nesting: 1: acoustic, 3: diffusive
    integer :: nNesting = 1

    !> number of dimensions for current run
    !! standard: 3, but if a 2d stencil was chosen, set to 2
    integer :: nDims = 3
    type(tem_time_type)   :: timing        !< global time information
    integer :: nSchemes = 0  ! total number of used schemes
    !> lattice unit time steps
    real(kind=rk) :: dtLB(globalMaxLevels) = -1._rk
    !> lattice unit spatial steps
    real(kind=rk) :: dxLB(globalMaxLevels) = -1._rk
    integer :: nLevels   !< number of levels    
    integer :: minLevel  !< global minimal level
    integer :: maxLevel  !< global maximal level
    !> is this level being updated at the current time step?
    !! Not all levels get updated with the nested time stepping
    logical :: levelActive( globalMaxLevels ) = .false.
    integer :: nBCs      !< number of boundary conditions
    integer :: nBCs_local !< number of local boundary conditions
    !> pointer from the original treeIDlist to the levelwise fluid list
    !! it has a size of tree%nElems
    integer, allocatable :: levelPointer(:)
    type( mus_stat_type ) :: stat
    logical :: comm_reduced = .false. !< Communicate all links?
  end type pdf_globInfo_type

end module mus_pdf_module
!******************************************************************************!


!******************************************************************************!
!>\page datastructures Data structures
!! The Octree data structure is mapped to a one-dimensional array in order to
!! have an efficient data structure on which the solver can act on in a 
!! performant way. An efficient representation of the elements and their 
!! neighbor relations is chosen.
!! The fluid elements are mapped to a one-dimensional array and the neighboring 
!! relations are introduced by an additional connectivity array. The access of a
!! neighbor element is performed by looking up the correct position of an 
!! element's link neighbor in the connectivity array, thus constituting an
!! indirect access.
!!
!! \image html  statevector_neighborlist.png
!!
!! The different dependencies of each link in an element require a thorough
!! treatment, when data is exchanged at domain boundaries. Only the links, which
!! point outside the domain have to be sent to neighbor partitions, and the 
!! links pointing inwards have to be filled with valid values from these.
