! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module contains data type and modules related to musubi
!! lattice to physical unit convertion and vice versa. \n
!! physics data type is defined for each scheme, it is defined
!! in scheme table as the following format: \n
!! physics = { setPhysicalQuantities = true,
!!             dt = dt_phy,
!!             rho0 = rho0_phy,
!!             viscosity = vis0_phy,
!!             bulk_viscosity = bulk_vis0_phy,
!!             molWeigh0 = mol_phy,
!!             temp0 = t_phy, } \n
!! Of these quantities, dt is mandetory for conversion.
!! Others can be omitted, thus use default values.
!! \author Kannan Masilamani
module mus_physics_module

  ! include treelm modules
  use env_module,          only: rk
  use tem_aux_module,      only: tem_abort
  use tem_tools_module,    only: tem_write, tem_horizontalSpacer
  use treelmesh_module,    only: treelmesh_type
  use tem_geometry_module, only: tem_ElemSizeLevel
  use tem_logging_module,  only: tem_log, tem_toStr

  ! include aotus modules
  use aotus_module,     only: flu_State, aot_get_val, aoterr_Fatal,            &
    &                         aoterr_NonExistent, aoterr_WrongType, flu_State
  use aot_table_module, only: aot_table_open, aot_table_close, aot_get_val
  use aot_out_module,   only: aot_out_type, aot_out_open_table,                &
    &                         aot_out_close_table, aot_out_val

  implicit none

  private

  public :: mus_physics_type
  public :: mus_load_physics
  public :: mus_create_funcStr
  public :: mus_physics_out

  !> This type contains the converstion factor for derived variables
  !! from lattice to physical.\n
  !! Their inverses can be used to convert physical to lattice units
  type mus_convertFac_type
    !> length(m) = dx
    real(kind=rk) :: length
    !> time (s) = dt
    real(kind=rk) :: time
    !> temperature (K)
    real(kind=rk) :: temp
    !> velocity(m/s) = dx/dt
    real(kind=rk) :: vel
    !> kinematic viscosity(m^2/s) = dx^2/dt
    real(kind=rk) :: visc
    !> dynamic viscosity(kg/m/s) = rho0*dx^2/dt
    real(kind=rk) :: dynVisc
    !> acceleration(m/s^2) = dx/dt^2
    real(kind=rk) :: accel
    !> Force(N)(kg m/s^2) = rho0*dx^4/dt^2
    real(kind=rk) :: force
    !> Pressure(N/m^2)(kg/m/s^2) = rho0*dx^2/dt^2
    real(kind=rk) :: press
    !> Strain Rate (1/s) = 1/dt
    real(kind=rk) :: strainRate
    !> specific Energy (rho*m^2/s^2) = rho0*dx^2/dt^2
    real(kind=rk) :: energy
    !> mass(kg) = rho0*dx^3
    real(kind=rk) :: mass
    !> mole(mol) = mass/molWeigh0
    real(kind=rk) :: mole
    !> mole flux(mol) = mol/m^2/s
    real(kind=rk) :: moleFlux
    !> mass flux(mol) = kg/m^2/s
    real(kind=rk) :: flux
    !> diffusivity(m^2/s) = dx^2/dt
    real(kind=rk) :: diffusivity
!    !> resistivity(s/m^2) = dt/dx^2
!    !! Inverse of diffusivity
!    real(kind=rk) :: resistivity
    ! newton, faraday, gas constant
    !> coulomb (C) = faraday* mole
    !! calcaulated from ref specific charge and mole
    real(kind=rk) :: coulomb
  end type mus_convertFac_type

  !> This type contains the basic SI units used to convert
  !! lattice to physical unit and vice versa
  type mus_physics_type
    !> if the conversion from physics to lattice is active
    !> reference length - discretization size of the coarsest level
    !! SI unit - meter
    real(kind=rk) :: dx = -1.0_rk
    real(kind=rk), allocatable :: dxLvl(:)
    !> reference time - time discretization for discretization size 
    !! of the coarsest level
    !! SI unit - seconds
    real(kind=rk) :: dt = -1.0_rk
    real(kind=rk), allocatable :: dtLvl(:)
    !> reference physical density
    !! SI unit - kg/m^3
    real(kind=rk) :: rho0 = -1.0_rk
    !> reference molecular weight
    !! SI unit - kg/mol
    real(kind=rk) :: molWeigh0 = -1.0_rk
    !> reference temperature
    !! SI unit - Kelvin
    real(kind=rk) :: temp0 = -1.0_rk
    !> conversion factor for derived variables
    !! This must exist for each level
    type( mus_convertFac_type ), allocatable :: fac(:)
    !> shear viscosity
    !! SI unit - m^2/s
    real(kind=rk) :: viscosity0 = -1._rk
    !> bulk viscosity
    !! SI unit - m^2/s
    real(kind=rk) :: bulk_viscosity0 = -1._rk
    !! faraday constant
    !! SI unit - C/mol
    real(kind=rk) :: faraday0 = 96485.3365_rk
  end type mus_physics_type

  contains

  !****************************************************************************!
  !> This routine loads the physics table inside each scheme table from
  !! musubi config file
  !!
  subroutine mus_load_physics( me, conf, parent, tree, scaling, dtRef, dxRef )
    !---------------------------------------------------------------------------
    !> physics type
    type( mus_physics_type ), intent(out) :: me
    !> flu state
    type( flu_State ) :: conf
    !> scheme handle
    integer, optional, intent(in) :: parent
    !> global treelm mesh
    type( treelmesh_type), intent(in) :: tree
    !> scaling: diffusive or acoustic?
    character(len=*), intent(in) :: scaling
    !> reference time step if none 
    real(kind=rk), optional, intent(in) :: dtRef
    !> reference spacestep if none 
    real(kind=rk), optional, intent(in) :: dxRef
    !---------------------------------------------------------------------------
    integer :: thandle
    integer :: iError, iLevel
    real(kind=rk) :: dt_Ref, dx_ref ! reference step if none 
    !---------------------------------------------------------------------------
    if( present( dtRef )) then
      dt_ref = dtRef
    else
      dt_ref = 1._rk 
    end if

    if( present( dxRef )) then
      dx_ref = dxRef
    else
      dx_ref = 1._rk 
    end if

    ! reference dx is always the coarsest level in the tree
    me%dx = real( tem_ElemSizeLevel( tree, tree%global%minLevel ), kind=rk)

    call tem_write(' Loading physics table ')
    if( present(parent) ) then
      call aot_table_open( L=conf, parent = parent, thandle=thandle,           &
        & key='physics' )
    else
      call aot_table_open( L=conf, thandle=thandle, key='physics' )
    end if

    if ( thandle .gt. 0 ) then
      ! Found a physics table.
      ! Activate the conversion
      call tem_write( ' Physics table is defined: Set all quantities like ')
      call tem_write( ' boundaries, initial conditions, fluid and species ')
      call tem_write( ' properties in phyiscal units ')

      ! load dt
      call aot_get_val( L = conf, thandle = thandle, key = 'dt',               &
        &               val = me%dt, ErrCode = iError )
      if (btest(iError, aoterr_Fatal)) then
        call tem_write('FATAL Error occured, while retrieving dt :')
        if (btest(iError, aoterr_NonExistent)) then
          call tem_write('Time step definition not found. Setting to dt=1')
          me%dt = dt_ref
        end if
        if (btest(iError, aoterr_WrongType))                                   &
          & call tem_write('Variable has wrong type!')
        call tem_write('STOPPING')
        call tem_abort()
      end if

      !load rho0
      call aot_get_val( L = conf, thandle = thandle, key = 'rho0',             &
        &               val = me%rho0, default = 1.0_rk, ErrCode = iError )
      if (btest(iError, aoterr_WrongType)) then
        call tem_write('Error occured, while retrieving rho0 :')
        call tem_write('Variable has wrong type!')
        call tem_write('STOPPING')
        call tem_abort()
      end if

      ! load viscosity0
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               key     = 'viscosity',                                 &
        &               val     = me%viscosity0,                               &
        &               default = -1.0_rk,                                     &
        &               ErrCode = iError )
      if (btest(iError, aoterr_WrongType)) then
        call tem_write('Error occured, while retrieving viscosity0 :')
        call tem_write('Variable has wrong type!')
        call tem_write('STOPPING')
        call tem_abort()
      end if

      ! load bulk_viscosity0
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               key     = 'bulk_viscosity',                            &
        &               val     = me%bulk_viscosity0,                          &
        &               default = -1.0_rk,                                     &
        &               ErrCode = iError )
      if (btest(iError, aoterr_WrongType)) then
        call tem_write('Error occured, while retrieving bulk_viscosity0 :')
        call tem_write('Variable has wrong type!')
        call tem_write('STOPPING')
        call tem_abort()
      end if

      ! load molWeig0
      call aot_get_val(L = conf, thandle = thandle, key = 'molWeigh0',         &
        &              val = me%molWeigh0, default = 1.0_rk, ErrCode = iError)
      if (btest(iError, aoterr_WrongType)) then
        call tem_write('Error occured, while retrieving molWeigh0 :')
        call tem_write('Variable has wrong type!')
        call tem_write('STOPPING')
        call tem_abort()
      end if

      ! load temp0
      call aot_get_val(L = conf, thandle = thandle, key = 'temp0',             &
        &              val = me%temp0, default = 1.0_rk, ErrCode = iError)
      if (btest(iError, aoterr_WrongType)) then
        call tem_write('Error occured, while retrieving temp0 :')
        call tem_write('Variable has wrong type!')
        call tem_write('STOPPING')
        call tem_abort()
      end if
    else
      ! No physics table defined. Deactivate conversion
      me%dt     = dt_ref
      me%dx     = dx_ref
      me%rho0   = 1._rk
      me%molWeigh0 = 1._rk
      me%temp0 = 1._rk
      call tem_write('WARNING: Physics table is not defined.')
      call tem_write('         Without physics table, musubi cannot convert'// &
        &            ' lattice to physical units and vice versa' )
    end if

    ! Assign dx and dt for each level
    allocate(me%dxLvl( tree%global%minLevel:tree%global%maxLevel ))
    allocate(me%dtLvl( tree%global%minLevel:tree%global%maxLevel ))
    select case( trim( scaling ))
    case( 'acoustic')
      call tem_write(' Acoustic scaling for dx and dt ')
      do iLevel = tree%global%minLevel, tree%global%maxLevel
        me%dxLvl(iLevel) = me%dx/2._rk**(iLevel-tree%global%minLevel)
        me%dtLvl(iLevel) = me%dt/2._rk**(iLevel-tree%global%minLevel)
      end do
    case( 'diffusive' )
      call tem_write(' Diffusive scaling for dx and dt ')
      do iLevel = tree%global%minLevel, tree%global%maxLevel
        me%dxLvl(iLevel) = me%dx/2._rk**(iLevel-tree%global%minLevel)
        me%dtLvl(iLevel) = me%dt/2._rk**(2*(iLevel-tree%global%minLevel))
      end do
    case default
      call tem_write('Error: wrong scaling provided! '//trim(scaling))
      call tem_write('must be either acoustic or diffusive')
      call tem_abort
    end select
    do iLevel = tree%global%minLevel, tree%global%maxLevel
      call tem_write(' Discretization on level', iLevel )
      call tem_write('    dx:   ',  me%dxLvl( iLevel ))
      call tem_write('    dt:   ',  me%dtLvl( iLevel ))
    end do

    call aot_table_close( L=conf, thandle=thandle )

    call tem_write('  rho0:   ', me%rho0 )
    call tem_write('  molWeigh0:', me%molWeigh0 )
    call tem_write('  Temp0:   ', me%temp0 )


    ! compute and store conversion factors in converstionFac type
    allocate( me%fac( tree%global%minLevel: tree%global%maxLevel ))
    call computeConvertFac( me = me, tree = tree )
    do iLevel = tree%global%minLevel, tree%global%maxLevel
      call tem_log( level = 10,                                              &
        &           msg   = 'conversion factors L='//                        &
        &                   trim( tem_toStr( iLevel )))
      call tem_log( level = 10,                                              &
        &           msg   = 'press '//                                       &
        &                   trim( tem_toStr( me%fac( iLevel )%press )))
      call tem_log( level = 10,                                              &
        &           msg   = 'vel   '//                                       &
        &                   trim( tem_toStr( me%fac( iLevel )%vel )))
    end do

  end subroutine mus_load_physics
!******************************************************************************!


!******************************************************************************!
  !> This routine computed conversion factors for lattice to physical units.
  !! inverse of this factors can be used to convert from physical to lattice
  !! units.\n
  !! Multiply these factors with the LB quantity to get the physical quantity
  !! Divide the physical quantity by these factors to get the LB units
  !!
  subroutine computeConvertFac( me, tree )
    !---------------------------------------------------------------------------
    type( mus_physics_type ), intent(inout) :: me !< physics type
    type(treelmesh_type), intent(in) :: tree !< the treelmesh tree
    !---------------------------------------------------------------------------
    integer :: iLevel
    !---------------------------------------------------------------------------

    do iLevel = tree%global%minLevel, tree%global%maxLevel
      ! length
      me%fac( iLevel )%length = me%dxLvl( iLevel )
      ! time
      me%fac( iLevel )%time = me%dtLvl( iLevel )
      ! temp
      me%fac( iLevel )%temp = me%temp0
      ! velocity
      me%fac( iLevel )%vel = me%dxLvl( iLevel )/me%dtLvl( iLevel )
      ! kinematic viscosity
      me%fac( iLevel )%visc = me%dxLvl( iLevel )**2/me%dtLvl( iLevel )
      ! dynamic viscosity
      me%fac( iLevel )%dynVisc = me%rho0*me%dxLvl( iLevel )**2/me%dtLvl( iLevel )
      ! acceleration
      me%fac( iLevel )%accel = me%dxLvl( iLevel )/me%dtLvl( iLevel )**2
      ! force
      me%fac( iLevel )%force = me%rho0*me%dxLvl( iLevel )**4                   &
        &                    / me%dtLvl( iLevel )**2
      ! pressure or shear stress
      me%fac( iLevel )%press = me%rho0 * me%dxLvl( iLevel )**2                 &
        &                    / me%dtLvl( iLevel )**2
      ! strain rate
      me%fac( iLevel )%strainRate = 1._rk / me%dtLvl( iLevel )
      ! Energy
      me%fac( iLevel )%energy = me%rho0*me%dxLvl( iLevel )**2                  &
        &                     / me%dtLvl( iLevel )**2
      ! Mass
      me%fac( iLevel )%mass = me%rho0*me%dxLvl( iLevel )**3
      ! mole
      me%fac( iLevel )%mole = me%fac( iLevel )%mass/me%molWeigh0
      ! coulomb
      me%fac( iLevel )%coulomb = me%faraday0 * me%fac( iLevel )%mole
      ! mole flux (mol/m^2/s)
      me%fac( iLevel )%moleFlux = me%fac( iLevel )%mole/(me%dxLvl( iLevel )**2 &
        &                       / me%dtLvl( iLevel ))
      ! mass flux (kg/m^2/s)  (density*velocity)
      me%fac( iLevel )%flux = me%rho0*me%dxLvl( iLevel )/me%dtLvl( iLevel )
      !diffusivity
      me%fac( iLevel )%diffusivity = me%dxLvl( iLevel )**2/me%dtLvl( iLevel )
    end do

  end subroutine computeConvertFac
!******************************************************************************!


!******************************************************************************!
  !> This routine creates musubi specific lua function to compute dx and dt.
  !!
  subroutine mus_create_funcStr( fun_str )
    !---------------------------------------------------------------------------
    !> This string contains lua functions to compute dt from visocosity or
    !! velocity
    character(len=*) :: fun_str
    !---------------------------------------------------------------------------
    character(len=256) :: dtVel, dtOmega, dtVisc, dxLevel, dxnL, omega_dt,     &
      &                   vel_dt
    !---------------------------------------------------------------------------
    write(dxLevel,*) 'function getdxFromLevel(t) '//                           &
      &              ' return t.len_bnd/math.pow(2,t.level) end'

    write(dxnL,*) 'function getdxFromnL(t)'//                                  &
      &           ' return t.len_p/t.nL end'

    write(dtVel,*) 'function getdtFromVel(t) return t.dx*t.u_l/t.u_p end'

    write(dtOmega,*) 'function getdtFromOmega(t) '//                           &
      &              ' return ((1.0/t.omega - 0.5)/3.0)*t.dx*t.dx/t.nu_p end'

    write(dtVisc,*) ' function getdtFromVisc(t)'//                             &
      &             ' return t.dx*t.dx*t.nu_l/t.nu_p end'

    write(omega_dt,*) 'function getOmegaFromdt(t) '//                          &
      &               ' return 1.0/(3.0*t.nu_p*t.dt/(t.dx*t.dx) + 0.5) end'

    write(vel_dt, *) 'function getVelFromdt(t) '//                             &
      &              ' return t.u_p*t.dt/t.dx end'

    write(fun_str,*) trim(dtVel)//' '//trim(dtOmega)//' '//trim(dtVisc)        &
      & //' '//trim(dxLevel)//' '//trim(dxnL)//' '//trim(omega_dt)             &
      & //' '//trim(vel_dt)

  end subroutine mus_create_funcStr
!******************************************************************************!


!******************************************************************************!
  !> This routine write physics convert factor into solver specific string in 
  !! lua format.
  !!
  subroutine mus_physics_out( me, conf, tree )
    !---------------------------------------------------------------------------
    type( mus_physics_type ), intent(in) :: me !< physics type
    type(aot_out_type) :: conf
    type(treelmesh_type), intent(in) :: tree
    !---------------------------------------------------------------------------
    integer :: iLevel
    !---------------------------------------------------------------------------

    call aot_out_open_table( put_conf = conf, tname = 'physics' )
    do iLevel = tree%global%minLevel, tree%global%maxLevel
      call aot_out_open_table( put_conf = conf )
      call aot_out_val( put_conf = conf, vname = 'length',                     &
        &               val = me%fac(iLevel)%length )
      call aot_out_val( put_conf = conf, vname = 'time',                       &
        &               val = me%fac(iLevel)%time )
      call aot_out_val( put_conf = conf, vname = 'density',                    &
        &               val = me%rho0 )
      call aot_out_val( put_conf = conf, vname = 'faraday',                    &
        &               val = me%faraday0 )
      call aot_out_val( put_conf = conf, vname = 'vel',                        &
        &               val = me%fac(iLevel)%vel )
      call aot_out_val( put_conf = conf, vname = 'visc',                       &
        &               val = me%fac(iLevel)%visc )
      call aot_out_val( put_conf = conf, vname = 'dynVisc',                    &
        &               val = me%fac(iLevel)%dynVisc )
      call aot_out_val( put_conf = conf, vname = 'accel',                      &
        &               val = me%fac(iLevel)%accel )
      call aot_out_val( put_conf = conf, vname = 'force',                      &
        &               val = me%fac(iLevel)%force )
      call aot_out_val( put_conf = conf, vname = 'press',                      &
        &               val = me%fac(iLevel)%press )
      call aot_out_val( put_conf = conf, vname = 'strainRate',                 &
        &               val = me%fac(iLevel)%strainRate )
      call aot_out_val( put_conf = conf, vname = 'energy',                     &
        &               val = me%fac(iLevel)%energy )
      call aot_out_val( put_conf = conf, vname = 'temp',                       &
        &               val = me%fac(iLevel)%temp )
      call aot_out_val( put_conf = conf, vname = 'mass',                       &
        &               val = me%fac(iLevel)%mass )
      call aot_out_val( put_conf = conf, vname = 'mole',                       &
        &               val = me%fac(iLevel)%mole )
      call aot_out_val( put_conf = conf, vname = 'molWeigh0',                  &
        &               val = me%molWeigh0 )
      call aot_out_val( put_conf = conf, vname = 'coulomb',                    &
        &               val = me%fac(iLevel)%coulomb )
      call aot_out_val( put_conf = conf, vname = 'moleFlux',                   &
        &               val = me%fac(iLevel)%moleFLux )
      call aot_out_val( put_conf = conf, vname = 'flux',                       &
        &               val = me%fac(iLevel)%flux )
      call aot_out_val( put_conf = conf, vname = 'diffusivity',                &
        &               val = me%fac(iLevel)%diffusivity )
      call aot_out_close_table( put_conf = conf )
    end do

    call aot_out_close_table( put_conf = conf )

  end subroutine mus_physics_out
!******************************************************************************!


end module mus_physics_module
!******************************************************************************!
