! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module provides the MUSUBI subroutines needed for the restart 
!! functionality.
!!
!! \author Simon Zimny 
!!
!! This module makes the restart IO functionality available for MUSUBI. When
!! using the write functionality the workflow is as follows
!! - open the file and write the 'normal header'
!! - run over the chunks and
!!   + serialize the data
!!   + write them to file
!! - close the file and write the 'last header'
!! For reading the restart files the workflow is as follows:
!! - open the file
!! - read the solver specific character
!! - loop over the chunks and
!!   + read the data from disc
!!   + unserialize the data and store it in the state vector
!! - close the file
!!
!! Further information on the treelm functions and the usage can be found in
!! the [tem_restart] and the [restart_usage] section.
!!
!! [tem_restart]: @ref tem_restart_module "tem_restart_module"
!! [restart_usage]: @ref restart_usage "Restart Usage"
!!
?? include 'header/lbm_macros.inc'

module mus_restart_module

  ! include treelm modules
  use mpi
  use env_module,            only: rk, io_buffer_size
  use treelmesh_module,      only: treelmesh_type
  use tem_time_module,       only: tem_time_type
  use tem_tools_module,      only: tem_write
  use tem_restart_module,    only: tem_restart_type, tem_restart_openRead,     &
    &                              tem_restart_closeRead, tem_restart_readData,&
    &                              tem_restart_openWrite,                      &
    &                              tem_restart_closeWrite,                     &
    &                              tem_restart_writeData  
  use tem_debug_module,      only: tem_debug, main_debug
  use tem_logging_module,    only: tem_toStr

  ! include musubi modules
  use mus_scheme_type_module, only: mus_scheme_type
  use mus_scheme_module,      only: mus_readSolSpecChar
  use mus_pdf_module,         only: pdf_globInfo_type
  use mus_buffer_module,      only: mus_rest_unserialize, mus_rest_serialize

  implicit none
  private

  public :: mus_writeRestart
  public :: mus_readRestart

contains


!******************************************************************************!
  !> Write the serialized buffer assembled in mus_serializeData to disk
  !!
  subroutine mus_writeRestart( pdfGlobal, restart, scheme, tree, timing )
    !--------------------------------------------------------------------------- 
    !> global pdf info
    type(pdf_globInfo_type), intent(inout) :: pdfGlobal
    !> restart information
    type(tem_restart_type), intent(inout) :: restart
    !> array of schemes including the data to be serialized and dumped
    type( mus_scheme_type ), intent(inout) :: scheme(:)
    !> mesh, provided in treelm format
    type(treelmesh_type), intent(in) :: tree
    !> current simulation time information
    type(tem_time_type), intent(inout) :: timing
    !--------------------------------------------------------------------------- 
    ! array for chunkwise memory bounded output
    real(kind=rk), allocatable :: chunk(:)
    ! counter variables
    integer :: iChunk
    !--------------------------------------------------------------------------- 
    allocate( chunk( io_buffer_size ))

    ! open the file, write header and prepare buffering
    call tem_restart_openWrite( me = restart,                                  &
      &                         tree = tree,                                   &
      &                         timing = timing )

    ! debug output
    if( main_debug%debugRestart )then
      call tem_debug( level = 1,                                               &
        &             msg   = 'DEBUG: WRITE RESTART' )
      call tem_debug( level = 1,                                               &
        &             msg   = 'minLevel '//                                    &
        &                     trim( tem_toStr( pdfGlobal%minLevel ))//         &
        &                     ' size '//trim( tem_toStr(                       &
        &             size( scheme( 1 )%pdf( pdfGlobal%minLevel )%state, 1))))
      call tem_debug( level = 1,                                               &
        &             msg   = 'maxLevel '//                                    &
        &                     trim( tem_toStr( pdfGlobal%maxLevel ))//         &
        &                     ' size '//trim( tem_toStr(                       &
        &             size( scheme( 1 )%pdf( pdfGlobal%maxLevel )%state, 1))))
    end if

    ! loop over the chunks
    do iChunk = 1, restart%nChunks
      ! prepare the data 
      call mus_rest_serialize( scheme, pdfGlobal, restart, tree, chunk,        &
        &                      iChunk )
  
      ! write the data to file
      call tem_restart_writeData( restart, chunk )
    end do
 
    ! close the file and write last header 
    call tem_restart_closeWrite( me     = restart,                             &
      &                          timing = timing,                              &
      &                          tree   = tree )
    
    deallocate( chunk )

    ! debug output
    if( main_debug%debugRestart )then
      call tem_debug( level = 1,                                               &
        &             msg   = 'END DEBUG: WRITE RESTART' )
    end if

  end subroutine mus_writeRestart
!******************************************************************************!


!******************************************************************************!
  !> Read the serialized restart file into the state vectors
  !!
  subroutine mus_readRestart( pdfGlobal, restart, scheme, tree ) 
    !--------------------------------------------------------------------------- 
    !> global pdf info
    type(pdf_globInfo_type), intent(inout) :: pdfGlobal
    !> restart information
    type(tem_restart_type), intent(inout) :: restart
    !> array of schemes including the data to be serialized and dumped
    type( mus_scheme_type ), intent(inout) :: scheme(:)
    !> mesh, provided in treelm format
    type(treelmesh_type), intent(in) :: tree
    !---------------------------------------------------------------------------
    ! local variables
    real(kind=rk), allocatable :: chunk(:)
    integer :: iChunk
    integer :: elemOff ! Offset in the overall number of elements
    !---------------------------------------------------------------------------

    allocate( chunk( io_buffer_size ))

    if( main_debug%debugRestart )then
      call tem_debug( level = 1,                                               &
        &             msg   = 'DEBUG: READ RESTART' )
    end if

    ! open the file, read header and prepare buffering
    call tem_restart_openRead( me = restart )

    ! read the solver specific character
    call mus_readSolSpecChar( restart = restart, scheme = scheme )

    elemOff = 0

    call tem_write('Reading restart...')
    ! loop over the chunks
    do iChunk = 1, restart%nChunks
      ! set the number of elements that are on the stack
      restart%nChunkElems = min(restart%chunkSize, tree%nElems-elemOff)

      !\todo: SZ: This function has to be extended such that it is possible to 
      !           read only data of one variable and store it in the chunk
      !           -> this should be used for initializing the flow in the
      !              mus_flow_module mus_init_flow subroutine as kind of IC
      ! read the data chunkwise from disc
      call tem_restart_readData( restart, chunk )

      !KJ: At this point, we have input scheme which wants to read restart and
      ! the depend instance defined within it (scheme->field->ic->depend) and
      ! the chunk for complete system of the requested variable system.
      
      !\todo: SZ: Maybe one has to modify the input for the following subroutine
      !           that the chunk might be reduced to only one variable (and not 
      !           all like now)
      ! unserialize the data
      call mus_rest_unserialize( scheme, pdfGlobal, restart, tree, chunk,      &
        &                        iChunk )
      elemOff = elemOff + restart%nChunkElems
    end do
 
    ! close the file
    call tem_restart_closeRead( me = restart )

    deallocate( chunk )

    ! debug output
    if( main_debug%debugRestart )then
      call tem_debug( level = 1,                                               &
        &             msg   = 'END DEBUG: READ RESTART' )
    end if

  end subroutine mus_readRestart
!******************************************************************************!


end module mus_restart_module
!******************************************************************************! 
