! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Some generic handy check routines to check the properties of the flow field 
!! and the current run
!! 
!! Performance estimation, check of the total density, check for NaNs
!!\author Manuel Hasert
?? include 'header/lbm_macros.inc'
module mus_tools_module

  ! include treelm modules
  use mpi
!$ use omp_lib
!  use, intrinsic :: ieee_arithmetic
!  use, intrinsic :: ieee_exceptions
  use env_module,             only: rk, long_k, newunit, PathLen, single_k
  use tem_comm_env_module,    only: process
  use tem_tools_module,       only: tem_write
  use tem_time_module,        only: tem_time_type, tem_time_output
  use tem_timeControl_module, only: tem_time_activeTimeStep,                   &
    &                               tem_timeControl_type,                      &
    &                               tem_time_updateTrigger,                    &
    &                               tem_timeControl_checkRuntime
  use tem_timer_module,       only: timerData, tem_startTimer, tem_stopTimer,  &
    &                               tem_getTimerVal, globalTimerHandle
  use tem_status_module,      only: tem_status_type, tem_status_communicate,   &
    &                               tem_stat_runEnd, tem_stat_steady_state,    &
    &                               tem_stat_nan_detected,                     &
    &                               tem_stat_maxWallReached 
  use tem_isNaN_module,       only: tem_isNaN
  use tem_debug_module,       only: main_debug
  use tem_logging_module,     only: tem_log, tem_toStr

  ! include musubi modules
  use mus_pdf_module,         only: pdf_globInfo_type 
  use mus_global_module,      only: param_type
  use mus_scheme_type_module, only: mus_scheme_type

  implicit none

  private

  public  :: perform_checks
  public  :: check_NaNs
  public  :: check_totalDensity
  public  :: dump_timing
  public  :: mus_time_init
  public  :: mus_time_homogenize
  public  :: mus_time_update
  public  :: mus_time_reset
  public  :: check_streaming_layout
  public  :: check_stopFile

  interface check_totalDensity
    module procedure check_scheme_density
    module procedure check_all_density
  end interface

contains

!******************************************************************************!
  !> Perform run-time checks if interval is active
  !!
  subroutine perform_checks( pdfGlobal, params, scheme, iScheme)
    !---------------------------------------------------------------------------
    type(pdf_globInfo_type)           :: pdfGlobal
    type(param_type), intent(inout)   :: params 
    type(mus_scheme_type), intent(in) :: scheme
    integer, intent(in)               :: iScheme
    !---------------------------------------------------------------------------
    logical :: checked
    !---------------------------------------------------------------------------
    checked = .false.
    ! Check for NaNs
    if ( main_debug%checkNans .and. .not.                                      &
      &  params%stat%bits( tem_stat_nan_detected ) ) then
      checked = .true.
      call tem_startTimer( params%isNanTimer )
      call check_nans(scheme,pdfGlobal,params%stat%bits(tem_stat_nan_Detected))
      call tem_stopTimer( params%isNanTimer )
      if( params%stat%bits( tem_stat_nan_detected ) ) then
        call tem_write( '   Found NaN at time = '//                            &
          &             trim( tem_toStr( pdfGlobal%timing%rtime ))//           &
          &             '       iTime = '//                                    &
          &             trim( tem_toStr( pdfGLobal%timing%iTime )))
      end if
    end if

    !  Check and output in the main intervals
    if ( tem_time_activeTimeStep( pdfGlobal%timing, params%timeControl)) then
      checked = .true.
      call check_totalDensity( scheme, pdfGlobal, params%stat )
      call check_stopFile( pdfGlobal, params%stat )
      call tem_timeControl_checkRuntime(                                       &
        &                 me      = params%timeControl,                        &
        &                 now     = tem_getTimerVal( globalTimerHandle ),      &
        &                 reached = params%stat%bits(tem_stat_maxWallReached) )

      !Update the trigger when we are at the last scheme
      if ( iScheme .eq. pdfGlobal%nSchemes ) then
        call tem_time_updateTrigger( me  = params%timeControl,                   &
          &                          now = pdfGlobal%timing)
      end if
    end if
    if( checked ) then
      if( params%stat%bits(tem_stat_nan_detected) .or.                         &
        & params%stat%bits(tem_stat_maxWallReached) )  then

          params%stat%bits( tem_stat_runEnd ) = .true.
      end if
      ! Communicate the tem_stat_runEnd with other processes 
      call tem_status_communicate( params%stat )
    end if

  end subroutine perform_checks
!******************************************************************************!


!******************************************************************************!
  !> Check if a stop file was encountered and communicate to other processes
  !!
  !! If stop file exists, the code stops in the next interval.
  !! the stop file is en empty file  with the name 'stop' 
  !! in the folder in which musubi is being run.
  !! Normally, this is the same folder where musubi.lua resides.
  !!
  subroutine check_stopFile( pdfGlobal, stat )
    !---------------------------------------------------------------------------
    type(pdf_globInfo_type) :: pdfGlobal
    type(tem_status_type)   :: stat
    !---------------------------------------------------------------------------
    character(len=PathLen) :: sig_filename
    logical :: exists
    !---------------------------------------------------------------------------
    sig_filename = './stop'
    exists = .false.
    if ( process%rank .eq. process%root ) then
      inquire(file=sig_filename, exist=exists)
      if( exists ) then
        stat%bits(tem_stat_runEnd) = .true. 
      end if
    endif

  end subroutine check_stopFile
!******************************************************************************!


!******************************************************************************!
  !> Check the state array for nans
  !!
  subroutine check_NaNs( scheme, pdfGlobal, nan_Detected )
    !---------------------------------------------------------------------------
    type(mus_scheme_type), intent(in) :: scheme
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    logical, intent(inout)  :: nan_Detected
    !---------------------------------------------------------------------------
    real(kind=rk)          :: value
    integer iDir, iLevel, iElem, iField, pos
    integer                :: found_nan
    !---------------------------------------------------------------------------

    found_nan = 0
    ! Check for nans only if none was found yet to avoid multiple outputs
    ! First appearing nan element is crucial!
    if( .not. nan_Detected ) then
      !> run over each entry in the state array and check if a nan was found
      do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
        found_nan = 0
        do iElem = 1, scheme%pdf( iLevel )%nElems_local
          found_nan = 0
          do iField = 1, scheme%nFields
            do iDir = 1, scheme%layout%stencil(1)%QQ
              pos = scheme%varSys%Variable%val( iField )%varPos( iDir )
              ! Calculate the density for each element each field
              ! Use post-collsion value to compute density
              value = scheme%pdf( iLevel )%state(                              &
   & ?SAVE?( pos, iElem, scheme%varSys%nScalars, scheme%pdf(iLevel)%neigh ),   &
                & scheme%pdf( iLevel )%nNext )
!              if( ieee_is_nan ( value )) then
              if( tem_isnan ( value )) then
                found_nan = iElem
              end if
            end do ! iDir
          end do ! iField 
        end do ! iElem 
        if( found_nan .gt. 0 ) then
          call tem_write( ' rk '//trim( tem_toStr( process%rank ))//           &
          &               ' NaN found at treeID '//trim( tem_toStr(            &
          &                scheme%glob%levelDesc(iLevel)%total( found_nan ))))
          nan_Detected = .true.
        end if
      end do ! iLevel
    end if

  end subroutine check_NaNs 
!******************************************************************************!


!******************************************************************************!
  !> Check the total density for a selected scheme and write to unit
  !!
  !! The output might be delayed by using arrays which are then dumped 
  !! to keep disc access more restricted
  !!
  subroutine check_scheme_density( scheme, pdfGlobal, stat )
    !---------------------------------------------------------------------------
    !>
    type(mus_scheme_type)   :: scheme
    !>
    type(pdf_globInfo_type) :: pdfGlobal
    !>
    type(tem_status_type )  :: stat
    !---------------------------------------------------------------------------
    real(kind=rk), allocatable :: dens(:) , total_dens(:)
    integer iDir, iLevel, ierror, iElem, iField, pos
    character(len=PathLen) :: outstring, buffer
    logical                :: local_tem_stat_runEnd
    !--------------------------------------------------------------------------
    allocate( dens( scheme%nFields ) )
    allocate( total_dens( scheme%nFields ) )
    total_dens = 0._rk
    dens = 0._rk
    ! Sum up all values in the complete vector to get the total density
    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      do iElem = 1, scheme%pdf( iLevel )%nElems_fluid
        ! compute density for each field
        do iField = 1, scheme%nFields
          do iDir = 1, scheme%layout%stencil(1)%QQ
            pos = scheme%varSys%Variable%val( iField )%varPos( iDir )
           ! Calculate the density for each element each field
           ! Use post-collsion value to compute density
            dens( iField ) = dens( iField ) +                                  &
              & scheme%pdf( iLevel )%state(                                    &
 & ?SAVE?( pos, iElem, scheme%varSys%nScalars, scheme%pdf(iLevel)%neigh ),     &
              & scheme%pdf( iLevel )%nNext )
          end do ! iDir
        end do ! iField 
      end do ! iElem 
    end do ! iLevel
 
    call mpi_reduce( dens, total_dens,scheme%nFields, mpi_double_precision,    &
                     mpi_sum, 0, process%comm, ierror )
    outstring = ''
    if ( process%rank .eq. process%root ) then
      write(outString,'( a,e15.4,a,i9,a)') 'time =', pdfGlobal%timing%rtime,   &
        & '       iTime = ', pdfGLobal%timing%iTime, new_line('A')
      write(outString, '(a,10a,22a)') trim(outString)//                        &
        & '     field ','| total density', new_line('A')
      buffer = ''
      do iField = 1, scheme%nFields
        write(buffer, '(a,i10,a,en20.10,a,a)' )                                &
          &        trim(buffer),iField,' | ', total_dens(iField), new_line('A')
      end do
      write(outString, '(a)') trim(outString)//trim(buffer)
      scheme%glob%totalDens = sum(total_dens)
    endif
    call tem_write( trim(outstring) )

    ! All ranks check, if there are NaN entries
    if( any( tem_isnan( dens ))  ) then
      call tem_write( ' Error: Total density is NaN on rank '//                &
        &             trim( tem_toStr( process%rank ))//'! Stopping... ' )
      stat%bits(tem_stat_runEnd) = .true. 
    endif
    if( stat%bits(tem_stat_steady_state) ) then
      call tem_write( ' Convergence achieved on rank '//                       &
        &             trim( tem_toStr( process%rank ))//'! Stopping... ' )
      stat%bits(tem_stat_runEnd) = .true. 
    endif

    local_tem_stat_runEnd = stat%bits( tem_stat_runEnd ) 

    !\todo: The reduce functionality below has to be extended to have an 
    ! array of logicals which can be reduced on a per need basis at diff
    ! or same intervals
!    call mpi_allreduce(local_tem_stat_runEnd, pdfGlobal%tem_stat_runEnd, 1, mpi_logical,  &
!      &             mpi_lor, process%comm, iError )

  end subroutine check_scheme_density 
!******************************************************************************!


!******************************************************************************!
  !> Check the total density for all schemes
  !!
  subroutine check_all_density( scheme, pdfGlobal, stat )
    !---------------------------------------------------------------------------
    type(mus_scheme_type)   :: scheme(:)
    type(pdf_globInfo_type) :: pdfGlobal
    type(tem_status_type)   :: stat
    !---------------------------------------------------------------------------
    integer :: iScheme
    !---------------------------------------------------------------------------

    do iScheme = 1, pdfGlobal%nSchemes
      call check_scheme_density( scheme    = scheme( iScheme ),                &
        &                        pdfGlobal = pdfGlobal,                        &
        &                        stat      = stat )
    enddo 

  end subroutine check_all_density 
!******************************************************************************!


!******************************************************************************!
  !> Performance results (MLUPs) are written to a file for statistical review
  !! The file-format is simple can be evaluated with gnuplot
  !!
  subroutine dump_timing( scheme, params, pdfGlobal )
    !---------------------------------------------------------------------------
    type(mus_scheme_type) :: scheme(:)
    type(param_type), intent(inout)     :: params 
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    !---------------------------------------------------------------------------
    real(kind=rk)        :: mainloopTime
    real(kind=rk)        :: mlups, mflups
    logical              :: file_exists
    character(15)        :: filename
    integer              :: fileunit, iTimer, iScheme, iLevel
    integer(kind=long_k) :: cellUpdates, fluidUpdates
    real(kind=rk),allocatable        :: timerVal(:)
    character(len=40),allocatable    :: timerLabel(:)
    character(len=PathLen)    :: header
    character(len=PathLen)    :: output
    !---------------------------------------------------------------------------

    if( process%rank .eq. process%root ) then
      allocate( timerVal( timerData%nTimers ) )
      allocate( timerLabel( timerData%nTimers ) )
 
      do iScheme = 1, pdfGlobal%nSchemes
 
        fileunit = newunit()
        
        mainloopTime = 0._rk
        header = '' 
        write(header,'(a,a1)') trim(header), '#'
        write(header,'(a,a16)') trim(header), 'Revision'
        write(header,'(a,a7)') trim(header), 'Scheme'
        write(header,'(a,a17)') trim(header), 'DomSize'
        write(header,'(a,a10)') trim(header), 'nProcs'
!$      write(header,'(a,a10)') trim(header), 'nThreads'
        write(header,'(a,a14)') trim(header), 'MFLUPs'
        write(header,'(a,a14)') trim(header), 'MLUPs'
        write(header,'(a,a14)') trim(header), 'timeTotal'
        write(header,'(a,a11)') trim(header), 'maxIter'
        write(header,'(a,a19)') trim(header), 'totalDens'
        write(header,'(a,a21)') trim(header), 'Casename'
  
        do iTimer = 1, timerData%nTimers
          timerVal( iTimer )   = timerData%value( iTimer )
          timerLabel( iTimer ) = 'time'//trim(timerData%label( iTimer ))
          !> \todo the identification, of which entry is really the main loop 
          !!  has to be done in a more general way
          if( trim(TimerLabel(iTimer)) == 'timeMainLoop' ) then
            mainloopTime = timerVal( iTimer )
          endif
          write(header,'(a,a16)') trim(header),trim(timerLabel( iTimer ))
        enddo

        cellUpdates  = 0_long_k
        fluidUpdates = 0_long_k

        ! Calculate the mlups with the level wise time steps
        do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
          cellUpdates = cellUpdates + scheme( iScheme )%glob                   &
            &                         %nElems_solveLevel( iLevel )             &
            &         / int( 2**( pdfGlobal%maxLevel - iLevel ), kind=long_k )
          fluidUpdates = fluidUpdates + scheme( iScheme )%glob                 &
            &                         %nElems_totalLevel( iLevel )             &
            &         / int( 2**( pdfGlobal%maxLevel - iLevel ), kind=long_k )
        enddo
        mlups = real(cellUpdates*pdfGlobal%timing%nIters,kind=rk)              &
          &          / (mainloopTime*1000000._rk) 
        mflups = real(fluidUpdates*pdfGlobal%timing%nIters,kind=rk)            &
          &          / (mainloopTime*1000000._rk) 

        output = ''
        write(output, '(a,a16)')    trim(output), trim(params%revision)
        write(output, '(a,i7)')     trim(output), iScheme
        write(output, '(a,i18)')    trim(output),                              &
          &                              scheme(iScheme)%glob%nElems_totalFluid
        write(output, '(a,i10)')    trim(output), process%comm_size
!$      write(output, '(a,i10)')    trim(output), omp_get_max_threads()
        write(output, '(a,en14.2)') trim(output), mflups
        write(output, '(a,en14.2)') trim(output), mlups
        write(output, '(a,en14.4)') trim(output), mainloopTime
        write(output, '(a,en11.0)') trim(output),                              &
          &                              real(params%timeControl%iMax, kind=rk)
        write(output, '(a,en19.9)') trim(output), scheme(iScheme)%glob%totalDens
        write(output, '(a,a1)')     trim(output), ' ' 
        write(output, '(a,a20)')    trim(output), trim(params%solver%simName)

        do iTimer = 1, timerData%nTimers
          write(output,'(a,f16.4)') trim(output),timerVal( iTimer )
        enddo
  
        filename='timing.res'
  
        inquire(file=filename,exist=file_exists)
        open(unit=fileunit,file=trim(filename),position='append') 
  
        if( .not. file_exists ) then
           write(fileunit,'(a)') trim(header)
        end if
        write(fileunit,'(a)') trim(output)
        close(fileunit)
      end do
    endif

  end subroutine dump_timing
!******************************************************************************!


!******************************************************************************!
  !> Check for the streaming layout.
  !! 
  subroutine check_streaming_layout( pdfGlobal )
    !---------------------------------------------------------------------------
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    !---------------------------------------------------------------------------

    ! Check, if advection layout is PUSH. if so, abort, because the current code
    ! can only handle PUSH with single level
    if( "?StreamName?" == 'push') then
      call tem_write(' Pushing pdfs to neighbor elements after collision ')
      if( pdfGlobal%minLevel .ne. pdfGlobal%maxLevel ) then
        call tem_write( ' ---------------------------------------------------------- ' ) 
        call tem_write( '           WARNING !!                                       ' ) 
        call tem_write( '    Push is not possible with multi-level execution.        ' ) 
        call tem_write( ' ---------------------------------------------------------- ' ) 
        call tem_write( '                                                            ' ) 
        
      endif
    else
      call tem_write(' Pulling pdfs from neighbor elements before collision ')
    endif

  end subroutine check_streaming_layout
!******************************************************************************!


!******************************************************************************!
  !> Reset the global time object
  !! 
  subroutine mus_time_reset( me )
    !---------------------------------------------------------------------------
    type(tem_time_type), intent(inout) :: me
    !---------------------------------------------------------------------------
    me%iTime = 0
    me%iStart= 0
    me%iStop = 0
    call mus_time_homogenize( me = me, dt = 1._rk, fromRealToInt = .false. )

  end subroutine mus_time_reset
!******************************************************************************!


!******************************************************************************!
  !> Set initial timing from the time Control object
  subroutine mus_time_init( me, timeControl )
    !---------------------------------------------------------------------------
    type(tem_time_type), intent(inout) :: me
    type(tem_timeControl_type), intent(inout) :: timeControl
    !---------------------------------------------------------------------------
    if( timeControl%useIterations ) then
      me%iStart  = timeControl%iMin
      me%iTime   = me%iStart
    else
      me%rStart  = timeControl%tMin  
      me%rTime   = me%rStart
    end if
  end subroutine mus_time_init
!******************************************************************************!


!******************************************************************************!
  !> Update the time step
  !! If no time step was defined, update the time by integer 1
  !!
  subroutine mus_time_update( me, dt )
    !---------------------------------------------------------------------------
    type(tem_time_type), intent(inout) :: me
    real(kind=rk), intent(in) :: dt
    !---------------------------------------------------------------------------
    real(kind=rk) :: dtPlus
    !---------------------------------------------------------------------------
    if( dt .le. 0._rk) then
      dtPlus = 1._rk
    else
      dtPlus = dt
    end if
    me%iTime  = me%iTime  + 1
    me%nIters = me%nIters + 1
    !KJ: The below statement was originally adding dt iteratively to rTime. This
    !has been changed to multiplication to ensure better precision per IEEE
    !standards. We found out inconsistency in the results with restart because
    !of the formal addition i.e. the rTime after a few iterations was incorrect
    !due to the addition operation instead of multiplication
    me%rTime  = me%iTime  * dtPlus

  end subroutine mus_time_update
!******************************************************************************!


!******************************************************************************!
  !> Convert time between real and integer time
  !! 
  subroutine mus_time_homogenize( me, dt, FromRealToInt, readRestart )
    !---------------------------------------------------------------------------
    type(tem_time_type), intent(inout) :: me
    real(kind=rk), intent(in) :: dt
    logical, intent(in), optional :: fromRealToInt
    logical, intent(in), optional :: readRestart
    !---------------------------------------------------------------------------
    logical :: task_realToInt, restart
    !---------------------------------------------------------------------------
    if( present( fromRealToInt )) then
      task_realToInt = fromRealToInt
    else
      task_realToInt = .false.
    endif
    
    if ( present(readRestart) ) then
      restart = readRestart
    else
      restart = .false.
    end if

    if( task_realToInt .or. readRestart ) then
      if( .not. me%iTime_set ) then
        me%iTime = nint( me%rTime/dt )
        me%iStart= nint( me%rStart/dt)
        me%iStop = nint( (me%rStop)/dt )
      end if
      if( readRestart ) then
        me%iStart = me%iTime
      end if
    else
      me%rTime = real( me%iTime, kind=rk )*dt
      me%rStart= real( me%iStart, kind=rk)*dt
      me%rStop = real( me%iStop, kind=rk )*dt
    endif

    call tem_log( level = 1,                                                   &
      &           msg   = ' Homogenizing real and int time' )
    if( task_realToInt) then
      call tem_log( level = 1,                                                 &
        &           msg   = '      real -> int ' )
    else
      call tem_log( level = 1,                                                 &
        &           msg   = '      int  -> real' )
    end if

    ! debug time output
    call tem_time_output( me )

  end subroutine mus_time_homogenize
!******************************************************************************!


end module mus_tools_module
!******************************************************************************!
