! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module provides the wrapper for calling the 
!! [tem_tracking] functionality within Treelm
!!
!! \author Kannan Masilamani, Simon Zimny, Manuel Hasert
!!
!! [tem_tracking]: @ref tem_tracking_module "tracking" functionality
!!
module mus_tracking_module

  ! include treelm modules
  use mpi
  use env_module,              only: rk, long_k, io_buffer_size
  use tem_reduction_module,    only: tem_reduction_open, tem_reduction_close,  &
    &                                tem_reduction_append
  use tem_transient_reduction_module,                                          &
    &                          only: tem_transient_reduction_open,             &
    &                                tem_transient_reduction_apply,            &
    &                                tem_transient_reduction_close
  use tem_restart_module,      only: tem_restart_writeData,                    &
    &                                tem_restart_openWrite,                    &
    &                                tem_restart_closeWrite
  use tem_tracking_module,     only: tem_tracking_type,                        &
    &                                tem_tracking_dumpAsciiTransient,          &
    &                                tem_tracking_dumpAsciiSpatial,            &
    &                                tem_trackingHarvester,                    &
    &                                tem_trackingAsciiSpatial,                 &
    &                                tem_trackingAsciiTransient,               &
    &                                tem_tracking_close_AsciiTransient,        &
    &                                tem_trackingConvergence
  use tem_var_system_module,   only: tem_var_system_type
  use tem_comm_env_module,     only: process
  use treelmesh_module,        only: treelmesh_type
  use tem_stencil_module,      only: tem_stencilHeader_type
  use tem_timeControl_module,  only: tem_time_activeTimeStep,                  &
    &                                tem_time_updateTrigger
  use tem_subTree_module,      only: tem_updatePropertyBits
  use tem_condition_module,    only: tem_convergence_evaluate
  use tem_subTree_type_module, only: tem_subTree_type, tem_treeIDfrom_subTree
  use tem_status_module,       only: tem_stat_steady_state,                    &
    &                                tem_stat_run_terminate

  ! include aotus modules
  use aotus_module, only: flu_state

  ! include musubi modules
  use mus_global_module,        only: param_type
  use mus_pdf_module,           only: pdf_data_type, pdf_globInfo_type
  use mus_scheme_layout_module, only: mus_scheme_layout_type
  use mus_scheme_type_module,   only: mus_scheme_type
  use mus_buffer_module,        only: mus_trans_serialize

  implicit none

  private

  public :: tracker

contains


!******************************************************************************!
  !> Loop over the tracker elements to calculate and dump requested quantities
  !!
  !! Usage of tracking objects and definitions can be found in 
  !! \ref tem_tracking_module
  !! Reductions can be applied. See \ref tem_reduction_module for details.
  subroutine tracker( scheme, pdf, track, params, layout, pdfGlobal, glob_tree )
    !---------------------------------------------------------------------------
    !> current scheme information to work on
    type(mus_scheme_type), intent(inout)  :: scheme
    !> array of pdfs including the state vectors
    type(pdf_data_type), intent(in)   :: pdf(:)
    !> tracking object containing all tracking relevant information
    type(tem_tracking_type ), intent(inout)    :: track(:)    
    !> parameter type
    type(param_type), intent(inout)       :: params
    !> layout information on the current scheme
    type(mus_scheme_layout_type), intent(in)  :: layout
    !> global pdf information
    type(pdf_globInfo_type), intent(inout)  :: pdfGlobal
    !> global tree
    type(treelmesh_type ), intent(in)  :: glob_tree
    !---------------------------------------------------------------------------
    ! counters
    integer :: iLog, iVal, iReduce, iPos
    ! temporal array to store the results
    real(kind=rk), allocatable :: res(:)
    ! number of elements to track within current tracker
    integer :: nElems
    ! Number of elements inside a chunk
    integer :: nChunkElems
    ! achieved the convergence?
    logical :: converged
    !---------------------------------------------------------------------------
    allocate( res( io_buffer_size ))

    ! Run over all tracking objects
    do iLog = 1,size(track)
      ! Determine, if it is now the time to perform the tracking for each 
      ! tracking object
      if(tem_time_activeTimeStep( pdfGlobal%timing,                            &
        &                         track( iLog )%header%timeControl )           &
        ! if run_terminate, then track again before exiting.
        ! However, if tmax was specified, don't track.
        & .or. (                                                               &
        ! if the code terminates soon ...
        &  ( params%stat%bits(tem_stat_run_terminate) .and.                    &
        ! ... then make sure it is really the last iteration due to 
        !     recursiveness
        & ( mod( pdfGlobal%timing%nIters, params%scaleFactor**                 &
        &                  (pdfGlobal%maxLevel - pdfGlobal%minLevel )) .eq. 0) &
        &  )                                                                   &
        &  .and.  (pdfGlobal%timing%iTime .le.                                 &
        & (track( iLog )%header%timeControl%iMax + pdfGlobal%timing%iStart)))  &
        & ) then
        ! Skip this tracking object, if there are no entries in the
        ! variable system
        if (track( iLog )%varSys%nScalars .lt. 1) cycle

        ! open transient reduction
        call tem_transient_reduction_open( me = track( iLog )%transientReduce, &
          &                                time = pdfGlobal%timing%rTime )

        ! open spatial reduction
        call tem_reduction_open( me     = track( iLog )%reduce,                &
          &                      varSys = track( iLog )%varSys)

        ! Call the matching function to get the quantity we want
        ! result goes into res
        if( track( iLog )%subTree%useGlobalMesh ) then

          ! prepare the header differently for using global tree and tracking 
          ! subTree. For harvester format, prepare the header.
          if( track( iLog )%header%fileformat .eq. tem_trackingHarvester ) then
            call tem_restart_openWrite( me     = track( iLog )%restart,        &
              &                         tree   = glob_tree,                    &
              &                         noDeps = .true.,                       &
              &                         timing = pdfGlobal%timing)
          end if

        else

          ! update the property bits according to the global mesh if this has
          ! changed
          call tem_updatePropertyBits( glob_tree, track( iLog )%subTree )

          ! For harvester format, prepare the header
          ! Here we pass the tracking subTree and the global tree to the header
          ! In tem_restart_openWrite the subTree will be written to disk
          if( track( iLog )%header%fileformat .eq. tem_trackingHarvester ) then
            call tem_restart_openWrite( me      = track( iLog )%restart,       &
              &                         tree    = glob_tree,                   &
              &                         timing  = pdfGlobal%timing,            &
              &                         subTree = track( iLog )%subTree,       &
              &                         label   = track( iLog )%               &
              &                                        header%label//'_',      &
              &                         noDeps  = .true. )

          end if
        end if

        ! track the variables, also do reduction inside
        call trackVariable( track       = track(iLog),                         &
          &                 pdf         = pdf,                                 &
          &                 globSys     = scheme%varSys,                       &
          &                 pdfGlobal   = pdfGlobal,                           &
          &                 res         = res,                                 &
          &                 glob_tree   = glob_tree,                           &
          &                 elemPos     = track( iLog )%levelPointer,          &
          &                 nChunkElems = nChunkElems,                         &
          &                 stencil     = layout%stencil(1),                   &
          &                 conf        = params%conf,                         &
          &                 subTree     = track( iLog )%subTree )

        ! close transient reduction
        call tem_transient_reduction_close( me = track( iLog )%transientReduce )
 
        ! Perform the global reduction on the data which was appended
        ! inside trackVariable by tem_reduction_append
        call tem_reduction_close( me   = track( iLog )%reduce,                 &
          &                       comm = track( iLog )%restart%comm )
 
        ! If a reduction is present in the tracking, then the output is
        ! changed completely to only the reduced values
        ! For each variable, a reduction has to be assigned
        if( track( iLog )%reduceActive ) then
          ! Re-assign the chunk here to the stuff which was produced in the
          ! reduction operation
          iPos = 0
          ! Reset the nChunkElems to 1
          nElems = 1
          nChunkElems = 1
          do iReduce = 1, size(track( iLog )%reduce )
            do iVal = 1, size( track( iLog )%reduce( iReduce )%val )
              iPos = iPos + 1
              res( iPos ) = track( iLog )%reduce( iReduce )%val( iVal )
            end do
          end do
        end if

        ! Now write to disk
        ! For the ascii formats, this is done after the chunk loop,
        ! as we anyway only allow one chunk to be filled.
        ! if it's larger, another
        select case( track( iLog )%header%fileformat )
          case( tem_trackingAsciiTransient )
            if( (track( iLog )%onlyRootDumps                                   &
              & .and. track( iLog )%subTree%global%myPart .eq. 0 )             &
              & .or. .not. track( iLog )%onlyRootDumps ) then
              call tem_tracking_dumpAsciiTransient(                            &
                &                    chunk       = res,                        &
                &                    me          = track( iLog ),              &
                &                    nChunkElems = nChunkElems,                &
                &                    timing      = pdfGlobal%timing )
            end if
            !close tracking ascii at last time step
            if( pdfGlobal%timing%iTime                                         &
              & .eq. track( iLog )%header%timeControl%iMax )                   &
              & call tem_tracking_close_AsciiTransient( me = track( iLog ))
          case( tem_trackingAsciiSpatial )
            call tem_tracking_dumpAsciiSpatial(                                &
              &                    me          = track( iLog ),                &
              &                    commGlob    = process,                      &
              &                    solver      = params%solver,                &
              &                    timing      = pdfGlobal%timing,             &
              &                    chunk       = res,                          &
              &                    nChunkElems = nChunkElems )
          case( tem_trackingHarvester )
            ! close the file and write last header 
            if( track( iLog )%subTree%useGlobalMesh ) then
              call tem_restart_closeWrite( me     = track( iLog )%restart,     &
                &                          timing = pdfGlobal%timing,          &
                &                          tree   = glob_tree )
            else
              call tem_restart_closeWrite( me      = track( iLog )%restart,    &
                &                          tree    = glob_tree,                &
                &                          timing  = pdfGlobal%timing,         &
                &                          subTree = track( iLog )%subTree )
            end if

          ! case for assessing the convergence of the solution
          case( tem_trackingConvergence)
            call tem_convergence_evaluate(                                     &
              &            me       = track( iLog )%header%convergence,        &
              &            state    = res,                                     &
              &            achieved = converged )
            ! If convergence is reached, publish a message
            if( converged ) then
              params%stat%bits(tem_stat_steady_state) = .true.
            end if
        end select
      end if  ! do tracking? interval, tmin, tmax check

      ! udpate trigger time which is the next time point to perform tracking
      call tem_time_updateTrigger( now = pdfGlobal%timing,                     &
        &                          me = track( iLog )%header%timeControl )

    end do ! iLog

    deallocate( res )

  end subroutine tracker
!******************************************************************************!


!******************************************************************************!
  !> this routine loops over the \ref tem_var_system_module "tracking variable 
  !! system" and call derive routines
  !!
  subroutine trackVariable( track, pdf, globSys, pdfGlobal, res, glob_tree,    &
    &                       elemPos, stencil, nChunkElems, conf, subTree )
    !---------------------------------------------------------------------------
    !> scheme variable system
    type(tem_tracking_type), intent(inout) :: track
    !> contains state vector of all levels and neigh array
    type(pdf_data_type), intent(in)   :: pdf(:)
    !> global scheme variable system
    type(tem_var_system_type), intent(in) :: globSys
    !> global pdf information
    type(pdf_globInfo_type), intent(in)    :: pdfGlobal
    !> array for the results
    real(kind=rk), intent(inout)  :: res(:)
    !> array element position in the state vector
    integer, intent(in) :: elemPos(:)
    !> global tree
    type(treelmesh_type), intent(in) :: glob_tree
    !> number of elements in the chunk
    integer, intent(out)        :: nChunkElems
    !> stencil information of the current scheme
    type(tem_stencilHeader_type), intent(in) :: stencil
    !> aotus type for solver specific information
    type( flu_state ), intent(in) :: conf
    !> optional subTree of given tree
    type(tem_subTree_type), optional, intent(in) :: subTree
    !---------------------------------------------------------------------------
    integer :: iChunk, nChunks, chunkSize
    integer :: iVar, nComp, nScalars
    integer :: iElem, res_size, maxnComp
    integer :: nElems ! number of elements to track
    integer :: input_buffer_size
    ! lower and upper bound for the list of elements
    integer :: bound(2)
    ! list of treeIDs to be tracked
    integer(kind=long_k), allocatable :: treeID(:)
    real(kind=rk), allocatable :: chunk(:)
    real(kind=rk), allocatable :: local_res(:)
    !---------------------------------------------------------------------------
    ! set the lower and upper bound for the element array in the subTree
    bound(1) = 1
    bound(2) = subTree%nElems

    call tem_treeIDfrom_subTree( subTree, glob_tree, treeID, bound )

    ! total number of elements to be tracked
    nElems = subTree%nElems

    ! number of elements in one input chunk
    chunkSize = track%restart%chunkSize
    input_buffer_size = track%restart%chunksize * globSys%nScalars
    nChunks = track%restart%nChunks

    nScalars = track%varSys%nScalars

    ! get the maximum number of components for all tracked variables to allocate
    ! the local result array
    maxnComp = maxval(track%varSys%variable%val(1:track%varSys%nVars)%nComponents)

    ! allocate the local results array with maximum number of elements
    allocate( local_res( maxnComp*chunkSize ))
    ! allocate the input buffer
    allocate( chunk( input_buffer_size ))

    ! Process all chunks to derive the quantities defined in the tracking object
    do iChunk = 1, nChunks

      ! Serialize the state quantities to provide input for 
      call mus_trans_serialize( pdf         = pdf,                             &
        &                       globSys     = globSys,                         &
        &                       chunk       = chunk,                           &
        &                       iChunk      = iChunk,                          &
        &                       chunkSize   = chunkSize,                       &
        &                       nChunkElems = nChunkElems,                     &
        &                       treeID      = treeID,                          &
        &                       elemPos     = elemPos )

      ! Compute the element lower and upper bound for the current chunk
      bound(1) = (iChunk-1)*chunkSize+1
      bound(2) = (iChunk-1)*chunkSize+nChunkElems

      ! Process all variables of the tracking object
      do iVar = 1, track%varSys%nVars

        ! get the number of components for variable iVar
        nComp = track%varSys%variable%val(iVar)%nComponents

        ! get the size of the needed part of the local_res array
        res_size = nComp*nChunkElems

        ! derive the quantities for all the elements of the current chunk
        call globSys%derive( track%varSys%variable%val( iVar )%mapping )%do(   &
          &                  input   = chunk,                                  &
          &                  bound   = bound,                                  &
          &                  tree    = glob_tree,                              &
          &                  subTree = subTree,                                &
          &                  requVar = track%varSys%Variable%val(iVar),        &
          &                  globSys = globSys,                                &
          &                  stencil = stencil,                                &
          &                  conf    = conf,                                   &
          &                  res     = local_res(1:res_size),                  &
          &                  bary    = track%bary(bound(1) : bound(2), : ),    &
          &                  time    = pdfGlobal%timing )

        ! Now convert LB unit quantities to physical, if defined.
        ! If LB units required, then toPhyiscal is 1
        !local_res = local_res*track%varSys%variable%val( iVar )%toPhysical

        ! copy the information to the right positions in the result array
        ! res contains results for all vatiables,
        ! local_res is only for one variable
        do iElem = 1, nChunkElems
          res( (iElem-1)*nScalars + track%varSys%variable%val(iVar)%varpos(1) :&
            &  (iElem-1)*nScalars + track%varSys%variable%val(iVar)%varpos(nComp) )&
            &  =    local_res( (iElem-1)*nComp+1 : iElem*nComp )
        end do

      end do ! iVar

      ! perform transient reduction
      call tem_transient_reduction_apply( me     = track%transientReduce,      &
        &                                 chunk  = res,                        &
        &                                 offset = bound(1) - 1,               &
        &                                 nChunkElems = nChunkElems,           &
        &                                 varSys = track%varSys )

      ! preform spatial reduction
      call tem_reduction_append( me     = track%reduce,                        &
        &                        chunk  = res,                                 &
        &                        nElems = nChunkElems,                         &
        &                        treeID = treeID(bound(1) : bound(2) ),        &
        &                        tree   = glob_tree,                           &
        &                        varSys = track%varSys )

      select case( track%header%fileformat )
      case( tem_trackingHarvester )
        ! update the nChunkElems in track%restart
        track%restart%nChunkElems = nChunkElems
        call tem_restart_writeData( chunk   = res,                             &
          &                         restart = track%restart )
      end select

    end do  ! iChunk

    ! deallocate chunk and the local result array
    deallocate(chunk)
    deallocate(local_res)

  end subroutine trackVariable
!******************************************************************************!


end module mus_tracking_module
!******************************************************************************!
