! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> M U S U B I  
!! The lattice Boltzmann solver within the APES suite
!! (c) 2012 German Research School for Simulation Sciences GmbH, Aachen.
!! (c) 2013 University of Siegen.
!! 
!! For a documentation, run ./waf gendoxy and find the documentation at
!! ./Documentation/html/index.html
program musubi   

  ! include treelm modules
  use mpi
  use env_module,              only: fin_env
  use tem_aux_module,          only: tem_start, tem_finalize
  use tem_tools_module,        only: tem_write
  use tem_timer_module,        only: tem_addTimer, tem_startTimer, tem_stopTimer
  use tem_timeControl_module,  only: tem_time_determineEnd
  use tem_time_module,         only: tem_time_dump
  use tem_status_module,       only: tem_stat_runEnd, tem_stat_run_terminate

  ! include musubi modules
  ! RESPECT THE ORDER !!!!!!!!
  use mus_pdf_module,    only: pdf_globInfo_type
  use mus_scheme_type_module, only: mus_scheme_type
  use mus_global_module, only: mus_init_global,param_type
  use mus_geom_module,     only: mus_geom_type 
  use mus_config_module,   only: mus_load_config
  use mus_tools_module,    only: check_totalDensity, dump_timing
  use mus_interpolate_header_module, only: interpolation_type
  use mus_flow_module,               only: mus_init_flow
  ! ESPECIALLY OF THE FOLLOWING MODULES
  use mus_construction_module,       only: mus_construct
  use mus_control_module,            only: mus_control_type, mus_init_control
  use mus_aux_module,                only: mus_init_aux, mus_finalize
   
  implicit none
  !-----------------------------------------------------------------------------
  type(mus_scheme_type), allocatable :: scheme(:) !< scheme types
  type(mus_geom_type)        :: geometry
  type(param_type)       :: params  
  type(interpolation_type) :: intp      
  type(pdf_globInfo_type) :: pdfGlobal
  type(mus_control_type) :: control
  integer :: ierr      
  !-----------------------------------------------------------------------------

  ! Initialize environment
  call tem_start('Musubi')
  call mus_init_global( params )

  ! Get settings and initialize behaviour and fluid domain
  call mus_load_config( geometry  = geometry,                                  &
    &                   boundary  = geometry%boundary,                         &
    &                   params    = params,                                    &
    &                   scheme    = scheme,                                    &
    &                   pdfGlobal = pdfGlobal,                                 &
    &                   intp      = intp )
  call mus_construct( params, geometry, scheme, pdfGlobal, intp )
  call mus_init_flow( scheme, geometry, params, pdfGlobal, intp )
  call mus_init_aux( scheme, params, geometry, intp, pdfGlobal )
  call mus_init_control( params, control )

  !-----------------------------------------------------------------------------
  ! Main time loop
  !
!$omp parallel
!$omp single
  call mpi_barrier( MPI_COMM_WORLD, ierr )
  call tem_startTimer( params%mainloopTimer )
!$omp end single
  mainloop: do while ( .not. tem_time_determineEnd( pdfGlobal%timing,          &
    &                        params%timeControl ))

    ! Iterate through the elements in a level-wise fashion 
    ! and advance the time steps
    call control%do_computation( scheme    = scheme,                           &
      &                          geometry  = geometry,                         &
      &                          params    = params,                           &
      &                          intp      = intp,                             &
      &                          iLevel    = pdfGlobal%minLevel,               &
      &                          pdfGlobal = pdfGlobal )

    if( params%stat%bits(tem_stat_run_terminate) ) exit mainLoop
    if( params%stat%bits(tem_stat_runEnd) ) then
      params%stat%bits(tem_stat_run_terminate) = .true.
    end if
  enddo mainloop
  !
  ! Finish main loop
  !-----------------------------------------------------------------------------
!$omp single
  call tem_stopTimer( params%mainloopTimer )
!$omp end single
!$omp end parallel

  call tem_write('Finished main loop with  tmax:', params%timeControl%tMax )
  call tem_write('                         iMax:', params%timeControl%iMax,    &
    &            after  = 1 ) 
  call check_totalDensity( scheme, pdfGlobal, params%stat )
  call mus_finalize( geometry%tree, scheme, params ) 
  call tem_finalize() 
  call dump_timing( scheme, params, pdfGlobal)
  call fin_env()

end program musubi  



!> \mainpage Introduction to Musubi
!! \image html apes_sub_small.png
!!
!! Introduction
!! =================
!!
!! Musubi is the multi-level parallel lattice Boltzmann solver within the APES suite. 
!! It is working on a linearized octree and uses efficient data structures 
!! allowing adaptive parallel simulations.
!! Musubi offers several collision kernels and is designed in a way to deal with huge
!! meshes very efficiently.
!!
!! Usage
!! -----
!! - [Quick Start: Building and Compiling Musubi](@ref buildmusubi) 
!! - [Configuring Simulation Runs](@ref mus_config_module)
!! - [Requirements](@ref tem_requirements)
!!
!! Getting Started
!! ---------------
!! - [Tutorials](@ref mus_tutorials_overview)
!!
!!   An introduction to the usage of the features of Musubi.
!!   You are guided through all required steps from generating meshes, 
!!   configuring the solver and post-processing the results.
!!
!! Basic Features
!! --------------
!! - [Boundary Conditions](@ref mus_bc_header_module)
!!  + [Space-dependent definitions](@ref tem_spatial_module::load_spatial)
!!  + [Transient definitions](@ref tem_transient_module::load_transient)
!!  + [Space-Time definitions](@ref tem_spacetime_fun_module::tem_load_spacetime_single)
!! - [Computation Procedure](@ref computationProcedure)
!! - [Restart](@ref tem_restart_module)
!! - [Tracking](@ref tem_tracking_module) 
!! - [Multilevel](@ref mus_interpolate_module)
!! - [Construction of the data structures](@ref mus_construction_module::mus_construct) 
!!
!! The Octree Data structure 
!! -------------------------
!! - [Treelm Framework](@ref treelmesh_module) 
!! - [Treelmesh type](@ref treelmesh_module::treelmesh_type) 
!! - [Distributed Octree](@ref tem_distributed_octree) 
