! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module contains scheme property type and module related to scheme prop
module mus_scheme_header_module

  ! include treelm modules
  use env_module,       only: LabelLen
  use tem_tools_module, only: tem_write, tem_horizontalSpacer
  use tem_aux_module,   only: tem_abort

  ! include aotus modules                                                      
  use aotus_module,     only: flu_State, aot_get_val, aoterr_NonExistent
  use aot_table_module, only: aot_table_open, aot_table_close
  use aot_out_module,   only: aot_out_type, aot_out_val, aot_out_open_table,   &
    &                         aot_out_close_table

  implicit none

  private

  public :: mus_scheme_header_type, mus_scheme_header_out
  public :: mus_load_scheme_header

  !> Datatype containing information to identify the scheme
  !!
  !! Combination of scheme kind, relaxation and layout%stencilKind 
  !! are used to choose the correct compute kernel for the 
  !! scheme
  type mus_scheme_header_type
    !> Character to store informations about the used scheme (e.g. 
    !! flow simulation)
    character(len=labelLen) :: label
    !> scheme kind, Ex: lbm
    character(len=labelLen) :: kind
    !> scheme layout, Ex: d3q19
    character(len=labelLen) :: layout
    !> scheme relaxation type Ex: BGK, MRT,...
    character(len=labelLen) :: relaxation
  end type mus_scheme_header_type

contains

!******************************************************************************!
  !> load scheme header info from lua file identify table or from scheme table
  !! or from config
  !!
  !! Load scheme label, kind, layoutKind and relaxation
  !! \verbatim
  !! identify = { label = 'schemeLabel', 
  !!              kind = 'simType',
  !!              layout = 'stencilLayout',
  !!              relaxation = 'relaxationType' }
  !! \endverbatim
  !! The label gives the current scheme a recognizable name for post-processing
  !! and variable processing.
  !! For a possible selection of the other parameters 
  !! - simType
  !! - \ref scheme_layout stencilLayout
  !! - relaxationType
  !!
  subroutine mus_load_scheme_header( me, conf, parent, scaling )
  !--------------------------------------------------------------------------
    !> returns scheme identify information
    type( mus_scheme_header_type ), intent(out) :: me  
    type( flu_State ) :: conf !< flu state
    !> parent handle if scheme table is defined
    integer, intent(in), optional :: parent
    !> scaling: diffusive or acoustic?
    character(len=*), intent(in) :: scaling 
    !--------------------------------------------------------------------------
    integer :: thandle !< handle for scheme identify table
    integer :: iError
    !---------------------------------------------------------------------------

    call tem_horizontalSpacer()
    call tem_write( 'Loading Scheme identify table: ')
    !default values
    me%label = ''
    me%kind = 'lbm'
    me%layout = 'd3q19'
    me%relaxation = 'bgk'

    if(present(parent)) then
      call aot_table_open( L       = conf,                                     &
        &                  parent  = parent,                                   &
        &                  thandle = thandle,                                  &
        &                  key = 'identify' )
    else
      call aot_table_open( L       = conf,                                     &
        &                  thandle = thandle,                                  &
        &                  key     = 'identify' )
    endif

    if( thandle .gt. 0 ) then
      !get label
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               key     = 'label',                                     &
        &               val     = me%label,                                    &
        &               default = '',                                          &
        &               ErrCode = iError )
      !get schemekind
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               key     = 'kind',                                      &
        &               val     = me%kind,                                     &
        &               default = 'lbm',                                       &
        &               ErrCode = iError )
      !get layoutkind
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               key     = 'layout',                                    &
        &               val     = me%layout,                                   &
        &               default = 'd3q19',                                     &
        &               ErrCode = iError )
  !    if (btest(iError, aoterr_NonExistent)) then
  !      call tem_write( 'scheme layout is not specified' )
  !      call tem_abort()
  !    end if  
      !get relaxation
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               key     = 'relaxation',                                &
        &               val     = me%relaxation,                               &
        &               default = 'bgk',                                       &
        &               ErrCode = iError )
  
    else
      call tem_write( 'Scheme Identify table not defined.')
      call tem_write( 'Setting default values for scheme..')
    endif 
   
    call aot_table_close( L=conf, thandle=thandle )

    call tem_write( 'label: '// trim( me%label ))
    call tem_write( 'kind: '// trim( me%kind ))
    call tem_write( 'Layout: '// trim( me%layout ))
    call tem_write( 'relaxation: '// trim( me%relaxation ))
    call tem_horizontalSpacer()

  end subroutine mus_load_scheme_header
!******************************************************************************!


!******************************************************************************!
  subroutine mus_scheme_header_out( me, conf )
  !-----------------------------------------------------------------------------
    !> returns scheme identify information
    type( mus_scheme_header_type ), intent(in) :: me  
    type(aot_out_type) :: conf
    !---------------------------------------------------------------------------

    ! the label does not have to be outputted (its the name of the
    ! outputted varSys)
    call aot_out_open_table( put_conf = conf, tname = 'identify' )
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'kind',                                       &
      &               val      = trim( me%kind ))
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'relaxation',                                 &
      &               val      = trim( me%relaxation ))
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'layout',                                     &
      &               val      = trim( me%layout ))
    call aot_out_close_table( put_conf = conf )
  end subroutine mus_scheme_header_out
!******************************************************************************!


end module mus_scheme_header_module
!******************************************************************************!
