! See copyright notice in the COPYRIGHT file.
!*****************************************************************************!
!> scheme_layout module, providing the scheme_layout datatype and the
!! functionality to read the lua files and to set the predefined stencils.
!!
!! \author Simon Zimny
!!
!! \section scheme_layout Scheme Layout
!! The layout includes all information about the stencil, the inverse
!! directions, the prevailing directions and the weights. 
!!
!! In MUSUBI one can choose among the following predefined layouts:
!!
!! -  'predefined_d3q19':
!!    The D3Q19 model. The advection relaxation kernel for this model is highly
!!    optimized.
!! -  'predefined_d3q7':
!!    The more simple D3Q7 model. The advection relaxation kernel for this
!!    model is highly optimized.
!!
!! In MUSUBI it is also possible to define a new layout in the lua file. This
!! feature is implemented to test new layouts. In case one wants to run
!! multiple simulations using this layout, it is highly recommended to
!! implement a new kernel (the following files have to be extended: 
!! \ref mus_flow_module "mus_flow_module" and
!! \ref mus_advrel_bgk_module "mus_advrel_bgk_module" or 
!! \ref mus_advrel_mrt_module "mus_advrel_mrt_module").
!!
!! Defining a new Layout
!! To define a new layout one has to set layout = 'new_stencil'. To define the
!! stencil the following information has to be provided:
!!
!!  -  the number of offsets (QQ)
!!  -  the offsets (disc_vel)
!!  -  the weights (weight)
!!  -  the inverse directions (inv_dir) as integers pointing on the belonging
!!     offsets
!!  -  the prevailing directions (prev_dir) as reals giving the distance in the
!!     different directions
!!
!! The following example shows how to define the standard D3Q19 as a new
!! layout:
!!
!! \verbatim
!! -- Scheme
!! scheme = {
!!  label = 'test',
!!  layout = 'new_stencil',
!!  -- Initial condition 
!!  initial_condition = {
!!                       density = 1.0,
!!                       velocityX = 0.0,
!!                       velocityY = 0.0,
!!                       velocityZ = 0.0 },
!!  -- Boundary conditions
!!  boundary_condition = {
!!  { label = 'wall',
!!    kind = 'inlet_ubb',
!!    velocityX = 0.03, velocityY = 0.0, velocityZ = 0.0 },
!!  { label = 'wall',
!!    kind = 'wall',
!!    velocityX = 0.0, velocityY = 0.0, velocityZ = 0.0 }
!!  },
!!  fluid = { relaxation_scheme = 'BGK',
!!            omega = 1.8, 
!!            rho0 = 1.0 },
!!  -- Defining the new Stencil 
!!  stencil = {
!!             QQ = 19,
!!             disc_vel = {
!!              {-1,0,0},{0,-1,0},{0,0,-1},{1,0,0},{0,1,0},{0,0,1},{0,-1,-1},{0,-1,1},{0,1,-1},
!!              {0,1,1},{-1,0,-1},{1,0,-1},{-1,0,1},{1,0,1},{-1,-1,0},{-1,1,0},{1,-1,0},{1,1,0},
!!              {0,0,0}
!!             },
!!             weight = {
!!              (1.0/18.0),(1.0/18.0),(1.0/18.0),(1.0/18.0),(1.0/18.0),(1.0/18.0),(1.0/36.0),(1.0/36.0),(1.0/36.0),
!!              (1.0/36.0),(1.0/36.0),(1.0/36.0),(1.0/36.0),(1.0/36.0),(1.0/36.0),(1.0/36.0),(1.0/36.0),(1.0/36.0),
!!              (1.0/3.0)
!!             },
!!             inv_dir = {
!!                         4,  5,  6,  1,  2,  3, 10,  9,  8,
!!                         7, 14, 13, 12, 11, 18, 17, 16, 15,
!!                        19
!!             },
!!             prev_dir = {
!!              {1.,0.,0.},{0.,1.,0.},{0.,0.,1.},{1.,0.,0.},{0.,1.,0.},{0.,0.,1.},
!!              {0.,0.5*math.sqrt(2.),0.5*math.sqrt(2.)},{0.,0.5*math.sqrt(2.),0.5*math.sqrt(2.)},
!!              {0.,0.5*math.sqrt(2.),0.5*math.sqrt(2.)},{0.,0.5*math.sqrt(2.),0.5*math.sqrt(2.)},
!!              {0.5*math.sqrt(2.),0.,0.5*math.sqrt(2.)},{0.5*math.sqrt(2.),0.,0.5*math.sqrt(2.)},
!!              {0.5*math.sqrt(2.),0.,0.5*math.sqrt(2.)},{0.5*math.sqrt(2.),0.,0.5*math.sqrt(2.)},
!!              {0.5*math.sqrt(2.),0.5*math.sqrt(2.),0.},{0.5*math.sqrt(2.),0.5*math.sqrt(2.),0.},
!!              {0.5*math.sqrt(2.),0.5*math.sqrt(2.),0.},{0.5*math.sqrt(2.),0.5*math.sqrt(2.),0.}
!!             }
!!  }
!! }
!! \endverbatim
!!
module mus_scheme_layout_module

  ! include treelm modules
  use env_module,            only: rk
  use tem_aux_module,        only: tem_abort
  use tem_stencil_module,    only: tem_stencilHeader_type, tem_loadStencil,   &
    &                              tem_identify_inverseDirections,            &
    &                              tem_identify_prevailDirections,            &
    &                              d3q19_cxDir, d3q7_cxDir, d3q6_cxDir,       &
    &                              d2q9_cxDir, tem_stencil_zeroPos,           &
    &                              tem_create_d3q19, tem_create_d3q7,         &
    &                              tem_create_d3q27, d3q27_cxDir,             &
    &                              tem_create_d3q6, tem_create_d2q9,          &
    &                              tem_create_d1q3
  use tem_tools_module,      only: tem_write
  use treelmesh_module,      only: treelmesh_type
  use tem_logging_module,    only: tem_toStr

  ! include aotus modules
  use aotus_module,     only: flu_state, aot_get_val
  use aot_table_module, only: aot_table_open, aot_table_close, aot_table_length
  use aot_out_module,   only: aot_out_type, aot_out_val

  ! include musubi modules
  use mus_moments_type_module, only: mus_moment_type

  implicit none

  private

  public :: mus_scheme_layout_type
  public :: mus_load_newLayout
  public :: mus_init_stencil
  public :: mus_destroy_stencil
  public :: mus_destroy_layout
  public :: mus_finish_stencil
  public :: mus_define_d3q19
  public :: mus_define_d3q27
  public :: mus_define_d3q7
  public :: mus_define_d3q6
  public :: mus_define_d2q9
  public :: mus_define_d1q3
  public :: mus_weights_out

  !> data structure containing all information related to the 
  !! compute stencil. Several stencils can be defined.
  !! \ref mus_moments_module Moments are directly related to the 
  !! stencil layout and are therefore defined here
  type mus_scheme_layout_type
    !> The list of stencil types ( for LBM this is 1, different in ateles )
    type( tem_stencilHeader_type ), allocatable :: stencil(:)
    !> The weights for the different discrete velocities
    real(kind=rk), allocatable :: weight(:)
    !> Prevailing directions
    real(kind=rk), allocatable :: prevailDir(:,:)
    !> Moment space definition
    type(mus_moment_type) :: moment
    !> New stencil definition loaded from config
    logical :: new_stencil = .false.
  end type mus_scheme_layout_type

contains

!******************************************************************************!
  !> load a new stencil definition from the lua file
  !!
  !! - a label
  !! - the stencil (predefined or new)
  !! - for a new defined stencil ( weights, inverse directions, prevailing
  !!   directions )
  !!
  subroutine mus_load_newLayout( me, parent_handle, conf )
    !---------------------------------------------------------------------------
    integer, intent(in), optional :: parent_handle
    type(mus_scheme_layout_type) :: me
    type(flu_State) :: conf
    !---------------------------------------------------------------------------
    ! defining local variables
    integer :: stencil_handle
    integer :: weight_handle
    integer :: nWeights, iWeight
    integer :: invDir_handle
    integer :: iInvDir, nInvDirs
    integer :: prevDir_handle
    integer :: iPrevDir, nPrevDirs
    integer :: prevDirCoo_handle
    integer :: iPrevDirCoo
    integer :: iError
    !---------------------------------------------------------------------------

    ! this is done due to the double usage of the stencils in ateles & musubi
    allocate( me%stencil(1) )

    if( present( parent_handle))then
      ! open stencil table
      call aot_table_open( L       = conf,                                     &
        &                  parent  = parent_handle,                            &
        &                  thandle = stencil_handle,                           &
        &                  key     = 'stencil' )
    else
      ! open stencil table
      call aot_table_open( L=conf, thandle=stencil_handle, key = 'stencil' )
    end if

    ! If a new stencil is defined, the flag is set to true to 
    ! differentiate between predefined and new stencil in initialization
    if (stencil_handle .gt. 0) then
      me%new_stencil = .true.

      call tem_write('Reading the new Layout...')

      ! Get the number of dimensions in which the stencil is defined
      call aot_get_val( L       = conf,                                        &
        &               thandle = stencil_handle,                              &
        &               val     = me%stencil(1)%nDims,                         &
        &               ErrCode = iError,                                      &
        &               key     = 'nDims',                                     &
        &               default = -1 )

      ! load the stencil
      call tem_loadStencil( stencil       = me%stencil(1),                     &
        &                   parent_handle = stencil_handle,                    &
        &                   conf          = conf )

      ! load additional information like weights, inv. directions, etc.
      call aot_table_open( L       = conf,                                     &
        &                  parent  = stencil_handle,                           &
        &                  thandle = weight_handle,                            &
        &                  key     = 'weight' )
      nWeights = aot_table_length( L = conf, thandle = weight_handle )
      if ( nWeights .eq. me%stencil(1)%QQ ) then
        allocate( me%weight( nWeights )) 

        do iWeight = 1, nWeights
          call aot_get_val( L       = conf,                                    &
            &               thandle = weight_handle,                           &
            &               val     = me%weight( iWeight ),                    &
            &               ErrCode = iError,                                  &
            &               pos     = iWeight )
        end do
        call aot_table_close( L=conf, thandle=weight_handle )
      else
        call tem_write( 'The number of defined weights does not match the '//  &
          &             'number of discrete velocities.' )
        call tem_write( trim( tem_toStr( nWeights))//' vs. '//                 &
          &             trim( tem_toStr( me%stencil(1)%QQ )))
        stop
      end if 
      call aot_table_open( L       = conf,                                     &
        &                  parent  = stencil_handle,                           &
        &                  thandle = invDir_handle,                            &
        &                  key     = 'inv_dir' )
      nInvDirs = aot_table_length( L = conf, thandle = invDir_handle )
      if ( nInvDirs .eq. me%stencil(1)%QQ ) then
        allocate( me%stencil(1)%cxDirInv( nInvDirs )) 

        do iInvDir = 1, nInvDirs
          call aot_get_val( L       = conf,                                    &
            &               thandle = invDir_handle,                           &
            &               val     = me%stencil(1)%cxDirInv( iInvDir ),       &
            &               ErrCode = iError,                                  &
            &               pos     = iInvDir )
        end do
        call aot_table_close( L = conf, thandle = invDir_handle )
      else
        call tem_identify_inverseDirections( me%stencil(1)%cxDirInv,           &
          &                                  me%stencil(1)%cxDir )
      end if 

      call aot_table_open( L       = conf,                                     &
        &                  parent  = stencil_handle,                           &
        &                  thandle = prevDir_handle,                           &
        &                  key     = 'prev_dir' )
      nPrevDirs = aot_table_length( L = conf, thandle = prevDir_handle )
      if(nPrevDirs .gt. 0)then
        allocate( me%prevailDir( 3, nPrevDirs ))
        do iPrevDir = 1, nPrevDirs
          call aot_table_open( L       = conf,                                 &
            &                  parent  = prevDir_handle,                       &
            &                  thandle = prevDirCoo_handle,                    &
            &                  pos     = iPrevDir )
          do iPrevDirCoo = 1, 3
            call aot_get_val( L       = conf,                                  &
              &               thandle = prevDirCoo_handle,                     &
              &               val     = me%prevailDir( iPrevDirCoo, iPrevDir ),&
              &               ErrCode = iError,                                &
              &               pos     = iPrevDirCoo )
          end do
          call aot_table_close( L = conf, thandle = prevDirCoo_handle ) 
        end do
        call aot_table_close( L=conf, thandle=prevDir_handle )
      else
        call tem_identify_prevailDirections( me%prevailDir,                    &
          &                                  me%stencil(1)%cxDir )
      end if 

      call tem_write( 'A new stencil has been defined successfully.' )
      if( me%stencil(1)%nDims .le. 0 ) then
        call tem_write('Error: number of dimensions is not given for '//       &
          &            'the stencil')
        call tem_write('Please define in the configuration file as ')
        call tem_write('  stencil = { nDims = 3, weights = ... }')
        call tem_abort
      end if
    end if

  end subroutine mus_load_newLayout
!******************************************************************************!


!******************************************************************************!
  !> Dump the weights in lua format.
  !!
  !! The style is:
  !! weights = {w1, w2, ... , wN}
  !!
  subroutine mus_weights_out( me, conf )
    !---------------------------------------------------------------------------
    !>musubi schemes stencil type
    real(kind=rk), intent(in) :: me(:)
    type(aot_out_type) :: conf
    !---------------------------------------------------------------------------
    call aot_out_val( put_conf     = conf,                                     &
      &               vname        = 'weight',                                 &
      &               val          = me,                                       &
      &               max_per_line = 2 )

  end subroutine mus_weights_out
!******************************************************************************!


!******************************************************************************!
  !> Prepare the stencils and related information
  !!
  !! allocate stencil for each boundary + pdf
  !! pdf stentil is stored at stencil(1) and rest for number of boundaries
  !!
  subroutine mus_init_stencil( stencil, nBC )
    !---------------------------------------------------------------------------
    !> musubi schemes stencil type
    type( tem_stencilHeader_type ), allocatable, intent(out) :: stencil(:)
    !> number of boundaries 
    integer, intent(in)  :: nBC
    !---------------------------------------------------------------------------
    call tem_write('Initializing stencil layout...',nBC+1) 
    if( allocated ( stencil )) deallocate(stencil)
    allocate( stencil( nBC + 1 ))

  end subroutine mus_init_stencil
!******************************************************************************!


!******************************************************************************!
  !> Destroy the stencil
  !!
  subroutine mus_destroy_stencil( stencil )
    !---------------------------------------------------------------------------
    !>musubi schemes stencil type
    type( tem_stencilHeader_type ), allocatable, intent(out) :: stencil(:)
    !---------------------------------------------------------------------------
    call tem_write('Deallocating stencil layout...') 
    if( allocated(stencil) ) deallocate( stencil )

  end subroutine mus_destroy_stencil
!******************************************************************************!


!******************************************************************************!
  !\todo: SZ: this subroutine needs to do s.th.
  !> Destroy the layout
  !!
  subroutine mus_destroy_layout( layout )
    !---------------------------------------------------------------------------
    !> musubi schemes layout type
    type( mus_scheme_layout_type ), intent(out) :: layout
    !---------------------------------------------------------------------------
    call tem_write('Deleting layout...') 

  end subroutine mus_destroy_layout
!******************************************************************************!


!******************************************************************************!
  !> Finish the stencil definition
  !!
  subroutine mus_finish_stencil( layout )
    !---------------------------------------------------------------------------
    !> musubi schemes layout type
    type( mus_scheme_layout_type ), intent(inout) :: layout
    !---------------------------------------------------------------------------
    integer  :: iStencil
    !---------------------------------------------------------------------------
    do iStencil = 1, 1
      layout%stencil( iStencil )%restPosition =                                &
        &                      tem_stencil_zeroPos( layout%stencil( iStencil ))
    enddo

  end subroutine mus_finish_stencil
!******************************************************************************!


!******************************************************************************!
  !> This subroutine sets the parameters for the predefined d3q19 stencil.
  !!
  subroutine mus_define_d3q19( layout, tree )
    !---------------------------------------------------------------------------
    !> scheme layout for pdf state
    type( mus_scheme_layout_type ), intent(inout)   :: layout
    !> treelm mesh
    type( treelmesh_type ), intent(in)    :: tree
    !---------------------------------------------------------------------------
    ! local variables
    integer :: iWeight
    !---------------------------------------------------------------------------
   
    ! fill the tem_stencilHeader_type
    call tem_create_d3q19( layout%stencil(1) )

    ! setting the prevailing directions (normalized major directions without
    ! sign)
    call tem_identify_prevailDirections( layout%prevailDir,                    &
      &                                  layout%stencil(1)%cxDir )

    allocate( layout%weight( layout%stencil(1)%QQ ))
   
    !Axial directions
    do iWeight=1, 6
      layout%weight( iWeight ) = 1._rk/18._rk
    end do
    !Diagonal directoins
    do iWeight=7, 18
      layout%weight( iWeight ) = 1._rk/36._rk
    end do
    !Center
    layout%weight(19) = 1._rk/ 3._rk 

    layout%stencil(1)%useAll = .true.
    layout%stencil(1)%nElems = tree%nElems

  end subroutine mus_define_d3q19
!******************************************************************************!


!******************************************************************************!
  !> This subroutine sets the parameters for the predefined d3q27 stencil.
  !!
  subroutine mus_define_d3q27( layout, tree )
    !---------------------------------------------------------------------------
    !> scheme layout for pdf state
    type( mus_scheme_layout_type ), intent(inout)   :: layout
    !> treelm mesh
    type( treelmesh_type ), intent(in)    :: tree
    !---------------------------------------------------------------------------
    ! local variables
    integer :: iWeight
    !---------------------------------------------------------------------------
   
    ! fill the tem_stencilHeader_type
    call tem_create_d3q27( layout%stencil(1) )

    ! setting the prevailing directions (normalized major directions without
    ! sign)
    call tem_identify_prevailDirections( layout%prevailDir,                    &
      &                                  layout%stencil(1)%cxDir )

    allocate( layout%weight( layout%stencil(1)%QQ ))
   
    call tem_write('D3Q27 weights are definitely wrong!')
    !Axial directions
    do iWeight=1, 6
      layout%weight( iWeight ) = 1._rk/18._rk
    end do
    !Edge directoins
    do iWeight=7, 18
      layout%weight( iWeight ) = 1._rk/36._rk
    end do
    !Corner   directions
    do iWeight=19, 26
      layout%weight( iWeight ) = 1._rk/36._rk
    end do
    !Center
    layout%weight(27) = 1._rk/ 3._rk 

    layout%stencil(1)%useAll = .true.
    layout%stencil(1)%nElems = tree%nElems

  end subroutine mus_define_d3q27
!******************************************************************************!


!******************************************************************************!
  !> This subroutine sets the parameters for the predefined d3q7 stencil.
  !!
  subroutine mus_define_d3q7( layout, tree )
    !---------------------------------------------------------------------------
    !> scheme layout for pdf state
    type( mus_scheme_layout_type ), intent(inout)   :: layout
    !> treelm mesh
    type( treelmesh_type ), intent(in)    :: tree
    !---------------------------------------------------------------------------
    ! local variables
    integer :: iWeight
    !---------------------------------------------------------------------------

    ! fill the tem_stencilHeader_type
    call tem_create_d3q7( layout%stencil(1) )

    ! setting the prevailing directions (normalized major directions without
    ! sign)
    call tem_identify_prevailDirections( layout%prevailDir,                    &
      &                                  layout%stencil(1)%cxDir )

    allocate( layout%weight( layout%stencil(1)%QQ ))

    do iWeight=1, 6
      layout%weight( iWeight ) = 1._rk/8._rk
    end do
    layout%weight(7) = 1._rk/ 4._rk 

    layout%stencil(1)%useAll = .true.
    layout%stencil(1)%nElems = tree%nElems
  end subroutine mus_define_d3q7
!******************************************************************************!


!******************************************************************************!
  !> This subroutine sets the parameters for the predefined d3q6
  !! layout%stencil(1), used by the Flekkoy model.
  !!
  subroutine mus_define_d3q6( layout, tree )
    !---------------------------------------------------------------------------
    !> scheme layout for pdf state
    type( mus_scheme_layout_type ), intent(inout)   :: layout
    !> treelm mesh
    type( treelmesh_type ), intent(in)    :: tree
    !---------------------------------------------------------------------------
    ! local variables
    integer :: iWeight
    !---------------------------------------------------------------------------

    ! fill the tem_stencilHeader_type
    call tem_create_d3q6( layout%stencil(1) )

    ! setting the prevailing directions (normalized major directions without
    ! sign)
    call tem_identify_prevailDirections( layout%prevailDir,                    &
      &                                  layout%stencil(1)%cxDir )

    allocate( layout%weight(layout%stencil(1)%QQ) )

    do iWeight=1, 6
      layout%weight(iWeight) = 1._rk/6._rk
    end do

    layout%stencil(1)%useAll = .true.
    layout%stencil(1)%nElems = tree%nElems

  end subroutine mus_define_d3q6
!******************************************************************************!


!******************************************************************************!
  !> This subroutine sets the parameters for the predefined d2q9 stencil.
  !!
  subroutine mus_define_d2q9( layout, tree )
    !---------------------------------------------------------------------------
    !> scheme layout for pdf state
    type( mus_scheme_layout_type ), intent(inout)   :: layout
    !> treelm mesh
    type( treelmesh_type ), intent(in)    :: tree
    !---------------------------------------------------------------------------
    ! local variables
    integer :: iWeight
    !---------------------------------------------------------------------------

    ! fill the tem_stencilHeader_type
    call tem_create_d2q9( layout%stencil(1) )

    ! setting the prevailing directions (normalized major directions without
    ! sign)
    call tem_identify_prevailDirections( layout%prevailDir,                    &
      &                                  layout%stencil(1)%cxDir )

    allocate( layout%weight( layout%stencil(1)%QQ ))
   
    !Axial directions
    do iWeight=1, 4
      layout%weight( iWeight ) = 1._rk/9._rk
    end do
    !Diagonal directoins
    do iWeight=5, 8
      layout%weight( iWeight ) = 1._rk/36._rk
    end do
    !Center
    layout%weight(9) = 4._rk/ 9._rk 

    layout%stencil(1)%useAll = .true.
    layout%stencil(1)%nElems = tree%nElems

  end subroutine mus_define_d2q9
!******************************************************************************!


!******************************************************************************!
  !> This subroutine sets the parameters for the predefined d2q9 stencil.
  !!
  subroutine mus_define_d1q3( layout, tree )
    !---------------------------------------------------------------------------
    !> scheme layout for pdf state
    type( mus_scheme_layout_type ), intent(inout)   :: layout
    !> treelm mesh
    type( treelmesh_type ), intent(in)    :: tree
    !---------------------------------------------------------------------------
    ! local variables
    integer :: iWeight
    !---------------------------------------------------------------------------

    ! fill the tem_stencilHeader_type
    call tem_create_d1q3( layout%stencil(1) )

    ! setting the prevailing directions (normalized major directions without
    ! sign)
    call tem_identify_prevailDirections( layout%prevailDir,                    &
      &                                  layout%stencil(1)%cxDir )

    allocate( layout%weight( layout%stencil(1)%QQ ))
   
    ! Weights from [1]  E. M. Viggen, The lattice Boltzmann method in
    ! acoustics, Scandinavian Symposium on Physical Acoustics, pp. 1–5, Mar.
    ! 2010.
    !Axial directions
    do iWeight=1, 2
      layout%weight( iWeight ) = 1._rk/6._rk
    end do
    !Center
    layout%weight(3) = 2._rk/ 3._rk 

    layout%stencil(1)%useAll = .true.
    layout%stencil(1)%nElems = tree%nElems

  end subroutine mus_define_d1q3
!******************************************************************************!


end module mus_scheme_layout_module
!******************************************************************************!

