! See copyright notice in the COPYRIGHT file.
?? include 'header/lbm_macros.inc'
!******************************************************************************!
!> Definition of the datatypes for the scheme implementation.
!!
!! For a detailed description see
!! \ref scheme_implementation "Scheme Implementation".
!!
!! \author Simon Zimny, Kannan Masilamani
!!
module mus_scheme_module

  ! include treelm modules
  use env_module,            only: rk, newUnit, LabelLen, PathLen
  use tem_aux_module,        only: tem_abort, tem_CheckLabel
  use treelmesh_module,      only: treelmesh_type
  use tem_bc_prop_module,    only: tem_bc_prop_type
  use tem_restart_module,    only: tem_restart_type
  use tem_tools_module,      only: tem_write, tem_horizontalSpacer
  use tem_tracking_module,   only: tem_load_tracking, tem_transform_solSpecChar
  use tem_variable_module,   only: tem_variable_type, tem_variable_from
  use tem_var_system_module, only: tem_var_system_type, tem_var_append,        &
    &                              tem_var_sys_mark, tem_var_intersect,        &
    &                              tem_var_sys_init
  use tem_source_module,     only: tem_sourceMap_type, tem_load_source,        &
    &                              tem_reduce_sourceVars
  use tem_derived_module,    only: access_stateVar
  use tem_meshInfo_module,   only: nMeshVars
  use tem_depend_module,     only: tem_load_depend_single, tem_init_depend
  use tem_logging_module,    only: tem_log, tem_toStr

  ! include aotus modules
  use aot_out_module,   only: aot_out_type, aot_out_open, aot_out_close
  use aotus_module,     only: flu_State, open_config_chunk
  use aot_table_module, only: aot_table_open, aot_table_close,                 &
    &                         aot_table_length
  use aot_path_module,  only: aot_path_type, assignment(=), aot_path_addNode

  ! include musubi modules
  use mus_pdf_module,           only: pdf_globInfo_type
  use mus_global_module,        only: param_type
  use mus_scheme_layout_module, only: mus_load_newLayout, mus_weights_out,     &
    &                                 mus_define_d3q19, mus_define_d3q7,       &
    &                                 mus_define_d2q9, mus_define_d3q6,        &
    &                                 mus_define_d1q3, mus_define_d3q27,       &
    &                                 mus_init_stencil, mus_finish_stencil
  use mus_scheme_header_module, only: mus_scheme_header_type,                  &
    &                                 mus_load_scheme_header,                  &
    &                                 mus_scheme_header_out
  use mus_scheme_type_module,   only: mus_scheme_type
  use mus_field_module,         only: mus_field_type, mus_load_fields,         &
    &                                 mus_field_out
  use mus_variable_module,      only: mus_init_varSys, mus_build_derVarList,   &
    &                                 mus_init_variables, mus_store_derVarPos
  use mus_source_module,        only: mus_build_sourceVarSys, mus_init_source
  use mus_physics_module,       only: mus_load_physics, mus_physics_out,       &
    &                                 mus_physics_type

  implicit none

  private

  public :: mus_readSolSpecChar
  public :: mus_writeSolSpecChar
  public :: mus_load_schemes
!  public :: mus_init_schemes
  public :: mus_init_sources
  public :: mus_init_stateVars
  public :: mus_kill_stateVars

  interface mus_writeSolSpecChar
    module procedure mus_writeSolSpecChar_array
    module procedure mus_writeSolSpecChar_single
  end interface 

contains

!******************************************************************************!
  !> load different schemes defined in lua file
  !!
  !! This routines checks whether schemes table 
  !! \verbatim
  !! scheme = { boundary_conditions = {...},
  !!            ...  }
  !! \endverbatim 
  !! is defined. If yes then it will
  !! load schemes table and set mus_scheme_type. 
  !! If no special scheme table is defined, default tables are loaded from the
  !! root level of the lua file
  !! fluid, boundary conditions and initial condiitions
  !! \verbatim
  !! fluid = { ... } 
  !! boundary_conditions = {...}
  !! initial_conditions = {...}
  !! \endverbatim 
  !!
  subroutine mus_load_schemes( me, boundary, conf, params, pdfGlobal, tree,    &
    &                          path )
    !---------------------------------------------------------------------------
    !> array of scheme types to be filled
    type( mus_scheme_type ), intent(out), allocatable :: me(:)
    !> boundary data from mesh
    type( tem_bc_prop_type ), intent(in):: boundary
    !> global parameter type
    type( param_type ), intent(inout) :: params
    !> flu state
    type( flu_State ) :: conf
    !> global pdf info
    type( pdf_globInfo_type) :: pdfGlobal
    !> global treelm mesh
    type( treelmesh_type), intent(in) :: tree
    !> lua path
    type( aot_path_type ),intent(in) :: path
    !---------------------------------------------------------------------------
    integer :: nSchemes, iScheme
    integer :: scheme_handle
    integer :: scheme_sub_handle
    ! lua paths
    type( aot_path_type ) :: schemes_path ! path to the scheme table
    type( aot_path_type ) :: scheme_sub_path ! path to entry in scheme table
    !---------------------------------------------------------------------------

    call tem_horizontalSpacer()
    call tem_write( 'Loading the schemes ...' )

    ! initialize all musubi derived variables
    call mus_init_variables()

    ! try to open the scheme table
    call aot_table_open( L=conf, thandle=scheme_handle, key='scheme' )
    ! set schemes path
    schemes_path = path
    call aot_path_addNode( me       = schemes_path,                            &
      &                    NodeType = 'table',                                 &
      &                    key      = 'scheme' )

    nSchemes = 1
    pdfGlobal%nSchemes = nSchemes
    allocate( me( nSchemes ))
    
    if (scheme_handle .eq. 0) then 
      ! scheme table is not defined 
      ! -> load single scheme with passing scheme handle
      ! call tem_write( 'No scheme table defined. &
      !  &              Setting number of schemes: ', nSchemes )
      call mus_load_scheme_single( me        = me(1),                          &
        &                          boundary  = boundary,                       &
        &                          conf      = conf,                           &
        &                          params    = params,                         &
        &                          pdfGlobal = pdfGlobal,                      &
        &                          tree      = tree,                           &
        &                          path      = path )
    else 
      ! check whether scheme is a single table or multiple table
      call aot_table_open( L       = conf,                                     &
        &                  parent  = scheme_handle,                            &
        &                  thandle = scheme_sub_handle,                        &
        &                  pos     = 1 )
      if ( scheme_sub_handle .eq. 0 ) then
        ! scheme is a single table, load its contents
        call aot_table_close( L=conf, thandle=scheme_sub_handle )
        ! call tem_write( 'Scheme is a single table' )
        call mus_load_scheme_single( me        = me(1),                        &
          &                          boundary  = boundary,                     &
          &                          conf      = conf,                         &
          &                          params    =  params,                      &
          &                          pdfGlobal = pdfGlobal,                    &
          &                          tree      = tree,                         &
          &                          path      = schemes_path,                 &
          &                          parent    = scheme_handle )
      else
        ! scheme is a multiple table
        call aot_table_close( L=conf, thandle=scheme_sub_handle )
        nSchemes = aot_table_length( L=conf, thandle=scheme_handle )
        !call tem_write( 'Number of Schemes used: ', nSchemes )
        deallocate( me )
        allocate( me( nSchemes ))
        pdfGlobal%nSchemes = nSchemes

        do iScheme = 1, nSchemes
          ! storing the number of the current scheme
          me(iScheme)%schemeID = iScheme

          call aot_table_open( L       = conf,                                 &
            &                  parent  = scheme_handle,                        &
            &                  thandle = scheme_sub_handle,                    &
            &                  pos     = iScheme )
          scheme_sub_path = schemes_path
          call aot_path_addNode( me       = scheme_sub_path,                   &
            &                    NodeType = 'table',                           &
            &                    pos      = iScheme ) 
          !call tem_write( 'Properties for scheme', iScheme )
          call tem_horizontalSpacer()
          call mus_load_scheme_single( me        = me( iScheme ),              &
            &                          boundary  = boundary,                   &
            &                          conf      = conf,                       &
            &                          params    = params,                     &
            &                          pdfGlobal = pdfGlobal,                  &
            &                          tree      = tree,                       &
            &                          path      = scheme_sub_path,            &
            &                          parent    = scheme_sub_handle )
          ! Check, if the names of the schemes are unique.
          ! If not, add an increasing number
          call tem_checkLabel( label = me(:)%header%label, nLabels = iScheme )
          call aot_table_close( L=conf, thandle=scheme_sub_handle )
        end do ! number of schemes

        ! if multiple schemes are defined.
        ! it is done after loading all scheme table since the global system
        ! in scheme might differ based on their scheme kind and in
        ! scheme depend we need to find the interection of required
        ! variable system from global var system of the depend scheme
        ! load scheme dependency for each scheme from lua file
        call mus_load_scheme_depend( me     = me,                              &
          &                          conf   = conf,                            &
          &                          parent = scheme_handle)

      end if !if scheme table is single or multiple
    end if ! if scheme is a table

    call aot_table_close( L=conf, thandle=scheme_handle )

    ! initialize the scheme depend for all schemes
    do iScheme = 1, nSchemes
      if ( allocated( me(iScheme)%depend) ) then
        call tem_init_depend( medepend    = me(iScheme)%depend,                &
          &                   varSys      = me(:)%varSys,                      &
          &                   resSysLabel = me(iScheme)%header%label,          &
          &                   label_list  = me(:)%header%label,                &
          &                   iScheme     = iScheme )
      end if
    end do

    ! write solver-specific character for passing solver-stuff to the
    ! treelm restart and tracking files
    call mus_writeSolSpecChar( scheme = me, tree = tree )

  end subroutine mus_load_schemes
!******************************************************************************!


!******************************************************************************!
  !> load single scheme defined in lua file with or without scheme handle
  !!
  subroutine mus_load_scheme_single( me, boundary, conf, params, pdfGlobal,    &
    &                                tree, path, parent )
    !---------------------------------------------------------------------------
    !> scheme type
    type( mus_scheme_type ), intent(inout) :: me
    !> boundary data from mesh
    type( tem_bc_prop_type ), intent(in):: boundary
    !> global parameter type
    type( param_type ), intent(inout) :: params
    !> flu state
    type( flu_State ) :: conf
    !> global pdf info
    type( pdf_globInfo_type) :: pdfGlobal
    !> global treelm mesh
    type( treelmesh_type), intent(in) :: tree
    !> lua path
    type( aot_path_type ),intent(in) :: path
    !> parent handle if scheme table is defined
    integer, intent(in), optional :: parent
    !--------------------------------------------------------------------------
    type( tem_variable_type ), allocatable :: tVar(:)
    integer :: loc_nTotal
    !--------------------------------------------------------------------------

    ! load scheme identifier table 'identify = { label = '', relaxation = '',...}'
    call mus_load_scheme_header( me      = me%header,                          &
      &                          conf    = conf,                               &
      &                          parent  = parent,                             &
      &                          scaling = params%scaling )

    ! load physics table for unit converstion
    call mus_load_physics( me      = me%physics,                               &
      &                    conf    = conf,                                     & 
      &                    parent  = parent,                                   &
      &                    tree    = tree,                                     & 
      &                    scaling = params%scaling )

    ! load fields from parent, including possible sources for each field
    call mus_load_fields( me           = me%field,                             &
      &                   nFields      = me%nFields,                           &
      &                   boundary     = boundary,                             &
      &                   conf         = conf,                                 &
      &                   pdfGlobal    = pdfGlobal,                            &
      &                   path         = path,                                 &
      &                   parent       = parent,                               &
      &                   schemeHeader = me%header,                            &
      &                   source       = me%source,                            &
      &                   physics      = me%physics )

    ! Compute lattice parameters from physics or vice versa 
    call mus_convertPhysics( physics      = me%physics,                        &
      &                      field        = me%field,                          &
      &                      nFields      = me%nFields,                        &
      &                      schemeHeader = me%header )

    ! load the global source variables
    call tem_load_source( me     = me%source(me%nFields + 1),                  &
      &                   conf   = conf,                                       &
      &                   parent = parent,                                     &
      &                   key    = 'source',                                   &
      &                   path   = path )

    ! Initialize the variable system
    call tem_var_sys_init( me%varSys )
    ! load tracking
    call tem_load_tracking( tCtrl          = me%trackingControl,               &
      &                     me             = me%track,                         &
      &                     conf           = conf,                             &
      &                     parent         = parent,                           &
      &                     path           = path,                             &
      &                     refTimeControl = params%timeControl,               &
      &                     prefix         = trim( me%header%label ))
  
    ! define the scheme stencil layout(set parameters, offsets, ...)
    call mus_init_stencil( stencil = me%layout%stencil,                        &
      &                    nBC     = size(me%field(1)%bc)*me%nFields )

    select case ( trim( me%header%layout ) )
      case( 'd2q9' )
        pdfGlobal%nDims = 2
        call mus_define_d2q9( layout = me%layout, tree = tree ) 
      case( 'd3q19' )
        pdfGlobal%nDims = 3
        call mus_define_d3q19( layout = me%layout, tree = tree ) 
      case( 'd3q27' )
        pdfGlobal%nDims = 3
        call mus_define_d3q27( layout = me%layout, tree = tree ) 
      case( 'd3q7' )
        pdfGlobal%nDims = 3
        call mus_define_d3q7( layout = me%layout, tree = tree ) 
      case( 'd3q6', 'flekkoy' )
        pdfGlobal%nDims = 3
        call mus_define_d3q6( layout = me%layout, tree = tree ) 
      case( 'd1q3' )
        pdfGlobal%nDims = 1
        call mus_define_d1q3( layout = me%layout, tree = tree ) 
      case ( 'new_stencil' )
        call tem_horizontalSpacer() 
        call tem_write( 'Reading the new layout...' )
        call mus_load_newLayout( me            = me%layout,                    &
          &                      parent_handle = parent,                       &
          &                      conf          = conf )
      case default
        call tem_horizontalSpacer() 
        call tem_write( 'The chosen scheme layout is not available. STOPPING!' )
        call tem_write( 'Choose one of:  ')
        call tem_write( '  - d2q9 ')
        call tem_write( '  - d3q19 ')
        call tem_write( '  - d3q7  ')
        call tem_write( '  - d3q6  ')
        call tem_write( '  - new_stencil  ')
        call tem_abort()
    end select
    ! finish stencil definition
    call mus_finish_stencil( layout = me%layout )

    call tem_write( 'Chosen stencil layout '//trim( me%header%layout ))
    call tem_write( '  Corresponding nDims ', pdfGlobal%nDims )

    call tem_horizontalSpacer()

    ! add analytical function loaded from lua file
    ! and in the end add error variable without dependency
    call tem_variable_from( me     = tVar,                                     &
      &                     conf   = conf,                                     &
      &                     parent = parent,                                   &
      &                     nTotal = loc_nTotal,                               &
      &                     key    = 'add_variable',                           &
      &                     path   = path )

    ! build a list of all derived variables (without prefixes)
    call mus_build_derVarList( derVarList = me%derVarList,                     &
      &                        schemeKind = me%header%kind,                    &
      &                        stencil    = me%layout%stencil(1) )

    ! initialize musubi state variables to the scheme variable system
    call mus_init_stateVars( me, size(tVar) )

    ! Now initialize all the available derived quantities
    call mus_init_varSys( derVarList = me%derVarList,                          &
      &                   var_sys    = me%varSys,                              &
      &                   label      = me%field(:)%label,                      &
      &                   tVar       = tVar )

    !\todo remove debug varsys out
!    call tem_var_sys_dump( me%varSys, dbgUnit )

    ! build derVarPos
    call mus_store_derVarPos( derVarPos = me%derVarPos,                        &
      &                       var_sys   = me%varSys,                           &
      &                       label     = me%field(:)%label )

  end subroutine mus_load_scheme_single
!******************************************************************************!


!******************************************************************************!
  !> Calls various initialization routines 
  !!
  !!\todo: SZ: THIS ROUTINE IS NOT USED IN THE CODE MAYBE WE CAN REMOVE IT!!!!!!
  !!
!  subroutine mus_init_schemes(me, pdfGlobal, tree)
!    !---------------------------------------------------------------------------
!    !> scheme type
!    type( mus_scheme_type ), intent(inout) :: me(:)
!    !> global pdf info
!    type( pdf_globInfo_type) :: pdfGlobal
!    !> global treelm mesh
!    type( treelmesh_type), intent(in) :: tree
!    !---------------------------------------------------------------------------
!    integer          :: iScheme
!    !---------------------------------------------------------------------------
!    call tem_write( 'init schemes' )
!    do iScheme = 1, size(me)
!      ! Initialize the variable system
!      call tem_var_sys_init( me(iScheme)%varSys )
!      
!      ! define the scheme stencil layout(set parameters, offsets, ...)
!      call mus_init_stencil(                                                   &
!        &         stencil = me(iScheme)%layout%stencil,                        &
!        &         nBC     = size(me(iScheme)%field(1)%bc)*me(iScheme)%nFields )
!      ! Initialize the predefined stencils only in case when new stencil is not
!      ! defined
!      if (.not. me(iScheme)%layout%new_stencil ) then
!        select case ( trim( me(iScheme)%header%layout ) )
!          case( 'd2q9' )
!            pdfGlobal%nDims = 2
!            call mus_define_d2q9( layout = me(iScheme)%layout, tree = tree ) 
!          case( 'd3q19' )
!            pdfGlobal%nDims = 3
!            call mus_define_d3q19( layout = me(iScheme)%layout, tree = tree ) 
!          case( 'd3q7' )
!            pdfGlobal%nDims = 3
!            call mus_define_d3q7( layout = me(iScheme)%layout, tree = tree ) 
!          case( 'd3q6', 'flekkoy' )
!            pdfGlobal%nDims = 3
!            call mus_define_d3q6( layout = me(iScheme)%layout, tree = tree ) 
!          case( 'd1q3' )
!            pdfGlobal%nDims = 1
!            call mus_define_d1q3( layout = me(iScheme)%layout, tree = tree ) 
!          case default
!            call tem_horizontalSpacer() 
!            call tem_write( 'The chosen scheme layout is not available. '//    &
!              &             'STOPPING!' )
!            call tem_write( 'Choose one of:  ')
!            call tem_write( '  - d2q9 ')
!            call tem_write( '  - d3q19 ')
!            call tem_write( '  - d3q7  ')
!            call tem_write( '  - d3q6  ')
!            call tem_write( '  - new_stencil  ')
!            call tem_abort()
!        end select
!      end if
!      
!      ! build a list of all derived variables (without prefixes)
!      call mus_build_derVarList( derVarList = me(iScheme)%derVarList,          &
!        &                        schemeKind = me(iScheme)%header%kind,         &
!        &                        stencil    = me(iScheme)%layout%stencil(1) )
!      
!      ! initialize musubi state variables to the scheme variable system
!      call mus_init_stateVars( me(iScheme), size(me(iScheme)%tVar) )
!
!      ! Now initialize all the available derived quantities
!      call mus_init_varSys( derVarList = me(iScheme)%derVarList,               &
!        &                   var_sys    = me(iScheme)%varSys,                   &
!        &                   label      = me(iScheme)%field(:)%label,           &
!        &                   tVar       = me(iScheme)%tVar )
!
!      ! build derVarPos
!      call mus_store_derVarPos( derVarPos = me(iScheme)%derVarPos,             &
!        &                       var_sys   = me(iScheme)%varSys,                &
!        &                       label     = me(iScheme)%field(:)%label )
!
!      ! Initialize the dependencies within this scheme (if there are any)
!      call tem_log( level = 1,                                                 &
!        &           msg   = 'depend allocated? '//                             &
!        &                   trim( tem_toStr( allocated( me(iScheme)%depend))))
!      if ( allocated( me(iScheme)%depend) ) then
!        call tem_init_depend( medepend    = me(iScheme)%depend,                &
!          &                   varSys      = me(:)%varSys,                      &
!          &                   resSysLabel = me(iScheme)%header%label,          &
!          &                   label_list  = me(:)%header%label,                &
!          &                   iScheme     = iScheme )
!      end if
!    end do ! iScheme
!
!
!    ! write solver-specific character for passing solver-stuff to the
!    ! treelm restart and tracking files
!    call mus_writeSolSpecChar( scheme = me, tree = tree )
!
!  end subroutine mus_init_schemes
!******************************************************************************!


!******************************************************************************!
  !> load multiple dependent variables of the scheme
  !!
  subroutine mus_convertPhysics( physics, field, nFields, schemeHeader )
    !---------------------------------------------------------------------------
    !>
    type( mus_physics_type ), intent(inout) :: physics
    !> number of fields defined in lua file
    integer, intent(in) :: nFields
    !>
    type( mus_field_type ), intent(inout) :: field( nFields )
    !> identifier of the scheme
    type( mus_scheme_header_type ), intent(in) :: schemeHeader
    !---------------------------------------------------------------------------
    integer :: iField
    !---------------------------------------------------------------------------
    select case( trim(schemeHeader%kind) )
    case('lbm', 'lbm_incomp')
      iField = 1
      if( physics%viscosity0 .lt. 0._rk ) then
        physics%viscosity0 =                                                   &
          &                field( iField )%fieldProp%fluid%kine_viscosityLB    &
          &                *physics%dx*physics%dx/physics%dt
        call tem_write('  Assigning physical viscosity from LB ')
        call tem_write('     kinematic shear viscosity:   ', physics%viscosity0)
      else
        call tem_write('  Assigning LB viscosity from physical ')
        call tem_write('     shear viscosity:      ',                          &
          &            real(physics%viscosity0))
        call tem_write(' Not yet implemented. This is still done in '//        &
          &            'mus_load_fluid')
        call tem_abort()
      end if 
    endselect
  end subroutine mus_convertPhysics
!******************************************************************************!


!******************************************************************************!
  !> load multiple dependent variables of the scheme
  !!
  subroutine mus_load_scheme_depend( me, conf, parent )
    !---------------------------------------------------------------------------
    !> scheme type
    type( mus_scheme_type ), intent(inout) :: me(:)
    !> flu state
    type( flu_State ) :: conf
    !>
    integer, intent(in) :: parent
    !---------------------------------------------------------------------------
    integer :: scheme_sub_handle, depend_handle, depend_sub_handle
    integer :: iScheme, iDepend, nDepend
    !---------------------------------------------------------------------------
    call tem_write( 'loading scheme depend table' )
    do iScheme = 1, size(me)
      call aot_table_open( L       = conf,                                     &
        &                  parent  = parent,                                   &
        &                  thandle = scheme_sub_handle,                        &
        &                  pos     = iScheme )
      
      call aot_table_open( L       = conf,                                     &
        &                  parent  = scheme_sub_handle,                        &
        &                  thandle = depend_handle,                            &
        &                  key     = 'depend' )
      if( depend_handle .gt. 0 ) then
        !check single table or multiple table
        call aot_table_open( L       = conf,                                   &
          &                  parent  = depend_handle,                          &
          &                  thandle = depend_sub_handle,                      &
          &                  pos     = 1 )
        if ( depend_sub_handle .eq. 0 ) then  
          allocate( me(iScheme)%depend(1) )
          call tem_load_depend_single( medepend = me(iScheme)%depend(1),       &
            &                          label    = me(iScheme)%header%label,    &
            &                          conf     = conf,                        &
            &                          thandle  = depend_handle )
        else
          call aot_table_close( L=conf, thandle=depend_sub_handle )
          nDepend = aot_table_length( L=conf, thandle=depend_handle )
          allocate( me(iScheme)%depend(nDepend) )
          do iDepend = 1, nDepend
            call aot_table_open( L       = conf,                               &
              &                  parent  = depend_handle,                      &
              &                  thandle = depend_sub_handle,                  &
              &                  pos     = iDepend )
            ! Load  variable system
            !\todo: SZ: why mus_load_depend_single and not mus_load_depend??
            call tem_load_depend_Single(                                       &
              &                       medepend = me(iScheme)%depend(iDepend),  &
              &                       label    = me(iScheme)%header%label,     &
              &                       conf     = conf,                         &
              &                       thandle  = depend_sub_handle )
            call aot_table_close( L = conf, thandle = depend_sub_handle )
          end do
        end if  
      else
        call tem_write( 'No depend table defined for scheme: '//               &
          & trim(me(iScheme)%header%label) )
      end if
      call aot_table_close( L=conf, thandle=depend_handle )

            
      call aot_table_close( L=conf, thandle=scheme_sub_handle )
    end do ! number of schemes
    
  end subroutine mus_load_scheme_depend
!******************************************************************************!


!******************************************************************************!
  !> Write solver-specific information (for e.g. the restart header file, which 
  !! is being handled in the Treelm) into a scratch file.
  !! The scratch file contents will then be read by the Treelm
  !! \todo KM: write stencil layout, fieldprop to solverspecific unit
  !!
  subroutine mus_writeSolSpecChar_array( scheme, tree )
   !----------------------------------------------------------------------------
   !> type holding state array, global information etc.
   type( mus_scheme_type ), intent(inout) :: scheme(:)
   !> global treelm mesh
   type( treelmesh_type), intent(in) :: tree
   !----------------------------------------------------------------------------
   integer :: iScheme
   !----------------------------------------------------------------------------
   ! Assemble solver specific lua table
   ! this might include a table of the field properties 
   do iScheme = 1, size(scheme) 
     call mus_writeSolSpecChar_single( scheme = scheme( iScheme ), tree = tree )
   enddo  

  end subroutine mus_writeSolSpecChar_array
!******************************************************************************!


!******************************************************************************!
  !> Write solver-specific information (for e.g. the restart header file, which 
  !! is being handled in the Treelm) into a scratch file.
  !! The scratch file contents will then be read by the Treelm
  !! \todo KM: write stencil layout, fieldprop to solverspecific unit
  !!
  subroutine mus_writeSolSpecChar_single( scheme, tree )
   !----------------------------------------------------------------------------
   !> type holding state array, global information etc.
   type( mus_scheme_type ), intent(inout) :: scheme
   !> global treelm mesh
   type( treelmesh_type), intent(in) :: tree
   !----------------------------------------------------------------------------
   integer :: nUnit
   type(aot_out_type) :: conf
   integer :: iChunk, strOff, nChunkChar, str_len, str_nChunk
   !----------------------------------------------------------------------------

   ! Assemble solver specific lua table
   ! this might include a table of the field properties 
   nUnit = newUnit()
   scheme%varSys%solSpecUnit = nUnit
   open(unit = nUnit, status='scratch')
   call aot_out_open( put_conf =  conf, outUnit = nUnit )
   !\todo write species properties
   call mus_scheme_header_out( me = scheme%header, conf = conf )
   call mus_field_out( me           = scheme%field,                            &
     &                 conf         = conf,                                    &
     &                 schemeHeader = scheme%header )
   call mus_weights_out( me = scheme%layout%weight, conf = conf )
   call mus_physics_out( me   = scheme%physics,                                &
     &                   conf = conf,                                          &
     &                   tree = tree )
   call aot_out_close( put_conf =  conf )

   ! transform the solver specific character to a single local string
   call tem_transform_solSpecChar( spc_unit = scheme%varSys%solSpecUnit,       &
     &                             string   = scheme%varSys%solspecString )
   ! open the solver specific string in lua format
   call open_config_chunk( L     = scheme%varSys%solSpec_conf,                 &
     &                     chunk = trim(scheme%varSys%solspecString) )
   call tem_log( level = 10,                                                   &
     &           msg   = 'solver specific string:' )
   ! KM: Fix for cray compiler writing to null unit with long string
   str_len = len(trim(scheme%varSys%solspecString))  
   str_nChunk = ceiling(real(str_len,kind=rk) / real(PathLen,kind=rk))
   do iChunk = 1, str_nChunk   
     strOff = ( (iChunk-1)*PathLen )
     nChunkChar = min(pathLen, str_len - strOff)
   call tem_log( level = 10,                                                   &
     & msg = trim(scheme%varSys%solspecString(strOff+1:(strOff+nChunkChar))) )
   end do

  end subroutine mus_writeSolSpecChar_single
!******************************************************************************!


!******************************************************************************!
  !\todo: what about this function??, 
  !> Take the solver specific information (from e.g. the restart header),
  !! and read it into the data structures using the lua read functions
  !!
  !! The information is stored in a lua style string.
  !!
  subroutine mus_readSolSpecChar( restart, scheme )
   !----------------------------------------------------------------------------
   !> the solver specific character
   type( tem_restart_type ) :: restart
   !> type holding state array, global information etc.
   type( mus_scheme_type ) :: scheme(:)
   !----------------------------------------------------------------------------
   integer :: iScheme, nSchemes ! scheme counter
   !----------------------------------------------------------------------------
   if( restart%comm%rank .eq. 0 ) then
     nSchemes = size( scheme )
     ! Read fluid data etc. from lua header file
     do iScheme = 1, nSchemes
       call tem_write(' Must read fluid props from restart header here')
     enddo
   endif

  end subroutine mus_ReadSolSpecChar
!******************************************************************************!


!******************************************************************************!
  !> Initialize the global variable system for this scheme.
  !!
  !! allocate function pointers, and append pdf to scheme%varSys
  subroutine mus_init_stateVars( scheme, extraVar )
    !---------------------------------------------------------------------------
    !> type holding state array, global information etc.
    type( mus_scheme_type ) :: scheme
    !> additional variables defined for this scheme
    integer, intent(in) :: extraVar
    !---------------------------------------------------------------------------
    integer :: pos, iField
    character(len=labelLen) ::  buffer
    !---------------------------------------------------------------------------
    call tem_write( 'Initializing global variable system ')
    
    ! allocate the varMap(function pointer) depending on possible multi-fields
    ! +1 in the end is for difference variable + number of mesh variables
    if( scheme%nFields .gt. 1 ) then
      allocate( scheme%varSys%derive(( 1 + size( scheme%derVarList ))          &
        &     * scheme%nFields + size( scheme%derVarList ) + extraVar + 1      &
        &     + nMeshVars ))
      ! +1 in the derVarPos is for total field variable like total density,
      ! total velocity   
      allocate( scheme%derVarPos( scheme%nFields + 1) )
    else
      allocate( scheme%varSys%derive( 1 + size( scheme%derVarList )            &
        &     + extraVar + 1 + nMeshVars ))
      allocate( scheme%derVarPos( 1 ) )
    end if

    do iField = 1, scheme%nFields
      write(buffer,'(a)') trim( scheme%field( iField )%label )//'pdf'
      call tem_var_append( sys         = scheme%varSys,                        &
        &                  varName     = trim(buffer),                         &
        &                  nComponents = scheme%layout%stencil(1)%QQ,          &
        &                  pos         = pos )
      scheme%varSys%derive( pos )%do => access_stateVar
    end do  

    ! Mark system as complete, so we can allocate the state vector
    ! and set the name of the variable system to the scheme label
    call tem_var_sys_mark( sys        = scheme%varSys,                         &
      &                    SystemName = trim( scheme%header%label )) 

  end subroutine mus_init_stateVars
!******************************************************************************!


!******************************************************************************!
  !> Destroy the arrays initialized in mus_init_stateVars
  !!
  subroutine mus_kill_stateVars( scheme )
    !---------------------------------------------------------------------------
    !> type holding state array, global information etc.
    type( mus_scheme_type ) :: scheme
    !---------------------------------------------------------------------------
    
    deallocate( scheme%varSys%derive)
    deallocate( scheme%derVarPos )
  end subroutine mus_kill_stateVars
!******************************************************************************!


!******************************************************************************!
  !> Initialize all source terms for the fields and the one for the scheme
  !!
  subroutine mus_init_sources( me, tree )
    !---------------------------------------------------------------------------
    !> scheme information
    type(mus_scheme_type) :: me
    !> The tree
    type(treelmesh_type)  :: tree
    !---------------------------------------------------------------------------
    integer :: iField
    integer :: mapSize
    character(len=labellen), allocatable :: tmp_prefix(:)
    !---------------------------------------------------------------------------

    call tem_write('Initializing the source terms ...')
    ! allocate and initialize the temporary array of prefixes
    ! size = nFields + 1 (for the scheme)
    allocate(tmp_prefix(me%nFields+1))

    tmp_prefix( 1:me%nFields )  = me%field( : )%label
    tmp_prefix(me%nFields + 1)  = ''

    
    ! initialize the basic source variable systems for each field
    call mus_build_sourceVarSys( me            = me%source,                    &
      &                          sourceVarList = me%sourceVarList,             &
      &                          stencil       = me%layout%stencil(1),         &
      &                          schemeKind    = trim(me%header%kind))

    do iField = 1, me%nFields + 1
      ! intersect the requested and the possible source variable systems
      call tem_var_intersect( systemA     = me%source(iField)%requ_varSys,     &
        &                     systemB     = me%source(iField)%poss_varSys,     &
        &                     resSys      = me%source(iField)%varSys,          &
        &                     resSysLabel = 'possInterRequ' )

    end do

    ! allocate the sourceMap with the number of variables in the global varSys
    ! plus the sum of those ones appended by the individual variable systems in
    ! the sources

    mapSize = me%varSys%variable%nVals
    do iField = 1, me%nFields + 1
      mapSize = mapSize + me%source( iField )%varSys%variable%nVals
    end do

    allocate( me%sourceMap%map( mapSize ))
    me%sourceMap%map = 0
    allocate( me%sourceMap%source( mapSize ))

    !\todo: SZ: remove the verbose statement
    call tem_reduce_sourceVars( me          = me%source,                       &
      &                         global_tree = tree )

    ! initialize the sources for all fields including the prefixes
    call mus_init_source( me            = me%source,                           &
      &                   global_varSys = me%varSys,                           &
      &                   sourceMap     = me%sourceMap,                        &
      &                   prefix        = tmp_prefix )

    ! copy the information back
!    me%source = tmp_sources

    ! update the varmap with the right function pointers
!    call mus_update_varmap( me = me%var,                     &
!      &                     varSys = me%varSys,              &
!      &                     sourceMap = me%sourceMap,        &
!      &                     sourceVarList = me%sourceVarList)

  end subroutine mus_init_sources
!******************************************************************************!


!******************************************************************************!
  !> 
  !!
!  subroutine mus_update_varmap( me, varSys, sourceMap, sourceVarList )
!    !------------------------------------------------------------------------ 
!    !> source information
!    type(mus_varMap_type), allocatable, intent(inout) :: me(:)
!    !> global variable system
!    type(tem_var_system_type), intent(in) :: varSys
!    !> source map to store the source type and the position of this source
!    !! variable in the global system
!    type(tem_sourceMap_type), intent(in) :: sourceMap
!    !> array of all possible source variables (without prefixes)
!    type(mus_variable_type), intent(in) :: sourceVarList(:)
!    !------------------------------------------------------------------------ 
!    integer :: iDir, pos, iLabel
!    integer :: iVar
!    integer :: iSource
!    character(len=labelLen) :: buffer
!    type(mus_varMap_type), allocatable :: tmp_varMap(:)
!    !------------------------------------------------------------------------ 
!
!    ! allocate the temporary array with the number of variables in the global varSys
!    allocate( tmp_varMap( varSys%variable%nVals ))
!    ! copy the function pointers from the old varmap
!    do iVar = 1, size(me)
!      tmp_varMap(iVar) = me(iVar)
!    end do
! 
!    ! set the function pointers for the source terms
!    do iVar = size(me) + 1 , varSys%variable%nVals 
!      ! get the pointer from the position of the 'pure' variable (no prefix)
!      ! in the poss_varSys (this is equal to the position in the sourceVarList)
!      pos = sourceMap%source( iVar )%varSys%variable%val(sourceMap%map(iVar))%mapping
!      tmp_varMap( iVar )%derive => sourceVarList( pos )%derive
!    end do
!
!    ! store the infromation in the varmap
!    deallocate(me)
!    allocate( me( varSys%variable%nVals ))
!    me = tmp_varMap
!
!  end subroutine mus_update_varmap
!******************************************************************************!


end module mus_scheme_module
!******************************************************************************!


!******************************************************************************!
!> \page scheme_implementation Scheme Implementation
!! The concept of schemes provides the user with a bigger flexibility. It is now
!! possible to run multiple simulations with different layouts on the same mesh.
!! In general the scheme implementation is able to run on old and new lua files.
!!
!! \subsection scheme_definition Scheme Definition
!!
!! The scheme is part of the individual solvers and contains all information about
!! initial conditions, boundary conditions, the \ref mus_scheme_layout_module 
!! "Scheme Layout", fluid and flow properties.
!! 
!! \subsection scheme_usage Usage
!!
!! To start a flow simulation one can either use the 'old' lua files or add the
!! table species with the following quantities:
!!
!! - a label
!! - the layout to be used
!! - the initial conditions
!! - the boundary conditions
!! - the fluid quantities
!!
!! \verbatim
!! -- Sample Flow Scheme Definition
!! scheme = {
!!  {  label = 'fluid',
!!   layout = 'predefined_d3q19',
!!   -- Initial condition 
!!   initial_condition = {
!!                        density = 1.0,
!!                        velocityX = 0.0,
!!                        velocityY = 0.0,
!!                        velocityZ = 0.0 },
!!   -- Boundary conditions
!!   boundary_condition = {
!!   { label = 'wall',
!!     kind = 'inlet_ubb',
!!     velocityX = 0.03, velocityY = 0.0, velocityZ = 0.0 },
!!   { label = 'wall',
!!     kind = 'wall',
!!     velocityX = 0.0, velocityY = 0.0, velocityZ = 0.0 }
!!   },
!!   fluid = { relaxation_scheme = 'BGK', 
!!             omega = 1.8, 
!!             rho0 = 1.0 } }
!! }
!! \endverbatim
!! 
