! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module contains the type definition(s) required in the scheme routines
!! In addition depend type and condition type for geometry increase routine 
!! are defined. Compute kernel definition is also defined in this module
!!
!! \author Simon Zimny, Kartik Jain
module mus_scheme_type_module

  ! include treelm modules
  use env_module,            only: rk, globalMaxLevels
  use tem_source_module,     only: tem_source_type
  use tem_tracking_module,   only: tem_tracking_type, tem_trackingControl_type
  use tem_var_system_module, only: tem_var_system_type
  use tem_variable_module,   only: tem_variable_type
  use tem_var_module,        only: tem_var_type
  use tem_source_module,     only: tem_sourceMap_type
  use treelmesh_module,      only: treelmesh_type
  use tem_depend_module,     only: tem_depend_type

  ! include musubi modules
  use mus_bc_header_module,     only: glob_boundary_type
  use mus_field_module,         only: mus_field_type
  use mus_field_prop_module,    only: mus_field_prop_type
  use mus_pdf_module,           only: pdf_data_type, pdf_global_type,          &
    &                                 pdf_globInfo_type
  use mus_scheme_layout_module, only: mus_scheme_layout_type
  use mus_scheme_header_module, only: mus_scheme_header_type
  use mus_variable_module,      only: mus_derVarPos_type
  use mus_physics_module,       only: mus_physics_type
  use mus_global_module,        only: param_type

  implicit none
  private
  
  public :: mus_scheme_type
  public :: kernel

  !> Datatype containing all information on the scheme.
  !!
  !! The mus_scheme_type contains of all information that are needed
  !! to run a simulation (including informations on the: fluid, boundary
  !! conditions, levelDescriptor, state vector, layout, diffusion info,
  !! tracking and restart).
  type mus_scheme_type
    !> the ID of the current scheme
    integer :: schemeID
    !> number of fields in the current scheme
    integer :: nFields = 0
    !> array of field type for each scheme
    type( mus_field_type ), allocatable :: field(:)
    !> array of boundary types contains elems of each boundary
    type( glob_boundary_type ), allocatable :: globBC(:)
    !> source information for the scheme and the fields (size: nFields + 1)
    type( tem_source_type ), allocatable :: source(:)
    !> identifier of the scheme
    type( mus_scheme_header_type ) :: header
    !> dependency of current scheme on other schemes
    type( tem_depend_type ), allocatable :: depend(:)
    !> one pdf_global_type including the level descriptor
    type( pdf_global_type ) :: glob
    !> pdf_data_types for every level including the state vector
    type( pdf_data_type ) :: pdf( globalMaxLevels )
    !> the scheme representation used in this scheme
    type( mus_scheme_layout_type ) :: layout
    !> function pointer to compute kernel
    procedure( kernel ),pointer,nopass :: compute => null() 
    !> the log type for the different schemes for the tracking
    type(tem_tracking_type),allocatable :: track(:)
    !> tracking status
    type(tem_trackingControl_type) :: trackingControl
    !> store position of derived variable each field and total field
    !! in the global system
    type(mus_derVarPos_type), allocatable :: derVarPos(:)
    !> global variable system definition
    type(tem_var_system_type) :: varSys
    !> map from the global varSys to the individual sources
    type( tem_sourceMap_type ) :: sourceMap
    !> array of all possible derived variables (without prefixes)
    type(tem_var_type), allocatable :: derVarList(:)
    !> array of all possible source variables (without prefixes)
    type(tem_var_type), allocatable :: sourceVarList(:)
    logical :: sourceActive = .false.
    !> contains basic SI units to convert from lattice to physical and 
    !! vice versa
    type(mus_physics_type) :: physics
    !> The initialized variable system
    type( tem_variable_type), allocatable :: tVar(:) 
  end type mus_scheme_type

  !> What does the kernel interface look like? 
  !! Every kernel's argument list must correspond to this one
  abstract interface
    subroutine kernel( iScheme, fieldProp, inState, outState, neigh, nStart,   &
      &                nEnd, level, layout, pdfGlobal, params, scheme, varSys, &
      &                tree )
      import :: rk, mus_field_prop_type, mus_scheme_layout_type,               &
        &       pdf_globInfo_type, mus_scheme_type, tem_var_system_type,       &
        &       treelmesh_type, param_type
      !> current scheme
      integer, intent(in) :: iScheme
      !> properties of the field
      type(mus_field_prop_type), intent(in) :: fieldProp(:)
      !> input pdf vector
      real(kind=rk), dimension(:), intent(in)  :: inState
      !> output pdf vector
      real(kind=rk), dimension(:), intent(out) :: outState
      !> connectivity vector 
      integer, dimension(:), intent(in) :: neigh
      !> start positions in pdf vector
      integer, intent(in) :: nStart
      !> end positions in pdf vector
      integer, intent(in) :: nEnd
      !> current level
      integer, intent(in) :: level
      !> the layout for this scheme
      type( mus_scheme_layout_type), intent(in) :: layout
      !> global info type
      type( pdf_globInfo_type ), intent(in) :: pdfGlobal
      !> global parameters
      type(param_type),intent(in) :: params
      !> array of all schemes
      type( mus_scheme_type ), intent(in) :: scheme(:)
      !> scheme variable system
      type( tem_var_system_type ), intent(in) :: varSys
      !> treelm mesh
      type( treelmesh_type ), intent(in) :: tree
    end subroutine kernel
  end interface

end module mus_scheme_type_module  
!******************************************************************************!
