module tem_precice_module
  use, intrinsic :: iso_c_binding
  use env_module,                only: rk, labelLen
  use aotus_module,              only: flu_State, aot_get_val, &
                                 &     aoterr_NonExistent, aoterr_fatal
  use aot_table_module,          only: aot_table_open, aot_table_close
  use aot_table_ops_module,      only: aot_table_length 
  use tem_aux_module,            only: tem_abort, tem_open_distconf
  use tem_tools_module,          only: tem_write, upper_to_lower
  use tem_comm_env_module,       only: process
  use tem_time_module,           only: tem_time_type
  use PreCICE_solver_if_module,  only: precicef_create, &
    &                                  precicef_initialize, &
    &                                  precicef_advance, &
    &                                  precicef_finalize, &
    &                                  precicef_set_read_pos, &
    &                                  precicef_set_write_pos, &
    &                                  precicef_ongoing, &
    &                                  precicef_get_data_id, &
    &                                  precicef_read_sdata, &
    &                                  precicef_read_vdata, &
    &                                  precicef_write_sdata, &
    &                                  precicef_write_vdata, &
    &                                  precicef_get_mesh_id
  implicit none

  private


  logical , parameter :: precice_available= .true.

  type tem_precice_type
  ! general info for precice
  ! name of solver, should be indetical to name in precice_configeFile
  character(len=labelLen) :: accessor
  character(len=labelLen) :: precice_configFile
  character(len=labelLen) :: mesh_name
  integer          :: meshID
  !> for reading data
  integer          :: n_readData=0 !--> length of datatable in Lua
  character(len=labelLen), allocatable :: read_data_name(:)
  integer, allocatable          :: read_dataID(:)
  !> for writing data
  integer          :: n_writeData=0 !--> length of datatable in Lua
  character(len=labelLen), allocatable :: write_data_name(:)
  integer, allocatable          :: write_dataID(:)
  !> Number of read positions ==> number of gauss points at bc 
  integer          :: n_read_points=0
  !> Number of write positions ==> number of gauss points at bc 
  integer          :: n_write_points=0
  end type tem_precice_type

  type(tem_precice_type) :: precice

  interface tem_spacetime_precice_for
    module procedure tem_spacetime_precice_scalar
    module procedure tem_spacetime_precice_vector
  end interface tem_spacetime_precice_for

  interface tem_precice_write
    module procedure tem_precice_write_scalar
    module procedure tem_precice_write_vector
  end interface tem_precice_write

  public :: tem_precice_load, tem_precice_init, tem_precice_create
  public :: tem_precice_advance
  public :: tem_precice_finalize
  public :: tem_precicef_ongoing
  public :: tem_spacetime_precice_for
  public :: tem_precice_write
  public :: tem_precicef_set_read_pos
  public :: tem_precicef_set_write_pos
  public :: precice
  public :: precice_available

contains

  subroutine tem_precice_load(conf)
    !---------------------------------------------------------------------------
    !> The filename to read the configuration from.
    type(flu_state), intent(in) :: conf
    !---------------------------------------------------------------------------
    integer                     :: thandle
    integer                     :: sub_thandle
    integer                     :: iError
    character(len=labelLen)     :: accessor
    character(len=labelLen)     :: precice_configFile
    integer                     :: i
    !---------------------------------------------------------------------------

    ! get the name of the solver and the config file from the lua script
    call aot_table_open(L = conf,  thandle = thandle, &
      &                 key = 'precice')

    if (thandle /= 0) then
      call aot_get_val(L = conf, thandle = thandle, key = 'accessor', &
        &              val = precice%accessor, default = 'Ateles', ErrCode = iError)
      call aot_get_val(L = conf, thandle = thandle, key = 'configFile', &
        &              val = precice%precice_configFile, default = 'config.xml', ErrCode = iError)
      call aot_get_val(L = conf, thandle = thandle, key = 'meshname', &
        &              val = precice%mesh_name, ErrCode = iError)
      if (btest(iError, aoterr_Fatal) ) then
        if (btest(iError, aoterr_NonExistent) ) then
         call tem_write('ERROR while init preCICE:' )
         call tem_write('      No mesh_name for exchange is specified in precice table' )
         call tem_write('      Stopping...')
         call tem_abort()
        else
         call tem_write('FATAL ERROR while init preCICE:' )
         call tem_write('      Stopping...')
         call tem_abort()
        end if
      end if

      ! get precice read data infos
      call aot_table_open( L       = conf,                                       &
        &                  parent  = thandle,                                    &
        &                  thandle = sub_thandle,                                &
        &                  key = 'exchange_data_read')

      if (sub_thandle .eq. 0) then
     !    call tem_write('ERROR while init preCICE:' )
         call tem_write('      No exchange data for reading specified in precice table' )
     !    call tem_write('      Stopping...')
     !    call tem_abort()
      else
         ! get the amount of exchange read data
         precice%n_readData =  aot_table_length(L = conf, thandle = sub_thandle)
         ! allocate the array to store the 'names' of this data
         allocate (precice%read_data_name(precice%n_readData))
         ! get the names of teh exchange data, need later on to get the dataID
         ! out of preCICE
         do i = 1, precice%n_readData
            call aot_get_val(L = conf, thandle = sub_thandle, &
              &              val =precice%read_data_name(i), ErrCode = iError, pos = i)
         end do
      end if
      call aot_table_close(L = conf, thandle = sub_thandle)

      ! get precice write data infos
      call aot_table_open( L       = conf,                                       &
        &                  parent  = thandle,                                    &
        &                  thandle = sub_thandle,                                &
        &                  key = 'exchange_data_write')

      if (sub_thandle .eq. 0) then
         call tem_write('      No exchange data for writting specified in precice table' )
      else
         ! geth the amount of exchange write data
         precice%n_writeData =  aot_table_length(L = conf, thandle = sub_thandle)
         ! allocate the array to store the 'names' of this data
         allocate (precice%write_data_name(precice%n_writeData))
         ! get the names of teh exchange data, need later on to get the dataID
         ! out of preCICE
         do i = 1, precice%n_writeData
            call aot_get_val(L = conf, thandle = sub_thandle, &
              &              val =precice%write_data_name(i), ErrCode = iError, pos = i)
         end do
      end if
      call aot_table_close(L = conf, thandle = sub_thandle)

      call tem_write('* preCICE asseccor = "'//trim(precice%accessor)//'"')
      call tem_write('* preCICE configFile = "'//trim(precice%precice_configFile)//'"' )
      call tem_write('* amout of exchange read data =' , precice%n_readData)
      do i = 1, precice%n_readData
         call tem_write('* preCICE read data = "'//trim(precice%read_data_name(i))//'"' )
      end do
      call tem_write('* amout of exchange write data =' , precice%n_writeData)
      do i = 1, precice%n_writeData
         call tem_write('* preCICE write data = "'//trim(precice%write_data_name(i))//'"' )
      end do
    else ! no precice table
      call tem_write('ERROR while init preCICE:' )
      call tem_write('      No Precice table is provided' )
      call tem_write('      Stopping...')
      call tem_abort()
    end if

    call aot_table_close(L = conf, thandle = thandle)
  end subroutine tem_precice_load 

  subroutine tem_precice_create
    !---------------------------------------------------------------------------
    character(len=labelLen, kind=c_char) :: c_accessor
    character(len=labelLen, kind=c_char) :: c_precice_configFile
    integer(kind=c_int) :: c_my_rank
    integer(kind=c_int) :: c_my_size
    integer(kind=c_int) :: c_accessorNameLength
    integer(kind=c_int) :: c_configFileNameLength
    character(len=labelLen, kind=c_char) :: c_MeshName
    integer(kind=c_int)    :: c_geometryNameLength
    integer(kind=c_int)    :: c_meshID
    character(len=labelLen, kind=c_char) :: c_dataName
    integer(kind=c_int)    :: c_dataNameLength
    integer(kind=c_int)    :: c_dataID
    integer                :: i
    !---------------------------------------------------------------------------
 
    c_accessor = precice%accessor
    c_precice_configFile = precice%precice_configFile
    c_my_rank = int(process%rank, kind=c_int)
    c_my_size = int(process%comm_size, kind=c_int)
    c_accessorNameLength = len_trim(precice%accessor)
    c_configFileNameLength = len_trim(precice%precice_configFile)

    call precicef_create(accessorName = c_accessor, &
      &                  configFileName = c_precice_configFile, &
      &                  solverProcessIndex = c_my_rank, &
      &                  solverProcessSize = c_my_size, &
      &                  lengthAccessorName = c_accessorNameLength, &
      &                  lengthConfigFileName = c_configFileNameLength)


    ! init the meshID
    c_geometryNameLength = len_trim(precice%mesh_name)
    c_meshName = precice%mesh_name

    call precicef_get_mesh_id(geometryName = c_meshName, &
      &                       meshID = c_meshID, &
      &                       lengthGeometryName = c_geometryNameLength) 
    precice%meshID=c_meshID
 
    !TODO if statement if there are reads
    if (precice%n_readData .gt. 0) then
       allocate(precice%read_dataID(precice%n_readData) )
       do i= 1, precice%n_readData
         c_dataNameLength = len_trim(precice%read_data_name(i))
         c_dataName = precice%read_data_name(i)
         call precicef_get_data_id(dataName = c_dataName, &
           &                       dataID   = c_dataID, &
           &                       lengthDataName = c_dataNameLength)
         precice%read_dataID(i) = c_dataID 
       end do
    end if
    
    write(*,*) 'ndata read = ', precice%n_readData
    write(*,*) 'dataIDs read = ', precice%read_dataID

    !init write data_id
    if (precice%n_writeData .gt. 0) then
       allocate(precice%write_dataID(precice%n_writeData) )
       do i= 1, precice%n_writeData
         c_dataNameLength = len_trim(precice%write_data_name(i))
         c_dataName = precice%write_data_name(i)
         call precicef_get_data_id(dataName = c_dataName, &
           &                       dataID   = c_dataID, &
           &                       lengthDataName = c_dataNameLength)
         precice%write_dataID(i) = c_dataID 
       end do
     end if
    
    write(*,*) 'ndata write = ', precice%n_writeData
    write(*,*) 'dataIDs write = ', precice%write_dataID
  
  end subroutine tem_precice_create
  
  subroutine tem_precice_init(timesteplimit)
    !---------------------------------------------------------------------------
    real(kind=rk), intent(inout) :: timestepLimit
    real(kind=c_double)    :: c_timestepLimit
    !---------------------------------------------------------------------------

    call tem_write('Initialize preCICE' )
    c_timestepLimit = timesteplimit
    call precicef_initialize(c_timestepLimit)
    timestepLimit=c_timestepLimit
  end subroutine tem_precice_init

  subroutine tem_precice_advance(timestepLimit)
    !---------------------------------------------------------------------------
    real(kind=rk), intent(inout) :: timestepLimit
    real(kind=c_double) :: c_timestepLimit
    !---------------------------------------------------------------------------

    c_timestepLimit = real(timestepLimit, kind=c_double)
    call precicef_advance(c_timestepLimit)
    ! do i need to convert the timestepLengthlimit from kind=c_double to kind=rk ??
    timestepLimit=c_timestepLimit
  end subroutine tem_precice_advance


  subroutine tem_precice_finalize()
    call precicef_finalize()
  end subroutine tem_precice_finalize
  
  subroutine tem_precicef_ongoing(isOngoing)
    !---------------------------------------------------------------------------
    integer, intent(inout)   :: isOngoing
    integer(kind=c_int) ::c_isOngoing  
    !---------------------------------------------------------------------------

  end subroutine tem_precicef_ongoing
  

  subroutine tem_precicef_set_read_pos(position, positionID )
    !---------------------------------------------------------------------------
    real(kind=rk), intent(in) :: position(3)
    integer, intent(out) :: positionID  
    !---------------------------------------------------------------------------
    integer(kind=c_int) :: c_meshID
    real(kind=c_double) :: c_pos(3)
    integer(kind=c_int) :: c_posID  
    !---------------------------------------------------------------------------

    c_meshID =  precice%meshID
    c_pos = real(position, kind=c_double)             
    call  precicef_set_read_pos(meshID = c_meshID, &
      &                         position = c_pos, &
      &                         positionID = c_posID ) 
    !! do i need to transfer the posID from c_int back to integer??
    positionID= int(c_posID) 

  endsubroutine tem_precicef_set_read_pos

  subroutine tem_precicef_set_write_pos(position, positionID )
    !---------------------------------------------------------------------------
    real(kind=rk), intent(in) :: position(3)
    integer, intent(out) :: positionID  
    !---------------------------------------------------------------------------
    integer(kind=c_int) :: c_meshID
    real(kind=c_double) :: c_pos(3)
    integer(kind=c_int) :: c_posID  
    !---------------------------------------------------------------------------

    c_meshID =  precice%meshID
    c_pos = real(position, kind=c_double)             
    call  precicef_set_write_pos(meshID = c_meshID, &
      &                         position = c_pos, &
      &                         positionID = c_posID ) 
    !! do i need to transfer the posID from c_int back to integer??
    positionID= int(c_posID) 

  endsubroutine tem_precicef_set_write_pos

  function tem_spacetime_precice_scalar(dataID, posID, npoints) result(res)
    !------------------------------------------------------------------------
    integer, intent(in)          :: dataID
    integer, intent(in)          :: posID(:)
    integer, intent(in)          :: npoints
    real(kind=rk) :: res(npoints) !< return value
    !------------------------------------------------------------------------
    integer(kind=c_int) :: c_dataID 
    integer(kind=c_int) :: c_dataNameLength
    integer(kind=c_int) :: c_positionID 
    real(kind=c_double) :: c_datavalue
    !------------------------------------------------------------------------
    integer :: ipoint
    !------------------------------------------------------------------------
 
    c_datavalue=10.0_c_double 

    do ipoint= 1, npoints  
    
      ! get and  convert int PosId to int_c Posid
      c_positionID= int(posID(iPoint), kind=c_int)
      ! get the data id from the dataid array (pressure,velocity)
      c_dataID= int(dataID, kind=c_int)

      call precicef_read_sdata( dataID = c_dataID, &
        &                       valueIndex = c_positionID, &
        &                       dataValue = c_datavalue)

      res(iPoint)=c_dataValue

    end do

  end function tem_spacetime_precice_scalar

  function tem_spacetime_precice_vector(dataID, posID, npoints, vector) result(res)
    !------------------------------------------------------------------------
    integer, intent(in)          :: dataID
    integer, intent(in)          :: posID(:)
   ! character(len=*), intent(in) :: dataName
    integer, intent(in)          :: npoints
    logical, intent(in)          :: vector
    real(kind=rk) :: res(npoints,3) !< return value
    !------------------------------------------------------------------------
    integer(kind=c_int) :: current_dataID 
    integer(kind=c_int) :: dataNameLength
    integer(kind=c_int) :: positionID 
    real(kind=c_double) :: datavalue(3)
    !------------------------------------------------------------------------
    integer :: ipoint
    !------------------------------------------------------------------------
    do ipoint= 1, npoints  
    
      ! get and  convert int PosId to int_c Posid
      positionID= int(posID(iPoint), kind=c_int)
      ! get the data id from the dataid array (pressure,velocity)
      current_dataID= int(dataID, kind=c_int)

      call precicef_read_vdata( dataID = dataID, &
        &                       valueIndex = positionID, &
        &                       dataValue = datavalue)

      ! TODO: convert real c_double to kind=rk ?? 
      res(iPoint, :)=dataValue(:)

    end do

  end function tem_spacetime_precice_vector
!******************************************************************************!

  subroutine tem_precice_write_scalar(dataID, posID, npoints, val) 
    !------------------------------------------------------------------------
    integer, intent(in)          :: dataID
    integer, intent(in)          :: posID(npoints)
    integer, intent(in)          :: npoints
    real(kind=rk), intent(in)    :: val(npoints)
    !------------------------------------------------------------------------
    integer(kind=c_int) :: c_dataID 
    integer(kind=c_int) :: c_dataName
    integer(kind=c_int) :: c_positionID 
    real(kind=c_double) :: c_datavalue
    !------------------------------------------------------------------------
    integer :: ipoint
    !------------------------------------------------------------------------

    write(*,*) 'in write_scalar'
    do ipoint= 1, npoints  
      
      ! get and  convert int PosId to int_c Posid
      c_positionID= int(posID(iPoint), kind=c_int)
      ! get the data id from the dataid array (pressure,velocity)
      c_dataID= int(dataID, kind=c_int)

      ! get and convert the value
      c_datavalue = real(val(ipoint), kind=c_double)

      call precicef_write_sdata( dataID = c_dataID, &
        &                        valueIndex = c_positionID, &
        &                        dataValue = c_datavalue)
  
    end do

  end subroutine tem_precice_write_scalar

  subroutine tem_precice_write_vector(dataID, posID, npoints, val) 
    !------------------------------------------------------------------------
    integer, intent(in)          :: dataID
    integer, intent(in)          :: posID(:)
    integer, intent(in)          :: npoints
    real(kind=rk), intent(in)    :: val(3,npoints)
    !------------------------------------------------------------------------
    integer(kind=c_int) :: c_dataID 
    integer(kind=c_int) :: c_dataName
    integer(kind=c_int) :: c_positionID 
    real(kind=c_double) :: c_datavalue(3)
    !------------------------------------------------------------------------
    integer :: ipoint
    !------------------------------------------------------------------------
    write(*,*) 'in write_vector'

    do ipoint= 1, npoints  
      
      ! get and  convert int PosId to int_c Posid
      c_positionID= int(posID(iPoint), kind=c_int)
      ! get the data id from the dataid array (pressure,velocity)
      c_dataID= int(dataID, kind=c_int)

      ! convert the value
      c_datavalue(:) = real(val(:,ipoint), kind=c_double)

      call precicef_write_vdata( dataID = c_dataID, &
        &                        valueIndex = c_positionID, &
        &                        dataValue = c_datavalue(:) )
  
    end do

  end subroutine tem_precice_write_vector

end module tem_precice_module
