! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Some auxilary routines, providing
!! frequently needed common tasks.
module tem_aux_module

  ! include treelm modules
  use mpi
  !$ use omp_lib
  use env_module,              only: rk, init_env, print_self_status, labelLen,&
     &                               printRuntimeInfo, pathLen
  use tem_comm_env_module,     only: tem_comm_env_init, tem_comm_env_fin,      &
    &                                process, tem_comm_env_type
  use tem_lua_requires_module, only: tem_require_track_rq,                     &
    &                                tem_get_required_lua,                     &
    &                                tem_pop_from_track_rq,                    &
    &                                tem_require_rq_store,                     &
    &                                tem_push_to_rq_store
  use tem_tools_module,        only: tem_write
  use tem_timer_module,        only: tem_addTimer, tem_startTimer,             &
    &                                tem_stopTimer, tem_getTimerVal,           &
    &                                tem_getTimerName, globalTimerHandle
  use tem_revision_module ! Providing parameters on the compilation environment

  ! include aotus modules
  use flu_binding,  only: flu_State, cbuf_type, flu_free_cbuf
  use aotus_module, only: open_config_buffer, open_config_file

  implicit none

  private

  public :: tem_start
  public :: tem_open_distconf
  public :: tem_open_distconf_array
  public :: tem_abort
  public :: tem_finalize
  public :: tem_unit_close
  public :: tem_checkLabel
  public :: tem_print_execInfo
  public :: utc_date_string

contains

!******************************************************************************!
  !> Initialize the environment. Should be the very first call in the program.
  !!
  subroutine tem_start(codeName, proc)
    !---------------------------------------------------------------------------
    character(len=*) :: codeName
    type(tem_comm_env_type), optional :: proc
    !---------------------------------------------------------------------------
    integer :: nProcs

    call init_env()
    if (present(proc)) then
      call tem_comm_env_init(proc)
      nProcs = proc%comm_size
    else
      call tem_comm_env_init(process)
      nProcs = process%comm_size
    end if

    call tem_write(" Starting up "//trim(codeName)//" with nprocs: ", nProcs )

    !$ call tem_write("               and nThreads pp: ", omp_get_max_threads())

    call tem_addTimer(globalTimerHandle,trim(codeName))
    call tem_startTimer(globalTimerHandle)

  end subroutine tem_start
!******************************************************************************!


!******************************************************************************!
  !> Read a Lua file on the first process and distribute it to all.
  !!
  !! \todo HK: Maybe deprecate and remove this routine in favor of
  !! TEM_open_distconf_array to avoid code duplication? Or keep it around and
  !! put a generic interface in place?
  !!
  !! This is a drop in replacement for open_config_file from Aotus and allows
  !! the scalable processing of Lua files, as they are read by a single process
  !! and then streamed to all in proc.
  !! There should be no restrictions on the Lua scripts themselves in this
  !! method, as it uses an overloading of the require mechanism in Lua itself to
  !! replace the file searches by lookups of buffered Lua code snippets.
  !! The execution of the Lua script itself is not changed.
  !!
  subroutine tem_open_distconf(L, fileName, proc)
    !---------------------------------------------------------------------------
    type(flu_State) :: L !< Opened Lua state with the loaded script.
    character(len=*), intent(in) :: fileName !< Name of the file to open.
    type(tem_comm_env_type), optional :: proc !< Process description to use.
    !---------------------------------------------------------------------------
    type(cbuf_type) :: scriptbuf
    type(cbuf_type) :: modbuf
    integer :: iError
    integer :: myrank
    integer :: comm
    integer :: nProcs
    integer :: nFiles
    integer :: bufsize
    integer :: iFile
    character(len=labelLen), allocatable :: req_label(:)
    character(len=pathLen), allocatable  :: req_file(:)
    !---------------------------------------------------------------------------

    if (present(proc)) then
      myrank = proc%rank
      comm = proc%comm
      nProcs = proc%comm_size
    else
      myrank = process%rank
      comm = process%comm
      nProcs = process%comm_size
    end if

    if (nProcs > 1) then

      if (myrank == 0) then
        ! Only rank 0 reads and executes the config file, while doing so, it
        ! keeps track of all the required files.
        call tem_require_track_rq(L)
        call open_config_file(L, trim(filename), buffer=scriptbuf)
        call tem_get_required_Lua( L, fileList = req_file,                     &
          &                        labelList = req_label)
        nFiles = size(req_Label)
      else
        ! Open the configuration on all other processes with the rq_store
        ! loaded:
        call tem_require_rq_store(L)
      end if

      ! Broadcast the number of required files
      ! MPI_BCAST(BUFFER, COUNT, DATATYPE, ROOT, COMM, IERROR)
      call MPI_Bcast(nFiles, 1, MPI_INTEGER, 0, comm, iError)

      if (myrank /= 0) allocate(req_label(nFiles))
      if (myrank /= 0) allocate(req_file(nFiles))

      ! Broadcast the module names
      call MPI_Bcast(req_label, nFiles*labelLen, MPI_CHARACTER, 0,             &
        &            comm, iError)
      ! Broadcast the file names
      call MPI_Bcast(req_file, nFiles*pathLen, MPI_CHARACTER, 0,               &
        &            comm, iError)

      ! Now go on opening all required files
      do iFile=1,nFiles
        if (myrank == 0) then
          call tem_pop_from_track_rq(L, trim(req_label(iFile)), modbuf)
          bufsize = size(modbuf%buffer)
        end if
        ! Broadcast the loaded script to all processes.
        ! MPI_BCAST(BUFFER, COUNT, DATATYPE, ROOT, COMM, IERROR)
        call MPI_Bcast(bufsize, 1, MPI_INTEGER, 0, comm, iError)
        if (myrank /= 0) allocate(modbuf%buffer(bufsize))
        call MPI_Bcast(modbuf%buffer, bufsize, MPI_CHARACTER, 0, comm, iError)
        if (myrank /= 0) then
          call tem_push_to_rq_store( L,                                        &
            &                        modname  = trim(req_label(iFile)),        &
            &                        filename = trim(req_file(iFile)),         &
            &                        buffer   = modbuf%buffer )
          deallocate(modbuf%buffer)
        else
          call flu_free_cbuf(modbuf)
        end if
      end do

      ! Broadcast the loaded script to all processes.
      if (myrank == 0) bufsize = size(scriptbuf%buffer)
      ! MPI_BCAST(BUFFER, COUNT, DATATYPE, ROOT, COMM, IERROR)
      call MPI_Bcast(bufsize, 1, MPI_INTEGER, 0, comm, iError)
      if (myrank /= 0) allocate(scriptbuf%buffer(bufsize))
      call MPI_Bcast(scriptbuf%buffer, bufsize, MPI_CHARACTER, 0, comm, iError)

      if (myrank /= 0) then
        call open_config_buffer(L = L, buffer = scriptbuf%buffer)
        deallocate(scriptbuf%buffer)
      else
        call flu_free_cbuf(scriptbuf)
      end if

    else

      ! Only a single process, no need for broadcasting.
      call open_config_file(L = L, filename = trim(fileName))

    end if

  end subroutine tem_open_distconf
!******************************************************************************!


!******************************************************************************!
  !> Open an array of Lua handles.
  !!
  !! This is a drop in replacement for open_config_file from Aotus and allows
  !! the scalable processing of Lua files, as they are read by a single process
  !! and then streamed to all in proc.
  !! There should be no restrictions on the Lua scripts themselves in this
  !! method, as it uses an overloading of the require mechanism in Lua itself to
  !! replace the file searches by lookups of buffered Lua code snippets.
  !! The execution of the Lua script itself is not changed.
  !!
  !! This variant of the routine opens an array of handles with the same script.
  !! It is used to provide individual Lua states, that can be used independently
  !! later on, for example in the evaluation of Lua functions for boundary
  !! conditions.
  !!
  subroutine tem_open_distconf_array(L, fileName, proc)
    !---------------------------------------------------------------------------
    type(flu_State) :: L(:) !< Opened Lua state with the loaded script.
    character(len=*), intent(in) :: fileName !< Name of the file to open.
    type(tem_comm_env_type), optional :: proc !< Process description to use.
    !---------------------------------------------------------------------------
    type(cbuf_type) :: scriptbuf
    type(cbuf_type) :: modbuf
    integer :: iError
    integer :: myrank
    integer :: comm
    integer :: nProcs
    integer :: nFiles
    integer :: nLuaStates
    integer :: bufsize
    integer :: iFile
    integer :: iState
    integer :: Lua_lb
    character(len=labelLen), allocatable :: req_label(:)
    character(len=pathLen), allocatable  :: req_file(:)
    !---------------------------------------------------------------------------

    nLuaStates = size(L)

    if (present(proc)) then
      myrank = proc%rank
      comm = proc%comm
      nProcs = proc%comm_size
    else
      myrank = process%rank
      comm = process%comm
      nProcs = process%comm_size
    end if

    if (myrank == 0) then
      Lua_lb = 2
    else
      Lua_lb = 1
    end if

    if (nProcs > 1) then

      if (myrank == 0) then
        ! Only rank 0 reads and executes the config file for the first Lua
        ! state, while doing so, it keeps track of all the required files.
        call tem_require_track_rq(L(1))
        call open_config_file(L(1), trim(filename), buffer=scriptbuf)
        call tem_get_required_Lua( L(1), fileList = req_file,                  &
          &                        labelList = req_label)
        nFiles = size(req_Label)

      end if
      ! Open the configuration on all other processes and the remaining local
      ! Lua states with the rq_store loaded:
      do iState=Lua_lb,nLuaStates
        call tem_require_rq_store(L(iState))
      end do

      ! Broadcast the number of required files
      ! MPI_BCAST(BUFFER, COUNT, DATATYPE, ROOT, COMM, IERROR)
      call MPI_Bcast(nFiles, 1, MPI_INTEGER, 0, comm, iError)

      if (myrank /= 0) allocate(req_label(nFiles))
      if (myrank /= 0) allocate(req_file(nFiles))

      ! Broadcast the module names
      call MPI_Bcast(req_label, nFiles*labelLen, MPI_CHARACTER, 0,             &
        &            comm, iError)
      ! Broadcast the file names
      call MPI_Bcast(req_file, nFiles*pathLen, MPI_CHARACTER, 0,               &
        &            comm, iError)

      ! Now go on opening all required files
      do iFile=1,nFiles
        if (myrank == 0) then
          call tem_pop_from_track_rq(L(1), trim(req_label(iFile)), modbuf)
          bufsize = size(modbuf%buffer)
        end if
        ! Broadcast the loaded script to all processes.
        ! MPI_BCAST(BUFFER, COUNT, DATATYPE, ROOT, COMM, IERROR)
        call MPI_Bcast(bufsize, 1, MPI_INTEGER, 0, comm, iError)
        if (myrank /= 0) allocate(modbuf%buffer(bufsize))
        call MPI_Bcast(modbuf%buffer, bufsize, MPI_CHARACTER, 0, comm, iError)
        do iState=Lua_lb,nLuaStates
          call tem_push_to_rq_store( L(iState),                                &
            &                        modname  = trim(req_label(iFile)),        &
            &                        filename = trim(req_file(iFile)),         &
            &                        buffer   = modbuf%buffer )
        end do
        if (myrank == 0) then
          call flu_free_cbuf(modbuf)
        else
          deallocate(modbuf%buffer)
        end if
      end do

      ! Broadcast the loaded script to all processes.
      if (myrank == 0) bufsize = size(scriptbuf%buffer)
      ! MPI_BCAST(BUFFER, COUNT, DATATYPE, ROOT, COMM, IERROR)
      call MPI_Bcast(bufsize, 1, MPI_INTEGER, 0, comm, iError)
      if (myrank /= 0) allocate(scriptbuf%buffer(bufsize))
      call MPI_Bcast(scriptbuf%buffer, bufsize, MPI_CHARACTER, 0, comm, iError)

      do iState=Lua_lb,nLuaStates
        call open_config_buffer(L = L(iState), buffer = scriptbuf%buffer)
      end do
      if (myrank == 0) then
        call flu_free_cbuf(scriptbuf)
      else
        deallocate(scriptbuf%buffer)
      end if

    else

      ! Only a single process, no need for broadcasting.
      do iState=1,nLuaStates
        call open_config_file(L = L(iState), filename = trim(fileName))
      end do

    end if

  end subroutine tem_open_distconf_array
!******************************************************************************!


!******************************************************************************!
  !> Abort the program with finalization of the environment
  !!
  subroutine tem_abort()
    !---------------------------------------------------------------------------
    integer :: iError
    !---------------------------------------------------------------------------

    ! \todo JZ: commented out the the tem_finalize here: In case that one rank
    ! while the other ranks are still waiting for communication the solver
    ! will not terminate.
    !call tem_finalize()
    call tem_write( "" )
    call tem_write( " !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
    call tem_write( "               Aborting. " )
    call tem_write( " !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
    call tem_write( "" )
    call mpi_abort(MPI_COMM_WORLD, 1, iError)

  end subroutine tem_abort
!******************************************************************************!


!******************************************************************************!
  !> Finalize the environment, should be the very last call in the program.
  !!
  subroutine tem_finalize(proc)
    !--------------------------------------------------------------------------
    ! The process communicator to close
    type(tem_comm_env_type), optional :: proc
    !---------------------------------------------------------------------------
    character(len=64) :: timerName
    real(kind=rk) :: timerValue
    type(tem_comm_env_type) :: local_proc
    !---------------------------------------------------------------------------

    if (present(proc)) then
      local_proc = proc
    else
      local_proc = process
    end if

    if ( printRuntimeInfo .and. (local_proc%rank == 0) ) then
      call print_self_status()
    end if

    call tem_comm_env_fin(proc = local_proc)
    call tem_stopTimer(globalTimerHandle)

    timerName  =  tem_getTimerName(globalTimerHandle)
    timerValue =  tem_getTimerVal(globalTimerHandle)
    call tem_write(" Done with " // trim(timerName) // " in [s]",              &
      &            timerValue, before = 2, after = 2)

  end subroutine tem_finalize
!******************************************************************************!


!******************************************************************************!
  !> Check, if a unit is open, and close it
  !!
  subroutine tem_unit_close(me)
    !------------------------------------------------------------------------
    !> the restart type to close
    integer :: me
    !------------------------------------------------------------------------
    logical :: nUnitOpened
    !------------------------------------------------------------------------
    ! Check, if any open units have to be closed
    inquire(unit= me, opened=nUnitOpened)
    if (nUnitOpened) then
      close( me )
    end if

  end subroutine tem_unit_close
!******************************************************************************!


!******************************************************************************!
  !> check whether the current label is already given
  !!
  subroutine tem_checkLabel(label, nLabels)
   !-------------------------------------------------------------------------
   !> holding array of labels with label(n) contains current label
   character(len=*), intent(in) :: label(:)
   !> Number of schemes already existing in the scheme array
   !! (they are being added currently, so we only have to compare against
   !! the ones coming before the current one, up to nSchemes-1)
   integer, intent(in) :: nLabels
   !-------------------------------------------------------------------------
   integer :: iLabel
   !-------------------------------------------------------------------------
   call tem_write(' Comparing the labels...')
   do ilabel = 1, nLabels - 1
     ! Check the labels here
     if ( trim(label(ilabel)) == trim(label(nLabels)) ) then
       call tem_write('Error: identical label have been encountered.')
       call tem_write('       Please specify a unique label for multiple '//   &
         &            'tables in the config file.')
       call tem_write("Example: scheme = {{ label = 'scheme1' , ...  }, ")
       call tem_write("                   { label = 'scheme2' , ...  }} ")
       call tem_write("                                                 ")
       call tem_write("Stopping...       ")
       call tem_abort()
      end if
   end do

  end subroutine tem_checkLabel
!******************************************************************************!


!******************************************************************************!
  !> Print information of the executable
  !!
  subroutine tem_print_execInfo()
    !---------------------------------------------------------------------------
    integer :: flagline
    !---------------------------------------------------------------------------

    call tem_write('+---------------------------------------------------------')
    call tem_write('| INFORMATION ON THE EXECUTABLE')
    call tem_write('| Revision of the code in this executable: '//             &
      &            trim(tem_solver_revision))
    call tem_write('|')
    call tem_write('| Compiled with '//trim(tem_FC_name)//' in version '//     &
      &            trim(tem_FC_version))
    call tem_write('| Using the command '//trim(tem_FC_command))
    call tem_write('| And the following flags:')
    do flagline = 1, tem_FC_nFlagLines
      call tem_write('| '//trim(tem_FC_flags(flagline)))
    end do
    call tem_write('|')
    call tem_write('| Build date: '//trim(tem_build_date))
    call tem_write('+---------------------------------------------------------')

  end subroutine tem_print_execInfo
!******************************************************************************!


!******************************************************************************!
  !> Function to create a human readable UTC date string.
  !!
  !! The resulting string has 26 characters.
  !!
  function utc_date_string() result(dat_string)
    !---------------------------------------------------------------------------
    character(len=26) :: dat_string
    !---------------------------------------------------------------------------
    integer, parameter :: year = 1
    integer, parameter :: month = 2
    integer, parameter :: day = 3
    integer, parameter :: hour = 5
    integer, parameter :: minute = 6
!    integer, parameter :: second = 7
    integer, parameter :: utc_diff = 4
    character(len=9) :: u_off_string
    integer :: off_min, off_hour
    integer :: dat(8)
    !---------------------------------------------------------------------------

    call date_and_time(values=dat)
    off_min = mod(dat(utc_diff),60)
    off_hour = dat(utc_diff)/60
    if (dat(utc_diff) >= 0) then
      write(u_off_string,'(a4,i2.2,a1,i2.2)') 'UTC+', off_hour, ':', off_min
    else
      write(u_off_string,'(a3,i3.2,a1,i2.2)') 'UTC', off_hour, ':', off_min
    end if
    write(dat_string,'(i4,4(a1,i2.2),a10)') dat(year), '-',                    &
      &     dat(month), '-', dat(day), ' ', dat(hour), ':',                    &
      &     dat(minute), ' '//u_off_string

  end function utc_date_string
!******************************************************************************!


end module tem_aux_module
!******************************************************************************!
