! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Provide load balancing algorithms
!!
!! The implementation is based on the Sparta Algorithm [1].
!!
!! Some elements which require further
!! treatment expose higher computational effort. This includes among others the
!! enforcing of boundary conditions, source terms, or the interpolation between
!! levels.  Depending on the selected behavior of elements, this result in
!! considerable higher effort on certain elements.
!! When running in parallel, it is important to balance the load between the
!! participating processing units equally in order to achieve a minimal time to
!! solution. It is always desired to minimize the waiting time of processing
!! units.
!! The initial disitribution of the elements among the processes assigns the
!! roughly the same amount of elements to each partition, regardless of their
!! cost.
!! This simple partitioning yields very good results for uniform grid
!! simulations, as the elements with additional behavior is of the order of
!! $\mathcal{O}(n^2)$ while the total number of elements are of order
!! $\mathcal{O}(n^3)$.
!! 
!! Especially when using locally refined grids, the computational cost of
!! elements varies considerably. Besides the increased update interval of the
!! smaller elements, the interpolation routines are very expensive compared to
!! the pure computation of an element.
!! 
!! To account for this variation of the computational cost, a load balancing
!! algorithm [1] is employed to homogenize the total cost among
!! the partitions. It is based on a space-filling curve, which fits nicely
!! into the TreElm framework. Additionally, each element of the tree is
!! assigned a weight representing the computational cost. The total cost is
!! then possibly equally distributed  among the partitions which results in a
!! re-partitioning of the mesh along the space-filling curve.
!! 
!! [1]  D. F. Harlacher, C. Siebert, H. Klimach, S.
!! Roller, and F. Wolf, 'Dynamic Load Balancing for Unstructured Meshes on
!! Space-Filling Curves,' presented at the IPDPSW '12 Proceedings of the 2012
!! IEEE 26th International Parallel and Distributed Processing Symposium
!! Workshops, Shanghai, 2012, pp. 1661–1669.
!!
!! \author Daniel Harlacher, Manuel Hasert
!!
module tem_balance_module

  ! include treelm modules
  use mpi
  use env_module,             only: rk, long_k, rk_mpi, long_k_mpi, labelLen
  use flu_binding,            only: flu_State
  use tem_timeControl_module, only: tem_timeControl_type, tem_load_TimeControl
  use tem_tools_module,       only: tem_write, tem_horizontalSpacer
  use tem_comm_env_module,    only: process
  use tem_logging_module,     only: tem_log, tem_toStr, logUnit

  ! include aotus modules
  use aotus_module,     only: aot_get_val
  use aot_table_module, only: aot_table_open, aot_table_close, aot_get_val

  implicit none
  private

  public :: tem_balance_type
  public :: tem_balance
  public :: tem_balance_evaluateImbalance
  public :: tem_balance_loadConfig

  !> Load balancing information and control 
  type tem_balance_type
    !> load balancing type
    character(len=labelLen) :: balanceKind
    !> is load balancing activated?
    logical :: active
    !> is dynamic load balancing activated?
    logical :: dynamic
    !> Imbalance: max( timeMainLoop( iProc )) - average( timeMainLoop )
    real( kind=rk ) :: imbalanceAbs
    !> Imbalance: max( timeMainLoop( iProc )) - average( timeMainLoop )/average
    real( kind=rk ) :: imbalanceRel
    !> base weight for each element which is set to be
    !! weight( element ) = baseWeight^( level(element) -refLevel )
    real( kind=rk ) :: baseWeight
    !> control about when to do dynamic balancing
    type(tem_timeControl_type) :: timeControl
  end type tem_balance_type

  contains

!******************************************************************************!
  !> Read all the configuration options for load balancing from the
  !! configuration.
  !!
  !! To activate the load balancing, use 
  !! \verbatim
  !! balance = { active = true, kind = 'simple', baseweight = 2. }
  !! \endverbatim
  !!
  subroutine tem_balance_loadConfig( me, conf, parent, key, defaultActive,     &
      &                              globalTime )
    !---------------------------------------------------------------------------
    !> The load balancing type
    type(tem_balance_type), intent(inout) :: me
    !> lua configuration handle
    type(flu_state) :: conf
    !> parent of the balance table
    integer, intent(in), optional :: parent
    !> key in the lua, if not balance
    character(len=*),intent(in), optional :: key
    !> The default option for activating load balancing 
    !! if nothing is given in the lua file
    logical, intent(in), optional :: defaultActive
    type(tem_timeControl_type)    :: globalTime
    !---------------------------------------------------------------------------
    integer :: iError, thandle
    logical :: local_defaultActive
    character(len=labelLen) :: localKey
    character(len=labelLen) :: local_defaultString
    !---------------------------------------------------------------------------
    if( present(key) ) then
      localKey = key
    else
      localKey = 'balance'
    endif
    if( present(defaultActive) ) then
      local_defaultActive = defaultActive
    else
      local_defaultActive = .false.
    endif
    local_defaultString = 'simple'
    me%dynamic = .false.

    call aot_table_open(L       = conf,                                        &
      &                 parent  = parent,                                      &
      &                 thandle = thandle,                                     &
      &                 key     = localKey )
    ! Check here if the table was successfully opened. 
    ! If not, try to read the keyword directly
    if( thandle .ne. 0 ) then
      ! table is open !
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               key     = 'active',                                    &
        &               val     = me%active,                                   &
        &               ErrCode = iError,                                      &
        &               default = local_defaultActive )

      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               key     = 'dynamic',                                   &
        &               val     = me%dynamic,                                  &
        &               ErrCode = iError,                                      &
        &               default = .false. )
    
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               key     = 'kind',                                      &
        &               val     = me%balanceKind,                              &
        &               ErrCode = iError,                                      &
        &               default = local_defaultString )
   
      !> Load the time controls for dynamic load balancing
      call tem_load_timeControl( me     = me%timeControl,                      &
        &                        conf   = conf,                                &
        &                        parent = thandle )

    else
      ! Now test if balance is given as a logical 
      call aot_get_val( L       = conf,                                        &
        &               key     = localKey,                                    &
        &               val     = me%active,                                   &
        &               ErrCode = iError,                                      &
        &               default = local_defaultActive )
      ! Was it given as a logical?
      if( iError .ne. 0 ) then
        ! There was an error, so not given as an error
        call aot_get_val( L       = conf,                                      &
          &               key     = localKey,                                  &
          &               val     = me%balanceKind,                            &
          &               ErrCode = iError,                                    &
          &               default = local_defaultString )
      end if
    end if
    call aot_table_close(L=conf, thandle=thandle)
    ! If the balance type is none, deactivate it
    if( trim( me%balanceKind ) == 'none' )  then
      me%active = .false.
    end if

    if( .not. me%dynamic ) then
      me%timeControl%iInterval  = globalTime%iMax
      me%timeControl%iMin       = globalTime%iMax
      me%timeControl%iMax       = globalTime%iMax
    end if

    if( me%active ) then
      write(logUnit(1),*) 'Load Balancing is active with kind '//            & 
        &                    trim( me%balanceKind )
    endif

  end subroutine tem_balance_loadConfig
!******************************************************************************!


!******************************************************************************!
  !> Evaluate the imbalance of all the processes
  !! by each rank.
  !!
  subroutine tem_balance_evaluateImbalance( me, CPUcost, nParts, comm)
    !---------------------------------------------------------------------------
    !> Load balance type to store the result to
    type(tem_balance_type), intent(inout) :: me
    !> Number of partitions, the mesh is partitioned into (= Number of MPI
    !! processes in comm).
    integer, intent(in) :: nParts
    !> MPI Communicator
    integer, intent(in) :: comm
    !> each process's cpu cost. Basis to evaluate the imbalance
    real(kind=rk), intent(in) :: CPUcost
    !---------------------------------------------------------------------------
    ! computation cost over all processes
    real(kind=rk) :: totalCPUcost
    ! maximum computation cost in one rank
    real(kind=rk) :: maxCPUcost
    ! average computation cost over all processes
    real(kind=rk) :: average
    integer :: iErr ! error handle
    !---------------------------------------------------------------------------

    ! Get the sum of total CPU cost
    call mpi_reduce( CPUcost, totalCPUcost, 1, rk_mpi, mpi_sum, 0, comm, iErr)
    average = totalCPUcost/real( nParts, kind=rk )

    ! Get the maximum process CPU cost
    call mpi_reduce( CPUcost, maxCPUcost, 1, rk_mpi, mpi_max, 0, comm, iErr)
    me%imbalanceAbs = maxCPUcost - average
    if( average .gt. 0._rk ) then
      me%imbalanceRel = (maxCPUcost - average)/average
    else
      me%imbalanceRel = 0._rk
    end if

    if( process%comm_size .gt. 1 ) then
      call tem_horizontalSpacer
      call tem_write(' Load-balancing evaluation ')
      call tem_write('   max. absolute imbalance  [s]:',                       &
        &                real(me%imbalanceAbs ))
      call tem_write('   max. relative imbalance  [%]:',                       &
        &                real(me%imbalanceRel*100._rk ))
      call tem_horizontalSpacer
    end if

  end subroutine tem_balance_evaluateImbalance
!******************************************************************************!


!******************************************************************************!
  !> Return splitting positions based on the weights provided
  !! by each rank.
  !! 
  !! This is the SPARTA algorithm which uses simple splitting based on 
  !! given weights for all elements in the mesh.
  !!
  subroutine tem_balance(weight, myPart, nParts, offset, comm)
    !---------------------------------------------------------------------------
    !> Sorted list of weights corresponding to treeID order
    real(kind=rk),intent(in)  :: weight(:)
    integer,intent(in) :: myPart !< Rank of the calling process
    !> Number of procs the distribution should span
    integer, intent(in) ::  nParts
    !> Array of offsets with the size nParts. Offset index starts at 0.
    !! This Array needs to be allocate and deallocated outside
    integer(kind=long_k),intent(out), allocatable :: offset(:)
    !> MPI Communicator
    integer, intent(in) :: comm
    !---------------------------------------------------------------------------
    integer :: iErr  ! MPI error variable
    integer :: iElem ! loop variables
    integer :: myElems ! Number of elements on a proc
                       !  Can be easily made intent(inout) if needed
    integer(kind=long_k) :: myElems_long
    real(kind=rk) :: w_sum, w_opt
    real(kind=rk) :: send, recv ! Send and receive buffers for MPI calls
    ! boundary values of the elements in which we search for splitters
    real(kind=rk) :: lower_boundary, upper_boundary
    real(kind=rk), allocatable :: send_sum(:)
    ! Array in which the splitter positions are stored
    integer, allocatable :: split(:) 
    ! Count variables that state which rank gets how many elements
    ! *_count(rank0, ...)
    ! Right now the size is nParts but that will be changed in
    ! the near future to avoid O(p) allocations
    integer, allocatable :: send_count(:), recv_count(:)
    ! local prefix sum array of myElems
    real(kind=rk), allocatable :: presum(:)
    integer :: rmin, rmax, lb, ub, left_off, mid
    real(kind=rk) :: opt_split, wsplit
    !---------------------------------------------------------------------------

    ! allocate offset array
    allocate( offset( nParts ))
    ! Get size of weight vector
    myElems = size(weight)

    ! Allocate Array for Prefix sum
    allocate(presum(myElems))
    ! Prefix sum over local weights. later on we will look for the splitter in
    ! this prefix sum
    presum(1) = weight(1)
    do iElem = 2,myElems
      presum(iElem) = presum(iElem-1) + weight(iElem)
    end do
    send = presum(myElems)
    ! sum up global total weight
    call MPI_ALLREDUCE(send, recv, 1, rk_mpi, mpi_sum, comm, iErr)

    w_sum = recv
    ! Calculate global optimum
    w_opt = w_sum/nParts

    ! Global prefix sum for weights
    call MPI_EXSCAN(send, recv, 1, rk_mpi, mpi_sum, comm, iErr)

    ! initialize splitter search
    lower_boundary=recv
    if (myPart == 0) lower_boundary = 0
    upper_boundary = lower_boundary + presum(myElems)

    allocate(send_sum(0:nParts))
    allocate(send_count(0:nParts-1))
    allocate(split(0:nParts-1))
    allocate(recv_count(0:nParts-1))

    send_count = 0
    split = 0
    recv_count = 0
    rmin = max(floor(lower_boundary/w_opt),0)
    rmax = min(ceiling(upper_boundary / w_opt),nParts-1)

    ! Do splitter search
    left_off = 1
    do iElem = rmin,rmax
      lb = left_off
      ub = myelems
      opt_split = (iElem+1)*w_opt
      if (iElem*w_opt < upper_boundary) then
      do
        mid = (lb+ub)/2
        wsplit = presum(mid) + lower_boundary
        if (wsplit == opt_split) exit
        if (wsplit < opt_split) then
           lb = mid
        else
           ub = mid
        end if
        ! exit if a single element was found, need to do this
        if (lb >= ub-1) exit
        !                    here, to have mid and wsplit set.
      end do
      if (ABS(wsplit - opt_split) > ABS(wsplit - opt_split - weight(mid))) then
        mid = mid - 1 ! return 0 if the splitter is left of the lower boundary
      else
        if (mid+1 <= myElems) then
          if (ABS(wsplit - opt_split)                                          &
            & > ABS(wsplit - opt_split + weight(mid+1))) then
            mid = mid + 1 ! return myElems at most
          end if
        else
          if (opt_split > upper_boundary) mid = myElems
        end if
      end if
      split(iElem) = mid
          send_count(iElem) = mid - left_off + 1
          left_off = mid + 1
      end if
    end do
    ! finished splitter search. Communciate results.
    ! Each process needs to know how many elements to receive from which process
    call mpi_alltoall(send_count, 1, mpi_integer, recv_count, 1, mpi_integer,  &
      &               comm, iErr)

    ! total number of my elements after exchanging elements
    myElems_long = int(sum(recv_count), kind=long_k)

    ! exchange total number of elements on all processes
    call mpi_allgather(myElems_long, 1, long_k_mpi, offset, 1, long_k_mpi,     &
      &                comm, iErr)

    ! Create offset list
    do iElem=2, nParts
      offset(iElem) = offset(iElem) + offset(iElem-1)
    end do

    deallocate(send_count, recv_count, send_sum, split)

  end subroutine tem_balance
!******************************************************************************!

end module tem_balance_module
!******************************************************************************!
