! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> The tem_bc_header is providing an overview of the boundary conditions
!! defined in the configuration of the solver.
!!
!! It provides collects the
!! labels and the kinds of the configured BCs, and connects them to the
!! corresponding boundaries defined in the mesh, given in the BC_prop.
!!
module tem_bc_header_module

  ! include treelm modules
  use env_module,         only: LabelLen
  use tem_aux_module,     only: tem_abort
  use tem_bc_prop_module, only: tem_bc_prop_type
  use tem_tools_module,   only: tem_write

  ! include aotus modules
  use aot_table_module, only: aot_table_open, aot_table_close, aot_get_val,    &
    &                         aot_table_length
  use flu_binding,      only: flu_State

  implicit none

  private

  public :: tem_bc_header_type
  public :: tem_load_bc_header, tem_open_bc

  !> This type describes the general, not solver specific, header information
  !! given in the Lua configuration of the solvers.
  type tem_bc_header_type
    integer :: nBCs !< Number of Bounary conditions in the Lua config
    !> Label for each boundary condition
    character(len=LabelLen), allocatable :: label(:)
    !> The kind of each boundary condition
    character(len=LabelLen), allocatable :: BC_kind(:)
    !> ID of the boundary in the mesh property, as defined by the bc_prop
    !! If an entry is not positive, there is no corresponding boundary
    !! condition found.
    integer, allocatable :: BC_ID(:)
  end type


contains


!******************************************************************************!
  !> This subroutine reads in the boundary conditions specified in the
  !! configuration file, and connects them to the corresponding entries in the
  !! treelmesh.
  !! If there boundary conditions in the mesh, for which no configuration is
  !! found, the program is aborted!
  !!
  subroutine tem_load_bc_header( me, conf, parentHandle, BC_prop )
    !---------------------------------------------------------------------------
    !> The boundary conditions to fill
    type(tem_bc_header_type) :: me
    !> A handle to the Lua configuration script, to read the data from
    type(flu_State) :: conf
    !> The boundary properties of the treelmesh, to connect the header to
    type(tem_BC_prop_type), intent(in) :: BC_prop
    !> handle for schemes table
    integer, intent(in),optional       :: parentHandle
    !---------------------------------------------------------------------------
    integer :: iBC
    integer :: bc_handle, myHandle
    integer :: iError
    !---------------------------------------------------------------------------

    call tem_write(' Loading BC header')
    ! if fluid informations in scheme table parentHandle /= 0
    call aot_table_open( L       = conf,                                       &
      &                  parent  = parentHandle,                               &
      &                  thandle = bc_handle,                                  &
      &                  key     = 'boundary_condition' )
    me%nBCs = aot_table_length(L=conf, thandle=bc_handle)
    call tem_write('    ... found nBCs ', me%nBCs)

    ! allocate the label, bc_kind and BC_ID arrays in me
    call tem_init_bc_header( me )

    do iBC = 1, me%nBCs
      call aot_table_open( L       = conf,                                     &
        &                  parent  = bc_handle,                                &
        &                  thandle = myHandle,                                 &
        &                  pos     = iBC )
      if (myHandle /= 0) then
        call aot_get_val( L       = conf,                                      &
          &               thandle = myHandle,                                  &
          &               val     = me%label(iBC),                             &
          &               ErrCode = iError,                                    &
          &               key     = 'label',                                   &
          &               default = 'unnamed' )

        call aot_get_val( L       = conf,                                      &
          &               thandle = myHandle,                                  &
          &               val     = me%BC_kind(iBC),                           &
          &               ErrCode = iError,                                    &
          &               key     = 'kind',                                    &
          &               default = 'none' )
        call tem_write('  Found BC '//trim(me%Label(iBC))//' of kind '//       &
          &            trim(me%BC_kind(iBC)))
      else
        call tem_write('WARNING: BC definition should be a table!')
        call tem_write('       : Assuming unnamed BC with kind none!')
        me%Label(iBC) = 'unnamed'
        me%BC_kind(iBC) = 'none'
      end if
      call aot_table_close(L=conf, thandle=myHandle)
    end do ! iBC

    call aot_table_close(L=conf, thandle=bc_handle)

    ! Now all the BCs defined in the config file are known,
    ! try to match them against the labels in the BC_prop.
    ! if a bc label can not be matched, treat it as a fatal error.
    call tem_match_bc_header( me = me, BC_prop = BC_prop, fatal = .true. )

    if ( me%nBCs .gt. 0 ) then
      call tem_write( ' Number of Boundaries: ', me%nBCs )
    else
      call tem_write( ' No Boundary conditions!' )
    end if

  end subroutine tem_load_bc_header
!******************************************************************************!


!******************************************************************************!
  !> This subroutine looks for a given label in the given boundary conditions
  !! table, and returns the according table handle.
  !! Note, that this should usually not be necessary, as the number of
  !! the header is given by the ordering in the bc_header_type, and you
  !! can use the desired position directly to look up a specific bc in the
  !! configuration script.
  !!
  subroutine tem_open_bc( label, bc_table, conf, thandle )
    !---------------------------------------------------------------------------
    !> The label to look for
    character(len=LabelLen), intent(in) :: label
    !> Handle to the boundary_condition table, to look in
    integer, intent(in) :: bc_table
    !> Handle of the Lua script to use
    type(flu_State) :: conf
    !> Returned handle of to the entry providing the requested label
    integer, intent(out) :: thandle
    !---------------------------------------------------------------------------
    character(len=LabelLen) :: bc_name
    integer :: myHandle
    integer :: iBC, iError
    integer :: nBCs
    !---------------------------------------------------------------------------

    nBCs = aot_table_length(L=conf, thandle=bc_table)
    ! Default thandle to nothing found.
    thandle = 0
    do iBC=1,nBCs
      call aot_table_open( L       = conf,                                     &
        &                  parent  = bc_table,                                 &
        &                  thandle = myHandle,                                 &
        &                  pos     = iBC )
      if (myHandle /= 0) then
        call aot_get_val( L       = conf,                                      &
          &               thandle = myHandle,                                  &
          &               val     = bc_name,                                   &
          &               ErrCode = iError,                                    &
          &               key     = 'label',                                   &
          &               default = 'unnamed' )
        if (adjustl(bc_name) == adjustl(label)) then
          ! Found the matching boundary table, returning it...
          thandle = myHandle
          exit
        end if
      end if
      call aot_table_close(L=conf, thandle=myHandle)
    end do

  end subroutine tem_open_bc
!******************************************************************************!


!******************************************************************************!
  !> This routine does the allocation job
  !!
  subroutine tem_init_bc_header( me )
    !---------------------------------------------------------------------------
    !> The boundary header
    type(tem_bc_header_type) :: me
    !---------------------------------------------------------------------------

    allocate(me%Label(me%nBCs))
    allocate(me%BC_kind(me%nBCs))
    allocate(me%BC_ID(me%nBCs))

    ! Set initial values for BC_ID
    me%BC_ID = -1

  end subroutine tem_init_bc_header
!******************************************************************************!


!******************************************************************************!
  !> This routine match the labels in me(tem_bc_header_type) against the labels
  !! in the bcProp(tem_BC_prop_type).\n
  !! If a bc label can not be matched, the code will STOP!
  !!
  subroutine tem_match_bc_header( me, BC_prop, fatal )
    !---------------------------------------------------------------------------
    !> The boundary header to fill
    type(tem_bc_header_type) :: me
    !> The boundary properties of the treelmesh, to connect the header to
    type(tem_BC_prop_type), intent(in) :: BC_prop
    !> whether treat it as a fatal error when a bc label not matched.
    logical, intent(in) :: fatal
    !---------------------------------------------------------------------------
    integer :: iBC, iBCtype
    character(len=LabelLen) :: proplabel
    logical :: MatchedBC
    !---------------------------------------------------------------------------

    call tem_write(' matching boundaries for nBCs', bc_prop%nBCtypes )
    ! Outside loop: loop over the bc labels in BC_prop
    do iBCtype = 1, BC_prop%nBCtypes

      proplabel = adjustl(BC_prop%BC_label(iBCtype))
      matchedBC = .false.
      ! Inside loop: loop over the bc labels in bc header
      do iBC = 1, me%nBCs
        if (me%BC_ID(iBC) == -1) then
          ! This ID has not been matched yet, check it now
          ! BC_ID is set to be -1 during initialization
          if (adjustl(me%label(iBC)) == proplabel) then
            ! Found a match, assign the configured BC, and leave the loop
            me%BC_ID(iBC) = iBCtype
            matchedBC = .true.
            exit
          end if
        end if
      end do ! iBC in me%label

      ! If this BC is not matched, give a error message and stop the code.
      if ( (.not. matchedBC) .and. fatal ) then
        call tem_write("ERROR:")
        call tem_write("No configuration found for boundary "//trim(proplabel) &
          &            //" from the BC property in the mesh!")
        call tem_write("This could also be caused by double definition "       &
          &            //"of the same BC in the mesh.")
        call tem_write("This is fatal, STOPPING execution now...")
        call tem_abort()
      end if
    end do ! iBCtype in bc_prop

    if (minval(me%BC_ID) == -1) then
      call tem_write("WARNING: There are more boundaries defined in the "      &
        &            //" config file, than in the mesh properties!")
!HK: Why should this be fatal?
!HK: If it indeed should be, this is not a WARNING, but an error in the
!HK: message above!
!HK!      call tem_write("This is fatal, STOPPING execution now...")
!HK!      call tem_abort()
    end if

  end subroutine tem_match_bc_header
!******************************************************************************!


end module tem_bc_header_module
!******************************************************************************!
