! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Definition of the boundary property for elements
!!
!! The has-boundary \ref tem_property_module "property" highlights an element 
!! that has at least one side with a boundary condition.
!! Therefore the neighbor on this side is not part of the computational domain,
!! and needs to be treated accordingly.
!! For all elements with the has-boundary property, the boundary conditions are
!! stored for all sides.
!! Yet this is done in a elementwise manner to allow arbitrary distribution of 
!! the elments to parallel processing units. The appropiate handling of the
!! boundaries is solver specific and left to them.
!!
!! Each boundary is identified by a unique number (ID), which starts from 1 and
!! ranges up to the number of existing boundaries (nBCtypes).
!! In the solvers, the labels of the boundaries attached to the boundary IDs in
!! the array of strings \verb+BC_label+ are analyzed in order to treat the
!! boundaries adequately.
!!
!! For each element with the has-boundary property, a boundary ID is assigned
!! for each of the 26 direct neighbor directions.
!! This is done in the two-dimensional array \verb+boundary_ID+,
!! where the first dimension has nSides entries and the second dimension has as
!! many entries as there elements with the has-boundary property on the local
!! partition.
!! For regular fluid neighbors, the boundary ID is set to zero in the respective
!! direction.
!! \image html boundary_data_structure.png
!!
!! To attach boundary conditions a bit is set for an element to indicate that
!! any of its sides has a boundary condition attached to it.
!! For all elements with this property and only those elements, boundary
!! conditions on every side of the element are stored.
!! This data is put in a separate file, containing the boundary conditions per
!! side for all elements with the boundary property.
!! The binary file has the same number of integer entries for all elements and
!! can therefore be easily accessed on an elemental basis.
!! In this list of elements the ordering from the complete mesh is preserved.
!! To describe the meaning of each boundary condition, indicated by an integer
!! number, there is an additional header file with labels for each type of
!! boundary condition in the mesh.
!! Only positive integers are considered as boundary entries, zeros indicate no
!! boundary in the corresponding direction, and negative numbers are reserved
!! for references on \emph{treeIDs} as neighboring element.
!! Directly referencing a certain \emph{treeID} instead of a boundary condition
!! on one of the sides of an element is used to implement periodic interfaces
!! within the universe cube.
!! Due to this layout of the boundary description, the boundaries can be read in
!! parallel with the following procedure:
!! - Count the local number of elements with boundary conditions (given by the
!!   property bits defined for all elements).
!! - Build a prefix summation across all partitions to find the offset of the
!!   local set of boundary conditions in the sparse set of global boundary
!!   conditions.
!! - Read this set of elemental boundary conditions from the file accordingly
!!   from the offset up to the number of local elements with boundary
!!   conditions.
!! 
!! With the strict ordering and uniform elemental treatment of data, the sparse
!! property information can be accessed efficiently.
!! Just a single collective prefix operation is required in the parallel system,
!! while the consumed disk storage is kept at a minimum.
!!
module tem_bc_module

  ! include treelm modules
  use env_module,               only: labelLen
  use tem_spacetime_fun_module, only: tem_spacetime_fun_type, tem_load_spacetime
  use tem_tools_module,         only: tem_write
  use tem_aux_module,           only: tem_abort

  ! include aotus modules
  use aot_table_module, only: aot_table_open, aot_table_close, aot_get_val
  use aot_path_module,  only: aot_path_type, assignment(=), aot_path_addNode
  use aotus_module,     only: aot_get_val, flu_State

  implicit none

  private

  public :: tem_bc_state_type
  public :: tem_load_bc_state

  !> boundary state type definition for boundary state variable
  !!
  !! State variables like  density, velocityX can be
  !! constant, transient, spatial, dirchlet/neumann.
  !! res = f(x,y,z)*g(t) or res = f(x,y,z,t)
  type tem_bc_state_type
    !> Name of the state, this boundary condition applies to
    !\todo KM: Not used
    character(len=LabelLen) :: state_name
    !> Style of this boundary condition
    !! dirichlet = set value itself
    !! neumann = set derivative of value
    character(len=LabelLen) :: style
    !> A flag to indicate that the state is properly defined
    logical :: isDefined
    !> spacetime function type includes transient and spatial
    type( tem_spacetime_fun_type ) :: spacetime
  end type tem_bc_state_type

contains

!******************************************************************************!
  !> function to load spatial, constant or transient boundary conditions to
  !! the boundary state type (or combinations of them).
  !! res = f(x,y,z)*g(t) or res = f(x,y,z,t)
  !! Valid definitions:
  !! Example given for state variable velocityX
  !! \verbatim
  !!  velocityX = 0.01
  !! \endverbatim
  !! \verbatim
  !! velocityX = { style = 'dirchlet',
  !!               transient = {predefined="linear", min_factor = 0.0,
  !!                      max_factor = 1.0, from_time = 0, to_time = 1000},
  !!               spatial = {predefined='parabol', shape = {object = 'line',
  !!                                   line = { origin={-2.0,0.0,0.0},direction={0.0,1.0,0.0}}}}
  !!--              ,spatial = {predefined='parabol', shape = {object = 'plane',
  !!--                             plane = { origin={-2.0,0.0,0.0},vecA={0.0,1.0,0.0},
  !!--                                         vecB={0.0,0.0,1.0} } } }
  !!             }
  !! \endverbatim
  !! [example given in subroutine load_spatial_parabola3d and load_trasient_predef_fun definition]
  !!
  subroutine tem_load_bc_state( bc, state_name, conf, bc_handle, path, nComp )
    !---------------------------------------------------------------------------
    !> The boundary to fill
    type(tem_bc_state_type), intent(inout) :: bc
    !> The state variable to set with this boundary condition
    character(len=*), intent(in) :: state_name
    type(flu_State),intent(in) :: conf !< Lua state
    !> Handle to the table describing the boundary
    integer, intent(in) :: bc_handle
    type(aot_path_type), intent(in) :: path
    !> position of spacetime fun in a table
    integer, intent(in), optional :: nComp
    !---------------------------------------------------------------------------
    integer :: iError
    integer, allocatable :: vError(:)
    integer :: state_handle
    type(aot_path_type) :: state_path
    ! local ncomp
    integer :: loc_nComp
    !---------------------------------------------------------------------------
    if( present( nComp ))then
      if( nComp .gt. 0)then
        loc_nComp = nComp
      else
        loc_nComp = 1
      end if
    else
      loc_nComp = 1
    end if

    ! Assume undefined boundary condition for this state.
    bc%isDefined = .false.

    call tem_write(' Loading bc state for '// trim(state_name))

    ! First try to read the state entry as a constant such as value = 1.
    allocate( bc%spacetime%const(1) )
    call aot_get_val( L       = conf,                                          &
      &               thandle = bc_handle,                                     &
      &               val     = bc%spacetime%const(1),                         &
      &               ErrCode = iError,                                        &
      &               key     = trim( state_name ))

    allocate(vError(1))
    vError = iError
    ! if is not a constant and defined as a table
    if ( iError .ne. 0 .and. loc_nComp .gt. 1) then
      deallocate(vError)
      deallocate( bc%spacetime%const )

      !check whether const is defined as a table
      call aot_get_val( L         = conf,                                      &
        &               thandle   = bc_handle,                                 &
        &               val       = bc%spacetime%const,                        &
        &               ErrCode   = vError,                                    &
        &               maxLength = loc_nComp,                                 &
        &               key       = trim(state_name))

    end if
    conststate: if ( any( vError.eq.0 ) )then

      ! There is just a single value set for this boundary condition, assume
      ! a constant set for the state of this variable.
      call tem_write('    TEM BC: '//trim(state_name)//' is constant')
      call tem_write('    constant: ', bc%spacetime%const)
      bc%isDefined = .true.
      bc%style = 'dirichlet'
      bc%spacetime%fun_kind = 'const'

    else conststate

      ! There is no plain constant value, so try to parse the table and find
      ! the various settings.
      call aot_table_open( L       = conf,                                     &
        &                  parent  = bc_handle,                                &
        &                  thandle = state_handle,                             &
        &                  key     = state_name )

      bc%isDefined = .true.

      if (state_handle.ne.0) then
        ! Found a table, now check for the style of this boundary condition
        ! Default to dirichlet.
        state_path = path
        call aot_path_addNode( state_path, NodeType='table', key=state_name )
        call aot_get_val( L       = conf,                                      &
          &               thandle = state_handle,                              &
          &               val     = bc%style,                                  &
          &               ErrCode = iError,                                    &
          &               key     = 'style',                                   &
          &               default = 'dirichlet' )

        call aot_table_close(L=conf, thandle=state_handle)
      end if

      ! Load spacetime function
      call tem_load_spacetime( me     = bc%spacetime,                          &
        &                      conf   = conf,                                  &
        &                      path   = path,                                  &
        &                      parent = bc_handle,                             &
        &                      key    = trim(state_name),                      &
        &                      nComp  = nComp )

      ! check if the spacetime function has been read
      if(trim(bc%spacetime%fun_kind).eq.'none')then
        call tem_write( 'The spacetime function for one boundary ' //          &
          &             'condition has not been loaded properly.' )
        call tem_abort()
      end if

    end if conststate

  end subroutine tem_load_bc_state
!******************************************************************************!


end module tem_bc_module
!******************************************************************************!
