! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> A module to provide informations on [boundary condition](@ref tem_bc_module )
!! property for elements.
!! 
module tem_bc_prop_module

  ! include treelm modules
  use mpi
  use env_module,          only: long_k, newunit, rk, tem_create_EndianSuffix
  use tem_prophead_module, only: tem_prophead_type
  use tem_property_module, only: tem_property_type, prp_hasBnd

  ! include aotus modules
  use aot_out_module,   only: aot_out_type, aot_out_val, aot_out_open,         &
    &                         aot_out_close, aot_out_open_table,               &
    &                         aot_out_close_table
  use aot_table_module, only: aot_table_open, aot_table_close
  use aotus_module,     only: open_config_file, aot_get_val, close_config,     &
    &                         flu_State

  implicit none

  private

  public :: tem_BC_prop_type
  public :: init_tem_bc_prop, dump_tem_BC_prop, load_tem_BC_prop
  public :: tem_empty_BC_prop, dump_tem_BC_propHeader
  public :: load_tem_BC_qVal

  !> The position of the bit in the bitfield to identify the property of having
  !! a boundary condition in an element.
  integer, parameter :: BC_bitID = 0

  !> Maximal allowed label length to identify the boundary conditions
  integer, parameter :: LabelLen = 80

  type tem_BC_prop_type
    !> Pointer to treelmesh_type%global%property
    type(tem_prophead_type),  pointer :: header => null()

    !> Number of sides in the elements, each side might be assigned a
    !! boundary condition ID
    !! The first 6 are dedicated to "direct" neighbors, and shall always be
    !! present.
    !! The next 12 are dedicated to neighbors where two of the indices in the
    !! 3 dimensional address have to be changed (neighbors at the edges of
    !! the cube)
    !! After these 18 entries, additional 8 entries might be used for neighbors
    !! at the corners.
    !!
    !! By default we should store all 26 neighbors, to provide an as complete
    !! neighborhood as possible to arbitrary solvers.
    integer :: nSides

    !> Number of different Boundary conditions used in the complete domain
    integer :: nBCtypes !< e.g. wall, inflow, outflow

    !> Array of labels identifying each of the boundary conditions.
    !! This array has a length of nBCtypes
    character(len=LabelLen), allocatable :: BC_label(:)

    !> Logical array indicating whether each boundary is high order wall
    !! JQ: This array has a length of nBCtypes
    logical, allocatable :: hasQVal(:)

    !> Actual q-value array for high order wall boundary conditions
    !! JQ: The fist dimension has a length of nSides
    !! The second dimension has a length of the number of elements with the
    !! boundary condition property. tem_property_type%nElems
    real(kind=rk), allocatable :: qVal(:,:)

    !> Actual boundary condition identification for all sides of elements with
    !! this property.
    !! The fist dimension has a length of nSides
    !! The second dimension has a length of the number of elements with the
    !! boundary condition property. tem_property_type%nElems
    !! An ID of 0 for a side indicates, that there is no boundary in that
    !! direction.
    integer(kind=long_k), allocatable :: boundary_ID(:,:)

    !> Pointer to treelmesh_type%property
    type(tem_property_type),  pointer :: property => null()

  end type tem_BC_prop_type

contains

!******************************************************************************!
  !> todo: Add description
  !!
  subroutine init_tem_bc_prop( header, property, bc )
    !---------------------------------------------------------------------------
    type(tem_prophead_type), pointer :: header(:)
    type(tem_property_type), pointer :: property(:)
    type(tem_bc_prop_type)           :: bc
    !---------------------------------------------------------------------------
    integer :: iprop
    !---------------------------------------------------------------------------
    
    ! run over all the headers and check for the index of prp_hasBnd
    do iprop=1, size(header)
      if (header(iprop)%bitpos == prp_hasBnd) then
        bc%header => header(iprop)
        bc%property => property(iprop)
        exit
      end if
    end do
    !
  end subroutine init_tem_bc_prop
!******************************************************************************!


!******************************************************************************!
  !> load bc property header from lua file, boundaryID from bnd.lsb
  !!
  subroutine load_tem_BC_prop( me, offset, nElems, basename, myPart, comm )
    !---------------------------------------------------------------------------
    !> Boundary condition construct to load the data into
    type(tem_BC_prop_type), intent(out) :: me
    !> Offset of the local set of elements in the global list
    integer(kind=long_k), intent(in) :: offset
    !> Local number of elements with this property
    integer, intent(in) :: nElems
    !> Name of the file, the data is stored in, will be appended with
    !! ".ascii" for the header information and ".lsb" or ".msb" for the
    !! binary data.
    character(len=*), intent(in) :: basename
    !> Partition to load
    integer, intent(in) :: myPart
    !> Communicator to use
    integer, intent(in), optional :: comm
    !---------------------------------------------------------------------------
    type( flu_State ) :: conf ! lua flu state to read lua file
    integer :: fUnit
    integer :: i, rl
    integer :: root
    integer :: locomm
    integer :: iError
    character(len=4) :: EndianSuffix
    character(len=256) :: headerfile
    character(len=256) :: datafile
    integer :: thandle
    !---------------------------------------------------------------------------

    root = 0

    if (present(comm)) then
      locomm = comm
    else
      locomm = MPI_COMM_WORLD
    end if

    headerfile = trim(basename)//'.lua'
    EndianSuffix = tem_create_EndianSuffix()
    datafile = trim(basename)//trim(EndianSuffix)

    if (myPart == root) then
      ! Read the header only on the root process, broadcast to all others
      ! open mesh header file
      call open_config_file( L = conf, filename = headerfile )
      call aot_get_val( L       = conf,                                        &
        &               key     = 'nSides',                                    &
        &               val     = me%nSides,                                   &
        &               ErrCode = iError )
      call aot_get_val( L       = conf,                                        &
        &               key     = 'nBCtypes',                                  &
        &               val     = me%nBCtypes,                                 &
        &               ErrCode = iError )
    end if

    call MPI_Bcast(me%nSides, 1, MPI_INTEGER, root, locomm, iError)
    call MPI_Bcast(me%nBCtypes, 1, MPI_INTEGER, root, locomm, iError)

    allocate(me%BC_label(me%nBCtypes))

    if (myPart == root) then
      ! Now read the list of boundary labels
      call aot_table_open( L = conf, thandle = thandle, key = 'bclabel' )
      do i=1,me%nBCtypes
        call aot_get_val( L       = conf,                                      &
          &               thandle = thandle,                                   &
          &               pos     = i,                                         &
          &               val     = me%BC_label(i),                            &
          &               ErrCode = iError )
      end do
      call aot_table_close( L = conf, thandle = thandle )
      call close_config( conf )
    end if

    call MPI_Bcast( me%BC_label, LabelLen*me%nBCtypes, MPI_CHARACTER,          &
      &             root, locomm, iError)

    allocate(me%boundary_ID(me%nSides, nElems))

    if (nElems > 0) then
      inquire(iolength=rl) me%boundary_ID(:,1)
      fUnit = newUnit()
      open(unit = fUnit, file = datafile, action = 'read',                     &
        &  access = 'direct', form = 'unformatted',                            &
        &  recl = rl, status = 'old')
      do i=1,nElems
        read(fUnit,rec=offset+i) me%boundary_ID(:,i)
      end do
      close(fUnit)
    end if


  end subroutine load_tem_BC_prop
!******************************************************************************!


!******************************************************************************!
  !> Initialize an empty boundary condition
  subroutine tem_empty_BC_prop( me )
    !---------------------------------------------------------------------------
    !> Boundary condition construct to load the data into
    type(tem_BC_prop_type), intent(out) :: me
    !---------------------------------------------------------------------------

    me%nSides = 0
    me%nBCtypes = 0
    allocate(me%BC_label(0))
    allocate(me%boundary_ID(0, 0))

  end subroutine tem_empty_BC_prop
!******************************************************************************!


!******************************************************************************!
  !> dump bc properties header information to lua file
  !!
  subroutine dump_tem_BC_propHeader( me, headerfile )
    !---------------------------------------------------------------------------
    !> Boundary condition construct to load the data into
    type(tem_BC_prop_type), intent(in) :: me
    !> name of the bc header lua file
    character(len=*), intent(in) :: headerfile
    !---------------------------------------------------------------------------
    integer :: i
    type(aot_out_type) :: conf ! aotus lua state to write output
    !---------------------------------------------------------------------------

    call aot_out_open( conf, headerfile )
    call aot_out_val( put_conf = conf, vname = 'nSides', val = me%nSides )
    call aot_out_val( put_conf = conf, vname = 'nBCtypes', val = me%nBCtypes )
    call aot_out_open_table( conf, 'bclabel' )
    do i = 1, me%nBCtypes
      call aot_out_val( conf, val = me%BC_label(i) )
    end do
    call aot_out_close_table(conf)
    ! close the mesh header file
    call aot_out_close(conf)

  end subroutine dump_tem_BC_propHeader
!******************************************************************************!


!******************************************************************************!
  !> dump bc properties
  subroutine dump_tem_BC_prop( me, offset, nElems, basename, myPart, comm )
    !---------------------------------------------------------------------------
    !> Boundary condition construct to load the data into
    type(tem_BC_prop_type), intent(in) :: me
    !> Offset of the local set of elements in the global list
    integer(kind=long_k), intent(in) :: offset
    !> Local number of elements with this property
    integer, intent(in) :: nElems
    !> Name of the file, the data is stored in, will be appended with
    !! ".ascii" for the header information and ".lsb" or ".msb" for the
    !! binary data.
    character(len=*), intent(in) :: basename
    !> Partition to dump
    integer, intent(in) :: myPart
    !> Communicator to use
    integer, intent(in), optional :: comm
    !---------------------------------------------------------------------------
    integer :: i
    integer :: fUnit
    integer :: rl
    integer :: root
    integer :: locomm
    character(len=256) :: headerfile
    character(len=256) :: datafile
    character(len=4) :: EndianSuffix
    !---------------------------------------------------------------------------

    root = 0

    if (present(comm)) then
      locomm = comm
    else
      locomm = MPI_COMM_WORLD
    end if

    headerfile = trim(basename)//'.lua'
    EndianSuffix = tem_create_EndianSuffix()
    datafile = trim(basename)//trim(EndianSuffix)

    if (myPart == root) then
      ! Only root partition needs to write the header
      !open up the mesh header lua file to dump the stuff using aotus library
      call dump_tem_BC_propHeader( me, headerfile )
    end if

    if (nElems > 0) then
      inquire(iolength=rl) me%boundary_ID(:,1)
      fUnit = newUnit()
      open(unit = fUnit, file = datafile, action = 'write',                    &
        &  access = 'direct', form = 'unformatted',                            &
        &  recl = rl, status = 'replace')
      do i = 1, nElems
       write(fUnit, rec=offset+i) me%boundary_ID(:,i)
       if (maxval(me%boundary_ID(:,i)).gt. me%nBCtypes)                        &
         &                  write(*,*) "BcID",me%boundary_ID(:,i)
      end do
      close(fUnit)
    end if

  end subroutine dump_tem_BC_prop
!******************************************************************************!


!******************************************************************************!
  !> load bc qVal header from lua file, qVal from bnddist.lsb
  !! JQ: nElems is the number of local elements with qVal
  !!
  subroutine load_tem_BC_qVal( me, offset, nElems, basename, myPart, comm )
    !---------------------------------------------------------------------------
    !> Boundary condition construct to load the data into
    type(tem_BC_prop_type), intent(inout) :: me
    !> Offset of the local set of elements in the global list
    integer(kind=long_k), intent(in) :: offset
    !> Local number of elements that have qVal
    integer, intent(in) :: nElems
    !> Name of the file, the data is stored in, will be appended with
    !! ".ascii" for the header information and ".lsb" or ".msb" for the
    !! binary data.
    character(len=*), intent(in) :: basename
    !> Partition to load
    integer, intent(in) :: myPart
    !> Communicator to use
    integer, intent(in), optional :: comm
    !---------------------------------------------------------------------------
    type( flu_State ) :: conf ! lua flu state to read lua file
    integer :: fUnit
    integer :: i, rl
    integer :: root
    integer :: locomm
    integer :: iError
    character(len=4) :: EndianSuffix
    character(len=256) :: headerfile
    character(len=256) :: datafile
    integer :: thandle
    !---------------------------------------------------------------------------

    root = 0

    if (present(comm)) then
      locomm = comm
    else
      locomm = MPI_COMM_WORLD
    end if

    ! set header file and binary file name
    headerfile = trim(basename)//'.lua'
    EndianSuffix = tem_create_EndianSuffix()
    datafile = trim(basename)//trim(EndianSuffix)

    allocate(me%hasQVal(me%nBCtypes))

    if (myPart == root) then
      ! Now read whether each boundary has q value
      call open_config_file(L = conf, filename = headerfile)
      call aot_table_open(L = conf, thandle = thandle, key = 'hasQVal')
      do i=1,me%nBCtypes
        call aot_get_val( L       = conf,                                      &
          &               thandle = thandle,                                   &
          &               pos     = i,                                         &
          &               val     = me%hasQVal(i),                             &
          &               ErrCode = iError )
      end do
      call aot_table_close( L = conf, thandle = thandle )
      call close_config( conf )
    end if

    call MPI_Bcast(me%hasQVal, me%nBCtypes, MPI_LOGICAL, root, locomm, iError)

    allocate(me%qVal(me%nSides, nElems))

    if (nElems > 0) then
      inquire(iolength=rl) me%qVal(:,1)
      fUnit = newUnit()
      open(unit = fUnit, file = datafile, action = 'read',                     &
        &  access = 'direct', form = 'unformatted',                            &
        &  recl = rl, status = 'old')
      do i=1,nElems
        read(fUnit,rec=offset+i) me%qVal(:,i)
      end do
      close(fUnit)
    end if

  end subroutine load_tem_BC_qVal
!******************************************************************************!


end module tem_bc_prop_module
!******************************************************************************!
