! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module provides the user with a simple geometrical object like point,
!! line, plane, box and box.
!!
!! \author Simon Zimny, Kartik Jain
!!
!! Detail description of shapes are given at \ref shapes
!!
module tem_canonicalND_module

  ! include treelm modules
  use env_module,            only: rk, long_k, globalMaxLevels, labelLen
  use tem_aux_module,        only: tem_abort
  use treelmesh_module,      only: treelmesh_type
  use tem_geometry_module,   only: tem_CoordOfReal, tem_IdOfCoord, tem_PosOfId
  use tem_tools_module,      only: tem_write
  use tem_dyn_array_module,  only: dyn_intArray_type, append
  use tem_topology_module,   only: tem_levelOf
  use tem_logging_module,    only: tem_log, tem_toStr, tem_logging_isActive, logUnit

  ! include aotus modules
  use aotus_module,     only: aot_get_val, aoterr_Fatal, aoterr_WrongType,     &
    &                         flu_State
  use aot_table_module, only: aot_table_open, aot_table_close, aot_table_length
  use aot_out_module,   only: aot_out_type, aot_out_val,                       &
    &                         aot_out_open_table, aot_out_close_table

  implicit none
  private

  public :: tem_canonicalND_type
  public :: tem_load_canonicalND
  public :: tem_getNextCoordOfcanonicalND
  public :: tem_canonicalND_out
  public :: tem_cano_initSubTree

  !> Definition of the canonicalND
  type tem_canonicalND_type
    real(kind=rk) :: origin(3)   !< origin of the canonicalND
    !< vector along the edge A (also defines size)
    !!1st dimension defines x,yz coord
    !! 2nd dimension vec number
    real(kind=rk) :: vec(3,3)    
    !> how many discrete points the canonicalND is divided into
    integer ::       segments(3)    
    !> spatial distribution of the points
    character(len=labellen) :: distribution 
    !> kind of canonicalND (line, plane, point, box)
    character( len=labellen ) :: kind
    logical :: active(3)     !< identify which vectors are active (not equal 0)
    !> dimension of canonical object 
    !! nDim=0 - point
    !! nDim=1 - line
    !! nDim=2 - plane
    !! nDim=3 - box
    integer :: nDim
    !> To choose what to do with intersection of box
    !! if only_surface = true than the only the surface of the object
    !! is intersected
    !! if only_surface = false then the whole object is intersected
    !! default is set to false
    logical :: only_surface
  end type tem_canonicalND_type

  !> interface to write out canonical shape(s) in lua format to a file 
  interface tem_canonicalND_out
    module procedure tem_canonicalND_out_scal
    module procedure tem_canonicalND_out_vec
  end interface tem_canonicalND_out

  !> interface to load canonical objects
  interface tem_load_canonicalND
    module procedure tem_load_onecanonicalND
    module procedure tem_load_canonicalND_vec
  end interface tem_load_canonicalND

  contains

!******************************************************************************!
  !> Loading canonicalNDs from the config file valid definitions:
  !!
  subroutine tem_load_canonicalND_vec( me, conf, thandle )
    !---------------------------------------------------------------------------
    !> the array of canonical objects which to read in (and first allocate)
    type(tem_canonicalND_type), allocatable, intent(out) :: me(:)
    !> lua config handle
    type(flu_state) :: conf 
    !> table handle from which to read
    integer, intent(in) :: thandle 
    !---------------------------------------------------------------------------
    ! lua handles
    integer :: canonicalNDs_thandle, single_thandle
    integer :: canonicalND_entries
    integer :: iCan
    integer :: ncanonicalNDs
    !---------------------------------------------------------------------------

    write(logUnit(3),*) 'Trying to read canoND shape...'


    call aot_table_open( L       = conf,                                       &
      &                  parent  = thandle,                                    &
      &                  thandle = canonicalNDs_thandle,                       &
      &                  key     = 'object' )
    canonicalND_entries = aot_table_length( L       = conf,                    &
      &                                     thandle = canonicalNDs_thandle )

    ! Now check, if the first entry is a table itself:
    call aot_table_open( L       = conf,                                       &
      &                  parent  = canonicalNDs_thandle,                       &
      &                  thandle = single_thandle,                             &
      &                  pos     = 1 )

    if (single_thandle == 0) then
      ! The entries are not tables themselves, try to interpret it as
      ! single canonical object.
      call aot_table_close( L = conf, thandle = single_thandle )
      ncanonicalNDs = 1
      allocate(me( ncanonicalNDs ))
      call tem_load_oneCanonicalND( me      = me(1),                           &
        &                           conf    = conf,                            &
        &                           thandle = canonicalNDs_thandle )
    else
      ! First entry is a table, assume all others are also tables,
      ! and properly define canonicalNDs coordinates, go on reading them.
      call aot_table_close( L = conf, thandle = single_thandle )
      ncanonicalNDs = canonicalND_entries
      allocate( me( ncanonicalNDs ))
      do iCan=1, ncanonicalNDs
        write(logUnit(3),*) 'Read Object: '//trim( tem_toStr( iCan ))//      & 
          &                   ' with ' 
        call aot_table_open( L       = conf,                                   &
          &                  parent  = canonicalNDs_thandle,                   &
          &                  thandle = single_thandle,                         &
          &                  pos     = iCan )
        call tem_load_oneCanonicalND( me      = me( iCan ),                    &
          &                           conf    = conf,                          &
          &                           thandle = single_thandle )
        call aot_table_close( L = conf, thandle = single_thandle )
      end do
    end if
    
    call aot_table_close( L = conf, thandle = canonicalNDs_thandle )


  end subroutine tem_load_canonicalND_vec
!******************************************************************************!


!******************************************************************************!
  !> Read one canonical object definition into a tem_canonicalND_type from a lua
  !! table.
  !!
  subroutine tem_load_oneCanonicalND( me, conf, thandle )
    !---------------------------------------------------------------------------
    !> contains canonicalND data
    type(tem_canonicalND_type), intent(out) :: me 
    !> lua state
    type(flu_state) :: conf 
    !> lua table identification
    integer, intent(in) :: thandle 
    !> output messages?
    !---------------------------------------------------------------------------
    ! error identifiers
    integer :: iError
    integer :: vError(3), errfatal(3)
    ! counter and total number of vectors
    integer :: iVec, nVecs   
    ! lua handles
    integer :: vec_thandle, halfvec_thandle, single_thandle, halfsingle_thandle
    real(kind=rk), allocatable :: local_vecs(:,:)
    real(kind=rk) :: local_length
    real(kind=rk) :: vecAbsSum
    character(len=labelLen) :: buffer
    !---------------------------------------------------------------------------

    errfatal = aotErr_Fatal

    ! initialize the array of vectors and the array of segments
    me%vec = 0._rk

    ! open the vec and halvec table and get its size
    call aot_table_open( L       = conf,                                       &
      &                  parent  = thandle,                                    &
      &                  thandle = vec_thandle,                                &
      &                  key     = 'vec' )
    call aot_table_open( L       = conf,                                       &
      &                  parent  = thandle,                                    &
      &                  thandle = halfvec_thandle,                            &
      &                  key     = 'halfvec' )
    ! check if the vec table is given
    if( vec_thandle .ne. 0 ) then
      nVecs = aot_table_length( L = conf, thandle = vec_thandle )

      ! Now check, if the first entry is a table itself:
      call aot_table_open( L       = conf,                                     &
        &                  parent  = vec_thandle,                              &
        &                  thandle = single_thandle,                           &
        &                  pos     = 1 )

      if (single_thandle == 0) then
        ! The entries are not tables themselves, try to interpret it as
        ! single vector.
        call aot_table_close( L = conf, thandle = single_thandle )
        nVecs = 1
        allocate(local_vecs( 3, nVecs ))
        local_vecs = 0._rk
        call aot_get_val( L       = conf,                                      &
          &               thandle = thandle,                                   &
          &               val     = local_vecs(:,1),                           &
          &               key     = 'vec',                                     &
          &               ErrCode = vError )
        if( any( btest( vError, errFatal ))) then
          call tem_write(' FATAL ERROR: in loading vec')
          call tem_write(' You should provide 3 numbers for this vector.')
          call tem_abort()
        end if
      else
        ! First entry is a table, assume all others are also tables,
        ! and properly define canonicalNDs coordinates, go on reading them.
        call aot_table_close( L = conf, thandle = single_thandle )
        allocate( local_vecs( 3, nVecs ))
        local_vecs = 0._rk
        do iVec=1, nVecs
          call aot_get_val( L       = conf,                                    &
            &               thandle = vec_thandle,                             &
            &               val     = local_vecs(:,iVec),                      &
            &               pos     = iVec,                                    &
            &               ErrCode = vError )
        end do
      end if
      if( any( btest( vError, errFatal ))) then
        call tem_write( ' FATAL ERROR: in loading vec' )
        call tem_abort()
      endif
    ! if vec table is not set try to read halfvec
    else if( halfvec_thandle .ne. 0 ) then
      call aot_table_open( L       = conf,                                     &
        &                  parent  = halfvec_thandle,                          &
        &                  thandle = halfsingle_thandle,                       &
        &                  pos     = 1 )
      nVecs = aot_table_length( L = conf, thandle = halfvec_thandle )
      if( halfsingle_thandle == 0 ) then
        ! The entries are not tables themselves, try to interpret it as
        ! single vector.
        call aot_table_close( L = conf, thandle = halfsingle_thandle )
        nVecs = 1
        allocate( local_vecs( 3, nVecs ))
        local_vecs = 0._rk
        call aot_get_val( L       = conf,                                      &
          &               thandle = thandle,                                   &
          &               val     = local_vecs(:,1),                           &
          &               key     = 'halfvec',                                 &
          &               ErrCode = vError )
        local_vecs = local_vecs * 2._rk
      else
        ! First entry is a table, assume all others are also tables,
        ! and properly define canonicalNDs coordinates, go on reading them.
        call aot_table_close( L = conf, thandle = halfsingle_thandle )
        allocate( local_vecs( 3, nVecs ))
        local_vecs = 0._rk
        do iVec=1, nVecs
          call aot_get_val( L       = conf,                                    &
            &               thandle = halfvec_thandle,                         &
            &               val     = local_vecs(:,iVec),                      &
            &               pos     = iVec,                                    &
            &               ErrCode = vError )
        end do
        local_vecs = local_vecs * 2._rk
      end if
      if( any( btest( vError, errFatal ))) then
        call tem_write( ' FATAL ERROR: in loading halfvec' )
        call tem_abort()
      endif
    ! if vec and halfvec table is not set, try to read length (for a cube)
    else
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               val     = local_length,                                &
        &               key     = 'length',                                    &
        &               default = 0.0_rk,                                      &
        &               ErrCode = iError )
      ! set the axis parallel entries of local_vecs to local_length
      nVecs = 3
      allocate( local_vecs( 3, nVecs ))
      local_vecs = 0._rk
      local_vecs(1,1) = local_length
      local_vecs(2,2) = local_length
      local_vecs(3,3) = local_length
    end if

    call aot_table_close( L = conf, thandle = vec_thandle )
    call aot_table_close( L = conf, thandle = halfvec_thandle )


    ! copy back the information from the local vector array
    me%vec(:,:nVecs) = local_vecs
    me%vec(:,nVecs+1:) = 0.0_rk

    ! read origin of the canonicalND
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               val     = me%origin,                                     &
      &               key     = 'origin',                                      &
      &               ErrCode = vError )
    if( any( btest( vError, errFatal ))) then
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               val     = me%origin,                                   &
        &               key     = 'center',                                    &
        &               ErrCode = vError )
      if (any(btest( vError, errFatal )) ) then
        call tem_write( 'ERROR reading the canonical objects, origin/center'// &
          &             ' is not well defined. STOPPING' )
        call tem_abort()
      endif
      me%origin = me%origin - me%vec(:,1)/2.0_rk - me%vec(:,2)/2.0_rk -        &
        &         me%vec(:,3)/2.0_rk
    end if

    me%segments = 1
    ! try loading segments as integer if fails try
    ! to loading segments as vector integer
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               val     = me%segments(1),                                &
      &               key     = 'segments',                                    &
      &               ErrCode = iError )
    if( btest( iError, aoterr_Fatal )) then
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               val     = me%segments,                                 &
        &               key     = 'segments',                                  &
        &               default = [ 1, 1, 1 ],                                 &
        &               ErrCode = vError )
      if( any( btest( vError, errFatal ))) then
        call tem_write( ' FATAL ERROR: in loading segments' )
        call tem_abort()
      endif
    endif

    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               val     = me%distribution,                               &
      &               ErrCode = iError,                                        &
      &               key     = 'distribution',                                &
      &               default = 'equal' )

    me%active = .false.
    me%nDim = 0

    ! check the kind and the active vectors
    do iVec = 1, 3
      vecAbsSum = me%vec(1,iVec) + me%vec(2,iVec) + me%vec(3,iVec)
      if (vecAbsSum .ne. 0._rk) then
        me%active(iVec) = .true.
        me%nDim = me%nDim + 1
      end if
    end do

    if(me%nDim .eq. 0) me%kind = 'point'
    if(me%nDim .eq. 1) me%kind = 'line'
    if(me%nDim .eq. 2) me%kind = 'plane'
    if(me%nDim .eq. 3) me%kind = 'box'

    !if canonical object is box then check for only_surface
    if(me%kind=='box') then
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               val     = me%only_surface,                             &
        &               ErrCode = iError,                                      &
        &               key     = 'only_surface',                              &
        &               pos     = 4,                                           &
        &               default = .false. )
   
      if (btest(iError, aoterr_WrongType)) then
        call tem_write( 'Error occured, while retrieving canonical box '//     &
          &             'only_surface' )
        call tem_write( 'Variable has wrong type!' )
        call tem_write( 'Should be a LOGICAL!' )
        call tem_abort()
      endif  
    endif

    ! \todo: DEB: Update this when converting arrays to strings is available
    write(logUnit(6),*) ' kind: '//trim( me%kind )
    if( tem_logging_isActive( level = 6 ))then
      call tem_write(' origin:', me%origin )
      do iVec=1,me%nDim
        write(buffer,'(a4,i1)') 'vec ', iVec
        call tem_write( trim(buffer), me%vec(:,iVec) )
      enddo
      do iVec=1,me%nDim
        write(buffer,'(a10,i1)') 'segments ', iVec
        call tem_write( trim(buffer), me%segments(iVec) )
      enddo
    end if
    write(logUnit(6),*) ' distribution '//trim( me%distribution )

  end subroutine tem_load_oneCanonicalND
!******************************************************************************!


!******************************************************************************!
  !> Return the next coordinate of the canonical shape.
  !!
  function tem_getNextCoordOfcanonicalND( me, iElem ) result( coord )
    !---------------------------------------------------------------------------
    !> current element within total amount of elements to process
    integer, intent(in) :: iElem
    !> current canonical type
    type( tem_canonicalND_type ) :: me 
    !> calulated real-world coordinate, which is returned
    real(kind=rk)       :: coord(3) 
    !---------------------------------------------------------------------------

      ! Start from the origin, step through the segments on the two
      ! defining vectors successively

      coord = me%origin                                                        &
            ! offset in direction of vector A                                  
        &   + real(mod(( iElem-1 ), me%segments( 1 )), kind=rk)                &
        &   / real( max(me%segments( 1 )-1,1), kind=rk)                        &
        &   * me%vec(:,1)                                                      &
            ! offset in direction of vector B
            + real(mod( (iElem-1)/me%segments( 1 ), me%segments( 2 )),         &
        &          kind=rk)                                                    &
        &   / real( max(me%segments( 2 )-1,1), kind=rk)                        &
        &   * me%vec(:,2)                                                      &
            ! offset in direction of vector C
            + real(mod( (iElem-1)/(me%segments( 1 )*me%segments( 2 )),         &
        &               me%segments( 3 )), kind=rk)                            &
        &   / real( max(me%segments( 3 )-1,1), kind=rk)                        &
        &   * me%vec(:,3)

  end function tem_getNextCoordOfcanonicalND
!******************************************************************************!


!******************************************************************************!
  !> Write out an array of canonical shapes in lua format
  !!
  subroutine tem_canonicalND_out_vec( me, conf )
    !---------------------------------------------------------------------------
    !> canonicalND types to write out
    type( tem_canonicalND_type ), intent(in) :: me(:)
    !> Aotus type handling the output to the file in lua format
    type(aot_out_type), intent(inout) :: conf
    !---------------------------------------------------------------------------
    ! counter
    integer :: i
    !---------------------------------------------------------------------------

    ! create a table with name canonicalND
    call aot_out_open_table( put_conf = conf, tname = 'canonicalND' )

    do i = 1, size(me)
      call tem_canonicalND_out_scal( me(i), conf )
    end do

    call aot_out_close_table( put_conf = conf )

  end subroutine tem_canonicalND_out_vec
!******************************************************************************!


!******************************************************************************!
  !> Write out a canonicalND shape in lua format
  !!
  subroutine tem_canonicalND_out_scal( me, conf )
    !---------------------------------------------------------------------------
    !> canonicalND types to write out
    type( tem_canonicalND_type ), intent(in) :: me
    !> Aotus type handling the output to the file in lua format
    type(aot_out_type), intent(inout) :: conf
    !---------------------------------------------------------------------------
    integer :: iVec
    !---------------------------------------------------------------------------

    ! create a table with name canonicalND if not exist
    if( conf%level .eq. 0 ) then
      call aot_out_open_table( put_conf = conf, tname = 'canonicalND' )
    else
      call aot_out_open_table( put_conf = conf )
    end if
    !write origin
    call aot_out_val( put_conf = conf, vname = 'origin', val = me%origin )

    !write vec
    if(me%nDim >0) then
      call aot_out_open_table( put_conf = conf, tname = 'vec' )
      do iVec=1,me%nDim
        call aot_out_val( put_conf = conf, val = me%vec(:,iVec) )
      enddo
      call aot_out_close_table( put_conf = conf )
     
      !write segments
      call aot_out_open_table( put_conf = conf, tname = 'segments' )

      do iVec=1,me%nDim
        call aot_out_val( put_conf = conf, val = me%segments(iVec) )
      enddo
      call aot_out_close_table( put_conf = conf )
    endif

    !write distribution
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'distribution',                               &
      &               val      = me%distribution )

    call aot_out_close_table( put_conf = conf )

  end subroutine tem_canonicalND_out_scal
!******************************************************************************!


!******************************************************************************!
  subroutine tem_cano_initSubTree( me, inTree, countElems, map2global )
    !---------------------------------------------------------------------------
    !> canonicalND objects on which to work
    type(tem_canonicalND_type ),intent(inout) :: me(:)
    !> Global tree
    type(treelmesh_type), intent(in) :: inTree
    !> How many elements there will be for each level in the track
    integer, intent( out ) :: countElems( globalMaxLevels )
    !> growing array for the map2global
    type(dyn_intArray_type), intent(inout) :: map2global
    !---------------------------------------------------------------------------
    integer :: iCano, tLevel, nElems, elemPos, iElem, dPos
    real(kind=rk) :: coord(3)
    integer(kind=long_k) :: treeID
    logical :: wasAdded
    !---------------------------------------------------------------------------

    call tem_write( 'Initializing canonical shapes')

    !Initialize countElems
    countElems = 0

    ! loop over all canoND objects
    do iCano = 1, size(me)
      ! \todo: SZ: set an upper limit in the load routine of the shapes,
      !            limit(*) = maximum number of elements in the directions
      !                       depending on minimum dx in the whole mesh
      !                       and edge size(*), *=x,y,z
      ! total number of elements
      nElems = me( iCano )%segments(1)                                         &
        &    * me( iCano )%segments(2)                                         &
        &    * me( iCano )%segments(3)

      ! Count how many elements are really local
      do iElem = 1, nElems
        coord(1:3) = tem_getNextCoordOfcanonicalND( me    = me( iCano ),       &
          &                                         iElem = iElem )
        ! Get the treeID on the highest level
        treeID = tem_IdOfCoord( tem_CoordOfReal( inTree, coord( 1:3 )))
        ! get position of the treeID in inTree
        elemPos = tem_PosOfId( treeID, inTree%treeID )
        if( elemPos .gt. 0 ) then
          ! append the position in inTree to the map (note that already existing
          ! ones are omitted)
          call append( me       = map2global,                                  &
            &          pos      = dpos,                                        &
            &          val      = elemPos,                                     &
            &          wasAdded = wasAdded )
          ! Count up if it was added
          if( wasAdded ) then
            tLevel   = tem_levelOf( treeID )
            countElems( tLevel ) = countElems( tLevel ) + 1
          endif ! wasAdded
        endif ! elemPos > 0
      end do ! iElem
    end do ! iCano
  end subroutine tem_cano_initSubTree
!******************************************************************************!


end module tem_canonicalND_module
!******************************************************************************!
!! \section shapes Shapes
!! The shape is defined as a canonical object. The different shapes are 
!! basically derivatives of a simple cube. 
!! A cube with no length in any direction is a point, a cube with length along 
!! only one axis is a line whereas a cube with length along two axes is simply 
!! treated as plane by the underlying code. 
!! 
!! The detailed examples for defining all shapes are discussed below.
!! 
!! \section point Point
!! 
!! A point is defined as the origin of the cube with no length vector. An 
!! example is shown below:
!! \verbatim
!! shape = {kind = 'canoND', object = {origin = {0.7, -1.9, 2.2}}}
!! \endverbatim
!! An illustrative diagram is shown in the image below:
!! \image html point.png
!! 
!! \section line Line
!! 
!! A line can be defined in two possible ways, 
!! by origin and vector or by definition the center and a vector.
!! 
!! \subsection line_def_dir 1. Line Definition by direction vector
!! 
!! We specify the origin coordinates (starting point) and direction vector of
!! the tracking line in the following way:
!! \verbatim
!! shape = {
!!   kind = 'canoND', 
!!   object = { 
!!     origin = {x,y,z}, 
!!     vec = {x1,y1,z1}, 
!!     segments = {n} 
!!     } }
!! \endverbatim
!!  The length of the line is calculated by taking distance of direction vector 
!!  from origin. A schematic diagram is shown below to provide intuition
!! 
!! \verbatim
!! origin---------> dir
!! \endverbatim
!! 
!! There is also a possibility of defining the number of segments the 
!! contiguous line is segmented into. This is of importance in the tracking
!! module. Spatial distribution of these segments on the line can also be
!! defined which currently can be only equal. If the distribution is not defined
!! then the default (equal) distribution is taken. The diagram below shows a 
!! line divided into five segments.
!!
!! \verbatim
!! |----|----|----|----|----|
!! \endverbatim
!!
!! Typically, a line will be defined as:
!! \verbatim
!! shape = {
!!   kind = 'canoND', 
!!   object = {
!!     -- the origin of the line
!!     origin = {0.,0.,0.},
!!     -- the length and direction vector of the line
!!     vec = {1.,0.,0.},
!!     -- with how many points does the contiguous line have to be 
!!     -- segmented (default = 10)
!!     segments = {10},
!!     -- how are the discrete points distributed 
!!     -- (default is equal if not defined) }
!!     distribution = 'equal' 
!!   }  
!! \endverbatim
!! 
!! The figure below shows a graphical representation of the above definition of
!! a line:
!!
!! \image html line1.png
!!  
!! \subsection line_def_center 2. Line Definition by center and direction vector
!! 
!! In this form of definition we can define the center coordinate of line and 
!! direction vector. The origin is thus calculated by reflecting the direction 
!! vector along center.
!! 
!! \verbatim
!! calculated origin<========= center --------->direction
!! \endverbatim
!! Referring to the diagram above, lets assume that center was defined as 
!! center = {0.,0.,0.} and direction = {1.,0.,0.}, then the origin will be 
!! calculated as origin = {-1.,0.,0.}.
!! Thus, general definition of a line by center in lua format will be:
!! 
!! \verbatim
!! shape = {
!!   kind = 'canoND', 
!!   object = { 
!!     center = {x,y,z}, 
!!     vec = {x1,y1,z1}, 
!!     segments = {n} 
!!     } } 
!! \endverbatim
!! 
!! In the diagram below, this definition of a line is illustrated:
!! 
!! \image html line2.png
!! 
!! Further, there can be a requirement to define a single or multiple lines. 
!! This can simply be done by placing origin and vector of other line within 
!! the object list. A generic form is shown below:
!! 
!! \verbatim
!! shape = {
!!   kind = 'canoND', 
!!   object = { 
!!   { origin = {x1, y1, z1}, 
!!     vec = {x', y', z'}, 
!!     segments = {n1} }, 
!!   { origin = {x2,y2,z2}, 
!!     vec = {x'',y'',z''} , 
!!     segments = {n2} } 
!!   } 
!! } 
!! \endverbatim
!! 
!! \section plane Plane
!! 
!! The various possible ways to define a plane are discussed below:
!! 
!! \subsection plane_def_one Plane Definition 1
!! 
!! The plane is defined by two vectors and position in x,y,z coordinate with 
!! reference to origin.
!! - origin - position to trasform plane from its origin
!! - vecA - vector A from its origin ( defines plane length )
!! - vecB - vector B from its origin ( defines plane width )
!! \verbatim
!! 
!! origin-----------> vecA
!! | -        |
!! |   -A+B   |
!! |     -    |
!! |       -  |
!! vecB v----------x
!! \endverbatim
!! 
!! For example plane at xy-plane normal to z-axis, offset from z-plane 
!! at 1.0, starting at x = -1.0 and y = -1.0 with length and width as 2.0 
!! will be defined as 
!! \verbatim
!! shape = {
!!   kind = 'canoND', 
!!   object = {
!!     origin={-1.0,-1.0,1.0}, 
!!     vec= { {2.0,0.0,0.0}, {0.0,2.0,0.0},
!!     segments = { {5}, {5}
!!     } 
!! },
!! bclabel='top',
!! refinementlevel = 5}
!! }
!! \endverbatim
!! 
!! <em>In this example and the ones below, it should be noted that bclabel 
!! and refinement level are important to be defined for seeder whereas for the 
!! solver, they are not required to be defined.</em>
!! 
!! Like a line, there is a possibility of defining the number of segments in 
!! 2 directions for a plane. The diagram below shows a plane divided into 5 
!! segments in direction A and 4 segments in direction B thus resulting in 4 
!! blocks inside.
!! \verbatim
!! |----|----|----|----|----|
!! |----|----|----|----|----|
!! |----|----|----|----|----|
!! |----|----|----|----|----|
!! \endverbatim
!! Similarly, there is a possibility of defining distribution of these segments 
!! which at the current moment is only equal. A typical Lua configurations 
!! script to define a plane will look like below:
!! \verbatim
!! shape = {
!!   kind = 'canoND', 
!!   object = { 
!!     origin={-1.0,-1.0,1.0},
!!     vec = { {2.0,0.0,0.0},{0.0,2.0,0.0} },
!!      segments = { {5}, {10} },
!!      distribution = 'equal'
!!   }                       
!! }
!! \endverbatim
!! 
!! The graphical image below shows a plane defined using the above form of 
!! definition. The plane is shown inside the bounding box with labels.
!! 
!! \image html plane_bb1.png
!! 
!! \subsection plane_def_two Plane Definition 2
!! 
!! Plane is defined by a combination of 2 triangles.
!! origin - common point for two triangles in a plane.\n
!! pointA - which defines vecA or 2nd point of 1st triangle.\n
!! pointB - which defined vecB or 2nd point of 2nd triangle.\n
!! \verbatim
!! origin---- pointA
!! | \        |
!! |   -A+B   |
!! |     \    |
!! |       \  |
!! pointB ----x
!! \endverbatim
!! 
!! The same example above for the definition 1 is now defined using points 
!! instead of vectors:
!! \verbatim
!! plane={
!!   { pointA={1.0,-1.0,1.0},
!!     pointB={-1.0,1.0,1.0},
!!     origin={-1.0,-1.0,1.0},
!!     bclabel='top',
!!     refinementlevel = 5}
!! }
!! \endverbatim
!! Taking a close look at this example reveals that whereas in first definition
!! coordinates are calculated according to the length desired by vectors, in 
!! this case coordinates are provided and direction is automatically calculated
!! by just joining these points.
!! 
!! \subsection plane_def_three Plane Definition 3
!! 
!! In this definition, the center of plane is defined as Cartesian 
!! coordinates and vectors A and B are defined for 2 combining triangles.
!! center - center of the plane.\n
!! vecA/pointA - which defines vecA or 2nd point of 1st triangle.\n
!! vecB/pointB - which defined vecB or 2nd point of 2nd triangle.\n
!! \verbatim   
!! _______vecB___ 
!! |       |      |
!! |       |      |
!! |       |______|vecA
!! |     center   |
!! |              |
!! |______________|_
!! \endverbatim 
!! 
!! The Lua format for defining such a plane is:
!! \verbatim
!! shape = {
!!   kind = 'canoND', 
!!   object = {
!!     center = {0.,0.,0.}, 
!!     vec = { {1.,0.,0.}, {0., 2., 0.}, 
!!     segments = { {5}, {5} } }
!!   bclabel='top',
!!   refinementlevel = 5 }
!! }
!! \endverbatim
!! 
!! In the image below, the above form of plane definition is graphically 
!! explained:
!! 
!! \image html plane_bb2.png
!!
!! \section box_def Box Definition
!!
!! The various possible ways to define a box are discussed below:
!! 
!! \subsection box_def_one Box Definition 1
!! 
!! The box is defined by three vectors and position in x,y,z coordinate with
!! reference to origin.
!! - origin - position to transform box from its origin
!! - vecA - vector A from its origin ( defines box length )
!! - vecB - vector B from its origin ( defines box width )
!! - vecC - vector C from its origin ( defines box height )
!! 
!! As an example a box starting at x = -1.0, y = -1.0 and z = -2.0
!! with length, width and height as 2.0 will be defined as
!! \verbatim
!! shape = {
!!   kind = 'canoND', 
!!   object = {
!!     origin={-1.0,-1.0,1.0}, 
!!     vec= { {2.0,0.0,0.0}, {0.0,2.0,0.0}, {0.0, 0.0, 2.0} }
!!     segments = { {5}, {5}, {5} }
!!     } 
!! },
!! bclabel='top',
!! refinementlevel = 5}
!! }
!! \endverbatim
!!
!! Like a plane, there is a possibility of defining the number of segments in 
!! 3 directions for a box. Similarly, there is a possibility of defining 
!! distribution of these segments which at the current moment is only equal.
!!
!! \subsection box_def_two Box Definition 2
!!
!! Box can also be defined by specifying the center and origin is then
!! calculated according to the length of half vectors
!!
!! The Lua format for defining such a box is:
!! \verbatim
!! shape = {
!!   kind = 'canoND', 
!!   object = {
!!     center={-1.0,-1.0,1.0}, 
!!     halfvec= { {1.0,0.0,0.0}, {0.0,1.0,0.0}, {0.0, 0.0, 1.0} }
!!     segments = { {5}, {5}, {5} }
!!     } 
!! },
!! bclabel='top',
!! refinementlevel = 5}
!! }
!! \endverbatim
!!
!! The above definition defines a box same as definition 1 but instead of full 
!! vectors, length of half vectors is given which is used in calculating the
!! origin.
