! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module creates the required treeID lists, from which then the state
!! vector and the neighbor information can be constructed.
!!
!! Also, the structures for the ghost and halo elements are created.
!! For the ghost cells, this means establishing the dependencies. The source
!! element positions for each ghost is stored. For the halo elements, the
!! information from where to get the element information is stored in the mpi
!! buffers.
!!
!! \author{ Harald Klimach, Manuel Hasert, Jens Zudrop }
!!
module tem_construction_module

  ! include treelm modules
  use mpi
  use env_module,            only: long_k, rk, globalMaxLevels,                &
    &                              io_buffer_size, pathLen
  use tem_aux_module,        only: tem_abort
  use tem_dyn_array_module,  only: dyn_longArray_type, dyn_intArray_type,      &
    &                              init, append, expand, destroy,              &
    &                              PositionOfVal
  use tem_grow_array_module, only: grw_longArray_type, grw_intArray_type,      &
    &                              init, append, expand, destroy
  use treelmesh_module,      only: treelmesh_type
  use tem_geometry_module,   only: tem_coordOfId, tem_idOfCoord, tem_tIdInfo,  &
    &                              tem_posOfPath
  use tem_tools_module,      only: tem_PositionInSorted, tem_write
  use tem_logging_module,    only: logUnit, tem_logging_isActive, tem_toStr
  use tem_debug_module,      only: tem_debug, main_debug, dbgUnit
  use tem_comm_env_module,   only: process, tem_comm_env_type
  use tem_comm_module,       only: tem_communication_type, tem_comm_dumpType,  &
    &                              tem_commPattern_type
  use tem_topology_module,   only: tem_LevelOf, tem_FirstIdAtLevel,            &
    &                              tem_PathComparison, tem_ParentOf,           &
    &                              tem_directChildren, tem_ChildNumber,        &
    &                              tem_PathOf, tem_path_type
  use tem_property_module,   only: prp_hasBnd, prp_missingNeigh
  use tem_bc_prop_module,    only: tem_bc_prop_type
  use tem_stencil_module,    only: tem_stencilHeader_type,                     &
    &                              tem_stencilElement_type,                    &
    &                              tem_stencil_map_toTreelmDef,                &
    &                              tem_stencil_getHeaderPos
  use tem_element_module,    only: tem_element_type, tem_element_dump,         &
    &                              init, append, PositionOfVal, truncate,      &
    &                              eT_fluidElem, eT_nonExistingElem,           &
    &                              eT_haloElem, destroy,                       &
    &                              eT_ghostFromCoarserElem, eT_minNumber,      &
    &                              eT_ghostFromFinerElem, eT_maxNumber,        &
    &                              eT_minNumRelevant, eT_undefinedElem,        &
    &                              eT_distributedGhostFromFinerElem
  use tem_halo_module,       only: dyn_haloArray_type, append, init, destroy,  &
    &                              tem_halo_type, PositionOfVal

  implicit none

  private

  public :: tem_levelDesc_type
  public :: tem_levelNeighbor_type
  public :: depSource_type

  public :: prp_nLists

  public :: tem_find_allElements
  public :: tem_build_horizontalDependencies
  public :: tem_build_verticalDependencies
  public :: tem_cleanupDependencyArrays
  public :: tem_cleanup_arrays
  public :: tem_init_elemLevels
  public :: tem_eTypeOfId
  public :: tem_eTypeOfTotalPos
  public :: tem_treeIDinTotal
  public :: tem_killArrays
  public :: tem_find_depProc
  public :: tem_elemList_dump
  public :: tem_debug_HorizontalDependencies
  public :: tem_dumpPropBits
  public :: tem_dumpTreeIDlists
  public :: tem_debugDependencies

  !> limit for searching neighbors
  !! * For acoustic scaling (2 coarser elements required) set to 1
  !! * For diffusive scaling (4 coarser elements required) set to 3
  integer, save :: nestingLimit

  !> Hash to quickly identify if an element was reconstructed before.
  !! If so, it is part of the hash
  integer(kind=long_k), allocatable, save :: hash(:)
  integer, allocatable, save :: hash_elemPos(:)
  !> Entries in the sh
  integer(kind=long_k) :: nHashes

  !> identification parameters for different lists
  integer, parameter :: prp_nLists    =  4
  !\todo replace by element%stencil%tID
  !> includes the direct neighbors of each tree ID
  type tem_levelNeighbor_type
    !> array of the neighbors in the resulting totalList.
    !! Use this one in the solver!
    integer,allocatable :: nghElems(:,:)
    !> treeIDs of neighbors in respective directions
    !! size should be nElems_stencil, nElems
    !! temporary array, which is used between findNeighbor and create_levelDescr
    !! for elemenet has prp_hasBnd, negative BC_ID is assigned to that direction
    !integer(kind=long_k),allocatable :: neigh(:,:)
  end type tem_levelNeighbor_type

  !> Type to specify the dependencies of ghost and halo cells.
  !! E.g.: used to specify which cells have to be known to be able
  !! to interpolate a ghost/halo cell
  type depSource_type
    !> the source level, from where the current ghost element
    !! gets the source elements for the interpolation
    integer                  ::  dependencyLevel = -1
    !> position of the source elements in the totalList
    type( grw_intArray_type ) :: elem
    !> position of the source elements in the subbuffer
    type( grw_intArray_type ) :: elemBuffer
    !> Interpolation weight for each source element specified above
    real(kind=rk), allocatable :: weight(:)
    real(kind=rk) :: coord(3)
    !> What's this stencil?
    integer, allocatable :: stencil(:,:)
    integer :: childNum
  end type depSource_type

  !> detailed information of a complete level of elements
  !! including all treeIDs, properties and neighbors as well as informations
  !! about ghost/halo cells and its dependencies for
  !! interpolation/reconstruction
  type tem_levelDesc_type

    !---------new types for element description---------------------
    type(tem_element_type) :: elem
    !> This list includes treeIDs for which additionally
    !! neighbors have to be identified
    type(dyn_longArray_type) :: require
    type(dyn_haloArray_type) :: halosPrc
    !> points to the position of the element in the elem array
    !type(grw_intArray_type) :: fluid
    !type(grw_intArray_type) :: halo

    !> ghost treeIDs on current level, i.e. elements that are virtual, but local
    !type(grw_intArray_type) :: ghostFromFiner
    !type(grw_intArray_type) :: ghostFromCoarser
    !---------------------------------------------------------------

    !> list of treeIDs for this level. Consists of ordered treeIDs of
    !! first fluid, then ghost, then halo elements.
    !! total:
    !! \verbatim
    !! +---------------------+--------------------+------------------+---------+
    !! |........fluid........|..ghostFromCoarser..|..ghostFromFiner..|...halo..|
    !! +---------------------+--------------------+------------------+---------+
    !! \endverbatim
    integer(kind=long_k), allocatable :: total(:)
    integer, allocatable :: totalPnt(:)
    !> list of property bits for this level. the same order as total list
    integer(kind=long_k), allocatable :: property(:)
    !> pointer from the levelDescriptor to the original treeID list
    !! ( only for fluids )
    integer, allocatable :: pntTID(:)

    !> neighbor relations for all fluid elements.
    !! Dimensions: 1: maxNeighbors    2: nElems_fluid (+nElems_ghost)
    !! We store the positions of the neighbor elements inside the total-list.
    !! If a fluid element does not have a neighbor in a direction
    !! (e.g. because of a boundary in that direction), we store the boundary ID
    !! as negative to indicate, that it is not a regular neighbor.
    type( tem_levelNeighbor_type ), allocatable :: neigh(:)

    !> Dependencies for ghost elements
    !! To reconstruct all the data you should
    !! iterate over this list and reconstruct the ghost elements
    !! with source element information from these data types
    !! data. Up = to coarser, down = to finer
    type( depSource_type ), allocatable :: depFromFiner(:)
    type( depSource_type ), allocatable :: depFromCoarser(:)
    !> Store the source elements for the ghostFromFiner
    type(dyn_intArray_type) :: sourceFromFiner
    !> Store the source elements for the ghostFromCoarser
    type(dyn_intArray_type) :: sourceFromCoarser
    !> Buffer storing intermediate values of the source elements for
    !! the interpolation
    real(kind=rk), allocatable :: intpBufFromFiner(:,:)
    real(kind=rk), allocatable :: intpBufFromCoarser(:,:)
    !> List to store high order interpolation ghost elements
    type(grw_intArray_type) :: intpFromCoarser_high
    !> List to store high order interpolation ghost elements
    type(grw_intArray_type) :: intpFromCoarser_low
    !> List to store high order interpolation ghost elements
    type(grw_intArray_type) :: intpFromFiner_high
    !> List to store high order interpolation ghost elements
    type(grw_intArray_type) :: intpFromFiner_low

    !> pointing to the position of boundary elements
    !! in the levelDescriptor total list
    type(dyn_intArray_type) :: bc_elemBuffer

    !> Offsets in the assembled lists for
    !! fluid (1), ghostFromCoarser(2), ghostFromFiner(3) and halo(4) elements
    !! for the assembled lists, i.e the totalList, invSorted, ...
    !! gets the values (0, nElems_fluid,
    !!   nElems_fluid+nElems_ghostCoarser,
    !!   nElems_fluid+nElems_ghostCoarser+nELems_ghostFiner)
    integer :: offset( 2, eT_minNumber:eT_maxNumber ) = 0

    type( tem_communication_type ) :: sendbuffer
    type( tem_communication_type ) :: sendbufferFromCoarser
    type( tem_communication_type ) :: sendbufferFromFiner
    type( tem_communication_type ) :: recvbuffer
    type( tem_communication_type ) :: recvbufferFromCoarser
    type( tem_communication_type ) :: recvbufferFromFiner

    ! number of elements
    integer :: nElems_fluid
    integer :: nElems_ghostFromCoarser
    integer :: nElems_ghostFromFiner
    integer :: nElems_halo
    integer :: nElems

  end type tem_levelDesc_type

  contains

!*******************************************************************************
  !> call this routine from your geometry initialization routine in the solver
  !! create all the necessary level-wise objects, such as element lists,
  !! dependencies
  !!
  !! 1.) build all dependencies for halos and ghost which
  !! are needed for interpolation/reconstruction (including MPI communications)
  !! 2.) build the pointers for each element to its neighbors/stencil elements.
  !! All this information is stored in tem_levelDesc_type
  !!
  subroutine tem_find_allElements( tree, levelDesc, levelPointer,              &
    &                              computeStencil, commPattern,                &
    &                              cleanup, reqNesting, dumpTreeIDlists )
    !---------------------------------------------------------------------------
    !>
    type(treelmesh_type), intent(inout) :: tree
    !>
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minLevel:)
    !> Pointer from treeIDlist entry to level-wise fluid part of total list
    integer, allocatable, intent(out) :: levelPointer(:)
    !>
    type(tem_stencilHeader_type) :: computeStencil(:)
    !>
    type(tem_commPattern_type), intent(in)    :: commPattern
    !>
    integer, intent(in), optional :: reqNesting
    !>
    logical, intent(in), optional :: cleanup
    !>
    logical, intent(in), optional :: dumpTreeIDlists
    !---------------------------------------------------------------------------
    ! Geometry and Tree related variables
    integer,allocatable :: eligible_child(:) ! possible children (can be 1...8)
    integer :: iLevel
    type(tem_path_type), allocatable :: pathFirst(:), pathLast(:)
    logical :: doAdditional
    logical :: clean_constructionArrays
    logical :: dumpLists
    !---------------------------------------------------------------------------
    if( present( dumpTreeIDlists )) then
      dumpLists = dumpTreeIDlists       
    else
      dumpLists = .false.
    end if
    if( present( cleanup )) then
      clean_constructionArrays = cleanup
    else
      clean_constructionArrays = .false.
    end if
    if( present( reqNesting )) then
      nestingLimit = reqNesting
    else
      nestingLimit = 1
    end if
    write(logUnit(3),*) 'Building up the element and total list.'

    allocate(pathFirst( tree%global%nParts ))
    allocate(pathLast( tree%global%nParts ))

    ! Build the pathlist for improved performance when finding local element
    ! positions since the tree is the same for every scheme the allocation only
    ! has to be performed once
    if ( .not. allocated(tree%pathList))then
      allocate( tree%pathList(tree%nElems) )
    end if

    ! Fill the pathList for each element in the treeID list
    tree%pathList =  tem_PathOf( tree%treeID )
    pathFirst     =  tem_PathOf( tree%Part_First )
    pathLast      =  tem_PathOf( tree%Part_Last )

    call init_arrays( levelPointer, tree, process, eligible_child )

    call build_levelElements( levelDesc, levelPointer, tree, process,          &
         &        computeStencil, pathFirst, pathLast, dumpLists )

    do iLevel = tree%global%minLevel, tree%global%maxLevel
      call identify_lists( tree, levelDesc, iLevel )
    end do

    call communicate_elements( tree, process, levelDesc, commPattern,          &
      &               pathFirst, pathLast, computeStencil, dumpLists)

    call check_additionalComm( levelDesc, process, doAdditional,               &
      &                        tree%global%minlevel ) 
   
    if( doAdditional ) then
      ! passing only first stencil as this is the required compute stencil
      call identify_additionalNeigh( tree, process, levelDesc,                 &
        &        pathFirst, pathLast, computeStencil(1) )
      call communicate_elements( tree, process, levelDesc, commPattern,        &
        &        pathFirst, pathLast, computeStencil, dumpLists )
    end if

    call assemble_lists( levelDesc, tree%global%minLevel )

    call update_elemPosToTotalPos( levelDesc, levelPointer, tree,              &
      &                            computeStencil )
    ! Warning: Truncation introduces a memory peak because of copy
    ! operations! Better not use...
    !call truncate_lists( levelDesc, tree%global%minLevel )

    if( clean_constructionArrays ) then
      call tem_cleanupDependencyArrays( levelDesc )
    endif

    write(logUnit(3),*) 'Finished building up element and total list. '
    deallocate(pathFirst)
    deallocate(pathLast)
    deallocate(hash)
    deallocate(hash_elempos)

  end subroutine tem_find_allElements
!******************************************************************************!


!******************************************************************************!
  !> subroutine to find neighbours of cells
  !!
  !! Typically every element requires information from its neighbors to perform
  !! an update of its state.
  !! All such required neighbors constitute a so called \ref tem_stencil_module
  !! "stencil".
  !! \ref tem_build_horizontalDependencies "horizontal relations", or the 
  !! connectivity of elements on the same refinement
  !! level in the tree, are therefore essential for the stencils in typical
  !! mesh-based numerical schemes.
  !! We now sketch the method to find the connectivity of elements in the mesh
  !! with the help of the described mesh layout.
  !! In a distributed mesh, the first distinction has to be made with respect to
  !! processor ownership.
  !! Is the element in question a \ref tem_find_depNeigh "local or remote
  !! element"?
  !!
  !! For all \ref tem_identify_local_element "local elements", their actual
  !! position in the sparse mesh has to be
  !! identified for a given 
  !! \ref treelmesh_module::treelmesh_type::treeid "treeID".
  !!
  !! A \ref tem_topology_module::tem_path_type "comparison of two nodes to
  !! decide their relative position" in the ordered list
  !! of elements has to take into account the hierarchy of the mesh.
  !!
  !! As in the actual simulation only leaves will be present, and no overlap of
  !! refinement levels is allowed, this is a sufficient determination.
  !! However, such overlapping regions might be created to enable the
  !! interpolation between different levels with virtual elements.
  !! These special elements are distinguished anyway, so this does not pose any
  !! problem in the search for neighborhood relations.
  !!
  !! An element can be either identified by its
  !! \ref treelmesh_module::treelmesh_type::treeid "treeID"
  !! or by a tuple with four integer entries for the spatial coordinates and the
  !! refinement level: \f$(x, y, z, L)\f$.
  !! This coordinate fully describes the spatial shape and position of an
  !! element and is important to determine spatial relations between elements.
  !! Conversion between treeIDs and coordinates can be achieved by the routines
  !! - \ref tem_geometry_module::tem_coordofid "Coordinate from TreeID"
  !! - \ref tem_geometry_module::tem_idofcoord "TreeID from Coordinate"
  !!
  !! The spatial indices are limited by the refinement level:
  !! \f$ x, y, z \in \mathbf{Z}_{\ge 0}: x, y, z < 2^L \f$.
  !! To avoid undefined coordinates, movements through the mesh by additions to
  !! indices are done in a modulo(\f$2^L\f$) safeguard resulting in an periodic
  !! universe cube.
  !! This enables the movement in the mesh in horizontal direction by index
  !! alteration and translation into a
  !! \ref treelmesh_module::treelmesh_type::treeid "treeID" again.
  !! The described procedure is completely reversible, enabling the construction
  !! of the \ref treelmesh_module::treelmesh_type::treeid "treeID" for any
  !! given coordinate.
  !! Thus, the conversion between this coordinate and the serialized 
  !! \ref treelmesh_module::treelmesh_type::treeid "treeID"
  !! encoding is fully described by the topology of the octree and the chosen
  !! space-filling curve ordering of elements, as explained above.
  !! 
  !! With this method we can describe the neighborhood of any given element
  !! independently of the actual solver by a simple list of relative offsets.
  !! In contrast to the generic horizontal relation, the vertical relation
  !! between child and parent nodes in the tree requires an interpolation
  !! operator between different refinement levels.
  !! This interpolation usually has to take into account solver specific
  !! requirements, but is otherwise quite isolated from the numerical operation
  !! on each refinement level.
  !! The \emph{TreElM} library offers the solver a level-wise view, as suggested
  !! by the properties described above.
  !! To find all required neighbors in the distributed octree, the solver merely
  !! has to provide its horizontal dependencies.
  !! These are described with the help of an element specific \ref
  !! tem_stencil_module "stencil".
  !! A stencil is basically a set of element-offsets \f$(s_x, s_y, s_z)\f$,
  !! describing the relative positions of all required elements for a given
  !! element.
  !!
  !! \ref tem_construction_example_stencil
  !!
  subroutine tem_init_elemLevels( me, boundary, tree, computeStencil )
    !---------------------------------------------------------------------------
    !> neighbor list containing all the neighbours for the
    !! cells given in treeidsubset. Result of this routine
    type(tem_levelDesc_type), allocatable, intent(out) :: me(:)
    !> boundaries for the elements with bnd property set
    type(tem_bc_prop_type), intent(in) :: boundary
    !> subset of tree ids for which the neighbours will be specified
    type(treelmesh_type), intent(in)   :: tree
    !> the given stencil
    type(tem_stencilHeader_type), intent(inout) :: computeStencil(:)
    !---------------------------------------------------------------------------
    integer :: addedPos
    type(tem_stencilElement_type) :: tStencil
    integer :: iElem, nElemsBnd, iQQN, iLevel, nElems, nProcs, hashpos
    integer :: x(4), nStencils, iStencil, elemPos, nStencilElems
    integer :: indElem
    logical :: wasAdded
    integer(kind=long_k) :: tOffset
    integer(kind=long_k), allocatable :: tStencilIDs(:)
    integer(kind=long_k) :: elemTID
    integer(kind=long_k), allocatable :: tComputeIDs(:)
    !---------------------------------------------------------------------------

    nElems = size(tree%treeID)
    nProcs = size(tree%Part_First)

    ! Find a suitable hash size.
    ! The hash lookup is O(1), while the identification grows with
    ! log(nElems) and log(nProcs), yet allocation, deallocation and
    ! filling will cause some overhead of O(nHashes), this is an attempt to
    ! find a compromise of these parameters.
    nHashes = int(max(nElems, nProcs, 64), kind=long_k)

    ! \todo Find the closest smaller prime for the hash size, to minimize the
    !       number of collisions?
    ! Limit the size of the hash:
    nHashes = min(nHashes, int(io_buffer_size, kind=long_k))

    allocate(hash(0:int(nHashes-1)))
    allocate(hash_elemPos(0:int(nHashes-1)))
    ! Reset the hash
    hash = -1_long_k
    hash_elemPos = -1

    write(logUnit(5),*) 'Allocating level Descriptor ...'
    allocate( me( tree%global%minLevel:tree%global%maxLevel ))

    do iLevel = tree%global%minLevel,tree%global%maxLevel
      call init(me = me(iLevel)%require, unique = .true.)
      write(logUnit(6),*) 'Initializing elemList on Level ', iLevel
      call init( me = me( iLevel )%elem )
    end do

    nElemsBnd = 0
    write(logUnit(4),*) 'Searching the neighbors for each element ...'
    nStencils = size( computeStencil )

    allocate( tComputeIDs( computeStencil(1)%QQN ))

    ! Loop over all given stencils
    stencilLoop: do iStencil = 1, nStencils
      write(logUnit(5),*) '  Identifying element neighbors on stencil ', &
        &                 iStencil

      call tem_stencil_map_toTreelmDef( me = computeStencil( iStencil ))
 
      if ( computeStencil(iStencil)%useAll ) then
        ! This stencil belongs to all elements in the tree
        nStencilElems = tree%nElems
      else
        if( iStencil == 1) then
          call tem_write('tem_init_elemLevels:')
          call tem_write('Error: The first stencil does not have the useAll '//&
            &            'flag set')
          call tem_write('       The first one is considered the compute '//   &
            &            'stencil and')
          call tem_write('       needs to be set for all elements.')
          call tem_abort()
        end if
        ! This stencil belongs to only a subset of the tree
        nStencilElems = computeStencil( iStencil )%nElems
      end if

      write(logUnit(6),*) '    nElems       ', nStencilElems
      write(logUnit(6),*) '    QQN          ', &
        &                 computeStencil(iStencil)%QQN
      if ( allocated(tStencilIDs) ) deallocate( tStencilIDs )
      allocate( tStencilIDs( computeStencil(iStencil)%QQN ) )

      ! Loop over all the elements connected to the current stencil 
      ! computeStencil( iStencil )
      elemLoop: do indElem = 1, nStencilElems
        ! Reset the empty stencil
        ! Create the temporary stencil for assigning the positions of the 
        ! neighbor treeIDs which are added in eleme%neighID
        ! also store the original stencil it depends on (depends = iStencil )
        call init( me        = tStencil,                                       &
          &        QQN       = computeStencil( iStencil )%QQN,                 &
          &        headerPos = iStencil )

        if( computeStencil( iStencil )%useAll ) then
          ! if the stencil is defined to be connected to all elements, 
          ! then simply loop over all the fluid elements
          iElem = indElem
        else
          ! if the stencil is defined for a subset, loop over the
          ! defined subset in %elem
          iElem = computeStencil( iStencil )%elem( indElem )
        end if
        ! assign fluid element with info from the tree list and properties
        ! from disk
        elemTID = tree%treeID( iElem )
        ! Get topological info about coordinates+level
        x = tem_CoordOfId( elemTID )
        tOffset = tem_FirstIDAtLevel( x(4) )

        ! Find position of this element in the elemList as it might be
        ! there already from previous stencil iterations
        hashpos = int(mod( elemTID, nHashes))
        if (hash(hashpos) /= elemTID) then ! cache miss
          call append( me         = me( x(4))%elem,                            &
            &          tID        = elemTID,                                   &
            &          pntTID     = iElem,                                     &
            &          eType      = eT_fluidElem,                              &
            &          nNeighIDs  = computeStencil( iStencil )%QQN,            &
            &          sourceProc = tree%global%myPart+1,                      &
            &          property   = tree%ElemPropertyBits( iElem ),            &
            &          pos        = elemPos,                                   &
            &          wasAdded   = wasAdded )
          hash(hashpos) = elemTID
          hash_elemPos( hashpos ) = elemPos
        else ! cache hit 
          elemPos = hash_elemPos( hashpos )
        end if

        ! Only perform the boundary check based on the treeID boundary
        ! information for stencils which work on all elements (so we can
        ! actually read out the boundary information from the tree )
        if( computeStencil( iStencil )%useAll .and. iStencil == 1) then
          ! First check, if I have prp_hasBnd
          ! If I don't take all the elements, then I can't analyze boundaries
          if( btest( me(x(4))%elem%property%val( elemPos ), prp_hasBnd )) then
          ! If yes, there might be no neighbors
            nElemsBnd = nElemsBnd +1
            me(x(4))%elem%boundaryPos%val( elemPos ) = nElemsBnd
          end if
        end if ! useAll elements? (boundary information)

        ! start searching for neighbors stored in nTreeID
        !! and create linked lists in neighborList
        ! \todo only nearest neighbors should be checked.
        tStencilIDs = 0_long_k

        directionLoop: do iQQN = 1, computeStencil( iStencil )%QQN
          if ( me(x(4))%elem%boundaryPos%val( elemPos ) .gt. 0                 &
            &  .and. computeStencil( iStencil )%map( iQQN ) .gt. 0 ) then
            ! If there is a boundary in the current direction
            ! ... add it to the temporary stencil
            if ( boundary%boundary_ID( computeStencil(iStencil)%map(iQQN),     &
              &                        me(x(4))%elem%boundaryPos%val(elemPos)  &
              &                      )  .gt. 0 ) then
              ! Set the negative boundary ID as neighbor, so we don't need
              !! more arrays for boundary information
              tStencilIDs( iQQN )                                              &
                &  =  -boundary                                                &
                &      %boundary_ID( computeStencil(iStencil)%map(iQQN),       &
                &                    me(x(4))%elem%boundaryPos%val(elemPos) )

            else if ( boundary                                                 &
              &       %boundary_ID( computeStencil(iStencil)%map(iQQN),        &
              &                     me(x(4))%elem%boundaryPos%val(elemPos)     &
              &                   )  .lt. 0 ) then
              ! Is a periodic neighbor! value is -treeID_periodicNeighbor
              ! Set the id given as treeID and make positive
              tStencilIDs( iQQN )                                              &
                &  = abs( boundary                                             &
                &         %boundary_ID( computeStencil(iStencil)%map(iQQN),    &
                &                       me(x(4))%elem%boundaryPos%val(elemPos) &
                &                     )                                        &
                &       )
            else
              ! ... there is NO boundary condition, so store the required
              ! neighbor
              ! Just store the neighboring treeID which will have to be
              ! found or reconstructed (might be an actual fluid element
              ! or ghost / halo)
              tStencilIDs( iQQN ) = tem_IdOfCoord(                             &
                &     [ x(1) + computeStencil( iStencil )%cxDir( 1, iQQN ),    &
                &       x(2) + computeStencil( iStencil )%cxDir( 2, iQQN ),    &
                &       x(3) + computeStencil( iStencil )%cxDir( 3, iQQN ),    &
                &       x(4) ], tOffset)
            endif
          else
            ! ... there is NO boundary condition, so store the required neighbor
            ! Just store the neighboring treeID which will have to be
            !! found or reconstructed (might be an actual fluid element
            !! or ghost / halo)
            tStencilIDs( iQQN )                                                &
              &  = tem_IdOfCoord( [ x(1) + computeStencil( iStencil )          &
              &                            %cxDir( 1, iQQN ),                  &
              &                     x(2) + computeStencil( iStencil )          &
              &                            %cxDir( 2, iQQN ),                  &
              &                     x(3) + computeStencil( iStencil )          &
              &                            %cxDir( 3, iQQN ),                  &
              &                     x(4) ],                                    &
              &                   tOffset )
          endif
          if( computeStencil( iStencil )%requireNeighNeigh                     &
            & .and. iStencil .gt. 1                                            &
            & .and. tStencilIDs(iQQN) .gt. 0 ) then
            ! If the neighbors of this stencil neigbors are required, add this 
            ! to the require array. This includes treeIDs for which additionall 
            ! neighbors have to be identified.
            call append( me  = me(x(4))%require,                               &
              &          val = tStencilIDs(iQQN),                              &
              &          pos = addedPos )
          end if
        end do directionLoop
        ! Append the stencil to the neighID array
        do iQQN = 1, computeStencil( iStencil )%QQN
          call append( me       = me(x(4))%elem%neighID%val( elemPos ),        &
            &          val      = tStencilIDs( iQQN ),                         &
            &          pos      = addedPos,                                    &
            &          wasAdded = wasAdded )
          tStencil%tIDpos( iQQN ) = addedPos
        end do

        if( computeStencil( iStencil )%QQN .gt. 0 ) then
          ! Append the temporary stencil to the element
          call append( me  = me(x(4))%elem%stencil%val( elemPos ),             &
            &          val = tStencil )
        end if
      end do elemLoop
    end do stencilLoop
    write(logUnit(4),*) 'Finished searching the neighbors for each element.'

    if (tem_logging_isActive(main_debug%logger, 5)) then
      do iLevel = tree%global%minLevel, tree%global%maxLevel
        call tem_elemList_dump( me      = me( iLevel )%elem,                  &
          &                     compact = .true.,                             &
          &                     nUnit   = dbgUnit(5),                         &
          &                     stencil = .true.,                             &
          &                     string  = 'after initializing level elements' )
        if( me( iLevel )%require%nVals .gt. 0 ) then
          call tem_require_dump( me      = me( iLevel )%require,               &
            &                    compact = .true.,                             &
            &                    nUnit   = dbgUnit(5),                         &
            &                    string  = 'after initializing level elements' )
        end if
      end do
    end if

  end subroutine tem_init_elemLevels
!******************************************************************************!


!******************************************************************************!
  !> Allocate and initialise most important arrays
  !!
  subroutine init_arrays( levelPointer, tree, prc, eligible_child )
    !---------------------------------------------------------------------------
    integer, allocatable, intent(out) :: levelPointer(:)
    type(treelmesh_type), intent(in) :: tree
    type(tem_comm_env_type), intent(inout) :: prc
    integer, allocatable, intent(out) :: eligible_child(:)
    !---------------------------------------------------------------------------

    ! allocate desciptors with minimal and maximal
    ! refinement level (of all MPI processs)
    allocate( levelPointer( tree%nElems ))
    levelPointer = -1 !< reset contents for easy debugging

    ! \todo This is LBM specific
    allocate( eligible_child(8) )
    eligible_child = (/1,2,3,4,5,6,7,8 /)

  end subroutine init_arrays
!******************************************************************************!


!******************************************************************************!
  !> Deallocate and de-initialise fluid, ghost and halo arrays
  !! todo: Give this subroutine s.th. to do
  !!
  subroutine tem_killArrays( levelDesc )
    !---------------------------------------------------------------------------
    type(tem_levelDesc_type), intent(inout) :: levelDesc(:)
    !---------------------------------------------------------------------------
    integer :: iLevel
    !---------------------------------------------------------------------------

    ! Allocate fluidList for each level
    do iLevel = 1, globalMaxLevels
    end do

  end subroutine tem_killArrays
!******************************************************************************!


!******************************************************************************!
  !>  Invoke the identify_elements for each neighbor of the stencil 
  !! and store the positions of the encountered elements
  !!
  recursive subroutine identify_stencilNeigh( iElem, iLevel, iStencil, tree,   &
    &                                         pathFirst, pathLast, levelDesc,  &
    &                                         prc, computeStencil, nesting )
    !---------------------------------------------------------------------------
    !> element position in levelDesc to identify
    integer, intent(in) :: iElem
    !> element level
    integer, intent(in) :: iLevel
    !> stencil within the element to act on 
    integer, intent(in) :: iStencil
    !> tree information
    type(treelmesh_type), intent(in) :: tree
    !>
    type(tem_path_type), intent(in) :: pathFirst(:)
    !>
    type(tem_path_type), intent(in) :: pathLast(:)
    !>
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minLevel:)
    !> process
    type(tem_comm_env_type), intent(in) :: prc
    !>
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !> nesting level
    integer, intent(in) :: nesting
    !---------------------------------------------------------------------------
    integer :: iStencilElem, elemPos
    integer :: neighIDpos
    integer(kind=long_k) :: nnTreeID
    !---------------------------------------------------------------------------
    ! identify all the compute neighbors of the current element
    do iStencilElem = 1, computeStencil%QQN
      neighIDpos = levelDesc( iLevel )%elem%stencil%val( iElem )               &
        &                               %val( iStencil )%tIDpos( iStencilElem )
      if( neighIDpos .gt. 0 ) then
        nnTreeID = levelDesc( iLevel )%elem%neighID%val( iElem )               &
          &                                                  %val( neighIDpos )
        ! This call might add new halo elements
        if ( nnTreeID .gt. 0_long_k ) then
          call identify_elements( nTreeID        = nnTreeID,                   &
            &                     tree           = tree,                       &
            &                     pathFirst      = pathFirst,                  &
            &                     pathLast       = pathLast,                   &
            &                     levelDesc      = levelDesc,                  &
            &                     elemPos        = elemPos,                    &
            &                     prc            = prc,                        &
            &                     computeStencil = computeStencil,             &
            &                     nesting        = nesting )
        else
          elemPos = 0
        end if
      else
        elemPos = 0
      end if
      ! And add the encountered neighbor elements to the current element's
      ! stencil neighbors
      levelDesc( iLevel )%elem%stencil%val( iElem )                            &
        &           %val( iStencil )%totalPos( iStencilElem ) = elemPos
    end do
 
  end subroutine identify_stencilNeigh
!******************************************************************************!


!******************************************************************************!
  !> Check, on which partition a given element is located add required elements 
  !! to corresponding lists:
  !! if remote, add to halo
  !! if ghost, add to resp. ghost list
  !!
  recursive subroutine identify_elements( nTreeID, tree, pathFirst, pathLast,  &
    &                                     levelDesc, elemPos, prc,             &
    &                                     computeStencil, nesting )
    !---------------------------------------------------------------------------
    !> neighboring treeID
    integer(kind=long_k), intent(in) :: nTreeID
    !> tree information
    type(treelmesh_type), intent(in) :: tree
    !>
    type(tem_path_type), intent(in) :: pathFirst(:)
    !>
    type(tem_path_type), intent(in) :: pathLast(:)
    !>
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minLevel:)
    !> process
    type(tem_comm_env_type), intent(in) :: prc
    !>
    integer, intent(out) :: elemPos
    !>
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !> nesting level
    integer, intent(in) :: nesting
    !---------------------------------------------------------------------------
    integer(kind=long_k) :: children(8) ! child elements
    integer :: nDepProcs ! processes that this neigh depend on
    integer :: depProc
    integer :: iChild, neighLevel
    type(tem_path_type) :: neighPath
    type(tem_stencilElement_type) :: emptyStencil(1)
    integer :: relFirst, relLast
    integer :: childPos   ! position of child element
    integer :: p_lb, p_ub ! process lower and upper bound
    integer :: nNesting ! new nesting level
    ! Position of the current element in the hash
    integer :: hashpos, elemNesting
    logical :: cacheHit
    logical :: updated ! was the element updated during identify_local_element
    !--------------------------------------------------------------------------=
    ! Increase the nesting
    nNesting = nesting + 1
    cacheHit = .false.

    write(dbgUnit(6),*) 'identifying required TreeID for ', &
      &                 nTreeID,                            &
      &                 ' nesting ', nesting 
    if( nTreeID .gt. 0 ) then
      neighLevel = tem_LevelOf( nTreeID )
      neighPath  = tem_PathOf( nTreeID )
      hashpos = int(mod(nTreeID, nHashes))
      if( hash(hashpos) .gt. -1_long_k ) then
        ! something is in the hash
        if( hash(hashpos) == nTreeID ) then
          if( nesting .eq. levelDesc( neighlevel )%elem%haloNesting%val(       &
            &                                   hash_elemPos( hashPos )) ) then
            cacheHit = .true.
          end if
          if( levelDesc( neighlevel )%elem%needsUpdate%val(                    &
            &                                   hash_elemPos( hashPos )) ) then
            cacheHit = .false.
           ! Set the needs update to false as it was now updated
           levelDesc( neighLevel )%elem%needsUpdate%val(                       &
             &                               hash_elemPos( hashPos )) = .false.
          end if
        end if
      end if
      !\todo: Reactivate the cache! The problem is that the nesting of an
      !element is updated to 0 without actually looping over its neighbors!
      !cacheHit = .false.
      cachemiss: if( .not. cacheHit ) then
        ! Probably did not hit this element yet, put it in the hash now,
        ! and identify it.
        ! (If this treeID is already in the hash, we definitely have
        !  already processed this treeID, and need nothing to do anymore).
        ! This leaves elements to be identified multiple times only in
        ! the case of hash collisions, which should be pretty few, for
        ! a sufficiently large hash.
        !-----------------------------------------------
        ! Cache miss

        nDepProcs = 0
        ! First check if this neighbor is in my local tree range
        p_lb = 1
        relFirst = tem_PathComparison(neighPath, PathFirst(prc%rank+1))
        relLast = tem_PathComparison(neighPath, PathLast(prc%rank+1))
        if (relFirst < 0) then
          ! The searched element is definitely left of myself
          ! Do only search up to my own range
          p_ub = prc%rank
        else
          p_ub = int(min(int(prc%comm_size, kind=long_k),                      &
            &            tree%global%nElems))
          if (relLast > 0) then
            ! The searched element is definitely right of myself
            ! Do only search beyond my own range
            p_lb = prc%rank+2
          else

            ! The element might be (partly) on my own partition
            if (relLast < 0) then
              ! The element is at least left of my right border
              ! Set the bounds to include myself
              p_ub = prc%rank + 1
            end if
            if (relFirst > 0) then
              ! The element is at least right of my left border
              ! Set the bounds to include myself
              p_lb = prc%rank + 1
            end if
          end if
        end if ! relFirst < 0

        if ((p_lb == p_ub) .and. (p_lb == prc%rank+1)) then
          ! The element is local, do not go on to other processes
          nDepProcs = 1
          depProc = prc%rank+1
        else
          ! Possibly NON-LOCAL Element...
          ! Every process COULD hold a part of the current neighbor element
          call tem_find_depProc(depProc, nDepProcs, &
            &                    neighPath, p_lb, p_ub, PathFirst, PathLast)
        end if ! local or non-local
        ! How many processes possess a part of the requested treeID
        if ( nDepProcs .eq. 1 ) then
          updated = .false.
          ! might be a local, halo of same level or a ghostFromCoarser
          call single_process_neighbor( nTreeID        = nTreeID,              &
            &                           levelDesc      = levelDesc,            &
            &                           tree           = tree,                 &
            &                           prc            = prc,                  &
            &                           iProc          = depProc,              &
            &                           minLevel       = tree%global%minlevel, &
            &                           foundLevel     = neighLevel,           &
            &                           elemPos        = elemPos,              &
            &                           updated        = updated,              &
            &                           computeStencil = computeStencil,       &
            &                           nesting        = nesting )
          elemNesting = min( nesting,                                          &
                  levelDesc( neighLevel )%elem%haloNesting%val( elemPos ) )
          if ( nesting .lt. levelDesc( neighLevel )%elem%haloNesting           &
            &                                            %val( elemPos )) then
            ! element needs updating as the current nesting is smaller than the
            ! one in the element property
            levelDesc( neighLevel )%elem%needsUpdate%val( elemPos ) = .true.
            levelDesc( neighLevel )%elem%haloNesting%val( elemPos )            &
              &  = elemNesting
            updated = .true.
          end if

          !updated = .true.
          if( updated ) then
            if ( levelDesc( neighLevel )%elem%eType%val( elemPos )             &
              &                               == eT_ghostFromCoarserElem ) then
              if( nesting .lt. nestingLimit ) then
                ! Create the direct neighbors of the ghostFromCoarser

                ! identify all the compute neighbors of the current element
                call identify_stencilNeigh( iElem          = elemPos,          &
                  &                         iLevel         = neighLevel,       &
                  &                         tree           = tree,             &
                  &                         iStencil       = 1,                &
                  &                         pathFirst      = pathFirst,        &
                  &                         pathLast       = pathLast,         &
                  &                         levelDesc      = levelDesc,        &
                  &                         prc            = prc,              &
                  &                         computeStencil = computeStencil,   &
                  &                         nesting        = elemNesting + 1 )

              end if ! nesting < 1?
        
              write(dbgUnit(6),*) 'creating all parent neighbors for ', &
                &              levelDesc(neighLevel)%elem%tID%val(elemPos)
              ! Create all elements required up to the actual existing fluid
              ! element these include the neighbors of the parents. In a level
              ! jump >1, these intermediate levels have to provide valid
              ! quantities over two of their computation updates to account for
              ! the recursive algorithm.
              call create_allParentNeighbors( targetElem     = elemPos,        &
                &                             nesting        = nesting,        &
                &                             level          = neighLevel,     &
                &                             tree           = tree,           &
                &                             computeStencil = computeStencil, &
                &                             levelDesc      = levelDesc,      &
                &                             pathFirst      = pathFirst,      &
                &                             pathLast       = pathLast,       &
                &                             prc            = prc )
        
            endif ! ghostFromCoarser?
          endif ! updated?

        elseif ( nDepProcs .gt. 1 ) then
          ! more than one depending processes
          ! create ghost and find children (can only have finer children)
          call init( me        = emptyStencil(1),                              &
            &        QQN       = computeStencil%QQN,                           &
            &        headerPos = 1 )
          ! Add to the level-wise ghost list
          ! append and store position of element for return
          call append( me      = levelDesc( neighLevel )%elem,                 &
               &       tID     = nTreeID, property = 0_long_k,                 &
               &       eType   = eT_distributedGhostFromFinerElem,             &
               &       stencil = emptyStencil,                                 &
               &       pos     = elemPos )

          !... and store the position in the ghost list !
          ! Now find all children of this ghost
          do iChild = 1,8
            children = tem_directChildren( nTreeID )
            call identify_elements( nTreeID        = children( iChild ),       &
              &                     tree           = tree,                     &
              &                     pathFirst      = pathFirst,                &
              &                     pathLast       = pathLast,                 &
              &                     levelDesc      = levelDesc,                &
              &                     prc            = prc,                      &
              &                     elemPos        = childPos,                 &
              &                     computeStencil = computeStencil,           &
              &                     nesting        = nNesting )
          end do
        else
          ! cell not existing. stop.
          write(*,*) '---------- WARNING!!! -------------------------'
          write(*,*) 'cell', nTreeID ,' not existing. ' //                     &
            &        'Number of dependencies', nDepProcs
          call tem_tIDinfo( me = nTreeID, tree = tree )
          write(*,*) 'WARNING!!! This should never occur and points to a bug'  &
            &        // ' or a buggy mesh.'
        end if
        if( elemPos .gt. 0 ) then
          ! Set the encountered element position to the hash
          hash(hashpos) = nTreeID
          hash_elemPos( hashpos ) = elemPos
        end if
        ! end cache miss 
        !-----------------------------------------------
      else cachemiss
        !-----------------------------------------------
        ! Cache hit
        elemPos = hash_elemPos( hashpos )
        levelDesc( neighLevel )%elem%haloNesting%val( elemPos )                &
          &  = min( levelDesc( neighLevel )%elem%haloNesting%val(elemPos),     &
          &         nesting )
        !-----------------------------------------------
      end if cachemiss
    else
      elemPos = int( nTreeID )
    end if ! nTreeID .gt. 0


  end subroutine identify_elements
!******************************************************************************!


!******************************************************************************!
  !> create all the neighbors of an element's parent
  !!
  !! Create all elements required up to the actual existing fluid
  !! element these include the neighbors of the parents. In a level
  !! jump >1, these intermediate levels have to provide valid
  !! quantities over two of their computation updates to account for
  !! the recursive algorithm.                                         
  !! 
  !! Here the fromCoarser interpolation should be handed in.
  !!
  recursive subroutine create_allParentNeighbors(                              &
    &                    targetElem, level, computeStencil, tree, levelDesc,   &
    &                    pathFirst, pathLast, prc, nesting )
    !---------------------------------------------------------------------------
    !> requested element position
    integer, intent(in) :: targetElem
    !> requested element level
    integer, intent(in) :: level
    !>
    type(tem_path_type), intent(in) :: pathFirst(:)
    !>
    type(tem_path_type), intent(in) :: pathLast(:)
    !> tree information
    type(treelmesh_type), intent(in) :: tree
    !>
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minLevel:)
    !> process
    type(tem_comm_env_type), intent(in) :: prc
    !>
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !> nesting level
    integer, intent(in) :: nesting
    !--------------------------------------------------------------------------=
    integer(kind=long_k) :: cTreeID  ! current tree ID
    integer :: coarserLevel, cPos, parentNesting
    !---------------------------------------------------------------------------
    !nNesting = nesting + 1
    coarserLevel = level
    ! The parent again must be computed, so the nesting has to be 0
    parentNesting = -1
    cPos = targetElem
    ! exit if we have reached the minimal level
    if( coarserLevel == tree%global%minlevel ) return

    ! current element treeID
    cTreeID = levelDesc( coarserLevel )%elem%tID%val( cPos )

    ! Get the parent of the current treeID
    coarserLevel = coarserLevel - 1
    cTreeID = tem_parentOf( cTreeID )

    ! ... and identify the parent (new current tree ID)
    call identify_elements( nTreeID        = cTreeID,                          &
      &                     tree           = tree,                             &
      &                     pathFirst      = pathFirst,                        &
      &                     pathLast       = pathLast,                         &
      &                     levelDesc      = levelDesc,                        &
      &                     elemPos        = cPos,                             &
      &                     prc            = prc,                              &
      &                     computeStencil = computeStencil,                   &
      &                     nesting        = parentNesting )

    if ( cPos <= 0 ) then
      write(dbgUnit(3),*) ' Element not found ', cTreeID
      write(dbgUnit(3),*) '   cPos         ', cPos
    end if

    ! identify the stencil neighbors of the parent.
    ! Here we should identify the fromCoarser interpolation stencil neighbors 
    ! instead of the compute stencil neighbors
    call identify_stencilNeigh( iElem          = cPos,                         &
      &                         iLevel         = coarserLevel,                 &
      &                         tree           = tree,                         &
      &                         iStencil       = 1,                            &
      &                         pathFirst      = pathFirst,                    &
      &                         pathLast       = pathLast,                     &
      &                         levelDesc      = levelDesc,                    &
      &                         prc            = prc,                          &
      &                         computeStencil = computeStencil,               &
      &                         nesting        = parentNesting )

  end subroutine create_allParentNeighbors
!******************************************************************************!


!******************************************************************************!
  !> Find partitions holding data on a given path
  !!
  !! Using a binary search over the processes first and last elements.
  !!
  !! Is the element in question a local or remote element?
  !! To look up a certain element by its
  !! \ref treelmesh_module::treelmesh_type::treeid "treeID" in the distributed
  !! list of elements, it is sufficient to know the splitting positions of all
  !! chunks.
  !! That is, the first and last
  !! \ref treelmesh_module::treelmesh_type::treeid "treeID" of each partition.
  !! With a binary search over the splitting positions any requested element can
  !! then be identified to be either outside the computational domain at all, or
  !! inside of one or several known partitions.
  subroutine tem_find_depProc( depProc, nDepProcs, neighPath, p_lb, p_ub,      &
    &                          PathFirst, PathLast)
    !---------------------------------------------------------------------------
    !> List of partitions
    integer, intent(out) :: depProc
    !> Number of partitions
    integer, intent(out) :: nDepProcs
    !> Element to look up
    type(tem_path_type), intent(in) :: neighPath
    !> Left partition bounds
    type(tem_path_type), intent(in) :: PathFirst(:)
    !> Right partition bounds
    type(tem_path_type), intent(in) :: PathLast(:)
    !> Left interval bound to search in
    integer, intent(in) :: p_lb
    !> Right interval bound to search in
    integer, intent(in) :: p_ub
    !---------------------------------------------------------------------------
    integer :: lb, ub
    integer :: foundProc, lastProc
    integer :: curProc
    integer :: pComp
    !---------------------------------------------------------------------------

    lb = p_lb
    ub = p_ub


    ! Find the partition, to which neighPath is definitely right of
    ! the first element.
    do
      curProc = (lb + ub) / 2
      if( curProc .le. 0 ) then
        ! Element not found on any proc
        foundProc = 0
        exit
      end if
      pComp = tem_PathComparison(neighPath, PathFirst(curProc))
      if (pComp > 0) then
        ! The first element of curProc is smaller than neighPath, therefore
        ! we do not have to search left of curProc. Can we even go to curProc+1?
        if (tem_PathComparison(neighPath, PathLast(curProc)) > 0) then
          ! The search element is not in the partition curProc at all, continue
          ! search with new lower bound of curProc+1.
          lb = min(curProc + 1, p_ub)
        else
          ! The left element is smaller than the search element and the right
          ! one is larger or equal to it, thus we found the first partition,
          ! which contains information on the element.
          foundProc = curProc
          exit
        end if
      else
        ! The first element of curProc is larger or equal to neighPath, thus
        ! we can restrict our search interval up to this partition.
        ub = curProc
        ! It is OK to not decrease the interval further here, as we always
        ! round down to find the middle, thus, if by this step we set
        ! ub = lb+1, the next iteration will test a new element, as we get
        ! curProc = (lb + lb + 1) / 2 = lb.
      end if
      if (ub <= lb) then
        ! Interval of size 0, found the first process to hold information on
        ! the element in neighpath.
        ! The lower bound was tested strictly, thus it is safe to set the first
        ! process to the lower bound here and exit.
        foundProc = lb
        exit
      end if
    end do

    ! Now linearly search for the last process spanned by the searched element.
    ! Usually this is just the single partition identified above or one more.
    lastProc = foundProc
    if( lastProc .gt. 0 ) then
      nDepProcs = 1
      depProc = lastProc
      do
        if (lastProc == p_ub) exit
        if (tem_PathComparison(neighPath, PathLast(lastProc)) < 0) exit
        if (tem_PathComparison(neighPath, PathFirst(lastProc+1)) < 0) exit
        lastProc = lastProc + 1
        nDepProcs = nDepProcs + 1
      end do
    else
      nDepProcs = 0
    end if

  end subroutine tem_find_depProc
!******************************************************************************!


!******************************************************************************!
  !> Determine the location (which process) of a requested element,
  !! which was identified to be located on one single process
  !! (can be local or remote)
  !! If it is located on a remote process: add to halo list
  !!                     local process: identify if ghost or fluid
  !!
  subroutine single_process_neighbor( nTreeID, levelDesc, tree, prc, iProc,    &
    &                                 minLevel, elemPos, foundLevel,           &
    &                                 computeStencil, nesting, updated )
    !---------------------------------------------------------------------------
    !> neighboring treeID
    integer(kind=long_k), intent(in) :: nTreeID
    !>
    integer, intent(in) :: minLevel
    !>
    type(tem_levelDesc_type), intent(inout) :: levelDesc(minLevel:)
    !> process
    type(tem_comm_env_type), intent(in) :: prc
    !>  requested process number
    integer, intent(in) :: iProc
    !> tree information
    type(treelmesh_type), intent(in) :: tree
    !>
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !>
    integer, intent(out) :: elemPos
    !>
    integer, intent(out) :: foundLevel
    !>
    integer, intent(in) :: nesting
    !> was the element updated in this call?
    logical, intent(out) :: updated
    !---------------------------------------------------------------------------
    type(tem_stencilElement_type) :: emptyStencil(1)
    integer :: neighLevel
    logical :: wasAdded
    !---------------------------------------------------------------------------
    neighLevel = tem_LevelOf( nTreeID ) ! Has to be same as tLevel!?
    if(( neighLevel .lt. minLevel )  &
      &                      .or. (neighLevel .gt. uBound(levelDesc,1)) ) then
      call tem_write( ' ERROR: level which is not included in the fluid'//   &
        &             ' tree was demanded in singleProcNeigh' )
      call tem_write('          treeID', nTreeID )
      call tem_abort()
    end if
    ! Set the element updated flag as a default to false
    updated = .false.

    ! If it is a remote cell on only one process -> regular halo
    if( iProc .ne. prc%rank + 1 )  then
      ! REMOTE
      call init( me = emptyStencil(1), QQN=computeStencil%QQN )
      call append( me          = levelDesc( neighLevel )%elem,                 &
        &          tID         = nTreeID,                                      &
        &          eType       = eT_haloElem,                                  &
        &          property    = 0_long_k,                                     &
        &          sourceProc  = iProc,                                        &
        &          haloNesting = nesting,                                      &
        &          stencil     = emptyStencil,                                 &
        &          pos         = elemPos,                                      &
        &          wasAdded    = wasAdded )

      if( .not. wasAdded ) then
        ! update the nesting to the current one
        if( nesting .lt.  levelDesc( neighLevel )%elem%haloNesting%            &
          &                                                val( elemPos )) then
          levelDesc( neighLevel )%elem%needsUpdate%val( elemPos ) = .true.
          levelDesc( neighLevel )%elem%haloNesting%val( elemPos )              &
            &  = min( levelDesc( neighLevel )%elem%haloNesting%val( elemPos ), &
            &         nesting )
          updated = .true.
        end if
      else
        updated = .true.
      end if
    else
      ! LOCAL
      ! Either a ghost or fluid cell
      call identify_local_element( nTreeID        = nTreeID,                   &
        &                          levelDesc      = levelDesc,                 &
        &                          tree           = tree,                      &
        &                          elemPos        = elemPos,                   &
        &                          level          = foundLevel,                &
        &                          minLevel       = minLevel,                  &
        &                          nesting        = nesting,                   &
        &                          updated        = updated,                   &
        &                          computeStencil = computeStencil )
    end if

  end subroutine single_process_neighbor
!******************************************************************************!


!******************************************************************************!
  !> Determine if the requested (local) nTreeID is fluid or ghost.
  !! If fluid: do nothing, as it will be added later on anyway (or already is)
  !!    ghostFromFiner (coarser than requested):
  !!        add all virtual children, i.e. all levels between requested treeID
  !!        and found one.
  !!        Remember: one level difference allowed only.
  !!    ghostFromCoarser (finer than requested):
  !!    not existing( localPos=0): add to halo (MH: why to %halo and not
  !!    halosPrc(iProc)%mine ?)
  !!
  subroutine identify_local_element( nTreeID, levelDesc, tree, minLevel, level,&
    &                                elemPos, nesting, updated, computeStencil )
    !---------------------------------------------------------------------------
    !> neighboring treeID
    integer(kind=long_k), intent(in) :: nTreeID
    !>
    integer, intent(in) :: minLevel
    !>
    type(tem_levelDesc_type), intent(inout) :: levelDesc(minLevel:)
    !>
    type(treelmesh_type), intent(in) :: tree
    !>
    integer, intent(in) :: nesting
    !>
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !>
    integer, intent(out) :: elemPos
    !>
    integer, intent(out) :: level
    !> was the element updated in this call?
    logical, intent(out) :: updated
    !---------------------------------------------------------------------------
    integer :: localPos, neighLevel, dPos, foundLevel
    integer(kind=long_k) :: foundTID
    type(tem_path_type) :: neighPath
    !---------------------------------------------------------------------------
    ! Set the element updated flag as a default to false
    updated = .false.

    ! Position of neighbor treeID
    neighLevel = tem_LevelOf( nTreeID )
    neighPath  = tem_PathOf(nTreeID)

    ! Return position in the treeIDlist
    localPos = tem_PosOfPath( neighPath, tree%pathList )
    ! By localPos we can determine how the element (might) exist locally:
    ! - fluid
    ! - ghostFromCoarser
    ! - ghostFromFiner
    if( localPos .gt. 0) then
      ! ghostFromCoarser or fluid
      foundTID = tree%treeID( localPos )

      foundLevel = tem_LevelOf( foundTID )
      if( foundLevel .eq. neighLevel ) then
        ! Fluid is already part of the elem list. Adding not necessary.
        updated = .false.
      elseif( foundLevel .lt. neighLevel ) then
        ! ghostFromCoarser
        ! existing fluid is coarser than current element
        ! Add all the children of found treeID down to me (intermediate levels)
        ! LBM specific. \todo hand over eligible child
        ! is added ghostFromCoarser
        call add_all_virtual_children(                                         &
          &                sourceID       = foundTID,                          &
          &                foundPos       = localPos,                          &
          &                elemPath       = neighPath,                         &
          &                sourceProperty = tree%ElemPropertyBits( localPos ), &
          &                sourceLevel    = foundLevel,                        &
          &                targetLevel    = neighLevel,                        &
          &                levelDesc      = levelDesc,                         &
          &                minlevel       = minLevel,                          &
          &                nesting        = nesting,                           &
          &                updated        = updated,                           &
          &                tree           = tree,                              &
          &                computeStencil = computeStencil )
      end if ! on same level?
    elseif( localPos .lt. 0) then
      ! ghostFromFiner
      ! Find all existing fluid cells within requested nTreeID position
      call fill_to_real_children( reqTreeID      = nTreeID,                    &
        &                         levelDesc      = levelDesc,                  &
        &                         minLevel       = minlevel,                   &
        &                         tree           = tree,                       &
        &                         foundPos       = dPos,                       &
        &                         updated        = updated,                    &
        &                         computeStencil = computeStencil )
    else
      write(*,*) 'Warning: element not existing',nTreeID,                      &
        &        'adding to nonexisting ...'
      call tem_tIDinfo( me = nTreeID, tree = tree )
      ! This case occurs, when a remote halo was added, which was not existing
      ! before.
      ! Halos are added in the transfer_treeIDs routine
      call append( me         = levelDesc( neighLevel )%elem,                  &
        &          tID        = nTreeID,                                       &
        &          eType      = eT_nonExistingElem,                            &
        &          sourceProc = tree%global%myPart+1,                          &
        &          pos        = dPos )

    endif ! localPos gt 0? coarser ghost or fluid?
    elemPos = PositionOfVal( me  = levelDesc( neighLevel )%elem%tID,           &
      &                      val = nTreeID )
    level = neighLevel

  end subroutine identify_local_element
!******************************************************************************!


!******************************************************************************!
  !> Find all the virtual children of the sourceTreeID down to the targetLevel
  !! and add to the level-wise ghostFromCoarser list in the level descriptor
  !!
  recursive subroutine add_all_virtual_children(                               &
    &       sourceID, sourceProperty, foundPos, elemPath, sourceLevel,         &
    &       targetLevel, levelDesc, minLevel, tree, computeStencil,            &
    &       nesting, updated )
    !---------------------------------------------------------------------------
    !>
    integer(kind=long_k), intent(in) :: sourceID
    !>
    integer(kind=long_k), intent(in) :: sourceProperty
    !>
    type(tem_path_type), intent(in)  :: elemPath
    !>
    integer,intent(in) ::  nesting
    !>
    integer,intent(in) ::  foundPos
    !>
    integer,intent(in) ::  sourceLevel
    !>
    integer,intent(in) ::  targetLevel
    !>
    integer, intent(in) :: minLevel
    !>
    type(tem_levelDesc_type), intent(inout) :: levelDesc(minLevel:)
    !>
    type(treelmesh_type), intent(in) :: tree
    !>
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !> was the element updated in this call?
    logical, intent(out) :: updated
    !---------------------------------------------------------------------------
    integer :: targetPos, iChild, iDir, nVals
    ! position of the existing (source) tID in the elem list
    integer :: sourcePos, levelPos 
    logical :: wasAdded
    logical :: childUpdated
    integer(kind=long_k) :: cTreeID     ! current treeID to identify
    integer(kind=long_k) :: childID     ! child treeID
    integer(kind=long_k) :: curNeighborID, coarserNeighborID
    integer(kind=long_k), allocatable :: tNeighID(:)
    type(tem_stencilElement_type) :: tStencil(1)
    integer :: x(4), curLevel, offset(4), xc(4), xElem(4), addedPos
    !---------------------------------------------------------------------------
    !Position of the coarser source element
    sourcePos = PositionOfVal( me  = levelDesc( sourceLevel )%elem%tID,        &
      &                        val = sourceID )
    allocate( tNeighID( computeStencil%QQN ))
    offset(4)   = 0
    ! By default, set that no element was updated
    updated = .false.
    childUpdated = .false.

    if( sourceLevel .lt. targetLevel ) then
      !children = tem_directChildren( sourceID )
      curLevel = sourceLevel + 1
      call init( me = tStencil(1), QQN = computeStencil%QQN, headerPos = 1 )
      !do iChild = 1, 8
        ! Add to the level-wise ghost list
        levelPos = targetLevel - sourceLevel
        cTreeID = elemPath%node( levelPos )
        ! determine which child it is
        iChild = tem_childNumber( cTreeID )
        xElem = tem_coordOfId( cTreeID ) !children( iChild ))
        call append( me          = levelDesc( curLevel )%elem,                 &
          &          tID         = cTreeID,                                    &
          &          property    = sourceProperty,                             &
          &          eType       = eT_ghostFromCoarserElem,                    &
          &          sourceProc  = tree%global%myPart+1,                       &
          &          stencil     = tStencil,                                   &
          &          pos         = targetPos,                                  &
          &          haloNesting = nesting,                                    &
          &          wasAdded    = wasAdded )

        if( wasAdded ) then
          updated = .true.
          ! inherit boundaries and neighbors from parent
          x(:) = tem_coordOfId( int(iChild, kind=long_k) )
          ! inherit the boundary infos from the parent
          tNeighID = 0_long_k
          do iDir = 1, 3
            call tem_find_BCs_fromCoarser( dir            = iDir,              &
              &                            childCoord     = x,                 &
              &                            sourceLevel    = sourceLevel,       &
              &                            sourcePos      = sourcePos,         &
              &                            neighID        = tNeighID,          &
              &                            minLevel       = minLevel,          &
              &                            levelDesc      = levelDesc,         &
              &                            computeStencil = computeStencil )
          enddo
          !
          do iDir = 1, computeStencil%QQN
            offset(1:3) = computeStencil%cxDir(1:3, iDir )
            !neighIDpos = levelDesc( curLevel )%elem%val( targetPos )%stencil%val(1)%tIDpos( iDir )
            !coarserNeighborID = levelDesc( curLevel )%elem%val( targetPos )%neighID%val( neighIDpos )
            coarserNeighborID = tNeighID( iDir )
            if( coarserNeighborID .gt. 0_long_k ) then
              ! find the corresponding children of the neighbor defined for my
              ! parent. Find the child in the corresponding direction
              xc(4) = 1
              ! get the corresponding neighbor in the offset direction, which
              ! is a child of the coarser neighbor element. Use +2 offset to get
              ! only positive numbers
              ! (child coordinates always range between {0..1})
              xc(1:3) = mod( x(1:3) + 2 + offset(1:3), 2 )
              ! get the child number of the corresponding neighbor within the
              ! coarser element. (is {1...8})
              childID  = tem_IdOfCoord( xc )
              ! calculate the child treeID from the coarser neighbor and the
              ! childID
              curNeighborID = coarserNeighborID*8_long_k + childID

            elseif( coarserNeighborID .eq. 0_long_k ) then
              ! If the neighbor is still 0, it must be a direct sibling,
              ! so we can directly compute its tID.
              ! find the neighbor according to stencil offset from me
              curNeighborID = tem_IdOfCoord( xElem + offset )
            else
              curNeighborID = coarserNeighborID
            endif
            ! append the neighbor ID ...
            call append( me  = levelDesc( curLevel )%elem%neighID              &
              &                                    %val( targetPos ),          &
              &          val = curNeighborID,                                  &
              &          pos = addedPos )
            ! ... and store this position in the stencil
            levelDesc( curLevel )%elem%stencil%val( targetPos )                &
              &                               %val(1)%tIDpos( iDir )           &
              &  = addedPos
          enddo
          ! Now all boundary conditions for the target elements are defined.
          nVals = levelDesc( curLevel )%elem%neighID%val( targetPos )%nVals
          if( minval( levelDesc( curLevel )%elem%neighID%val( targetPos )      &
            &          %val( 1: nVals )) .lt. 0 ) then
            ! Update the boundary condition property bit if any boundary
            ! condition is present
            levelDesc( curLevel )%elem%property%val( targetPos )               &
              &  = ibset(levelDesc( curLevel )%elem%property%val(targetPos),   &
              &          prp_hasBnd)
          else
            ! Unset the property bit
            levelDesc( curLevel )%elem%property%val( targetPos )               &
              &  = ibclr(levelDesc( curLevel )%elem%property%val(targetPos),   &
              &          prp_hasBnd)
          end if
        else
          ! Existing element encountered.
          updated = .false.
        endif
        ! Overwrite the eventually existing nesting with the smallest value.
        ! The smallest nesting determines if further neighbors have to be
        ! retrieved in communicate_elements
        if( nesting .lt. levelDesc( curLevel )%elem%haloNesting%               &
          &                                              val( targetPos )) then
          ! needs update
          updated = .true.
          levelDesc( curLevel )%elem%needsUpdate%val( targetPos ) = .true.
          levelDesc( curLevel )%elem%haloNesting%val( targetPos )              &
            &       = min( nesting, levelDesc( curLevel )%elem%haloNesting%    &
            &                                                 val( targetPos ))
        end if
        ! In any case we have to recurse down to the target level
        ! lower levels might not yet exist.
        !MH: What about childUpdated?
        call add_all_virtual_children( sourceID       = cTreeID,               &
          &                            foundPos       = targetPos,             &
          &                            elemPath       = elemPath,              &
          &                            nesting        = nesting,               &
          &                            sourceLevel    = curLevel,              &
          &                            targetLevel    = targetLevel,           &
          &                            levelDesc      = levelDesc,             &
          &                            minLevel       = minLevel,              &
          &                            updated        = childUpdated,          &
          &                            tree           = tree,                  &
          &                            sourceProperty = sourceProperty,        &
          &                            computeStencil = computeStencil )
      !end do
    end if
    updated =( updated .or. childUpdated )

  end subroutine add_all_virtual_children
!******************************************************************************!


!******************************************************************************!
  !> Inherit the neighborhood from the sourceELem to the targetElem
  !! Note that targetElem is inout, as it might have already values assigned.
  !!
  subroutine tem_find_BCs_fromFiner( childPos, sourceLevel, targetLevel,       &
    &                                targetPos, levelDesc, minLevel,           &
    &                                computeStencil )
    !---------------------------------------------------------------------------
    !>
    integer, intent(in) :: childPos(8)
    !>
    integer, intent(in) :: sourceLevel
    !>
    integer, intent(in) :: targetLevel
    !>
    integer, intent(in) :: targetPos
    !>
    integer, intent(in) :: minLevel
    !>
    type(tem_levelDesc_type ) :: levelDesc(minLevel:)
    !>
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !---------------------------------------------------------------------------
    integer :: dir
    ! Tangential direction iterators
    integer :: iDirX, iDirY, iDir
    integer :: dirX !< first tangential direction
    integer :: dirY !< second tangential direction
    integer :: iStencilElem  !< stencil element iterator
    integer :: childCoord(4)
    ! the type of the current stencil link (side, edge, corner)
    integer :: linktype
    integer :: iDirNormal
    integer :: iStencil, addedPos
    type(tem_stencilElement_type) :: tStencil
    integer(kind=long_k) :: tNeighID
    !---------------------------------------------------------------------------
    if  ( .not. allocated( levelDesc( targetLevel )%elem%stencil%              &
      &                                              val(targetPos)%val )) then
      call init( me     = levelDesc( targetLevel )%elem%stencil%val(targetPos),&
        &        length = 1 )
    endif
    call init( me = tStencil, QQN = computeStencil%QQN, headerPos = 1 )
    call append( me  = levelDesc( targetLevel )%elem%stencil%val(targetPos),   &
      &          val = tStencil )
    if( levelDesc( targetLevel )%elem%neighID%val(targetPos)%nVals .gt. 0 ) then
      write(dbgUnit(2),*) 'Warning: tem_find_BCs_fromFiner has already' &
        &                       //' encountered'
      write(dbgUnit(2),*) '         existing neighbor IDs for the stencil'
    end if
    iStencil = levelDesc( targetLevel )%elem%stencil%val( targetPos )%nVals

    childCoord(4) = 1
    do iStencilElem = 1, computeStencil%QQN
      ! Reset current direction
      tNeighID = -huge( 0_long_k )
      ! Count the non-zero entries to decide the kind of the current
      ! link direction in the stencil. (only direct neighbors)
      ! Distinguishing three different kinds of neighbors:
      ! SIDE, EDGE and CORNER
      linktype = sum( computeStencil%cxDir( :, iStencilElem )**2)
      select case(linktype)
      case(1)
        ! 1 non-zero
        ! SIDE, needs four children
        ! identify the normal direction of the side
        dir = maxloc( abs( computeStencil%cxDir(:, iStencilElem )), 1)
        ! Convert the stencil direction to child direction
        iDirNormal = stencilToChild(computeStencil%cxDir(dir, iStencilElem ))
        childCoord( dir ) = iDirNormal
        ! identify the tangential directions
        dirX = mod( dir, 3) + 1
        dirY = mod( dir+1, 3) + 1
        do iDirY = 0, 1
          childCoord( dirY ) = iDirY
          do iDirX = 0, 1
            childCoord( dirX ) = iDirX
            call update_childNeighborID( neighID      = tNeighID,              &
              &                          childCoord   = childCoord,            &
              &                          childPos     = childPos,              &
              &                          iStencil     = iStencil,              &
              &                          iStencilElem = iStencilElem,          &
              &                          levelDesc    = levelDesc(sourceLevel))
          end do
        end do

      case(2)
        ! 2 non-zeroes
        ! EDGE, needs two children
        ! Determine the 0-direction (dimension with the zero entry)
        dir = minloc( abs( computeStencil%cxDir(:,iStencilElem )), 1)
        ! Get the edge directions
        dirX = mod( dir, 3) + 1
        dirY = mod( dir+1, 3) + 1
        childCoord( dirX ) = stencilToChild(computeStencil                     &
          &                                 %cxDir(dirX,iStencilElem))
        childCoord( dirY ) = stencilToChild(computeStencil                     &
          &                                 %cxDir(dirY,iStencilElem))
        do iDir = 0, 1
          childCoord( dir ) = iDir
          call update_childNeighborID( neighID      = tNeighID,                &
            &                          childCoord   = childCoord,              &
            &                          childPos     = childPos,                &
            &                          iStencil     = iStencil,                &
            &                          iStencilElem = iStencilElem,            &
            &                          levelDesc    = levelDesc( sourceLevel ))
        enddo

      case(3)
        ! No zero at all, all three directions have a offset
        ! CORNER, just a single child is connected to this link
        do iDir = 1, 3
          childCoord( iDir ) = stencilToChild( computeStencil                  &
            &                                  %cxDir(iDir,iStencilElem) )
        enddo
        call update_childNeighborID( neighID      = tNeighID,                  &
          &                          childCoord   = childCoord,                &
          &                          childPos     = childPos,                  &
          &                          iStencil     = iStencil,                  &
          &                          iStencilElem = iStencilElem,              &
          &                          levelDesc    = levelDesc( sourceLevel ))
      end select
      ! Append the required treeID
      call append( me  = levelDesc( targetLevel )%elem%neighID%val(targetPos), &
        &          val = tNeighID,                                             &
        &          pos = addedPos )
      levelDesc( targetLevel )%elem%stencil%val(targetPos)                     &
        &                                  %val(iStencil)%tIDpos(iStencilElem) &
        &  = addedPos

    end do

  end subroutine tem_find_BCs_fromFiner
!******************************************************************************!


!******************************************************************************!
  !> Inherit the neighborhood from the sourceELem to the targetElem
  !! Note that targetElem is inout, as it might have already values assigned.
  !!
  subroutine tem_find_BCs_fromCoarser( dir, childCoord, sourceLevel,           &
    &                                  sourcePos, neighID, computeStencil,     &
    &                                  levelDesc, minLevel )
    !---------------------------------------------------------------------------
    !>
    integer, intent(in) :: sourceLevel
    !>
    integer, intent(in) :: sourcePos
    !>
    integer, intent(in) :: dir
    !>
    integer, intent(in) :: minLevel
    !>
    integer(kind=long_k) :: neighID(:)
    !>
    integer, intent(in) :: childCoord(4)
    !>
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !>
    type(tem_levelDesc_type), intent(in) :: levelDesc(minLevel:)
    !---------------------------------------------------------------------------
    ! Tangential direction iterators
    integer :: iDirX, iDirY
    integer :: dirX ! first tangential direction
    integer :: dirY ! second tangential direction
    integer :: curDir ! child face normal direction, always {-1,1}
    integer :: myLink(4) ! child link direction
    integer :: parentLink(3) ! each entry must be either {-1,0,1}
    integer :: iChildStencil ! child stencil element iterator
    integer :: iStencilElem  ! stencil element iterator
    integer :: iStencil
    !---------------------------------------------------------------------------

    curDir = childToStencil( childCoord( dir ))
    myLink( dir ) = curDir
    myLink(4) = 0
    parentLink(dir) = curDir
    ! identify the tangential directions
    dirX = mod( dir, 3) + 1
    dirY = mod( dir+1, 3) + 1
    iStencil = 1

    ! Iterate over all nine child links of the surface with the normal pointing
    ! into dir
    do iDirY = -1, 1
      myLink( dirY ) = iDirY
      ! select the valid directions to look for the neighbors in the second
      ! tangential direction
      if( childCoord( dirY ) == 0) then
        parentLink( dirY ) = min( iDirY, 0 )
      else
        parentLink( dirY ) = max( iDirY, 0 )
      end if
      do iDirX = -1, 1
        myLink( dirX ) = iDirX
        if( childCoord( dirX ) == 0) then
          parentLink( dirX ) = min( iDirX, 0 )
        else
          parentLink( dirX ) = max( iDirX, 0 )
        end if
        iChildStencil = 0
        do iStencilElem = 1, computeStencil%QQN
          if  (      computeStencil%cxDir( 1, iStencilElem ) .eq. myLink( 1 )  &
            &  .and. computeStencil%cxDir( 2, iStencilElem ) .eq. myLink( 2 )  &
            &  .and. computeStencil%cxDir( 3, iStencilElem ) .eq. myLink( 3 )  &
            & ) then
            ! Found matching stencil entry for current child direction
            iChildStencil = iStencilElem
          endif
        enddo
        if( iChildStencil .gt. 0 ) then
          do iStencilElem = 1, computeStencil%QQN
            if (      computeStencil%cxDir(1, iStencilElem) .eq. parentLink(1) &
              & .and. computeStencil%cxDir(2, iStencilElem) .eq. parentLink(2) &
              & .and. computeStencil%cxDir(3, iStencilElem) .eq. parentLink(3) &
              &  ) then
              ! Found matching stencil entry for current child direction
              ! Set the parent's neighbor here. Later we replace the
              ! parent's neighbor by the current level's neighbor
              !targetElem%neighID%val( targetElem%stencil%val(iStencil)%tIDpos( iChildStencil )) =  &
              neighID( iChildStencil )                                         &
                &  = levelDesc( sourceLevel )                                  &
                &    %elem%neighID%val(sourcePos)                              &
                &                 %val( levelDesc( sourceLevel )               &
                &                       %elem%stencil%val(sourcePos)           &
                &                                    %val(iStencil)            &
                &                                    %tIDpos(iStencilElem)&    
                &                     )
!                sourceElem%neighID%val( sourceElem%stencil%val(iStencil)%tIDpos( iStencilElem ))
            end if
          end do
        end if
      end do  
    end do  

  end subroutine tem_find_BCs_fromCoarser
!******************************************************************************!


!******************************************************************************!
  !> Find all the children of the requested reqTreeID
  !! add to the level-wise ghostFromFine list in the level descriptor
  !!
  recursive subroutine fill_to_real_children( reqTreeID, levelDesc,            &
    &                                         minLevel, tree, updated,         &
    &                                         foundPos, computeStencil )
    !---------------------------------------------------------------------------
    !> requested treeID
    integer(kind=long_k), intent(in) :: reqTreeID
    !>
    integer, intent(in) :: minLevel
    !>
    type(tem_levelDesc_type), intent(inout) :: levelDesc(minLevel:)
    !>
    type(treelmesh_type), intent(in) :: tree
    !>
    integer, intent(out) :: foundPos
    !> was the current element updated in this call?
    logical, intent(out) :: updated
    !>
    type( tem_stencilHeader_type ), intent(in) :: computeStencil
    !---------------------------------------------------------------------------
    integer :: iChild, addedPos, reqLevel
    integer(kind=long_k) :: children(8), property
    logical :: wasAdded, childUpdated
    integer :: childPos(8), neighLevel
    type(tem_path_type) :: neighPath
    !---------------------------------------------------------------------------
    ! Set as not updated by default
    updated = .false.

    ! Create the ghostFromFiner
    reqLevel        = tem_LevelOf( reqTreeID )
    call append( me         = levelDesc( reqLevel )%elem,                      &
      &          tID        = reqTreeID,                                       &
      &          eType      = eT_ghostFromFinerElem,                           &
      &          property   = 0_long_k,                                        &
      &          sourceProc = tree%global%myPart+1,                            &
      &          pos        = addedPos,                                        &
      &          wasAdded   = wasAdded )

    if( wasAdded ) then
      updated = .true.
      children = tem_directChildren( reqTreeID )
      childPos = 0 ! reset child positions. non-existing children are 0
      ! reset property
      property = 0_long_k
      do iChild = 1, 8
        ! What's the configuration of the requested element
      !  childPos( iChild ) = PositionOfVal( levelDesc( reqLevel + 1 )%elem%tID,&
      !                                      children(iChild))
        neighLevel = reqLevel + 1
        neighPath  = tem_PathOf( children( iChild ))

        ! Return position in the treeIDlist
        childPos( iChild ) = tem_PosOfPath( neighPath, tree%pathList )

        if( childPos( iChild ) .lt. 0 ) then
           ! finer elements! add them.
           call fill_to_real_children( reqTreeID      = children( iChild ),    &
             &                         levelDesc      = levelDesc,             &
             &                         minLevel       = minlevel,              &
             &                         tree           = tree,                  &
             &                         updated        = childUpdated,          &
             &                         computeStencil = computeStencil,        &
             &                         foundPos       = childPos( iChild ))
           ! Unify all properties of the children
           property = ieor( property,                                          &
           &   levelDesc( reqLevel + 1)%elem%property%val( childPos( iChild )))
           updated = ( updated .or. childUpdated )
        else
          ! Fluid cells are already in the element list (Not existing cells do
          ! not have to be added). If the element is a fluid, we have to get its
          ! position in the levelwise element list of the level descriptor.
          childPos( iChild ) = PositionOfVal( levelDesc( reqLevel + 1 )%elem%  &
            &                                            tID, children(iChild))
        end if
      end do

      ! Now reconstruct current element's neighborhood
      ! based on the children's information
      call tem_find_BCs_fromFiner( childPos       = childPos,                  &
        &                          sourceLevel    = reqLevel + 1,              &
        &                          targetLevel    = reqLevel,                  &
        &                          targetPos      = addedPos,                  &
        &                          levelDesc      = levelDesc,                 &
        &                          minLevel       = tree%global%minLevel,      &
        &                          computeStencil = computeStencil )

    else
      ! ghostFromFiner element was not added
      ! Hence, we found available information which now has to be returned
      updated = .false.
    end if

    ! Return the position of where the element was encountered
    foundPos = addedPos

  end subroutine fill_to_real_children
!******************************************************************************!


!******************************************************************************!
  !> Assemble the fluid list and identify neighbor relations
  !! start with building up the ghost and halo element collection as well
  !!
  subroutine build_levelElements( levelDesc, levelPointer, tree, prc,          &
    &                             computeStencil, pathFirst, pathLast,         &
    &                             dumpLists )
    !---------------------------------------------------------------------------
    !>
    type(treelmesh_type), intent(inout) :: tree
    !>
    type(tem_levelDesc_type), intent(inout) :: levelDesc( tree%global%minLevel:)
    !>
    integer, intent(inout) :: levelPointer(:)
    !>
    type(tem_comm_env_type), intent(inout) :: prc
    !>
    type(tem_stencilHeader_type) :: computeStencil(:)
    !>
    type(tem_path_type), intent(in) :: pathFirst(:), pathLast(:)
    !>
    logical, intent(in) :: dumpLists
    !--------------------------------------------------------------------------=
    integer(kind=long_k) :: nTreeID   ! neighboring treeID
    integer(kind=long_k) :: elemTID   ! neighboring treeID
    integer :: nElems, nProcs, elemPos, elemPos_new
    integer :: iElem, iNeighElem, iLevel, iStencil
    integer :: tLevel
    ! Nesting level to make sure all neighbors of first row of ghost elements
    ! are reconstructed
    integer :: nesting
    ! position of where to read the stencil neighbor treeid in the element
    integer :: neighPos
    !---------------------------------------------------------------------------
    nElems = size(tree%treeID)
    nProcs = size(tree%Part_First)

    ! Reset the has
    hash = -1_long_k
    hash_elemPos = -1

    write(logUnit(2),*) 'Building level-wise fluid list ...'

    ! Create levelpointer: create the pointer from the original
    ! treeID list to the level-wise lists
    do iElem = 1, nElems
      tLevel = tem_LevelOf( tree%treeID( iElem ) )
      elemTID = tree%treeID( iElem )
      levelPointer( iElem ) = PositionOfVal( me  = levelDesc(tLevel)%elem%tID, &
        &                                    val = elemTID )
    end do

    do iLevel = tree%global%minLevel, tree%global%maxLevel
      levelDesc( iLevel )%nElems_fluid = levelDesc( iLevel )%elem%tID%nVals
    end do
    ! Now iterate over all the neighbor lists and add the required elements if
    ! necessary
    do iLevel = tree%global%minLevel, tree%global%maxLevel
      do iElem = 1, levelDesc( iLevel )%nElems_fluid
        do iStencil = 1, levelDesc( iLevel )%elem%stencil%val(iElem )%nVals
          do iNeighElem = 1, levelDesc( iLevel )%elem                          &
            &                         %stencil%val( iElem )%val( iStencil )%QQN
            !MH: Same as in Line 2092 (identify_additionalNeigh)
            neighPos =levelDesc( iLevel )%elem%stencil%val( iElem )            &
              &                           %val( iStencil )%tIDpos( iNeighElem ) 
            if( neighPos .gt. 0 ) then
              nTreeID = levelDesc( iLevel )%elem%neighID%val( iElem )%         &
                &                                               val( neighPos )
              nesting = 0
              call identify_elements( nTreeID        = nTreeID,                &
                &                     tree           = tree,                   &
                &                     pathFirst      = pathFirst,              &
                &                     pathLast       = pathLast,               &
                &                     levelDesc      = levelDesc,              &
                &                     elemPos        = elemPos,                &
                &                     prc            = prc,                    &
                &                     nesting        = nesting,                &
                &                     computeStencil = computeStencil(1) )
            else
              elemPos = 0 
            end if
            levelDesc( iLevel )%elem%stencil%val( iElem )%val( iStencil )      &
              &                               %totalPos( iNeighElem ) = elemPos
          end do ! iNeighElem!
        end do ! iStencil
      end do ! iElem
    end do ! iLevel

    ! Find neighbors of neighbors from the require list
    ! the require list
    ! MH: What exactly is the require list for?
    do iLevel = tree%global%minLevel, tree%global%maxLevel
      do iElem = 1, levelDesc( iLevel )%require%nVals
        iStencil = 1
        elemPos = PositionOfVal( me  = levelDesc( iLevel )%elem%tID,           &
          &                      val = levelDesc( iLevel )%require%val( iElem ))
        if( elemPos .gt. 0 ) then
         if( levelDesc(iLevel )%elem%eType%val( elemPos ) .gt. 0) then
           do iNeighElem = 1, levelDesc( iLevel )%elem                         &
             &                      %stencil%val( elemPos )%val( iStencil )%QQN
              neighPos = levelDesc( iLevel )%elem%stencil%val( elemPos )       &
                &                         %val( iStencil )%tIDpos( iNeighElem ) 
              if( neighPos .gt. 0 ) then
                nTreeID = levelDesc( iLevel )%elem%neighID%val( elemPos )%     &
                  &                                             val( neighPos )
                nesting = 0
                call identify_elements( nTreeID        = nTreeID,              &
                  &                     tree           = tree,                 &
                  &                     pathFirst      = pathFirst,            &
                  &                     pathLast       = pathLast,             &
                  &                     levelDesc      = levelDesc,            &
                  &                     elemPos        = elemPos_new,          &
                  &                     prc            = prc,                  &
                  &                     nesting        = nesting,              &
                  &                     computeStencil = computeStencil(1) )
              else
                elemPos_new = 0 
              end if
              levelDesc( iLevel )%elem%stencil%val( elemPos )%val( iStencil )  &
                &                         %totalPos( iNeighElem ) = elemPos_new
           end do
          end if
        end if
      end do
    end do ! iLevel

    do iLevel = tree%global%minLevel, tree%global%maxLevel
      if (dumpLists) then
        call tem_elemList_dump( me      = levelDesc( iLevel )%elem, &
          &                     nUnit   = dbgUnit(0),               &
          &                     stencil = .true.,                   &
          &                     string  = 'after build level elm'   )
      end if
    end do ! iLevel

  end subroutine build_levelElements
!******************************************************************************!


!******************************************************************************!
  !> Return the element type of a treeID by defining the position in the
  !! totalList
  !!
  function tem_eTypeOfTotalPos( pos, levelDesc ) result( eType )
    !---------------------------------------------------------------------------
    !> the element you are looking for
    integer, intent(in) :: pos
    !> the descriptor you use for searching
    type(tem_levelDesc_type), intent(in) :: levelDesc
    !>
    integer :: eType
    !---------------------------------------------------------------------------
    integer :: levelPos
    !---------------------------------------------------------------------------
    eType   = eT_undefinedElem
    levelPos = PositionOfVal( me  = levelDesc%elem%tID,                        &
      &                       val = levelDesc%total( pos ))
    if( levelPos .gt. 0 ) then
      eType = levelDesc%elem%eType%val( levelPos )
    end if

  end function tem_eTypeOfTotalPos
!******************************************************************************!


!******************************************************************************!
  !> Return the element type of a treeID .
  !!
  function tem_eTypeOfId( tID, levelDesc ) result( eType )
    !---------------------------------------------------------------------------
    !> the element you are looking for
    integer(kind=long_k), intent(in) :: tID
    !> the descriptor you use for searching
    type(tem_levelDesc_type), intent(in) :: levelDesc
    !>
    integer :: eType
    !---------------------------------------------------------------------------
    integer :: levelPos
    !---------------------------------------------------------------------------
    eType   = eT_undefinedElem
    levelPos = PositionOfVal( me  = levelDesc%elem%tID,                        &
      &                       val = tID )
    if( levelPos .gt. 0 ) then
      eType = levelDesc%elem%eType%val( levelPos )
    end if

  end function tem_eTypeOfId
!******************************************************************************!


!******************************************************************************!
  !> Returns the absolute position in the total list of a given treeID
  !! opposed to PosOfId, where the relative position in one of the separate
  !! lists is returned. Herefore, total list has to be created beforehand.
  !!
  function tem_treeIDinTotal( tID, levelDesc, eType ) result( elemPos )
    !---------------------------------------------------------------------------
    !> the element you are looking for
    integer(kind=long_k), intent(in) :: tID
    !> the descriptor you use for searching
    type(tem_levelDesc_type), intent(in) :: levelDesc
    !>
    integer, intent(in), optional :: eType
    !>
    integer :: elemPos
    !---------------------------------------------------------------------------
    integer :: eType_loc, levelPos
    !---------------------------------------------------------------------------
    if( present( eType )) then
      eType_loc = eType
    else
      levelPos = PositionOfVal( me  = levelDesc%elem%tID,                      &
        &                       val = tID )
      if( levelPos .gt. 0 ) then
        eType_loc = levelDesc%elem%eType%val( levelPos )
      else
        eType_loc = 0
      end if
    end if
    elemPos = 0
    if( eType_loc .gt. 0 ) then
      elemPos = tem_PositionInSorted( me    = levelDesc%total,                 &
        &                             lower = levelDesc%offset(1, eType_loc)+1,&
        &                             upper = levelDesc%offset(2, eType_loc),  &
        &                             val   = tID )
    end if
    elemPos = max( elemPos, 0 )

  end function tem_treeIDinTotal
!******************************************************************************!


!******************************************************************************!
  !> create the intermediate, static list totalPnt, which holds pointers to the 
  !! elem%TID list, but in an ordered fashion. The order is the same as it will
  !! be in the total list later on, i.e.: fluid, ghostFC, ghostFF, halo.
  !! this four sub-lists are within sorted by their treeID.
  !! Additionally, the process-wise collections of halo elements are collected
  !! into halosPrc by grouping the treeIDs according to their belonging process
  !!
  subroutine identify_lists( tree, levelDesc, iLevel)
    !---------------------------------------------------------------------------
    !>
    type(treelmesh_type), intent(in) :: tree
    !>
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minlevel:)
    !>
    integer, intent(in) ::  iLevel
    !---------------------------------------------------------------------------
    integer :: iElem, indElem, iProc
    integer :: dPos ! Position where the object was stored in the dynamic list
    integer :: procPos ! Position where the process is stored in dynamic list
    type( tem_halo_type ) :: tProc ! temporary halo type
    logical :: wasAdded ! was the element added to the list
    integer :: iPnt( eT_minNumber:eT_maxNumber ), eType, iVal
    integer(kind=long_k) :: tID
    !---------------------------------------------------------------------------
    ! Destroy lists
    call destroy( me = levelDesc( iLevel )%halosPrc  )
    ! init lists
    call init( me = levelDesc( iLevel )%halosPrc, unique = .true. )

    ! 1. count nelems
    levelDesc( iLevel )%nElems_fluid = 0
    levelDesc( iLevel )%nElems_ghostFromCoarser = 0
    levelDesc( iLevel )%nElems_ghostFromFiner = 0
    levelDesc( iLevel )%nElems_halo  = 0
    do indElem = 1, levelDesc( iLevel )%elem%tID%nVals
      iElem = levelDesc( iLevel )%elem%tID%sorted( indElem )
      select case( levelDesc( iLevel )%elem%eType%val( iElem ) )
      case( eT_fluidElem )
        levelDesc( iLevel )%nElems_fluid = levelDesc( iLevel )%nElems_fluid + 1
      case( eT_ghostFromCoarserElem )
        levelDesc( iLevel )%nElems_ghostFromCoarser =                          &
          &                     levelDesc( iLevel )%nElems_ghostFromCoarser + 1
      case( eT_ghostFromFinerElem, eT_distributedGhostFromFinerElem )
        levelDesc( iLevel )%nElems_ghostFromFiner =                            &
          &                       levelDesc( iLevel )%nElems_ghostFromFiner + 1
      case( eT_haloElem )
        levelDesc( iLevel )%nElems_halo = levelDesc( iLevel )%nElems_halo + 1
      end select     
    end do
    levelDesc( iLevel )%nElems =                                               &
      &                        + levelDesc( iLevel )%nElems_fluid              &
      &                        + levelDesc( iLevel )%nElems_ghostFromCoarser   &
      &                        + levelDesc( iLevel )%nElems_ghostFromFiner     &
      &                        + levelDesc( iLevel )%nElems_halo
    if( allocated( levelDesc( iLevel )%totalPnt ))                             &
      &     deallocate( levelDesc( iLevel)%totalPnt )
    allocate( levelDesc( iLevel )%totalPnt( levelDesc( iLevel )%nElems ))

    call set_offsets( me       = levelDesc( iLevel )%offset(:,:),              &
      &               nFluids  = levelDesc( iLevel )%nElems_fluid,             &
      &               nGhostFC = levelDesc( iLevel )%nElems_ghostFromCoarser,  &
      &               nGhostFF = levelDesc( iLevel )%nElems_ghostFromFiner,    &
      &               nHalos   = levelDesc( iLevel )%nElems_halo )

    ! Reset pointers to current eType element in the total list
    iPnt = 0
    levelDesc( iLevel )%totalPnt = -1
    do indElem = 1, levelDesc( iLevel )%elem%tID%nVals
      ! Access sorted list to maintain locality of space-filling curve order
      iElem = levelDesc( iLevel )%elem%tID%sorted( indElem )
      tID = levelDesc( iLevel )%elem%tID%val( iElem )
      eType = levelDesc( iLevel )%elem%eType%val( iElem ) 
      if ( eType .eq. eT_distributedGhostFromFinerElem ) then
        eType =eT_ghostFromFinerElem
      end if
      ! increase counter for current element
      iPnt( eType ) = iPnt( eType ) + 1
      ! is the eType a required element (fluid, ghost or halo)?
      ! Get rid of nonExistent elements here.
      if( eType .ge. eT_minNumRelevant) then
        ! Add the pointer 
        levelDesc( iLevel )%totalPnt(                                          &
          & levelDesc( iLevel )%offset(1,eType) + iPnt( eType ) ) = iElem
        ! And add the halosPrc for halo elements
        if ( eType .eq. eT_haloElem ) then
          ! get the process from where to get the element
          iProc = levelDesc( iLevel )%elem%sourceProc%val( iElem )
          ! Create an entry in the process list for the current source process
          tProc%partnerProc = iProc
          call append( me  = levelDesc( iLevel )%halosPrc,                     &
            &          val = tProc, pos = procPos, wasAdded = wasAdded )
          ! If the source process was not in the list before, I need to
          ! initialize the element list inside the current source process entry
          if( wasAdded ) then
            call init( me  = levelDesc( iLevel )%halosPrc%val( procPos )%mine, &
              &        unique = .true. )
          end if
          ! ... and add the current halo element to the respective process entry
          call append( me  = levelDesc( iLevel )%halosPrc%val( procPos )%mine, &
            &          val = iElem,                                            &
            &          pos = dPos )
        end if
      end if
    end do
    
    ! Security check
    ! Check if there are no entries < 1
    do iElem = 1, levelDesc( iLevel )%nElems
      if( levelDesc( iLevel )%totalPnt( iElem ) .lt. 1 ) then
        write(dbgUnit(1),*)"Error: Found index < 1 in the totalPnt array."
        write(dbgUnit(1),*)'Abort!'
        write(dbgUnit(1),*)'offset: ' // trim(tem_toStr(levelDesc(iLevel) &
          &                                            %offset(1,:),     &
          &                                            '; '))            
        do iVal = 1, levelDesc( iLevel )%nElems
          write(dbgUnit(1),*)'totalPnt: ' &
            &                // trim(tem_toStr(levelDesc(iLevel) &
            &                                  %totalPnt(iVal))) 
        end do
        call tem_abort
      end if
    enddo

  end subroutine identify_lists
!******************************************************************************!


!******************************************************************************!
  !> Create the level-wise total lists. they consist of
  !! fluid elements + ghost elements + halo elements
  !! and are sorted by the treeIDs
  !!
  subroutine truncate_lists( levelDesc, minlevel )
    !---------------------------------------------------------------------------
    !>
    integer, intent(in) :: minlevel
    !>
    type(tem_levelDesc_type), intent(inout) :: levelDesc(minlevel:)
    !---------------------------------------------------------------------------
    integer :: iLevel
    !---------------------------------------------------------------------------
    do iLevel = minLevel, ubound( levelDesc, 1 )
      call truncate( me = levelDesc( iLevel )%elem )
    end do 
  end subroutine truncate_lists
!******************************************************************************!


!******************************************************************************!
  !>  Set the offsets
  !!
  subroutine set_offsets( me, nFluids, nGhostFC, nGhostFF, nHalos )
    !---------------------------------------------------------------------------
    !>
    integer, intent(out) :: me( 2, eT_minNumber:eT_maxNumber)
    !>
    integer, intent(in) :: nFluids, nGhostFC, nGhostFF, nHalos
    !---------------------------------------------------------------------------
    integer ::  nTotal
    !---------------------------------------------------------------------------
    nTotal = nFluids + nGhostFC+ nGhostFF+ nHalos
    ! Offset for accessing totallist, invsorted etc. arrays for
    !! fluids, ghosts and halos
    ! store the offset for the beginning ...
    me( 1, eT_fluidElem ) = 0
    me( 1, eT_ghostFromCoarserElem )  = nFluids
    me( 1, eT_ghostFromFinerElem )  = nFluids + nGhostFC
    me( 1, eT_haloElem )  = nFluids + nGhostFC + nGhostFF
    ! ... and the end of the list
    me( 2, eT_fluidElem ) =  me( 1, eT_ghostFromCoarserElem )
    me( 2, eT_ghostFromCoarserElem ) = me( 1, eT_ghostFromFinerElem )
    me( 2, eT_ghostFromFinerElem )   = me( 1, eT_haloElem )
    me( 2, eT_haloElem )   =           nTotal
    me( :, eT_distributedGhostFromFinerElem ) = me( :, eT_GhostFromFinerElem )

  end subroutine set_offsets
!******************************************************************************!


!******************************************************************************!
  !> Create the level-wise total lists.
  !!
  !! They consist of 
  !! fluid elements + ghost elements + halo elements 
  !! and are sorted by the treeIDs
  !! Also create the property lists
  !!
  subroutine assemble_lists( levelDesc, minlevel )
    !---------------------------------------------------------------------------
    !> Minimal level in the mesh
    integer, intent(in) :: minlevel
    !> Level descriptor to fill
    type(tem_levelDesc_type), intent(inout) :: levelDesc(minlevel:)
    !---------------------------------------------------------------------------
    integer :: iLevel, iIndex
    !---------------------------------------------------------------------------
    write(logUnit(2),*) 'Assembling total list ...'

    do iLevel = minlevel, ubound(levelDesc,1)
      allocate( levelDesc( iLevel)%total( levelDesc( iLevel )%nElems ))
      allocate( levelDesc( iLevel)%property(levelDesc( iLevel )%nElems ))
      allocate( levelDesc( iLevel)%pntTID( levelDesc( iLevel )%nElems_fluid ))

      call set_offsets( me       = levelDesc( iLevel )%offset(:,:),            &
        &               nFluids  = levelDesc( iLevel )%nElems_fluid,           &
        &               nGhostFC = levelDesc( iLevel )%nElems_ghostFromCoarser,&
        &               nGhostFF = levelDesc( iLevel )%nElems_ghostFromFiner,  &
        &               nHalos   = levelDesc( iLevel )%nElems_halo )

      write(logUnit(5),*) '  Level           ', iLevel
      write(logUnit(5),*) '    nElems_total  ', levelDesc(iLevel)%nElems
      write(logUnit(5),*) '    nElems_fluid  ', levelDesc(iLevel)%nElems_fluid
      write(logUnit(5),*) '    nElems_gfC    ', levelDesc(iLevel) &
        &                                       %nElems_ghostFromCoarser
      write(logUnit(5),*) '    nElems_gfF    ', levelDesc(iLevel) &
        &                                       %nElems_ghostFromFiner
      write(logUnit(5),*) '    nElems_halo   ', levelDesc(iLevel) &
        &                                       %nElems_halo

      ! Counter across all lists
      do iIndex = 1, levelDesc( iLevel )%nElems
        levelDesc( iLevel)%total( iIndex ) = levelDesc( iLevel )               &
          &    %elem%tID%val( levelDesc( iLevel )%totalPnt(iIndex) )
        levelDesc( iLevel)%property( iIndex ) = levelDesc( iLevel )            &
          &    %elem%property%val( levelDesc( iLevel )%totalPnt(iIndex))
        if( iIndex .le. levelDesc( iLevel )%nElems_fluid ) then
          levelDesc( iLevel)%pntTID( iIndex )  = levelDesc( iLevel )           &
            &    %elem%pntTID%val( levelDesc( iLevel )%totalPnt( iIndex ))
        end if
      end do

    end do ! iLevel

  end subroutine assemble_lists
!******************************************************************************!


!******************************************************************************!
  !> identify additionally required neighbor elements
  !! run over the 'require' list of elements, which was accumulated before in
  !! init_elemLevels. The list includes neighbor elements of stencil neighbors,
  !! for stencils with the requireNeighNeigh attribute set. 
  !! This is needed for example for LBM boundary stencil elements, which in turn
  !! require their compute stencil neighborhood to allow PULL operations from
  !! there
  !!
  subroutine identify_additionalNeigh( tree, prc, levelDesc, pathFirst,        &
    &                                  pathLast, computeStencil )
    !---------------------------------------------------------------------------
    !>
    type(treelmesh_type), intent(in) :: tree
    !>
    type(tem_comm_env_type), intent(inout) :: prc
    !>
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minlevel:)
    !>
    type(tem_path_type), intent(in) :: pathFirst(:)
    !>
    type(tem_path_type), intent(in) :: pathLast(:)
    !> the compute stencil, for which the additional neighbors are reconstructed
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !---------------------------------------------------------------------------
    integer :: iLevel, iElem, neighPos, iStencil, iNeighElem, indElem
    integer :: nesting
    integer(kind=long_k) :: nTreeID
    !---------------------------------------------------------------------------
    ! The position of the compute stencil
    iStencil = 1
    do iLevel = tree%global%minlevel, tree%global%maxLevel
      ! Run over the additionally required element list
      do indElem = 1, levelDesc( iLevel )%require%nVals 
        ! get the position of the treeID in the element list
        iElem = PositionOfVal( me  = levelDesc( iLevel )%elem%tID,             &
          &                    val = levelDesc( iLevel )%require%val( indElem ))
        ! get the element position 
        if( iElem .gt. 0 ) then
          if ( levelDesc( iLevel )%elem%eType%val( iElem ) .gt. 0) then
            ! Run over all the neighbors of the compute stencil
            do iNeighElem = 1, levelDesc( iLevel )%elem%stencil%val( iElem )   &
                &                                          %val( iStencil )%QQN
              ! position
              neighPos = levelDesc( iLevel )%elem%stencil%val( iElem )         &
                &                         %val( iStencil )%tIDpos( iNeighElem ) 
              if( neighPos .gt. 0 ) then
                nTreeID = levelDesc( iLevel )%elem%neighID%val( iElem )%       &
                  &                                             val( neighPos )
                nesting = 0
                call identify_elements( nTreeID        = nTreeID,              &
                  &                     tree           = tree,                 &
                  &                     pathFirst      = pathFirst,            &
                  &                     pathLast       = pathLast,             &
                  &                     levelDesc      = levelDesc,            &
                  &                     elemPos        = neighPos,             &
                  &                     prc            = prc,                  &
                  &                     nesting        = nesting,              &
                  &                     computeStencil = computeStencil )
              else
                neighPos = 0 
              end if
              levelDesc( iLevel )%elem%stencil%val( iElem )%val( iStencil )    &
                &                %totalPos( iNeighElem ) = neighPos
            end do
          else
            write(logUnit(2),*) 'not found required element in mesh ', nTreeID
          endif  
        end if
      end do
    end do

  end subroutine identify_additionalNeigh
!******************************************************************************!


!******************************************************************************!
  !> exchange the requested treeIDs between all MPI processs
  !!
  subroutine communicate_elements( tree, prc, levelDesc, commPattern,          &
    &                  pathFirst, pathLast, computeStencil, dumpLists )
    !---------------------------------------------------------------------------
    !>
    type(treelmesh_type), intent(in) :: tree
    !>
    type(tem_comm_env_type), intent(inout) :: prc
    !>
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minlevel:)
    !>
    type(tem_commPattern_type), intent(in)    :: commPattern
    !>
    type(tem_path_type), intent(in) :: pathFirst(:), pathLast(:)
    !>
    type(tem_stencilHeader_type), intent(in) :: computeStencil(:)
    !>
    logical, intent(in) :: dumpLists
    !---------------------------------------------------------------------------
    integer :: iLevel, iErr, nProcs, iProc, iHalo, elemPos
    integer,allocatable :: nHalos(:)
    integer :: nIterations
    logical :: redo ! locally indicate if another iteration has to be performed
    logical :: redo_global    ! global indicator for another iteration
    !---------------------------------------------------------------------------

    write(logUnit(3),*) '-------------------------------------------'
    write(logUnit(3),*) 'Communicating elements ...'
    !----------- exchange number of requesting / requested treeIDs -------------
    nIterations = 0
    do  !iIter = 1, 1 ! exchange halos until no new elements are created
      nIterations = nIterations + 1
      write(logUnit(4),*) 'Halo count exchange iteration: ', nIterations
      redo = .false.
      ! communicate nHalos and allocate levelDesc%buffer
      call communicate_nElemsToTransfer( tree      = tree,                     &
        &                                prc       = prc,                      &
        &                                levelDesc = levelDesc )
      !
      !---done.--- exchange number of requesting / requested treeIDs ----------!
      write(logUnit(5),*) '  Done communicating nElems to Transfer'

      !-----------                                                   ----------!
      !
      ! now we request the halo cells from the mpi processs,
      ! so they can send us their dependencies for these halo cells.
      ! We allow only dependencies with a difference of one level
      ! from lower to higher refinement levels and an arbitrary difference
      ! of refinement level from higher to lower refinement level.
      ! Since we get only the leaves as a dependency from the other mpi process
      ! we have to figure out later which cells we have to add locally
      !
      ! 1)  get the number of cells we will receive from the active processes,
      !     since we have
      !     to communicate only to the mpi processs which have/need informations
      ! 2)  exchange the treeIDs of the halos
      !
      ! First collapse the levelDesc%halos.
      ! We remove the processes where we don't have any elements to exchange
      ! with.
      ! Inverse Communciation (send to sourceProc, recv from targetProc)
      do iLevel = tree%global%minLevel, tree%global%maxLevel
        nProcs = levelDesc( iLevel )%halosPrc%nVals
        if( allocated( nHalos )) deallocate( nHalos )
        allocate( nHalos( nProcs ))
        ! Send treeIDs
        write(logUnit(5),*) '  Requesting remote halos on level ', iLevel
        call request_remoteHalos( levelDesc      = levelDesc,                  &
          &                       tree           = tree,                       &
          &                       iLevel         = iLevel,                     &
          &                       pathFirst      = pathFirst,                  &
          &                       pathLast       = pathLast,                   &
          &                       computeStencil = computeStencil(1),          &
          &                       prc            = process )
        if (dumpLists) then
          call tem_elemList_dump( me = levelDesc( iLevel )%elem,      &
            &                     nUnit = dbgUnit(0),                 &
            &                     stencil = .true.,                   &
            &                     string = 'after request remotHalos' )
        end if
        write(logUnit(5),*) '  Done requesting remote halos.'

        nHalos(:nProcs) = levelDesc( iLevel )%halosPrc%val(:nProcs )%mine%nVals
        write(logUnit(5),*) '  Identifying lists'
        call identify_lists( tree, levelDesc, iLevel )
        if( any( levelDesc( iLevel )%halosPrc%val(1:nProcs)%mine%nVals         &
          &      .ne. nHalos(:) )                                              &
          &      .or. nProcs .ne. levelDesc( iLevel )%halosPrc%nVals)          &
          redo = .true.
      end do
      write(logUnit(6),*) '  Allreduce to check if changes occurred on any' &
        &                 //' process'
      ! JUROPA work-around for crash in the mpi_allreduce
      call mpi_barrier( prc%comm, iErr )
      ! Determine among all neighbor processes, if further iterations required
      call mpi_allreduce( redo, redo_global, 1, mpi_logical, mpi_lor,          &
        &                 prc%comm, iErr )
      if( .not. redo_global ) exit
    end do
    write(logUnit(3),*) 'Done exchanging number of elements to communicate.'
     
    !!   Now each process knows, which halos are requested.
    !!   Continue with identifying the actual leaf elements, which
    !!   are then communicated

    write(logUnit(3),*) 'Return halo counts and Redefine halos ...'
    do iLevel = tree%global%minLevel, tree%global%maxLevel
      write(logUnit(5),*) '  Returning halo counts on level ', iLevel
      ! Receive the number of really existing halo elements
      call return_haloCounts( sendbuffer = levelDesc( iLevel )%sendbuffer,     &
        &                     recvbuffer = levelDesc( iLevel )%recvbuffer,     &
        &                     prc        = process )
      call return_haloCounts( sendbuffer = levelDesc( iLevel )                 &
        &                                  %sendbufferFromCoarser,             &
        &                     recvbuffer = levelDesc( iLevel )                 &
        &                                  %recvbufferFromCoarser,             &
        &                     prc        = process )
      call return_haloCounts( sendbuffer = levelDesc( iLevel )                 &
        &                                  %sendbufferFromFiner,               &
        &                     recvbuffer = levelDesc( iLevel )                 &
        &                                  %recvbufferFromFiner,               &
        &                     prc        = process )

      ! reset the halos in the elem list
      do iProc = 1, levelDesc( iLevel )%halosPrc%nVals
        ! First declare all local halos as non-existent, and set only those
        ! actually provided by the remote process.
        do iHalo = 1, levelDesc( iLevel )%halosPrc%val( iProc )%mine%nVals
          elemPos = levelDesc( iLevel )%halosPrc%val( iProc )%mine%val( iHalo )
          levelDesc( iLevel )%elem%eType%val( elemPos ) = eT_nonExistingElem
        end do
      end do

      write(logUnit(5),*) '  Redefining halos ... '
      ! Receive the number of really existing halo elements 
      call redefine_halos( levelDesc      = levelDesc( iLevel ),               &
        &                  sendbuffer     = levelDesc( iLevel )%sendbuffer,    &
        &                  recvbuffer     = levelDesc( iLevel )%recvbuffer,    &
        &                  commPattern    = commPattern,                       &
        &                  computeStencil = computeStencil,                    &
        &                  prc            = process )
      
      call redefine_halos( levelDesc      = levelDesc( iLevel ),               &
        &                  sendbuffer     = levelDesc( iLevel )                &
        &                                             %sendbufferFromCoarser,  &
        &                  recvbuffer     = levelDesc( iLevel )                &
        &                                             %recvbufferFromCoarser,  &
        &                  commPattern    = commPattern,                       &
        &                  computeStencil = computeStencil,                    &
        &                  prc            = process )
      
      call redefine_halos( levelDesc      = levelDesc( iLevel ),               &
        &                  sendbuffer     = levelDesc( iLevel )                &
        &                                              %sendbufferFromFiner,   &
        &                  recvbuffer     = levelDesc( iLevel )                &
        &                                              %recvbufferFromFiner,   &
        &                  commPattern    = commPattern,                       &
        &                  computeStencil = computeStencil,                    &
        &                  prc            = process )
      call identify_lists( tree, levelDesc, iLevel )
    end do
    do iLevel = tree%global%minLevel, tree%global%maxLevel
      if (dumpLists) then
      call tem_elemList_dump( me      = levelDesc( iLevel )%elem,   &
        &                     nUnit   = dbgUnit(0),                 &
        &                     stencil = .true.,                     &
        &                     string  = 'after redefine remotHalos' )
      end if
    end do
    write(logUnit(3),*) 'Done with communication of elements. '
    write(logUnit(3),*) '-------------------------------------------'

  end subroutine communicate_elements
!******************************************************************************!


!******************************************************************************!
  !> Map requested halo to a position in my local fluid list or
  !! add recursively ghosts until I reach valid fluid elements
  !! return type of added element in levelPos(2)
  !! Also, non-existing elements are reported as such (levelPos(2))
  !!
  subroutine identify_halo( haloTreeID, elemPos, haloLevel, levelDesc, tree,   &
    &                       updated,   nesting, minLevel, computeStencil )
    !---------------------------------------------------------------------------
    !>
    type(treelmesh_type), intent(in) :: tree
    !>
    integer, intent(in) :: minLevel
    !>
    integer, intent(in) :: nesting
    !> neighboring treeID
    integer(kind=long_k), intent(in) :: haloTreeID
    !> type and position in list of found treeID
    integer, intent(out) :: elemPos
    !>
    integer, intent(out) :: haloLevel
    !>
    logical, intent(out) :: updated  
    !>
    type(tem_levelDesc_type), intent(inout) :: levelDesc(minlevel:)
    !>
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !---------------------------------------------------------------------------
    haloLevel = tem_levelOf( haloTreeID )
    elemPos = PositionOfVal( me  = levelDesc( halolevel )%elem%tID,            &
      &                      val = haloTreeID )
    
    if( elemPos .eq. 0 ) then
      ! requested halo does not exist and has to be created somehow
      call identify_local_element( nTreeID        = haloTreeID,                &
        &                          levelDesc      = levelDesc,                 &
        &                          tree           = tree,                      &
        &                          elemPos        = elemPos,                   &
        &                          level          = haloLevel,                 &
        &                          nesting        = nesting,                   &
        &                          updated        = updated,                   &
        &                          minLevel       = minLevel,                  &
        &                          computeStencil = computeStencil )
    end if

  end subroutine identify_halo
!******************************************************************************!


!******************************************************************************!
  !> deallocate all dispensable dynamic data structures and arrays
  !!
  subroutine tem_cleanupDependencyArrays( levelDesc )
    !---------------------------------------------------------------------------
    !> the level descriptor
    type(tem_levelDesc_type) , intent(inout) :: levelDesc(:) 
    !---------------------------------------------------------------------------
    integer :: iLevel
    !---------------------------------------------------------------------------

    do iLevel = 1,size(levelDesc)
      ! Kill empty fluid, ghost and halo lists
      call destroy( me = levelDesc( iLevel )%halosPrc )
      call destroy( me = levelDesc( iLevel )%require )
    end do

  end subroutine tem_cleanupDependencyArrays   
!******************************************************************************!


!******************************************************************************!
  !> deallocate the stencil treeID neighbor arrays for each element
  !! This routine can only be called after build_horizontalDependencies
  !! 
  subroutine tem_cleanup_arrays( levelDesc )
    !---------------------------------------------------------------------------
    !> the level descriptor
    type(tem_levelDesc_type) , intent(inout) :: levelDesc(:) 
    !---------------------------------------------------------------------------
    integer :: iLevel
    !---------------------------------------------------------------------------

    ! loop over all levels and destroy the corresponding elem types
    do iLevel = 1,size(levelDesc)
      call destroy( me = levelDesc( iLevel )%elem )
    end do

  end subroutine tem_cleanup_arrays   
!******************************************************************************!


!******************************************************************************!
  !> Build the vertical dependencies of ghost elements
  !! 
  subroutine tem_build_verticalDependencies( levelDesc, minlevel ) 
    !---------------------------------------------------------------------------
    !> the level descriptor
    type(tem_levelDesc_type), intent(inout) :: levelDesc(minlevel:)
    !>
    integer, intent(in) :: minlevel
    !---------------------------------------------------------------------------
    integer(kind=long_k) :: cTreeID, childTreeID(8)
    integer :: iLevel, iElem, iChild
    integer :: elemPos, offset
    !---------------------------------------------------------------------------
    write(logUnit(3),*) 'Build vertical dep ...'

    ! Start with GhostFromCoarser
    do iLevel = minLevel+1,ubound(levelDesc,1)
      allocate( levelDesc( iLevel )%depFromCoarser(                            &
        &       levelDesc( iLevel )%nElems_ghostFromCoarser))
      offset = levelDesc( iLevel )%offset(1,eT_ghostFromCoarserElem)
      do iElem = 1, levelDesc( iLevel )%nElems_ghostFromCoarser
        elemPos = levelDesc( iLevel )%totalPnt( offset+iElem )      
        cTreeID = levelDesc( iLevel )%elem%tID%val( elemPos )
        elemPos = tem_treeIDinTotal(                                           &
          &       tem_parentOf( cTreeID ), levelDesc( iLevel-1 ))
        ! Add the requested TreeID to the halos and add dependencies
        call appendGhostDependency(                                            &
          &           sourcePos   = elemPos,                                   &
          &           sourceLevel = iLevel-1,                                  &
          &           tgtDep      = levelDesc( iLevel )%depFromCoarser( iElem ))
      end do
    end do

    ! Continue with GhostFromFiner
    do iLevel = minLevel,ubound(levelDesc,1)
      allocate( levelDesc( iLevel )%depFromFiner(                              &
        &       levelDesc( iLevel )%nElems_ghostFromFiner ))
      offset = levelDesc( iLevel )%offset(1,eT_ghostFromFinerElem)
      do iElem = 1, levelDesc( iLevel )%nElems_ghostFromFiner
        elemPos = levelDesc( iLevel )%totalPnt( offset + iElem )
        cTreeID = levelDesc( iLevel )%elem%tID%val( elemPos )
        childTreeID(:) = tem_directChildren( cTreeID )
        do iChild = 1, 8
          elemPos = tem_treeIDinTotal( childTreeID( iChild ),                  &
            &                          levelDesc( iLevel+1 ))
          if( elemPos .gt. 0 ) then
            ! if the child treeID exists, add dependency
            call appendGhostDependency(                                        &
              &         sourcePos   = elemPos,                                 &
              &         sourceLevel = iLevel+1,                                &
              &         tgtDep      = levelDesc( iLevel )%depFromFiner( iElem ))
          end if
        end do
      end do
    end do

  end subroutine tem_build_verticalDependencies
!******************************************************************************!


!******************************************************************************!
  !> Building neighor array
  !!
  subroutine tem_build_horizontalDependencies( iStencil, levelDesc, tree,      &
    &                                          computeStencil, check )
    !---------------------------------------------------------------------------
    !> Tree representation of your mesh.
    type(treelmesh_type), intent(in) :: tree
    !> Level descriptor for each level of your mesh (starting from minimum
    !! level).
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minLevel:)
    !> Index of your neighbor list.
    integer, intent(in) :: iStencil
    !> The stencil you build the horizontal dependencies for.
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !> check dependencies
    logical, intent(in), optional :: check
    !---------------------------------------------------------------------------
    logical :: do_check   ! check dependencies
    integer :: iLevel, iIndex
    integer :: nElemsWithNeigh
    !---------------------------------------------------------------------------

    if( present( check )) then
      do_check  = check   
    else 
      do_check  = .false.
    end if
    write(logUnit(3),*) 'Building neighbor array for stencil', iStencil
    do iLevel = tree%global%minLevel,tree%global%maxLevel
      
      if( computeStencil%useAll ) then
        nElemsWithNeigh = levelDesc(iLevel)%nElems
      else
        nElemsWithNeigh = computeStencil%nElems
      end if
      write(logUnit(4),*) '     nElems to treat ', nElemsWithNeigh
      
      allocate( levelDesc( iLevel )%neigh( iStencil )%                         &
        &          nghElems( computeStencil%QQN, nElemsWithNeigh ) )
      iIndex = 0
      if( computeStencil%QQN .gt. 0 ) then
        ! we init everything by default with 0 to find a bug as fast as possible
        levelDesc( iLevel )%neigh( iStencil )%nghElems(:,:) = 0
        if (computeStencil%useAll) then
!          write(logUnit(5),*) '     fluid elements  ', iIndex
          call tem_build_listHorizontalDep(                                    &
            &                  levelDesc      = levelDesc( iLevel ),           &
            &                  computeStencil = computeStencil,                &
            &                  iStencil       = iStencil,                      &
            &                  elem           = levelDesc( iLevel )%elem,      &
            &                  list           = levelDesc( iLevel )%totalPnt,  &
            &                  nElems         = levelDesc( iLevel )%nElems,    &
            &                  iIndex         = iIndex )
        
        else ! stencil not using all elements
!          write(logUnit(5),*) '     not all elems   ', iIndex
          call tem_build_treeHorizontalDep(                                    &
            &         levelDesc      = levelDesc( iLevel ),                    &
            &         computeStencil = computeStencil,                         &
            &         iStencil       = iStencil,                               &
            &         list           = computeStencil%elemLvl( iLevel )%val,   &
            &         nElems         = computeStencil%elemLvl( iLevel )%nVals, &
            &         tree           = tree,                                   &
            &         iIndex         = iIndex )
        end if
      end if
      
    end do ! iLevel
    
    if ( do_check .or. tem_logging_isActive(10) ) then
      call tem_debug_HorizontalDependencies( iStencil       = iStencil,        &
        &                                    levelDesc      = levelDesc,       &
        &                                    tree           = tree,            &
        &                                    computeStencil = computeStencil,  &
        &                                    check          = do_check )
    end if
    
  end subroutine tem_build_horizontalDependencies
!******************************************************************************!


!******************************************************************************!
  !> Building neighor array
  !!
  subroutine tem_debug_HorizontalDependencies( iStencil, levelDesc, tree,      &
    &                                          computeStencil, check )
    !---------------------------------------------------------------------------
    !> Tree representation of your mesh.
    type(treelmesh_type), intent(in) :: tree
    !> Level descriptor for each level of your mesh (starting from min level).
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minLevel:)
    !> Index of your neighbor list.
    integer, intent(in) :: iStencil
    !> The stencil you build the horizontal dependencies for.
    type(tem_stencilHeader_type) :: computeStencil 
    !>
    logical, intent(in) :: check
    !---------------------------------------------------------------------------
    integer :: iElem, iLevel, nElems, nEntries, iVal, elemPos, tIDpos
    integer :: neighPos, stencilSize, stencilPos
    integer(kind=long_k) :: tID, bufID, refID
    character(len=1024) :: buffer
    logical :: abort_loc, warning, skipElem
    !---------------------------------------------------------------------------

    write(dbgUnit(1),*)'Check neighbor positions for stencil ' &
      &                // trim(tem_toStr(iStencil)) 
    abort_loc = .false.
    warning = .false.
    write(logUnit(0),*)' Checking, if the horizontal dependencies were set'// &
      &                ' correctly'

    do iLevel = tree%global%minLevel,tree%global%maxLevel
      nEntries = size( levelDesc( iLevel )%neigh(iStencil)%nghElems, 1)
      nElems = size( levelDesc( iLevel )%neigh( iStencil )%nghElems, 2)
      if( nEntries .gt. 0 ) then
        do iElem = 1, nElems
          abort_loc    = .false.
          warning  = .false.
          skipElem = .false.
          ! First get the treeID
          if( computeStencil%useAll ) then
            tIDPos =iElem
            tID = levelDesc( iLevel )%total( tIDpos )
          else
            if(  computeStencil%elemLvl( iLevel )%nVals .le. 0 ) exit
            tIDPos = computeStencil%elemLvl( iLevel )%val( iElem )
            tID = tree%treeID( tIDpos )
          end if
          ! Get the position of the treeID in the element list
          elemPos = PositionOfVal(levelDesc( iLevel )%elem%tID, tID )
          stencilPos = tem_stencil_getHeaderPos(                               &
            &           me  = levelDesc( iLevel )%elem%stencil%val( elemPos ), &
            &           val = iStencil )
          stencilSize = size( levelDesc( iLevel )%elem%stencil%val( elemPos )  &
            &                                       %val( stencilPos )%tIDpos )
          if (nEntries .ne. stencilSize) then
            write(dbgUnit(1),*)'stencil entries do not match for tID '  &
              &                 // trim(tem_toStr(tID)) // ': '         &
              &                 // trim(tem_toStr(nEntries))// ' vs. '  &
              &                 // trim(tem_toStr(stencilSize))         
 
          end if
          write( buffer, '(i9, a)' ) tID,' ngh  '
          do iVal = 1, nEntries
            neighPos = levelDesc( iLevel )%neigh(iStencil)%                    &
              &                                         nghElems( iVal, iElem )
            tIDpos = levelDesc( iLevel )%elem%stencil%val( elemPos )%val(      &
              &                                     stencilPos )%tIDpos( iVal )
            if( tIDpos .gt. 0 ) then
              refID = levelDesc( iLevel )%elem%neighID%val( elemPos )%val(     &
                &                                                      tIDpos )
            else
              refID = 0
            end if
            if( neighPos .gt. 0 ) then
              bufID = levelDesc( iLevel )%total( neighPos )
            else
              bufID = neighPos
            endif
            write(buffer,'(a,i9)') trim(buffer),  bufID
            if( refID .ne. bufID ) then
              write(buffer,'(a,a,i0,a)') trim(buffer),  ' (',refID,')'
              warning = .true.
              if( bufID .gt. 0_long_k ) then
                abort_loc  = .true.
                exit  
              end if
            end if
          end do
          write(dbgUnit(6),*) trim(buffer)
          if (abort_loc .and. check) then
            write(dbgUnit(1),*)' Non matching stencil entries!!' 
            write(dbgUnit(1),*)'    original treeID:         ' & 
              &               //trim(tem_toStr( tID)) 
            write(dbgUnit(1),*)'    required stencil treeID: ' &  
              &               //trim(tem_toStr(refID)) 
            write(dbgUnit(6),*) '    encountered treeID:      '& 
              &               //trim(tem_toStr(bufID)) 
            write(logUnit(0),*)' Found non-matching treeID in the neighborhood'
            write(logUnit(0),*)'    original treeID:         ', tID
            write(logUnit(0),*)'    required stencil treeID: ', refID
            write(logUnit(0),*)'    encountered treeID:      ', bufID
            call tem_abort()
          end if
        end do
      end if
    end do

  end subroutine tem_debug_HorizontalDependencies
!******************************************************************************!


!******************************************************************************!
  !> Update the neighor arrays depending on what is given in the element stencil
  !! 
  !! The array levelDesc( iLevel )%neigh( iStenci )%nghElems( 1:QQN, 1:nElems )
  !! is being filled up here
  !!
  subroutine tem_build_treeHorizontalDep( iStencil, levelDesc, computeStencil, &
    &                                     list, nElems, iIndex, tree )
    !---------------------------------------------------------------------------
    !> Level descriptor for each level of your mesh (starting from min level).
    type(tem_levelDesc_type), intent(inout) :: levelDesc
    !> tree information
    type(treelmesh_type), intent(in) :: tree
    !> Index of your neighbor list.
    integer, intent(in) :: iStencil
    !> The stencil you build the horizontal dependencies for.
    type(tem_stencilHeader_type), intent(in) :: computeStencil 
    !>
    integer, intent(inout) :: iIndex
    !>
    integer, intent(in) :: nElems
    !>
    integer, intent(in) :: list(:)
    !---------------------------------------------------------------------------
    integer :: iElem, iStencilElem
    integer :: elemPos, stencilPos, levelPos, totalPos
    integer(kind=long_k) :: tID
    logical :: missingNeigh
    !---------------------------------------------------------------------------
    ! run over the given element list to update
    do iElem = 1, nElems
      missingNeigh = .false.
      tID = tree%treeID( list( iElem ))
      ! Get the position of the treeID in the element list
      elemPos = PositionOfVal( me = levelDesc%elem%tID, val = tID )
      if( elemPos .gt. 0 ) then
        totalPos = tem_treeIDinTotal(                                          &
          &                    tID       = tID,                                &
          &                    levelDesc = levelDesc,                          &
          &                    eType     = levelDesc%elem%eType%val( elemPos ))
        stencilPos = tem_stencil_getHeaderPos(                                 &
          &                       me  = levelDesc%elem%stencil%val( elemPos ), &
          &                       val = iStencil )
        if( stencilPos .gt. 0 ) then
          do iStencilElem=1,levelDesc%elem%stencil%val(elemPos)                &
            &                                             %val(stencilPos)%QQN
            ! Totalpos has before already been updated, so just read from
            ! totalPos
            levelPos = levelDesc%elem%stencil%val(elemPos)                     &
              &                              %val(stencilPos)                  &
              &                              %totalPos(iStencilElem)
            if( levelPos .le. 0 .and. computeStencil%requireNeighNeigh ) then
              missingNeigh = .true.
            end if
            levelDesc%neigh( iStencil )%nghElems( iStencilElem, iElem )        &
              &                                                     = levelPos
          end do
        end if
        if( missingNeigh ) then
! write(*,*) 'missing neighbor for element ', iElem
          ! Set the property of the current element to have missing neighbors
          levelDesc%property( totalPos )                                       &
            &  = ibset( levelDesc%property( totalPos ), prp_missingNeigh )
        end if
      else
        write(dbgUnit(1),*)'Warning!!!' 
        write(dbgUnit(1),*)'element not found in build_treehorizontalDep' 
        write(dbgUnit(1),*)'That should not occur.' 
      end if
    end do

  end subroutine tem_build_treeHorizontalDep
!******************************************************************************!
 

!******************************************************************************!
  !> Update the neighor arrays depending on what is given in the element stencil
  !! 
  !! The array levelDesc( iLevel )%neigh( iStenci )%nghElems( 1:QQN, 1:nElems )
  !! is being filled up here
  subroutine tem_build_listHorizontalDep( iStencil, levelDesc, computeStencil, &
    &                                     elem, list, nElems, iIndex )
    !---------------------------------------------------------------------------
    !> Level descriptor for each level of your mesh (starting from min level).
    type(tem_levelDesc_type), intent(inout) :: levelDesc
    !> Index of your neighbor list.
    integer, intent(in) :: iStencil
    !> The stencil you build the horizontal dependencies for.
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !> element list
    type(tem_element_type), intent(in) :: elem          
    !>
    integer, intent(inout) :: iIndex
    !>
    integer, intent(in) :: nElems
    !>
    integer, intent(in) :: list(:)
    !---------------------------------------------------------------------------
    integer :: iElem, iStencilElem
    integer :: elemPos, levelPos, neighPos, neighVal
    integer(kind=long_k) :: nTreeID ! neighbor treeID value to treat
    logical :: invalid
    !---------------------------------------------------------------------------
    ! run over the given element list to update
    do iElem = 1, nElems
      elemPos = list( iElem )
      ! Check if there are enough stencils given in the element
      if( levelDesc%elem%stencil%val( elemPos )%nVals .ge. iStencil ) then
        ! Increase the index element counter, if a stencil was encountered
        iIndex = iIndex + 1
        ! Run over each entry of the stencil
        do iStencilElem = 1, levelDesc%elem%stencil%val( elemPos )%            &
          &                                                val( iStencil )%QQN
          neighPos = levelDesc%elem%stencil%val( elemPos )%val( iStencil )     &
            &                                           %tIDpos( iStencilElem )
          invalid = .false.
          if( neighPos .gt. 0 ) then
            nTreeID = levelDesc%elem%neighID%val( elemPos )%val( neighPos )
            if( nTreeID .lt. 0_long_k ) then
              ! If the entry in the stencil is < 0, it must be a boundary ID
              ! -> simply set
              neighVal = int(nTreeID)
            else
              ! If the entry is > 0, test if there is also an entry for the
              ! element position directly (%elem array)
              if( allocated( levelDesc%elem%stencil%val( elemPos )             &
                &                            %val( iStencil )%totalPos) ) then
                ! If there is an entry, test, if it is valid
                levelPos = levelDesc%elem%stencil%val( elemPos )               &
                &                    %val( iStencil )%totalPos( iStencilElem )
                if( levelPos .gt. 0 ) then
                  ! If there is a valid entry, get the position of this element
                  ! in the total list
                  neighVal = levelPos 
                else 
                  ! If the entry in %elem was not valid, use the tID to find
                  invalid = .true.
                end if ! levelPos > 0
              else 
                ! if there was no %elem entry given at all, directly use the
                ! tID to find the element position in the total list
                invalid = .true.
              end if ! allocated totalPos
            end if ! nTreeID > 0
!            invalid = .true.
            ! If above anything went wrong, ...
            if( invalid ) then
              ! ... try to find it with the slowest method: just look up the
              ! tID in the total list
              neighVal = tem_treeIDinTotal( nTreeID, levelDesc )
            end if
          else ! neighPos < 0
            neighVal = neighPos
          end if ! neighPos > 0
          levelDesc%neigh(iStencil)%nghElems( iStencilElem, iIndex ) = neighVal
        end do ! stencil element
      end if ! nStencils > iStencil
    end do ! iElem
  end subroutine tem_build_listHorizontalDep
!******************************************************************************!


!******************************************************************************!
  !> Communicate with all existing process the number of requested halo elements
  !!
  !! After this routine, each process knows how many processes there are to
  !! communicate with and how many elements have to be transferred
  !!
  subroutine communicate_nElemsToTransfer( tree, prc, levelDesc )
    !---------------------------------------------------------------------------
    !>
    type(treelmesh_type), intent(in) :: tree
    !>
    type(tem_comm_env_type), intent(inout) :: prc
    !>
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minlevel:)
    !---------------------------------------------------------------------------
    ! the number of halos which will be sent to this mpi process for each level
    ! and for each process:
    ! first dimension is the level, second one is the mpi process
    integer, allocatable :: nHalosHis( :,:)
    integer, allocatable :: nHalosMine(:,:)
    integer :: iProc, iLevel, ierr
    integer :: iIndexMine, iIndexHis
    integer :: nLevels, sourceProc
    !---------------------------------------------------------------------------

    ! allocate the send buffer. 
    ! Contains the number of elements for each level, which are requested 
    ! from the specific process
    allocate( nHalosMine( tree%global%minLevel:tree%global%maxLevel,           &
      &                   prc%comm_size ))
    ! allocate the recv buffer. 
    ! Will contain the number of elements for each level, which other processes 
    ! request from me. This one is filled in the mpi_alltoall call
    allocate( nHalosHis( tree%global%minLevel:tree%global%maxLevel,            &
      &                  prc%comm_size ))

    ! Reset the number of requested and requesting halos
    nHalosMine(:,:) = 0
    nHalosHis(:,:)  = 0

    ! Delete possible pre-allocated arrays from previous exchanges
    do iLevel = tree%global%minLevel,tree%global%maxLevel
      if( allocated( levelDesc( iLevel )%recvbuffer%proc ))                    &
        &                     deallocate( levelDesc( iLevel )%recvbuffer%proc )
      if( allocated( levelDesc( iLevel )%recvbuffer%nElemsProc ))              &
        &               deallocate( levelDesc( iLevel )%recvbuffer%nElemsProc ) 
      if( allocated( levelDesc( iLevel )%sendbuffer%proc ))                    &
        &                     deallocate( levelDesc( iLevel )%sendbuffer%proc )
      if( allocated( levelDesc( iLevel )%sendbuffer%nElemsProc ))              &
        &                deallocate(levelDesc( iLevel )%sendbuffer%nElemsProc ) 
    end do

    !---------------------------------------------------------------------------
    ! Fill the send buffer, i.e. number of elements, which I request from
    ! all other processes.
    do iLevel = tree%global%minLevel, tree%global%maxLevel
      do iProc = 1, levelDesc( iLevel )%halosPrc%nVals
        sourceProc = levelDesc( iLevel )%halosPrc%val( iProc )%partnerProc
        nHalosMine(iLevel, sourceProc) =                                       &
          &      levelDesc( iLevel )%halosPrc%val( iProc )%mine%nVals
      end do !< iProc
    end do   !< iLevel

    ! JUROPA MPI WORKAROUND !
    ! This barrier is just here to workaround a deadlock problem in
    ! Parastation MPI on Juropa, which results in the consequent all
    ! to all to never complete on some processes, if PSP_ONDEMAND=1
    ! is set.
    !> \todo HK: remove this workaround for the buggy MPI implementation
    !! on Juropa again!
    call mpi_barrier(prc%comm, iErr) ! JUROPA
    write(logUnit(3),*) ''
    ! Exchange the number of requested treeIDs
    nLevels = tree%global%maxLevel - tree%global%minLevel + 1
    call mpi_alltoall( nHalosMine, nLevels, mpi_integer,                       &
      &                nHalosHis, nLevels, mpi_integer,                        &
      &                prc%comm, ierr )
    write(logUnit(3),*) '    Finished all-to-all communication for the'   &
      &                 //' number of elements to exchange.'

    ! JUROPA MPI WORKAROUND
    ! Security barrier for the mentioned Juropa shortcoming before the
    ! all to all.
    call mpi_barrier(prc%comm, iErr) ! JUROPA
    write(logUnit(5),*) '    Done.'

    do iLevel = tree%global%minLevel,tree%global%maxLevel
      ! now check which processs really ask for information and allocate buffer
      levelDesc( iLevel )%recvbuffer%nProcs                                    &
        &  = count( nHalosMine( iLevel, : ) .gt. 0 )
      levelDesc( iLevel )%sendbuffer%nProcs                                    &
        &  = count( nHalosHis( iLevel, : ) .gt. 0 )
      allocate( levelDesc( iLevel )%recvbuffer%proc(                           &
        &           levelDesc( iLevel )%recvbuffer%nProcs ))
      allocate( levelDesc( iLevel )%recvbuffer%nElemsProc(                     &
        &           levelDesc( iLevel )%recvbuffer%nProcs))
      allocate( levelDesc( iLevel )%sendbuffer%proc(                           &
        &           levelDesc( iLevel )%sendbuffer%nProcs ))
      allocate( levelDesc( iLevel )%sendbuffer%nElemsProc(                     &
        &           levelDesc( iLevel )%sendbuffer%nProcs ))
      iIndexMine = 0
      iIndexHis  = 0
      do iProc = 1, prc%comm_size
        if( nHalosMine( iLevel, iProc ) .gt. 0 ) then 
          ! Store the processes numbers to receive from
          iIndexMine = iIndexMine + 1
          levelDesc( iLevel )%recvbuffer%nElemsProc( iIndexMine )              &
            & = nHalosMine( iLevel, iProc )
          levelDesc( iLevel )%recvbuffer%proc( iIndexMine ) = iProc - 1
        end if
        if( nHalosHis( iLevel, iProc ) .gt. 0 ) then 
          ! Store the processes numbers to send to
          iIndexHis = iIndexHis + 1
          levelDesc( iLevel )%sendbuffer%nElemsProc( iIndexHis )               &
            & = nHalosHis( iLevel, iProc )
          levelDesc( iLevel )%sendbuffer%proc( iIndexHis ) = iProc - 1
        end if
      end do
      ! Copy all headers (information about target and source procs)
      ! to bufferFromCoarser and bufferFromFiner
      levelDesc( iLevel )%sendbufferFromCoarser = levelDesc( iLevel )%sendbuffer
      levelDesc( iLevel )%sendbufferFromFiner   = levelDesc( iLevel )%sendbuffer
      levelDesc( iLevel )%recvbufferFromCoarser = levelDesc( iLevel )%recvbuffer
      levelDesc( iLevel )%recvbufferFromFiner   = levelDesc( iLevel )%recvbuffer
    end do ! iLevel
    write(logUnit(3),*) '  Done. Finished comm nElems to transfer.'
 
  end subroutine communicate_nElemsToTransfer
!******************************************************************************!


!******************************************************************************!
  !> Inverse Communication: Communicate, which elements each process needs from
  !! me.
  !!
  !! In this routine, we send the treeIDs of the halo elements to the
  !! processes, where they are located. Later on, we fill these halos locally
  !! with information from these processes (sourceProcs). In this routine
  !! however, we now SEND information to these sourceProcs, so do not get
  !! confused here.
  !!
  subroutine request_remoteHalos( levelDesc, prc, tree, iLevel, computeStencil,&
    &                             pathFirst, pathLast )
    !---------------------------------------------------------------------------
    !>
    type(treelmesh_type), intent(in) :: tree
    !>
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minlevel: )
    !>
    type(tem_comm_env_type), intent(in) :: prc
    !>
    integer, intent(in) :: iLevel
    !>
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !>
    type(tem_path_type), intent(in) :: pathFirst(:), pathLast(:)
    !---------------------------------------------------------------------------
    integer :: iProc, iErr, iElem, elemPos, procPos
    integer :: haloLevel
    integer, allocatable :: rq_handle(:)
    integer, allocatable :: status(:,: )
    integer :: nCommunications
    integer :: nesting
    type( grw_longArray_type ), allocatable :: sendBuf_long(:)
    type( grw_longArray_type ), allocatable :: recvBuf_long(:)
    type( grw_intArray_type ), allocatable :: sendBuf_int(:)
    type( grw_intArray_type ), allocatable :: recvBuf_int(:)
    type( tem_halo_type ) :: tProc
    logical :: wasAdded
    logical :: updated
    integer, parameter :: message_flag_long = 24
    integer, parameter :: message_flag_int  = 25
    !---------------------------------------------------------------------------
    ! two communications: treeID and nesting
    nCommunications = 2*(levelDesc( iLevel )%sendbuffer%nProcs                 &
      &             + levelDesc( iLevel )%recvbuffer%nProcs)
    allocate( rq_handle( nCommunications ) )
    allocate( status( mpi_status_size, nCommunications ) )
    allocate( recvBuf_long( levelDesc( iLevel )%sendbuffer%nProcs ))
    allocate( recvBuf_int( levelDesc( iLevel )%sendbuffer%nProcs ))
    !---------------------------------------------------------------------
    ! Warning: Inverse Communication ! (send to source, recv from target)
    !
    ! The following is CONFUSING, because I have to RECEIVE the SEND buffers.
    ! Allocate receive buffers, which are later on used as information, 
    ! which elements to SEND to which neighbors, that's why mustSend
    rq_handle(:) = MPI_REQUEST_NULL
    !> We have to use nHalosHis and the targetProc here!!
    do iProc = 1, levelDesc( iLevel )%sendbuffer%nProcs
      ! Try to add the target process to the process list
      ! It might not be exist there yet
      tProc%partnerProc = levelDesc( iLevel )%sendbuffer%proc( iProc )+1
      call append( me       = levelDesc( iLevel )%halosPrc,                    &
        &          val      = tProc,                                           &
        &          pos      = procPos,                                         &
        &          wasAdded = wasAdded )
      ! Allocate the buffers
      call init( me     = recvBuf_long( iProc ),                               &
        &        length = levelDesc( iLevel )%sendbuffer%nElemsProc( iProc ))
      call init( me     = recvBuf_int( iProc ),                                &
        &        length = levelDesc( iLevel )%sendbuffer%nElemsProc( iProc ))
      
      ! Receive the element tree IDs
      call mpi_irecv( recvBuf_long( iProc )%val,                               &
        &             recvBuf_long( iProc )%ContainerSize,                     &
        &             mpi_integer8,                                            &
        &             levelDesc( iLevel )%sendbuffer%proc(iProc),              &
        &             message_flag_long,                                       &
        &             prc%comm,                                                &
        &             rq_handle( iProc),                                       &
        &             iErr )
      ! Receive the element nestings
      call mpi_irecv( recvBuf_int( iProc )%val,                                &
        &             recvBuf_int( iProc )%ContainerSize,                      &
        &             mpi_integer,                                             &
        &             levelDesc( iLevel )%sendbuffer%proc(iProc),              &
        &             message_flag_int,                                        &
        &             prc%comm,                                                &
        &             rq_handle( iProc+nCommunications/2 ),                    &
        &             iErr  )                                              
      ! Update the number of elements inside the growing array of the recv 
      ! buffer
      recvBuf_long( iProc )%nVals = recvBuf_long( iProc )%ContainerSize
      recvBuf_int( iProc )%nVals  = recvBuf_long( iProc )%ContainerSize
    end do ! iProc
   
    ! I send the NEEDED elements to the neighbor processes.
    allocate( sendBuf_long( levelDesc( iLevel )%recvbuffer%nProcs ))
    allocate( sendBuf_int( levelDesc( iLevel )%recvbuffer%nProcs ))
    do iProc = 1, levelDesc( iLevel )%recvbuffer%nProcs
      tProc%partnerProc = levelDesc( iLevel )%recvbuffer%proc( iProc ) + 1
      ! Get the position of the process in the dynamic halos list (might be
      ! unordered)
      procPos = PositionOfVal( me = levelDesc( iLevel )%halosPrc, val = tProc )
      call init(  me     = sendBuf_long( iProc ),                              &
        &         length = levelDesc( iLevel )%recvbuffer%nElemsProc( iProc ))
      call init(  me     = sendBuf_int( iProc ),                               &
        &         length = levelDesc( iLevel )%recvbuffer%nElemsProc( iProc ))
      ! Collect the halo treeIDs into send buffers for all the processes to
      ! request from
      do iElem = 1, levelDesc( iLevel )%recvbuffer%nElemsProc( iProc )
        elemPos = levelDesc( iLevel )%halosPrc%val( procPos )%mine%val( iElem )
        call append( me  = sendBuf_long( iProc ),                              &
          &          val = levelDesc( iLevel )%elem%tID%val( elemPos ) )
        call append( me  = sendBuf_int( iProc ),                               &
          &          val = levelDesc( iLevel )%elem%haloNesting%val( elemPos ) )
      enddo
      ! Send treeIDs
      call mpi_isend(                                                          &
       &        sendBuf_long( iProc )%val,                                     &
       &        sendBuf_long( iProc )%nVals,                                   &
       &        mpi_integer8,                                                  &
       &        levelDesc( iLevel )%recvbuffer%proc( iProc ),                  &
       &        message_flag_long,                                             &
       &        prc%comm,                                                      &
       &        rq_handle( iProc + levelDesc( iLevel )%sendbuffer%nProcs),     &
       &        ierr  )
      ! Send nesting
      call mpi_isend(                                                          &
       &        sendBuf_int( iProc )%val,                                      &
       &        sendBuf_int( iProc )%nVals,                                    &
       &        mpi_integer,                                                   &
       &        levelDesc( iLevel )%recvbuffer%proc( iProc ),                  &
       &        message_flag_int,                                              &
       &        prc%comm,                                                      &
       &        rq_handle( iProc + levelDesc( iLevel )%sendbuffer%nProcs       &
       &                         + nCommunications/2),                         &
       &        ierr  )
    end do ! iProc

    call mpi_waitall( nCommunications, rq_handle, status, ierr)
    write(logUnit(5),*) '  Received requested halo elements successfully.'
    ! Requested halo elements were received. 
    !---------------------------------------------------------------------------

    !---------------------------------------------------------------------------
    ! Now identify the requested halos.
    ! Arrays might still be allocated from previous communication iteration
    if( allocated( levelDesc( iLevel )%sendbuffer%elemPos))                    &
      &                    deallocate( levelDesc( iLevel )%sendbuffer%elemPos )
    allocate( levelDesc( iLevel )%sendbuffer%elemPos(                          &
      &                                levelDesc( iLevel )%sendbuffer%nProcs ))
    if( allocated( levelDesc( iLevel )%sendbufferFromCoarser%elemPos))         &
      &         deallocate( levelDesc( iLevel )%sendbufferFromCoarser%elemPos )
    allocate( levelDesc( iLevel )%sendbufferFromCoarser%elemPos(               &
      &                                levelDesc( iLevel )%sendbuffer%nProcs ))
    if( allocated( levelDesc( iLevel )%sendbufferFromFiner%elemPos ))          &
      &           deallocate( levelDesc( iLevel )%sendbufferFromFiner%elemPos )
    allocate( levelDesc( iLevel )%sendbufferFromFiner%elemPos(                 &
      &                                levelDesc( iLevel )%sendbuffer%nProcs ))

    ! Add elements of received buffers to elem
    do iProc = 1, levelDesc( iLevel )%sendbuffer%nProcs
      ! Allocate the buffer for the element position indices
      call init( me = levelDesc( iLevel )%sendbuffer%elemPos(iProc) )
      call init( me = levelDesc( iLevel )%sendbufferFromCoarser%elemPos(iProc) )
      call init( me = levelDesc( iLevel )%sendbufferFromFiner%elemPos(iProc) )
      do iElem = 1, recvBuf_long( iProc )%nVals
        nesting = recvBuf_int( iProc )%val( iElem )
        call identify_halo( haloTreeID     = recvBuf_long( iProc )%val(iElem), &
          &                 elemPos        = elemPos,                          &
          &                 halolevel      = haloLevel,                        &
          &                 levelDesc      = levelDesc,                        &
          &                 nesting        = nesting,                          &
          &                 updated        = updated,                          &
          &                 tree           = tree,                             &
          &                 minLevel       = tree%global%minLevel,             &
          &                 computeStencil = computeStencil )                          
        if( elemPos .gt. 0 ) then
          if( nesting .lt. nestingLimit .and.                                  &
            &  levelDesc( iLevel )%elem%eType%val( elemPos )                   &
            &                                 == eT_ghostFromCoarserElem ) then
            ! identify all the compute neighbors of the current element
            call identify_stencilNeigh( iElem          = elemPos,              &
              &                         iLevel         = iLevel,               &
              &                         tree           = tree,                 &
              &                         iStencil       = 1,                    &
              &                         pathFirst      = pathFirst,            &
              &                         pathLast       = pathLast,             &
              &                         levelDesc      = levelDesc,            &
              &                         prc            = prc,                  &
              &                         computeStencil = computeStencil,       &
              &                         nesting        = nesting + 1 )
          end if
 
          if( recvBuf_int( iProc )%val( iElem ) .lt.                           &
            &   levelDesc( haloLevel )%elem%haloNesting%val( elemPos )) then
            levelDesc( haloLevel )%elem%needsUpdate%val( elemPos ) = .true.
            levelDesc( haloLevel )%elem%haloNesting%val( elemPos ) =           &
               min( recvBuf_int( iProc )%val( iElem ),                         &
                    levelDesc( haloLevel )%elem%haloNesting%val( elemPos ) )
          end if
          ! only add, if the element was added locally
          select case( levelDesc( iLevel )%elem%eType%val( elemPos ) ) 
          ! Depending on the type of the element, add to the 
          ! regular buffer, bufferFromCoarser, bufferFromFiner
          case( eT_fluidElem )
            call append( me  = levelDesc( iLevel )%sendbuffer%elemPos( iProc ),&
              &          val = elemPos )
          case( eT_ghostFromCoarserElem ) 
            call append( me  = levelDesc( iLevel )%sendbufferFromCoarser       &
              &                                             %elemPos( iProc ), &
              &          val = elemPos )

            if( levelDesc( haloLevel )%elem%haloNesting%val( elemPos )         &
              &                                        .lt. nestingLimit ) then
              call create_allParentNeighbors(                                  &
                &      targetElem     = elemPos,                               &
                &      nesting        = levelDesc( haloLevel )                 &
                &                            %elem%haloNesting%val( elemPos ), &
                &      level          = iLevel,                                &
                &      tree           = tree,                                  &
                &      computeStencil = computeStencil,                        &
                &      levelDesc      = levelDesc,                             &
                &      pathFirst      = pathFirst,                             &
                &      pathLast       = pathLast,                              &
                &      prc            = prc )
            end if

          case( eT_ghostFromFinerElem )
            call append( me  = levelDesc( iLevel )%sendbufferFromFiner         &
              &                %elemPos( iProc ),                              &
              &          val = elemPos )
          case( eT_distributedGhostFromFinerElem )
             call tem_write(' Found distributed ghost From Finer in request'// &
               &            ' remote Halos')
             call tem_write(' This case should not occur!')
             call tem_abort()
          end select
        end if
      end do
      levelDesc( iLevel )%sendbuffer%nElemsProc( iProc )                       &
        &           = levelDesc( iLevel )%sendbuffer%elemPos( iProc )%nVals 
      levelDesc( iLevel )%sendbufferFromCoarser%nElemsProc( iProc )            &
        &          = levelDesc( iLevel )%sendbufferFromCoarser                 &
        &                               %elemPos( iProc )%nVals 
      levelDesc( iLevel )%sendbufferFromFiner%nElemsProc( iProc )              &
        &          = levelDesc( iLevel )%sendbufferFromFiner                   &
        &                               %elemPos( iProc )%nVals 
    end do
    ! Now each Process knows, which elements to send to others
    !---------------------------------------------------------------------
    write(logUnit(5),*) 'Finished requesting remote halos'

  end subroutine request_remoteHalos
!******************************************************************************!


!******************************************************************************!
  !> Report the actually existing elements, which were requested as halos
  !! from remote
  !! 
  subroutine return_haloCounts( sendbuffer, recvbuffer, prc )
    !---------------------------------------------------------------------------
    !>
    type(tem_communication_type), intent(in) :: sendbuffer
    !>
    type(tem_communication_type), intent(inout) :: recvbuffer
    !>
    type(tem_comm_env_type), intent(in) :: prc
    !---------------------------------------------------------------------------
    integer :: iProc ! process iterator
    integer :: iErr  ! error flag
    integer, parameter :: nVals        = 1 ! number of values to comm. = 1
    integer, parameter :: message_tag = 23 ! arbitrary message flag
    integer, allocatable :: rq_handle(:)
    integer, allocatable :: status(:,:)
    integer :: nCommunications ! amount of communciations
    !---------------------------------------------------------------------------

    nCommunications = sendbuffer%nProcs + recvbuffer%nProcs
    allocate( rq_handle( nCommunications ) )
    allocate( status( mpi_status_size, nCommunications ) )
    !---------------------------------------------------------------------
    !! first report the number of actual halos
    rq_handle(:) = MPI_REQUEST_NULL

    do iProc = 1, recvbuffer%nProcs
      ! Receive the number of actual elements
      call mpi_irecv( recvbuffer%nElemsProc( iProc ),                          &
       &              nVals,                                                   &
       &              mpi_integer,                                             &
       &              recvbuffer%proc(iProc),                                  &
       &              message_tag,                                             &
       &              prc%comm,                                                &
       &              rq_handle(iProc),                                        &
       &              iErr  )
    end do ! iProc

    ! I send the number of actual existing halo elements to the requester.
    do iProc = 1, sendbuffer%nProcs
      call mpi_isend( sendbuffer%nElemsProc( iProc ),                          &
       &              nVals,                                                   &
       &              mpi_integer,                                             &
       &              sendbuffer%proc( iProc ),                                &
       &              message_tag,                                             &
       &              prc%comm,                                                &
       &              rq_handle(iProc + recvbuffer%nProcs),                    &
       &              iErr )
    end do ! iProc

    call mpi_waitall( nCommunications, rq_handle, status, iErr) 
    ! Now we know the number of how actually existent cells, to receive
    ! allocate the recv buffers and communicate
    !---------------------------------------------------------------------

  end subroutine return_haloCounts 
!******************************************************************************!


!******************************************************************************!
  !> Check if additional communications have to be performed
  !! 
  subroutine check_additionalComm( levelDesc, prc, doAdditional, minlevel )
    !---------------------------------------------------------------------------
    !>
    integer, intent(in) :: minlevel
    !>
    type(tem_levelDesc_type), intent(inout) :: levelDesc(minlevel:)
    !>
    type(tem_comm_env_type), intent(in) :: prc
    !>
    logical, intent(out) :: doAdditional
    !---------------------------------------------------------------------------
    integer :: iError, iLevel
    logical :: do_local
    !---------------------------------------------------------------------------
    doAdditional = .false.
    do_local = .false.
    do iLevel = minlevel, ubound( levelDesc, 1 )
      do_local = ( do_local .or. (levelDesc( iLevel )%require%nVals .gt. 0))
    end do
    ! JUROPA work-around for crash in the mpi_allreduce
    call mpi_barrier( prc%comm, iError )
    call mpi_allreduce( do_local, doAdditional, 1, mpi_logical,                &
      &                 mpi_LOR, prc%comm, iError )
    if( doAdditional ) then
      write(dbgUnit(1),*)'Perform additional exchanges for required neighbors' 
    end if

  end subroutine check_additionalComm
!******************************************************************************!
 

!******************************************************************************!
  !> Report the actually existing elements, which were requested as halos
  !! from remote
  !! 
  subroutine redefine_halos( levelDesc, sendbuffer, recvbuffer, prc,           &
    &                        commPattern, computeStencil )
    !---------------------------------------------------------------------------
    !>
    type(tem_levelDesc_type), intent(inout) :: levelDesc
    !>
    type(tem_communication_type), intent(inout) :: sendbuffer, recvbuffer
    !>
    type(tem_comm_env_type), intent(in) :: prc
    !>
    type(tem_commPattern_type)    :: commPattern
    !>
    type(tem_stencilHeader_type)  :: computeStencil(:)
    !---------------------------------------------------------------------------
    integer :: iProc, tPos, iElem, nElems
    integer :: elemCount
    integer(kind=long_k), allocatable :: recvBuf(:)
    integer :: message_flag
    logical :: wasAdded
    integer :: iStencil
    !---------------------------------------------------------------------------
    message_flag = 3

    if( allocated( sendbuffer%buf_long)) deallocate( sendbuffer%buf_long )
    allocate( sendbuffer%buf_long( sendbuffer%nProcs ))

    if( allocated( recvbuffer%buf_long)) deallocate( recvbuffer%buf_long )
    allocate( recvbuffer%buf_long( recvbuffer%nProcs ))
    if( allocated( recvBuffer%elemPos)) deallocate( recvBuffer%elemPos)
    allocate( recvbuffer%elemPos( recvbuffer%nProcs ))

    allocate( recvBuf(sum(recvbuffer%nElemsProc)) )

    elemCount = 0
    do iProc = 1, recvbuffer%nProcs
      nElems = recvbuffer%nElemsProc(iProc)
      call init(  me     = recvBuffer%elemPos( iProc ),                        &
        &         length = nElems )
      do iElem=1,nElems
        elemCount = elemCount + 1
        call append( me  = recvbuffer%elemPos( iProc ),                        &
          &          val = elemCount )
      end do
    end do

    if (tem_logging_isActive(main_debug%logger, 10)) then
      write(dbgUnit(0),*) '  SEND BUFFER '
      call tem_comm_dumpType( me = sendbuffer,    &
        &                     nUnit = dbgUnit(10) )
      write(dbgUnit(0),*) ' '
      write(dbgUnit(0),*) '  RECV BUFFER '
      call tem_comm_dumpType( me = recvbuffer,    &
        &                     nUnit = dbgUnit(10) )
    end if

    !---------------------------------------------------------------------
    ! initialize receive buffer
    do iProc = 1, recvbuffer%nProcs
      call commPattern%initBuf_long( me    = recvBuffer%buf_long( iProc ),     &
        &                            pos   = recvBuffer%elemPos( iProc )%val,  &
        &                            nVals = recvBuffer%nElemsProc( iProc ))
    end do

    ! initialize send buffer
    do iProc = 1, sendbuffer%nProcs
      call commPattern%initBuf_long( me    = sendBuffer%buf_long( iProc ),     &
        &                            pos   = sendBuffer%elemPos( iProc )%val,  &
        &                            nVals = sendBuffer%nElemsProc( iProc ))
    end do

    call commPattern%exchange_long( send         = sendbuffer,                 &
      &                             recv         = recvbuffer,                 &
      &                             state        = recvbuf,                    &
      &                             message_flag = message_flag,               &
      &                             send_state   = levelDesc%elem%tID%val )

    ! Now I received all the halos, which are really existing
    ! and required on the remote processes
    !---------------------------------------------------------------------

    !---------------------------------------------------------------------
    ! Add all received elements 
    ! Set the eType accordingly or leave as nonExisting if it was not received
    elemCount = 0
    do iProc = 1, recvbuffer%nProcs
      nElems = recvBuffer%nElemsProc( iProc )
      call commPattern%finBuf_long(me = recvbuffer%buf_long(iProc))
      call init(  me     = recvBuffer%elemPos( iProc ),                        &
        &         length = nElems )
      do iElem = 1, nElems
        elemCount = elemCount + 1
        call append( me         = levelDesc%elem,                              &
          &          tID        = recvBuf( elemCount ),                        &
          &          sourceProc = recvbuffer%proc( iProc )+1,                  &
          &          eType      = eT_haloElem,                                 &
          &          pos        = tPos,                                        &
          &          wasAdded   = wasAdded )
        if( wasAdded ) then
          ! New halo element received
        else
          ! If it was not added, it was there before.
          ! Then simply set the element type to halo (was initialized with
          ! eT_nonExistingElem above )
          levelDesc%elem%eType%val( tPos ) = eT_haloElem
        end if
        call append( me  = recvbuffer%elemPos( iProc ),                        &
          &          val = tPos )
      end do
      call commPattern%initBuf_long( me    = recvbuffer%buf_long( iProc ),     &
        &                            pos   = recvbuffer%elemPos( iProc )%val,  &
        &                            nVals = recvBuffer%nElemsProc( iProc ))
    end do

    deallocate(recvbuf)

    ! Allocate integer buffers
    if( allocated( recvbuffer%buf_int )) deallocate( recvbuffer%buf_int )
    allocate( recvbuffer%buf_int( recvbuffer%nProcs ))
    do iProc = 1, recvbuffer%nProcs
      call commPattern%initBuf_int( me    = recvbuffer%buf_int( iProc ),       &
        &                           pos   = recvbuffer%elemPos( iProc )%val,   &
        &                           nVals = recvBuffer%nElemsProc( iProc ))
    end do
    if( allocated( sendbuffer%buf_int )) deallocate( sendbuffer%buf_int )
    allocate( sendbuffer%buf_int( sendbuffer%nProcs ))
    do iProc = 1, sendbuffer%nProcs
      call commPattern%initBuf_int( me    = sendbuffer%buf_int( iProc ),       &
        &                           pos   = sendbuffer%elemPos( iProc )%val,   &
        &                           nVals = sendBuffer%nElemsProc( iProc ))
    end do
    ! added and initialized the receive buffers for extended properties of the
    ! halos
    !---------------------------------------------------------------------

    !---------------------------------------------------------------------
    ! communicate the properties of the halos
    call commPattern%exchange_long(                                            &
      &                         send         = sendbuffer,                     &
      &                         recv         = recvbuffer,                     &
      &                         state        = levelDesc%elem%property%val(:), &
      &                         message_flag = 5 )
    ! communicate the stencil neighbors
    ! the stencil includes the boundary information
    !>\todo HK: communicate all stencils, that are shared by all elements here.
    !!          See #1083.
    if ( recvbuffer%nProcs .ne. 0 .or. sendbuffer%nProcs .ne. 0 ) then
      nElems = levelDesc%elem%tID%nVals
      do iStencil = 1, size(computeStencil)
        if( .not. computeStencil( iStencil )%useAll ) exit
        call tem_stencil_communicate( send           = sendbuffer,             &
          &                           recv           = recvbuffer,             &
          &                           elem           = levelDesc%elem,         &
          &                           prc            = prc,                    &
          &                           commPattern    = commPattern,            &
          &                           computeStencil = computeStencil,         &
          &                           iStencil       = iStencil )
      end do
    end if

    write(logUnit(5),*) '  Done communicating stencil buffer'
    !---------------------------------------------------------------------

  end subroutine redefine_halos
!******************************************************************************!


!******************************************************************************!
  !> Communicate the complete stencil
  !!
  !! Currently, this assumes same stencils for all participating elements
  !!
  subroutine tem_stencil_communicate( send, recv, elem, computeStencil,        &
    &                                 prc, commPattern, iStencil )
    !---------------------------------------------------------------------------
    !>
    type( tem_communication_type ), intent(inout) :: send, recv
    !>
    type(tem_commPattern_type), intent(in)     :: commPattern
    !>
    type(tem_element_type), intent(inout)  :: elem
    !>
    type(tem_comm_env_type), intent(in) :: prc
    !>
    type(tem_stencilHeader_type), intent(in) :: computeStencil(:)
    !> amount of values to communicate
    integer, intent(in) :: iStencil
    !---------------------------------------------------------------------------
    integer :: iStencilElem, iElem, nElems
    integer(kind=long_k), allocatable :: buffer(:)
    integer, allocatable :: buffer_int(:)
    type(tem_stencilElement_type ) :: tStencil
    integer :: addedPos, iProc, elemPos, neighPos, stencilPos
    logical :: throwWarning, wasAdded
    !---------------------------------------------------------------------------
    ! Number of local elements in element list
    nElems = elem%tID%nVals
    ! reset warning flag
    throwWarning = .false.

    allocate( buffer( nElems ))
    allocate( buffer_int( nElems ))
!    !---------This is not needed for useAll stencils---------------------------
!    do iElem = 1, nElems
!      ! First check, if there is already a stencil
!      if( elem%stencil%val( iElem )%nVals .lt. 1) then
!        buffer_int( iElem ) = 0
!      else
!        buffer_int( iElem ) = elem%stencil%val( iElem )%val( iStencil )%QQN 
!      end if
!    end do
!
!    ! Exchange the number of stencil entries
!    call commPattern%exchange_int( send = send, recv = recv,                &
!      &                            state = buffer_int, message_flag = 0 )
!
!    ! Set all elements which were received from proc
!    do iProc = 1, recv%nProcs
!      do iElem = 1, recv%nElemsProc( iProc )
!        elemPos = recv%elemPos( iProc )%val( iElem )
!        call init( me = tStencil, QQN = buffer_int( elemPos ), headerPos = iStencil ) 
!        call append( me = elem%stencil%val( elemPos ), val = tStencil )
!      end do
!    end do
!    !----------This is not needed for useAll stencils--------------------------
    ! Set the temporary (empty) stencil
    call init( me        = tStencil,                                           &
      &        QQN       = computeStencil(iStencil)%QQN ,                      &
      &        headerPos = iStencil ) 
    do iProc = 1, recv%nProcs
      do iElem = 1, recv%nElemsProc( iProc )
        elemPos = recv%elemPos( iProc )%val( iElem )
        if( elem%stencil%val( elemPos )%nVals .gt. 0 ) then
          throwWarning = .true.
          ! Find the stencil corresponding to the current one (iStencil )
          stencilPos = tem_stencil_getHeaderPos(                               &
            &                               me  = elem%stencil%val( elemPos ), &
            &                               val = iStencil )
          ! if none was found, use the most appropriate one (headerPos = 0)
          if( stencilPos .le. 0 ) then
            stencilPos = tem_stencil_getHeaderPos(                             &
              &                             me  = elem%stencil%val( elemPos ), &
              &                             val = 0 )
            elem%stencil%val( elemPos )%val( stencilPos )%headerPos = iStencil
          end if
        else
          call append( me = elem%stencil%val( elemPos ), val = tStencil )
        end if
      end do
    end do

    !--------------------------------------------------------------------------
    ! Do the communication for each stencil tID entry
    do iStencilElem = 1, computeStencil(iStencil)%QQN
      ! Fill the send buffer
      do iProc = 1, send%nProcs
        do iElem = 1, send%nElemsProc( iProc )
          elemPos = send%elemPos( iProc )%val( iElem )
          neighPos = elem%stencil%val( elemPos )%val( iStencil )%              &
            &                                            tIDpos( iStencilElem )
          buffer( elemPos ) = elem%neighID%val( elemPos )%val( neighPos )
        end do
      end do
      ! Do the exchange between all participating processes
      call commPattern%exchange_long( send         = send,                     &
        &                             recv         = recv,                     &
        &                             state        = buffer,                   &
        &                             message_flag = 0 )
      ! Read from the buffer
      do iProc = 1, recv%nProcs
        do iElem = 1, recv%nElemsProc( iProc )
          elemPos = recv%elemPos( iProc )%val( iElem )
          ! Find the stencil corresponding to the current one (iStencil )
          stencilPos = tem_stencil_getHeaderPos(                               &
            &                               me  = elem%stencil%val( elemPos ), &
            &                               val = iStencil )

          if( stencilPos .gt. 0 ) then 
            ! Only append elements, if there is only one stencil for the element
            ! Append the treeID to the element's neighIDs
            call append( me       = elem%neighID%val( elemPos ),               &
              &          val      = buffer( elemPos),                          &
              &          pos      = addedPos,                                  &
              &          wasAdded = wasAdded )
            ! ... and store the appended position to the stencil
            elem%stencil%val( elemPos )%val( stencilPos )%                     &
              &                               tIDpos( iStencilElem ) = addedPos
          end if
        end do
      end do
    end do
    if (throwWarning) then
      write(logUnit(1),*) ' !!!  WARNING !!! :'
      write(logUnit(1),*) ' TreElm was asked to communicate more stencils. '
      write(logUnit(1),*) ' Only the first stencils are communicated.'
      write(logUnit(1),*) ' This can happen in the loop for recursively' &
        &              // ' identifying additional halo elements'
    end if
    
  end subroutine tem_stencil_communicate
!******************************************************************************!


!******************************************************************************!
  !> Update the found dependencies, which were built for non-ordered lists
  !! Out of fluid, ghost and halo lists, the totalList is constructed in an
  !! ordered fashion. The element order as in the TotalList is later passed on 
  !! to the solver.
  !!
  subroutine update_elemPosToTotalPos( levelDesc, levelPointer, tree,          &
    &                                  computeStencil )
    !---------------------------------------------------------------------------
    !>
    type(treelmesh_type), intent(in) :: tree
    !>
    integer, intent(inout) :: levelPointer(:)
    !>
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minlevel:)
    !>
    type(tem_stencilHeader_type), intent(inout) :: computeStencil(:)
    !---------------------------------------------------------------------------
    integer :: iLevel, iElem, iStencil, iStencilElem, pos
    integer :: nUpdates, iError
    integer(kind=long_k) :: nTreeID
    logical :: needUpdate
    !---------------------------------------------------------------------------
    write(logUnit(3),*) 'Updating dependencies ...'

    ! Update the levelPointer
    do iElem = 1, tree%nElems
      iLevel = tem_levelOf( tree%treeID( iElem ))
      levelPointer( iElem ) = tem_PositionInSorted(                            &
        &                      me    = levelDesc(iLevel )%total,               &
        &                      val   = tree%treeID( iElem ),                   &
        &                      upper = levelDesc( iLevel )%nElems_fluid )
    end do

    ! Update the element positions from the position in the original 
    ! elem list to the total list 
    do iLevel = tree%global%minlevel, tree%global%maxlevel
      call update_buffer_elemPos( buffer    = levelDesc( iLevel )%sendbuffer,  &
        &                         levelDesc = levelDesc( iLevel ),             &
        &                         iError    = iError )
      if (iError .gt. 0) &
        write(dbgUnit(1),*)  'error in sendbuffer '//trim(tem_toStr(iError)) &
        &                    // ' l' // trim(tem_toStr(iLevel)) 
      call update_buffer_elemPos( buffer    = levelDesc( iLevel )%recvbuffer,  &
        &                         levelDesc = levelDesc( iLevel ),             &
        &                         iError    = iError )
      if( iError .gt. 0 ) &
        write(dbgUnit(1),*) 'error in recvbuffer '//trim(tem_toStr(iError)) &
        &                    // ' l' // trim(tem_toStr(iLevel)) 
      call update_buffer_elemPos(                                              &
        &            buffer    = levelDesc( iLevel )%sendbufferFromCoarser,    &
        &            levelDesc = levelDesc( iLevel ),                          &
        &            iError    = iError )
      if( iError .gt. 0 ) &
        write(dbgUnit(1),*) 'error in sendFCbuff '//trim(tem_toStr(iError)) &
        &                    // ' l' // trim(tem_toStr(iLevel)) 
      call update_buffer_elemPos(                                              &
        &            buffer    = levelDesc( iLevel )%recvbufferFromCoarser,    &
        &            levelDesc = levelDesc( iLevel ),                          &
        &            iError    = iError )      
      if( iError .gt. 0 ) &
        write(dbgUnit(1),*) 'error in recvFCbuff '//trim(tem_toStr(iError)) &
        &                    // ' l' // trim(tem_toStr(iLevel)) 
      call update_buffer_elemPos(                                              &
        &            buffer    = levelDesc( iLevel )%sendbufferFromFiner,      &
        &            levelDesc = levelDesc( iLevel ),                          &
        &            iError    = iError )
      if( iError .gt. 0 ) &
        write(dbgUnit(1),*) 'error in sendFFbuff '//trim(tem_toStr(iError)) &
        &                    // ' l' // trim(tem_toStr(iLevel)) 
      call update_buffer_elemPos(                                              &
        &            buffer    = levelDesc( iLevel )%recvbufferFromFiner,      &
        &            levelDesc = levelDesc( iLevel ),                          &
        &            iError    = iError )
      if( iError .gt. 0 ) &
        write(dbgUnit(1),*) 'error in recvFFbuff '//trim(tem_toStr(iError)) &
        &                    // ' l' // trim(tem_toStr(iLevel)) 
    end do
    write(logUnit(4),*) 'Updating the stencil entries'
    do iStencil = 2, size( computeStencil )
      if( .not. computeStencil( iStencil )%useAll ) then
        do iElem = 1, computeStencil( iStencil )%nElems
          computeStencil( iStencil )%elem( iElem ) =                           &
            &   levelPointer( computeStencil( iStencil )%elem( iElem ))
        end do
      end if
    end do
    ! Update the totalPos in the element-stencils
    write(logUnit(4),*) ' Updating entries in the element stencils...'
    nUpdates =  0
    do iLevel = tree%global%minlevel, tree%global%maxlevel
      do iElem = 1, levelDesc(iLevel)%elem%tID%nVals
        do iStencil = 1, levelDesc( iLevel )%elem%stencil%val( iElem )%nVals
          do iStencilElem = 1, levelDesc( iLevel )%elem%stencil%               &
            &                                  val( iElem )%val( iStencil )%QQN
            needUpdate = .false.
            pos = levelDesc( iLevel )%elem%stencil%val( iElem )%               &
              &                        val( iStencil )%totalPos( iStencilElem )
            if( pos .gt. 0 ) then
              nTreeID = levelDesc( iLevel )%elem%tID%val( pos )
              if( pos .gt. levelDesc( iLevel )%nElems ) needUpdate = .true.
              if( .not. needUpdate ) then
                if( levelDesc( iLevel )%total( pos ) .ne. nTreeID )            &
                  &                                         needUpdate = .true.
              end if
              if( needUpdate ) then
                ! Totalpos needs update!
                nUpdates = nUpdates + 1
                levelDesc( iLevel )%elem%stencil%val( iElem )%                 &
                  &                   val( iStencil )%totalPos( iStencilElem ) &
                  &  = tem_treeIDinTotal( nTreeID, levelDesc( iLevel ))
              end if
            end if
          end do
        end do
      end do
    end do
    write(logUnit(3),*) '     Updated nEntries = ', nUpdates
    write(logUnit(3),*) ' Finished updating dependencies'

   end subroutine update_elemPosToTotalPos
!******************************************************************************!


!******************************************************************************!
  !> Update the position of the elements inside the buffers from the original
  !! tID list to the later totalList
  !!
  subroutine update_buffer_elemPos( buffer, levelDesc, iError )
    !---------------------------------------------------------------------------
    !>
    type(tem_communication_type), intent(inout) :: buffer
    !>
    type(tem_levelDesc_type), intent(inout) :: levelDesc
    !>
    integer, intent(out) :: iError
    !---------------------------------------------------------------------------
    integer :: iElem, iProc, elemPos, iVal
    !---------------------------------------------------------------------------
    iError = 0
    do iProc = 1, buffer%nProcs
      do iElem = 1, buffer%elemPos( iProc )%nVals
        elemPos = buffer%elemPos( iProc )%val( iElem ) 
        buffer%elemPos( iProc )%val( iElem ) =                                 &
        &     tem_treeIDinTotal( tID       = levelDesc%elem%tID%val( elemPos ),&
        &                        eType     = levelDesc%elem%eType%val(elemPos),&
        &                        levelDesc = levelDesc )
        if( buffer%elemPos( iProc )%val( iElem ) .lt. 1 ) then
          write(dbgUnit(2),*) ' tID ' &
            &                // trim(tem_toStr(levelDesc%elem &
            &                                           %tID%val(eleMPos)))   &
            &                // ' in buffer not found' 
          iError = iProc*10000000+iElem
          do iVal = 1, levelDesc%elem%tid%nVals
            if( levelDesc%elem%tid%val( iVal ) .eq. levelDesc%elem%tID%        &
              &                                            val( eleMPos )) then
              write(dbgUnit(2),*)'found at' // trim(tem_toStr(iVal)) 
            end if
          end do
          do iVal = 1, size(levelDesc%total)
            if( levelDesc%total( iVal ) .eq. levelDesc%elem%tID%               &
              &                                            val( eleMPos )) then
              write(dbgUnit(2),*) 'found in total at' // trim(tem_toStr(iVal)) 
            end if
          end do
        end if
      end do
    end do
 
  end subroutine update_buffer_elemPos
!******************************************************************************!


!******************************************************************************!
  !> add here the dependency for interpolation between the levels
  !! For each target cell, there are one or more source cells.
  !! The source cell can be of type fluid, ghost or halo.
  !! We save the type to update the correct element position later on,
  !! when the lists have been assembled.
  !!
  subroutine appendGhostDependency( sourcePos, sourceLevel, tgtDep )
    !---------------------------------------------------------------------------
    !> position of the source cell within its list inside the
    integer, intent(in) :: sourcePos
    !> level of the source ghost cell
    integer, intent(in) :: sourceLevel
    !> array of level descriptors which holds the informations
    !! for all the levels. this methods adds new informations to
    !! levelDesc(:)%source -> inout
    type(depSource_type), intent(inout) :: tgtDep
    !---------------------------------------------------------------------------

    ! check if the current dependency list already contains information
    if( tgtDep%elem%nVals .eq. 0 ) then
      call init( me=tgtDep%elem, length = 8 )
      tgtDep%dependencyLevel = sourceLevel
    end if

    ! append the new entry to the dependency list at the end+1
    call append( me  = tgtDep%elem,                                            &
         &       val = sourcePos )

  end subroutine appendGhostDependency
!******************************************************************************!


!******************************************************************************!
  !> Convert a non-zero stencil direction {-1,1} to child coordinate {0,1}
  !!
  function stencilToChild( stencilCoord ) result( childCoord)
    !---------------------------------------------------------------------------
    integer, intent(in) :: stencilCoord
    integer :: childCoord
    !---------------------------------------------------------------------------

    childCoord = (stencilCoord + 1)/2

  end function stencilToChild
!******************************************************************************!


!******************************************************************************!
  !> Convert a child coordinate {0,1} to non-zero stencil direction {-1,1}
  !!
  function childToStencil( childCoord ) result( stencilCoord)
    !---------------------------------------------------------------------------
    integer, intent(in) :: childCoord
    integer :: stencilCoord
    !---------------------------------------------------------------------------

    stencilCoord = (childCoord*2) -1

  end function childToStencil
!******************************************************************************!


!******************************************************************************!
  !> Update the link into a given direction, based on the childs neighbor
  !! relations.
  !! Define here the trumping rule to decide, which of the neighbors or
  !! boundarie is taken for the ghostFromFiner element
  !!
  subroutine update_childNeighborID( neighID, childCoord, childPos,            &
    &                                iStencilElem, levelDesc, iStencil)
    !---------------------------------------------------------------------------
    !>
    integer(kind=long_k),intent(inout) :: neighID
    !>
    integer, intent(in) :: childCoord(4)
    !>
    integer, intent(in) :: childPos(8)
    !>
    integer, intent(in) :: iStencilElem
    !>
    integer, intent(in) :: iStencil
    !>
    type(tem_levelDesc_type), intent(in) :: levelDesc
    !---------------------------------------------------------------------------
    integer :: childID, stencilTIDpos
    !---------------------------------------------------------------------------

    childID = int(tem_idOfCoord( childCoord ))
    ! childPos holds the positions of the child treeIDs in levelDesc%elem
    ! only if the current child element really exists, get information from its
    ! stencil
    if (childPos(childID) .gt. 0) then
      ! if child at current position exists,
      ! take the max tID neighbor parent or the
      ! min boundaryID -> trumping rule (bIDs are stored as negative integers)
      stencilTIDpos = levelDesc%elem%stencil%val( childPos( childID ))%val(1)  &
        &            %tIDpos( iStencilElem )
      neighID = max( neighID,                                                  &
        & tem_parentOf( levelDesc%elem%neighID%val( childPos( childID ))%val(  &
        &            stencilTIDpos )))
    else
    end if
  end subroutine update_childNeighborID
!******************************************************************************!


!******************************************************************************!
  !> write out the complete list of elements of a given level
  subroutine tem_elemList_dump( me, nUnit, string, stencil, stencilPos, &
    &                           compact )
    !---------------------------------------------------------------------------
    type( tem_element_type ), intent(in) :: me
    integer, intent(in) :: nUnit
    character(len=*), intent(in) :: string
    logical, intent(in), optional :: stencil
    logical, intent(in), optional :: stencilPos
    logical, intent(in), optional :: compact
    !---------------------------------------------------------------------------
    integer :: iElem
    !---------------------------------------------------------------------------
    write(nUnit, *) '========================================================='
    write(nUnit, *) '==               ', trim(string)
    write(nUnit, *) '========================================================='
    do iElem = 1, me%tID%nVals
      call tem_element_dump( me         = me,                                  &
        &                    elemPos    = iElem,                               &
        &                    nUnit      = nUnit,                               &
        &                    header     = (mod(iElem,32)==1),                  &
        &                    compact    = compact,                             &
        &                    stencil    = stencil,                             &
        &                    stencilPos = stencilPos)
    end do
    write(nUnit, *) '==  DONE         ', trim(string)
    write(nUnit, *) '========================================================='

  end subroutine tem_elemList_dump
!******************************************************************************!


!******************************************************************************!
  !> write out the complete list of elements of a given level
  !!
  subroutine tem_require_dump( me, nUnit, string, compact )
    !---------------------------------------------------------------------------
    type( dyn_longArray_type ), intent(in) :: me
    integer, intent(in) :: nUnit
    character(len=*), intent(in) :: string
    logical, intent(in), optional :: compact
    !---------------------------------------------------------------------------
    integer :: iElem
    integer :: nEntries, iLoop, iEntry
    character(len=pathLen) :: buffer
    !---------------------------------------------------------------------------
    nEntries = 15
    write(nUnit, *) '========================================================='
    write(nUnit, *) '==      REQUIRE  ', trim(string)
    write(nUnit, *) '========================================================='
    iElem = 0
    do iLoop = 1, ceiling( real(me%nVals)/real(nEntries) )
      buffer = ''
      do iEntry = 1, nEntries
        iElem = iElem + 1
        if( iElem .ge. me%nVals ) exit
        write(buffer, '(a, i10)') trim(buffer), me%val( iElem )
      end do
      write(nUnit,*) trim(buffer)
      if( iElem .ge. me%nVals ) exit
    end do
    write(nUnit, *) '==  DONE         ', trim(string)
    write(nUnit, *) '========================================================='

  end subroutine tem_require_dump
!******************************************************************************!


!******************************************************************************!
  !> output all level-wise treeIDs in a clean way
  !!
  subroutine tem_dumpTreeIDlists( levelDesc, levelPointer, minlevel )
    !---------------------------------------------------------------------------
    !> minimum level of the global tree
    integer, intent(in) :: minlevel
    !> level pointer from the original tree to the level wise total list
    integer, intent(in) :: levelPointer(:)
    !> level descriptor
    type(tem_levelDesc_type), intent(in) :: levelDesc(minlevel:)
    !---------------------------------------------------------------------------
    integer :: iLevel
    !---------------------------------------------------------------------------
    write(dbgUnit(2),*) '' 
    write(dbgUnit(2),*) '' 
    write(dbgUnit(2),*) '================================================' &
      &                // '=================' 
    write(dbgUnit(2),*) '=======            A L L  TreeID  L I S T S     ' &
      &                // '      ==========' 
    write(dbgUnit(2),*) '================================================' &
      &                // '=================' 
    do iLevel = minLevel, minLevel+size(levelDesc)-1

      if (size(levelDesc( iLevel )%total) .gt. 0) then
        write(dbgUnit(3),*) 'Level ' // trim(tem_toStr(iLevel))  
        write(dbgUnit(3),*) '==============================================' &  
          &                // '===================' 
        write(dbgUnit(3),*) 'nVals in total list ' &  
          &                // trim(tem_toStr(size(levelDesc(iLevel)%total))) 
!      call tem_print_array( title = 'total',         &
!                        me = levelDesc(iLevel)%total,&
!                        itemLength=8,                &
!                        lineLength = 128 )
!
!      if( levelDesc( iLevel )%fluid%nvals .gt. 0 ) &
!      call tem_print_array( title = 'fluid',         &
!                        me = levelDesc( iLevel )%fluid%val(          &
!                                1:levelDesc( iLevel )%fluid%nVals ), &
!                        itemLength=8,                                &
!                        lineLength = 128 )
!
!      if( levelDesc( iLevel )%pntTID%nvals .gt. 0 ) &
!      call tem_print_array( title = 'pntTID',                                           &
!                        me = int( levelDesc( iLevel )%pntTID%val(                &
!                                     1:levelDesc( iLevel )%pntTID%nvals ), kind=long_k),         &
!                        itemLength=8,                                               &
!                        lineLength = 128 )
!
!      if( levelDesc( iLevel )%ghostFromCoarser%nvals .gt. 0 ) &
!      call tem_print_array( title = 'ghostFromCoarser', &
!                        me = levelDesc( iLevel )%ghostFromCoarser%val(   &
!                                1:levelDesc( iLevel )%ghostFromCoarser%nvals),&
!                        itemLength=8,     &
!                        lineLength = 128 )
!      if( levelDesc( iLevel )%ghostFromFiner%nvals .gt. 0 ) &
!      call tem_print_array( title = 'ghostFromFiner', &
!                        me = levelDesc( iLevel )%ghostFromFiner%val(   &
!                                1:levelDesc( iLevel )%ghostFromFiner%nvals),&
!                        itemLength=8,     &
!                        lineLength = 128 )
!      if( levelDesc( iLevel )%halo%nvals .gt. 0 ) &
!      call tem_print_array( title = 'halo',                                            &
!                          me = levelDesc( iLevel )%halo%val(                      &
!                                  1:levelDesc( iLevel )%halo%nVals ),                &
!                          itemLength=8,                                               &
!                          lineLength = 128 )
      endif
    enddo ! iLevel

  end subroutine tem_dumpTreeIDlists
!******************************************************************************!


!******************************************************************************!
  !> output the found dependencies
  !!
  !! \todo: DEB: Adapt this subroutine or kick it out
  !!
  subroutine tem_debugDependencies( levelDesc, levelPointer, minlevel )
    !---------------------------------------------------------------------------
    !> minimal level
    integer, intent(in) :: minlevel
    !> level descriptor
    type(tem_levelDesc_type), intent(in) :: levelDesc(minlevel:)
    !> level pointer
    integer, intent(in) :: levelPointer(:)
    !---------------------------------------------------------------------------
    integer :: iLevel
    !---------------------------------------------------------------------------
!  ! Debugging
!  write(dbgUnit,*)
!  write(dbgUnit,*)
!  write(dbgUnit,*) '================================================================='
!  write(dbgUnit,*) '=======            A L L   L I S T S                   =========='
!  write(dbgUnit,*) '================================================================='
!  do iLevel = minlevel, ubound( levelDesc, 1 )
!
!    if(size(levelDesc( iLevel )%total) .gt. 0)then
!    write(dbgUnit,'(a,i3)') 'Level',iLevel
!    write(dbgUnit,'(a)') &
!   '===================================================================='
!    write(dbgUnit,'(a,i8)') 'nvals',size(levelDesc( iLevel )%total)
!    write(dbgUnit,*) '---------------------------------------------------------------'
!    write(dbgUnit,'(a)') 'TOTAL:'
!
!    write(dbgUnit,*) '---------------------------------------------------------------'
!    !write(dbgUnit,'(a14,512i8)') 'pntTID:',&
!    !            levelDesc( iLevel )%pntTID%val(1:levelDesc( iLevel )%pntTID%nvals)
!    write(dbgUnit,'(a14,512i8)') 'fluid:',levelDesc( iLevel )%totalPnt(   &
!                1:levelDesc( iLevel )%nElems_fluid)
!    write(dbgUnit,*) '---------------------------------------------------------------'
!    if( levelDesc( iLevel )%ghostFromCoarser%nvals .gt. 0) then
!      write(dbgUnit,'(a14,512i8)') 'ghostFC:',levelDesc( iLevel )%totalPnt(   &
!                1:levelDesc( iLevel )%nElems_ghostFromCoarser)
!      write(dbgUnit,*) &
!          '------------------------------------------------------------------'
!    endif
!    if( levelDesc( iLevel )%ghostFromFiner%nvals .gt. 0) then
!      write(dbgUnit,'(a14,512i8)') 'ghostFF:',levelDesc( iLevel )%totalPnt(   &
!                1:levelDesc( iLevel )%nElems_ghostFromFiner)
!      write(dbgUnit,*) &
!          '------------------------------------------------------------------'
!    endif
!    if( levelDesc( iLevel )%halo%nvals .gt. 0) then
!      write(dbgUnit,'(a14,512i8)') 'halo:',levelDesc( iLevel )%halo%   &
!                val(1:levelDesc( iLevel )%halo%nvals)
!      write(dbgUnit,*) &
!          '------------------------------------------------------------------'
!    endif
! endif
! enddo
!

  end subroutine tem_debugDependencies
!******************************************************************************!


!******************************************************************************!
  !> output all level-wise PropertyBits in a clean way
  !!
  subroutine tem_dumpPropBits( levelDesc, minlevel )
    !---------------------------------------------------------------------------
    !> minimum level of the global tree
    integer, intent(in) :: minlevel
    !> level descriptor
    type(tem_levelDesc_type), intent(in) :: levelDesc(minlevel:)
    !---------------------------------------------------------------------------
    integer :: iLevel
    !---------------------------------------------------------------------------
    write(dbgUnit(2),*)'' 
    write(dbgUnit(2),*)'' 
    write(dbgUnit(2),*) '================================================' &
      &                // '=================' 
    write(dbgUnit(2),*) '======        Property Bits Lists               ' &
      &                // '=================' 
    write(dbgUnit(2),*) '================================================' &
      &                // '=================' 
    do iLevel = minLevel, minLevel+size(levelDesc)-1

!      if(size(levelDesc( iLevel )%ttlProp) .gt. 0)then
!      write(dbgUnit,'(a,i3)') 'Level',iLevel
!      write(dbgUnit,'(a)') &
!     '===================================================================='
!      write(dbgUnit,'(a,i8)') 'nvals in total ',size(levelDesc( iLevel )%ttlProp)
!      call tem_print_array( title = 'total property',  &
!                        me = levelDesc(iLevel)%ttlProp,&
!                        itemLength=8,                &
!                        lineLength = 128 )
!
!      if( levelDesc( iLevel )%fldProp%nvals .gt. 0 ) &
!      call tem_print_array( title = 'fluid Property',         &
!                        me = levelDesc( iLevel )%fldProp%val(          &
!                                1:levelDesc( iLevel )%fluid%nVals ), &
!                        itemLength=8,                                &
!                        lineLength = 128 )

!    write(dbgUnit,'(a14,512i8)') 'sort fluid:',levelDesc( iLevel )%fluid%sorted(  &
!                1:levelDesc( iLevel )%fluid%nvals)
!    write(dbgUnit,'(a14,512i8)') 'invsort fluid:',levelDesc( iLevel )%invSorted(   &
!                1:levelDesc( iLevel )%fluid%nvals)
!      if( levelDesc( iLevel )%ghostFromCoarser%nvals .gt. 0 ) &
!      call tem_print_array( title = 'ghostFromCoarser', &
!                        me = levelDesc( iLevel )%ghostFromCoarser%val(   &
!                                1:levelDesc( iLevel )%ghostFromCoarser%nvals),&
!                        itemLength=8,     &
!                        lineLength = 128 )
!      if( levelDesc( iLevel )%ghostFromFiner%nvals .gt. 0 ) &
!      call tem_print_array( title = 'ghostFromFiner', &
!                        me = levelDesc( iLevel )%ghostFromFiner%val(   &
!                                1:levelDesc( iLevel )%ghostFromFiner%nvals),&
!                        itemLength=8,     &
!                        lineLength = 128 )
!      if( levelDesc( iLevel )%halo%nvals .gt. 0 ) &
!      call tem_print_array( title = 'halo property',                  &
!                          me = levelDesc( iLevel )%haloProp%val(      &
!                                  1:levelDesc( iLevel )%halo%nVals ), &
!                          itemLength=8,                               &
!                          lineLength = 128 )
!      endif
    enddo ! iLevel

  end subroutine tem_dumpPropBits
!******************************************************************************!

end module tem_construction_module
!******************************************************************************!
!> \page tem_construction_example_stencil Example for the stencil construction
!! To build the stencil for a given element, all required 
!! \ref treelmesh_module::treelmesh_type::treeid "treeIDs" can be obtained with
!! the following procedure:
!! First the integer coordinate tuple \f$(x,y,z,L)\f$ of the
!! \ref treelmesh_module::treelmesh_type::treeid "treeID", for which the stencil
!! has to be obtained, is computed.
!! Then for each stencil element the 
!! \ref treelmesh_module::treelmesh_type::treeid "treeID" is obtained by the 
!! following steps:
!! - Add the stencil offset to the coordinate of the central element to obtain
!!   the coordinate \f$(x + s_x, y + s_y, z + s_z, L)\f$ for the stencil element
!! - Convert the obtained coordinate back to a
!!   \ref treelmesh_module::treelmesh_type::treeid "treeID".
!! To illustrate this procedure with a specific example, lets consider a two
!! dimensional mesh.
!! We would like to find the right neighbor of the element with 
!! \ref treelmesh_module::treelmesh_type::treeid "treeID" \f$40\f$.
!! An illustration of the geometrical layout for this example is given in Figure
!! \ref{fig:rightNeighbor}.
!! As we are looking at a two dimensional mesh, there are \f$4^L\f$ elements on
!! each refinement level \f$L\f$.
!! \image html find_rightNeighbor.png
!! 
!! The first information, that is needed, is the level of the given 
!! \ref treelmesh_module::treelmesh_type::treeid "treeID".
!! To find this, and the Morton index on that level, we subtract the elements
!! on coarse levels, until the reduced index is smaller than the next number of
!! elements to subtract:
!! 
!! - Level 0: \f$40 - 4^0 = 39\f$
!! - Level 1: \f$39 - 4^1 = 35\f$
!! - Level 2: \f$35 - 4^2 = 19\f$
!! - Level 3: \f$4^3 > 19\f$
!! - \ref treelmesh_module::treelmesh_type::treeid "treeID" \f$40\f$ is located
!!   on Level 3 and has a Morton index of \f$19\f$.
!! 
!! The binary representation of \f$19\f$ is \f$010011_b\f$.
!! Due to the bit-interleaving rule, odd bits correspond to the X coordinate and
!! even bits correspond to the Y coordinate.
!! Thus we obtain for X \f$101_b = 5\f$ and for Y \f$001_b = 1\f$.
!! Now we found the coordinate tuple \f$(x, y, L) = (5, 1, 3)\f$ of
!! \ref treelmesh_module::treelmesh_type::treeid "treeID" \f$40\f$ and can
!! simply add the offset \f$(1, 0, 0)\f$, that represents the right neighbor.
!! By this we obtain the coordinate of this neighbor to be \f$(6, 1, 3)\f$,
!! resulting in a binary representation of \f$(110_b, 001_b)\f$.
!! After interleaving the bits, the Morton index of the right neighbor is found
!! to be \f$010110_b = 22\f$
!! Finally, by adding the offset, we obtain the 
!! \ref treelmesh_module::treelmesh_type::treeid "treeID" of the right neighbor
!! as \f$22 + 4^0 + 4^1 + 4^2 = 43\f$.
!! Note, that there is no restriction on the offsets, that are to be applied in
!! the neighborhood search, and due to the periodic universe all offset
!! locations are well defined.
