! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Module to provide simple growing data structures.
!! The dynamic arrays provided by this module are
!! capable of handling lists of values, which might
!! need to grow over time.
!! Removal of entries is not possible directly.
!! The complete module might be put into a CoCo Text
!! template, to create new modules of this object
!! for different types. For now, two different
!! templates are used for the declaration part and
!! the implementation part.
?? include 'arrayMacros.inc'
!!
module tem_grow_array_module

  ! include treelm modules
  use env_module, only: long_k, rk, minLength, zeroLength

  implicit none

  type intArray2d_type
    integer, allocatable :: val(:,:)
  end type

  type realVector3D_type
    real(kind=rk) :: component(3)
  end type

!-----------------------------------------------------------------
! Growing array (GA)
! tname ... indicates type of dynamic array (long, int, real, ...)

?? copy :: GA_decltxt(logical, logical)
?? copy :: GA_decltxt(long, integer(kind=long_k))
?? copy :: GA_decltxt(int, integer)
?? copy :: GA_decltxt(real, real(kind=rk))
?? copy :: GA_decltxt(dtInt2d, type(intArray2d_type))
?? copy :: GA_decltxt(realVec3D, type(realVector3D_type))

! Growing array
!-----------------------------------------------------------------


!-----------------------------------------------------------------
! Growing array (GA2d)
! tname ... indicates type of dynamic array (long, int, real, ...)
! 2d array, but only the second dimension can grow.
!
?? text :: GA2d_decltxt(tname, tstring)
  type grw_?tname?Array_type
    integer :: nVals = 0
    integer :: ContainerSize = 0
    integer :: ContainerWidth = 0
    ?tstring?, allocatable :: Val(:,:)
  end type

  !> Initialize the dynamic array
  interface init
    module procedure init_GA2d_?tname?
  end interface

  !> Destroy the dynamic array
  interface destroy
    module procedure destroy_GA2d_?tname?
  end interface

  !> Append a value to the dynamic array
  !! and return its position.
  interface append
    module procedure append_SingleGA2d_?tname?
    module procedure append_ArrayGA2d_?tname?
  end interface

  !> Increase the size of the container
  !! for the array.
  interface expand
    module procedure expand_GA2d_?tname?
  end interface

?? end text GA2d_decltxt

?? copy :: GA2d_decltxt(logical2d, logical)
?? copy :: GA2d_decltxt(long2d, integer(kind=long_k))
?? copy :: GA2d_decltxt(int2d, integer)
?? copy :: GA2d_decltxt(real2d, real(kind=rk))

! Growing array 2D
!-----------------------------------------------------------------




contains

!-----------------------------------------------------------------
! Growing array (GA)
! tname ... indicates type of dynamic array (long, int, real, ...)

?? copy :: GA_impltxt(logical, logical)
?? copy :: GA_impltxt(long, integer(kind=long_k))
?? copy :: GA_impltxt(int, integer)
?? copy :: GA_impltxt(real, real(kind=rk))
?? copy :: GA_impltxt(dtInt2d, type(intArray2d_type))
?? copy :: GA_impltxt(realVec3D, type(realVector3D_type))
! Growing array
!-----------------------------------------------------------------



!******************************************************************************!
!-----------------------------------------------------------------
! 2d Array, which can grow in second dimension only (GA2d)
! tname ... indicates type of dynamic array (long, int, real, ...)
?? text :: GA2d_impltxt(tname, tstring, invalidval)
  subroutine init_GA2d_?tname?(me, width, length)
    !---------------------------------------------------------------------------
    !> Dynamic Array to init
    type(grw_?tname?Array_type), intent(out) :: me
    !> width of the container
    integer, intent(in)           :: width
    !> Initial length of the container
    integer, intent(in), optional :: length
    !---------------------------------------------------------------------------

    me%ContainerWidth = width
    if (present(length)) me%ContainerSize = length

    allocate(me%Val(me%ContainerWidth,me%ContainerSize))
    !> Reset all values (only positive values are valid)
    me%val(:,:) = ?invalidval?

  end subroutine init_GA2d_?tname?
!******************************************************************************!


!******************************************************************************!
  !> Destroy the 2d growing array
  !!
  subroutine destroy_GA2d_?tname?(me)
    !---------------------------------------------------------------------------
    !> Dynamic Array to init 
    type(grw_?tname?Array_type), intent(inout) :: me
    !---------------------------------------------------------------------------

    deallocate(me%val)
    me%nvals  = 0
    me%ContainerSize = 0

  end subroutine destroy_GA2d_?tname?
!******************************************************************************!


!******************************************************************************!
  !> Append a single value to the growing 2D array.
  !!
  subroutine append_SingleGA2d_?tname?(me, val, pos1, pos2)
    !---------------------------------------------------------------------------
    !> Array to append the value to 
    type(grw_?tname?Array_type) :: me
    !> Value to append
    ?tstring?, intent(in) :: val
    !> Position in first dimension (cannot grow)
    integer, intent(in) :: pos1
    !> Position in Second dimension (can grow)
    integer, intent(in) :: pos2
    !---------------------------------------------------------------------------

    ! Value to append is larger than all existing ones,
    ! just put it to the end of the list, this captures
    ! also the case of empty lists.
    ! In this case FoundPos = me%nVals + 1 holds.
    if (pos2 > me%ContainerSize) then
      ! Expand the array, if its boundary is reached
      call expand(me)
    end if
    me%nVals = pos2
    me%val(pos1,pos2) = val

  end subroutine append_SingleGA2d_?tname?
!******************************************************************************!


!******************************************************************************!
  !> Append an array of values to the growing 2D array.
  !!
  subroutine append_ArrayGA2d_?tname?(me, val, pos )
    !---------------------------------------------------------------------------
    !> Array to append the value to
    type(grw_?tname?Array_type) :: me
    !> Array of values to append
    ?tstring?, intent(in) :: val(:)
    !> Position in Second dimension (can grow)
    integer, intent(in), optional :: pos
    !---------------------------------------------------------------------------
    integer :: newPos
    !---------------------------------------------------------------------------

    newPos = me%nVals + 1
    if( present( pos ) )  newPos = pos
    ! Value to append is larger than all existing ones,
    ! just put it to the end of the list, this captures
    ! also the case of empty lists.
    ! In this case FoundPos = me%nVals + 1 holds.
    if (newPos > me%ContainerSize) then
      ! Expand the array, if its boundary is reached
      call expand(me)
    end if
    me%nVals = newPos
    me%val( : , newPos ) = val(:)

  end subroutine append_ArrayGA2d_?tname?
!******************************************************************************!


!******************************************************************************!
  !> Expand the growing 2D array
  subroutine expand_GA2d_?tname?( me )
    !---------------------------------------------------------------------------
    !> Array to resize
    type(grw_?tname?Array_type) :: me
    !---------------------------------------------------------------------------
    ?tstring?, allocatable :: swpval(:,:)
    !---------------------------------------------------------------------------

    allocate(swpval(me%ContainerWidth,me%nVals))
    swpval(:,:) = me%Val(:,:me%nVals)
    deallocate(me%Val)

    me%ContainerSize = max(minLength, me%ContainerSize*2)
    allocate(me%Val(me%ContainerWidth,me%ContainerSize))
    !> First reset all entries
    me%Val(:,:) = ?invalidval?
    me%Val(:,:me%nVals) = swpval(:,:)
    deallocate(swpval)

  end subroutine expand_GA2d_?tname?
!******************************************************************************!

?? end text GA2d_impltxt

?? copy :: GA2d_impltxt(logical2d, logical, .false.)
?? copy :: GA2d_impltxt(long2d, integer(kind=long_k), -1_long_k)
?? copy :: GA2d_impltxt(int2d, integer, -1)
?? copy :: GA2d_impltxt(real2d, real(kind=rk), -1.0_rk)

! Growing array 2D
!-----------------------------------------------------------------


end module tem_grow_array_module
!******************************************************************************!
