! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Some generic matrix and vector function
!! 
module tem_math_module

  ! include treelm modules
  use env_module,       only: rk
  use tem_aux_module,   only: tem_abort
  use tem_tools_module, only: tem_write

  implicit none

  private

  public :: invert_matrix
  public :: matrix_type
  public :: tem_intp_trilinearReduced
  public :: cross_product3D

  !> This derived type encapsulates the definition of the used interpolation
  !! method
  type matrix_type
    !> inverted matrix to solve linear system of equation
    real(kind=rk), allocatable :: A(:,:)
    !> how many entries are in the 2d matrix?
    integer :: nEntries(2)       
  end type matrix_type

  interface tem_intp_trilinearReduced
    module procedure tem_intp_trilinearReduced_scal
    module procedure tem_intp_trilinearReduced_vect
  end interface

  contains

!******************************************************************************!
  !> Returns the inverse of a matrix calculated by finding the LU
  !! decomposition.  Depends on LAPACK.
  !!
  function invert_matrix(A) result(Ainv)
    !---------------------------------------------------------------------------
    !> Matrix to invert
    real(kind=rk), dimension(:,:), intent(in) :: A
    !> inverse of A
    real(kind=rk), dimension(size(A,1),size(A,2)) :: Ainv
    !---------------------------------------------------------------------------
    real(kind=rk), dimension(size(A,1)) :: work  ! work array for LAPACK
    integer, dimension(size(A,1)) :: ipiv   ! pivot indices
    integer :: n, info
    ! External procedures defined in LAPACK
    external DGETRF
    external DGETRI
    !---------------------------------------------------------------------------

    ! Store A in Ainv to prevent it from being overwritten by LAPACK
    Ainv = A
    n = size(A,1)

    ! DGETRF computes an LU factorization of a general M-by-N matrix A
    ! using partial pivoting with row interchanges.
    call DGETRF(n, n, Ainv, n, ipiv, info)

    if (info /= 0) then
      call tem_write( 'Matrix is numerically singular!' )
      call tem_abort
    end if

    ! DGETRI computes the inverse of a matrix using the LU factorization
    ! computed by DGETRF.
    call DGETRI(n, Ainv, n, ipiv, work, n, info)

    if (info /= 0) then
      call tem_write( 'Matrix inversion failed!' )
      call tem_abort
    end if
  end function invert_matrix
!******************************************************************************!


!******************************************************************************!
  !> This function returns the tri-linearly interpolated values from the seven
  !! source points to the target position located at targetCoord. 
  !! The source points are arranged in a square from (0,0,0)x(1,1,1)
  !! The order of the source points are according to the morton curve
  !!  1        2        3        4
  !! (0,0,0); (1,0,0); (0,1,0); (1,1,0)
  !!  5        6        7        
  !! (0,0,1); (1,0,1); (0,1,1)
  !! 
  function tem_intp_trilinearReduced_scal( srcVal, targetCoord ) result( phi )
    !---------------------------------------------------------------------------
    !> source values of the square corners
    real(kind=rk), intent(in) :: srcVal(7)
    !> interpolation location within the square
    real(kind=rk), intent(in) :: targetCoord(3)
    !> interpolated value
    real(kind=rk) :: phi
    !---------------------------------------------------------------------------
    real(kind=rk) :: phi_northFront, phi_southFront
    real(kind=rk) :: phi_northBack, phi_southBack
    real(kind=rk) :: phi_front, phi_back 
    !---------------------------------------------------------------------------
    ! Linear interpolation on the cube front side (z = 0 )
    phi_northFront = (1._rk - targetCoord(1))* srcVal(3)                       &
      &                     + targetCoord(1) * srcVal(4)
    phi_southFront = (1._rk - targetCoord(1))* srcVal(1)                       &
      &                     + targetCoord(1) * srcVal(2)
    ! Linear interpolation on the cube back side (z = 1 )
    phi_northBack  = srcVal(7) !(1._rk - targetCoord(1))* srcVal(7) ! + targetCoord(1)*srcVal(8)
    phi_southBack  = (1._rk - targetCoord(1))* srcVal(5)                       &
      &                     + targetCoord(1) * srcVal(6)
    ! Linear interpolation on the cube front side (z = 0 )
    phi_front = (1._rk - targetCoord(2))* phi_southFront                       &
      &                + targetCoord(2) * phi_northFront
    ! Linear interpolation on the cube back side (z = 1 )
    phi_back  = (1._rk - targetCoord(2))* phi_southBack                        &
      &                + targetCoord(2) * phi_northBack 
    phi       = (1._rk - targetCoord(3))* phi_front                            &
      &                + targetCoord(3) * phi_back 

  end function tem_intp_trilinearReduced_scal
!******************************************************************************!


!******************************************************************************!
  !> This function returns the tri-linearly interpolated values from the seven
  !! source points to the target position located at targetCoord. 
  !! The source points are arranged in a square from (0,0,0)x(1,1,1)
  !! The order of the source points are according to the morton curve
  !!  1        2        3        4
  !! (0,0,0); (1,0,0); (0,1,0); (1,1,0)
  !!  5        6        7        
  !! (0,0,1); (1,0,1); (0,1,1)
  !! 
  function tem_intp_trilinearReduced_vect( srcVal, targetCoord, nSize )        &
    &                                                             result( phi )
    !---------------------------------------------------------------------------
    !> vector size
    integer, intent(in) :: nSize
    !> source values of the square corners
    real(kind=rk), intent(in) :: srcVal(nSize, 7)
    !> interpolation location within the square
    real(kind=rk), intent(in) :: targetCoord(3)
    !> interpolated value
    real(kind=rk) :: phi( nSize )
    !---------------------------------------------------------------------------
    real(kind=rk) :: phi_northFront, phi_southFront
    real(kind=rk) :: phi_northBack, phi_southBack
    real(kind=rk) :: phi_front, phi_back 
    integer :: iSize
    !---------------------------------------------------------------------------
    do iSize = 1, nSize
      ! Linear interpolation on the cube front side (z = 0 )
      phi_northFront = (1._rk - targetCoord(1))* srcVal(iSize,3)               &
        &                     + targetCoord(1) * srcVal(iSize,4)
      phi_southFront = (1._rk - targetCoord(1))* srcVal(iSize,1)               &
        &                     + targetCoord(1) * srcVal(iSize,2)
      ! Linear interpolation on the cube back side (z = 1 )
      phi_northBack  = srcVal(iSize,7) !(1._rk - targetCoord(1))* srcVal(7) ! + targetCoord(1)*srcVal(8)
      phi_southBack  = (1._rk - targetCoord(1))* srcVal(iSize,5)               &
        &                     + targetCoord(1) * srcVal(iSize,6)
      ! Linear interpolation on the cube front side (z = 0 )
      phi_front = (1._rk - targetCoord(2))* phi_southFront                     &
        &                + targetCoord(2) * phi_northFront
      ! Linear interpolation on the cube back side (z = 1 )
      phi_back  = (1._rk - targetCoord(2))* phi_southBack                      &
        &                + targetCoord(2) * phi_northBack 
      phi( iSize ) = (1._rk - targetCoord(3))* phi_front                       &
        &                   + targetCoord(3) * phi_back 
    enddo

  end function tem_intp_trilinearReduced_vect
!******************************************************************************!


!******************************************************************************!
  !> This function calculate the cross product of two 3D vector
  !!
  FUNCTION cross_product3D(a, b) result( cross )
    !---------------------------------------------------------------------------
    !> resulting cross produkt
    real(kind=rk), DIMENSION(3) :: cross
    !> input vector a
    real(kind=rk), DIMENSION(3), INTENT(IN) :: a
    !> input vector b
    real(kind=rk), DIMENSION(3), INTENT(IN) :: b
    !---------------------------------------------------------------------------

    cross(1) = a(2) * b(3) - a(3) * b(2)
    cross(2) = a(3) * b(1) - a(1) * b(3)
    cross(3) = a(1) * b(2) - a(2) * b(1)

  END FUNCTION cross_product3D
!******************************************************************************!


end module tem_math_module
!******************************************************************************!
