! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Parameter lists for Treelm
!! 
!! This module holds the numbering of elements and subelements acoording 
!! to the treelm numbering (Morton-curve).
!!
!! \author Kannan Masilamani, Harald Klimach, Manuel Hasert, Jems Zudrop, Simon Zimny, ...
!!
module tem_param_module
  use env_module, only:rk

  implicit none

  integer,parameter :: nMaxLevels = 20 !< max refinement levels

  !> Neighboring convention
  integer,parameter :: qQQQ     = 26    !< number of directions

  integer,parameter :: q__W     = 1     !< west               x-
  integer,parameter :: q__S     = 2     !< south              y-
  integer,parameter :: q__B     = 3     !< bottom             z-
  integer,parameter :: q__E     = 4     !< east               x+
  integer,parameter :: q__N     = 5     !< north              y+
  integer,parameter :: q__T     = 6     !< top                z+
  integer,parameter :: q_BS     = 7     !< bottom south       z-,y-
  integer,parameter :: q_TS     = 8     !< top south          z+,y-
  integer,parameter :: q_BN     = 9     !< bottom north       z-,y+
  integer,parameter :: q_TN     = 10    !< top north          z+,y+
  integer,parameter :: q_BW     = 11    !< bottom west        x-,z-
  integer,parameter :: q_BE     = 12    !< bottom east        x+,z-
  integer,parameter :: q_TW     = 13    !< top west           x-,z+
  integer,parameter :: q_TE     = 14    !< top east           x+,z+
  integer,parameter :: q_SW     = 15    !< south west         y-,x-
  integer,parameter :: q_NW     = 16    !< north west         y+,x-
  integer,parameter :: q_SE     = 17    !< south east         y-,x+
  integer,parameter :: q_NE     = 18    !< north east         y+,x+
  integer,parameter :: qBSW     = 19    !< bottom south west  x-,y-,z-
  integer,parameter :: qBSE     = 20    !< bottom south east  x+,y-,z-
  integer,parameter :: qBNW     = 21    !< bottom north west  x-,y+,z-
  integer,parameter :: qBNE     = 22    !< bottom north east  x+,y+,z-
  integer,parameter :: qTSW     = 23    !< top south west     x-,y-,z+
  integer,parameter :: qTSE     = 24    !< top south east     x+,y-,z+
  integer,parameter :: qTNW     = 25    !< top north west     x-,y+,z+
  integer,parameter :: qTNE     = 26    !< top north east     x+,y+,z+

  integer, dimension(qQQQ),parameter  :: qInvDir =   (/                        &
    &     q__E, q__N, q__T, q__W, q__S, q__B,                                  &
    &     q_TN, q_BN, q_TS, q_BS,                                              &
    &     q_TE, q_TW, q_BE, q_BW,                                              &
    &     q_NE, q_SE, q_NW, q_SW,                                              &
    &     qTNE, qTNW, qTSE, qTSW,                                              &
    &     qBNE, qBNW, qBSE, qBSW    /)

  !> mapping the first six neihbor directions to x, y, z (1,2,3)
  integer, parameter :: qAxis(q__T) = [1,2,3,1,2,3]

  !> Spatial offset according to the directions above for q__W, ...
  integer, dimension(qQQQ,3),parameter  :: qOffset =                           &
  reshape((/-1, 0, 0, 1, 0, 0, 0, 0, 0, 0,-1,+1,-1,+1,-1,-1,+1,+1,-1,+1,-1,+1,-1,+1,-1,+1,      &
    &        0,-1, 0, 0, 1, 0,-1,-1,+1,+1, 0, 0, 0, 0,-1,+1,-1,+1,-1,-1,+1,+1,-1,-1,+1,+1,      &
    &        0, 0,-1, 0, 0, 1,-1,+1,-1,+1,-1,-1,+1,+1, 0, 0, 0, 0,-1,-1,-1,-1,+1,+1,+1,+1/),(/qQQQ,3/))

  !> direction names of q__W, ... very useful for debuggin
  character(len=3), parameter           :: qDirName(qQQQ) = [                  &
     &        '  W', '  S', '  B', '  E', '  N', '  T',                        &
     &        ' BS', ' TS', ' BN', ' TN', ' BW', ' BE',                        &
     &        ' TW', ' TE', ' SW', ' NW', ' SE', ' NE',                        &
     &        'BSW', 'BSE', 'BNW', 'BNE',                                      &
     &        'TSW', 'TSE', 'TNW', 'TNE'                                       &
     &        ]


!-----------------------------------------------------------------------!
  ! Children conventions
  ! Numbering of children, starting in the lower, left, bottom corner.
  !> Spatial position of the child relative to the parent barycenter
  !! in integer coordinates
  integer, dimension(8,3),parameter  :: childPosition =                  &
                                 reshape((/-1,+1,-1,+1,-1,+1,-1,+1,      &
                                   &       -1,-1,+1,+1,-1,-1,+1,+1,      &
                                   &       -1,-1,-1,-1, 1, 1, 1, 1/),(/8,3/))
  ! Transformation matrices for mapping the children
  integer, dimension(3,3,8),parameter  :: tem_rotationMatrix =             &
     reshape((/ 0,-1, 0,      &  ! 0 0 0
       &       -1, 0, 0,      &
       &        0, 0,-1,      &
       !        2nd entry in 3rd dimension
       &         1, 0, 0,     &  ! 1 0 0
       &         0,-1, 0,     &
       &         0, 0,-1,     &
       !        3rd entry in 3rd dimension
       &        -1, 0, 0,     &  ! 0 1 0
       &         0, 1, 0,     &
       &         0, 0,-1,     &
       !        4th entry in 3r!        4th entry in 3rd dimension
       &         0, 1, 0,     &!&         0, 0,-1,     &  ! 1 1 0
       &         1, 0, 0,     &!&         0, 1, 0,     &
       &         0, 0,-1,     &!&         1, 0, 0,     &
       !        5th entry in 3rd dimension
       &        -1, 0, 0,     &  ! 0 0 1
       &         0,-1, 0,     &
       &         0, 0, 1,     &
       !        6th entry in 3rd dimension
       &         0,-1, 0,     &!&         1, 0, 0,     &  ! 1 0 1
       &         1, 0, 0,     &!&         0, 0, 1,     &
       &         0, 0, 1,     &!&         0,-1, 0,     &
       !        7th entry in 3rd dimension
       &         0, 1, 0,     &  ! 0 1 1
       &        -1, 0, 0,     &
       &         0, 0, 1,     &
       !        8th entry  1 1 1
       &         1, 0, 0,     &  ! 1 1 1
       &         0, 1, 0,     &
       &         0, 0, 1      /),(/3,3,8/))

  ! MH: Why is this numbering scheme not according to the treelm general
  ! numbering?
  integer, dimension(7,3),parameter  :: parentNeighbor =                 &
                                 reshape((/ 1, 0, 0, 1, 0, 1, 1,         &
                                   &        0, 1, 0, 1, 1, 0, 1,         &
                                   &        0, 0, 1, 1, 1, 1, 0/),(/7,3/))
  ! map the apes numbering of the vertices to the one used by vtk 
  ! element type 12 (VTK_HEXAHEDRON), 8 vertices
  ! our numbering:                  vtk numbering:
  !   3       4                       4       3
  !   ---------                       ---------
  !   |       |                       |       |
  !   |       |                       |       |
  !   ---------                       ---------
  !   1       2                       1       2
  !
  !                                                     0 1 2 3 4 5 6 7
  integer, dimension(8), parameter :: vtk8_vrtxMap = (/ 1,2,4,3,5,6,8,7 /)

  ! map the apes numbering of the vertices to the one used by vtk
  ! element type 25 (VTK_QUADRATIC_HEXAHEDRON), 20 vertices
  integer, dimension(20), parameter :: vtk20_vrtxMap =                         &
  !                        0 1 2 3 4 5 6 7 8  9 10 11 12 13 14 15 16 17 18 19
    &                   (/ 1,2,6,5,3,4,8,7,9,15,17,13,12,16,20,14,10,11,19,18 /)
!
!------------------------------------------------------------------------------!
!
  ! State variable parameters
  integer, parameter :: prp_state     = 0         !< the mesoscopic state (pdfs)
  integer, parameter :: prp_density   = 1         !< fluid density
  integer, parameter :: prp_pressure  = 2         !< fluid pressure
  integer, parameter :: prp_wss       = 3         !< Wall Shear Stress
  integer, parameter :: prp_velMag    = 10        !< velocity magnitude
  integer, parameter :: prp_velX      = 11        !< velocity_X
  integer, parameter :: prp_velY      = 12        !< velocity_Y
  integer, parameter :: prp_velZ      = 13        !< velocity_Z


  ! Mathematical constants

  !> constant PI value
  real(kind=rk), parameter :: PI = 3.14159265358979323846_rk


  ! some other constants

  ! musubi related
  !> the speed of sound cs in various ways
  real(kind=rk), parameter :: cs       = 0.57735026918962576451_rk
  real(kind=rk), parameter :: csInv    = 1.73205080756887729352_rk
  real(kind=rk), parameter :: cs2inv   = 3._rk
  real(kind=rk), parameter :: cs2      = 1._rk/3._rk
  real(kind=rk), parameter :: cs4      = 1._rk/9._rk
  real(kind=rk), parameter :: t2cs4inv = 4.5_rk
  real(kind=rk), parameter :: t2cs2inv = 1.5_rk
  real(kind=rk), parameter :: cs4inv   = 9._rk

  ! mathematical constants
  real(kind=rk), parameter :: sqrt2      = 1.4142135623731_rk
  real(kind=rk), parameter :: divSqrt2_2 = 0.707106781186548_rk
  real(kind=rk), parameter :: sqrt3      = 1.73205080756887729352_rk

  real(kind=rk), parameter :: div1_2  = 1._rk /  2._rk
  real(kind=rk), parameter :: div1_3  = 1._rk /  3._rk
  real(kind=rk), parameter :: div1_4  = 1._rk /  4._rk
  real(kind=rk), parameter :: div1_6  = 1._rk / 6._rk
  real(kind=rk), parameter :: div1_8  = 1._rk /  8._rk
  real(kind=rk), parameter :: div1_9  = 1._rk /  9._rk
  real(kind=rk), parameter :: div1_18 = 1._rk / 18._rk
  real(kind=rk), parameter :: div1_36 = 1._rk / 36._rk
  real(kind=rk), parameter :: div1_54 = 1._rk / 54._rk


  real(kind=rk), parameter :: div4_3  = 4._rk /  3._rk
  real(kind=rk), parameter :: div4_9  = 4._rk /  9._rk

  real(kind=rk), parameter :: div3_2  = 3._rk /  2._rk
  real(kind=rk), parameter :: div3_4  = 3._rk /  4._rk
  real(kind=rk), parameter :: div3_4h = 3._rk / 4.5_rk
  real(kind=rk), parameter :: div3_8  = 3._rk /  8._rk
  real(kind=rk), parameter :: div3_16 = 3._rk /  16._rk

  real(kind=rk), parameter :: div9_16 = 9._rk /  16._rk

end module tem_param_module
!******************************************************************************!
