! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> plane type definition and routines used to generate planes
module tem_plane_module

  ! include treelm modules
  use env_module,         only: rk
  use tem_aux_module,     only: tem_abort
  use tem_bc_prop_module, only: tem_bc_prop_type
  use tem_tools_module,   only: tem_write
  use tem_logging_module, only: tem_log, tem_toStr, logUnit

  ! include aotus modules
  use aot_table_module, only: aot_table_open, aot_table_close, aot_table_length
  use aotus_module,     only: aot_get_val, flu_State, aoterr_Fatal,            &
    &                         aoterr_NonExistent, aoterr_WrongType
  use aot_out_module,   only: aot_out_type, aot_out_val,                       &
    &                         aot_out_open_table, aot_out_close_table

  implicit none
  private
  public :: tem_plane_type
  public :: tem_load_plane
  public :: tem_plane_out
  public :: tem_getNextCoordOfPlane
  public :: tem_plane_toSTLdata

!> Definition of a plane object
  type tem_plane_type
    !> origin of the plane object
    real(kind=rk) :: origin(3)
    !> vector along the edge A of plane (also defines size)
    real(kind=rk) :: vecA(3)
    !> vector along the edge B of plane (also defines size)
    real(kind=rk) :: vecB(3)
    !> how many discrete points the plane is divided into
    integer ::       segments(2)
    !> spatial distribution of the points on the plane
    character(len=40) :: distribution
    !> maximum number of refinement levels
    integer :: nRefine
    !> identification number
    integer :: id
    !>  identification of boundary type
    integer :: boundary_type
  end type tem_plane_type

  interface tem_plane_out
    module procedure tem_plane_out_scal
    module procedure tem_plane_out_vec
  end interface tem_plane_out

  contains

!---------------------------------------------------------------------------!
!> load plane information from config file
!!
!> Valid definitions:
!! None at all, creates no planes
!! Definition 1:
!! plane is defined by two vectors and position in x,y,z coordinate
!! with reference to origin.
!! - origin - position to trasform plane from its origin
!! - vecA- vector A from its origin ( defines plane length )
!! - vecB- vector B from its origin ( defines plane width )
!!\verbatim
!!  origin-----------> vecA
!!        | -        |
!!        |   -A+B   |
!!        |     -    |
!!        |       -  |
!!   vecB v----------x
!!\endverbatim
!!
!! Example:
!! plane at xy-plane normal to z-axis,
!! offset from z-plane at 1.0,
!! starting at x = -1.0 and y = -1.0,
!! length and width of plane is 2.0.
!! \verbatim
!! geom={plane={{vecA={2.0,0.0,0.0},
!!      vecB={0.0,2.0,0.0},
!!      origin={-1.0,-1.0,1.0},
!!      bclabel='top',
!!      refinementlevel = 5}
!!       }}
!! \endverbatim
!!\n
!! Definition 2:\n
!! center - center of the plane.\n
!! halfvecA - defines halfvecA from center.\n
!! halfvecB - defines halfvecB from center.\n
!!\verbatim
!!
!!  ____halfvecB___
!! |       |      |
!! |       |      |
!! |       |______|halfvecA
!! |     center   |
!! |              |
!! |______________|_
!!\endverbatim
!! User can specify halfvecA and halfvecB together with origin and
!! also vecA and vecB with center
!---------------------------------------------------------------------------!

!******************************************************************************!
  !> Loading planes from the config file
  !! valid definitions:
  !! * None at all, picks the default (origin of bounding cube)
  !! * plane = { 1.0, 0.0, -1.0 } -- single plane
  !! * plane = { {1.0, 0.0, -1.0} } -- single plane
  !! * plane = { {1.0, 0.0, -1.0},
  !!             {2.0, 0.0, -1.0} } -- multiple planes
  subroutine tem_load_plane( me, conf, thandle, bc )
    !---------------------------------------------------------------------------
    !> the array of planes which to read in (and first allocate)
    type(tem_plane_type), allocatable, intent(out) :: me(:)
    !> the lua config handle
    type(flu_state) :: conf
    !> table handle from which to read
    integer, intent(in) :: thandle
    !> boundary property
    type(tem_bc_prop_type), optional, intent(inout) :: bc
    !---------------------------------------------------------------------------
    integer :: planes_thandle
    integer :: single_thandle
    integer :: plane_entries
    integer :: iPlane
    integer :: nPlanes
    !---------------------------------------------------------------------------

    call aot_table_open( L       = conf,                                       &
      &                  parent  = thandle,                                    &
      &                  thandle = planes_thandle,                             &
      &                  key     = 'plane' )
    plane_entries = aot_table_length( L = conf, thandle = planes_thandle )

    ! Now check, if the first entry is a table itself:
    call aot_table_open( L       = conf,                                       &
      &                  parent  = planes_thandle,                             &
      &                  thandle = single_thandle,                             &
      &                  pos     = 1 )

    if (single_thandle == 0) then
      ! The entries are not tables themselves, try to interpret it as
      ! single plane.
      call aot_table_close( L = conf, thandle = single_thandle )
      nPlanes = 1
      allocate(me( nPlanes ))
      call tem_load_onePlane( me(1), conf, planes_thandle, bc )

    else
      ! First entry is a table, assume all others are also tables,
      ! and properly define plane coordinates, go on reading them.
      call aot_table_close(L=conf, thandle=single_thandle)
      nPlanes = plane_entries
      allocate(me(nPlanes))
      do iPlane=1,nPlanes
        call aot_table_open( L       = conf,                                   &
          &                  parent  = planes_thandle,                         &
          &                  thandle = single_thandle,                         &
          &                  pos     = iPlane)
        call tem_load_onePlane( me(iPlane), conf, single_thandle, bc )
      end do
    end if
    call aot_table_close(L=conf, thandle=planes_thandle)

    do iPlane = 1, nPlanes
      write(logUnit(1),*) 'Read Plane: '//trim( tem_toStr( iPlane ))//' with'
      write(logUnit(1),*) ' origin: '//trim( tem_toStr(                      &
        &                   val = me(iPlane)%origin,                           &
        &                   sep = ',' ))
      write(logUnit(1),*) ' vecA: '//trim( tem_toStr(                        &
        &                   val = me(iPlane)%vecA,                             &
        &                   sep = ',' ))
      write(logUnit(1),*) ' vecB: '//trim( tem_toStr(                        &
        &                   val = me(iPlane)%vecB,                             &
        &                   sep = ',' ))
      if (present(bc))                                                         &
           write(logUnit(1),*) 'Boundary_label:'//                           &
        &                        trim(bc%BC_label(me(iPlane)%boundary_type))
      write(logUnit(1),*) ' nRefine: '//trim( tem_toStr( me(iPlane)%nRefine))
      write(logUnit(1),*) ' segments1: '//                                   &
        &                   trim( tem_toStr( me(iPlane)%segments(1) ))
      write(logUnit(1),*) ' segments2: '//                                   &
        &                   trim( tem_toStr( me(iPlane)%segments(2) ))
      write(logUnit(1),*) ' distribution '//trim( me(iPlane)%distribution)
    end do

  end subroutine tem_load_plane
!******************************************************************************!


!******************************************************************************!
  !> Read one plane definition into a tem_plane_type from a lua table plane = {}
  !!
  subroutine tem_load_onePlane(me, conf, thandle, bc)
    !---------------------------------------------------------------------------
    !> contains plane datas
    type(tem_plane_type), intent(out) :: me
    !> lua state
    type(flu_state) :: conf
    !> lua table identification
    integer, intent(in) :: thandle
    !> global boundary property type
    type(tem_bc_prop_type), optional, intent(inout):: bc
    !---------------------------------------------------------------------------
    integer :: iError
    integer :: vError(3), errfatal(3)
    character(len=128) :: buffer
    !---------------------------------------------------------------------------

    errfatal = aotErr_Fatal

    !read vecA if not defined read halfvecA
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               val     = me%vecA,                                       &
      &               key     = 'vecA',                                        &
      &               ErrCode = vError )

    if (any(btest( vError, errFatal )) ) then
      !read halfvecA
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               val     = me%vecA,                                     &
        &               key     = 'halfvecA',                                  &
        &               ErrCode = vError )
      if (any(btest( vError, errFatal )) ) then
        write(*,*) 'ERROR reading the plane, vecA/halfvecA is not well'
        write(*,*) 'defined. STOPPING'
        call tem_abort()
      endif
      me%vecA = me%vecA * 2.0_rk
    end if

    !read vecB if not defined read halfvecB if not halfvecA read pointB
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               val     = me%vecB,                                       &
      &               key     = 'vecB',                                        &
      &               ErrCode = vError )

    if (any(btest( vError, errFatal )) ) then
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               val     = me%vecB,                                     &
        &               key     = 'halfvecB',                                  &
        &               ErrCode = vError )
      if (any(btest( vError, errFatal )) ) then
        write(*,*) 'ERROR reading the plane, vecB/vecB is not well'
        write(*,*) 'defined. STOPPING'
        call tem_abort()
      endif
      me%vecB = me%vecB * 2.0_rk
    end if

    !read position of plane
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               val     = me%origin,                                     &
      &               key     = 'origin',                                      &
      &               ErrCode = vError )
    if (any(btest( vError, errFatal )) ) then
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               val     = me%origin,                                   &
        &               key     = 'center',                                    &
        &               ErrCode = vError )
      if (any(btest( vError, errFatal )) ) then
        write(*,*) 'ERROR reading the plane, origin/center is not well'
        write(*,*) 'defined. STOPPING'
        call tem_abort()
      endif
      me%origin = me%origin - me%vecA/2.0_rk - me%vecB/2.0_rk
    end if

    ! Do we have to read in boundary attributes for this plane as forSeeder?
    if( present( bc )) then
      ! Boundary condition definition
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               val     = buffer,                                      &
        &               ErrCode = iError,                                      &
        &               key     = 'bclabel',                                   &
        &               default = 'wall' )

      if (btest(iError, aoterr_Fatal)) then
        write(*,*) 'FATAL Error occured, while retrieving boundary_type:'
        if (btest(iError, aoterr_NonExistent))                                 &
          &                                 write(*,*) 'Variable not existent!'
        if (btest(iError, aoterr_WrongType))                                   &
          &                               write(*,*) 'Variable has wrong type!'
        write(*,*) 'STOPPING'
        call tem_abort()
      else
        if (btest(iError, aoterr_NonExistent))                                 &
          & write(*,*) 'Variable boundary_type not set in configuration, '//   &
          &            'using default value!'
      end if

      bc%nBCtypes = bc%nBCtypes + 1
      write(*,*) 'Creating boundary', bc%nBCtypes
      bc%BC_label( bc%nBCtypes ) = trim(buffer)
      me%boundary_type = bc%nBCtypes
    endif

    ! Read in of refinement level
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               val     = me%nRefine,                                    &
      &               ErrCode = iError,                                        &
      &               key     = 'refinementlevel',                             &
      &               pos     = 5,                                             &
      &               default = -1 )

    ! Read in the shape segmentation
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               val     = me%segments(1),                                &
      &               ErrCode = iError,                                        &
      &               key     = 'segments1',                                   &
      &               default = 10 )
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               val     = me%segments(2),                                &
      &               ErrCode = iError,                                        &
      &               key     = 'segments2',                                   &
      &               default = 10 )
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               val     = me%distribution,                               &
      &               ErrCode = iError,                                        &
      &               key     = 'distribution',                                &
      &               default = 'equal' )


  end subroutine tem_load_onePlane
!******************************************************************************!


!******************************************************************************!
  !> This subroutine convert planes into two triangles
  !!
  !! This subroutine convert planes defined in config.lua into two triangles.
  !! then nodes indicies and triangle indices are stored as stl_data_type
  subroutine tem_plane_toSTLdata( plane, nodes, tri_node)
    !--------------------------------------------------------------------------!
    !> plane data type
    type(tem_plane_type), intent(in) :: plane
    !> six nodes (x,y,z) to define 2 triangles
    real(kind=rk), intent(out) :: nodes(3,6)
    !> node indicies of 2 triangles
    integer,intent(out) :: tri_node(3,2)
    !--------------------------------------------------------------------------!
    ! resultant of vector A and B to find 4th node to define plane
    real(kind=rk) :: vecAB(3)
    !--------------------------------------------------------------------------!

    call tem_write(" Converting plane datas to stl data type ...")

    tri_node(:,1) = (/ 1, 2, 3 /)
    tri_node(:,2) = (/ 4, 5, 6 /)

    ! resultant of vecA and vecB
    vecAB = plane%vecA+plane%vecB

    !1st Triangle nodes - numbering is done in anticlockwise direction
    nodes(1:3,1) = plane%origin
    nodes(1:3,2) = plane%origin + plane%vecA
    nodes(1:3,3) = plane%origin + vecAB
    !2nd triangle nodes
    nodes(1:3,4) = plane%origin
    nodes(1:3,5) = plane%origin + vecAB
    nodes(1:3,6) = plane%origin + plane%vecB

  end subroutine tem_plane_toSTLdata
!******************************************************************************!


!******************************************************************************!
  !> Return the next coordinate of the plane
  !!
  !! More complex shapes are identified by a point-wise processing of the shape
  !! The objects are described by a set of points. These sets are defined by an
  !! amount of subdivisions
  function tem_getNextCoordOfPlane( me, iElem ) result(coord)
    !---------------------------------------------------------------------------
    !> current element within total amount of elements to process
    integer, intent(in) :: iElem
    !> current plane
    type(tem_plane_type ) :: me
    !> calulated real-world coordinate, which is returned
    real(kind=rk)       :: coord(3)
    !---------------------------------------------------------------------------

      ! Start from the origin, step through the segments on the two
      ! defining vectors successively
      ! \todo  Not clear how it is done
      coord = me%origin                                                        &
            ! first calculate the vecA-offset by taking into account the
            ! number of segments in vecA direction, and use the modulo (rest),
            ! so we can proceed just like for a line, see tem_getNextCoordOfLine
        &   + real(mod( iElem - 1, me%segments(1) ) ,kind=rk)                  &
        &   / real(me%segments(1)-1,kind=rk)                                   &
        &   * me%vecA(:)                                                       &
            ! and add the offset in vecB is computed by taking the int of
            ! iElem/segments(1)
        &   + real(   ( (iElem-1)/ me%segments(1) ) ,kind=rk)                  &
        &   / real(me%segments(2)-1,kind=rk)                                   &
        &   * me%vecB(:)

  end function tem_getNextCoordOfPlane
!******************************************************************************!


!******************************************************************************!
  !> Write a array of lines to lua file
  !!
  subroutine tem_plane_out_vec( me, conf )
    !---------------------------------------------------------------------------
    !> plane types to write out
    type( tem_plane_type ), intent(in) :: me(:)
    !> Aotus type handling the output to the file in lua format
    type(aot_out_type), intent(inout) :: conf
    !---------------------------------------------------------------------------
    integer :: i
    !---------------------------------------------------------------------------

    ! create a table with name plane
    call aot_out_open_table( put_conf = conf, tname = 'plane' )

    do i = 1, size(me)
      call tem_plane_out_scal( me(i), conf )
    end do

    call aot_out_close_table( put_conf = conf )

  end subroutine tem_plane_out_vec
!******************************************************************************!


!******************************************************************************!
  !> Write a plane to lua file
  !!
  subroutine tem_plane_out_scal( me, conf )
    type( tem_plane_type ), intent(in) :: me !< plane types to write out
    !> Aotus type handling the output to the file in lua format
    type(aot_out_type), intent(inout) :: conf

    ! create a table with name plane if not exist
    if( conf%level .eq. 0 ) then
      call aot_out_open_table( put_conf = conf, tname = 'plane' )
    else
      call aot_out_open_table( put_conf = conf )
    end if
    ! write origin
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'origin',                                     &
      &               val      = me%origin )

    ! write vecA
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'vecA',                                       &
      &               val      = me%vecA )

    ! write vecB
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'vecB',                                       &
      &               val      = me%vecB )
    ! write segments1
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'segment1',                                   &
      &               val      = me%segments(1) )
    ! write segments2
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'segment2',                                   &
      &               val      = me%segments(2) )
    ! write distribution
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'distribution',                               &
      &               val      = me%distribution )

    call aot_out_close_table( put_conf = conf )

  end subroutine tem_plane_out_scal
!******************************************************************************!


end module tem_plane_module
!******************************************************************************!
