! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!>
!!
module tem_polynomial_module

  ! include treelm modules
  use env_module, only: rk

  implicit none

  private

  public :: polynomial, set_divU, set_coefficients

  contains

!******************************************************************************!
  !> Evaluate a 3d polynomial with given coefficients at a coordinate 
  !!
  function polynomial( a, coord, order, nDims ) result( res )
    !---------------------------------------------------------------------------
    !> coefficients
    real(kind=rk), intent(in) :: a(:)
    real(kind=rk), intent(in) :: coord(3)
    integer, intent(in) :: order
    integer, intent(in) :: nDims
    real(kind=rk) :: res
    !---------------------------------------------------------------------------
    real(kind=rk) :: x,y,z
    !---------------------------------------------------------------------------

    x = coord(1)
    y = coord(2)
    z = coord(3)

    select case( order )
    case( 1 )
      if ( nDims .eq. 3 ) then
        res = a(1) + a(2)*x + a(3)*y + a(4)*z
      else
        res = a(1) + a(2)*x + a(3)*y
      end if
    case( 2)
      if ( nDims .eq. 3 ) then
        res = a(1) + a(2)*x + a(3)*y + a(4)*z + a(5)*x**2 + a(6)*y**2          &
          & + a(7)*z**2 + a(8)*x*y + a(9)*y*z + a(10)*z*x
      else
        res = a(1) + a(2)*x + a(3)*y + a(4)*x**2 + a(5)*y**2  + a(6)*x*y 
      end if
    end select

  end function polynomial
!******************************************************************************!

 
!******************************************************************************!
  !> set the last velocity to satisfy the divergence(u) property
  !! div(u) = 0
  !! 3d: dux/dx + duy/dy + duz/dz = 0
  !!     set ux and uy, get uz by
  !!     duz/dz = -dux/dx - duy/dy
  !!     uz = int (-dux/dx - duy/dy ) dz
  !! 2d: dux/dx + duy/dy = 0
  !!     set ux and get uy by
  !!     duy/dy = -dux/dx 
  !!     uy = int (-dux/dx ) dy
  !!
  subroutine set_divU( coeff, Sxxcoeff, nStresses, order, nDims, nDofs,        &
    &                  uOut, sxxOut, coord ) 
    !---------------------------------------------------------------------------
    !>
    integer, intent(in) :: order
    !>
    integer, intent(in) :: nDims
    !>
    integer, intent(in) :: nDofs
    !> coefficients
    real(kind=rk), intent(out) :: uOut
    !> Velocity moment coefficients
    real(kind=rk), intent(inout) :: coeff(nDofs,nDims)
    !> Number of Shear stress moments
    integer, intent(in) :: nStresses
    !> Shear stress moment coefficients
    real(kind=rk), optional, intent(inout) :: Sxxcoeff(nDofs,nStresses )
    !>
    real(kind=rk), optional, intent(out) :: SxxOut(nStresses )
    !>
    real(kind=rk), intent(in) :: coord(3)
    !---------------------------------------------------------------------------
    logical :: setStress 
    integer :: iSxx
    !---------------------------------------------------------------------------

    setStress = .false.
    if(  present( Sxxcoeff ) .and. present( SxxOut )) then
      setStress = .true.
      ! Set all coefficients for the shear rate to 0
      Sxxcoeff(:,:) = 0._rk
    end if

    select case( order )
    case( 1 )
      ! Linear case
      if ( nDims .eq. 3 ) then
        ! ux = coeff(1,1) + coeff(2,1)*x + coeff(3,1)*y + coeff(4,1)*z
        ! uy = coeff(1,2) + coeff(2,2)*x + coeff(3,2)*y + coeff(4,2)*z
        ! uz = coeff(1,3) + coeff(2,3)*x + coeff(3,3)*y + coeff(4,3)*z
        !  -> replace coeff(4,3) = -( coeff(2,1) + coeff(3,2))
        coeff(4,3) = -( coeff(2,1) + coeff(3,2) )
        if( setStress ) then
          ! Sxx = 2*dux/dx
          ! Syy = 2*duy/dy
          ! Szz = 2*duz/dz
          ! Sxy = dux/dy + duy/dx
          ! Syz = duz/dy + duy/dz
          ! Sxz = dux/dz + duz/dx
          Sxxcoeff(1,1) = coeff(2,1)*2._rk
          Sxxcoeff(1,2) = coeff(3,2)*2._rk
          Sxxcoeff(1,3) = coeff(4,3)*2._rk
          Sxxcoeff(1,4) = coeff(3,1) + coeff(2,2)
          Sxxcoeff(1,5) = coeff(4,2) + coeff(3,3)
          Sxxcoeff(1,6) = coeff(4,1) + coeff(2,3)
        end if
      else
        ! ux = coeff(1,1) + coeff(2,1)*x + coeff(3,1)*y
        ! uy = coeff(1,2) + coeff(2,2)*x + coeff(3,2)*y
        !  -> replace coeff(3,2) = -coeff(2,1)
        coeff(3,2) = -coeff(2,1)
        ! duxdx = coeff(2,1)
        ! duydx = coeff(2,2)
        ! duxdy = coeff(3,1)
        ! duydy = coeff(3,2)
        if( setStress ) then
          ! Sxx = 2*dux/dx
          ! Syy = 2*duy/dy
          ! Sxy = dux/dy + duy/dx
          Sxxcoeff(1,1) = coeff(2,1)*2._rk
          Sxxcoeff(1,2) = coeff(3,2)*2._rk
          Sxxcoeff(1,3) = coeff(3,1) + coeff(2,2)
        end if
      end if
    case( 2)
      if ( nDims .eq. 3 ) then
        ! ux = coeff(1,1) + coeff(2,1)*x + coeff(3,1)*y + coeff(4,1)*z
        !    + coeff(5,1)*x^2 + coeff(6,1)*y^2 + coeff(7,1)*z^2
        !    + coeff(8,1)*x*y + coeff(9,1)*y*z + coeff(10,1)*z*x
        ! uy = coeff(1,2) + coeff(2,2)*x + coeff(3,2)*y + coeff(4,2)*z
        !    + coeff(5,2)*x^2 + coeff(6,2)*y^2 + coeff(7,2)*z^2
        !    + coeff(8,2)*x*y + coeff(9,2)*y*z + coeff(10,2)*z*x
        ! uz = coeff(1,3) + coeff(2,3)*x + coeff(3,3)*y + coeff(4,3)*z
        !    + coeff(5,3)*x^2 + coeff(6,3)*y^2 + coeff(7,3)*z^2
        !    + coeff(8,3)*x*y + coeff(9,3)*y*z + coeff(10,3)*z*x
        ! div(u) = dux/dx + duy/dy + duz/dz 
        ! -> uz = - int( dux/dx + duy/dy) dz
        !    uz = -(coeff(2,1) + coeff(3,2))*z - (2*coeff(5,1)+coeff(8,2))*xz
        !       -(coeff(8,1)+2*coeff(6,2))*yz -0.5*(coeff(10,1)+coeff(9,2))*z^2
        !       + C
        coeff(4,3)  = -coeff(2,1) - coeff(3,2) 
        coeff(7,3)  = -0.5_rk*(coeff(10,1) + coeff(9,2))
        coeff(9,3)  = -coeff(8,1) - 2._rk*coeff(6,2)
        coeff(10,3) = -2._rk*coeff(5,1) - coeff(8,2) 
        if( setStress ) then
          Sxxcoeff(1,1) = 2._rk*coeff(2,1)    ! Sxx 1
          Sxxcoeff(2,1) = 4._rk*coeff(5,1)    ! Sxx x
          Sxxcoeff(3,1) = 2._rk*coeff(8,1)    ! Sxx y
          Sxxcoeff(4,1) = 2._rk*coeff(10,1)   ! Sxx z
          Sxxcoeff(1,2) = 2._rk*coeff(3,2)    ! Syy 1
          Sxxcoeff(2,2) = 2._rk*coeff(8,2)    ! Syy x
          Sxxcoeff(3,2) = 4._rk*coeff(6,2)    ! Syy y
          Sxxcoeff(4,2) = 2._rk*coeff(9,2)    ! Syy z
          Sxxcoeff(1,3) = 2._rk*coeff(4,3)    ! Szz 1
          Sxxcoeff(2,3) = 2._rk*coeff(10,3)   ! Szz x
          Sxxcoeff(3,3) = 2._rk*coeff(9,3)    ! Szz y
          Sxxcoeff(4,3) = 4._rk*coeff(7,3)    ! Szz z
          Sxxcoeff(1,4) = coeff(3,1)+coeff(2,2)        ! Sxy 1
          Sxxcoeff(2,4) = coeff(8,1)+coeff(5,2)*2._rk  ! Sxy x
          Sxxcoeff(3,4) = coeff(8,2)+coeff(6,1)*2._rk  ! Sxy y
          Sxxcoeff(4,4) = coeff(9,1)+coeff(10,2)       ! Sxy z
          Sxxcoeff(1,5) = coeff(3,3)+coeff(4,2)        ! Syz 1
          Sxxcoeff(2,5) = coeff(8,3)+coeff(10,2)       ! Syz x
          Sxxcoeff(3,5) = coeff(9,2)+coeff(6,3)*2._rk  ! Syz y
          Sxxcoeff(4,5) = coeff(9,3)+coeff(7,2)*2._rk  ! Syz z
          Sxxcoeff(1,6) = coeff(4,1)+coeff(2,3)        ! Sxz 1
          Sxxcoeff(2,6) = coeff(10,1)+coeff(5,3)*2._rk ! Sxz x
          Sxxcoeff(3,6) = coeff(8,3)+coeff(9,1)        ! Sxz y
          Sxxcoeff(4,6) = coeff(10,3)+coeff(7,1)*2._rk ! Sxz z
        end if
      else
        ! ux = coeff(1,1) + coeff(2,1)*x + coeff(3,1)*y 
        !    + coeff(4,1)*x^2 + coeff(5,1)*y^2 + coeff(6,1)*x*y
        ! uy = coeff(1,2) + coeff(2,2)*x + coeff(3,2)*y
        !    + coeff(4,2)*x^2 + coeff(5,2)*y^2 + coeff(6,2)*x*y
        ! duxdx = coeff(2,1) + 2*coeff(4,1)*x + coeff(6,1)*y
        ! duxdy = coeff(3,1) + 2*coeff(5,1)*y + coeff(6,1)*x
        ! duydx = coeff(2,2) + 2*coeff(4,2)*x + coeff(6,2)*y
        ! duydy = coeff(3,2) + 2*coeff(5,2)*y + coeff(6,2)*x
        coeff(3,2) = -coeff(2,1)
        coeff(6,2) = -2.0_rk*coeff(4,1)
        coeff(5,2) = -0.5_rk*coeff(6,1)
        if( setStress ) then
          ! Sxx = 2*dux/dx
          ! Syy = 2*duy/dy
          ! Sxy = dux/dy + duy/dx
          Sxxcoeff(1,1) = coeff(2,1)*2._rk  ! Sxx 1
          Sxxcoeff(2,1) = coeff(4,1)*4._rk  ! Sxx x
          Sxxcoeff(3,1) = coeff(6,1)*2._rk  ! Sxx y
          Sxxcoeff(1,2) = coeff(3,2)*2._rk  ! Syy 1
          Sxxcoeff(2,2) = coeff(6,2)*2._rk  ! Syy x
          Sxxcoeff(3,2) = coeff(5,2)*4._rk  ! Syy y
          Sxxcoeff(1,3) = coeff(3,1)+coeff(2,2) ! Sxy 1
          Sxxcoeff(2,3) = coeff(6,1)+2._rk*coeff(4,2) ! Sxy x
          Sxxcoeff(3,3) = 2._rk*coeff(5,1)+coeff(6,2) ! Sxy y
        end if
      end if
    case default
      write(*,*) 'wrong order. select 1 or 2'
      stop
    end select

    uOut = polynomial( a     = coeff(:, nDims ),                               &
      &                coord = coord,                                          &
      &                order = order,                                          &
      &                nDims = nDims )
    if( setStress ) then
      do iSxx = 1, nStresses
        SxxOut(iSxx) = polynomial( a     = SxxCoeff(:, iSxx ),                 &
          &                        coord = coord,                              &
          &                        order = order,                              &
          &                        nDims = nDims )
      end do
    end if

  end subroutine set_divU
!******************************************************************************!


!******************************************************************************!
  !> Set the polynomial coefficients for each moment. 
  !! This is more or less random. 
  !! Negative values for the density would be bad for the LBM 
  !! But here it should not hurt.
  !!
  !! The seed is being set in tem_init_env
  !!
  subroutine set_coefficients( me, nDOFS, QQ, valMin, valMax )
    !---------------------------------------------------------------------------
    real(kind=rk), intent(out), allocatable :: me(:,:) 
    integer, intent(in) :: nDOFs
    integer, intent(in) :: QQ
    real(kind=rk), intent(in) :: valMin
    real(kind=rk), intent(in) :: valMax
    !---------------------------------------------------------------------------
    real(kind=rk) :: test
    integer :: iVal, iDof
    !---------------------------------------------------------------------------
    ! minimum range to get a random number
    if( allocated( me )) deallocate(me)
    allocate( me( nDOFs, QQ ))
    do iVal = 1, QQ
      do iDof = 1, nDOFs
        call random_number(test)
        call random_number(test)
        call random_number(test)
        me( iDof, iVal ) = (test*(valMax-valMin))+valMin
        !if( iDof .eq. 1 ) me( iDof, iVal ) = me( iDof, iVal ) + 0._rk
      end do
      !write(*,*) iVal, 'coeff', me(:, iVal )
    end do

  end subroutine set_coefficients
!******************************************************************************!


end module tem_polynomial_module
!******************************************************************************!
