! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module applies reductions to a set of quantities
!!
!! This module takes a set of input quantities and reduces them 
!! according to the chosen operation.
!! You can choose between the following operations:
!! - average: Average of all occurring values
!! - sum: Sum of all occurring values
!! - max: Maximum of all occurring values
!! - min: Minimum of all occurring values
!! - l2norm: L2 norm of all values, i.e. L_2(x) = sqrt( sum( x_i^2 ))
!! - linfnorm: L_inf norm of all values, i.e. L_inf(x) = max( | x_i | )
!!
!! \verbatim
!! reduction = {
!!               'norm' -- 'average' or 'sum'
!!             } \endverbatim
!! 
!! The table of reductions has to correspond to the number of entries in the
!! tracking variable table. For each variable, specify a reduction as
!! \verbatim
!! tracking = {
!!   variable = {{'pressure'}, {'velMag'}, {'density'}},
!!   reduction = { 'average', 'max', l2norm' }
!! }
!! \endverbatim
!!
!! Usage of this module
!! --------------------
!! The tracking module handles the reductions. In the solver, the reduction
!! must then be inkoved in the compute loop
!! 
!! Embedding in Treelm
!! -------------------
!! Reductions are usually defined in tracking objects. Thus,
!! the tracking module handles the loading and closing of the reductions.
!! The following steps need to be performed within the tracking module
!! # Treelm: Load the configuration from the lua file
!!   with tem_load_reduction
!! # Treelm: Initialize the reductions with tem_init_reduction
!! # Solver: 
!!    # Open the reduction in the current time step tem_reduction_open
!!    # invoke the append operation to the array on which the reduction
!!       will be applied with tem_append_reduction
!!       This can be performed for a number of chunks
!!    # Close the reduction with tem_close_reduction
!!   
!! \author Vyacheslav Korchagin, Manuel Hasert
!!
module tem_reduction_module

  ! include treelm modules
  use mpi
  use env_module,            only: rk, rk_mpi, labelLen, long_k
  use tem_comm_env_module,   only: tem_comm_env_type
  use treelmesh_module,      only: treelmesh_type
  use tem_tools_module,      only: tem_write, upper_to_lower
  use tem_geometry_module,   only: tem_GetLocalBoundingCube, tem_elemSize
  use tem_var_system_module, only: tem_var_system_type
  use tem_logging_module,    only: tem_log, tem_toStr, logUnit

  ! include aotus modules
  use aotus_module,     only: aot_get_val, flu_State
  use aot_table_module, only: aot_table_open, aot_table_close,                 &
    &                         aot_table_length
  implicit none

  private

  public :: tem_reduction_type
  public :: tem_load_reduction
  public :: tem_reduction_init
  public :: tem_reduction_open
  public :: tem_reduction_append
  public :: tem_reduction_close
  public :: tem_reduction_toChunk
  public :: tem_GetRealBoundingCube

  !> This data type is providing the input for the reduction routines
  !! It must be filled by the solver, before the reduction is called
  !! It exists on each process
  type tem_reduction_type
    !> Which operation to perform on the  list of elements
    character(len=labelLen) :: operation = ''
    !> amount of components of the quantity to reduce
    integer :: nComponents
    !> the result from the reduction operation
    !! size: nComponents
    real(kind=rk), allocatable :: val(:)
    !> communicator on which to work on
    integer :: comm
    !> how many elements have been included into the reduction (so far)
    integer :: nElems
  end type tem_reduction_type

contains

!******************************************************************************!
  !> read configuration file
  !!
  subroutine tem_load_reduction(conf, me, parent, key, iError )
    !---------------------------------------------------------------------------
    !> handle for lua file
    type(flu_State)        :: conf
    !> the reduction file to fill
    type(tem_reduction_type),intent(inout), allocatable :: me(:)
    !> handle for reduce table
    integer, optional,intent(in)   :: parent
    !> which key to open
    character(len=*),optional,intent(in) :: key
    !> handle for reduce table
    integer, optional,intent(out)        :: iError
    !---------------------------------------------------------------------------
    integer          :: nEntries, handle, iErr, iPos
    character(len=labelLen) :: localKey
    !---------------------------------------------------------------------------
    if( present( key )) then
      localKey = key
    else
      localKey = 'reduction'
    endif
    nEntries = 1
    allocate( me( nEntries))
    iPos = 1

    ! First check, if it is NOT a table
    if (present( parent )) then
      ! ...but still inside a table
      call tem_load_reduction_single( conf    = conf,                          &
        &                             handle  = parent,                        &
        &                             me      = me( iPos ),                    &
        &                             key     = localKey,                      &
        &                             iError = iErr )
    else
      ! ... or at the root level
      call aot_get_val( L       = conf,                                        &
        &               key     = localKey,                                    &
        &               val     = me(iPos)%operation,                          &
        &               ErrCode = iErr,                                        &
        &               default = '' )
    endif

    if( iErr .ne. 0 .or. trim(me(iPos)%operation)=='' ) then
      ! ok, that failed. Let's try opening as a table.
      call aot_table_open( L       = conf,                                     &
        &                  thandle = handle,                                   &
        &                  parent  = parent,                                   &
        &                  key     = localKey)
      nEntries = aot_table_length( L = conf, thandle = handle )
      deallocate( me )
      allocate( me( nEntries ))
      do iPos = 1, nEntries
        call tem_load_reduction_single( conf    = conf,                        &
          &                             handle  = handle,                      &
          &                             me      = me( iPos ),                  &
          &                             pos     = iPos,                        &
          &                             iError  = iErr )
      enddo
      call aot_table_close(conf, handle)
      if( iErr .ne. 0 ) then
        ! Error, allocate with 0 size
        nEntries = 0
        deallocate( me )
        allocate( me(nEntries))
      endif
    endif

    write(logUnit(1),*)' nReductions loaded: '//trim( tem_toStr( nEntries )) 

    if( present( iError )) iError = iErr

  end subroutine tem_load_reduction
!******************************************************************************!


!******************************************************************************!
  !> Read a single entry of reductions from the lua file
  !!
  subroutine tem_load_reduction_single( conf, me, handle, key, pos, iError )
    !---------------------------------------------------------------------------
    !> handle for lua file
    type(flu_State),intent(inout) :: conf
    !> reduction type to be filled
    type(tem_reduction_type),intent(inout) :: me
    !> handle for reduce table
    integer, optional,intent(in) :: handle
    !> which key to open
    character(len=*),optional,intent(in) :: key
    !> position to load from in config file
    integer,optional,intent(in) :: pos
    !> handle for reduce table
    integer, optional,intent(out) :: iError
    !---------------------------------------------------------------------------
    integer :: iErr
    !---------------------------------------------------------------------------
    if( present( key )) then
      call aot_get_val( L       = conf,                                        &
        &               thandle = handle,                                      &
        &               val     = me%operation,                                &
        &               ErrCode = iErr,                                        &
        &               key     = key )
    elseif( present( pos ))then
      call aot_get_val( L       = conf,                                        &
        &               thandle = handle,                                      &
        &               val     = me%operation,                                &
        &               ErrCode = iErr,                                        &
        &               pos     = pos )
    else
      iErr = -1
    endif
    me%operation = upper_to_lower( me%operation ) 
    ! Check the chosen operation
    select case( me%operation )
      case('sum')
      case('average')
      case('l2norm','l2_norm')
        me%operation = 'l2norm'
      case('linfnorm', 'linf_norm', 'l_inf_norm')
        me%operation = 'linfnorm'
      case('max','maximum')
        me%operation = 'max'
      case('min','minimum')
        me%operation = 'min'
      case('none')
      case default
!        call tem_write(' Error: The chosen reduction '//trim(me%operation)//   &
!          &            ' is not defined.')
        iErr = -1
    end select
    if( present( iError )) then
      iError = iErr
    endif

  end subroutine tem_load_reduction_single
!******************************************************************************!


!******************************************************************************!
  !> Initialize reduction objects according to the variable systems
  !!
  subroutine tem_reduction_init( me, varSys )
    !---------------------------------------------------------------------------
    !> array of reductions to initialize
    type( tem_reduction_type ), intent(inout), allocatable :: me(:)
    !> corresponding variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !---------------------------------------------------------------------------
    type( tem_reduction_type ), allocatable :: tmpRdc(:)
    integer :: iRdc
    !---------------------------------------------------------------------------

    call tem_write(' Initializing reduction ...')
    ! check if the amount of reduction greater than amount of variables
    if( size( me ) .gt. varSys%nVars ) then
      call tem_write('  The amount of reduction is greater than the amount '// &
        &            'of available variables!')
      allocate( tmpRdc( varSys%nVars ))
      do iRdc = 1, varSys%nVars
        tmpRdc( iRdc ) = me( iRdc )
      end do
      deallocate( me )
      allocate( me( varSys%nVars ))
      do iRdc = 1, varSys%nVars
        me( iRdc ) = tmpRdc( iRdc )
      end do
      deallocate( tmpRdc )
    endif

    do iRdc = 1, size( me )
      me( iRdc )%nComponents = varSys%variable%val( iRdc )%nComponents
      allocate( me( iRdc )%val( me(iRdc)%nComponents ))
      me( iRdc )%val(:) = 0._rk
    end do

  end subroutine tem_reduction_init
!******************************************************************************!


!******************************************************************************!
  !> Prepare the reduction data type
  !!
  !! \todo HK: what is going on here?
  !!           This stuff reads, as if we have a reduction per variable?
  !!           However above, we read a set of reduction objects from the
  !!           configuration, thus it seems to be independent of the number
  !!           of variables? Somehow there is a missing link between the
  !!           reduction and the variable it should act on?
  !!           How is this supposed to work? Please clarify, document or
  !!           change the code!
  !! MH: Ok, the current way it is intended to work is the following
  !!  you define a tracking object with a given set of variables
  !!  and define a number of reduction operations.
  !!  The reduction operations work on the variables starting from the first
  !!  up to the number of defined reductions
  !!  example:
  !!  tracking = { variable = {'density', 'velocity'},
  !!               reduction = 'average', ... }
  !!  Here the number of reductions is smaller than the number of variables
  !!  which results in that only the density is taken an average.
  !!
  subroutine tem_reduction_open(me, varSys)
    !---------------------------------------------------------------------------
    !> The reduction type to work on. All definitions should be
    !! present in here
    type( tem_reduction_type ), intent(inout) :: me(:)
    !> variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !---------------------------------------------------------------------------
    integer :: iReduce
    !---------------------------------------------------------------------------

    do iReduce = 1, size( me )
      ! Reset the number of treated elements.
      ! Must be counted up during _append
      me( iReduce )%nElems = 0

      ! Take the components of the connected variable system
      me( iReduce )%nComponents = varSys%variable%val( iReduce )%nComponents
      if( .not. allocated( me( iReduce )%val ))                                &
        & allocate( me( iReduce )%val( me(iReduce)%nComponents ))

      select case( me( iReduce )%operation )
        case('max')
          me( iReduce )%val(:) = -huge(1._rk)
        case('min')
          me( iReduce )%val(:) = huge(1._rk)
        case default ! ('sum', 'average', 'l2norm', 'none', ...
          me( iReduce )%val(:) = 0._rk
      end select
    end do

  end subroutine tem_reduction_open
!******************************************************************************!


!******************************************************************************!
  !> Local chunk-wise reduction
  !!
  !! Chunk-wise local reduction, which has to be performed for each
  !! chunk until all elements have been treated within the reduction.
  !! After all chunks are treated (=_append has been finished),
  !! the global reduction has to be performed by tem_reduction_close
  !!
  subroutine tem_reduction_append(me, chunk, nElems, varSys, tree, treeID)
    !---------------------------------------------------------------------------
    !> The reduction type to work on. All definitions should be
    !! present in here
    type( tem_reduction_type ), intent(inout) :: me(:)
    !> number of elements the chunk has
    integer, intent(in) :: nElems
    !> number of scalar entries there are per element
    type( tem_var_system_type ), intent(in) :: varSys
    !>
    real(kind=rk), intent(in) :: chunk(:)
    !> the global tree
    type(treelmesh_type), intent(in) :: tree
    !> The list of treeIDs of the current chunk
    integer(kind=long_k), intent(in) :: treeID( nElems )
    !---------------------------------------------------------------------------
    real(kind=rk) :: temp, volume
    integer :: iVar, iComp, iPos, varPos, iElem
    !---------------------------------------------------------------------------

    ! Run over all variables in the chunk
    do iVar = 1, size( me )
      me( iVar )%nElems = me( iVar )%nElems + nElems
      ! count up the me%nElems here as well
      ! update the me%val
      do iComp = 1, me( iVar )%nComponents
        varPos = varSys%variable%val( iVar )%varPos(iComp)
        temp = 0._rk
        select case( me( iVar )%operation )
          ! Upon extending the reductions, make sure to add them in the
          ! tem_load_reduction  routine
          case('sum', 'average')
            ! Take the values corresponding to the current component of the
            ! variable with step (varSys%nScalars)
            temp = sum(chunk( varPos :nElems*varSys%nScalars:varSys%nScalars))
            me( iVar )%val( iComp ) = me( iVar )%val( iComp ) + temp

          case('l2norm')
            ! Sum up the squares of elemets of chunk into a temp variable
            ! Total error L2 on the domain Omega is with the error function u 
            ! $L2(\Omega) =  \sqrt(\int_\Omega |u|^2 ))
            !             =  \sqrt(\sum_i \int_\Omega_i |u|^2 ))
            !             =  \sqrt(\sum_i \Omega_i \cdot |u|^2 ))
            ! -> We need to compute the volume of each element and weight the 
            !    element error with it to obtain a consistent L2 norm for
            !    multilevel meshes
            do iPos = varPos, nElems*varSys%nScalars, varSys%nScalars
              iElem  = ((iPos-1)/varSys%nScalars) + 1
              volume = tem_elemSize( tree, treeID( iElem ) )**3
              temp = temp + volume*chunk( iPos )*chunk( iPos )
            enddo
            me( iVar )%val( iComp ) = me( iVar)%val( iComp) + temp

          case('linfnorm')
            ! L_inf norm evaluates the maximum of the absolute occurring value
            me( iVar )%val( iComp ) = max( me( iVar )%val( iComp ),            &
              &           maxval( abs(chunk( varPos : nElems*varSys%nScalars : &
              &                   varSys%nScalars) )))

          case('max')
            ! maximum between previous value and current
            me( iVar )%val( iComp ) = max(me( iVar )%val( iComp ),             &
              & maxval( chunk( varPos:nElems*varSys%nScalars:varSys%nScalars )))

          case('min')
            ! minimum between previous value and current
            me( iVar )%val( iComp ) = min(me(iVar)%val(iComp),                 &
              & minval( chunk( varPos:nElems*varSys%nScalars:varSys%nScalars )))
          case('none')
            me( iVar )%val( iComp ) = 0._rk
        end select
      enddo
    enddo

  end subroutine tem_reduction_append
!******************************************************************************!


!******************************************************************************!
  !> Perform the global reduction
  !!
  !! After the local reductions have been performed (in _append),
  !! the results must be communicated between processes.
  !!
  subroutine tem_reduction_close(me, comm)
    !---------------------------------------------------------------------------
    !> The reduction type to work on. All definitions should be
    !! present in here
    type( tem_reduction_type ), intent(inout) :: me(:)
    !> communicator for the processes participating in this restart (might
    !! be only a subset)
    type(tem_comm_env_type) :: comm
    !---------------------------------------------------------------------------
    integer :: i, nComp, ierr, globalnElems
    real(kind=rk), allocatable :: buff(:)
    !---------------------------------------------------------------------------

    !loop over all tracking objects
    do i = 1, size(me)
      globalnElems = 0
      ! get global number of elements in this reduction to rank 0
      call mpi_reduce( me(i)%nElems, globalnElems,                             &
        &              1, mpi_integer, mpi_sum, comm%root, comm%comm, iErr)

      ! get number of components
      nComp = me(i)%nComponents

      allocate(buff(nComp))
      buff = 0.0_rk

      !choose reduction operation and perform it
      select case( me( i )%operation )
        !sum all values
        case('sum')
          call mpi_reduce( me(i)%val, buff,                                    &
            &              nComp, rk_mpi, mpi_sum, comm%root, comm%comm, iErr)
          me(i)%val = buff

        !sum all values and devide by number of elements
        case('average')
          call mpi_reduce( me(i)%val, buff,                                    &
          &                nComp, rk_mpi, mpi_sum, comm%root, comm%comm, iErr)
          me(i)%val(:) = buff / real( globalnElems, kind = rk)

        !sum all values(sum of squares) and extract a square root
        case('l2norm')
          call mpi_reduce( me(i)%val, buff,                                    &
          &                nComp, rk_mpi, mpi_sum, comm%root, comm%comm, iErr)
          me(i)%val(:) = sqrt(buff)

        !maximium over all values
        case('max','linfnorm')
          call mpi_reduce( me(i)%val, buff,                                    &
          &                nComp, rk_mpi, mpi_max, comm%root, comm%comm, iErr)
          me(i)%val = buff

        !minimum over all values
        case('min')
          call mpi_reduce( me(i)%val, buff,                                    &
          &                nComp, rk_mpi,mpi_min, comm%root, comm%comm, iErr)
          me(i)%val = buff
      end select
      deallocate( buff )
    enddo

  end subroutine tem_reduction_close
!******************************************************************************!


!******************************************************************************!
  !> Transfer reduction results to array chunk
  !!
  subroutine tem_reduction_toChunk( me, chunk, nChunkElems )
    !---------------------------------------------------------------------------
    !>
    type(tem_reduction_type), intent(in) :: me(:)
    !>
    real(kind=rk), intent(inout) :: chunk(:)
    !>
    integer :: nChunkElems
    !---------------------------------------------------------------------------
    integer :: iReduce, iPos, iComp
    !---------------------------------------------------------------------------

    nChunkElems = 1
    iPos = 0
    do iReduce = 1, size( me )
      do iComp = 1, me(iReduce)%nComponents
        chunk( iPos+iComp ) = me(iReduce)%val( iComp )
      enddo
      iPos = iPos + me(iReduce)%nComponents
    enddo

  end subroutine tem_reduction_toChunk
!******************************************************************************!


!******************************************************************************!
  !> Calculate the real bounding box around the fluid domain
  !! and return only to root (efficiency reasons)
  !!
  function tem_GetRealBoundingCube( tree ) result( boundingCube )
    !---------------------------------------------------------------------------
    !>
    type(treelmesh_type)   :: tree
    !>
    real(kind=rk) :: boundingCube(3,2)
    !---------------------------------------------------------------------------
    real(kind=rk) :: tBounding(3)
    integer :: iErr
    !---------------------------------------------------------------------------
    ! Calculate process-local bounding cube
    boundingCube = tem_GetLocalBoundingCube( tree )
    ! Exchange with neighbors
    call mpi_reduce( boundingCube(:,1), tBounding, 3, rk_mpi, mpi_min,         &
      &              0, tree%global%comm, iErr  )
    boundingCube(:,1) = tBounding
    call mpi_reduce( boundingCube(:,2), tBounding, 3, rk_mpi, mpi_max,         &
      &              0, tree%global%comm, iErr  )
    boundingCube(:,2) = tBounding

  end function tem_GetRealBoundingCube
!******************************************************************************!


end module tem_reduction_module
!******************************************************************************!
