! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module provides the main IO facilities to read and write large amounts
!! of data during the simulation.
!!
!! \author Simon Zimny, Harald Klimach
!!
!! Especially it enables the dumping of the simulation status
!! (e.g. the state vector) to disk and reading the information later on to
!! restart the simulation at a certain point of time.
!! Additionally informations like solver, version, number of elements, etc.
!! are stored in a seperate header file.
!!
!! The general procedure is as follows:
!! - read the configuration from the lua file with tem_load_restart
!! - initialize the read/write restart with tem_init_restart
!! - read restart
!! - write restart
!!
!! Find more information about usage of restart in the section '[restart_usage]'
!!
!! \section restart_format The Restart / Harvester Format
!!
!! \todo Add here an extensive description of the restart format
!!
!! \section restart_usage The Restart Usage
!!
!! To use the restart functionality the user has to define the io buffersize 
!!
!! \verbatim
!! -- IO buffer size in MB (default = 8)
!! io_buffer_size = 1
!! \endverbatim
!!
!! and the restart settings in the lua configuration file.
!!
!! \verbatim
!! -- Restart settings
!! restart = { read = 'restart/gaussian_lastHeader.lua',  -- file to restart from
!!             write = 'restart/',                        -- prefix to write the files to
!!             time = { min = 0, max = 10, interval = 10} -- timing definitions (either iterations or simulation time)
!!           }
!! \endverbatim
!!
!! If only the write option shall be used the identifier read has to be removed 
!! and vice versa.
!!
!! [restart_format]: @ref restart_format "The Restart / Harvester Format"
!! [restart_usage]: @ref restart_usage "Restart Usage"
!!
module tem_restart_module

  ! include treelm modules
  use mpi
  use env_module,              only: LabelLen, PathLen, rk_mpi, long_k, rk,    &
    &                                tem_create_endianSuffix, pathSep,         &
    &                                io_buffer_size, globalMaxLevels
  use treelmesh_module,        only: treelmesh_type, load_tem, dump_treelmesh
  use tem_global_module,       only: tem_global_type, tem_mesh_out
  use tem_aux_module,          only: tem_abort, tem_open_distconf
  use tem_comm_env_module,     only: tem_comm_env_type
  use tem_solveHead_module,    only: tem_solveHead_type, tem_solverTag
  use tem_timeControl_module,  only: tem_timeControl_type, tem_load_TimeControl
  use tem_time_module,         only: tem_generate_timeStamp, tem_time_type,    &
    &                                tem_time_out, tem_time_load
  use tem_tools_module,        only: tem_write
  use tem_var_system_module,   only: tem_var_system_type, tem_var_sys_out,     &
    &                                tem_var_sys_load, tem_var_sys_debug
  use tem_subTree_type_module, only: tem_subTree_type, tem_dump_subTree

  ! include aotus modules
  use aotus_module,     only: flu_State, aot_get_val, close_config
  use aot_out_module,   only: aot_out_type, aot_out_val,                       &
    &                         aot_out_open, aot_out_close,                     &
    &                         aot_out_open_table, aot_out_close_table
  use aot_table_module, only: aot_table_open, aot_table_close

  implicit none

  private

  public :: tem_restart_type
  public :: tem_restartControl_type
  public :: tem_restartHeader_type
  public :: tem_load_restart
  public :: tem_init_restart
  public :: tem_init_read_restart
  public :: tem_restart_finalize
  public :: tem_restart_finalize_single
  public :: tem_restart_writeData
  public :: tem_restart_writeHeader
  public :: tem_restart_readData
  public :: tem_restart_readData_single
  public :: tem_restart_openRead
  public :: tem_restart_openRead_single
  public :: tem_restart_closeRead
  public :: tem_restart_closeRead_single
  public :: tem_restart_openWrite
  public :: tem_restart_closeWrite
  public :: tem_restart_readHeader

  !> Control the behavior of the restart, like at which point in time etc.
  type tem_restartControl_type
    !> is the restart read active?
    logical :: readRestart
    !> is the restart write active?
    logical :: writeRestart
    !> read the restart file from following file
    character(len=PathLen) :: readFileName
    !> write the restart file into a file with the following prefix
    character(len=PathLen) :: writePrefix
    !> control about when to do the restart
    type(tem_timeControl_type) :: timeControl
  end type

  !> Define quantities like the prefix, the mesh and the timestamp
  type tem_restartHeader_type
    !> a unique tag for the solver including version
    character(len=LabelLen) :: solverTag = 'unknown'
    !> solver creating the input data (Ateles, Musubi)
    character(len=LabelLen) :: solverKind
    !> the simulation name
    character(len=LabelLen) :: simName = ''
    !> name of the binary files (this variable is set for every timestamp
    !! differently, used for read and write)
    character(len=PathLen), allocatable :: binName(:)
    !> prefix for the binary files (includes directory and file prefix)
    character(len=PathLen), allocatable :: binPrefix(:)
    !> prefix for the header files (includes directory and file prefix)
    character(len=PathLen) :: headerPrefix
    !> the current time generated from tem_time_module
    character(len=16) :: timestamp
    !> mesh directory
    character(len=pathLen) :: meshDir
    !> Number of elements as read from the restart header
    integer       :: nElems
  end type

  !> The restart type defining everything related to the
  !! disk input/output
  type tem_restart_type
    !> communicator for the processes participating in this restart (might be
    !! only a subset of the global communicator)
    type(tem_comm_env_type) :: comm
    !> actual number of elements in the current chunk (= chunkSize or
    ! tree%nElems-(nChunks-1)*ChunkSize)
    integer :: nChunkElems
    !> number of elements that fit in the buffer
    integer :: chunkSize
    !> number of chunks per var system
    integer :: nChunks
    !> the maximum number of chunks to dump (max(nChunks))
    integer :: maxnChunks
    !> Control the behavior of the restart, like at which point in time etc.
    type(tem_restartControl_type) :: controller
    !> Define quantities like the prefix, the mesh and the timestamp
    type(tem_restartHeader_type)  :: header
    !> array of units to write binary data to
    integer, allocatable :: binaryUnit(:)
    !> handle for the MPI type to describe the view for IO in binary IO
    integer, allocatable :: ftype(:)
    !> handle for the MPI type describing the vector of data in each element
    integer, allocatable :: vectype(:)
    !> How many variable systems are defined for this restart
    integer :: nSys = 0
    !> global variable systems for all schemes
    type(tem_var_system_type), allocatable :: varSys(:)
    !> Number of degrees of freedom for each of the scalars entries in
    !! the varibale system.
    integer :: nDofs = 1
  end type

contains

!******************************************************************************!
  !> Read all necessary information for the restart from the lua config file.
  !!
  !! Include this routine into your general configuration load routine.
  !! The configuration looks as follows
  !!\verbatim
  !! restart = { read = 'restart/lastHeader.lua', -- Which file to restart from,
  !!                                              -- if any
  !!             write = 'restart/', -- Where to write the restart files to,
  !!                                 -- if any
  !!             time = { min = 0, max = 10, interval = 10} -- when to output
  !!             }
  !!\endverbatim
  !! Here, the restart is loaded from restart/lastHeader.lua and reads in the
  !! related data and configuration.
  !! Restart files are written out in restart/ folder
  !!
  subroutine tem_load_restart( me, conf, tree, timing, proc, parent_table,     &
    &                          key, refTimeControl, levelweight, dynamic )
    !---------------------------------------------------------------------------
    !> restart type to be filled
    type(tem_restart_type), intent(inout)             :: me
    !> lua configuration file
    type(flu_state)                                   :: conf
    !> mesh, provided in treelm format
    type(treelmesh_type)                              :: tree
    !> the timing for re-setting the times
    type(tem_time_type), intent(inout)                :: timing
    !> process
    type( tem_comm_env_type )                         :: proc
    !> optional parent handle
    integer, optional, intent(in)                     :: parent_table
    !> optional key for table
    character(len=*), optional, intent(in)            :: key
    !> Reference time control object, which passes properties to the tracker's
    !! timeControl object (mainly for useIterations property)
    type( tem_timeControl_type ),intent(in), optional :: refTimeControl
    !> the weights for each level
    real(kind=rk), optional, intent(in) :: levelweight(globalMaxLevels)
    !> Dynamic flag to check if dynamic load balancing is present
    logical, optional, intent(in)                     :: dynamic 
    !---------------------------------------------------------------------------
    character(len=32) :: localKey
    integer :: restart_table
    integer :: iError
    !---------------------------------------------------------------------------

    if (present(key)) then
      ! The table to look for is not named restart, look for this different
      ! key.
      localKey = key
    else
      ! Use the default name restart for the table.
      localKey = 'restart'
      ! Set current folder as default prefix for writing
      me%controller%writePrefix = './'
    end if

    me%controller%readRestart  = .false.
    me%controller%writeRestart = .false.

    ! Attempt to open the restart table (within another table, if a parent is
    ! given).
    call aot_table_open( L       = conf,                                       &
      &                  thandle = restart_table,                              &
      &                  parent  = parent_table,                               &
      &                  key     = trim(localKey) )

    ! If the restart table is present, the parameters are loaded.
    ! In case of dynamic load balancing, parameters are loaded 
    ! in a different manner i.e. timings etc. are read from balance table
    if (restart_table .ne. 0 ) then 
      ! Successfully opened the table.

      ! First we get all the informations of the read table.
      ! Reading the filename to read the restart data from.
      call aot_get_val( L       = conf,                                        &
        &               thandle = restart_table,                               &
        &               key     = 'read',                                      &
        &               val     = me%controller%readFileName,                  &
        &               ErrCode = iError ) 

      ! Set the restart flag
      me%controller%readRestart = (iError == 0)

      if (me%controller%readRestart) then
        ! Successfully obtained a filename to restart from, now go on and read
        ! the data from its header.
        call tem_write( "-----------------------------" )
        call tem_write( "Restart read parameters: " )
        call tem_write( "  filename : "//trim(me%controller%readFileName) )
        call tem_write( "-----------------------------" )
        call tem_restart_readHeader( me          = me,                         &
          &                          tree        = tree,                       &
          &                          timing      = timing,                     &
          &                          levelweight = levelweight,                &
          &                          proc        = proc,                       &
          &                          readMesh    = .true.)
      end if ! If reading restart

      ! Now we get all the information about writing restart data.
      call aot_get_val( L       = conf,                                        &
        &               thandle = restart_table,                               &
        &               key     = 'write',                                     &
        &               val     = me%controller%writePrefix,                   &
        &               ErrCode = iError )
      me%controller%writeRestart = (iError .eq. 0)

      if (me%controller%writeRestart) then
        ! Restart output location specified, attempt to find the points in time
        ! when to output.
        call tem_write( "-----------------------------" )
        call tem_write( "Restart write parameters: " )
        call tem_write( "  folder   : "//trim(me%controller%writePrefix) )
        call tem_write( "-----------------------------" )
        ! Read the time intervals from the lua config.
        call tem_load_timeControl( me             = me%controller%timeControl, &
          &                        conf           = conf,                      &
          &                        parent         = restart_table,             &
          &                        refTimeControl = refTimeControl )
      end if 

      call aot_table_close( L = conf, thandle = restart_table )
    end if ! If restart table present

    call aot_table_close( L=conf, thandle=restart_table )

    ! If dynamic load balancing is activated, just the read and write flags are
    ! set because they will be required and timings of balance table take
    ! precedence. 
    if( present(dynamic) ) then
      if( dynamic ) then
        ! Set the name of file to be read
        me%controller%readFileName = trim( me%header%headerPrefix )//'_lastHeader.lua'
        call tem_write("Reading restart file for dynamic load balancing")
        call tem_write( "  filename : "//trim(me%controller%readFileName) )
        ! Read the header to get various parameters
        call tem_restart_readHeader( me          = me,                         &
          &                          tree        = tree,                       &
          &                          timing      = timing,                     &
          &                          levelweight = levelweight,                &
          &                          proc        = proc,                       &
          &                          readMesh    = .true.)
      end if
    end if

  end subroutine tem_load_restart
!******************************************************************************!


!******************************************************************************!
  !> Initialize the global restart data type and prepare for the restart output
  !!
  !! This routine is called as well for read as for write restart tasks
  !! It must be called after tem_load_restart and before the first call to any
  !! of tem_restart_*
  !! This routine is called only when restart is read from the restart table.
  !! If restart is performed from the initial condition table, then a 
  !! different routine is invoked
  !!
  !! ATTENTION: The currently used MPI_SubArray allows only default integers
  !!            in the element counts and offsets, this limits the overall
  !!            problem size to roughly 2 billion elements right now!
  !!
  subroutine tem_init_restart( me, solver, nSys, varSys, tree, subTree,        &
    &                          nDofs, chunkSize, dynamic )
    !---------------------------------------------------------------------------
    !> The restart object to initialize.
    type(tem_restart_type), intent(inout) :: me
    !> Details about the solver.
    type(tem_solveHead_type), optional, intent(in) :: solver
    !> Number of variable systems associated with this output.
    !! This is ignored, if the data is actually provided by reading a restart.
    integer, intent(in) :: nSys
    !> Description of each variable system, one data file will be generated
    !! for each system. This is ignored, if the data is provided by reading
    !! a restart.
    type(tem_var_system_type), intent(in) :: varSys(nSys)
    !> Mesh, provided in treelm format
    type(treelmesh_type), intent(in) :: tree
    !> optional subTree of the given tree
    type(tem_subTree_type), optional, intent(in) :: subTree
    !> number of degrees of freedom for each variable of the equation system
    integer, optional, intent(in) :: nDofs
    !> use predefined chunkSize
    integer, optional, intent(in) :: chunkSize
    !> Dynamic load balancing flag
    logical, optional, intent(in) :: dynamic
    !---------------------------------------------------------------------------
    integer :: iSys ! counter for the var systems
    ! error variables
    integer :: iError, ErrErr 
    character(len=100) :: IOError
    integer :: resultlen = 100
    integer :: globElems
    integer :: elemOff
    integer :: locElems
    logical :: dyn_loadbal
    !---------------------------------------------------------------------------

    ! Inherit communicator information from the tree we are to act on in this
    ! restart object:
    if( present( subTree ))then
      me%comm%comm      = subTree%global%comm
      me%comm%rank      = subTree%global%myPart
      me%comm%comm_size = subTree%global%nParts
    else
      me%comm%comm      = tree%global%comm
      me%comm%rank      = tree%global%myPart
      me%comm%comm_size = tree%global%nParts
    end if
    me%comm%root      = 0

    ! IF dynamic load balancing is present, set the flag
    if( present( dynamic ) ) then
      dyn_loadbal = dynamic 
    else
      dyn_loadbal = .false.
    end if

    ! If this restart object is not meant to read data, set the output data
    ! according to the provided ar/interguments.
    ! Otherwise this stuff should already be present and properly filled
    ! by the restart header.
    if ( .not. me%controller%readRestart .and. .not. dyn_loadbal ) then
      me%nSys = nSys
      allocate( me%varSys(nSys) )
      me%varSys = varSys
      allocate( me%header%binName(nSys) )
    end if

    !KM: workaround to dump solverspecific unit when loading varsys from 
    !restart header
    if (me%controller%readRestart .or. dyn_loadbal)                            &
      &                        me%varSys(:)%solSpecUnit = varSys(:)%solSpecUnit

    ! if provided reset the simulation name and solver tag
    if (present(solver)) then
      me%header%simName   = solver%simName
      me%header%solverTag = tem_solverTag(solver)
    end if

    if(present(nDofs)) then
      me%nDofs = nDofs
    else
      me%nDofs = 1
    end if

    ! Set the prefix for the header file.
    me%header%headerPrefix = trim(me%controller%writePrefix)//                 &
      &                      trim(me%header%simName)

    ! Only allocate when dynamic load balancing is not required
    if ( .not. dyn_loadbal ) then
      allocate( me%header%binPrefix(me%nSys) )
      allocate( me%ftype(me%nSys) )
      allocate( me%vectype(me%nSys) )
      allocate( me%binaryUnit(me%nSys) )
    end if

    ! Reset all the values of the chunk information
    me%nChunkElems = 0
    me%chunkSize   = 0
    me%nChunks     = 0
    me%maxnChunks  = 0

    ! Translation of long_k integers to integers, as required by MPI.
    ! ATTENTION, this limits the ability of outputting restarts to 2 billion
    ! total elements!
    if(present( subTree ))then
      globElems = int(subTree%global%nElems)
      elemOff = int(subTree%elemOffset)
      locElems = int(subTree%nElems)
    else
      globElems = int(tree%global%nElems)
      elemOff = int(tree%elemOffset)
      locElems = int(tree%nElems)
    end if

    do iSys = 1, me%nSys
      ! Set the prefix for all binary files.
      me%header%binPrefix(iSys) = trim(me%controller%writePrefix)//            &
        &                         trim(me%varSys(iSys)%SystemName)//'_'

      ! A contiguous type to describe the vector per element.
      ! MPI_TYPE_CONTIGUOUS(COUNT, OLDTYPE, NEWTYPE, IERROR)
      call MPI_Type_contiguous( me%varSys( iSys )%nScalars*me%nDofs, rk_mpi,   &
        &                       me%vectype(iSys), iError)
      call MPI_Type_commit( me%vectype(iSys), iError )

      ! A subarray to describe the view of this process on the file.
      ! MPI_TYPE_CREATE_SUBARRAY(ndims, array_of_sizes, array_of_subsizes,
      !                          array_of_starts, order, oldtype, newtype)
      call MPI_Type_create_subarray( 1, [ globElems ], [ locElems ],           &
        &                            [ elemOff ], MPI_ORDER_FORTRAN,           &
        &                            me%vectype(iSys), me%ftype(iSys), iError)
      if (iError .ne. MPI_SUCCESS) then
        call MPI_ERROR_STRING( iError, IOError, resultlen, ErrErr )
        call tem_write( 'Read Type contiguous MPI WARNING: '// trim( IOError ))
        call tem_abort()
      end if
      ! commit the new contiguous type
      call MPI_Type_commit( me%ftype(iSys), iError )

      if (iError .ne. MPI_SUCCESS) then
        call MPI_ERROR_STRING( iError, IOError, resultlen, ErrErr )
        call tem_write( 'Read Type Commit MPI WARNING: '// trim( IOError ))
        call tem_abort()
      end if
    end do

    ! Get the total number of chunks necessary to write data-set through IO
    ! buffer to disk.
    call tem_restart_getTotalChunks( restart   = me,                           &
      &                              tree      = tree,                         &
      &                              subTree   = subTree,                      &
      &                              chunkSize = chunkSize )

  end subroutine tem_init_restart
!******************************************************************************!


!******************************************************************************!
  !> This subroutine initializes the restart in case of reading initial 
  !! conditions from restart file. The dependent scheme id is passed as input.
  !\ Author: Kartik Jain
  !!
  subroutine tem_init_read_restart( me, solver, varSys, tree, nDofs, iSys,     &
    &                                 chunkSize )
    !---------------------------------------------------------------------------
    !> The restart object to initialize.
    type(tem_restart_type), intent(inout) :: me
    !> Details about the solver.
    type(tem_solveHead_type), optional, intent(in) :: solver
    !> array of variable systems
    type(tem_var_system_type), intent(in) :: varSys(:)
    !> Mesh, provided in treelm format
    type(treelmesh_type), intent(in) :: tree
    !> number of degrees of freedom for each variable of the equation system
    integer, optional, intent(in) :: nDofs
    !> The id of variable system in global system
    integer, intent(in)       :: iSys
    !> use predefined chunkSize
    integer, optional, intent(in) :: chunkSize
    !---------------------------------------------------------------------------
    ! error variables
    integer :: iError, ErrErr 
    character(len=100) :: IOError
    integer :: resultlen = 100
    integer :: globElems
    integer :: elemOff
    !------------------------------------------------------------------------===

    ! Inherit communicator information from the tree we are to act on in this
    ! restart object:
    me%comm%comm      = tree%global%comm
    me%comm%rank      = tree%global%myPart
    me%comm%comm_size = tree%global%nParts
    me%comm%root      = 0
    
    !KM: workaround to dump solverspecific unit when loading varsys from 
    !restart header
!    if (me%controller%readRestart) &
!      & me%varSys(:)%solSpecUnit = varSys%solSpecUnit

    ! if provided reset the simulation name and solver tag
    if (present(solver)) then
      me%header%simName   = solver%simName
      me%header%solverTag = tem_solverTag(solver)
    end if

    if(present(nDofs)) then
      me%nDofs = nDofs
    else
      me%nDofs = 1
    end if

    ! Set the prefix for the header file.
    me%header%headerPrefix = trim(me%controller%writePrefix)//                 &
      &                      trim(me%header%simName)

    if( .not. allocated( me%header%binPrefix))                                 &
      &                  allocate( me%header%binPrefix(1) )
    if( .not. allocated( me%ftype))                                            &
      &                  allocate( me%ftype(1) )
    if( .not. allocated( me%vectype))                                          &
      &                  allocate( me%vectype(1) )
    if( .not. allocated( me%binaryUnit))                                       &
      &                  allocate( me%binaryUnit(1) )

    ! Reset all the values of the chunk information
    me%nChunkElems = 0
    me%chunkSize   = 0
    me%nChunks     = 0
    me%maxnChunks  = 0

    ! Translation of long_k integers to integers, as required by MPI.
    ! ATTENTION, this limits the ability of outputting restarts to 2 billion
    ! total elements!
    globElems = int(tree%global%nElems)
    elemOff = int(tree%elemOffset)

    ! Set the prefix for all binary files.
    me%header%binPrefix(1) = trim(me%controller%writePrefix)//                 &
      &                      trim(me%varSys(iSys)%SystemName)//'_'

    ! A contiguous type to describe the vector per element.
    ! MPI_TYPE_CONTIGUOUS(COUNT, OLDTYPE, NEWTYPE, IERROR)
    call MPI_Type_contiguous( me%varSys( iSys )%nScalars*me%nDofs, rk_mpi,     &
      &                       me%vectype(1), iError)
    call MPI_Type_commit( me%vectype(1), iError )

    ! A subarray to describe the view of this process on the file.
    ! MPI_TYPE_CREATE_SUBARRAY(ndims, array_of_sizes, array_of_subsizes,
    !                          array_of_starts, order, oldtype, newtype)
    call MPI_Type_create_subarray( 1, [ globElems ], [ tree%nElems ],          &
      &                            [ elemOff ], MPI_ORDER_FORTRAN,             &
      &                            me%vectype(1), me%ftype(1), iError)
    if (iError .ne. MPI_SUCCESS) then
      call MPI_ERROR_STRING( iError, IOError, resultlen, ErrErr )
      call tem_write( 'Read Type contiguous MPI WARNING: '// trim( IOError ))
      call tem_abort()
    end if
    ! commit the new contiguous type
    call MPI_Type_commit( me%ftype(1), iError )

    if (iError .ne. MPI_SUCCESS) then
      call MPI_ERROR_STRING( iError, IOError, resultlen, ErrErr )
      call tem_write( 'Read Type Commit MPI WARNING: '// trim( IOError ))
      call tem_abort()
    end if

    ! Get the total number of chunks necessary to write data-set through IO
    ! buffer to disk.
    call tem_restart_getTotalChunks( restart   = me,                           &
      &                              tree      = tree,                         &
      &                              chunkSize = chunkSize )
    
    end subroutine tem_init_read_restart
!******************************************************************************!


!******************************************************************************!
  !> This subroutine dumps the given chunk to a given position in the restart
  !! file.
  !!
  subroutine tem_restart_writeData( restart, chunk )
    !---------------------------------------------------------------------------
    !> The restart object describing how and what to output.
    type(tem_restart_type), intent(inout) :: restart
    !> The data to output.
    !! It is organized as a serialized array of all scalar entries of all
    !! variable systems. Where first all the data for the nElems of the first
    !! variable system is provided. Within each variable system the data is
    !! organized elementwise.
    real(kind=rk), intent(in) :: chunk(:)
    !---------------------------------------------------------------------------
    integer :: ioStatus( mpi_status_size )
    integer :: iError, ErrErr
    character(len=100) :: IOError
    integer :: resultlen = 100
    integer :: iSys
    integer :: offset        ! offset for the different var systems
    integer :: nWords
    !---------------------------------------------------------------------------
    offset = 1
    ! For all variable systems defined in the restart header...
    do iSys = 1, restart%nSys
      nWords = restart%nChunkElems * restart%varSys(iSys)%nScalars             &
        &    * restart%nDofs
      ! Now write the collected data (from the state array and prob. derived
      ! quantities) into the file, using the view defined in
      ! tem_restart_openWrite
      ! arguments:
      ! file handle = binary unit opened in mpi_file_open
      ! initial address of buffer = first entry to dump within the chunk
      !                             this is ad
      call mpi_file_write_all( restart%binaryUnit(iSys), chunk(offset),        &
        &                      restart%nChunkElems, restart%vectype(iSys),     &
        &                      iostatus, iError )

      if (iError .ne. mpi_success) then
        call mpi_error_string( iError, IOError, resultlen, ErrErr )
        call tem_write( 'Write All MPI WARNING: '//trim(IOError) )
        call tem_abort()
      end if

      ! Increment the offset for the next variable system.
      offset = offset + nWords
    end do

  end subroutine tem_restart_writeData
!******************************************************************************!


!******************************************************************************!
  !> Read data from a restart file.
  !!
  subroutine tem_restart_readData( restart, chunk )
    !---------------------------------------------------------------------------
    !>
    type(tem_restart_type), intent(inout) :: restart
    !>
    real(kind=rk), intent(out) :: chunk(:)
    !---------------------------------------------------------------------------
    ! defining local variables
    integer :: nWords
    integer :: iSys
    integer :: offset        ! offset for the different var systems
    !---------------------------------------------------------------------------

    ! \todo:KJ: For the new restart only the requested variable system should 
    ! be read. 
    offset = 1

    do iSys = 1, restart%nSys
      call tem_restart_readData_single( restart, chunk, iSys, offset )
      nWords = restart%varSys(iSys)%nScalars*restart%nChunkElems*restart%nDofs
      offset = offset + nWords
    end do

  end subroutine tem_restart_readData
!******************************************************************************!


!******************************************************************************!
  !> subroutine which reads data from restart file corresponding to the input
  !! variable system
  !!
  subroutine tem_restart_readData_single( restart, chunk, iSys, offset )
    !---------------------------------------------------------------------------
    !>
    type(tem_restart_type), intent(inout) :: restart
    !>
    real(kind=rk), intent(out)            :: chunk(:)
    !>
    integer, intent(in)                   :: iSys
    !>
    integer, intent(in)                   :: offset
    !---------------------------------------------------------------------------
    ! defining local variables
    integer :: iostatus( MPI_STATUS_SIZE )
    integer :: iError, ErrErr
    character(len=100) :: IOError
    integer :: resultlen = 100
    !---------------------------------------------------------------------------
    
    ! MPI_FILE_READ_ALL(fh, buf, count, datatype, status)

    call mpi_file_read_all( restart%binaryUnit(iSys), chunk(offset),           &
      &                     restart%nChunkElems, restart%vectype(iSys),        &
      &                     iostatus, iError )

    if(iError .ne. MPI_SUCCESS)then
      call MPI_ERROR_STRING( iError, IOError, resultlen, ErrErr )
      call tem_write( 'Read All MPI WARNING: '// trim( IOError ))
      call tem_write( 'Aborting. ')
      call tem_abort()
    end if

  end subroutine tem_restart_readData_single
!******************************************************************************!


!******************************************************************************!
  !> open the restart dump file and write out the 'normal' restart header.
  !!
  subroutine tem_restart_openRead(me)
    !---------------------------------------------------------------------------
    !> the restart information
    type(tem_restart_type) :: me
    !---------------------------------------------------------------------------
    integer :: iSys !< var system counter
    !---------------------------------------------------------------------------
    call tem_write(' Open read... ')

    ! read the restart header file
    ! prepare everything for the serialization of the data
    do iSys = 1, me%nSys
      ! Invoke the routine to open single variable system from restart file
      call tem_restart_openRead_single(me, iSys)
    end do

  end subroutine tem_restart_openRead
!******************************************************************************!


!******************************************************************************!
  !> Open the restart dump file and set file view for the input variable system
  !!
  subroutine tem_restart_openRead_single(me, iSys)
    !---------------------------------------------------------------------------
    !> the restart information
    type(tem_restart_type) :: me
    !> The ID of variable system requested
    integer, intent(in)    :: iSys
    !---------------------------------------------------------------------------
    ! variables to catch possible MPI I/O errors
    integer :: iError, ErrErr
    character(len=100) :: IOError
    integer :: resultlen = 100
    !---------------------------------------------------------------------------
    
    ! open the binary file for MPI I/O
    call MPI_FILE_OPEN( me%comm%comm,                                          &
      &                 trim( me%header%binName(iSys) ),                       &
      &                 MPI_MODE_RDONLY,                                       &
      &                 MPI_INFO_NULL, me%binaryUnit( iSys ),                  &
      &                 iError )

    if (iError .ne. MPI_SUCCESS) then
      call MPI_ERROR_STRING( iError, IOError, resultlen, ErrErr )
      call tem_write( 'Read File Open MPI WARNING: '// trim( IOError ))
      call tem_abort()
    end if

    ! set the view of each process on the file opened above
    call MPI_FILE_SET_VIEW( me%binaryUnit(iSys), 0_MPI_OFFSET_KIND,            &
      &                     me%vectype(iSys), me%ftype(iSys), "native",        &
      &                     MPI_INFO_NULL, iError )

    if(iError .ne. MPI_SUCCESS)then
      call MPI_ERROR_STRING( iError, IOError, resultlen, ErrErr )
      call tem_write( 'Read File Set view MPI WARNING: '// trim( IOError ))
      call tem_abort()
    end if

  end subroutine tem_restart_openRead_single
!******************************************************************************!


!******************************************************************************!
  !> This subroutine closes the restart dump file and writes the last header.
  subroutine tem_restart_closeRead( me, iSys_idx )
    !---------------------------------------------------------------------------
    !> the restart infotmation
    type(tem_restart_type)        :: me
    !>
    integer, optional, intent(in) :: iSys_idx
    !---------------------------------------------------------------------------
    integer :: iSys ! var systems counter
    ! temporary buffer array for empty reads
    real(kind=rk) :: chunk(1)
    integer :: iChunk
    !---------------------------------------------------------------------------

    call tem_write('Close Read...')

    ! If local process has performed less reads than global maximum required,
    ! we need to complete as many empty reads as the difference to maxnChunks
    ! to satisfy collective MPI-IO operations.
    me%nChunkElems = 0
    do iChunk = me%nChunks+1, me%maxnChunks
      call tem_restart_readData(restart = me, chunk = chunk)
    end do

    ! If iSys is provided as an input, we close the file only for that
    ! particular iSys, otherwise we loop over all the systems and close 
    ! the restart files
    if( present( iSys_idx ) ) then
      ! Invoke the routine to close current variable system file
      call tem_restart_closeRead_single(me, iSys_idx)
    else
      do iSys = 1, me%nSys
        ! Invoke the routine to close current variable system file
        call tem_restart_closeRead_single(me, iSys)
      end do
    end if 

  end subroutine tem_restart_closeRead
!******************************************************************************!


!******************************************************************************!
  !> Close the restart dump file corresponding to a particular variable system
  !!
  subroutine tem_restart_closeRead_single(me, iSys)
    !---------------------------------------------------------------------------
    !> the restart infotmation
    type(tem_restart_type) :: me
    !> The ID of input variable system
    integer, intent(in)    :: iSys
    !---------------------------------------------------------------------------
    ! variables to catch possible MPI I/O errors
    integer :: iError, ErrErr
    character(len=100) :: IOError
    integer :: resultlen = 100
    !---------------------------------------------------------------------------
    
    ! now close the binary file
    call MPI_File_close(me%binaryUnit(iSys), iError)
    if(iError .ne. MPI_SUCCESS)then
      call MPI_ERROR_STRING(iError, IOError, resultlen, ErrErr)
      call tem_write( 'Read Close File MPI WARNING: '// trim(IOError) )
      call tem_abort()
    end if

  end subroutine tem_restart_closeRead_single
!******************************************************************************!


!******************************************************************************!
  !> open the restart dump file and write out the 'normal' restart header.
  !!
  subroutine tem_restart_openWrite( me, tree, timing, subTree, label, noDeps )
    !---------------------------------------------------------------------------
    !> the restart infotmation
    type(tem_restart_type) :: me
    !> mesh, provided in treelm format
    type(treelmesh_type) :: tree
    !> current simulation time information
    type(tem_time_type),intent(in) :: timing
    !> optional subTree of the given tree
    type(tem_subTree_type), optional, intent(inout) :: subTree
    !> additional label for the filename (needed for tracking in harvester
    !! format)
    character(len=labelLen), optional, intent(in) :: label
    !>
    logical, optional, intent(in) :: noDeps
    !---------------------------------------------------------------------------
    ! variables to catch possible MPI I/O errors
    integer :: iError, ErrErr
    character(len=100) :: IOError
    integer :: resultlen = 100
    integer :: iSys ! var system counter
    integer :: pos
    character(len=pathLen) :: prefix
    logical :: meshChange_loc
    type(tem_global_type) :: global_loc
    !---------------------------------------------------------------------------

    ! Update the timestamp
    me%header%timestamp = tem_generate_timeStamp( timing )

    if(present( subTree ))then
      global_loc = subTree%global
    else
      global_loc = tree%global
    endif

    meshChange_loc = global_loc%meshChange

    ! communicate wether the mesh has changed since last time dumping it
    call MPI_ALLREDUCE( meshChange_loc, global_loc%meshChange, 1,              &
      &                 MPI_LOGICAL, MPI_LOR, global_loc%comm, iError )

    ! if the mesh has changed ...
    if( global_loc%meshChange)then
      ! ... change the dirname and set the meshChange to false
      global_loc%meshChange = .false.
      ! get the position of the last path seperator
      pos = INDEX(trim(global_loc%dirname), pathSep, .true.)
      if(present(label))then
        prefix = trim(global_loc%dirname(1:pos))//trim(label)//'_'
      else
        prefix = trim(global_loc%dirname(1:pos))
      end if
      write(global_loc%dirname,'(a)') trim(prefix)//                           &
        &                                       trim( me%header%timestamp )//'_'
      ! ... and dump it
      if(present( subTree ))then
        call tem_dump_subTree( subTree, tree )
      else
        call dump_treelmesh( tree )
      end if
    end if

    ! copy back the global information to the tree or subTree
    if(present( subTree ))then
      subTree%global = global_loc
    else
      tree%global = global_loc
    endif

    ! prepare everything for the serialization of the data
    do iSys = 1, me%nSys

      ! define the name of the file to write the binary data to
      write(me%header%binName(iSys),'(a)') trim( me%header%binPrefix( iSys ))  &
        &                               // trim( me%header%timestamp )         &
        &                               // tem_create_EndianSuffix()

      ! open the binary file for MPI I/O
      call MPI_FILE_OPEN( me%comm%comm,                                        &
        &                 trim( me%header%binName(iSys) ),                     &
        &                 MPI_MODE_WRONLY+MPI_MODE_CREATE,                     &
        &                 MPI_INFO_NULL, me%binaryUnit( iSys ),                &
        &                 iError )

      if (iError .ne. MPI_SUCCESS) then
        call MPI_ERROR_STRING( iError, IOError, resultlen, ErrErr )
        call tem_write( 'File Open MPI WARNING: '// trim( IOError ))
        call tem_abort()
      end if

      ! set the view of each process on the file opened above
      call MPI_FILE_SET_VIEW( me%binaryUnit(iSys), 0_MPI_OFFSET_KIND,          &
        &                     me%vectype(iSys), me%ftype(iSys), "native",      &
        &                     MPI_INFO_NULL, iError )

      if(iError .ne. MPI_SUCCESS)then
        call MPI_ERROR_STRING( iError, IOError, resultlen, ErrErr )
        call tem_write( 'File Set view MPI WARNING: '// trim(IOError) )
        call tem_abort()
      end if

    end do

    ! write out a regular restart header
    call tem_restart_writeHeader( me      = me,                                &
      &                           tree    = tree,                              &
      &                           nSys    = me%nSys,                           &
      &                           subTree = subTree,                           &
      &                           timing  = timing,                            &
      &                           noDeps  = noDeps,                            &
      &                           varSys  = me%varSys )

  end subroutine tem_restart_openWrite
!******************************************************************************!


!******************************************************************************!
  !> This subroutine closes the restart dump file and writes the last header.
  !!
  subroutine tem_restart_writeHeader( me, tree, nSys, varSys, subTree,         &
    &                                 timing, lastHeader, noDeps )
    !---------------------------------------------------------------------------
    !> the restart header info
    type(tem_restart_type) :: me
    !> mesh, provided in treelm format
    type(treelmesh_type) :: tree
    !> the number of var systems
    integer, intent(in) :: nSys
    !> the used var systeme
    type(tem_var_system_type), intent(inout) :: varSys(:)
    !> optional subTree of the given tree
    type(tem_subTree_type), optional, intent(in) :: subTree
    !>
    type(tem_time_type), intent(in) :: timing
    !> is this header a last header
    logical, optional :: lastHeader
    !>
    logical, optional, intent(in) :: noDeps
    !---------------------------------------------------------------------------
    integer :: iSys !< counter for the var systems
    character(len=1024) :: filename
    type(aot_out_type) :: conf !< aotus lua state to write output
    logical :: isLast
    type(tem_global_type) :: global_loc
    integer(kind=long_k) :: nElems
    !---------------------------------------------------------------------------

    if (present(lastHeader)) then
      isLast = lastHeader
    else
      isLast = .false.
    end if

    ! now write out the header file
    ! only one process (root) writes out the header file
    if (me%comm%rank .eq. 0) then
      ! choose the right filename
      if (isLast) then
        filename = trim( me%header%headerPrefix )// '_lastHeader.lua'
      else
        filename = trim( me%header%headerPrefix )//'_header_'                  &
         &      // trim(me%header%timestamp)//'.lua'
      end if
      ! Open up the restart header lua file, so we can write the stuff using
      ! the aotus library
      call aot_out_open(conf, filename)

      ! the binary names; for multiple schemes a list of binaries have to be
      ! written
      call aot_out_open_table(conf, 'binary_name')
      do iSys = 1, nSys
        call aot_out_val( put_conf = conf,                                     &
          &               val      = trim(me%header%binName(iSys)) )
      end do
      call aot_out_close_table(conf)

      if( present( subTree ))then
        global_loc = subTree%global
        nElems = subTree%global%nElems
      else
        global_loc = tree%global
        nElems = tree%global%nElems
      end if
      
      ! the mesh info
      call tem_mesh_out( me = global_loc, conf = conf )

      ! the time stamp
      !call aot_out_val( conf, trim(me%header%timestamp), 'time_point')
      call tem_time_out( conf = conf, me = timing, key = 'time_point' )
      ! the total number of elements
      call aot_out_val( conf, nElems, 'nElems')
      ! the number of dofs for each scalar variable of the equation system
      call aot_out_val( conf, me%nDofs, 'nDofs')
      ! the solver tag (solver name and version)
      call aot_out_val( conf, trim( me%header%solverTag), 'solver')
      ! the variable systems (one per scheme), incl. the solver specific
      ! information from the solSpecUnit
      call tem_var_sys_out( sys       = varSys,                                &
        &                   conf      = conf,                                  &
        &                   ssUnit    = varSys(:)%solSpecUnit,                 &
        &                   dumpnVars = .true.,                                &
        &                   noDeps    = noDeps  )
      ! close the restart header file
      call aot_out_close(conf)
    end if
  end subroutine tem_restart_writeHeader
!******************************************************************************!


!******************************************************************************!
  !> read the restart header lua file and hand the information to the
  !! required data types, re-set the time ...
  !!
  subroutine tem_restart_readHeader( me, tree, timing, proc, levelweight,      &
    &                                readMesh )
    !---------------------------------------------------------------------------
    !> the restart header info
    type(tem_restart_type) :: me
    !> mesh, provided in treelm format
    type(treelmesh_type) :: tree
    !> the timing for re-setting the times
    type(tem_time_type), intent(inout) :: timing
    !>
    type( tem_comm_env_type ) :: proc
    !> Do we need to load the mesh from disk?
    !! This is only required for initial restart loading, but not when we 
    !! load the states for the fields
    logical, optional, intent(in) :: readMesh
    !>
    real(kind=rk), optional, intent(in) :: levelweight(globalMaxLevels)
    !---------------------------------------------------------------------------
    integer :: thandle
    integer :: iSys, iError
    type( flu_State ) :: conf
    character(len=labelLen) :: buffer
    integer :: myPart
    integer :: nParts
    integer :: comm
    !---------------------------------------------------------------------------
    call tem_write( 'Opening Restart Header '//me%controller%readFileName )

    myPart = proc%rank
    nParts = proc%comm_size
    comm = proc%comm

    ! Open the restart header file
    call tem_open_distconf( L        = conf,                                   &
      &                     filename = trim(me%controller%readFileName),       &
      &                     proc     = proc)
    ! Load the number of elements from the restart header file for sanity check 
    call aot_get_val( L       = conf,                                          &
      &               key     = 'nElems',                                      &
      &               val     = me%header%nElems,                              &
      &               ErrCode = iError,                                        &
      &               default = 1 )

    ! Load the solver name
    call aot_get_val( L       = conf,                                          &
      &               key     = 'solver',                                      &
      &               val     = buffer,                                        &
      &               ErrCode = iError,                                        &
      &               default = '' )
    me%header%solverTag = trim( buffer )
    call tem_write( 'Solver: '// trim(me%header%solverTag ))
    ! Load the number of dofs for each scalar variable of the equation system
    call aot_get_val( L       = conf,                                          &
      &               key     = 'nDofs',                                       &
      &               val     = me%nDofs,                                      &
      &               ErrCode = iError,                                        &
      &               default = 1 )

    ! Load the timestamp from the header
    call tem_time_load( conf = conf, key = 'time_point', me = timing )
    me%header%timestamp = tem_generate_timestamp( timing%rTime )
    timing%rStart = timing%rTime
    call tem_write( 'Restarting from point in time: ', timing%rTime )
    call tem_write( '                        iTime: ', timing%iTime )

    ! Load the variable systems
    call tem_var_sys_load( conf = conf, sys = me%varSys )
    
    me%nSys = size( me%varSys )

    call tem_var_sys_debug( sys   = me%varSys,                                 &
      &                     level = 1 )

    call aot_table_open( L=conf, thandle = thandle, key = 'binary_name' )

    ! MH: Adding workaround for crashing Musubi with nLevels > 1
    if( allocated( me%header%binName)) then
      if( size(me%header%binName) .ne. me%nSys ) then
        deallocate( me%header%binName )
        allocate( me%header%binName( me%nSys ))
      end if
    else
      allocate( me%header%binName( me%nSys ))
    end if

    ! Load the binary file names for each variable system defined
    do iSys = 1, me%nSys
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               pos     = iSys,                                        &
        &               val     = me%header%binName( iSys ),                   &
        &               ErrCode = iError,                                      &
        &               default = '' )
    enddo

    call aot_table_close( L = conf, thandle = thandle )

    ! Load the tree from the "mesh = 'mesh/'" entry in the header file
    if ( present(readMesh) ) then
      if (readMesh) then
        call load_tem( me          = tree,                                     &
          &            conf        = conf,                                     &
          &            myPart      = myPart,                                   &
          &            levelweight = levelweight,                              &
          &            nParts      = nParts,                                   &
          &            comm        = comm,                                     &
          &            meshDir     = me%header%meshDir )
      end if
    end if
    
    ! If the number of nElems from restart header is different from those in the
    ! tree, that means that a wrong mesh has been loaded and we stop the 
    ! simulation at this point
    if ( tree%global%nElems .ne. me%header%nElems ) then
      call tem_write( 'Number of elements in restart header different '//      &
        &             'from mesh' )
      call tem_write( 'Aborting Simulation ..... ' )
      call tem_abort() 
    end if
    
    call close_config( conf )

  end subroutine tem_restart_readHeader
!******************************************************************************!


!******************************************************************************!
  !> Complete a number of empty writes (due to higher amount of mpi_file_writes
  !! from other processes to finalize the write process), close the restart dump
  !! file and write the last header.
  !!
  subroutine tem_restart_closeWrite( me, tree, subTree, timing )
    !---------------------------------------------------------------------------
    !> the restart information
    type(tem_restart_type) :: me
    !> mesh, provided in treelm format
    type(treelmesh_type) :: tree
    !> optional subTree of given tree
    type(tem_subTree_type), optional :: subTree
    !> The timing object holding the simulation time
    type(tem_time_type), intent(in) :: timing
    !---------------------------------------------------------------------------
    ! variables to catch possible MPI I/O errors
    integer :: iError, ErrErr
    character(len=100) :: IOError
    integer :: resultlen = 100
    integer :: iSys ! var systems counter
    integer :: iChunk ! chunk counter
    ! temporary buffer array for empty writes
    real(kind=rk) :: chunk(1)
    !---------------------------------------------------------------------------
    ! Check, if the number of calls to mpi_file_write_all corresponds to
    ! the maximum number throughout all processes
    ! if local process has performed less writes, we need to complete
    ! as many empty writes as the difference to maxnChunks
    me%nChunkElems = 0
    do iChunk = me%nChunks+1, me%maxnChunks
      call tem_restart_writeData( restart = me, chunk = chunk )
    end do

    do iSys = 1, me%nSys
      ! now close the binary file
      call MPI_File_close(me%binaryUnit(iSys), iError)
      if (iError .ne. MPI_SUCCESS) then
        call MPI_ERROR_STRING(iError, IOError, resultlen, ErrErr)
        call tem_write( 'MPI WARNING: '// trim(IOError) )
        call tem_abort()
      end if
    end do

    call MPI_Barrier( me%comm%comm, iError )

    ! Now write out the 'last' header which points to the last successful
    ! restart file.
    call tem_restart_writeHeader( me         = me,                             &
      &                           tree       = tree,                           &
      &                           nSys       = me%nSys,                        &
      &                           subTree    = subTree,                        &
      &                           timing     = timing,                         &
      &                           varSys     = me%varSys,                      &
      &                           lastHeader = .true. )

  end subroutine tem_restart_closeWrite
!******************************************************************************!


!******************************************************************************!
  !> calculate the maximum number of elements which fit into the output buffer
  !! = chunk
  !!
  !! for a given set of variable systems with their nScalar values to dump
  !! Also, calculate the number of local chunks required to dump all the data
  !! = nChunks
  !! Finally find the globally largest number of nChunks
  !!
  subroutine tem_restart_getTotalChunks( restart, tree, subTree, chunkSize )
    !---------------------------------------------------------------------------
    !> the restart type
    type( tem_restart_type ), intent(inout) :: restart
    !> mesh, provided in treelm format
    type(treelmesh_type), intent(in) :: tree
    !> optional subTree
    type(tem_subTree_type), optional, intent(in) :: subTree
    !> optional predefined chunksize
    integer, optional, intent(in) :: chunkSize
    !---------------------------------------------------------------------------
    integer :: nTotalScalars
    integer :: locElems
    integer :: iError ! MPI error
    integer :: rank
    integer :: comm
    !---------------------------------------------------------------------------
    ! Get the number of total scalars.
    nTotalScalars = sum(restart%varSys(:)%nScalars)
    if ( nTotalScalars .eq. 0 ) then
      call tem_write( '!! Error !! No variable found in restart var_sys !!' )
      call tem_write( 'May be variable label in requested system is not ' //   &
        &             'found in global varsys' )
      call tem_abort()
    endif

    if(present( subTree ))then
      locElems = subTree%nElems
      comm = subTree%global%comm
    else
      locElems = tree%nElems
      comm = tree%global%comm
    end if

    if( present( chunkSize ))then
      restart%chunkSize = chunkSize
    else
      ! Get the number of elements that fit into the IO buffer.
      restart%chunkSize = min( io_buffer_size                                  &
        &                     / (nTotalScalars*restart%nDofs),                 &
        &                    locElems )
    end if

    ! check if at least one complete element fits into the buffer
    ! if not abort since no valid output can be garanteed
    if (restart%chunkSize .le. 0) then
      call tem_write('The chosen io_buffer_size of ', io_buffer_size)
      call tem_write('is too small for outputting ', nTotalScalars)
      call tem_write('scalar values with ', restart%nDofs)
      call tem_write('degrees of freedom!')
      call tem_write('Please increase the io_buffer_size to at least (MB) ',   &
        &            real(nTotalScalars*restart%nDofs)/real(131072))
      call tem_abort()
    endif

    ! Get the number of chunks which are needed to dump all values to disk.
    ! This needs to be rounded up, to cover also a possible last incomplete
    ! chunk.
    restart%nChunks = ceiling( real(locElems, kind=rk)                         &
      &                       / real(restart%chunkSize, kind=rk) )

    ! identify the maximum number of chunks throughout all processes
    ! and store that into restart%maxnChunks
    call MPI_Allreduce( restart%nChunks, restart%maxnChunks, 1,                &
      &                 MPI_INTEGER, MPI_MAX, comm, iError )

    call MPI_Comm_Rank( comm, rank, iError )

  end subroutine tem_restart_getTotalChunks
!******************************************************************************!


!******************************************************************************!
  !> Finalizing a restart object.
  !!
  subroutine tem_restart_finalize( me )
    !---------------------------------------------------------------------------
    !> the restart type to close
    type( tem_restart_type ), intent(inout) :: me
    !---------------------------------------------------------------------------
    integer :: iSys
    !---------------------------------------------------------------------------

    do iSys=1,me%nSys
      call tem_restart_finalize_single(me, iSys)
    end do

  end subroutine tem_restart_finalize
!******************************************************************************!


!******************************************************************************!
  !> Finalizing a restart object for one system
  !!
  subroutine tem_restart_finalize_single( me, iSys )
    !---------------------------------------------------------------------------
    !> the restart type to close
    type( tem_restart_type ), intent(inout) :: me
    !> current variable system
    integer, intent(in)                     :: iSys
    !---------------------------------------------------------------------------
    integer :: iError, ErrErr
    character(len=100) :: IOError
    integer :: resultlen = 100
    !---------------------------------------------------------------------------

    ! free the contiguous type
    call MPI_Type_free(me%ftype(iSys), iError)
    if(iError .ne. MPI_SUCCESS)then
      call MPI_ERROR_STRING(iError, IOError, resultlen, ErrErr)
      call tem_write( 'Read Free Type MPI WARNING: '// trim(IOError) )
      call tem_abort()
    end if
    call MPI_Type_free(me%vectype(iSys), iError)

  end subroutine tem_restart_finalize_single
!******************************************************************************!


end module tem_restart_module
!******************************************************************************!
