! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Geometrical shape definitions for pre-defined bodies
!!
!! \author ...
!!
!! This module provides data types and definitions for pre-defined geometrical
!! objects such as points, lines, planes, cubes, spheres and more.
!!
!! \section shapes Shapes
!! - [canond]
!! - All elements
!! All elements in the tree are part of the tracking object.
!! Here you can define global [tem_reduction] for tracking the global state of
!! a simulation
!!
!! [canond]: @ref tem_canonicalnd_module "canoND"
!! [tem_reduction]: @ref tem_reduction_module "reduction operations"
!!
module tem_shape_module

  ! include treelm modules
  use env_module,             only: rk, long_k, labelLen, pathLen
  use tem_aux_module,         only: tem_abort
  use tem_tools_module,       only: tem_write
  use tem_property_module,    only: prp_hasBnd, prp_fluidify
  use tem_canonicalND_module, only: tem_canonicalND_type, tem_canonicalND_out, &
    &                               tem_load_canonicalND,                      &
    &                               tem_getNextCoordOfcanonicalND
  use tem_logging_module,     only: tem_log, tem_toStr, logUnit

  ! include aotus modules
  use aot_table_module, only: aot_table_open, aot_table_close,                 &
    &                         aot_table_length, aot_get_val
  use aotus_module,     only: flu_State, aoterr_NonExistent
  use aot_out_module,   only: aot_out_type, aot_out_open_table,                &
    &                         aot_out_close_table

  implicit none

  private

  public :: tem_shape_type
  public :: tem_load_shape
  public :: tem_global_shape
  public :: tem_canonicalND_shape
  public :: tem_property_shape
  public :: tem_local_shape
  public :: tem_getNextCoordOfShapePoint
  public :: tem_shape_out
  
  !> Parameters for different tracking shapes
  !> Global mesh
  integer, parameter :: tem_global_shape = 0
  !> canonical ND object
  integer, parameter :: tem_canonicalND_shape = 1
  !> elements that has a certain property
  integer, parameter :: tem_property_shape = 2
  !> elements on the local partition
  integer, parameter :: tem_local_shape = 3
  ! add further possible object types here like
  ! spherical objects (may include spheres, circles, cylinders, ...)
  ! tetrahedrical objects (may include tetrahedrons, triangles, ...)
  ! also add the types below in the shpae type

  !> Complete shape definitions
  type tem_shape_type
    !> a kind of the shape defined.
    character(len=labelLen) :: kind
    !> a identification for the shape
    integer :: shapeID = 0
    !> canonical definition
    type(tem_canonicalND_type),allocatable :: canoND(:)
    !> property bits
    integer(kind=long_k) :: propBits = 0_long_k
  end type tem_shape_type

  interface tem_shape_out
    module procedure tem_shape_out_scal
    module procedure tem_shape_out_vec
  end interface tem_shape_out

  interface tem_load_shape
    module procedure tem_load_shapes
    module procedure tem_load_shape_single
  end interface tem_load_shape

  contains

!******************************************************************************!
  !> Read in an arbitrary shapes from a lua file defined as multiple tables
  !!
  !! read a shape like for example inside a tracking table
  !! \verbatim
  !! tracking = {  
  !!             {variable = { 'velocity' },\n
  !!              shape = { kind = 'canoND',\n
  !!                        object = { origin = { 1.0, 1.0, 1.0 },\n
  !!                                   vec = { 2.0, 2.0, 2.0 },\n
  !!                                   segments = { 10, 20, 30 } }\n
  !!                      }}\n
  !!            } -- tracking table\n
  !! \endverbatim
  !! elements that has a certain property can also be tracked.
  !! This feature enables us to track boundary elements.
  !! \verbatim
  !! tracking = { \n
  !!              { variable = { 'velocity' },\n
  !!                shape = { kind = 'property',\n
  !!                          property = {'boundary'} },\n
  !!                      }
  !!              }\n
  !!            } -- tracking table\n
  !! \endverbatim
  !!
  subroutine tem_load_shapes( me, conf, parent, key, iError )
    !---------------------------------------------------------------------------
    !> array of shape type defined in a lua file
    type(tem_shape_type), allocatable, intent(out) :: me(:)
    !> lua config file to load shape from
    type(flu_state) :: conf
    !> optional parent handle
    integer, optional, intent(in) :: parent
    !> optional key to load from
    character(len=*), optional, intent(in) :: key
    !> error flag
    integer, intent(out), optional :: iError
    !---------------------------------------------------------------------------
    character(len=32) :: localKey
    integer :: nShapes  ! number of shape table entries
    integer :: iShape, shape_table, sub_table
    !---------------------------------------------------------------------------
    if( present( key )) then
      localKey = key
    else
      localKey = 'shape'
    endif

    ! open the table
    ! shape = {}
    call aot_table_open( L       = conf,                                       &
      &                  thandle = shape_table,                                &
      &                  parent  = parent,                                     &
      &                  key     = trim( localKey ))

    call aot_table_open( L       = conf,                                       &
      &                  parent  = shape_table,                                &
      &                  thandle = sub_table,                                  &
      &                  pos     = 1 )

    if ( sub_table .eq. 0 ) then ! shape is a single table
      ! load table from parent shape_table
      call aot_table_close(L=conf, thandle=sub_table)
      allocate( me( 1 ))
      call tem_load_shape_single( me        = me(1),                           &
        &                         conf      = conf,                            &
        &                         sub_table = shape_table,                     &
        &                         iError    = iError )
    else ! multiple table
      call aot_table_close( L=conf, thandle=sub_table )
      ! open the first entry in the shape table
      ! shape = {{this entry}, {second entry}}
      ! get the number of entries in  the shape table
      nShapes = aot_table_length( L=conf, thandle=shape_table )
      call tem_write( 'Number of shapes defined: ', nShapes )
      allocate( me( nShapes ))
      do iShape = 1, nShapes
        call aot_table_open( L       = conf,                                   &
          &                  thandle = sub_table,                              &
          &                  parent  = shape_table,                            &
          &                  pos     = iShape )
        call tem_load_shape_single( me        = me(iShape),                    &
          &                         conf      = conf,                          &
          &                         sub_table = sub_table,                     &
          &                         iError    = iError )
        call aot_table_close( L = conf, thandle = sub_table )
      end do
    end if

    call aot_table_close( L=conf, thandle=shape_table )

  end subroutine tem_load_shapes
!******************************************************************************!


!******************************************************************************!
  !> Read in an arbitrary shape from a lua file defined in a single table
  !!
  !! read a shape like for example inside a tracking table
  !! \verbatim
  !! tracking = { variable = { 'velocity' },\n
  !!              shape = { kind = 'canoND',\n
  !!                        object = { origin = { 1.0, 1.0, 1.0 },\n
  !!                                   vec = { 2.0, 2.0, 2.0 },\n
  !!                                   segments = { 10, 20, 30 } }\n
  !!                      }\n
  !!            } -- tracking table\n
  !! \endverbatim
  !!
  subroutine tem_load_shape_single( me, conf, key, parent, sub_table, iError )
    !---------------------------------------------------------------------------
    !> shape type defined in a lua file
    type(tem_shape_type), intent(out) :: me
    !> lua state
    type(flu_state) :: conf
    !> optional key to load from
    character(len=*), optional, intent(in) :: key
    !> optional parent handle
    integer, optional, intent(in) :: parent
    !> shape table handle
    integer, optional, intent(in) :: sub_table
    !> error flag
    integer, optional, intent(out) :: iError
    !---------------------------------------------------------------------------
    character(len=PathLen) :: buffer
    integer :: iError_loc
    integer :: shape_table
    character(len=32) :: localKey
    !---------------------------------------------------------------------------

    if( present( iError )) iError = 0
    if( present( key )) then
      localKey = key
    else
      localKey = 'shape'
    endif

    ! open the table
    ! shape = {}
    if( present( sub_table )) then
      shape_table = sub_table
    else
      call aot_table_open( L       = conf,                                     &
        &                  thandle = shape_table,                              &
        &                  parent  = parent,                                   &
        &                  key     = trim(localKey) )
    endif

    if (shape_table .gt. 0 ) then

      ! Get the entry 'kind'
      ! shape = {{kind='test'}}
      call aot_get_val( L       = conf,                                        &
        &               thandle = shape_table,                                 &
        &               val     = buffer,                                      &
        &               ErrCode = iError_loc,                                  &
        &               key     = 'kind',                                      &
        &               default = 'none' )

      me%kind = trim(buffer)
      call tem_write(' Loading shape kind: '//trim(me%kind))
      ! Read in the type and configuration of tracking shape
      ! (point, line, plane, ...)
      select case( trim(buffer) )
      case('global', 'all', 'globalmesh')
        me%shapeID = tem_global_shape

      case('canonicalND', 'canoND')
        me%shapeID = tem_canonicalND_shape
        call tem_load_canonicalND( me      = me%canoND,                        &
          &                        conf    = conf,                             &
          &                        thandle = shape_table )

      case('property')
        me%shapeID = tem_property_shape
        call tem_shape_load_propLabel( me, conf, shape_table )

      ! If find unknown tracking type, code will stop!
      case default
        call tem_write( '  Unknown Tracking type specified! '//trim(buffer) )
        call tem_write( '  Choose: ')
        call tem_write( '    shape = { kind = "canoND", ')
        call tem_write( '              object = {origin = {-0.5,0.,0.}, ')
        call tem_write( '              vec = {1.0, 0.0, 0.0},' )
        call tem_write( '              segments = 1000 } } or' )
        call tem_write( '    shape = { kind = "all" } or' )
        call tem_write( '    shape = { kind = "property",' )
        call tem_write( '              property = {"boundary"}}')
        call tem_write( '  Stopping ...' )
        call tem_abort()
      end select

      if( .not. present( parent ))                                             &
        &       call aot_table_close( L = conf, thandle = shape_table )

      write(logUnit(3),*) 'shape ID: '//trim( tem_toStr( me%shapeID ))

    else
      if( present( iError )) iError = ibset( iError, aoterr_NonExistent )
    endif

  end subroutine tem_load_shape_single
!******************************************************************************!


!******************************************************************************!
  !> Write a array of shapes to lua file
  !!
  subroutine tem_shape_out_vec( me, conf )
    !---------------------------------------------------------------------------
    !> shape types to write out
    type( tem_shape_type ), intent(in) :: me(:)
    !> Aotus type handling the output to the file in lua format
    type(aot_out_type), intent(inout) :: conf
    !---------------------------------------------------------------------------
    integer :: iVal
    !---------------------------------------------------------------------------

    ! create a table with name shape if not exist
    if( conf%level .eq. 0 ) then
      call aot_out_open_table( put_conf = conf, tname = 'shape' )
    else
      call aot_out_open_table( put_conf = conf )
    endif

    do iVal = 1, size(me)
      call tem_shape_out_scal( me(iVal), conf )
    end do

    call aot_out_close_table( put_conf = conf )

  end subroutine tem_shape_out_vec
!******************************************************************************!


!******************************************************************************!
  !> Write a shape to lua file
  !!
  subroutine tem_shape_out_scal( me, conf )
    !---------------------------------------------------------------------------
    !> shape types to write out
    type( tem_shape_type ), intent(in) :: me
    !> Aotus type handling the output to the file in lua format
    type(aot_out_type), intent(inout) :: conf
    !---------------------------------------------------------------------------

    ! create a table with name shape if not exist
    if( conf%level .eq. 0 ) then
      call aot_out_open_table( put_conf = conf, tname = 'shape' )
    else
      call aot_out_open_table( put_conf = conf )
    end if

    ! choose output shape from shapeID
    select case( me%shapeID )
      case(tem_canonicalND_shape)
        call tem_canonicalND_out(me%canoND, conf)

    ! \todo: SZ: add further select cases for spheres, tetrahedrical objects
    end select

    call aot_out_close_table( put_conf = conf )

  end subroutine tem_shape_out_scal
!******************************************************************************!


!******************************************************************************!
  !> Return the next coordinate of the point shape
  !!
  !! It actually returns just the point coordinate. However, this function is
  !! used for other shapes as well, where more than one point have to be
  !! returned.
  !!
  function tem_getNextCoordOfShapePoint( me, iElem, iShape ) result(coord)
    !---------------------------------------------------------------------------
    !> current element within total amount of elements to process
    integer, intent(in) :: iElem
    !> current shape to treat
    integer, intent(in) :: iShape
    !> shape type
    type(tem_shape_type ) :: me
    !> calulated real-world coordinate, which is returned
    real(kind=rk)       :: coord(3) 
    !---------------------------------------------------------------------------
    select case( me%shapeID )
      case( tem_canonicalND_shape )
          coord = tem_getNextCoordOfcanonicalND( me%canoND( iShape ), iElem )

      ! \todo: SZ: add further select cases for spheres, tetrahedrical objects
      case default
        call tem_write( ' shape not found, maybe wrong shapeID set, or' //     &
          &             'not set at all?' )
        call tem_abort
    end select

  end function tem_getNextCoordOfSHapePoint
!******************************************************************************!


!******************************************************************************!
  !> Loading property labels from the config file,
  !! set the property bits accordingly
  !!
  subroutine tem_shape_load_propLabel(me, conf, thandle )
    !---------------------------------------------------------------------------
    !> shape type
    type(tem_shape_type) :: me
    !> lua config handle
    type(flu_state) :: conf
    !> table handle from which to read
    integer, intent(in) :: thandle
    !---------------------------------------------------------------------------
    ! lua handles
    integer :: propLabel_handle, nPropLabels
    character(len=labelLen) :: labelBuff
    integer :: iLabel, iErr
    !---------------------------------------------------------------------------

    call aot_table_open( L       = conf,                                       &
      &                  parent  = thandle,                                    &
      &                  thandle = propLabel_handle,                           &
      &                  key     = 'property' )
    ! get the number of property labels
    nPropLabels = aot_table_length( L = conf, thandle = propLabel_handle )

    do iLabel = 1, nPropLabels

      ! Now read in property labels
      call aot_get_val( L       = conf,                                        &
        &               thandle = propLabel_handle,                            &
        &               val     = labelBuff,                                   &
        &               ErrCode = iErr,                                        &
        &               pos     = iLabel )
      call tem_write( '  name of property label: '//labelBuff )
      ! set propBits according to property labels
      if( iErr == 0 ) then
        select case( trim(labelBuff) )
        case( 'boundary' )
          me%propBits = ibset( me%propBits, prp_hasBnd )
        case( 'solidified' )
          me%propBits = ibset( me%propBits, prp_fluidify )
        case default
          call tem_write( '  Unknown property label in a shape specified! '//  &
            &             trim(labelBuff) )
          call tem_write( '  Ignored it. ')
        endselect
      endif
    enddo

    call aot_table_close( L = conf, thandle = propLabel_handle )

  end subroutine tem_shape_load_propLabel
!******************************************************************************!


end module tem_shape_module
!******************************************************************************!
