! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> \brief datatypes for source-terms
!!
!! \author Simon Zimny 
!!
!! This module defines the datatypes to read source-terms from the configuration
!! file. The source terms are solver specific, a possible definition looks like
!! this:
!!  \verbatim
!! 
!!  -- lua function that defines the parameters of the source
!!  function testInjection2(x, y, z, t)
!!    if tmax/2.-1. < t and  t < tmax then
!!      return t+10.
!!    else
!!      return 0.
!!    end
!!  end
!!
!!  source =
!!  {
!!    -- define one source
!!    {
!!      name = 'equalInjection',  -- type of the source
!!      ncomponents = 1,          -- number of scalar values that "define" the source
!!      equalInjection =          -- the name of the source
!!      {
!!        -- where does the source act
!!        shape = {kind = 'canoND', object= {origin = {2.0,1.0,1.0}} },
!!        -- specify the space-time function, that returns the necessary parameters (vector with ncomponents entries)
!!        kind = 'lua_fun',
!!        fun = testInjection2
!!      }
!!    },
!!    -- define another source (this time without labels)
!!    {
!!      'injection',
!!      1,
!!      {
!!        shape = {kind = 'canoND', object= {origin = {0.0,1.0,1.0}} },
!!        kind = 'lua_fun',
!!        fun = testInjection2
!!      }
!!    }
!! \endverbatim
!! 
module tem_source_module

  ! include treelm modules
  use tem_aux_module,        only: tem_abort
  use treelmesh_module,      only: treelmesh_type
  use tem_tools_module,      only: tem_write, tem_horizontalSpacer
  use tem_variable_module,   only: tem_variable_type, tem_variable_out,        &
    &                              assignment(=), PositionOfVal,               &
    &                              tem_variable_from
  use tem_var_system_module, only: tem_var_system_type, tem_var_sys_init,      &
    &                              tem_var_sys_debug, tem_var_append,          &
    &                              tem_var_sys_mark
  use tem_subTree_module,    only: tem_create_subTree_of
  use tem_grow_array_module, only: grw_intArray_type, init, append
  use tem_debug_module,      only: main_debug, tem_debug, dbgUnit
  use tem_logging_module,    only: tem_log, tem_toStr, logUnit
  use tem_shape_module,      only: tem_global_shape

  ! include aotus modules
  use aotus_module,     only: flu_State
  use aot_path_module,  only: aot_path_type, assignment(=)
  use aot_out_module,   only: aot_out_type, aot_out_open, aot_out_close

  implicit none
  private

  public :: tem_source_type
  public :: tem_load_source
  public :: tem_sourceMap_type
  public :: tem_reduce_sourceVars
  public :: tem_reduce_sourceVars_fromType
  public :: assignment(=)

  !> Datatype including the requested variable system for source terms (rhs)
  !! and its functions.
  !!
  type tem_source_type
    !> possible variable system
    type( tem_var_system_type ) :: poss_varSys
    !> requested variable system
    type( tem_var_system_type ) :: requ_varSys
    !> resulting variable system (poss intersected with requ)
    type( tem_var_system_type ) :: varSys
    !> array of source variables (incl. spacetime functions and trees)
    type( tem_variable_type ), allocatable :: sourceVar(:)
    !> number of source variables
    integer :: nSourceVars
    !> array of maps (from the resulting var sys to the array of source vars)
    !! if a variable is defined multiple times in the sourceVar array the map
    !! points points to both entries
    type(grw_intArray_type), allocatable :: map(:)
  end type tem_source_type

  !> datatype containing
  !! - pointer to a source type
  !! - map to the position of a variable in the variable
  !! system in the source type
  type tem_sourceMap_type
    !> pointer to a source type
    type(tem_source_type), pointer, dimension(:) :: source => NULL()
    !> map to the position of a variable in the variable system
    integer, allocatable :: map(:)
  end type tem_sourceMap_type

  interface assignment(=)
    module procedure copySource
  end interface assignment(=)

  interface tem_reduce_sourceVars
    module procedure tem_reduce_sourceVars_single
    module procedure tem_reduce_sourceVars_vector
  end interface tem_reduce_sourceVars

  interface tem_reduce_sourceVars_fromType
    module procedure tem_reduce_sourceVar_fromType_single
    module procedure tem_reduce_sourceVar_fromType_vector
  end interface tem_reduce_sourceVars_fromType

contains


!******************************************************************************!
  !> Copy 'right' source type to 'left' source type.
  !!
  subroutine copySource(left, right)
    !---------------------------------------------------------------------------
    !> source type to copy from
    type(tem_source_type), intent(out) :: left
    !> source type to copy to
    type(tem_source_type), intent(in) :: right
    !---------------------------------------------------------------------------

    left%poss_varSys = right%poss_varSys
    left%requ_varSys = right%requ_varSys
    left%varSys = right%varSys
    if(allocated(right%sourceVar)) then
      allocate(left%sourceVar(size(right%sourceVar)))
      left%sourceVar = right%sourceVar
    endif
    left%nSourceVars = right%nSourceVars
    if(allocated(right%map)) then
      allocate(left%map(size(right%map)))
      left%map = right%map
    endif
  end subroutine copySource
!******************************************************************************!


!******************************************************************************!
  !> Load fill source type from the config fila in lua format.
  !!
  subroutine tem_load_source(me, conf, parent, key, path )
    !---------------------------------------------------------------------------
    !> source information
    type(tem_source_type), intent(inout) :: me
    !> lua config file
    type(flu_state) :: conf
    !> optional parent handle
    integer, optional, intent(in) :: parent
    !> optional key for the tbale
    character(len=*), optional, intent(in) :: key
    !> Path through the Lua script
    type(aot_path_type), intent(in) :: path
    !---------------------------------------------------------------------------
    character(len=32) :: localKey
    integer :: iVar, iSourceVar, ifun
    type(aot_path_type) :: source_path ! lua path to the source table
    integer :: iError
    ! debug variables
    type(aot_out_type) :: out_conf !< Handle for the file
    ! number of variables in the source table
    integer :: nTotal
    !---------------------------------------------------------------------------
    call tem_horizontalSpacer()
    write(logUnit(1),*)'Loading the sources ...' 

    if( present( key )) then
      localKey = key
    else
      localKey = 'source'
    endif

    ! set the local source_path to the given one qand initialize nTotal
    source_path = path
    nTotal = 0
    ! load the source variables
    if( present(parent) ) then
      call tem_variable_from( me      = me%sourceVar,                          &
        &                     conf    = conf,                                  &
        &                     parent  = parent,                                &
        &                     nTotal  = nTotal,                                &
        &                     iError  = iError,                                &
        &                     key     = localkey,                              &
        &                     path    = source_path )
    else
      call tem_variable_from( me      = me%sourceVar,                          &
        &                     conf    = conf,                                  &
        &                     nTotal  = nTotal,                                &
        &                     iError  = iError,                                &
        &                     key     = localkey,                              &
        &                     path    = source_path )
    end if
    if(ierror .eq. -1)then
      me%nSourceVars = 0
    else
      me%nSourceVars = size(me%sourceVar)
    end if

    do iSourceVar = 1, me%nSourceVars
      do iFun = 1, size(me%SourceVar(iSourceVar)%st_fun)
        ! check if the spacetime function has been read
        if(trim(me%SourceVar(iSourceVar)%st_fun(iFun)%fun_kind).eq.'none')then
          call tem_write( 'The spacetime function for ' //                     &
            &             trim(me%SourceVar(iSourceVar)%label) //              &
            &             ' has not been loaded properly.' )
          call tem_abort()
        end if
      end do
    end do

    ! build a requested variable system from the list of source terms
    call tem_build_requSourceVarSys( me )

    ! write out debug output
    ! \todo: DEB: Maybe use a wrapper for tem_variable_out the same style as
    !             used for tem_var_sys_dump -> tem_var_sys_debug
    if( main_debug%dumpSource )then
      call aot_out_open( put_conf = out_conf, &
        &                outUnit = main_debug%logger%funit(0) )
      write(dbgUnit(1),*) 'The source variables are: ' 
      write(dbgUnit(1),*) 'nSourceVars: '//trim(tem_toStr(me%nSourceVars))
      do iVar = 1, me%nSourceVars
        call tem_variable_out( me = me%sourceVar(iVar), conf = out_conf )
      end do
      call aot_out_close( out_conf )
    end if

    call tem_horizontalSpacer()

  end subroutine tem_load_source
!******************************************************************************!


!******************************************************************************!
  !> Build a variable system from the list of sources.
  !!
  subroutine tem_build_requSourceVarSys( me )
    !---------------------------------------------------------------------------
    !> source entities
    type( tem_source_type ),intent(inout) :: me
    !---------------------------------------------------------------------------
    ! counter variable
    integer :: iVar
    integer :: pos
    !---------------------------------------------------------------------------

    ! initialise the requested variable system
    call tem_var_sys_init( me%requ_varSys )

    ! build the requested varSys from the array of sources
    do iVar = 1, me%nSourceVars
      call tem_var_append( sys = me%requ_varSys,                               &
        &                  var = me%sourceVar(iVar),                           &
        &                  pos = pos )
    end do

    call tem_var_sys_mark( sys = me%requ_varSys, SystemName = 'requSources')

    if( main_debug%dumpSource ) then
      call tem_var_sys_debug( sys   = me%requ_varSys,                          &
        &                     level = 1 )
    endif

  end subroutine tem_build_requSourceVarSys
!******************************************************************************!


!******************************************************************************!
  !> Reduce the list source variables
  !!
  subroutine tem_reduce_sourceVar_fromType_vector( me )
    !---------------------------------------------------------------------------
    !> source entities
    type( tem_source_type ),intent(inout) :: me(:)
    !---------------------------------------------------------------------------
    ! temporary source entities
    type( tem_variable_type ), allocatable :: tmp_sourceVar(:)
    ! temporary position
    integer :: tmp_pos
    ! array of active positions
    integer, allocatable :: tmp_activePos(:,:)
    ! counter variables
    integer :: iVar, iVal, iSourceType
    integer :: nSources
    !---------------------------------------------------------------------------
    call tem_write( 'Reducing the source variables ...' )

    allocate( tmp_activePos( size(me), maxval( me(:)%nSourceVars )))
    tmp_activePos(:,:) = 0

    write(logUnit(6),*) 'REDUCE SOURCE VAR nSOURCETYPES: '//                 &
      &                   trim( tem_toStr( size( me )))

    ! first reduce by the varSys -> all sources without valid transformations
    ! are neglected
    do iSourceType = 1, size( me )
      nSources = 0
      ! allocate and set the temporary array
      allocate( tmp_sourceVar( size( me(iSourceType)%sourceVar )))
      do iVar = 1, size(me(iSourceType)%sourceVar)
        ! get the position of the source variable in me(iSourceType)%varSys
        tmp_pos = PositionOfVal( me = me( iSourceType )%varSys%variable,       &
          &                      val = me( iSourceType )%sourceVar( iVar ))

        if( tmp_pos .gt. 0 )then
           nSources = nSources + 1
           tmp_sourceVar( nSources ) = me( iSourceType )%sourceVar( iVar )
           tmp_activePos( iSourceType, nSources ) = tmp_pos
        end if
      end do

      ! all sources which are not part of the possible variable system
      ! are neglected
      if( nSources .lt. size( me(iSourceType)%sourceVar ))then
        deallocate( me( iSourceType )%sourceVar )
        allocate( me( iSourceType )%sourceVar( nSources ))
        me( iSourceType )%sourceVar = tmp_sourceVar( 1:nSources )
        me( iSourceType )%nSourceVars = nSources
      end if
      me( iSourceType )%nSourceVars = nSources

      ! deallocate the temporary array
      deallocate( tmp_sourceVar )
    end do

    ! set the map from source%varSys to the sourceVar array
    do iSourceType = 1, size( me )
      ! allocate the map
      if ( allocated( me( iSourceType )%map ))                                 &
        &  deallocate( me( iSourceType )%map )
      allocate( me( iSourceType )%map( me(iSourceType)%varSys%variable%nVals ))

      ! initialize each growing array
      do iVar = 1, me( iSourceType )%varSys%variable%nVals
        call init( me( iSourceType )%map( iVar ))
      end do

      do iVar = 1, size( me( iSourceType )%sourceVar )
        call append( me  = me( iSourceType )%                                  &
          &                           map( tmp_activePos(iSourceType, iVar )), &
          &          val = iVar )
      end do
    end do

    deallocate( tmp_activePos )

    if( main_debug%dumpSource )then
      write(dbgUnit(1),*) '------------------------------------------'//   &
        &                     '-------' 
      write(dbgUnit(1),*) 'MAP: ' 
      write(dbgUnit(1),*) '' 
      write(dbgUnit(1),*)  '' 
      do iSourceType = 1, size(me)
        do iVar = 1, me( iSourceType )%varSys%variable%nVals
          do iVal = 1, me( iSourceType )%map(iVar)%nVals
            write(dbgUnit(1),*) trim( tem_toStr(                           &
              &                        me( iSourceType )%map(iVar)%Val(iVal)))
          end do
        end do
        write(dbgUnit(1),*) ''
      end do
      write(dbgUnit(1),*) '' 
      write(dbgUnit(1),*) '' 
      write(dbgUnit(1),*) '------------------------------------------'//   &
        &                     '-------' 
    end if

  end subroutine tem_reduce_sourceVar_fromType_vector
!******************************************************************************!


!******************************************************************************!
  !> Reduce the source var list by the type and the shape.
  !!
  subroutine tem_reduce_sourceVars_vector( me, global_tree )
    !---------------------------------------------------------------------------
    !> source entities
    type( tem_source_type ),intent(inout) :: me(:)
    !> the global tree
    type(treelmesh_type), intent(in) :: global_tree
    !---------------------------------------------------------------------------
    ! temporary source entities
    type( tem_variable_type ), allocatable :: tmp_sourceVar(:)
    ! active sources
    logical, allocatable :: sourceVarActive(:)
    ! number of active sources
    integer :: nActiveSourceVars
    ! counters
    integer :: iVar, iFun, iSource
    integer :: local_iVar
    integer :: maxsourceSize
    logical :: useGlobalMesh
    !---------------------------------------------------------------------------
    call tem_write('Building the trees for the sources ...')

    ! reduce the sourceVar list by the variable system of provided source terms
    call tem_reduce_sourceVar_fromType_vector( me = me )

    maxsourceSize = 0
    do iSource=1,size(me)
      maxsourceSize = max(size(me(iSource)%sourceVar), maxsourceSize)
    end do

    ! allocate the array describing if a source variable is used
    allocate( sourceVarActive(maxsourceSize))

    ! loop over all source types
    do iSource = 1, size( me )
      ! size of me(iSource)%sourceVar might have been decreased by the previous
      ! reduce subroutine

      ! initialize the array of active source variables to not active
      sourceVarActive = .false.
      ! initialize the number of active sources to 0
      nActiveSourceVars = 0
      ! loop over all source variables in the source(iSource)
      do iVar = 1, size( me(iSource)%sourceVar )

        ! build the trees for this
        !-----------------------------------------------------------------------
        ! identify source elements
        !-----------------------------------------------------------------------
        !\todo: SZ: the prefix is defined identically for all sourceVars with
        !           the same name in case we want to output this tree we might
        !           have to change this
        useGlobalMesh = .false.
        do iFun = 1, size( me(iSource)%sourceVar(iVar)%st_fun )

          !\todo: SZ: the tree in st_fun type has to be a subTree
          call tem_create_subTree_of(                                          &
            &      inTree  = global_tree,                                      &
            &      subTree = me(iSource)%sourceVar(iVar)%st_fun(iFun)%subTree, &
            &      inShape = me(iSource)%sourceVar(iVar)%st_fun(iFun)%geom,    &
            &      prefix  = trim( me(iSource)%sourceVar(iVar)%label )//'_' )

          if( any(me(iSource)%sourceVar(iVar)%st_fun(iFun)%geom(:)%shapeID     &
            & .eq. tem_global_shape ))                                         &
            & useGlobalMesh = .true.
        end do

        ! now check if the number of elements is gt 0, this means that
        ! elements are given
        if ( me(iSource)%sourceVar(iVar)%st_fun(iFun)%subTree%nElems .gt. 0    &
          & .or. useGlobalMesh )  then
          nActiveSourceVars = nActiveSourceVars + 1
          sourceVarActive( iVar ) = .true.
        end if

      end do

      ! now reduce the sourceVar array for this type

      ! allocate the temporary array of source variables with nActiveSourcVars
      allocate( tmp_sourceVar( nActiveSourceVars ))

      local_iVar = 0
      ! loop over all source variables and check if the logical is true
      do iVar = 1, size( me(iSource)%sourceVar )
        if( sourceVarActive( iVar ))then
          local_iVar = local_iVar + 1
          ! copy the source variable to the temporary list
          tmp_sourceVar( local_iVar ) = me(iSource)%sourceVar( iVar )
        end if
      end do

      ! reallocate the original source var array with its new size
      deallocate( me(iSource)%sourceVar )
      allocate( me(iSource)%sourceVar( nActiveSourceVars ))

      ! now copy the information from the temporary list of source vars back
      me(iSource)%sourceVar = tmp_sourceVar


      ! deallocate the array of temporary arrays
      deallocate( tmp_sourceVar )
    end do

    deallocate( sourceVarActive )

    ! hack, use this function to update me%map
    call tem_reduce_sourceVar_fromType_vector( me = me )

  end subroutine tem_reduce_sourceVars_vector
!******************************************************************************!


!******************************************************************************!
  !> Reduce the list source variables
  !!
  subroutine tem_reduce_sourceVar_fromType_single( me )
    !---------------------------------------------------------------------------
    !> source entities
    type( tem_source_type ),intent(inout) :: me
    !---------------------------------------------------------------------------
    ! temporary source entities
    type( tem_variable_type ), allocatable :: tmp_sourceVar(:)
    ! temporary position
    integer :: tmp_pos
    ! array of active positions
    integer, allocatable :: tmp_activePos(:)
    ! counter variables
    integer :: iVar, iVal
    integer :: nSources
    !---------------------------------------------------------------------------
    call tem_write( 'Reducing the source variables ...' )

    allocate( tmp_activePos(me%nSourceVars) )
    tmp_activePos(:) = 0

    ! first reduce by the varSys -> all sources without valid transformations
    ! are neglected
    nSources = 0
    ! allocate and set the temporary array
    allocate( tmp_sourceVar( size( me%sourceVar )))
    do iVar = 1, size(me%sourceVar)
      ! get the position of the source variable in me%varSys
      tmp_pos = PositionOfVal( me  = me%varSys%variable,                       &
        &                      val = me%sourceVar( iVar ))

      if( tmp_pos .gt. 0 )then
         nSources = nSources + 1
         tmp_sourceVar( nSources ) = me%sourceVar( iVar )
         tmp_activePos( nSources ) = tmp_pos
      end if
    end do

    ! all sources which are not part of the possible variable system
    ! are neglected
    if( nSources .lt. size( me%sourceVar ))then
      deallocate( me%sourceVar )
      allocate( me%sourceVar( nSources ))
      me%sourceVar = tmp_sourceVar( 1:nSources )
    end if
    me%nSourceVars = nSources

    ! deallocate the temporary array
    deallocate( tmp_sourceVar )

    ! set the map from source%varSys to the sourceVar array
    ! allocate the map
    if ( allocated( me%map ))                                                  &
      &  deallocate( me%map )
    allocate( me%map( me%varSys%variable%nVals ))

    ! initialize each growing array
    do iVar = 1, me%varSys%variable%nVals
      call init( me%map( iVar ))
    end do

    do iVar = 1, size( me%sourceVar )
      call append( me  = me%map( tmp_activePos( iVar )),                       &
        &          val = iVar )
    end do

    deallocate( tmp_activePos )

    if( main_debug%dumpSource )then
      write(dbgUnit(1),*) '------------------------------------------'//   &
        &                     '-------' 
      
      write(dbgUnit(1),*) 'MAP: '
      write(dbgUnit(1),*) '' 
      write(dbgUnit(1),*)'' 
      do iVar = 1, me%varSys%variable%nVals
        do iVal = 1, me%map(iVar)%nVals
          write(dbgUnit(1),*) trim( tem_toStr( me%map(iVar)%Val(iVal)))
          end do
        end do
      write(dbgUnit(1),*) '' 
      write(dbgUnit(1),*) '' 
      write(dbgUnit(1),*) '' 
      write(dbgUnit(1),*) '------------------------------------------'//   &
        &                     '-------' 
    end if

  end subroutine tem_reduce_sourceVar_fromType_single
!******************************************************************************!


!******************************************************************************!
  !> Reduce the source var list by the type and the shape.
  !!
  subroutine tem_reduce_sourceVars_single( me, global_tree )
    !---------------------------------------------------------------------------
    !> source entities
    type( tem_source_type ),intent(inout) :: me
    !> the global tree
    type(treelmesh_type), intent(in) :: global_tree
    !---------------------------------------------------------------------------
    ! temporary source entities
    type( tem_variable_type ), allocatable :: tmp_sourceVar(:)
    ! active sources
    logical, allocatable :: sourceVarActive(:)
    ! number of active sources
    integer :: nActiveSourceVars
    ! counters
    integer :: iVar, iFun
    integer :: local_iVar
    logical :: useGlobalMesh
    !---------------------------------------------------------------------------
    call tem_write('Building the trees for the sources ...')

    ! reduce the sourceVar list by the variable system of provided source terms
    call tem_reduce_sourceVar_fromType_single( me = me )

    ! allocate the array describing if a source variable is used
    allocate( sourceVarActive( size(me%sourceVar) ) )

    ! size of me%sourceVar might have been decreased by the previous
    ! reduce subroutine

    ! initialize the array of active source variables to not active
    sourceVarActive = .false.
    ! initialize the number of active sources to 0
    nActiveSourceVars = 0
    ! loop over all source variables in the source
    do iVar = 1, size( me%sourceVar )

      ! build the trees for this
      !-----------------------------------------------------------------------
      ! identify source elements
      !-----------------------------------------------------------------------
      !\todo: SZ: the prefix is defined identically for all sourceVars with
      !           the same name in case we want to output this tree we might
      !           have to change this
      useGlobalMesh = .false.
      do iFun = 1, size( me%sourceVar(iVar)%st_fun )
        !\todo: SZ: the tree in st_fun type has to be a subTree
        call tem_create_subTree_of(                                            &
          &      inTree  = global_tree,                                        &
          &      subTree = me%sourceVar(iVar)%st_fun(iFun)%subTree,            &
          &      inShape = me%sourceVar(iVar)%st_fun(iFun)%geom,               &
          &      prefix  = trim( me%sourceVar(iVar)%label )//'_' )
        if( any(me%sourceVar(iVar)%st_fun(iFun)%geom(:)%shapeID                &
          & .eq. tem_global_shape ) ) &
          & useGlobalMesh = .true.
      end do

      ! now check if the number of elements is gt 0, this means that
      ! elements are given
      if ( any(me%sourceVar(iVar)%st_fun(:)%subTree%nElems .gt. 0)             &
         & .or. useGlobalMesh) then
        nActiveSourceVars = nActiveSourceVars + 1
        sourceVarActive( iVar ) = .true.
      end if

    end do

    ! now reduce the sourceVar array for this type

    ! allocate the temporary array of source variables with nActiveSourcVars
    allocate( tmp_sourceVar( nActiveSourceVars ))

    local_iVar = 0
    ! loop over all source variables and check if the logical is true
    do iVar = 1, size( me%sourceVar )
      if( sourceVarActive( iVar ))then
        local_iVar = local_iVar + 1
        ! copy the source variable to the temporary list
        tmp_sourceVar( local_iVar ) = me%sourceVar( iVar )
      end if
    end do

    ! reallocate the original source var array with its new size
    deallocate( me%sourceVar )
    allocate( me%sourceVar( nActiveSourceVars ))

    ! now copy the information from the temporary list of source vars back
    me%sourceVar = tmp_sourceVar


    ! deallocate the array of temporary arrays
    deallocate( tmp_sourceVar )

    deallocate( sourceVarActive )

    ! hack, use this function to update me%map
    call tem_reduce_sourceVar_fromType_single( me = me )

  end subroutine tem_reduce_sourceVars_single
!******************************************************************************!


end module tem_source_module
!******************************************************************************!
