! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> A module to deal with generic space-time functions.
!!
!! It allows the definition of arbitrary space-time
!! dependent functions, and the multiplication of
!! a spatial function with a transient function as
!! a often required special case.
!!
!! Space-Time functions might be predefined as Fortran functions, simple
!! constants or Lua functions.
!! They might return scalars or one-dimensional arrays.
!! If only a single value is expected from a Lua function, the Lua function
!! is supposed to return a scalar value. Irregardless if the function is
!! invoked by an array valued routine or a plain scalar one.
!! Otherwise, the Lua function has to return a table with the correct number
!! of entries identified by position.
!!
!! Note, that this makes the interface in the Lua script the same:
!! - whenever a single return value is expected, a plain scalar should be
!!   returned.
!! - otherwise a table should be returned
module tem_spacetime_fun_module

  ! include treelm modules
  use env_module,              only: LabelLen, rk, long_k
  use tem_subTree_type_module, only: tem_subTree_type
  use treelmesh_module,        only: treelmesh_type
  use tem_aux_module,          only: tem_abort
  use tem_geometry_module,     only: tem_baryOfId
  use tem_spatial_module,      only: tem_spatial_type, load_spatial,           &
    &                                tem_spatial_for
  use tem_tools_module,        only: tem_write
  use tem_transient_module,    only: tem_transient_type, load_transient,       &
    &                                tem_transient_for
  use tem_shape_module,        only: tem_shape_type, tem_load_shape
  use tem_time_module,         only: tem_time_type

  use tem_precice_module,      only: tem_spacetime_precice_for 
  use tem_miescatter_module,   only: tem_miescatter_field_type,                &
                                   & tem_load_miescatter_displacementfieldz,   &
                                   & tem_load_miescatter_magneticfieldx,       &
                                   & tem_load_miescatter_magneticfieldy,       &
                                   & tem_eval_miescatter_magnx,                &
                                   & tem_eval_miescatter_magny,                &
                                   & tem_eval_miescatter_displz

  ! include aotus modules
  use aotus_module,     only: aot_get_val, aot_top_get_val,                    &
    &                         aoterr_Fatal, aoterr_WrongType,                  &
    &                         aoterr_NonExistent, flu_State
  use aot_fun_module,   only: aot_fun_type, aot_fun_open, aot_fun_close,       &
    &                         aot_fun_put, aot_fun_do
  use aot_path_module,  only: aot_path_type, aot_path_addNode,                 &
    &                         aot_path_open, aot_path_close,                   &
    &                         assignment(=)
  use aot_table_module, only: aot_table_open, aot_table_close,                 &
    &                         aot_table_length

  implicit none

  private

  public :: tem_spacetime_fun_type
  public :: tem_load_spacetime
  public :: tem_spacetime_for

  type tem_spacetime_fun_type
    !> The function kind
    !!
    !! Should be either:
    !! \li 'none': Not defined at all
    !! \li 'const': Constant for all (x,y,z,t)
    !! \li 'combined': This returns spatial(x,y,z)*transient(t)
    !! \li 'lua_fun': Function defined in the Lua script
    !! \li Add predefined functions here
    character(len=labelLen) :: fun_kind

    !> spatial restrictions
    type(tem_shape_type), allocatable :: geom(:)

    !> subTree build from the shapes
    type(tem_subTree_type) :: subTree

    !> constant value for (x, y, z, t)
    real(kind=rk), allocatable :: const(:)

    !> Description of composite spatial fun
    type(tem_spatial_type) :: spatial

    !> Composite transient fun
    type(tem_transient_type) :: transient

    !> Path to space-time Lua function
    type(aot_path_type) :: LuaPath

    !> Space-time function for Mie-series solution
    !! of electrodynamic scattering at dielectric sphere.
    type(tem_miescatter_field_type) :: mie_fun

  end type

  interface tem_spacetime_lua_for
    module procedure tem_spacetime_lua_for_treeIds
    module procedure tem_spacetime_lua_vector_for_treeIds
    module procedure tem_spacetime_lua_for_coord
    module procedure tem_spacetime_lua_vector_for_coord
  end interface tem_spacetime_lua_for

  interface tem_spacetime_for
    module procedure tem_spacetime_for_treeIds
    module procedure tem_spacetime_vector_for_treeIds
    module procedure tem_spacetime_for_coord
    module procedure tem_spacetime_vector_for_coord
    module procedure tem_spacetime_for_stcoord
  end interface tem_spacetime_for

  interface tem_load_spacetime
    module procedure tem_load_spacetime_single
    module procedure tem_load_spacetime_table
  end interface tem_load_spacetime

contains


!******************************************************************************!
  !> This routine loads the spacetime from the given key or position
  !!
  !! \todo Return ERROR Code
  subroutine tem_load_spacetime_table( me, conf, path, parent, key, pos, nComp )
    !---------------------------------------------------------------------------
    !> spacetime fun information
    type(tem_spacetime_fun_type), intent(inout), allocatable :: me(:)
    !> lua state handle
    type(flu_State) :: conf 
    !> path through the Lua script
    type(aot_path_type), intent(in) :: path 
    !> aotus parent handle
    integer, intent(inout), optional :: parent
    !> name of the variable which is defined as spacetime function
    character(len=*), intent(in), optional :: key
    !> position of spacetime fun in a table
    integer, intent(in), optional :: pos 
    !> position of spacetime fun in a table
    integer, intent(in), optional :: nComp
    !---------------------------------------------------------------------------
    ! aotus handle
    integer :: thandle, thandle2
    ! counter variables
    integer :: nSt, iSt
    ! path through the Lua script
    type(aot_path_type) :: st_path
    !---------------------------------------------------------------------------

    st_path = path
    ! test if the spacetime is stored in multiple tables

    ! at first try to open two tables
    ! if yes -> close the later one get the length and read the st functions
    !           one by one
    ! if no  -> close both tables and read it directly as a single value
    call aot_table_open( L       = conf,                                       &
      &                  thandle = thandle,                                    &
      &                  parent  = parent,                                     &
      &                  key     = trim(key),                                  &
      &                  pos     = pos )
    call aot_table_open( L       = conf,                                       &
      &                  thandle = thandle2,                                   &
      &                  parent  = thandle,                                    &
      &                  pos     = 1 )

    if (thandle2 /= 0) then
      ! Multiple space time functions defined as a table of tables.
      call aot_table_close( L = conf, thandle = thandle2 )
      call aot_path_addNode( me       = st_path,                               &
        &                    NodeType = 'table',                               &
        &                    key      = key,                                   &
        &                    pos      = pos )
      nSt = aot_table_length( L=conf, thandle=thandle )
      allocate( me( nSt ))

      ! loop over all spacetime functions and read them one by one
      do iSt = 1, nSt
        call tem_load_spacetime_single( me     = me(iSt),                      &
          &                             conf   = conf,                         &
          &                             path   = st_path,                      &
          &                             parent = thandle,                      &
          &                             nComp  = nComp,                        &
          &                             pos    = iSt )
      end do
      call aot_table_close( L = conf, thandle = thandle )
    else
      ! Just a single space time function, go on reading it directly.
      allocate(me(1))
      call aot_table_close( L = conf, thandle = thandle2 )
      call aot_table_close( L = conf, thandle = thandle )

      ! read in a single spacetime function
      call tem_load_spacetime_single( me     = me(1),                          &
        &                             conf   = conf,                           &
        &                             path   = st_path,                        &
        &                             parent = parent,                         &
        &                             key    = key,                            &
        &                             nComp  = nComp,                          &
        &                             pos    = pos )

    end if

  end subroutine tem_load_spacetime_table
!******************************************************************************!


!******************************************************************************!
  !> This routine loads the spacetime variable from the given key
  !!
  !!\todo: update description no more planes -> canoND
  !!Example for source variable
  !!\verbatim
  !!injection = { shape = { object = 'plane', plane = {origin = {0.0,5.0,5.0} ,
  !!                              vecA = {0.0,10.0,0.0},  vecB = {0.0,0.0,10.0},
  !!                              segments1 = 100, segments2 = 100 }},
  !!              kind = 'lua_fun', fun =  chargeDensitySpaceTime
  !!              -- here scalar function
  !! --           kind = 'combined', spatial = {}, transient ={}
  !!--            kind = 'const', const = 1.0
  !!                            }
  !!\endverbatim
  !!!\verbatim
  !!density = {   kind = 'lua_fun', fun =  chargeDensitySpaceTime
  !!              -- here scalar function
  !! --           kind = 'combined', spatial = {}, transient ={}
  !!--            kind = 'const', const = 1.0
  !!                            }
  !!or
  !! density = 1.0
  !! or
  !! density = chargeDensitySpaceTime
  !!\endverbatim
  !!
  !!
  !\todo: SZ: This function is for now only defined for given shape and time
  !!          objects to use it for the bc we might need to change it
  !!
  subroutine tem_load_spacetime_single( me, conf, path, parent, key, pos,      &
    &                                   nComp )
    !---------------------------------------------------------------------------
    !> spacetime fun information
    type(tem_spacetime_fun_type), intent(out) :: me
    !> lua state type
    type(flu_State) :: conf
    !> path through the lua script
    type(aot_path_type), intent(in) :: path
    !> aotus parent handle
    integer, intent(in), optional :: parent
    !> name of the variable which is defined as spacetime function
    character(len=*), intent(in), optional :: key
    !> position of spacetime fun in a table
    integer, intent(in), optional :: pos
    !> position of spacetime fun in a table
    integer, intent(in), optional :: nComp
    !---------------------------------------------------------------------------
    ! path through the Lua script
    type(aot_fun_type) :: fun
    ! aotus handle
    integer :: thandle
    ! error variables
    integer :: iError, iError_shape
    integer, allocatable :: vError(:)
    ! do we require a shape definition in the current context?
    logical :: requireShape
    ! local ncomp
    integer :: loc_nComp
    !---------------------------------------------------------------------------

    if( present( nComp ))then
      if( nComp .gt. 0)then
        loc_nComp = nComp
      else
        loc_nComp = 1
      end if
    else
      loc_nComp = 1
    end if

    requireShape = .false.

    ! First test for a lua function
    if (present(parent)) then
      call aot_fun_open( L      = conf,                                        &
        &                parent = parent,                                      &
        &                fun    = fun,                                         &
        &                key    = key,                                         &
        &                pos    = pos )
    else
      call aot_fun_open( L = conf, fun = fun, key = key )
    end if

    ! copy the current path
    me%LuaPath = path

    if_fun: if( fun%handle /= 0 ) then
      call tem_write( 'Space time for '//trim(key)                             &
        &             //' is defined as a lua function')
      ! There is a function defined in Lua.
      me%fun_kind = 'lua_fun'
      ! Append a final function node to the path
      call aot_path_addNode( me       = me%LuaPath,                            &
        &                    NodeType = 'function',                            &
        &                    key      = key,                                   &
        &                    pos      = pos )
      call aot_fun_close( L=conf, fun=fun )

    else

      call aot_fun_close( L=conf, fun=fun )
      ! Not a function, try to interpret it as a table:
      call aot_table_open( L       = conf,                                     &
        &                  thandle = thandle,                                  &
        &                  parent  = parent,                                   &
        &                  key     = key,                                      &
        &                  pos     = pos )

      if_tab: if (thandle /= 0) then
        ! Succeeded in reading the given Lua object as a table.
        call aot_path_addNode( me       = me%LuaPath,                          &
          &                    NodeType = 'table',                             &
          &                    key      = key,                                 &
          &                    pos      = pos )

        ! read the shape
        call tem_load_shape( me      = me%geom,                                &
          &                  conf    = conf,                                   &
          &                  parent  = thandle,                                &
          &                  iError  = iError_shape )

        call aot_get_val( L       = conf,                                      &
          &               thandle = thandle,                                   &
          &               key     = 'kind',                                    &
          &               val     = me%fun_kind,                               &
          &               default = 'unknown',                                 &
          &               ErrCode = iError )
        if( iError .ne. 0 ) then
          call tem_write('Error in retrieving the function kind.')
          call tem_abort
        end if
        call tem_write(' Space time for '//trim(key)//' is defined as '        &
          &            //trim(me%fun_kind))

        select case(trim(adjustl(me%fun_kind)))
        case('combined')
          ! A typical case of a space-time function that can be represented
          ! a multiplicative combination of a transient and a spatial part.
          call load_spatial( me     = me%spatial,                              &
            &                conf   = conf,                                    &
            &                parent = thandle,                                 &
            &                path   = me%LuaPath )
          call load_transient( me     = me%transient,                          &
            &                  conf   = conf,                                  &
            &                  parent = thandle,                               &
            &                  path   = me%LuaPath )

        case('lua_fun')
          ! The spacetime function is defined in the Lua script
          call aot_fun_open( L      = conf,                                    &
            &                parent = thandle,                                 &
            &                fun    = fun,                                     &
            &                key    = 'fun' )
          ! Append a final function node to the path
          call aot_path_addNode( me       = me%LuaPath,                        &
            &                    NodeType = 'function',                        &
            &                    key      = 'fun' )
          call aot_fun_close( L=conf, fun=fun )
          !\todo HK: what is this requireShape actually good for? Only the
          !!         lua_fun seems to activate it, why not the other ones?
          !!         If it can be omitted for the other ones, why does it
          !!         have to be defined for the lua_fun?
          !!         Should not all cases be treated similarly? It is weird to
          !!         require the user to remember the subtle differences of
          !!         the various cases...
          !!         But if all are treated the same way, there is no need for
          !!         the logical, and if only a single case is to be treated
          !!         differently, this can easily be tested aswell, there is
          !!         again no need for this requireShape variable at all!
          !> \todo JZ: I just replaced requireShape by false. Currently, I don't
          !! see a reason why lua functions can only work with shapes. However, 
          !! we should check carefully if this is really true.
          !requireShape = .true.
          requireShape = .false.
          call tem_write( 'Space time for '//trim(key)                         &
            &             //' is defined as a lua function')

        case('const')
          ! The spacetime function is just a constant, get its value.
          call aot_get_val( L         = conf,                                  &
            &               thandle   = thandle,                               &
            &               val       = me%const,                              &
            &               maxLength = loc_nComp,                             &
            &               key       = 'const',                               &
            &               ErrCode   = vError )
          if (size(me%const) == 0) then
            ! The constant is probably not an array, try to interpret it
            ! directly as a scalar value:
            allocate(me%const(1), vError(1))
            call aot_get_val( L       = conf,                                  &
              &               thandle = thandle,                               &
              &               val     = me%const(1),                           &
              &               key     = 'const',                               &
              &               ErrCode = vError(1) )
            if (btest(vError(1), aoterr_Fatal)) then
              call tem_write('FATAL Error occured in definition of a space'//  &
                &            '-time function')
              call tem_write('while retrieving a scalar space time constant!')
              call tem_write('STOPPING')
              call tem_abort()
            end if
          end if
          call tem_write( 'variable  is a constant' )

          ! catching possible errors
          if (any(btest(vError, aoterr_WrongType))) then
            call tem_write('FATAL Error occured in definition of a space'//    &
              &            '-time function')
            call tem_write('while retrieving space time constant:')
            call tem_write('Variable has wrong type (should be a real number)!')
            call tem_write('STOPPING')
            call tem_abort()
          end if
          deallocate(vError)

        case('miescatter_displacementfieldz')
          call tem_load_miescatter_displacementfieldz(conf = conf, &
                                                     & thandle = thandle , &
                                                     & me = me%mie_fun)
        case('miescatter_magneticfieldx')
          call tem_load_miescatter_magneticfieldx(conf = conf, &
                                                     & thandle = thandle , &
                                                     & me = me%mie_fun)
        case('miescatter_magneticfieldy')
          call tem_load_miescatter_magneticfieldy(conf = conf, &
                                                     & thandle = thandle , &
                                                     & me = me%mie_fun)
        case default
          ! If you introduce new predefined functions, add their loading 
          ! routine in a seperate case branch here.
          call tem_write( 'ERROR in definition of a space'//                   &
            &             '-time function:' )
          call tem_write( 'Selected an unknown space-time function' )
          call tem_write( trim( me%fun_kind ))
          !>\todo As the constant might be a vector now, we actually would
          !!      need to try to interpret the table as a vector in this default
          !!      branch.
        end select

        if( btest( iError_shape, aoterr_nonExistent ) .and. requireShape ) then
          call tem_write( 'Fatal Error: Shape is not defined in the spatial'// &
            &             ' definition, but required.' )
          call tem_abort()
        endif
        call aot_table_close( L=conf, thandle=thandle )

      else
        ! Not provided as a table, try to interpret it as a constant factor.
        allocate(me%const(1))
        if (present(parent)) then
          call aot_get_val( L       = conf,                                    &
            &               thandle = parent,                                  &
            &               val     = me%const(1),                             &
            &               key     = key,                                     &
            &               pos     = pos,                                     &
            &               ErrCode = iError )
        else
          call aot_get_val( L       = conf,                                    &
            &               key     = key,                                     &
            &               val     = me%const(1),                             &
            &               ErrCode = iError )
        end if

        ! catching possible errors
        if (btest(iError, aoterr_WrongType)) then
          call tem_write('FATAL Error occured in definition of a space'//      &
            &            '-time function')
          call tem_write('while retrieving space time constant:')
          call tem_write('Variable has wrong type (should be a real number)!')
          call tem_write('STOPPING')
        end if

        if (btest(iError, aoterr_NonExistent)) then
          me%fun_kind = 'none'
        else
          call tem_write(' Space time for '//trim(key)//' is a constant =',    &
            &              me%const)
          me%fun_kind = 'const'
        end if

      end if if_tab

    end if if_fun

  end subroutine tem_load_spacetime_single
!******************************************************************************!


!******************************************************************************!
  !> This function compute space time function for given list of treeIDs
  !! \todo pass subtree with treeIDs lists instead of treeIds and tree
  !!
  function tem_spacetime_for_treeIDs( me, treeIDs, time, tree, n, conf )       &
    &                       result(res)
    !---------------------------------------------------------------------------
    !> Spacetime function to evaluate
    type(tem_spacetime_fun_type) :: me
    !> Global treelmesh to look for positions in
    type(treelmesh_type), intent(in) ::tree
    !> Number of values to return
    integer, intent(in) :: n
    !> TreeIDs where to evaluate the function
    integer(kind=long_k), intent( in ) :: treeIDs(n)
    !> timer object incl. the current time information
    type(tem_time_type), intent( in )  :: time
    type(flu_State), intent(in), optional :: conf !< lua state
    !> return value of the function
    real(kind=rk) :: res(n)
    !---------------------------------------------------------------------------

    select case(trim(adjustl(me%fun_kind)))
    case('none')
      res = 0.0_rk
    case('const')
      res = me%const(1)
    case('lua_fun')
      res = tem_spacetime_lua_for( me%LuaPath, treeIDs, time, tree, n, conf)
    case('combined')
      res = tem_spatial_for( me%spatial, treeIDs, tree, n, conf )              &
        & * tem_transient_for( me%transient, time, conf )
    case default
      call tem_write('ERROR: Unknown spatial function in tem_spacetime_vector_for_treeIDs.')
      call tem_write('Aborting...')
      call tem_abort()
    end select
  end function tem_spacetime_for_treeIDs
!******************************************************************************!


!******************************************************************************!
  !> This function compute space time function that give bach a table of results
  !! for given list of treeIDs
  !! \todo pass subtree with treeIDs lists instead of treeIds and tree
  !!
  function tem_spacetime_vector_for_treeIDs( me, treeIDs, time, tree, n, ncomp,&
    &                                        conf) result(res)
    !---------------------------------------------------------------------------
    !> Spacetime function to evaluate
    type(tem_spacetime_fun_type) :: me
    !> Global treelmesh to look for positions in
    type(treelmesh_type), intent(in) ::tree
    !> Number of tables to return
    integer, intent(in) :: n
    !> Number of values in a Table
    integer, intent(in) :: ncomp
    !> TreeIDs where to evaluate the function
    integer(kind=long_k), intent( in ) :: treeIDs(n)
    !> timer object incl. the current time information
    type(tem_time_type), intent( in )  :: time
    type(flu_State), intent(in), optional :: conf !< lua state
    !> return value of the function
    real(kind=rk) :: res(n,ncomp)
    !---------------------------------------------------------------------------
    ! counter
    integer :: i
    !---------------------------------------------------------------------------

    select case(trim(adjustl(me%fun_kind)))
    case('none')
      res = 0.0_rk
    case('const')
      do i = 1, nComp
        res(:,i) = me%const(i)
      end do
    case('lua_fun')
      res = tem_spacetime_lua_for( me%LuaPath, treeIDs, time,                  &
        &                          tree, n, ncomp, conf )
    case('combined')
        write(*,*) 'NOT SUPPORTED YET'
        stop
    !!\todo: DH:implement ncomp for spatial
    !  res = tem_spatial_for(me%spatial, treeIDs, tree, n, ncomp, conf) &
    !    &  * tem_transient_for( me%transient, time, conf )
    case default
      call tem_write('ERROR: Unknown spatial function in tem_spacetime_vector_for_treeIDs.')
      call tem_write('Aborting...')
      call tem_abort()
    end select
  end function tem_spacetime_vector_for_treeIDs
!******************************************************************************!


!******************************************************************************!
  !> This function computes the space time function for given list of space-time
  !! coordinates.
  !!
  function tem_spacetime_for_stcoord( me, stCoord, n, conf ) result(res)
    !---------------------------------------------------------------------------
    !> Spacetime function to evaluate
    type(tem_spacetime_fun_type) :: me
    !> Number of values to return
    integer, intent(in) :: n
    !> barycentric Ids of an elements.
    !! 1st index goes over number of elements and
    !! 2nd index goes over x,y,z,t coordinates
    real(kind=rk), intent( in ) :: stCoord(n,4)
    type(flu_State), intent(in), optional :: conf !< lua state
    !> return value of the function
    real(kind=rk) :: res(n)
    !---------------------------------------------------------------------------
    real(kind=rk) :: spaceCoord(1,3)
    integer :: iPoint
    type(tem_time_type) :: time
    !---------------------------------------------------------------------------

    !> \todo JZ: i added the subroutine for the space-time formulation.
    !! Below you see, that we create a tem_time_type variable in each loop.
    !! The code can be faster if we implement this routine in a nicer way.

    do iPoint = 1, n
      time%rTime = stcoord(iPoint, 4)
      res(iPoint:iPoint)                                                       &
        & = tem_spacetime_for_coord(me, spaceCoord, time, 1, conf)
    end do

  end function tem_spacetime_for_stcoord
!******************************************************************************!


!******************************************************************************!
  !> This function computes the space time function for a given list of
  !! coordinates
  !!
  function tem_spacetime_for_coord(me, coord, time, n, conf) result(res)
    !---------------------------------------------------------------------------
    !> Spacetime function to evaluate
    type(tem_spacetime_fun_type) :: me
    !> Number of values to return
    integer, intent(in) :: n
    !> barycentric Ids of an elements.
    !! 1st index goes over number of elements and
    !! 2nd index goes over x,y,z coordinates
    real(kind=rk), intent( in ) :: coord(n,3)
    !> timer object incl. the current time information
    type(tem_time_type), intent( in )  :: time
    type(flu_State), intent(in), optional :: conf !< lua state
    !> return value of the function
    real(kind=rk) :: res(n)
    !---------------------------------------------------------------------------

    select case(trim(adjustl(me%fun_kind)))
    case('none')
      res = 0.0_rk
    case('const')
      res = me%const(1)
    case('lua_fun')
      res = tem_spacetime_lua_for(me%LuaPath, coord, time, n, conf)
!    case('precice')
!      res = tem_spacetime_precice_for( posID, dataName)
    case('combined')
      res = tem_spatial_for( me%spatial, coord, n, conf ) &
        & * tem_transient_for( me%transient, time, conf )
    case('miescatter_displacementfieldz')
      res = tem_eval_miescatter_displz( me = me%mie_fun,   &
                                      & coord = coord,     &
                                      & time = time%rTime, &
                                      & n = n)
    case('miescatter_magneticfieldx')
      res = tem_eval_miescatter_magnx( me = me%mie_fun,   &
                                      & coord = coord,     &
                                      & time = time%rTime, &
                                      & n = n)
    case('miescatter_magneticfieldy')
      res = tem_eval_miescatter_magny( me = me%mie_fun,   &
                                      & coord = coord,     &
                                      & time = time%rTime, &
                                      & n = n)
    case default
      call tem_write('ERROR: Unknown spatial function in tem_spacetime_for_coord.')
      call tem_write('Aborting...')
      call tem_abort()
    end select

  end function tem_spacetime_for_coord
!******************************************************************************!


!******************************************************************************!
  !> This function computes the space time function that gives back an array
  !! for a given list of coordinates
  !!
  function tem_spacetime_vector_for_coord( me, coord, time, n, ncomp, conf )   &
    &                             result(res)
    !---------------------------------------------------------------------------
    !> Spacetime function to evaluate
    type(tem_spacetime_fun_type) :: me
    !> Number of arrays to return
    integer, intent(in) :: n
    !> Number of entrys in each array
    integer, intent(in) :: ncomp
    !> barycentric Ids of an elements.
    !! 1st index goes over number of elements and
    !! 2nd index goes over x,y,z coordinates
    real(kind=rk), intent( in ) :: coord(n,3)
    !> timer object incl. the current time information
    type(tem_time_type), intent( in )  :: time
    type(flu_State), intent(in), optional :: conf !< lua state
    !> return value of the function
    real(kind=rk) :: res(n,ncomp)
    !---------------------------------------------------------------------------
    ! counter
    integer :: i
    !---------------------------------------------------------------------------

    select case(trim(adjustl(me%fun_kind)))
    case('none')
      res = 0.0_rk
    case('const')
      do i = 1, nComp
        res(:,i) = me%const(i)
      end do
    case('lua_fun')
      res = tem_spacetime_lua_for( me%LuaPath, coord, time, n, ncomp, conf )
    case('precice')
      !res = tem_spacetime_precice_for(me%LuaPath, coord, time, n, ncomp, conf)
    case('combined')
        write(*,*) 'NOT IMPLEMENTED YET'
        stop
    !! \todo: DH: implement ncomp for spatial
    !  res = tem_spatial_for(me%spatial, coord, n, ncomp, conf) &
    !    &  * tem_transient_for( me%transient, time, conf )
    case default
      call tem_write('ERROR: Unknown spatial function in tem_spacetime_vector_for_coord.')
      call tem_write('Aborting...')
      call tem_abort()
    end select

  end function tem_spacetime_vector_for_coord
!******************************************************************************!


!******************************************************************************!
  !> \brief This function invokes the Lua function for barycentric coordinates
  !! of an element specified by treeIds
  !!
  function tem_spacetime_lua_for_treeIds( me, treeIds, time, tree, n, conf )   &
    &                           result(res)
    !---------------------------------------------------------------------------
    !> aotus path type in tem spatial type
    type(aot_path_type) :: me
    !> global treelm mesh
    type(treelmesh_type), intent(in) ::tree
    !> number of return values
    integer, intent(in) :: n
    !> treeIds of elements in given level
    integer(kind=long_k), intent( in ) :: treeIds(n)
    !> timer object incl. the current time information
    type(tem_time_type), intent( in )  :: time
    !> return value
    real(kind=rk) :: res(n)
    !> lua state
    type(flu_State), intent(in), optional :: conf
    !---------------------------------------------------------------------------
    type(aot_fun_type) :: fun
    integer :: iError
    integer :: i, j
    real(kind=rk) :: coord(3)
    type(flu_State) :: loc_conf
    logical :: openLua
    !---------------------------------------------------------------------------

    if (present(conf)) then
      loc_conf = conf
      openLua = .false.
    else
      openLua = .true.
    end if

    call aot_path_open(me, conf=loc_conf, fun=fun, openLua=openLua)

    do i=1,n
      coord = tem_BaryOfId( tree, treeIds(i) )
      do j=1,3
        call aot_fun_put(L=loc_conf, fun=fun, arg=coord(j))
      end do
      call aot_fun_put(L=loc_conf, fun=fun, arg=time%rTime)
      call aot_fun_do(L=loc_conf, fun=fun, nresults=1)
      call aot_top_get_val(L=loc_conf, val=res(i), ErrCode=iError)
    end do

    call aot_path_close(me, conf=loc_conf, fun=fun, closeLua=openLua)

  end function tem_spacetime_lua_for_treeIds
!******************************************************************************!


!******************************************************************************!
  !> This function invokes the Lua function for barycentric coordinates
  !! of an element specified by treeIds and returns an array with the given
  !! number of components.
  !!
  !! Note, that the returned object by the Lua function has to be a table,
  !! except if there is only one component. For arrays of length 1 the Lua
  !! return value has to be a simple scalar, not a table!
  function tem_spacetime_lua_vector_for_treeIds( me, treeIds, time, tree, n,   &
    &                                            ncomp, conf ) result(res)
    !---------------------------------------------------------------------------
    !> aotus path type in tem spatial type
    type(aot_path_type) :: me
    !> global treelm mesh
    type(treelmesh_type), intent(in) ::tree
    !> number of return values
    integer, intent(in) :: n
    !> number of components per returned value
    integer, intent(in) :: ncomp
    !> treeIds of elements in given level
    integer(kind=long_k), intent( in ) :: treeIds(n)
    !> timer object incl. the current time information
    type(tem_time_type), intent( in )  :: time
    !> return value
    real(kind=rk) :: res(n,ncomp)
    !> lua state
    type(flu_State), intent(in), optional :: conf
    !---------------------------------------------------------------------------
    type(aot_fun_type) :: fun
    integer :: iError(ncomp)
    integer :: i, j
    real(kind=rk) :: coord(3)
    type(flu_State) :: loc_conf
    logical :: openLua
    !---------------------------------------------------------------------------

    if (present(conf)) then
      loc_conf = conf
      openLua = .false.
    else
      openLua = .true.
    end if

    call aot_path_open(me, conf=loc_conf, fun=fun, openLua=openLua)

    do i=1,n
      coord = tem_BaryOfId( tree, treeIds(i) )
      do j=1,3
        call aot_fun_put(L=loc_conf, fun=fun, arg=coord(j))
      end do
      call aot_fun_put(L=loc_conf, fun=fun, arg=time%rTime)
      call aot_fun_do(L=loc_conf, fun=fun, nresults=1)
      if (nComp == 1) then
        call aot_top_get_val(L=loc_conf, val=res(i,1), ErrCode=iError(1))
      else
        call aot_top_get_val(L=loc_conf, val=res(i,:), ErrCode=iError)
      end if
    end do

    call aot_path_close(me, conf=loc_conf, fun=fun, closeLua=openLua)

  end function tem_spacetime_lua_vector_for_treeIds
!******************************************************************************!


!******************************************************************************!
  !> \brief This function invokes the Lua function for a given coordinate.
  !!
  function tem_spacetime_lua_for_coord( me, coord, time, n, conf ) result(res)
    !---------------------------------------------------------------------------
    !> aotus path type in tem spatial type
    type(aot_path_type) :: me
    !> number of return values
    integer, intent(in) :: n
    !> barycentric Ids of an elements.
    !! 1st index goes over number of elements and
    !! 2nd index goes over x,y,z coordinates
    real(kind=rk), intent( in ) :: coord(n,3)
    !> timer object incl. the current time information
    type(tem_time_type), intent( in )  :: time
    !> return value
    real(kind=rk) :: res(n)
    !> lua state
    type(flu_State), intent(in), optional :: conf
    !---------------------------------------------------------------------------
    type(aot_fun_type) :: fun
    integer :: iError
    integer :: i, j
    type(flu_State) :: loc_conf
    logical :: openLua
    !---------------------------------------------------------------------------

    if (present(conf)) then
      loc_conf = conf
      openLua = .false.
    else
      openLua = .true.
    end if

    call aot_path_open(me, conf=loc_conf, fun=fun, openLua=openLua)

    do i=1,n
      do j=1,3
        call aot_fun_put(L=loc_conf, fun=fun, arg=coord(i,j))
      end do
      call aot_fun_put(L=loc_conf, fun=fun, arg=time%rTime)
      call aot_fun_do(L=loc_conf, fun=fun, nresults=1)
      call aot_top_get_val(L=loc_conf, val=res(i), ErrCode=iError)
    end do

    call aot_path_close(me, conf=loc_conf, fun=fun, closeLua=openLua)

  end function tem_spacetime_lua_for_coord
!******************************************************************************!


!******************************************************************************!
  !> This function invokes the Lua function for a given coordinate and returns
  !! an array valued result.
  !!
  !! Note, that the returned object by the Lua function has to be a table,
  !! except if there is only one component. For arrays of length 1 the Lua
  !! return value has to be a simple scalar, not a table!
  function tem_spacetime_lua_vector_for_coord( me, coord, time, n, ncomp, conf)&
    &                                          result(res)
    !------------------------------------------------------------------------
    !> aotus path type in tem spatial type
    type(aot_path_type) :: me
    !> number of return values
    integer, intent(in) :: n
    integer, intent(in) :: ncomp !< number of components returned for each value
    !> barycentric Ids of an elements.
    !! 1st index goes over number of elements and
    !! 2nd index goes over x,y,z coordinates
    real(kind=rk), intent( in ) :: coord(n,3)
    !> timer object incl. the current time information
    type(tem_time_type), intent( in )  :: time
    !> return value
    real(kind=rk) :: res(n,ncomp)
    !> lua state
    type(flu_State), intent(in), optional :: conf
    !---------------------------------------------------------------------------
    type(aot_fun_type) :: fun
    integer :: iError(ncomp)
    integer :: i, j
    type(flu_State) :: loc_conf
    logical :: openLua
    !---------------------------------------------------------------------------

    if (present(conf)) then
      loc_conf = conf
      openLua = .false.
    else
      openLua = .true.
    end if

    call aot_path_open(me, conf=loc_conf, fun=fun, openLua=openLua)

    do i=1,n
      do j=1,3
        call aot_fun_put(L=loc_conf, fun=fun, arg=coord(i,j))
      end do
      call aot_fun_put(L=loc_conf, fun=fun, arg=time%rTime)
      call aot_fun_do(L=loc_conf, fun=fun, nresults=1)
      if (nComp == 1) then
        call aot_top_get_val(L=loc_conf, val=res(i,1), ErrCode=iError(1))
      else
        call aot_top_get_val(L=loc_conf, val=res(i,:), ErrCode=iError)
      end if
    end do

    call aot_path_close(me, conf=loc_conf, fun=fun, closeLua=openLua)

  end function tem_spacetime_lua_vector_for_coord
!******************************************************************************!


end module tem_spacetime_fun_module
!******************************************************************************!
