! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!>  This module provides the description of space dependent functions.
!!
!! These can include: constant, Lua and predefined Fortran functions.
!! Currently supports:
!! - const
!! - lua_fun
!! - parabol2d
!! - \ref parabol3d
!! Remember boundaries can be defined with 
!! \ref tem_spacetime_fun_module::tem_load_spacetime_single "spacetime"
!! combined conditions ( in space and time) the functions defined here, can be
!! accessed in the boundary declarations with the help of the combined kind.
!! \author Kannan Masilamani
!!
module tem_spatial_module

  ! include treelm modules
  use env_module,             only: rk, long_k, LabelLen, globalMaxLevels
  use tem_aux_module,         only: tem_abort
  use tem_tools_module,       only: tem_write, upper_to_lower
  use treelmesh_module,       only: treelmesh_type
  use tem_geometry_module,    only: tem_BaryOfId
  use tem_canonicalND_module, only: tem_canonicalND_type
  use tem_shape_module,       only: tem_shape_type, tem_load_shape
  use tem_ic_predefs_module,  only: ic_gausspulse_type, load_ic_gausspulse,    &
    &                               ic_gausspulse_for,                         &
    &                               ic_2dcrvp_type, load_ic_2dcrvp,            &
    &                               ic_2dcrvpX_for, ic_2dcrvpY_for,            &
    &                               ic_2dcrvpPressure_for
  use tem_logging_module,     only: tem_log, tem_toStr, logUnit
  use tem_miescatter_module,  only: tem_miescatter_field_type,                 &
                                  & tem_load_miescatter_displacementfieldz,    &
                                  & tem_load_miescatter_magneticfieldx,        &
                                  & tem_load_miescatter_magneticfieldy,        &
                                  & tem_eval_miescatter_displz,                &
                                  & tem_eval_miescatter_magnx,                 &
                                  & tem_eval_miescatter_magny 

  ! include aotus modules
  use aotus_module,     only: flu_State, aoterr_Fatal, aoterr_NonExistent,     &
    &                         aoterr_WrongType, aot_top_get_val, aot_get_val
  use aot_table_module, only: aot_table_open, aot_table_close
  use aot_fun_module,   only: aot_fun_type, aot_fun_open, aot_fun_close,       &
    &                         aot_fun_put, aot_fun_do
  use aot_path_module,  only: aot_path_type, assignment(=), aot_path_addNode,  &
    &                         aot_path_open, aot_path_close

  implicit none

  private

  public :: tem_spatial_type
  public :: load_spatial
  public :: tem_spatial_lua_for
  public :: tem_spatial_parabol3d_for
  public :: tem_spatial_parabol2d_for
  public :: tem_spatial_for

  !> stores values of spatial term during initialization to reduce
  !! computations during main loop.
  type spatial_value_type
    !> array for the values
    real(kind=rk), allocatable :: val(:)
  end type spatial_value_type

  !> defines parabola functions
  !! shape kind defines 2d or 3d parabola
  type spatial_parabol_type
    !> define point, line or plane spatial profile
    type( tem_shape_type ) :: geometry
    !> magnitude of the spatial value
    real(kind=rk) :: amplitude
  end type spatial_parabol_type

  !> Defines a random spatial distribution within a given interval.
  type spatial_random_type
    !> Minimal value to produce
    real(kind=rk) :: val_min
    !> Maximal value to produce
    real(kind=rk) :: val_max
    !> Length of the value interval
    real(kind=rk) :: val_range
  end type spatial_random_type

  !> Defines a stationary solution of the Euler equation by the Hopf Fibration.
  type spatial_hopf_type
    !> Minimal value to produce
    real(kind=rk) :: radius
    !> Maximal value to produce
    real(kind=rk) :: p0
  end type spatial_hopf_type

  !> contains spatial state information
  type tem_spatial_type
    !> kind: how is the spatial defined?
    !! - 'none' = no spatial modifier defined
    !! - 'const' = a constant factor
    !! - 'lua_fun' = defined as a lua function
    !! - 'parabol' = parabolic function
    !!               shape = {object = line} defines 2d parabola
    !!               shape = {object = plane} defines 3d parabola
    !! Further predefined functions might be added here
    character(len=LabelLen) :: kind

    !> constant spatial value
    !!\todo HK: Need to declare this const as allocatable array, to allow
    !!          vectorial definitions.
    real(kind=rk) :: const

    !> lua path type to use lua functions
    type( aot_path_type ) :: LuaPath

    !> defines gausspulse
    type( ic_gausspulse_type ) :: gausspulse

    !> 2d co-rotating vortex pair
    type( ic_2dcrvp_type ) :: crvp

    !> Parabol type
    type( spatial_parabol_type ) :: parabol

    !> Random type
    type( spatial_random_type ) :: random

    !> Hopf Fibration type
    type( spatial_hopf_type ) :: hopf

    !> Spatial function for Mie-series solution
    !! of electrodynamic scattering at dielectric sphere.
    type(tem_miescatter_field_type) :: mie_fun

    !> store spatial value on each level
    type( spatial_value_type ) :: stored( globalMaxLevels )

    !> to store spatial values during initialization
    logical :: toStore = .false.
  end type tem_spatial_type

  interface tem_spatial_lua_for
    module procedure tem_spatial_lua_for_treeIDs
    module procedure tem_spatial_lua_for_coord
    module procedure tem_spatial_lua_vector_for_treeIDs
    module procedure tem_spatial_lua_vector_for_coord
  end interface tem_spatial_lua_for

  interface tem_spatial_parabol2d_for
    module procedure tem_spatial_parabol2d_for_treeIDs
    module procedure tem_spatial_parabol2d_for_coord
  end interface tem_spatial_parabol2d_for

  interface tem_spatial_parabol3d_for
    module procedure tem_spatial_parabol3d_for_treeIDs
    module procedure tem_spatial_parabol3d_for_coord
  end interface tem_spatial_parabol3d_for

  interface tem_spatial_for
    module procedure tem_spatial_for_treeIDs
    module procedure tem_spatial_for_coord
    module procedure tem_spatial_vector_for_treeIDs
    module procedure tem_spatial_vector_for_coord
  end interface tem_spatial_for

contains


!******************************************************************************!
  !>  This subroutine load spatial boundary state variable.
  !!
  !! If spatial is defined as block than read block for predefined Fortran
  !! function variables else it is defined as constant.
  !! Valid definitions:
  !! - Constant
  !! \verbatim
  !! spatial = 1.0
  !! \endverbatim
  !! - lua_function
  !! \verbatim
  !! spatial = lua_fun_name
  !! \endverbatim
  !! Note. Lua function take 3 input arguments (x,y,z) i.e barycentric
  !! coordinates of an element
  !! - Predefined Fortran function
  !! \verbatim
  !! spatial  = {predefined = "fun_name", fun_parameters}
  !! [example given in subroutine load_spatial_parabola3d definition]
  !! \endverbatim
  !!
  subroutine load_spatial( me, conf, parent, key, path, defaultValue )
    !---------------------------------------------------------------------------
    !> spatial boundary state type
    type(tem_spatial_type), intent(out) :: me
    !> lua state type
    type(flu_State) :: conf
    !> aotus parent handle
    integer, intent(in) :: parent
    !> aotus path type
    type(aot_path_type), intent(in) :: path
    !> state variable key string defined in lua
    character(len=*), intent(in), optional :: key
    !> What should be set s a default value for the quantities if no 
    !! quantity was given in the lua file
    real(kind=rk), intent(in), optional :: defaultValue
    !---------------------------------------------------------------------------
    integer :: thandle
    integer :: iError
    type(aot_fun_type) :: fun
    character(len=labelLen) :: local_key
    real(kind=rk) :: local_default
    !--------------------------------------------------------------------------

    if(present(defaultValue)) then
      local_default = defaultValue
    else
      local_default = 1._rk
    endif

    if(present(key)) then
      local_key = trim(key)
    else
      local_key = 'spatial'
    endif

    me%LuaPath = path
    ! First test for a lua function
    call aot_fun_open(L=conf, parent=parent, fun=fun, key=trim(local_key))

    if (fun%handle /= 0) then
      ! There was a function found for the spatial entry, go on using it
      call tem_write('   Spatial for '//trim(local_key)//' is a lua function')
      me%kind = 'lua_fun'
      call aot_path_addNode( me       = me%LuaPath,                            &
        &                    NodeType = 'function',                            &
        &                    key      = trim( local_key ))
    else
      call aot_table_open( L       = conf,                                     &
        &                  thandle = thandle,                                  &
        &                  parent  = parent,                                   &
        &                  key     = trim( local_key ))
      if (thandle .ne. 0) then
        ! The definition is provided as a table for a predefined function
        call aot_get_val( L       = conf,                                      &
          &               thandle = thandle,                                   &
          &               key     = 'predefined',                              &
          &               val     = me%kind,                                   &
          &               default = 'unknown',                                 &
          &               ErrCode = iError)
        call aot_path_addNode( me       = me%LuaPath,                          &
          &                    NodeType = 'table',                             &
          &                    key      = trim(local_key) )

        call tem_write('   Spatial for "'//trim(local_key)//'" kind is "'     &
          &             //trim(me%kind)//'"')

        select case(upper_to_lower(trim(adjustl(me%kind))))
        case('lua_fun')
          call tem_write( '  is a Lua function')
          call aot_fun_open( L      = conf,                                    &
            &                parent = thandle,                                 &
            &                fun    = fun,                                     &
            &                key    = trim(local_Key) )
          ! Append a final function node to the path
          call aot_path_addNode( me       = me%LuaPath,                        &
            &                    NodeType = 'function',                        &
            &                    key      = trim(local_key))
          call aot_fun_close(L=conf, fun=fun)
        case('parabol')
          call tem_write( '  is a parabolic profile')
          call load_spatial_parabol( me      = me%parabol,                     &
            &                        conf    = conf,                           &
            &                        thandle = thandle )
        case('random')
          call tem_write( '  is a random distribution')
          call load_spatial_random( me      = me%random,                       &
            &                       conf    = conf,                            &
            &                       thandle = thandle )
        case('gausspulse')
          call tem_write( '  is a gaussian pulse')
          call load_ic_gausspulse( conf    = conf,                             &
            &                      thandle = thandle,                          &
            &                      me      = me%gausspulse )
!HK: Not implemented yet.
!HK!        case('hopf')
!HK!          call tem_write( '  is a Hopf fibration solution')
!HK!          call load_spatial_hopf( me = me%hopf, conf = conf, &
!HK!            &                     thandle = thandle )
        case('crvpx')
          call tem_write( '  is a Spinning vortex pair X component')
          call load_ic_2dcrvp( conf    = conf,                                 &
            &                  thandle = thandle,                              &
            &                  me      = me%crvp )
        case('crvpy')
          call tem_write( '  is a Spinning vortex pair Y component')
          call load_ic_2dcrvp( conf   = conf,                                  &
            &                 thandle = thandle,                               &
            &                  me     = me%crvp )
        case('crvppressure')
          call tem_write( '  is a Spinning vortex pair density')
          call load_ic_2dcrvp( conf    = conf,                                 &
            &                  thandle = thandle,                              &
            &                  me=me%crvp )
        case('miescatter_displacementfieldz')
          call tem_load_miescatter_displacementfieldz(conf = conf,             &
                                               & thandle = thandle ,           &
                                               & me = me%mie_fun)
        case('miescatter_magneticfieldx')
          call tem_load_miescatter_magneticfieldx(conf = conf,                 &
                                               & thandle = thandle ,           &
                                               & me = me%mie_fun)
        case('miescatter_magneticfieldy')
          call tem_load_miescatter_magneticfieldy(conf = conf,                 &
                                               & thandle = thandle ,           &
                                               & me = me%mie_fun)
        case default
          call tem_write('ERROR in definition of the spatial'//                &
            &            ' boundary Conditions:')
          call tem_write('Selected an unknown spatial function: '//            &
            &            trim(me%kind))
          call tem_abort()
        end select

        !check whether to store spatial values during initliazation or not
        call aot_get_val( L       = conf,                                      &
          &               thandle = thandle,                                   &
          &               key     = 'store',                                   &
          &               val     = me%toStore,                                &
          &               default = .false.,                                   &
          &               ErrCode = iError )
      else
        ! The spatial is not defined as a table, try to interpret it as
        ! a constant factor.
        call aot_get_val( L       = conf,                                      &
          &               thandle = parent,                                    &
          &               key     = trim(local_key),                           &
          &               val     = me%const,                                  &
          &               default = local_default,                             &
          &               ErrCode = iError )
        if (btest(iError, aoterr_WrongType)) then
          call tem_write('FATAL Error occured in definition of the spatial'//  &
            &            ' boundary conditions')
          call tem_write('while retrieving spatial constant:')
          call tem_write('Variable has wrong type (should be a real number)!')
          call tem_write('STOPPING')
          call tem_abort()
        end if
        if (btest(iError, aoterr_NonExistent)) then
          write(logUnit(1),*) ' WARNING: variable is non-existent.'//        &
            &                   ' Taking default value "none".' 
          me%kind = 'none'
        else
          me%kind = 'const'
        end if
        write(logUnit(1),*) '   Spatial for '//trim(local_key)//             &
          &                   ' is constant='//trim( tem_toStr( me%const ))
      end if
      call aot_table_close(L=conf, thandle=thandle)
    end if
    call aot_fun_close(L=conf, fun=fun)

  end subroutine load_spatial
!******************************************************************************!


!******************************************************************************!
  !>   This subroutine load 3d parabola type variables from LUA file.
  !!
  !! Specify ref_value, if spatial block is defined.
  !!Valid definition:
  !! 1. 2D parabola
  !! \verbatim
  !! spatial = {predefined='parabol', shape = { center/origin={-5.0, 0.0, 0.0},
  !!                                            vec={0.0, 1.0, 0.0}},
  !!                                            amplitude = 0.01}
  !! where, center - line center.
  !!        origin - line origin. (define either center or origin)
  !!        vec    - length of the line.
  !! \endverbatim
  !! Parabolic 2D profile at channel inlet.
  !! \image html parabol2d.png
  !! 2. 3D parabola
  !! \verbatim
  !! spatial = {predefined='parabol', shape = { center/origin={-5.0, 0.0, 0.0},
  !!                                            vec={{0.0, 1.0, 0.0},
  !!                                                 {1.0, 0.0, 0.0}},
  !!                                    amplitude = 0.01}
  !! where, center - plane center.
  !!        origin - plane origin. (define either center or origin)
  !!        vec    - length of the plane in 2 direction.
  !!        amplitude - maximum value
  !! \endverbatim
  !! Parabolic 3D profile at channel inlet.
  !! \image html parabolic3D.png
  !! Example: 
  !! Inlet velocity at inlet plane with velocity along x-dir with a maximum
  !! velocity 0.08.
  !! Inlet plane is normal to x-dir and located at offset of {-5,0,0} from
  !! origin. The plane spans -1 to 1 in both y and z-dir.
  !!
  !! \verbatim
  !! boundary_condition =
  !!         {
  !!           { label = 'inlet',
  !!             kind = 'inlet_ubb',
  !!             velocityX = { kind = 'combined',
  !!                          spatial = {predefined='parabol',
  !!                                     shape = { origin={-5.0, 0.0, 0.0},
  !!                                     vec={{0.0, 1.0, 0.0},
  !!                                          {0.0, 0.0, 1.0} }},
  !!                                     amplitude = 0.01}
  !!                         },
  !!             velocityY = 0.0
  !!             velocityZ = 0.0
  !!           }
  !!         }
  !! \endverbatim
  !!\verbatim
  !!  _______vecB____
  !! |       |      |
  !! |       |      |
  !! |       |______|vecA
  !! |     center   |
  !! |              |
  !! |______________|_
  !!\endverbatim
  !!
  subroutine load_spatial_parabol( me, conf, thandle )
    !---------------------------------------------------------------------------
    !> parabola3d spatial datas
    type(spatial_parabol_type),intent(out) :: me
    !> lua state type
    type(flu_State) :: conf
    !> aotus parent handle
    integer, intent(in) :: thandle
    !---------------------------------------------------------------------------
    integer :: iError
    !---------------------------------------------------------------------------

    !load geometry
    call tem_load_shape( conf    = conf,                                       &
      &                  parent  = thandle,                                    &
      &                  me      = me%geometry )
    !amplitude
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               key     = 'amplitude',                                   &
      &               val     = me%amplitude,                                  &
      &               ErrCode = iError,                                        &
      &               default = 1._rk )
    call tem_write( ' * amplitude =', me%amplitude )

  end subroutine load_spatial_parabol
!******************************************************************************!


!******************************************************************************!
  !> Reading the definition for a random distribution function.
  !!
  !! This spatial function will return random numbers within a given interval.
  !! The interval has to be specified by 'min' and 'max' in the Lua
  !! configuration.
  !! Resulting values will be given by: min + (max-min)*random_number
  !! The range defaults to min=0.0 - max=1.0.
  !!
  subroutine load_spatial_random( me, conf, thandle )
    !---------------------------------------------------------------------------
    !> random spatial specification
    type(spatial_random_type),intent(out) :: me
    !> lua state type
    type(flu_State) :: conf
    !> aotus parent handle
    integer, intent(in) :: thandle
    !---------------------------------------------------------------------------
    integer :: iError
    !---------------------------------------------------------------------------

    ! val_min
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               key     = 'min',                                         &
      &               val     = me%val_min,                                    &
      &               ErrCode = iError,                                        &
      &               default = 0.0_rk )
    if( btest( iError, aoterr_Fatal ))then
      call tem_write('FATAL Error occured in definition of the spatial'//      &
        &            ' function')
      call tem_write('while retrieving the min for random numbers!')
      call tem_write('STOPPING')
      call tem_abort()
    end if
    if( btest( iError, aoterr_NonExistent ))then
      write(logUnit(1),*) ' WARNING: min is non-existent.'//                 &
        &                   ' Using default value 0.0.'
    end if
    call tem_write( ' * val_min =', me%val_min )

    ! val_max
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               key     = 'max',                                         &
      &               val     = me%val_max,                                    &
      &               ErrCode = iError,                                        &
      &               default = 1.0_rk )
    if( btest( iError, aoterr_Fatal ))then
      call tem_write('FATAL Error occured in definition of the spatial'//      &
        &            ' function')
      call tem_write('while retrieving the max for random numbers!')
      call tem_write('STOPPING')
      call tem_abort()
    end if
    if( btest( iError, aoterr_NonExistent ))then
      write(logUnit(1),*) ' WARNING: min is non-existent.'//                 &
        &                   ' Using default value 1.0.'
    end if
    call tem_write( ' * val_max =', me%val_max )

    me%val_range = me%val_max - me%val_min

  end subroutine load_spatial_random
!******************************************************************************!


!******************************************************************************!
  !>  This function computes 3d parabola profile from treeIDs of an element.
  !!
  !! This profile is defined
  !! by element barycentric coordinate and 3d parabola parameters.
  !! 3D parabola profile at given plane is computed in the following way:
  !! - Project barycentric coordinate vector in a given plane via
  !! \f$ \alpha = ((baryCoord - center) \cdot vecA)/\sqrt{ vecA\cdot vecA }\f$
  !! - Compute spatial value using
  !! \f$ res = (1+\alpha)*(1-\alpha)*(1+\beta)*(1-\beta) \f$ 
  !! Example:
  !! Parabolic 3D profile at channel inlet.
  !! \image html parabolic3D.png 
  !! 
  function tem_spatial_parabol3d_for_treeIds( me, treeIds, tree, n ) result(res)
    !---------------------------------------------------------------------------
    !> contains plane parameter for 3d parabola
    type( tem_canonicalND_type ) :: me
    !> global treelm mesh
    type( treelmesh_type ), intent( in ) ::tree
    !> number of return values
    integer, intent(in) :: n
    !> treeIds of elements in given level
    integer(kind=long_k), intent( in ) :: treeIds(n)
    !> return value of a function
    real(kind=rk) :: res(n)
    !---------------------------------------------------------------------------
    real(kind=rk) :: coord(3), alpha, beta, diff(3)
    real(kind=rk) :: vecAsqr, vecBsqr
    real(kind=rk) :: center(3), halfvec(2,3)
    !loop variables
    integer :: iDir, jDir
    !---------------------------------------------------------------------------

    jDir = 0
    do iDir = 1, 3
      if( me%active( iDir )) then
        jDir = jDir + 1
        halfvec( jDir, : ) = me%vec( :, iDir ) / 2._rk
      end if
    end do


    center = me%origin + halfvec( 1, : ) + halfvec( 2, : )

    vecAsqr = dot_product( halfvec( 1, : ), halfvec( 1, : ) )
    vecBsqr = dot_product( halfvec( 2, : ), halfvec( 2, : ) )

    !loop over number of return values
    do iDir = 1, n

      !barycentric coordinate
      coord = tem_BaryOfId( tree, treeIds(iDir) )

      !distance between parabola center and barycentric coordinates
      diff = coord - center
      !projection of diff in a plane on vecA
      alpha =  dot_product( diff, halfvec( 1, : ) )                            &
        &   / vecAsqr

      !projection of diff in a plane on vecB
      beta =  dot_product( diff, halfvec( 2, : ) )                             &
        &  / vecBsqr

      res( iDir )  = ( 1.0_rk-alpha ) * ( 1.0_rk+alpha )                       &
        &          * ( 1.0_rk-beta ) * ( 1.0_rk+beta )

      if ( abs(alpha) .gt. 1.0_rk .or.                                         &
        &  abs(beta) .gt. 1.0_rk ) then
        res( iDir ) = 0.0_rk
      end if
    end do

  end function tem_spatial_parabol3d_for_treeIds
!******************************************************************************!


!******************************************************************************!
  !>  This function computes 3d parabola profile from coord of an element.
  !!
  !! This profile is defined
  !! by element barycentric coordinate and 3d parabola parameters.
  !! 3D parabola profile at given plane is computed in the following way:
  !! - Project barycentric coordinate vector in a given plane via
  !! \f$ \alpha = ((baryCoord - center) \cdot vecA)/\sqrt{ vecA\cdot vecA }\f$
  !! - Compute spatial value using
  !! \f$ res = (1+\alpha)*(1-\alpha)*(1+\beta)*(1-\beta) \f$ 
  !! Example:
  !! Parabolic 3D profile at channel inlet.
  !! \image html parabolic3D.png 
  !! 
  !\todo use projection of point on line and point on plane
  !!
  function tem_spatial_parabol3d_for_coord( me, coord, n ) result(res)
    !---------------------------------------------------------------------------
    !> contains parameter for 3d parabola
    type( tem_canonicalND_type ) :: me
    !> the number of return values
    integer, intent(in) :: n
    !> barycentric Ids of an elements.
    !! 1st index goes over number of elements and
    !! 2nd index goes over x,y,z coordinates
    real(kind=rk), intent( in ) :: coord(n,3)
    !> return value of a function
    real(kind=rk) :: res(n)
    !---------------------------------------------------------------------------
    real(kind=rk) :: alpha, beta, diff(3)
    real(kind=rk) :: vecAsqr, vecBsqr
    real(kind=rk) :: center(3), halfvec(2,3)
    integer :: iDir, jDir
    !---------------------------------------------------------------------------

    jDir = 0
    do iDir = 1, 3
      if( me%active( iDir )) then
        jDir = jDir + 1
        halfvec( jDir, : ) = me%vec( :, iDir ) / 2._rk
      end if
    end do

    center = me%origin + halfvec( 1, : ) + halfvec( 2, : )

    vecAsqr = dot_product( halfvec( 1, : ), halfvec( 1, : ) )
    vecBsqr = dot_product( halfvec( 2, : ), halfvec( 2, : ) )

    !loop over number of return values
    do iDir = 1, n

      !distance between parabola center and barycentric coordinates
      diff = coord(iDir,:) - center
      !projection of diff in a plane on vecA
      alpha =  dot_product( diff, halfvec( 1, : ) )                            &
        &   / vecAsqr

      ! projection of diff in a plane on vecB
      beta =  dot_product( diff, halfvec( 2, : ) )                             &
        &  / vecBsqr

      res(iDir)  = ( 1.0_rk-alpha ) * ( 1.0_rk+alpha )                         &
        &        * ( 1.0_rk-beta ) * ( 1.0_rk+beta )

      if ( abs(alpha) .gt. 1.d0 .or.                                           &
        &  abs(beta) .gt. 1.d0 ) then
        res(iDir) = 0.0_rk
      end if

    end do

  end function tem_spatial_parabol3d_for_coord
!******************************************************************************!


!******************************************************************************!
  !>  This function computes 2d parabola profile from treeIds of elements
  !!
  !! This profile is defined
  !! by element barycentric coordinate and 2d parabola parameters.
  !! 2D parabola profile at given plane is computed in the following way:
  !! - Project barycentric coordinate vector in a given plane via
  !! \f$ alpha = ((baryCoord - center) \cdot vecA)/\sqrt(vecA\cdot vecA)\f$
  !! - Compute spatial value using
  !! \f$ res = (1+alpha)*(1-alpha) \f$
  !! Actual parabola 2d formula: \f$ (y-y_0) = a(x-x_0)^2 \f$
  !! parabola open towards opposite x and with vertex at (1,0) is given by
  !! \f$ y = (1-x)^2 \f$
  !! 
  function tem_spatial_parabol2d_for_treeIds( me, treeIds, tree, n ) result(res)
    !---------------------------------------------------------------------------
    !> contains parameters for 2d parabola
    type( tem_canonicalND_type ) :: me
    !> global treelm mesh
    type( treelmesh_type ), intent( in ) ::tree
    !> number of return values
    integer, intent( in ) :: n
    !> treeIds of elements in given level
    integer(kind=long_k), intent( in ) :: treeIds(n)
    !> return value of a function
    real(kind=rk) :: res(n)
    !---------------------------------------------------------------------------
    real(kind=rk) :: coord(3), alpha, diff(3)
    real(kind=rk) :: vecAsqr, center(3), halfdir(3)
    !loop variables
    integer :: iDir
    !---------------------------------------------------------------------------

    ! since this is a line only one of the three vectors in vec
    ! is active
    halfdir = 0.0_rk
    do iDir = 1, 3
      if( me%active( iDir ))                                                   &
        &        halfdir = me%vec( :, iDir )/2.0_rk
    end do
    center = me%origin + halfdir
    vecAsqr = dot_product( halfdir, halfdir )
    !loop over number of return values
    do iDir = 1, n

      !barycentric coordinate
      coord = tem_BaryOfId( tree, treeIds(iDir) )

      !distance between parabola center and barycentric coordinates
      diff = coord - center

      !projection of diff in a plane on vecA
      alpha = dot_product( diff, halfdir )                                    &
        &   / vecAsqr

      res(iDir)  = ( 1.0_rk-alpha ) * ( 1.0_rk+alpha )

      if ( abs(alpha) .gt. 1.0d0 ) then
        res(iDir) = 0.0_rk
      end if

    end do

  end function tem_spatial_parabol2d_for_treeIds
!******************************************************************************!


!******************************************************************************!
  !>  This function computes 2d parabola profile from coord of elements
  !!
  !! This profile is defined
  !! by element barycentric coordinate and 2d parabola parameters.
  !! 2D parabola profile at given plane is computed in the following way:
  !! - Project barycentric coordinate vector in a given plane via
  !! \f$ alpha = ((baryCoord - center) \cdot vecA)/\sqrt(vecA\cdot vecA)\f$
  !! - Compute spatial value using
  !! \f$ res = (1+alpha)*(1-alpha) \f$
  !! 
  function tem_spatial_parabol2d_for_coord( me, coord, n ) result(res)
    !---------------------------------------------------------------------------
    !> contains line parameters for 2d parabola
    type( tem_canonicalND_type ) :: me
    !> number of return values
    integer, intent( in ) :: n
    !> barycentric Ids of an elements.
    !! 1st index goes over number of elements and
    !! 2nd index goes over x,y,z coordinates
    real(kind=rk), intent( in ) :: coord(n,3)
    !> return value of a function
    real(kind=rk) :: res(n)
    !---------------------------------------------------------------------------
    real(kind=rk) :: alpha, diff(3)
    real(kind=rk) :: vecAsqr, center(3), halfdir(3)
    ! loop variable
    integer :: iDir
    !---------------------------------------------------------------------------

    halfdir = 0.0_rk
    do iDir = 1, 3
      if( me%active( iDir ))                                                   &
        &        halfdir = me%vec( :, iDir )/2.0_rk
    end do

    vecAsqr = dot_product( halfdir, halfdir )

    center = me%origin + halfdir
    !loop over number of return values
    do iDir = 1, n

      !distance between parabola center and barycentric coordinates
      diff = coord(iDir,:) - center

      !projection of diff in a plane on vecA
      alpha =  dot_product( diff, halfdir ) / vecAsqr

      res(iDir)  = ( 1.0_rk-alpha ) * ( 1.0_rk+alpha )

      if ( abs(alpha) .gt. 1.0d0 ) then
        res(iDir) = 0.0_rk
      end if

    end do

  end function tem_spatial_parabol2d_for_coord
!******************************************************************************!


!******************************************************************************!
  !> Produce random numbers
  !!
  function tem_spatial_random_for( me, n ) result(res)
    !---------------------------------------------------------------------------
    !> Interval definition to get the random values from.
    type(spatial_random_type), intent(in) :: me
    !> number of return values
    integer, intent(in) :: n
    !> return value of a function
    real(kind=rk) :: res(n)
    !---------------------------------------------------------------------------
    call random_number(res)
    res = me%val_min + me%val_range*res

  end function tem_spatial_random_for
!******************************************************************************!


!******************************************************************************!
  !> This function invokes the Lua function,
  !! in which the barycentric coordinates are computed from treeIds of an
  !! element.
  !!
  !! Lua function defined in the script is
  !! connected to the conf handle and return the result of the function.
  !! The Lua function takes barycentric coordinate as input argument
  !! i.e fun_name(x,y,z)
  function tem_spatial_lua_for_treeIds( me, conf, treeIds, tree, n ) result(res)
    !---------------------------------------------------------------------------
    !> aotus path type in tem spatial type
    type( aot_path_type ) :: me
    !> global treelm mesh
    type( treelmesh_type ), intent( in ) ::tree
    !> number of return values
    integer, intent( in ) :: n
    !> treeIds of elements in given level
    integer(kind=long_k), intent( in ) :: treeIds(n)
    !> lua state
    type(flu_State), optional :: conf
    !> return value
    real(kind=rk) :: res(n)
    !---------------------------------------------------------------------------
    type(aot_fun_type) :: fun
    integer :: iError
    integer :: iDir, jDir
    real(kind=rk) :: coord(3)
    type(flu_State) :: loc_conf
    logical :: openLua
    !---------------------------------------------------------------------------

    if (present(conf)) then
      loc_conf = conf
      openLua = .false.
    else
      openLua = .true.
    end if

    call aot_path_open(me, conf=loc_conf, fun=fun, openLua=openLua)

    do iDir=1,n
      coord = tem_BaryOfId( tree, treeIds(iDir) )
      do jDir=1,3
        call aot_fun_put(L=loc_conf, fun=fun, arg=coord(jDir))
      end do
      call aot_fun_do(L=loc_conf, fun=fun, nresults=1)
      call aot_top_get_val(L=loc_conf, val=res(iDir), ErrCode=iError)
    end do

    call aot_path_close(me, conf=loc_conf, fun=fun, closeLua=openLua)

  end function tem_spatial_lua_for_treeIds
!******************************************************************************!


!******************************************************************************!
  !>  This function invokes the vectorial Lua function,
  !! in which the barycentric coordinates are computed from treeIds of an
  !! element.
  !!
  !! Lua function defined in the script is
  !! connected to the conf handle and return the result of the function.
  !! The Lua function takes barycentric coordinate as input argument
  !! i.e fun_name(x,y,z)
  function tem_spatial_lua_vector_for_treeIds( me, conf, treeIds, tree,        &
    &                                          n, ncomp ) result(res)
    !---------------------------------------------------------------------------
    !> aotus path type in tem spatial type
    type(aot_path_type) :: me
    !> global treelm mesh
    type(treelmesh_type), intent(in) ::tree
    !> number of return values
    integer, intent(in) :: n
    !> Number of components in each returned value
    integer, intent(in) :: ncomp
    !> treeIds of elements in given level
    integer(kind=long_k), intent(in) :: treeIds(n)
    !> lua state
    type(flu_State), optional :: conf
    !> return value
    real(kind=rk) :: res(n,ncomp)
    !---------------------------------------------------------------------------
    type(aot_fun_type) :: fun
    integer :: iError(nComp)
    integer :: iDir, jDir
    real(kind=rk) :: coord(3)
    type(flu_State) :: loc_conf
    logical :: openLua
    !---------------------------------------------------------------------------

    if (present(conf)) then
      loc_conf = conf
      openLua = .false.
    else
      openLua = .true.
    end if

    call aot_path_open(me, conf=loc_conf, fun=fun, openLua=openLua)

    do iDir=1,n
      coord = tem_BaryOfId( tree, treeIds(iDir) )
      do jDir=1,3
        call aot_fun_put(L=loc_conf, fun=fun, arg=coord(jDir))
      end do
      call aot_fun_do(L=loc_conf, fun=fun, nresults=1)
      call aot_top_get_val(L=loc_conf, val=res(iDir,:), ErrCode=iError)
    end do

    call aot_path_close(me, conf=loc_conf, fun=fun, closeLua=openLua)

  end function tem_spatial_lua_vector_for_treeIds
!******************************************************************************!


!******************************************************************************!
  !> This function invokes the Lua function,
  !! which takes barycentric coordinates of an element.
  !!
  !! Lua function defined in the script is
  !! connected to the conf handle and return the result of the function.
  !! The Lua function takes barycentric coordinate as input argument
  !! i.e fun_name(x,y,z)
  !!
  function tem_spatial_lua_for_coord( me, conf, coord, n ) result(res)
    !---------------------------------------------------------------------------
    !> aotus path type in tem spatial type
    type( aot_path_type ) :: me
    !> number of return values
    integer, intent( in ) :: n
    !> barycentric Ids of an elements.
    !! 1st index goes over number of elements and
    !! 2nd index goes over x,y,z coordinates
    real(kind=rk), intent( in ) :: coord(n,3)
    !> lua state
    type(flu_State), optional :: conf
    !> return value
    real(kind=rk) :: res(n)
    !---------------------------------------------------------------------------
    type(aot_fun_type) :: fun
    integer :: iError
    integer :: iDir, jDir
    type(flu_State) :: loc_conf
    logical :: openLua
    !---------------------------------------------------------------------------

    if (present(conf)) then
      loc_conf = conf
      openLua = .false.
    else
      openLua = .true.
    end if

    call aot_path_open(me, conf=loc_conf, fun=fun, openLua=openLua)

    do iDir = 1, n
      do jDir = 1, 3
        call aot_fun_put( L=loc_conf, fun=fun, arg=coord(iDir,jDir) )
      end do
      call aot_fun_do(L=loc_conf, fun=fun, nresults=1)
      call aot_top_get_val(L=loc_conf, val=res(iDir), ErrCode=iError)
    end do

    call aot_path_close(me, conf=loc_conf, fun=fun, closeLua=openLua)

  end function tem_spatial_lua_for_coord
!******************************************************************************!


!******************************************************************************!
  !> This function invokes the vectorial Lua function,
  !! which takes barycentric coordinates of an element.
  !!
  !! Lua function defined in the script is
  !! connected to the conf handle and return the result of the function.
  !! The Lua function takes barycentric coordinate as input argument
  !! i.e fun_name(x,y,z)
  !!
  function tem_spatial_lua_vector_for_coord( me, conf, coord, n, ncomp )       &
    &      result(res)
    !---------------------------------------------------------------------------
    !> aotus path type in tem spatial type
    type(aot_path_type) :: me
    !> number of return values
    integer, intent(in) :: n
    !> number of components in the resulting vector
    integer, intent(in) :: ncomp
    !> barycentric Ids of an elements.
    !! 1st index goes over number of elements and
    !! 2nd index goes over x,y,z coordinates
    real(kind=rk), intent(in) :: coord(n,3)
    !> lua state
    type(flu_State), optional :: conf
    !> return value
    real(kind=rk) :: res(n,ncomp)
    !---------------------------------------------------------------------------
    type(aot_fun_type) :: fun
    integer :: iError(ncomp)
    integer :: iDir, jDir
    type(flu_State) :: loc_conf
    logical :: openLua
    !---------------------------------------------------------------------------

    if (present(conf)) then
      loc_conf = conf
      openLua = .false.
    else
      openLua = .true.
    end if

    call aot_path_open(me, conf=loc_conf, fun=fun, openLua=openLua)

    do iDir = 1, n
      do jDir = 1, 3
        call aot_fun_put( L=loc_conf, fun=fun, arg=coord(iDir,jDir) )
      end do
      call aot_fun_do(L=loc_conf, fun=fun, nresults=1)
      call aot_top_get_val(L=loc_conf, val=res(iDir,:), ErrCode=iError)
    end do

    call aot_path_close(me, conf=loc_conf, fun=fun, closeLua=openLua)

  end function tem_spatial_lua_vector_for_coord
!******************************************************************************!


!******************************************************************************!
  !> This function invokes different spatial boundary kinds like constant, lua
  !! function or predefined Fortran function for given treeIDs
  !!
  !! If a spatial block is not defined and a temporal block is defined in the
  !! lua file, the return value = ref_value.
  !! If both spatial and temporal block are not defined in the lua file, the
  !! return value = 1.0_rk.
  !! based spatial_kind(kind).
  !! case 1. const - set constant value
  !! case 2. lua_fun - lua function
  !! case 3. gausspulse - fortran gauss pulse function
  !! case 4. 2dcrvp     - fortran spinning vortex function
  !! case 5. parabol    - fotran parabolic function
  !!
  function tem_spatial_for_treeIDs( me, treeIds, tree, n, conf ) result( res )
    !--------------------------------------------------------------------------
    !> spatial type for given boundary state
    type( tem_spatial_type ) :: me
    !> global treelm mesh
    type( treelmesh_type ), intent( in ) ::tree
    !> number of return values
    integer, intent( in ) :: n
    !> treeIds of elements in given level
    integer(kind=long_k), intent( in ) :: treeIds(n)
    !> lua state
    type(flu_State), optional, intent(in) :: conf
    !> return value of a function
    real( kind=rk ) :: res(n)
    !--------------------------------------------------------------------------

    select case( trim(adjustl(me%kind)) )
    case( 'none' )
      res = me%const

    case( 'const' )
      res = me%const

    case( 'lua_fun' )
      res = tem_spatial_lua_for( me%LuaPath, conf, treeIds, tree, n)

    case( 'parabol' )
      select case( trim(adjustl(me%parabol%geometry%canoND(1)%kind)) )
      case('line')
        res = tem_spatial_parabol2d_for( me%parabol%geometry%canoND(1),        &
          &                              treeIds, tree, n )
      case('plane')
        res = tem_spatial_parabol3d_for( me%parabol%geometry%canoND(1),        &
          &                              treeIds, tree, n )
      endselect
      res = res*me%parabol%amplitude

    case ('random')
      res = tem_spatial_random_for(me%random, n)
!    case( 'gausspulse' )
!      res = ic_gausspulse_for(desc%gausspulse, coord, n)
!    case( '2dcrvpX' )
!      res = ic_2dcrvpX_for(desc%crvp, coord, n)
!    case( '2dcrvpY' )
!      res = ic_2dcrvpY_for(desc%crvp, coord, n)
    case default
      call tem_write('ERROR: Unknown spatial function in tem_spatial_for_treeIDs.')
      call tem_write('Aborting...')
      call tem_abort()
    end select

   end function tem_spatial_for_treeIDs
!******************************************************************************!


!******************************************************************************!
  function tem_spatial_vector_for_treeIDs( me, treeIds, tree, n, ncomp, conf ) &
    &      result(res)
    !--------------------------------------------------------------------------=
    !> spatial type for given boundary state
    type(tem_spatial_type) :: me
    !> global treelm mesh
    type(treelmesh_type), intent( in ) ::tree
    !> number of return values
    integer, intent(in) :: n
    !> number of components in the resulting vector
    integer, intent(in) :: ncomp
    !> treeIds of elements in given level
    integer(kind=long_k), intent(in) :: treeIds(n)
    !> lua state
    type(flu_State), optional, intent(in) :: conf
    !> return value of a function
    real(kind=rk) :: res(n,ncomp)
    !--------------------------------------------------------------------------=

    select case( trim(adjustl(me%kind)) )
    case( 'none' )
      res = me%const
    case( 'const' )
      res = me%const
    case( 'lua_fun' )
      res = tem_spatial_lua_for( me%LuaPath, conf, treeIds, tree, n, ncomp )
    case default
      call tem_write('ERROR: No vectorial routine for spatial functions of')
      call tem_write('       kind "' // trim(adjustl(me%kind)) // '"') 
      call tem_write('       available! Have a look in the tem_spatial_module')
      call tem_write('       for implemented vectorial functions.')
      call tem_write('Aborting...')
      call tem_abort()
    end select

   end function tem_spatial_vector_for_treeIDs
!******************************************************************************!


!******************************************************************************!
  !> This function invokes different spatial boundary kinds like constant, lua
  !! function or predefined Fortran function for given coord
  !!
  !! If a spatial block is not defined and a temporal block is defined in the
  !! lua file, the return value = ref_value.
  !! If both spatial and temporal block are not defined in the lua file, the
  !! return value = 1.0_rk.
  !! based spatial_kind(kind).
  !! - case 1. const - set constant value
  !! - case 2. lua_fun - lua function
  !! - case 3. gausspulse - fortran gauss pulse function
  !! - case 4. 2dcrvp     - fortran spinning vortex function
  !! - case 5. parabol    - fotran parabolic function
  !!
  function tem_spatial_for_coord( me, coord, n, conf ) result( res )
    !---------------------------------------------------------------------------
    !> spatial type for given boundary state
    type(tem_spatial_type), intent(in) :: me
    !> number of return values
    integer, intent( in ) :: n
    !> barycentric Ids of an elements.
    !! 1st index goes over number of elements and
    !! 2nd index goes over x,y,z coordinates
    real(kind=rk), intent( in ) :: coord(n,3)
    !> lua state
    type(flu_State), optional, intent(in) :: conf
    !> return value of a function
    real( kind=rk ) :: res(n)
    !---------------------------------------------------------------------------
    type(aot_path_type) :: luaPath

    select case( trim(adjustl(me%kind)) )
    case( 'none' )
      res = me%const
    case( 'const' )
      res = me%const
    case( 'lua_fun' )
      luaPath = me%LuaPath
      res = tem_spatial_lua_for(LuaPath, conf, coord, n)
    case ('random')
      res = tem_spatial_random_for(me%random, n)
    case( 'parabol' )
      select case( trim(adjustl(me%parabol%geometry%canoND(1)%kind)) )
      case('line')
        res = tem_spatial_parabol2d_for( me%parabol%geometry%canoND(1),        &
          &                              coord, n)
      case('plane')
        res = tem_spatial_parabol3d_for( me%parabol%geometry%canoND(1),        &
          &                              coord, n)
      endselect
      res = res*me%parabol%amplitude
    case( 'gausspulse' )
      res = ic_gausspulse_for(me%gausspulse, coord, n)
    case( 'crvpX' )
      res = ic_2dcrvpX_for(me%crvp, coord, n)
    case( 'crvpY' )
      res = ic_2dcrvpY_for(me%crvp, coord, n)
    case( 'crvpPressure' )
      res = ic_2dcrvpPressure_for(me%crvp, coord, n)
    case('miescatter_displacementfieldz')
      res = tem_eval_miescatter_displz( me = me%mie_fun, &
                                      & coord = coord,   &
                                      & time = 0.0_rk,   &
                                      & n = n)
    case('miescatter_magneticfieldx')
      res = tem_eval_miescatter_magnx( me = me%mie_fun, &
                                      & coord = coord,   &
                                      & time = 0.0_rk,   &
                                      & n = n)
    case('miescatter_magneticfieldy')
      res = tem_eval_miescatter_magny( me = me%mie_fun, &
                                      & coord = coord,   &
                                      & time = 0.0_rk,   &
                                      & n = n)
    case default
      call tem_write('ERROR: Unknown spatial function in tem_spatial_for_coord.')
      call tem_write('Aborting...')
      call tem_abort()
    end select

   end function tem_spatial_for_coord
!******************************************************************************!


!******************************************************************************!
  !> This function invokes different spatial boundary kinds like constant, lua
  !! function or predefined Fortran function for given coord
  !!
  !! If a spatial block is not defined and a temporal block is defined in the
  !! lua file, the return value = ref_value.
  !! If both spatial and temporal block are not defined in the lua file, the
  !! return value = 1.0_rk.
  !! based spatial_kind(kind).
  !! - case 1. const - set constant value
  !! - case 2. lua_fun - lua function
  !! - case 3. gausspulse - fortran gauss pulse function
  !! - case 4. 2dcrvp     - fortran spinning vortex function
  !! - case 5. parabol    - fotran parabolic function
  function tem_spatial_vector_for_coord( me, coord, n, ncomp, conf )           &
    &      result( res )
    !---------------------------------------------------------------------------
    !> spatial type for given boundary state
    type(tem_spatial_type) :: me
    !> number of return values
    integer, intent(in) :: n
    !> number of components per returned value
    integer, intent(in) :: ncomp
    !> barycentric Ids of an elements.
    !! 1st index goes over number of elements and
    !! 2nd index goes over x,y,z coordinates
    real(kind=rk), intent(in) :: coord(n,3)
    !> lua state
    type(flu_State), optional, intent(in) :: conf
    !> return value of a function
    real( kind=rk ) :: res(n,ncomp)
    !---------------------------------------------------------------------------

    select case( trim(adjustl(me%kind)) )
    case( 'none' )
      res = me%const
    case( 'const' )
      res = me%const
    case( 'lua_fun' )
      res = tem_spatial_lua_for(me%LuaPath, conf, coord, n, ncomp)
    case default
      call tem_write('ERROR: No vectorial routine for spatial functions of')
      call tem_write('       kind "' // trim(adjustl(me%kind)) // '"') 
      call tem_write('       available! Have a look in the tem_spatial_module')
      call tem_write('       for implemented vectorial functions.')
      call tem_write('Aborting...')
      call tem_abort()
    end select

   end function tem_spatial_vector_for_coord
!******************************************************************************!


end module tem_spatial_module
!******************************************************************************!
