! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Stencil definitions for the elements
!!
!! \author Simon Zimny, Manuel Hasert, Kannan Masilmani, ...
!!
!! A stencil is basically a set of element-offsets $(s_x, s_y, s_z)$,
!! describing the relative positions of all required elements for a given 
!! element.
!!
!!\todo: SZ: set the right reference
!! \ref tem_construction_example_stencil
!!
?? include 'arrayMacros.inc' 
!!
module tem_stencil_module

  ! include treelm modules
  use env_module,            only: rk, long_k, zeroLength, minLength
  use tem_tools_module,      only: tem_write
  use tem_aux_module,        only: tem_abort
  use tem_param_module,      only: qQQQ, qOffset
  use tem_comm_module,       only: tem_communication_type, tem_commPattern_type
  use tem_comm_env_module,   only: tem_comm_env_type
  use tem_grow_array_module, only: grw_intArray_type
  use tem_dyn_array_module,  only: dyn_longArray_type
  use tem_logging_module,    only: logUnit, tem_toStr

  ! include aotus modules
  use aotus_module,     only: flu_State
  use aot_table_module, only: aot_table_open, aot_table_close,                 &
    &                         aot_table_length, aot_get_val

  implicit none

  private

  public :: tem_stencilHeader_type
  public :: tem_stencilElement_type
  public :: tem_stencil_dump
  public :: tem_stencil_getHeaderPos
  public :: tem_loadStencil
  public :: tem_identify_prevailDirections
  public :: tem_identify_inverseDirections
  public :: tem_stencil_map_toTreelmDef
  public :: tem_stencil_zeroPos
  public :: d3q19_cxDir, d3q27_cxDir
  public :: d3q7_cxDir, d3q6_cxDir, d2q9_cxDir, d1q3_cxDir
  public :: tem_create_stencil
  public :: tem_create_d3q19, tem_create_d3q27
  public :: tem_create_d3q7, tem_create_d3q6, tem_create_d2q9
  public :: tem_create_d1q3
  public :: grw_stencilHeaderArray_type
  public :: grw_stencilElementArray_type
  public :: init, destroy, append, expand, empty, truncate

  ! cartesian directions of discrete velocities
  integer, dimension( 3, 19 ),parameter  :: d3q19_cxDir =                      &
  ! axis parallel offsets
  reshape((/ -1,  0,  0,   &        ! West(W)
              0, -1,  0,   &        ! South(S)
              0,  0, -1,   &        ! Bottom(B)
              1,  0,  0,   &        ! East(E)
              0,  1,  0,   &        ! North(N)
              0,  0,  1,   &        ! Top(T)
  ! diagonal offsets 
              0, -1, -1,   &        ! Bottom South(BS)
              0, -1,  1,   &        ! Top South(TS)
              0,  1, -1,   &        ! Bottom North(BN)
              0,  1,  1,   &        ! Top North(TN)
             -1,  0, -1,   &        ! Bottom West(BW)
              1,  0, -1,   &        ! Bottom East(BE)
             -1,  0,  1,   &        ! Top West(TW)
              1,  0,  1,   &        ! Top East(TE)
             -1, -1,  0,   &        ! South West(SW)
             -1,  1,  0,   &        ! North West(NW)
              1, -1,  0,   &        ! South East(SE)
              1,  1,  0,   &        ! North East(NE)
  ! central offset
              0,  0,  0/),(/3,19/)) ! Center(C)

! cartesian directions of discrete velocities
  integer, dimension( 3, 27 ),parameter  :: d3q27_cxDir =                      &
  ! axis parallel offsets
  reshape((/ -1,  0,  0,   &        ! West(W)
              0, -1,  0,   &        ! South(S)
              0,  0, -1,   &        ! Bottom(B)
              1,  0,  0,   &        ! East(E)
              0,  1,  0,   &        ! North(N)
              0,  0,  1,   &        ! Top(T)
  ! diagonal offsets 
              0, -1, -1,   &        ! Bottom South(BS)
              0, -1,  1,   &        ! Top South(TS)
              0,  1, -1,   &        ! Bottom North(BN)
              0,  1,  1,   &        ! Top North(TN)
             -1,  0, -1,   &        ! Bottom West(BW)
              1,  0, -1,   &        ! Bottom East(BE)
             -1,  0,  1,   &        ! Top West(TW)
              1,  0,  1,   &        ! Top East(TE)
             -1, -1,  0,   &        ! South West(SW)
             -1,  1,  0,   &        ! North West(NW)
              1, -1,  0,   &        ! South East(SE)
              1,  1,  0,   &        ! North East(NE)
  ! corner offsets 
             -1, -1, -1,   &        ! Bottom South West(BSW)
             -1, -1,  1,   &        ! Top South West(TSW)
             -1,  1, -1,   &        ! Bottom North West(BNW)
             -1,  1,  1,   &        ! Top North West(TNW)
              1, -1, -1,   &        ! Bottom South East(BSE)
              1, -1,  1,   &        ! Top South East(TSE)
              1,  1, -1,   &        ! Bottom North East(BNE)
              1,  1,  1,   &        ! Top North East(TNE)
  ! central offset
              0,  0,  0/),(/3,27/)) ! Center(C)

  integer, dimension( 3, 7 ),parameter  :: d3q7_cxDir =                        &
  ! axis parallel offsets
  reshape((/ -1,  0,  0,   &       ! West(W)
              0, -1,  0,   &       ! South(S)
              0,  0, -1,   &       ! Bottom(B)
              1,  0,  0,   &       ! East(E)
              0,  1,  0,   &       ! North(N)
              0,  0,  1,   &       ! Top(T)
  ! central offset
              0,  0,  0/),(/3,7/)) ! Center(C)

  integer, dimension( 3, 6 ),parameter  :: d3q6_cxDir =                        &
  ! axis parallel offsets
  reshape((/ -1,  0,  0,   &        ! West(W)
              0, -1,  0,   &        ! South(S)
              0,  0, -1,   &        ! Bottom(B)
              1,  0,  0,   &        ! East(E)
              0,  1,  0,   &        ! North(N)
              0,  0,  1 /),(/3,6/)) ! Top(T)

  ! cartesian directions of discrete velocities
  integer, dimension( 3, 9 ),parameter  :: d2q9_cxDir =                        &
  ! axis parallel offsets
  reshape((/ -1,  0,  0,   &       ! West(W)
              0, -1,  0,   &       ! South(S)
              1,  0,  0,   &       ! East(E)
              0,  1,  0,   &       ! North(N)
  ! diagonal offsets
             -1, -1,  0,   &       ! South West(SW)
             -1,  1,  0,   &       ! North West(NW)
              1, -1,  0,   &       ! South East(SE)
              1,  1,  0,   &       ! North East(NE)
  ! central offset
              0,  0,  0/),(/3,9/)) ! Center(C)

  ! cartesian directions of discrete velocities
  integer, dimension( 3, 3 ),parameter  :: d1q3_cxDir =                        &
  ! axis parallel offsets
  reshape((/ -1,  0,  0,   &       ! West(W)
              1,  0,  0,   &       ! East(E)
              0,  0,  0/),(/3,3/)) ! Center(C)

  !> Stencil definitions
  !! 
  !! A stencil is basically a set of element-offsets $(s_x, s_y, s_z)$,
  !! describing the relative positions of all required elements for
  !! a given element.
  type tem_stencilHeader_type

    !> a stencil label
    character(len=32) :: label

    !> number of directions
    integer :: QQ

    !>\todo HK: do we really need this? It is mostly confusing, and solver
    !!          specific!
    !!
    !! number of directions excluding the central (0,0,0)
    integer :: QQN
    integer :: nDims = 0

    !> cartesian directions of discrete velocities 
    !! Size is (3,nDir)
    integer, allocatable :: cxDir(:,:)

    !> inverted cartesian direction indices.
    !! this is well defined for symmetric stencils only.
    integer, allocatable :: cxDirInv(:)

    !> mapping of stencil entries to treelm definition, if possible
    !! entry is zero if nothing defined,
    !! -1 if a stencil entry with entries further than neighbor are encountered
    integer, allocatable :: map(:)

    !> Rest-density position in stencil
    integer :: restPosition

    !> the stencil on which the current one depends
    !! this is required for boundary stencils which elements actually require
    !! the neighbors of the compute stencil
    integer :: depStencil

    !> Stencil kind
    integer :: stencilKind

    !> use this stencil for all elements
    logical :: useAll

    !> requires valid neighbors of the stencil neighbors
    logical :: requireNeighNeigh = .false.

    !> the number of elements using this stencil
    integer :: nElems

    !> list of elements on which this stencil should be applied
    type(grw_intArray_type), allocatable :: elemLvl(:)

    !> list of elements on which this stencil should be applied
    integer, allocatable :: elem(:)

  end type tem_stencilHeader_type

  !> Element stencil definition
  type tem_stencilElement_type

    !> number of entries in pos
    integer :: QQN

    !> position in the tem_element_type::neighID( iStencil )
    integer, allocatable :: tIDpos(:)

    !> the stencil on which the current one depends
    !! this is required for boundary stencils which elements actually require
    !! the neighbors of the compute stencil
    integer :: headerPos

    !> position in the tem_element_type::neighID( iStencil )
    integer, allocatable :: totalPos(:)

    !> Require compute stencil neighbors?
    logical :: requireNeighNeigh = .false.

  end type tem_stencilElement_type


! declare the growing array of stencil type
?? copy :: GA_decltxt( stencilElement, type(tem_stencilElement_type))
?? copy :: GA_decltxt( stencilHeader, type(tem_stencilHeader_type))
  interface init
    module procedure init_stencilHeader
    module procedure init_stencilElement
  end interface 

  interface destroy
    module procedure empty_stencil
    module procedure destroy_stencilElement
  end interface 

  interface empty
    module procedure empty_stencil
  end interface 

  interface tem_stencil_dump
    module procedure tem_stencilHeader_dump
    module procedure tem_stencilElement_dump
  end interface 

  contains

!******************************************************************************!
  !> Include the subroutines for the growing array.
?? copy :: GA_impltxt(stencilElement, type(tem_stencilElement_type))
?? copy :: GA_impltxt(stencilHeader,  type(tem_stencilHeader_type))

!******************************************************************************!
  !> Load the stencil configuration from the lua file
  !!
  subroutine tem_loadStencil( stencil, parent_handle, conf )
    !---------------------------------------------------------------------------
    !> stencil type to fill
    type(tem_stencilHeader_type) :: stencil
    !> handle of the parent table
    integer, intent(in)    :: parent_handle
    !> lua state type 
    type(flu_State) :: conf
    !---------------------------------------------------------------------------
    integer :: nCartDiscVel, iCartDiscVel
    integer :: nDiscreteVel, iDiscreteVel
    integer :: discVel_handle
    integer :: discVel_cart_handle
    integer :: iError
    !---------------------------------------------------------------------------
    ! Usually, the stencil to read applies for all elements
    ! (Compute stencil)
    stencil%useAll = .true.

    ! read the number of discrete velocities QQ
    ! \todo{ MH: Do we really need to read QQ explicitely?
    ! Isn't it enough to just count the number of entries below?}
    call aot_get_val( L       = conf,                                          &
      &               thandle = parent_handle,                                 &
      &               val     = stencil%QQ,                                    &
      &               ErrCode = iError,                                        &
      &               key     = 'QQ' )
    ! save the number of discrete velocities without the resting velocity
    stencil%QQN = stencil%QQ - 1

    ! read the discrete velocities
    call aot_table_open( L       = conf,                                       &
      &                  parent  = parent_handle,                              &
      &                  thandle = discVel_handle,                             &
      &                  key     = 'disc_vel' )

    ! Check the length of the discrete velocity table
    nDiscreteVel = aot_table_length( L = conf, thandle = discVel_handle )

    if (nDiscreteVel .eq. stencil%QQ ) then
      ! If the number of discrete velocities matches the given stencil QQ
      allocate( stencil%cxDir(3,stencil%QQ) )

      do iDiscreteVel = 1, nDiscreteVel

        call aot_table_open( L       = conf,                                   &
          &                  parent  = discVel_handle,                         &
          &                  thandle = discVel_cart_handle,                    &
          &                  pos     = iDiscreteVel )
        nCartDiscVel = aot_table_length( L       = conf,                       &
          &                              thandle = discVel_cart_handle )

        if (nCartDiscVel .eq. 3) then

          do iCartDiscVel=1, nCartDiscVel
            call aot_get_val(                                                  &
              &         L       = conf,                                        &
              &         thandle = discVel_cart_handle,                         &
              &         val     = stencil%cxDir( iCartDiscVel, iDiscreteVel ), &
              &         ErrCode = iError,                                      &
              &         pos     = iCartDiscVel )
          end do
        else
          call tem_write( "Not every velocity has three cartesian "//          &
            &             "coordinates!!")
        end if

      end do
      call aot_table_close( L = conf, thandle = discVel_cart_handle )

    else 
      call tem_write("The number of discrete velocities for the new stencil "//&
        &            "does not match QQ!!")
      stop
    end if
  end subroutine tem_loadStencil
!******************************************************************************!


!******************************************************************************!
  !> Delete all entries in the stencil
  !!
  subroutine empty_stencil( me )
    !---------------------------------------------------------------------------
    !> stencil header type to empty
    type( tem_stencilHeader_type ), intent(out) :: me 
    !---------------------------------------------------------------------------
    me%QQ  = 0
    me%QQN = 0
    if( allocated( me%cxDir )) deallocate( me%cxDir )
    if( allocated( me%cxDirInv )) deallocate( me%cxDirInv )
    if( allocated( me%map )) deallocate( me%map )
    if( allocated( me%elem )) deallocate( me%elem )
    if( allocated( me%elemLvl )) deallocate( me%elemLvl )

  end subroutine empty_stencil
!******************************************************************************!


!******************************************************************************!
  !> This subroutine fills the array of inverse directions according to the
  !! given array of directions.
  !!
  subroutine tem_identify_inverseDirections( me, cxDir )
    !---------------------------------------------------------------------------
    !> array of positions of the inverse offsets in cxDir
    integer, intent(out) :: me(:)
    !> array of given offsets
    integer, intent(in)  :: cxDir(:,:)
    !---------------------------------------------------------------------------
    ! counter variable
    integer :: iOffset, iCompare
    !---------------------------------------------------------------------------

    do iOffset = 1, size( me )
      do iCompare = 1, size( me )
        if( (( -cxDir( 1, iOffset )) .eq. ( cxDir( 1, iCompare ))) .and.       &
          & (( -cxDir( 2, iOffset )) .eq. ( cxDir( 2, iCompare ))) .and.       &
          & (( -cxDir( 3, iOffset )) .eq. ( cxDir( 3, iCompare ))) )then
          me( iOffset ) = iCompare
        end if
      end do
    end do
  end subroutine tem_identify_inverseDirections
!******************************************************************************!


!******************************************************************************!
  !> This subroutine fills the array of prevailing directions according to the
  !! given array of directions.
  !!
  subroutine tem_identify_prevailDirections( me, cxDir )
    !---------------------------------------------------------------------------
    !> growing array of prevailing directions
    real( kind=rk ), allocatable, intent(out) :: me(:,:)
    !> array of given offsets
    integer, intent(in)  :: cxDir(:,:)
    !---------------------------------------------------------------------------
    real(kind=rk),allocatable :: tmpPrevailDir(:,:)
    real(kind=rk) :: length
    ! counters
    integer :: iOffset, iDir
    integer :: nEntries ! size of array
    integer :: counter  
    !---------------------------------------------------------------------------

    nEntries = size( cxDir, 2 )
    allocate( tmpPrevailDir( 3, nEntries ))
    counter = 0
    do iOffset = 1, nEntries
      length = 0._rk
      do iDir = 1, 3
        length = length + abs(real(cxDir( iDir, iOffset), kind=rk))
      end do
      if(length .eq. 0._rk)then
          tmpPrevailDir( :, iOffset) = 0._rk
      else
        counter = counter + 1
        !\todo exclude duplicate vectors here
        do iDir = 1, 3
          tmpPrevailDir(iDir, iOffset) =                                       &
          &                        abs( real(cxDir( iDir, iOffset ), kind=rk)/ &
          &                        real(sqrt(length), kind=rk) )
        end do
      end if
    end do
    allocate( me( 3, counter ))
    do iOffset = 1, counter
      me(:, iOffset) = tmpPrevailDir( :, iOffset)
    enddo

  end subroutine tem_identify_prevailDirections
!******************************************************************************!


!******************************************************************************!
  !> This subroutine fills the array of prevailing directions according to the
  !! given array of directions.
  !!
  subroutine tem_stencil_map_toTreelmDef( me )
    !---------------------------------------------------------------------------
    !> stencil to map
    type(tem_stencilHeader_type) :: me
    !---------------------------------------------------------------------------
    ! counters
    integer :: iBC, iStencilElem
    !---------------------------------------------------------------------------

    if( allocated( me%map )) deallocate( me%map )
    allocate( me%map( me%QQ ))
    me%map = 0
    do iStencilElem = 1,me%QQ
      if( maxval( abs( me%cxDir( :,iStencilElem ))) .le. 1) then
        ! Now match the stencil offset with the treelm definition offsets
boundaryLoop:  do iBC = 1, qQQQ
          if( me%cxDir( 1, iStencilElem ) .eq. qOffset( iBC, 1 )               &
        .and. me%cxDir( 2, iStencilElem ) .eq. qOffset( iBC, 2 )               &
        .and. me%cxDir( 3, iStencilElem ) .eq. qOffset( iBC, 3 ) ) then
            me%map( iStencilElem ) = iBC
            exit boundaryLoop
          endif
        enddo boundaryLoop
      else ! stencil is larger than one
        me%map( iStencilElem ) = 0
      endif
    enddo

  end subroutine tem_stencil_map_toTreelmDef
!******************************************************************************!


!******************************************************************************!
!  !> Communicate the complete stencil
!  !!
!  !! Currently, this assumes same stencils for all participating elements
!  subroutine tem_stencil_communicate( send, recv, stencil, neighID, prc, commPattern, nStencilElems )
!    !--------------------------------------------------------------------------
!    type( tem_communication_type ), intent(inout) :: send, recv
!    type(tem_commPattern_type), intent(in)     :: commPattern
!    type(tem_stencilElement_type), intent(inout)  :: stencil(:)
!    type(dyn_longArray_type), intent(inout)  :: neighID(:)
!    type(tem_comm_env_type), intent(in) :: prc
!    integer, intent(in) :: nStencilElems !< amount of values to communicate
!    !--------------------------------------------------------------------------
!    integer :: iStencilElem, iElem
!    integer(kind=long_k), allocatable :: buffer(:)
!    !--------------------------------------------------------------------------
!
!    call tem_write('Exchanging stencil does not work yet!')
!    call tem_abort
!    allocate( buffer( size( stencil )))
!    ! First exchange the number of stencil entries
!    call commPattern%exchange_int( send = send, recv = recv,                &
!      &                            state = stencil(:)%QQN, message_flag = 0 )
!
!    ! then exchange all the elements of the tID entry
!    do iStencilElem = 1, nStencilElems
!      do iElem = 1, size( stencil )
!        !buffer(iElem) = stencil(iElem)%neighID%val( stencil( iElem )%tIDpos( iStencilElem ))
!      enddo
!      call commPattern%exchange_long( send = send, recv = recv,                &
!        &                             state = buffer, message_flag = 0 )
!      do iElem = 1, size( stencil )
!        !stencil(iElem)%tID( iStencilElem ) = buffer( iElem )
!      enddo
!    end do
!    
!  end subroutine tem_stencil_communicate
!******************************************************************************!


!******************************************************************************!
  !> Write element information to disk
  !!
  subroutine tem_stencilHeader_dump( me, nUnit )
    !---------------------------------------------------------------------------
    !> stencil header to be dumped
    type( tem_stencilHeader_type ), intent(in) :: me
    integer, optional, intent(in) :: nUnit
    !---------------------------------------------------------------------------
    integer :: iDir
    integer :: dmpU
    !---------------------------------------------------------------------------
   
    ! Default to write to logUnit
    dmpU = logUnit(3)

    if (present(nUnit)) then
      dmpU = nUnit
    end if

    write(dmpU, *) '   Stencil  '
    write(dmpU, *) '     QQ   : ', me%QQ
    write(dmpU, *) '     QQN  : ', me%QQN

    if( allocated( me%cxDir )) then
      write(dmpU, *) '     Offset directions:'
      do iDir=1,me%QQN
        write(dmpU, *) '       cxDir(:,', iDir, ') = ' &
            &          //trim( tem_toStr( val = me%cxDir(:,iDir), &
            &                             sep = ','               ) )
      end do
    end if
    write(dmpU, *) '     useAll: ', tem_toStr(me%useAll)
    if( .not. me%useAll ) then
      write(dmpU, *) '     nElems: ', me%nElems
      if( allocated( me%elem  ))  &
      write(dmpU, *) '     Elems : ' &
          &          //trim( tem_toStr(val = me%elem, &
          &                            sep = ','      ) )
    end if

  end subroutine tem_stencilHeader_dump
!******************************************************************************!


!******************************************************************************!
  !> Write element information to disk
  !!
  subroutine tem_stencilElement_dump( me, nUnit, neighID, tIDonly )
    !---------------------------------------------------------------------------
    !> stencil element to be dumped
    type( tem_stencilElement_type ), intent(in) :: me
    !> unit to dump to
    integer, intent(in) :: nUnit
    !> neighbor ID
    integer(kind=long_k), allocatable, intent(in), optional :: neighID(:)
    !> only use tree IDs
    logical, intent(in), optional :: tIDonly
    !---------------------------------------------------------------------------
    integer :: iVal
    integer(kind=long_k) :: bufID
    logical :: tIDonly_loc
    logical :: neighID_loc
    !---------------------------------------------------------------------------

    if( present( tIDonly )) then
      tIDonly_loc = tIDonly
    else 
      tIDonly_loc = .false.
    endif

    if( .not. tIDonly_loc ) then 
      write(nUnit,*) '   Stencil  '
      write(nUnit,*) '     QQN    ', me%QQN, '  headerPos ', me%headerPos
      if ( allocated(me%tIDpos) ) &
        &  write(nUnit, *) '     pos    ', me%tIDpos
    end if

    neighID_loc = .false.
    if( present( neighID  )) then
      if( allocated( neighID )) then
        if( allocated( me%tIDpos )) then
          neighID_loc = .true. 
        end if
      end if
    end if
    if( neighID_loc  ) then
      if( size( me%tIDpos ) .gt. 0 ) then 
        if( maxval( me%tIDpos ) .gt. 0 ) then
          write(unit=nUnit,advance='no',fmt='(a,i3,a)') '     neighID (',    &
            &                                             me%headerPos, ') ' 
          do iVal = 1, size( me%tIDpos )
            if( me%tIDpos( iVal ) .gt. 0 ) then
              bufID = neighID( me%tIDpos( iVal))
            else
              bufID = 0 
            end if
            write(unit=nUnit,advance='no',fmt='(i15)') bufID
          end do
          write(nUnit,*) ''
        end if
      end if
    end if

  end subroutine tem_stencilElement_dump
!******************************************************************************!


!******************************************************************************!
  !> This subroutine creates the predefined d3q19 stencil.
  !!
  subroutine tem_create_d3q19( stencil )
    !---------------------------------------------------------------------------
    !> stencil type to be defined
    type( tem_stencilHeader_type), intent(inout) :: stencil
    !---------------------------------------------------------------------------

    call tem_write('  ...d3q19') 
    ! save the number of directions and the number of directions 
    ! excluding the zero velocity
    stencil%QQ  = 19
    stencil%QQN = 18
    stencil%nDims = 3
    stencil%label = 'd3q19'

    ! save the cartesian directions of the discrete velocities     
    allocate( stencil%cxDir( 3, stencil%QQ ))
    allocate( stencil%cxDirInv( stencil%QQ ))

    stencil%cxDir = d3q19_cxDir
    call tem_identify_inverseDirections( stencil%cxDirInv, stencil%cxDir )

    stencil%useAll = .true.

  end subroutine tem_create_d3q19
!******************************************************************************!


!******************************************************************************!
  !> This subroutine creates the predefined d3q19 stencil.
  !!
  subroutine tem_create_d3q27( stencil )
    !---------------------------------------------------------------------------
    !> stencil type to be defined
    type( tem_stencilHeader_type), intent(inout) :: stencil
    !---------------------------------------------------------------------------

    call tem_write('  ...d3q27') 
    ! save the number of directions and the number of directions 
    ! excluding the zero velocity
    stencil%QQ  = 27
    stencil%QQN = 26
    stencil%nDims = 3
    stencil%label = 'd3q27'

    ! save the cartesian directions of the discrete velocities     
    allocate( stencil%cxDir( 3, stencil%QQ ))
    allocate( stencil%cxDirInv( stencil%QQ ))

    stencil%cxDir = d3q27_cxDir
    call tem_identify_inverseDirections( stencil%cxDirInv, stencil%cxDir )

    stencil%useAll = .true.

  end subroutine tem_create_d3q27
!******************************************************************************!


!******************************************************************************!
  !> This subroutine creates the predefined d3q7 stencil.
  !!
  subroutine tem_create_d3q7( stencil )
    !---------------------------------------------------------------------------
    !> stencil type to be defined
    type( tem_stencilHeader_type), intent(inout) :: stencil
    !---------------------------------------------------------------------------

    call tem_write('  ...d3q7') 
    ! save the number of directions and the number of directions 
    ! excluding the zero velocity
    stencil%QQ  = 7
    stencil%QQN = 6
    stencil%nDims = 3
    stencil%label = 'd3q7'

    ! save the cartesian directions of the discrete velocities     
    allocate( stencil%cxDir( 3, stencil%QQ ))
    allocate( stencil%cxDirInv( stencil%QQ ))

    ! setting the offsets

    ! axis parallel offsets
    stencil%cxDir = d3q7_cxDir

    ! setting the position of the inverse directions in the cxDir array
    call tem_identify_inverseDirections( stencil%cxDirInv, stencil%cxDir )

  end subroutine tem_create_d3q7
!******************************************************************************!


!******************************************************************************!
  !> This subroutine creates the predefined d3q6 stencil.
  !!
  subroutine tem_create_d3q6( stencil )
    !---------------------------------------------------------------------------
    !> stencil type to be defined
    type( tem_stencilHeader_type), intent(inout) :: stencil
    !---------------------------------------------------------------------------

    call tem_write('  ...d3q6') 
    ! save the number of directions and the number of directions 
    ! excluding the zero velocity
    stencil%QQ  = 6
    stencil%QQN = 6
    stencil%nDims = 3
    stencil%label = 'd3q6'

    ! save the cartesian directions of the discrete velocities     
    allocate( stencil%cxDir( 3, stencil%QQ ))
    allocate( stencil%cxDirInv( stencil%QQ ))

    ! setting the offsets

    ! axis parallel offsets
    stencil%cxDir = d3q6_cxDir

    ! setting the position of the inverse directions in the cxDir array
    call tem_identify_inverseDirections( stencil%cxDirInv, stencil%cxDir )

  end subroutine tem_create_d3q6
!******************************************************************************!


!******************************************************************************!
  !> This subroutine creates the predefined d2q9 stencil.
  !!
  subroutine tem_create_d2q9( stencil )
    !---------------------------------------------------------------------------
    !> stencil type to be defined
    type( tem_stencilHeader_type), intent(inout) :: stencil
    !---------------------------------------------------------------------------

    call tem_write('  ...d2q9') 
    ! save the number of directions and the number of directions 
    ! excluding the zero velocity
    stencil%QQ  = 9
    stencil%QQN = 8
    stencil%nDims = 2
    stencil%label = 'd2q9'

    ! save the cartesian directions of the discrete velocities    
    if( allocated( stencil%cxDir )) deallocate(stencil%cxDir )
    if( allocated( stencil%cxDirInv )) deallocate(stencil%cxDirInv )
    allocate( stencil%cxDir( 3, stencil%QQ ))
    allocate( stencil%cxDirInv( stencil%QQ ))

    stencil%cxDir = d2q9_cxDir
  
    call tem_identify_inverseDirections( stencil%cxDirInv, stencil%cxDir )

    stencil%useAll = .true.

  end subroutine tem_create_d2q9
!******************************************************************************!

!******************************************************************************!
  !> This subroutine creates the predefined d1q3 stencil.
  !!
  subroutine tem_create_d1q3( stencil )
    !---------------------------------------------------------------------------
    !> stencil type to be defined
    type( tem_stencilHeader_type), intent(inout) :: stencil
    !---------------------------------------------------------------------------

    call tem_write('  ...d1q3') 
    ! save the number of directions and the number of directions 
    ! excluding the zero velocity
    stencil%QQ  = 3
    stencil%QQN = 2
    stencil%nDims = 1
    stencil%label = 'd1q3'

    ! save the cartesian directions of the discrete velocities    
    allocate( stencil%cxDir( 3, stencil%QQ ))
    allocate( stencil%cxDirInv( stencil%QQ ))

    stencil%cxDir = d1q3_cxDir
  
    call tem_identify_inverseDirections( stencil%cxDirInv, stencil%cxDir )

    stencil%useAll = .true.

  end subroutine tem_create_d1q3
!******************************************************************************!


!******************************************************************************!
  !> This subroutine creates the required stencil.
  !!
  subroutine tem_create_stencil( stencil, stencilKind )
    !---------------------------------------------------------------------------
    !> stencil type to be defined
    type( tem_stencilHeader_type), intent(out) :: stencil
    !> stencil kind to decide which create function to call
    character(len=*) :: stencilKind
    !---------------------------------------------------------------------------
    select case (trim(stencilKind))
      case ('d3q19')
        call tem_create_d3q19(stencil)  
      case ('d3q27')
        call tem_create_d3q27(stencil)  
      case ('d3q7')
        call tem_create_d3q7(stencil)
      case ('d3q6')
        call tem_create_d3q6(stencil)
      case ('d2q9')
        call tem_create_d2q9(stencil)
      case ('default')
        call tem_write("Stencil Name Not defined yet!!")
        call tem_abort()
    end select
  end subroutine tem_create_stencil
!******************************************************************************!


!******************************************************************************!
  !> Identify the zero-position in the stencil
  !! Return -1 if not found in stencil
  !!
  function tem_stencil_zeroPos( me ) result( pos )
    !---------------------------------------------------------------------------
    !> stencil to map
    type(tem_stencilHeader_type),intent(in) :: me
    !> Position of zero-entry in the stencil
    integer :: pos
    !---------------------------------------------------------------------------
    ! counter
    integer :: iStencilElem
    !---------------------------------------------------------------------------
    pos = -1
    do iStencilElem = 1, me%QQ
      if( me%cxDir( 1, iStencilElem ) .eq. 0 .and.                             &
        & me%cxDir( 2, iStencilElem ) .eq. 0 .and.                             &
        & me%cxDir( 3, iStencilElem ) .eq. 0 ) then
        pos = iStencilElem
      endif
    enddo
  end function tem_stencil_zeroPos
!******************************************************************************!


!******************************************************************************!
  !> initialize stencil
  !!
  subroutine init_stencilHeader( me, QQN, QQ, useAll )
    !---------------------------------------------------------------------------
    !> stencil header to be initialized
    type( tem_stencilHeader_type ), intent(out) :: me
    !> number of discrete velocities in the model (without the center one)
    integer, intent(in) :: QQN
    !> number of discrete velocities in the model (incl. the center one)
    integer, intent(in), optional :: QQ
    !> use this stencil for all elements?
    logical, intent(in), optional :: useAll
    !---------------------------------------------------------------------------
    !---------------------------------------------------------------------------
    me%QQN = QQN
    if( present( QQ )) me%QQ  = QQ
    if( present( useAll )) me%useAll = useAll
 
  end subroutine init_stencilHeader
!******************************************************************************!


!******************************************************************************!
  !> initialize stencil
  !!
  subroutine init_stencilElement( me, QQN, headerPos, tIDpos )
    !---------------------------------------------------------------------------
    !> stencil element type to be initialized
    type( tem_stencilElement_type ), intent(out) :: me
    !> number of discrete velocities in the model (without the center one)
    integer, intent(in) :: QQN
    !> 
    integer, intent(in), optional :: headerPos
    !>
    integer, intent(in), optional :: tIDpos(:)
    !---------------------------------------------------------------------------
    !---------------------------------------------------------------------------
    me%QQN = QQN
    if( allocated( me%tIDpos ) ) deallocate( me%tIDpos )
    allocate( me%tIDpos( me%QQN ) )
    if( present( tIDpos )) then
      me%tIDpos  = tIDpos
    else
      me%tIDpos  = 0
    end if
    me%headerPos = 0
    if( present( headerPos )) me%headerPos = headerPos
    allocate( me%totalPos( me%QQN ))
    me%totalPos = 0
 
  end subroutine init_stencilElement
!******************************************************************************!


!******************************************************************************!
  !> destroy stencil
  !!
  subroutine destroy_stencilElement( me )
    !---------------------------------------------------------------------------
    !> stencil element type to be destroyed
    type( tem_stencilElement_type ), intent(out) :: me
    !---------------------------------------------------------------------------
    !---------------------------------------------------------------------------
    me%QQN = 0
    if( allocated( me%tIDpos ) ) deallocate( me%tIDpos )
    me%headerPos = 0
    if( allocated( me%totalPos ) ) deallocate( me%totalPos )
 
  end subroutine destroy_stencilElement
!******************************************************************************!


!******************************************************************************!
  !> find position
  !!
  function tem_stencil_getHeaderPos( me, val ) result( headerPos )
    !---------------------------------------------------------------------------
    !>
    type( grw_stencilElementArray_type ), intent(in) :: me
    !>
    integer, intent(in) :: val
    !>
    integer :: headerPos
    !---------------------------------------------------------------------------
    integer :: iStencil
    !---------------------------------------------------------------------------
    headerPos = 0
    ! Run over all the stencils
    do iStencil = 1, me%nVals
      if( me%val( iStencil )%headerPos .eq. val ) then
        headerPos = iStencil
      end if
    end do
 
  end function tem_stencil_getHeaderPos
!******************************************************************************!


end module tem_stencil_module
!******************************************************************************!
