! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module provides a description to steer time dependent facilities.
!!
!! It is for example required for the tracking and restart facilities, to
!! describe when to do the respective outputs.
!!
module tem_timeControl_module

  ! include treelm modules
  use env_module,         only: rk, labelLen
  use tem_tools_module,   only: tem_write, tem_horizontalSpacer
  use tem_time_module,    only: tem_time_type
  use tem_logging_module, only: tem_log, tem_toStr, logUnit

  ! include aotus modules
  use aotus_module,     only: flu_State
  use aot_table_module, only: aot_table_open, aot_table_close, aot_get_val
  use aot_out_module,   only: aot_out_type, aot_out_val,                       &
    &                         aot_out_open_table, aot_out_close_table

  implicit none
  private

  public :: tem_timeControl_type
  public :: tem_timeControl_dump
  public :: tem_timeControl_out
  public :: tem_timeControl_checkRuntime
  public :: tem_load_timeControl
  public :: tem_timeControlToInt
  public :: tem_time_activeTimeStep
  public :: tem_time_updateTrigger
  public :: tem_time_determineEnd
  public :: tem_time_determine_lastslot

  !> Measure and control the current run time
  type tem_runtime_type
    !> Time it took for initialization
    real(kind=rk) :: init   = 0._rk
    !> Number of intervals already done
    integer       :: nIntervals  = 0
    !> max wall clock limit in seconds
    real(kind=rk) :: limit
  end type tem_runtime_type

  !> This type describes controlling quantities for time dependent facilities.
  !!
  !! there are
  !! - iteration-wise measures (integers) and
  !! - simulation-time measures (reals)
  type tem_timeControl_type
    !> This flag indicates, wether the time is measured in iterations.
    !!
    !! If it is .true., the integer quantities below are used for the control
    !! decision, otherwise the real quantities are used.
    logical :: useIterations

    ! Simulation-time measure (reals)
    real(kind=rk) :: tInterval !< Interval of simulation time.
    !> Simulation time from where on the controller should be active.
    real(kind=rk) :: tMin
    !> Simulation time up to which the controller should be active.
    real(kind=rk) :: tMax
    !> Keep track of the next point in time to trigger this timer.
    real(kind=rk) :: trigger_time

    ! Iteration-wise measure (integers)
    integer :: iInterval !< Interval in terms of iterations.
    !> Iteration from where on the controller should be active.
    integer :: iMin
    !> Iteration after which, the controller is inactive again.
    integer :: iMax
    !> Keep track of the next iteration at which to trigger this timer. 
    integer :: trigger_iter
    !> Run time (wall clock) limits and current 
    type( tem_runtime_type ) :: runTime
  end type

  contains

!******************************************************************************!
  !> This subroutine reads all time control dependend data from the lua config
  !! file. Depending on the logical useIterations simulation time or iterations
  !! are read.
  !!
  subroutine tem_load_timeControl( me, conf, parent, key, refTimeControl )
    !---------------------------------------------------------------------------
    !> time control type to fill from config file
    type(tem_timeControl_type), intent(inout) :: me
    !> lua config file
    type(flu_state) :: conf
    !> parent table to read from
    integer, intent(in), optional :: parent
    !> key to read from
    character(len=*),intent(in), optional :: key
    !> Reference time control object, which passes properties to the tracker's
    !! timeControl object (mainly for useIterations property)
    type( tem_timeControl_type ),intent(in), optional :: refTimeControl
    !---------------------------------------------------------------------------
    ! defining local variables
    integer :: iError, thandle
    character(len=labelLen) :: localKey
    ! Default value for the useIteration. If refTimeControl is given, take from
    ! there
    logical :: iterationDefault
    !---------------------------------------------------------------------------
    if( present(key) ) then
      localKey = key
    else
      localKey = 'time'
    endif
    if( present(refTimeControl) ) then
      iterationDefault = refTimeControl%useIterations
    else
      iterationDefault = .true.
    endif

    call aot_table_open( L       = conf,                                       &
      &                  parent  = parent,                                     &
      &                  thandle = thandle,                                    &
      &                  key     = localKey )
    me%iInterval = 1
    me%tInterval = 0._rk
    me%trigger_iter = 0
    me%trigger_time = 0._rk

    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               key     = 'useIterations',                               &
      &               val     = me%useIterations,                              &
      &               ErrCode = iError,                                        &
      &               default = iterationDefault )

    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               key     = 'wall_max',                                    &
      &               val     = me%runTime%limit,                              &
      &               ErrCode = iError,                                        &
      &               default = -1._rk )
    if( me%runTime%limit .gt. 0._rk ) then
      call tem_write('    Wall clock limit set to:  ', me%runTime%limit )
    end if

    if( me%useIterations ) then


      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               key     = 'interval',                                  &
        &               val     = me%iInterval,                                &
        &               default = 1,                                           &
        &               ErrCode = iError)

      ! Do not allow negative interval specifications.
      me%iInterval = max(me%iInterval, 1)

      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               key     = 'min',                                       &
        &               val     = me%iMin,                                     &
        &               default = 0,                                           &
        &               ErrCode = iError )
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               key     = 'max',                                       &
        &               default = huge( me%iMax),                              &
        &               val     = me%iMax,                                     &
        &               ErrCode = iError )
      !> \todo HK: Should a missing max for the time interval be fatal?
      !!           We might run into undefined behavior here!
      !!           Either we need a default, or abort...

      ! Negative timestep limit indicates to do as many time steps as possible.
      if (me%iMax .lt. 0) then
        me%iMax = huge( me%iMax )
      end if

      ! Limit the interval to the difference between last und first time step.
      ! (0-based difference)
      ! Interval is not allowed to be zero (used in modulo operation)
      me%iInterval    = max( min(me%iInterval, me%iMax - me%iMin), 1 )
      me%trigger_iter = me%iMin
    else
      !> \todo HK: It is even more unclear for the real time definition, here
      !!           we might get undefined behavior for interval and max!
      !!           Need to think about sane defaults (least surprises to the
      !!           user), and conditions upon which to abort!
      !!
      !!           I now set both to 1.0_rk per default, but this is totally
      !!           arbitrary, maybe we shoul abort on missing definitions here.
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               key     = 'interval',                                  &
        &               val     = me%tInterval,                                &
        &               default = 1.0_rk,                                      &
        &               ErrCode = iError )

      ! Limit interval specification to non negative numbers.
      me%tInterval = max(me%tInterval, 0._rk)
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               key     = 'min',                                       &
        &               val     = me%tMin,                                     &
        &               default = 0.0_rk,                                      &
        &               ErrCode = iError )
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               key     = 'max',                                       &
        &               val     = me%tMax,                                     &
        &               default = 1.0_rk,                                      &
        &               ErrCode = iError )
      me%trigger_time = me%tMin
    end if

    call tem_timeControl_dump( me )

    call aot_table_close( L = conf, thandle = thandle )

  end subroutine tem_load_timeControl
!******************************************************************************!


!******************************************************************************!
  !> Determine if at the current time step the action defined within timeControl
  !! is active. Returns true if the action has to be perfomed
  !!
  function tem_time_activeTimeStep(timing, timeControl, reqInt )               &
    &            result( isActive )
    !---------------------------------------------------------------------------
    !> current timings
    type(tem_time_type), intent(in) :: timing
    !> time control type to check for
    type(tem_timeControl_type), intent(in) :: timeControl
    !> Required interval, in which the update MUST occur.
    !! This is required for the musubi multilevel, where the time step should
    !! only be determined active, when the end of the largest cycle is reached.
    integer, intent(in), optional :: reqInt
    !> is this type active?
    logical :: isActive
    !---------------------------------------------------------------------------
    isActive = .false.
    if( timeControl%useIterations ) then
      if (  timing%iTime .le. timeControl%iMax   .and.                         &
        &   timing%iTime .ge. timeControl%iMin ) then  
        isActive =  (timing%iTime .ge. timeControl%trigger_iter)
        if ( isActive ) then
          if( present( reqInt )) then
            isActive = (mod( timing%nIters, reqInt) .eq. 0)
          end if
        end if     
      end if     
    else
      if (timeControl%trigger_time <= timeControl%tMax) then
        isActive = (timing%rTime >= timeControl%trigger_time)
      end if
    end if

  end function tem_time_activeTimeStep
!******************************************************************************!


!******************************************************************************!
  !> Update the triggers in the time control object, if they were surpassed
  !! by the current time.
  !!
  subroutine tem_time_updateTrigger(me, now)
    !---------------------------------------------------------------------------
    !> time control type incl. the trigger
    type(tem_timeControl_type), intent(inout) :: me
    !> current timings
    type(tem_time_type), intent(in) :: now
    !---------------------------------------------------------------------------

    if (me%useIterations) then
      if (now%iTime >= me%trigger_iter) then
        me%trigger_iter = max(me%trigger_iter + me%iInterval, now%iTime)
      end if
    else
      if (now%rTime >= me%trigger_time) then
        me%trigger_time = max(me%trigger_time + me%tInterval, now%rTime)
      end if
    end if
  end subroutine tem_time_updateTrigger
!******************************************************************************!


!******************************************************************************!
  !> Determine if the end has been reached. 
  !! Returns true if the time has passed over the defined tmax
  !!
  function tem_time_determineEnd(now, timeControl ) result( isEnd )
    !---------------------------------------------------------------------------
    !> current timings
    type(tem_time_type), intent(in) :: now
    !> time control type to check for
    type(tem_timeControl_type), intent(in) :: timeControl
    !> are the conditions in the control type satisfied
    logical :: isEnd
    !---------------------------------------------------------------------------
    isEnd = .false.

    if( timeControl%useIterations ) then
      if( now%iTime .ge. timeControl%iMax ) then
        isEnd =  .true.
      endif     
    else
      if( now%rTime .ge. timeControl%tMax ) then
        isEnd = .true.
      end if
    end if
    ! Is any wall clock limit given?
    if( timeControl%runTime%limit .gt. 0 ) then

    end if

  end function tem_time_determineEnd
!******************************************************************************!
  !> Determine the last slot from where the simulation will initiate. This is
  !! useful particularly in dynamic load balancing. This function returns the 
  !! last iMin after which there will be no further load balancing intervals
  function tem_time_determine_lastslot(timeControl) result(lastslot)
    !---------------------------------------------------------------------------
    !> time control type to check for
    type(tem_timeControl_type), intent(in) :: timeControl
    !> Return type integer
    integer :: lastslot
    !---------------------------------------------------------------------------
    lastslot = timeControl%iMin + (( timeControl%iMax - timeControl%iMin)/     &
      &                              timeControl%iInterval)              *     &
      &                           timeControl%iInterval   

  end function tem_time_determine_lastslot
!******************************************************************************!


!******************************************************************************!


!******************************************************************************!
  !> write out the time control object settings on the root process only
  !!
  !! \todo: DEB: Add a level here?
  !!
  subroutine tem_timeControl_dump( me )
    !---------------------------------------------------------------------------
    !> control type to be dumped
    type(tem_timeControl_type), intent(in) :: me
    !---------------------------------------------------------------------------
    character(len=labelLen) :: buffer
    !---------------------------------------------------------------------------

    if( me%useIterations ) then
      call tem_write( 'Timing control with usage of iterations' )
      write(buffer,'(3(a,i7),a)') 'tMin: ', me%iMin, '    tMax: ', me%iMax
      call tem_write( ' '//trim(buffer))
      write(buffer,'(3(a,i7),a)') 'tInterval: ', me%iInterval, 'trigger iter:',&
        &                         me%trigger_iter
    else
      call tem_write( 'Timing control' )
      write(buffer,'(3(a,e7.2),a)') 'tMin: ', me%tMin, '    tMax: ', me%tMax,  &
        &                           '     tInterval: ', me%tInterval
    endif
    call tem_write( ' '//trim(buffer))

  end subroutine tem_timeControl_dump
!******************************************************************************!


!******************************************************************************!
  !> This routine write information in timeControl type into lua file
  !!
  subroutine tem_timeControl_out( me , conf )
    !---------------------------------------------------------------------------
    !> contains time control information
    type(tem_timeControl_type), intent(in) :: me
    !> aotus lua state to write output
    type(aot_out_type) :: conf
    !---------------------------------------------------------------------------

    call aot_out_open_table( put_conf = conf, tname = 'time')
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'useIterations',                              &
      &               val      = me%useIterations )

    if ( me%useIterations ) then
    !write integer values
      call aot_out_val( put_conf = conf, vname = 'min', val = me%imin)
      call aot_out_val( put_conf = conf, vname = 'max', val = me%imax)
      call aot_out_val( put_conf = conf, vname = 'interval', val = me%iInterval)
    else
      call aot_out_val( put_conf = conf, vname = 'min', val = me%tmin)
      call aot_out_val( put_conf = conf, vname = 'max', val = me%tmax)
      call aot_out_val( put_conf = conf, vname = 'interval', val = me%tInterval)
    end if

    call aot_out_close_table( put_conf = conf )

  end subroutine tem_timeControl_out
!******************************************************************************!


!******************************************************************************!
  !> Convert time between real and integer time
  !! 
  subroutine tem_timeControlToInt( me, dt )
    !---------------------------------------------------------------------------
    !> time control type to be converted
    type(tem_timeControl_type), intent(inout) :: me
    !> temporal stepping
    real(kind=rk), intent(in) :: dt
    !---------------------------------------------------------------------------

    if( .not. me%useIterations ) then
      me%useIterations = .true.
      me%iInterval  = max(1, int(me%tInterval/dt))
      me%iMax       = ceiling(me%tMax/dt)
      me%iMin       = int(me%tMin/dt)
      me%trigger_time = int(me%trigger_time/dt)
    end if

    write(logUnit(1),*) ' Converting the time control object from real '//   &
      &                   'to int' 
    call tem_timeControl_dump( me )

  end subroutine tem_timeControlToInt
!******************************************************************************!


!******************************************************************************!
  !> Determine if maximum runTime was reached by the current time or within the
  !! next interval.
  !!
  !! Below you see the time line of the execution of the code.
  !! It starts with the initialization phase up to me%runTime%init
  !! The compute loop then starts, with the check interval chosen in the main
  !! time object 
  !! \verbatim time = { tmax = 10., interval = 1. } \endverbatim
  !! Checks are performed in each of the intervals.
  !! The number of intervals are counted up in me%runTime%nIntervals.
  !! The average duration of an interval is computed by the total compute time
  !! up to now = ( now - me%runTime%init ) divided by the number of intervals
  !! checked.
  !! Then we can estimate the duration of the next check interval. 
  !! If the wall clock limit is reached within the next check interval, we must
  !! cleanly exit the code now.
  !!
  !! Time line
  !!\verbatim
  !!                    1      2      3      4      5      6        nIntervals
  !! |-----------------||------|------|------|------|------|-x----||
  !!      |                 |                                |     L tEnd
  !!    init              dtInterval                     wall_clock_limit
  !!\endverbatim
  !!
  subroutine tem_timeControl_checkRuntime(me, now, reached )
    !---------------------------------------------------------------------------
    !> control type to be checked
    type(tem_timeControl_type), intent(inout) :: me
    !> current timings
    real(kind=rk), intent(in) :: now
    !> has the end of this run been reached?
    logical, intent(inout) :: reached
    !---------------------------------------------------------------------------
    real(kind=rk) :: dtInterval, computeTime, tNext
    !---------------------------------------------------------------------------
    ! Only if we have not yet reached the wall clock limit, test
    if( .not. reached .and. me%runTime%limit .gt. 0._rk ) then
      me%runTime%nIntervals = me%runTime%nIntervals + 1
      computeTime = (now - me%runTime%init) 
      if( me%runTime%nIntervals .gt. 0 ) then
        dtInterval = computeTime / me%runTime%nIntervals
      else 
        dtInterval = 0._rk
      end if
      tNext = dtInterval*(me%runTime%nIntervals + 1)
      if( tNext + me%runTime%init .gt.  me%runTime%limit ) then
        reached = .true.
      end if
    end if

    if( reached ) then
      call tem_horizontalSpacer( before = 1 )
      call tem_write('        time per interval [s]:  ', real( dtInterval ))
      call tem_write('  current wall clock time [s]:  ', real( now ))
      call tem_write('  predicted next interval [s]:  ',                       &
        &            real( tNext + me%runTime%init ))
      call tem_write(' reached wall clock limit [s]:  ',                       &
        &            real( me%runTime%limit ))
      call tem_horizontalSpacer( after = 1 ) 
    end if

  end subroutine tem_timeControl_checkRuntime
!******************************************************************************!


end module tem_timeControl_module
!******************************************************************************!
