! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Routines and definitions relating to the current, maximum and starting time
!! for the solvers
module tem_time_module

  ! include treelm modules
  use env_module,         only: rk, output_unit, labelLen
  use tem_tools_module,   only: tem_write
  use tem_logging_module, only: logUnit

  ! include aotus modules
  use aotus_module,     only: flu_State, aot_get_val
  use aot_out_module,   only: aot_out_type, aot_out_val,                       &
    &                         aot_out_open_table, aot_out_close_table
  use aot_table_module, only: aot_table_open, aot_table_close

  implicit none

  private

  public :: tem_time_type
  public :: tem_time_dump
  public :: tem_time_out 
  public :: tem_time_load 
  public :: tem_time_output
  public :: tem_time_reset
  public :: tem_generate_timeStamp
  public :: tem_extract_timeFromtimeStamp

  !> Global solver time information
  !!
  !! - iteration-wise measures (integers) and
  !! - simulation-time measures (reals)
  type tem_time_type
    integer :: nIters = 0   !< number of performed iterations

    ! iteration-wise (integer)
    integer :: iTime  = 0   !< current iteration = nIters + start_iter
    integer :: iStart = 0   !< iteration to start
    logical :: iTime_set = .false.

    ! simulation-time-wise (real)
    real(kind=rk) :: rTime   = 0.0_rk !< the global simulation time
!    ! Start and stopp should not be part of the time_type.
!    ! Currently only timeControl%start / stop are being used. 
!    ! Will be removed soon
    real(kind=rk) :: rStart  = 0.0_rk !< the starting time of the simulation
    real(kind=rk) :: rStop !< the final simulation time
    integer :: iStop  = 0   !< maximum number of iterations
  end type tem_time_type

  interface tem_generate_timeStamp
    module procedure tem_generate_timeStampFromReal
    module procedure tem_generate_timeStampFromTimeType
  end interface tem_generate_timeStamp

  contains

!******************************************************************************
  !> write the time information to the unit given
  !!
  subroutine tem_time_output( me, loglevel )
    !--------------------------------------------------------------------------
    !> time type to be outputted
    type( tem_time_type ), intent(in) :: me
    integer, optional, intent(in) :: logLevel
    !--------------------------------------------------------------------------
    integer :: ll

    if (present(logLevel)) then
      ll = logLevel
    else
      ll = 1
    end if

    write(logUnit(ll),*) ' Solver time information '
    write(logUnit(ll),*) '     # iterations:   ', me%nIters
    write(logUnit(ll),*) '     # iStart:       ', me%iStart
    write(logUnit(ll),*) '     # iStop:        ', me%iStop
    write(logUnit(ll),*) '     # iTime:        ', me%iTime
    write(logUnit(ll),*) '       rStart:       ', me%rStart
    write(logUnit(ll),*) '       rStop:        ', me%rStop
    write(logUnit(ll),*) '       rTime:        ', me%rTime
  end subroutine tem_time_output
!******************************************************************************


!******************************************************************************
  !> write the time information to the unit given
  !!
  subroutine tem_time_dump( me, outUnit )
    !--------------------------------------------------------------------------
    !> time type to be dumped
    type( tem_time_type ), intent(in) :: me
    !> unit to write to
    integer, intent(in), optional :: outUnit
    !--------------------------------------------------------------------------
    integer :: nUnit
    !--------------------------------------------------------------------------
    if( present( outUnit )) then
      nUnit = outUnit
    else
      nUnit = output_unit
    endif
    write( nUnit, *) 'Timing '
    write( nUnit, *) ' nIters         ', me%nIters
    write( nUnit, *) ' iStart         ', me%iStart
    write( nUnit, *) ' iStop          ', me%iStop
    write( nUnit, *) ' iTime          ', me%iTime
    write( nUnit, *) ' rStart         ', me%rStart
    write( nUnit, *) ' rStop          ', me%rStop
    write( nUnit, *) ' rTime          ', me%rTime

  end subroutine tem_time_dump
!******************************************************************************!


!******************************************************************************!
  !> Reading the time object from a lua table
  subroutine tem_time_load(me, conf, key, parent)
    !---------------------------------------------------------------------------
    !> time object to write into the lua file
    type(tem_time_type), intent(inout) :: me
    !> aotus type handling the output to the file in lua format
    type(flu_state), intent(inout) :: conf
    !> parent handle
    integer, intent(in), optional    :: parent
    !> table key to load from
    character(len=*), intent(in), optional :: key
    !---------------------------------------------------------------------------
    character(len=labelLen) :: loc_key
    integer :: iErr, thandle
    !---------------------------------------------------------------------------
    if( present( key )) then
      loc_key    = key   
    else
      loc_key    = 'time'
    endif
    call aot_table_open( L       = conf,                                       &
      &                  parent  = parent,                                     &
      &                  thandle = thandle,                                    &
      &                  key     = loc_key )
    ! Check here if the table was successfully opened. 
    ! If not, try to read the keyword directly
    if( thandle .ne. 0 ) then
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               val     = me%iTime,                                    &
        &               key     = 'iTime',                                     &
        &               ErrCode = iErr )
      if( iErr .eq. 0 ) then
        me%iTime_set = .true.
      end if
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               val     = me%rTime,                                    &
        &               key     = 'rTime',                                     &
        &               ErrCode = iErr )
      call aot_table_close( L = conf, thandle = thandle )
    else
      me%iTime_set =  .false.
      call aot_get_val( L       = conf,                                        &
        &               key     = loc_key,                                     &
        &               val     = me%rTime,                                    &
        &               ErrCode = iErr )
    end if

  end subroutine tem_time_load
!******************************************************************************!


!******************************************************************************!
  !> Writing the time object to a lua table
  !!
  subroutine tem_time_out( me, conf, key )
    !---------------------------------------------------------------------------
    !> time object to write into the lua file
    type(tem_time_type), intent(in) :: me
    !> aotus type handling the output to the file in lua format
    type(aot_out_type), intent(inout) :: conf
    !> tbale key to read from
    character(len=*), intent(in), optional :: key
    !---------------------------------------------------------------------------
    character(len=labelLen) :: loc_key
    !---------------------------------------------------------------------------
    if( present( key )) then
      loc_key    = key   
    else
      loc_key    = 'time'
    endif

    call aot_out_open_table( put_conf = conf, tname = loc_key )
    call aot_out_val( put_conf = conf,                                         &
      &               val      = me%iTime,                                     &
      &               vname    = 'iTime' )
    call aot_out_val( put_conf = conf,                                         &
      &               val      = me%rTime,                                     &
      &               vname    = 'rTime' )
    call aot_out_close_table( put_conf = conf )

  end subroutine tem_time_out
!******************************************************************************!


!******************************************************************************!
  !> generate a unique time stamp from the information in time_type
  !!
  !! this basically generates a string identifying the solution time and
  !! writing in a meaningful format, so that it can be easily recognized
  !!
  function tem_generate_timeStampFromReal( fromTime  ) result( timeStamp )
    !---------------------------------------------------------------------------
    !> current point in time
    real( kind=rk ), intent(in) :: fromTime
    !> time stamp generated
    character(len=16)     :: timeStamp
    !---------------------------------------------------------------------------
    ! procedure to set the filename in a good way
    write(timestamp,'(en12.3)') fromTime
    ! adjust the time stamp left in the string
    timestamp = adjustl( timestamp)

  end function  tem_generate_timeStampFromReal
!******************************************************************************!


!******************************************************************************!
  !> generate a unique time stamp from the information in time_type
  !!
  !! this basically generates a string identifying the solution time and
  !! writing in a meaningful format, so that it can be easily recognized
  !!
  function tem_generate_timeStampFromTimeType( fromTime  ) result( timeStamp )
    !---------------------------------------------------------------------------
    !> current timings
    type( tem_time_type ), intent(in) :: fromTime
    !> time stamp generated
    character(len=16)     :: timeStamp
    !---------------------------------------------------------------------------

    timestamp = tem_generate_timeStampFromReal( fromTime%rTime )

  end function  tem_generate_timeStampFromTimeType
!******************************************************************************!


!******************************************************************************!
  !> extract information about the timing from a timestamp
  !!
  !! It takes a string which was generated by the function above and reads the
  !! actual time out of it.
  !!
  function tem_extract_TimeFromtimeStamp( timeStamp ) result( toTime )
    !--------------------------------------------------------------------------
    !> time stamp to extract time from
    character(len=*), intent(in)     :: timeStamp
    !> time type to store time in
    type( tem_time_type ) :: toTime
    !--------------------------------------------------------------------------

    read( timestamp, *) toTime%rTime

  end function  tem_extract_TimeFromTimeStamp
!******************************************************************************!


!******************************************************************************!
  !> Reset the global time object
  !!
  subroutine tem_time_reset( me )
    !--------------------------------------------------------------------------
    !> time type to reset
    type(tem_time_type), intent(inout) :: me
    !--------------------------------------------------------------------------
    me%iTime = 0
    me%iStart= 0
    me%iStop = 0
    me%nIters= 0

    me%rTime = 0._rk
    me%rStart= 0._rk
    me%rStop = 0._rk

  end subroutine tem_time_reset
!******************************************************************************!


end module tem_time_module
!******************************************************************************!
