! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Timer infrastructure methods for the TreElM module
!!
!! This module provides a convenience timing functionality for coarse grain
!! measurements by encapsulating mpi_wtime and providing a set of timers to
!! track.
!!
module tem_timer_module

  ! include treelm modules
  use mpi
  use env_module,       only: rk
  use tem_tools_module, only: tem_write

  implicit none
  private

  public :: tem_addTimer
  public :: tem_startTimer
  public :: tem_stopTimer
  public :: tem_writeTimer
  public :: tem_writeTimers
  public :: tem_getTimerName
  public :: tem_getTimerVal
  public :: tem_timer_type
  public :: globalTimerHandle
  public :: timerData

  integer, parameter :: maxTimers = 64
  !> Timer handle for the overall run time
  integer :: globalTimerHandle

  type tem_timer_type
    !> number of used timers
    integer :: nTimers = 0

    !> Start timing values
    real(kind=rk) :: tStart(maxTimers) = 0._rk

    !>  timer running?
    logical(kind=rk) :: running(maxTimers) = .false.

    !> timing value
    real(kind=rk) :: value(maxTimers) = 0._rk

    !> Fluctuations in a timer value (parallel)
    real(kind=rk) :: fluc (maxTimers) = 0._rk

    !> timing object labels
    character(len=40) :: label(maxTimers)
  end type tem_timer_type

  type(tem_timer_type), save :: timerData
  ! A timer type for internal timing analyses, this will not be written to disk

  contains

!******************************************************************************!
  !> Setup a new timer object, reset the values and give it a label
  !! for later identification
  !!
  subroutine tem_addTimer( timerHandle, timerName, me )
    !---------------------------------------------------------------------------
    !> optional timer object
    type(tem_timer_type), optional, intent(inout) :: me
    !> Name for this timer
    character(len=*), intent(in) :: timerName
    !> A handle to reference this timer
    integer, intent(out) :: timerHandle
    !---------------------------------------------------------------------------

    if ( present(me) ) then
        if(me%nTimers .ge. maxTimers) then
          call tem_write( "Exceeded number of timer slots. stopping." )
          stop
        endif
        me%nTimers = me%nTimers + 1
        timerHandle = me%nTimers
        me%label(timerHandle) = timerName
    else ! timer data is written to disk
      if(timerData%nTimers .ge. maxTimers) then
        call tem_write( "Exceeded number of timer slots. stopping." )
        stop
      endif

      timerData%nTimers = timerData%nTimers + 1
      timerHandle = timerData%nTimers
      timerData%label(timerHandle) = timerName
      ! created a new timer handle and set the name
    end if

  end subroutine tem_addTimer
!******************************************************************************!


!******************************************************************************!
  !> Start the timer for the given timer handle
  !!
  subroutine tem_startTimer(timerHandle, me)
    !--------------------------------------------------------------------------
    !> optional timer object
    type(tem_timer_type), optional, intent(inout) :: me
    !> Handle of the timer to start
    integer, intent(in) :: timerHandle
    !--------------------------------------------------------------------------

    if ( present(me) ) then
      me%tStart(timerHandle) = mpi_wtime()
      me%running(timerHandle) = .true.
    else
      timerData%tStart(timerHandle) = mpi_wtime()
      timerData%running(timerHandle) = .true.
    end if

  end subroutine tem_startTimer
!******************************************************************************!


!******************************************************************************!
  !> Stop the timer for the given timer handle and update the timings
  !!
  subroutine tem_stopTimer(timerHandle, me )
    !---------------------------------------------------------------------------
    !> optional timer object
    type(tem_timer_type), optional, intent(inout) :: me
    !> Handle of the timer to stop
    integer, intent(in) :: timerHandle
    !---------------------------------------------------------------------------
    real(kind=rk) :: endTime
    !---------------------------------------------------------------------------

    if ( present(me) ) then
        !> Get the current time
        endTime = mpi_wtime()
        me%running(timerHandle) = .false.
        !> And add the measured difference to the last value
        me%value(timerHandle) =  me%value(timerHandle) + endTime               &
          &                   -  me%tStart(timerHandle)

    else
      !> Get the current time
      endTime = mpi_wtime()
      timerData%running(timerHandle) = .false.
      !> And add the measured difference to the last value
      timerData%value(timerHandle) = timerData%value(timerHandle) + endTime    &
        &                          - timerData%tStart(timerHandle)
    end if

  end subroutine tem_stopTimer
!******************************************************************************!


!******************************************************************************!
  !> Write out the timer name and its value
  !!
  subroutine tem_writeTimer( timerHandle, outunit )
    !--------------------------------------------------------------------------
    !> handle of the timer to be dumped
    integer, intent(in) :: timerHandle
    !> unit to write to
    integer, optional, intent(in) :: outunit
    !--------------------------------------------------------------------------
    integer :: loc_unit
    !--------------------------------------------------------------------------

    if( present( outunit ))then
      loc_unit = outunit
    else
      ! default is the screen output (unit 6)
      loc_unit = 6
    end if

    write(loc_unit,'(3a,f15.4,a)') ' Run-time ',                               &
      &                            trim(timerData%label(timerHandle)),         &
      &                            ':', timerData%value(timerHandle), ' s'

  end subroutine tem_writeTimer
!******************************************************************************!


!******************************************************************************!
  !> Write out all timers including name and values to the given outunit
  !!
  subroutine tem_writeTimers( outunit )
    !---------------------------------------------------------------------------
    !> unit to dump timers to
    integer, optional, intent(in) :: outunit
    !---------------------------------------------------------------------------
    integer :: loc_unit
    integer :: iTimer
    !---------------------------------------------------------------------------

    if( present( outunit ))then
      loc_unit = outunit
    else
      ! default is the screen output (unit 6)
      loc_unit = 6
    end if

    do iTimer = 1, timerData%nTimers
      call tem_writeTimer( iTimer, loc_unit )
    end do

  end subroutine tem_writeTimers
!******************************************************************************!


!******************************************************************************!
  !> Write out the timer name and its value
  !!
  function tem_getTimerName( timerHandle, me ) result( timerName )
    !---------------------------------------------------------------------------
    !> optional timer object 
    type(tem_timer_type), optional, intent(inout) :: me
    !> timer handle
    integer, intent(in) :: timerHandle
    !> name of the timer
    character(len=64) :: timerName
    !---------------------------------------------------------------------------
    if ( present( me) ) then
      timerName = trim( me%label(timerHandle) )
    else
      timerName = trim(timerData%label(timerHandle))
    end if

  end function tem_getTimerName
!******************************************************************************!


!******************************************************************************!
  !> Write out the timer name and its value
  !!
  function tem_getTimerVal( timerHandle, me ) result( retValue )
    !---------------------------------------------------------------------------
    !> optional timer object
    type(tem_timer_type), optional, intent(inout) :: me
    !> timer handle
    integer, intent(in) :: timerHandle
    !> timer value
    real(kind=rk) :: retValue
    !---------------------------------------------------------------------------
    real(kind=rk) :: endTime
    !---------------------------------------------------------------------------
    if( present(me) ) then
      if( .not. me%running(timerHandle) ) then
        retValue = me%value(timerHandle)
      else
        endTime = mpi_wtime()
        !> And add the measured difference to the last value
        retValue = endTime - me%tStart(timerHandle)
      end if
    else 
      if( .not. timerData%running(timerHandle) ) then
        retValue = timerData%value(timerHandle)
      else
        endTime = mpi_wtime()
        !> And add the measured difference to the last value
        retValue = endTime - timerData%tStart(timerHandle)
      end if
    end if

  end function tem_getTimerVal
!******************************************************************************!


end module tem_timer_module
!******************************************************************************!
