! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> A tracking output (e.g. for macroscopic values) for tracking certain flow
!! Quantities over a period of time at certain locations in the grid.
!!
!! See the \ref tracking_mod "detailed tracking module description".
!!
!! \author Manuel Hasert, Simon Zimny
!!
module tem_tracking_module

  ! incude treelm modules
  use env_module,                     only: rk, long_k, newunit, labelLen,     &
    &                                       pathLen, pathSep
  use treelmesh_module,               only: treelmesh_type
  use tem_aux_module,                 only: tem_abort, tem_unit_close
  use tem_comm_env_module,            only: process, tem_comm_env_type
  use tem_geometry_module,            only: tem_BaryOfId
  use tem_restart_module,             only: tem_restart_type, tem_init_restart,&
    &                                       tem_restart_finalize
  use tem_reduction_module,           only: tem_reduction_type,                &
    &                                       tem_load_reduction,                &
    &                                       tem_reduction_init,                &
    &                                       tem_reduction_toChunk
  use tem_transient_reduction_module, only: tem_transient_reduction_type,      &
    &                                       tem_load_transient_reduction,      &
    &                                       tem_transient_reduction_init
  use tem_shape_module,               only: tem_shape_type, tem_load_shape,    &
    &                                       tem_shape_out
  use tem_subTree_module,             only: tem_create_subTree_of
  use tem_subTree_type_module,        only: tem_subTree_type, tem_dump_subTree
  use tem_solveHead_module,           only: tem_solveHead_type, tem_solverTag
  use tem_time_module,                only: tem_generate_timeStamp,            &
    &                                       tem_time_type
  use tem_timeControl_module,         only: tem_timeControl_type,              &
    &                                       tem_load_timeControl,              &
    &                                       tem_timeControl_out
  use tem_var_system_module,          only: tem_var_system_type,               &
    &                                       tem_var_intersect,                 &
    &                                       tem_var_sys_load, tem_var_sys_out, &
    &                                       tem_writeSolverSpec
  use tem_tools_module,               only: tem_write, tem_horizontalSpacer,   &
    &                                       upper_to_lower
  use tem_topology_module,            only: tem_levelOf
  use tem_condition_module,           only: tem_convergence_type,              &
    &                                       tem_load_convergence

  ! include aotus modules
  use aotus_module,     only: flu_State
  use aot_table_module, only: aot_table_open, aot_table_close,                 &
    &                         aot_table_length, aot_get_val
  use aot_out_module,   only: aot_out_type, aot_out_val,                       &
    &                         aot_out_open, aot_out_close
  use aot_path_module,  only: aot_path_type, aot_path_addNode, assignment(=)

  implicit none

  private

  public :: tem_tracking_type
  public :: tem_trackingControl_type
  public :: tem_load_tracking
  public :: tem_init_tracker
  public :: tem_tracking_finalize
  public :: tem_tracking_header_Ascii
  public :: tem_tracking_close_AsciiTransient
  public :: tem_tracking_dumpAsciiTransient
  public :: tem_tracking_dumpAsciiSpatial
  public :: tem_transform_solSpecChar

  !> Parameter for harvester output
  integer, public, parameter :: tem_trackingHarvester      = 2
  !> Parameter for gnuplot ascii, writing a complete spatial
  !! representation at each time step into a new file
  integer, public, parameter :: tem_trackingAsciiSpatial   = 3
  !> parameter for ascii output, writing only one row consisting
  !! of all the elements with all variable values in that row
  !! into an initially opened file
  integer, public, parameter :: tem_trackingAsciiTransient = 4
  !> Parameter for convergence
  integer, public, parameter :: tem_trackingConvergence    = 5
  !> Parameter for precice spatial coupling
  integer, public, parameter :: tem_trackingPreciceSpatial = 6

  !> type for storing a list of elements to track and their values
  type tem_elemList_type
    !> Number of elements in the list
    integer :: nElems
    !> List of elements (actually the position of the
    !! element in glob%levelDesc( level )%total
    integer, allocatable :: elem(:)
  end type tem_elemList_type

  !> General information about the tracking entities
  type tem_trackingControl_type
    !> output status, activated?
    logical :: active   = .true.
    !> number of tracking entities
    integer :: nTrackings = 0
  end type tem_trackingControl_type

  !> Contains all header information about tracking
  type tem_trackingHeader_type
    !> log object labels
    character(len=labelLen)  :: label
    !> folder to store files to
    character(len=pathLen) :: prefix
    !> a unique tag for the solver including version
    character(len=LabelLen) :: solverTag
    !> from the user requested system of variables
    type(tem_var_system_type) :: requestedSys
    !> stores time control parameters
    type(tem_timeControl_type) :: timeControl
    !< simple gnuplot or partial tree mesh
    integer :: fileformat
    !> unit to write to the ascii file
    integer :: unit  = -1
    !> tracking shapes
    type(tem_shape_type), allocatable  :: geometry(:)
    !> type to check for convergence
    type(tem_convergence_type) :: convergence
  end type tem_trackingHeader_type

  !> Tracking entity definition
  type tem_tracking_type
    !> tracking header for collecting the properties from the lua file
    type(tem_trackingHeader_type) :: header
    !> Variable system definition, i.e. list of variables to track
    type(tem_var_system_type)  :: varSys
    !> map the requested variables to the offered ones
    integer,allocatable :: varMap(:)
    !> Restart type for saving in harvester format
    type(tem_restart_type)  :: restart
    !> sub-tree resulting from the elements within the tracking shape
    !! The sub-tree also holds the sub-communicator
    type(tem_subTree_type)    :: subTree
    !> Store the barycenters for the linearized tree elements
    !! it has a size of ( nElems, 3 )
    real(kind=rk), allocatable :: bary(:,:)
    !> store where the treeIDs of the sub-mesh are located in the levelwise
    !! arrays of levelDesc( iLevel )%total
    integer, allocatable    :: levelPointer(:)
    !> does only the root process of this tracking object have to write to disk?
    !! this is the case for spatial reduction where a transient ascii file is
    !! written
    logical :: onlyRootDumps
    !> Is there any reduction active inside this tracker?
    logical :: reduceActive
    !> reduction type which saves results from reduction
    type(tem_reduction_type), allocatable  :: reduce(:)
    !> is there any transient reduction active?
    !! Then this tracking must be updated every timestep!
    logical :: transientReduceActive
    !> transient reductions which collect and reduce data over several timesteps
    type(tem_transient_reduction_type), allocatable :: transientReduce(:)
  end type tem_tracking_type

  interface assignment(=)
    module procedure Copy_tracking
  end interface

  contains

!******************************************************************************!
  !> Read the tracker configuration from the main lua file
  !!
  !! Setup the values for the tracking entities
  !!
  subroutine tem_load_tracking( tCtrl, me, conf, parent, path, prefix,         &
    &                           refTimeControl )
    !---------------------------------------------------------------------------
    !> list of the tracking entities to create
    type( tem_tracking_type ), allocatable :: me(:)
    !> general control parameters
    type( tem_trackingControl_type ) :: tCtrl
    !> handle of the lua config file
    type( flu_state ) :: conf
    !> Reference time control object, which passes properties to the tracker's
    !! timeControl object (mainly for useIterations property)
    type( tem_timeControl_type ),intent(in), optional :: refTimeControl
    !> if the tracking table is a child-table of some other table,
    !! use the parent as a reference
    integer, optional :: parent
    !> lua path
    type( aot_path_type ),intent(in), optional :: path
    !> prefix for tracking label to differiate tracking for different scheme
    !! with same tracking label
    character(len=*), intent(in), optional :: prefix
    !---------------------------------------------------------------------------
    integer :: tc_handle, sub_handle
    integer :: iTrack, nTracks
    type( aot_path_type ) :: track_path, track_sub_path
    !---------------------------------------------------------------------------

    ! Generally assume that tracking is not defined
    tCtrl%active = .false.

    ! Read the number of trackings in the lua file
    call aot_table_open( L       = conf,                                       &
      &                  thandle = tc_handle,                                  &
      &                  key     = 'tracking',                                 &
      &                  parent  = parent)

    if (tc_handle .eq. 0) then
      call tem_write( 'No Tracking entities found!' )
      call aot_table_close(L=conf, thandle=tc_handle)
      call tem_horizontalSpacer()
      nTracks = 0
      tCtrl%nTrackings = nTracks
      allocate( me( nTracks ) )
      return
    else ! track entity exists.
      ! add this path to track path
      if( present( path ) ) then
        track_path = path
        call aot_path_addNode( me       = track_path,                          &
          &                    NodeType = 'table',                             &
          &                    key      = 'tracking' )
      end if
      tCtrl%active = .true.
    end if

    ! Check whether tracking had a subtable
    ! If no, then it is a single table, load single tracking entry
    ! else load multiple tables, open tracking subtable
    call aot_table_open( L       = conf,                                       &
      &                  parent  = tc_handle,                                  &
      &                  thandle = sub_handle,                                 &
      &                  pos     = 1 )

    ! Only single table
    if (sub_handle .eq. 0) then
      nTracks = 1
      tCtrl%nTrackings = nTracks
      call tem_write( 'Tracking is a single table' )
      allocate( me( nTracks ) )
      if( present( path ) ) then
        call tem_load_trackingHeader( conf           = conf,                   &
          &                           sub_handle     = tc_handle,              &
          &                           me             = me(1),                  &
          &                           prefix         = prefix,                 &
          &                           path           = track_path,             &
          &                           refTimeControl = refTimeControl )
      else
        call tem_load_trackingHeader( conf           = conf,                   &
          &                           sub_handle     = tc_handle,              &
          &                           me             = me(1),                  &
          &                           prefix         = prefix,                 &
          &                           refTimeControl = refTimeControl )
      end if
      call aot_table_close(L=conf, thandle=sub_handle)
    else ! Multiple table
      call aot_table_close(L=conf, thandle=sub_handle)
      nTracks = aot_table_length(L=conf, thandle=tc_handle)
      tCtrl%nTrackings = nTracks
      ! Allocate the defined number of tracking entities
      allocate( me( nTracks ))
      call tem_write( 'Number of Tracking entities: ', nTracks )

      ! Loop over all the definitions and assign the variables from the lua
      ! file on the tem_tracking_type.
      ! Inside this routine it will open tracking subtable. Each subtable 
      ! contains one or more tracking variables the stuff is done in the 
      ! routine tem_load_trackingHeader
      do iTrack = 1, nTracks
        call tem_write( 'Loading tracker ', iTrack )
        call aot_table_open( L       = conf,                                   &
          &                  parent  = tc_handle,                              &
          &                  thandle = sub_handle,                             &
          &                  pos     = iTrack)
        if( present( path )) then
          track_sub_path = track_path
          call aot_path_addNode( me       = track_sub_path,                    &
            &                    NodeType = 'table',                           &
            &                    pos      = iTrack )
          call tem_load_trackingHeader( conf           = conf,                 &
            &                           sub_handle     = sub_handle,           &
            &                           me             = me(iTrack),           &
            &                           path           = track_sub_path,       &
            &                           prefix         = prefix,               &
            &                           refTimeControl = refTimeControl )
        else
          call tem_load_trackingHeader( conf           = conf,                 &
            &                           sub_handle     = sub_handle,           &
            &                           me             = me(iTrack),           &
            &                           prefix         = prefix,               &
            &                           refTimeControl = refTimeControl )
        end if
        call aot_table_close(L=conf, thandle=sub_handle)
        call tem_write( 'Done tracker ', iTrack )
      end do
    end if ! sub_handle

    call aot_table_close(L=conf, thandle=tc_handle) ! close tracking table
    call tem_horizontalSpacer()

  end subroutine tem_load_tracking
!******************************************************************************!


!******************************************************************************!
  !> Read the tracker variables from tracking subtables defined in
  !! configuration from the main lua file
  !!
  !! If tracking is just a single table with single tracking entry
  !! then load only one tracking log exists with
  !! one or more variables using tem_load_trackingHeader_single.
  !! Else if tracking is table of many log then allocate log and load each 
  !! log type using tem_load_trackingHeader_single
  !! Setup the values for the tracking entities
  !!
  subroutine tem_load_trackingHeader( me, conf, sub_handle, path, prefix,      &
    &                                 refTimeControl )
    !---------------------------------------------------------------------------
    !> list of the tracking entities to create
    type( tem_tracking_type ), intent(out) :: me
    !> handle of the lua config file
    type( flu_state ) :: conf
    !> table sub-handle for the tracking table
    integer, intent(in) :: sub_handle
    !> lua path
    type( aot_path_type ),intent(in), optional :: path
    !> Reference time control object, which passes properties to the tracker's
    !! timeControl object (mainly for useIterations property)
    type( tem_timeControl_type ),intent(in), optional :: refTimeControl
    !> prefix for tracking label to differiate tracking for different scheme
    !! with same tracking label
    character(len=*), intent(in), optional :: prefix
    !---------------------------------------------------------------------------
    integer :: iError            ! error flag handle
    character(len=32) :: buffer  ! temporary variable for string treatment
    !---------------------------------------------------------------------------

    call aot_get_val( L       = conf,                                          &
      &               thandle = sub_handle,                                    &
      &               val     = me%header%label,                               &
      &               ErrCode = iError,                                        &
      &               key     = 'label',                                       &
      &               default = 'unnamed_track')
    if(present(prefix)) then
      if( trim(prefix) .ne. '' )                                               &
        &write(me%header%label, '(a)') trim(prefix)//'_'//trim(me%header%label)
    end if

    call tem_write('Tracking entity: '//trim( me%header%label ))

    ! Load single variable system
    ! \todo remove duplicate calls
    if( present( path )) then
      call tem_var_sys_load( conf    = conf,                                   &
        &                    sys     = me%header%requestedSys,                 &
        &                    thandle = sub_handle,                             &
        &                    path    = path,                                   &
        &                    sysName = trim(me%header%label)//'_sys' )
    else
      call tem_var_sys_load( conf    = conf,                                   &
        &                    sys     = me%header%requestedSys,                 &
        &                    thandle = sub_handle,                             &
        &                    sysName = trim(me%header%label)//'_sys' )
    end if

    ! load time control to output tracking
    call tem_load_timeControl( conf           = conf,                          &
      &                        parent         = sub_handle,                    &
      &                        me             = me%header%timeControl,         &
      &                        refTimeControl = refTimeControl )

    ! Where to store the tracking file?
    call aot_get_val( L       = conf,                                          &
      &               thandle = sub_handle,                                    &
      &               val     = me%header%prefix,                              &
      &               ErrCode = iError,                                        &
      &               key     = 'folder',                                      &
      &               default = '.'//pathSep)


    ! load tracking object shapes like point, line, plane
    call tem_load_shape( conf = conf, parent = sub_handle,                     &
                         me = me%header%geometry )

    if( size( me%header%geometry) .lt. 1) then
      call tem_write('The geometrical objects for the tracker are not '//      &
        &            'defined correctly.')
      call tem_abort()
    end if

    ! load reductions
    me%reduceActive = .false.
    call tem_load_reduction( conf    = conf,                                   &
      &                      parent  = sub_handle,                             &
      &                      me      = me%reduce )
    if( size( me%reduce ) .gt. 0 ) me%reduceActive = .true.

    if( me%reduceActive ) then
      ! Check if the number of reductions correspond to the number of variables
      ! in the system
      if( size( me%reduce) .ne. me%header%requestedSys%nVars ) then
        call tem_write('The number of defined reductions does not '//          &
          &            'correspond to the ')
        call tem_write('number of variables in the system. ')
        call tem_abort()
      end if
    end if

    ! load transient reductions
    me%transientReduceActive = .false.
    call tem_load_transient_reduction( conf    = conf,                         &
      &                                parent  = sub_handle,                   &
      &                                me      = me%transientReduce )
    if( size( me%transientReduce ) .gt. 0 ) me%transientReduceActive = .true.

    if( me%transientReduceActive ) then
      ! Check if the number of reductions correspond to the number of variables
      ! in the system
      if( size( me%transientReduce) .ne. me%header%requestedSys%nVars ) then
        call tem_write('The number of defined transient reductions does '//    &
          &            'not correspond to the ')
        call tem_write('number of variables in the system. ')
        call tem_abort()
      end if
    end if

    ! get tracking format
    call aot_get_val( L       = conf,                                          &
      &               thandle = sub_handle,                                    &
      &               val     = buffer,                                        &
      &               ErrCode = iError,                                        &
      &               key     = 'format',                                      &
      &               default = 'ascii' )
    buffer = upper_to_lower( buffer )
    if( me%reduceActive ) then
      if( trim(buffer) .ne. 'ascii' .and. trim(buffer) .ne. 'convergence') then
        call tem_write('Warning: ')
        call tem_write('Reduction is defined in current tracking object')
        call tem_write('Activating asciiTransient, as it is the only '//       &
          &            'format which')
        call tem_write('makes sense here.')
        buffer = 'ascii'
      end if
    end if

     ! Decide, which format is being written to disk
    select case( trim(buffer))
      case( 'ascii','asciitransient')
        me%header%fileformat = tem_trackingAsciiTransient
      case( 'asciispatial')
        ! write an ascii file for each time step and write all three barycenters
        ! each line has corrdinates and variables of only ONE element
        ! #       coordX  coordY  coordZ    var1       var2        var3
        me%header%fileformat = tem_trackingAsciiSpatial
      case( 'harvester','binary','hvs','harvest')
        ! ... in the harvester format, meaning
        ! there is a mesh in treelm format and corresponding to it
        ! there are restart files which hold the state information for the
        ! elements in the mesh for the requested time step
        me%header%fileformat = tem_trackingHarvester
      case( 'precice')
        ! ... in the harvester format, meaning
        ! there is a mesh in treelm format and corresponding to it
        ! there are restart files which hold the state information for the
        ! elements in the mesh for the requested time step
        me%header%fileformat = tem_trackingPreciceSpatial
      case( 'convergence' )
        ! if convergence is defined we load the condition table
        me%header%fileFormat = tem_trackingConvergence
        call tem_load_convergence( me      = me%header%convergence,            &
          &                        conf    = conf,                             &
          &                        parent  = sub_handle )
      case default
        call tem_write( '  Unknown Tracking Format specified! ' )
        call tem_write( '  Choose between ascii and harvester. Stopping!' )
        call tem_abort()
    end select

  end subroutine tem_load_trackingHeader
!******************************************************************************!


!******************************************************************************!
  !> Initialize the tracker entities
  !!
  !! Identify, how many and which elements exist on my local process and are
  !! requested from the trackers
  !! Empty tracking entities are removed, so the track(:) might be re-allocated
  !! \verbatim scheme = { ... tracking = { ... }} \endverbatim
  !!
  subroutine tem_init_tracker( me, tCtrl, levelPointer, tree, solver, varSys,  &
    &                          nDofs )
    !---------------------------------------------------------------------------
    !> tracking entities
    type(tem_tracking_type ),intent(inout),allocatable :: me(:)
    !> global tracking parameters
    type(tem_trackingControl_type ),intent(inout) :: tCtrl
    !> Global mesh from which the elements are identified and then stored to
    !! sub-meshes inside the trackers
    type(treelmesh_type), intent(in) :: tree
    !> Pointer from the global treeID list to the level-wise positions of the
    !! elements, optional due to harvester usage without levelwise information
    integer, optional, intent(in) :: levelPointer(:)
    !> Global solver information
    type(tem_solveHead_type ),intent(in) :: solver
    !> solver-provided variable systems
    type(tem_var_system_type ) :: varSys
    !> The number of dofs for each scalar variable of the equation system
    integer, intent(in), optional :: nDofs
    !---------------------------------------------------------------------------
    integer :: iLog, iElem, nElems, elemPos
    ! temporary tracker array
    type( tem_tracking_type ),allocatable :: tempTrack(:)
    integer :: nTracks
    integer :: tLevel
    integer :: iError
    logical :: writeTreeToDisk
    ! Define limit for max number of tracking entities in ascii file
    integer, parameter :: nElemLimit_ascii = 200
    ! temporary treeID
    integer(kind=long_k) :: tTreeID
    type(tem_var_system_type), allocatable :: tmpSys(:)
    integer :: currentNDofs
    ! store levelwise information
    logical :: levelwise
    !---------------------------------------------------------------------------

    if(present(nDofs)) then
      currentNDofs = nDofs
    else
      currentNDofs = 1
    end if

    ! fill levelwise information
    if( present( levelPointer ))then
      levelwise = .true.
    else
      levelwise = .false.
    end if

    call tem_horizontalSpacer
    call tem_write('Setting up the tracking infrastructure')
    call tem_horizontalSpacer
    nTracks = 0

    ! Allocate the temporary track
    allocate( tempTrack(size(me)) )
    allocate( tmpSys(1) )

    if( tCtrl%active ) then
      do iLog = 1, size(me)
        call tem_write( 'Initializing tracking object '//                      &
          &             trim( me( iLog )%header%label))
        ! map variables
        call tem_var_intersect( systemA = me( iLog )%header%requestedSys,      &
          &                     systemB = varSys,                              &
          &                     resSys  = me( iLog )%varSys,                   &
          &                     iError  = iError )
        if( iError .gt. 0 ) then
          call tem_write(' Intersection was not entirely completed. ')
          call tem_write(' Continuing with partial variables...     ')
        end if
        ! Store solver tag in trackingHeader
        me( iLog )%header%solverTag = tem_solverTag( solver )

        !\todo: SZ: check wether the st_fun is defined (maybe luaPath???) if
        !           yes use st_fun%tree else use global tree
        !           what about multiple variables in me(ilog)%varSys????
        !-----------------------------------------------------------------------
        ! identify tracker elements
        !-----------------------------------------------------------------------
        call tem_create_subTree_of( inTree = tree,                             &
          &                         subTree = me( iLog )%subTree,              &
          &                         inShape = me( iLog )%header%geometry,      &
          &                         prefix = trim(me(iLog)%header%prefix)//    &
          &                                  trim( me(iLog)%header%label )//'_')

        ! get rid of the empty track in order to avoid empty writes to disk
        if ( me( iLog )%subTree%nElems .gt. 0                                  &
          &                  .and. .not. me( iLog )%subTree%useGlobalMesh )then
          nTracks = nTracks + 1 !< Increase number of log

          ! Copy all entries from the log derived type to the temporary one.
          ! this might not work on all compilers!!
          tempTrack( nTracks ) = me( iLog )
          if( levelwise )then
            allocate( tempTrack( nTracks )%levelPointer(                       &
              &                                    me( iLog )%subTree%nElems ))
          end if

          if( levelwise )then
            ! Find the corresponding elements
            do iElem =1, me( iLog )%subTree%nElems
              ! get the position in the global tree (from subTree%map2global)
              elemPos = me( iLog )%subTree%map2global( iElem )
              ! Determine the level of the treeID
              tLevel = tem_levelOf( tree%treeID( elemPos ))
              ! Store the element position in the level-wise total list
              elemPos = levelPointer( elemPos )
              ! Create a reference treeID-sorted to the position in 
              ! levelDesc%total
              temptrack( nTracks )%levelPointer( iElem ) = elemPos
            end do
          end if
        else if( me( iLog )%subTree%useGlobalMesh )then
          nTracks = nTracks + 1
          tempTrack( nTracks ) = me( iLog )
          allocate( tempTrack( nTracks )%levelPointer( size( levelPointer )))
          temptrack( nTracks )%levelPointer = levelPointer
        end if
      end do  ! nTracks

      deallocate( me )
      allocate( me( nTracks ))


      do iLog = 1, nTracks
        ! Copy the stuff from the temporary track
        me( iLog ) = temptrack( iLog )

        !-----------------------------------------------------------------------
        ! Initialize restart header and communicator
        ! the tem_tracking_header_Ascii needs initialized restart type
        !-----------------------------------------------------------------------
        me( iLog )%restart%controller%writePrefix =                            &
          &     trim( me( iLog )%header%prefix )//                             &
          &     trim( me(iLog)%header%label )   //'_'

        ! this restart object is not meant to read data!
        me( iLog )%restart%controller%readRestart = .false.

        ! init the restart typed file format
        ! should be called when the first time the mesh is dumped
        tmpsys(1) = me( iLog )%varSys

        if ( me( iLog )%subTree%useGlobalMesh ) then
          call tem_init_restart( me     = me( iLog )%restart,                  &
            &                    solver = solver,                              &
            &                    nSys   = 1,                                   &
            &                    varSys = tmpsys,                              &
            &                    tree   = tree,                                &
            &                    nDofs  = currentNDofs )

          ! Assign barycenters
          nElems = tree%nElems
          allocate( me( iLog )%bary( nElems, 3 ) )
          do iElem = 1, nElems
            tTreeID = tree%treeID( iElem )
            me(iLog)%bary( iElem, : ) = tem_BaryOfId( tree, tTreeID )
          end do
        else ! not use global mesh
          call tem_init_restart( me      = me( iLog )%restart,                 &
            &                    solver  = solver,                             &
            &                    nSys    = 1,                                  &
            &                    varSys  = tmpsys,                             &
            &                    tree    = tree,                               &
            &                    subTree = me( iLog )%subTree,                 &
            &                    nDofs   = currentNDofs )

          ! Assign barycenters
          nElems = me( iLog )%subTree%nElems
          allocate( me( iLog )%bary( nElems, 3 ) )
          do iElem = 1, nElems
            tTreeID = tree%treeID( me( iLog )%subTree%map2global( iElem ))
            me(iLog)%bary( iElem, : ) = tem_BaryOfId( tree, tTreeID )
          end do
        end if ! useGlobalMesh

        ! Initialize reduction
        call tem_reduction_init( me( iLog )%reduce, me(iLog)%varSys )
        ! initialize transient reductions
        call tem_transient_reduction_init( me( iLog )%transientReduce,         &
          &                                nElems,                             &
          &                                me(iLog)%varSys,                    &
          &                                me( iLog )%header%timeControl%tMin )

        ! Is it only the root process of the tracking object who has to
        ! write to disk? this is (so far) only the case for spatial
        ! reductions
        me( iLog )%onlyRootDumps = me( iLog )%reduceActive

        ! Depending on the selected format, intitialize the files and headers
        select case( me( iLog )%header%fileformat )
        case( tem_trackingAsciiTransient )
          writeTreeToDisk = .false.
          ! Check if there are too many entries for an ascii line
          if( me( iLog )%subTree%nElems*me(iLog)%varSys%nVars                  &
            &                                .gt. nElemLimit_ascii             &
            &                         .and. .not. me( iLog )%reduceActive) then
            call tem_write('Error in tracking object ', iLog)
            call tem_write('There are n entities to track:',                   &
              &            me( iLog )%subTree%nElems * me(iLog)%varSys%nVars)
            call tem_write('Limit:                        ', nElemLimit_ascii )
            call tem_write('Error: Too many entries for the tracking entity '//&
              &            'as it is in ascii format')
            call tem_write('       All this information must be packed into '//&
              &            'one line.               ')
            call tem_write('Solution1: Use the Harvester format. It is way '// &
              &            'more efficient.           ')
            call tem_write('Solution2: reduce number of variables, number '//  &
              &            'of elements(segments)      ')
            call tem_abort
          end if
          if( (me( iLog )%onlyRootDumps                                        &
            &                  .and. me( iLog )%subTree%global%myPart .eq. 0 ) &
            &             .or. .not. me( iLog )%onlyRootDumps ) then
            ! open ascii file transient output and write first line to define
            ! datas in the column
            call tem_tracking_open_AsciiTransient( me       = me( iLog ),      &
              &                                    commGlob = process )
            ! Write the file headers
            call tem_tracking_header_Ascii( me       = me( iLog ),             &
              &                             commGlob = process,                &
              &                             solver   = solver,                 &
              &                             ssUnit   = varSys%solSpecUnit )
          end if

        case( tem_trackingAsciiSpatial )
          writeTreeToDisk = .false.
          ! Write the file headers
          call tem_tracking_header_Ascii( me       = me( iLog ),               &
            &                             commGlob = process,                  &
            &                             solver   = solver,                   &
            &                             ssUnit   = varSys%solSpecUnit )

        case( tem_trackingHarvester )
          ! This should only occur, if really elements are here, and so I must 
          ! be part of the communicator
          writeTreeToDisk = .not. me(iLog)%subTree%useGlobalMesh
          ! Prepare the solver specific information for the restart header file
        case( tem_trackingConvergence )
          writeTreetoDisk = .false.
        case default
          writeTreeToDisk = .false.
        end select

        if (writeTreeToDisk) then
          call tem_dump_subTree( me( iLog )%subTree, tree )
        end if
      end do
    end if ! if tracking active

    deallocate( temptrack )
    deallocate(tmpsys)

  end subroutine tem_init_tracker
!******************************************************************************!


!******************************************************************************!
  !> close the ascii transient output unit
  !!
  subroutine tem_tracking_close_AsciiTransient( me )
    !---------------------------------------------------------------------------
    !> tracker object
    type(tem_tracking_type ), intent(inout) :: me
    !---------------------------------------------------------------------------
    !local variable
    logical :: nUnitOpened
    !---------------------------------------------------------------------------

    ! check if tracking unit is open
    inquire( unit = me%header%unit, opened =  nUnitOpened )
    if ( nUnitOpened ) close( me%header%unit )

  end subroutine tem_tracking_close_AsciiTransient
!******************************************************************************!



!******************************************************************************!
  !> open the ascii transient output unit
  !!
  subroutine tem_tracking_open_AsciiTransient( me, commGlob )
    !---------------------------------------------------------------------------
    !> tracker object
    type(tem_tracking_type ), intent(inout) :: me
    !> Global communicator type for global rank information
    type(tem_comm_env_type ), intent(in) :: commGlob
    !---------------------------------------------------------------------------
    character(len=254) :: logName
    character(len=1024)  :: buffer
    !---------------------------------------------------------------------------


    if ( me%subTree%global%myPart .eq. 0 ) then
      write(buffer,'(i5.5)') commGlob%rank
       ! Include the rank name in the file name to avoid overwriting of files
       ! by different processes
      write(logName,'(a)') trim(me%header%prefix)//                            &
         &                 trim( me%header%label )//                           &
         &                 '_p'//trim(buffer)//'.res'

      me%header%unit = newunit()
      open( unit     = me%header%unit,                                         &
        &   file     = trim(logName),                                          &
        &   position = 'append',                                               &
        &   status   = 'replace' )

      write ( me%header%unit , '(a,i7)' ) '# Rank of the process: ',           &
        &                                 commGlob%rank
      write( buffer, '(a)' ) '#'
      write( buffer, '(a,a19)' ) trim(buffer),'time'
      call tem_tracking_getHeader( me, buffer )
      ! ... and write the header row to the file
      write ( me%header%unit , '(a)' ) trim(buffer)
    end if

  end subroutine tem_tracking_open_AsciiTransient
!******************************************************************************!


!******************************************************************************!
  !> Write the header of the ascii tracking files
  !!
  !! This writes a header with detailed information into the ascii file for the
  !! defined tracking object it writes into the unit, which was opened before 
  !! in \ref tem_init_tracker
  !!
  subroutine tem_tracking_header_Ascii( me, commGlob, solver, ssUnit )
    !---------------------------------------------------------------------------
    !> tracking object
    type(tem_tracking_type ), intent(inout) :: me
    !> Global communicator type for global rank information
    type(tem_comm_env_type ), intent(in) :: commGlob
    !> Global solver information
    type(tem_solveHead_type ),intent(in) :: solver
    !> The solver specific unit
    integer, intent(in)                  :: ssUnit
    !---------------------------------------------------------------------------
    character(len=1024) :: logName
    type(aot_out_type) :: conf !< aotus lua state to write output
    character(len=1024)  :: buffer
    !---------------------------------------------------------------------------

    write( buffer, '(a)' ) trim(me%header%prefix)//trim(me%header%label)
    write(logName,'(a)') trim(buffer)//'.lua'

    write( buffer, '(a)' ) trim(buffer)//'_p*'

    if ( me%subTree%global%myPart .eq. 0 ) then
      !! Write the header only on the root process
      !open up the mesh header lua file to dump the stuff using aotus library
      call aot_out_open( conf, logName )
      select case( me%header%fileformat)
        case( tem_trackingAsciiTransient )
          call aot_out_val( put_conf = conf,                                   &
            &               vname    = 'format',                               &
            &               val      = 'AsciiTransient' )
        case( tem_trackingAsciiSpatial )
          call aot_out_val( put_conf = conf,                                   &
            &               vname    = 'format',                               &
            &               val      = 'AsciiSpatial' )
      end select
      call aot_out_val( put_conf = conf,                                       &
        &               vname    = 'solver',                                   &
        &               val      = trim(me%header%solverTag) )
      call aot_out_val( put_conf = conf,                                       &
        &               vname    = 'simname',                                  &
        &               val      = trim(solver%simname) )
      call aot_out_val( put_conf = conf,                                       &
        &               vname    = 'label',                                    &
        &               val      = trim(me%header%label) )
      call aot_out_val( put_conf = conf,                                       &
        &               vname    = 'prefix',                                   &
        &               val      = trim(me%header%prefix) )
      call aot_out_val( put_conf = conf,                                       &
        &               vname    = 'rank',                                     &
        &               val      = commGlob%rank )
      call aot_out_val( put_conf = conf,                                       &
        &               vname    = 'nprocs',                                   &
        &               val      = commGlob%comm_size )
      call aot_out_val( put_conf = conf,                                       &
        &               vname    = 'resultfile',                               &
        &               val      = trim(buffer) )
      call aot_out_val( put_conf = conf,                                       &
        &               vname    = 'nElems',                                   &
        &               val      = me%subTree%global%nElems)
      ! write timeControl info
      call tem_timeControl_out( me%header%timeControl, conf )
      ! write shapes
      call tem_shape_out( me%header%geometry, conf )
      ! write variable systems
      if ( ssUnit .ge. 0 ) then
        call tem_var_sys_out( me%varSys, conf, ssUnit )
      else
        call tem_var_sys_out( me%varSys, conf )
      end if
!      ! write solve specific info is already in restart header
!      if ( ssUnit .ge. 0 ) &
!        & call tem_writeSolverSpec( outUnit = conf%outUnit, &
!        &                           ssUnit = ssUnit)
      ! close aotus file
      call aot_out_close( conf )
    end if

  end subroutine tem_tracking_header_Ascii
!******************************************************************************!


!******************************************************************************!
  !> Write single log for the right scheme into its ascii file.
  !!
  subroutine tem_tracking_dumpAsciiTransient( me, chunk, nChunkElems, timing)
    !---------------------------------------------------------------------------
    !> the tracking log to dump to disk
    type(tem_tracking_type ) :: me
    !> chunk to be dumped
    real(kind=rk)  :: chunk(:)
    !> how many elements are in the chunk
    integer, intent(in)   :: nChunkElems
    !> the current global time
    type(tem_time_type )  :: timing
    !---------------------------------------------------------------------------
    integer :: iElem, iScalar, nScalars
    character(len=4000) :: log_output !< output buffer
    !---------------------------------------------------------------------------

    if ( me%subTree%global%myPart .eq. 0 ) then
  !    if( size( me%reduce ) .gt. 0 ) then
  !      call tem_reduction_toChunk( me%reduce, chunk, nChunkElems )
  !    end if
      nScalars = me%varSys%nScalars

      ! First assemble the complete row consisting of the time ...
      write( log_output, '(e21.8)' ) timing%rTime
      ! and all the scalars entries of the variable systems for each elements
      do iElem = 1, nChunkElems
        do iScalar = 1, nScalars
          write( log_output, '(a,e21.10)' ) trim(log_output),                  &
            &                             chunk( (iElem-1)*nScalars + iScalar )
        end do
      end do
      ! ... then write into the open ascii file
      write ( me%header%unit , '(a)' ) trim(log_output)
    end if

  end subroutine tem_tracking_dumpAsciiTransient
!******************************************************************************!


!******************************************************************************!
  !> Write a spatial representation for elements into an ascii tracking file
  !!
  !! Each time this routine is called, a new file is written
  !! Filename: {tracking_folder}{tracking_label}_spatial_{proc}_{timestamp}.res
  !!     e.g.: tracking/lineProbe_spatial_00001_01_01378.1.res
  !! Each process writes its own files
  !!
  subroutine tem_tracking_dumpAsciiSpatial( me, commGlob, solver, chunk,       &
    &                                       nChunkElems, timing )
    !---------------------------------------------------------------------------
    !> tracking lgo to write to disk
    type(tem_tracking_type ), intent(inout) :: me
    !> solver information
    type(tem_solveHead_type ), intent(in) :: solver
    !> Global communicator type for global rank information
    type(tem_comm_env_type ), intent(in) :: commGlob
    !> data chunk to be dumped
    real(kind=rk)  :: chunk(:)
    !> number of elements in the chunk
    integer, intent(in)   :: nChunkElems
    !> current global time
    type(tem_time_type ), intent(in)  :: timing
    !---------------------------------------------------------------------------
    integer :: iElem
    integer :: iScalar, nScalars
    character(len=pathLen)  :: logName  ! ascii file name
    character(len=1024)  :: buffer
    character(len=labelLen) :: timestamp
    !---------------------------------------------------------------------------
    nScalars = me%varSys%nScalars
    ! Write the rank into the string buffer
    write(buffer,'(i5.5)') commGlob%rank

    timestamp = tem_generate_timeStamp( timing )

    ! Generate ascii file name
    write(logName,'(a)') trim(me%header%prefix)//trim( me%header%label )//     &
      &                  '_p'//trim(buffer)//'_t'//trim(timestamp)//'.res'
    me%header%unit = newunit()
    open( unit   = me%header%unit,                                             &
      &   file   = trim(logName),                                              &
      &   recl   = 1024,                                                       &
      &   status = 'replace' )

    ! in case a reduction is defined no coordinates will be printed
    if( size( me%reduce ) .gt. 0)then
      ! First write the header
      write ( me%header%unit , '(a,i7,a)' ) '# Rank of the process: ',         &
        &                                   commGlob%rank, ' reduction active'
      write( buffer, '(a)' ) '#'
    else
      write ( me%header%unit , '(a,i7)' ) '# Rank of the process: ',           &
        &                                 commGlob%rank
      write ( buffer, '(a)' ) '#         coordX          coordY          coordZ'
    end if
    call tem_tracking_getHeader( me, buffer )
    ! ... and write the header row to the file
    write ( me%header%unit , '(a)' ) trim(buffer)

    ! Then gather contents into buffer, and write buffer to file
    do iElem = 1, nChunkElems
      ! in case no reduction is defined write the barycenters else empty the
      ! buffer
      if( size( me%reduce ) .gt. 0)then
        write( buffer, '(a)' ) ''
      else
        write( buffer, '(3e16.8)' ) me%bary(iElem, 1:3)
      end if
      do iScalar = 1, nScalars
        write( buffer, '(a,e20.12)' ) trim(buffer),                            &
          &                               chunk( (iElem-1)*nScalars + iScalar )
      end do
      write ( me%header%unit , '(a)' ) trim(buffer)
    end do

    close( unit=me%header%unit)

  end subroutine tem_tracking_dumpAsciiSpatial
!******************************************************************************!


!******************************************************************************!
  !> This routine write a variable labels into buffer and this buffer is written
  !! into ascii result header for for spatial and transient.
  !!
  subroutine tem_tracking_getHeader( me, buffer )
    !---------------------------------------------------------------------------
    !> tracking lgo to write to disk
    type(tem_tracking_type ), intent(inout) :: me 
    !> buffer containing first column of the header
    character(len=*), intent(inout)  :: buffer 
    !---------------------------------------------------------------------------
    !local variable
    character(len=pathLen)  :: buffer2
    integer :: iVar, iComp
    character(len=4)  :: reduceChar
    !---------------------------------------------------------------------------

    if( size( me%reduce ) .gt. 0 ) then
      reduceChar = '_red'
    else
      reduceChar = ''
    end if
    do iVar = 1, me%varSys%nVars
      if ( me%varSys%Variable%val( iVar )%nComponents .gt. 1 ) then
        do iComp = 1, me%varSys%Variable%val( iVar )%nComponents
          write( buffer2, '(3a,i2.2)' )                                        &
            &                       trim(me%varSys%Variable%val( iVar )%label),&
            &                       trim(reduceChar),                          &
            &                       '_',iComp
          write( buffer, '(a,a20)' ) trim(buffer), trim(buffer2)
        end do
      else
        write(buffer, '(a,a20,a,a4)') trim(buffer),                            &
          &                         trim(me%varSys%Variable%val( iVar )%label),&
          &                         trim(reduceChar)
      end if
    end do

  end subroutine tem_tracking_getHeader
!******************************************************************************!


!******************************************************************************!
  !> Close all the units of the tracking objects
  !!
  subroutine tem_tracking_finalize( me )
    !---------------------------------------------------------------------------
    !> tracker object to close
    type(tem_tracking_type ), intent(inout) :: me(:)
    !---------------------------------------------------------------------------
    integer :: iTrack
    !---------------------------------------------------------------------------

    do iTrack = 1, size(me)
      ! Close solver specific units
      call tem_unit_close( me( iTrack )%header%unit )
      ! finalize the restart for harvester output
      if( me( iTrack )%header%fileformat .eq. tem_trackingHarvester )then
        call tem_restart_finalize( me( iTrack )%restart )
      end if
    end do

  end subroutine tem_tracking_finalize
!******************************************************************************!


!******************************************************************************!
  !> Linearize the information from file into a single string.
  !!
  subroutine tem_transform_solSpecChar( spc_unit, string )
    !---------------------------------------------------------------------------
    !> solver specific unit
    integer, intent(in) :: spc_unit
    !> linearized string
    character(len=*) :: string
    !---------------------------------------------------------------------------
    integer :: old_pos, length
    integer :: io
    logical :: nUnitOpened
    character(len=PathLen) :: loc_string
    !---------------------------------------------------------------------------
    string = ' '
    inquire(unit= spc_unit, opened=nUnitOpened)
    if (nUnitOpened) then
      ! Go to the start of the scratch file.
      rewind(spc_unit)
      old_pos = 0
      do
        ! Read all contents of the scratch file until eof is encountered
        read(spc_unit, '(a)', iostat=io) loc_string

        if ( io .eq. 0 ) then
          length = LEN_TRIM(loc_string)
          string( old_pos+1 : old_pos+length) = trim(loc_string)
          old_pos = old_pos+length
        else
          ! Reached end of the file, exit the loop.
          exit
        end if
      end do
    else
      call tem_write(' WARNING: no solver specific unit was opened before!')
      call tem_write('          Restart header file might be incomplete   ')
    end if

  end subroutine tem_transform_solSpecChar
!******************************************************************************!


!******************************************************************************!
  !> This function provides the assignment operator of two tracking.
  !! Temporary Solution as CRAY compiler dont have = Operator
  !! Copying a tracking Components into other tracking
  !!
  subroutine Copy_tracking(left,right)
    !---------------------------------------------------------------------------
    !> tracking to copy to
    type(tem_tracking_type), intent(out) :: left
    !> tracking to copy from
    type(tem_tracking_type), intent(in) :: right
    !---------------------------------------------------------------------------

    left%header = right%header
    left%varSys = right%varSys
    if(allocated(right%varMap)) then
      allocate(left%varMap(size(right%varMap)))
      left%varMap = right%varMap
    endif
    left%restart = right%restart
    left%subTree = right%subTree
    if(allocated(right%bary)) then
      allocate(left%bary(size(right%bary,1),size(right%bary,2)))
      left%bary=right%bary
    endif
    if(allocated(right%levelPointer)) then
      allocate(left%levelPointer(size(right%levelPointer)))
      left%levelPointer = right%levelPointer
    endif
    left%onlyRootDumps = right%onlyRootDumps
    left%reduceActive = right%reduceActive
    if(allocated(right%reduce)) then
      allocate(left%reduce(size(right%reduce)))
      left%reduce = right%reduce
    endif
    left%transientReduceActive = right%transientReduceActive
    if(allocated(right%transientReduce)) then
      allocate(left%transientReduce(size(right%transientReduce)))
      left%transientReduce = right%transientReduce
    endif  
  end subroutine Copy_tracking
!******************************************************************************!


end module tem_tracking_module
!******************************************************************************!
!> \page tracking_mod Tracking Module Description
!! \section tracking Tracking
!! 
!! During the runtime of simulations, the user needs to keep track of the 
!! simultation state without being obliged to write out the complete domain 
!! to disk.
!! This is achieved by tracking objects, at which the state is evaluated 
!! at certain defined intervals and then written to disk.
!! 
!! \subsection tracking_format Tracking format
!! 
!! We support three formats for tracking.
!! <em>Note: We strongly encourage you to use the Harvester format. </em>
!! Herefore you need to run Harvester on the simulation folder to create 
!! files for the Visualization tools.
!! - \ref tem_tracking_dumpasciitransient "ascii" - 
!! Transient data: simple gnuplot-style 
!! A simple ascii file which creates a small header and then writes at each 
!! defined time step a column wise representation of all variables and points. 
!! This is not efficient and should only be used for simple point-tracking 
!! objects.
!! Each process writes its own file.
!! - \ref tem_tracking_dumpasciispatial "asciiSpatial" - 
!! Spatial Data: This format stores solution data along with coordinates in 
!! single data file for every timestep. Therefore, there would be multiple 
!! files depending on the total number of time-steps. 
!! It is ideal for tracking *line* and *plane*.
!! 
!! - \ref tem_restart_module::restart_format "Harvester Format"
!! Writes all participating elements to a treelm mesh and creates for each 
!! required time step a restart file.
!! This is a scalable tracking object which works regardless of number of 
!! processes.
!! \verbatim format='harvester', folder = './tracking/'\endverbatim 
!! 
!! <em>Note: We will demonstrate each format in the examples of tracking point, 
!! line and plane in the following sections</em>
!! 
!! \subsection tracking_variabletypes Tracking variable types
!! 
!! - \ref tem_variable_module "variables": 
!!  all variables defined in the \ref tem_var_system_module "global variable 
!!                   system" can be tracked.
!!                   See \ref tem_variable_module::tem_variable_from 
!!                   "how to define variables"
!! - derived variables: all variables defined in the global variable system 
!!    can be tracked
!! - spacetime: as analytical result, can be defined in lua file
!! - \ref tem_reduction_module "reductions": 
!!     e.g. difference, sum or average of defined variables
!! 
!! \subsection tracking_shapes Tracking shapes
!! 
!! We show the tracking with simple shapes here. 
!! Get an overview over all existing shapes \ref tem_shape_module "here".
!! 
!! - zero-dimensional: point 
!! - one-dimensional: line 
!! - two-dimensional: plane
!! - three-dimensional: cubes, cylinders
!! 
!! So far point, line, plane are implemented in \ref tem_shape_module "shapes".
!! 
!! \subsection tracking_usage Usage
!! 
!! The user specifies the desired parameters in the configuration file 
!! (in the scheme table) as
!! \verbatim 
!! -- if you want to have a analytical solution as reference, you can add 
!! --  a spacetime variable first
!! -- define single new (spacetime) variable
!! add_variable = { name='spacetime', ncomponents=1, spacetime=luaFun }
!! -- define multiple new (spacetime) variables
!! add_variable = {                                                   
!! {name='spacetime', ncomponents=1, spacetime=luaFun },
!! {name='spacetime2', ncomponents=1, spacetime=luaFun }
!! }
!! 
!! -- Tracking
!! tracking = {
!! { label = 'probe_error',
!! variable = {
!!   { 'velocity' },
!!   { name='difference', 
!!     ncomponents=1, 
!!     dep={'density','spacetime'}},
!!   { name='difference', 
!!     ncomponents=1, 
!!     dep={'wss','spacetime2'}}},
!! folder = './tracking/',
!! format = 'ascii',
!! time = {interval = 1, min = 1000, max = 4000 },
!! reduction = {'l2norm'},  
!! shape = { kind = 'canoND', 
!!           object = { origin ={0.,0.,0.},
!!                      vec = {1.,0.,0.}, 
!!                      segments={100}} }
!! } 
!! \endverbatim 
!! Here, a point at the location {0.,0.,0.} is defined as a pressure probe. 
!! The interval is set to 1, meaning at each time step, the pressure has to 
!! be written.
!! The probing starts at iteration 1000 and lasts until 4000.
!! The tracking write both pressure and velocityX in a single file.
!! Use ref_value to normalize the output.
!! \verbatim 
!! # Name:       GaussPulse_TestSimulation     Date 2011-060-27  Time 11:43 CEST
!! # Probe type: Point at {0.0 0.0 0.0} 
!! # Quantities: pressure, velocityX
!! # tMin:       1000,     tMax:    4000,     Interval: 1
!! # ref_value: 1.0, 0.01
!! #   iTime     rTime      Pressure           velocityX
!! 1000          1000.0     1.0241818          0.00158287
!! 1001          1001.0     1.0251427          0.00152617
!! ...                                      
!! 4000          4000.0     1.7162511          0.01273673
!! \endverbatim 
!! 
!! The shape is defined as a canonical object. 
!! The different shapes are basically derivatives of a simple cube. 
!! A cube with no length in any direction is a point, a cube with length along 
!! only one axis is a line whereas a cube with length along two axes is simply 
!! treated as plane by the underlying code. 
!! One example each for Point, Line and Plane is shown below.
!! 
!! \subsection tracking_point Point Tracking
!! 
!! <em>Example: tracking a point in ASCII format</em>
!! 
!! \verbatim               
!! tracking = { label = 'Velocity', 
!! variable = {'velocity'},   -- options: density, velocity
!! 
!! --   Tracking over a point
!! shape={kind = 'canoND', origin = {0.5,0.5,0.5} },
!! 
!! time = {min = 0, max = -1, interval = 1},     -- max = -1 assigns max=tmax
!! format = 'ascii',
!! folder = 'tracking/'}
!! }
!! \endverbatim 
!! 
!! Once, the above section is executed within the musubi.lua file, an output in 
!! the asciiTransient format is generated in the output folder specified 
!! (tracking in this particular example). 
!! Now, we open the result file inside tracking folder, the structure of which 
!! looks like below:
!! 
!! \verbatim 
!! # Rank of the process:       0
!! #               time     velocity_01     velocity_02     velocity_03
!! 1.0000  0.000000000000  0.000000000000  0.000000000000
!! 2.0000  0.000000000000  0.000000000000  0.000000000000
!! 3.0000  0.000000000000  0.000000000000  0.000000000000
!! 4.0000  0.000000000000  0.000000000000  0.000000000000
!! ...
!! ...
!! 1997.0000 -0.006065558329  0.000443400622 -0.000003481789
!! 1998.0000 -0.006063901533  0.000443795703 -0.000003286949
!! 1999.0000 -0.006061596302  0.000443327890 -0.000003534752
!! 2000.0000 -0.006059665212  0.000443720604 -0.000003339771
!! \endverbatim 
!! Here, it can be noticed that 3 velocity components for all time steps at the 
!! point (0.5, 0.5, 0.5) are stored. The interval was chosen as 1, so they are 
!! tracked for each time step. If the interval was chosen as say 10, then the 
!! tracking would have been done after 10 timesteps.
!! 
!! 
!! \subsection tracking_line Line Tracking
!! 
!! We need to set the origin and the direction vector, 
!! as e.g. {origin={0.,0.,0.}, vec={1.0,1.0,1.0}}.
!! The line is represented as a list of points in the tracking facility. 
!! The number of subdivisions and its distribution of the points on the line 
!! needs to be set, as e.g. @segments=500, distribution='equal'@. 
!! Make sure to specify enough segments to take into account all the required 
!! elements on the line. 
!! Elements are added only once, so it might be beneficial to specify much more 
!! segments than actually needed to consider elements on the highest levels as 
!! well.
!! 
!! <em>Example: Tracking a line in ASCII Spatial format</em>
!! 
!! \verbatim tracking = { 
!!   label = 'line_probe',
!!   variable = {'state'}, -- state (=pdfs), pressure, density etc
!!   shape = { kind = 'canoND', 
!!             object={
!!               origin = {0.5,0.5,0.5}, 
!!               vec = { {0.0,0.2,0.0}},
!!               segments = {300},
!!               distribution='equal'} 
!!           },
!!   format='asciispatial', 
!!   folder='tracking/',
!!   time = {interval = 5, min = 0, max = 800 } }
!! }\endverbatim 
!! 
!! After, the above tracking is executed within musubi.lua file multiple res 
!! files are obtained inside the tracking folder each specific to particular 
!! time step. If you open any one of the @res@ files, there will be six data 
!! columns: three coordinate and three velocity components. 
!! An example is shown below: 
!! \verbatim 
!! # Rank of the process:       0
!! #         coordX          coordY          coordZ     velocity_01     velocity_02     velocity_03
!! 0.51562500      0.51562500      0.51562500 -0.000998448727 -0.000118182367  0.000007560959
!! 0.51562500      0.54687500      0.51562500 -0.000922746216  0.000011023243 -0.000000130462
!! 0.51562500      0.57812500      0.51562500 -0.001083956104  0.000038534363 -0.000003774649
!! ...
!! 0.51562500      0.64062500      0.51562500 -0.001343981219 -0.000131107478 -0.000004381864
!! 0.51562500      0.67187500      0.51562500 -0.001261956611  0.000081027056  0.000000215364
!! 0.51562500      0.70312500      0.51562500 -0.001733303068  0.000060154818  0.000005793207
!! \endverbatim 
!! 
!! Out of three coordinate components, only 'coordY' is varying and 'coordX', 
!! 'coordZ' remain constant. This is according to our tracking input with 
!! @origin = {0.5, 0.5, 0.5}, vec = {0.0, 0.2, 0.0}@.
!! You can now make a line plot with any one of velocity components with 
!! respect to @coordY@ in "gnuplot":http://www.gnuplot.info/.
!! 
!! \subsection tracking_plane Plane Tracking
!! 
!! We need to set the origin and the direction vectors, as e.g. 
!! object = {origin={0.,0.,0.}, vec={{1.0,1.0,1.0},{0.0,-1.0,1.0}}}.
!! The plane is, just as the line, represented as a list of points in the 
!! tracking facility. 
!! The number of subdivisions in each direction and its distribution of the 
!! points on the plane needs to be set, 
!! as e.g. @segments1=500, segments2=300, distribution='equal'@. 
!! Make sure to specify enough segments to take into account all the required 
!! elements on the line. 
!! Elements are added only once, so it might be beneficial to specify much more 
!! segments than actually needed to consider elements on the highest levels 
!! as well.
!! 
!! <em> Example: tracking a plane in harvester format</em>
!! 
!! \verbatim tracking = { 
!!   label = 'plane_probe',
!!   variable = {'state'}, -- state (=pdfs), pressure, density etc
!!   shape = { kind = 'canoND', 
!!             object={
!!               origin = {-75.,-15.,0.}, 
!!               vec = { {150.,0,0.}, {0.,30,0.}},
!!               segments = {300, 300},
!!               distribution='equal'} 
!!           },
!!   format='harvester', 
!!   folder='tracking/',
!!   time = {interval = 5, min = 0, max = 800 } }
!! }\endverbatim 
!! 
!! \subsection tracking_global_mesh Global mesh tracking
!! 
!! <em> Example: tracking the whole mesh</em>
!! 
!! \verbatim tracking = { 
!!   ...
!!   shape = { kind = 'global' },
!!   ...
!! }\endverbatim 
!! 
!! \subsection tracking_boundary Boundary elements tracking
!! 
!! <em> Example: tracking the boundary elements</em>
!! 
!! \verbatim tracking = { 
!!   ...
!!   shape = { kind = 'property', property = {'boundary'} },
!!   ...
!! }\endverbatim 
!! 
!! \subsection tracking_design Design and Implementation
!! 
!! The user-specified tracking shapes have to be translated to a set of
!! elements, on which the defined quantities are calculated and written to disk
!! in a suitable format.
!! 
!! All elements are converted to a set of points.
!! For each point, the correspoinding treeID is added to the element list of
!! the tracker as \verbatim track%elems( iLevel )%elem(1:nElems)\endverbatim 
!! Each element will only be added once.
!! All these elements are then treated in the main loop within the @tracker@
!! routine.
