! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module gathers different transient effect treatments.
!! like: constant, lua function and predefined Fortran function. \n
!! Currently supports:
!! \li const
!! \li lua_fun
!! \li linear
!! \li smooth
!! Remember boundaries can be defined with superposed conditions in
!! space and time. Therefore the function defined here acts as a
!! factor applied to the overall boundary value.
!!
module tem_transient_module

  ! include treelm modules
  use env_module,       only: rk, LabelLen
  use tem_aux_module,   only: tem_abort
  use tem_tools_module, only: tem_write
  use tem_param_module, only: PI
  use tem_time_module,  only: tem_time_type

  ! include aotus modules
  use aot_path_module,  only: aot_path_type, assignment(=), aot_path_addNode,  &
    &                         aot_path_open, aot_path_close
  use aotus_module,     only: aoterr_Fatal, aoterr_NonExistent, flu_State,     &
    &                         aoterr_WrongType, aot_top_get_val, aot_get_val
  use aot_table_module, only: aot_table_open, aot_table_close
  use aot_fun_module,   only: aot_fun_type, aot_fun_open, aot_fun_close,       &
    &                         aot_fun_put, aot_fun_do

  implicit none

  private

  public :: tem_transient_type
  public :: load_transient, tem_transient_for

  !> contains information for predefined transient functions
  type tem_predef_transient_type
    !> start time
    real(kind=rk) :: from_time
    !> end time
    real(kind=rk) :: to_time
    !> minimum factor at start time
    real(kind=rk) :: min_factor
    !> maximum factor at end time
    real(kind=rk) :: max_factor
  end type tem_predef_transient_type

  !> defines different transient types like const, lua func or predefined func
  type tem_transient_type
    !> transient kind
    character(len=LabelLen) :: kind
    !> constant value
    real(kind=rk) :: const
    !> stores path to lua function
    type( aot_path_type ) :: LuaPath
    !> contains informaion for predefined functions
    type( tem_predef_transient_type ) :: predef_fun
  end type tem_transient_type

  contains

!******************************************************************************!
  !> This subroutine load transient table defined for a boundary.\n
  !!
  !! If transient is defined as lua function then set kind = transient_lua
  !! else if transient block is defined then load transient table for predefined
  !! Fortran function variables and set kind = transient_\a function_name
  !! else transient is a constant value and set kind = transient_const.\n
  !! \n
  !! Valid definitions:
  !! \li Constant
  !! \verbatim
  !! transient = 1.0
  !! \endverbatim
  !! \li lua_function
  !! \verbatim
  !! transient = 'linear'
  !! \endverbatim
  !! Example: \a linear lua function
  !! \verbatim
  !!function linear(iTime)
  !!  local to_time = 1000
  !!  if iTime < to_time then
  !!    return iTime/to_time
  !!  else
  !!    return 1.0
  !!  end
  !!end
  !! \endverbatim
  !! \li Predefined Fortran function
  !! \verbatim
  !! transient = {predefined="Fortranfun_name",
  !!              min_factor = 0.0, max_factor = 1.0,
  !!              from_time = 0.0, to_time = 1000.0}
  !! \endverbatim
  !! \n
  subroutine load_transient( me, conf, parent, path, key )
    !---------------------------------------------------------------------------
    !> boundary transient type
    type(tem_transient_type), intent(out) :: me
    !> lua state
    type(flu_State) :: conf
    !> parent handle contains transient table
    integer, intent(in) :: parent
    !> lua path to transient lua functin
    type(aot_path_type), intent(in) :: path
    !> state variable key string defined in lua
    character(len=*), intent(in), optional :: key
    !---------------------------------------------------------------------------
    type(aot_fun_type) :: fun
    integer :: thandle
    integer :: iError
    character(len=labelLen) :: local_key
    !---------------------------------------------------------------------------

    if(present(key)) then
      local_key = trim(key)
    else
      local_key = 'transient'
    endif

    ! First test for a lua function
    call aot_fun_open( L      = conf,                                          &
      &                parent = parent,                                        &
      &                fun    = fun,                                           &
      &                key    = trim( local_key ))

    if (fun%handle /= 0) then
      ! This transient modifier is defined as a Lua function
      me%kind = 'lua_fun'
      me%LuaPath = path
      call tem_write('    Defined lua transient function')
      call aot_path_addNode( me       = me%LuaPath,                            &
        &                    NodeType = 'function',                            &
        &                    key      = trim( local_key ))
    else
      ! It is not defined as a function, try to interpret it as a table
      call aot_table_open( L       = conf,                                     &
        &                  thandle = thandle,                                  &
        &                  parent  = parent,                                   &
        &                  key     = trim( local_key ))
      if (thandle /= 0) then
        call aot_get_val( L       = conf,                                      &
          &               thandle = thandle,                                   &
          &               key     = 'predefined',                              &
          &               val     = me%kind,                                   &
          &               default = 'unknown',                                 &
          &               ErrCode = iError)

        call tem_write('    Defined predefined transient function '//          &
          &            '"'//trim(me%kind)//'"')
        select case(trim(me%kind))
        case('linear', 'smooth')
          ! Load the standard parameters necessary to describe the
          ! transient behavior of the BC.
          call load_transient_predef_fun( me      = me%predef_fun,             &
            &                             conf    = conf,                      &
            &                             thandle = thandle )
        case default
          call tem_write('ERROR in definition of the transient '//             &
            &            ' boundary Conditions:')
          call tem_write('Selected an unknown transient boundary')
          call tem_write(trim(me%kind))
          call tem_abort()
        end select
      else
        ! As the entry for the variable is neither a function nor a table, try
        ! to interpret it as a constant variable
        call aot_get_val( L       = conf,                                      &
          &               thandle = parent,                                    &
          &               key     = trim(local_key),                           &
          &               val     = me%const,                                  &
          &               ErrCode = iError )
        if (btest(iError, aoterr_WrongType)) then
          call tem_write('FATAL Error occured in definition of the transient'//&
            &            ' boundary conditions')
          call tem_write('while retrieving transient constant:')
          call tem_write('Variable has wrong type (should be a real number)!')
          call tem_write('STOPPING')
          call tem_abort()
        end if
        if (btest(iError, aoterr_NonExistent)) then
          ! "transient" variable not specified
          me%kind = 'none'
          me%const = 1.0_rk
        else
          me%kind = 'const'
        end if
      end if
    end if

  end subroutine load_transient
!******************************************************************************!


!******************************************************************************!
  !> This subroutine load standard transient function variables from LUA file.
  !!
  !! Default values: 
  !! - min_factor = 0.0
  !! - max_factor = 1.0
  !! - from_time = 0
  !! - to_time = tmax/2
  !! 
  !! Valid definition:
  !! - linear function
  !! \verbatim
  !! transient = {predefined="linear", min_factor = 0.0, max_factor = 1.0,
  !! from_time = 0, to_time = 1000}
  !! \endverbatim
  !! Example: Transient inlet velocity which starts from 0 to 1000 with maximum
  !! value 0.08 is shown below for linear and smooth with definition and image.
  !! \verbatim
  !! boundary_condition =
  !!         {
  !!           { label = 'inlet',
  !!             kind = 'inlet_ubb',
  !!             velocityX = {ref_value = 0.08
  !!                ,transient = {predefined="linear", min_factor = 0.0,
  !!                      max_factor = 1.0, from_time = 0, to_time = 1000}}
  !!--                ,transient = {predefined="smooth", min_factor = 0.0,
  !!--                      max_factor = 1.0, from_time = 0, to_time = 1000}}
  !!             velocityY = 0.0
  !!             velocityZ = 0.0
  !!           }
  !!         }
  !! \endverbatim
  !! \image html transient.png
  !! 
  subroutine load_transient_predef_fun( me, conf, thandle )
    !---------------------------------------------------------------------------
    !> transient predefined fun type
    type(tem_predef_transient_type),intent(inout) :: me
    !> lua state type
    type(flu_State) :: conf
    !> aotus parent handle
    integer, intent(in) :: thandle
    !---------------------------------------------------------------------------
    ! local variables
    integer :: iError
    !---------------------------------------------------------------------------

    !min_factor
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               key     = 'min_factor',                                  &
      &               val     = me%min_factor,                                 &
      &               ErrCode = iError,                                        &
      &               default = 0.0_rk )
    if (btest(iError, aoterr_Fatal)) then
      call tem_write('FATAL Error occured, while retrieving min_factor :')
      if (btest(iError, aoterr_NonExistent))                                   &
        & call tem_write('Variable not existent!')
      if (btest(iError, aoterr_WrongType))                                     &
        & call tem_write('Variable has wrong type!')
      call tem_write('STOPPING')
      call tem_abort()
    end if

    !max_factor
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               key     = 'max_factor',                                  &
      &               val     = me%max_factor,                                 &
      &               ErrCode = iError,                                        &
      &               default = 1.0_rk )
    if (btest(iError, aoterr_Fatal)) then
      call tem_write('FATAL Error occured, while retrieving max_factor :')
      if (btest(iError, aoterr_NonExistent))                                   &
        & call tem_write('Variable not existent!')
      if (btest(iError, aoterr_WrongType))                                     &
        & call tem_write('Variable has wrong type!')
      call tem_write('STOPPING')
      call tem_abort()
    end if

    !from_time
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               key     = 'from_time',                                   &
      &               val     = me%from_time,                                  &
      &               ErrCode = iError,                                        &
      &               default = 0.0_rk )
    if (btest(iError, aoterr_Fatal)) then
      call tem_write('FATAL Error occured, while retrieving from_time :')
      if (btest(iError, aoterr_NonExistent))                                   &
        & call tem_write('Variable not existent!')
      if (btest(iError, aoterr_WrongType))                                     &
        & call tem_write('Variable has wrong type!')
      call tem_write('STOPPING')
      call tem_abort()
    end if

    ! to_time
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               key     = 'to_time',                                     &
      &               val     = me%to_time,                                    &
      &               ErrCode = iError,                                        &
      &               default = 1.0_rk )
    if (btest(iError, aoterr_Fatal)) then
      call tem_write('FATAL Error occured, while retrieving to_time :')
      if (btest(iError, aoterr_NonExistent))                                   &
        & call tem_write('Variable not existent!')
      if (btest(iError, aoterr_WrongType))                                     &
        & call tem_write('Variable has wrong type!')
      call tem_write('STOPPING')
      call tem_abort()
    end if

  end subroutine load_transient_predef_fun
!******************************************************************************!


!******************************************************************************!
  !> This function returns value of linear function which is defined by
  !! from_time, to_time, min_factor and max_factor
  !!
  function transient_linear_for( me, time ) result(res)
    !---------------------------------------------------------------------------
    !> transient linear type
    type(tem_predef_transient_type) :: me
    !> timer object incl. the current time information
    type(tem_time_type), intent( in )  :: time
    !> return value of a function
    real(kind=rk) :: res
    !---------------------------------------------------------------------------

    if (time%rTime .lt. me%from_time) then
      res = me%min_factor
    else if (time%rTime .gt. me%to_time) then
      res = me%max_factor
    else
      res = me%min_factor + (me%max_factor - me%min_factor)                    &
        & * (  (time%rTime - me%from_time) / (me%to_time - me%from_time) )
    end if
  end function transient_linear_for
!******************************************************************************!


!******************************************************************************!
  !> This function returns value of smooth sin function which is defined by
  !! from_time, to_time, min_factor and max_factor
  !!
  function transient_smooth_for( me, time ) result(res)
    !---------------------------------------------------------------------------
    !> transient smooth type
    type(tem_predef_transient_type) :: me
    !> timer object incl. the current time information
    type(tem_time_type), intent( in )  :: time
    !> return value of a function
    real(kind=rk) :: res
    !---------------------------------------------------------------------------

    if (time%rTime .lt. me%from_time) then
      res = me%min_factor
    else if (time%rTime .gt. me%to_time) then
      res = me%max_factor
    else
      res = me%min_factor                                                      &
        & + (me%max_factor - me%min_factor)                                    &
        & * sin( PI * 0.5_rk * (time%rTime - me%from_time)                     &
        &                    / (me%to_time - me%from_time))
    end if
  end function transient_smooth_for
!******************************************************************************!


!******************************************************************************!
  !> This function evaluate lua function and return its result
  !!
  function transient_lua_for(me, time, conf) result(res)
    !---------------------------------------------------------------------------
    !> lua path to transient lua function
    type(aot_path_type) :: me
    !> timer object incl. the current time information
    type(tem_time_type), intent( in )  :: time
    !> optional lua state
    type(flu_State), optional :: conf
    !> return value
    real(kind=rk) :: res
    !---------------------------------------------------------------------------
    ! local variables
    type(aot_fun_type) :: fun
    integer :: iError
    type(flu_State) :: loc_conf
    logical :: openLua
    !---------------------------------------------------------------------------

    if (present(conf)) then
      loc_conf = conf
      openLua = .false.
    else
      openLua = .true.
    end if

    call aot_path_open( me, conf=loc_conf, fun=fun, openLua=openLua )

    call aot_fun_put( L=loc_conf, fun=fun, arg=time%rTime )
    call aot_fun_do( L=loc_conf, fun=fun, nresults=1 )
    call aot_top_get_val( L=loc_conf, val=res, ErrCode=iError )

    call aot_path_close( me, conf=loc_conf, fun=fun, closeLua=openLua )

  end function transient_lua_for
!******************************************************************************!


!******************************************************************************!
  !> This function invokes the type of the boundary such as constant, lua or
  !! predefined Fortran function.
  !!
  !! If transient block is not defined than it return value = 1.0.
  function tem_transient_for( transient, time, conf ) result( res )
    !---------------------------------------------------------------------------
    !> boundary state
    type( tem_transient_type ) :: transient
    !> timer object incl. the current time information
    type(tem_time_type), intent( in )  :: time
    !> lua state
    type(flu_State), intent(in), optional :: conf
    !> return value of a function
    real(kind=rk) :: res
    !---------------------------------------------------------------------------

    select case( trim(adjustl(transient%kind)) )
    case( 'none' )
      res = transient%const
    case( 'const' )
      res = transient%const
    case( 'lua_fun' )
      res = transient_lua_for( transient%LuaPath, time, conf )
    case( 'linear' )
      res = transient_linear_for( transient%predef_fun, time )
    case( 'smooth' )
      res = transient_smooth_for( transient%predef_fun, time )
    end select

   end function tem_transient_for
!******************************************************************************!


end module tem_transient_module
!******************************************************************************!
