! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!>\brief This module reduces values in time
!! \author Melven Zoellner
!! Module takes data from all timesteps and different cells
!! (one after another, but may store data of previous timesteps)
!! and reduces the the data to one value per cell for each variable.
!! Currently the only implemented operations are average, min, max
!! and sum.
!! Desired operation should be specified as\verbatim
!! transient_reduction = {
!!                         'min', 'max', 'sum' or 'average', 'none'
!!                       } \endverbatim
!!
module tem_transient_reduction_module

  ! include treelm modules
  use env_module,            only: rk, labelLen
  use tem_aux_module,        only: tem_abort
  use tem_tools_module,      only: tem_write
  use tem_var_system_module, only: tem_var_system_type
  use tem_logging_module,    only: tem_log, tem_toStr, logUnit

  ! include aotus modules
  use aotus_module,     only: aot_get_val, flu_State
  use aot_table_module, only: aot_table_open, aot_table_close, aot_table_length
  implicit none

  private

  public :: tem_transient_reduction_type
  public :: tem_load_transient_reduction
  public :: tem_transient_reduction_init
  public :: tem_transient_reduction_open
  public :: tem_transient_reduction_apply
  public :: tem_transient_reduction_close
  public :: tem_transient_reduction_reset


  !> all data needed for a transient reduction,
  !! operation to perform and necessary data from previous timesteps
  type tem_transient_reduction_type
    !> Which operation to perform
    character(len=labelLen) :: operation = ''
    !> number of elements
    integer :: nElems = 0
    !> number of components
    integer :: nComponents = 0
    !> number of "recorded" previous timesteps
    integer :: nTimes = 0
    !> index of current timestep entry (so the following arrays
    !! may be addressed in a circular way in time)
    integer :: iTime = 0
    !> array to store data from previous timesteps
    !! size (max nTimes, nComponents*nElems)
    real(kind=rk), allocatable :: val(:,:)
    !> array to store time at previous (and current) timestep
    !! size: (0:max nTimes+1)
    real(kind=rk), allocatable :: time(:)
  end type tem_transient_reduction_type

  contains

!******************************************************************************!
  !> read configuration file
  !!
  subroutine tem_load_transient_reduction( conf, me, parent, key, iError )
    !---------------------------------------------------------------------------
    !> handle for lua file
    type(flu_State)                      :: conf
    !>
    type(tem_transient_reduction_type), intent(out), allocatable :: me(:)
    !> handle for reduce table
    integer, optional,intent(in)         :: parent
    !> which key to open
    character(len=*),optional,intent(in) :: key
    !> handle for reduce table
    integer, optional,intent(out)        :: iError
    !---------------------------------------------------------------------------
    integer                        :: nEntries, handle, iErr, iPos, subhandle
    character(len=labelLen)              :: localKey
    !---------------------------------------------------------------------------
    if( present( key )) then
      localKey = key
    else
      localKey = 'transient_reduction'
    endif
    nEntries = 1
    allocate( me( nEntries))
    iPos = 1

    ! First check, if it is NOT a table
    if (present( parent )) then
      ! ...but still inside a table
      call tem_load_transient_reduction_single( conf    = conf,                &
        &                                       handle  = parent,              &
        &                                       me      = me(iPos),            &
        &                                       key     = localKey,            &
        &                                       iError  = iErr )
    else
      ! ... or at the root level
      call aot_get_val( L       = conf,                                        &
        &               key     = localKey,                                    &
        &               val     = me(iPos)%operation,                          &
        &               ErrCode = iErr,                                        &
        &               default = '' )
    endif

    if( iErr .ne. 0 .or. trim(me(iPos)%operation)=='' ) then
      ! ok, that failed. Let's try opening as a table.
      call aot_table_open( L       = conf,                                     &
        &                  thandle = handle,                                   &
        &                  parent  = parent,                                   &
        &                  key     = localKey )

      ! count number of entries
      nEntries = aot_table_length(L=conf, thandle=handle)
      deallocate( me )
      allocate( me( nEntries ))

      ! try to open subtables
      call aot_table_open( L       = conf,                                     &
        &                  parent  = handle,                                   &
        &                  thandle = subhandle,                                &
        &                  pos     = 1 )

      if ( subhandle .eq. 0 ) then
        ! only a single table, but multiple entries, read by positions
        call aot_table_close( L = conf, thandle = subhandle )
        do iPos = 1, nEntries
          call tem_load_transient_reduction_single( conf    = conf,            &
            &                                       handle  = handle,          &
            &                                       me      = me(iPos),        &
            &                                       pos     = iPos,            &
            &                                       iError  = iErr )
        enddo
        if( iErr .ne. 0 ) then
          ! Error, allocate with 0 size
          nEntries = 0
          deallocate( me )
          allocate( me(nEntries))
        endif
      else ! it has subtables, get into subtable and read by key
        call aot_table_close( L = conf, thandle = subhandle )
        do iPos = 1, nEntries
          call aot_table_open( L       = conf,                                 &
            &                  parent  = handle,                               &
            &                  thandle = subhandle,                            &
            &                  pos     = iPos )
          call tem_load_transient_reduction_single( conf    = conf,            &
            &                                       handle  = subhandle,       &
            &                                       me      = me(iPos),        &
            &                                       key     = localKey,        &
            &                                       iError  = iErr )
          call aot_table_close( L = conf, thandle = subhandle )
        enddo

      endif ! open subtable
      call aot_table_close(conf, handle)
    endif ! open transient_reduction table

    do iPos = 1, nEntries
      write(logUnit(1),*) ' Load transient reduction: '//                    &
        &                   trim(me(iPos)%operation)//'. nTimes: '//           &
        &                   trim( tem_toStr( me(iPos)%nTimes ))
    enddo

    if( present( iError )) iError = iErr

  end subroutine tem_load_transient_reduction
!******************************************************************************!


!******************************************************************************!
  !> Read a single entry of transient reductions from the lua file
  !!
  subroutine tem_load_transient_reduction_single( conf, me, handle, key, pos,  &
    &                                             iError )
    !---------------------------------------------------------------------------
    !> handle for lua file
    type(flu_State),intent(inout)                    :: conf
    !>
    type(tem_transient_reduction_type),intent(inout) :: me
    !> handle for reduce table
    integer, optional,intent(in)                     :: handle
    !> which key to open
    character(len=*),optional,intent(in)             :: key
    !> which key to open
    integer,optional,intent(in)                      :: pos
    !> handle for reduce table
    integer, optional,intent(out)                    :: iError
    !---------------------------------------------------------------------------
    integer :: iErr
    !---------------------------------------------------------------------------
    if( present( key )) then
      call aot_get_val( L       = conf,                                        &
        &               thandle = handle,                                      &
        &               val     = me%nTimes,                                   &
        &               ErrCode = iErr,                                        &
        &               key     = "nTimes",                                    &
        &               default = 0 )
      call aot_get_val( L       = conf,                                        &
        &               thandle = handle,                                      &
        &               val     = me%operation,                                &
        &               ErrCode = iErr,                                        &
        &               key     = key )
    elseif( present( pos ))then
      call aot_get_val( L       = conf,                                        &
        &               thandle = handle,                                      &
        &               val     = me%operation,                                &
        &               ErrCode = iErr,                                        &
        &               pos     = pos )
    else
      iErr = -1
    endif

    if( present( iError )) then
      iError = iErr
    endif

  end subroutine tem_load_transient_reduction_single
!******************************************************************************!


!******************************************************************************!
  !> Initialize transient reduction objects according to the number of values we
  !! get each timestep
  !!
  subroutine tem_transient_reduction_init( me, nElems, varSys, time)
    !---------------------------------------------------------------------------
    !> transient reduction type to be initialized
    type(tem_transient_reduction_type ),intent(inout) :: me(:)
    !> number of elements
    integer, intent(in)                               :: nElems
    !> corresponding variable system
    type(tem_var_system_type), intent(in)             :: varSys
    !> current time information
    real(kind=rk), intent(in)                         :: time
    !---------------------------------------------------------------------------
    integer :: iRed
    !---------------------------------------------------------------------------

    call tem_write('Initializing transient reductions ...')

    do iRed = 1, size(me)
      ! number of values
      me(iRed)%nElems = nElems
      me(iRed)%nComponents = varSys%variable%val(iRed)%nComponents

      select case(me(iRed)%operation)

        case('none')

        case('min')
          allocate(me(iRed)%val(1,me(iRed)%nElems*me(iRed)%nComponents))
          me(iRed)%val(1,:) = huge(1.0_rk)

        case('max')
          allocate(me(iRed)%val(1,me(iRed)%nElems*me(iRed)%nComponents))
          me(iRed)%val(1,:) = tiny(1.0_rk)

        case('sum')
          allocate(me(iRed)%val(1,me(iRed)%nElems*me(iRed)%nComponents))
          me(iRed)%val(1,:) = 0.0_rk

        case('average')
          allocate(me(iRed)%val(1,me(iRed)%nElems*me(iRed)%nComponents))
          me(iRed)%val(1,:) = 0.0_rk
          allocate(me(iRed)%time(0:2)) ! store t0 and current time
          ! t0 is reference time point
          ! t1 is the last reduction time point
          ! t2 is current time
          me(iRed)%time(0) = time
          me(iRed)%time(1:2) = 0.0_rk

        case('osi') ! oscillatory shear index
          allocate(me(iRed)%val( me(iRed)%nTimes, me(iRed)%nElems*me(iRed)%nComponents))
          me(iRed)%val(:,:) = 0.0_rk

        case default
          call tem_write('unknown transient reduction operation: '//           &
            &            me(iRed)%operation )
          call tem_abort()

      end select
    end do

  end subroutine tem_transient_reduction_init
!******************************************************************************!


!******************************************************************************!
  !> resets the "recorded" data of the reduction
  !!
  subroutine tem_transient_reduction_reset( me )
    !---------------------------------------------------------------------------
    !> transient reduction type
    type(tem_transient_reduction_type ),intent(inout) :: me(:)
    !---------------------------------------------------------------------------
    integer :: iRed
    !---------------------------------------------------------------------------

    do iRed = 1, size(me)
      select case(me(iRed)%operation)
        case('min')
          me(iRed)%val(1,:) = huge(1.0_rk)

        case('max')
          me(iRed)%val(1,:) = tiny(1.0_rk)

        case('sum')
          me(iRed)%val(1,:) = 0.0_rk

        case('average')
          ! reset reference time point
          me(iRed)%val(1,:) = 0.0_rk
          me(iRed)%time(0) = me(iRed)%time(0) + me(iRed)%time(1)
          me(iRed)%time(1) = 0.0_rk

      end select
    end do

  end subroutine tem_transient_reduction_reset
!******************************************************************************!


!******************************************************************************!
  !> "open" the reduction to prepare it
  !!
  subroutine tem_transient_reduction_open( me, time )
    !---------------------------------------------------------------------------
    !> transient reduction type
    type(tem_transient_reduction_type ),intent(inout) :: me(:)
    !>
    real(kind=rk), intent(in)                         :: time
    !---------------------------------------------------------------------------
    integer :: iRed
    !---------------------------------------------------------------------------

    do iRed = 1, size(me)
      select case(me(iRed)%operation)
        case('average')
          me(iRed)%time(2) = time-me(iRed)%time(0)
      end select
    end do

  end subroutine tem_transient_reduction_open
!******************************************************************************!


!******************************************************************************!
  !> apply the transient reduction
  !!
  subroutine tem_transient_reduction_apply( me, chunk, offset, nChunkElems,    &
    &                                       varSys )
    !---------------------------------------------------------------------------
    !> The transient reduction type to work on
    type( tem_transient_reduction_type ), intent(inout) :: me(:)
    !> apply the reduction to the data from the current timestep
    real(kind=rk), intent(inout)                        :: chunk(:)
    !> offset of chunk in total number of elemts
    integer, intent(in)                                 :: offset
    !> number of elements the chunk has
    integer, intent(in)                                 :: nChunkElems
    !> number of scalar entries there are per element
    type( tem_var_system_type ), intent(in)             :: varSys
    !---------------------------------------------------------------------------
    integer :: iRed, iComp, iElem, varPos, chunkVarPos
    !---------------------------------------------------------------------------

    ! calculate result and update me%val
    do iRed = 1, size( me )
      varPos = offset*me( iRed )%nComponents
      do iElem = 1, nChunkElems
        do iComp = 1, me( iRed )%nComponents
          chunkVarPos = (iElem-1)*varSys%nScalars                              &
            &         + varSys%variable%val( iRed )%varPos(iComp)
          varPos = varPos + 1
          !varPos = (offset + iElem - 1)*varSys%variable%val( iRed )%nComponents + iComp

          select case( me( iRed )%operation )

            case('none')

            case('min')
              me(iRed)%val(1,varPos) = min( me(iRed)%val(1,varPos),            &
                &                           chunk(chunkVarPos))
              chunk(chunkVarPos) = me(iRed)%val(1,varPos)

            case('max')
              me(iRed)%val(1,varPos) = max( me(iRed)%val(1,varPos),            &
                &                           chunk(chunkVarPos))
              chunk(chunkVarPos) = me(iRed)%val(1,varPos)

            case('sum')
              me(iRed)%val(1,varPos) = me(iRed)%val(1,varPos)                  &
                &                    + chunk(chunkVarPos)
              chunk(chunkVarPos) = me(iRed)%val(1,varPos)

            case('average')
              ! average = ( t1*val + (t2-t1)*chunk ) / t2
              me(iRed)%val(1,varPos) = 1./me(iRed)%time(2)                     &
                &                    * ( me(iRed)%time(1)                      &
                &                    * me(iRed)%val(1,varPos)                  &
                &                    + (me(iRed)%time(2)                       &
                &                    - me(iRed)%time(1))                       &
                &                    * chunk(chunkVarPos) )
              chunk(chunkVarPos) = me(iRed)%val(1,varPos)

          end select
        end do
      end do
    end do

  end subroutine tem_transient_reduction_apply
!******************************************************************************!


!******************************************************************************!
  !> close the reduction after it has been applied to all chunks
  !!
  subroutine tem_transient_reduction_close( me )
    !---------------------------------------------------------------------------
    !> transient reduction type to be closed
    type(tem_transient_reduction_type ),intent(inout) :: me(:)
    !---------------------------------------------------------------------------
    integer :: iRed
    !---------------------------------------------------------------------------

    do iRed = 1, size(me)
      select case(me(iRed)%operation)
        case('average')
          ! record current time for future usage
          me(iRed)%time(1) = me(iRed)%time(2)
      end select
    end do

  end subroutine tem_transient_reduction_close
!******************************************************************************!


end module tem_transient_reduction_module
!******************************************************************************!
