! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module provides functionality for description of variable systems.
!!
!! \author Harald Klimach, Jens Zudrop, Simon Zimny, Manuel Hasert, Kannan Masilmani, ...
!!
!! The description of the variable system heavily relies on the order in
!! which the variables are defined.
!! The first nVar variables describe the essential state, which is required
!! for restarts. Afterwards might follow an arbitrary list of derived
!! quantities.
!! In order to fill the system, you first have to call tem_var_sys_init,
!! and subsequently append the essential variables with tem_var_append first.
!! After these are complete, the SystemName and the number of variables
!! describing the state can be recorded with tem_var_sys_mark.
!! When doing an output, only the first nVars variables will be written!
!!
module tem_var_system_module

  ! include treelm modules
  use env_module,              only: LabelLen, PathLen, rk, SolSpecLen,        &
    &                                tem_err_warning
  use tem_aux_module,          only: tem_abort
  use tem_variable_module,     only: tem_variable_type, tem_variable_out,      &
    &                                tem_variable_from, assignment(=), append, &
    &                                dyn_varArray_type, init, PositionOfVal    
  use tem_tools_module,        only: tem_write, tem_horizontalSpacer
  use tem_stencil_module,      only: tem_stencilHeader_type
  use tem_time_module,         only: tem_time_type
  use treelmesh_module,        only: treelmesh_type
  use tem_subTree_type_module, only: tem_subTree_type
  use tem_debug_module,        only: tem_debug_type, main_debug, tem_debug,dbgUnit
  use tem_logging_module,      only: tem_logging_type, tem_logging_isActive,   &
    &                                tem_logging_unit, tem_log, logUnit

  ! include aotus modules
  use flu_binding,      only: flu_pop, flu_next, flu_isNumber, flu_toNumber,   &
    &                         flu_isNoneOrNil, flu_getfield, flu_toLString
  use aot_table_module, only: aot_table_open, aot_table_close,                 &
    &                         aot_table_length, aot_get_val
  use aot_out_module,   only: aot_out_type, aot_out_open_table, aot_out_val,   &
    &                         aot_out_close_table, aot_out_open, aot_out_close
  use aot_path_module,  only: aot_path_type
  use aotus_module,     only: flu_state, open_config_chunk

  implicit none
  private

  public :: tem_var_system_type
  public :: derive
  public :: tem_var_sys_init, tem_var_append, tem_var_sys_mark
  public :: tem_var_intersect, tem_var_union, tem_var_complement, tem_var_partOf
  public :: tem_var_sys_load, tem_var_sys_out, tem_var_sys_dump, tem_var_sys_log
  public :: tem_writeSolverSpec, tem_var_sys_debug


  !-----------------------------------------------------------------------------
  !> Datatype containing mapping of variables to function pointers
  !!
  type tem_varMap_type
    procedure( derive ),pointer,nopass :: do => null()
  end type tem_varMap_type
  !-----------------------------------------------------------------------------


  !-----------------------------------------------------------------------------
  !> The variable system description type
  !!
  !! gathers the information about all variables in the deployed equation
  !! system. In conjunction with schemes, you need one such description for
  !! each scheme, and they all should be gathered into an array, which is
  !! handed over to the restart output.
  !! Besides the actually available variables, this description also provides
  !! information on the additionally obtainable quantities, as provided by the
  !! solver. However elementwise data, which is required for restarting has to
  !! be given in the first nVars entries.
  type tem_var_system_type
    !> A descriptive name for this system of variables.
    character(len=LabelLen) :: SystemName

    !> Number of variables, which are actually existent / needed for restart.
    !!
    !! This definition is mostly useful to the solver, as there the values
    !! up to nVars usually can be obtained directly, while those afterwards
    !! might require some additional transformations.
    integer :: nVars

    !> Total available Variables, state vector variables have to come first.
    !!
    !! This is organized in a growing array, to allow easy extension with new
    !! variables. However it is important to keep the order in mind, and
    !! remember, that state variables have to be the first nVar variables.
    !! Number of totally available variables can be accessed in
    !! Variable%nVals
    type(dyn_varArray_type) :: Variable

    !> Number of scalar variables, which are actually existent / needed for
    !! restart.
    !!
    !! Computed from \f$ \sum^nVars_i( Variable\%val(i)\%nComponents ) \f$
    integer :: nScalars

    !> Array of function pointers for transformations (only defined in the
    !! globSys)
    type( tem_varMap_type ), allocatable :: derive(:)

    !> has the current variable system been initialized?
    logical :: initialized = .false.

    !> the unit of the scratch file, where the solver specific stuff was written
    integer :: solSpecUnit = -1
    !> solver specific string read as one part from the lua file
    character(len=SolSpecLen) :: solSpecString
    !> flu state which loads solver specific string
    type(flu_State) :: solSpec_conf
  end type tem_var_system_type
  !-----------------------------------------------------------------------------


  !-----------------------------------------------------------------------------
  abstract interface
    !> Interface description for the on-the-fly calculation of a single derived
    !! quantity for a set of elements.
    !!
    !! Various equation systems have to implement the actual functionality and
    !! provide function pointers with this interface declaration in the variable
    !! system.
    !! This is used by all solvers.
    !!
    !! Please note the option, that these functions might be used in-place.
    !! If no separate res array is provided, the input will be overwritten with
    !! derived quantity.
    !! Note, that the in-place operation only works fine, if the input system
    !! is larger than the output system.
    subroutine derive( input, bound, tree, subTree, requVar, globSys, stencil, &
      &                conf, res, bary, time, ndof )
      import  :: tem_var_system_type, tem_subTree_type, tem_variable_type,     &
        &        treelmesh_type, tem_stencilHeader_type, flu_state,            &
        &        tem_time_type, rk

      !> State of all state variables in the buffer (input)
      !! If the computation should be done in place and no separate res array
      !! is provided as argument, this input will be overwritten.
      !! Output values for all elements in the buffer (output)
      !! In state nElems * globSys%nScalars values are stored as follows:
      !! |  stateVars  |  stateVars  |  ...   |    stateVars    |   stateVars  |
      !! |-------------|-------------|--------|-----------------|--------------|
      !! |  iElem = 1  |  iElem = 2  |  ...   |iElem = nElems -1|iElem = nElems|
      !! The results are stored in the same manner in the same array, but with
      !! the reduced state size of the derived quantity.
      !! If the entries in the derived quantity are larger than the original
      !! system in the input, the operation can not be done in-place, and a
      real(kind=rk), target, intent(inout)    :: input(:)

      !> lower and upper bound of the element arrays to be processed from the 
      !! subTree
      !! subTree%useGlobalMesh = .true. -> tree%treeID(:)
      !! subTree%useLocalMesh  = .true. -> subTree%treeID(:)
      !! both false                     -> tree%treeID(subTree%map2global(:))
      !! 
      !! bound(2) minus bound(1) has to be equal to the number of elements to
      !! be processed in this chunk
      integer, intent(in) :: bound(2)

      !> global tree information
      type(treelmesh_type), intent(in) :: tree

      !> subTree to work on
      type(tem_subTree_type), intent(in) :: subTree

      !> Required variable.
      type( tem_variable_type ), intent(in) :: requVar

      !> Global variable system.
      type( tem_var_system_type ), intent(in) :: globSys

      !> Stencil information.
      type( tem_stencilHeader_type ) :: stencil

      !> Aotus type for main configuration file.
      type( flu_state ), intent(in) :: conf

      !> An optional array to put results in, instead of computing everything in
      !! place.
      real(kind=rk), optional, target, intent(out)    :: res(:)

      !> Array of barycenters of the elements.
      real(kind=rk), optional, intent(in) :: bary(:,:)

      !> Description of the simulation time.
      type(tem_time_type), optional, intent(in)  :: time

      !> Optional argument for the dof
      integer, optional, intent(in) :: ndof
    end subroutine derive
  end interface
  !-----------------------------------------------------------------------------


  !-----------------------------------------------------------------------------
  interface tem_var_append
    module procedure tem_var_appendName
    module procedure tem_var_appendVar
  end interface

  interface tem_var_sys_load
    module procedure tem_var_sys_load_table
    module procedure tem_var_sys_load_single
  end interface

  interface tem_var_sys_out
    module procedure tem_varSys_out_scal
    module procedure tem_varSys_out_vect
  end interface

  interface tem_var_sys_dump
    module procedure tem_var_sys_dump_scal
    module procedure tem_var_sys_dump_vect
  end interface

  interface tem_var_sys_log
    module procedure tem_var_sys_log_scal
    module procedure tem_var_sys_log_vect
  end interface

  interface tem_var_sys_debug
    module procedure tem_var_sys_debug_scal
    module procedure tem_var_sys_debug_vect
  end interface
  !-----------------------------------------------------------------------------

  contains

!******************************************************************************!
  !> Load variable system(s) from a lua file
  !!
  subroutine tem_var_sys_load_table( conf, sys, parent, key, iError )
    !---------------------------------------------------------------------------
    !> Lua handle connected to the script to read the table from
    type(flu_State) :: conf
    !> The variable system to read in
    type(tem_var_system_type), intent(out), allocatable :: sys(:)
    !> A parent table handle in which to look the current variables up
    integer, intent(in), optional :: parent
    !> load from this key
    character(len=*), optional, intent(in) :: key
    !> optional return error
    integer, optional, intent(inout) :: iError
    !---------------------------------------------------------------------------
    character(len=LabelLen) :: localkey
    integer :: thandle, subhandle
    integer :: nSys, iSys, iErr
    !---------------------------------------------------------------------------
    if( present( key )) then
      localKey = key
    else
      localKey = 'varsys'
    end if
    call aot_table_open( L       = conf,                                       &
      &                  parent  = parent,                                     &
      &                  thandle = thandle,                                    &
      &                  key     = localKey )
    if( thandle .lt. 1 ) then
      iErr = -1
      return
    end if
    ! Now the 'varsys' table is opened. Let's first check, if
    ! there is a single variable system directly inside, or if there are several,
    ! so we have to open a table first and check whether it has table val 
    ! 'systemname'
    allocate( sys( 1 ))
    call tem_var_sys_load_single( conf    = conf,                              &
      &                           thandle = thandle,                           &
      &                           iError  = iErr,                              &
      &                           sys     = sys( 1 ))
    if( iErr .ne. 0 ) then
      deallocate( sys )
      ! it wasn't directly in the varsys table.
      ! let's try opening the first entry as a table again
      nSys = aot_table_length( L = conf, thandle = thandle )
      allocate( sys(nSys))
      do iSys = 1, nSys
        call aot_table_open( L       = conf,                                   &
          &                  parent  = thandle,                                &
          &                  thandle = subhandle,                              &
          &                  pos     = iSys )
        call tem_var_sys_load_single( conf    = conf,                          &
          &                           thandle = subhandle,                     &
          &                           iError  = iErr,                          &
          &                           sys     = sys( iSys ))
        call aot_table_close( L = conf, thandle = subhandle )
       end do
    end if

    call aot_table_close( L = conf, thandle = thandle )

    if( present( iError )) iError = iErr

  end subroutine tem_var_sys_load_table
!******************************************************************************!


!******************************************************************************!
  !> Load a single variable system from a lua table
  !!
  subroutine tem_var_sys_load_single( conf, sys, thandle, iError, sysName, path)
    !---------------------------------------------------------------------------
    !> Lua handle connected to the script to read the table from
    type(flu_State) :: conf
    !> The variable system to initialize.
    type(tem_var_system_type), intent(out) :: sys
    !> A parent table handle in which to look the current variables up
    integer, intent(in) :: thandle
    !> optional return error
    integer, optional, intent(inout) :: iError
    !> name of the system
    character(len=*), optional, intent(in) :: sysName
    !> lua path
    type( aot_path_type ),intent(in), optional :: path
    !---------------------------------------------------------------------------
    ! The variable to read from a Lua script, and fill
    type(tem_variable_type), allocatable :: tVar(:)
    integer :: iVar, iErr, nTotalScalars, pos
    !---------------------------------------------------------------------------
    ! Initialise a variable system
    call tem_var_sys_init( sys = sys )
    ! Get SystemName from subroutine argument or from lua file
    if( present( sysName )) then
      sys%SystemName = sysName
    else
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               key     = 'systemname',                                &
        &               val     = sys%SystemName,                              &
        &               ErrCode = iErr )
      if( iErr .ne. 0) then
        if( present( iError )) then
          iError = iErr
        end if
        return
      end if
    end if
    ! How many variable entries are there in the current variable system?
    ! currently only 0
    nTotalScalars = 0

    ! Get variables, store them in tVar
    if( present( path )) then
      call tem_variable_from( me     = tVar,                                   &
        &                     conf   = conf,                                   &
        &                     nTotal = nTotalScalars,                          &
        &                     parent = thandle,                                &
        &                     iError = iErr,                                   &
        &                     path   = path )
    else
      call tem_variable_from( me     = tVar,                                   &
        &                     conf   = conf,                                   &
        &                     nTotal = nTotalScalars,                          &
        &                     parent = thandle,                                &
        &                     iError = iErr )
    end if

    ! Append variables to variable system
    do iVar = 1, size( tVar )
      call tem_var_append( sys = sys, var = tVar( iVar ), pos = pos )
    end do

    ! Get number of variables and total scalars from lua file
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               key     = 'nVars',                                       &
      &               val     = sys%nVars,                                     &
      &               ErrCode = iErr,                                          &
      &               default = -1 )

    ! Store the total number of variable components within the current variable
    ! system
    if( sys%nVars .gt. 0 ) then
      call tem_var_sys_mark( sys        = sys,                                 &
        &                    SystemName = sys%SystemName,                      &
        &                    nVars      = sys%nVars )
    else
      call tem_var_sys_mark( sys = sys, SystemName = sys%SystemName )
    end if

    call tem_readSolverSpec( conf = conf, parent = thandle, sys = sys )

    write(logUnit(1),*) 'Loaded Variable system '//sys%SystemName

end subroutine tem_var_sys_load_single
!******************************************************************************!


!******************************************************************************!
  !> Use this routine to initialize the variable system.
  !!
  !! This sets back all entries, and returns an empty variable system.
  !!
  subroutine tem_var_sys_init(sys, expected_length)
    !---------------------------------------------------------------------------
    !> The variable system to initialize.
    type(tem_var_system_type), intent(out) :: sys
    !> This optional parameter might be used, to allocate the growing array
    !! with a larger initial array (default is 0).
    integer, intent(in), optional :: expected_length
    !---------------------------------------------------------------------------

    sys%nVars = 0
    call init( me = sys%Variable, unique = .true., length = expected_length )
    sys%nScalars = 0
    sys%initialized = .true.

  end subroutine tem_var_sys_init
!******************************************************************************!


!******************************************************************************!
  !> The tem_var_appendVar routine is used to append a variable pre-defined
  !! variable to the variable system. Return the position, where it was added.
  !!
  subroutine tem_var_appendVar( sys, var, pos )
    !---------------------------------------------------------------------------
    !> Variable system to add the variable to, it has to be initialized
    !! properly.
    type(tem_var_system_type), intent(inout) :: sys
    !> The variable to add, which was defined before
    type(tem_variable_type) :: var
    !> Returned position of the variable in the system
    integer, intent(out) :: pos
    !---------------------------------------------------------------------------
    ! counters
    integer :: iComp, iDep, iVar
    integer :: totalScalar, nComp
    logical :: wasAdded, toAppend
    ! tmp variable for the dependencies
    type(tem_variable_type) :: tmp_Depvar
    ! position of the depending variable
    integer                 :: dep_pos
    !---------------------------------------------------------------------------
    ! initialise the out variables
    toAppend = .true.
    pos = 0

    if( allocated( var%depVars ) .and. sys%nVars .gt. 0 ) then
      pos = PositionOfVal( me = sys%variable, val = var )
      ! is the variable already added to the variable system ?
      ! yes -> return the position (pos = position in varSys)
      ! no -> check if the required variables are provided
      !          and store them temporary in an array
      if( pos .le. 0 )then
        ! if no -> check if the required variables are provided
        !          and store them temporary in an array

        ! if depVar is dependent on all then
        ! assign all the variables in the sys%nVars
        if( var%depVars(1) .eq. 'all'  ) then
          deallocate( var%varDep )
          allocate( var%varDep( sys%nVars ) )
          do iVar = 1, sys%nVars
            var%varDep( iVar ) = iVar
          end do
        else
          ! loop over all depending variables and check if they are in
          ! the given varSys (this makes shure that only 'valid' variables are
          ! appended)
          do iDep = 1, size( var%depVars )
            if( toAppend )then
              ! create a temporary variable of the depending variable
              ! setting nComponents to -1 garanties that only the VarName is 
              ! checked
              tmp_DepVar%label = trim(var%depVars(iDep))
              tmp_DepVar%nComponents = -1

              ! get position of this dependent variable
              dep_pos = PositionOfVal( me = sys%variable, val = tmp_DepVar )
              if( dep_pos .gt. 0 )then
                var%varDep( iDep ) = dep_pos
              else
                toAppend = .false.
              end if
            end if
          end do ! loop tmp_var%depVars
        end if ! depVars(1) .eq. all
      else
        toAppend = .false.
      end if ! pos less than 0
    end if ! var%depVars exists

    if( toAppend ) then
      ! Append the variable using the derived type constructor, with the
      ! gathered information about the variable.
      call append( me       = sys%Variable,                                    &
        &          val      = var,                                             &
        &          pos      = pos,                                             &
        &          wasAdded = wasAdded )

      nComp = var%nComponents

      if (wasAdded .and. nComp .gt. 0) then
        ! totalScalar is the sum of components in the list UP TO NOW
        ! It is needed to compute the convenience field
        ! varpos pointing from the variable to its entries in a globally
        ! serialized list of variable values.
        totalScalar = sum(sys%Variable%val(:pos-1)%nComponents)

        if( .not. allocated( sys%Variable%val( pos )%varPos ) )                &
          &  allocate( sys%Variable%val( pos )%varPos( nComp ))
        ! Build the reference from components of this variable to the components
        ! in the serialized list of all components. They have to be appended
        ! directly after the components added so far.
        do iComp = 1, nComp
          sys%Variable%val( pos )%varPos( iComp ) = iComp + totalScalar
        end do
      end if

    end if ! toAppend

  end subroutine tem_var_appendVar
!******************************************************************************!


!******************************************************************************!
  !> The tem_var_append routine is used to append a variable identified by
  !! VarName and nComponents to the given variable system.
  !!
  subroutine tem_var_appendName( sys, VarName, nComponents, pos, depVar )
    !---------------------------------------------------------------------------
    !> Variable system to add the variable to, it has to be initialized
    !! properly.
    type(tem_var_system_type), intent(inout) :: sys
    !> Name of appended variable
    character(len=*), intent(in) :: VarName
    !> Components in the appended variable
    integer, intent(in) :: nComponents
    !> Returned position of the variable in the system
    integer, intent(out) :: pos
    !> names of its dependent variables
    character(len=*), optional, intent(in) :: depVar(:)
    !---------------------------------------------------------------------------
    ! counters
    integer :: i, iDep, iVar, iComp
    integer :: dep_pos, totalScalar
    logical :: toAppend ! append the variable??
    type(tem_variable_type) :: tmp_var    ! tmp variable to add
    type(tem_variable_type) :: tmp_Depvar ! tmp variable for the dependencies
    !---------------------------------------------------------------------------

    ! initialise the out variables
    toAppend = .true.
    pos = 0

    ! define a temporary variable given by name and nComponents
    tmp_var%label = trim(VarName)
    tmp_var%nComponents = nComponents
    allocate( tmp_var%varPos( nComponents ))
    !allocate sub_index
    allocate( tmp_var%sub_index( nComponents ) )
    do iComp = 1, nComponents
      tmp_var%sub_index(iComp) = iComp
    end do

    ! find the position of variable with deplabel in the dynamic variable array
    ! up to now. such that derived variable can be dependent on other derived
    ! variables. If found then set the position of that variable in the varDep
    ! else add position of all variable defined till mark into varDep array
    if( present( depVar ) .and. sys%nVars .gt. 0 ) then
      ! transfer names of its dependent variables if depVar is given
      allocate( tmp_var%depVars( size( depVar )))
      tmp_var%depVars = depVar

      ! \todo: SZ: after marking a variable system only derived variables
      !            with matching dependencies shall be added 
      pos = PositionOfVal( me = sys%variable, val = tmp_var )
      ! is the variable already added to the variable system ?
      ! yes -> return the position (pos = position in varSys)
      if( pos .le. 0)then
        ! if no -> check if the required variables are provided
        !          and store them temporary in an array

        ! if depVar is dependent on all then
        ! assign all the variables in the sys%nVars
        if( depVar(1) .eq. 'all'  ) then
          allocate( tmp_var%varDep( sys%nVars ))
          do iVar = 1, sys%nVars
            tmp_var%varDep( iVar ) = iVar
          end do
        else
          ! loop over all depending variables and check if they are in
          ! the given varSys (this makes shure that only 'guilty' variables are
          ! appended)
          allocate( tmp_var%varDep( size(depVar) ))
          do iDep = 1, size( depVar )
            if( toAppend )then
              ! create a temporary variable of the depending variable
              ! setting nComponents to -1 garanties that only the VarName
              ! is checked
              tmp_DepVar%label = trim(tmp_var%depVars(iDep))
              tmp_DepVar%nComponents = -1

              ! get position of this dependent variable
              dep_pos = PositionOfVal( me = sys%variable, val = tmp_DepVar )
              if( dep_pos .gt. 0 )then
                tmp_var%varDep( iDep ) = dep_pos
              else
                toAppend = .false.
              end if
            end if
          end do ! loop tmp_var%depVars
        end if ! depVars(1) .eq. all
      else
        toAppend = .false.
      end if ! pos less than 0
    end if ! present( depVar )

    ! if the requirements are fullfilled append the variable using the derived
    ! type constructor, with the gathered information about the variable.
    if( toAppend )then
      call append(me  = sys%Variable,                                          &
        &         val = tmp_var,                                               &
        &         pos = pos)

      ! Total number of scalar entries untill the previous pos
      ! is given by the sum of components in the list up to now(pos-1),
      ! and is needed to compute the convenience field
      ! varpos pointing from the variable to its entries in a globally
      ! serialized list of variable values.
      totalScalar = sum(sys%Variable%val(:pos-1)%nComponents)

      ! Build the reference from components of this variable to the components
      ! in the serialized list of all components. They have to be appended
      ! directly after the components added so far.
      do i = 1, nComponents
        sys%Variable%val( pos )%varPos( i ) = i + totalScalar
      end do

    end if ! toAppend

  end subroutine tem_var_appendName
!******************************************************************************!


!******************************************************************************!
  !> This subroutine marks the equation system as fully defined.
  !!
  !! That is, the first nVars required for a full description of the state
  !! vector and for a restart are all described. With SystemName, some
  !! identifying string can be connected to the variables, which might be
  !! useful for later evaluations.
  !!
  subroutine tem_var_sys_mark(sys, SystemName, nVars)
    !---------------------------------------------------------------------------
    !> The variable sytem to mark.
    type(tem_var_system_type), intent(inout) :: sys
    !> A descriptive label for the variable system.
    character(len=*), intent(in) :: SystemName
    !> Optional parameter to describe the number of variables describing the
    !! state.
    !!
    !! If not, set it is assumed that all variables, put into the object
    !! so far are part of the state vector.
    integer, intent(in), optional :: nVars
    !---------------------------------------------------------------------------
    ! counter
    integer :: iVar
    !---------------------------------------------------------------------------

    sys%SystemName = trim(SystemName)
    if (present(nVars)) then
      sys%nVars = nVars
    else
      sys%nVars = sys%Variable%nVals
    end if

    sys%nScalars = 0
    do iVar = 1, sys%nVars
      sys%nScalars = sys%nScalars + sys%Variable%val(iVar)%nComponents
    end do

  end subroutine tem_var_sys_mark
!******************************************************************************!


!******************************************************************************!
  !> Write the system of variables description into a Lua file.
  !! use the aotus_out functions for doing so, in order to obtain a neatly
  !! formatted lua file
  !!
  subroutine tem_varSys_out_scal(sys, conf, ssUnit, dumpnVars, noDeps)
    !---------------------------------------------------------------------------
    !> Aotus type handling the output to the file in lua format
    type(aot_out_type), intent(inout) :: conf
    !> Variable system to write out
    type(tem_var_system_type), intent(in) :: sys
    integer, optional, intent(in) :: ssUnit !< solver specific unit
    !> dump nVars (true) or nVals (false)??
    logical, optional, intent(in) :: dumpnVars 
    !> do not dump the dependencies
    logical, optional, intent(in) :: noDeps
    !---------------------------------------------------------------------------
    integer :: i
    ! number of variables to be dumped (nVals or nVars)
    integer :: nDump
    !---------------------------------------------------------------------------
    ! set number of variables to be dumped
    nDump = sys%variable%nVals

    ! Overwrite it, if dumpnVars is set to .true.
    if (present(dumpnVars)) then
      if (dumpnVars) nDump = sys%nVars
    end if

    if( conf%level .eq. 0) then
      call aot_out_open_table( put_conf = conf, tname = 'varsys' )
    else
      call aot_out_open_table( put_conf = conf )
    end if
    call aot_out_val( put_conf = conf, val = trim(sys%systemName),             &
      &               vname = 'systemname')

    call aot_out_open_table( put_conf = conf, tname = 'variable' )
    do i = 1, nDump
      call tem_variable_out( me     = sys%Variable%Val(i),                     &
        &                    conf   = conf,                                    &
        &                    noDeps = noDeps )
    end do
    call aot_out_close_table( put_conf = conf )
    call aot_out_val( put_conf = conf, val = sys%nScalars, vname = 'nScalars' )
    call aot_out_val( put_conf = conf, val = sys%nVars, vname = 'nVars' )
    if(present(ssUnit))then
      if( ssUnit .ne. -1 ) then
        call aot_out_open_table( put_conf = conf, tname = 'solverDefinition' )
        call tem_writeSolverSpec( outUnit = conf%outUnit,                      &
          &                       ssUnit  = ssUnit )
        call aot_out_close_table( put_conf = conf )
      end if
    end if
    call aot_out_close_table( put_conf = conf )

  end subroutine tem_varSys_out_scal
!******************************************************************************!


!******************************************************************************!
  !> Write an array of systems of variables description into a Lua file.
  !!
  subroutine tem_var_sys_dump_scal( sys, outUnit, noDeps )
    !---------------------------------------------------------------------------
    !> Variable system to write out
    type(tem_var_system_type), intent(inout) :: sys
    !> file unit to write to
    integer, intent(inout) :: outUnit
    !> do not dump the dependencies
    logical, optional, intent(in) :: noDeps
    !---------------------------------------------------------------------------
    ! Aotus type handling the output to the file in lua format
    type(aot_out_type) :: conf
    !---------------------------------------------------------------------------
    call aot_out_open( put_conf = conf, outUnit = outUnit )
    call tem_var_sys_out( sys = sys, conf = conf, noDeps = noDeps  )
    call aot_out_close( conf )

  end subroutine tem_var_sys_dump_scal
!******************************************************************************!


!******************************************************************************!
  !>
  !!
  subroutine tem_var_sys_log_scal(sys, level, logger, noDeps )
    !---------------------------------------------------------------------------
    !> Variable system to write out
    type(tem_var_system_type), intent(inout) :: sys
    !> level of output
    integer, intent(in) :: level
    !> optional logging type to use
    type(tem_logging_type), optional, intent(inout) :: logger
    !> do not dump the dependencies
    logical, optional, intent(in) :: noDeps
    !---------------------------------------------------------------------------
    integer :: outunit
    !---------------------------------------------------------------------------

    outunit = tem_logging_unit( logger )

    if( tem_logging_isActive( logger, level ))then
      call tem_var_sys_dump( sys     = sys,                                    &
        &                    outUnit = outunit,                                & 
        &                    noDeps  = noDeps )
    end if

  end subroutine tem_var_sys_log_scal
!******************************************************************************!


!******************************************************************************!
  !>
  !!
  subroutine tem_var_sys_debug_scal(sys, level, debug, noDeps )
    !---------------------------------------------------------------------------
    !> Variable system to write out
    type(tem_var_system_type), intent(inout) :: sys
    !> level of output
    integer, intent(in) :: level
    !> optional debug type to use
    type(tem_debug_type), optional, intent(inout) :: debug
    !> do not dump the dependencies
    logical, optional, intent(in) :: noDeps
    !---------------------------------------------------------------------------
    type( tem_debug_type ) :: local_debug
    !---------------------------------------------------------------------------

    if( present( debug ))then
      local_debug = debug
    else
      local_debug = main_debug
    end if

    call tem_var_sys_log( sys    = sys,                                        &
      &                   level  = level,                                      &
      &                   logger = local_debug%logger,                         &
      &                   noDeps = noDeps )

  end subroutine tem_var_sys_debug_scal
!******************************************************************************!


!******************************************************************************!
  !> Write an array of systems of variables description into a Lua file.
  !!
  subroutine tem_var_sys_dump_vect(sys, outUnit, noDeps )
    !---------------------------------------------------------------------------
    !> Variable system to write out
    type(tem_var_system_type), intent(inout) :: sys(:)
    !> file unit to write to
    integer, intent(inout) :: outUnit
    !> do not dump the dependencies
    logical, optional, intent(in) :: noDeps
    !---------------------------------------------------------------------------
    !> Aotus type handling the output to the file in lua format
    type(aot_out_type) :: conf
    !---------------------------------------------------------------------------
    call aot_out_open( put_conf = conf, outUnit = outUnit )
    call tem_var_sys_out( sys = sys, conf = conf, noDeps = noDeps  )
    call aot_out_close( conf )

  end subroutine tem_var_sys_dump_vect
!******************************************************************************!


!******************************************************************************!
  !>
  !!
  subroutine tem_var_sys_log_vect(sys, level, logger, noDeps )
    !---------------------------------------------------------------------------
    !> Variable system to write out
    type(tem_var_system_type), intent(inout) :: sys(:)
    !> level of output
    integer, intent(in) :: level
    !> optional logging type to use
    type(tem_logging_type), optional, intent(inout) :: logger
    !> do not dump the dependencies
    logical, optional, intent(in) :: noDeps
    !---------------------------------------------------------------------------
    integer :: outunit
    !---------------------------------------------------------------------------

    outunit = tem_logging_unit( logger )

    if( tem_logging_isActive( logger, level ))then
      call tem_var_sys_dump( sys     = sys,                                    &
        &                    outUnit = outunit,                                &
        &                    noDeps  = noDeps )
    end if

  end subroutine tem_var_sys_log_vect
!******************************************************************************!


!******************************************************************************!
  !>
  !!
  subroutine tem_var_sys_debug_vect(sys, level, debug, noDeps )
    !---------------------------------------------------------------------------
    !> Variable system to write out
    type(tem_var_system_type), intent(inout) :: sys(:)
    !> level of output
    integer, intent(in) :: level
    !> optional debug type to use
    type(tem_debug_type), optional, intent(inout) :: debug
    !> do not dump the dependencies
    logical, optional, intent(inout) :: noDeps
    !---------------------------------------------------------------------------
    type( tem_debug_type ) :: local_debug
    !---------------------------------------------------------------------------

    if( present( debug ))then
      local_debug = debug
    else
      local_debug = main_debug
    end if

    call tem_var_sys_log( sys    = sys,                                        &
      &                   level  = level,                                      &
      &                   logger = local_debug%logger,                         &
      &                   noDeps = noDeps )

  end subroutine tem_var_sys_debug_vect
!******************************************************************************!


!******************************************************************************!
  !> Write an array of systems of variables description into a Lua file.
  !!
  subroutine tem_varSys_out_vect( sys, conf, ssUnit, dumpnVars, noDeps )
    !---------------------------------------------------------------------------
    !> Variable system to write out
    type(tem_var_system_type), intent(in) :: sys(:)
    !> Aotus type handling the output to the file in lua format
    type(aot_out_type), intent(inout) :: conf
    !> solver specific unit
    integer, optional, intent(in) :: ssUnit(:)
    !> dump nVars (true) or nVals (false)??
    logical, optional, intent(in) :: dumpnVars 
    !> do not dump the dependencies
    logical, optional, intent(in) :: noDeps
    !---------------------------------------------------------------------------
    ! counters
    integer :: iSys, nSys
    !---------------------------------------------------------------------------
    if( conf%level .eq. 0) then
      call aot_out_open_table( put_conf = conf, tname = 'varsys' )
    else
      call aot_out_open_table( put_conf = conf )
    end if
    nSys = size(sys)

    if( present( ssUnit ))then
      do iSys = 1, nSys
        call tem_varSys_out_scal( sys       = sys(iSys),                       &
          &                       conf      = conf,                            &
          &                       ssUnit    = ssUnit(iSys),                    &
          &                       dumpnVars = dumpnVars,                       &
          &                       noDeps    = noDeps )
      end do
    end if
    call aot_out_close_table( put_conf = conf )

  end subroutine tem_varSys_out_vect
!******************************************************************************!


!******************************************************************************!
  !> Intersection of requested and source system.
  !!
  !! This subroutine calculates the intersection of systemA and systemB,
  !! save the resulting system to resSys.
  !! (\f$ A \cap B \f$).
  !!
  subroutine tem_var_intersect( systemA, systemB, resSys, resSysLabel, iError )
    !---------------------------------------------------------------------------
    !> requested variable system
    type(tem_var_system_type ) :: systemA
    !> available variable system
    type(tem_var_system_type ) :: systemB
    !> resulting variable system
    type(tem_var_system_type ), intent(out) :: resSys
    !> error message
    integer, optional, intent(out) :: iError
    !> name of the result variable system
    character(Len=*), intent(in), optional :: resSysLabel
    !---------------------------------------------------------------------------
    ! counters
    integer :: iVar, pos, iDep, iComp
    integer :: varPos
    integer :: nComponents
    type(tem_variable_type) :: tVar, tVar2
    !---------------------------------------------------------------------------
    resSys%nScalars = 0
    if( present( iError )) iError = 0

    write(dbgUnit(3),*) 'Intersection of SystemA'

    call tem_var_sys_debug( sys = systemA,                                     &
      &                     level = 3 )
    write(dbgUnit(3),*) 'and SystemB'
    call tem_var_sys_debug( sys = systemB,                                     &
      &                     level = 3 )

    ! Variable system will be created, so initializing first...
    call tem_var_sys_init( sys = resSys )

    ! Try to find all variables from the requested system SystemA in the
    ! given SystemB
    do iVar = 1, systemA%variable%nVals
      ! Try to find position of variable in SystemB
      varPos = PositionOfVal( me  = systemB%variable,                          &
                              val = systemA%variable%val( iVar ))

      ! mapping gt 0 if variable in system A is found in systemB
      if( varPos .gt. 0 ) then
        ! Sum up the total amount of components
        ! there might be a defined sub-index
        if( allocated( systemA%variable%val( iVar )%sub_index )) then
          if( maxval( systemA%variable%val( iVar )%sub_index ) .gt.            &
            &         systemB%variable%val( varPos )%nComponents) then
            call tem_write(' defined index to a component in variable '//      &
              &            trim( systemA%variable%val( iVar )%label )//        &
              &            ' is larger than the actual amount of components')
            call tem_abort( )
          end if
          ! if there is a valid sub_index array, then the number of components
          ! is the size of the sub-index
          nComponents = size( systemA%variable%val( iVar )%sub_index )
        else
          ! if no such sub-index array exists, all the components of this
          ! variable are requested
          nComponents = systemB%variable%val( varPos )%nComponents
        end if

        ! here we set correct number of ncomponents in case if it was -1
        tVar = systemA%Variable%val( iVar )
        tVar%nComponents = nComponents
        if(.not. allocated(tVar%sub_index)) then
          allocate(tVar%sub_index( nComponents ) )
          do iComp = 1, nComponents
            tVar%sub_index(iComp) = iComp
          end do
        end if

        ! deallocate the varPos
        if( allocated( tVar%varPos )) deallocate( tVar%varPos )

        allocate( tVar%varPos(nComponents) )

        ! ... and attach to the resulting system
        call tem_var_append( sys = resSys, var = tVar, pos = pos )

        ! set the mapping
        resSys%variable%val( iVar )%mapping = varPos

        ! if dep var is defined for required varsys than
        ! find the position of this depvar in the available system
        ! and store the position in the resultant system
        if (allocated(systemA%Variable%val(iVar)%depVars)) then
          do iDep = 1, size(systemA%Variable%val(iVar)%depVars)
            tVar2%label = systemA%Variable%val(iVar)%depVars(iDep)
            tVar2%nComponents = -1
            varPos = PositionOfVal( me = systemB%variable, val = tVar2 )
            if( varPos .gt. 0 )                                                &
              & resSys%variable%val(iVar)%varDep(iDep) = varPos
          end do
        end if
      else
        call tem_horizontalSpacer()
        call tem_write( 'WARNING:' )
        call tem_write( 'Requested variable '//                                &
          &             trim( systemA%variable%val(iVar)%label ))
        call tem_write( 'was not found in the given system. Ignoring...' )
        call tem_horizontalSpacer()
        if( present( iError )) iError = ibset( iError, tem_err_warning )
      end if
    end do

    !set solver spec unit
    if(systemB%solSpecUnit .gt. 1) resSys%solSpecUnit = systemB%solSpecUnit

    ! to complete this we have to mark the resulting system
    if ( present(resSysLabel) ) then
      call tem_var_sys_mark( sys=resSys, SystemName=trim( resSysLabel ))
    else
      call tem_var_sys_mark( sys=resSys, SystemName='intersection' )
    end if

    write(dbgUnit(3),*) 'Intersection result is:'
    call tem_var_sys_debug( sys = resSys,                                      &
      &                     level = 3 )
    write(dbgUnit(3),*) 'Done. Variable System Intersection.'

  end subroutine tem_var_intersect
!******************************************************************************!


!******************************************************************************!
  !> Union of systemA and systemB.
  !!
  !! This subroutine calculates the union of systemA and systemB
  !! (\f$ A \cup B \f$).
  !!
  subroutine tem_var_union( systemA, systemB, resSys )
    !---------------------------------------------------------------------------
    !> requested variable system
    type(tem_var_system_type ) :: systemA
    !> available variable system
    type(tem_var_system_type ) :: systemB
    !> resulting variable system
    type(tem_var_system_type ), intent(out) :: resSys
    !---------------------------------------------------------------------------
    ! counters
    integer :: iVarA, iVarB
    integer :: pos
    !---------------------------------------------------------------------------
    ! initialiye the resulting system
    call tem_var_sys_init( sys = resSys )

    ! loop over all variables of systemA and append them to the resulting 
    ! variable system
    do iVarA = 1, systemA%variable%nVals
      call tem_var_append( sys = resSys,                                       &
        &                  var = systemA%variable%val(iVarA),                  &
        &                  pos = pos )
    end do
    do iVarB = 1, systemB%variable%nVals
      call tem_var_append( sys = resSys,                                       &
        &                  var = systemB%variable%val(iVarB),                  &
        &                  pos = pos )
    end do

    ! to complete this we have to mark the resulting system
    call tem_var_sys_mark( sys = resSys, SystemName='union' )

    write(dbgUnit(3),*)'UNION of' 
    call tem_var_sys_debug( sys = systemA,                                     &
      &                     level = 3 )
    write(dbgUnit(3),*) 'and'
    call tem_var_sys_debug( sys = systemB,                                     &
      &                     level = 3 )
    write(dbgUnit(3),*) 'is:'
    call tem_var_sys_debug( sys = resSys,                                      &
      &                     level = 3 )

  end subroutine tem_var_union
!******************************************************************************!


!******************************************************************************!
  !> Complement of systemA and systemB.
  !!
  !! This subroutine calculates the complement of systemA and systemB
  !! (\f$ A \setminus B \f$).
  !!
  subroutine tem_var_complement( systemA, systemB, resSys )
    !---------------------------------------------------------------------------
    !> requested variable system
    type(tem_var_system_type ) :: systemA
    !> available variable system
    type(tem_var_system_type ) :: systemB
    !> resulting variable system
    type(tem_var_system_type ), intent(out) :: resSys
    !---------------------------------------------------------------------------
    ! counter
    integer :: iVarA
    integer :: pos
    !---------------------------------------------------------------------------
    ! initialiye the resulting system
    call tem_var_sys_init( sys = resSys )
    resSys%SystemName = 'complement'

    ! find the positions of all variables in systemB in systemA
    ! if position = 0 --> variable not in systemA
    do iVarA = 1, systemA%variable%nVals
      pos = PositionOfVal( me  = systemB%variable,                             &
                           val = systemA%variable%val( iVarA ))
      if(pos .eq. 0)then
        call tem_var_append( sys = resSys,                                     &
          &                  var = systemA%variable%val(iVarA),                &
          &                  pos = pos )
      end if
    end do
    ! to complete this we have to mark the resulting system
    call tem_var_sys_mark( sys = resSys, SystemName = 'complement' )

    write(dbgUnit(3),*)'COMPLEMENT of'
    call tem_var_sys_debug( sys = systemA,                                     &
      &                     level = 3 )
    write(dbgUnit(3),*)'and'
    call tem_var_sys_debug( sys = systemB,                                     &
      &                     level = 3 )
    write(dbgUnit(3),*)'is:'
    call tem_var_sys_debug( sys = resSys,                                      &
      &                     level = 3 )

  end subroutine tem_var_complement
!******************************************************************************!


!******************************************************************************!
  !> Is systemB part of systemA.
  !!
  !! This function checks if systemB is part of systemA by checking each entry
  !! of systemB in systemA.
  !!
  function tem_var_partOf( systemA, systemB ) result(isPart)
    !---------------------------------------------------------------------------
    !> requested variable system
    type(tem_var_system_type ) :: systemA
    !> available variable system
    type(tem_var_system_type ) :: systemB
    !> return value if B is part of A
    logical :: isPart
    !---------------------------------------------------------------------------
    ! counter
    integer :: iVarB, counter
    integer :: pos
    !---------------------------------------------------------------------------

    counter = 0

    ! loop over all variables from systemB and get their positions in systemA
    do iVarB = 1, systemB%variable%nVals
      pos = PositionOfVal( me  = systemA%variable,                             &
                           val = systemB%variable%val( iVarB ))
      ! check if the variable is actually part of systemA,
      ! if yes --> output = true
      if(pos .gt. 0)then
        counter = counter + 1
      end if
    end do

    ! if the number of equal elements is equal to the number variables from
    ! systemB
    ! --> B is part of A
    if(counter .eq. systemB%variable%nVals)then
      isPart = .true.
    else
      isPart = .false.
    end if

  end function tem_var_partOf
!******************************************************************************!


!******************************************************************************!
  !> Read the solver specific stuff from the scratch unit and write it to the
  !! restart header file
  !!
  subroutine tem_writeSolverSpec( outUnit, ssUnit, closeUnit)
    !--------------------------------------------------------------------------
    !> unit to write to
    integer, intent(in) :: outUnit
    !> unit of the solver specific scratch file
    integer, intent(in) :: ssUnit
    !> close the outunit??
    logical, optional, intent(in) :: closeUnit
    !--------------------------------------------------------------------------
    integer :: io
    logical :: nUnitOpened ! was the scratch unit open??
    character(len=PathLen) :: buffer
    character(len=2) :: conca
    !--------------------------------------------------------------------------
    ! make a string out of the solver spec string in the lua file
    conca = ''
    inquire(unit= ssUnit, opened=nUnitOpened)
    if (nUnitOpened) then
      ! Go to the start of the scratch file.
      rewind(ssUnit)
      do
        ! Read all contents of the scratch file until eof is encountered
        read(ssUnit, '(a)', iostat=io) buffer
        if ( io .eq. 0 ) then
          write(outUnit,'(a)') trim(conca)//'" '//trim(buffer)//' "'
          conca = '..'
        else
          ! Reached end of the file, exit the loop.
          exit
        end if
      end do
      ! Close the output unit, if desired. Usually, we have to write more
      ! restart headers and so will need to leave this unit open
      if ( present(closeUnit) ) then
        if (closeUnit) close(ssUnit)
      end if
    else
      call tem_write( ' WARNING: no solver specific unit was opened before!' )
      call tem_write( '          Restart header file might be incomplete   ' )
    end if

  end subroutine tem_writeSolverSpec
!******************************************************************************!


!******************************************************************************!
  !> Read the solver specific stuff from the restart lua file and store it in
  !! the solver specific string of the belonging restart system.
  !!
  subroutine tem_readSolverSpec( conf, parent, sys )
    !---------------------------------------------------------------------------
    !> aotus type for restart lua file
    type( flu_state ), intent(inout) :: conf
    !> parent handle
    integer, intent(in)              :: parent
    !> var sys to store the string to
    type( tem_var_system_type ), intent(inout) :: sys
    !---------------------------------------------------------------------------
    integer :: solSpec_handle
    integer :: iError
    !---------------------------------------------------------------------------
    ! open the solver definition table
    call aot_table_open( L       = conf,                                       &
      &                  parent  = parent,                                     &
      &                  thandle = solSpec_handle,                             &
      &                  key     = 'solverDefinition' )

    call aot_get_val( L       = conf,                                          &
      &               thandle = solSpec_handle,                                &
      &               val     = sys%solSpecString,                             &
      &               ErrCode = iError,                                        &
      &               pos     = 1,                                             &
      &               default = '' ) 
    
    ! close the solver definition table again
    call aot_table_close( L= conf, thandle = solSpec_handle )

    call open_config_chunk( L     = sys%solSpec_conf,                          &
      &                     chunk = trim( sys%solSpecString ))

  end subroutine tem_readSolverSpec
!******************************************************************************!


end module tem_var_system_module
!******************************************************************************!
