! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module provides the description of variables
!!
!! \author Harald Klimach, Jens Zudrop, Simon Zimny, Manuel Hasert, Kannan Masilmani, ...
!!
!! It is used to describe the single variables in the deployed equation system,
!! and provide the means to describe requested output variables in the tracking
!! module.
!!
?? include 'arrayMacros.inc'
!!
module tem_variable_module

  ! include treelm modules
  use env_module,               only: rk, LabelLen, minLength, zeroLength
  use tem_spacetime_fun_module, only: tem_spacetime_fun_type,                  &
    &                                 tem_load_spacetime
  use tem_tools_module,         only: tem_write
  use tem_logging_module,       only: tem_log, tem_toStr, logUnit

  ! include aotus modules
  use aotus_module,     only: flu_State
  use aot_table_module, only: aot_table_open, aot_table_close,                 &
    &                         aot_table_length, aot_get_val
  use aot_out_module,   only: aot_out_type, aot_out_open, aot_out_close,       &
    &                         aot_out_val, aot_out_open_table,                 &
    &                         aot_out_close_table
  use aot_path_module,  only: aot_path_type, aot_path_addNode, assignment(=)


  implicit none

  private

  public :: tem_variable_type, tem_variable_from
  public :: tem_variable_from_single
  public :: tem_variable_out, tem_variable_dump
  ! growing array of variable types
  public :: grw_vararray_type
  ! dynamic array of variable types
  public :: dyn_vararray_type, init, destroy, append, expand, positionofval,   &
    &       sortedposofval
  public :: operator(==), operator(/=)
  public :: operator(<), operator(<=), operator(>), operator(>=)
  public :: assignment(=)


  !-----------------------------------------------------------------------------
  !> The tem_variable_type describes a single variable
  !!
  !! Which might possess several components to allow the representation of
  !! vectors.
  type tem_variable_type
    character(len=LabelLen) :: label !< A name for this variable
    !> Number of components of this variable, this can be used to describe
    !! vectorial quantities like the velocity, with nComponents = number of
    !! dimensions, or any other vectors, like for example the PDF in LBM.
    !!
    !! Possible solver specific interpretations of variable types for given
    !! number of Components:
    !!  1 = scalar
    !!  2 = complex
    !!  3 = vector
    !! 19 = PDF?
    !! if nComponents < 1 means there is an arbitrary number of components
    integer :: nComponents
    !> A pointer to the overall array of serialized entries in the equation
    !! system. It has as many entries, as there are components. And is available
    !! for convenient access in the overall state definition.
    integer, allocatable :: varPos(:)
    !> A pointer to the requested varPos, \todo: SZ: in the given varsys ??
    !! Example:
    !! velocity([1,3])
    integer, allocatable :: sub_index(:)
    !> A pointer to the position of dependent state variable
    !! in the dynamic array
    integer, allocatable :: varDep(:)
    !> A list of variable names this variable is depending on
    !! \todo: SZ: the varDep and depVars can be easily confused.
    !!            can we use another name?
    character(len=LabelLen),allocatable :: depVars(:)
    !> A list of tem spacetime function type
    type( tem_spacetime_fun_type ), allocatable :: st_fun(:)
    !> map of this variable to the same variable in another varSys 
    !! (mainly globSys)
    integer :: mapping
    !> Conversion factor to derive physical quantities with dimensions
    real(kind=rk) :: toPhyiscal = 1._rk
  end type
  !-----------------------------------------------------------------------------


! declare the growing and dynamic array of variable type
?? copy :: GA_decltxt(Var, type(tem_variable_type))
?? copy :: DA_decltxt(var, type(tem_variable_type))


  !-----------------------------------------------------------------------------
  interface assignment(=)
    module procedure Copy_var
  end interface

  interface operator(==)
    module procedure isEqual
  end interface

  interface operator(/=)
    module procedure isUnequal
  end interface

  interface operator(<)
    module procedure isSmaller
  end interface

  interface operator(<=)
    module procedure isSmallerOrEqual
  end interface

  interface operator(>)
    module procedure isGreater
  end interface

  interface operator(>=)
    module procedure isGreaterOrEqual
  end interface

  interface tem_variable_dump
    module procedure tem_variable_dump_single
    module procedure tem_variable_dump_list
  end interface
  !-----------------------------------------------------------------------------

contains
!******************************************************************************!
  !> Include the subroutines for the growing array.
?? copy :: GA_impltxt(Var, type(tem_variable_type))

!******************************************************************************!
  !> Include the subroutines for the dynamic array.
?? copy :: DA_impltxt(var, type(tem_variable_type), type(tem_variable_type))

!******************************************************************************!
  !> The tem_variable_out allows the output of a list of variables.
  !!
  !! The data is written into the file, the lunit is connected to.
  !! It is formatted as a Lua table.
  !!
  subroutine tem_variable_dump_list( me, outUnit )
    !---------------------------------------------------------------------------
    !> variable to write into the lua file
    type(tem_variable_type), intent(in) :: me(:)
    !> unit to write to
    integer, intent(in) :: outUnit
    !---------------------------------------------------------------------------
    ! aotus type handling the output to the file in lua format
    type(aot_out_type) :: conf
    ! counter
    integer :: iVar
    !---------------------------------------------------------------------------

    call aot_out_open( put_conf = conf, outUnit = outUnit )
    do iVar = 1, size( me )
      call tem_variable_out( conf = conf, me = me( iVar ))
    enddo
    call aot_out_close( put_conf = conf )

  end subroutine tem_variable_dump_list
!******************************************************************************!


!******************************************************************************!
  !> The tem_variable_out allows the output of a variable.
  !!
  !! The data is written into the file, the lunit is connected to.
  !! It is formatted as a Lua table.
  !!
  subroutine tem_variable_dump_single(me, outUnit )
    !---------------------------------------------------------------------------
    !> variable to write into the lua file
    type(tem_variable_type), intent(in) :: me
    !> unit to write to
    integer, intent(in) :: outUnit
    !---------------------------------------------------------------------------
    ! aotus type handling the output to the file in lua format
    type(aot_out_type) :: conf
    !---------------------------------------------------------------------------

    call aot_out_open( put_conf = conf, outUnit = outUnit )
    call tem_variable_out( conf = conf, me = me)
    call aot_out_close( put_conf = conf )

  end subroutine tem_variable_dump_single
!******************************************************************************!


!******************************************************************************!
  !> The tem_variable_out allows the output of the variable.
  !!
  !! The data is written into the file, the lunit is connected to.
  !! It is formatted as a Lua table.
  !!
  subroutine tem_variable_out(me, conf, noDeps)
    !---------------------------------------------------------------------------
    !> variable to write into the lua file
    type(tem_variable_type), intent(in) :: me
    !> aotus type handling the output to the file in lua format
    type(aot_out_type), intent(inout) :: conf
    !> dump dependencies?
    logical, intent(in), optional :: noDeps
    !---------------------------------------------------------------------------
    logical :: loc_noDeps
    !---------------------------------------------------------------------------

    call aot_out_open_table( put_conf = conf )
    call aot_out_val( put_conf = conf,                                         &
      &               val      = trim(me%label),                               &
      &               vname    = 'name' )
    call aot_out_val( put_conf = conf,                                         &
      &               val      = me%nComponents,                               &
      &               vname    = 'ncomponents' )
    if( present( noDeps )) then
      loc_noDeps = noDeps
    else
      loc_noDeps = .false.
    endif

    if( .not. loc_noDeps ) then
      ! write the dependent variables out
      if( allocated( me%varPos ))                                              &
        &  call aot_out_val( put_conf = conf,                                  &
        &                    val      = me%varPos,                             &
        &                    vname    = 'varpos' )
      if( allocated( me%varDep )) then
        if ( minval(me%varDep) .ne. -1)                                        &
          &  call aot_out_val( put_conf = conf,                                &
          &                    val      = me%varDep,                           &
          &                    vname    = 'varDep' )
      endif
      if( allocated( me%depVars))                                              &
        &  call aot_out_val( put_conf = conf,                                  &
        &                    val      = me%depVars,                            &
        &                    vname    = 'depVars' )
    end if
    call aot_out_close_table( put_conf = conf )

  end subroutine tem_variable_out
!******************************************************************************!


!******************************************************************************!
  !> This function returns a filled variable description read from a Lua script.
  !!
  !! An nTotal needs to be provided to allow the filling of the varPos array
  !! for backreference from each of the variables components to their
  !! corresponding entries in the overall serialized state vector.
  !! The variable is read from the position given in the pos argument, an
  !! access via key words is currently not supported.
  !! the lua table for a variable has to look something like
  !! \verbatim variables = {{'density',1} }\endverbatim
  !! \verbatim { { name = 'density', ncomponents = 1} }\endverbatim
  !!
  subroutine tem_variable_from_single( me, conf, nTotal, thandle, iError,      &
    &                                  path, key )
    !---------------------------------------------------------------------------
    !> The variable to read from the Lua script(conf) and fill
    type(tem_variable_type), intent(out) :: me
    !> Lua handle connected to the script to read the table from
    type(flu_State), intent(inout) :: conf
    !> Number of total components in the parent equation system up to now
    integer, optional, intent(inout) :: nTotal
    !> A parent table handle in which to look the current variable up
    integer, intent(in) :: thandle
    !> An optional error code for the calling routine
    integer, intent(out), optional :: iError
    !> key for a single variable
    character(len=*),intent(in), optional :: key
    !> lua path
    type( aot_path_type ),intent(in), optional :: path
    !---------------------------------------------------------------------------
    integer :: iErr
    integer :: nIndices, iComp, iIndex
    integer :: var_handle = 0
    integer :: local_thandle, subIndex_handle
    type( aot_path_type ) :: local_path
    !---------------------------------------------------------------------------

    me%label = ''
    if(present(path)) local_path = path
    ! if a key is given open the belonging table
    if( present(key) )then
      call aot_table_open( L       = conf,                                     &
        &                  thandle = var_handle,                               &
        &                  parent  = thandle,                                  &
        &                  key     = trim( key ))
      local_thandle = var_handle
      call aot_path_addNode( me       = local_path,                            &
        &                    NodeType = 'table',                               &
        &                    key      = trim( key ))
    else
      local_thandle = thandle
    end if

    ! Get the name of the variable
    call aot_get_val( L       = conf,                                          &
      &               thandle = local_thandle,                                 &
      &               val     = me%label,                                      &
      &               ErrCode = iErr,                                          &
      &               key     = 'name' )

    if( iErr .ne. 0 ) then
      ! There is an error reading the name , try read by positions
      ! first read label
      call aot_get_val( L       = conf,                                        &
        &               thandle = local_thandle,                               &
        &               val     = me%label,                                    &
        &               ErrCode = iErr,                                        &
        &               pos     = 1)
      write(logUnit(1),*) 'Variable label: '//trim( me%label )

      if( trim(me%label) .eq. '')  iErr = -1
      if( iErr .ne. 0 ) then
        write(logUnit(1),*) 'Unable to load "label" with pos.' 
        if( present( iError )) iError = iErr
        return
      end if
      ! second read nComponents
      call aot_get_val( L       = conf,                                        &
        &               thandle = local_thandle,                               &
        &               val     = me%nComponents,                              &
        &               ErrCode = iErr,                                        &
        &               pos     = 2,                                           &
        &               default = -1 )

      ! third load the lua function
      call tem_write( 'loading the spacetime functions for variable '          &
        &             //trim(me%label)//' from position 3')

      call tem_load_spacetime( me     = me%st_fun,                             &
        &                      conf   = conf,                                  &
        &                      path   = local_path,                            &
        &                      parent = local_thandle,                         &
        &                      nComp  = me%nComponents,                        &
        &                      pos    = 3 )

    else
      write(logUnit(1),*) 'Variable label: '//trim( me%label )
      ! Get the number of components for this variable
      call aot_get_val( L       = conf,                                        &
        &               thandle = local_thandle,                               &
        &               val     = me%nComponents,                              &
        &               ErrCode = iErr,                                        &
        &               key     = 'ncomponents',                               &
        &               default = -1 )
      if( iErr .ne. 0 ) then
        write(logUnit(1),*) 'No ncomponents specified for variable name'//   &
          &                 trim( me%label )
        if( present( iError )) iError = iErr
      endif

      call tem_write( 'loading the spacetime functions for variable '//        &
        &             trim( me%label ))
      call tem_load_spacetime( me     = me%st_fun,                             &
        &                      conf   = conf,                                  &
        &                      path   = local_path,                            &
        &                      parent = local_thandle,                         &
        &                      nComp  = me%nComponents,                        &
        &                      key    = trim( me%label ))

    endif !not from position

    ! Get the dependent variables
    call aot_table_open( L       = conf,                                       &
      &                  parent  = local_thandle,                              &
      &                  thandle = subIndex_handle,                            &
      &                  key     = 'dep' )
    if( subIndex_handle /= 0 ) then
      nIndices = aot_table_length( L = conf, thandle = subIndex_handle )
      allocate( me%depVars( nIndices ))
      allocate( me%varDep( nIndices ))
      me%varDep = -1
      do iIndex = 1, nIndices
        call aot_get_val( L       = conf,                                      &
          &               thandle = subIndex_handle,                           &
          &               val     = me%depVars( iIndex ),                      &
          &               ErrCode = iErr,                                      &
          &               pos     = iIndex )
      enddo
    endif ! if dep table is exist
    call aot_table_close( L = conf, thandle = subIndex_handle) ! close dep table


    ! get the sub-components, which are requested
    ! Try to open the variable table
    call aot_table_open( L       = conf,                                       &
      &                  parent  = local_thandle,                              &
      &                  thandle = subIndex_handle,                            &
      &                  key     = 'sub_index' )
    if (subIndex_handle /= 0) then
      nIndices = aot_table_length( L = conf, thandle = subIndex_handle )
      allocate( me%sub_index( nIndices ))
      do iIndex = 1, nIndices
        call aot_get_val( L       = conf,                                      &
          &               thandle = subIndex_handle,                           &
          &               val     = me%sub_index( iIndex ),                    &
          &               ErrCode = iError,                                    &
          &               pos     = iIndex )
      enddo
    endif
    call aot_table_close( L = conf, thandle = subIndex_handle )


    !\todo: SZ: does filling the varpos without a variable system make sense??
    !           this is not needed for the source variables, varPos is updated when
    !           appending a variable
    if(( iErr .eq. 0) .and. present(nTotal))then
      allocate( me%varPos( max( me%nComponents, 0 )))
      ! count up the total number of scalar variable entries
      do iComp = 1, me%nComponents
        me%varPos( iComp ) = iComp + nTotal
      end do
    endif
    if( present( iError )) iError = iErr

    ! close the variable table
    if( present(key) )then
      call aot_table_close( L = conf, thandle = local_thandle )
    end if

  end subroutine tem_variable_from_single
!******************************************************************************!


!******************************************************************************!
  !> This function returns a filled variable description read from a Lua script.
  !!
  !! The variable is read from the position given in the pos argument, an
  !! access via key words is currently not supported.
  !! the lua table for a variable has to look something like
  !! \verbatim variables = {{'density',1} }\endverbatim
  !! \verbatim variables = { { label = 'density', ncomponents = 1} }\endverbatim
  !! An nTotal needs to be provided for the variable system, which includes
  !! the current variable(s), to allow the filling of the varPos array
  !! for backreference from each of the variables components to their
  !! corresponding entries in the overall serialized state vector.
  !!
  subroutine tem_variable_from( me, conf, parent, nTotal, iError, key, path )
    !---------------------------------------------------------------------------
    !> The variable to read from a Lua script, and fill
    type(tem_variable_type), intent(out), allocatable :: me(:)
    !> Lua handle connected to the script to read the table from
    type(flu_State) :: conf
    !> Number of total components in the parent equation system up to now
    integer, optional, intent(inout) :: nTotal
    !> A parent table handle in which to look the current variable up
    integer, intent(in), optional :: parent
    !> An optional error code for the calling routine
    integer, intent(out), optional :: iError
    !> A parent table handle in which to look the current variable up
    character(len=*), intent(in), optional :: key
    type( aot_path_type ),intent(in), optional :: path !< lua path
    !---------------------------------------------------------------------------
    type( aot_path_type ) :: var_path, var_sub_path !< lua path
    character(len=LabelLen) :: local_key
    integer :: varhandle, varsubhandle
    integer :: iErr, iVar
    integer :: nVars
    !---------------------------------------------------------------------------
    if( present( key )) then
      local_key = key
    else
      local_key = 'variable'
    endif

    ! Try to open the variable table
    call aot_table_open( L       = conf,                                       &
      &                  parent  = parent,                                     &
      &                  thandle = varhandle,                                  &
      &                  key     = trim( local_key ))

    ! Reset number of variables
    nVars  = 0

    ! variable table exist but check whether it is a single or sub-tables
    if (varhandle > 0) then
      if( present( path ) ) then
        var_path = path
        call aot_path_addNode( me       = var_path,                            &
          &                    NodeType = 'table',                             &
          &                    key      = trim( local_key ))
      endif
      call aot_table_open( L       = conf,                                     &
        &                  parent  = varhandle,                                &
        &                  thandle = varsubhandle,                             &
        &                  pos     = 1 )
      ! It is only a single table
      if ( varsubhandle .eq. 0 ) then
        call aot_table_close( L = conf, thandle = varsubhandle )
        nVars  = 1
        allocate( me( nVars ))
        ! Get the name and nComponents for this single variable
        ! If iErr = 0 then variable is a single table else multiple table
        call tem_variable_from_single( me      = me(1),                      &
          &                            conf    = conf,                       &
          &                            path    = var_path,                   &
          &                            nTotal  = nTotal,                     &
          &                            thandle = varhandle,                  &
          &                            iError  = iErr )

      else  ! It has sub-tables
        call aot_table_close( L = conf, thandle = varsubhandle )
        nVars = aot_table_length( L = conf, thandle = varhandle )
        allocate( me( nVars ))
        do iVar = 1, nVars
          call aot_table_open( L       = conf,                                 &
            &                  parent  = varhandle,                            &
            &                  thandle = varsubhandle,                         &
            &                  pos     = iVar )
          if( present( path ) ) then
            var_sub_path = var_path
            call aot_path_addNode( me       = var_sub_path,                    &
              &                    NodeType = 'table',                         &
              &                    pos      = iVar )
            call tem_variable_from_single( me      = me(iVar),                 &
              &                            conf    = conf,                     &
              &                            nTotal  = nTotal,                   &
              &                            thandle = varsubhandle,             &
              &                            path    = var_sub_path,             &
              &                            iError  = iErr )
          else
            call tem_variable_from_single( me      = me(iVar),                 &
              &                            conf    = conf,                     &
              &                            nTotal  = nTotal,                   &
              &                            thandle = varsubhandle,             &
              &                            iError  = iErr )
          endif
          call aot_table_close( L = conf, thandle = varsubhandle )
          if( iErr .ne. 0) then
            if( present( iError )) iError = iErr
          endif
          if(present(nTotal))then
            nTotal = me( iVar )%nComponents + nTotal
          end if
        enddo
      endif ! if variable is single or sub-tables
    else ! No variable table is defined
      iErr = -1
      allocate( me( nVars ))
    end if ! if variable exist
    call aot_table_close( L = conf, thandle = varhandle )

    if( present( iError )) iError = iErr
  end subroutine tem_variable_from
!******************************************************************************!


!******************************************************************************!
  !> This function provides the assignment operator of two variables.
  !! Temporary Solution as CRAY compiler dont have = Operator
  !! Copying a Variable Components into other Variable
  !!
  subroutine Copy_var(left, right)
    !---------------------------------------------------------------------------
    !> variable to copy to
    type(tem_variable_type), intent(out) :: left
    !> variable to copy from
    type(tem_variable_type), intent(in) :: right
    !---------------------------------------------------------------------------

    ! First compare the string of the variable names
    left%label = right%label
    left%nComponents = right%nComponents
    if ( allocated(right%varPos) ) then
      allocate( left%varPos(size(right%varPos)) )
      left%varPos = right%varPos
    end if
    if ( allocated(right%sub_index) ) then
      allocate( left%sub_index(size(right%sub_index)) )
      left%sub_index = right%sub_index
    end if
    if ( allocated(right%varDep) ) then
      allocate( left%varDep(size(right%varDep)) )
      left%varDep = right%varDep
    end if

    if ( allocated( right%depVars )) then
      allocate( left%depVars(size(right%depVars)) )
      left%depVars = right%depVars
    end if
    if ( allocated( right%st_fun )) then
      allocate( left%st_fun( size( right%st_fun )))
      left%st_fun = right%st_fun
    end if

  end subroutine Copy_var
!******************************************************************************!


!******************************************************************************!
  !> This function provides the test for equality of two variables.
  !!
  !! Two variables are considered to be equal, if their name and number of
  !! components are equal.
  !!
  function isEqual(left, right) result(equality)
    !---------------------------------------------------------------------------
    !> variable to compare
    type(tem_variable_type), intent(in) :: left
    !> variable to compare against
    type(tem_variable_type), intent(in) :: right
    !> is equal??
    logical :: equality
    !---------------------------------------------------------------------------

    equality = .false.
    ! First compare the string of the variable names
    if (trim(left%label) == trim(right%label)) then
      ! if the string is the same, then check for the same number of components
      ! if either one is negative, then assume equality
      equality = (left%nComponents == right%nComponents)                       &
        &   .or. ( left%nComponents*right%nComponents .lt. 0 )
    end if

  end function isEqual
!******************************************************************************!



!******************************************************************************!
  !> This function provides the test for unequality of two variables.
  !!
  !! Two variables are considered to be equal, if their name and number of
  !! components are equal.
  !!
  function isUnequal(left, right) result(unequality)
    !---------------------------------------------------------------------------
    !> variable to compare
    type(tem_variable_type), intent(in) :: left
    !> variable to compare against
    type(tem_variable_type), intent(in) :: right
    !> is unequal??
    logical :: unequality
    !---------------------------------------------------------------------------

    unequality = .true.
    ! first compare the variable names
    if (trim(left%label) == trim(right%label)) then
      unequality = (left%nComponents /= right%nComponents)                     &
        &   .and. ( left%nComponents*right%nComponents .gt. 0 )
    end if

  end function isUnequal
!******************************************************************************!


!******************************************************************************!
  !> This function provides a comparison of two variables.
  !!
  !! A lexical ordering is used for the names, and as second criteria the
  !! number of components are used.
  !!
  function isSmaller(left, right) result(small)
    !---------------------------------------------------------------------------
    !> variable to compare
    type(tem_variable_type), intent(in) :: left
    !> variable to compare against
    type(tem_variable_type), intent(in) :: right
    !> is smaller??
    logical :: small
    !---------------------------------------------------------------------------

    small = .false.
    if (trim(left%label) < trim(right%label)) then
      small = .true.
    elseif( trim(left%label) == trim(right%label) ) then
      small = (left%nComponents < right%nComponents)                           &
        &   .and. ( left%nComponents*right%nComponents .gt. 0 )
    end if

  end function isSmaller
!******************************************************************************!


!******************************************************************************!
  !> This function provides a comparison of two variables.
  !!
  !! A lexical ordering is used for the names, and as second criteria the
  !! number of components are used.
  !!
  function isSmallerOrEqual(left, right) result(small)
    !---------------------------------------------------------------------------
    !> variable to compare
    type(tem_variable_type), intent(in) :: left
    !> variable to compare against
    type(tem_variable_type), intent(in) :: right
    !> is smaller??
    logical :: small
    !---------------------------------------------------------------------------

    small = .false.
    if (trim(left%label) <= trim(right%label)) then
      small = (left%nComponents <= right%nComponents)                          &
        &   .or. ( left%nComponents*right%nComponents .lt. 0 )
    end if

  end function isSmallerOrEqual
!******************************************************************************!


!******************************************************************************!
  !> This function provides a comparison of two variables.
  !!
  !! A lexical ordering is used for the names, and as second criteria the
  !! number of components are used.
  !!
  function isGreater(left, right) result(great)
    !---------------------------------------------------------------------------
    !> variable to compare
    type(tem_variable_type), intent(in) :: left
    !> variable to compare against
    type(tem_variable_type), intent(in) :: right
    !> is greater??
    logical :: great
    !---------------------------------------------------------------------------

    great = .false.
    if (trim(left%label) > trim(right%label)) then
      great = .true.
    elseif(trim(left%label) == trim(right%label) ) then
      great = (left%nComponents > right%nComponents)                           &
        &   .and. ( left%nComponents*right%nComponents .gt. 0 )
    end if

  end function isGreater
!******************************************************************************!


!******************************************************************************!
  !> This function provides a comparison of two variables.
  !!
  !! A lexical ordering is used for the names, and as second criteria the
  !! number of components are used.
  !!
  function isGreaterOrEqual(left, right) result(great)
    !---------------------------------------------------------------------------
    !> variable to compare
    type(tem_variable_type), intent(in) :: left
    !> variable to compare against
    type(tem_variable_type), intent(in) :: right
    !> is greater??
    logical :: great
    !---------------------------------------------------------------------------

    great = .false.
    if (trim(left%label) >= trim(right%label)) then
      great = (left%nComponents >= right%nComponents)                          &
        &   .or. ( left%nComponents*right%nComponents .lt. 0 )
    end if

  end function isGreaterOrEqual
!******************************************************************************!


end module tem_variable_module
!******************************************************************************!
